(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* This Unit is added by Jrg Mrkwitschka				*)
(* Uni-GH Siegen										*)
(* last change : 21-Nov-91								*)


(* ------------------------------------------------------------------- *)
unit CP_PointsFB;
(* ------------------------------------------------------------------- *)


{$SETC AclosestPair = TRUE }


(* ------------------------------------------------------------------- *)
interface
(* ------------------------------------------------------------------- *)


uses
	pointVector;


(* ------------------------------------------------------------------- *)
implementation
(* ------------------------------------------------------------------- *)

function pointVector.closestPairModified (ask: boolean): lineSegment;
	const
		below = left;
		above = right;
	type
		referenceArrayH = ^referenceArrayP;
		referenceArrayP = ^referenceArray;
		referenceArray = array[seqIndex] of reference;
	var
		yTable: dictionary;
		dummy: priorityQueue;
		squaredDelta: extended;
		currentPoint, closestLeft, closestRight: point2d;
		tail: seqIndex;
		current: seqIndex1;
		found: boolean;
		references: referenceArrayH;
		l, checkedConnection: lineSegment;
		check1, check2: reference;
		initialGuessing: Boolean;

{$IFC AnimationEnabled AND AclosestPair }
		oldDelta: extended;

       (* ------------------------------------------------------------------- *)

	procedure drawStatus (p: point2d; delta: extended);
		begin
			drawThickVertical(p.getX);
			drawVertical(p.getX - delta);
(*    drawCircle(p.getX, p.getY, delta - 1 / currentFrame.mx, 180, 180); *)
(*    drawLineSegment(p.getX, p.getY + delta, p.getX - delta, p.getY + delta); *)
(*    drawLineSegment(p.getX, p.getY - delta, p.getX - delta, p.getY - delta); *)
		end; (* drawStatus *)
{$ENDC }


        (* ------------------------------------------------------------------- *)


	procedure checkNeighbours (p1, p2: point2d);
		var
			nSqDelta: extended;
		begin

	(* first check y-Distance *)
			nSqDelta := p1.squaredDistanceX(p2);

{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin	(* draw the cloest  pair found jet *)
				checkedConnection.p := point2d(p1);
				checkedConnection.q := point2d(p2);
				checkedConnection.interactiveFlash;
			end;
{$ENDC }

			if nSqDelta < squaredDelta then begin
				squaredDelta := nSqDelta;

{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					waitForClick(animationFlag[AClosestPairItem]);
					closestRight.drawLine(closestLeft);		(* delete the  old line *)
				end;
{$ENDC }

				closestLeft := p1; (* left and right may not always be correct, but this will be fixed later *)
				closestRight := p2;

{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then		(* draw new line *)
					closestRight.drawLine(closestLeft);
{$ENDC }

			end;		(* euclidean squaredDistance *)
		end;		(* checkNeighbours *)

(* ------------------------------------------------------------------- *)

	procedure preprocessingStandard;
		begin
			closestLeft := point2d(elements^^[1]);
			closestRight := point2d(elements^^[2]);
			references^^[1] := yTable.insertObj(closestLeft, comparePointYX, found);
			references^^[2] := yTable.insertObj(closestRight, comparePointYX, found);
			if found then
				current := length + 1
			else begin
				squaredDelta := closestRight.squaredDistanceX(closestLeft);
				tail := 1;
				current := 3;
			end;

		end;

       (* ------------------------------------------------------------------- *)

	procedure preprocessingInitialGuessing;
		begin
			squaredDelta := 100000000;
			closestPairInitialGuessing(2, squaredDelta, closestLeft, closestRight);
			if squaredDelta = 0 then			(* 2 => 1-2,3-4,5-6,... *)
				current := length + 1				(* 1 => 1-2,2-3,4-5,... *)
			else begin
				references^^[1] := yTable.insertObj(point2d(elements^^[1]), comparePointYX, found);
				tail := 1;							(* insert the first Point *)
				current := 2;
			end;
		end;

       (* ------------------------------------------------------------------- *)

	begin
		if length >= 2 then begin
			references := referenceArrayH(dynamicArray(length, sizeOf(reference)));
	(* allocate memory for the refferenceArray *)

	(* initialize and select datastructure of the y-Table *)

			if ask then
				if getXY(0, false, dummy, length, yTable, true) then
					initialGuessing := stdInitialGuessing
				else begin
					closestPairModified := nil;
					exit(closestPairModified);
				end
			else begin
				initialGuessing := globalInitialGuessing;
				currentVectorLength := length;
				yTable := dictionary(newInitObj(avlTreeType));
			end;

		(*  preprocessing, like sorting and guessing of Delta *)
			sortAll(pointLessThanXY);	(* QuickSort the elements (increasing x) *)
										(*    heapSort(pointGreaterThanXY);	 --> use heapSort instead of QuickSort  *)

	(*startTime :=  TickCount; *)
	(* is used to determine the time used by the pure algorithm *)

			if initialGuessing then
				preprocessingInitialGuessing	(* Procedurecall whithout used parameters ! sorry *)
			else
				preprocessingStandard;

{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then		(* draw the cloest  pair found jet *)
				closestRight.drawLine(closestLeft);
			new(checkedConnection);							(* used for animation in checkNeighbours *)
			failNil(Handle(checkedConnection));
{$ENDC }

			while current <= length do begin	(* plane-sweep WHILE-loop *)

{$IFC AnimationEnabled }
				if userAbort then begin
{$IFC THINK_PASCAL }
					disposHandle(Handle(references));
{$ELSEC }
					DisposeHandle(Handle(references));
{$ENDC }
					yTable.free;
					closestPairModified := nil;
					exit(closestPairModified);
				end;
{$ENDC }

				currentPoint := point2d(elements^^[current]);	(* select current point *)

{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					oldDelta := sqrt(squaredDelta);
					drawStatus(currentPoint, oldDelta);	(* draw the status, two vertical lines ... *)
				end;
{$ENDC }

				while squaredDelta <= currentPoint.squaredDxX(point2d(elements^^[tail])) do begin

					if current - tail > 2 then begin			(* more then two points are active *)
						check1 := yTable.next(references^^[tail], above);
						check2 := yTable.next(references^^[tail], below);
						if (check1 <> nil) & (check2 <> nil) then
							checkNeighbours(point2d(yTable.getObject(check1)), point2d(yTable.getObject(check2)));
			(* check the point below and above tail         *)
					end;

					yTable.delete(references^^[tail]);	(* deactivate (delete in y-table) the points,   *)
													(*    which are not in the delta-surrounding   *)

{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then begin
{$IFC THINK_PASCAL }
						PenPat(gray);							(* paint the inactivated point gray             *)
						elements^^[tail].interactiveOutput;
						PenPat(black);
{$ELSEC }
						PenPat(qd.gray);							(* paint the inactivated point gray             *)
						elements^^[tail].interactiveOutput;
						PenPat(qd.black);
{$ENDC }
						waitForClick(animationFlag[AClosestPairItem]);
					end;
{$ENDC }

					tail := tail + 1;
				end;	(* of deleting WHILE- loop *)

				references^^[current] := yTable.insertObj(currentPoint, comparePointYX, found);
		(* activate (insert in y-Table)  current Point         *)
				if found then begin		(* <=> point with same coordinates allready exists *)
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then begin
						waitForClick(animationFlag[AClosestPairItem]);
						closestRight.drawLine(closestLeft);
					end;
{$ENDC }

					closestLeft := currentPoint;
					closestRight := point2d(yTable.getObject(references^^[current]));
					current := length;
				end
				else if (tail < current) then begin		(* check current Point against active points *)
					check1 := yTable.next(references^^[current], above);
					check2 := yTable.next(references^^[current], below);
					if check1 <> nil then
						checkNeighbours(point2d(yTable.getObject(check1)), currentPoint);
					if check2 <> nil then
						checkNeighbours(point2d(yTable.getObject(check2)), currentPoint);
				end;

				current := current + 1;

{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					waitForClick(animationFlag[AClosestPairItem]);
					drawStatus(currentPoint, oldDelta);			(* delete the old status- lines !! *)
				end;
{$ENDC }

			end;		(* of while current <= length *)

{$IFC THINK_PASCAL }
			disposHandle(Handle(references));
{$ELSEC }
			DisposeHandle(Handle(references));
{$ENDC }
			if yTable.infoString <> '' then
				displayInfoString(concat('Y', yTable.infoString));
			yTable.free;

{$IFC AnimationEnabled AND AclosestPair }
			dispose(checkedConnection);
{$ENDC }

			new(l);
			failNil(Handle(l));
			if (closestLeft.getX < closestRight.getX) | ((closestLeft.getX = closestRight.getX) & (closestLeft.getY < closestRight.getY)) then begin
				l.p := point2d(closestLeft.duplicate);
				l.q := point2d(closestRight.duplicate);
			end
			else begin
				l.q := point2d(closestLeft.duplicate);
				l.p := point2d(closestRight.duplicate);
			end;
			closestPairModified := l;

		end;

	end;  (* of closestPairModified *)

(* ------------------------------------------------------------------- *)

procedure pointVector.closestPairInitialGuessing (stepSize: stepIndex; var squaredDelta: extended; var cl, cr: point2d);
	var
		p1, p2: point2d;
		count1, count2: seqIndex2;
		newSquaredDelta: extended;
	begin

		count1 := 1;
		count2 := 2;

		while count2 <= length do begin										(* scan thru list *)
			p1 := point2d(elements^^[count1]);
			p2 := point2d(elements^^[count2]);
			newSquaredDelta := p1.squaredDistanceX(p2);
			if newSquaredDelta < squaredDelta then begin
				squaredDelta := newSquaredDelta;
				cl := p1;
				cr := p2;
			end;
			count1 := count1 + stepSize;
			count2 := count2 + stepSize;
		end;

	end; (* of closestPairInitialGuessing *)

(* ------------------------------------------------------------------- *)

end.
