(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(*	By Ernst Hinz, Mnster.								*)

unit ClosestPairOnline;

(* Implementierte On - Line Closest Pair Algorithmen fuer Dim = k(max 6)		*)
(*--------------------------------------------------------------------------	*)
(* ClosestPairSimple     : 22.09.92  letzte Aenderung : 08.01.93				*)
(* ClosestPairBinaryTree : 02.12.92  letzte Aenderung : 08.01.93				*)
(* ClosestPairBetaTree1  : 14.01.93  letzte Aenderung : 14.01.93				*)
(* ClosestPairBetaTree2  : 29.01.93  letzte Aenderung : 01.02.93				*)

interface

uses
	dDimPointVector, kBoxTree;

const
	Radius = 3;							(* Radius der gezeichneten Punkte *)

var
	P, Q: dDimPoint; 					(* aktuelles closest pair *)
	Delta, newDelta: extended;			(* Delta = Abstand closest pair *)

implementation
(***************** Universelle Prozeduren und Funktionen ******************* *)

procedure AnimationClosestPair (choice: integer; p1, p2: dDimPoint);
(* Animation fuer on--line ClosestPair Algorithmen. *)
	var
		x, y: extended;
		rp1, rp2: realPoint;
	begin (* AnimationClosestPair *)
		case choice of
			1:  begin (* "Loeschen der Punkte" vom Bildschirm *)
				rp1 := p1.toRealPoint;
				x := rp1.getX;
				y := rp1.getY;
				drawPoint(x, y, Radius + 1);
				drawPoint(x, y, Radius - 1);
				rp1.shallowFree;
			end;
			2:  begin (* Einfuegen der Punkte (on - line) *)
				rp1 := p1.toRealPoint;
				x := rp1.getX;
				y := rp1.getY;
				drawPoint(x, y, Radius);
				rp1.shallowFree;
				waitForClick(animationFlag[AClosestPairItem]);
			end;
			3:  begin (* Zeichnen der Strecke der zwei kontrollierten Punkte *)
				rp1 := p1.toRealPoint;
				rp2 := p2.toRealPoint;
				rp1.drawLine(rp2);
				waitForClick(animationFlag[AClosestPairItem]);
				rp1.drawLine(rp2);
				rp1.shallowFree;
				rp2.shallowFree;
			end;
			4:  begin (*  Loeschen des alten closest pair vom Bildschirm *)
				rp1 := p1.toRealPoint;
				rp2 := p2.toRealPoint;
				x := rp1.getX;
				y := rp1.getY;
				drawPoint(x, y, Radius + 1);
				x := rp2.getX;
				y := rp2.getY;
				drawPoint(x, y, Radius + 1);
				rp1.shallowFree;
				rp2.shallowFree;
			end;
			5:  begin (* Zeichnen des neuen closest pair auf den Bildschirm *)
				rp1 := p1.toRealPoint;
				rp2 := p2.toRealPoint;
				x := rp1.getX;
				y := rp1.getY;
				drawPoint(x, y, Radius + 1);
				x := rp2.getX;
				y := rp2.getY;
				drawPoint(x, y, Radius + 1);
				rp1.shallowFree;
				rp2.shallowFree;
				waitForClick(animationFlag[AClosestPairItem]);
			end;
			6:  begin (* Zeichnen der HilfsAnfragepunkte *)
				rp1 := p1.toRealPoint;
				x := rp1.getX;
				y := rp1.getY;
				drawPoint(x, y, 2);
				waitForClick(animationFlag[AClosestPairItem]);
				drawPoint(x, y, 2);
				rp1.shallowFree
			end
		end (* case *)
	end; (* AnimationClosestPair *)

function Inside (p: dDimPoint): boolean;
	var
		i: integer;
		IsIn: boolean;
	begin
		IsIn := true;
		i := 0;
		while IsIn and (i <= k) do begin
			i := i + 1;
			IsIn := Interval(RootBoxMin, RootBoxMax, p.coordinates[i])
		end;
		Inside := IsIn
	end;

function MyPower (Base, Value: integer): longint;
(* Berechnet die Anzahl der Anfragepunkte ( Base HOCH Value) *)
	var
		i: integer;
		Res: longint;
	begin (* MyPower *)
		Res := 1;
		if Value > 0 then
			for i := 1 to Value do
				Res := Res * Base;
		MyPower := Res
	end; (* MyPower *)

function CreateQueryPoint (m: longint; Update: dDimPoint; var dp: dDimPoint): dDimPoint;
(* Berechnet aus Update einen neuen Anfragepunkt *)
	var
		l, j: longint;
	begin (* CreateQueryPoint *)
		for j := 1 to k do begin
			l := m div MyPower(3, k - j);
			case l mod 3 of
				0: 
					dp.coordinates[j] := Update.coordinates[j] - Delta;
				1: 
					dp.coordinates[j] := Update.coordinates[j];
				2: 
					dp.coordinates[j] := Update.coordinates[j] + Delta;
			end; (* case *)
		end; (* for *)
		CreateQueryPoint := dp;
	end;(* CreateQueryPoint *)

function EqualPoints (p1, p2: dDimPoint): boolean;
(* Ueberprueft p1 und p2 auf Gleichheit *)
	var
		i: dimensionRange;
		gleich: boolean;
	begin (* EqualPoints *)
		i := 1;
		gleich := false;
		repeat
			if p1.coordinates[i] = p2.coordinates[i] then
				gleich := true
			else
				gleich := false;
			i := i + 1;
		until (i > k) or not gleich;
		EqualPoints := gleich
	end;(* EqualPoints *)

(**************************  ALGORITHMUS 1 ********************************* *)
procedure dDimPointVector.ClosestPairSimple;
	var
		n: seqIndex;
		i: longint;

	procedure OutputClosestPair;
(* Ausgabe des closest pair (P,Q). *)
		begin (* OutputClosestPair *)
			currentResult.append(P.duplicate);
			currentResult.append(Q.duplicate);
			if k = 2 then
				displayInfoString(StringOf('Delta=', Delta : 7 : 3, ' P=(', P.coordinates[1] : 7 : 3, ',', P.coordinates[2] : 7 : 3, ') Q=(', Q.coordinates[1] : 7 : 3, ',', Q.coordinates[2] : 7 : 3, ')'))
			else
				displayInfoString(StringOf('Delta  = ', Delta : 7 : 3));
			self.free
		end;(* OutputClosestPair *)

	begin (* ClosestPairSimple *)
		Delta := 100000;
		k := dDimPoint(elements^^[1]).dimension;
		if animationFlag[AClosestPairItem] then begin
			for i := 1 to length do
				AnimationClosestPair(1, dDimPoint(elements^^[i]), dDimPoint(elements^^[i]));
			waitForClick(animationFlag[AClosestPairItem])
		end;
		if length > 1 then begin
			n := 1;
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(2, dDimPoint(elements^^[n]), dDimPoint(elements^^[n]));
			repeat
				n := n + 1;
				if animationFlag[AClosestPairItem] then
					AnimationClosestPair(2, dDimPoint(elements^^[n]), dDimPoint(elements^^[n]));
				for i := 1 to n - 1 do begin
					newDelta := dDimPoint(elements^^[n]).distance(dDimPoint(elements^^[i]));
					if animationFlag[AClosestPairItem] then
						AnimationClosestPair(3, dDimPoint(elements^^[n]), dDimPoint(elements^^[i]));
					if newDelta < Delta then begin
						if animationFlag[AClosestPairItem] and (n > 2) then
							AnimationClosestPair(4, P, Q);
						P := dDimPoint(elements^^[n]);
						Q := dDimPoint(elements^^[i]);
						Delta := newDelta;
						if animationFlag[AClosestPairItem] then
							AnimationClosestPair(5, P, Q);
					end (* if *)
				end (* for *)
			until n = length;
			OutputClosestPair;
		end; (* if *)
	end;(* ClosestPairSimple *)

(************************** ENDE    ALOGRITHMUS 1 ************************** *)

(*****************************  ALGORITHMUS 2 ****************************** *)
procedure dDimPointVector.ClosestPairBinaryTree;
	var
		Tree: BinarykBoxTree;

	function ReadPoint (j: integer): dDimPoint;
(* Eingabe fuer die k - dim. Punkte *)
		begin (* ReadPoint *)
			ReadPoint := dDimPoint(elements^^[j]);
		end; (* ReadPoint *)

	procedure InitStructure;
(*Initialisiert die Datenstruktur BinarykBoxTree, das Closest Pair und Delta *)
		var
			i: integer;
		begin (* InitStructure *)
			if animationFlag[AClosestPairItem] then begin
				for i := 1 to length do
					AnimationClosestPair(1, dDimPoint(elements^^[i]), dDimPoint(elements^^[i]));
				waitForClick(animationFlag[AClosestPairItem])
			end;
			new(Tree);
			failNil(Handle(Tree));
			Tree.init;
			P := ReadPoint(1);
			k := P.dimension;
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(2, P, P);
			Q := ReadPoint(2);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(2, Q, Q);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(3, P, Q);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(5, P, Q);
			Delta := P.distance(Q);
			Tree.Insert(P);
			Tree.Insert(Q)
		end; (* InitStructure *)

	procedure UpDateClosestPair (Update: dDimPoint);
(* Ueberprueft, ob der Punkt Update mit einem bereits eingefuegten Punkt ein *)
(* neues Closest Pair bildet.                                               *)
		var
			Node: BinarykBoxNode;
			QueryPoint: dDimPoint;
			Help: ListElt;
			m, MaxQuery: longint;
		begin (* UpDateClosestPair *)
			new(QueryPoint);
			failNil(Handle(QueryPoint));
			MaxQuery := MyPower(3, k);  (* MaxQuery = Anzahl der Anfragepunkte *)
			for m := 1 to MaxQuery do begin
				QueryPoint := CreateQueryPoint(m, Update, QueryPoint);(* Liefert Anfragepunkt *)
				if animationFlag[AClosestPairItem] then
					AnimationClosestPair(6, QueryPoint, QueryPoint);
				if Inside(QueryPoint) then begin
					Node := Tree.Search(QueryPoint);
					if EqualPoints(QueryPoint, Update) or not Node.kBox.Include(Update) then begin
						Help := Node.Head;
						while Help <> nil do begin
							newDelta := Update.distance(Help.Element);
							if animationFlag[AClosestPairItem] then
								AnimationClosestPair(3, Update, Help.Element);
							if newDelta < Delta then begin
								if animationFlag[AClosestPairItem] then
									AnimationClosestPair(4, P, Q);
								P := Update;
								Q := Help.Element;
								if animationFlag[AClosestPairItem] then
									AnimationClosestPair(5, P, Q);
								Delta := newDelta;
							end; (* if *)
							Help := Help.Next
						end;(* while *)
					end; (* if *)
				end; (* if *)
			end; (* for *)
			QueryPoint.free;
		end; (* UpDateClosestPair *)

	procedure InsertPoints;
(* Einfuege - Algorithmus *)
		var
			UpdatePoint: dDimPoint;
			i: integer;
		begin (* InsertPoints *)
			if length > 2 then begin
				for i := 3 to length do begin
					UpdatePoint := ReadPoint(i);           (* Input eines neuen Punkt *)
					if animationFlag[AClosestPairItem] then
						AnimationClosestPair(2, UpdatePoint, UpdatePoint);
					UpDateClosestPair(UpdatePoint);    (* Aktualisierung Closest Pair *)
					Tree.Insert(UpdatePoint);     (* Aktualisierung des Restes *)
				end; (* for *)
			end; (* if *)
		end;(* InsertPoints *)

	procedure OutputClosestPair;
(* Ausgabe des closest pair (P,Q). *)
		begin (* OutputClosestPair *)
			currentResult.append(P.duplicate);
			currentResult.append(Q.duplicate);
			if k = 2 then
				displayInfoString(StringOf('Delta=', Delta : 7 : 3, ' P=(', P.coordinates[1] : 7 : 3, ',', P.coordinates[2] : 7 : 3, ') Q=(', Q.coordinates[1] : 7 : 3, ',', Q.coordinates[2] : 7 : 3, ')'))
			else
				displayInfoString(StringOf('Delta  = ', Delta : 7 : 3));
			if animationFlag[AClosestPairItem] then
				waitForClick(animationFlag[AClosestPairItem]);
			self.free
		end;(* OutputClosestPair *)

	begin(* ClosestPairBinaryTree *)
		InitStructure;                                         (* Initialisierung *)
		InsertPoints;                                   (* Einfuege - Algorithmus *)
		OutputClosestPair;                            (* Ausgabe des Closest Pair *)
		Tree.freeAll;                        (* Freigabe des Baum - Speicherplatz *)
	end;(* ClosestPairBinaryTree *)
(************************** ENDE    ALOGRITHMUS 2 ************************** *)

(*****************************  ALGORITHMUS 3 ****************************** *)
procedure dDimPointVector.ClosestPairBetaTree1;
	var
		Tree: BetakBoxTree;
	function ReadPoint (j: integer): dDimPoint;
(* Eingabe fuer die k - dim. Punkte *)
		begin (* ReadPoint *)
			ReadPoint := dDimPoint(elements^^[j]);
		end; (* ReadPoint *)

	procedure InitStructure;
(*Initialisiert die Datenstruktur BetakBoxTree, das Closest Pair und Delta  *)
		var
			i: integer;
		begin (* InitStructure *)
			if animationFlag[AClosestPairItem] then begin
				for i := 1 to length do
					AnimationClosestPair(1, dDimPoint(elements^^[i]), dDimPoint(elements^^[i]));
				waitForClick(animationFlag[AClosestPairItem])
			end;
			new(Tree);
			failNil(Handle(Tree));
			Tree.init;
			P := ReadPoint(1);
			k := P.dimension;
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(2, P, P);
			Q := ReadPoint(2);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(2, Q, Q);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(3, P, Q);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(5, P, Q);
			Delta := P.distance(Q);
			Tree.Insert(P);
			Tree.Insert(Q)
		end; (* InitStructure *)

	procedure UpDateClosestPair (Update: dDimPoint);
(* Ueberprueft, ob der Punkt Update mit einem bereits eingefuegten Punkt ein *)
(* neues Closest Pair bildet.                                               *)
		var
			Node: BetakBoxNode;
			QueryPoint: dDimPoint;
			Help: ListElt;
			m, MaxQuery: longint;
		begin (* UpDateClosestPair *)
			new(QueryPoint);
			failNil(Handle(QueryPoint));
			MaxQuery := MyPower(3, k);  (* MaxQuery = Anzahl der Anfragepunkte *)
			for m := 1 to MaxQuery do begin
				QueryPoint := CreateQueryPoint(m, Update, QueryPoint);(* Liefert Anfragepunkt *)
				if animationFlag[AClosestPairItem] then
					AnimationClosestPair(6, QueryPoint, QueryPoint);
				if Inside(QueryPoint) then begin
					Node := BetakBoxNode(Tree.Search(QueryPoint));
					if EqualPoints(QueryPoint, Update) or not Node.kBox.Include(Update) then begin
						Help := Node.Head;
						while Help <> nil do begin
							newDelta := Update.distance(Help.Element);
							if animationFlag[AClosestPairItem] then
								AnimationClosestPair(3, Update, Help.Element);
							if newDelta < Delta then begin
								if animationFlag[AClosestPairItem] then
									AnimationClosestPair(4, P, Q);
								P := Update;
								Q := Help.Element;
								if animationFlag[AClosestPairItem] then
									AnimationClosestPair(5, P, Q);
								Delta := newDelta;
							end; (* if *)
							Help := Help.Next
						end;(* while *)
					end; (* if *)
				end; (* if *)
			end; (* for *)
			QueryPoint.free;
		end; (* UpDateClosestPair *)

	procedure InsertPoints;
(* Einfuege - Algorithmus *)
		var
			UpdatePoint: dDimPoint;
			i: integer;
		begin (* InsertPoints *)
			if length > 2 then begin
				for i := 3 to length do begin
					UpdatePoint := ReadPoint(i);           (* Input eines neuen Punkt *)
					if animationFlag[AClosestPairItem] then
						AnimationClosestPair(2, UpdatePoint, UpdatePoint);
					UpDateClosestPair(UpdatePoint);    (* Aktualisierung Closest Pair *)
					Tree.Insert(UpdatePoint);     (* Aktualisierung des Restes *)
				end; (* for *)
			end; (* if *)
		end;(* InsertPoints *)

	procedure OutputClosestPair;
(* Ausgabe des closest pair (P,Q). *)
		begin (* OutputClosestPair *)
			currentResult.append(P.duplicate);
			currentResult.append(Q.duplicate);
			if k = 2 then
				displayInfoString(StringOf('Delta=', Delta : 7 : 3, ' P=(', P.coordinates[1] : 7 : 3, ',', P.coordinates[2] : 7 : 3, ') Q=(', Q.coordinates[1] : 7 : 3, ',', Q.coordinates[2] : 7 : 3, ')'))
			else
				displayInfoString(StringOf('Delta  = ', Delta : 7 : 3));
			if animationFlag[AClosestPairItem] then
				waitForClick(animationFlag[AClosestPairItem]);
			self.free;
		end;(* OutputClosestPair *)

	begin(* ClosestPairBetaTree 1 *)
		InitStructure;                                         (* Initialisierung *)
		InsertPoints;                                   (* Einfuege - Algorithmus *)
		OutputClosestPair;                            (* Ausgabe des Closest Pair *)
		Tree.freeAll;                        (* Freigabe des Baum - Speicherplatz *)
	end;(* ClosestPairBetaTree *)
(************************** ENDE    ALOGRITHMUS 3 ************************** *)

(*****************************  ALGORITHMUS 4 ****************************** *)
procedure dDimPointVector.ClosestPairBetaTree2;
	var
		Tree: BetakBoxTree;
	function ReadPoint (j: integer): dDimPoint;
(* Eingabe fuer die k - dim. Punkte *)
		begin (* ReadPoint *)
			ReadPoint := dDimPoint(elements^^[j]);
		end; (* ReadPoint *)

	procedure InitStructure;
(*Initialisiert die Datenstruktur BetakBoxTree, das Closest Pair und Delta  *)
		var
			i: integer;
		begin (* InitStructure *)
			if animationFlag[AClosestPairItem] then begin
				for i := 1 to length do
					AnimationClosestPair(1, dDimPoint(elements^^[i]), dDimPoint(elements^^[i]));
				waitForClick(animationFlag[AClosestPairItem])
			end;
			new(Tree);
			failNil(Handle(Tree));
			Tree.init;
			P := ReadPoint(1);
			k := P.dimension;
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(2, P, P);
			Q := ReadPoint(2);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(2, Q, Q);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(3, P, Q);
			if animationFlag[AClosestPairItem] then
				AnimationClosestPair(5, P, Q);
			Delta := P.distance(Q);
			Tree.Insert2(P);
			Tree.Insert2(Q)
		end; (* InitStructure *)

	procedure UpDateClosestPair (Update: dDimPoint);
(* Ueberprueft, ob der Punkt Update mit einem bereits eingefuegten Punkt ein *)
(* neues Closest Pair bildet.                                               *)
		var
			Node: BetakBoxNode;
			QueryPoint: dDimPoint;
			Help: ListElt;
			m, MaxQuery: longint;
		begin (* UpDateClosestPair *)
			new(QueryPoint);
			failNil(Handle(QueryPoint));
			MaxQuery := MyPower(3, k);  (* MaxQuery = Anzahl der Anfragepunkte *)
			for m := 1 to MaxQuery do begin
				QueryPoint := CreateQueryPoint(m, Update, QueryPoint);(* Liefert Anfragepunkt *)
				if animationFlag[AClosestPairItem] then
					AnimationClosestPair(6, QueryPoint, QueryPoint);
				if Inside(QueryPoint) then begin
					Node := BetakBoxNode(Tree.Search(QueryPoint));
					if EqualPoints(QueryPoint, Update) or not Node.kBox.Include(Update) then begin
						Help := Node.Head;
						while Help <> nil do begin
							newDelta := Update.distance(Help.Element);
							if animationFlag[AClosestPairItem] then
								AnimationClosestPair(3, Update, Help.Element);
							if newDelta < Delta then begin
								if animationFlag[AClosestPairItem] then
									AnimationClosestPair(4, P, Q);
								P := Update;
								Q := Help.Element;
								if animationFlag[AClosestPairItem] then
									AnimationClosestPair(5, P, Q);
								Delta := newDelta;
							end; (* if *)
							Help := Help.Next
						end;(* while *)
					end; (* if *)
				end; (* if *)
			end; (* for *)
			QueryPoint.free;
		end; (* UpDateClosestPair *)

	procedure InsertPoints;
(* Einfuege - Algorithmus *)
		var
			UpdatePoint: dDimPoint;
			i: integer;
		begin (* InsertPoints *)
			if length > 2 then begin
				for i := 3 to length do begin
					UpdatePoint := ReadPoint(i);           (* Input eines neuen Punkt *)
					if animationFlag[AClosestPairItem] then
						AnimationClosestPair(2, UpdatePoint, UpdatePoint);
					UpDateClosestPair(UpdatePoint);    (* Aktualisierung Closest Pair *)
					Tree.Insert2(UpdatePoint);    (* Aktualisierung des Restes *)
				end; (* for *)
			end; (* if *)
		end;(* InsertPoints *)

	procedure OutputClosestPair;
(* Ausgabe des closest pair (P,Q). *)
		begin (* OutputClosestPair *)
			currentResult.append(P.duplicate);
			currentResult.append(Q.duplicate);
			if k = 2 then
				displayInfoString(StringOf('Delta=', Delta : 7 : 3, ' P=(', P.coordinates[1] : 7 : 3, ',', P.coordinates[2] : 7 : 3, ') Q=(', Q.coordinates[1] : 7 : 3, ',', Q.coordinates[2] : 7 : 3, ')'))
			else
				displayInfoString(StringOf('Delta  = ', Delta : 7 : 3));
			if animationFlag[AClosestPairItem] then
				waitForClick(animationFlag[AClosestPairItem]);
			self.free;
		end;(* OutputClosestPair *)

	begin(* ClosestPairBetaTree *)
		InitStructure;                                         (* Initialisierung *)
		InsertPoints;                                   (* Einfuege - Algorithmus *)
		OutputClosestPair;                            (* Ausgabe des Closest Pair *)
		Tree.freeAll;                             (* Freigabe des Baum - Speicherplatz *)
	end;(* ClosestPairBetaTree *)
(************************** ENDE    ALOGRITHMUS 4 ************************** *)

end.
