(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit DistanceDistribution;

interface

uses
	pointVector;

implementation

procedure disposeVertex (o: obj);
	begin
		o.freeAll;
	end;

function histogram.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, histogram) & inherited equal(o);
	end;

function histogram.memberType (o: obj): boolean;
	begin
		memberType := member(o, histogram);
	end;

function histogram.getType: objectType;
	begin
		getType := histogramType;
	end;

function histogram.getTextType: Str255;
	begin
		getTextType := StringOf('histogram(', length : 1, ')');
	end;

procedure histogram.initFromGenerator (numberOfValues, numberOfBins: longInt; width, height: extended; function generateValue: real; procedure resetGenerator);
	const
		originX = 50;
		originY = 150;
	type
		longIntArrayH = ^longIntArrayP;
		longIntArrayP = ^longIntArray;
		longIntArray = array[seqIndex] of longInt;
		exitActionType = (disposeValues, noDisposeValues, disposeResult);
	var
		value, minValue, maxValue, lambda, y: real;
		counters: longIntArrayH;
		i, index, maxCounter: longInt;
		values: realArrayH;
		rectangleWidth: extended;

	procedure add (newX, newY: real);
		var
			rp, p: realPoint;
		begin
			if length > 0 then begin
				p := realPoint(elements^^[length]);
				if (p.x = newX) & (p.y = newY) then
					exit(add);
				if (length > 1) & (realPoint(elements^^[length - 1]).y = p.y) & (p.y = newY) then begin
					p.x := newX;
					exit(add);
				end;
			end;
			new(rp);
			failNil(Handle(rp));
			rp.x := newX;
			rp.y := newY;
			append(rp);
		end;

	procedure conditionalExit (condition: boolean; exitAction: exitActionType);
		begin
			if condition then begin
				if exitAction = disposeValues then
{$IFC THINK_PASCAL }
					disposHandle(Handle(values));
				disposHandle(Handle(counters));
{$ELSEC }
					DisposeHandle(Handle(values));
				DisposeHandle(Handle(counters));
{$ENDC }
				if exitAction = disposeResult then begin
					forAll(disposeVertex);
{$IFC THINK_PASCAL }
					disposHandle(Handle(elements));
{$ELSEC }
					DisposeHandle(Handle(elements));
{$ENDC }
				end;
				currentVectorLength := 0;
				init;
				exit(initFromGenerator);
			end;
		end;

	begin (* histogram.initFromGenerator *)
		counters := longIntArrayH(dynamicArray(numberOfBins, sizeOf(longInt)));
		for i := 1 to numberOfBins do
			counters^^[i] := 0;
		minValue := 1E99;
		maxValue := 0;
		resetGenerator;
		values := realArrayH(dynamicArrayNoCheck(numberOfValues, sizeOf(real)));
		if values = nil then begin
			for i := 1 to numberOfValues do begin
{$IFC AnimationEnabled }
				conditionalExit(userAbort, noDisposeValues);
{$ENDC }
				value := generateValue;
				if value > maxValue then
					maxValue := value;
				if value < minValue then
					minValue := value;
			end;
			conditionalExit(minValue = maxValue, noDisposeValues);
			lambda := numberOfBins / (maxValue - minValue);
			resetGenerator;
			for i := 1 to numberOfValues do begin
{$IFC AnimationEnabled }
				conditionalExit(userAbort, noDisposeValues);
{$ENDC }
{$IFC THINK_PASCAL }
				index := trunc(lambda * (generateValue - minValue) + 1);
{$ELSEC }
				index := truncLong(lambda * (generateValue - minValue) + 1);
{$ENDC }
				if index > numberOfBins then
					index := numberOfBins;
				counters^^[index] := counters^^[index] + 1;
			end;
		end
		else begin
			for i := 1 to numberOfValues do begin
{$IFC AnimationEnabled }
				conditionalExit(userAbort, disposeValues);
{$ENDC }
				value := generateValue;
				if value > maxValue then
					maxValue := value;
				if value < minValue then
					minValue := value;
				values^^[i] := value;
			end;
			conditionalExit(minValue = maxValue, disposeValues);
			lambda := numberOfBins / (maxValue - minValue);
			for i := 1 to numberOfValues do begin
{$IFC AnimationEnabled }
				conditionalExit(userAbort, disposeValues);
{$ENDC }
{$IFC THINK_PASCAL }
				index := trunc(lambda * (values^^[i] - minValue) + 1);
{$ELSEC }
				index := truncLong(lambda * (values^^[i] - minValue) + 1);
{$ENDC }
				if index > numberOfBins then
					index := numberOfBins;
				counters^^[index] := counters^^[index] + 1;
			end;
{$IFC THINK_PASCAL }
			disposHandle(Handle(values));
{$ELSEC }
			DisposeHandle(Handle(values));
{$ENDC }
		end;

		maxCounter := 0;
		for i := 1 to numberOfBins do
			if counters^^[i] > maxCounter then
				maxCounter := counters^^[i];

		currentVectorLength := 2 * numberOfBins + 2;
		init;
		rectangleWidth := width / numberOfBins;
		lambda := height / maxCounter;
		add(originX, originY);
		for i := 1 to numberOfBins do begin
 {$IFC AnimationEnabled }
			conditionalExit(userAbort, disposeResult);
{$ENDC }
			y := originY - counters^^[i] * lambda;
			add((i - 1) * rectangleWidth + originX, y);
			add(i * rectangleWidth + originX, y);
		end;
		add(originX + width, originY);
{$IFC THINK_PASCAL }
		disposHandle(Handle(counters));
{$ELSEC }
		DisposeHandle(Handle(counters));
{$ENDC }
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function histogram.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	begin
		dragVertex := false;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure histogram.interactiveOutput;
	var
		save: PenState;
		poly: PolyHandle;
		x, y: integer;
		i: seqIndex;

	procedure getXYCoordinates (i: seqIndex);
		var
			dummy: boolean;
		begin
			with point2d(ith(i)) do
				dummy := worldToWindow(getX, getY, x, y);
		end;

	begin (* histogram.interactiveOutput *)
		if length > 2 then begin
			GetPenState(save);
{$IFC THINK_PASCAL }
			PenPat(dkGray);
{$ELSEC }
			PenPat(qd.dkGray);
{$ENDC }
			poly := OpenPoly;
			getXYCoordinates(length);
			MoveTo(x, y);
			for i := 1 to length do begin
				getXYCoordinates(i);
				LineTo(x, y);
			end;
			ClosePoly;
			PaintPoly(poly);
			KillPoly(poly);
			SetPenState(save);
		end;
	end;

procedure histogram.interactiveHighlight;
	begin
		inherited interactiveOutput;
	end;

{$IFC OptionalCommand }
function pointVector.distanceDistribution (resolution: longInt; width, height: extended): histogram;
	var
		i, j: longInt;
		result: histogram;

	function generateValue: real;
		begin
			generateValue := point2d(elements^^[i]).distanceLpX(point2d(elements^^[j]));
			j := j + 1;
			if j > length then begin
				i := i + 1;
				j := i + 1;
			end;
		end;

	procedure resetGenerator;
		begin
			i := 1;
			j := 2;
		end;

	begin (* pointVector.distanceDistribution *)
		new(result);
		failNil(Handle(result));
		result.initFromGenerator(length * (length - 1) div 2, resolution, width, height, generateValue, resetGenerator);
		distanceDistribution := result;
	end;
{$ENDC }

end.
