(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit GFUtilities;

{**********************************************									}
{																				}
{		Project		:	MacGFT													}
{		Module		:	Utilities												}
{																				}
{		Autor		:	Bjoern Beeli	, BB									}
{						Informatik, ETH, CH-8092 Zuerich						}
{																				}
{		Datum		:	01.03.91												}
{																				}
{		Sprache		:	THINK Pascal 3.0.1										}
{		Maschine	:	Apple Macitosh II										}
{																				}
{**********************************************									}
{																				}
{		Beschreibung	:														}
{																				}
{		In dieses Modul befinden sich folgende Utilities:						}
{		- ErrorHandling:														}
{			Damit koennen Fehlermeldungen behandelt werden.						}
{																				}
{		- FileNameHandling:														}
{			Damit kann der Benutzer interactiv das zu ladende File-				}
{			Directory	 bzw. File selectieren.									}
{			Diese Prozeduren werden von den GridFile-Modulen nicht				}
{			gebraucht, sie sind fuer die GridFile-Applikationen gedacht.		}
{																				}
{		Diese Unit muss vor dem ersten Gebrauch mit 'UTILITIESinit' 			}
{		initialisiert werden.													}
{																				}
{**********************************************									}
{																				}
{		Aenderungen	:															}
{																				}
{		Datum		Initialen	Kommentar										}
{		13.05.91	BB			Rueckgabe des Directory-Pfades					}
{																				}
{**********************************************}

interface

uses
	GeoBenchUtility;

{**********************************************}

const

{**********************************************}
{**********************************************}
{	Error handling:								 	}

{ allgmeine Fehler: }
	noErr = 0;						{ kein Fehler }
	errNumber = 201;				{ falsche Zahl }
	errIndex = 202;					{ falscher Index (zu gross oder zu klein) }
	errAddr = 203;					{ falsche Adresse }
	errNIL = 204;					{ Zugriff auf NIL-Pointer }
	errPtr = 205;					{ falscher Pointer}

{ Value-Fehler: }
	errType = 210;					{ falscher ValueType }
	errFormat = 211;				{ falsches Format }

{ Attribute-Fehler: }
	errDim = 220;					{ falsche Dimension }
	errProfile = 221;				{ FormatFehler im Profile }

{ Buffer-Fehler: }
	errBufferNil = 230;				{ abstrakter DatenTyp 'pBuffer' ist NIL }
	errBufferNotNil = 231;			{ abstrakter DatenTyp 'pBuffer' ist nicht NIL }
	errBufferFixed = 232;			{ BufferBlock ist noch gefixt. }
	errBufferNotFixed = 233;		{ BufferBlock ist nicht gefixt. }
	errNotInBuffer = 234;			{ Block befindet sich nicht im Buffer. }
	errBufferEOF = 235;				{ EOF: keine weiteren Bloecke vorhanden. }
	errBufferFreeList = 236;		{ gelesener Block gehoert zur FreiListe. }

{ Bucket-Fehler: }
	errBktAddr = 240;				{ falsche BucketAdresse }
	errBktNotInBuff = 241;			{ Bucket nicht im Buffer }
	errBktFull = 242;				{ Bucket ist voll }

{ Directory-Fehler: }
	errDirPageAddr = 250;			{ falsche DirPageAdresse }
	errDirPageNotInBuff = 251;		{ DirPage nicht im Buffer }
	errDirPageFull = 252;			{ DirPage ist voll }
	errDirPageEmpty = 253;			{ DirPage ist leer }

{ Scales-Fehler: }
	errSclExactFind = 260;			{ Such-Schluessel auserhalb des Bereichs. }

{ GFMain-Fehler: }
	errWrongKey = 270;				{ LowKey > HighKey . }
	errTestFail = 271;				{ interner StrukturFehler }

{**********************************************}
{**********************************************}

type
	tModuleName = string[15];
	tProcName = string[20];

	{ Fehler-Code ist ein INTEGER-Type:}
	{ Bedeutung:}
	{ = 0 (noErr): kein Fehler aufgetreten}
	{ < 0 : System-Fehler aufgetreten -> Inside Mac, Appendix A}
	{ > 200 : GFT-Fehler aufgetreten}

	tError = record
			moduleName: tModuleName; { Modul/Prozedur wo der Fehler aufgetreten ist. }
			procName: tProcName;
			id: INTEGER;					{  Art des Fehlers  ( id = noErr, dann ist kein Fehler aufgetreten) }
		end;


{**********************************************}
{*********** ErrorHandling *************************}

procedure ERRORshow (err: tError; inProc: tProcName);
{ Diese Prozedur zeigt die FehlerNummer an und in welchem Modul/Prozedur }
{ der Fehler aufgetreten ist. Danach wird der Fehler geloescht. In 'inProc' kann der Prozedurename }
{ angegeben werden, in welcher die Prozedur ERRORshow aufgerufen wurde.}

procedure ERRORset (var err: tError; moduleName: tModuleName; procName: tProcName; errId: INTEGER);
{ Diese Prozedur setzt den entsprechenden Fehler, falls schon ein Fehler gesetzt ist, wird dieser }
{ angezeigt und danach der neue Fehler gesetzt. }
{ Falls 'errId' = 'noErr' ist, dann wird nichts gemacht }

procedure ERRORclear (var err: tError);
{ Dies Prozedur loescht den Fehler. }


{**********************************************}
{************** Selection ************************}

function PathNameFromWD (vRefNum: longint): Str255;

function PathNameFromDirID (DirID: longint; vRefnum: integer): Str255;

function SELECTdirectory (var dirRefNum: INTEGER): BOOLEAN;
{ Diese Funktion fragt den Benutzer nach dem Directory, wo sich die Daten befinden. }
{ Wenn der Abbruch-Knopf gedrueckt wurde, wird FALSE zurueckgegeben, }
{ sonst TRUE . In 'dirRefNum' wird die Nummer des Directorys zurueckgegeben. }
{ Inside Macintosh: Standard File Package, Package Manager }

function SELECTfile (var fullFileName: Str255; var volume: INTEGER; fileType: OSType): BOOLEAN;
{ Diese Funktion fragt den Benutzer nach einem File. Der Benutzer kann ein File selektieren. }
{ Der FileName, der FileName mit dem Path und die Volume-Nummer wird zurueckgegeben. }
{ Wenn der Abbruch-Knopf gedrueckt wurde, wird FALSE zurueckgegeben. }
{ Information: }
{ Inside Macintosh: Standard File Package, Package Manager }

function SELECTnewFile (prompt: Str255; origName: Str255; var fullFileName: Str255; var volume: INTEGER; var override: BOOLEAN): BOOLEAN;
{ Diese Funktion fragt den Benutzer nach einem neu zu erzeugenden File. Der }
{ Benutzer kann ein File selektieren oder einen neuen Filename eingeben. }
{ Der FileName, der FileName mit dem Path  und die Volume-Nummer wird zurueckgegeben. }
{ Wenn der Abbruch-Knopf gedrueckt wurde, wird FALSE zurueckgegeben. }
{ Information: }
{ Inside Macintosh: Standard File Package, Package Manager }

{$IFC NOT THINK_PASCAL }
function NewFileName(title, suggestion: Str255): Str255;
function OldFileName: Str255;
{$ENDC }

{**********************************************}

function GetFileName (path: Str255): Str255;
{ Diese Funktion extrahiert den FileNamen aus dem Pfad. }

{**********************************************}
{**********************************************}

procedure UTILITIESinit;
{ Diese Prozedur initialisiert das Modul Utilities. Diese Prozedur darf nur }
{ einmal aufgerufen werden. }

{**********************************************}
{**********************************************}
{**********************************************}

implementation

{$IFC NOT THINK_PASCAL }
uses Packages, StandardFile;
{$ENDC }


procedure ERRORclear (var err: tError);
	begin
		with err do begin
			moduleName := '';
			procName := '';
			id := 0;
		end;
	end; { ERRORclear }

procedure ERRORshow (err: tError; inProc: tProcName);
	begin
		with err do begin
			if id <> 0 then begin
				fatalError(StringOf('in GridFile ', procName, ' No.: ', id : 2, '-', inProc));
				ERRORclear(err);
			end;
		end;
	end; { ERRORshow }

procedure ERRORset (var err: tError; moduleName: tModuleName; procName: tProcName; errId: INTEGER);
	begin
		if errId <> noErr then begin
			if err.id <> 0 then begin
				ERRORshow(err, procName);
			end;
			err.moduleName := moduleName;
			err.procName := procName;
			err.id := errId;
		end;
	end; { ERRORset }

{**********************************************}
{**********************************************}
	const
		kHWCfgFlags = $B22;		{ Used to see if A/UX is running						}

	type
		PtrToWord = ^integer;
		PtrToLong = ^LongInt;

	var
		haveAUX: Boolean;						{ Set to TRUE if A/UX is running }

procedure UTILITIESinit;
	begin
		InitPack(stdFile);
{ Test: A/UX }
		haveAUX := BAnd(PtrToWord(kHWCfgFlags)^, $0200) <> 0;
	end; { UTILITIESinit }

{**********************************************}
{**********************************************}

{ Die folgenden Prozeduren sind aus dem Inside Macintosh uebernommen worden. }

	const
		selDirDlogId = 1003;	{ fuer 'SELECTdirectory' }
		selDirBtnId = 11;		{ fuer 'SELECTdirectory' }
		selDirCurBtnId = 12;	{ fuer 'SELECTdirectory' }

	type
		VCBPtr = ^VCB;
		DefVCBPtr = ^VCBPtr;

	var
		selDirOk: BOOLEAN;		{ fuer 'SELECTdirectory' }
		reply: SFReply;			{ fuer 'SELECTdirectory' }
		gDirID: LONGINT; 		{Directory ID waehrend SF Calls }


function GetDiskBlockSize (gVRefNum: INTEGER): INTEGER;
{ Diese Funktion gibt die Groesse eines DiskBlocks auf dem Volume mit der Referenznummer}
{ 'gVRefNum' zurueck. (vielfaches von 512Bytes). }
	begin
		if SetVol(nil, gVRefNum) = noErr then begin
			GetDiskBlockSize := DefVCBPtr($352)^^.vcbAlBlkSiz;
		end
		else
			GetDiskBlockSize := 0;
	end; { GetDiskBlockSize }


{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function GetDirHook (item: INTEGER; theDlog: DialogPtr): INTEGER;
	begin
		GetDirHook := item;
		case item of
			-1: 
				selDirOk := false;
			selDirBtnId:  begin
				if LongInt(reply.fType) <> 0 then begin
					gDirID := LongInt(reply.fType);
					GetDirHook := getCancel;
					selDirOk := true;
				end; (* IF *)
			end; (* CASE selDirBtnId *)
			selDirCurBtnId:  begin
				gDirID := PtrToLong($398)^;
				GetDirHook := getCancel;
				selDirOk := true;
			end; (* CASE selDirCurBtnId *)
			otherwise
				;
		end; (* CASE *)
	end; (* GetDirHook *)
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function IsFolder (p: ParmBlkPtr): BOOLEAN;
	begin
		IsFolder := not (BTst(p^.ioFlAttrib, 4));
	end; (* IsFolder *)

function SELECTdirectory (var dirRefNum: INTEGER): BOOLEAN;
	var
		typeList: SFTypeList;
		vRefNum: INTEGER; 	{ VolumeReferenzNummer (Inside Mac -> File Manager)}
	begin
		SELECTdirectory := false;
{$IFC THINK_PASCAL }
		SFPGetFile(Point($00640064), '', @IsFolder, -1, typeList, @GetDirHook, reply, selDirDlogId, nil);
{$ELSEC }
		SFPGetFile(Point($00640064), '', NewFileFilterProc(ProcPtr(@IsFolder)), -1, @typeList, NewDlgHookProc(@GetDirHook), reply, selDirDlogId, nil);
{$ENDC }
		if selDirOk then begin
			vRefNum := -(PtrToWord($214)^);
			if OpenWD(vRefNum, gDirID, $00112233, dirRefNum) = noErr then begin
{    gDiskBlockSize := GetDiskBlockSize(vRefNum);}
				SELECTdirectory := true;
			end;
		end; (* IF *)
	end; (* SELECTdirectory *)

{**********************************************}
{**********************************************}
{#		Apple Macintosh Developer Technical Support}
{#		Standard File Sample Application		StdFile}

(** PathNameFromDirID ************************************ *)
(*}
{(*	Given a DirID and real vRefnum, this routine will create and return the}
{(*	full pathname that corresponds to it. It does this by calling PBGetCatInfo}
{(*	for the given directory, and finding out its name and the DirID of its}
{(*	parent. It the performs the same operation on the parent, sticking ITS}
{(*	name onto the beginning of the first directory. This whole process is}
{(*	carried out until we have processed the root directory (identified with}
{(*	a DirID of 2.}
{(*}
{(*	NOTE: This routine is now A/UX friendly. A/UX likes sub-directories}
{(*				separated by slashes in a pathname. This routine automatically}
{(*				uses colons or slashes as separators based on the value of the}
{(*				global gHasAUX.  This global must be initialized correctly for}
{(*				this routine to do its thing. However, because of this dependancy}
{(*				on the idiosyncracies of file systems, generating full pathnames}
{(*				for other than display purposes is discouraged; it's changed in}
{(*				the past when A/UX was implemented, and it may change again in}
{(*				the future it support for other file systems such as ProDOS,}
{(*				MS-DOS, or OS/2 are added.}
{(*}
{(***************************************************** *)

function PathNameFromDirID (DirID: longint; vRefNum: integer): Str255;
	var
		block: CInfoPBRec;
		directoryName, fullPathName: Str255;
	begin
		fullPathName := '';
		with block do begin
			ioNamePtr := @directoryName;
			ioDrParID := DirID;
		end;
		repeat
			with block do begin
				ioVRefNum := vRefNum;
				ioFDirIndex := -1;
				ioDrDirID := block.ioDrParID;
			end;
{$IFC THINK_PASCAL }
			testOSError(PBGetCatInfo(@block, FALSE), noErr);
{$ELSEC }
			testOSError(PBGetCatInfoSync(@block), noErr);
{$ENDC }
			if haveAUX then begin
				if directoryName[1] <> '/' then begin
						{ If this isn't root (i.e. "/"), append a slash ('/') }
					directoryName := concat(directoryName, '/');
				end;
			end
			else begin
				directoryName := concat(directoryName, ':');
			end;
			fullPathName := concat(directoryName, fullPathName);
		until (block.ioDrDirID = 2);
		PathNameFromDirID := fullPathName;
	end; { PathNameFromDirID }


(** PathNameFromWD ************************************** *)
(*}
{(*	Given an HFS working directory, this routine returns the full pathname}
{(*	that corresponds to it. It does this by calling PBGetWDInfo to get the}
{(*	VRefNum and DirID of the real directory. It then calls PathNameFromDirID,}
{(*	and returns its result.}
{(*}
{(**************************************************** *)

function PathNameFromWD (vRefNum: longint): Str255;
	var
		myBlock: WDPBRec;
	begin
			{}
{			{ PBGetWDInfo has a bug under A/UX 1.1.  If vRefNum is a real vRefNum}
{			{ and not a wdRefNum, then it returns garbage.  Since A/UX has only 1}
{			{ volume (in the Macintosh sense) and only 1 root directory, this can}
{			{ occur only when a file has been selected in the root directory (/).}
{			{ So we look for this and hardcode the DirID and vRefNum. }

		if (haveAUX) & (vRefNum = -1) then begin
			PathNameFromWD := PathNameFromDirID(2, -1);
		end
		else begin
			with myBlock do begin
				ioNamePtr := nil;
				ioVRefNum := vRefNum;
				ioWDIndex := 0;
				ioWDProcID := 0;
			end;
			{ Change the Working Directory number in vRefnum into a real vRefnum }
			{ and DirID. The real vRefnum is returned in ioVRefnum, and the real }
			{ DirID is returned in ioWDDirID. }
{$IFC THINK_PASCAL }
			testOSError(PBGetWDInfo(@myBlock, FALSE), noErr);
{$ELSEC }
			testOSError(PBGetWDInfoSync(@myBlock), noErr);
{$ENDC }
			with myBlock do
				PathNameFromWD := PathNameFromDirID(ioWDDirID, ioWDVRefNum)
		end;
	end; { PathNameFromWD }

{**********************************************}

function NotOpenFilter (p: ParmBlkPtr): BOOLEAN;
{ zeigt nur geschlossene Files an. }
	begin
		NotOpenFilter := BTst(p^.ioFlAttrib, 7);
	end; { NotOpenFilter }

{**********************************************}

function SELECTfile (var fullFileName: Str255; var volume: INTEGER; fileType: OSType): BOOLEAN;
	var
		typeList: SFTypeList;
		reply: SFReply;
	begin
		typeList[0] := fileType;	{ Filter auf FileType }
{$IFC THINK_PASCAL }
		SFGetFile(Point($00640064), '', @NotOpenFilter, 1, typeList, nil, reply);
{$ELSEC }
		SFGetFile(Point($00640064), '', NewFileFilterProc(ProcPtr(@NotOpenFilter)), 1, @typeList, nil, reply);
{$ENDC }
		if reply.good then begin
			fullFileName := CONCAT(PathNameFromWD(reply.vRefNum), reply.fName);
			volume := reply.vRefNum;
		end;
		SELECTfile := reply.good;
	end; { SELECTfile }

{**********************************************}

function SELECTnewFile (prompt: Str255; origName: Str255; var fullFileName: Str255; var volume: INTEGER; var override: BOOLEAN): BOOLEAN;
	var
		reply: SFReply;
		refNum: INTEGER;
		err: OSErr;
	begin
		SFPutFile(Point($00640064), prompt, origName, nil, reply);
		if reply.good then begin
			fullFileName := CONCAT(PathNameFromWD(reply.vRefNum), reply.fName);
			volume := reply.vRefNum;

			err := FSOpen(reply.fName, reply.vRefNum, refNum);	{ exist File ?}
			if err = noErr then begin
				err := FSClose(refNum);
				override := true;
			end
			else
				override := false;
		end;
		SELECTnewFile := reply.good;
	end; { SELECTnewFile }

{**********************************************}

function GetFileName (path: Str255): Str255;
{ Gibt den FileNamen aus dem Pfad zurueck. }
	var
		i: INTEGER;
	begin
		repeat
			i := POS(':', path);
			if i = 0 then
				LEAVE;
			DELETE(path, 1, i);
		until FALSE;
		GetFileName := path;
	end; { GetFileName }

{**********************************************}

{$IFC NOT THINK_PASCAL }
function NewFileName(title, suggestion: Str255): Str255;
	var
		fullFileName: Str255;
		volume: integer;
		override: boolean;
	begin
		fullFileName := '';
		volume := 0;
		override := false;
		if SELECTnewFile(title, suggestion, fullFileName, volume, override) then
			NewFileName := GetFileName(fullFileName)
		else
			NewFileName := '';
	end;

function OldFileName: Str255;
	var
		fullFileName: Str255;
		volume: integer;
	begin
		if SELECTfile(fullFileName, volume, 'DATA') then
			OldFileName := GetFileName(fullFileName)
		else
			OldFileName := '';
	end;
{$ENDC }

end. { Utilities }
