(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit GeoBenchAppleEvents;

interface

uses
{$IFC NOT THINK_PASCAL }
	Files, 
{$ENDC }
	AppleEvents;

function HandleOAPP (var theAppleEvent: AppleEvent; var reply: AppleEvent; handlerRefcon: LongInt): OSErr;
function HandleODOC (var theAppleEvent: AppleEvent; var reply: AppleEvent; handlerRefcon: LONGINT): OSErr;
function HandlePDOC (var theAppleEvent: AppleEvent; var reply: AppleEvent; handlerRefcon: LONGINT): OSErr;
function HandleQUIT (var theAppleEvent: AppleEvent; var reply: AppleEvent; handlerRefcon: LONGINT): OSErr;
function ProcessItem (myFSSPtr: FSSpecPtr): OSErr;
function ProcessFolder (myFSSPtr: FSSpecPtr): OSErr;
procedure InstallOtherEvents;
procedure OpenApp;
procedure QuitApp;
function PreFlightDocs (opening: Boolean; itemCount: integer; var userDataHandle: Handle): Boolean;
procedure OpenDoc (myFSSPtr: FSSpecPtr; opening: Boolean; userDataHandle: Handle);
procedure PostFlightDocs (opening: Boolean; itemCount: Integer; userDataHandle: Handle);
procedure InitAEVTStuff;
procedure DSInitGlobals;

{	set qWalkFolders to true if you want each item of a dropped folder processed	}
{	as an individual item															}
const
qWalkFolders = false;

var
gOApped: Boolean;
{$IFC THINK_PASCAL }
OAPPHandlerUPP, ODOCHandlerUPP, PDOCHandlerUPP, QUITHandlerUPP: EventHandlerProcPtr;
{$ELSEC }
OAPPHandlerUPP, ODOCHandlerUPP, PDOCHandlerUPP, QUITHandlerUPP: AEEventHandlerUPP;
{$ENDC }

implementation

uses
{$IFC NOT THINK_Pascal}
	StringCompare, 
{$ENDC}
	GeoBenchFileMenu, GestaltEqu;
{	This routine does all initialization for AEM, including the}
{	creation and then population of the dispatch table.}
{$S Initialize}

{$IFC THINK_Pascal}
function NewAEEventHandlerProc (proc: ProcPtr): ProcPtr;
	begin
		NewAEEventHandlerProc := proc;
	end;
{$ENDC}

procedure InitAEVTStuff;
	var
		aevtErr: OSErr;
	begin
		aevtErr := noErr;
		if (aevtErr = noErr) then begin
			OAPPHandlerUPP := NewAEEventHandlerProc(@HandleOAPP);
			aevtErr := AEInstallEventHandler(kCoreEventClass, kAEOpenApplication, OAPPHandlerUPP, 0, false);
		end;
		if (aevtErr = noErr) then begin
			ODOCHandlerUPP := NewAEEventHandlerProc(@HandleODOC);
			aevtErr := AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments, ODOCHandlerUPP, 0, false);
		end;
		if (aevtErr = noErr) then begin
			PDOCHandlerUPP := NewAEEventHandlerProc(@HandlePDOC);
			aevtErr := AEInstallEventHandler(kCoreEventClass, kAEPrintDocuments, PDOCHandlerUPP, 0, false);
		end;
		if (aevtErr = noErr) then begin
			QUITHandlerUPP := NewAEEventHandlerProc(@HandleQUIT);
			aevtErr := AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, QUITHandlerUPP, 0, false);
		end;
		if (aevtErr = noErr) then
			InstallOtherEvents;
		if (aevtErr <> noErr) then
			;		{ report an error if you are so included}
	end;



{	This routine is a utility routine for checking that all required }
{	parameters in the Apple event have been used.}
{$S Main}
function GotRequiredParams (var theAppleEvent: AppleEvent): OSErr;
	var
		typeCode: DescType;
		actualSize: Size;
		retErr, err: OSErr;
	begin
		err := AEGetAttributePtr(theAppleEvent, keyMissedKeywordAttr, typeWildCard, typeCode, nil, 0, actualSize);
		if (err = errAEDescNotFound) then	{ we got all the required params: all is ok}
			retErr := noErr
		else if (err = noErr) then
			retErr := errAEEventNotHandled
		else
			retErr := err;
		GotRequiredParams := retErr;
	end;

{	This routine is the handler for the oapp (Open Application) event.}
{	It first checks the number of parameters to make sure we got them all }
{	(even though we don't want any) and then calls the OpenApp userProc in QSUserProcs.}
{	Finally it checks to see if the caller wanted a reply & sends one, setting any error.}
{$S Main}
{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function HandleOAPP (var theAppleEvent: AppleEvent; var reply: AppleEvent; handlerRefcon: LongInt): OSErr;
	var
		errStr: Str255;
	begin
		testOSError(GotRequiredParams(theAppleEvent), noErr);
		OpenApp;		{ pass it on to the app specific routine}
		if reply.dataHandle <> nil then { a reply is sought }
			begin
			errStr := 'Opening';
			HandleOAPP := AEPutParamPtr(reply, 'errs', 'TEXT', Ptr(@errStr[1]), length(errStr));
		end
		else
			HandleOAPP := noErr;
	end;

{	This routine is the handler for the quit (Quit Application) event.}
{	It first checks the number of parameters to make sure we got them all }
{	(even though we don't want any) and then calls the QuitApp userProc in QSUserProcs.}
{	Finally it checks to see if the caller wanted a reply & sends one, setting any error.}
{$S Main}
function HandleQUIT (var theAppleEvent: AppleEvent; var reply: AppleEvent; handlerRefcon: LONGINT): OSErr;
	var
		errStr: Str255;
	begin
		testOSError(GotRequiredParams(theAppleEvent), noErr);
		QuitApp;		{ pass it on to the app specific routine}
		if reply.dataHandle <> nil then { a reply is sought }
			begin
			errStr := 'Quiting';
			HandleQUIT := AEPutParamPtr(reply, 'errs', 'TEXT', Ptr(@errStr[1]), length(errStr));
		end
		else
			HandleQUIT := noErr;
	end;

{	This routine is the low level processing routine for both the }
{	odoc (Open Document) and pdoc (Print Document) events.}
{	This routine is the key one, since this is how we get the list of}
{	files/folders/disks to process.  The first thing to do is the get the}
{	list of files, and then make sure that's all the parameters (should be!).}
{	We then send call the PreflightDocs routine (from DSUserProcs), process}
{	each file in the list by calling OpenDoc (again in DSUserProcs), and finally}
{	call PostflightDocs (you know where) and setting a return value.}
{$S Main}
function _HandleDocs (var theAppleEvent: AppleEvent; var reply: AppleEvent; opening: Boolean): OSErr;
	var
		err: OSErr;
		myFSS: FSSpec;
		docList: AEDescList;
		index, itemsInList: LongInt;
		actualSize: Size;
		keywd: AEKeyword;
		typeCode: DescType;
		userDataHandle: Handle;
	begin
		testOSError(AEGetParamDesc(theAppleEvent, keyDirectObject, typeAEList, docList), noErr);
		testOSError(GotRequiredParams(theAppleEvent), noErr);
{		How many items do we have?. }
{ NOTE: Moved here in DS 2.0 due to requests for this info in preflighter }
		err := AECountItems(docList, itemsInList);
		testOSError(err, noErr);
		if (PreFlightDocs(opening, itemsInList, userDataHandle)) then begin 	{ let the app do any preflighting it might need}
			for index := 1 to itemsInList do begin
				testOSError(AEGetNthPtr(docList, index, typeFSS, keywd, typeCode, @myFSS, sizeof(myFSS), actualSize), noErr);
				OpenDoc(@myFSS, opening, userDataHandle);	{call the open routine}
			end;
			PostFlightDocs(opening, itemsInList, userDataHandle);	{ cleanup time}
		end
		else
			err := errAEEventNotHandled;	{ tells AEM that we didn't handle it!}
		testOSError(AEDisposeDesc(docList), noErr);
		_HandleDocs := err;
	end;

{	This routine is the handler for the odoc (Open Document) event.}
{	The odoc event simply calls the common _HandleDocs routines, which will}
{	do the dirty work of parsing the AEVT & calling the userProcs.}
{$S Main}
function HandleODOC (var theAppleEvent: AppleEvent; var reply: AppleEvent; handlerRefcon: LONGINT): OSErr;
	begin
		HandleODOC := _HandleDocs(theAppleEvent, reply, TRUE);	{call the low level routine}
	end; { HandleODOC }

{	This routine is the handler for the pdoc (Print Document) event.}
{	The pdoc event like the odoc simply calls the common _HandleDocs routines}
function HandlePDOC (var theAppleEvent: AppleEvent; var reply: AppleEvent; handlerRefcon: LONGINT): OSErr;
	begin
		HandlePDOC := _HandleDocs(theAppleEvent, reply, FALSE);	{call the low level routine}
	end; { HandlePDOC }
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

{$S Main}
{---------------------------------------------------------------------------------------------------}
{	This routine is called during init time.}
{	}
{	It allows you to install more AEVT Handlers beyond the standard four}
{---------------------------------------------------------------------------------------------------}
procedure InstallOtherEvents;
{---------------------------------------------------------------------------------------------------}
	begin
	end;

{$S Main}
{---------------------------------------------------------------------------------------------------}
{	This routine is called when an OAPP event is received.}
{	}
{	Currently, all it does is set the gOApped flag, so you know that}
{	you were called initally with no docs, and therefore you shouldn't }
{	quit when done processing any following odocs.}
{---------------------------------------------------------------------------------------------------}
procedure OpenApp;
{---------------------------------------------------------------------------------------------------}
	var
		w: windowIndex;
	begin
		gOApped := true;
		w := makeWindow(niceRect(false), '', false, false, 0);
	end;

{$S Main}
{---------------------------------------------------------------------------------------------------}
{	This routine is called when an QUIT event is received.}
{	We simply set the global done flag so that the main event loop can}
{	gracefully exit.  We DO NOT call ExitToShell for two reasons:}
{	1) It is a pretty ugly thing to do, but more importantly}
{	2) The Apple event manager will get REAL upset!}
{---------------------------------------------------------------------------------------------------}
procedure QuitApp;
{---------------------------------------------------------------------------------------------------}
	begin
		SkelWhoa;
	end;



{$S Main}
{---------------------------------------------------------------------------------------------------}
{	This routine is the first one called when an ODOC or PDOC event is received.}
{	In this routine you would place code used to setup structures, etc. }
{	which would be used in a 'for all docs' situation (like "Archive all}
{	dropped files")}
{	}
{	Obviously, the opening boolean tells you whether you should be opening}
{	or printing these files based on the type of event recieved.}
{	}
{	NEW IN 2.0!}
{	The itemCount parameter is simply the number of items that were dropped on}
{	the application and that you will be processing.  This gives you the ability}
{	to do a single preflight for memory allocation needs, rather than doing it}
{	once for each item as in previous versions.}
{	}
{	userDataHandle is a handle that you can create & use to store your own}
{	data structs.  This dataHandle will be passed around to the other }
{	odoc/pdoc routines so that you can get at your data without using}
{	globals - just like the new StandardFile.  }
{	}
{	We also return a boolean to tell the caller if you support this type}
{	of event.  By default, our dropboxes don't support the pdoc, so when}
{	opening is FALSE, we return FALSE to let the caller send back the}
{	proper error code to the AEManager.}
{	}
{	You will probably want to remove the #pragma unused (currently there to fool the compiler!)}
{---------------------------------------------------------------------------------------------------}
{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function PreFlightDocs (opening: Boolean; itemCount: integer; var userDataHandle: Handle): Boolean;
{---------------------------------------------------------------------------------------------------}
	begin

		PreFlightDocs := opening;

	{ we support opening, but not printing - see above}
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

{	Is the file represented by this FSSpec really a folder? }
{-------------------------------------------------------------------------------}
function FSpIsFolder (theFSSpec: FSSpecPtr): Boolean;
{-------------------------------------------------------------------------------}
	var
		err: OSErr;
		pb: CInfoPBRec;
		fName: Str255;
	begin
		if (theFSSpec^.parID = fsRtParID) then	{ it's a volume!!}
			FSpIsFolder := true;

		BlockMove(@theFSSpec^.name, @fName, 32);
		with pb do begin
			ioDirID := theFSSpec^.parID;
			ioCompletion := nil;
			ioNamePtr := @fName;
			ioVRefNum := theFSSpec^.vRefNum;
			ioFDirIndex := 0;
			ioFVersNum := 0;
		end;
		err := PBGetCatInfoSync(@pb);

		if (err = noErr) then begin
			if BTST(longint(pb.ioFlAttrib), ioDirMask) then
				FSpIsFolder := (true);
		end;
		FSpIsFolder := (false)

	end;

{$S Main}
{---------------------------------------------------------------------------------------------------}
{	This routine is called for each file passed in the ODOC event.}
{	}
{	In this routine you would place code for processing each file/folder/disk that}
{	was dropped on top of you.}
{	}
{	You will probably want to remove the #pragma unused (currently there to fool the compiler!)}
{---------------------------------------------------------------------------------------------------}
{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure OpenDoc (myFSSPtr: FSSpecPtr; opening: Boolean; userDataHandle: Handle);
{---------------------------------------------------------------------------------------------------}
	var
		err: OSErr;
	begin
		err := noErr;
		if qWalkFolders then begin
			{	For this case we need to determine if the FSSpec is a file or folder.}
			{	If it's a folder, we then need to process each item in that folder,}
			{	otherwise just process the item.}
			if FSpIsFolder(myFSSPtr) then
				err := ProcessFolder(myFSSPtr)
			else
				err := ProcessItem(myFSSPtr);
		end
		else begin
		{For this case we just call ProcessItem on the FSSpec above.}
			err := ProcessItem(myFSSPtr);
		end
	{ you should probably do something if you get back an error ;)}
	end;


{$S Main}
{---------------------------------------------------------------------------------------------------}
{	This routine is the last routine called as part of an ODOC event.}
{	}
{	In this routine you would place code to process any structures, etc. }
{	that you setup in the PreflightDocs routine.}
{}
{	NEW IN 2.0!}
{	The itemCount parameter was the number of items that you processed.}
{	It is passed here just in case you need it ;)  }
{	}
{	If you created a userDataHandle in the PreFlightDocs routines, this is}
{	the place to dispose of it since the Shell will NOT do it for you!}
{	}
{	You will probably want to remove the #pragma unusued (currently there to fool the compiler!)}
{---------------------------------------------------------------------------------------------------}
procedure PostFlightDocs (opening: Boolean; itemCount: Integer; userDataHandle: Handle);
{---------------------------------------------------------------------------------------------------}
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }


{---------------------------------------------------------------------------------------------------}
{	This routine gets called for each item (which could be either a file or a folder)}
{	that the caller wants dropped.  The determining factor is the definition of the }
{	qWalkFolder compiler directive.   Either way, the item in question should be}
{	processed as a single item and not "dissected" into component units (like subfiles}
{	of a folder!)}
{---------------------------------------------------------------------------------------------------}
function ProcessItem (myFSSPtr: FSSpecPtr): OSErr;
{---------------------------------------------------------------------------------------------------}

	var
		info: FInfo;
	begin
	{ do something here}
		testOSError(FSpGetFInfo(myFSSPtr^, info), noErr);
		if (info.fdType = gbFileType) & (info.fdCreator = gbCreator) then
			inputFile(CONCAT(PathNameFromDirID(myFSSPtr^.parID, myFSSPtr^.vRefNum), myFSSPtr^.name), myFSSPtr^.vRefNum);
		ProcessItem := noErr;
	end;

{---------------------------------------------------------------------------------------------------}
{	This routine gets called for any folder (or disk) that the caller wants }
{	processed as a set of component items, instead of as a single entity.}
{	The determining factor is the definition of the qWalkFolder compiler directive.}
{---------------------------------------------------------------------------------------------------}
function ProcessFolder (myFSSPtr: FSSpecPtr): OSErr;
{---------------------------------------------------------------------------------------------------}
	var
		err: OSErr;
		index, oldIndex, localIndex: Integer;
		localFSSpec, curFSSpec: FSSpec;
		cipb: CInfoPBRec;
		fName, vFName: Str255;
		dirID, origDirID: LongInt;
		foundPosition: Boolean;
	begin
		err := noErr;

 	{ copy the source locally to avoid recursion problems}
{$IFC THINK_Pascal}
		BlockMove(Ptr(myFSSPtr), @localFSSpec, sizeof(FSSpec));
	{	get the dirID for THIS folder, not it's parent!}
		BlockMove(@localFSSpec.name, @fName, 32);
{$ELSEC}
		BlockMoveData(Ptr(myFSSPtr), @localFSSpec, sizeof(FSSpec));
	{	get the dirID for THIS folder, not it's parent!}
		BlockMoveData(@localFSSpec.name, @fName, 32);
{$ENDC}
		cipb.ioCompletion := nil;
		cipb.ioNamePtr := @fName;
		cipb.ioVRefNum := localFSSpec.vRefNum;
		cipb.ioFDirIndex := 0;	{ use the dir & vRefNum;}
		cipb.ioDirID := localFSSpec.parID;
		err := PBGetCatInfoSync(@cipb);

		if (err = noErr) then begin
			origDirID := cipb.ioDrDirID; { copy the sucker}
			index := 1;

		{ index through all contents of this folder}
			while (err = noErr) do begin
				dirID := origDirID;
				localIndex := index;
				fName := '';
				cipb.ioCompletion := nil;
				cipb.ioNamePtr := @fName;
				cipb.ioVRefNum := localFSSpec.vRefNum;
				cipb.ioFDirIndex := localIndex;	{ use a real index}
				cipb.ioDirID := dirID;
				err := PBGetCatInfoSync(@cipb);

				if (err = noErr) then begin
{$IFC THINK_Pascal}
					BlockMove(@fName, @curFSSpec.name, 32);
{$ELSEC}
					BlockMoveData(@fName, @curFSSpec.name, 32);
{$ENDC}
					curFSSpec.vRefNum := cipb.ioVRefNum;
					curFSSpec.parID := dirID;
{Check to see if this entry is a folder.}
					if BAND(cipb.ioFlAttrib, ioDirMask) = ioDirMask then begin
						err := ProcessFolder(@curFSSpec);
					end
					else
						err := ProcessItem(@curFSSpec);

				{	If we've had an error, get out! }
					if (err <> noErr) then
						leave;

				{ dirID = origDirID;	}
					localIndex := index;
				{Now take into account new files being created}
				{in the current directory & messing up our index.}
				{See Dev.CD Vol. XI:Tools & Apps (Moof!):Misc Utilities:}
				{Disinfectant & Source 2.5.1:Sample:Notes:Scan Alg	}

					vFName := '';
					cipb.ioCompletion := nil;
					cipb.ioNamePtr := @vFName;
					cipb.ioVRefNum := localFSSpec.vRefNum;
					cipb.ioFDirIndex := localIndex;	{ use a real index}
					cipb.ioDirID := dirID;
					err := PBGetCatInfoSync(@cipb);
					oldIndex := index;
					if (err = noErr) then begin
					{	If they're equal - same place, go to next }
						if (EqualString(vFName, fName, false, false)) then
							index := index + 1;
					end;

				{	If we didn't advance, then perhaps a file was created or deleted }
					if (oldIndex = index) then begin
						oldIndex := index;	{ save off the old }
						index := 0;		{ and start at the beginning }
						err := noErr;
						vFName := '';
						foundPosition := false;

						while (not foundPosition) do begin
							index := index + 1;
							vFName := '';
							cipb.ioCompletion := nil;
							cipb.ioNamePtr := @vFName;
							cipb.ioVRefNum := localFSSpec.vRefNum;
							cipb.ioFDirIndex := index;	{ now use a real index }
							cipb.ioDirID := dirID;
							err := PBGetCatInfoSync(@cipb);

							if (err = fnfErr) then begin   { we've just been deleted}
								index := oldIndex;
								foundPosition := true;
								err := noErr;	{ have to remember to reset this!}
							end;

					{	found same file & same index position }
					{	so try the next item }
							if ((not foundPosition) and EqualString(fName, vFName, false, false)) then begin
								index := index + 1;
								foundPosition := true;
							end;
						end;
					end;
				end;
			end;
		end;

		ProcessFolder := err;
	end;

procedure DSInitGlobals;
	var
		aLong: LongInt;
	begin
		gOApped := false;	{ probably not since users are supposed to DROP things!}
		assert(Gestalt(gestaltAppleEventsAttr, aLong) = noErr, 'Apple Events available');
		InitAEVTStuff;
	end;

end.
