(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit GeoBenchUtility;

interface

uses
{$IFC THINK_PASCAL }
	sane, 
{$ELSEC }
	fp, Resources, Devices, Dialogs, Controls, ToolUtils, DiskInit, saveDrawing, Sound, QD3DMath,
{$ENDC }
	script, AppleEvents;

(* system error *)
procedure testOSError (result, exception: OSErr);

procedure fatalError (text: Str255);

procedure nonFatalError (text: Str255);

procedure failNil (h: Handle);

function failNilF (h: Handle): Handle;

function newHandleFailNil (size: longInt): Handle;

procedure unimplementedMethod (name: Str255);
(* warns the user that an unimplemented method is called *)

procedure DoCautionAlert (alertNumber: integer);

procedure wrongVersion;

procedure assert (condition: boolean; t: Str255);
(* If 'condition = FALSE' the user gets the warning that the assertion 't' has failed.											*)
(* The user can abort the program or continue although continuation often returns in unexpected failure.				*)

var
startTime, startTimeCorrections: longInt;			(* For timing the duration of operations and random creations.		*)
isSystemSeven: boolean;								(* true iff system 7 or higher of the MacOS is running.				*)
rebuildOperationsMenu, redrawScreen: boolean;
(* set to true by a command, if the operations menu needs to be rebuild or the screen needs to be redrawn, e.g. the	*)
(* command changes 'currentArguments'. 																							*)

(* random numbers *)

{$IFC NOT THINK_PASCAL }
function truncLong (x: extended): longInt;
function roundLong (x: extended): longInt;
procedure PaintOval2 (t, l, b, r: integer);
procedure PaintRect2 (t, l, b, r: integer);
procedure PaintArc2 (t, l, b, r, startAngle, arcAngle: integer);
procedure FrameRect2 (t, l, b, r: integer);
procedure FrameOval2 (t, l, b, r: integer);
procedure FrameArc2 (t, l, b, r, startAngle, arcAngle: integer);
procedure IOCheck (on: boolean);
const pi = kQ3Pi;
{$ENDC }

procedure initRandom;

procedure randomize;

function goodRandomizerAvailable: boolean;

function randomReal: extended;

function randomLongInt: longInt;

function randomRange (beginRange, endRange: longInt): longInt; (* beginRange <= randomRange < endRange *)

function getRandomSeed: extended;

procedure setRandomSeed (seed: extended);


(* dialog support *)
procedure GetDText (theDlg: DialogPtr; item: integer; var str: Str255);

procedure SetDText (theDlg: DialogPtr; item: integer; str: Str255);

function GetDCtl (theDlg: DialogPtr; item: integer): boolean;

procedure SetDCtl (theDlg: DialogPtr; item: integer; value: boolean);

procedure SetDProc (theDlg: DialogPtr; item: integer; p: ProcPtr);

procedure setAndHighlightX (theDlg: DialogPtr; item: integer; x: extended; precision: integer);

procedure setAndHighlight (theDlg: DialogPtr; item: integer; x: extended);

procedure ActivateItem (theDlg: DialogPtr; item: integer; active: boolean);

procedure HiliteIfAllowed (pDlog: DialogPtr; pItem: integer; howLong: longInt; var vItem: integer);

function dragDialogFilter (theDialog: DialogPtr; var theEvent: EventRecord; var itemHit: integer): boolean;

procedure niceWindowPosition (w: WindowPtr; adjust: boolean);

procedure showDefaultButton (theDlg: DialogPtr; item: integer);

function askUserBoolean (question: Str255): boolean;

function userAgreesToAbort: boolean;

(* animation *)
const
animationMenuString = 'All Animations On;All Animations Off;(-';
allAnimationsOnItem = 1;
allAnimationsOffItem = allAnimationsOnItem + 1;

minAnimation = allAnimationsOffItem + 2;

(* begin Marco Ciot *)
ARelativeNeighborhoodItem = minAnimation;
{$IFC Tutorial }
ACenterOfGravityItem = ARelativeNeighborhoodItem + 1;
AClipWithRectangleItem = ACenterOfGravityItem + 1;
{$ELSEC }
AClipWithRectangleItem = ARelativeNeighborhoodItem + 1;
{$ENDC }
(* end Marco Ciot *)
AClosestPairItem = AClipWithRectangleItem + 1;
AContourOfRectanglesItem = AClosestPairItem + 1;
AConvexHullItem = AContourOfRectanglesItem + 1;
AConvexPolygonDistanceItem = AConvexHullItem + 1;
ADiameterItem = AConvexPolygonDistanceItem + 1;

{$IFC Chemistry }
ADistanceGeometryItem = ADiameterItem + 1;
ALineSegmentIntersectionItem = ADistanceGeometryItem + 1;
{$ELSEC }
ALineSegmentIntersectionItem = ADiameterItem + 1;
{$ENDC }

ALowerEnvelopeItem = ALineSegmentIntersectionItem + 1;
AMinimalAreaDiskItem = ALowerEnvelopeItem + 1;
AMinimumSpanningTreeItem = AMinimalAreaDiskItem + 1;
ANearestNeighborsItem = AMinimumSpanningTreeItem + 1;
APolygonIntersectionItem = ANearestNeighborsItem + 1;
ASearchItem = APolygonIntersectionItem + 1;
AShortestPathItem = ASearchItem + 1;
ATangentItem = AShortestPathItem + 1;
ATravelingSalesmanItem = ATangentItem + 1;
ATriangulationItem = ATravelingSalesmanItem + 1;
AVoronoiDiagramItem = ATriangulationItem + 1;
AWindingNumberItem = AVoronoiDiagramItem + 1;

maxAnimation = AWindingNumberItem;

type
animationVector = array[minAnimation..maxAnimation] of boolean;

var
animationFlag: animationVector;
animationMenu, appleMenu: MenuHandle;
continueButton: DialogPtr;
animationIsOn, doBeep: boolean;
abortFlag, abortEnable: boolean;	(*	set by animation dialog-box and checked by 'userAbort' *)
movieWait: real;					(*	speed of the animation *)

procedure beginAnimation;

procedure waitForClick (var animate: boolean);

procedure endAnimation;

function animationEnabled: boolean;

procedure setAllAnimationFlag (value: boolean);

procedure initAnimation;

function AbortInQueue: Boolean;

function AbortInQueueUserClick (var userClicked: boolean; var whichPort: GrafPtr; var whichPoint: Point): Boolean;

function userAbort: boolean;

(* How to add user interrupts to geometric routines. The boolean function 'userAbort' in		*)
(* 'GeoBenchUtility'  is 'true' if the user is currently holding 'command' and '.' down. This	*)
(* condition should be checked frequently. If 'userAbort = true' the routine should exit and	*)
(* reclaim all memory it has allocated. The result should be 'nil', although the implementor	*)
(* can check for an user interrupt by checking 'abortFlag'. If a routine is to be executed		*)
(* without interrupts, set the boolean 'abortEnable' to false, execute the routine and set		*)
(* 'abortEnable' back to its original value. Note also that a call to 'userAbort' advances the	*)
(* the handle of the stop watch.																			*)


(* cursor handling *)
type
cursorType = (arrowCursor, stopwatchCursor, littleCrossCursor);

const
version0 = 1908196100 + 19;
version1 = version0;
version2 = version1 + 17;
version3 = version2 + 13;
version4 = version3 + 1;
version5 = version4 + 1991;
version6 = version4 + 1; {!}
version7 = version6 + 1;
version8 = version7 + 1;
version9 = version8 + 1;
version10 = version9 + 1;
version11 = version10 + 1;
{$IFC THINK_PASCAL }
thisVersion = version11;
{$ELSEC }
version12 = version11+100;
thisVersion = version12;
{$ENDC }

{$IFC THINK_PASCAL }
paletteWidth = 18;		(*	for Metrowerks defined in saveWindow	*)
scrollBarWidth = 15;	(*	for Metrowerks defined in saveWindow	*)
{$ENDC }


commandKeyValue = 55;
shiftKeyValue = 56;
optionKeyValue = 58;
controlKeyValue = 59;

var
startMemory: longInt;
infoWindow, currentGeoBenchWindow: WindowPtr;	(* initialized by geoBenchMain 						*)
currentVersion: longInt;						(* version number of file currently open (if any)	*)
globalKeyMap: KeyMap;							(* used as a temporary variable						*)
lpMetric: extended;								(* p of the lp-metric to be used					*)

procedure cursorSet (c: cursorType);

procedure beginCursor (c: cursorType);

procedure endCursor;

procedure initMemoryMeter;

procedure advanceMemoryMeter;

procedure redrawMemoryMeter;

procedure advanceStopWatchHands;

procedure advanceStopWatchHandsNoMem;

function drawingRect (w: WindowPtr): Rect;

procedure clearWindow (w: WindowPtr);

procedure clearCurrentGeoBenchWindow;

procedure doNothingP;

function cap (x: char): char;

function capString (x: Str255): Str255;

function getDateAndTimeString: Str255;

procedure setAbortTime (newAbortTime: longInt);

function getAbortTime: longInt;

type
stringP = ^Str255;
stringH = ^stringP;

function GetConstResource (resourceNumber: integer): longInt;

procedure initGeoBenchUtility;

{	TransSkel version 2.02 - Transportable application skeleton											}
{																										}
{	TransSkel is public domain and was originally written in LightSpeed C by:							}
{																										}
{			Paul DuBois																					}
{			Wisconsin Regional Primate Research Center													}
{			1220 Capital Court																			}
{			Madison WI  53706  USA																		}
{																										}
{	UUCP:	[allegra,ihnp4,seismo]!uwvax!rhesus!dubois													}
{	ARPA:	dubois@rhesus.primate.wisc.edu																}
{	The Pascal Version of TransSkel is public domain and was ported and changed by						}
{																										}
{			Owen Hartnett																				}
{			hm Software																				}
{			163 Richard Drive																			}
{			Tiverton, RI 02878																			}
{																										}
{	CSNET:	omh@cs.brown.edu.CSNET 																		}
{	ARPA:		omh%cs.brown.edu																		}
{	UUCP:		[ihnp4,allegra]!brunix !omh																}
{																										}
{	This version of TransSkel written for Lightspeed Pascal.  Lightspeed Pascal is a					}
{	trademark of:																						}
{			THINK Technologies, Inc																		}
{			420 Bedford Street  Suite 350																}
{			Lexington, MA  02173  USA																	}
{																										}
{  History																								}
{  06/13/86	Beta version. (pd) 																			}
{  08/27/86	Version number changed to 1.01.(pd)															}
{  			v1.0 DoGrow bug fixed - the port at the point of the										}
{  			InvalRect could have been anything; the fix is to set										}
{  			the port to the grown window first.  This also explains										}
{  			why the kludge to DoActivate in v1.0 worked.(pd)											}
{  10/02/86	Version number changed to 1.02, as a result of adding										}
{			modifications by David W. Berry (well!dwb@lll-lcc.arpa)										}
{			for supporting window zooming.  Also used his modifications									}
{			for supporting modeless dialogs (though not in the same										}
{			form).  Dialogs can be #define'd on or off.(pd)												}
{																										}
{12/ 28 / 86 Version number changed to 1.03 . Modified to work under LightspeedC v . 2.01 				}
{			- took out definitions for window zooming stuff , as it is now supported by the compiler	}
{ 			directly . Also declared DoZoom static , fixing an oversight . ( pd )						}
{ 01 / 18 / 86 Put a SetPort into DoZoom - ZoomWindow requires port to be								}
{			set to window being zoomed . ( pd )															}
{ 02 / 05 / 86 Version number changed to 1.04 . Big change : port setting behavior made explicit		}
{			 - the only persistant switch occurs when a window comes active . This changes 				}
{			underlying programming model ( see manual for detailed discussion ) . Thanks to				}
{			 Duane Williams for pointing out that this should be done . Typedef 'd 						}
{			integer/long variables to Integer, Longint to facilitate coversion to other C 				}
{			compilers . More complete type-casting done . LightspeedC does a lot of it 					}
{			automatically , other compilers may not . ( pd - this version never released ) 				}
{03 / 02 / 87 Fixed bug whereby clicks in drag region of non - active windows may not bring 			}
{			window to front . Seems to be due to DragWindow calling StillDown to see  if mouse is		}
{			stilll down .    If the machine was busy    otherwise when click occurred and 				}
{			mouse already up when DragWindow is called , the click ends up being ignored . 				}
{			Thanks to Roger Humphrey for finding this one .												}
{																										}
{* * * Changes implemented first by omh to Pascal Version												}
{																										}
{ 12/24/86  Finished first Pascal version.   Dialogs cannot be defined off.	 (omh)						}
{																										}
{4 / 18 / 87 Changed Desk Accessory code so it 's more tolerant of memory								}
{			conditions for desk accessories . ( omh ) 													}
{7 / 12 / 87 Added "cache " code to GetWDHandler . Now TransSkel figures 								}
{			that an event is most likely to occur for the same window as the previous 					}
{			event . Thus the WindowPtr and WDHandle for events are cached and examined 					}
{			to avoid searching through the handler list . ( omh ) 										}
{7 / 12 / 87 Excised the notorious "SetPort "excess . As pointed out by Duane Williams ,				}
{			 SetPort traps abounded unnecessarily in version 1.02 . These have been eliminated 			}
{			now with two exceptions . First , the port is set when a window handler 					}
{			is installed . The justification for this is that when a handler is installed , it 			}
{			is likely that further processing will be done on it immediately . The application gets 	}
{			control immediately after the handler is installed anyway , so this behavior can be 		}
{			manually overridden where necessary . Second , when a window is activated , the port is		}
{			set to it . This follows the model of keeping the port in sync with the 					}
{			active window . ( omh ) 																	}
{7 / 14 / 87 Added grow zone function installation    and MoreMasters to SkelInit , 					}
{			which now requires two parameters . The first indicates the number of times to call 		}
{			MoreMasters . The second is a ProcPtr indicating a user - supplied grow zone				}
{			function to be called when memory problems occur .  If nil , no grow zone					}
{			function is    installed . ( omh ) 															}
{7 / 14 / 87 SkelMenu , SkelWindow , and SkelDialog now return zero or non - zero to indicate 			}
{			failure or success of handler allocation . This could break * all * previous TransSkel 		}
{			applications ( as will the change to SkelInit , above . Please see the section "How to 		}
{			adapt old TransSkel to New "    in the manual    for detailed specifications on 			}
{			how to convert your old programs . TransSkel becomes more memory conscious					}
{			with these changes . The functions SkelMenu , SkelWindow , and 								}
{			SkelDialog are the only routines which actually allocate memory . Since they may be 		}
{			called at any time , knowing that you have enough memory becomes important . Thus , 		}
{			these routines return a value to indicate what happened .  If they return zero ,			}
{			 then memory allocation failed . ( omh ) 													}
{10 / 21 / 87 Added another parameter to SkelMenu:  drawBar: Boolean . This tells SkelMenu 				}
{			whether to draw the menu bar after adding the Menu . This is done to eliminate 				}
{			the menus popping up one at a time . Simply call SkelMenu  with drawBar false				}
{			until the    last time you call SkelMenu , then call it (for the last menu )				}
{			with drawBar true . ( omh ) 																}
{10 / 26 / 87 Removed declarations for zoom - in and zoom - out . Added Pascal 							}
{			changes ( above ) to C version . ( omh )													}
{																										}
{    02 / 02 / 88 Merged pd 's 1.04 changes with those of omh, above, to create							}
{			release version 2.0 . Fixed bug whereby cmd - key equivalents								}
{			for menu selections would execute twice if DA window in front . Thanks 						}
{			to Don Fredkin and Julian Vrieslander for finding this one , and to Don for the				}
{			 best fix . ( pd ) 																			}
{ 10/28/88 Removed all New Rom calls. 																	}
{ 10/28/88 Added support for conditional compilation for dialogs and MPW support.  By setting 			}
{			the Think_pascal flag to false, TransSkel will run under MPW. Now correctly written			}
{			for LSP 2.0																					}
{  2/11/89	V. 2.01 Added fix to DoDialog, inserted statement DoDialog:=false at beginning of			}
{			procedure. This fixed bug that Dodialog might return without being correctly set. pd 		}
{  2/11/89	Added Paul DuBois fix for the call to DoEvent in SkelMain.  This should correct 			}
{			the non-blinking and allround bad stuff when DoEvent isn't called for null events and 		}
{			there's a dialogwind in front. Also, short circuit Booleans in SkelMain.					}
{  3/12/89	Finally added Paul Dubois code to segment in DoDialog.  Also added MultiFinder				}
{			aware code from Calvin Teague <CAL@STAR.STANFORD.EDU> 										}
{ 3/20/89	Fixed a bug in SkelRmveWind which only set oldWind to nil but still left a value in			}
{			oldWindHandler this caused problems when there was no window on screen but an event			}
{			for which a previous window had a handler for occurred.  Thanks to Norman Franke			}
{			nf0i+@andre.cmu.edu for "cacheing" this bug. 												}
{ 3/25/89	Folded a statement in DoDialog right into the if statement that really needed it and got	}
{			rid of a bug thereby.  Changed Boolean expressions in "if" statements to short-circuits		}
{			for better speed. 																			}
{																										}
{ _____________________________________________________________________________________________________ }

{ 2/8/91		Procedure passed as pSelect in SkelMenu has now two integer parameters: menuID and item.}
{	to			SkelMenu was extended by an option to determine whether a menu is a hierarchical submenu or	}
{ 7/31/91	a regular pull-down menu. SkelRmveMenu has now a drawBar boolean parameter like SkelMenu.	}
{				Resizing a window now triggers the call of procedure pGrow. The procedure pUpdate does *not*	}
{				take any parameters any more. Specifying one leads to undetermined crashes!						}
{																																	}
{				More information about those changes see in 'Changes to TransSkel', Part of 'Gelber Bericht' 	}

{$SETC supportDialogs:= true }
	{ Set to false to disallow modeless dialog support and save code space 		}


{$IFC NOT THINK_PASCAL }
type
mouseProcType = procedure ( thePoint : Point; theTime: longint; theMods: integer);
keyProcType = procedure ( theChar : char; theMods: integer);
eventProcType = procedure ( theitem : integer; ch: char; var theEvent: EventRecord);
eventFuncType = function (var theEvent: EventRecord): Boolean;
booleanProcType = procedure ( myBool : Boolean );
longintProcType = procedure ( myInt : longint );
int2LongintProcType = procedure ( myInt1 , myInt2 : integer; myLongInt: longInt);
menuProcType = procedure ( myMenu : MenuHandle );
noArgProcType = procedure ( );
{$ENDC }

procedure SkelInit (NoMasters: integer; myGrowZone: ProcPtr);
procedure SkelMain;
procedure SkelWhoa;
procedure SkelClobber;
{$IFC THINK_PASCAL }
function SkelMenu (theMenu: MenuHandle; pSelect: ProcPtr; pClobber: ProcPtr; drawBar, hierarchical: boolean): boolean;
procedure SkelApple (aboutTitle: Str255; aboutProc, deskAccProc, resumeProc: ProcPtr);
function SkelWindow (theWind: WindowPtr; pMouse, pKey, pUpdate, pGrow, pActivate, pClose, pClobber, pIdle: ProcPtr; frontOnly: boolean): boolean;
{$ELSEC }
function SkelMenu (theMenu: MenuHandle; pSelect: int2LongintProcType; pClobber: menuProcType; drawBar, hierarchical: boolean): boolean;
procedure SkelApple (aboutTitle: Str255; aboutProc: int2LongintProcType; deskAccProc: noArgProcType; resumeProc: booleanProcType);
function SkelWindow (theWind: WindowPtr; pMouse: mouseProcType; pKey: keyProcType; pUpdate, pGrow: noArgProcType; pActivate: booleanProcType; pClose: noArgProcType; pClobber, pIdle: noArgProcType; frontOnly: Boolean): Boolean;
{$ENDC }
procedure SkelRmveMenu (theMenu: MenuHandle; drawBar: boolean);
procedure SkelRmveWind (theWind: WindowPtr);
{$IFC supportDialogs }
{$IFC THINK_PASCAL }
function SkelDialog (theDialog: DialogPtr; pEvent, pClose, pClobber, pUpdate: ProcPtr): boolean;
{$ELSEC }
function SkelDialog (theDialog: DialogPtr; pEvent: eventProcType; pClobber, pClose, pUpdate: noArgProcType): boolean;
{$ENDC }
procedure SkelRmveDlog (theDialog: DialogPtr);
{$ENDC}
procedure SkelGrowBounds (theWind: WindowPtr; hLo, vLo, hHi, vHi: integer);
procedure SkelEventMask (mask: integer);
procedure SkelGetEventMask (var mask: integer);
{$IFC THINK_PASCAL }
procedure SkelBackground (p: ProcPtr);
procedure SkelGetBackground (var p: ProcPtr);
procedure SkelEventHook (p: ProcPtr);
procedure SkelGetEventHook (var p: ProcPtr);
{$ELSEC }
procedure SkelBackground (p: noArgProcType);
procedure SkelGetBackground (var p: noArgProcType);
procedure SkelEventHook (p: eventFuncType);
procedure SkelGetEventHook (var p: eventFuncType);
{$ENDC }
{$IFC supportDialogs }
procedure SkelDlogMask (mask: integer);
procedure SkelGetDlogMask (var mask: integer);
{$ENDC}

{ Added following two routines for Resume Events }
procedure SkelSetSleep (time: longInt);
procedure SkelGetSleep (var time: longInt);
procedure SkelSetMouseRgn (mouse: RgnHandle);
procedure SkelGetMouseRgn (var mouse: RgnHandle);

{ Added following two routines - omh }
{$IFC THINK_PASCAL }
procedure SkelSetMenuEnable (pMenuEnableProc: ProcPtr);
procedure SkelGetMenuEnable (var pMenuEnableProc: ProcPtr);
procedure SkelSetWindowZoom (theWind: WindowPtr; pZoomProc: ProcPtr);
procedure SkelGetWindowZoom (theWind: WindowPtr; var pZoomProc: ProcPtr);
{$ELSEC }
procedure SkelSetMenuEnable (pMenuEnableProc: booleanProcType);
procedure SkelGetMenuEnable (var pMenuEnableProc: booleanProcType);
procedure SkelSetWindowZoom (theWind: WindowPtr; pZoomProc: noArgProcType);
procedure SkelGetWindowZoom (theWind: WindowPtr; var pZoomProc: noArgProcType);
{$ENDC }

procedure DoEvent (theEvent: EventRecord);

implementation

{The following procedures are Pascal "glue" that allows Pascal to call a Procedure		}
{ from a ProcPtr.  It is similar to (*p) () construct used in the C dialect.  Different		}
{ procedures are necessary for the reason of Pascal's strongly typed parameter		}
{ list.  Fortunately, there are not too many calls which use different param lists		}

{$IFC THINK_PASCAL }
procedure callpMouse (thePoint: Point; theTime: longInt; theMods: integer; myProc: ProcPtr);
inline
	$205f, $4e90;

procedure callpKey (theChar: char; theMods: integer; myProc: ProcPtr);
inline
	$205f, $4e90;

procedure callpEvent (theItem: integer; theChar: char; var theEvent: EventRecord; myProc: ProcPtr);
inline
	$205f, $4e90;

function callOtherEvent (var theEvent: EventRecord; myProc: ProcPtr): boolean;
inline
	$205f, $4e90;

procedure callpBoolean (myBool: boolean; myProc: ProcPtr);
inline
	$205f, $4e90;

procedure callPLongInt (myLongInt: longInt; myProc: ProcPtr);
inline
	$205f, $4e90;

procedure callpInt2LongInt (myInt1, myInt2: integer; myLongInt: longInt; myProc: ProcPtr);
inline
	$205f, $4e90;

{ Handle removal of menus. 	}
procedure callpMenu (myMenu: MenuHandle; myProc: ProcPtr);
inline
	$205f, $4e90;

{ For all the Procedures that are called with no arguments							}
procedure callPNoArg (myProc: ProcPtr);
inline
	$205f, $4e90;
{$ENDC }

var
	minIgnore: longInt;
	WNEIsImplemented: boolean;
	goodDraggingRect: Rect;

procedure correctTimer (localStartTime: longInt);
	var
		delta: longInt;
	begin
		delta := TickCount - localStartTime;
		if delta > minIgnore then begin
			startTime := startTime + delta - 1;
			startTimeCorrections := startTimeCorrections + 1;
		end;
	end;

function longIntToString (x: longInt; fieldLength: integer): Str255;
	var
		result: Str255;
	begin
		result := StringOf(x : 1);
		while length(result) < fieldLength do
			result := Concat('0', result);
		longIntToString := result;
	end;

function getDateAndTimeString: Str255;
	var
		d: DateTimeRec;

	function twoDigits (x: integer): Str255;
		begin
			twoDigits := longIntToString(x, 2);
		end;

	begin (* getDateAndTimeString *)
		GetTime(d);
		with d do
			getDateAndTimeString := stringOf(twoDigits(month), '/', twoDigits(day), '/', year : 4, ' at ', twoDigits(hour), ':', twoDigits(minute), ':', twoDigits(second));
	end;

procedure clearCurrentGeoBenchWindow;
	begin
		clearWindow(currentGeoBenchWindow);
	end;

function drawingRect (w: WindowPtr): Rect;
	var
		r: Rect;
	begin
		r := w^.portRect;
		r.left := r.left + paletteWidth;
		r.right := r.right - scrollBarWidth;
		r.bottom := r.bottom - scrollBarWidth;
		drawingRect := r;
	end;

procedure clearWindow (w: WindowPtr);
	var
		tmp: GrafPtr;
	begin
		GetPort(tmp);
		SetPort(w);
		EraseRect(drawingRect(w));
		SetPort(tmp);
	end;

procedure doNothingP;
	begin
	end;

(* system error *)
	const
		fileProblemAlert = 1002;
		versionAlertResource = 1003;
		unimplementedMethodAlert = 1005;
		fatalErrorAlert = 1006;
		nonFatalErrorAlert = 1008;
		haltItem = 3;
		reservedSize = 2048;
		outOfMemoryAlert = 1011;

	type
		memBlock = array[1..reservedSize] of Byte;
		memBlockP = ^memBlock;
		memBlockH = ^memBlockP;

	var
		cursorX, cursorY: integer;
		oldTick: longInt;
		reservedMemory: memBlockH;

procedure testOSError (result, exception: OSErr);
	begin
		if (result <> noErr) & (result <> exception) then begin
			beginCursor(arrowCursor);
			ParamText(stringOf(result : 1), '', '', '');
			if StopAlert(fileProblemAlert, nil) = haltItem then
				halt;
			endCursor;
		end;
	end;

procedure fatalError (text: Str255);
	begin
		{$IFC THINK_PASCAL }
		DisposHandle(Handle(reservedMemory));
		{$ELSEC }
		DisposeHandle(Handle(reservedMemory));
		{$ENDC }
		beginCursor(arrowCursor);
		ParamText(text, '', '', '');
		if StopAlert(fatalErrorAlert, nil) = haltItem then
			halt;
		endCursor;
	end;

procedure failNil (h: Handle);
	var
		d: integer;
	begin
		if h = nil then begin
			{$IFC THINK_PASCAL }
			DisposHandle(Handle(reservedMemory));
			{$ELSEC }
			DisposeHandle(Handle(reservedMemory));
			{$ENDC }
			beginCursor(arrowCursor);
			d := StopAlert(outOfMemoryAlert, nil);
			halt;
		end;
	end;

function failNilF (h: Handle): Handle;
	begin
		failNil(h);
		failNilF := h;
	end;

function newHandleFailNil (size: longInt): Handle;
	begin
		newHandleFailNil := failNilF(NewHandle(size));
	end;

procedure DoCautionAlert (alertNumber: integer);
	var
		t: integer;
	begin
		beginCursor(arrowCursor);
		t := CautionAlert(alertNumber, nil);
		endCursor;
	end;

procedure nonFatalError (text: Str255);
	begin
		ParamText(text, '', '', '');
		DoCautionAlert(nonFatalErrorAlert);
	end;

procedure unimplementedMethod (name: Str255);
	begin
		beginCursor(arrowCursor);
		ParamText(name, '', '', '');
		if StopAlert(unimplementedMethodAlert, nil) = haltItem then
			halt;
		endCursor;
	end;

procedure wrongVersion;
	begin
		DoCautionAlert(versionAlertResource);
	end;

procedure assert (condition: boolean; t: Str255);
	begin
		if not condition then
			fatalError(concat('Assertion ''', t, ''' failed.'));
	end;


(* random numbers *)
	const
		necessaryClicks = 32;
		mNecessaryClicks = -necessaryClicks;
		randomScale = 2147483647;			(* 2^31-1 *)
		minusRandomScale = -randomScale;
	var
		randomSeed: extended;
		randomizerLongInt, lastTime: longInt;
		clickCount: integer;

function goodRandomizerAvailable: boolean;
	begin
		goodRandomizerAvailable := clickCount = 0;
	end;

procedure randomize;
	begin
		if randomizerLongInt > 0 then
			if randomizerLongInt < randomScale then
				randomSeed := randomizerLongInt
			else
				randomSeed := TickCount
		else if randomizerLongInt < 0 then
			if randomizerLongInt > minusRandomScale then
				randomSeed := -randomizerLongInt
			else
				randomSeed := TickCount
		else
			randomSeed := TickCount;
		clickCount := mNecessaryClicks;
	end;

function randomReal: extended;
	begin
		randomReal := randomx(randomSeed) / randomScale;
	end;

{$IFC NOT THINK_PASCAL }
function truncLong (x: extended): longInt;
	begin
		truncLong := roundtol(trunc(x));
	end;

function roundLong (x: extended): longInt;
	begin
		roundLong := roundtol(x);
	end;

procedure PaintOval2 (t, l, b, r: integer);
	var
		x: Rect;
	begin
		x.bottom := b;
		x.left := l;
		x.top := t;
		x.right := r;
		PaintOval(x);
	end;

procedure FrameOval2 (t, l, b, r: integer);
	var
		x: Rect;
	begin
		x.bottom := b;
		x.left := l;
		x.top := t;
		x.right := r;
		FrameOval(x);
	end;

procedure PaintRect2 (t, l, b, r: integer);
	var
		x: Rect;
	begin
		x.bottom := b;
		x.left := l;
		x.top := t;
		x.right := r;
		PaintRect(x);
	end;

procedure FrameRect2 (t, l, b, r: integer);
	var
		x: Rect;
	begin
		x.bottom := b;
		x.left := l;
		x.top := t;
		x.right := r;
		FrameRect(x);
	end;

procedure FrameArc2 (t, l, b, r, startAngle, arcAngle: integer);
	var
		x: Rect;
	begin
		x.bottom := b;
		x.left := l;
		x.top := t;
		x.right := r;
		FrameArc(x, startAngle, arcAngle);
	end;

procedure PaintArc2 (t, l, b, r, startAngle, arcAngle: integer);
	var
		x: Rect;
	begin
		x.bottom := b;
		x.left := l;
		x.top := t;
		x.right := r;
		PaintArc(x, startAngle, arcAngle);
	end;

{$PUSH }
{$PRAGMAC warn_unusedarg off }
procedure IOCheck (on: boolean);
	var
		x: extended;
	begin
		readString('2', x); (* set IOResult to 0 *)
		assert(ioResult = 0, StringOf('ioResult = 0 <> ', ioResult));
	end;
{$POP }
{$ENDC }

function randomLongInt: longInt;
	begin
		{$IFC THINK_PASCAL }
		randomLongInt := trunc(randomX(randomSeed));
		{$ELSEC }
		randomLongInt := truncLong(randomx(randomSeed));
		{$ENDC }
	end;

function randomRange (beginRange, endRange: longInt): longInt; (* beginRange <= randomRange < endRange *)
	begin
		{$IFC THINK_PASCAL }
		randomRange := trunc(randomReal * (endRange - beginRange)) + beginRange;
		{$ELSEC }
		randomRange := truncLong(randomReal * (endRange - beginRange)) + beginRange;
		{$ENDC }
	end;

function getRandomSeed: extended;
	begin
		getRandomSeed := randomSeed;
	end;

procedure setRandomSeed (seed: extended);
	begin
		randomSeed := seed;
	end;

procedure updateRandomizer (time: longInt);
	begin
		if time - lastTime > 10 then begin
			lastTime := time;
			clickCount := clickCount + 1;
			if clickCount >= 0 then
				clickCount := 0;
			randomizerLongInt := 2 * randomizerLongInt + time mod 2;
		end;
	end;

procedure initRandom;
	begin
		randomSeed := 19081961;
		randomizerLongInt := TickCount;
		clickCount := mNecessaryClicks;
		lastTime := 0;
	end;


(* dialog support *)
procedure GetDText (theDlg: DialogPtr; item: integer; var str: Str255);
	var
		itemHandle: Handle;
		itemType: integer;
		itemRect: Rect;
	begin
		{$IFC THINK_PASCAL }
		GetDItem(theDlg, item, itemType, itemHandle, itemRect);
		GetIText(itemHandle, str);
		{$ELSEC }
		GetDialogItem(theDlg, item, itemType, itemHandle, itemRect);
		GetDialogItemText(itemHandle, str);
		{$ENDC }
	end;

procedure SetDText (theDlg: DialogPtr; item: integer; str: Str255);
	var
		itemHandle: Handle;
		itemType: integer;
		itemRect: Rect;
	begin
		{$IFC THINK_PASCAL }
		GetDItem(theDlg, item, itemType, itemHandle, itemRect);
		SetIText(itemHandle, str);
		{$ELSEC }
		GetDialogItem(theDlg, item, itemType, itemHandle, itemRect);
		SetDialogItemText(itemHandle, str);
		{$ENDC }
	end;

function GetDCtl (theDlg: DialogPtr; item: integer): boolean;
	var
		itemHandle: Handle;
		itemType: integer;
		itemRect: Rect;
	begin
		{$IFC THINK_PASCAL }
		GetDItem(theDlg, item, itemType, itemHandle, itemRect);
		GetDCtl := GetCtlValue(ControlHandle(itemHandle)) = 1;
		{$ELSEC }
		GetDialogItem(theDlg, item, itemType, itemHandle, itemRect);
		GetDCtl := GetControlValue(ControlHandle(itemHandle)) = 1;
		{$ENDC }
	end;

procedure SetDCtl (theDlg: DialogPtr; item: integer; value: boolean);
	var
		itemHandle: Handle;
		itemType: integer;
		itemRect: Rect;
	begin
		{$IFC THINK_PASCAL }
		GetDItem(theDlg, item, itemType, itemHandle, itemRect);
		SetCtlValue(ControlHandle(itemHandle), integer(value));
		{$ELSEC }
		GetDialogItem(theDlg, item, itemType, itemHandle, itemRect);
		SetControlValue(ControlHandle(itemHandle), integer(value));
		{$ENDC }
	end;

procedure SetDProc (theDlg: DialogPtr; item: integer; p: ProcPtr);
	var
		itemHandle: Handle;
		itemType: integer;
		itemRect: Rect;
	begin
		{$IFC THINK_PASCAL }
		GetDItem(theDlg, item, itemType, itemHandle, itemRect);
		SetDITem(theDlg, item, itemType, Handle(p), itemRect);
		{$ELSEC }
		GetDialogItem(theDlg, item, itemType, itemHandle, itemRect);
		SetDialogItem(theDlg, item, itemType, Handle(NewUserItemProc(p)), itemRect);
		{$ENDC }
	end;

procedure setAndHighlightX (theDlg: DialogPtr; item: integer; x: extended; precision: integer);
	begin
		if trunc(x) = x then
{$IFC THINK_PASCAL }
			SetDText(theDlg, item, stringOf(trunc(x) : 1))
{$ELSEC }
			SetDText(theDlg, item, stringOf(truncLong(x) : 1))
{$ENDC }
		else
			SetDText(theDlg, item, stringOf(x : 1 : precision));
		{$IFC THINK_PASCAL }
		SelIText(theDlg, item, 0, 32767);
		{$ELSEC }
		SelectDialogItemText(theDlg, item, 0, 32767);
		{$ENDC }
	end;

procedure setAndHighlight (theDlg: DialogPtr; item: integer; x: extended);
	begin
		setAndHighlightX(theDlg, item, x, 8);
	end;

procedure ActivateItem (theDlg: DialogPtr; item: integer; active: boolean);
	var
		controlItem: ControlHandle;
		ignoreBox: Rect;
		ignoreType: integer;
	begin
		{$IFC THINK_PASCAL }
		GetDItem(theDlg, item, ignoreType, Handle(controlItem), ignoreBox);
		{$ELSEC }
		GetDialogItem(theDlg, item, ignoreType, Handle(controlItem), ignoreBox);
		{$ENDC }
		if active then
			HiliteControl(controlItem, 0)
		else
			HiliteControl(controlItem, 255);
	end;

procedure showDefaultButton (theDlg: DialogPtr; item: integer);
	var
		itemType, diameter: integer;
		itemHandle: Handle;
		itemRect: Rect;
		safe: GrafPtr;
	begin
		GetPort(safe);
		SetPort(theDlg);
		{$IFC THINK_PASCAL }
		GetDItem(theDlg, item, itemType, itemHandle, itemRect);
		{$ELSEC }
		GetDialogItem(theDlg, item, itemType, itemHandle, itemRect);
		{$ENDC }
		diameter := (itemRect.bottom - itemRect.top) div 2;
		if diameter < 16 then
			diameter := 16;
		PenSize(3, 3);
		InsetRect(itemRect, -4, -4);
		FrameRoundRect(itemRect, diameter, diameter);
		SetPort(safe);
	end;

(* animation *)
	const
		continueButtonDialogResource = 304;
		continueItem = 1;
		movieItem = 2;
		slowerItem = 3;
		fasterItem = 4;
		stopItem = 5;
		offItem = 6;
		abortItem = 7;
		saveFrameItem = 8;
		standardMovieWait = 60;
		waitChange = 1.33;
		shortFlash = 5;

	var
		singleStep, dialogVisible, dialogInstalled: boolean;

function buttonPressedInDialog: boolean;
	var
		theEvent: EventRecord;
		evtPort: GrafPtr;
	begin
		buttonPressedInDialog := EventAvail(mDownMask, theEvent) & (FindWindow(theEvent.where, evtPort) >= 0) & (evtPort = continueButton);
	end;

procedure wait (howLong: real);
	var
		t: longInt;
	begin
		{$IFC THINK_PASCAL }
		t := TickCount + round(howLong);
		{$ELSEC }
		t := TickCount + roundtol(howLong);
		{$ENDC }
		repeat
		until (TickCount >= t) | buttonPressedInDialog;
	end;

procedure removeDialog;
	begin
		if dialogInstalled then begin
			SkelRmveDlog(continueButton);
			dialogInstalled := false;
		end;
	end;

	var
		frameNumber: longInt;

procedure saveAnimationWindow;
	var
		save: GrafPtr;
	begin
		GetPort(save);
		SetPort(currentGeoBenchWindow);
		SaveDrawing(Concat('Frame', longIntToString(frameNumber, 4)));
		frameNumber := frameNumber + 1;
		SetPort(save);
	end;

procedure movieOff;
	begin
		ActivateItem(continueButton, continueItem, true);
		ActivateItem(continueButton, movieItem, true);
		ActivateItem(continueButton, fasterItem, false);
		ActivateItem(continueButton, slowerItem, false);
		ActivateItem(continueButton, saveFrameItem, true);
		ActivateItem(continueButton, stopItem, false);
		ActivateItem(continueButton, offItem, true);
		ActivateItem(continueButton, abortItem, true);
		singleStep := true;
		removeDialog;
	end;

	type
		booleanP = ^boolean;
	var
		animateP: booleanP;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoAnimationDialog (item: integer; ch: char; var theEvent: EventRecord);
	begin
		if ch = 'O' then
			HiliteIfAllowed(continueButton, offItem, shortFlash, item)
		else if ch = 'S' then
			HiliteIfAllowed(continueButton, slowerItem, shortFlash, item)
		else if ch = 'F' then
			HiliteIfAllowed(continueButton, fasterItem, shortFlash, item)
		else if ch = 'M' then
			HiliteIfAllowed(continueButton, stopItem, shortFlash, item)
		else if ch = 'P' then
			HiliteIfAllowed(continueButton, saveFrameItem, shortFlash, item);
		case item of
			offItem:  begin
				endAnimation;
				animateP^ := false;
			end;
			slowerItem:  begin
				movieWait := movieWait * waitChange;
				wait(movieWait);
			end;
			fasterItem:  begin
				movieWait := movieWait / waitChange;
				wait(movieWait);
			end;
			saveFrameItem: 
				saveAnimationWindow;
			stopItem: 
				movieOff;
			abortItem:  begin
				endAnimation;
				animateP^ := false;
				abortFlag := true;
			end;
			otherwise
		end;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure UpdateAnimation;
	begin
		DrawDialog(continueButton);
		showDefaultButton(continueButton, continueItem);
	end;

procedure installDialog;
	var
		dummy: boolean;
		save: GrafPtr;
	begin
		if not dialogInstalled then begin
			GetPort(save);
			dummy := SkelDialog(continueButton, @DoAnimationDialog, nil, nil, @UpdateAnimation);
			SetPort(save);
			dialogInstalled := true;
		end;
	end;

procedure movieOn;
	begin
		ActivateItem(continueButton, continueItem, false);
		ActivateItem(continueButton, movieItem, false);
		ActivateItem(continueButton, fasterItem, true);
		ActivateItem(continueButton, slowerItem, true);
		ActivateItem(continueButton, saveFrameItem, true);
		ActivateItem(continueButton, stopItem, true);
		ActivateItem(continueButton, offItem, true);
		ActivateItem(continueButton, abortItem, true);
		singleStep := false;
		installDialog;
	end;

procedure beginAnimation;
	begin
		dialogVisible := false;
		dialogInstalled := false;
	end;

procedure niceWindowPosition (w: WindowPtr; adjust: boolean);
	var
		dialogX, dialogY, t: integer;
	begin
		if (FrontWindow <> nil) & (FrontWindow <> infoWindow) then begin
			dialogX := WindowPeek(FrontWindow)^.strucRgn^^.rgnBBox.right;
			dialogY := WindowPeek(FrontWindow)^.contRgn^^.rgnBBox.top;
			if adjust then begin
				dialogX := dialogX + 8;
				dialogY := dialogY - 11;
			end;
			with w^.portRect do begin
				{$IFC THINK_PASCAL }
				t := screenBits.bounds.right + left - right;
				{$ELSEC }
				t := qd.screenBits.bounds.right + left - right;
				{$ENDC }
				if adjust then
					t := t - 8;
				if dialogX > t then
					dialogX := t;
				{$IFC THINK_PASCAL }
				t := screenBits.bounds.bottom + top - bottom;
				{$ELSEC }
				t := qd.screenBits.bounds.bottom + top - bottom;
				{$ENDC }
				if adjust then
					t := t - 8;
				if dialogY > t then
					dialogY := t;
			end;
			MoveWindow(w, dialogX, dialogY, false);
		end;
	end;

procedure HiliteIfAllowed (pDlog: DialogPtr; pItem: integer; howLong: longInt; var vItem: integer);
	var
		lHdl: Handle;
		lLong: longInt;
		lType: integer;
		lRect: Rect;
	begin
		{$IFC THINK_PASCAL }
		GetDItem(pDlog, pItem, lType, lHdl, lRect);
		{$ELSEC }
		GetDialogItem(pDlog, pItem, lType, lHdl, lRect);
		{$ENDC }
		if ControlHandle(lHdl)^^.contrlHilite = 0 then begin
			HiliteControl(ControlHandle(lHdl), 1);
			Delay(howLong, lLong);
			HiliteControl(ControlHandle(lHdl), 0);
			vItem := pItem;
		end;
	end;

function cap (x: char): char;
	const
		delta = ord('A') - ord('a');
	begin
		if ('a' <= x) & (x <= 'z') then
			cap := chr(ord(x) + delta)
		else
			cap := x;
	end;

function capString (x: Str255): Str255;
	var
		i: 0..255;
	begin
		for i := 1 to length(x) do
			x[i] := cap(x[i]);
		capString := x;
	end;

function dragDialogFilterGeneral (theDialog: DialogPtr; var theEvent: EventRecord; var itemHit: integer; animation: boolean): boolean;
	var
		evtPort, save: GrafPtr;
		chInt: integer;
		ch: char;
	begin
		GetPort(save);
		dragDialogFilterGeneral := false;
		with theEvent do
			case what of
				keyDown:  begin
					chInt := BitAnd(message, charCodeMask);
					if (chInt = 3) | (chInt = 13) then begin
						HiliteIfAllowed(theDialog, continueItem, shortFlash, itemHit);
						dragDialogFilterGeneral := true;
					end
					else if animation & (BitAnd(modifiers, cmdKey) > 0) then begin
						ch := cap(chr(chInt));
						if ch = '.' then begin
							HiliteIfAllowed(theDialog, abortItem, shortFlash, itemHit);
							dragDialogFilterGeneral := true;
						end
						else if ch = 'C' then begin
							HiliteIfAllowed(theDialog, continueItem, shortFlash, itemHit);
							dragDialogFilterGeneral := true;
						end
						else if ch = 'M' then begin
							HiliteIfAllowed(theDialog, movieItem, shortFlash, itemHit);
							dragDialogFilterGeneral := true;
						end
						else if ch = 'O' then begin
							HiliteIfAllowed(theDialog, offItem, shortFlash, itemHit);
							dragDialogFilterGeneral := true;
						end
						else if ch = 'P' then begin
							HiliteIfAllowed(theDialog, saveFrameItem, shortFlash, itemHit);
							dragDialogFilterGeneral := true;
						end;
					end;
				end;
				mouseDown:  begin
					updateRandomizer(theEvent.when);
					if (FindWindow(where, evtPort) = inDrag) & (evtPort = theDialog) then begin
						DragWindow(evtPort, where, goodDraggingRect);
						itemHit := 0;
						dragDialogFilterGeneral := true;
					end
					else if (evtPort <> theDialog) then begin
						DoEvent(theEvent);
						itemHit := 0;
						dragDialogFilterGeneral := true;
					end;
				end;
				otherwise
					if not IsDialogEvent(theEvent) then begin
						DoEvent(theEvent);
						itemHit := 0;
						dragDialogFilterGeneral := true;
					end;
			end;
		if FrontWindow <> theDialog then begin
			SelectWindow(theDialog);
			showDefaultButton(theDialog, continueItem);
		end;
		SetPort(save);
	end;

function dragDialogFilter (theDialog: DialogPtr; var theEvent: EventRecord; var itemHit: integer): boolean;
	begin
		dragDialogFilter := dragDialogFilterGeneral(theDialog, theEvent, itemHit, false);
	end;

function dragDialogFilterAnimation (theDialog: DialogPtr; var theEvent: EventRecord; var itemHit: integer): boolean;
	begin
		dragDialogFilterAnimation := dragDialogFilterGeneral(theDialog, theEvent, itemHit, true);
	end;

procedure waitForClick (var animate: boolean);
	var
		item: integer;
		theEvent: EventRecord;
		localStartTime: longInt;
	begin
		if not dialogVisible then begin
			continueButton := GetNewDialog(continueButtonDialogResource, nil, WindowPtr(-1));
			niceWindowPosition(WindowPtr(continueButton), false);
			SelectWindow(continueButton);
			ShowWindow(continueButton);
			showDefaultButton(continueButton, continueItem);
			movieOff;
			dialogVisible := true;
		end;
		if singleStep then begin
			localStartTime := TickCount;
			if not WindowPeek(currentGeoBenchWindow)^.visible then
				ShowWindow(currentGeoBenchWindow);
			repeat
				beginCursor(arrowCursor);
{$IFC THINK_PASCAL }
				ModalDialog(@dragDialogFilterAnimation, item);
{$ELSEC }
				ModalDialog(NewModalFilterProc(@dragDialogFilterAnimation), item);
{$ENDC }
				endCursor;
				if item = 0 then
					cycle;
				case item of
					offItem:  begin
						endAnimation;
						animate := false;
					end;
					movieItem:  begin
						movieOn;
						animateP := @animate;
					end;
					slowerItem:  begin
						movieWait := movieWait * waitChange;
						wait(movieWait);
					end;
					fasterItem:  begin
						movieWait := movieWait / waitChange;
						wait(movieWait);
					end;
					saveFrameItem: 
						saveAnimationWindow;
					stopItem: 
						movieOff;
					abortItem:  begin
						endAnimation;
						animate := false;
						abortFlag := true;
					end;
					otherwise
				end;
			until not EventAvail(mDownMask, theEvent);
			correctTimer(localStartTime);
		end
		else if userAbort then begin
			endAnimation;
			animate := false;
			abortFlag := true;
		end
		else begin
			localStartTime := TickCount;
			wait(movieWait);
			correctTimer(localStartTime);
		end;
	end;

procedure endAnimation;
	begin
		if dialogVisible then begin
			HideWindow(continueButton);
			removeDialog;
			{$IFC THINK_PASCAL }
			DisposDialog(continueButton);
			{$ELSEC }
			DisposeDialog(continueButton);
			{$ENDC }
			continueButton := nil;
			dialogVisible := false;
		end;
	end;

function animationEnabled: boolean;
	var
		i: integer;
	begin
		i := minAnimation;
		while (i <= maxAnimation) & not animationFlag[i] do
			i := i + 1;
		animationEnabled := i <= maxAnimation;
	end;


(* cursor handling *)
	const
		maxCursorStack = 32;
		stopWatchesID = 128;
		stopWatchTime = 10;
		memXOrg = 15;
		memYOrg = 85;
		memWidth = 370;
		memHeight = 100;
		memLeft = memXOrg;
		memRight = memXOrg + memWidth - 1;

	type
		cursorStackIndex = 1..maxCursorStack;

	var
		cursorSP: cursorStackIndex;
		cursorStack: array[cursorStackIndex] of cursorType;
		stopWatchHands: 0..7;
		oldUserAbortTick, thisTick: longInt;
		memXPos: integer;
		memoryMeter: array[memLeft..memRight] of Byte;

procedure drawMeterFrame;
	const
		numberOfGridLines = 4;
	var
		r: Rect;
		i, y: integer;
	begin
		SetRect(r, memXOrg, memYOrg, memXOrg + memWidth, memYOrg + memHeight);
		FrameRect(r);
		for i := 1 to numberOfGridLines do begin
			{$IFC THINK_PASCAL }
			y := memYOrg + round(i * memHeight / (numberOfGridLines + 1));
			{$ELSEC }
			y := memYOrg + roundtol(i * memHeight / (numberOfGridLines + 1));
			{$ENDC }
			MoveTo(memXOrg, y);
			LineTo(memXOrg + memWidth - 1, y);
		end;
		MoveTo(0, memYOrg);
		{$IFC THINK_PASCAL }
		writeDraw('0%');
		{$ELSEC }
		DrawString('0%');
		{$ENDC }
		MoveTo(0, memYOrg + memHeight);
		{$IFC THINK_PASCAL }
		writeDraw('100%');
		{$ELSEC }
		DrawString('100%');
		{$ENDC }
	end;

procedure redrawMemoryMeter;
	var
		i: integer;
	begin
		drawMeterFrame;
		for i := memLeft to memXPos - 1 do begin
			MoveTo(i, memYOrg);
			LineTo(i, memoryMeter[i]);
		end;
	end;

procedure initMemoryMeter;
	var
		save: GrafPtr;
		r: Rect;
	begin
		GetPort(save);
		SetPort(infoWindow);
		SetRect(r, memXOrg, memYOrg, memXOrg + memWidth, memYOrg + memHeight);
		EraseRect(r);
		drawMeterFrame;
		memXPos := memXOrg;
		SetPort(save);
	end;

procedure advanceMemoryMeter;
	var
		save: GrafPtr;
		pos: Byte;
	begin
		if memXPos >= memXOrg + memWidth then
			initMemoryMeter;
		GetPort(save);
		SetPort(infoWindow);
		MoveTo(memXPos, memYOrg);
		{$IFC THINK_PASCAL }
		pos := memYOrg + trunc(memHeight * ((startMemory - FreeMem) / startMemory));
		{$ELSEC }
		pos := memYOrg + truncLong(memHeight * ((startMemory - FreeMem) / startMemory));
		{$ENDC }
		memoryMeter[memXPos] := pos;
		LineTo(memXPos, pos);
		SetPort(save);
		memXPos := memXPos + 1;
	end;

procedure advanceStopWatchHands;
	begin
		if cursorStack[cursorSP] = stopwatchCursor then begin
			thisTick := TickCount;
			if thisTick - oldTick > stopWatchTime then begin
				if stopWatchHands = 7 then
					stopWatchHands := 0
				else
					stopWatchHands := stopWatchHands + 1;
				SetCursor(GetCursor(stopWatchesID + stopWatchHands)^^);
				advanceMemoryMeter;
				oldTick := thisTick;
			end;
		end;
	end;

procedure advanceStopWatchHandsNoMem;
	begin
		if cursorStack[cursorSP] = stopwatchCursor then begin
			thisTick := TickCount;
			if thisTick - oldTick > stopWatchTime then begin
				if stopWatchHands = 7 then
					stopWatchHands := 0
				else
					stopWatchHands := stopWatchHands + 1;
				SetCursor(GetCursor(stopWatchesID + stopWatchHands)^^);
				oldTick := thisTick;
			end;
		end;
	end;

procedure cursorSet (c: cursorType);
	begin
		case c of
			arrowCursor: 
				{$IFC THINK_PASCAL }
				SetCursor(arrow);
				{$ELSEC }
			SetCursor(qd.arrow);
				{$ENDC }
			stopwatchCursor:  begin
				SetCursor(GetCursor(stopWatchesID + stopWatchHands)^^);
				oldTick := TickCount;
			end;
			littleCrossCursor: 
				SetCursor(GetCursor(crossCursor)^^);
		end;
	end;

procedure beginCursor (c: cursorType);
	begin
		assert(cursorSP < maxCursorStack, 'cursor SP < max. cursor SP');
		cursorSP := cursorSP + 1;
		cursorStack[cursorSP] := c;
		if cursorStack[cursorSP - 1] <> c then
			cursorSet(c);
	end;

procedure endCursor;
	begin
		assert(cursorSP > 1, 'cursor SP > 1');
		cursorSP := cursorSP - 1;
		if cursorStack[cursorSP] <> cursorStack[cursorSP + 1] then
			cursorSet(cursorStack[cursorSP]);
	end;

function GetConstResource (resourceNumber: integer): longInt;
	var
		h: stringH;
		x: longInt;

	procedure simpleAssert (condition: boolean);
		begin
			assert(condition, StringOf('STR ', resourceNumber : 1, ' is defined.'));
			if not condition then
				x := 0;
		end;

	begin
		h := stringH(GetResource('STR ', resourceNumber));
		simpleAssert(h <> nil);
		IOCheck(false);
		readString(h^^, x);
		simpleAssert(ioResult = 0);
		IOCheck(true);
		GetConstResource := x;
	end;

function haveEvent (var theEvent: EventRecord): boolean;
	begin
		if WNEIsImplemented then
			haveEvent := WaitNextEvent(everyEvent, theEvent, 0, nil)
		else
			haveEvent := GetNextEvent(everyEvent, theEvent);
	end;

function AbortInQueueUserClick (var userClicked: boolean; var whichPort: GrafPtr; var whichPoint: Point): Boolean;
	var
		theEvent: EventRecord;
		evtPart: integer;
	begin
		AbortInQueueUserClick := false;
		userClicked := false;
		while haveEvent(theEvent) do
			case theEvent.what of
				keyDown: 
					if BitAnd(theEvent.modifiers, cmdKey) > 0 then begin
						if BitAnd(theEvent.message, charCodeMask) = ord('.') then begin
							FlushEvents(keyDownMask + keyUpMask, 0); (* Flush all all keydown- and keyup events from the event queue. *)
							AbortInQueueUserClick := True;
						end
						else if (continueButton <> nil) & (FrontWindow = continueButton) then
							DoEvent(theEvent);
					end
					else
						SysBeep(1);
				mouseDown:  begin
					updateRandomizer(theEvent.when);
					evtPart := FindWindow(theEvent.where, whichPort);
					if (evtPart = inDrag) | (evtPart = inSysWindow) | (evtPart = inMenuBar) | (continueButton <> nil) & (whichPort = continueButton) then begin
						beginCursor(arrowCursor);
						DoEvent(theEvent);
						endCursor;
					end
					else if evtPart = inContent then begin
						userClicked := true;
						whichPoint := theEvent.where;
						GlobalToLocal(whichPoint);
					end;
				end;
				otherwise
					DoEvent(theEvent);
			end;
	end;

function AbortInQueue: boolean;
	var
		userClicked: boolean;
		whichPort: GrafPtr;
		whichPoint: Point;
	begin
		AbortInQueue := AbortInQueueUserClick(userClicked, whichPort, whichPoint);
	end;

	var
		abortTime, longComputation: longInt;

procedure setAbortTime (newAbortTime: longInt);
	begin
		abortTime := newAbortTime;
	end;

function getAbortTime: longInt;
	begin
		getAbortTime := abortTime;
	end;

function askUserBoolean (question: Str255): boolean;
	begin
		ParamText(question, '', '', '');
		askUserBoolean := Alert(1022, nil) = 4;
	end;

function userAgreesToAbort: boolean;
	begin
		beginCursor(arrowCursor);
		userAgreesToAbort := (TickCount - startTime <= longComputation) | askUserBoolean('Do you really want to abort?');
		endCursor;
	end;

function userAbort: boolean;
	var
		newAbortFlag: boolean;
		localThisTick, saveStartTime: longInt;
		save: GrafPtr;
	begin
		advanceStopWatchHands;
		if abortEnable then
			if abortFlag then
				userAbort := true
			else begin
				thisTick := TickCount;
				if thisTick - oldUserAbortTick > abortTime then begin
					oldUserAbortTick := thisTick;
					localThisTick := thisTick;
					saveStartTime := startTime;
					GetPort(save);
					newAbortFlag := AbortInQueue & userAgreesToAbort;
					SetPort(save);
					startTime := saveStartTime;
					abortFlag := abortFlag | newAbortFlag;
					userAbort := abortFlag;
					correctTimer(localThisTick);
				end
				else
					userAbort := false;
			end
		else
			userAbort := false;
	end;

procedure setAllAnimationFlag (value: boolean);
	var
		i: minAnimation..maxAnimation;
	begin
		for i := minAnimation to maxAnimation do
			animationFlag[i] := value;
	end;

procedure initAnimation;

	procedure menuEntry (item: integer; text: Str255);
		begin
			{$IFC THINK_PASCAL }
			SetItem(animationMenu, item, text);
			{$ELSEC }
			SetMenuItemText(animationMenu, item, text);
			{$ENDC }
			EnableItem(animationMenu, item);
		end;

	begin
		setAllAnimationFlag(false);
(* begin Marco Ciot *)
		menuEntry(ARelativeNeighborhoodItem, 'Relative Neighborhood');
(* end Marco Ciot *)
{$IFC Tutorial }
		menuEntry(ACenterOfGravityItem, 'Center Of Gravity');
{$ENDC }
		menuEntry(AClipWithRectangleItem, 'Clip With Rectangle');
		menuEntry(AClosestPairItem, 'Closest Pair');
		menuEntry(AContourOfRectanglesItem, 'Contour Of Rectangles');
		menuEntry(AConvexHullItem, 'Convex Hull');
		menuEntry(AConvexPolygonDistanceItem, 'Convex Polygon Distance');
		menuEntry(ADiameterItem, 'Diameter');
{$IFC Chemistry }
		menuEntry(ADistanceGeometryItem, 'Distance Geometry');
{$ENDC }
		menuEntry(ALineSegmentIntersectionItem, 'Line Segment Intersection');
		menuEntry(ALowerEnvelopeItem, 'Lower Envelope');
		menuEntry(AMinimalAreaDiskItem, 'Minimal Area Disk');
		menuEntry(AMinimumSpanningTreeItem, 'Minimum Spanning Tree');
		menuEntry(ANearestNeighborsItem, 'Nearest Neighbors');
		menuEntry(APolygonIntersectionItem, 'Polygon Intersection');
		menuEntry(ASearchItem, 'Search In Structure');
		menuEntry(AShortestPathItem, 'Shortest Path');
		menuEntry(ATangentItem, 'Tangent');
		menuEntry(ATravelingSalesmanItem, 'Traveling Salesman');
		menuEntry(ATriangulationItem, 'Triangulation');
		menuEntry(AVoronoiDiagramItem, 'Voronoi Diagram');
		menuEntry(AWindingNumberItem, 'Winding Number');
	end;

procedure initGeoBenchUtility;
	var
		error: OSErr;
		theWorld: SysEnvRec;
	begin
		reservedMemory := memBlockH(NewHandle(reservedSize));
		error := SysEnvirons(1, theWorld);
		isSystemSeven := (error = noErr) & (theWorld.systemVersion >= 7 * 256);
		cursorSP := 1;
		cursorStack[1] := arrowCursor;
		InitCursor;
		cursorSet(arrowCursor);
		stopWatchHands := 0;
		initRandom;
		oldTick := 0;
		oldUserAbortTick := 0;
		abortFlag := false;
		abortEnable := true;
		memXPos := memXOrg;
		movieWait := standardMovieWait;
		currentVersion := 0;
		lpMetric := 2;
		continueButton := nil;
		minIgnore := GetConstResource(202);
		abortTime := GetConstResource(211);
		longComputation := GetConstResource(212);
		assert(minIgnore >= 0, 'minIgnore  0');
		frameNumber := 0;
		currentGeoBenchWindow := nil;
	end;


{	Window and Menu handler types, constants, variables.						}

{	whList and mhList are the lists of window and menu handlers.				}
{	whClobOnRmve and mhClobOnRmve are true if the handler disposal proc			}
{	is to be called when a handler is removed.  They are temporarily set		}
{	false when handlers are installed for windows or menus that already			}
{	have handlers - the old handler is removed WITHOUT calling the				}
{	disposal proc.																}

{	Default lower limits on window sizing of 80 pixels both directions is		}
{	sufficient to allow text windows room to draw a grow box and scroll			}
{	bars without having the thumb and arrows overlap.  These values may			}
{	be changed if such a constraint is undesirable with SkelGrowBounds.			}
{	Default upper limits are for the Macintosh, not the Lisa, but are set		}
{	per machine in SkelInit.													}

	const
		mBarHeight = 20;	{ menu bar height.  All window sizing}

	type
		WHandlerPtr = ^WHandler;
		WHandlerHnd = ^WHandlerPtr;
		{$IFC THINK_PASCAL }
		WHandler = record
				whWind: WindowPtr;		{ window/dialog to be handled		}
				whClobber: ProcPtr;		{ data structure disposal proc		}
				whMouse: ProcPtr;		{ mouse-click handler proc			}
				whKey: ProcPtr;			{ key-click handler proc			}
				whUpdate: ProcPtr;		{ update handler proc				}
				whGrowP: ProcPtr;
				whActivate: ProcPtr;	{ activate event handler proc		}
				whClose: ProcPtr;		{ close "event" handler proc		}
				whZoom: ProcPtr;
				whIdle: ProcPtr;		{ main loop proc					}
{$IFC supportDialogs }
				whEvent: ProcPtr;		{ dialog event proc					}
{$ENDC }
				whGrow: Rect;			{ limits on window sizing			}
				whFrontOnly: boolean;	{ true = idle only when active		}
				whNext: WHandlerHnd;	{ next window handler				}
			end;
		{$ELSEC }
		WHandler = record
				whWind: WindowPtr;			{ window/dialog to be handled		}
				whClobber: noArgProcType;	{ data structure disposal proc		}
				whMouse: mouseProcType;		{ mouse-click handler proc			}
				whKey: keyProcType;			{ key-click handler proc			}
				whUpdate: noArgProcType;	{ update handler proc				}
				whGrowP: noArgProcType;
				whActivate: booleanProcType;{ activate event handler proc		}
				whClose: noArgProcType;		{ close "event" handler proc		}
				whZoom: noArgProcType;
				whIdle: noArgProcType;		{ main loop proc					}
{$IFC supportDialogs }
				whEvent: eventProcType;		{ dialog event proc					}
{$ENDC }
				whGrow: Rect;			{ limits on window sizing			}
				whFrontOnly: boolean;	{ true = idle only when active		}
				whNext: WHandlerHnd;	{ next window handler				}
			end;
		{$ENDC }

		MHandlerPtr = ^MHandler;
		MHandlerHnd = ^MHandlerPtr;

{$IFC THINK_PASCAL }
		MHandler = record
				mhID: integer;				{ menu id							}
				mhSelect: ProcPtr;			{ item selection handler proc		}
				mhClobber: ProcPtr;			{ menu disposal handler proc		}
				mhNext: MHandlerHnd;		{ next menu handler					}
{$ELSEC }
				MHandler = record
						mhID: integer;					{ menu id							}
						mhSelect: int2LongintProcType;	{ item selection handler proc		}
						mhClobber: menuProcType;		{ menu disposal handler proc		}
						mhNext: MHandlerHnd;			{ next menu handler					}
{$ENDC }
					end;

				var
					whList: WHandlerHnd;		{ list of menu handlers 			}
					whClobOnRmve: boolean;
					growRect: Rect;
					mhList: MHandlerHnd;
					mhClobOnRmve: boolean;

{	Variables for default Apple menu handler.  appleAboutProc			}
{	is the procedure to execute if there is an About... item and it's	}
{	chosen from the Apple menu.  If doAboutItems>0 , then the menu		}
{	contains the About... item, otherwise it's just desk accessories.	}

					doAboutItems: integer;
{$IFC THINK_PASCAL }
					appleAboutProc, deskAccOpenProc, mfResumeProc: ProcPtr;
{$ELSEC }
					appleAboutProc: int2LongintProcType;
					deskAccOpenProc: noArgProcType;
					mfResumeProc: booleanProcType;
{$ENDC }


{	Miscellaneous														}

{	screenPort points to the window manager port.						}

{	doneFlag determines when SkelMain returns.  It is set by calling	}
{	SkelWhoa(), which the host does to request a halt.					}

{	pBkgnd points to a background procedure, to be run during event		}
{	processing.  Set it with SkelBackground.  If nil, there's no		}
{	procedure.															}

{	pEvent points to an event-inspecting hook, to be run whenever an	}
{	event occurs.  Set it with SkelEventHook.  If nil, there's no		}
{	procedure.															}

{	eventMask controls the event types requested in the GetNextEvent	}
{	call in SkelMain.													}

{	diskInitPt is the location at which the disk initialization dialog	}
{	appears, if an uninitialized disk is inserted.						}

					screenPort: GrafPtr;
					doneFlag: integer;
{$IFC THINK_PASCAL }
					pBkgnd: ProcPtr;
					pEvent: ProcPtr;
{$ELSEC }
					pBkgnd: noArgProcType;
					pEvent: eventFuncType;
{$ENDC }
					eventMask: integer;
					diskInitPt: Point;

{$IFC supportDialogs }

{	Events that are passed to dialogs.  Others are ignored.					}
{	Standard mask passes , mousedown, keydown, autokey, update,		}
{	activate and null events.  Null events are controlled by bit 0.			}

					dlogEventMask: integer;
{$ENDC}
					pEventflag: boolean;

{	"caching" global variables.  previous version would search down the window }
{	list for every event it found.  Now, if the event happened to the same window }
{	as last time, GetWDHandler will just do a simple compare }
{	and return the last window handler.  This speeds up multiple window applications }
{	immensely, at only a slight cost when you activate a new window (one }
{	compare!)  If you don't like it, use the old version. }

					oldWindow: WindowPtr;
					oldWDHandler: WHandlerHnd;

{$IFC THINK_PASCAL }
					pMenuEnable: ProcPtr;							{  ProcPtr to Menu Enabler routine }
{$ELSEC }
					pMenuEnable: booleanProcType;					{  Menu Enabler routine }
{$ENDC }
{	Global for built in "Grow Zone" function  }


					sleep: integer;
					mouseRgn: RgnHandle;


{ --------------------------------------------------------------------	}
{						Internal (private) Routines										}
{ --------------------------------------------------------------------	}



{	Get handler associated with user or dialog window.								}
{	Return nil if window doesn't belong to any known handler.						}
{	This routine is absolutely fundamental to TransSkel.								}
function GetWDHandler (theWind: WindowPtr): WHandlerHnd;
	var
		h: WHandlerHnd;
	begin
		h := whList;
		GetWDHandler := nil;
		if theWind = oldWindow then			{  caching code 	}
			GetWDHandler := oldWDHandler
		else
			while h <> nil do
				if h^^.whWind = theWind then begin
					oldWindow := theWind;			{ Load in new values for new window }
					oldWDHandler := h;
					GetWDHandler := h;
					h := nil;
				end
				else
					h := WHandlerHnd(h^^.whNext);
	end;

{ Get Handler associated with user window.  Return nil if window doesn't		}
{  have a Handler.																		}
function GetWHandler (theWind: WindowPtr): WHandlerHnd;
	var
		h: WHandlerHnd;
	begin
		h := GetWDHandler(theWind);
		GetWHandler := nil;
		if (h <> nil) & (WindowPeek(theWind)^.windowKind <> dialogKind) then
			GetWHandler := h;
	end;

{$IFC supportDialogs }

{	Get handler associated with dialog window.						}
{	Return nil if window doesn't belong to any known handler.	}
function GetDHandler (theDialog: WindowPtr): WHandlerHnd;
	var
		h: WHandlerHnd;
	begin
		h := GetWDHandler(theDialog);
		GetDHandler := nil;
		if (h <> nil) & (WindowPeek(theDialog)^.windowKind = dialogKind) then
			GetDHandler := h;
	end;
{$ENDC}

{	General menu-handler.  Just passes selection to the handler's					}
{	select routine.  If the select routine is nil, selecting items from				}
{	the menu is a nop.																}
procedure DoMenuCommand (command, when: longInt);
	var
		menu: integer;
		mh: MHandlerHnd;
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: int2LongintProcType;
{$ENDC }
	begin
		mh := mhList;
		menu := HiWord(command);
		while mh <> nil do begin
			p := mh^^.mhSelect;
			if (menu = mh^^.mhID) & (p <> nil) then begin
				{$IFC THINK_PASCAL }
				callpInt2LongInt(menu, LoWord(command), when, p);
				{$ELSEC }
				p(menu, LoWord(command), when);
				{$ENDC }
				mh := nil;
			end
			else
				mh := mh^^.mhNext;
		end;
		HiliteMenu(0);
	end;

{	Apple menu handler																	}

{	DoAppleItem:  If the first item was chosen, and there's an "About..."		}
{	item, call the procedure associated with it (if not nil).  If there				}
{	is no "About..." item or the item was not the first one, then open			}
{	the associated desk accessory.  The port is saved and restored				}
{	because OpenDeskAcc does not always preserve it correctly.				}
procedure DoAppleItem (menu, item: integer; when: longInt);
	var
		curPort: GrafPtr;
		str: Str255;
		ignore: integer;
	begin
		if (doAboutItems > 0) & (item <= doAboutItems) then begin
			if appleAboutProc <> nil then
{$IFC THINK_PASCAL }
				callPInt2LongInt(menu, item, when, appleAboutProc);
{$ELSEC }
			appleAboutProc(menu, item, when);
{$ENDC }
		end
		else begin
			if deskAccOpenProc <> nil then
{$IFC THINK_PASCAL }
				callPNoArg(deskAccOpenProc);
{$ELSEC }
			deskAccOpenProc;
{$ENDC }
			GetPort(curPort);
{$IFC THINK_PASCAL }
			GetItem(appleMenu, item, str);
{$ELSEC }
			GetMenuItemText(appleMenu, item, str);
{$ENDC }
			ignore := OpenDeskAcc(str);
			SetPort(curPort);
		end;
	end;

{	DoAppleClobber disposes of the Apple menu.}
{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoAppleClobber (m: MenuHandle);
	begin
		DisposeMenu(appleMenu);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

{ -------------------------------------------------------------------- 	}
{						Window-handler routing routines									}
{																								}
{	Each routine sets the port to the handler's window before executing			}
{	the handler procedure.																	}
{ -------------------------------------------------------------------- 	}


{	Pass local mouse coordinates, click time, and the modifiers flag				}
{	word to the handler.  Should not be necessary to set the port, as				}
{   	the click is passed to the active window's hander.									}
procedure DoMouse (h: WHandlerHnd; theEvent: EventRecord);
	var
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: mouseProcType;
{$ENDC }
		thePt: Point;
	begin
		if h <> nil then begin
			p := h^^.whMouse;
			if p <> nil then begin
				thePt := theEvent.where;
				GlobalToLocal(thePt);
{$IFC THINK_PASCAL }
				callpMouse(thePt, theEvent.when, theEvent.modifiers, p);
{$ELSEC }
				p(thePt, theEvent.when, theEvent.modifiers);
{$ENDC }
			end;
		end;
	end;

{	Pass the character and the modifiers flag word to the handler.}
{	Should not be necessary to set the port, as the  click is passed to the}
{	active window's handler. }
procedure DoKey (h: WHandlerHnd; ch: char; mods: integer);
	var
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: keyProcType;
{$ENDC }
	begin
		if h <> nil then begin
			p := h^^.whKey;
			if p <> nil then
{$IFC THINK_PASCAL }
				callpKey(ch, mods, p);
{$ELSEC }
			p(ch, mods);
{$ENDC }
		end;
	end;

{	Call the window updating procedure, passing to it an indicator whether the		}
{	window has been resized or not.  Then clear the flag, assuming the update		}
{	proc took whatever action was necessary to respond to resizing.					}
{}
{	If the handler doesn't have any update proc, the Begin/EndUpdate stuff			}
{	is still done, to clear the update region.  Otherwise the Window Manager			}
{	will keep generating update events for the window, stalling updates of			}
{	other windows.																				}

{	Make sure to save and restore the port, as it's not always the active window	}
{	that's updated.																				}
procedure DoUpdate (rh: WHandlerHnd);
	var
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: noArgProcType;
{$ENDC }
		updPort, tmpPort: GrafPtr;
	begin
		if rh <> nil then begin
			GetPort(tmpPort);
			updPort := rh^^.whWind;
			SetPort(updPort);
			BeginUpdate(updPort);
			p := rh^^.whUpdate;
			if p <> nil then
{$IFC THINK_PASCAL }
				callPNoArg(p);
{$ELSEC }
			p;
{$ENDC }
			EndUpdate(updPort);
			SetPort(tmpPort);
		end;
	end;

{	Pass activate/deactivate notification to handler.  On activate, set the port to			}
{	the window coming active.  Normally this is done by the user clicking in a window.		}
{	*** BUT ***																						}
{	Under certain conditions, a deactivate may be generated for a window					}
{	that has not had the port set to it by a preceding activate.  If an							}
{	application puts up window A, then window B in front of A, then starts					}
{	processing events, the first events will be a deactivate for A and an activate			}
{	for B.  Since it therefore can't be assumed the port was set to A by an activate,		}
{	the port needs to be set for deactivates as well.												}
{																											}
{	This matters not a whit for the more usual cases that occur.  If a deactivate for		}
{	one window is followed by an activate for another, the port will still be switched		}
{	properly to the newly active window.  If no activate follows the deactivate, the			}
{	deactivated window is the last one, and it doesn't matter what the port ends up set to,	}
{	anyway.																								}
procedure DoActivate (h: WHandlerHnd; active: boolean);
	var
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: booleanProcType;
{$ENDC }
	begin
		if h <> nil then begin
			SetPort(h^^.whWind); (* always set the port to the correct one *)
			p := h^^.whActivate;
			if p <> nil then
{$IFC THINK_PASCAL }
				callpBoolean(active, p);
{$ELSEC }
			p(active);
{$ENDC }
		end
	end;

{	Execute a window handler's close proc.  The close box for handlers					}
{	for temp windows that want to remove themselves when the window						}
{	is closed can call SkelRmveWind to dispose of the window							}
{	and remove the handler from the window handler list.  Thus, windows					}
{	may be dynamically created and destroyed without filling up the						}
{	handler list with a bunch of invalid handlers.										}

{	If the handler doesn't have a close proc, just hide the window.						}
{	The host should provide some way of reopening the window (perhaps					}
{	a menu selection).  Otherwise the window will be lost from user						}
{	control if it is hidden, since it won't receive user-initiated events.				}

{	Since the close box of only the active window may be clicked, it					}
{	is not necessary to set the port.																	}

{	This is called both for regular and dialog windows.									}
procedure DoClose (rh: WHandlerHnd);
	var
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: noArgProcType;
{$ENDC }
	begin
		if rh <> nil then begin
			p := rh^^.whClose;
			if (p <> nil) then
{$IFC THINK_PASCAL }
				callPNoArg(p)
{$ELSEC }
				p
{$ENDC }
			else
				HideWindow(rh^^.whWind);
		end;
	end;

{	Execute a window Handler's clobber proc.  This is called both for regular and dialog windows.		}
{	Must save, set and restore port, since any window (not just active one) may be clobbered		}
{	at any time.																										}
{																														}
{	Don't need to check whether handler is nil, as in other handler procedures, since this is only		}
{	called by SkelRmveWind with a known valid handler.														}
procedure DoClobber (h: WHandlerHnd);
	var
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: noArgProcType;
{$ENDC }
		curPort: GrafPtr;
	begin
		if h <> nil then begin
			GetPort(curPort);
			SetPort(h^^.whWind);
			p := h^^.whClobber;
			if p <> nil then
{$IFC THINK_PASCAL }
				callPNoArg(p);
{$ELSEC }
			p;
{$ENDC }
			SetPort(curPort);
		end;
	end;

{$IFC supportDialogs }

{	Handle event if it's for a dialog.  The event must be one of									}
{	those that is passed to dialogs according to dlogEventMask.									}
{	This mask can be set so that disk-inserts, for instance, don't								}
{	get eaten up.																							}
{																											}
{	Examine event and set port if dialog window is coming active (for							}
{	normal windows, DoActivate sets the port, there's no such thing for						}
{	dialogs, so it's done here.)  When this is done, the trio of GetPort/SetPort/SetPort	}
{	calls surrounding the ProcPtr call doesn't appear to be necessary, but on grounds		}
{	of cautiousness, I left them in.  If you want to live dangerously, comment them out.	}
{	Note: In the C version, these are commented out, in Pascal, they're active.				}

function forgetModifier (theEvent: EventRecord): char;
{$IFC THINK_PASCAL }
	var
		someState: LongInt;
{$ELSEC }
	var
		someState:UInt32;
{$ENDC }
	begin
		if isSystemSeven then begin
			someState := 0;
{$IFC THINK_PASCAL }
			forgetModifier := char(BAnd(KeyTrans(Ptr(GetEnvirons(smKCHRCache)), BSR(BAnd(theEvent.message, keyCodeMask), 8), someState), $FF));
{$ELSEC }
			forgetModifier := char(BAnd(KeyTranslate(Ptr(GetScriptManagerVariable(smKCHRCache)), BSR(BAnd(theEvent.message, keyCodeMask), 8), someState), $FF));
{$ENDC }
		end
		else
			forgetModifier := char(BitAnd(theEvent.message, charCodeMask));
	end;

function DoDialog (theEvent: EventRecord): boolean;
	var
		dh: WHandlerHnd;
		theDialog: DialogPtr;
		what, item: integer;
		tmpPort: GrafPtr;
		testMe: longInt;
		ch: char;
	begin

{	handle command keys before they get to IsDialogEvent}

{ Added following line 2/11/88 }
		DoDialog := false;
		what := theEvent.what;
		testMe := BitShift(longInt(1), what);
		testMe := BitAnd(testMe, longInt(dlogEventMask));
		if ((what = keyDown) | (what = autoKey)) & (BitAnd(theEvent.modifiers, cmdKey) <> 0) then begin
			ch := forgetModifier(theEvent);
			dh := GetDHandler(FrontWindow);
			if (dh <> nil) & (dh^^.whEvent <> nil) then begin
				GetPort(tmpPort);
{$IFC THINK_PASCAL }
				callpEvent(-1, cap(ch), theEvent, dh^^.whEvent);
{$ELSEC }
				dh^^.whEvent(-1, cap(ch), theEvent);
{$ENDC }
				SetPort(tmpPort);
				DoDialog := true;
			end
			else begin
 (* the following two lines added to check for command key equivalents *)
				if pMenuEnable <> nil then
{$IFC THINK_PASCAL }
					callPBoolean(true, pMenuEnable);
{$ELSEC }
				pMenuEnable(true);
{$ENDC }
				DoMenuCommand(MenuKey(ch), theEvent.when);
				if pMenuEnable <> nil then
{$IFC THINK_PASCAL }
					callPBoolean(false, pMenuEnable);
{$ELSEC }
				pMenuEnable(false);
{$ENDC }
				DoDialog := true;
			end;
		end
		else if testMe > 0 then
			if IsDialogEvent(theEvent) then begin
				GetPort(tmpPort);

				if what = updateEvt then
					DoUpdate(GetDHandler(WindowPtr(theEvent.message)));

{ Pascal version of Paul DuBois ugly programming award semi-finalist code follows }
				if (what = activateEvt) & (BitAnd(theEvent.modifiers, activeFlag) > 0) & (WindowPeek(theEvent.message)^.windowKind = dialogKind) then
					SetPort(GrafPtr(theEvent.message));
				if DialogSelect(theEvent, theDialog, item) then begin
					dh := GetDHandler(theDialog);
					if (dh <> nil) & (dh^^.whEvent <> nil) then
{$IFC THINK_PASCAL }
						callpEvent(item, chr(0), theEvent, dh^^.whEvent);
{$ELSEC }
					dh^^.whEvent(item, chr(0), theEvent);
{$ENDC }
				end;
				DoDialog := true;
				SetPort(tmpPort);
			end
			else
				DoDialog := false;
	end;
{$ENDC}

{ --------------------------------------------------------------------	}
{							Event-handling routines										}
{ --------------------------------------------------------------------	}

{	Size a window.  If the window has a handler, use the grow limits				}
{	in the handler record, otherwise use the defaults.									}

{	Since the grow region of only the active window may be clicked, it should	}
{	not be necessary to set the port.														}
procedure DoGrow (h: WHandlerHnd; thePort: GrafPtr; startPt: Point);
	var
		r: Rect;
		growRes: longInt;
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: noArgProcType;
{$ENDC }
	begin
		if h = nil then
			r := growRect
		else
			r := h^^.whGrow;
		growRes := GrowWindow(thePort, startPt, r);
		if growRes <> 0 then begin
			SizeWindow(thePort, LoWord(growRes), HiWord(growRes), true);
			p := h^^.whGrowP;
			if p <> nil then
{$IFC THINK_PASCAL }
				callPNoArg(p);
{$ELSEC }
			p;
{$ENDC }
		end;
	end;


{	This is the routine called by the main event loop, when a high level			}
{	event is found.  Since we only deal with Apple events, and not other			}
{	high level events, we just pass everything onto the AEM via AEProcessAppleEvent	}
procedure DoHighLevelEvent (event: EventRecord);
	var
		err: OSErr;
	begin
		err := AEProcessAppleEvent(event);
		if err = errAEEventNotHandled then
			nonFatalError('Unhandled Apple Event (probably ''Print...'')!')
		else
			testOSError(err, noErr);
	end;

{	General event handler}
procedure DoEvent (theEvent: EventRecord);
	var
		evtPt: Point;
		evtPort: GrafPtr;
		evtPart: integer;
		evtChar: char;
		evtMods: integer;
		h: WHandlerHnd;
		ignore: integer;
{$IFC THINK_PASCAL }
		p: ProcPtr;
{$ELSEC }
		p: noArgProcType;
{$ENDC }

	begin
{$IFC THINK_PASCAL }
		if (pEvent <> nil) & callOtherEvent(theEvent, pEvent) then
{$ELSEC }
			if (pEvent <> nil) & pEvent(theEvent) then
{$ENDC }
				exit(DoEvent);
{$IFC supportDialogs }
		if not DoDialog(theEvent) then
{$ENDC}
			begin
			evtPt := theEvent.where;
			evtMods := theEvent.modifiers;
			case theEvent.what of
				kHighLevelEvent: 
					DoHighLevelEvent(theEvent);
				nullEvent: 
					;

{	Mouse click.  Get the window that the click occurred in, and the				}
{	part of the window.	Get WDHandler is called here, not GetWHandler, since		}
{	we need the handler for a window which might turn out to be a dialog window,	}
{	e.g., if the click is in a close box.}
				mouseDown:  begin
					updateRandomizer(theEvent.when);
					evtPart := FindWindow(evtPt, evtPort);
					h := GetWDHandler(evtPort);
					case evtPart of

{	Click in a desk accessory window.  Pass back to the system.}
						inSysWindow:  begin
							if deskAccOpenProc <> nil then
{$IFC THINK_PASCAL }
								callPNoArg(deskAccOpenProc);
{$ELSEC }
							deskAccOpenProc;
{$ENDC }
							SystemClick(theEvent, evtPort);
						end;

{	Click in menu bar.  Track the mouse and execute selected command,}
{	if any.}
						inMenuBar:  begin
							if pMenuEnable <> nil then
{$IFC THINK_PASCAL }
								callPBoolean(true, pMenuEnable);
{$ELSEC }
							pMenuEnable(true);
{$ENDC }
							DoMenuCommand(MenuSelect(evtPt), theEvent.when);
							if pMenuEnable <> nil then
{$IFC THINK_PASCAL }
								callPBoolean(false, pMenuEnable);
{$ELSEC }
							pMenuEnable(false);
{$ENDC }
						end;

{	Click in grow box.  Resize window.}
						inGrow: 
							DoGrow(h, evtPort, evtPt);

{	Click in title bar.  Drag the window around.  Leave at least												}
{	4 pixels visible in both directions.	Bug fix: The window, if not front, is								}
{	selected first to make sure it's at least activated (unless the command key is down - see Inside	}
{	Macintosh).  DragWindow seems to call StillDown first, so that clicks in drag regions while		}
{	machine is busy don't otherwise bring window to front if the mouse is already up by the time		}
{	DragWindow is called.																							}
						inDrag:  begin
							if (evtPort <> FrontWindow) & (BitAnd(evtMods, cmdKey) = 0) then
								SelectWindow(evtPort);
							DragWindow(evtPort, evtPt, goodDraggingRect);
						end;

{	Click in close box.  Call the close proc if the window has one.											}
						inGoAway: 
							if TrackGoAway(evtPort, evtPt) then
								DoClose(GetWDHandler(evtPort));

{	Click in content region.  If the window wasn't frontmost (active),										}
{	just select it, otherwise pass the click to the window's mouse											}
{	click handler.																									}
						inContent: 
							if evtPort = FrontWindow then
								DoMouse(h, theEvent)
							else
								SelectWindow(evtPort);

{	Click in zoom box.  Track the click and then zoom the window if											}
{	necessary																										}
						inZoomIn, inZoomOut: 
							if TrackBox(evtPort, evtPt, evtPart) then begin
								p := h^^.whZoom;
								if p <> nil then
{$IFC THINK_PASCAL }
									callPNoArg(p);
{$ELSEC }
								p;
{$ENDC }
							end;
						otherwise
							;
					end;	{mousedown}
				end;

{	Key event.  If the command key was down, process as menu item	}
{	selection, otherwise pass the character and the modifiers flags	}
{	to the active window's key handler.										}

{	If dialogs are supported, there's no check for command-key			}
{	equivalents, since that would have been checked in DoDialog.		}

				keyDown, autoKey:  begin
					evtChar := char(BitAnd(theEvent.message, charCodeMask));
					if BitAnd(evtMods, cmdKey) > 0 then begin

{ 3/20/89 Added following If clause to check for menu enable routine }
						if pMenuEnable <> nil then
{$IFC THINK_PASCAL }
							callPBoolean(true, pMenuEnable);
{$ELSEC }
						pMenuEnable(true);
{$ENDC }
						DoMenuCommand(MenuKey(evtChar), theEvent.when);
						if pMenuEnable <> nil then
{$IFC THINK_PASCAL }
							callPBoolean(false, pMenuEnable);
{$ELSEC }
						pMenuEnable(false);
{$ENDC }
					end
					else
						DoKey(GetWHandler(FrontWindow), evtChar, evtMods);
				end;

{	Update a window.}
				updateEvt: 
					DoUpdate(GetWHandler(WindowPtr(theEvent.message)));

{	Activate or deactivate a window.}
				activateEvt: 
					DoActivate(GetWHandler(WindowPtr(theEvent.message)), BitAnd(theEvent.modifiers, activeFlag) <> 0);

{	handle Suspend/Resume Events }
{$IFC THINK_PASCAL }
				app4Evt:  begin
{$ELSEC }
					osEvt: 
					begin
{$ENDC }
						if (BSR(theEvent.message, 24) = 1) & (mfResumeProc <> nil) then
{$IFC THINK_PASCAL }
							callPBoolean(BitAnd(theEvent.message, 1) = 1, mfResumeProc);
{$ELSEC }
						mfResumeProc(BitAnd(theEvent.message, 1) = 1);
{$ENDC }
					end;

{	handle inserts of uninitialized disks}
					diskEvt: 
					if HiWord(theEvent.message) <> noErr then begin
						DILoad;
						ignore := DIBadMount(diskInitPt, theEvent.message);
						DIUnload;
					end;
					otherwise
						;
				end;
			end;
		end;

{ --------------------------------------------------------------------	}
{						Interface (public) Routines											}
{ --------------------------------------------------------------------	}


{	Initialize the various Macintosh Managers.																		}
{	Set default upper limits on window sizing.																		}
{	FlushEvents does NOT toss disk insert events, so that disks													}
{	inserted while the application is starting up don't result														}
{	in dead drives.																										}
{	NoMasters is the number of times to call MoreMasters.  gzProc is the address of a user - provided	}
{	grow zone function procedure to call if memory gets tight.  Pass nil if none to be used.					}

procedure SkelInit (NoMasters: integer; myGrowZone: ProcPtr);
	var
		i: integer;
	begin

{	For non-Lightspeed Pascal users, the following inits are included as a compile time option,	}
{	See the $SETC definition at the beginning of the unit.													}

{$IFC false }

		InitGraf(@thePort);
		InitFonts;
		InitWindows;
		InitMenus;
		TEInit;
		InitDialogs(nil);
		MaxApplZone;
{$ENDC}

		FlushEvents(everyEvent - diskMask, 0);
		if myGrowZone <> nil then
			SetGrowZone(myGrowZone);
		for i := 1 to NoMasters do begin
			advanceStopWatchHandsNoMem;
			MoreMasters;
		end;

		whList := nil;
		whClobOnRmve := true;
		SetRect(growRect, 80, 80, 512, 342 - mBarHeight);
		mhList := nil;
		mhClobOnRmve := true;
		appleAboutProc := nil;
		doAboutItems := 0;
		doneFlag := 0;
		pBkgnd := nil;
		pEvent := nil;
		pEventflag := false;
		eventMask := everyEvent;
		diskInitPt.v := 120;
		diskInitPt.h := 100;
{$IFC supportDialogs }
		dlogEventMask := $16f;
{$ENDC}

{	Set upper limits of window sizing to machine screen size.  Allow}
{	for the menu bar.}

		GetWMgrPort(screenPort);
		growRect.right := screenPort^.portRect.right;
		growRect.bottom := screenPort^.portRect.bottom - mBarHeight;

		goodDraggingRect := screenPort^.portRect;
		goodDraggingRect.top := goodDraggingRect.top + mBarHeight;		{ Skip down past menu bar	}
		InsetRect(goodDraggingRect, 4, 4);

{	Set caching global variables to nil }

		oldWindow := nil;
		oldWDHandler := nil;

{	New globals for waitnextevent }

		sleep := 0;
		mouseRgn := nil;
		pMenuEnable := nil;
	end;

{	Main loop.}

{	Task care of DA's with SystemTask.																				}
{	Run background task if there is one.																				}
{	If there is an event, check for an event hook.  If there isn't													}
{	one defined, or if there is but it returns false, call the														}
{	general event handler.  (Hook returns true if TransSkel should												}
{	ignore the event.)																									}
{	If no event, call the "no-event" handler for the front window and for										}
{	any other windows with idle procedures that are always supposed											}
{	to run.  This is done in such a way that it is safe for idle procs												}
{	to remove the handler for their own window if they want (unlikely,										}
{	but...)  This loop doesn't check whether the window is really												}
{	a dialog window or not, but it doesn't have to, because such													}
{	things always have a nil idle proc.																				}
{																															}
{	doneFlag is reset upon exit.  This allows it to be called														}
{	repeatedly, or recursively.																						}

{	Null events are looked at (in SkelMain)																			}
{	and passed to the event handler.  This is necessary to make sure											}
{	DialogSelect gets called repeatedly, or the caret won't blink if												}
{	a dialog has any editText items.  Null events are not passed to any event-inspecting hook that may	}
{	be installed.																											}

	const
		WNETrapNum = $60;			{ WaitNextEvent Trap Number }
		UnImplTrapNum = $9f;			{ Unimplemented trap Number }

procedure SkelMain;
	var
		theEvent: EventRecord;
		wh, wh2: WHandlerHnd;
		w: WindowPtr;
		haveEvent: boolean;
		tmpPort: GrafPtr;
	begin
		WNEIsImplemented := NGetTrapAddress(WNETrapNum, ToolTrap) <> NGetTrapAddress(UnImplTrapNum, ToolTrap);
		while doneFlag = 0 do begin
			if not WNEIsImplemented then
				SystemTask;
			if pBkgnd <> nil then
{$IFC THINK_PASCAL }
				callPNoArg(pBkgnd);
{$ELSEC }
			pBkgnd;
{$ENDC }
{	GetNextEvent calls SystemEvent to handle some DA events and returns false if the event			}
{	was handled.  However, in such cases the event record will still have the event that occurred,	}
{	*not* a null event, as you might reasonably expect.  So it's not enough to look at haveEvent.		}

{	Previous versions figured (wrongly) that haveEvent = false meant a null event had occurred,		}
{	and passed it through to DoEvent and DoDialog, so that caret-blinking in dialog TextEdit items		}
{	would occur.  But cmd-key equivalents while DA windows were in front, in particular,				}
{	allowed already-processed DA events to get into DoEvent (because haveEvent was false),			}
{	and they got handled twice because when the event record was examined, lo and behold, it		}
{	had a cmd-key event!  So now this logic is used:															}
{																														}
{	If have a real event, and there's no event hook or there is but it doesn't handle the event, OR		}
{	if the "non-event" is a true nullEvent, then process it.													}

			if WNEIsImplemented then
				haveEvent := WaitNextEvent(eventMask, theEvent, sleep, mouseRgn)
			else
				haveEvent := GetNextEvent(eventMask, theEvent);

{ following line fixed from version 1.02 & 1.03, and 2.00 (in 2.01 changed to short circurts) 	}
			if haveEvent | (theEvent.what = nullEvent) then
				DoEvent(theEvent);
			if not haveEvent then begin
				wh := whList;
				GetPort(tmpPort);
				while wh <> nil do begin
					wh2 := wh^^.whNext;
					w := wh^^.whWind;
					if (w = FrontWindow) | not wh^^.whFrontOnly then begin
						if not WNEIsImplemented then
							SystemTask;
						if wh^^.whIdle <> nil then begin
							SetPort(wh^^.whWind);
{$IFC THINK_PASCAL }
							callPNoArg(wh^^.whIdle);
{$ELSEC }
							wh^^.whIdle;
{$ENDC }
						end;
					end;
					wh := wh2;
				end;
				SetPort(tmpPort);
			end;
		end;
		doneFlag := 0;
	end;

{	Tell SkelMain to stop}

procedure SkelWhoa;
	begin
		doneFlag := 1;
	end;

{	Clobber all the menu, window and dialog handlers}

procedure SkelClobber;
	begin
		oldWDHandler := nil;
		oldWindow := nil;
		while whList <> nil do
			SkelRmveWind(whList^^.whWind);
		while mhList <> nil do
{$IFC THINK_PASCAL }
			SkelRmveMenu(GetMHandle(mhList^^.mhID), false);
{$ELSEC }
		SkelRmveMenu(GetMenuHandle(mhList^^.mhID), false);
{$ENDC }
	end;

{ --------------------------------------------------------------------	}
{						Menu-handler interface routines					}
{ --------------------------------------------------------------------	}




{	Install handler for a menu.  Remove any previous handler for it.		}
{	Pass the following parameters:											}

{	theMenu	Handle to the menu to be handled.  Must be created by host.		}
{	pSelect	Proc that handles selection of items from menu.  If this is		}
{			nil, the menu is installed, but nothing happens when items		}
{			are selected from it.											}
{	pClobber Proc for disposal of handler's data structures.  Usually		}
{			nil for menus that remain in menu bar until program				}
{			termination.													}

{	The menu is installed and drawn in the menu bar.						}

{ 	Return false if no handler could be allocated, true if successful.		}
{$IFC THINK_PASCAL }
function SkelMenu (theMenu: MenuHandle; pSelect: ProcPtr; pClobber: ProcPtr; drawBar, hierarchical: boolean): boolean;
{$ELSEC }
	function SkelMenu (theMenu: MenuHandle; pSelect: int2LongintProcType; pClobber: menuProcType; drawBar, hierarchical: boolean): boolean;
{$ENDC }
		var
			mh: MHandlerHnd;
			myHand: Handle;
		begin
			mhClobOnRmve := false;
			SkelRmveMenu(theMenu, false);
			mhClobOnRmve := true;
			myHand := newHandleFailNil(sizeOf(MHandler));
			SkelMenu := false;
			if myHand <> nil then begin
				SkelMenu := true;					{ show we really got the memory }
				mh := MHandlerHnd(myHand);
				mh^^.mhNext := mhList;
				mhList := MHandlerHnd(myHand);
				mh^^.mhID := theMenu^^.menuID;	{ get menu id number }
				mh^^.mhSelect := pSelect;		{ install selection handler }
				mh^^.mhClobber := pClobber;		{ install disposal handler }
				if hierarchical then
					InsertMenu(theMenu, -1)		{ put menu in hierarchical part of menu list}
				else
					InsertMenu(theMenu, 0);		{ put menu at end of menu bar }
				if drawBar then
					DrawMenuBar;
			end;
		end;

{	Remove a menu handler.  This calls the handler's disposal routine			}
{	and then takes the handler out of the handler list and disposes				}
{	of it.																					}

{	Note that the menu MUST be deleted from the menu bar before calling		}
{	the clobber proc, because the menu bar will end up filled with				}
{	garbage if the menu was allocated with NewMenu (see discussion of		}
{	DisposeMenu in Menu Manager section of Inside Macintosh).					}
	procedure SkelRmveMenu (theMenu: MenuHandle; drawBar: boolean);
		var
			mID: integer;
			h, h2: MHandlerHnd;
{$IFC THINK_PASCAL }
			p: ProcPtr;
{$ELSEC }
			p: menuProcType;
{$ENDC }
			returnFlag: boolean;
		begin
			mID := theMenu^^.menuID;
			returnFlag := false;
			if mhList <> nil then begin
				if mhList^^.mhID = mID then begin
					h2 := mhList;
					mhList := h2^^.mhNext;
				end
				else begin
					h := mhList;
					while (h <> nil) & not returnFlag do begin
						h2 := h^^.mhNext;
						if (h2 = nil) then begin
							h := nil;
							returnFlag := true;
						end
						else if h2^^.mhID = mID then begin
							h^^.mhNext := h2^^.mhNext;
							h := nil;
						end;
						if h <> nil then
							h := h2;
					end;
				end;
				if not returnFlag then begin
					DeleteMenu(mID);
					if drawBar then
						DrawMenuBar;
					p := h2^^.mhClobber;
					if mhClobOnRmve & (p <> nil) then
{$IFC THINK_PASCAL }
						callpMenu(theMenu, p);
					disposHandle(Handle(h2));
{$ELSEC }
					p(theMenu);
					DisposeHandle(Handle(h2));
{$ENDC }
				end;
			end;
		end;

{	Install a handler for the Apple menu.}

{	SkelApple is called if TransSkel is supposed to handle the apple												}
{	menu itself.  The title is the title of the first item.  If nil,													}
{	then only desk accessories are put into the menu.  If not nil, then											}
{	the title is entered as the first item, followed by a gray line,												}
{	then the desk accessories.																							}

{ 	SkelApple does not cause the menubar to be drawn, so if the Apple menu is the only menu,				}
{	DrawMenuBar must be called afterward.																		}

{	No value is returned, unlike SkelMenu.  It is assumed that SkelApple will be called so early in the	}
{	application that the call to SkelMenu is virtually certain to succeed.										}
{$IFC THINK_PASCAL }
	procedure SkelApple (aboutTitle: Str255; aboutProc, deskAccProc, resumeProc: ProcPtr);
{$ELSEC }
		procedure SkelApple (aboutTitle: Str255; aboutProc: int2LongintProcType; deskAccProc: noArgProcType; resumeProc: booleanProcType);
{$ENDC }
			var
				appleTitle: Str255;
				dummy: boolean;
				i: integer;
			begin
				appleTitle := ' ';
				appleTitle[1] := char($14);
				appleMenu := NewMenu(1, appleTitle);
				if aboutTitle <> '' then begin
					doAboutItems := 1;
					for i := 1 to length(aboutTitle) do
						if aboutTitle[i] = ';' then
							doAboutItems := doAboutItems + 1;
					AppendMenu(appleMenu, aboutTitle);
					AppendMenu(appleMenu, '(-');
					appleAboutProc := aboutProc;
				end;
{$IFC THINK_PASCAL }
				AddResMenu(appleMenu, 'DRVR');
				dummy := SkelMenu(appleMenu, @DoAppleItem, @DoAppleClobber, false, false);
{$ELSEC }
				AppendResMenu(appleMenu, 'DRVR');
				dummy := SkelMenu(appleMenu, DoAppleItem, DoAppleClobber, false, false);
{$ENDC }
				deskAccOpenProc := deskAccProc;
				mfResumeProc := resumeProc;
			end;

{ --------------------------------------------------------------------	}
{					Window-handler interface routines									}
{ --------------------------------------------------------------------	}



{	Install handler for a window.  Remove any previous handler for it.				}
{	Pass the following parameters:														}

{	theWind	Pointer to the window to be handled.  Must be created by host.		}
{	pMouse	Proc to handle mouse clicks in window.  The proc will be				}
{			passed the point (in local coordinates), the time of the					}
{			click, and the modifier flags word.											}
{	pKey	Proc to handle key clicks in window.  The proc will be passed			}
{			the character and the modifier flags word.									}
{	pUpdate	Proc for updating window.  TransSkel brackets calls to update		}
{			procs with calls to BeginUpdate and EndUpdate, so the visRgn			}
{			is set up correctly.  A flag is passed indicating whether the				}
{			window was resized or not.  BY CONVENTION, the entire portRect		}
{			is invalidated when the window is resized.  That way, the				}
{			handler's update proc can redraw the entire content region				}
{			without interference from BeginUpdate/EndUpdate.  The flag			}
{			is set to false after the update proc is called; the							}
{			assumption is made that it will notice the resizing and					}
{			respond appropriately.															}
{	pActivate Proc to execute when window is activated or deactivated.			}
{			A boolean is passed to it which is true if the window is					}
{			coming active, false if it's going inactive.									}
{	pClose	Proc to execute when mouse clicked in close box.  Useful				}
{			mainly to temp window handlers that want to know when to				}
{			self-destruct (with SkelRmveWind).											}
{	pClobber Proc for disposal of handler's data structures							}
{	pIdle	Proc to execute when no events are pending.								}
{	frontOnly True if pIdle should execute on no events only when					}
{			theWind is frontmost, false if executes all the time.  Note				}
{			that if it always goes, everything else may be slowed down!			}

{	If a particular procedure is not needed (e.g., key events are						}
{	not processed by a handler), pass nil in place of the appropriate				}
{	procedure address.																		}

{	Return false  if no handler could be allocated, true if successful.				}
{$IFC THINK_PASCAL }
		function SkelWindow (theWind: WindowPtr; pMouse, pKey, pUpdate, pGrow, pActivate, pClose, pClobber, pIdle: ProcPtr; frontOnly: boolean): boolean;
{$ELSEC }
			function SkelWindow (theWind: WindowPtr; pMouse: mouseProcType; pKey: keyProcType; pUpdate, pGrow: noArgProcType; pActivate: booleanProcType; pClose: noArgProcType; pClobber, pIdle: noArgProcType; frontOnly: Boolean): Boolean;
{$ENDC }
				var
					hHand: WHandlerHnd;
				begin
					whClobOnRmve := false;
					SkelRmveWind(theWind);
					whClobOnRmve := true;

{	Get new handler, attach to list of handlers.  It is attached to the beginning of the list, which is simpler;	}
{	the order should be irrelevant to the host, anyway.																}

					hHand := WHandlerHnd(newHandleFailNil(sizeOf(WHandler)));
					SkelWindow := false;
					if hHand <> nil then begin
						hHand^^.whNext := whList;
						whList := hHand;
						with hHand^^ do begin
							SkelWindow := true;					{ Show that we got the memory }
							whWind := theWind;
							whMouse := pMouse;
							whKey := pKey;
							whUpdate := pUpdate;
							whGrowP := pGrow;
							whActivate := pActivate;
							whClose := pClose;
							whZoom := nil;
							whClobber := pClobber;
							whIdle := pIdle;
							whFrontOnly := frontOnly;
							whGrow := growRect;
						end;
					end;
					SetPort(theWind);
				end;

{	Remove a window handler.  This calls the handler's disposal routine										}
{	and then takes the handler out of the handler list and disposes												}
{	of it.																													}

{	SkelRmveWind is also called by SkelRmveDlog.																	}

{	Note that if the window cache variable is set to the window whose handler is being clobbered, the	}
{	variable must be zeroed.																							}

			procedure SkelRmveWind (theWind: WindowPtr);
				var
					h, h2: WHandlerHnd;
					returnFlag: boolean;
				begin
					if theWind = oldWindow then begin
						oldWindow := nil;
						oldWDHandler := nil;
					end;

					if whList <> nil then begin
						returnFlag := false;
						if whList^^.whWind = theWind then begin
							h2 := whList;
							whList := whList^^.whNext;
						end
						else begin
							h := whList;
							while (h <> nil) & not returnFlag do begin
								h2 := h^^.whNext;
								if h2 = nil then begin
									h := nil;
									returnFlag := true;
								end
								else if h2^^.whWind = theWind then begin
									h^^.whNext := h2^^.whNext;
									h := nil;
								end;
								if h <> nil then
									h := h2;
							end;
						end;
						if not returnFlag then begin
							if whClobOnRmve then
								DoClobber(h2);
{$IFC THINK_PASCAL }
							disposHandle(Handle(h2));
{$ELSEC }
							DisposeHandle(Handle(h2));
{$ENDC }
						end;
					end;
				end;

{$IFC supportDialogs }

{ --------------------------------------------------------------------	}
{					Dialog-handler interface routines										}
{ --------------------------------------------------------------------	}



{	Install a dialog handler.  Remove any previous handler for it.					}
{	SkelDialog calls SkelWindow as a subsidiary to install a window					}
{	handler, then sets the event procedure on return.								}

{	Pass the following parameters:													}

{	theDialog	Pointer to the dialog to be handled.  Must be created				}
{			by host.																}
{	pEvent	Event-handling proc for dialog events.									}
{	pClose	Proc to execute when mouse clicked in close box.  Useful				}
{			mainly to dialog handlers that want to know when to						}
{			self-destruct (with SkelRmveDlog).										}
{	pClobber Proc for disposal of handler's data structures							}

{	If a particular procedure is not needed, pass nil in place of					}
{	the appropriate procedure address.												}

{	Return false if no handler could be allocated, true if successful.				}
{$IFC THINK_PASCAL }
			function SkelDialog (theDialog: DialogPtr; pEvent, pClose, pClobber, pUpdate: ProcPtr): boolean;
{$ELSEC }
				function SkelDialog (theDialog: DialogPtr; pEvent: eventProcType; pClobber, pClose, pUpdate: noArgProcType): boolean;
{$ENDC }
					var
						wh: WHandlerHnd;
						aBool: boolean;
					begin
						aBool := SkelWindow(theDialog, nil, nil, pUpdate, nil, nil, pClose, pClobber, nil, false);
						if aBool then begin
							wh := GetWDHandler(theDialog);
							wh^^.whEvent := pEvent;
						end;
						SkelDialog := aBool;
					end;

{	Remove a dialog and its handler}
				procedure SkelRmveDlog (theDialog: DialogPtr);
					begin
						SkelRmveWind(theDialog);
					end;
{$ENDC}
{ --------------------------------------------------------------------	}
{					Miscellaneous interface routines										}
{ --------------------------------------------------------------------	}


{	Override the default sizing limits for a window, or, if theWind					}
{	is nil, reset the default limits used by SkelWindow.								}
				procedure SkelGrowBounds (theWind: WindowPtr; hLo, vLo, hHi, vHi: integer);
					var
						h: WHandlerHnd;
						r: Rect;
					begin
						if theWind = nil then
							SetRect(growRect, hLo, vLo, hHi, vHi)
						else begin
							h := GetWHandler(theWind);
							if h <> nil then begin
								SetRect(r, hLo, vLo, hHi, vHi);
								h^^.whGrow := r;
							end;
						end;
					end;

{	Set the event mask.}
				procedure SkelEventMask (mask: integer);
					begin
						eventMask := mask;
					end;

{	Return the event mask.}
				procedure SkelGetEventMask (var mask: integer);
					begin
						mask := eventMask;
					end;

{	Install a background task.  If p is nil, the current task is}
{	disabled.}
{$IFC THINK_PASCAL }
				procedure SkelBackground (p: ProcPtr);
{$ELSEC }
					procedure SkelBackground (p: noArgProcType);
{$ENDC }
						begin
							pBkgnd := p;
						end;

{	Return the current background task.  Return nil if none.}
{$IFC THINK_PASCAL }
					procedure SkelGetBackground (var p: ProcPtr);
{$ELSEC }
						procedure SkelGetBackground (var p: noArgProcType);
{$ENDC }
							begin
								p := pBkgnd;
							end;

{	Install an event-inspecting hook.  If p is nil, the hook is}
{	disabled.}
{$IFC THINK_PASCAL }
						procedure SkelEventHook (p: ProcPtr);
{$ELSEC }
							procedure SkelEventHook (p: eventFuncType);
{$ENDC }
								begin
									pEvent := p;
								end;

{$IFC THINK_PASCAL }
							procedure SkelGetEventHook (var p: ProcPtr);
{$ELSEC }
								procedure SkelGetEventHook (var p: eventFuncType);
{$ENDC }
									begin
										p := pEvent;
									end;

{$IFC supportDialogs }

{	Set the mask for event types that will be passed to dialogs.		}
{	Bit 1 is always set, so that null events will be passed.			}
{	If this is not done, the caret does not blink in editText items.	}
								procedure SkelDlogMask (mask: integer);
									begin
										dlogEventMask := BitOr(mask, 1);
									end;

{	Return the current dialog event mask.}
								procedure SkelGetDlogMask (var mask: integer);
									begin
										mask := dlogEventMask;
									end;
{$ENDC}

{ Waitnextevent extensions }

{	Set sleep time }
								procedure SkelSetSleep (time: longInt);
									begin
										sleep := time;
									end;

{	Return the WaitNextEvent sleep time. }
								procedure SkelGetSleep (var time: longInt);
									begin
										time := sleep;
									end;

{	Set the WaitNextEvent mouse region handle. }
								procedure SkelSetMouseRgn (mouse: RgnHandle);
									begin
										mouseRgn := mouse;
									end;

{	Return the WaitNextEvent mouse region handle. }
								procedure SkelGetMouseRgn (var mouse: RgnHandle);
									begin
										mouse := mouseRgn;
									end;

{$IFC THINK_PASCAL }
								procedure SkelSetMenuEnable (pMenuEnableProc: ProcPtr);
									begin
										pMenuEnable := pMenuEnableProc;
									end;

								procedure SkelGetMenuEnable (var pMenuEnableProc: ProcPtr);
									begin
										pMenuEnableProc := pMenuEnable;
									end;
{$ELSEC }
								procedure SkelSetMenuEnable (pMenuEnableProc: booleanProcType);
									begin
										pMenuEnable := pMenuEnableProc;
									end;

								procedure SkelGetMenuEnable (var pMenuEnableProc: booleanProcType);
									begin
										pMenuEnableProc := pMenuEnable;
									end;
{$ENDC }

{$IFC THINK_PASCAL }
								procedure SkelSetWindowZoom (theWind: WindowPtr; pZoomProc: ProcPtr);
{$ELSEC }
									procedure SkelSetWindowZoom (theWind: WindowPtr; pZoomProc: noArgProcType);
{$ENDC }
										var
											wh: WHandlerHnd;
										begin
											wh := whList;
											while (wh <> nil) & (wh^^.whWind <> theWind) do
												wh := wh^^.whNext;
											if wh <> nil then
												wh^^.whZoom := pZoomProc;
										end;

{$IFC THINK_PASCAL }
									procedure SkelGetWindowZoom (theWind: WindowPtr; var pZoomProc: ProcPtr);
{$ELSEC }
										procedure SkelGetWindowZoom (theWind: WindowPtr; var pZoomProc: noArgProcType);
{$ENDC }
											var
												wh: WHandlerHnd;
											begin
												wh := whList;
												while (wh <> nil) & (wh^^.whWind <> theWind) do
													wh := wh^^.whNext;
												if wh <> nil then
													pZoomProc := wh^^.whZoom;
											end;

											end.
