(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit QDPicture;

{$setc usingDevices := true }

interface

uses
	basicGeometry;

const
	noMemMsg = 'Not enough memory for this operation. Sorry.';

type
	QDPicture = object(rectangle)
			pic: PicHandle;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			function isSelected (where: obj): boolean;
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure free;
			override;
			function duplicate: obj;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
{$IFC GFEnabled }
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}
		end;

	ColorQDPicture = object(QDPicture)
			palette: integer; (* resource ID *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure interactiveOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			function memberType (o: obj): boolean;
			override;
		end;

procedure OpenQDPicture (left, top, right, bottom: integer);
function CloseQDPicture (color: boolean): QDPicture;

function OpenBMQDPicture (left, top, right, bottom: integer): boolean;
function CloseBMQDPicture: QDPicture;

function OpenColorBMQDPicture (left, top, right, bottom, palette: integer): boolean;
function CloseColorBMQDPicture: ColorQDPicture;

procedure HSBColor (hue, sat, bri: real; var rgb: RGBColor);

procedure initQDPicture;

implementation

{$IFC THINK_PASCAL }
function GetNewPalette (PaletteID: integer): Handle;
inline
	$AA92;
function GetPalette (srcWindow: WindowPtr): Handle;
inline
	$AA96;
procedure SetPalette (dstWindow: WindowPtr; srcPalette: Handle; cUpdates: boolean);
inline
	$AA95;
procedure Palette2CTab (srcPalette: Handle; dstCTab: CTabHandle);
inline
	$AAA0;
{$ELSEC }
uses Palettes;
{$ENDC }

var
	currentPicture: PicHandle;
	currentPalette: integer;
	savePort: GrafPtr;
	saveGDevice, offscreen: GDHandle;
	currentPort, currentCPort: GrafPort;

function handleLessThan (a, b: Handle): boolean;
	type
		bytePtr = ^Byte;
	var
		ap, bp, ae, be: longInt;
	begin
		ap := longInt(a^);
		bp := longInt(b^);
		ae := ap + GetHandleSize(a);
		be := bp + GetHandleSize(b);
		while (ap < ae) & (bp < be) & (bytePtr(ap)^ = bytePtr(bp)^) do begin
			ap := ap + 1;
			bp := bp + 1;
		end;
		handleLessThan := (bp < be) & ((ap = ae) | (ap < ae) & (bytePtr(ap)^ < bytePtr(bp)^));
	end;

function QDPicture.genericLessThan (o: obj): boolean;
	var
		c: QDPicture;
	begin
		c := QDPicture(o);
		genericLessThan := handleLessThan(Handle(pic), Handle(c.pic));
	end;

procedure defaultPicture (p: QDPicture);
	var
		r: Rect;
	begin
		SetRect(r, 0, 0, 100, 100);
		p.pic := OpenPicture(r);
		MoveTo(0, 0);
		LineTo(100, 100);
		MoveTo(0, 100);
		LineTo(100, 0);
		FrameOval(r);
		ClosePicture;
	end;

function QDPicture.getType: objectType;
	begin
		getType := QDPictureType;
	end;

function ColorQDPicture.getType: objectType;
	begin
		getType := ColorQDPictureType;
	end;

function QDPicture.getTextType: Str255;
	begin
		getTextType := 'QuickDraw Picture';
	end;

function ColorQDPicture.getTextType: Str255;
	begin
		getTextType := 'Color QuickDraw Picture'
	end;

function QDPicture.memberType (o: obj): boolean;
	begin
		memberType := member(o, QDPicture);
	end;

function ColorQDPicture.memberType (o: obj): boolean;
	begin
		memberType := member(o, ColorQDPicture);
	end;

procedure QDPicture.interactiveOutput;
	var
		r: Rect;
	begin
		if not currentFrame.proj3d & worldToWindow(left, bottom, r.left, r.top) & worldToWindow(right, top, r.right, r.bottom) then
			DrawPicture(pic, r)
		else
			inherited interactiveOutput;
	end;

procedure ColorQDPicture.interactiveOutput;
	begin
{$IFC THINK_PASCAL }
		if colorAvailable & (GetPalette(thePort) = nil) then
			SetPalette(thePort, GetNewPalette(palette), true);
{$ELSEC }
		if colorAvailable & (GetPalette(qd.thePort) = nil) then
			SetPalette(qd.thePort, GetNewPalette(palette), true);
{$ENDC }
		inherited interactiveOutput;
	end;

function QDPicture.isSelected (where: obj): boolean;
	var
		d: boolean;
	function sect1D (l1, h1, l2, h2: real): boolean;
		begin
			if h2 < h1 then
				h1 := h2;
			if l2 > l1 then
				l1 := l2;
			sect1D := l1 <= h1;
		end;
	begin
		if member(where, rectangle) then
			with rectangle(where) do
				d := sect1D(left, right, self.left, self.right) & sect1D(bottom, top, self.bottom, self.top)
		else
			d := false;
		isSelected := d;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function QDPicture.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	begin
		dragVertex := false;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure QDPicture.interactiveHighlight;
	var
		r: Rect;
	begin
		PenSize(3, 3);
		if not currentFrame.proj3d & worldToWindow(left, bottom, r.left, r.top) & worldToWindow(right, top, r.right, r.bottom) then begin
			r.left := r.left - 2;
			r.right := r.right + 2;
			r.top := r.top - 2;
			r.bottom := r.bottom + 2;
			FrameRect(r);
		end
		else
			inherited interactiveOutput;
		PenSize(1, 1);
	end;

procedure QDPicture.free;
	begin
		KillPicture(pic);
		shallowFree;
	end;

function QDPicture.duplicate: obj;
	var
		o: obj;
	begin
		o := shallowDuplicate;
		if HandToHand(Handle(QDPicture(o).pic)) <> noErr then
			failNil(nil);
		duplicate := o;
	end;

procedure QDPicture.fileOutput;
	begin
		inherited fileOutput;
		writeHandle(Handle(pic));
	end;

procedure QDPicture.fileInput;
	begin
		inherited fileInput;
		readHandle(Handle(pic));
	end;

procedure QDPicture.textFileOutput;
	begin
		inherited textFileOutput;
		writeTextHandle(Handle(pic));
	end;

procedure QDPicture.textFileInput;
	begin
		inherited textFileInput;
		readTextHandle(Handle(pic));
	end;

procedure ColorQDPicture.fileOutput;
	begin
		writeInteger(palette);
		inherited fileOutput;
	end;

procedure ColorQDPicture.fileInput;
	begin
		readInteger(palette);
		inherited fileInput;
		if not colorAvailable then begin
			nonFatalError('There is a color picture in this file, but your Macintosh is not able to display color. Sorry.');
{$IFC THINK_PASCAL }
			disposHandle(Handle(pic));
{$ELSEC }
			DisposeHandle(Handle(pic));
{$ENDC }
			defaultPicture(self);
		end;
	end;

procedure ColorQDPicture.textFileOutput;
	begin
		writeTextInteger(palette);
		inherited textFileOutput;
	end;

procedure ColorQDPicture.textFileInput;
	begin
		readTextInteger(palette);
		inherited textFileInput;
		if not colorAvailable then begin
			nonFatalError('There is a color picture in this file, but your Macintosh is not able to display color. Sorry.');
{$IFC THINK_PASCAL }
			disposHandle(Handle(pic));
{$ELSEC }
			DisposeHandle(Handle(pic));
{$ENDC }
			defaultPicture(self);
		end;
	end;

{$IFC GFEnabled }
function QDPicture.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := false;
	end; { canUseGridfile }
{$ENDC}

procedure OpenQDPicture (left, top, right, bottom: integer);
	var
		r: Rect;
	begin
		SetRect(r, left, top, right, bottom);
		currentPicture := OpenPicture(r);
		ClipRect(r);
	end;

function CloseQDPicture (color: boolean): QDPicture;
	var
		p: QDPicture;
		cp: ColorQDPicture;
	begin
		ClosePicture;
		if color then begin
			new(cp);
			failNil(Handle(cp));
			p := cp;
		end
		else begin
			new(p);
			failNil(Handle(p));
		end;
		with currentPicture^^.picFrame do begin
			p.left := left;
			p.right := right;
			p.bottom := top;
			p.top := bottom;
		end;
		p.pic := currentPicture;
		CloseQDPicture := p;
	end;

function OpenBMQDPicture (left, top, right, bottom: integer): boolean;
	var
		r: Rect;
	begin
		OpenBMQDPicture := false;
		SetRect(r, left, top, right, bottom);
		GetPort(savePort);
		OpenPort(@currentPort);
		InitPort(@currentPort);
		SetPort(@currentPort);
		with currentPort.portBits do begin
			bounds := r;
			rowBytes := (right - left + 15) div 16 * 16;
			baseAddr := NewPtr(longInt(bottom - top) * rowBytes);
			if baseAddr = nil then begin
				nonFatalError(noMemMsg);
				ClosePort(@currentPort);
				SetPort(savePort);
				exit(OpenBMQDPicture);
			end;
		end;
		ClipRect(r);
{$IFC THINK_PASCAL }
		RectRgn(thePort^.visRgn, r);
{$ELSEC }
		RectRgn(qd.thePort^.visRgn, r);
{$ENDC }
		EraseRect(r);
		OpenBMQDPicture := true;
	end;

function CloseBMQDPicture: QDPicture;
	var
		p: QDPicture;
	begin
		with currentPort.portBits do begin
			OpenQDPicture(bounds.left, bounds.top, bounds.right, bounds.bottom);
{$IFC THINK_PASCAL }
			CopyBits(currentPort.portBits, thePort^.portBits, bounds, bounds, srcXor, nil);
{$ELSEC }
			CopyBits(currentPort.portBits, qd.thePort^.portBits, bounds, bounds, srcXor, nil);
{$ENDC }
			p := CloseQDPicture(false);
{$IFC THINK_PASCAL }
			disposPtr(baseAddr);
{$ELSEC }
			DisposePtr(baseAddr);
{$ENDC }
		end;
		ClosePort(@currentPort);
		SetPort(savePort);
		CloseBMQDPicture := p;
	end;

procedure HSBColor (hue, sat, bri: real; var rgb: RGBColor);
	function toInt (r: real): integer;
		var
			ri: longInt;
		begin
			if r > 1 then
				r := 1
			else if r < 0 then
				r := 0;
{$IFC THINK_PASCAL }
			ri := round(r * 65535);
{$ELSEC }
			ri := roundLong(r * 65535);
{$ENDC }
			if ri > maxInt then
				toInt := ri - 65536
			else
				toInt := ri;
		end;
		var
			i, b, p, q, t: integer;
	begin
		hue := 6 * hue;
		if hue < 0 then
			hue := hue + 6;
{$IFC THINK_PASCAL }
		i := trunc(hue);
{$ELSEC }
		i := truncLong(hue);
{$ENDC }
		hue := hue - i;
		b := toInt(bri);
		p := toInt(bri * (1 - sat));
		q := toInt(bri * (1 - sat * hue));
		t := toInt(bri * (1 - sat * (1 - hue)));
		with rgb do
			case i mod 6 of
				0:  begin
					red := b;
					green := t;
					blue := p;
				end;
				1:  begin
					red := q;
					green := b;
					blue := p;
				end;
				2:  begin
					red := p;
					green := b;
					blue := t;
				end;
				3:  begin
					red := p;
					green := q;
					blue := b;
				end;
				4:  begin
					red := t;
					green := p;
					blue := b;
				end;
				5:  begin
					red := b;
					green := p;
					blue := q;
				end;
			end;
	end;

	var
		outP: integer;
		outParr: array[0..20] of longInt;

function OpenColorBMQDPicture (left, top, right, bottom, palette: integer): boolean;
	var
		r: Rect;
		bigPtr: Ptr;
		pmPtr: PixMapPtr;
	begin
		OpenColorBMQDPicture := false;
		if not colorAvailable then begin
			nonFatalError('No color available on this Macintosh. Sorry.');
			exit(OpenColorBMQDPicture);
		end;
		currentPalette := palette;
		SetRect(r, left, top, right, bottom);
		GetPort(savePort);
		saveGDevice := GetGDevice;
		with offscreen^^ do begin
			gdType := 0;
			gdResPref := 4;
			gdRect := r;
			gdReserved := 0;
		end;
		pmPtr := offscreen^^.gdPMap^;
		Palette2CTab(GetNewPalette(palette), pmPtr^.pmTable);
		MakeITable(pmPtr^.pmTable, offscreen^^.gdITable, 4);
		with pmPtr^ do begin
			bounds := r;
			rowBytes := (right - left + 15) * pixelSize div 16 * 2;
		end;
		bigPtr := NewPtr(longInt(bottom - top) * pmPtr^.rowBytes);
		if bigPtr = nil then begin
			nonFatalError(noMemMsg);
			exit(OpenColorBMQDPicture);
		end;
		with pmPtr^ do begin
			baseAddr := bigPtr;
			rowBytes := -maxInt - 1 + rowBytes;
		end;
		SetGDevice(offscreen);
		OpenCPort(@currentCPort);
		InitCPort(@currentCPort);
		SetPort(@currentCPort);
		ClipRect(r);
{$IFC THINK_PASCAL }
		RectRgn(thePort^.visRgn, r);
{$ELSEC }
		RectRgn(qd.thePort^.visRgn, r);
{$ENDC }
		EraseRect(r);
		OpenColorBMQDPicture := true;
	end;

function CloseColorBMQDPicture: ColorQDPicture;
	var
		pmPtr: PixMapPtr;
		p: ColorQDPicture;
	begin
		SetPort(savePort);
		SetGDevice(saveGDevice);
		pmPtr := offscreen^^.gdPMap^;
		with pmPtr^ do
			OpenQDPicture(bounds.left, bounds.top, bounds.right, bounds.bottom);
		with pmPtr^ do
{$IFC THINK_PASCAL }
			CopyBits(GrafPort(currentCPort).portBits, thePort^.portBits, bounds, bounds, srcXor, nil);
{$ELSEC }
			CopyBits(GrafPort(currentCPort).portBits, qd.thePort^.portBits, bounds, bounds, srcXor, nil);
{$ENDC }
		p := ColorQDPicture(CloseQDPicture(true));
		p.palette := currentPalette;
{$IFC THINK_PASCAL }
		disposPtr(pmPtr^.baseAddr);
{$ELSEC }
		DisposePtr(pmPtr^.baseAddr);
{$ENDC }
		CloseCPort(@currentCPort);
		CloseColorBMQDPicture := p;
	end;

procedure initQDPicture;
	var
		p: QDPicture;
		cp: ColorQDPicture;
	begin
		new(p);
		failNil(Handle(p));
		prototypes[QDPictureType] := p;
		new(cp);
		failNil(Handle(cp));
		prototypes[ColorQDPictureType] := cp;
		if colorAvailable then begin
			offscreen := NewGDevice(0, -1);
			with offscreen^^.gdPMap^^ do begin
				pmVersion := 0;
				pixelType := 0;
				pixelSize := 8;
				cmpCount := 1;
				cmpSize := 8;
				packType := 0;
				packSize := 0;
				planeBytes := 0;
				pmReserved := 0;
				SetRect(bounds, 0, 0, 0, 0);
				rowBytes := 0;
				baseAddr := nil;
			end;
		end;
	end;

end.
