(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit basicGeometry;

interface

uses
	getUserParameter;

const
	lpInfinity = 99;

type
	point2d = object(obj)
			procedure init;
			override;
			procedure randomChange;  (* The global variable 'currentRandomConstraint' is interpreted as a rectangle. *)
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			procedure interactiveOutput;
			override;
			procedure textOutput;
			override;
			procedure windowOutput;
			override;
			procedure interactiveHighlight;
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure construct (px, py, newX, newY: integer);
			override;
			procedure windowToWorld;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			function convertToType (t: pointTypeRange): obj;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function getBoundingObject: obj;
			override;
			function convertToConvexPolygon: obj;
			override;
			function convertToConvexPolygonPossible: boolean;
			override;
			function HBGLessThan (o: obj): boolean;
			override;
			function getPointType: objectType;
			override;
			function getLeftPoint: obj;
			override;
			function getRightPoint: obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function getX: extended;								(* Get the value of the x-coordinate. *)
			procedure setX (newX: extended);					(* Set the value of the x-coordinate. *)
			function getY: extended;								(* Get the value of the y-coordinate. *)
			procedure setY (newY: extended);					(* Set the value of the y-coordinate. *)
			procedure setXY (newX, newY: extended);			(* Sets the value of the x- and y-coordinate. *)
			procedure getXY (var newX, newY: extended);	(* Gets the value of the x- and y-coordinate. *)
			function xLessThan (p: point2d): boolean;			(* Is 'self.getX < p.getX' ? *)
			function xEqual (p: point2d): boolean;					(* Is 'self.getX = p.getX' ? *)
			function yLessThan (p: point2d): boolean;			(* Is 'self.getY < p.getY' ? *)
			function yEqual (p: point2d): boolean;					(* Is 'self.getY = p.getY' ? *)
			function equalPoint (p: point2d): boolean;			(*	Is 'self.getX = p.getX & self.getY = p.getY' *)

(* Drawing *)
			procedure drawLine (p: point2d); (* Draws a line from 'self' to 'p'. *)
			procedure drawLineHighlight (p: point2d; highlight: boolean); (* Draws a line from 'self' to 'p'. *)
			procedure labelPoint (l: Str255); (* Draws the string 'l' below 'self'. *)
			procedure drawSimpleLine (p: point2d);
(* inserted by PV *)
			procedure drawRectangle (p: point2d);
			procedure drawCrossOfRectangle (p, middleP: point2d);
(* *)

			function whichSideX (p, q: point2d): extended;
(* Determines on which side of the directed line segment from 'p' to 'q' the point 'self' lies.	*)
(*	< 0: 'self' lies to the left of the directed line segment 'pq'										*)
(*	= 0: 'self' lies on the directed line segment 'pq'													*)
(*	> 0: 'self' lies to the right of the directed line segment 'pq'  									*)
(* whichSideX' is also twice the signed area of the triangle with vertices 'self', 'p' and 'q'.	*)

			function leftOf (p, q: point2d): boolean;
(* Bestimmt, ob self links der Geraden 'pq' liegt *)
(* Eingefuegt von Michael Vollenbroeker 2.11.92 *)

			function rightOf (p, q: point2d): boolean;
(* Bestimmt, ob self rechts der Geraden 'pq' liegt *)
(* Eingefuegt von Michael Vollenbroeker 2.11.92 *)

			function onLine (p, q: point2d): boolean;
(* Bestimmt, ob 'self' auf der Geraden 'pq' liegt *)
(* Eingefuegt von Michael Vollenbroeker 20.11.92 *)

			function whichSideSign (p, q: point2d): signType;
(* Computes sign(whichSideX) and rounds it to '0' if the exact result cannot be determined.	*)

			function crossProductX (p, q, r: point2d): extended;
(* The cross product ('self' - 'p')  ('r' - 'q'). *)

			function scalarProductX (p: point2d): extended;
(* The scalar product ('self.x' * 'p.x') + ('self.y' * 'p.y') ;            added by FB					*)

			function distanceX (p: point2d): extended;
(* The Euclidean distance between 'self' and 'p'. *)

			function squaredDistanceX (p: point2d): extended;
(* The squared Euclidean distance between 'self' and 'p'. *)

			function squaredDxX (p: point2d): extended;
(* The squared difference in x-coordinates between 'self' and 'p'. *)

			function squaredDyX (p: point2d): extended;
(* The squared difference in y-coordinates between 'self' and 'p'. *)

			function distanceLpX (p: point2d): extended;
(* The Lp-distance between 'self' and 'p'. *)

			function distancePLpX (p: point2d): extended;
(* The p-th power of the Lp-distance between 'self' and 'p'. If 'lpMetric' = 'lpInfinity',	*)
(*	the Lp-distance between 'self' and 'p' is returned.											*)

			function dxPLpX (p: point2d): extended;
(* The p-th power of difference in x-coordinates between 'self' and 'p'.						*)
(*	If 'lpMetric' = 'lpInfinity', the absolute value of the difference in x-coordindates		*)
(*	between 'self' and 'p' is returned. 																*)

			function dyPLpX (p: point2d): extended;
(* The p-th power of difference in y-coordinates between 'self' and 'p'.						*)
(*	If 'lpMetric' = 'lpInfinity', the absolute value of the difference in y-coordindates		*)
(*	between 'self' and 'p' is returned. 																*)

			procedure circleCenterX (p, q: point2d; var cx, cy: extended);
(* The point ('cx', 'cy') is the center of the circle through 'self', 'p' and 'q'. *)

			function intersectLineSegmentX (q, r, s, left, right: point2d): boolean;
(* True iff the segment 'self q' and 'rs' intersect. The left intersection point is 'left', the right *)
(* intersection point is 'right' (lexicographically). *)

			function intersectLineX (q, r, s, left, right: point2d): boolean;
(* True iff the line 'self q' and the line segment 'rs' intersect. The left intersection point is 'left', the right *)
(* intersection point is 'right' (lexicographically). *)

			function lineIntersectsLine (q, r, s, intersectionPoint: point2d): boolean;
(* True, iff the lines 'self q' and 'r s' intersect. The intersection is 'intersectionPoint'. *)

			procedure projectOnLine (endPoint, candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
(*	'projectionPoint' becomes the projection of 'candidatePoint' onto the line 'self endPoint'.			*)
(*	'distanceToLine' becomes the distance of 'candidatePoint' to the line 'self endPoint'.				*)
(*	'whichSide' is the the result of 'candidatePoint.whichSideX(self, endPoint)'.							*)
(*	'lambdaSign' is -1 if the projection point lies on the infinite ray emanating from 'self',			*)
(*	'lambdaSign' is 0 if the projection point lies on the segment 	'self endPoint' and						*)
(*	'lambdaSign' is +1 if the projection point lies on the infinite ray emanating from 'endPoint'.		*)
(*	'projectionPoint' can be 'nil'. In this case the projection point is not computed.						*)

			function orthogonalProjection (endPoint, candidatePoint, projectionPoint: point2d): boolean;
(* True, falls die orthogonale Projektion von candidatePoint auf das Liniensegment 'self endPoint' existiert. *)
(* projectionPoint ist die orthogonale Projektion auf die Gerade 'self endPoint' .                                           *)
(* Eingefuegt von Michael Vollenbroeker 20.11.92 *)

			function distanceToLineSegmentX (p, q: point2d; var closestX, closestY: extended): extended;
(* added by FB *)
(* the Euclidean distance between 'self' and the line segment with endpoints p and q; closestX *)
(* and closestY become the coordinates of the point on the line segment closest to 'self'         *)

			function xPlusYX: extended; (* Computes 'x'+'y'. *)
			function xMinusYX: extended; (* Computes 'x'-'y'. *)

			function rayEvalX (p: point2d; cosSlope, sinSlope: extended): extended;
(* Evaluates the positive ray emanating from 'p' with the given slope at 'self.getX'. *)

			function bisectorEvalX (p, q: point2d): extended;
(* Evaluates the bisector of 'p' and 'q' at 'self.getX'. *)

			function intersectRayBisectorX (cosSlope, sinSlope: extended; p, q, intersection: point2d): boolean;
(* Tests whether the positive ray emanating from 'self' with the given slope intersects the bisector given by 'p' and 'q'. *)

			function intersectBisectorBisectorX (p, q, r, intersection: point2d): boolean;
(* Tests whether the bisector('self','p') intersects bisector('q','r') where |setOf('self', p, q, r) | = 3 *)

			function intersectRayRayX (cosLower, sinLower: extended; p: point2d; cosUpper, sinUpper: extended; intersection: point2d): boolean;
(* Tests whether the rays ('self', 'cosLower', 'sinLower') and ('p', 'cosUpper', 'sinUpper') intersect. *)

			function middle (p: point2d): point2d;
(* Computes the point ('self.getX + p.getX', 'self.getY + p.getY') / 2. *)

			function createLineSegment (p: point2d): lineSegment;
(* Creates a line segment from 'self' to 'p'. *)

(* inserted by PV *)
			function createDirectedLineSegment (p: point2d): directedLineSegment;

			function createCircle (p, q: point2d): circle;
(* Creates the circle through 'self', 'p' and 'q' if they are not collinear. *)

(* ------------ inserted by JM  ----------------------------------------------------------------------- *)

			function DxX (p: point2d): extended; (* self.x - p.x.   *)

			function DyX (p: point2d): extended; (* self.y - p.y.   *)

			procedure setPoint (newp: point2d);
(* sets:   self.x:= newp.x    and     self.y:=newp.y  *)

(* ---------------------------------------------------------------------------------------------------------------------------   *)

		end; (* point2d *)

	realPoint = object(point2d)
			x, y: real;
			function getX: extended;
			override;
			procedure setX (newX: extended);
			override;
			function getY: extended;
			override;
			procedure setY (newY: extended);
			override;
			function xLessThan (p: point2d): boolean;
			override;
			function xEqual (p: point2d): boolean;
			override;
			function yLessThan (p: point2d): boolean;
			override;
			function yEqual (p: point2d): boolean;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function equal (o: obj): boolean;
			override;
			function equalPoint (p: point2d): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function middle (p: point2d): point2d;
			override;
			function whichSideX (p, q: point2d): extended;
			override;
			function crossProductX (p, q, r: point2d): extended;
			override;
			function scalarProductX (p: point2d): extended;
			override;                                                                        (* added by FB *)
			function distanceX (p: point2d): extended;
			override;
			function squaredDxX (p: point2d): extended;
			override;
			function squaredDyX (p: point2d): extended;
			override;
			function squaredDistanceX (p: point2d): extended;
			override;
			function distanceLpX (p: point2d): extended;
			override;
			function distancePLpX (p: point2d): extended;
			override;
			function dxPLpX (p: point2d): extended;
			override;
			function dyPLpX (p: point2d): extended;
			override;
			procedure circleCenterX (p, q: point2d; var cx, cy: extended);
			override;
			function intersectLineSegmentX (q, r, s, left, right: point2d): boolean;
			override;
			function lineIntersectsLine (q, r, s, intersectionPoint: point2d): boolean;
			override;
			function intersectLineX (q, r, s, left, right: point2d): boolean;
			override;
			function distanceToLineSegmentX (p, q: point2d; var closestX, closestY: extended): extended;
			override;														  (* added by FB *)
			function xPlusYX: extended;
			override;
			function xMinusYX: extended;
			override;
			function rayEvalX (p: point2d; cosSlope, sinSlope: extended): extended;
			override;
			function bisectorEvalX (p, q: point2d): extended;
			override;
			function intersectRayBisectorX (cosSlope, sinSlope: extended; p, q, intersection: point2d): boolean;
			override;
			function intersectBisectorBisectorX (p, q, r, intersection: point2d): boolean;
			override;
			function intersectRayRayX (cosLower, sinLower: extended; p: point2d; cosUpper, sinUpper: extended; intersection: point2d): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function verticalDistance (o: obj): extended;
			override;
			function distanceO (o: obj; var minP, minQ: obj): extended;
			override;
			procedure projectOnLine (endPoint, candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
			override;
{$IFC GFEnabled }
			function convertToParameterPoint: obj;
			override;
			function convertFromParameterPoint (pP: obj): obj;
			override;
			procedure createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
			override;
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}
(* ---------------------------------- inserted by JM ----------------------------------------   *)
			function DxX (p: point2d): extended;
			override;
			function DyX (p: point2d): extended;
			override;
			procedure setPoint (newP: point2d);
			override;
(* ------------------------------------------------------------------------------------------------  *)
		end;

	markedRealPoint = object(realPoint)
			mark: strings;
			procedure init;
			override;
			procedure free;
			override;
			function duplicate: obj;
			override;
			procedure interactiveOutput;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function equal (o: obj): boolean;
			override;
			function equalPoint (p: point2d): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function getInfo: obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
			function getBoundingObject: obj;
			override;
			procedure initLocationMark (o: obj; s: Str255; translation: boolean);
(* location is 'o.centerOfGravity' and value of 'mark' is 's' *)
			function toRealPoint: realPoint;
(* ignore mark *)
		end;

	longIntPoint = object(point2d)
			x, y: longInt;
			function getX: extended;
			override;
			procedure setX (newX: extended);
			override;
			function getY: extended;
			override;
			procedure setY (newY: extended);
			override;
			function xLessThan (p: point2d): boolean;
			override;
			function xEqual (p: point2d): boolean;
			override;
			function yLessThan (p: point2d): boolean;
			override;
			function yEqual (p: point2d): boolean;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function equal (o: obj): boolean;
			override;
			function equalPoint (p: point2d): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function middle (p: point2d): point2d;
			override;
			function whichSideX (p, q: point2d): extended;
			override;
			function crossProductX (p, q, r: point2d): extended;
			override;
			function scalarProductX (p: point2d): extended; (* added by FB *)
			override;
			function distanceX (p: point2d): extended;
			override;
			function squaredDxX (p: point2d): extended;
			override;
			function squaredDyX (p: point2d): extended;
			override;
			function squaredDistanceX (p: point2d): extended;
			override;
			function distanceLpX (p: point2d): extended;
			override;
			function distancePLpX (p: point2d): extended;
			override;
			function dxPLpX (p: point2d): extended;
			override;
			function dyPLpX (p: point2d): extended;
			override;
			function intersectLineSegmentX (q, r, s, left, right: point2d): boolean;
			override;
			function intersectLineX (q, r, s, left, right: point2d): boolean;
			override;
			function lineIntersectsLine (q, r, s, intersectionPoint: point2d): boolean;
			override;
			function distanceToLineSegmentX (p, q: point2d; var closestX, closestY: extended): extended; (* added by FB *)
			override;
			function xPlusYX: extended;
			override;
			function xMinusYX: extended;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure circleCenterX (p, q: point2d; var cx, cy: extended);
			override;
			function verticalDistance (o: obj): extended;
			override;
			function distanceO (o: obj; var minP, minQ: obj): extended;
			override;
			procedure projectOnLine (endPoint, candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
			override;
(* ---------------------------------- inserted by JM ----------------------------------------   *)
			function DxX (p: point2d): extended;
			override;
			function DyX (p: point2d): extended;
			override;
			procedure setPoint (newP: point2d);
			override;
(* ------------------------------------------------------------------------------------------------  *)
		end;

	floatPoint = object(point2d)
			x, y: float;
			function getX: extended;
			override;
			procedure setX (newX: extended);
			override;
			function getY: extended;
			override;
			procedure setY (newY: extended);
			override;
			function xLessThan (p: point2d): boolean;
			override;
			function xEqual (p: point2d): boolean;
			override;
			function yLessThan (p: point2d): boolean;
			override;
			function yEqual (p: point2d): boolean;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function equal (o: obj): boolean;
			override;
			function equalPoint (p: point2d): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function whichSideX (p, q: point2d): extended;
			override;
			function whichSideSign (p, q: point2d): signType;
			override;
			function crossProductX (p, q, r: point2d): extended;
			override;
			function scalarProductX (p: point2d): extended;  (* added by FB *)
			override;
			function distanceX (p: point2d): extended;
			override;
			function squaredDxX (p: point2d): extended;
			override;
			function squaredDyX (p: point2d): extended;
			override;
			function squaredDistanceX (p: point2d): extended;
			override;
			function distanceLpX (p: point2d): extended;
			override;
			function distancePLpX (p: point2d): extended;
			override;
			function dxPLpX (p: point2d): extended;
			override;
			function dyPLpX (p: point2d): extended;
			override;
			procedure circleCenterX (p, q: point2d; var cx, cy: extended);
			override;
			function intersectLineSegmentX (q, r, s, left, right: point2d): boolean;
			override;
			function intersectLineX (q, r, s, left, right: point2d): boolean;
			override;
			function lineIntersectsLine (q, r, s, intersectionPoint: point2d): boolean;
			override;
			function distanceToLineSegmentX (p, q: point2d; var closestX, closestY: extended): extended; (* added by FB *)
			override;
			function xPlusYX: extended;
			override;
			function xMinusYX: extended;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function middle (p: point2d): point2d;
			override;
			function verticalDistance (o: obj): extended;
			override;
			function distanceO (o: obj; var minP, minQ: obj): extended;
			override;
			procedure projectOnLine (endPoint, candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
			override;
(* ---------------------------------- inserted by JM ----------------------------------------   *)
			function DxX (p: point2d): extended;
			override;
			function DyX (p: point2d): extended;
			override;
			procedure setPoint (newP: point2d);
			override;
(* ------------------------------------------------------------------------------------------------  *)
		end;

	idRealPoint = object(realPoint)
			id: seqIndex;
			function getTextType: Str255;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function equal (o: obj): boolean;
			override;
			function equalPoint (p: point2d): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function genericLessThan (o: obj): boolean;
			override;
{$IFC GFEnabled }
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}
		end;

	lineSegment = object(obj)
			p, q: point2d;
			procedure init;	(* The global variable 'currentPointType' determines which kind of points are used. *)
			override;
			function duplicate: obj;
			override;
			procedure free;
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure windowOutput;
			override;
			procedure windowToWorld;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure textOutput;
			override;
			procedure randomChange;  (* The global variable 'currentRandomConstraint' is interpreted as a rectangle. *)
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			procedure construct (px, py, newX, newY: integer);
			override;
			procedure showObjectParameter;
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function convertToType (t: pointTypeRange): obj;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			function randomPointCover (n: longInt): obj;
			override;
			function getInfo: obj;
			override;
			function getBoundingObject: obj;
			override;
			function perimeter: extended;
			override;
			function convertToConvexPolygon: obj;
			override;
			function convertToConvexPolygonPossible: boolean;
			override;
			function HBGLessThan (o: obj): boolean;
			override;
			function verticalDistance (o: obj): extended;
			override;
			function getPointType: objectType;
			override;
			function distanceO (o: obj; var minP, minQ: obj): extended;
			override;
			function getLeftPoint: obj;
			override;
			function getRightPoint: obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function whichSideX (r: point2d): extended;
(* Determines on which side of the directed line segment (from 'p' to 'q') the point 'r' lies.			*)
(*	< 0: 'r' lies to the left of the directed line segment 'self'												*)
(*	= 0: 'r' lies on the directed line segment 'self'															*)
(*	> 0: 'r' lies to the right of the directed line segment 'self'  												*)
(* whichSideX' is also twice the signed area of the triangle with vertices 'r', 'self.p' and 'self.q'. 	*)
(* This function is similar to 'point2d.whichSideX' and introduced here for convenience.				*)

			function intersectLineSegmentX (r, l: lineSegment): boolean;
(* True iff the segment 'self' and 'r' intersect. The intersection point is 'l', where 'l.p'  'l.q'		*)
(* lexicographically ('l.p' = 'l.q' is possible and very common). 											*)

			function intersectLineX (r, l: lineSegment): boolean;
(* True iff the segment 'self' and the line defined by 'r' intersect. The intersection point is 'l',		*)
(*	where 'l.p'  'l.q' lexicographically ('l.p' = 'l.q' is possible and very common). 						*)

			function lineIntersectsLine (l: lineSegment; intersectionPoint: point2d): boolean;
(* True, iff the lines trough 'self' and 'l' intersect. The intersection point is 'intersectionPoint'. *)

			function length: extended; (* Computes the length of the segment. *)
			function lengthMP: markedRealPoint;

			procedure projectOnLine (candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
(*	'projectionPoint' becomes the projection of 'candidatePoint' onto the line 'self'.					*)
(*	'distanceToLine' becomes the distance of 'candidatePoint' to the line 'self'.						*)
(*	'whichSide' is the the result of 'whichSideX(candidatePoint)'.										*)
(*	'lambdaSign' is -1 if the projection point lies on the infinite ray emanating from 'self.p',		*)
(*	'lambdaSign' is 0 if the projection point lies on the segment 	'self' and							*)
(*	'lambdaSign' is +1 if the projection point lies on the infinite ray emanating from 'self.q'.		*)

			function distanceToPointX (candPoint: point2d; var closestX, closestY: extended): extended;
(* added by FB ,'self' and 'candPoint' have to be disjoint											 *)
(* the Euclidean distance between 'self' and 'candPoint'; closestX and closestY become the *)
(* coordinates of the point on 'self' closest to 'candPoint'    			 						 *)

			function distanceLineSegmentsX (l: lineSegment; var shortestConnection: lineSegment): extended;
(* added by FB ; 'self' and 'l' have to be disjoint													 *)
(* the Euclidean distance between 'self' and 'l'; 'closestConnection' becomes the shortest  *)
(* connecting line segment between 'self' and 'l'                      									 *)

			function collinear (l: lineSegment): boolean;
(* Tests whether 'self' and 'l' are parallel. *)

{$IFC GFEnabled }
			function convertToParameterPoint: obj;
			override;
			function convertFromParameterPoint (pP: obj): obj;
			override;
			procedure createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
			override;
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}
		end;

(* inserted by PV *)

	directedLineSegment = object(lineSegment)
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure windowOutput;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
(* the segment is drawn as an arc *)
		end;

(* *)

	circle = object(obj)
			x, y, radius: real;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure constructWorld (mouse: Point);
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure randomChange; (* The global variable 'currentRandomConstraint' is interpreted as a rectangle. *)
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			function area: extended;
			override;
			function perimeter: extended;
			override;
			procedure construct (px, py, newX, newY: integer);
			override;
			procedure showObjectParameter;
			override;
			procedure windowOutput;
			override;
			procedure windowToWorld;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			function randomPointCover (n: longInt): obj;
			override;
			function inclRectangle (bottomS, leftS, topS, rightS: extended): integer;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function getBoundingObject: obj;
			override;
			function HBGLessThan (o: obj): boolean;
			override;
			function verticalDistance (o: obj): extended;
			override;
			function getPointType: objectType;
			override;
			function distanceO (o: obj; var minP, minQ: obj): extended;
			override;
			function getLeftPoint: obj;
			override;
			function getRightPoint: obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function getCenter: point2d;
	(*	Creates the center as a point. 'currentPointType' determines which kind of point is created. *)

			function pointCoverAngle (n: longInt; a: real): obj;
	(* Like 'pointCover' but result is rotated by 'a' (radians) *)

			function getRadius: markedRealPoint;
	(* Create a 'markedRealPoint' marked with the radius. *)

			function inCircle (p: point2d): boolean;
	(* Tests whether 'p' is in the circle. *)

			function inside (p: point2d): boolean;
	(* Determines whether a given point 'p' lies inside 'self' (which could be a regular n-gon). *)

			function distanceToPerimeterX (p: point2d; var distance: extended): boolean;
	(* Computes the distance of 'p' to the perimeter of the circle in 'distance'. 'distanceToPerimeterX = true' iff	*)
	(*	the point 'p' lies inside the circle (or regularNGon).																	*)

			procedure makeFrom2Points (p, q: point2d);
	(* Creates the smallest circle through the two points 'p' and q'. *)

			procedure makeFrom3Points (p, q, r: point2d);
	(* Creates the smallest circle through the three points 'p', 'q' and 'r'. *)

			procedure makeDisk (l: list);
	(* Creates the smallest circle through the points in 'l'. *)

			procedure makeFromPointCircle (p: point2d; c: circle);
	(* Creates the smallest circle through 'p' and 'c' with 'not inCircle(p)'. *)

{$IFC GFEnabled }
			function convertToParameterPoint: obj;
			override;
			function convertFromParameterPoint (pP: obj): obj;
			override;
			procedure createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
			override;
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}
		end;

var
	constrainedRectangle: boolean; (* used for 'rectangle.construct' *)
	widthByHeight: real; (* constrains |left-right| / |top-bottom| *)

type
	rectangle = object(obj)
			left, bottom, right, top: real; (* (left  right) and (bottom  top) *)
			function isSelected (where: obj): boolean;
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			procedure randomChange; (* The global variable 'currentRandomConstraint' is interpreted as a rectangle. *)
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			function area: extended;
			override;
			function perimeter: extended;
			override;
			procedure construct (px, py, newX, newY: integer);
			override;
			procedure constructWorld (mouse: Point);
			override;
			procedure showObjectParameter;
			override;
			procedure windowOutput;
			override;
			procedure windowToWorld;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			function inclRectangle (bottomS, leftS, topS, rightS: extended): integer;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function getBoundingObject: obj;
			override;
			function convertToConvexPolygon: obj;
			override;
			function convertToConvexPolygonPossible: boolean;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
{$IFC GFEnabled }
			function convertToParameterPoint: obj;
			override;
			function convertFromParameterPoint (pP: obj): obj;
			override;
			procedure createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
			override;
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}

			procedure initFromOppositePoints (p, q: point2d);
(* initializes the rectangle from two opposite points *)

			procedure orderVertices;
(* establish the object invariant: 'left <=right' and 'bottom <= top' *)

		end;

const
	maxDimension1 = maxDimension + 1;

type
	dimensionRange1 = 1..maxDimension1;
	coordinateVector = array[dimensionRange] of real;

	dDimCircle = object(obj)
			dimension: dimensionRange; (* dimension >=2 *)
			coordinates: coordinateVector;
			radius: real;
			procedure init;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure randomChange; (* The global variable 'currentRandomConstraint' is interpreted as a rectangle. *)
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			procedure construct (px, py, newX, newY: integer);
			override;
			procedure windowOutput;
			override;
			procedure windowToWorld;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function toCircle: circle;
(* Project on 2-d circle *)

			function getRadius: markedRealPoint;

			function inCircle (p: dDimPoint): boolean;
(* Tests whether the d-dimensional point 'p' is inside the d-dimensional circle 'self'.			*)

			function inCircleEps (p: dDimPoint; tolerance: extended): boolean;
(* Tests whether the d-dimensional point 'p' is in the d-dimensional circle 'self' whose		*)
(* radius is enlarged by a factor of (1 + 'tolerance'). 												*)

			procedure makeDisk (l: list);
(*  Creates the smallest disk in d-space having the d-space points from 'l' on the boundary. 	*)

			procedure makeFromPointCircle (p: dDimPoint; c: dDimCircle);
(* Creates the smallest circle through 'p' and 'c' with 'not inCircle(p)'. 							*)

			function randomPoint: dDimPoint;
(* Creates a uniformly distributed random point inside the disk. 									*)


		end;

	dDimPoint = object(obj)
			dimension: dimensionRange; (* dimension >=2 *)
			coordinates: coordinateVector;
			procedure init;
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure randomChange; (* The global variable 'currentRandomConstraint' is interpreted as a rectangle. *)
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			procedure construct (px, py, newX, newY: integer);
			override;
			procedure windowOutput;
			override;
			procedure windowToWorld;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function toRealPoint: realPoint;
(* Project the d-dimensional point on the first two dimensions. *)

			function distance (p: dDimPoint): extended;
(* Computes the Euclidean distance between the d-dimensional points 'self' and 'p'. *)

			procedure randomInCircle (c: dDimCircle);
(* Changes the coordinates uniformly distributed inside the circle 'c'.		*)

{$IFC GFEnabled }
			function convertToParameterPoint: obj;
			override;
			function convertFromParameterPoint (pP: obj): obj;
			override;
			procedure createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
			override;
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}
		end;

{$IFC GFEnabled }
type
	fieldsT = array[dimensionRange] of obj;

	pPoint = object(obj)
			dimension: dimensionRange;
			fields: fieldsT;

			procedure init;
			override;
			function duplicate: obj;
			override;
			procedure free;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;

			procedure describe;
			override;
			procedure execute (item: integer);
			override;
		end;
{$ENDC }


function pointLessThanXY (p, q: point2d): boolean;
(* Lexicographical comparison, first X then Y. *)

function pointLessThanYX (p, q: point2d): boolean;
(* Lexicographical comparison, first Y then X. *)

function pointGreaterThanXY (p, q: point2d): boolean;

function pointGreaterThanYX (p, q: point2d): boolean;

function comparePointXY (p, q: point2d): relationObj;

function inverseComparePointXY (p, q: point2d): relationObj;

function comparePointYX (p, q: point2d): relationObj;

function inverseComparePointYX (p, q: point2d): relationObj;

procedure orderPointsGreaterThanXY (var p, q: point2d);

procedure dynamicExecution2 (obj1, obj2: obj; procedure computeResult (obj1, obj2: obj; var x: obj); highlightResult, showType: boolean);

procedure interactiveChange (title: Str255; o: obj; function select (x: obj): seqIndex0; procedure change (y: obj; s: seqIndex));

procedure initBasicGeometry;

function lexicographicLessThan (a, b: coordinateVector; dim: dimensionRange): boolean;

function inRange (a, b, x: extended): boolean;

const
userTolerance = 3;

procedure getSelectionRectangle (mouse: Point; r: rectangle);


{**********************************************}
{**********************************************}

var
pointRadius, pointRadiusHighlight: integer;

implementation

procedure interactiveChange (title: Str255; o: obj; function select (x: obj): seqIndex0; procedure change (y: obj; s: seqIndex));
	var
		selected: seqIndex0;
		oldPort, thisPort: GrafPtr;
		oldFrame, thisFrame: wFrame;
	begin
		displayInfoStringCommand(title);
		GetPort(oldPort);
		oldFrame := currentFrame;
		additionalHighlight := o;
		o.interactiveHighlight;
		selected := select(o);
		while selected > 0 do begin
			rebuildOperationsMenu := true;
			doBeep := false;
			o.interactiveOutput;
			change(o, selected);
			o.interactiveOutput;
			selected := select(o);
		end;
		additionalHighlight := nil;
		GetPort(thisPort);
		thisFrame := currentFrame;
		SetPort(oldPort);
		currentFrame := oldFrame;
		o.interactiveHighlight;
		SetPort(thisPort);
		currentFrame := thisFrame;
		displayInfoString('');
	end;

function inRange (a, b, x: extended): boolean;
	begin
		if a < b then
			inRange := (a <= x) & (x <= b)
		else
			inRange := (b <= x) & (x <= a);
	end;

function point2d.genericLessThan (o: obj): boolean;
	var
		p: point2d;
	begin
		p := point2d(o);
		genericLessThan := (getX < p.getX) | (getX = p.getX) & (getY < p.getY);
	end;

function point2d.pointCover (n: longInt): obj;
	var
		v: vector;
		i: seqIndex0;
		p: point2d;
	begin
		p := point2d(convertToType(currentPointType));
		currentVectorLength := n;
		v := vector(newInitObj(pointVectorType));
		v.append(p);
		for i := 2 to n do
			v.append(p.duplicate);
		pointCover := v;
	end;

function point2d.convertToType (t: pointTypeRange): obj;
	var
		result: point2d;
	begin
		result := point2d(newInitObj(t));
		result.setX(getX);
		result.setY(getY);
		convertToType := result;
	end;

procedure point2d.init;
	begin
		setX(0.0);
		setY(0.0);
	end;

function point2d.getX: extended;
	begin
		unimplementedMethod('point2d.getX');
		getX := 0;
	end;

function point2d.getY: extended;
	begin
		unimplementedMethod('point2d.getY');
		getY := 0;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure point2d.setX (newX: extended);
	begin
		unimplementedMethod('point2d.setX');
	end;

procedure point2d.setY (newY: extended);
	begin
		unimplementedMethod('point2d.setY');
	end;

function point2d.xLessThan (p: point2d): boolean;
	begin
		unimplementedMethod('point2d.xLessThan');
		xLessThan := false;
	end;

function point2d.xEqual (p: point2d): boolean;
	begin
		unimplementedMethod('point2d.xEqual');
		xEqual := false;
	end;

function point2d.yLessThan (p: point2d): boolean;
	begin
		unimplementedMethod('point2d.yLessThan');
		yLessThan := false;
	end;

function point2d.yEqual (p: point2d): boolean;
	begin
		unimplementedMethod('point2d.yEqual');
		yEqual := false;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure point2d.setXY (newX, newY: extended);
	begin
		setX(newX);
		setY(newY);
	end;

procedure point2d.getXY (var newX, newY: extended);
	begin
		newX := getX;
		newY := getY;
	end;

function point2d.equalPoint (p: point2d): boolean;
	begin
		equalPoint := equal(p);
	end;

function point2d.HBGLessThan (o: obj): boolean;
	begin
		HBGLessThan := yLessThan(point2d(o));
	end;

function point2d.getPointType: objectType;
	begin
		getPointType := getType;
	end;

function point2d.getLeftPoint: obj;  (* Eingefuegt von Michael Vollenbroeker *)
	begin
		getLeftPoint := duplicate;
	end;

function point2d.getRightPoint: obj;
	begin
		getRightPoint := duplicate;
	end;

procedure point2d.randomChange;
	var
		p: realPoint;
	begin
		if currentRandomConstraint <> nil then
			if member(currentRandomConstraint, rectangle) then
				with rectangle(currentRandomConstraint) do begin
					setX((right - left) * randomReal + left);
					setY((bottom - top) * randomReal + top);
				end
			else if member(currentRandomConstraint, circle) then
				with circle(currentRandomConstraint) do begin
					new(p);
					failNil(Handle(p));
					repeat
						p.x := x + radius * (2 * randomReal - 1);
						p.y := y + radius * (2 * randomReal - 1);
					until inside(p);
					setX(p.x);
					setY(p.y);
					p.free;
				end;
	end;

	const
		ProjectC = 0;
		ProjectDynamicC = 1;
		LineSegmentC = 2;
		RotateC = 3;
		CircleC = 4;
		MarkedPointC = 5;
		WhichSideC = 6;
		WhichSideSignC = 7;
		ScaleC = 8;
		EditMarkC = 9;
		MiddleC = 10;
		RemoveMarkC = 11;
		WhichSideDynamicC = 12;
		WhichSideSignDynamicC = 13;
		WhichSideThreePointsC = 14;
		WhichSideSignThreePointsC = 15;
		EuclideanDistanceC = 16;
		LpDistanceC = 17;
		Circle2C = 18;

procedure point2d.describe;
	begin
		with currentArguments do begin
			if (length = 2) & member(ith(1), point2d) & sameType2(ith(1), ith(2)) then begin
				menuEntry('Middle', MiddleC);
				menuEntry('Line Segment', LineSegmentC);
				menuEntry('Euclidean Distance', EuclideanDistanceC);
				menuEntry('Lp Distance', LpDistanceC);
				menuEntry('Circle', Circle2C);
			end;
			if (length = 2) & (member(ith(1), point2d) | member(ith(2), point2d)) then begin
				menuEntry('Rotate...', RotateC);
				menuEntry('Scale...', ScaleC);
			end;
			if (length = 3) & member(ith(1), point2d) & sameType3(ith(1), ith(2), ith(3)) then begin
				menuEntry('Which Side', WhichSideThreePointsC);
				menuEntry('Which Side Sign', WhichSideSignThreePointsC);
				if point2d(ith(1)).whichSideX(point2d(ith(2)), point2d(ith(3))) <> 0 then
					menuEntry('Circle', CircleC);
			end;
			if (length = 1) & member(ith(1), point2d) then
				menuEntry('Marked Point...', MarkedPointC);
			if (length = 1) & member(ith(1), markedRealPoint) then begin
				menuEntry('Edit Mark...', EditMarkC);
				menuEntry('Remove Mark', RemoveMarkC);
			end;
			if (length = 2) & (member(ith(1), point2d) & member(ith(2), lineSegment) & sameType2(ith(1), lineSegment(ith(2)).p) | member(ith(2), point2d) & member(ith(1), lineSegment) & sameType2(ith(2), lineSegment(ith(1)).p)) then begin
				menuEntry('Which Side', WhichSideC);
				menuEntry('Which Side Sign', WhichSideSignC);
				menuEntry('Which Side Dynamic', WhichSideDynamicC);
				menuEntry('Which Side Sign Dynamic', WhichSideSignDynamicC);
				menuEntry('Project', ProjectC);
				menuEntry('Project Dynamic', ProjectDynamicC);
			end;
		end;
	end;

procedure point2d.execute (item: integer);

	function callRotate (p: point2d; o: obj): obj;
		var
			angle: extended;
			toRotate: obj;
			x, y: extended;
		begin
			angle := 0;
			if getUserParameter1('Angle (in degrees)', 0, angle, nil) & (angle <> 0) then begin
				x := p.getX;
				y := p.getY;
				toRotate := o.duplicateAll;
				toRotate.rotate(x, y, oneDegree * angle);
				callRotate := toRotate;
			end
			else
				callRotate := nil;
		end;

	procedure computeWhichSide (p: point2d; l: lineSegment; var mp: markedRealPoint);
		var
			w: extended;
		begin
			w := l.whichSideX(p);
			new(mp);
			failNil(Handle(mp));
			mp.init;
			mp.x := p.getX;
			mp.y := p.getY;
			mp.mark.value := stringOf(w : 8 : 4);
		end;

	procedure computeWhichSideSign (p: point2d; l: lineSegment; var mp: markedRealPoint);
		begin
			new(mp);
			failNil(Handle(mp));
			mp.init;
			mp.x := p.getX;
			mp.y := p.getY;
			mp.mark.value := stringOf(p.whichSideSign(l.p, l.q) : 1);
		end;

	procedure computeProjection (p: point2d; l: lineSegment; var mp: markedRealPoint);
		var
			projectionPoint: point2d;
			d, w: extended;
			lambda: signType;
		begin
			projectionPoint := point2d(p.duplicate);
			new(mp);
			failNil(Handle(mp));
			mp.init;
			l.p.projectOnLine(l.q, p, projectionPoint, d, w, lambda);
			mp.mark.value := stringOf(d : 8 : 3, ' ', w : 8 : 3, ' ', lambda : 1);
			mp.setXY(projectionPoint.getX, projectionPoint.getY);
			projectionPoint.free;
		end;

	procedure callProjection (p: point2d; l: lineSegment);
		var
			mp: markedRealPoint;
		begin
			computeProjection(p, l, mp);
			currentResult.append(mp);
		end;

	function callScale (p: point2d; o: obj): obj;
		var
			toScale: obj;
			lambda: extended;
			x, y: extended;
		begin
			lambda := 1;
			if getUserParameter1('Lambda', 1, lambda, nil) then begin
				x := p.getX;
				y := p.getY;
				toScale := o.duplicateAll;
				toScale.translate(-x, -y);
				toScale.scale(lambda);
				toScale.translate(x, y);
				callScale := toScale;
			end
			else
				callScale := nil;
		end;

	function callMarkedPoint (p: point2d): markedRealPoint;
		var
			r: markedRealPoint;
			m: Str255;
		begin
			m := '';
			if getUserString1('Mark', m) then begin
				new(r);
				failNil(Handle(r));
				r.init;
				r.x := p.getX;
				r.y := p.getY;
				r.mark.value := m;
				callMarkedPoint := r;
			end
			else
				callMarkedPoint := nil;
		end;

	function callWhichSide (p: point2d; l: lineSegment; sign: boolean): markedRealPoint;
		var
			r: markedRealPoint;
			s: signType;
		begin
			new(r);
			failNil(Handle(r));
			r.init;
			r.x := p.getX;
			r.y := p.getY;
			if sign then begin
				s := p.whichSideSign(l.p, l.q);
				if s = +1 then
					r.mark.value := '+1'
				else
					r.mark.value := stringOf(s : 1);
			end
			else
				r.mark.value := stringOf(p.whichSideX(l.p, l.q) : 5 : 3);
			callWhichSide := r;
		end;

	function callWhichSideThreePoints (p, q, r: point2d; sign: boolean): markedRealPoint;
		var
			l: lineSegment;
		begin
			new(l);
			failNil(Handle(l));
			l.p := q;
			l.q := r;
			callWhichSideThreePoints := callWhichSide(p, l, sign);
			dispose(l);
		end;

	procedure callEditMark (p: markedRealPoint);
		var
			newMark: Str255;
		begin
			newMark := p.mark.value;
			if getUserString1('Edit Mark', newMark) & (newMark <> p.mark.value) then begin
				p.interactiveOutput;
				p.mark.value := newMark;
				p.interactiveOutput;
				rebuildOperationsMenu := true;
			end;
		end;

	function createCircle2 (p1, p2: point2d): circle;
		var
			c: circle;
		begin
			new(c);
			failNil(Handle(c));
			c.makeFrom2Points(p1, p2);
			createCircle2 := c;
		end;

	begin (* point2d.execute *)
		with currentArguments do
			case item of
				EuclideanDistanceC: 
					displayInfoString(StringOf('Euclidean Distance = ', point2d(ith(1)).distanceX(point2d(ith(2))) : 10 : 5));
				LpDistanceC: 
					displayInfoString(StringOf('Lp Distance = ', point2d(ith(1)).distanceLpX(point2d(ith(2))) : 10 : 5));
				WhichSideThreePointsC, WhichSideSignThreePointsC: 
					currentResult.append(callWhichSideThreePoints(point2d(ith(1)), point2d(ith(2)), point2d(ith(3)), item = WhichSideSignThreePointsC));
				MiddleC: 
					currentResult.append(point2d(ith(1)).middle(point2d(ith(2))));
				LineSegmentC: 
					currentResult.append(point2d(ith(1)).createLineSegment(point2d(ith(2))));
				RotateC: 
					if member(ith(1), point2d) then
						currentResult.appendNonNil(callRotate(point2d(ith(1)), ith(2)))
					else
						currentResult.appendNonNil(callRotate(point2d(ith(2)), ith(1)));
				ScaleC: 
					if member(ith(1), point2d) then
						currentResult.appendNonNil(callScale(point2d(ith(1)), ith(2)))
					else
						currentResult.appendNonNil(callScale(point2d(ith(2)), ith(1)));
				CircleC: 
					currentResult.appendNonNil(point2d(ith(1)).createCircle(point2d(ith(2)), point2d(ith(3))));
				Circle2C: 
					currentResult.appendNonNil(createCircle2(point2d(ith(1)), point2d(ith(2))));
				MarkedPointC: 
					currentResult.appendNonNil(callMarkedPoint(point2d(ith(1))));
				EditMarkC: 
					callEditMark(markedRealPoint(ith(1)));
				RemoveMarkC: 
					currentResult.append(markedRealPoint(ith(1)).toRealPoint);
				ProjectDynamicC: 
					if member(ith(1), point2d) then
						dynamicExecution2(point2d(ith(1)), lineSegment(ith(2)), computeProjection, false, true)
					else
						dynamicExecution2(point2d(ith(2)), lineSegment(ith(1)), computeProjection, false, true);
				WhichSideDynamicC: 
					if member(ith(1), point2d) then
						dynamicExecution2(point2d(ith(1)), lineSegment(ith(2)), computeWhichSide, false, true)
					else
						dynamicExecution2(point2d(ith(2)), lineSegment(ith(1)), computeWhichSide, false, true);
				WhichSideSignDynamicC: 
					if member(ith(1), point2d) then
						dynamicExecution2(point2d(ith(1)), lineSegment(ith(2)), computeWhichSideSign, false, true)
					else
						dynamicExecution2(point2d(ith(2)), lineSegment(ith(1)), computeWhichSideSign, false, true);
				WhichSideC, WhichSideSignC: 
					if member(ith(1), point2d) then
						currentResult.append(callWhichSide(point2d(ith(1)), lineSegment(ith(2)), item = WhichSideSignC))
					else
						currentResult.append(callWhichSide(point2d(ith(2)), lineSegment(ith(1)), item = WhichSideSignC));
				ProjectC: 
					if member(ith(1), point2d) then
						callProjection(point2d(ith(1)), lineSegment(ith(2)))
					else
						callProjection(point2d(ith(2)), lineSegment(ith(1)))
			end;
	end;

procedure point2d.centerOfGravity (var cx, cy: extended);
	begin
		cx := getX;
		cy := getY;
	end;

procedure point2d.textOutput;
	begin
		write('(', getX : 4 : 1, ',', getY : 4 : 1, ') ');
	end;

procedure point2d.interactiveOutput;
	begin
		drawPoint(getX, getY, pointRadius);
	end;

procedure point2d.windowOutput;
	begin
{$IFC THINK_PASCAL }
		drawOval(round(getX), round(getY), pointRadius);
{$ELSEC }
		drawOval(roundLong(getX), roundLong(getY), pointRadius);
{$ENDC }
	end;

procedure point2d.interactiveHighlight;
	begin
		drawPoint(getX, getY, pointRadiusHighlight);
	end;

function point2d.isSelected (where: obj): boolean;
	var
		x, y: real;
	begin
		x := getX;
		y := getY;
		if member(where, rectangle) then
			isSelected := (rectangle(where).left <= x) & (x <= rectangle(where).right) & (rectangle(where).bottom <= y) & (y <= rectangle(where).top)
		else
			isSelected := false;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure point2d.construct (px, py, newX, newY: integer);
	begin
		setX(newX);
		setY(newY);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure point2d.windowToWorld;
	begin
		with currentFrame do begin
			setX((getX - dx) / mx);
			setY((getY - dy) / my);
		end;
	end;

procedure point2d.translate (dx, dy: extended);
	begin
		setX(getX + dx);
		setY(getY + dy);
	end;

procedure point2d.rotate (xOrigin, yOrigin, alpha: extended);
	var
		x, y: extended;
	begin
		x := getX;
		y := getY;
		rotatePoint(xOrigin, yOrigin, alpha, x, y);
		setX(x);
		setY(y);
	end;

procedure point2d.scale (lambda: extended);
	begin
		setX(lambda * getX);
		setY(lambda * getY);
	end;

procedure point2d.reflect (a, b, c: extended);
	var
		x, y: extended;
	begin
		x := getX;
		y := getY;
		reflectPoint(a, b, c, x, y);
		setX(x);
		setY(y);
	end;

procedure point2d.drawLine (p: point2d);
	begin
		drawLineSegment(getX, getY, p.getX, p.getY);
	end;

procedure point2d.drawLineHighlight (p: point2d; highlight: boolean);
	begin
		drawLine(p);
		if highlight then begin
			PenSize(3, 3);
			drawLine(p);
			PenSize(1, 1);
		end;
	end;

procedure point2d.drawSimpleLine (p: point2d);
	begin
{$IFC THINK_PASCAL }
		MoveTo(round(getX), round(getY));
		LineTo(round(p.getX), round(p.getY));
{$ELSEC }
		MoveTo(roundLong(getX), roundLong(getY));
		LineTo(roundLong(p.getX), roundLong(p.getY));
{$ENDC }
	end;


(* inserted by PV *)

procedure point2d.drawRectangle (p: point2d);
(* Draws a rectangle with corner self left bottom and corner p right top. *)
	begin
		if equalPoint(p) then begin
	(* The box consists of one point. *)
			PenMode(patOr);
			interactiveHighlight;
			PenMode(patXor);
		end
		else
			drawRectangle2d(getX, getY, p.getX, p.getY);
	end;   { drawRectangle }

procedure point2d.drawCrossOfRectangle (p, middleP: point2d);
(* Draws a cross through the middlepoint of the rectangle with corner self left bottom and corner p right top. *)
	var
		m1, m2: realPoint;
	begin
		new(m1);
		failNil(Handle(m1));
		new(m2);
		failNil(Handle(m2));
		m1.x := middleP.getX;
		m1.y := getY;
		m2.x := middleP.getX;
		m2.y := p.getY;
		point2d(m1).drawLine(point2d(m2));
		m1.x := getX;
		m1.y := middleP.getY;
		m2.x := p.getX;
		m2.y := middleP.getY;
		point2d(m1).drawLine(point2d(m2));
		m1.free;
		m2.free
	end;   { drawCrossOfRectangle }

(* *)

procedure point2d.labelPoint (l: Str255);
	var
		ix, iy: integer;
	begin
		if worldToWindow(getX, getY, ix, iy) then begin
			MoveTo(ix, iy - 5);
{$IFC THINK_PASCAL }
			writeDraw(l);
			ExtendDrawingSheet(ix + StringWidth(l), iy - 5 - thePort^.txSize);
{$ELSEC }
			DrawString(l);
			ExtendDrawingSheet(ix + StringWidth(l), iy - 5 - qd.thePort^.txSize);
{$ENDC }
		end;
	end;

function point2d.whichSideSign (p, q: point2d): signType;
	begin
		whichSideSign := sign(whichSideX(p, q));
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function point2d.whichSideX (p, q: point2d): extended;
	begin
		unimplementedMethod('point2d.whichSideX');
		whichSideX := 0;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function point2d.leftOf (p, q: point2d): boolean; (* Eingefuegt von Michael Vollenbroeker 2.11.92 *)
	begin
		leftOf := whichSideX(p, q) < 0;
	end;

function point2d.rightOf (p, q: point2d): boolean; (* Eingefuegt von Michael Vollenbroeker 2.11.92 *)
	begin
		rightOf := whichSideX(p, q) > 0;
	end;

function point2d.onLine (p, q: point2d): boolean; (* Eingefuegt von Michael Vollenbroeker 20.11.92 *)
	begin
		onLine := whichSideX(p, q) = 0;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function point2d.crossProductX (p, q, r: point2d): extended;
	begin
		unimplementedMethod('point2d.crossProductX');
		crossProductX := 0;
	end;

function point2d.scalarProductX (p: point2d): extended; (* added by FB *)
	begin
		unimplementedMethod('point2d.scalarProductX');
		scalarProductX := 0;
	end;

function point2d.distanceX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.distanceX');
		distanceX := 0;
	end;

function point2d.squaredDxX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.squaredDxX');
		squaredDxX := 0;
	end;

function point2d.squaredDyX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.squaredDyX');
		squaredDyX := 0;
	end;

function point2d.squaredDistanceX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.squaredDistanceX');
		squaredDistanceX := 0;
	end;

function point2d.distanceLpX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.distanceLpX');
		distanceLpX := 0;
	end;

function point2d.distancePLpX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.distancePLpX');
		distancePLpX := 0;
	end;

function point2d.dxPLpX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.dxPLpX');
		dxPLpX := 0;
	end;

function point2d.dyPLpX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.dyPLpX');
		dyPLpX := 0;
	end;

procedure point2d.circleCenterX (p, q: point2d; var cx, cy: extended);
	begin
		unimplementedMethod('point2d.circleCenterX');
		cx := 0;
		cy := 0;
	end;

function point2d.intersectLineSegmentX (q, r, s, left, right: point2d): boolean;
	begin
		unimplementedMethod('point2d.intersectLineSegmentX');
		intersectLineSegmentX := false;
	end;

function point2d.intersectLineX (q, r, s, left, right: point2d): boolean;
	begin
		unimplementedMethod('point2d.intersectLineX');
		intersectLineX := false;
	end;

function point2d.lineIntersectsLine (q, r, s, intersectionPoint: point2d): boolean;
	begin
		unimplementedMethod('point2d.lineIntersectsLine');
		lineIntersectsLine := false;
	end;

procedure point2d.projectOnLine (endPoint, candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
	begin
		unimplementedMethod('point2d.projectOnLine');
		distanceToLine := 0;
		whichSide := 0;
		lambdaSign := 0;
	end;

function point2d.distanceToLineSegmentX (p, q: point2d; var closestX, closestY: extended): extended; (* added by FB *)
	begin
		unimplementedMethod('point2d.distanceToLineSegmentX');
		closestX := 0;
		closestY := 0;
		distanceToLineSegmentX := 0;
	end;

function point2d.middle (p: point2d): point2d;
	begin
		unimplementedMethod('point2d.middle');
		middle := nil;
	end;

function point2d.rayEvalX (p: point2d; cosSlope, sinSlope: extended): extended;
	begin
		unimplementedMethod('point2d.rayEvalX');
		rayEvalX := 0;
	end;

function point2d.bisectorEvalX (p, q: point2d): extended;
	begin
		unimplementedMethod('point2d.bisectorEvalX');
		bisectorEvalX := 0;
	end;

function point2d.intersectRayBisectorX (cosSlope, sinSlope: extended; p, q, intersection: point2d): boolean;
	begin
		unimplementedMethod('point2d.intersectRayBisectorX');
		intersectRayBisectorX := false;
	end;

function point2d.intersectBisectorBisectorX (p, q, r, intersection: point2d): boolean;
	begin
		unimplementedMethod('point2d.intersectBisectorBisectorX');
		intersectBisectorBisectorX := false;
	end;

function point2d.intersectRayRayX (cosLower, sinLower: extended; p: point2d; cosUpper, sinUpper: extended; intersection: point2d): boolean;
	begin
		unimplementedMethod('point2d.intersectRayRayX');
		intersectRayRayX := false;
	end;

function point2d.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 1;
	end;
{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }

function point2d.orthogonalProjection (endPoint, candidatePoint, projectionPoint: point2d): boolean;
	var
		d, w: extended;
		l: signType;
	begin
		projectOnLine(endPoint, candidatePoint, projectionPoint, d, w, l);
		orthogonalProjection := l = 0;
	end;

function point2d.xPlusYX: extended;
	begin
		unimplementedMethod('point2d.xPlusYX');
		xPlusYX := 0;
	end;

function point2d.xMinusYX: extended;
	begin
		unimplementedMethod('point2d.xMinusYX');
		xMinusYX := 0;
	end;

function point2d.createLineSegment (p: point2d): lineSegment;
	var
		l: lineSegment;
	begin
		new(l);
		failNil(Handle(l));
		l.p := point2d(duplicate);
		l.q := point2d(p.duplicate);
		createLineSegment := l;
	end;

(* inserted by PV *)

function point2d.createDirectedLineSegment (p: point2d): directedLineSegment;
	var
		l: directedLineSegment;
	begin
		new(l);
		failNil(Handle(l));
		l.p := point2d(duplicate);
		l.q := point2d(p.duplicate);
		createDirectedLineSegment := l;
	end;

(* *)

function point2d.createCircle (p, q: point2d): circle;
	var
		cx, cy: extended;
		c: circle;
	begin
		if whichSideX(p, q) = 0 then
			createCircle := nil
		else begin
			circleCenterX(p, q, cx, cy);
			new(c);
			failNil(Handle(c));
			c.x := cx;
			c.y := cy;
			c.radius := sqrt(sqr(getX - cx) + sqr(getY - cy));
			createCircle := c;
		end;
	end;

(* -------------------- inserted by JM  ------------------------------------------------------------ *)

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function point2d.DxX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.DxX');
		DxX := 0;
	end;

function point2d.DyX (p: point2d): extended;
	begin
		unimplementedMethod('point2d.DyX');
		DyX := 0;
	end;

procedure point2d.setPoint (newp: point2d);
	begin
		unimplementedMethod('point2d.setPoint');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

(* ------------------------------------------------------------------------------------------------------- *)


function point2d.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(2);
		appendObjInfoExtended(v, 'X-Coordinate', getX, false);
		appendObjInfoExtended(v, 'Y-Coordinate', getY, false);
		getInfo := v;
	end;

procedure point2d.setInfo (o: obj);
	var
		v: vector;
	begin
		v := vector(o);
		setX(objInfoExtended(v.elements^^[1]).e);
		setY(objInfoExtended(v.elements^^[2]).e);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure int.interactiveInput (px, py: integer);
	begin
		write('Value = ');
		readLn(value);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure int.interactiveOutput;
	begin
		write(' ', value : 1);
	end;

procedure int.textOutput;
	begin
		interactiveOutput;
	end;

procedure int.textFileInput;
	begin
		readTextLongInt(value);
	end;

procedure int.textFileOutput;
	begin
		writeTextLongInt(value);
	end;

procedure int.fileInput;
	begin
		readLongInt(value);
	end;

procedure int.fileOutput;
	begin
		writeLongInt(value);
	end;

function int.getType: objectType;
	begin
		getType := intType;
	end;

function int.getTextType: Str255;
	begin
		getTextType := 'Int';
	end;

procedure int.randomChange;
	begin
		if currentRandomConstraint <> nil then
			if member(currentRandomConstraint, int) then
				value := randomLongInt mod int(currentRandomConstraint).value
			else
				value := randomLongInt;
	end;

function int.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, int) & (int(o).value = value);
	end;

function int.memberType (o: obj): boolean;
	begin
		memberType := member(o, int);
	end;

function realNumber.getTextType: Str255;
	begin
		getTextType := 'Real';
	end;

function realNumber.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(1);
		appendObjInfoExtended(v, 'Value', value, false);
		getInfo := v;
	end;

procedure realNumber.setInfo (o: obj);
	begin
		value := objInfoExtended(vector(o).elements^^[1]).e;
	end;

procedure sequenceIndex.setvalue (n: seqIndex);  (* Eingefuegt von Michael Vollenbroeker 30.11.92 *)
	begin
		value := n;
	end;

function sequenceIndex.getvalue: seqIndex;  (* Eingefuegt von Michael Vollenbroeker 30.11.92 *)
	begin
		getvalue := value;
	end;

procedure strings.textFileInput;
	begin
		readTextString(value);
	end;

procedure strings.textFileOutput;
	begin
		writeTextString(value);
	end;

procedure strings.fileInput;
	begin
		readStr255(value);
	end;

procedure strings.fileOutput;
	begin
		writeStr255(value);
	end;

function strings.getType: objectType;
	begin
		getType := stringsType;
	end;

function strings.getTextType: Str255;
	begin
		getTextType := StringOf('String(', length(value) : 1, ')');
	end;

function strings.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, strings) & (strings(o).value = value);
	end;

function strings.memberType (o: obj): boolean;
	begin
		memberType := member(o, strings);
	end;

function strings.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(1);
		appendObjInfoString(v, 'Value', value, false);
		getInfo := v;
	end;

procedure strings.setInfo (o: obj);
	begin
		value := objInfoString(vector(o).elements^^[1]).s;
	end;

function pointLessThanXY (p, q: point2d): boolean;
	begin
		pointLessThanXY := p.xLessThan(q) | p.xEqual(q) & p.yLessThan(q);
	end;

function pointLessThanYX (p, q: point2d): boolean;
	begin
		pointLessThanYX := p.yLessThan(q) | p.yEqual(q) & p.xLessThan(q);
	end;

function pointGreaterThanXY (p, q: point2d): boolean;
	begin
		pointGreaterThanXY := q.xLessThan(p) | q.xEqual(p) & q.yLessThan(p);
	end;

function pointGreaterThanYX (p, q: point2d): boolean;
	begin
		pointGreaterThanYX := q.yLessThan(p) | q.yEqual(p) & q.xLessThan(p);
	end;

procedure orderPointsGreaterThanXY (var p, q: point2d);
	var
		t: point2d;
	begin
		if pointGreaterThanXY(p, q) then begin
			t := p;
			p := q;
			q := t;
		end;
	end;

function comparePointXY (p, q: point2d): relationObj;
	begin
		if p.xLessThan(q) then
			comparePointXY := lessThanObj
		else if q.xLessThan(p) then
			comparePointXY := greaterThanObj
		else if p.yLessThan(q) then
			comparePointXY := lessThanObj
		else if q.yLessThan(p) then
			comparePointXY := greaterThanObj
		else
			comparePointXY := equalObj;
	end;

function inverseComparePointXY (p, q: point2d): relationObj;
	begin
		if p.xLessThan(q) then
			inverseComparePointXY := greaterThanObj
		else if q.xLessThan(p) then
			inverseComparePointXY := lessThanObj
		else if p.yLessThan(q) then
			inverseComparePointXY := greaterThanObj
		else if q.yLessThan(p) then
			inverseComparePointXY := lessThanObj
		else
			inverseComparePointXY := equalObj;
	end;

function comparePointYX (p, q: point2d): relationObj;
	begin
		if p.yLessThan(q) then
			comparePointYX := lessThanObj
		else if q.yLessThan(p) then
			comparePointYX := greaterThanObj
		else if p.xLessThan(q) then
			comparePointYX := lessThanObj
		else if q.xLessThan(p) then
			comparePointYX := greaterThanObj
		else
			comparePointYX := equalObj;
	end;

function inverseComparePointYX (p, q: point2d): relationObj;
	begin
		if p.yLessThan(q) then
			inverseComparePointYX := greaterThanObj
		else if q.yLessThan(p) then
			inverseComparePointYX := lessThanObj
		else if p.xLessThan(q) then
			inverseComparePointYX := greaterThanObj
		else if q.xLessThan(p) then
			inverseComparePointYX := lessThanObj
		else
			inverseComparePointYX := equalObj;
	end;

procedure dynamicExecution2 (obj1, obj2: obj; procedure computeResult (obj1, obj2: obj; var x: obj); highlightResult, showType: boolean);
	var
		v: vector;

	function doComputeResult (v: vector): obj;
		var
			result: obj;
		begin
			computeResult(v.elements^^[1], v.elements^^[2], result);
			doComputeResult := result;
		end;

	begin (* dynamicExecution2 *)
		v := newVector(2);
		v.append(obj1);
		v.append(obj2);
		v.dynamicExecution(doComputeResult, highlightResult, showType);
		v.free;
	end;

procedure getSelectionRectangle (mouse: Point; r: rectangle);
	begin
		r.left := mouse.h - userTolerance;
		r.right := mouse.h + userTolerance;
		r.bottom := mouse.v - userTolerance;
		r.top := mouse.v + userTolerance;
		r.windowToWorld;
	end;

procedure sequence.dynamicExecutionHighlightShowType (function computeResult (objects: sequence): obj);
	begin
		dynamicExecution(computeResult, true, true);
	end;

procedure sequence.dynamicExecutionNoHighlightShowType (function computeResult (objects: sequence): obj);
	begin
		dynamicExecution(computeResult, false, true);
	end;

procedure initBasicGeometry;
	var
		m: markedRealPoint;
		r: realPoint;
		x: rectangle;
		c: circle;
		l: lineSegment;
		f: floatPoint;
		li: longIntPoint;
		s: strings;
		i: int;
		d: dDimPoint;
		dl: directedLineSegment;
		dd: dDimCircle;
		idr: idRealPoint;
	begin
		new(r);
		failNil(Handle(r));
		prototypes[realPointType] := r;
		new(m);
		failNil(Handle(m));
		prototypes[markedRealPointType] := m;
		new(r);
		failNil(Handle(r));
		prototypes[point2dType] := r;
		new(x);
		failNil(Handle(x));
		prototypes[rectangleType] := x;
		new(c);
		failNil(Handle(c));
		prototypes[circleType] := c;
		new(l);
		failNil(Handle(l));
		prototypes[lineSegmentType] := l;
		new(f);
		failNil(Handle(f));
		prototypes[floatPointType] := f;
		new(li);
		failNil(Handle(li));
		prototypes[longIntPointType] := li;
		new(s);
		failNil(Handle(s));
		prototypes[stringsType] := s;
		new(d);
		failNil(Handle(d));
		prototypes[dDimPointType] := d;
		new(i);
		failNil(Handle(i));
		prototypes[intType] := i;
		new(dl);
		failNil(Handle(dl));
		prototypes[directedLineSegmentType] := dl;
		new(dd);
		failNil(Handle(dd));
		prototypes[dDimCircleType] := dd;
		new(idr);
		failNil(Handle(idr));
		prototypes[idRealPointType] := idr;
		constrainedRectangle := false;
	end;

function lexicographicLessThan (a, b: coordinateVector; dim: dimensionRange): boolean;
	var
		i: dimensionRange1;
	begin
		i := 1;
		while (i <= dim) & (a[i] = b[i]) do
			i := i + 1;
		lexicographicLessThan := (i <= dim) & (a[i] < b[i]);
	end;

end.
