(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit browse;

interface

{$IFC NetWorkEnabled }
uses
	TCPStreams, 
	pointVector;
{$ELSEC }
uses
	pointVector;
{$ENDC }

type
	browseStatus = (browseOk, browseCancel, browseUp, browsePrevious, browseNext);

var
	browseChange: boolean;

procedure enterBrowse;
function browse (infoVector: vector; defaultProtection, nextPossible, prevPossible: boolean): browseStatus;
procedure exitBrowse;

implementation

{$IFC NOT THINK_PASCAL }
uses Lists, Fonts;
{$ENDC }

const
	maxBrowseListLength = 550;
	lineHeight = 15;
	browseListLeft = 56;
	browseListRight = 446 - scrollBarWidth;
	browseListBottom = 147;
	browseListTop = 12;
	browseDialogRes = 7274;
	okItem = 1;
	setItem = 2;
	cancelItem = 3;
	browseListItem = 4;
	textItem = 5;
	downItem = 6;
	upItem = 7;
	nextItem = 8;
	doubleClickItem = 10;
	prevItem = 9;
	column1Width = 15; (* in characters *)
	column2Width = 28;
	column3Width = 15;
	longFlash = 15;

var
	currentList: ListHandle;
	listDialog: DialogPtr;

function formatExtended (x: extended): string;
	begin
		if trunc(x) = x then
{$IFC THINK_PASCAL }
			formatExtended := stringOf(trunc(x) : 1)
{$ELSEC }
			formatExtended := stringOf(truncLong(x) : 1)
{$ENDC }
		else
			formatExtended := stringOf(x : 1 : 8);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure doBrowseList (theWindow: WindowPtr; itemNo: integer);
	var
		r: Rect;
	begin
		SetRect(r, browseListLeft - 1, browseListTop - 1, browseListRight + 1, browseListBottom + 1);
{$IFC THINK_PASCAL }
		FillRect(r, white);
{$ELSEC }
		FillRect(r, qd.white);
{$ENDC }
		FrameRect(r);
		LUpdate(theWindow^.visRgn, currentList);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure disposeBrowseList;
	begin
		if currentList <> nil then begin
			LDispose(currentList);
			currentList := nil;
		end;
	end;
	
procedure enterBrowse;
	begin
{$IFC THINK_PASCAL }
		SetDAFont(monaco);
		listDialog := GetNewDialog(browseDialogRes, nil, WindowPtr(-1));
		SetDAFont(systemFont);
{$ELSEC }
		SetDialogFont(kFontIDMonaco);
		listDialog := GetNewDialog(browseDialogRes, nil, WindowPtr(-1));
		SetDialogFont(systemFont);
{$ENDC }
		niceWindowPosition(WindowPtr(listDialog), false);
		SelectWindow(listDialog);
		ShowWindow(listDialog);
		showDefaultButton(listDialog, okItem);
		SetPort(listDialog);
		SetDProc(listDialog, browseListItem, @doBrowseList);
		currentList := nil;
		browseChange := false;
		SetDText(listDialog, textItem, '');
		ActivateItem(listDialog, setItem, false);
		beginCursor(arrowCursor);
	end; (* enterBrowse *)

procedure exitBrowse;
	begin
		endCursor;
		HideWindow(listDialog);
{$IFC THINK_PASCAL }
		DisposDialog(listDialog);
{$ELSEC }
		DisposeDialog(listDialog);
{$ENDC }
		disposeBrowseList;
	end; (* exitBrowse *)

{$IFC NOT THINK_PASCAL }
{$PRAGMAC global_optimizer on }
{$ENDC }
function browse (infoVector: vector; defaultProtection, nextPossible, prevPossible: boolean): browseStatus;
	var
		item, oldSelection: integer;
		theEvent: EventRecord;
		selectedCell: Point;
		dummy, continue, noNext: boolean;
		newDescVect: vector;
		rView: Rect;
		status: browseStatus;
		o: objInfoElement;

	function fill (s: Str255; l: integer): Str255;
		begin
			if length(s) > l then begin
				s := copy(s, 1, l);
				s[length(s)] := '';
			end
			else if length(s) > 0 then
				while length(s) < l do
					s := concat(s, ' ');
			fill := s;
		end;

	procedure setCell (i: integer; o: objInfoElement; truncate: longInt);
		var
			column2, column3: Str255;
			cell: Point;
		begin
			if truncate > 0 then
				o.cellInfo := StringOf(' and ', truncate : 1, ' more elements ')
			else if member(o, objInfoObj) then
				o.cellInfo := concat(fill(o.desc, column1Width), ' ', fill(objInfoObj(o).o.getTextType, column2Width + column3Width))
			else begin
				if member(o, objInfoExtended) then begin
					column2 := 'Extended';
					column3 := formatExtended(objInfoExtended(o).e);
				end
				else if member(o, objInfoString) then begin
					column2 := 'String';
					column3 := objInfoString(o).s;
				end;
				o.cellInfo := concat(fill(o.desc, column1Width), ' ', fill(column2, column2Width), ' ', fill(column3, column3Width));
			end;
			SetPt(cell, 0, i);
			LSetCell(pointer(ord(@o.cellInfo) + 1), length(o.cellInfo), cell, currentList);
		end;

	procedure createBrowseList;
		var
			dataBounds: Rect;
			csize: Point;
			i, elementsInList, listSize: integer;
		begin
			if infoVector.length <= maxBrowseListLength then begin
				elementsInList := infoVector.length;
				listSize := elementsInList;
			end
			else begin
				elementsInList := maxBrowseListLength;
				listSize := maxBrowseListLength + 1;
			end;
			SetRect(dataBounds, 0, 0, 1, listSize);
			SetPt(csize, browseListRight - browseListLeft, lineHeight);
			disposeBrowseList;
			currentList := LNew(rView, dataBounds, csize, 0, listDialog, false, false, false, true);
			if infoVector <> nil then
				for i := 1 to elementsInList do
					setCell(i - 1, objInfoElement(infoVector.elements^^[i]), 0);
			if listSize > elementsInList then
				setCell(elementsInList, objInfoElement(infoVector.elements^^[listSize]), infoVector.length - elementsInList);
{$IFC THINK_PASCAL }
			LDoDraw(true, currentList);
{$ELSEC }
			LSetDrawingMode(true, currentList);
{$ENDC }
			InvalRect(rView);
		end; (* createBrowseList *)

	begin
		SetRect(rView, browseListLeft, browseListTop, browseListRight, browseListBottom);
		continue := true;
		createBrowseList;
		ActivateItem(listDialog, downItem, false);
		ActivateItem(listDialog, upItem, true);
		ActivateItem(listDialog, nextItem, nextPossible);
		ActivateItem(listDialog, prevItem, prevPossible);

		while continue do begin
{$IFC THINK_PASCAL }
			ModalDialog(@dragDialogFilter, item);
{$ELSEC }
			ModalDialog(NewModalFilterProc(@dragDialogFilter), item);
{$ENDC }
{$IFC XYZApplication }
(* sonst Probleme mit THINK Pascal windows ! *)
			while CheckUpdate(theEvent) do
				DoEvent(theEvent);
{$ENDC }
			if item = browseListItem then begin
				dummy := GetNextEvent(mDownMask, theEvent);
				GlobalToLocal(theEvent.where);
				if LClick(theEvent.where, theEvent.modifiers, currentList) then
					item := doubleClickItem;
				oldSelection := selectedCell.v;
				selectedCell.v := 0;
				selectedCell.h := 0;
				if not LGetSelect(true, selectedCell, currentList) then begin
					selectedCell.v := -1; (* no cell is selected *)
					selectedCell.h := -1;
				end;
				ActivateItem(listDialog, downItem, (selectedCell.v >= 0) & (selectedCell.v < maxBrowseListLength));
				if selectedCell.v <> oldSelection then begin
					SetDText(listDialog, textItem, '');
					ActivateItem(listDialog, setItem, false);
				end;
			end;

			if item = downItem then begin
				item := doubleClickItem
			end;

			repeat
				noNext := true;
				if (item = doubleClickItem) & (selectedCell.v >= 0) & (selectedCell.v < maxBrowseListLength) then begin
					o := objInfoElement(infoVector.elements^^[selectedCell.v + 1]);
					if member(o, objInfoExtended) then begin
						setAndHighlightX(listDialog, textItem, objInfoExtended(o).e, 15);
						ActivateItem(listDialog, setItem, not (o.protect | defaultProtection));
					end
					else if member(o, objInfoString) then begin
						SetDText(listDialog, textItem, objInfoString(o).s);
{$IFC THINK_PASCAL }
						SelIText(listDialog, textItem, 0, 32767);
{$ELSEC }
						SelectDialogItemText(listDialog, textItem, 0, 32767);
{$ENDC }
						ActivateItem(listDialog, setItem, not (o.protect | defaultProtection));
					end
					else if member(o, objInfoObj) then begin
						SetDText(listDialog, textItem, '');
						newDescVect := vector(objInfoObj(o).o.getInfo);
						if newDescVect <> nil then begin
							status := browse(newDescVect, o.protect | defaultProtection, (selectedCell.v + 1 < infoVector.length) & (selectedCell.v + 1 < maxBrowseListLength) & member(infoVector.elements^^[selectedCell.v + 2], objInfoObj), (selectedCell.v > 0) & member(infoVector.elements^^[selectedCell.v], objInfoObj));
							if status <> browseCancel then
								objInfoObj(o).o.setInfo(newDescVect);
							newDescVect.freeAll;
							case status of
								browseOk:  begin
									continue := false;
									browse := browseOk;
								end;
								browseCancel:  begin
									continue := false;
									browse := browseCancel;
								end;
								browseUp:  begin
									createBrowseList;
									LSetSelect(true, selectedCell, currentList);
									LAutoScroll(currentList);
									SetDText(listDialog, textItem, '');
									ActivateItem(listDialog, setItem, false);
									ActivateItem(listDialog, downItem, true);
									ActivateItem(listDialog, upItem, true);
									ActivateItem(listDialog, nextItem, nextPossible);
									ActivateItem(listDialog, prevItem, prevPossible);
									InvalRect(rView); (* redraw *)
								end;
								browseNext:  begin
									noNext := false;
									selectedCell.v := selectedCell.v + 1;
								end;
								browsePrevious:  begin
									noNext := false;
									selectedCell.v := selectedCell.v - 1;
								end;
							end; (* case *)
						end;
					end;
				end; (* doubleClick *)
			until noNext;

			case item of
				setItem:  begin
					o := objInfoElement(infoVector.elements^^[selectedCell.v + 1]);
					if member(o, objInfoExtended) then begin
{$IFC THINK_PASCAL }
						dummy := getValue(listDialog, textItem, objInfoExtended(o).e, objInfoExtended(o).e, noConstraint);
{$ELSEC }
						dummy := getValue(listDialog, textItem, objInfoExtended(o).e, objInfoExtended(o).e, noConstraint);
{$ENDC }
						setCell(selectedCell.v, o, 0);
						browseChange := true;
					end
					else if member(o, objInfoString) then begin
						GetDText(listDialog, textItem, Str255(objInfoString(o).s));
						setCell(selectedCell.v, o, 0);
						browseChange := true;
					end;
				end;
				cancelItem:  begin
					browse := browseCancel;
					continue := false;
				end;
				okItem:  begin
					browse := browseOk;
					continue := false;
				end;
				upItem:  begin
					browse := browseUp;
					continue := false;
				end;
				nextItem:  begin
					browse := browseNext;
					continue := false;
{$IFC THINK_PASCAL }
					LDoDraw(false, currentList);
{$ELSEC }
					LSetDrawingMode(false, currentList);
{$ENDC }
				end;
				prevItem:  begin
					browse := browsePrevious;
					continue := false;
{$IFC THINK_PASCAL }
					LDoDraw(false, currentList);
{$ELSEC }
					LSetDrawingMode(false, currentList);
{$ENDC }
				end;
				otherwise
			end;

		end; (* while *)

		disposeBrowseList;
	end; (* browse *)

end.
