(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit circle;

interface

uses
{$IFC GFEnabled }
	GFValues, 
{$ENDC}
	pointVector;

implementation

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function circle.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 3;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function circle.area: extended;
	begin
		area := sqr(radius) * pi;
	end;

function circle.perimeter: extended;
	begin
		perimeter := 2 * radius * pi;
	end;

function circle.pointCoverAngle (n: longInt; a: real): obj;
	var
		result: vector;
		i: seqIndex0;
		p: point2d;
		alpha, cosAlpha, sinAlpha, cosSum, sinSum, newCosSum: extended;
	begin
		result := newPointVector(n);
		alpha := 2 * pi / n;
		cosAlpha := cos(alpha);
		sinAlpha := sin(alpha);
		cosSum := cos(a);
		sinSum := sin(a);
		for i := n downto 1 do begin
{$IFC AnimationEnabled }
			advanceStopWatchHands;
{$ENDC }
			p := point2d(newObj(currentPointType));
			p.setX(x + radius * cosSum);
			p.setY(y + radius * sinSum);
			result.elements^^[i] := p;
			newCosSum := cosAlpha * cosSum - sinAlpha * sinSum;
			sinSum := sinAlpha * cosSum + cosAlpha * sinSum;
			cosSum := newCosSum;
		end;
		result.length := n;
		pointCoverAngle := result;
	end;

procedure circle.centerOfGravity (var cx, cy: extended);
	begin
		cx := x;
		cy := y;
	end;

function circle.pointCover (n: longInt): obj;
	begin
		pointCover := pointCoverAngle(n, 0);
	end;

function circle.genericLessThan (o: obj): boolean;
	var
		c: circle;
	begin
		c := circle(o);
		genericLessThan := (x < c.x) | (x = c.x) & ((y < c.y) | (y = c.y) & (radius < c.radius));
	end;

procedure circle.construct (px, py, newX, newY: integer);
	var
		dx, dy: real;
	begin
		x := px;
		y := py;
		dx := newX - px;
		dy := newY - py;
		radius := round(sqrt(sqr(dx) + sqr(dy)));
	end;

function circle.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	begin
		GetKeys(globalKeyMap);
		if globalKeyMap[optionKeyValue] & isSelected(where) then begin
			dragVertex := true;
			dragVertexWorld(mouse, highlight);
		end
		else
			dragVertex := false;
	end;

procedure circle.constructWorld (mouse: Point);
	var
		newX, newY: extended;
	begin
		windToWorld(mouse.h, mouse.v, newX, newY);
		radius := sqrt(sqr(newX - x) + sqr(newY - y));
	end;

procedure circle.showObjectParameter;
	begin
		displayInfoString(StringOf('Radius = ', radius : 8 : 3, '.'));
	end;

procedure circle.windowToWorld;
	begin
		with currentFrame do begin
			x := (x - dx) / mx;
			y := (y - dy) / my;
			radius := 2 * radius / (mx + my);
		end;
	end;

function circle.isSelected (where: obj): boolean;
	var
		r2, dlb, dlt, drb, drt: real;
		enclosed: boolean;
	function d (x1, y1, x2, y2: real): real;
		begin
			d := sqr(x1 - x2) + sqr(y1 - y2);
		end;
	function pair (a, b: real): boolean;
		begin
			pair := (a > r2) & (b < r2) | (a < r2) & (b > r2);
		end;
	begin
		if member(where, rectangle) then begin
			with rectangle(where) do begin
				r2 := sqr(radius);
				dlb := d(x, y, left, bottom);
				dlt := d(x, y, left, top);
				drb := d(x, y, right, bottom);
				drt := d(x, y, right, top);
				enclosed := (left <= x - radius) & (x + radius <= right) & (bottom <= y - radius) & (y + radius <= top);
			end;
			isSelected := enclosed | pair(dlb, dlt) | pair(dlt, drt) | pair(drt, drb) | pair(drb, dlb);
		end
		else
			isSelected := false;
	end;

procedure circle.windowOutput;
	begin
{$IFC THINK_PASCAL }
		FrameArc(round(y - radius), round(x - radius), round(y + radius), round(x + radius), 0, 360);
{$ELSEC }
		FrameArc2(roundLong(y - radius), roundLong(x - radius), roundLong(y + radius), roundLong(x + radius), 0, 360);
{$ENDC }
	end;

procedure circle.interactiveOutput;
	begin
		if radius > 0 then
			drawCircle(x, y, radius, 0, 360)
		else
			drawPoint(x, y, pointRadius);
	end;

procedure circle.interactiveHighlight;
	begin
		if radius > 0 then
			inherited interactiveHighlight
		else
			drawPoint(x, y, pointRadiusHighlight);
	end;

procedure circle.textFileInput;
	begin
		readTextReal(x);
		readTextReal(y);
		readTextReal(radius);
	end;

procedure circle.textFileOutput;
	begin
		writeTextReal(x);
		writeTextReal(y);
		writeTextReal(radius);
	end;

procedure circle.fileInput;
	begin
		readReal(x);
		readReal(y);
		readReal(radius);
	end;

procedure circle.fileOutput;
	begin
		writeReal(x);
		writeReal(y);
		writeReal(radius);
	end;

function circle.distanceToPerimeterX (p: point2d; var distance: extended): boolean;
	var
		d: extended;
	begin
		d := radius - sqrt(sqr(p.getX - x) + sqr(p.getY - y));
		if d > 0 then begin
			distanceToPerimeterX := true;
			distance := d;
		end
		else begin
			distanceToPerimeterX := false;
			distance := -d;
		end;
	end;

procedure circle.randomChange;
	var
		p: realPoint;
		c: circle;
		d, m: extended;
		dummy: boolean;
	begin
		if currentRandomConstraint <> nil then
			if member(currentRandomConstraint, rectangle) then
				with rectangle(currentRandomConstraint) do begin
					m := min(right - left, top - bottom);
					repeat
						x := (right - left) * randomReal + left;
						y := (bottom - top) * randomReal + top;
						radius := m * randomReal;
					until (x - radius > left) & (x + radius < right) & (y - radius > bottom) & (y + radius < top);
				end
			else if member(currentRandomConstraint, circle) then begin
				c := circle(currentRandomConstraint);
				new(p);
				failNil(Handle(p));
				repeat
					p.randomChange;
					radius := c.radius * randomReal;
					dummy := c.distanceToPerimeterX(p, d);
				until radius < d;
				x := p.x;
				y := p.y;
				p.free;
			end;
	end;

function circle.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, circle) & (circle(o).x = x) & (circle(o).y = y) & (circle(o).radius = radius);
	end;

function circle.memberType (o: obj): boolean;
	begin
		memberType := member(o, circle);
	end;

procedure circle.translate (dx, dy: extended);
	begin
		x := x + dx;
		y := y + dy;
	end;

procedure circle.rotate (xOrigin, yOrigin, alpha: extended);
	var
		xx, yy: extended;
	begin
		xx := x;
		yy := y;
		rotatePoint(xOrigin, yOrigin, alpha, xx, yy);
		x := xx;
		y := yy;
	end;

procedure circle.scale (lambda: extended);
	begin
		radius := lambda * radius;
	end;

procedure circle.reflect (a, b, c: extended);
	var
		xx, yy: extended;
	begin
		xx := x;
		yy := y;
		reflectPoint(a, b, c, xx, yy);
		x := xx;
		y := yy;
	end;

(* following function added by "Von0auf100", Thema QuadTrees *)
function circle.inclRectangle (bottomS, leftS, topS, rightS: extended): integer;
	const
		all = 1;
		part = 0;
		no = -1;
	var
		sqrRadius: extended;
	begin (* circle.inclRectangle *)
		sqrRadius := radius * radius;
		if (y <= bottomS - radius) | (topS + radius <= y) | (x <= leftS - radius) | (rightS + radius <= x) | (x < leftS) & ((y < bottomS) & (sqr(x - leftS) + sqr(y - bottomS) >= sqrRadius) | (topS < y) & (sqr(x - leftS) + sqr(y - topS) >= sqrRadius)) | (rightS < x) & ((y < bottomS) & (sqr(x - rightS) + sqr(y - bottomS) >= sqrRadius) | (topS < y) & (sqr(x - rightS) + sqr(y - topS) >= sqrRadius)) then
			inclRectangle := no
		else if (sqr(x - leftS) + sqr(y - bottomS) <= sqrRadius) & (sqr(x - rightS) + sqr(y - bottomS) <= sqrRadius) & (sqr(x - rightS) + sqr(y - topS) <= sqrRadius) & (sqr(x - leftS) + sqr(y - topS) <= sqrRadius) then
			inclRectangle := all
		else
			inclRectangle := part
	end; (* circle.inclRectangle *)
(* end of added function *)

function circle.getType: objectType;
	begin
		getType := circleType;
	end;

function circle.getTextType: Str255;
	begin
		getTextType := 'Circle';
	end;

function circle.getRadius: markedRealPoint;
	var
		mp: markedRealPoint;
	begin
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.mark.value := stringOf(radius : 5 : 3);
		mp.x := x;
		mp.y := y;
		getRadius := mp;
	end;

function circle.inCircle (p: point2d): boolean;
	begin
		inCircle := (sqr(x - p.getX) + sqr(y - p.getY) <= sqr(radius));
	end;

function circle.inside (p: point2d): boolean;
	begin
		inside := inCircle(p);
	end;

procedure circle.makeFrom2Points (p, q: point2d);
	begin
		x := (p.getX + q.getX) / 2;
		y := (p.getY + q.getY) / 2;
		radius := sqrt(sqr(p.getX - q.getX) + sqr(p.getY - q.getY)) / 2;
	end;

procedure circle.makeFrom3Points (p, q, r: point2d);
	var
		cx, cy: extended;
	begin
		p.circleCenterX(q, r, cx, cy);
		x := cx;
		y := cy;
		radius := sqrt(sqr(p.getX - x) + sqr(p.getY - y));
	end;

procedure circle.makeDisk (l: list);
	begin
		case l.length of
			0:  begin
				x := 0;
				y := 0;
				radius := -1.0;
			end;
			1:  begin
				x := point2d(l.ith(1)).getX;
				y := point2d(l.ith(1)).getY;
				radius := 0.0;
			end;
			2: 
				makeFrom2Points(point2d(l.ith(1)), point2d(l.ith(2)));
			3: 
				makeFrom3Points(point2d(l.ith(1)), point2d(l.ith(2)), point2d(l.ith(3)));
		end;
	end;

procedure circle.makeFromPointCircle (p: point2d; c: circle);
	var
		dpc: extended;
	begin
		dpc := sqrt(sqr(p.getX - c.x) + sqr(p.getY - c.y));
		x := (p.getX + c.x + c.radius * (c.x - p.getX) / dpc) / 2;
		y := (p.getY + c.y + c.radius * (c.y - p.getY) / dpc) / 2;
		radius := (c.radius + dpc) / 2;
	end;

function circle.HBGLessThan (o: obj): boolean;  (* Eingefuegt von Michael Vollenbroeker 3.12.92 *)
	begin
		HBGLessThan := y < circle(o).y;
	end;

function circle.verticalDistance (o: obj): extended;  (* Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	var
		c: circle;
	begin
		c := circle(o);
		if y < c.y then
			if c.y - c.radius - (y + radius) > 0 then
				verticalDistance := c.y - c.radius - (y + radius)
			else
				verticalDistance := 0
		else if y - radius - (c.y + c.radius) > 0 then
			verticalDistance := y - radius - (c.y + c.radius)
		else
			verticalDistance := 0
	end;

function circle.getPointType: objectType;  (* Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	begin
		getPointType := realPointType;
	end;

function circle.distanceO (o: obj; var minP, minQ: obj): extended;   (* Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	var
		c: circle;
		m1, m2, mp: point2d;
		dm1m2: extended;
	begin
		c := circle(o);
		m1 := point2d(newObj(realPointType));
		failNil(Handle(m1));
		m2 := point2d(newObj(realPointType));
		failNil(Handle(m2));
		m1.setX(self.x);
		m1.setY(self.y);  (* Mittelpunkt von 'self' *)
		m2.setX(c.x);
		m2.setY(c.y);      (* Mittelpunkt von 'c' *)
		dm1m2 := m1.distanceX(m2);
		if dm1m2 = 0 then begin
			point2d(minP).setPoint(m1);
			point2d(minQ).setPoint(m1);
			distanceO := 0;
			m1.freeAll;
			m2.freeAll;
			exit(distanceO);
		end;
		point2d(minP).setX(self.x + (m2.getX - m1.getX) * self.radius / dm1m2);
		point2d(minP).setY(self.y + (m2.getY - m1.getY) * self.radius / dm1m2);
		point2d(minQ).setX(c.x + (m1.getX - m2.getX) * c.radius / dm1m2);
		point2d(minQ).setY(c.y + (m1.getY - m2.getY) * c.radius / dm1m2);
		if (dm1m2 - c.radius - self.radius > 0) then
			distanceO := dm1m2 - c.radius - self.radius
		else begin
			mp := point2d(minP).middle(point2d(minQ));
			point2d(minP).setPoint(mp);
			point2d(minQ).setPoint(mp);
			distanceO := 0;
			mp.freeAll;
		end;
		m1.freeAll;
		m2.freeAll;
	end;

function circle.getLeftPoint: obj;  (* Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	var
		p: realPoint;
	begin
		new(p);
		failNil(Handle(p));
		p.x := x - radius;
		p.y := y;
		getLeftPoint := p;
	end;

function circle.getRightPoint: obj;  (* Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	var
		p: realPoint;
	begin
		new(p);
		failNil(Handle(p));
		p.x := x + radius;
		p.y := y;
		getRightPoint := p;
	end;

{$IFC GFEnabled }
function circle.convertToParameterPoint: obj;
	var
		pP: pPoint;
		r: Creal;
	begin
		NEW(pP);
		failNil(Handle(pP));
		pP.dimension := 3;
		NEW(r);
		failNil(Handle(r));
		r.value := x;
		pP.fields[1] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := y;
		pP.fields[2] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := radius;
		pP.fields[3] := r;
		convertToParameterPoint := obj(pP);
	end; { convertToParameterPoint }

function circle.convertFromParameterPoint (pP: obj): obj;
	var
		c: circle;
	begin
		NEW(c);
		failNil(Handle(c));
		c.x := Creal(pPoint(pP).fields[1]).value;
		c.y := Creal(pPoint(pP).fields[2]).value;
		c.radius := Creal(pPoint(pP).fields[3]).value;
		convertFromParameterPoint := obj(c);
	end; { convertFromParameterPoint }

procedure circle.createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
	var
		c: circle;
	begin
		whichAttributes[1] := true;
		whichAttributes[2] := true;
		whichAttributes[3] := false;
		NEW(c);
		failNil(Handle(c));
		c.x := rectangle(rect).left;
		c.y := rectangle(rect).bottom;
		lowpP := pPoint(c.convertToParameterPoint);
		c.x := rectangle(rect).right;
		c.y := rectangle(rect).top;
		highpP := pPoint(c.convertToParameterPoint);
		c.free;
	end; { createQuery }

function circle.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := true;
	end; { canUseGridfile }
{$ENDC}

{********************************************}

function testSeries (l: list): boolean;
	var
		e: elementH;
		listOk: boolean;
		d1, d2: longInt;
	begin
		e := l.first^^.next;
		listOk := true;
		IOCheck(false);
		repeat
			if member(e^^.value, markedRealPoint) then begin
				readString(markedRealPoint(e^^.value).mark.value, d1, d2);
				listOk := ioResult = 0;
			end
			else
				listOk := false;
			e := e^^.next;
		until (e = nil) | not listOk;
		IOCheck(true);
		testSeries := listOk;
	end;

	const
		RadiusC = 1;
		CenterC = 2;
		ConvexHullExperimentC = 3;
		EnvelopeExperimentC = 4;
		EnvelopeExperimentSeriesC = 5;

procedure circle.describe;
	begin
		with currentArguments do begin
			if (length = 1) & member(ith(1), circle) then begin
				menuEntry('Center', CenterC);
				menuEntry('Radius', RadiusC);
{$IFC OptionalCommand }
				menuEntry('CH Rnd Pts In Circle Exp...', ConvexHullExperimentC);
				menuEntry('Low Env In Circle Exp...', EnvelopeExperimentC);
{$ENDC }
			end;
			if (length = 1) & member(ith(1), rectangle) then begin
{$IFC OptionalCommand }
				menuEntry('CH Rnd Pts In Rect Exp...', ConvexHullExperimentC);
				menuEntry('Low Env In Rect Exp...', EnvelopeExperimentC);
{$ENDC }
			end
			else if (length > 1) then begin
{$IFC OptionalCommand }
				if member(ith(1), circle) & testSeries(currentArguments) then
					menuEntry('Low Env In Circle Exp Series', EnvelopeExperimentSeriesC);
				if member(ith(1), rectangle) & testSeries(currentArguments) then
					menuEntry('Low Env In Rect Exp Series', EnvelopeExperimentSeriesC);
{$ENDC }
			end;
		end;
	end;

function doEnvelopeExperiment (o: obj; numOfObjects: longInt; var numOfExperiments: longInt): extended;
	var
		i, n, edges: longInt;
		r: lineSegmentVector;
		p: polyLine;
	begin
		n := 0;
		edges := 0;
		new(r);
		failNil(Handle(r));
		r.elements := nil;
		currentPointType := realPointType;
		r.preset(numOfObjects, lineSegmentType);
		currentRandomConstraint := o;
		for i := 1 to numOfExperiments do begin
{$IFC AnimationEnabled }
			if userAbort then
				leave;
{$ENDC }
			n := n + 1;
			r.randomChange;
{$IFC AnimationEnabled }
			abortEnable := false;
{$ENDC }
			p := polyLine(r.lowerEnvelope);
{$IFC AnimationEnabled }
			displayInfoString(stringOf('LE = ', p.length - 2 : 1));
{$ENDC }
			edges := edges + p.length - 2;
			p.freeAll;
{$IFC AnimationEnabled }
			abortEnable := true;
{$ENDC }
		end;
		r.freeAll;
		numOfExperiments := n;
		if n = 0 then
			doEnvelopeExperiment := 0
		else
			doEnvelopeExperiment := edges / n;
	end;

procedure callEnvelopeExperiment (o: obj);
	var
		numOfObjects, numOfExperiments: longInt;
		result: extended;
	begin
		if getUserParameter2('Number of line segments:', 'Number of experiments:', numberOfObjects, numberOfExperiments, numberOfObjects, numberOfExperiments, @isSeqIndex, @isNatural) then begin
{$IFC THINK_PASCAL }
			numOfObjects := trunc(numberOfObjects);
			numOfExperiments := trunc(numberOfExperiments);
{$ELSEC }
			numOfObjects := truncLong(numberOfObjects);
			numOfExperiments := truncLong(numberOfExperiments);
{$ENDC }
			result := doEnvelopeExperiment(o, numOfObjects, numOfExperiments);
			displayInfoString(stringOf('Ls=', numOfObjects : 1, ', Exps=', numOfExperiments : 1, ', LE=', result : 6 : 3));
		end;
	end;

function callEnvelopeExperimentSeries (l: list): vector;
	var
		mp: markedRealPoint;
		v: vector;
		e: elementH;
		numOfObjects, numOfExperiments, startTime: longInt;
		result: extended;
	begin
		v := newVector(l.length - 1);
		e := l.first^^.next;
		repeat
			mp := markedRealPoint(e^^.value.duplicate);
			ReadString(mp.mark.value, numOfObjects, numOfExperiments);
			startTime := TickCount;
			result := doEnvelopeExperiment(l.ith(1), numOfObjects, numOfExperiments);
			mp.mark.value := StringOf('Ls=', numOfObjects : 1, ', Exps=', numOfExperiments : 1, ', LE=', result : 6 : 3, ', Time=', (TickCount - startTime) / 60 : 7 : 3);
			v.append(mp);
			e := e^^.next;
		until e = nil;
		callEnvelopeExperimentSeries := v;
	end;

procedure callConvexHullExperiment (o: obj);
	var
		i, edges, n, k, e: longInt;
		hullLength, eliminated: seqIndex0;
		r: pointVector;
		abortEnableSave: boolean;
	begin
		if getUserParameter2('Number of points:', 'Number of experiments:', numberOfObjects, numberOfExperiments, numberOfObjects, numberOfExperiments, @isSeqIndex, @isNatural) then begin
			edges := 0;
			n := 0;
			e := 0;
{$IFC THINK_PASCAL }
			k := trunc(numberOfObjects);
{$ELSEC }
			k := truncLong(numberOfObjects);
{$ENDC }
			r := createRealPointVector(k);
			currentRandomConstraint := o;
{$IFC THINK_PASCAL }
			for i := 1 to trunc(numberOfExperiments) do begin
{$ELSEC }
				for i := 1 to truncLong(numberOfExperiments) do begin
{$ENDC }
{$IFC AnimationEnabled }
					if userAbort then
						leave;
{$ENDC }
					n := n + 1;
					r.randomChange;
{$IFC AnimationEnabled }
					abortEnableSave := abortEnable;
					abortEnable := false;
{$ENDC }
					r.convexHull(true, false, hullLength, eliminated);
{$IFC AnimationEnabled }
					abortEnable := abortEnableSave;
{$ENDC }
					edges := edges + hullLength;
					e := e + eliminated;
				end;
				r.freeAll;
				displayInfoString(stringOf('Pts=', k : 1, ', Exps=', n : 1, ', CH=', edges / n : 6 : 3, ', Eli=', e / n : 6 : 3));
			end;
		end;

function circle.getCenter: point2d;
	var
		p: point2d;
	begin
		p := point2d(newObj(currentPointType));
		p.setX(x);
		p.setY(y);
		getCenter := p;
	end;

function circle.randomPointCover (n: longInt): obj;
	var
		result: pointVector;
		i: seqIndex;
		p: point2d;
		angle: real;
	begin
		result := newPointVector(n);
		for i := 1 to n do begin
			p := point2d(newInitObj(currentPointType));
			angle := 2 * pi * randomReal;
			p.setXY(x + radius * cos(angle), y + radius * sin(angle));
			result.append(p);
		end;
		randomPointCover := result;
	end;

procedure circle.execute (item: integer);
	begin
		with currentArguments do
			case item of
				RadiusC: 
					currentResult.append(circle(ith(1)).getRadius);
				CenterC: 
					currentResult.append(circle(ith(1)).getCenter);
				ConvexHullExperimentC: 
					callConvexHullExperiment(ith(1));
				EnvelopeExperimentC: 
					callEnvelopeExperiment(ith(1));
				EnvelopeExperimentSeriesC: 
					currentResult.appendNonNil(callEnvelopeExperimentSeries(currentArguments));
			end;
	end;

function circle.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(3);
		appendObjInfoExtended(v, 'X-Center', x, false);
		appendObjInfoExtended(v, 'Y-Center', y, false);
		appendObjInfoExtended(v, 'Radius', radius, false);
		getInfo := v;
	end;

procedure circle.setInfo (o: obj);
	var
		v: vector;
	begin
		v := vector(o);
		x := objInfoExtended(v.elements^^[1]).e;
		y := objInfoExtended(v.elements^^[2]).e;
		radius := abs(objInfoExtended(v.elements^^[3]).e);
	end;

end.
