(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit closestPairLineSegments;

(* this unit added by Frank Bartling, 						*)
(* Universitt-GH Siegen, Germany 							*)
(* Version 1.10:  27-Juni-91								*)

{$SETC AclosestPair = TRUE }

interface

uses
	lineSegmentVector;

implementation

function lineSegmentVector.simpleClosestPair: obj;
	var
		i, j: seqIndex;
		l, r, shortestConnection, shortestConnectLine: lineSegment;
		delta, dis: extended;
	begin
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then
			interactiveHighlight;
{$ENDC }
		l := lineSegment(elements^^[1].duplicate);
		delta := 100000000;
		new(shortestConnection);
		failNil(Handle(shortestConnection));
		shortestConnection.init;
		for i := 1 to length do begin
			r := lineSegment(elements^^[i]);
{$IFC AnimationEnabled AND AClosestPair }
			if animationFlag[AClosestPairItem] then
				r.interactiveHighlight;
{$ENDC }
			for j := i + 1 to length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					shortestConnection.free;
					l.free;
					simpleClosestPair := nil;
					exit(simpleClosestPair);
				end;
{$ENDC }
{$IFC AnimationEnabled AND AClosestPair }
				if animationFlag[AClosestPairItem] then begin
					elements^^[j].interactiveHighlight;
					waitForClick(animationFlag[AClosestPairItem]);
				end;
{$ENDC }
				if lineSegment(elements^^[j]).intersectLineSegmentX(r, l) then
					with l do
						if p.equalPoint(q) then begin
{$IFC AnimationEnabled AND AClosestPair }
							if animationFlag[AClosestPairItem] then begin
								p.interactiveHighlight;
								waitForClick(animationFlag[AClosestPairItem]);
							end;
{$ENDC }
							simpleClosestPair := p.duplicate;
							l.free;
							shortestConnection.free;
							exit(simpleClosestPair);
						end
						else begin

{$IFC AnimationEnabled AND AClosestPair }
							if animationFlag[AClosestPairItem] then begin
								l.interactiveHighlight;
								waitForClick(animationFlag[AClosestPairItem]);
							end;
{$ENDC }
							shortestConnection.free;
							simpleClosestPair := l;
							exit(simpleClosestPair);
						end
				else begin
					new(shortestConnectLine);
					failNil(Handle(shortestConnectLine));
					shortestConnectLine.init;
					dis := lineSegment(elements^^[j]).distanceLineSegmentsX(r, shortestConnectLine);
					if dis < delta then begin
						delta := dis;
						shortestConnection.free;
						shortestConnection := lineSegment(shortestConnectLine.duplicate);
					end;
					shortestConnectLine.free;
				end;
{$IFC AnimationEnabled AND AClosestPair }
				if animationFlag[AClosestPairItem] then
					elements^^[j].interactiveHighlight;
{$ENDC }
			end;
{$IFC AnimationEnabled AND AClosestPair }
			if animationFlag[AClosestPairItem] then
				r.interactiveHighlight;
{$ENDC }
		end;
		l.free;
		simpleClosestPair := shortestConnection;
	end;



	type
		cpEventType = (startEvent, endEvent);

		cpEvent = object(obj)
				where: point2d;
				procedure processEvent (yTable: dictionary; var intersection, shortestConnection: lineSegment; var delta: extended);
				function getEventType: cpEventType;
			end;

		cpStartEvent = object(cpEvent)
				l: lineSegment;
				e: cpEndEvent;
				procedure processEvent (yTable: dictionary; var intersection, shortestConnection: lineSegment; var delta: extended);
				override;
				function getEventType: cpEventType;
				override;
			end;

		cpEndEvent = object(cpEvent)
				toKill: reference; (* into y-table *)
				procedure processEvent (yTable: dictionary; var intersection, shortestConnection: lineSegment; var delta: extended);
				override;
				function getEventType: cpEventType;
				override;
			end;

function cpEvent.getEventType: cpEventType;
	begin
		unimplementedMethod('cpEvent.getEventType');
		getEventType := startEvent;
	end;

function cpStartEvent.getEventType: cpEventType;
	begin
		getEventType := startEvent;
	end;

function cpEndEvent.getEventType: cpEventType;
	begin
		getEventType := endEvent;
	end;

function compareLineSegment (a, b: lineSegment): relationObj;
	begin
		if (b.p.getX <= a.p.getX) & (a.p.getX <= b.q.getX) then
			case sign(b.whichSideX(a.p)) of
				-1: 
					compareLineSegment := greaterThanObj;
				0: 
					compareLineSegment := relationObj(1 + sign(b.whichSideX(a.q)));
				+1: 
					compareLineSegment := lessThanObj
			end
		else if a.p.getX > b.q.getX then
			if a.p.getY >= b.q.getY then
				compareLineSegment := greaterThanObj
			else
				compareLineSegment := lessThanObj
	end;

function lessThancpEvent (a, b: cpEvent): boolean;
	begin
		lessThancpEvent := pointLessThanXY(a.where, b.where) | a.where.equal(b.where) & (a.getEventType < b.getEventType);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure cpEvent.processEvent (yTable: dictionary; var intersection, shortestConnection: lineSegment; var delta: extended);
	begin
		unimplementedMethod('cpEvent.processEvent');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure checkDistance (d: dictionary; a, b: reference; var intersection, shortestConnection: lineSegment; var delta: extended);
	var
		s, r, shortestConnectLine, x: lineSegment;
		dis: extended;
{$IFC AnimationEnabled AND AclosestPair }
		l: list;
{$ENDC }
	begin
		if (a <> nil) & (b <> nil) then begin
			s := lineSegment(d.getObject(a));
			r := lineSegment(d.getObject(b));
			x := lineSegment(s.duplicate);
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				l := newEmptyList;
				l.append(s);
				l.append(r);
				l.interactiveFlash;
				waitForClick(animationFlag[AClosestPairItem]);
				l.free;
			end;
{$ENDC }
			if s.intersectLineSegmentX(r, x) then begin
				intersection := x;
				delta := 0.0;
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					intersection.interactiveHighlight;
					waitForClick(animationFlag[AClosestPairItem]);
				end;
{$ENDC }
			end
			else begin
				x.free;
				new(shortestConnectLine);
				failNil(Handle(shortestConnectLine));
				shortestConnectLine.init;
				dis := s.distanceLineSegmentsX(r, shortestConnectLine);
				if dis < delta then begin
					delta := dis;
					shortestConnection.free;
					shortestConnection := lineSegment(shortestConnectLine.duplicate);
				end;
				shortestConnectLine.free;
			end;
		end;
	end;

procedure cpStartEvent.processEvent (yTable: dictionary; var intersection, shortestConnection: lineSegment; var delta: extended);
	var
		whereRef: reference;
		direction: directionType;
		oldDelta: extended;
	begin
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then begin
			drawThickVertical(where.getX);
			drawVertical(where.getX - delta);
			oldDelta := delta;
			l.interactiveFlash;
			l.interactiveHighlight;
			waitForClick(animationFlag[AClosestPairItem]);
			l.interactiveHighlight;
		end;
{$ENDC }
		if yTable.find(l, compareLineSegment, whereRef, direction) then begin
			intersection := lineSegment(l.duplicate);
			assert(l.intersectLineSegmentX(lineSegment(yTable.getObject(whereRef)), intersection), 'l intersect yTable.l');
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				intersection.interactiveHighlight;
				waitForClick(animationFlag[AClosestPairItem]);
			end;
{$ENDC }
		end
		else begin
			whereRef := yTable.insert(l, whereRef, direction);
			e.toKill := whereRef;
			checkDistance(yTable, whereRef, yTable.next(whereRef, left), intersection, shortestConnection, delta);
			if intersection = nil then
				checkDistance(yTable, whereRef, yTable.next(whereRef, right), intersection, shortestConnection, delta);
		end;
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then begin
			drawThickVertical(where.getX);
			drawVertical(where.getX - oldDelta);
		end;
{$ENDC }
	end;

procedure cpEndEvent.processEvent (yTable: dictionary; var intersection, shortestConnection: lineSegment; var delta: extended);
{$IFC AnimationEnabled AND AclosestPair }
	var
		l: lineSegment;
{$ENDC }
	begin
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then begin
			l := lineSegment(yTable.getObject(toKill));
			l.interactiveFlash;
			l.interactiveHighlight;
			waitForClick(animationFlag[AClosestPairItem]);
			l.interactiveHighlight;
			l.interactiveOutput;
		end;
{$ENDC }
		checkDistance(yTable, yTable.next(toKill, left), yTable.next(toKill, right), intersection, shortestConnection, delta);
		yTable.delete(toKill);
	end;

function lineSegmentVector.planeSweepClosestPair (ask: boolean): obj;
	var
		yTable: dictionary;
		dummy: priorityQueue;
		xQueueStart, xQueueEnd: vector;
		intersection, shortestConnection: lineSegment;
		i, tail: seqIndex1;
		delta, oldDelta: extended;

	procedure buildTwoXQueues (lv: lineSegmentVector; var xQueueStart, xQueueEnd: vector);
		var
			i: seqIndex0;
			cpStartH: cpStartEvent;
			cpEndH: cpEndEvent;
			l: lineSegment;
		begin
			currentVectorLength := length;
			for i := 1 to lv.length do begin
				l := lineSegment(elements^^[i]);
				orderPointsGreaterThanXY(l.p, l.q);
				new(cpEndH);
				failNil(Handle(cpEndH));
				cpEndH.where := l.q;
				cpEndH.toKill := nil; (* will be set by processing the startEvent *)
				xQueueEnd.append(cpEndH);
				new(cpStartH);
				failNil(Handle(cpStartH));
				cpStartH.where := l.p;
				cpStartH.l := l;
				cpStartH.e := cpEndH;
				xQueueStart.append(cpStartH);
			end;
			xQueueStart.sortAll(lessThancpEvent);
			xQueueEnd.sortAll(lessThancpEvent);
		end;


	begin
		if ask then begin
			if not getXY(0, false, dummy, length, yTable, false) then begin
				planeSweepClosestPair := nil;
				exit(planeSweepClosestPair);
			end;
		end
		else begin
			currentVectorLength := length;
			yTable := dictionary(newInitObj(avlTreeType));
		end;
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then
			interactiveHighlight;
{$ENDC }
		new(xQueueStart);
		failNil(Handle(xQueueStart));
		xQueueStart.init;
		new(xQueueEnd);
		failNil(Handle(xQueueEnd));
		xQueueEnd.init;
		buildTwoXQueues(self, xQueueStart, xQueueEnd);
		intersection := nil;
		delta := 10000000000.0;
		i := 1;
		tail := 1;
		new(shortestConnection);
		failNil(Handle(shortestConnection));
		shortestConnection.init;
		while (i <= xQueueStart.length) & (intersection = nil) do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				shortestConnection.free;
				xQueueStart.freeAll;
				xQueueEnd.freeAll;
				yTable.free;
				planeSweepClosestPair := nil;
				exit(planeSweepClosestPair);
			end;
{$ENDC }
			while (cpEvent(xQueueEnd.elements^^[tail]).where.getX <= (cpEvent(xQueueStart.elements^^[i]).where.getX - delta)) do begin
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					drawThickVertical(cpEvent(xQueueStart.elements^^[i]).where.getX);
					drawVertical(cpEvent(xQueueStart.elements^^[i]).where.getX - delta);
					oldDelta := delta;
				end;
{$ENDC }
				cpEvent(xQueueEnd.elements^^[tail]).processEvent(yTable, intersection, shortestConnection, delta);
				tail := tail + 1;
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					drawThickVertical(cpEvent(xQueueStart.elements^^[i]).where.getX);
					drawVertical(cpEvent(xQueueStart.elements^^[i]).where.getX - oldDelta);
				end;
{$ENDC }
			end;
			cpEvent(xQueueStart.elements^^[i]).processEvent(yTable, intersection, shortestConnection, delta);
			i := i + 1;
		end;

		while (tail <= xQueueStart.length - 2) & (intersection = nil) do begin
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				drawThickVertical(cpEvent(xQueueEnd.elements^^[tail]).where.getX);
			end;
{$ENDC }
			cpEvent(xQueueEnd.elements^^[tail]).processEvent(yTable, intersection, shortestConnection, delta);
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				drawThickVertical(cpEvent(xQueueEnd.elements^^[tail]).where.getX);
			end;
{$ENDC }
			tail := tail + 1;
		end;
		xQueueStart.freeAll;
		xQueueEnd.freeAll;
		if yTable.infoString <> '' then
			displayInfoString(concat('Y', yTable.infoString));
		yTable.free;
		if (intersection = nil) then begin
			planeSweepClosestPair := lineSegment(shortestConnection.duplicate);
			shortestConnection.free
		end
		else if (not intersection.p.equalPoint(intersection.q)) then begin
			shortestConnection.free;
			planeSweepClosestPair := lineSegment(intersection.duplicate);
			intersection.free
		end
		else begin
			shortestConnection.free;
			planeSweepClosestPair := intersection.p.duplicate;
			intersection.free;
		end;
	end;

end.
