(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit convexPolygon;

interface

uses
	pointVector;

implementation

{$SETC AconvexTangent= TRUE }
{$SETC AconvexDiameter = TRUE }
{$SETC AconvexIntersection = TRUE }

const
	IntersectC = 1;
	TangentC = 2;
	DiameterC = 3;
	SimpleConvexPolyClosestPairC = 4;
	ConvexPolyClosestPairC = 5;
	ConvexPolyClosestPairAskC = 6;
	ConvexPolySimpleClosestPairC = 7;
	SimpleMinimalDistanceC = 8;
	MinimalDistanceC = 9;
	TestAllCyclicShiftsC = 10;

	SimpleMinimalDistanceDynamicC = 11;
	MinimalDistanceDynamicC = 12;

	IntersectDynamicC = 13;
	TangentDynamicC = 14;
	PlaneSweepClosestPairC = 15;
	PlaneSweepClosestPairDynamicC = 16;
	ConvexPolyClosestPairDynamicC = 17;

function lengthOk (x: convexPolygon): boolean;
	begin
		lengthOk := x.length > 0;
	end;

procedure convexPolygon.describe;

	function separated (a, b: convexPolygon): boolean;
		begin
			if pointLessThanXY(point2d(a.elements^^[1]), point2d(b.elements^^[1])) then
				separated := pointLessThanXY(point2d(a.elements^^[a.findExtreme(pointGreaterThanXY)]), point2d(b.elements^^[1]))
			else
				separated := pointLessThanXY(point2d(b.elements^^[b.findExtreme(pointGreaterThanXY)]), point2d(a.elements^^[1]))
		end;

	begin
		with currentArguments do begin
			if (length > 2) & currentArgumentsSameType & (currentArguments.ith(1).getType in [circleType, realPointType, floatPointType, longIntPointType, convexPolygonType, lineSegmentType]) then begin
				menuEntry('Plane Sweep Closest Pair', PlaneSweepClosestPairC);
				menuEntry('Plane Sweep Closest Pair Dynamic', PlaneSweepClosestPairDynamicC);
			end;
			if (length > 2) & currentArgumentsSameType & member(ith(1), convexPolygon) & sameProperty(lengthOk) then begin       (* ---- inserted by JM ----- *)
				beginSubMenu('Closest Pair');
				menuEntry('Plane Sweep', ConvexPolyClosestPairC);                    (* <<< ---- inserted by JM ----- *)
				menuEntry('Plane Sweep Ask...', ConvexPolyClosestPairAskC);
				menuEntry('Plane Sweep Dynamic', ConvexPolyClosestPairDynamicC);
				menuEntry('Simple N2Dist', SimpleConvexPolyClosestPairC);
				menuEntry('Simple LogDist', ConvexPolySimpleClosestPairC);
				endSubMenu;
			end                                                                              (* ----------------------- *)
			else if (length = 1) & member(ith(1), convexPolygon) then
				menuEntry('Diameter', DiameterC)
			else if (length = 2) & member(ith(1), convexPolygon) & member(ith(2), convexPolygon) & (convexPolygon(ith(1)).length > 0) & (convexPolygon(ith(2)).length > 0) then begin
				if member(convexPolygon(ith(1)).elements^^[1], realPoint) & member(convexPolygon(ith(2)).elements^^[1], realPoint) then begin
					menuEntry('Convex Intersection', IntersectC);
					menuEntry('Convex Intersection Dynamic', IntersectDynamicC);
					menuEntry('Convex Intersection All Shifts', TestAllCyclicShiftsC);
				end;
				if separated(convexPolygon(ith(1)), convexPolygon(ith(2))) then begin
					menuEntry('Tangent', TangentC);
					menuEntry('Tangent Dynamic', TangentDynamicC);
				end;
				beginSubMenu('Minimal Distance');
				menuEntry('Simple', SimpleMinimalDistanceC);
				menuEntry('Simple Dynamic', SimpleMinimalDistanceDynamicC);
				menuEntry('Binary', MinimalDistanceC);
				menuEntry('Binary Dynamic', MinimalDistanceDynamicC);     (*  <<<<------ inserted by JM *)
				endSubMenu;
			end
			else if (length = 1) & member(ith(1), convexPolygon) & (convexPolygon(ith(1)).length > 0) then
				menuEntry('Diameter', DiameterC);
		end;
	end;

procedure orientPolygon (cp: convexPolygon);
	begin
		if not cp.counterclockwise then  (*  changed on 26.6.93 by JM because ConvexPolygon.GetVertex() was changed to other order *)
			cp.revert;
	end;

function simpleClosestPairPS (l: list): lineSegment;
	var
		h: homogenVector;
		r, s: lineSegment;
		rLength, sLength: extended;
		i, j: seqIndex0;
		cp: convexPolygon;
	begin
		r := nil;
		new(h);
		failNil(Handle(h));
		h.initFromList(l);
		for i := 1 to h.length do begin
			cp := convexPolygon(h.elements^^[i]);
			cp.findExtremalPoints;
			for j := 1 to i - 1 do begin
				s := cp.minimalDistance(convexPolygon(h.elements^^[j]), sLength);
{$IFC AnimationEnabled }
				if userAbort then begin
					h.free;
					if r <> nil then
						r.free;
					simpleClosestPairPS := nil;
					exit(simpleClosestPairPS);
				end;
{$ENDC }
				if r = nil then begin
					r := s;
					rLength := sLength;
				end
				else if sLength < rLength then begin
					r.free;
					r := s;
					rLength := sLength;
				end
				else
					s.free;
			end;
		end;
		h.free;
		simpleClosestPairPS := r;
	end;

procedure computeResultIntersect (cp, cq: convexPolygon; var result: obj);
	begin
		result := cp.intersect(cq, false);
	end;

procedure convexPolygon.execute (item: integer);

	procedure callIntersect (cp, cq: convexPolygon);
		begin
			currentResult.appendNonNil(cp.intersect(cq, true));
		end;

	procedure computeResultTangent (cp, cq: convexPolygon; var result: obj);
		var
			upperP, upperQ, lowerP, lowerQ: seqIndex0;
			upper, lower: lineSegment;
			r: homogenVector;
		begin
			cp.tangent(cq, upperP, upperQ, lowerP, lowerQ);
			if userAbort | (upperP = 0) then
				result := nil
			else begin
				new(upper);
				failNil(Handle(upper));
				upper.p := point2d(cp.elements^^[upperP].duplicate);
				upper.q := point2d(cq.elements^^[upperQ].duplicate);
				new(lower);
				failNil(Handle(lower));
				lower.p := point2d(cp.elements^^[lowerP].duplicate);
				lower.q := point2d(cq.elements^^[lowerQ].duplicate);
				r := newHomogenVector(2);
				r.append(upper);
				r.append(lower);
				result := r;
			end;
		end;

	procedure callTangent (cp, cq: convexPolygon);
		var
			result: obj;
		begin
			computeResultTangent(cp, cq, result);
			currentResult.appendNonNil(result);
		end;


(* ----------------------- inserted by JM, changed by PS, MV, JM  ------------------- *)

	procedure appendPointOrLineSegment (dl: lineSegment);
		begin
{$IFC AnimationEnabled }
			if userAbort then
				exit(appendPointOrLineSegment);
{$ENDC }
			if dl.p.equalPoint(dl.q) then begin
				currentResult.appendNonNil(point2d(dl.p.duplicate));
				dl.free;
			end
			else
				currentResult.append(dl);
		end;

		const
			simple = 0;
			binary = 1;

	function getDistance (cp, cq: convexPolygon; findExtremePoints: boolean; version: integer): lineSegment;
		var
			dl: lineSegment;
			d: extended;
		begin
			case version of
				simple: 
					dl := cp.simpleMinimalDistance(cq, d);
				binary:  begin
					if findExtremePoints then begin
						if not cp.counterclockwise then   (*  NOT inserted  by JM on 25.6.93 because ConvexPolygon.GetVertex() was changed to other order *)
							cp.revert;
						cp.findExtremalPoints;
						if not cq.counterclockwise then     (*  NOT inserted  by JM on 25.6.93 because ConvexPolygon.GetVertex() was changed to other order *)
							cq.revert;
						cq.findExtremalPoints;
					end;
					dl := cp.minimalDistance(cq, d);
				end;
			end;
			getDistance := dl;
		end;

	procedure callMinimalDistance (cp, cq: convexPolygon; version: integer);
		begin
			appendPointOrLineSegment(getDistance(cp, cq, true, version));
		end;

	procedure callMinimalDistanceDynamic (cp, cq: convexPolygon; version: integer);

		procedure computeDistance (cp, cq: convexPolygon; var result: obj);
			var
				dl: lineSegment;
			begin
				dl := getDistance(cp, cq, false, version);
				if dl.p.equalPoint(dl.q) then begin
					result := dl.p.duplicate;
					dl.free;
				end
				else
					result := dl;
			end;

		begin
			if version = binary then begin
				if not cp.counterclockwise then (*  NOT inserted  by JM on 25.6.93 because ConvexPolygon.GetVertex() was changed to other order *)
					cp.revert;
				cp.findExtremalPoints;
				if not cq.counterclockwise then (*  NOT inserted  by JM on 25.6.93 because ConvexPolygon.GetVertex() was changed to other order *)
					cq.revert;
				cq.findExtremalPoints;
			end;
			dynamicExecution2(cp, cq, computeDistance, false, true);
		end;

	procedure callClosestPair (ask, simple: boolean);
		var
			dl: lineSegment;
		begin
			currentArguments.forAll(orientPolygon);
			if simple then
				dl := simpleClosestPair
			else
				dl := closestPairConvexPolygons(ask);
			appendPointOrLineSegment(dl);
		end;

(* ------------------------------------------------------------------------------------------------ *)

	procedure callDiameter (c: convexPolygon);
		var
			iMax, jMax, iMin, jMin: seqIndex0;
			s: lineSegment;
		begin
			if c.length > 0 then begin
				c.diameter(iMax, jMax, iMin, jMin);
{$IFC AnimationEnabled }
				if userAbort then
					exit(callDiameter);
{$ENDC }
				new(s);
				failNil(Handle(s));
				s.p := point2d(c.elements^^[iMax].duplicate);
				s.q := point2d(c.elements^^[jMax].duplicate);
				currentResult.append(s);
				new(s);
				failNil(Handle(s));
				s.p := point2d(c.elements^^[iMin].duplicate);
				s.q := point2d(c.elements^^[jMin].duplicate);
				currentResult.append(s);
			end;
		end;

	procedure doAppend (x: obj);
		begin
			currentResult.append(x);
		end;

	procedure callTestAllCyclicShifts (cp, cq: convexPolygon);
		var
			i, j: seqIndex;
			result: vector;
		begin
			if (cp.length > 1) & (cq.length > 1) then begin
				result := newVector(cp.length * cq.length);
				for i := 1 to cp.length do begin
					for j := 1 to cq.length do begin
{$IFC AnimationEnabled AND AconvexIntersection }
						if animationFlag[APolygonIntersectionItem] & ((i > 1) | (j > 1)) then begin
							clearCurrentGeoBenchWindow;
							cp.interactiveOutput;
							cp.interactiveHighlight;
							cq.interactiveOutput;
							cq.interactiveHighlight;
						end;
{$ENDC }
						result.append(cp.intersect(cq, false));
						cq.cyclicShift(2);
					end;
					cp.cyclicShift(2);
				end;
				result.eliminateDuplicates;
				result.forAll(doAppend);
				result.free;
			end;
		end;

	procedure findVertices (cp: convexPolygon);
		begin
			cp.findExtremalPoints;                                  (*  changed on 5.7.93 by JM because convexPolygon.findExtremVertices() has a bug *)
		end;

	procedure conditionalRevert (cp: convexPolygon);  (* sorgt dafuer, dass Polygone entgegen dem Uhrzeigersinn gespeichert sind *)
  (* Eingefuegt von Michael Vollenbroeker 9.12.92 *)
		begin
			if not cp.counterclockwise then
				cp.revert;
		end;

	procedure callClosestPairGeneral;
		begin
			if member(currentArguments.ith(1), convexPolygon) then begin
				currentArguments.forAll(conditionalRevert);
				currentArguments.forAll(findVertices);
			end;
			appendPointOrLineSegment(lineSegment(closestPair));
		end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
	function doComputeClosestPairGeneral (x: sequence): obj;
		var
			result: lineSegment;
		begin
			result := lineSegment(x.closestPair);
			if result.p.equalPoint(result.q) then begin
				doComputeClosestPairGeneral := result.p.duplicate;
				result.free;
			end
			else
				doComputeClosestPairGeneral := result;
		end;

	function doComputeClosestPair (x: sequence): obj;
		var
			result: lineSegment;
		begin
			result := lineSegment(closestPairConvexPolygons(false));
			if result.p.equalPoint(result.q) then begin
				doComputeClosestPair := result.p.duplicate;
				result.free;
			end
			else
				doComputeClosestPair := result;
		end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

	procedure callClosestPairGeneralDynamic;
		begin
			if member(currentArguments.ith(1), convexPolygon) then begin
				currentArguments.forAll(conditionalRevert);
				currentArguments.forAll(findVertices);
			end;
			currentArguments.dynamicExecutionHighlightShowType(doComputeClosestPairGeneral);
		end;

	procedure callClosestPairDynamic;
		begin
			currentArguments.forAll(orientPolygon);
			currentArguments.dynamicExecutionHighlightShowType(doComputeClosestPair);
		end;

	begin
		case item of
			PlaneSweepClosestPairC: 
				callClosestPairGeneral;
			PlaneSweepClosestPairDynamicC: 
				callClosestPairGeneralDynamic;
			ConvexPolyClosestPairDynamicC: 
				callClosestPairDynamic;
			IntersectDynamicC: 
				dynamicExecution2(currentArguments.ith(1), currentArguments.ith(2), computeResultIntersect, true, true);
			TangentDynamicC: 
				dynamicExecution2(currentArguments.ith(1), currentArguments.ith(2), computeResultTangent, true, true);
			TestAllCyclicShiftsC: 
				callTestAllCyclicShifts(convexPolygon(currentArguments.ith(1)), convexPolygon(currentArguments.ith(2)));
			IntersectC: 
				callIntersect(convexPolygon(currentArguments.ith(1)), convexPolygon(currentArguments.ith(2)));
			TangentC: 
				callTangent(convexPolygon(currentArguments.ith(1)), convexPolygon(currentArguments.ith(2)));
			MinimalDistanceC, SimpleMinimalDistanceC:                                                                     (* <---- inserted by JM *)
				callMinimalDistance(convexPolygon(currentArguments.ith(1)), convexPolygon(currentArguments.ith(2)), item - SimpleMinimalDistanceC);
			MinimalDistanceDynamicC, SimpleMinimalDistanceDynamicC:                                           (* <---- inserted by JM *)
				callMinimalDistanceDynamic(convexPolygon(currentArguments.ith(1)), convexPolygon(currentArguments.ith(2)), item - SimpleMinimalDistanceDynamicC);
			DiameterC: 
				callDiameter(convexPolygon(currentArguments.ith(1)));
			ConvexPolyClosestPairC, ConvexPolyClosestPairAskC, SimpleConvexPolyClosestPairC:   (* <----- inserted by JM *)
				callClosestPair(item = ConvexPolyClosestPairAskC, item = SimpleConvexPolyClosestPairC);
			ConvexPolySimpleClosestPairC: 
				currentResult.appendNonNil(simpleClosestPairPS(currentArguments));
		end;
	end;

procedure convexPolygon.randomChange;
	const
		infinity = 10000;
		twoPi = 6.28318530718;
	var
		i, j, k: seqIndex0;
		angles: homogenVector;
		angleRN: realNumber;
		p, pp, q, qq: point2d;
		el: objArrayH;
		alpha, cosAlpha, sinAlpha, originX, originY, polygonRadius, rl, rr, rb, rt, circleRadius: extended;
{$IFC ChecksEnabled }
		count: seqIndex1;
{$ENDC }
		isRectangle: boolean;

	function rayIntersect (px, py, qx, qy: extended): extended;
		var
			dx, dy, d: extended;
		begin
			dx := qx - px;
			dy := py - qy;
			d := cosAlpha * dy + sinAlpha * dx;
			if d = 0 then
				rayIntersect := infinity
			else begin
				d := ((px - originX) * dy + (py - originY) * dx) / d;
				if d < 0 then
					rayIntersect := infinity
				else
					rayIntersect := d;
			end;
		end;

	procedure setPoint (z: point2d; minT, maxT: extended);
		var
			realT: extended;
		begin
			assert(minT >= 0, 'minT >= 0');
			if (0 < minT) & (minT < maxT) & (maxT < infinity) then begin
				realT := randomReal * (maxT - minT) + minT;
				z.setXY(originX + realT * cosAlpha, originY + realT * sinAlpha);
			end;
		end;

	function getMaxT (p1x, p1y, p2x, p2y, q1x, q1y, q2x, q2y: extended): extended;
		var
			boundary: extended;
		begin
			if isRectangle then
				boundary := min(min(rayIntersect(rl, rb, rl, rt), rayIntersect(rl, rt, rr, rt)), min(rayIntersect(rr, rt, rr, rb), rayIntersect(rr, rb, rl, rb)))
			else
				boundary := circleRadius;
			getMaxT := min(min(rayIntersect(p1x, p1y, p2x, p2y), rayIntersect(q1x, q1y, q2x, q2y)), boundary);
		end;

	begin (* convexPolygon.randomChange *)
		if currentRandomConstraint <> nil then begin
			isRectangle := member(currentRandomConstraint, rectangle);
			if (length > 2) & (isRectangle | member(currentRandomConstraint, circle)) then begin
				if isRectangle then
					with rectangle(currentRandomConstraint) do begin
						rl := left;
						rr := right;
						rb := bottom;
						rt := top;
						polygonRadius := min(rr - rl, rt - rb) / 2;
					end
				else
					with circle(currentRandomConstraint) do begin
						circleRadius := radius;
						polygonRadius := randomReal * circleRadius;
					end;
				currentRandomConstraint.centerOfGravity(originX, originY);
				angles := newHomogenVector(length);
				for i := 1 to length do begin
					new(angleRN);
					failNil(Handle(angleRN));
					angleRN.value := twoPi * randomReal;
					angles.append(angleRN);
				end;
				angles.sortAll(lessThanRealNumber);
				for i := 1 to length do begin
					alpha := realNumber(angles.elements^^[i]).value;
					point2d(elements^^[i]).setXY(originX + polygonRadius * cos(alpha), originY + polygonRadius * sin(alpha));
				end;
				for j := 1 to length do begin
{$IFC THINK_PASCAL }
					i := trunc(randomReal * length) + 1;
{$ELSEC }
					i := truncLong(randomReal * length) + 1;
{$ENDC }
					pp := point2d(cyclicIth(i - 2));
					p := point2d(cyclicIth(i - 1));
					q := point2d(cyclicIth(i + 1));
					qq := point2d(cyclicIth(i + 2));
					alpha := realNumber(angles.elements^^[i]).value;
					cosAlpha := cos(alpha);
					sinAlpha := sin(alpha);
					setPoint(point2d(elements^^[i]), rayIntersect(p.getX, p.getY, q.getX, q.getY), getMaxT(p.getX, p.getY, pp.getX, pp.getY, q.getX, q.getY, qq.getX, qq.getY));
				end;
				el := objArrayH(dynamicArray(length + 1, sizeOf(obj)));
				k := findExtreme(pointLessThanXY);
				for i := 1 to length do begin
					el^^[i] := elements^^[k];
					if k = 1 then
						k := length
					else
						k := k - 1;
				end;
{$IFC THINK_PASCAL }
				disposHandle(Handle(elements));
{$ELSEC }
				DisposeHandle(Handle(elements));
{$ENDC }
				elements := el;
				angles.freeAll;
{$IFC ChecksEnabled }
				assert(isConvex(count) <= polygonIsConvexPlus, 'Random polygon is convex.');
{$ENDC }
			end
			else
				inherited randomChange;
		end;
	end;

procedure convexPolygon.establishConvexity;
	var
		i, hullLength: seqIndex0;
{$IFC animationEnabled }
		saveAnimationFlag: boolean;
{$ENDC }
	begin
{$IFC animationEnabled }
		saveAnimationFlag := animationFlag[AConvexHullItem];
		animationFlag[AConvexHullItem] := false;
		abortEnable := false;
{$ENDC }
		inherited convexHull(true, false, hullLength, i);(* 'i' is used as dummy *)
{$IFC animationEnabled }
		abortEnable := true;
		animationFlag[AConvexHullItem] := saveAnimationFlag;
{$ENDC }
		for i := hullLength + 1 to length do
			elements^^[i].free;
		allocateElements(hullLength);
	end;

procedure convexPolygon.windowInput (px, py: integer);
	begin
		inherited windowInput(px, py);
		establishConvexity;
	end;

function convexPolygon.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	begin
		if inherited dragVertex(where, mouse, highlight) then begin
			dragVertex := true;
			interactiveOutput;
			if highlight then
				interactiveHighlight;
			establishConvexity;
			interactiveOutput;
			if highlight then
				interactiveHighlight;
		end
		else
			dragVertex := false;
	end;

function convexPolygon.counterclockwise: boolean;
	var
		dummy: seqIndex1;
	begin
		counterclockwise := isConvex(dummy) >= polygonIsConvexDegenerate;
	end;

procedure convexPolygon.setInfo (o: obj);
	begin
		inherited setInfo(o);
		establishConvexity;
	end;

function convexPolygon.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, convexPolygon) & inherited equal(o);
	end;

function convexPolygon.getType: objectType;
	begin
		getType := convexPolygonType;
	end;

function convexPolygon.getTextType: Str255;
	begin
		getTextType := concat('Convex Polygon', getTextBaseType);
	end;

function convexPolygon.memberType (o: obj): boolean;
	begin
		memberType := member(o, convexPolygon) & memberBaseType(o);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure convexPolygon.convexHull (eliminate, verbose: boolean; var hullLength, eliminated: seqIndex0);
	begin
		hullLength := length;
		eliminated := 0;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function convexPolygon.convexHullDivideAndConquer: convexPolygon;
	begin
		convexHullDivideAndConquer := convexPolygon(duplicateAll);
	end;

procedure convexTangent (ps, qs: vector; var upperP, upperQ, lowerP, lowerQ: seqIndex0);
	var
		limit: longInt;

	procedure advance (ps, qs: vector; var i, j: seqIndex0);
		var
			finis: boolean;
			k: longInt;
		begin
			finis := false;
			k := 0;
			repeat
				k := k + 1;
				if k > limit then begin
					upperP := 0;
					upperQ := 0;
					lowerP := 0;
					lowerQ := 0;
					exit(convexTangent);
				end;
{$IFC AnimationEnabled }
				if userAbort then
					exit(convexTangent);
{$ENDC }
{$IFC AnimationEnabled AND AconvexTangent }
				if animationFlag[ATangentItem] then begin
					point2d(ps.elements^^[i]).drawLine(point2d(qs.elements^^[j]));
					waitForClick(animationFlag[ATangentItem]);
					point2d(ps.elements^^[i]).drawLine(point2d(qs.elements^^[j]));
				end;
{$ENDC }
				if point2d(qs.elements^^[j]).whichSideSign(point2d(qs.elements^^[j + 1]), point2d(ps.elements^^[i])) = -1 then begin
					if j = qs.length then
						j := 1
					else
						j := j + 1
				end
				else if point2d(ps.elements^^[i]).whichSideSign(point2d(ps.elements^^[i + 1]), point2d(qs.elements^^[j])) = +1 then begin
					if i = ps.length then
						i := 1
					else
						i := i + 1
				end
				else
					finis := true;
			until finis;
{$IFC AnimationEnabled AND AconvexTangent }
			if animationFlag[ATangentItem] then
				point2d(ps.elements^^[i]).drawLine(point2d(qs.elements^^[j]));
{$ENDC }
		end; (* advance *)

	begin
		if (ps.length > 0) & (qs.length > 0) then begin
			ps.elements^^[ps.length + 1] := ps.elements^^[1];
			qs.elements^^[qs.length + 1] := qs.elements^^[1];
			upperP := 1;
			upperQ := 1;
			limit := 2 * (ps.length + qs.length);
			if pointLessThanXY(point2d(ps.elements^^[upperP]), point2d(qs.elements^^[upperQ])) then
				advance(ps, qs, upperP, upperQ)
			else
				advance(qs, ps, upperQ, upperP);
			lowerP := ps.findExtreme(pointGreaterThanXY);
			lowerQ := qs.findExtreme(pointGreaterThanXY);
			if pointGreaterThanXY(point2d(ps.elements^^[lowerP]), point2d(qs.elements^^[lowerQ])) then
				advance(ps, qs, lowerP, lowerQ)
			else
				advance(qs, ps, lowerQ, lowerP);
		end;
	end; (* convexTangent *)

procedure convexPolygon.tangent (c: convexPolygon; var upperP, upperQ, lowerP, lowerQ: seqIndex0);
	begin
		convexTangent(self, c, upperP, upperQ, lowerP, lowerQ);
	end;

function convexPolygon.inside (p: point2d): boolean;
	var
		onBoundary: boolean;
	begin
{$IFC AnimationEnabled }
		abortEnable := false;
{$ENDC }
		inside := (windingNumber(p, onBoundary) = 1) | onBoundary;
{$IFC AnimationEnabled }
		abortEnable := true;
{$ENDC }
	end; (* inside *)

procedure convexPolygon.diameter (var iMax, jMax, iMin, jMin: seqIndex0);
	var
		p, pp: point2d;
		pi, qi, q0i: seqIndex;
		dummy: seqIndex1;
		finis1, finis2: boolean;
		maxDistance, minDistance, t: extended;
{$IFC AnimationEnabled AND AconvexDiameter }
		qiOld: seqIndex;
{$ENDC }

	function next (i: seqIndex): seqIndex;
		begin
			if i = length then
				next := 1
			else
				next := i + 1;
		end; (* next *)

	procedure pair (pi, qi: seqIndex);
		var
			newDistance: extended;
		begin
{$IFC AnimationEnabled AND AconvexDiameter }
			if animationFlag[ADiameterItem] then begin
				point2d(elements^^[qiOld]).labelPoint('q');
				point2d(elements^^[qi]).labelPoint('q');
				qiOld := qi;
				p.drawLine(point2d(elements^^[qi]));
				waitForClick(animationFlag[ADiameterItem]);
			end;
{$ENDC }
			newDistance := point2d(elements^^[pi]).squaredDistanceX(point2d(elements^^[qi]));
			if newDistance > maxDistance then begin
				maxDistance := newDistance;
				iMax := pi;
				jMax := qi;
			end;
			if newDistance < minDistance then begin
				minDistance := newDistance;
				iMin := pi;
				jMin := qi;
			end;
		end; (* pair *)

	begin
		if length > 2 then begin
			if isConvex(dummy) = polygonIsConvexPlus then
				revert;
			maxDistance := 0;
			minDistance := 1.0E99;
			p := point2d(elements^^[length]);
			pp := point2d(elements^^[1]);
			elements^^[length + 1] := pp;
			pi := 1;
			qi := 1;
			while p.whichSideX(pp, point2d(elements^^[qi + 1])) > p.whichSideX(pp, point2d(elements^^[qi])) do
				qi := qi + 1;
			q0i := qi;
{$IFC AnimationEnabled AND AconvexDiameter }
			if animationFlag[ADiameterItem] then begin
				point2d(elements^^[qi]).labelPoint('q');
				qiOld := qi;
			end;
{$ENDC }
			finis1 := false;
{$IFC AnimationEnabled AND AconvexDiameter }
			if animationFlag[ADiameterItem] then begin
				p.labelPoint('p');
				pp.labelPoint('p+');
			end;
{$ENDC }
			repeat
{$IFC AnimationEnabled AND AconvexDiameter }
				if animationFlag[ADiameterItem] then begin
					p.labelPoint('p');
					pp.labelPoint('p+');
				end;
{$ENDC }
				p := pp;
				pp := point2d(elements^^[pi + 1]);
{$IFC AnimationEnabled AND AconvexDiameter }
				if animationFlag[ADiameterItem] then begin
					p.labelPoint('p');
					pp.labelPoint('p+');
				end;
{$ENDC }
				pair(pi, qi);
				finis2 := false;
				repeat
{$IFC AnimationEnabled }
					if userAbort then
						exit(diameter);
{$ENDC }
					t := p.whichSideX(pp, point2d(elements^^[next(qi)])) - p.whichSideX(pp, point2d(elements^^[qi]));
					if t < 0 then
						finis2 := true
					else if (pi = q0i) & (qi = length) then
						finis1 := true
					else begin
						qi := next(qi);
						pair(pi, qi);
						finis2 := t = 0;
					end;
				until finis2 | finis1;
				pi := pi + 1;
			until finis1;
		end
		else begin
			iMax := ord(length > 0);
			jMax := length;
			iMin := iMax;
			jMin := jMax;
		end;
	end; (* diameter *)

{$SETC FixByORourke = FALSE }

{$IFC FixByORourke }
procedure previousPoint (var pqIndex: seqIndex0; pq: convexPolygon; var pqHead, pqTail: realPoint);
	begin
		pqIndex := pqIndex - 1;
		if pqIndex = 0 then
			pqIndex := pq.length;
		pqHead := realPoint(pq.elements^^[pqIndex]);
		if pqIndex = 1 then
			pqTail := realPoint(pq.elements^^[pq.length])
		else
			pqTail := realPoint(pq.elements^^[pqIndex - 1]);
	end;
{$ENDC }

procedure convexIntersection (p, q: convexPolygon; var r: convexPolygon; verbose: boolean);
	var
		pHead, pTail, qHead, qTail, intersection: realPoint;
		pIndex, qIndex, pIndexFirstIntersection, qIndexFirstIntersection: seqIndex0;
		cp: extended;
		inside: (pInside, qInside, undefined);
		finis: boolean;
		pCount, qCount, iterationCount, maxIterationCount: longInt;

	procedure assignPointVector (var lhs: convexPolygon; rhs: convexPolygon);
		var
			i: seqIndex0;
		begin
			for i := 1 to rhs.length do
				lhs.elements^^[i] := rhs.elements^^[i].duplicate;
			lhs.length := rhs.length;
		end; (* assignPointVector *)

	procedure output (nextPoint: realPoint);
		begin
			if (r.length = 0) | (not nextPoint.equalPoint(point2d(r.elements^^[r.length])) & not nextPoint.equalPoint(point2d(r.elements^^[1]))) then begin
				if (r.length > 1) & nextPoint.equalPoint(point2d(r.elements^^[2])) then begin
					finis := true;
{$IFC FixByORourke }
					assert(false, 'Fix by O''Rourke ok');
{$ENDC }
				end
				else begin
					r.append(nextPoint.duplicate);
{$IFC AnimationEnabled AND AconvexIntersection }
					if animationFlag[APolygonIntersectionItem] then begin
						nextPoint.interactiveOutput;
						nextPoint.interactiveFlash;
					end;
{$ENDC }
				end;
			end;
		end; (* output *)

	procedure twoDim (var p, q, r: convexPolygon);  (* p.length=2 *)
		var
			i: seqIndex;
			left, right: realPoint;
		begin
			q.elements^^[q.length + 1] := q.elements^^[1];
			i := 1;
			new(left);
			failNil(Handle(left));
			left.init;
			new(right);
			failNil(Handle(right));
			right.init;
			while (i <= q.length) & (r.length < 2) do begin
				if realPoint(p.elements^^[1]).intersectLineSegmentX(realPoint(p.elements^^[2]), point2d(q.elements^^[i]), realPoint(q.elements^^[i + 1]), point2d(left), point2d(right)) then begin
					output(left);
					output(right);
				end;
				i := i + 1;
			end;
			right.free;
			left.free;
			if r.length = 2 then
				orderPointsGreaterThanXY(point2d(r.elements^^[1]), point2d(r.elements^^[2]));
		end; (* twoDim *)

	begin
{$IFC AnimationEnabled AND AconvexIntersection }
		if animationFlag[APolygonIntersectionItem] then begin
			p.interactiveHighlight;
			q.interactiveHighlight;
		end;
{$ENDC }
		r.length := 0;
		if (p.length > 2) & (q.length > 2) then begin

{$IFC AnimationEnabled AND AconvexIntersection }
			iterationCount := 0;
{$ENDC }
			pHead := realPoint(p.elements^^[2]);
			pTail := realPoint(p.elements^^[1]);
			pIndex := 2;
			pIndexFirstIntersection := 0;
			qHead := realPoint(q.elements^^[2]);
			qTail := realPoint(q.elements^^[1]);
			qIndex := 2;

{$IFC FixByORourke }
			finis := false;
			repeat
				if pTail.equalPoint(qHead) then
					previousPoint(pIndex, p, pHead, pTail)
				else if pHead.equalPoint(qTail) then
					previousPoint(qIndex, q, qHead, qTail)
				else if pTail.equalPoint(qTail) then begin
					previousPoint(pIndex, p, pHead, pTail);
					previousPoint(qIndex, q, qHead, qTail)
				end
				else
					finis := true;
			until finis;
{$ENDC }

			inside := undefined;
			pCount := 0;
			qCount := 0;
			finis := false;
			new(intersection);
			failNil(Handle(intersection));
			intersection.init;
			maxIterationCount := 2 * (p.length + q.length);
			repeat
{$IFC AnimationEnabled }
				if userAbort then begin
					intersection.free;
					r.freeAll;
					r := nil;
					exit(convexIntersection);
				end;
{$ENDC }
				cp := pTail.crossProductX(pHead, qTail, qHead);

{$IFC AnimationEnabled AND AconvexIntersection }
				iterationCount := iterationCount + 1;
				if animationFlag[APolygonIntersectionItem] then begin
					case inside of
						undefined: 
							displayInfoString('Inside = Undefined');
						pInside: 
							displayInfoString('Inside = P');
						qInside: 
							displayInfoString('Inside = Q');
					end;
					pTail.drawLine(pHead);
					qTail.drawLine(qHead);
					PenSize(3, 3);
					pTail.drawLine(pHead);
					qTail.drawLine(qHead);
					pHead.labelPoint('p');
					qHead.labelPoint('q');
					waitForClick(animationFlag[APolygonIntersectionItem]);
					pTail.drawLine(pHead);
					qTail.drawLine(qHead);
					pHead.labelPoint('p');
					qHead.labelPoint('q');
					PenSize(1, 1);
					pTail.drawLine(pHead);
					qTail.drawLine(qHead);
				end;
{$ENDC }
				if (cp <> 0) & pTail.intersectLineSegmentX(pHead, qTail, qHead, point2d(intersection), point2d(intersection)) then
					if (pIndexFirstIntersection = pIndex) & (qIndexFirstIntersection = qIndex) then
						finis := true
					else begin
						if pIndexFirstIntersection = 0 then begin
							pIndexFirstIntersection := pIndex;
							qIndexFirstIntersection := qIndex;
						end;
						output(intersection);
						if qTail.whichSideX(qHead, pHead) >= 0 then
							inside := pInside
						else
							inside := qInside;
					end;
				if not finis then
					if (cp >= 0) & (qTail.whichSideX(qHead, pHead) >= 0) | (cp < 0) & (pTail.whichSideX(pHead, qHead) < 0) then begin
						if inside = qInside then
							output(qHead);
						qTail := qHead;
						if qIndex = q.length then
							qIndex := 1
						else
							qIndex := qIndex + 1;
						qHead := realPoint(q.elements^^[qIndex]);
						qCount := qCount + 1;
					end
					else begin
						if inside = pInside then
							output(pHead);
						pTail := pHead;
						if pIndex = p.length then
							pIndex := 1
						else
							pIndex := pIndex + 1;
						pHead := realPoint(p.elements^^[pIndex]);
						pCount := pCount + 1;
					end;
			until finis | (iterationCount > maxIterationCount) | (pIndexFirstIntersection = 0) & ((pCount > p.length) | (qCount > q.length));
{$IFC AnimationEnabled AND AconvexIntersection }
			if verbose then
				displayInfoString(StringOf('Iterations = ', iterationCount : 1));
{$ENDC }
			intersection.free;
			if pIndexFirstIntersection > 0 then
			else if q.inside(realPoint(p.elements^^[1])) then
				assignPointVector(r, p)
			else if p.inside(realPoint(q.elements^^[1])) then
				assignPointVector(r, q);
		end
		else if (p.length = 2) & (q.length > 1) then
			twoDim(p, q, r)
		else if (q.length = 2) & (p.length > 1) then
			twoDim(q, p, r)
		else if (p.length = 1) & q.inside(realPoint(p.elements^^[1])) then
			assignPointVector(r, p)
		else if (q.length = 1) & p.inside(realPoint(q.elements^^[1])) then
			assignPointVector(r, q);
	end;

function convexPolygon.intersect (x: convexPolygon; verbose: boolean): convexPolygon;
	var
		dummy: seqIndex1;
		result: convexPolygon;
	begin
		if isConvex(dummy) = polygonIsConvexPlus then
			revert;
		if x.isConvex(dummy) = polygonIsConvexPlus then
			x.revert;
		currentVectorLength := length + x.length;
		new(result);
		failNil(Handle(result));
		result.init;
		convexIntersection(self, x, result, verbose);
{$IFC AnimationEnabled }
		if userAbort then begin
			intersect := nil;
			exit(intersect);
		end;
{$ENDC }
		if result.length = 0 then begin
			result.freeAll;
			intersect := nil;
		end
		else begin
			result.allocateElements(result.length);
			intersect := result;
		end;
	end;

function convexPolygon.verticalDistance (o: obj): extended;
	var
		Q: convexPolygon;
	begin
		Q := convexPolygon(o);
		if (point2d(elements^^[bottom]).getY > point2d(Q.elements^^[Q.top]).getY) then
			verticalDistance := point2d(elements^^[bottom]).DyX(point2d(Q.elements^^[Q.top]))
		else if (point2d(elements^^[top]).getY < point2d(Q.elements^^[Q.bottom]).getY) then
			verticalDistance := point2d(Q.elements^^[Q.bottom]).DyX(point2d(elements^^[top]))
		else
			verticalDistance := 0;
	end;

function convexPolygon.getPointType: objectType;
	begin
		getPointType := point2d(self.elements^^[1]).getType;
	end;

function convexPolygon.distanceO (o: obj; var minP, minQ: obj): extended;
	begin
		distanceO := distance(convexPolygon(o), point2d(minP), point2d(minQ));
	end;

function convexPolygon.getLeftPoint: obj;
	var
		p: point2d;
	begin
		p := point2d(newInitObj(getPointType));
		p.setPoint(point2d(elements^^[left]));
		getLeftPoint := p;
	end;

function convexPolygon.getRightPoint: obj;
	var
		p: point2d;
	begin
		p := point2d(newInitObj(getPointType));
		p.setPoint(point2d(self.elements^^[right]));
		getRightPoint := p;
	end;

end.
