(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit dDimCircle;

interface

uses
{$IFC GFEnabled }
	GFValues, 
{$ENDC }
	basicGeometry;

implementation

{$SETC AminimalDisk = TRUE }

const
	undefinedRadius = -1.0;

procedure dDimPoint.centerOfGravity (var cx, cy: extended);
	begin
		cx := coordinates[1];
		cy := coordinates[2];
	end;

function dDimPoint.toRealPoint: realPoint;
	var
		p: realPoint;
	begin
		new(p);
		failNil(Handle(p));
		p.x := coordinates[1];
		p.y := coordinates[2];
		toRealPoint := p;
	end;

function dDimPoint.pointCover (n: longInt): obj;
	var
		p: realPoint;
	begin
		p := toRealPoint;
		pointCover := p.pointCover(n);
		p.free;
	end;

function dDimPoint.genericLessThan (o: obj): boolean;
	var
		d: dDimPoint;
	begin
		d := dDimPoint(o);
		genericLessThan := (dimension < d.dimension) | (dimension = d.dimension) & lexicographicLessThan(coordinates, d.coordinates, dimension);
	end;

procedure dDimPoint.init;
	var
		i: dimensionRange;
	begin
		dimension := currentDimension;
		for i := 1 to dimension do
			coordinates[i] := 0;
	end;

function dDimPoint.isSelected (where: obj): boolean;
	var
		p: realPoint;
	begin
		p := toRealPoint;
		isSelected := p.isSelected(where);
		p.free;
	end;

procedure dDimPoint.interactiveOutput;
	var
		p: realPoint;
	begin
		p := toRealPoint;
		p.interactiveOutput;
		p.free;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function dDimPoint.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 1;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure dDimPoint.interactiveHighlight;
	var
		p: realPoint;
	begin
		p := toRealPoint;
		p.interactiveHighlight;
		p.free;
	end;

procedure dDimPoint.textFileInput;
	var
		dd: longInt;
		i: dimensionRange;
	begin
		readTextLongInt(dd);
		dimension := dd;
		for i := 1 to dimension do
			readTextReal(coordinates[i]);
	end;

procedure dDimPoint.textFileOutput;
	var
		i: dimensionRange;
	begin
		case currentTextFileOutputMethod of
			0:  begin
				writeTextLongInt(dimension);
				for i := 1 to dimension do
					writeTextReal(coordinates[i]);
			end;
			otherwise
				for i := 1 to dimension do
					writeTextReal(coordinates[i]);
		end;
	end;

procedure dDimPoint.fileInput;
	var
		i: dimensionRange;
		b: Byte;
	begin
		readByte(b);
		dimension := dimensionRange(b);
		for i := 1 to dimension do
			readReal(coordinates[i]);
	end;

procedure dDimPoint.fileOutput;
	var
		i: dimensionRange;
	begin
		writeByte(Byte(dimension));
		for i := 1 to dimension do
			writeReal(coordinates[i]);
	end;

procedure dDimPoint.randomChange;
	var
		i: dimensionRange;
		c: dDimCircle;
	begin
		if currentRandomConstraint <> nil then
			if member(currentRandomConstraint, rectangle) then
				with rectangle(currentRandomConstraint) do begin
					coordinates[1] := (right - left) * randomReal + left;
					for i := 2 to dimension do
						coordinates[i] := (top - bottom) * randomReal + bottom;
				end
			else if member(currentRandomConstraint, dDimCircle) & (dDimCircle(currentRandomConstraint).dimension = dimension) then
				randomInCircle(dDimCircle(currentRandomConstraint))
			else if member(currentRandomConstraint, circle) then
				with circle(currentRandomConstraint) do begin
					new(c);
					failNil(Handle(c));
					c.dimension := dimension;
					c.coordinates[1] := x;
					c.coordinates[2] := y;
					for i := 3 to dimension do
						c.coordinates[i] := 0;
					c.radius := radius;
					randomInCircle(c);
					c.free;
				end;
	end;

function dDimPoint.equal (o: obj): boolean;
	var
		i: dimensionRange1;
	begin
		if (o <> nil) & member(o, dDimPoint) & (dDimPoint(o).dimension = dimension) then begin
			i := 1;
			while (i <= dimension) & (dDimPoint(o).coordinates[i] = coordinates[i]) do
				i := i + 1;
			equal := i > dimension;
		end
		else
			equal := false;
	end;

function dDimPoint.memberType (o: obj): boolean;
	begin
		memberType := member(o, dDimPoint) & (dDimPoint(o).dimension = dimension);
	end;

procedure dDimPoint.translate (dx, dy: extended);
	begin
		coordinates[1] := coordinates[1] + dx;
		coordinates[2] := coordinates[2] + dy;
	end;

procedure dDimPoint.rotate (xOrigin, yOrigin, alpha: extended);
	var
		x, y: extended;
	begin
		x := coordinates[1];
		y := coordinates[2];
		rotatePoint(xOrigin, yOrigin, alpha, x, y);
		coordinates[1] := x;
		coordinates[2] := y;
	end;

procedure dDimPoint.scale (lambda: extended);
	var
		i: dimensionRange;
	begin
		for i := 1 to dimension do
			coordinates[i] := coordinates[i] * lambda;
	end;

procedure dDimPoint.reflect (a, b, c: extended);
	var
		x, y: extended;
	begin
		x := coordinates[1];
		y := coordinates[2];
		reflectPoint(a, b, c, x, y);
		coordinates[1] := x;
		coordinates[2] := y;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure dDimPoint.construct (px, py, newX, newY: integer);
	begin
		coordinates[1] := newX;
		coordinates[2] := newY;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure dDimPoint.windowOutput;
	var
		p: realPoint;
	begin
		p := toRealPoint;
		p.windowOutput;
		p.free;
	end;

procedure dDimPoint.windowToWorld;
	begin
		with currentFrame do begin
			coordinates[1] := (coordinates[1] - dx) / mx;
			coordinates[2] := (coordinates[2] - dy) / my;
		end;
	end;

function dDimPoint.getType: objectType;
	begin
		getType := dDimPointType;
	end;

function dDimPoint.getTextType: Str255;
	begin
		getTextType := stringOf('D Point(', dimension : 1, ')');
	end;

{$IFC GFEnabled }
function dDimPoint.convertToParameterPoint: obj;
	var
		p: pPoint;
		r: Creal;
		id: INTEGER;
	begin
		NEW(p);
		failNil(Handle(p));
		p.dimension := dimension;
		for id := 1 to dimension do begin
			NEW(r);
			failNil(Handle(r));
			r.value := coordinates[id];
			p.fields[id] := r;
		end;
		convertToParameterPoint := obj(p);
	end; { convertToParameterPoint }

function dDimPoint.convertFromParameterPoint (pP: obj): obj;
	var
		d: dDimPoint;
		id: INTEGER;
	begin
		NEW(d);
		failNil(Handle(d));
		d.dimension := pPoint(pP).dimension;
		for id := 1 to pPoint(pP).dimension do
			d.coordinates[id] := Creal(pPoint(pP).fields[id]).value;
		convertFromParameterPoint := obj(d);
	end; { convertFromParameterPoint }

procedure dDimPoint.createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
	var
		d: dDimPoint;
		id: INTEGER;
	begin
		whichAttributes[1] := true;
		whichAttributes[2] := true;
		for id := 3 to self.dimension do	{ only the first two attributes are defined. }
			whichAttributes[id] := false;
		NEW(d);
		failNil(Handle(d));
		d.dimension := self.dimension;
		d.coordinates[1] := rectangle(rect).left;
		d.coordinates[2] := rectangle(rect).bottom;
		lowpP := pPoint(d.convertToParameterPoint);
		d.coordinates[1] := rectangle(rect).right;
		d.coordinates[2] := rectangle(rect).top;
		highpP := pPoint(d.convertToParameterPoint);
		d.free;
	end; { createQuery }

function dDimPoint.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := true;
	end; { canUseGridfile }
{$ENDC}

procedure getCoordinateVector (v: vector; dimension: dimensionRange; var coordinates: coordinateVector);
	var
		i: dimensionRange;
	begin
		for i := 1 to dimension do
			appendObjInfoExtended(v, stringOf('X', i : 1), coordinates[i], false);
	end;

procedure setCoordinateVector (v: vector; dimension: dimensionRange; var coordinates: coordinateVector);
	var
		i: dimensionRange;
	begin
		for i := 1 to dimension do
			coordinates[i] := objInfoExtended(v.elements^^[i]).e;
	end;

function dDimPoint.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(dimension);
		getCoordinateVector(v, dimension, coordinates);
		getInfo := v;
	end;

procedure dDimPoint.setInfo (o: obj);
	begin
		setCoordinateVector(vector(o), dimension, coordinates);
	end;


function dDimCircle.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(dimension + 1);
		getCoordinateVector(v, dimension, coordinates);
		appendObjInfoExtended(v, 'Radius', radius, false);
		getInfo := v;
	end;

procedure dDimCircle.setInfo (o: obj);
	begin
		setCoordinateVector(vector(o), dimension, coordinates);
		radius := objInfoExtended(vector(o).elements^^[dimension + 1]).e;
	end;


function dDimCircle.toCircle: circle;
	var
		c: circle;
	begin
		new(c);
		failNil(Handle(c));
		c.x := coordinates[1];
		c.y := coordinates[2];
		c.radius := radius;
		toCircle := c;
	end;

function dDimCircle.pointCover (n: longInt): obj;
	var
		c: circle;
	begin
		c := toCircle;
		pointCover := c.pointCover(n);
		c.free;
	end;

function dDimCircle.genericLessThan (o: obj): boolean;
	var
		d: dDimCircle;
	begin
		d := dDimCircle(o);
		genericLessThan := (dimension < d.dimension) | (dimension = d.dimension) & ((radius < d.radius) | (radius = d.radius) & lexicographicLessThan(coordinates, d.coordinates, dimension));
	end;

procedure dDimCircle.init;
	var
		i: dimensionRange;
	begin
		dimension := currentDimension;
		for i := 1 to dimension do
			coordinates[i] := 0;
		radius := undefinedRadius;
	end;

function dDimCircle.isSelected (where: obj): boolean;
	var
		c: circle;
	begin
		c := toCircle;
		isSelected := c.isSelected(where);
		c.free;
	end;

procedure dDimCircle.interactiveOutput;
	var
		c: circle;
	begin
		c := toCircle;
		c.interactiveOutput;
		c.free;
	end;

procedure dDimCircle.interactiveHighlight;
	var
		c: circle;
	begin
		c := toCircle;
		c.interactiveHighlight;
		c.free;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function dDimCircle.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 2;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure dDimCircle.textFileInput;
	var
		i: dimensionRange;
		dd: longInt;
	begin
		readTextLongInt(dd);
		dimension := dd;
		for i := 1 to dimension do
			readTextReal(coordinates[i]);
		readTextReal(radius);
	end;

procedure dDimCircle.textFileOutput;
	var
		i: dimensionRange;
	begin
		writeTextLongInt(dimension);
		for i := 1 to dimension do
			writeTextReal(coordinates[i]);
		writeTextReal(radius);
	end;

procedure dDimCircle.fileInput;
	var
		i: dimensionRange;
		b: Byte;
	begin
		readByte(b);
		dimension := dimensionRange(b);
		for i := 1 to dimension do
			readReal(coordinates[i]);
		readReal(radius);
	end;

procedure dDimCircle.fileOutput;
	var
		i: dimensionRange;
	begin
		writeByte(Byte(dimension));
		for i := 1 to dimension do
			writeReal(coordinates[i]);
		writeReal(radius);
	end;

procedure dDimCircle.randomChange;
	var
		i: dimensionRange;
	begin
		if (currentRandomConstraint <> nil) & member(currentRandomConstraint, rectangle) then
			with rectangle(currentRandomConstraint) do begin
				coordinates[1] := (right - left) * randomReal + left;
				for i := 2 to dimension do
					coordinates[i] := (top - bottom) * randomReal + bottom;
				radius := min(min(coordinates[1] - left, right - coordinates[1]), min(coordinates[2] - bottom, top - coordinates[2])) * randomReal;
			end;
	end;

function dDimCircle.equal (o: obj): boolean;
	var
		i: dimensionRange1;
	begin
		if (o <> nil) & member(o, dDimCircle) & (dDimCircle(o).dimension = dimension) & (dDimCircle(o).radius = radius) then begin
			i := 1;
			while (i <= dimension) & (dDimCircle(o).coordinates[i] = coordinates[i]) do
				i := i + 1;
			equal := i > dimension;
		end
		else
			equal := false;
	end;

procedure dDimCircle.centerOfGravity (var cx, cy: extended);
	begin
		cx := coordinates[1];
		cy := coordinates[2];
	end;

function dDimCircle.memberType (o: obj): boolean;
	begin
		memberType := member(o, dDimCircle) & (dDimCircle(o).dimension = dimension);
	end;

procedure dDimCircle.translate (dx, dy: extended);
	begin
		coordinates[1] := coordinates[1] + dx;
		coordinates[2] := coordinates[2] + dy;
	end;

procedure dDimCircle.rotate (xOrigin, yOrigin, alpha: extended);
	var
		x, y: extended;
	begin
		x := coordinates[1];
		y := coordinates[2];
		rotatePoint(xOrigin, yOrigin, alpha, x, y);
		coordinates[1] := x;
		coordinates[2] := y;
	end;

procedure dDimCircle.scale (lambda: extended);
	begin
		radius := radius * lambda;
	end;

procedure dDimCircle.reflect (a, b, c: extended);
	var
		x, y: extended;
	begin
		x := coordinates[1];
		y := coordinates[2];
		reflectPoint(a, b, c, x, y);
		coordinates[1] := x;
		coordinates[2] := y;
	end;

procedure dDimCircle.construct (px, py, newX, newY: integer);
	var
		dx, dy: real;
	begin
		coordinates[1] := px;
		coordinates[2] := py;
		dx := newX - px;
		dy := newY - py;
		radius := round(sqrt(sqr(dx) + sqr(dy)));
	end;

procedure dDimCircle.windowOutput;
	begin
{$IFC THINK_PASCAL }
		FrameArc(round(coordinates[2] - radius), round(coordinates[1] - radius), round(coordinates[2] + radius), round(coordinates[1] + radius), 0, 360);
{$ELSEC }
		FrameArc2(roundLong(coordinates[2] - radius), roundLong(coordinates[1] - radius), roundLong(coordinates[2] + radius), roundLong(coordinates[1] + radius), 0, 360);
{$ENDC }
	end;

procedure dDimCircle.windowToWorld;
	begin
		with currentFrame do begin
			coordinates[1] := (coordinates[1] - dx) / mx;
			coordinates[2] := (coordinates[2] - dy) / my;
		end;
	end;

function dDimCircle.getType: objectType;
	begin
		getType := dDimCircleType;
	end;

function dDimCircle.getTextType: Str255;
	begin
		getTextType := stringOf('D Circle(', dimension : 1, ', ', radius : 5 : 3, ')');
	end;

function dDimCircle.getRadius: markedRealPoint;
	var
		mp: markedRealPoint;
	begin
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.mark.value := stringOf(radius : 5 : 3);
		mp.x := coordinates[1];
		mp.y := coordinates[2];
		getRadius := mp;
	end;


	type
		dimensionRange0 = 0..maxDimension;
		matrix0 = array[dimensionRange0] of coordinateVector;
		matrix = array[dimensionRange, dimensionRange] of extended;

function distanceL2 (d: dimensionRange; var x, y: coordinateVector): real;
	var
		i: dimensionRange;
		s: extended;
	begin
		s := sqr(x[1] - y[1]);
		for i := 2 to d do
			s := s + sqr(x[i] - y[i]);
		distanceL2 := sqrt(s);
	end;

function dDimPoint.distance (p: dDimPoint): extended;
	begin
		assert(dimension = p.dimension, 'self.dimension = p.dimension');
		distance := distanceL2(dimension, coordinates, p.coordinates);
	end;

function scalarProduct (d: dimensionRange; a, b: coordinateVector): real;
	var
		s: extended;
		i: dimensionRange;
	begin
		s := a[1] * b[1];
		for i := 2 to d do
			s := s + a[i] * b[i];
		scalarProduct := s;
	end;

procedure subtract (d: dimensionRange; var x, y: coordinateVector);(* x:=x-y *)
	var
		i: dimensionRange;
	begin
		for i := 1 to d do
			x[i] := x[i] - y[i];
	end;

procedure add (d: dimensionRange; var x, y: coordinateVector; lambda: real);(* x:=x+lambda*y *)
	var
		i: dimensionRange;
	begin
		for i := 1 to d do
			x[i] := x[i] + lambda * y[i];
	end;

procedure createLinearSystem (d, k: dimensionRange; var p: matrix0; var a: matrix; var b: coordinateVector);
	var
		i, j: dimensionRange;
	begin
		for i := 1 to k do
			subtract(d, p[i], p[0]);
		for i := 1 to k do begin
			b[i] := -scalarProduct(d, p[i], p[i]);
			a[i, i] := -2 * b[i];								(* on diagonal *)
			for j := i + 1 to k do
				a[j, i] := 2 * scalarProduct(d, p[i], p[j]);	(* below diagonal *)
		end;
	end;

procedure combine (d, k: dimensionRange; var p: matrix0; var lambda, c: coordinateVector);
	var
		i: dimensionRange;
	begin
		c := p[0];
		for i := 1 to k do
			add(d, c, p[i], lambda[i]);
	end;

procedure solveLinearSystem (n: dimensionRange; var a: matrix; var bx: coordinateVector);
(* solve the symmetric positive definite system a*x=b using the Cholesky factorization. *)
(* The solution vector overwrites the right hand side *)
	var
		i, j, k: dimensionRange0;
		s: extended;
	begin
		for k := 1 to n do begin
			assert(a[k, k] > 0, 'a[k,k] > 0');
			a[k, k] := sqrt(a[k, k]);
			for i := k + 1 to n do begin
				a[i, k] := a[i, k] / a[k, k];
				for j := k + 1 to i do
					a[i, j] := a[i, j] - a[i, k] * a[j, k];
			end;
		end;
		for i := 1 to n do begin
			s := bx[i];
			for j := 1 to i - 1 do
				s := s - a[i, j] * bx[j];
			bx[i] := s / a[i, i];
		end;
		for i := n downto 1 do begin
			s := bx[i];
			for k := i + 1 to n do
				s := s + a[k, i] * bx[k];
			bx[i] := -s / a[i, i];
		end;
	end;

procedure dDimCircle.makeDisk (l: list);
	var
		p: matrix0;
		a: matrix;
		b: coordinateVector;
		k: dimensionRange0;

	procedure moveToP (x: dDimPoint);
		begin
			p[k] := x.coordinates;
			k := k + 1;
		end;

	begin
		if l.length = 0 then
			init
		else if l.length = 1 then begin
			dimension := dDimPoint(l.ith(1)).dimension;
			coordinates := dDimPoint(l.ith(1)).coordinates;
			radius := 0;
		end
		else begin
			k := 0;
			l.forAll(moveToP);
			k := k - 1;
			dimension := dDimPoint(l.ith(1)).dimension;
			createLinearSystem(dimension, k, p, a, b);
			solveLinearSystem(k, a, b);
			combine(dimension, k, p, b, coordinates);
			radius := distanceL2(dimension, coordinates, p[0]);
		end;
	end;

function dDimCircle.inCircle (p: dDimPoint): boolean;
	var
		radius2, s: real;
		i: dimensionRange1;
	begin
		if radius = undefinedRadius then
			inCircle := false
		else begin
			assert(p.dimension = dimension, 'p.dimension=dimension');
			radius2 := sqr(radius);
			s := sqr(coordinates[1] - p.coordinates[1]);
			i := 2;
			while (s <= radius2) & (i <= dimension) do begin
				s := s + sqr(coordinates[i] - p.coordinates[i]);
				i := i + 1;
			end;
			inCircle := (i > dimension) & (s <= radius2);
		end;
	end;

function dDimCircle.inCircleEps (p: dDimPoint; tolerance: extended): boolean;
	var
		radius2, s: extended;
		i: dimensionRange1;
	begin
		if radius = undefinedRadius then
			inCircleEps := false
		else begin
			assert(p.dimension = dimension, 'p.dimension=dimension');
			radius2 := sqr(radius * (1 + tolerance));
			s := sqr(coordinates[1] - p.coordinates[1]);
			i := 2;
			while (s <= radius2) & (i <= dimension) do begin
				s := s + sqr(coordinates[i] - p.coordinates[i]);
				i := i + 1;
			end;
			inCircleEps := (i > dimension) & (s <= radius2);
		end;
	end;

procedure dDimCircle.makeFromPointCircle (p: dDimPoint; c: dDimCircle);
	var
		i: dimensionRange;
		lambda, dpc: real;
	begin
		assert(p.dimension = c.dimension, 'p.dimension=c.dimension');
		dimension := c.dimension;
		dpc := distanceL2(dimension, p.coordinates, c.coordinates);
		lambda := c.radius / dpc;
		for i := 1 to dimension do
			coordinates[i] := (p.coordinates[i] + c.coordinates[i] + lambda * (c.coordinates[i] - p.coordinates[i])) / 2;
		radius := (c.radius + dpc) / 2;
	end;

procedure dDimPoint.randomInCircle (c: dDimCircle);
	const
		dimensionCutoff = 4;
	var
		i: dimensionRange0;
		d, dx, x1, x2, sqrSum, s: extended;

	procedure normalDeviate (var x1, x2, sqrSum: extended);
		var
			v1, v2, s: extended;
		begin
			repeat
				v1 := 2 * randomReal - 1;
				v2 := 2 * randomReal - 1;
				s := sqr(v1) + sqr(v2);
			until s < 1;
			sqrSum := -2 * ln(s);
			s := sqrt(sqrSum / s);
			x1 := v1 * s;
			x2 := v2 * s;
		end;

	begin
		assert(dimension = c.dimension, 'dimension = c.dimension');
		if dimension < dimensionCutoff then begin
			d := 0;
			i := 0;
			repeat
				dx := 2 * randomReal - 1;
				d := d + sqr(dx);
				if d < 1.0 then begin
					i := i + 1;
					coordinates[i] := c.coordinates[i] + c.radius * dx;
				end
				else begin
					d := 0;
					i := 0;
				end;
			until i = dimension;
		end
		else begin
			s := 0;
			for i := 1 to dimension div 2 do begin
				normalDeviate(x1, x2, sqrSum);
				s := s + sqrSum;
				coordinates[2 * i - 1] := x1;
				coordinates[2 * i] := x2;
			end;
			if odd(dimension) then begin
				normalDeviate(x1, x2, sqrSum);
				coordinates[dimension] := x1;
				s := s + sqr(x1);
			end;
			if dimension = 2 then
				s := c.radius * sqrt(randomReal / s)
			else
				s := c.radius * exp(ln(randomReal) / dimension) / sqrt(s);
			for i := 1 to dimension do
				coordinates[i] := coordinates[i] * s + c.coordinates[i];
		end;
	end;

function dDimCircle.randomPoint: dDimPoint;
	var
		p: dDimPoint;
	begin
		new(p);
		failNil(Handle(p));
		p.dimension := dimension;
		p.randomInCircle(self);
		randomPoint := p;
	end;

	const
		MakeDiskC = 1;

procedure dDimCircle.describe;
	begin
		with currentArguments do
			if (length >= 2) & member(ith(1), dDimPoint) & currentArgumentsSameType & (length <= dDimPoint(ith(1)).dimension + 1) then
				menuEntry('Make Disk', MakeDiskC);
	end;

procedure dDimCircle.execute (item: integer);

	function callMakeDisk (l: list): dDimCircle;
		var
			c: dDimCircle;
		begin
			new(c);
			failNil(Handle(c));
			c.makeDisk(l);
			callMakeDisk := c;
		end;

	begin
		case item of
			MakeDiskC: 
				currentResult.append(callMakeDisk(currentArguments));
		end;
	end;

end.
