(* XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit dDimPointVector;

interface

uses
	pointVector;

type
	minMax = record
			min, max: seqIndex;
			minValue, maxValue: real;
			delta: extended;
		end;

	dDimPointVector = object(homogenVector)
			function equal (o: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;

			procedure minMaxInDimension (whichDimension: dimensionRange; var minMaxPair: minMax);
			procedure largestPairToFront;

			function minimalDisk (tolerance: extended; initialize, verbose: boolean): dDimCircle;
(* Computes the minimal area disk that contains the given points. The radius is correct within		*)
(* a factor of (1 + 'tolerance'), i.e. the radius is exact if tolerance = 0'. If 'initialize = TRUE' the	*)
(* algorithm is initialized with the pair of points that have the largest coordinate difference in		*)
(* some direction.																									*)

			function simpleMinimalDisk (shuffle, initialize: boolean): dDimCircle;
(* Computes a disk containing all given points using a heuristic that is guaranteed to produce a		*)
(* circle whose radius is at most twice as large as the radius of the minimal area disk. If 			*)
(* 'shuffle = TRUE' the points are randomly shuffled before the algorithm starts since it depends of	*)
(* the order of the points. The parameter 'tolerance' has the same meaning as in the previous		*)
(* method.																											*)

			function worstSimpleMinimalDisk: dDimCircle;
(* Rearranges the points in such a way that the worst case order for the previous enclosing disk	*)
(* algorithm is achieved. This algorithm has running time proportional to the factorial of the			*)
(* number of given points. 																						*)

			function containedInCircle (c: dDimCircle): boolean;
(* Tests whether all given points lie inside the d-dimensional circle 'c'. 									*)

			function projectOnXY: pointVector;
(* Projects the given points on the x-y-plane and produces a collection of 2-dimensional points		*)
(* (type 'point2d'). 																								*)

			function projectOnXiXj (i, j: seqIndex): pointVector;
(* Projects the given points on the xi-xj-plane and produces a collection of 2-dimensional points		*)
(* (type 'point2d'). 																								*)
(*************************************************************************** *)
(*                                                                          *)
(*           INS-POS : PointAlgorithms                                      *)
(*                                                                          *)
(*                                                                          *)
(* please declare your point algorithm here                                 *)
(* mask:                                                                    *)
(*      procedure MyAlgor(...);                                             *)
(*                                                                          *)
(*************************************************************************** *)
(*    AENDERUNG  : 22.12.92                                                 *)
			procedure ClosestPairSimple;
(*    AENDERUNG  : 22.12.92                                                 *)
(* Algorithmus 1 :                                                          *)
(* Die Prozedur ClosestPairSimple ist ein O(n*n) - Algorithmus. Bei jedem   *)
(* neueingefuegten Punkt p1 werden die Abstaende newDelta zu allen bereits  *)
(* vorhandenen Punkten berechnet. Existiert ein newDelta, das kleiner als   *)
(* Delta ist, so wird Delta := newDelta,P :=p1 und Q:=p2, wobei p2 der zu   *)
(* p1 naechste Punkt ist.                                                   *)

			procedure ClosestPairBinaryTree;
(* Aenderung :  14.01.93 *)
(* Algorithmus 2 :                                                          *)
(* Die Prozedur ClosestPairBinaryTree ist ein O(n*n) - Algorithmus. Zuerst  *)
(* werden die ersten beiden Punkte in 'InitStructure' als Closest Pair      *)
(* definiert und in die Datenstruktur 'BinarykBoxTree' eingefuegt. Der      *)
(* BinarykBoxTree liefert eine hierarchische Unterteilung des Raumes in     *)
(* k - Boxen, die in den Knoten des Baumes gespeichert sind. Die Blaetter   *)
(* speichern zusaetzlich die Punkte, die in der zugehoerigen k - Box liegen *)
(* Um nun festzustellen, ob ein neueingefuegter Punkt new einen geringeren  *)
(* Abstand zu einem anderen Punkt als 'Delta' (Abstand Closest Pair P,Q)    *)
(* hat, muessen nur die k - Boxen ueberprueft werden, die Punkte enthalten  *)
(* koennen, welche hoechstens Delta von new entfernt sind. Das sind maximal *)
(* 3 HOCH k k - Boxen, also 3 HOCH k Blaetter, die kontrolliert  werden     *)
(* muessen. Das Ueberpruefen der Punkte in einem Blatt kostet also  O(1).   *)
(* Da der BinarykBoxTree aber zu einer linearen Liste entarten kann, kostet *)
(* es O(n) (n = Anzahl Punkte) ein Blatt zu finden.                         *)

			procedure ClosestPairBetaTree1;
(* Aenderung :  29.01.93 *)
(* Algorithmus 3 :                                                          *)
(* Die Prozedur ClosestPairBetaTree ist ein O(n*log(n)) - Algorithmus. Der  *)
(* Algorithmus arbeitet genauso wie der ClosestPairBinaryTree - Algorithmus. *)
(* mit der Ausnahme, dass der BetaTree balanciert wird. Um dies zu erreichen *)
(* wird in jedem Knoten das Gewicht des Teilbaumes mitgefuehrt, dessen      *)
(* Wurzel der Knoten ist. Wird nun ein Teilbaum zu schwer, dann wird er mit *)
(* Hilfe eines Binaerbaumes neubalanciert.                                  *)

			procedure ClosestPairBetaTree2;
(*  Aenderung : 20.01.93 *)
(* Algorithmus 4 :                                                          *)
(* Wie Algorithmus 3, nur anderer Balancierungs - Algorithmus.                  *)

			procedure MLP;
(*************************************************************************** *)
(* Put the implementation of your algorithm into an other                   *)
(*of your files. This makes Geobench not to compile                         *)
(*whole source every time.                                                  *)
(*                                                                          *)
(* WARNING: Don't forget the needed modification at                         *)
(*        label "PointAlgoMenu"                                             *)
(*************************************************************************** *)

		end;

var
	ricDefaultDimension, defaultTolerance, defaultNumberOfVertices, defaultXi, defaultXj: extended;
	miniDiskCounter, inCircleCounter: longInt;

procedure initdDimPointVector;


implementation

{$SETC AminimalDisk = TRUE }

type
	destroyAction = (doNothing, doFree, doDispose);

const
	MiniDiskC = 1;
	SimpleMiniDiskC = 2;
	InCircleC = 3;
	ProjectOnXYC = 4;
	RandomInCircleC = 5;
	MiniDiskToleranceC = 6;
	MiniDiskSeriesC = 7;
	SimpleMiniDiskSeriesC = 8;
	SimpleMiniDiskNoShuffleC = 9;
	WorstSimpleMiniDiskC = 10;
	NGonC = 11;
	SimpleMiniDiskSeriesNoInitC = 12;
	MiniDiskSeriesNoInitC = 13;
	ProjectOnXiXjC = 14;
	SimpleC = 15;
	BinaryC = 16;
	BetaC1 = 17;
	BetaC2 = 18;
	SpecialC = 19;

{***********************************************}

procedure MakeDimPointVector (item: INTEGER; var dpv: dDimPointVector; var destroy: destroyAction);

	function todDimPointVector (x: vector): dDimPointVector;
		var
			dpv: dDimPointVector;
		begin
			new(dpv);
			failNil(Handle(dpv));
			dpv.length := x.length;
			dpv.elements := x.elements;
			destroy := doDispose;
			todDimPointVector := dpv;
		end;

	function ListToVector: dDimPointVector;
		var
			dpv: dDimPointVector;
		begin
			new(dpv);
			failNil(Handle(dpv));
			dpv.initFromList(currentArguments);
			destroy := doFree;
			ListToVector := dpv;
		end;

	begin { MakeDimPointVector }
		if currentArguments.length > 1 then begin
			if item = InCircleC then
				if member(currentArguments.ith(1), vector) then
					dpv := todDimPointVector(vector(currentArguments.ith(1)))
				else
					dpv := todDimPointVector(vector(currentArguments.ith(2)))
			else
				dpv := ListToVector
		end
		else if member(currentArguments.ith(1), dDimPointVector) then begin
			dpv := dDimPointVector(currentArguments.ith(1));
			destroy := doNothing;
		end
		else if member(currentArguments.ith(1), vector) then
			dpv := todDimPointVector(vector(currentArguments.ith(1)))
		else
			dpv := ListToVector;
	end; { MakeDimPointVector }

{***********************************************}

function dDimPointVector.memberType (o: obj): boolean;
	begin
		memberType := member(o, dDimPointVector) & memberBaseType(o);
	end;

function dDimPointVector.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, dDimPointVector) & inherited equal(o);
	end;

function dDimPointVector.getType: objectType;
	begin
		getType := dDimPointVectorType;
	end;

function dDimPointVector.getTextType: Str255;
	begin
		if length = 0 then
			getTextType := stringOf('D Point Vector(', length : 1, ')')
		else
			getTextType := stringOf('D Point Vector(', dDimPoint(elements^^[1]).dimension : 1, ', ', length : 1, ')');
	end;

function md (l: doubleList; n: seqIndex0; r: list; d: dimensionRange; tolerance: extended): dDimCircle;
	var
		disk: dDimCircle;
		e: doubleElementH;
		i: seqIndex0;
	begin
		miniDiskCounter := miniDiskCounter + 1;
		new(disk);
		failNil(Handle(disk));
		disk.makeDisk(r);
		if r.length <= d then begin
			e := l.first;
			for i := 1 to n do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					disk.free;
					md := nil;
					exit(md);
				end;
{$ENDC }
{$IFC AnimationEnabled AND AMinimalDisk }
				if animationFlag[AMinimalAreaDiskItem] & (r.length = 0) then
					with e^^.value do begin
						interactiveFlash;
						interactiveOutput;
						interactiveHighlight;
					end;
{$ENDC }
				inCircleCounter := inCircleCounter + 1;
				if not disk.inCircleEps(dDimPoint(e^^.value), tolerance) then begin
{$IFC AnimationEnabled AND AMinimalDisk }
					if animationFlag[AMinimalAreaDiskItem] & (disk.radius > 0) & (r.length = 0) then
						disk.interactiveHighlight;
{$ENDC }
					r.push(e^^.value);
					disk.free;
					disk := md(l, i - 1, r, d, tolerance);
{$IFC AnimationEnabled }
					if disk = nil then begin
						md := nil;
						exit(md);
					end;
{$ENDC }
{$IFC AnimationEnabled AND AMinimalDisk }
					if animationFlag[AMinimalAreaDiskItem] & (disk.radius > 0) then
						if r.length = 1 then begin
							disk.interactiveHighlight;
							waitForClick(animationFlag[AMinimalAreaDiskItem]);
						end
						else
							disk.interactiveOutput;
{$ENDC }
					r.pop;
					l.moveToFront(e);
				end;
				e := e^^.succ
			end;
		end;
		md := disk;
	end;

function dDimPointVector.minimalDisk (tolerance: extended; initialize, verbose: boolean): dDimCircle;
	var
		e: list;
		l: doubleList;

	procedure moveToL (x: obj);
		begin
			l.append(x);
		end;

	begin
{$IFC AnimationEnabled AND AMinimalDisk }
		if animationFlag[AMinimalAreaDiskItem] then
			interactiveOutput;
{$ENDC }
		randomShuffleAll;
		if initialize then
			largestPairToFront;
		e := newEmptyList;
		miniDiskCounter := 0;
		inCircleCounter := 0;
		new(l);
		failNil(Handle(l));
		l.init;
		forAll(moveToL);
		minimalDisk := md(l, length, e, dDimPoint(elements^^[1]).dimension, tolerance);
		l.free;
		e.free;
		if verbose then
			displayInfoString(stringOf('#miniDisk=', miniDiskCounter : 1, ', #inCircle = ', inCircleCounter : 1));
	end;

function dDimPointVector.worstSimpleMinimalDisk: dDimCircle;
	var
		maxCircle: dDimCircle;
		worstCase: dDimPointVector;
		i: seqIndex;

	procedure experiment (v: vector);
		var
			c: dDimCircle;
		begin
			c := dDimPointVector(v).simpleMinimalDisk(false, false);
{$IFC AnimationEnabled }
			if c = nil then begin
				if maxCircle <> nil then begin
					worstCase.free;
					maxCircle.free;
				end;
				worstSimpleMinimalDisk := nil;
				exit(worstSimpleMinimalDisk);
			end;
{$ENDC }
			if maxCircle = nil then begin
				maxCircle := c;
				worstCase := dDimPointVector(v.duplicate);
			end
			else if c.radius > maxCircle.radius then begin
				maxCircle.free;
				maxCircle := c;
				worstCase.free;
				worstCase := dDimPointVector(v.duplicate);
			end
			else
				c.free;
		end;

	begin
		if length = 0 then
			worstSimpleMinimalDisk := nil
		else begin
			maxCircle := nil;
			forAllPermutations(length, experiment);
			for i := 1 to length do
				elements^^[i] := worstCase.elements^^[i];
			worstCase.free;
			worstSimpleMinimalDisk := maxCircle;
		end;
	end;

function dDimPointVector.simpleMinimalDisk (shuffle, initialize: boolean): dDimCircle;
	var
		c: dDimCircle;
		i: seqIndex0;
		l: list;
	begin
{$IFC AnimationEnabled AND AMinimalDisk }
		if animationFlag[AMinimalAreaDiskItem] then
			interactiveOutput;
{$ENDC }
		if shuffle then
			randomShuffleAll;
		if initialize then
			largestPairToFront;
		l := newEmptyList;
		l.append(elements^^[1]);
		l.append(elements^^[2]);
		new(c);
		failNil(Handle(c));
		c.makeDisk(l);
		l.free;
{$IFC AnimationEnabled AND AMinimalDisk }
		if animationFlag[AMinimalAreaDiskItem] then begin
			c.interactiveHighlight;
			waitForClick(animationFlag[AMinimalAreaDiskItem]);
		end;
{$ENDC }
		for i := 3 to length do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				c.free;
				simpleMinimalDisk := nil;
				exit(simpleMinimalDisk);
			end;
{$ENDC }
{$IFC AnimationEnabled AND AMinimalDisk }
			if animationFlag[AMinimalAreaDiskItem] then
				elements^^[i].interactiveFlash;
{$ENDC }
			if not c.inCircle(dDimPoint(elements^^[i])) then begin
{$IFC AnimationEnabled AND AMinimalDisk }
				if animationFlag[AMinimalAreaDiskItem] then
					c.interactiveHighlight;
{$ENDC }
				c.makeFromPointCircle(dDimPoint(elements^^[i]), c);
{$IFC AnimationEnabled AND AMinimalDisk }
				if animationFlag[AMinimalAreaDiskItem] then begin
					c.interactiveHighlight;
					waitForClick(animationFlag[AMinimalAreaDiskItem]);
				end;
{$ENDC }
			end;
		end;
		simpleMinimalDisk := c;
	end;

function dDimPointVector.containedInCircle (c: dDimCircle): boolean;
	var
		i: seqIndex1;
	begin
		i := 1;
		while (i <= length) & c.inCircleEps(dDimPoint(elements^^[i]), 0.001) do
			i := i + 1;
		containedInCircle := i > length;
	end;

function checkDim (x: extended): boolean;
	begin
		checkDim := (trunc(x) = x) & (1 <= x) & (x <= 6);
	end;

procedure dDimPointVector.MLP;
	var
		i, j, k: longInt;
		default: extended;
		Point: dDimPoint;
		result: dDimPointVector;
	begin
		default := 2;
		if getUserParameter1('Dimension <=  6', default, default, @checkDim) then begin
			result := dDimPointVector(duplicateAll);
{$IFC THINK_PASCAL }
			k := trunc(default);
{$ELSEC }
			k := truncLong(default);
{$ENDC }
			for i := 1 to length do begin
				for j := 2 to k do begin
					Point := dDimPoint(result.elements^^[i]);
					Point.dimension := k;
					Point.coordinates[j] := 100;
				end;
				Point.coordinates[1] := i;
			end;
			currentResult.append(result);
		end;
		free;
	end;

function dDimPointVector.projectOnXY: pointVector;
	begin
		projectOnXY := projectOnXiXj(1, 2);
	end;

function dDimPointVector.projectOnXiXj (i, j: seqIndex): pointVector;
	var
		pv: pointVector;

	procedure project (x: dDimPoint);
		var
			p: realPoint;
		begin
{$IFC AnimationEnabled }
			if userAbort then begin
				pv.freeAll;
				projectOnXiXj := nil;
				exit(projectOnXiXj);
			end;
{$ENDC }
			new(p);
			failNil(Handle(p));
			p.x := x.coordinates[i];
			p.y := x.coordinates[j];
			pv.append(p);
		end;

	begin
		pv := newPointVector(length);
		forAll(project);
		projectOnXiXj := pv;
	end;

procedure dDimPointVector.minMaxInDimension (whichDimension: dimensionRange; var minMaxPair: minMax);
	var
		t: real;
		i: seqIndex;
	begin
		with minMaxPair do begin
			min := 1;
			minValue := dDimPoint(elements^^[1]).coordinates[whichDimension];
			max := 1;
			maxValue := minValue;
			for i := 2 to length do begin
				t := dDimPoint(elements^^[i]).coordinates[whichDimension];
				if t < minValue then begin
					min := i;
					minValue := t;
				end
				else if t > maxValue then begin
					max := i;
					maxValue := t;
				end;
			end;
			delta := dDimPoint(elements^^[max]).distance(dDimPoint(elements^^[min]));
		end;
	end;

procedure dDimPointVector.largestPairToFront;
	var
		minMaxPair, newMinMaxPair: minMax;
		d: dimensionRange;
	begin
		if length > 2 then begin
			minMaxInDimension(1, minMaxPair);
			for d := 2 to dDimPoint(elements^^[1]).dimension do begin
				minMaxInDimension(d, newMinMaxPair);
				if newMinMaxPair.delta > minMaxPair.delta then
					minMaxPair := newMinMaxPair;
			end;
			swap(1, minMaxPair.min);
			if minMaxPair.max = 1 then
				swap(2, minMaxPair.min)
			else
				swap(2, minMaxPair.max);
		end;
	end;

function numberOk (x: extended): boolean;
	begin
		numberOk := (1 <= x) & (x <= 20000) & (x = trunc(x));
	end;

function dimOk (x: extended): boolean;
	begin
		dimOk := (2 <= x) & (x <= maxDimension) & (x = trunc(x));
	end;

function randomInCircle (c: circle): dDimPointVector;
	var
		dc: dDimCircle;
		i, n: longInt;
		d: dimensionRange;
		dpv: dDimPointVector;
	begin
		if getUserParameter2('Number of points:', 'Dimension:', numberOfObjects, ricDefaultDimension, numberOfObjects, ricDefaultDimension, @numberOk, @dimOk) then begin
{$IFC THINK_PASCAL }
			n := trunc(numberOfObjects);
			d := trunc(ricDefaultDimension);
{$ELSEC }
			n := truncLong(numberOfObjects);
			d := truncLong(ricDefaultDimension);
{$ENDC }
			currentVectorLength := n;
			new(dpv);
			failNil(Handle(dpv));
			dpv.init;
			currentDimension := d;
			new(dc);
			failNil(Handle(dc));
			dc.init;
			dc.radius := c.radius;
			dc.coordinates[1] := c.x;
			dc.coordinates[2] := c.y;
			for i := 1 to n do
				dpv.append(dc.randomPoint);
			dc.free;
			randomInCircle := dpv;
		end
		else
			randomInCircle := nil;
	end;

procedure dDimPointVector.describe;

	function isdDimPointVector (v: vector): boolean;
		begin
			isdDimPointVector := (v.length > 1) & (member(v, dDimPointVector) | member(v.elements^^[1], dDimPoint));
		end;

	begin
		with currentArguments do
			if (length = 1) & currentArguments1IsHomogenVector & isdDimPointVector(vector(ith(1))) | (length > 1) & currentArgumentsSameType & member(first^^.value, dDimPoint) then begin
				beginSubMenu('Minimal Area Disk');
				menuEntry('Randomized', MiniDiskC);
				menuEntry('Randomized Tolerance...', MiniDiskToleranceC);
				menuEntry('Randomized Series...', MiniDiskSeriesC);
				menuEntry('Randomized Series No Init...', MiniDiskSeriesNoInitC);
				menuEntry('Approximate', SimpleMiniDiskC);
				menuEntry('Approximate Worst Case', WorstSimpleMiniDiskC);
				menuEntry('Approximate No Shuffle', SimpleMiniDiskNoShuffleC);
				menuEntry('Approximate Series...', SimpleMiniDiskSeriesC);
				menuEntry('Approximate Series No Init...', SimpleMiniDiskSeriesNoInitC);
				endSubMenu;
				menuEntry('Project On XY', ProjectOnXYC);
				menuEntry('Project On XI XJ...', ProjectOnXiXjC);
				if length > 1 then begin
					beginSubMenu('Closest Pair On-Line');
					menuEntry('Simple', SimpleC);
					menuEntry('Binary K-Box Tree', BinaryC);
					menuEntry('Beta K-Box Tree 1', BetaC1);
					menuEntry('Beta K-Box Tree 2', BetaC2);
					menuEntry('MLP', SpecialC);
					endSubMenu;
				end;
			end
			else if (length = 2) & (currentArguments1IsHomogenVector & isdDimPointVector(vector(ith(1))) & member(ith(2), dDimCircle) | currentArguments2IsHomogenVector & isdDimPointVector(vector(ith(2))) & member(ith(1), dDimCircle)) then
				menuEntry('In Circle', InCircleC)
			else if (length = 1) & member(ith(1), circle) then begin
				menuEntry('Random In Circle...', RandomInCircleC);
				menuEntry('Regular N-Gon...', NGonC);
			end;
	end;

	var
		maxD: dimensionRange;

function checkDimension (x: extended): boolean;
	begin
		checkDimension := (trunc(x) = x) & (1 <= x) & (x <= maxD);
	end;

procedure dDimPointVector.execute (item: integer);
	var
		dpv: dDimPointVector;
		destroy: destroyAction;

	function callContainedInCircle (p: dDimPointVector; c: dDimCircle): markedRealPoint;
		var
			mp: markedRealPoint;
		begin
			new(mp);
			failNil(Handle(mp));
			mp.init;
			mp.mark.value := stringOf(p.containedInCircle(c));
			mp.x := c.coordinates[1];
			mp.y := c.coordinates[2];
			callContainedInCircle := mp;
		end;

	procedure callMiniDisk (c: dDimCircle);
		begin
			if c <> nil then begin
				currentResult.append(c);
				currentResult.append(c.getRadius);
			end;
		end;

	procedure callMiniDiskTolerance (dpv: dDimPointVector; initialize: boolean);
		begin
			if getUserParameter1('Tolerance:', defaultTolerance, defaultTolerance, nil) then
				callMiniDisk(dpv.minimalDisk(defaultTolerance, initialize, true));
		end;

	procedure callMiniDiskSeries (dpv: dDimPointVector; initialize: boolean);
		var
			inCircleSum, miniDiskSum: extended;
			i: longInt;
			c: dDimCircle;
		begin
			if getUserParameter1('Number of experiments:', numberOfExperiments, numberOfExperiments, @isNatural) then begin
				inCircleSum := 0;
				miniDiskSum := 0;
				c := nil;
{$IFC THINK_PASCAL }
				for i := 1 to trunc(numberOfExperiments) do begin
{$ELSEC }
				for i := 1 to truncLong(numberOfExperiments) do begin
{$ENDC }
					if c <> nil then
						c.free;
					c := dpv.minimalDisk(0, initialize, false);
{$IFC AnimationEnabled }
					if c = nil then
						exit(callMiniDiskSeries);
{$ENDC }
					inCircleSum := inCircleSum + inCircleCounter;
					miniDiskSum := miniDiskSum + miniDiskCounter;
				end;
				callMiniDisk(c);
				displayInfoString(stringOf('avg. miniDisk=', miniDiskSum / numberOfExperiments : 7 : 1, ',  avg. inCircle = ', inCircleSum / numberOfExperiments : 7 : 1));
			end;
		end;

	procedure callSimpleMiniDiskSeries (dpv: dDimPointVector; initialize: boolean);
		var
			quotient, cMinRadius: extended;
			i: longInt;
			j: seqIndex0;
			c, cWorst: dDimCircle;
			dpvWorst: dDimPointVector;
		begin
			if getUserParameter1('Number of experiments:', numberOfExperiments, numberOfExperiments, @isNatural) then begin
				c := dpv.minimalDisk(0, true, false);
{$IFC AnimationEnabled }
				if c = nil then
					exit(callSimpleMiniDiskSeries);
{$ENDC }
				callMiniDisk(c);
				startTime := TickCount;
				cMinRadius := c.radius;
				quotient := 1.0;
				c := nil;
				cWorst := nil;
				dpvWorst := nil;
{$IFC THINK_PASCAL }
				for i := 1 to trunc(numberOfExperiments) do begin
{$ELSEC }
				for i := 1 to truncLong(numberOfExperiments) do begin
{$ENDC }
					if c <> nil then
						c.free;
					c := dpv.simpleMinimalDisk(true, initialize);
{$IFC AnimationEnabled }
					if c = nil then begin
						if cWorst <> nil then
							cWorst.free;
						if dpvWorst <> nil then
							dpvWorst.free;
						exit(callSimpleMiniDiskSeries);
					end;
{$ENDC }
					if c.radius / cMinRadius > quotient then begin
						if cWorst <> nil then
							cWorst.free;
						if dpvWorst <> nil then
							dpvWorst.free;
						dpvWorst := dDimPointVector(dpv.duplicate);
						cWorst := c;
						quotient := c.radius / cMinRadius;
						displayInfoString(stringOf(i : 1, '. Off by ', 100 * (quotient - 1) : 5 : 2, '%'));
						c := nil;
					end;
				end;
				if dpvWorst <> nil then begin
					for j := 1 to dpv.length do
						dpv.elements^^[j] := dpvWorst.elements^^[j];
					dpvWorst.free;
				end;
				if c <> nil then
					c.free;
				if cWorst <> nil then
					callMiniDisk(cWorst);
			end;
		end;

	function makeNGon (c: circle): dDimPointVector;
		var
			l: seqIndex;
			result: dDimPointVector;
			i: seqIndex0;
			p: dDimPoint;
			alpha, cosAlpha, sinAlpha, cosSum, sinSum, newCosSum: extended;
		begin
			if getUserParameter1('Number of vertices:', defaultNumberOfVertices, defaultNumberOfVertices, @isSeqIndex) then begin
{$IFC THINK_PASCAL }
				l := trunc(defaultNumberOfVertices);
{$ELSEC }
				l := truncLong(defaultNumberOfVertices);
{$ENDC }
				currentVectorLength := l;
				new(result);
				failNil(Handle(result));
				result.init;
				alpha := 2 * pi / l;
				cosAlpha := cos(alpha);
				sinAlpha := sin(alpha);
				cosSum := 1;
				sinSum := 0;
				for i := 1 to l do begin
{$IFC AnimationEnabled }
					if userAbort then begin
						result.freeAll;
						makeNGon := nil;
						exit(makeNGon);
					end;
{$ENDC }
					new(p);
					failNil(Handle(p));
					p.dimension := 2;
					p.coordinates[1] := c.x + c.radius * cosSum;
					p.coordinates[2] := c.y + c.radius * sinSum;
					result.append(p);
					newCosSum := cosAlpha * cosSum - sinAlpha * sinSum;
					sinSum := sinAlpha * cosSum + cosAlpha * sinSum;
					cosSum := newCosSum;
				end;
				makeNGon := result;
			end
			else
				makeNGon := nil;
		end;

	function callProjectOnXiXj (dpv: dDimPointVector): pointVector;
		begin
			maxD := dDimPoint(dpv.elements^^[1]).dimension;
			if defaultXi > maxD then
				defaultXi := maxD;
			if defaultXj > maxD then
				defaultXj := maxD;
			if getUserParameter2(stringOf('Dimension I (1 <= I <= ', maxD : 1, ')'), stringOf('Dimension I (1 <= J <= ', maxD : 1, ')'), defaultXi, defaultXj, defaultXi, defaultXj, @checkDimension, @checkDimension) then
{$IFC THINK_PASCAL }
				callProjectOnXiXj := dpv.projectOnXiXj(trunc(defaultXi), trunc(defaultXj))
{$ELSEC }
				callProjectOnXiXj := dpv.projectOnXiXj(truncLong(defaultXi), truncLong(defaultXj))
{$ENDC }
			else
				callProjectOnXiXj := nil;
		end;

	begin
		if item = RandomInCircleC then
			currentResult.appendNonNil(randomInCircle(circle(currentArguments.ith(1))))
		else if item = NGonC then
			currentResult.appendNonNil(makeNGon(circle(currentArguments.ith(1))))
		else if (item = SimpleC) | (item = BinaryC) | (item = BetaC1) | (item = BetaC2) | (item = SpecialC) then begin
			MakeDimPointVector(item, dpv, destroy);
			case item of
				SimpleC: 
					dpv.ClosestPairSimple;
				BinaryC: 
					dpv.ClosestPairBinaryTree;
				BetaC1: 
					dpv.ClosestPairBetaTree1;
				BetaC2: 
					dpv.ClosestPairBetaTree2;
				SpecialC: 
					dpv.MLP
			end;
		end
(*************************************************************************** *)
		else begin
			MakeDimPointVector(item, dpv, destroy);
			case item of
				MiniDiskC: 
					callMiniDisk(dpv.minimalDisk(0, true, true));
				MiniDiskToleranceC: 
					callMiniDiskTolerance(dpv, true);
				SimpleMiniDiskC, SimpleMiniDiskNoShuffleC: 
					callMiniDisk(dpv.simpleMinimalDisk(item = SimpleMiniDiskC, item = SimpleMiniDiskC));
				WorstSimpleMiniDiskC: 
					callMiniDisk(dpv.worstSimpleMinimalDisk);
				MiniDiskSeriesC, MiniDiskSeriesNoInitC: 
					callMiniDiskSeries(dpv, item = MiniDiskSeriesC);
				SimpleMiniDiskSeriesC, SimpleMiniDiskSeriesNoInitC: 
					callSimpleMiniDiskSeries(dpv, item = SimpleMiniDiskSeriesC);
				InCircleC: 
					if member(currentArguments.ith(1), dDimCircle) then
						currentResult.append(callContainedInCircle(dpv, dDimCircle(currentArguments.ith(1))))
					else
						currentResult.append(callContainedInCircle(dpv, dDimCircle(currentArguments.ith(2))));
				ProjectOnXYC: 
					currentResult.appendNonNil(dpv.projectOnXY);
				ProjectOnXiXjC: 
					currentResult.appendNonNil(callProjectOnXiXj(dpv));
			end;
			if destroy = doFree then
				dpv.free
			else if destroy = doDispose then
				dispose(dpv);
		end;
	end;

procedure initdDimPointVector;
	var
		x: dDimPointVector;
	begin
		new(x);
		failNil(Handle(x));
		prototypes[dDimPointVectorType] := x;
		ricDefaultDimension := 2;
		defaultTolerance := 0.001;
		defaultNumberOfVertices := 17;
		defaultXi := 1;
		defaultXj := 2;
	end;

end.
