(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit dialogEngine;

interface

uses
	sequence;

type
	BooleanP = ^boolean;
	IntegerP = ^integer;
	ExtendedP = ^extended;
	Str255P = ^Str255;

{$IFC NOT THINK_PASCAL }
	booleanFuncT = function(): boolean;
 	extendedBooleanFuncT = function(x: extended): boolean;
{$ENDC }

	interactiveHandler = object(obj)
			isVisible: ProcPtr;
			dialog: DialogPtr;
			item: integer;
			visible, oldVisible: boolean;
			procedure show;
			procedure showValue;
			procedure update (clickedItem: integer);
			procedure restore;
			procedure setDefault;
			function isAcceptable: boolean;
			procedure makeVisible (yes: boolean);
			procedure beginVisible;
			procedure endVisible;
			procedure install (d: DialogPtr);
		end;

function processDialogBox (resource, defaultButton, revertButton, setButton: integer; function activateOkButton: boolean; handlers: vector): boolean;
(* The dialog box must confirm to the following standards: ok=1, cancel=2. The resource number					*)
(*	is 'resource'. 'handlers' is a vector of elements created by the following functions.	 If there is not a				*)
(*	default button, then set 'defaultButton' to a value less than or equal to 2.												*)

function getBooleanHandler (item: integer; value: BooleanP; default: boolean; isVisible: ProcPtr): interactiveHandler;
(*	Check box item with number 'item' corresponds to the boolean pointed to by 'value'. The default is 'default'	*)
(*	and 'function isVisible:boolean' determines whether the checkbox should be visible or not. If 'isVisible = nil'	*)
(*	then the item is always visible.																								*)

function getEnumerationHandler (item: integer; value: IntegerP; default: integer; isVisible: ProcPtr; highValue: integer): interactiveHandler;
(*	Radio buttons with numbers 'item' to 'item+highValue' corresponding to an integer pointed to by 'value' with	*)
(*	has the values '0' to 'highValue'. 																								*)

function s (x: longInt; l: list): list;

function getEnumerationExtendedHandler (radioButtons: list; value: IntegerP; default: integer; isVisible: ProcPtr): interactiveHandler;

function t (s: Str255; l: list): list;

function getEnumerationMenuHandler (item: integer; value: IntegerP; default: integer; isVisible: ProcPtr; menuItems: list): interactiveHandler;

function getExtendedHandler (item: integer; value: ExtendedP; default: extended; isVisible: ProcPtr; accept: ProcPtr): interactiveHandler;
(* Field for number entry. The function accept(x:extended):boolean determines whether the number should be		*)
(*	accepted.	If 'accept = nil' then all numbers are accepted.																	*)

function getStringHandler (item: integer; value: Str255P; default: Str255; isVisible: ProcPtr): interactiveHandler;
(* Field for string entry.																											*)

function getStaticTextHandler (item: integer; title: Str255; titleValid: boolean; isVisible: ProcPtr): interactiveHandler;
(* Used for static text item with number 'item'. *)

function getButtonHandler (item: integer; isVisible, whatToDo: ProcPtr): interactiveHandler;

function getValue (d: DialogPtr; item: integer; default: extended; var value: extended; function check (x: extended): boolean): boolean;

procedure initDialogEngine;

var
processPriorityUpdateEvents: boolean;

implementation

const
	okItem = 1;
	cancelItem = 2;
	longFlash = 15;

type
	boolHandler = object(interactiveHandler)
			value: BooleanP;
			oldValue, default: boolean;
			procedure showValue;
			override;
			procedure update (clickedItem: integer);
			override;
			procedure restore;
			override;
			procedure setDefault;
			override;
		end;

	enumHandler = object(interactiveHandler)
			value: IntegerP;
			oldValue, default, highValue: integer;
			procedure showValue;
			override;
			procedure update (clickedItem: integer);
			override;
			procedure restore;
			override;
			procedure setDefault;
			override;
			procedure makeVisible (yes: boolean);
			override;
		end;

	enumExtendedHandler = object(interactiveHandler)
			value: IntegerP;
			oldValue, default: integer;
			radioButtons: list;
			procedure free;
			override;
			procedure showValue;
			override;
			procedure update (clickedItem: integer);
			override;
			procedure restore;
			override;
			procedure setDefault;
			override;
			procedure makeVisible (yes: boolean);
			override;
		end;

	enumMenuHandler = object(interactiveHandler)
			value: IntegerP;
			oldValue, default, menuId: integer;
			menuItems: list; (* of strings *)
			menu: MenuHandle;
			procedure free;
			override;
			procedure restore;
			override;
			procedure setDefault;
			override;
			procedure showValue;
			override;
			procedure install (d: DialogPtr);
			override;
		end;

	numHandler = object(interactiveHandler)
			value: ExtendedP;
			oldValue, default: extended;
			accept: ProcPtr;
			procedure showValue;
			override;
			procedure setDefault;
			override;
			function isAcceptable: boolean;
			override;
			procedure restore;
			override;
		end;

	stringHandler = object(interactiveHandler)
			value: Str255P;
			oldValue, default: Str255;
			procedure showValue;
			override;
			function isAcceptable: boolean;
			override;
			procedure restore;
			override;
			procedure setDefault;
			override;
		end;

	staticTextHandler = object(interactiveHandler)
			title: Str255;
			titleValid: boolean;
			procedure showValue;
			override;
		end;

	buttonHandler = object(interactiveHandler)
			whatToDo: ProcPtr;
			procedure update (clickedItem: integer);
			override;
		end;

var
	gPictHandle: PicHandle;
	gPictPoint: Point;
	currentHandlers: vector;
	menuIdGenerator: integer;

function getValue (d: DialogPtr; item: integer; default: extended; var value: extended; function check (x: extended): boolean): boolean;
	var
		number: Str255;
		v: extended;
	begin
		GetDText(d, item, number);
		IOCheck(false);
		readString(number, v);
		if (ioResult > 0) | not check(v) then begin
			SysBeep(1);
			value := default;
			getValue := false;
		end
		else begin
			value := v;
			getValue := true;
		end;
		IOCheck(true);
	end;

function interactiveHandler.isAcceptable: boolean;
	begin
		isAcceptable := true;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure interactiveHandler.update (clickedItem: integer);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure interactiveHandler.restore;
	begin
	end;

procedure interactiveHandler.setDefault;
	begin
	end;

{$IFC THINK_PASCAL }
function callFExtendedBoolean (x: extended; myProc: ProcPtr): boolean;
inline
	$205f, $4e90;

function callFBoolean (myProc: ProcPtr): boolean;
inline
	$205f, $4e90;

procedure callP (myProc: ProcPtr);
inline
	$205f, $4e90;
{$ELSEC }
function callFExtendedBoolean (x: extended; myProc: ProcPtr): boolean;
	var f: extendedBooleanFuncT;
	begin
		f := extendedBooleanFuncT(myProc);
		callFExtendedBoolean := f(x);
	end;
	

function callFBoolean (myProc: ProcPtr): boolean;
	var f: booleanFuncT;
	begin
		f := booleanFuncT(myProc);
		callFBoolean := f();
	end;

procedure callP (myProc: ProcPtr);
	var p: noArgProcType;
	begin
		p := noArgProcType(myProc);
		p();
	end;
{$ENDC }

function numHandler.isAcceptable: boolean;
	function check (x: extended): boolean;
		begin
			check := (accept = nil) | callFExtendedBoolean(x, accept);
		end;
	begin
		isAcceptable := getValue(dialog, item, value^, value^, check);
	end;

function stringHandler.isAcceptable: boolean;
	begin
		GetDText(dialog, item, value^);
		isAcceptable := true;
	end;

procedure DoShow (dialog: DialogPtr; item: integer; yes: boolean);
	begin
		if yes then
{$IFC THINK_PASCAL }
			ShowDItem(dialog, item)
		else
			HideDItem(dialog, item);
{$ELSEC }
			ShowDialogItem(dialog, item)
		else
			HideDialogItem(dialog, item);
{$ENDC }
	end;

procedure interactiveHandler.makeVisible (yes: boolean);
	begin
		DoShow(dialog, item, yes);
		visible := yes;
	end;

procedure interactiveHandler.beginVisible;
	begin
		oldVisible := visible;
		if not visible then
			makeVisible(true);
	end;

procedure interactiveHandler.endVisible;
	begin
		visible := oldVisible;
		if not visible then
			makeVisible(false);
	end;

procedure interactiveHandler.show;
	var
		result: boolean;
	begin
		result := (isVisible = nil) | callFBoolean(isVisible);
		if result <> visible then
			makeVisible(result);
	end;

procedure interactiveHandler.install (d: DialogPtr);
	begin
		dialog := d;
	end;

(* From Markus Julen *)
procedure MenuUserItemProc (pDlog: DialogPtr; pItem: integer);
	var
		lRect, lRectOrg: Rect;
		lPort: GrafPtr;
		lHdl: Handle;
		lWidth, lType: integer;
		lPen: PenState;
		lStr: Str255;
		i: seqIndex0;
	function findPItem (x: interactiveHandler): boolean;
		begin
			findPItem := pItem = x.item;
		end;
	begin
{$IFC THINK_PASCAL }
		GetDItem(pDlog, pItem, lType, lHdl, lRectOrg);
{$ELSEC }
		GetDialogItem(pDlog, pItem, lType, lHdl, lRectOrg);
{$ENDC }
		lRect := lRectOrg;
		GetPort(lPort);
		SetPort(pDlog);
		GetPenState(lPen);
		PenNormal;
		with lRect do begin
			bottom := bottom - 1;
			right := right - 1;
			lWidth := right - left - 23;
		end;
		EraseRect(lRect);
		FrameRect(lRect);
		with lRect do begin
			MoveTo(right, top + 2);
			LineTo(right, bottom);
			LineTo(left + 2, bottom);
		end;
		i := currentHandlers.exists(findPItem);
		assert((i > 0) & member(currentHandlers.elements^^[i], enumMenuHandler), 'Item found');
		with enumMenuHandler(currentHandlers.elements^^[i]) do
{$IFC THINK_PASCAL }
			GetItem(menu, value^ + 1, lStr);
{$ELSEC }
			GetMenuItemText(menu, value^ + 1, lStr);
{$ENDC }
		lStr := Concat(lStr, ' ');
{$PUSH }
{$R- }
		while (StringWidth(lStr) > lWidth) & (lStr[0] > Chr(0)) do begin
			lStr[0] := Chr(Ord(lStr[0]) - 1);
			lStr[Ord(lStr[0])] := '';
		end;
{$POP }
		with lRect do
			MoveTo(left + 15, bottom - 5);
		DrawString(lStr);
		with lRect do begin
			right := right - 4;
			left := right - gPictPoint.h;
			top := top + (bottom - top - gPictPoint.v) div 2;
			bottom := top + gPictPoint.v;
		end;
		DrawPicture(gPictHandle, lRect);
		SetPenState(lPen);
		ValidRect(lRectOrg);
		SetPort(lPort);
	end;

function DialogFilter (pDialog: DialogPtr; var vEvent: EventRecord; var vItem: integer): boolean;
	var
		lChar: char;
		lPoint: Point;
		lPort: GrafPtr;
		lPart, lType, oldSelection: integer;
		lHdl: Handle;
		lRect: Rect;
		i: seqIndex0;
	function findVItem (x: interactiveHandler): boolean;
		begin
			findVItem := vItem = x.item;
		end;
	begin
		DialogFilter := false;
		with vEvent do
			if what = keyDown then begin
				lChar := Chr(BAND(message, $000000FF));
				if BTST(modifiers, 8) then begin
					if (lChar = chr(13)) | (lChar = '.') then begin
						HiliteIfAllowed(pDialog, cancelItem, longFlash, vItem);
						DialogFilter := true;
					end;
				end
				else begin
					if (lChar = chr(13)) | (lChar = chr(27)) | (lChar = chr(3)) then begin
						HiliteIfAllowed(pDialog, okItem, longFlash, vItem);
						DialogFilter := true;
					end;
				end;
			end
			else if what = mouseDown then begin
				lPoint := vEvent.where;
				GetPort(lPort);
				SetPort(pDialog);
				GlobalToLocal(lPoint);
				SetPort(lPort);
{$IFC THINK_PASCAL }
				vItem := Succ(FindDItem(pDialog, lPoint));
{$ELSEC }
				vItem := Succ(FindDialogItem(pDialog, lPoint));
{$ENDC }
				i := currentHandlers.exists(findVItem);
				if (i > 0) & member(currentHandlers.elements^^[i], enumMenuHandler) then begin
{$IFC THINK_PASCAL }
					GetDItem(pDialog, vItem, lType, lHdl, lRect);
{$ELSEC }
					GetDialogItem(pDialog, vItem, lType, lHdl, lRect);
{$ENDC }
					lPoint := lRect.topLeft;
					SetPort(pDialog);
					LocalToGlobal(lPoint);
					lPoint := Point(longInt(lPoint) + $00010001);
					EraseRect(lRect);
					SetPort(lPort);
					with enumMenuHandler(currentHandlers.elements^^[i]) do begin
						oldSelection := value^ + 1;
						CheckItem(menu, oldSelection, true);
						lPart := BAND(PopUpMenuSelect(menu, lPoint.v, lPoint.h, oldSelection), $0000FFFF);
						CheckItem(menu, oldSelection, false);
						DialogFilter := (lPart <> 0) & (lPart <> oldSelection);
						if lPart <> 0 then
							value^ := lPart - 1;
					end;
					MenuUserItemProc(pDialog, vItem);
				end;
			end;
	end;

procedure SetUserItem (pDlog: DialogPtr; pItem: integer; pProc: univ Ptr);
	var
		lType: integer;
		lHandle: Handle;
		lBox: Rect;
	begin
{$IFC THINK_PASCAL }
		GetDItem(pDlog, pItem, lType, lHandle, lBox);
		SetDItem(pDlog, pItem, lType, Handle(pProc), lBox);
{$ELSEC }
		GetDialogItem(pDlog, pItem, lType, lHandle, lBox);
		SetDialogItem(pDlog, pItem, lType, Handle(NewUserItemProc(pProc)), lBox);
{$ENDC }
	end;

procedure enumMenuHandler.install (d: DialogPtr);
	var
		result: MenuHandle;
	procedure appendToMenu (s: strings);
		begin
			AppendMenu(result, s.value);
		end;
	begin
		inherited install(d);
		result := NewMenu(menuIdGenerator, '');
		menuId := menuIdGenerator;
		menuIdGenerator := menuIdGenerator + 1;
		menuItems.forAll(appendToMenu);
		InsertMenu(result, -1);
		SetUserItem(d, item,@MenuUserItemProc);
		menu := result;
	end;

procedure enumMenuHandler.free;
	begin
		menuItems.freeAll;
		DeleteMenu(menuId);
		menuIdGenerator := menuIdGenerator - 1;
		DisposeMenu(menu);
		shallowFree;
	end;

procedure interactiveHandler.showValue;
	begin
	end;

procedure boolHandler.showValue;
	begin
		beginVisible;
		SetDCtl(dialog, item, value^);
		endVisible;
	end;

procedure enumHandler.makeVisible (yes: boolean);
	var
		i: integer;
	begin
		for i := item to item + highValue do
			DoShow(dialog, i, yes);
		visible := yes;
	end;

procedure enumExtendedHandler.makeVisible (yes: boolean);
	procedure DoShowItem (x: int);
		begin
			DoShow(dialog, x.value, yes);
		end;
	begin
		radioButtons.forAll(DoShowItem);
		visible := yes;
	end;

procedure enumExtendedHandler.free;
	begin
		radioButtons.freeAll;
		shallowFree;
	end;

procedure enumHandler.showValue;
	var
		i: integer;
	begin
		beginVisible;
		for i := item to item + highValue do
			SetDCtl(dialog, i, i = item + value^);
		endVisible;
	end;

procedure enumExtendedHandler.showValue;
	var
		i: integer;
	procedure showItem (x: int);
		begin
			SetDCtl(dialog, x.value, i = value^);
			i := i + 1;
		end;
	begin
		beginVisible;
		i := 0;
		radioButtons.forAll(showItem);
		endVisible;
	end;

procedure numHandler.showValue;
	begin
		beginVisible;
		setAndHighlight(dialog, item, value^);
		endVisible;
	end;

procedure stringHandler.showValue;
	begin
		beginVisible;
		SetDText(dialog, item, value^);
{$IFC THINK_PASCAL }
		SelIText(dialog, item, 0, 32767);
{$ELSEC }
		SelectDialogItemText(dialog, item, 0, 32767);
{$ENDC }
		endVisible;
	end;

procedure staticTextHandler.showValue;
	begin
		if titleValid then begin
			beginVisible;
			SetDText(dialog, item, title);
			endVisible;
		end;
	end;

procedure enumMenuHandler.showValue;
	begin
		beginVisible;
		MenuUserItemProc(dialog, item);
		endVisible;
	end;



procedure boolHandler.update (clickedItem: integer);
	begin
		if item = clickedItem then begin
			value^ := not value^;
			showValue;
		end;
	end;

procedure enumHandler.update (clickedItem: integer);
	begin
		if (item <= clickedItem) & (clickedItem <= item + highValue) then begin
			value^ := clickedItem - item;
			showValue;
		end;
	end;

procedure enumExtendedHandler.update (clickedItem: integer);
	var
		i: integer;
	function clickedItemExists (x: int): boolean;
		begin
			clickedItemExists := x.value = clickedItem;
		end;
	begin
		i := radioButtons.exists(clickedItemExists);
		if i > 0 then begin
			value^ := i - 1;
			showValue;
		end;
	end;

procedure buttonHandler.update (clickedItem: integer);
	begin
		if item = clickedItem then
			callP(whatToDo);
	end;


procedure boolHandler.restore;
	begin
		value^ := oldValue;
	end;

procedure enumHandler.restore;
	begin
		value^ := oldValue;
	end;

procedure enumExtendedHandler.restore;
	begin
		value^ := oldValue;
	end;

procedure enumMenuHandler.restore;
	begin
		value^ := oldValue;
	end;

procedure numHandler.restore;
	begin
		value^ := oldValue;
	end;

procedure stringHandler.restore;
	begin
		value^ := oldValue;
	end;


procedure boolHandler.setDefault;
	begin
		value^ := default;
	end;

procedure enumHandler.setDefault;
	begin
		value^ := default;
	end;

procedure enumExtendedHandler.setDefault;
	begin
		value^ := default;
	end;

procedure enumMenuHandler.setDefault;
	begin
		value^ := default;
	end;

procedure numHandler.setDefault;
	begin
		value^ := default;
	end;

procedure stringHandler.setDefault;
	begin
		value^ := default;
	end;

function getBooleanHandler (item: integer; value: BooleanP; default: boolean; isVisible: ProcPtr): interactiveHandler;
	var
		result: boolHandler;
	begin
		assert((item <> okItem) & (item <> cancelItem), 'Legal items');
		new(result);
		failNil(Handle(result));
		result.item := item;
		result.isVisible := isVisible;
		result.visible := true;
		result.value := value;
		result.default := default;
		result.oldValue := value^;
		getBooleanHandler := result;
	end;

function getEnumerationHandler (item: integer; value: IntegerP; default: integer; isVisible: ProcPtr; highValue: integer): interactiveHandler;
	var
		result: enumHandler;
	begin
		assert((item <> okItem) & (item <> cancelItem), 'Legal items');
		new(result);
		failNil(Handle(result));
		result.item := item;
		result.isVisible := isVisible;
		result.visible := true;
		result.value := value;
		result.default := default;
		result.oldValue := value^;
		result.highValue := highValue;
		getEnumerationHandler := result;
	end;

function s (x: longInt; l: list): list;
	var
		i: int;
	begin
		new(i);
		failNil(Handle(i));
		i.value := x;
		if l = nil then
			l := newEmptyList;
		l.push(i);
		s := l;
	end;

function getEnumerationExtendedHandler (radioButtons: list; value: IntegerP; default: integer; isVisible: ProcPtr): interactiveHandler;
	var
		result: enumExtendedHandler;
	begin
		new(result);
		failNil(Handle(result));
		result.item := -1;
		result.isVisible := isVisible;
		result.visible := true;
		result.value := value;
		result.default := default;
		result.oldValue := value^;
		result.radioButtons := radioButtons;
		getEnumerationExtendedHandler := result;
	end;

function t (s: Str255; l: list): list;
	var
		i: strings;
	begin
		new(i);
		failNil(Handle(i));
		i.value := s;
		if l = nil then begin
			new(l);
			failNil(Handle(l));
			l.init;
		end;
		l.push(i);
		t := l;
	end;

function getEnumerationMenuHandler (item: integer; value: IntegerP; default: integer; isVisible: ProcPtr; menuItems: list): interactiveHandler;
	var
		result: enumMenuHandler;
	begin
		new(result);
		failNil(Handle(result));
		result.item := item;
		result.isVisible := isVisible;
		result.visible := true;
		result.value := value;
		result.default := default;
		result.oldValue := value^;
		result.menuItems := menuItems;
		getEnumerationMenuHandler := result;
	end;

function getExtendedHandler (item: integer; value: ExtendedP; default: extended; isVisible: ProcPtr; accept: ProcPtr): interactiveHandler;
	var
		result: numHandler;
	begin
		assert((item <> okItem) & (item <> cancelItem), 'Legal items');
		new(result);
		failNil(Handle(result));
		result.item := item;
		result.isVisible := isVisible;
		result.visible := true;
		result.value := value;
		result.default := default;
		result.oldValue := value^;
		result.accept := accept;
		getExtendedHandler := result;
	end;

function getStringHandler (item: integer; value: Str255P; default: Str255; isVisible: ProcPtr): interactiveHandler;
	var
		result: stringHandler;
	begin
		assert((item <> okItem) & (item <> cancelItem), 'Legal items');
		new(result);
		failNil(Handle(result));
		result.item := item;
		result.isVisible := isVisible;
		result.visible := true;
		result.value := value;
		result.default := default;
		result.oldValue := value^;
		getStringHandler := result;
	end;

function getStaticTextHandler (item: integer; title: Str255; titleValid: boolean; isVisible: ProcPtr): interactiveHandler;
	var
		result: staticTextHandler;
	begin
		new(result);
		failNil(Handle(result));
		result.item := item;
		result.isVisible := isVisible;
		result.visible := true;
		result.title := title;
		result.titleValid := titleValid;
		getStaticTextHandler := result;
	end;

function getButtonHandler (item: integer; isVisible, whatToDo: ProcPtr): interactiveHandler;
	var
		result: buttonHandler;
	begin
		new(result);
		failNil(Handle(result));
		result.item := item;
		result.isVisible := isVisible;
		result.visible := true;
		result.whatToDo := whatToDo;
		getButtonHandler := result;
	end;

procedure processEvents;
	var
		theEvent: EventRecord;
		dummy: boolean;
	begin
		while EventAvail(activMask + updateMask, theEvent) do begin
			dummy := GetNextEvent(activMask + updateMask, theEvent);
			DoEvent(theEvent);
		end;
	end;

function processDialogBox (resource, defaultButton, revertButton, setButton: integer; function activateOkButton: boolean; handlers: vector): boolean;
	var
		dialog: DialogPtr;
		item: integer;
		timer: longInt;
		i: seqIndex01;
		result: boolean;
	begin
		timer := TickCount;
		currentHandlers := handlers;
		beginCursor(arrowCursor);
		dialog := GetNewDialog(resource, nil, WindowPtr(-1));
		assert(dialog <> nil, 'Resource exists.');
		niceWindowPosition(WindowPtr(dialog), true);
		for i := 1 to handlers.length do
			interactiveHandler(handlers.elements^^[i]).install(dialog);
		for i := 1 to handlers.length do
			with interactiveHandler(handlers.elements^^[i]) do begin
				showValue;
				show;
			end;
		SelectWindow(dialog);
		ShowWindow(dialog);
		showDefaultButton(dialog, okItem);
		repeat
			ActivateItem(dialog, okItem, activateOkButton);
{$IFC THINK_PASCAL }
			ModalDialog(@DialogFilter, item);
{$ELSEC }
			ModalDialog(NewModalFilterProc(@DialogFilter), item);
{$ENDC }
			if (item = okItem) | (setButton > 2) & (item = setButton) then begin
				i := 1;
				while (i <= handlers.length) & interactiveHandler(handlers.elements^^[i]).isAcceptable do
					i := i + 1;
				if i > handlers.length then
					result := item = okItem
				else begin
					result := false;
					for i := 1 to handlers.length do
						interactiveHandler(handlers.elements^^[i]).restore;
				end;
				leave;
			end
			else if item = cancelItem then begin
				for i := 1 to handlers.length do
					interactiveHandler(handlers.elements^^[i]).restore;
				result := false;
				leave;
			end
			else if (defaultButton > 2) & (item = defaultButton) then begin
				for i := 1 to handlers.length do
					interactiveHandler(handlers.elements^^[i]).setDefault;
				for i := 1 to handlers.length do
					with interactiveHandler(handlers.elements^^[i]) do begin
						showValue;
						show;
						update(defaultButton);
					end;
			end
			else if (revertButton > 2) & (item = revertButton) then begin
				for i := 1 to handlers.length do
					interactiveHandler(handlers.elements^^[i]).restore;
				for i := 1 to handlers.length do
					with interactiveHandler(handlers.elements^^[i]) do begin
						showValue;
						show;
					end;
			end
			else begin
				for i := 1 to handlers.length do
					interactiveHandler(handlers.elements^^[i]).update(item);
				for i := 1 to handlers.length do
					interactiveHandler(handlers.elements^^[i]).show;
			end;
		until false;
		HideWindow(dialog);
{$IFC THINK_PASCAL }
		DisposDialog(dialog);
{$ELSEC }
		DisposeDialog(dialog);
{$ENDC }
		endCursor;
		if result & processPriorityUpdateEvents then
			processEvents;
		processDialogBox := result;
		handlers.freeAll;
		startTime := startTime + TickCount - timer;
	end;

procedure initDialogEngine;
	begin
		gPictHandle := GetPicture(129);
		with gPictHandle^^.picFrame, gPictPoint do begin
			h := right - left;
			v := bottom - top;
		end; (* WITH *)
		menuIdGenerator := 999;
	end;

end.
