(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit dictPriorityQueue;

interface

uses
	priorityQueue;

implementation

procedure dictPriorityQueue.reset;
	begin
		minRef := nil;
		numberOfElements := 0;
		d.reset;
	end;

procedure dictPriorityQueue.init;
	begin
		minRef := nil;
		numberOfElements := 0;
		d := dictionary(newInitObj(currentDictionaryType));
	end;

procedure dictPriorityQueue.free;
	begin
		d.free;
		shallowFree;
	end;

procedure dictPriorityQueue.freeAll;
	begin
		d.freeAll;
		shallowFree;
	end;

procedure dictPriorityQueue.interactiveOutput;
	var
		t: obj;
	begin
		write('Minimum = ');
		if minRef = nil then
			writeLn('NIL')
		else begin
			t := getObject(minRef);
			t.interactiveOutput;
			writeLn;
		end;
		d.interactiveOutput;
	end;

function dictPriorityQueue.getObject (where: reference): obj;
	begin
		getObject := d.getObject(where);
	end;

function dictPriorityQueue.insert (x: obj; function compare (a, b: obj): relationObj): reference;
	var
		t: reference;
		found: boolean;
	begin
		numberOfElements := numberOfElements + 1;
		t := d.insertObj(x, compare, found);
		insert := t;
		if minRef = nil then
			minRef := t
		else if compare(x, d.getObject(minRef)) <= equalObj then
			minRef := d.extreme(left);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure dictPriorityQueue.delete (x: reference; function compare (a, b: obj): relationObj);
	begin
		numberOfElements := numberOfElements - 1;
		if minRef = x then begin
			minRef := d.next(minRef, right);
			d.delete(x);
		end
		else
			d.delete(x);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function dictPriorityQueue.minimum: reference;
	begin
		minimum := minRef;
	end;

function dictPriorityQueue.ith (i: longInt): reference;
	begin
		ith := d.ith(i);
	end;

function dictPriorityQueue.isEmpty: boolean;
	begin
		isEmpty := d.isEmpty;
	end;

procedure dictPriorityQueue.forAll (procedure whatToDo (x: obj));
	begin
		d.forAll(whatToDo, left);
	end;

function dictPriorityQueue.find (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType): boolean;
	begin
		find := d.find(x, compare, where, direction);
	end;

function dictPriorityQueue.next (x: reference; direction: directionType): reference;
	begin
		next := d.next(x, direction);
	end;


end.
