(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit EyeInterface;

interface

uses
	geoBenchUtility;

type

	Cube = record
			xmin, ymin, zmin: Integer;
			xmax, ymax, zmax: Integer;
		end;

	EyeDesc = record
			x, y, z: Longint;
			vx, vy, vz: Longint;
			ProjectionType: Integer; (* 0 = parallel, 1 = Central *)
			hidden: boolean; (* true = visibility enabled *)
		end;

{$IFC false THEN }
var
	usePict: Boolean;
(* this global should be put in prefs File. It controls if a passed *)
(* picture for seteye should be displayed or not. value changes *)
(* if user clicks in 'use object scetch' checkbox                        *)
(* NOTE: if var is placed in prefs file, delete initialization code *)
(*            in InitItems                                                               *)
{$ENDC }
const (* Error codes *)
	OK = 0;
	ApplySettings = 1;
	Err = -1;

(* SetEye :  Bring up dialog for setting interactively viewing parameters.  *)
(*                ENTRY : theEye initialized to current viewing parameters       *)
(*                             Bounds3D contains minimal and maximal coordinates  *)
(*                                            of objects to view. these will be used to     *)
(*                                            calibrate the setting box and zoom settings *)
(*                             TopPic is handle to picture containing a view straight *)
(*                                         down onto the object or nil if no picture         *)
(*                             SidePic is handle to picture showing object seen from *)
(*                                         one side (straight down x-achsis)                   *)
(*               EXIT : SetEye will return an integer value indicating the folowing *)
(*                          0 : OK, everything done. Usually followed by a redraw      *)
(*                               of the object with new parameters                               *)
(*                          1 : Apply current parameters: redraw object with            *)
(*                                current params and recall SetEye procedure                *)
(*                        -1 :  Error. Ignore all changes. SetEye will also return -1 if *)
(*                                User clicked Cancel Button                                           *)

function SetEye (var theEye: EyeDesc; Bounds3D: Cube; DlgP: DialogPtr): Integer;

implementation

{$IFC NOT THINK_PASCAL }
uses NumberFormatting;
{$ENDC }

type
	Vector = record
			x, y, z: Real
		end;

(* The global variable statusPtr holds the current status and is placed in the window's refCon-field				*)
(* When the control is first passed to SetEye, the status is set. Then the control can switch several			*)
(* times between the calling routine that updates the GeoBench Window and the SetEye procedure.				*)
(* After the user pressed OK or CANCEL, the statusPtr is deallocated.												*)
	statusDesc = record
			Eye: EyeDesc;									(* Current Projection Descriptor									*)
			de: Vector; 										(* Direction Eyepoint - Viewpoint									*)
			theCube: Cube;									(* Bounding Box of the object											*)
			zoomCube: Cube;								(* Borders of the objects (Eye und Viewpoint, Bounding Box)	*)
														(* that have to be drawn in the panels								*)
			zoomfactor: Vector;							(* Current Zoomfactor for the panels								*)
			ext, eyt, ezt: Handle; 							(* All ...t variables are handles to their dialog items				*)
			vxt, vyt, vzt: Handle;
			dt, phit, ZoomInt, ZoomOutt: Handle;
			xyplanet, xzplanet, autoApplyt: Handle;
			parallelt, centralt, hiddenLinet: Handle;
			xw, yw, zw: integer;							(* Dimensions of the panels 											*)
			xypRect, xzpRect: Rect; 						(* Rectangles that contain the panels 								*)
			okRect: Rect;
			autoApplyOn: boolean;							(* Is the auto apply button on?										*)
			thePattern: Pattern;							(* Pattern for the object rects										*)
		end;

const
	PatternList = 9000;
	ObjectPatternID = 3;
	ShiftCode = 56;
	okButton = 1;
	cancel = 2;
	apply = 3;
	xypanel = 5;
	xzpanel = 4;
	parallel = 6;
	Central = 7;
	ZoomIn = 8;
	ZoomOut = 9;
	Ex = 10;
	Ey = 11;
	Ez = 12;
	Vx = 13;
	Vy = 14;
	Vz = 15;
	dist = 16;
	SetCloser = 18;
	SetFurther = 17;
	ResetPanel = 19;
	hiddenLine = 20;
(*UseObjectScetch = 21; *)
	AutoApply = 21;
	DontCare = 0;

var
	statusPtr: ^statusDesc;
	itemHit: Integer;
	Dlg: DialogPtr;
{$IFC false THEN }
	HasPic: Boolean;									(* Are Top- and SidePic present ? 									*)
	topPict, sidePict: PicHandle;
	UseObjectScetcht: Handle;
{$ENDC }


function max (a, b: longint): longint;
	begin
		if a > b then
			max := a
		else
			max := b;
	end;


function min (a, b: longint): longint;
	begin
		if a < b then
			min := a
		else
			min := b
	end;

procedure getValue (item: integer; default: longint; var value: longint);
	var
		number: Str255;
		v: extended;
	begin
		GetDText(Dlg, item, number);
		IOCheck(false);
		readString(number, v);
		if (ioResult > 0) or (abs(v) > 32000) then begin
			SysBeep(1);
			value := default;
		end
		else
{$IFC THINK_PASCAL }
			value := round(v);
{$ELSEC }
			value := roundLong(v);
{$ENDC }
		IOCheck(true);
	end;

(* CalcBounds : Calculate the bounds for the 3d panels. *)
(* after this execution, zoomcube will be initialized *)
procedure CalcBounds;
	var
		lox, hix, loy, hiy, loz, hiz: integer;
	begin
		lox := min(statusPtr^.theCube.xmin, statusPtr^.Eye.x);
		lox := min(lox, statusPtr^.Eye.vx); (* calculate the absolute minima and maxima *)
		hix := max(statusPtr^.theCube.xmax, statusPtr^.Eye.x);
		hix := max(hix, statusPtr^.Eye.vx);
		loy := min(statusPtr^.theCube.ymin, statusPtr^.Eye.y);
		loy := min(loy, statusPtr^.Eye.vy);
		hiy := max(statusPtr^.theCube.ymax, statusPtr^.Eye.y);
		hiy := max(hiy, statusPtr^.Eye.vy);
		loz := min(statusPtr^.theCube.zmin, statusPtr^.Eye.z);
		loz := min(loz, statusPtr^.Eye.vz);
		hiz := max(statusPtr^.theCube.zmax, statusPtr^.Eye.z);
		hiz := max(hiz, statusPtr^.Eye.vz);
		with statusPtr^.zoomCube do begin
			xmin := lox;
			ymin := loy;
			zmin := loz;
			xmax := hix;
			ymax := hiy;
			zmax := hiz;
		end;
	end;


procedure CalcZoomFactor;
	begin
		with statusPtr^.zoomCube do begin
			statusPtr^.zoomfactor.x := 1 / ((xmax - xmin) / statusPtr^.xw);
			statusPtr^.zoomfactor.y := 1 / ((ymax - ymin) / statusPtr^.yw);
			statusPtr^.zoomfactor.z := 1 / ((zmax - zmin) / statusPtr^.zw);
		end;
	end;


(* Convertzx : convert a point in Worldcoordinates first to panel-coordinates *)
(* in zx-panel and then to screen coords                                                         *)
function ConvertPointzx (var aPoint: Point): boolean;
	var
		thePoint: record
				h, v: longint
			end;
	begin
		thePoint.h := longint(aPoint.h) - longint(statusPtr^.zoomCube.xmin);
{$IFC THINK_PASCAL }
		thePoint.h := round(statusPtr^.zoomfactor.x * thePoint.h);	(* panel-relative screen *)
{$ELSEC }
		thePoint.h := roundLong(statusPtr^.zoomfactor.x * thePoint.h);	(* panel-relative screen *)
{$ENDC }
		thePoint.h := thePoint.h + statusPtr^.xzpRect.left;				(* now screen absolute *)
		thePoint.v := longint(aPoint.v) - longint(statusPtr^.zoomCube.zmin);
{$IFC THINK_PASCAL }
		thePoint.v := round(statusPtr^.zoomfactor.z * thePoint.v);
{$ELSEC }
		thePoint.v := roundLong(statusPtr^.zoomfactor.z * thePoint.v);
{$ENDC }
		thePoint.v := statusPtr^.xzpRect.bottom - thePoint.v;
		ConvertPointzx := false;
		if (abs(thePoint.h) < 32000) & (abs(thePoint.v) < 32000) then begin
			aPoint.h := thePoint.h;
			aPoint.v := thePoint.v;
			ConvertPointzx := true;
		end;
	end;

function RevertPointzx (var aPoint: Point): boolean;
	var
		thePoint: record
				h, v: longint
			end;
	begin
		thePoint.h := longint(aPoint.h) - longint(statusPtr^.xzpRect.left);		(* screen relative -> screen absolute *)
{$IFC THINK_PASCAL }
		thePoint.h := round(thePoint.h / statusPtr^.zoomfactor.x);				(* screen absolute -> world relative *)
{$ELSEC }
		thePoint.h := roundLong(thePoint.h / statusPtr^.zoomfactor.x);				(* screen absolute -> world relative *)
{$ENDC }
		thePoint.h := thePoint.h + statusPtr^.zoomCube.xmin;						(* absolute *)
		thePoint.v := -longint(aPoint.v) + longint(statusPtr^.xzpRect.bottom);	(* screen relative -> screen absolute *)
{$IFC THINK_PASCAL }
		thePoint.v := round(thePoint.v / statusPtr^.zoomfactor.z);				(* screen absolute -> world relative *)
{$ELSEC }
		thePoint.v := roundLong(thePoint.v / statusPtr^.zoomfactor.z);				(* screen absolute -> world relative *)
{$ENDC }
		thePoint.v := thePoint.v + statusPtr^.zoomCube.zmin;						(* absolute *)
		RevertPointzx := false;
		if (abs(thePoint.h) < 32000) & (abs(thePoint.v) < 32000) then begin
			aPoint.h := thePoint.h;
			aPoint.v := thePoint.v;
			RevertPointzx := true;
		end;
	end;

function ConvertPointxy (var aPoint: Point): boolean;
	var
		thePoint: record
				h, v: longint
			end;
	begin
		thePoint.h := longint(aPoint.h) - longint(statusPtr^.zoomCube.xmin);
{$IFC THINK_PASCAL }
		thePoint.h := round(statusPtr^.zoomfactor.x * thePoint.h);
{$ELSEC }
		thePoint.h := roundLong(statusPtr^.zoomfactor.x * thePoint.h);
{$ENDC }
		thePoint.h := thePoint.h + statusPtr^.xypRect.left;
		thePoint.v := longint(aPoint.v) - longint(statusPtr^.zoomCube.ymin);
{$IFC THINK_PASCAL }
		thePoint.v := round(statusPtr^.zoomfactor.y * thePoint.v);
{$ELSEC }
		thePoint.v := roundLong(statusPtr^.zoomfactor.y * thePoint.v);
{$ENDC }
		thePoint.v := statusPtr^.xypRect.top + thePoint.v;
		ConvertPointxy := false;
		if (abs(thePoint.h) < 32000) & (abs(thePoint.v) < 32000) then begin
			aPoint.h := thePoint.h;
			aPoint.v := thePoint.v;
			ConvertPointxy := true;
		end;
	end;

function RevertPointxy (var aPoint: Point): boolean;
	var
		thePoint: record
				h, v: longint
			end;
	begin
		thePoint.h := longint(aPoint.h) - longint(statusPtr^.xypRect.left);	(* screen relative -> screen absolute *)
{$IFC THINK_PASCAL }
		thePoint.h := round(thePoint.h / statusPtr^.zoomfactor.x);			(* screen absolute -> world relative *)
{$ELSEC }
		thePoint.h := roundLong(thePoint.h / statusPtr^.zoomfactor.x);			(* screen absolute -> world relative *)
{$ENDC }
		thePoint.h := thePoint.h + statusPtr^.zoomCube.xmin;					(* absolute *)
		thePoint.v := longint(aPoint.v) - longint(statusPtr^.xypRect.top);	(* screen relative -> screen absolute *)
{$IFC THINK_PASCAL }
		thePoint.v := round(thePoint.v / statusPtr^.zoomfactor.y);			(* screen absolute -> world relative *)
{$ELSEC }
		thePoint.v := roundLong(thePoint.v / statusPtr^.zoomfactor.y);			(* screen absolute -> world relative *)
{$ENDC }
		thePoint.v := thePoint.v + statusPtr^.zoomCube.ymin;					(* absolute *)
		RevertPointxy := false;
		if (abs(thePoint.h) < 32000) & (abs(thePoint.v) < 32000) then begin
			aPoint.h := thePoint.h;
			aPoint.v := thePoint.v;
			RevertPointxy := true;
		end;
	end;

function CalcZoomOut: boolean;
	var
		dx, dy, dz, ddx, ddy, ddz: longint;
		result: boolean;
	begin
		with statusPtr^.zoomCube do begin
			dx := longint(xmax) - longint(xmin);
			dy := longint(ymax) - longint(ymin);
			dz := longint(zmax) - longint(zmin);
			dx := dx div 2;
			dy := dy div 2;
			dz := dz div 2;
			ddx := xmax + dx;
			ddy := ymax + dy;
			ddz := zmax + dz;
			result := false;
			if (abs(ddx) < 32000) & (abs(ddy) < 32000) & (abs(ddz) < 32000) then begin
				result := true;
				xmax := ddx;
				ymax := ddy;
				zmax := ddz;
			end;
			ddx := xmin - dx;
			ddy := ymin - dy;
			ddz := zmin - dz;
			if (abs(ddx) < 32000) & (abs(ddy) < 32000) & (abs(ddz) < 32000) then begin
				xmin := ddx;
				ymin := ddy;
				zmin := ddz;
			end
			else
				result := false;
		end;
		statusPtr^.zoomfactor.x := statusPtr^.zoomfactor.x / 2;
		statusPtr^.zoomfactor.y := statusPtr^.zoomfactor.y / 2;
		statusPtr^.zoomfactor.z := statusPtr^.zoomfactor.z / 2;
		CalcZoomOut := result;
	end;

function CalcZoomIn: boolean;
	var
		dx, dy, dz, ddx, ddy, ddz: longint;
		result: boolean;
	begin
		with statusPtr^.zoomCube do begin
			dx := longint(xmax) - longint(xmin);
			dy := longint(ymax) - longint(ymin);
			dz := longint(zmax) - longint(zmin);
			dx := dx div 4;
			dy := dy div 4;
			dz := dz div 4;
			ddx := xmax - dx;
			ddy := ymax - dy;
			ddz := zmax - dz;
			result := false;
			if (abs(ddx) < 32000) & (abs(ddy) < 32000) & (abs(ddz) < 32000) then begin
				result := true;
				xmax := ddx;
				ymax := ddy;
				zmax := ddz;
			end;
			ddx := xmin + dx;
			ddy := ymin + dy;
			ddz := zmin + dz;
			if (abs(ddx) < 32000) & (abs(ddy) < 32000) & (abs(ddz) < 32000) then begin
				xmin := ddx;
				ymin := ddy;
				zmin := ddz;
			end
			else
				result := false;
		end;
		statusPtr^.zoomfactor.x := statusPtr^.zoomfactor.x * 2;
		statusPtr^.zoomfactor.y := statusPtr^.zoomfactor.y * 2;
		statusPtr^.zoomfactor.z := statusPtr^.zoomfactor.z * 2;
		CalcZoomIn := result;
	end;


(* Checks whether we still can zoom in. We cannot zoom further if  eye or	*)
(* viewpoint will fall outside either panel  or the objectrect grows too big.	*)
(* Also, both panels have to be fully reverteable.									*)

function CheckZoomIn: Boolean; (* FALSE means cannot zoom further *)
	var
		eyePoint, viewPoint: Point;
		objectRect: Rect;
		oldEye: EyeDesc;
		oldZoomCube: Cube;
		oldZoomFactor: Vector;
		result: boolean;
	begin
		oldEye := statusPtr^.Eye;
		oldZoomCube := statusPtr^.zoomCube;
		oldZoomFactor := statusPtr^.zoomfactor;
		result := CalcZoomIn;
		if result then begin
	 (* Check, whether eyePoint and viewPoint are both in the resulting panels *)
	 (* First, check zx-panel *)
			eyePoint.h := statusPtr^.Eye.x;
			eyePoint.v := statusPtr^.Eye.z;
			result := ConvertPointzx(eyePoint);
			viewPoint.h := statusPtr^.Eye.vx;
			viewPoint.v := statusPtr^.Eye.vz;
			result := result & ConvertPointzx(viewPoint);
			objectRect := statusPtr^.xzpRect;
			objectRect.topLeft.v := objectRect.topLeft.v - 1;
			objectRect.topLeft.h := objectRect.topLeft.h - 1;
			objectRect.botRight.v := objectRect.botRight.v + 1;
			objectRect.botRight.h := objectRect.botRight.h + 1;
			result := result & (PtInRect(eyePoint, objectRect)) & (PtInRect(viewPoint, objectRect));
	 (* Now check xy-panel *)
			eyePoint.h := statusPtr^.Eye.x;
			eyePoint.v := statusPtr^.Eye.y;
			result := result & ConvertPointxy(eyePoint);
			viewPoint.h := statusPtr^.Eye.vx;
			viewPoint.v := statusPtr^.Eye.vy;
			result := result & ConvertPointxy(viewPoint);
			objectRect := statusPtr^.xypRect;
			objectRect.topLeft.v := objectRect.topLeft.v - 1;
			objectRect.topLeft.h := objectRect.topLeft.h - 1;
			objectRect.botRight.v := objectRect.botRight.v + 1;
			objectRect.botRight.h := objectRect.botRight.h + 1;
			result := result & (PtInRect(eyePoint, objectRect)) & (PtInRect(viewPoint, objectRect));
			if result then begin
	   (* Check, whether the object rect is converteable. *)
				SetRect(objectRect, statusPtr^.theCube.xmin, statusPtr^.theCube.zmin, statusPtr^.theCube.xmax, statusPtr^.theCube.zmax);
				with objectRect do begin
					result := ConvertPointzx(topLeft);
					result := result & ConvertPointzx(botRight);
				end;
				SetRect(objectRect, statusPtr^.theCube.xmin, statusPtr^.theCube.ymin, statusPtr^.theCube.xmax, statusPtr^.theCube.ymax);
				with objectRect do begin
					result := result & ConvertPointxy(objectRect.topLeft);
					result := result & ConvertPointxy(objectRect.botRight);
				end;
			end;
		end;
		statusPtr^.Eye := oldEye;
		statusPtr^.zoomCube := oldZoomCube;
		statusPtr^.zoomfactor := oldZoomFactor;
		CheckZoomIn := result;
	end;

(* checkzoomOut : Check if we cannot zoom further out. this happens if zoomfactor is smaller than a *)
(*                          specific number *)
function CheckZoomOut: Boolean;
	var
		oldEye: EyeDesc;
		oldZoomCube: Cube;
		oldZoomFactor: Vector;
		result: boolean;
		objectRect: Rect;
		diameter: real;
	begin
		oldEye := statusPtr^.Eye;
		oldZoomCube := statusPtr^.zoomCube;
		oldZoomFactor := statusPtr^.zoomfactor;
		result := CalcZoomOut;
		if result then begin
	 (* Check, whether the new panel area is fully reverteable *)
			objectRect := statusPtr^.xzpRect;
			result := RevertPointzx(objectRect.topLeft);
			result := result & RevertPointzx(objectRect.botRight);
			diameter := sqr(longint(objectRect.right) - longint(objectRect.left));
			diameter := sqrt(diameter + sqr(longint(objectRect.bottom) - longint(objectRect.top)));
			result := result & (diameter < 32000);
			objectRect := statusPtr^.xypRect;
			result := result & RevertPointxy(objectRect.topLeft);
			result := result & RevertPointxy(objectRect.botRight);
			diameter := sqr(longint(objectRect.right) - longint(objectRect.left));
			diameter := sqrt(diameter + sqr(longint(objectRect.bottom) - longint(objectRect.top)));
			result := result & (diameter < 32000.0);
		end;
		statusPtr^.Eye := oldEye;
		statusPtr^.zoomCube := oldZoomCube;
		statusPtr^.zoomfactor := oldZoomFactor;
		CheckZoomOut := result;
	end;

(* DrawEyePoint: Draw a circle around eyepoint with radius 4. Point is in local coordinates *)
(*                        Clipping region must be set previously *)

procedure DrawEyePoint (thePoint: Point);

	var
		EyeRect: Rect;

	begin
		with EyeRect do begin
			top := thePoint.v;
			left := thePoint.h;
			bottom := thePoint.v + 1;
			right := thePoint.h + 1;
		end;
		InsetRect(EyeRect, -4, -4);
		PaintOval(EyeRect);
	end;

procedure DrawViewPoint (thePoint: Point);

	var
		ViewRect: Rect;

	begin
		with ViewRect do begin
			top := thePoint.v;
			left := thePoint.h;
			bottom := thePoint.v + 1;
			right := thePoint.h + 1;
		end;
		InsetRect(ViewRect, -4, -4);
		FrameOval(ViewRect);
		with ViewRect do begin
			MoveTo(left, top);
			LineTo(right, bottom);
			MoveTo(right, top);
			LineTo(left, bottom);
		end;
	end;

(* xyPanleItem : Procedure to draw the xyPanel *)
{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DrawxzPanelItem (theWindow: WindowPtr; theItem: Integer);
	var
		objectRect: Rect;
		temp: integer;
		thePoint: Point;
		SaveRgn: RgnHandle;
		ok: boolean;
	begin
		SaveRgn := NewRgn;
		GetClip(SaveRgn);
		ClipRect(statusPtr^.xzpRect);
		EraseRect(statusPtr^.xzpRect);
	(* draw the object in the pane corresponding to its zoomsize *)
		SetRect(objectRect, statusPtr^.theCube.xmin, statusPtr^.theCube.zmin, statusPtr^.theCube.xmax, statusPtr^.theCube.zmax);
		ok := ConvertPointzx(objectRect.topLeft);
		ok := ConvertPointzx(objectRect.botRight);
		if objectRect.top > objectRect.bottom then begin
			temp := objectRect.top;
			objectRect.top := objectRect.bottom;
			objectRect.bottom := temp;
		end;
{$IFC false THEN }
		if usePict then
			DrawPicture(sidePict, objectRect)
		else begin
{$ENDC }
			PenPat(statusPtr^.thePattern);
			PaintRect(objectRect);
{$IFC THINK_PASCAL }
			PenPat(black);
{$ELSEC }
			PenPat(qd.black);
{$ENDC }
			FrameRect(objectRect);
{$IFC false THEN }
		end;
{$ENDC }
		FrameRect(statusPtr^.xzpRect);		(* now draw frame *)
		thePoint.h := statusPtr^.Eye.x;		(* now draw eye and viewpoint *)
		thePoint.v := statusPtr^.Eye.z;
		ok := ConvertPointzx(thePoint);
		DrawEyePoint(thePoint);
		MoveTo(thePoint.h, thePoint.v);
		thePoint.h := statusPtr^.Eye.vx;
		thePoint.v := statusPtr^.Eye.vz;
		ok := ConvertPointzx(thePoint);
		LineTo(thePoint.h, thePoint.v);
		DrawViewPoint(thePoint);
		SetClip(SaveRgn);
		DisposeRgn(SaveRgn);
	end;

(* yzPanleItem : Procedure to draw the xyPanel *)
procedure DrawxyPanelItem (theWindow: WindowPtr; theItem: Integer);
	var
		objectRect: Rect;
		temp: integer;
		thePoint: Point;
		SaveRgn: RgnHandle;
		ok: boolean;

	begin
		SaveRgn := NewRgn;
		GetClip(SaveRgn);
		ClipRect(statusPtr^.xypRect);
		EraseRect(statusPtr^.xypRect);
		SetRect(objectRect, statusPtr^.theCube.xmin, statusPtr^.theCube.ymin, statusPtr^.theCube.xmax, statusPtr^.theCube.ymax);
		ok := ConvertPointxy(objectRect.topLeft);
		ok := ConvertPointxy(objectRect.botRight);
		if objectRect.top > objectRect.bottom then begin
			temp := objectRect.top;
			objectRect.top := objectRect.bottom;
			objectRect.bottom := temp;
		end;

{$IFC false THEN }
		if usePict then
			DrawPicture(topPict, objectRect)
		else begin
{$ENDC }
			PenPat(statusPtr^.thePattern);
			PaintRect(objectRect);
{$IFC THINK_PASCAL }
			PenPat(black);
{$ELSEC }
			PenPat(qd.black);
{$ENDC }
			FrameRect(objectRect);
{$IFC false THEN }
		end;
{$ENDC }
		FrameRect(statusPtr^.xypRect);
		thePoint.h := statusPtr^.Eye.x;		(* now draw eye and viewpoint *)
		thePoint.v := statusPtr^.Eye.y;
		ok := ConvertPointxy(thePoint);
		DrawEyePoint(thePoint);
		MoveTo(thePoint.h, thePoint.v);
		thePoint.h := statusPtr^.Eye.vx;
		thePoint.v := statusPtr^.Eye.vy;
		ok := ConvertPointxy(thePoint);
		LineTo(thePoint.h, thePoint.v);
		DrawViewPoint(thePoint);
		SetClip(SaveRgn);
		DisposeRgn(SaveRgn);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }


procedure RedrawPanels;
	begin
		EraseRect(statusPtr^.xypRect);
		EraseRect(statusPtr^.xzpRect);
		DrawxzPanelItem(Dlg, DontCare);
		DrawxyPanelItem(Dlg, DontCare);
	end;

(* UpdateTextItems converts all Eye-Coords and View coords to  strings *)
(* and replaces them with the previous contents *)
procedure UpdateTextItems;
	var
		Str: Str255;
		dd: Real;
		dx, dy, dz: LongInt;
		delta: Integer;

	begin
		NumToString(statusPtr^.Eye.x, Str);
{$IFC THINK_PASCAL }
		SetIText(statusPtr^.Ext, Str);
		NumToString(statusPtr^.Eye.y, Str);
		SetIText(statusPtr^.Eyt, Str);
		NumToString(statusPtr^.Eye.z, Str);
		SetIText(statusPtr^.Ezt, Str);
		NumToString(statusPtr^.Eye.vx, Str);
		SetIText(statusPtr^.vxt, Str);
		NumToString(statusPtr^.Eye.vy, Str);
		SetIText(statusPtr^.vyt, Str);
		NumToString(statusPtr^.Eye.vz, Str);
		SetIText(statusPtr^.vzt, Str);
{$ELSEC }
		SetDialogItemText(statusPtr^.ext, Str);
		NumToString(statusPtr^.Eye.y, Str);
		SetDialogItemText(statusPtr^.eyt, Str);
		NumToString(statusPtr^.Eye.z, Str);
		SetDialogItemText(statusPtr^.ezt, Str);
		NumToString(statusPtr^.Eye.vx, Str);
		SetDialogItemText(statusPtr^.vxt, Str);
		NumToString(statusPtr^.Eye.vy, Str);
		SetDialogItemText(statusPtr^.vyt, Str);
		NumToString(statusPtr^.Eye.vz, Str);
		SetDialogItemText(statusPtr^.vzt, Str);
{$ENDC }
		dx := statusPtr^.Eye.x - statusPtr^.Eye.vx;
		dy := statusPtr^.Eye.y - statusPtr^.Eye.vy;
		dz := statusPtr^.Eye.z - statusPtr^.Eye.vz;
		dd := sqrt(dx * dx + dy * dy + dz * dz);
{$IFC THINK_PASCAL }
		delta := round(dd);
		NumToString(delta, Str);
		SetIText(statusPtr^.dt, Str);
{$ELSEC }
		delta := roundLong(dd);
		NumToString(delta, Str);
		SetDialogItemText(statusPtr^.dt, Str);
{$ENDC }
	end;

(* Get all Dialog items from Dialog  *)
procedure GetItems;

	var
		dummyInt: Integer;
		dummyRect: Rect;
		dummyHandle: Handle;

	begin
{$IFC THINK_PASCAL }
		GetDItem(Dlg, okButton, dummyInt, dummyHandle, statusPtr^.okRect);
		InsetRect(statusPtr^.okRect, -4, -4);
		GetDItem(Dlg, xypanel, dummyInt, statusPtr^.xyplanet, statusPtr^.xypRect);
		GetDItem(Dlg, xzpanel, dummyInt, statusPtr^.xzplanet, statusPtr^.xzpRect);
		GetDItem(Dlg, central, dummyInt, statusPtr^.centralt, dummyRect);
		GetDItem(Dlg, parallel, dummyInt, statusPtr^.parallelt, dummyRect);
		GetDItem(Dlg, Ex, dummyInt, statusPtr^.Ext, dummyRect);
		GetDItem(Dlg, Ey, dummyInt, statusPtr^.Eyt, dummyRect);
		GetDItem(Dlg, Ez, dummyInt, statusPtr^.Ezt, dummyRect);
		GetDItem(Dlg, Vx, dummyInt, statusPtr^.Vxt, dummyRect);
		GetDItem(Dlg, Vy, dummyInt, statusPtr^.Vyt, dummyRect);
		GetDItem(Dlg, Vz, dummyInt, statusPtr^.Vzt, dummyRect);
		GetDItem(Dlg, dist, dummyInt, statusPtr^.dt, dummyRect);
		GetDItem(Dlg, hiddenLine, dummyInt, statusPtr^.hiddenLinet, dummyRect); (* check boxes *)
(*GetDitem(Dlg, UseObjectScetch, dummyInt, UseObjectScetcht, dummyRect); *)
		GetDItem(Dlg, AutoApply, dummyInt, statusPtr^.autoApplyt, dummyRect);
		GetDItem(Dlg, ZoomIn, dummyInt, statusPtr^.ZoomInt, dummyRect);
		GetDItem(Dlg, ZoomOut, dummyInt, statusPtr^.ZoomOutt, dummyRect);
{$ELSEC }
		GetDialogItem(Dlg, okButton, dummyInt, dummyHandle, statusPtr^.okRect);
		InsetRect(statusPtr^.okRect, -4, -4);
		GetDialogItem(Dlg, xypanel, dummyInt, statusPtr^.xyplanet, statusPtr^.xypRect);
		GetDialogItem(Dlg, xzpanel, dummyInt, statusPtr^.xzplanet, statusPtr^.xzpRect);
		GetDialogItem(Dlg, Central, dummyInt, statusPtr^.centralt, dummyRect);
		GetDialogItem(Dlg, parallel, dummyInt, statusPtr^.parallelt, dummyRect);
		GetDialogItem(Dlg, Ex, dummyInt, statusPtr^.ext, dummyRect);
		GetDialogItem(Dlg, Ey, dummyInt, statusPtr^.eyt, dummyRect);
		GetDialogItem(Dlg, Ez, dummyInt, statusPtr^.ezt, dummyRect);
		GetDialogItem(Dlg, Vx, dummyInt, statusPtr^.vxt, dummyRect);
		GetDialogItem(Dlg, Vy, dummyInt, statusPtr^.vyt, dummyRect);
		GetDialogItem(Dlg, Vz, dummyInt, statusPtr^.vzt, dummyRect);
		GetDialogItem(Dlg, dist, dummyInt, statusPtr^.dt, dummyRect);
		GetDialogItem(Dlg, hiddenLine, dummyInt, statusPtr^.hiddenLinet, dummyRect); (* check boxes *)
(*GetDitem(Dlg, UseObjectScetch, dummyInt, UseObjectScetcht, dummyRect); *)
		GetDialogItem(Dlg, AutoApply, dummyInt, statusPtr^.autoApplyt, dummyRect);
		GetDialogItem(Dlg, ZoomIn, dummyInt, statusPtr^.ZoomInt, dummyRect);
		GetDialogItem(Dlg, ZoomOut, dummyInt, statusPtr^.ZoomOutt, dummyRect);
{$ENDC }
		GetIndPattern(statusPtr^.thePattern, PatternList, ObjectPatternID);
		statusPtr^.xw := statusPtr^.xypRect.right - statusPtr^.xypRect.left; (* calc unit that is the width of the panel *)
		statusPtr^.yw := statusPtr^.xypRect.bottom - statusPtr^.xypRect.top;
		statusPtr^.zw := statusPtr^.xzpRect.bottom - statusPtr^.xzpRect.top;
	end;


(* Set Initial Values for Dialog Items and set up callback procs for useritems *)
procedure InitItems;
	begin
{$IFC THINK_PASCAL }
		SetDItem(Dlg, xypanel, userItem, @DrawxyPanelItem, statusPtr^.xypRect);
		SetDItem(Dlg, xzpanel, userItem, @DrawxzpanelItem, statusPtr^.xzpRect);
		UpdateTextItems;
		if statusPtr^.Eye.projectionType = 0 then (* parallel *)
			SetCtlValue(ControlHandle(statusPtr^.parallelt), 1)
		else
			SetCtlValue(ControlHandle(statusPtr^.centralt), 1);
		if statusPtr^.Eye.hidden then
			SetCtlValue(ControlHandle(statusPtr^.hiddenLinet), 1)
		else
			SetCtlValue(ControlHandle(statusPtr^.hiddenLinet), 0);
		if statusPtr^.autoApplyOn then
			SetCtlValue(ControlHandle(statusPtr^.autoApplyt), 1)
		else
			SetCtlValue(ControlHandle(statusPtr^.autoApplyt), 0);
		if statusPtr^.Eye.hidden then
			SetCtlValue(ControlHandle(statusPtr^.hiddenLinet), 1);
{$ELSEC }
		SetDialogItem(Dlg, xypanel, userItem, Handle(NewUserItemProc(@DrawxyPanelItem)), statusPtr^.xypRect);
		SetDialogItem(Dlg, xzpanel, userItem, Handle(NewUserItemProc(@DrawxzPanelItem)), statusPtr^.xzpRect);
		UpdateTextItems;
		if statusPtr^.Eye.ProjectionType = 0 then (* parallel *)
			SetControlValue(ControlHandle(statusPtr^.parallelt), 1)
		else
			SetControlValue(ControlHandle(statusPtr^.centralt), 1);
		if statusPtr^.Eye.hidden then
			SetControlValue(ControlHandle(statusPtr^.hiddenLinet), 1)
		else
			SetControlValue(ControlHandle(statusPtr^.hiddenLinet), 0);
		if statusPtr^.autoApplyOn then
			SetControlValue(ControlHandle(statusPtr^.autoApplyt), 1)
		else
			SetControlValue(ControlHandle(statusPtr^.autoApplyt), 0);
		if statusPtr^.Eye.hidden then
			SetControlValue(ControlHandle(statusPtr^.hiddenLinet), 1);
{$ENDC }

{$IFC FALSE THEN }
		if (TopPict = nil) or (SidePict = nil) then
			HasPic := FALSE (* user has passed no picture *)
		else
			HasPic := TRUE;
		if HasPic then begin
			SetCtlValue(ControlHandle(UseObjectScetcht), 1);
			HiliteControl(ControlHandle(UseObjectScetcht), 0);
			UsePict := True;
		end
		else begin
			SetCtlValue(ControlHandle(UseObjectScetcht), 0);
			HiliteControl(ControlHandle(UseObjectScetcht), 255);
			UsePict := False;
		end;
		UsePict := False;
{$ENDC }
	end;


(* User clicked one of the Radiobuttons *)
procedure ChangeProjection (theItem: Integer);

	begin
		case theItem of
			parallel:  begin
{$IFC THINK_PASCAL }
				SetCtlValue(ControlHandle(statusPtr^.parallelt), 1);
				SetCtlValue(ControlHandle(statusPtr^.centralt), 0);
				statusPtr^.Eye.projectionType := 0;
			end;

			central:  begin
				SetCtlValue(ControlHandle(statusPtr^.parallelt), 0);
				SetCtlValue(ControlHandle(statusPtr^.centralt), 1);
{$ELSEC }
				SetControlValue(ControlHandle(statusPtr^.parallelt), 1);
				SetControlValue(ControlHandle(statusPtr^.centralt), 0);
				statusPtr^.Eye.ProjectionType := 0;
			end;

			Central:  begin
				SetControlValue(ControlHandle(statusPtr^.parallelt), 0);
				SetControlValue(ControlHandle(statusPtr^.centralt), 1);
{$ENDC }
				statusPtr^.Eye.ProjectionType := 1;
			end;
			otherwise
				;
		end; (* case *)
	end;

(* User clicked AutoApply CheckBox. *)
procedure ChangeAutoApply;

	var
		theVal: Integer;

	begin
{$IFC THINK_PASCAL }
		theVal := GetCtlValue(ControlHandle(statusPtr^.autoApplyt));
		if theVal = 0 then begin
			SetCtlValue(ControlHandle(statusPtr^.autoApplyt), 1);
			statusPtr^.autoApplyOn := true;
		end
		else begin
			SetCtlValue(ControlHandle(statusPtr^.autoApplyt), 0);
			statusPtr^.autoApplyOn := false;
		end;
{$ELSEC }
		theVal := GetControlValue(ControlHandle(statusPtr^.autoApplyt));
		if theVal = 0 then begin
			SetControlValue(ControlHandle(statusPtr^.autoApplyt), 1);
			statusPtr^.autoApplyOn := true;
		end
		else begin
			SetControlValue(ControlHandle(statusPtr^.autoApplyt), 0);
			statusPtr^.autoApplyOn := false;
		end;
{$ENDC }
	end;



(* User clicked in Hidden Line Checkbox. handle this *)
procedure ChangeHiddenAttribute;

	var
		theVal: Integer;

	begin
{$IFC THINK_PASCAL }
		theVal := GetCtlValue(ControlHandle(statusPtr^.hiddenLinet));
		if theVal = 0 then begin
			SetCtlValue(ControlHandle(statusPtr^.hiddenLinet), 1);
			statusPtr^.Eye.hidden := true;
		end
		else begin
			SetCtlValue(ControlHandle(statusPtr^.hiddenLinet), 0);
			statusPtr^.Eye.hidden := false;
		end;
{$ELSEC }
		theVal := GetControlValue(ControlHandle(statusPtr^.hiddenLinet));
		if theVal = 0 then begin
			SetControlValue(ControlHandle(statusPtr^.hiddenLinet), 1);
			statusPtr^.Eye.hidden := true;
		end
		else begin
			SetControlValue(ControlHandle(statusPtr^.hiddenLinet), 0);
			statusPtr^.Eye.hidden := false;
		end;
{$ENDC }
	end;

{$IFC false THEN }
(* User clicked in Use Object Scetch Checkbox. handle this *)
procedure ChangeScetchAttribute;
	var
		theVal: Integer;
	begin
		theVal := GetCtlValue(ControlHandle(UseObjectScetcht));
		if theVal = 0 then begin
			SetCtlValue(ControlHandle(UseObjectScetcht), 1);
			UsePict := TRUE;
		end
		else begin
			SetCtlValue(ControlHandle(UseObjectScetcht), 0);
			UsePict := FALSE;
		end;
		RedrawPanels;
	end;
{$ENDC }


(* User changed one of the Text-Based Parameters. Recalc projection           *)
(* distance. Update also the Viewing User items to reflect changes               *)
(* Update also Eye (local unit) data struct to reflect changes directly            *)

procedure CalcProjectionDist;

	var
		dx, dy, dz: Longint;
		dd: longint;
		newEye: EyeDesc;
		incorrect: boolean;

	begin
		newEye := statusPtr^.Eye;
		with statusPtr^ do begin
			getValue(Ex, Eye.x, newEye.x);
			getValue(Vx, Eye.vx, newEye.vx);
			dx := newEye.x - newEye.vx;
			getValue(Ey, Eye.y, newEye.y);
			getValue(Vy, Eye.vy, newEye.vy);
			dy := newEye.y - newEye.vy;
			getValue(Ez, Eye.z, newEye.z);
			getValue(Vz, Eye.vz, newEye.vz);
			dz := newEye.z - newEye.vz;
{$IFC THINK_PASCAL }
			dd := round(sqrt(dx * dx + dy * dy + dz * dz));
{$ELSEC }
			dd := roundLong(sqrt(dx * dx + dy * dy + dz * dz));
{$ENDC }
			incorrect := ((dz = dy) & (dy = dx) & (dx = 0)) | (abs(dd) > 32000);
			if not incorrect then
				statusPtr^.Eye := newEye;
		end; (*with *)
		with statusPtr^.Eye do begin
			dx := x - vx;
			dy := y - vy;
			dz := z - vz;
{$IFC THINK_PASCAL }
			dd := round(sqrt(dx * dx + dy * dy + dz * dz));
{$ELSEC }
			dd := roundLong(sqrt(dx * dx + dy * dy + dz * dz));
{$ENDC }
			UpdateTextItems;
			RedrawPanels;
		end; (* with *)
	end;


(* CalcDirection calculates the unitary (len = 1) vector of the      *)
(* viewing direction  (from Eye to Viewpoint)  de.                        *)

procedure CalcDirection;

	var
		dx, dy, dz, dd: longint;

	begin
		with statusPtr^.Eye do begin

			dx := vx - x;
			dy := vy - y;
			dz := vz - z;
{$IFC THINK_PASCAL }
			dd := round(sqrt(dx * dx + dy * dy + dz * dz));
{$ELSEC }
			dd := roundLong(sqrt(dx * dx + dy * dy + dz * dz));
{$ENDC }
			if dd = 0 then begin
				dd := 1;
				statusPtr^.de.x := 1;
				dx := 1;
				statusPtr^.de.y := 0;
				statusPtr^.de.z := 0;
			end;
			statusPtr^.de.x := dx / dd;
			statusPtr^.de.y := dy / dd;
			statusPtr^.de.z := dz / dd;
		end;
	end;


(* CalcViewPoint recalcs the viewpoint. Given old direction it will *)
(* set the new viepoint according to distance in the same direction *)
procedure CalcViewPoint;
	var
		displacement, def, dx, dy, dz: longint;
		Str: Str255;
	begin
		CalcDirection; (* Set up the de (delta Real) vector for new calculations *)
		dx := statusPtr^.Eye.x - statusPtr^.Eye.vx;
		dy := statusPtr^.Eye.y - statusPtr^.Eye.vy;
		dz := statusPtr^.Eye.z - statusPtr^.Eye.vz;
{$IFC THINK_PASCAL }
		def := round(sqrt(dx * dx + dy * dy + dz * dz));
{$ELSEC }
		def := roundLong(sqrt(dx * dx + dy * dy + dz * dz));
{$ENDC }
		getValue(dist, def, displacement);
		if (displacement > 0) & (displacement < 32000) then begin
			with statusPtr^.Eye do begin
{$IFC THINK_PASCAL }
				vx := x + round(displacement * statusPtr^.de.x);
				vy := y + round(displacement * statusPtr^.de.y);
				vz := z + round(displacement * statusPtr^.de.z);
{$ELSEC }
				vx := x + roundLong(displacement * statusPtr^.de.x);
				vy := y + roundLong(displacement * statusPtr^.de.y);
				vz := z + roundLong(displacement * statusPtr^.de.z);
{$ENDC }
			end;
		end
		else
			displacement := def;
		NumToString(statusPtr^.Eye.vx, Str);
{$IFC THINK_PASCAL }
		SetIText(statusPtr^.vxt, Str);
		NumToString(statusPtr^.Eye.vy, Str);
		SetIText(statusPtr^.vyt, Str);
		NumToString(statusPtr^.Eye.vz, Str);
		SetIText(statusPtr^.vzt, Str);
		NumToString(displacement, Str);
		SetIText(statusPtr^.dt, Str);
{$ELSEC }
		SetDialogItemText(statusPtr^.vxt, Str);
		NumToString(statusPtr^.Eye.vy, Str);
		SetDialogItemText(statusPtr^.vyt, Str);
		NumToString(statusPtr^.Eye.vz, Str);
		SetDialogItemText(statusPtr^.vzt, Str);
		NumToString(displacement, Str);
		SetDialogItemText(statusPtr^.dt, Str);
{$ENDC }
		RedrawPanels;
	end;


(* move viewpoint 50% closer in viewing direction to Eye *)

procedure HandleCloser;
	var
		displacement, def, dx, dy, dz: longint;
		Str: Str255;
	begin
		CalcDirection; (* Set up the de (delta Real) vector for new calculations *)
		dx := statusPtr^.Eye.x - statusPtr^.Eye.vx;
		dy := statusPtr^.Eye.y - statusPtr^.Eye.vy;
		dz := statusPtr^.Eye.z - statusPtr^.Eye.vz;
{$IFC THINK_PASCAL }
		def := round(sqrt(dx * dx + dy * dy + dz * dz));
{$ELSEC }
		def := roundLong(sqrt(dx * dx + dy * dy + dz * dz));
{$ENDC }
		getValue(dist, def, displacement);
		if abs(displacement) > 3 then begin
			displacement := displacement div 2;
			with statusPtr^.Eye do begin
{$IFC THINK_PASCAL }
				vx := x + round(displacement * statusPtr^.de.x);
				vy := y + round(displacement * statusPtr^.de.y);
				vz := z + round(displacement * statusPtr^.de.z);
{$ELSEC }
				vx := x + roundLong(displacement * statusPtr^.de.x);
				vy := y + roundLong(displacement * statusPtr^.de.y);
				vz := z + roundLong(displacement * statusPtr^.de.z);
{$ENDC }
			end;
		end
		else
			displacement := def;
		NumToString(statusPtr^.Eye.vx, Str);
{$IFC THINK_PASCAL }
		SetIText(statusPtr^.vxt, Str);
		NumToString(statusPtr^.Eye.vy, Str);
		SetIText(statusPtr^.vyt, Str);
		NumToString(statusPtr^.Eye.vz, Str);
		SetIText(statusPtr^.vzt, Str);
		NumToString(displacement, Str);
		SetIText(statusPtr^.dt, Str);
{$ELSEC }
		SetDialogItemText(statusPtr^.vxt, Str);
		NumToString(statusPtr^.Eye.vy, Str);
		SetDialogItemText(statusPtr^.vyt, Str);
		NumToString(statusPtr^.Eye.vz, Str);
		SetDialogItemText(statusPtr^.vzt, Str);
		NumToString(displacement, Str);
		SetDialogItemText(statusPtr^.dt, Str);
{$ENDC }
		RedrawPanels;
	end;

(* move viewpoint away in viewing direction from Eye *)
procedure HandleFurther;

	var
		dx, dy, dz, displacement, def: longint;
		Str: Str255;

	begin
		CalcDirection; (* Set up the de (delta Real) vector for new calculations *)
		dx := statusPtr^.Eye.x - statusPtr^.Eye.vx;
		dy := statusPtr^.Eye.y - statusPtr^.Eye.vy;
		dz := statusPtr^.Eye.z - statusPtr^.Eye.vz;
{$IFC THINK_PASCAL }
		def := round(sqrt(dx * dx + dy * dy + dz * dz));
{$ELSEC }
		def := roundLong(sqrt(dx * dx + dy * dy + dz * dz));
{$ENDC }
		getValue(dist, def, displacement);
		if abs(displacement) < 16000 then begin
			displacement := displacement * 2;
			with statusPtr^.Eye do begin
{$IFC THINK_PASCAL }
				dx := x + round(displacement * statusPtr^.de.x);
				dy := y + round(displacement * statusPtr^.de.y);
				dz := z + round(displacement * statusPtr^.de.z);
{$ELSEC }
				dx := x + roundLong(displacement * statusPtr^.de.x);
				dy := y + roundLong(displacement * statusPtr^.de.y);
				dz := z + roundLong(displacement * statusPtr^.de.z);
{$ENDC }
				if (abs(dx) < 32000) & (abs(dy) < 32000) & (abs(dz) < 32000) then begin
					vx := dx;
					vy := dy;
					vz := dz;
				end;
			end;
		end
		else
			displacement := def;
		NumToString(statusPtr^.Eye.vx, Str);
{$IFC THINK_PASCAL }
		SetIText(statusPtr^.vxt, Str);
		NumToString(statusPtr^.Eye.vy, Str);
		SetIText(statusPtr^.vyt, Str);
		NumToString(statusPtr^.Eye.vz, Str);
		SetIText(statusPtr^.vzt, Str);
		NumToString(displacement, Str);
		SetIText(statusPtr^.dt, Str);
{$ELSEC }
		SetDialogItemText(statusPtr^.vxt, Str);
		NumToString(statusPtr^.Eye.vy, Str);
		SetDialogItemText(statusPtr^.vyt, Str);
		NumToString(statusPtr^.Eye.vz, Str);
		SetDialogItemText(statusPtr^.vzt, Str);
		NumToString(displacement, Str);
		SetDialogItemText(statusPtr^.dt, Str);
{$ENDC }
		RedrawPanels;
	end;

procedure HandleZoomIn;
	var
		dummy: boolean;

	begin
		dummy := CalcZoomIn;
		RedrawPanels;
		if not CheckZoomIn then (* cannot zoomIn *)
			HiliteControl(ControlHandle(statusPtr^.ZoomInt), 255);
		HiliteControl(ControlHandle(statusPtr^.ZoomOutt), 0);
	end;

procedure HandleZoomOut;
	var
		dummy: boolean;
	begin
		dummy := CalcZoomOut;
		RedrawPanels;
		if not CheckZoomOut then
			HiliteControl(ControlHandle(statusPtr^.ZoomOutt), 255);
		HiliteControl(ControlHandle(statusPtr^.ZoomInt), 0);
	end;


(* trackboth : track eye and viewpoint *)

function TrackBoth (item1, item2: Rect; start: Point; limit, slop: Rect): Point;

	var
		itemRgn: RgnHandle;
		result: longint;

	begin
		itemRgn := NewRgn;
		OpenRgn;
		FrameOval(item1);
		FrameOval(item2);
		MoveTo((item1.left + item1.right) div 2, (item1.top + item1.bottom) div 2); (* third region: connector *)
		LineTo((item2.left + item2.right) div 2, (item2.top + item2.bottom) div 2);
		LineTo((item2.left + item2.right) div 2 + 1, (item2.top + item2.bottom) div 2 + 1);
		LineTo((item1.left + item1.right) div 2 + 1, (item1.top + item1.bottom) div 2 + 1);
		LineTo((item1.left + item1.right) div 2, (item1.top + item1.bottom) div 2); (* and back again to close region *)
		CloseRgn(itemRgn);
		result := DragGrayRgn(itemRgn, start, limit, slop, noConstraint, nil);
		TrackBoth := start;
		if HiWord(result) = $8000 then (* no change *)
		else begin
			TrackBoth.v := HiWord(result) + start.v;
			TrackBoth.h := LoWord(result) + start.h;
		end;
		DisposeRgn(itemRgn);
	end;

(* TrackBall will track a circle within the given region *)
(*                 See IM I-294 for description of parameters *)

function TrackBall (Item: Rect; start: Point; limit, slop: Rect): Point;

	var
		itemRgn: RgnHandle;
		result: longint;

	begin
		itemRgn := NewRgn;
		OpenRgn;
		FrameOval(Item);
		CloseRgn(itemRgn);
		result := DragGrayRgn(itemRgn, start, limit, slop, noConstraint, nil);
		TrackBall := start;
		if HiWord(result) = $8000 then (* no change *)
		else begin
			TrackBall.v := HiWord(result) + start.v;
			TrackBall.h := LoWord(result) + start.h;
		end;
		DisposeRgn(itemRgn);
	end;

(* User Clicked in yz-Panel. Look if clicked in eye or viewpoint and if so, drag to new *)
(* coordinates *)

procedure HandleXZClick;

	var
		eyeRect, viewRect: Rect;
		thePoint: Point;
		savePoint: Point; (* save for converted points *)
		delta: Point; (* contains really a vector pointing from eye to viewpoint *)
		limit, slop: Rect;
		SaveRgn: RgnHandle;
		shift: Boolean;
		theKeys: KeyMap;
		change, ok: boolean; (* change is true if a change happened *)
		oldEye: EyeDesc;

	begin
(* first, set up the two rects that contain the local coordinates for eye and viewpoint *)
		oldEye := statusPtr^.Eye;
		change := false;
		thePoint.h := statusPtr^.Eye.x;
		thePoint.v := statusPtr^.Eye.z;
		ok := ConvertPointzx(thePoint); (* convert to local screen coordinates *)

		with eyeRect do begin
			top := thePoint.v;
			left := thePoint.h;
			bottom := thePoint.v + 1;
			right := thePoint.h + 1;
		end;
		InsetRect(eyeRect, -4, -4);
		delta := thePoint; (* save point for calculation of delta *)

		thePoint.h := statusPtr^.Eye.vx;
		thePoint.v := statusPtr^.Eye.vz;
		ok := ConvertPointzx(thePoint);

		with viewRect do begin
			top := thePoint.v;
			left := thePoint.h;
			bottom := thePoint.v + 1;
			right := thePoint.h + 1;
		end;
		InsetRect(viewRect, -4, -4);
		delta.h := thePoint.h - delta.h;
		delta.v := thePoint.v - delta.v; (* calculate vector from eye to view *)

(* now set up limit and slop rect *)
		limit := statusPtr^.xzpRect;
		slop := limit;
		InsetRect(slop, -4, -4);

(* now set clipping rectangle for dragging *)
		SaveRgn := NewRgn;
		GetClip(SaveRgn);
		ClipRect(statusPtr^.xzpRect);

(* look if shift was pressed *)
		GetKeys(theKeys);
		shift := theKeys[ShiftCode];

(* now look if a click was within either eye or viewpoint *)
		GetMouse(thePoint); (* already local *)
		if shift then begin
			if PtInRect(thePoint, eyeRect) then
      (* User clicked to move the Eye. Track and update *)
				begin
				thePoint := TrackBoth(eyeRect, viewRect, thePoint, limit, slop);
				savePoint := thePoint;
				ok := RevertPointzx(thePoint); (* recalc to world coordinates *)
				statusPtr^.Eye.x := thePoint.h;
				statusPtr^.Eye.z := thePoint.v;
				savePoint.h := savePoint.h + delta.h;
				savePoint.v := savePoint.v + delta.v;
				ok := RevertPointzx(savePoint);
				statusPtr^.Eye.vx := savePoint.h;
				statusPtr^.Eye.vz := savePoint.v;
				change := true;
			end
			else if PtInRect(thePoint, viewRect) then begin
				thePoint := TrackBoth(eyeRect, viewRect, thePoint, limit, slop);
				savePoint := thePoint;
				ok := RevertPointzx(thePoint); (* recalc to world coordinates *)
				statusPtr^.Eye.vx := thePoint.h;
				statusPtr^.Eye.vz := thePoint.v;
				savePoint.h := savePoint.h - delta.h;
				savePoint.v := savePoint.v - delta.v;
				ok := RevertPointzx(savePoint);
				statusPtr^.Eye.x := savePoint.h;
				statusPtr^.Eye.z := savePoint.v;
				change := true;
			end;
		end (* if shift *)
		else begin (* not shift *)
			if PtInRect(thePoint, eyeRect) then
      (* User clicked to move the Eye. Track and update *)
				begin
				thePoint := TrackBall(eyeRect, thePoint, limit, slop);
				ok := RevertPointzx(thePoint); (* recalc to world coordinates *)
				statusPtr^.Eye.x := thePoint.h;
				statusPtr^.Eye.z := thePoint.v;
				if ((statusPtr^.Eye.x = statusPtr^.Eye.vx) & (statusPtr^.Eye.y = statusPtr^.Eye.vy) & (statusPtr^.Eye.z = statusPtr^.Eye.vz)) then begin
					statusPtr^.Eye := oldEye;
				end;
				change := true;
			end
			else if PtInRect(thePoint, viewRect) then begin
				thePoint := TrackBall(viewRect, thePoint, limit, slop);
				ok := RevertPointzx(thePoint); (* recalc to world coordinates *)
				statusPtr^.Eye.vx := thePoint.h;
				statusPtr^.Eye.vz := thePoint.v;
				if ((statusPtr^.Eye.x = statusPtr^.Eye.vx) & (statusPtr^.Eye.y = statusPtr^.Eye.vy) & (statusPtr^.Eye.z = statusPtr^.Eye.vz)) then begin
					statusPtr^.Eye := oldEye;
				end;
				change := true;
			end;
		end; (* else not shift *)

(* restore clip and redraw if change true  *)
		SetClip(SaveRgn);
		if change then begin
			RedrawPanels;
			UpdateTextItems; (* fill these values into Text Items *)
		end;
		DisposeRgn(SaveRgn);
	end;

procedure HandleXYClick;

	var
		eyeRect, viewRect: Rect;
		thePoint: Point;
		delta: Point;
		savePoint: Point;
		limit, slop: Rect;
		SaveRgn: RgnHandle;
		shift: Boolean;
		theKeys: KeyMap;
		change, ok: boolean; (* change is true if a change happened *)
		oldEye: EyeDesc;

	begin
(* first, set up the two rects that contain the local coordinates for eye and viewpoint *)
		oldEye := statusPtr^.Eye;
		change := false;
		thePoint.h := statusPtr^.Eye.x;
		thePoint.v := statusPtr^.Eye.y;
		ok := ConvertPointxy(thePoint); (* convert to local screen coordinates *)
		with eyeRect do begin
			top := thePoint.v;
			left := thePoint.h;
			bottom := thePoint.v + 1;
			right := thePoint.h + 1;
		end;
		InsetRect(eyeRect, -4, -4);
		delta := thePoint;

		thePoint.h := statusPtr^.Eye.vx;
		thePoint.v := statusPtr^.Eye.vy;
		ok := ConvertPointxy(thePoint);
		with viewRect do begin
			top := thePoint.v;
			left := thePoint.h;
			bottom := thePoint.v + 1;
			right := thePoint.h + 1;
		end;
		InsetRect(viewRect, -4, -4);
		delta.h := thePoint.h - delta.h;
		delta.v := thePoint.v - delta.v; (* calculate vector from eye to view *)

(* now set up limit and slop rect *)
		limit := statusPtr^.xypRect;
		slop := limit;
		InsetRect(slop, -4, -4);

(* now set clipping rectangle for dragging *)
		SaveRgn := NewRgn;
		GetClip(SaveRgn);
		ClipRect(statusPtr^.xypRect);

(* look if shift was pressed *)
		GetKeys(theKeys);
		shift := theKeys[ShiftCode];
	(* now look if a click was within either eye or viewpoint *)
		GetMouse(thePoint); (* already local *)
		if shift then begin
			if PtInRect(thePoint, eyeRect) then
      (* User clicked to move the Eye. Track and update *)
				begin
				thePoint := TrackBoth(eyeRect, viewRect, thePoint, limit, slop);
				savePoint := thePoint;
				ok := RevertPointxy(thePoint); (* recalc to world coordinates *)
				statusPtr^.Eye.x := thePoint.h;
				statusPtr^.Eye.y := thePoint.v;
				savePoint.h := savePoint.h + delta.h;
				savePoint.v := savePoint.v + delta.v;
				ok := RevertPointxy(savePoint);
				statusPtr^.Eye.vx := savePoint.h;
				statusPtr^.Eye.vy := savePoint.v;
				change := TRUE;
			end
			else if PtInRect(thePoint, viewRect) then begin
				thePoint := TrackBoth(eyeRect, viewRect, thePoint, limit, slop);
				savePoint := thePoint;
				ok := RevertPointxy(thePoint); (* recalc to world coordinates *)
				statusPtr^.Eye.vx := thePoint.h;
				statusPtr^.Eye.vy := thePoint.v;
				savePoint.h := savePoint.h - delta.h;
				savePoint.v := savePoint.v - delta.v;
				ok := RevertPointxy(savePoint);
				statusPtr^.Eye.x := savePoint.h;
				statusPtr^.Eye.y := savePoint.v;
				change := TRUE;
			end;
		end (* if shift *)
		else begin (* not shift *)
			if PtInRect(thePoint, eyeRect) then
      (* User clicked to move the Eye. Track and update *)
				begin
				thePoint := TrackBall(eyeRect, thePoint, limit, slop);
				ok := RevertPointxy(thePoint); (* recalc to world coordinates *)
				statusPtr^.Eye.x := thePoint.h;
				statusPtr^.Eye.y := thePoint.v;
				if ((statusPtr^.Eye.x = statusPtr^.Eye.vx) & (statusPtr^.Eye.y = statusPtr^.Eye.vy) & (statusPtr^.Eye.z = statusPtr^.Eye.vz)) then begin
					statusPtr^.Eye := oldEye;
				end;
				change := TRUE;
			end
			else if PtInRect(thePoint, viewRect) then begin
				thePoint := TrackBall(viewRect, thePoint, limit, slop);
				ok := RevertPointxy(thePoint); (* recalc to world coordinates *)
				statusPtr^.Eye.vx := thePoint.h;
				statusPtr^.Eye.vy := thePoint.v;
				if ((statusPtr^.Eye.x = statusPtr^.Eye.vx) & (statusPtr^.Eye.y = statusPtr^.Eye.vy) & (statusPtr^.Eye.z = statusPtr^.Eye.vz)) then begin
					statusPtr^.Eye := oldEye;
				end;
				change := TRUE;
			end;
		end; (* else not shift *)
	(* restore clip and redraw if change true  *)
		SetClip(SaveRgn);
		if change then begin
			RedrawPanels;
			UpdateTextItems; (* fill these values into Text Items *)
		end;
		DisposeRgn(SaveRgn);
	end;


procedure HandleResetPanel;
	begin
		CalcBounds;
		CalcZoomFactor;
		RedrawPanels;
		if not CheckZoomIn then
			HiliteControl(ControlHandle(statusPtr^.ZoomInt), 255)
		else
			HiliteControl(ControlHandle(statusPtr^.ZoomInt), 0);
		if not CheckZoomOut then
			HiliteControl(ControlHandle(statusPtr^.ZoomOutt), 255)
		else
			HiliteControl(ControlHandle(statusPtr^.ZoomOutt), 0);
	end;


function MyFilter (theDialog: DialogPtr; var theEvent: EventRecord; var itemHit: integer): boolean;
	const
		charCodeMask = $000000FF;
	var
		theWindow: WindowPtr;
		dragRect: Rect;
		charCode: integer;
	begin
		MyFilter := false;
		if (theEvent.what = mouseDown) & (FindWindow(theEvent.where, theWindow) = inDrag) & (theWindow = theDialog) then begin
{$IFC THINK_PASCAL }
			with screenBits.bounds do
{$ELSEC }
			with qd.screenBits.bounds do
{$ENDC }
				SetRect(dragRect, 4, 24, right - 4, bottom - 4);
			DragWindow(theWindow, theEvent.where, dragRect);
			MyFilter := true;
			itemHit := apply;
		end;
		if (theEvent.what = keyDown) then begin
			charCode := BitAnd(theEvent.message, charCodeMask);
			if (charCode = $0D) or (charCode = $03) then begin
				MyFilter := true;
				itemHit := okButton;
			end;
		end;
	end;

(* Main Routine. Bring up Dialog, Init all values and handle *)
(* all actions (via switch to subroutines)                           *)

(* SetEye :  Bring up dialog for setting interactively viewing parameters.  *)
(*                ENTRY : theEye initialized to current viewing parameters       *)
(*                             Bounds3D contains minimal and maximal coordinates  *)
(*                                            of objects to view. these will be used to     *)
(*                                            calibrate the setting box and zoom settings *)
(*                             TopPic is handle to picture containing a view straight *)
(*                                         down onto the object or nil if no picture         *)
(*                             SidePic is handle to picture showing object seen from *)
(*                                         one side (straight down x-achsis)                   *)
(*               EXIT : SetEye will return an integer value indicating the folowing *)
(*                          0 : OK, everything done. Usually followed by a redraw      *)
(*                               of the object with new parameters                               *)
(*                          1 : Apply current parameters: redraw object with            *)
(*                                current params and recall SetEye procedure                *)
(*                        -1 :  Error. Ignore all changes. SetEye will also return -1 if *)
(*                                User clicked Cancel Button                                           *)

function SetEye (var theEye: EyeDesc; Bounds3D: Cube; DlgP: DialogPtr): Integer;

	var
		wPeek: WindowPeek;
		autoApp: boolean;

	begin
		Dlg := DlgP;
		wPeek := pointer(FrontWindow);					(* Is SetEye called for the first time? *)
		if (pointer(wPeek^.refCon) = nil) then begin
	 (* yes it is *)
			new(statusPtr);
			failNil(Handle(statusPtr));
			statusPtr^.Eye := theEye;
			statusPtr^.theCube := Bounds3D;
			wPeek^.refCon := longint(statusPtr);
			GetItems; 						(* Get Dialog Items and put their Handles into globals 	*)
			CalcBounds;						(* calculates zoomCube 										*)
			CalcZoomFactor;				(* calculates zoomFactor such that everything is visible	*)
			InitItems;						(* Put Initial Values into the fields 							*)
		end
		else begin
	 (* no it isn't *)
			statusPtr := pointer(wPeek^.refCon);
		end;
		if not CheckZoomIn then
			HiliteControl(ControlHandle(statusPtr^.ZoomInt), 255)
		else
			HiliteControl(ControlHandle(statusPtr^.ZoomInt), 0);
		if not CheckZoomOut then
			HiliteControl(ControlHandle(statusPtr^.ZoomOutt), 255)
		else
			HiliteControl(ControlHandle(statusPtr^.ZoomOutt), 0);

		PenSize(3, 3);
		FrameRoundRect(statusPtr^.okRect, 16, 16);
		PenSize(1, 1);
		repeat
{$IFC THINK_PASCAL }
			ModalDialog(@myFilter, itemHit);
{$ELSEC }
			ModalDialog(NewModalFilterProc(@MyFilter), itemHit);
{$ENDC }
			case itemHit of
				parallel, Central:  (* a click in Radiobuttons [parallel,central] *)
					ChangeProjection(itemHit);
				hiddenLine: 
					ChangeHiddenAttribute;
(*UseObjectScetch: *)
(*ChangeScetchAttribute; *)
				AutoApply: 
					ChangeAutoApply;
				Ex, Ey, Ez, Vx, Vy, Vz: 
					CalcProjectionDist;
				dist: (* user changed distance. have to recalc viewpoint *)
					CalcViewPoint;
				ZoomOut: 
					HandleZoomOut;
				ZoomIn: 
					HandleZoomIn;
				SetCloser: 
					HandleCloser;
				SetFurther: 
					HandleFurther;
				xypanel: 
					HandleXYClick;
				xzpanel: 
					HandleXZClick;
				ResetPanel: 
					HandleResetPanel;
				otherwise (* do nothing *)
					begin
				end;
			end; (*case *)
			autoApp := statusPtr^.autoApplyOn & (itemHit <> ZoomIn) & (itemHit <> ZoomOut) & (itemHit <> ResetPanel);
			autoApp := autoApp & (itemHit <> okButton) & (itemHit <> cancel) & (itemHit <> apply);
		until (itemHit = okButton) | (itemHit = cancel) | (itemHit = apply) | autoApp;
		if autoApp then
			itemHit := apply;
		case itemHit of
			okButton:  begin
				theEye := statusPtr^.Eye;
				SetEye := 0;
				wPeek^.refCon := longint(nil);
				dispose(statusPtr);
			end;
			cancel:  begin
				wPeek^.refCon := longint(nil);
				dispose(statusPtr);
				SetEye := -1;
			end;
			apply:  begin
				theEye := statusPtr^.Eye;
				wPeek^.refCon := longint(statusPtr);
				SetEye := 1;
			end;
		end; (* case *)
	end;

end. (* implementation *)
