(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit firstIntersect;

{$SETC ALineSegmentIntersection = TRUE }

interface

uses
	lineSegmentVector;

implementation

function lineSegmentVector.simpleFirstIntersect: obj;
	var
		i, j: seqIndex;
		l, r: lineSegment;
	begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			interactiveHighlight;
{$ENDC }
		l := lineSegment(elements^^[1].duplicate);
		for i := 1 to length do begin
			r := lineSegment(elements^^[i]);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then
				r.interactiveHighlight;
{$ENDC }
			for j := i + 1 to length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					l.free;
					simpleFirstIntersect := nil;
					exit(simpleFirstIntersect);
				end;
{$ENDC }
{$IFC AnimationEnabled AND ALineSegmentIntersection }
				if animationFlag[ALineSegmentIntersectionItem] then begin
					elements^^[j].interactiveHighlight;
					waitForClick(animationFlag[ALineSegmentIntersectionItem]);
				end;
{$ENDC }
				if lineSegment(elements^^[j]).intersectLineSegmentX(r, l) then
					with l do
						if p.equalPoint(q) then begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
							if animationFlag[ALineSegmentIntersectionItem] then begin
								p.interactiveHighlight;
								waitForClick(animationFlag[ALineSegmentIntersectionItem]);
							end;
{$ENDC }
							simpleFirstIntersect := p.duplicate;
							l.free;
							exit(simpleFirstIntersect);
						end
						else begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
							if animationFlag[ALineSegmentIntersectionItem] then begin
								l.interactiveHighlight;
								waitForClick(animationFlag[ALineSegmentIntersectionItem]);
							end;
{$ENDC }
							simpleFirstIntersect := l;
							exit(simpleFirstIntersect);
						end;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
				if animationFlag[ALineSegmentIntersectionItem] then
					elements^^[j].interactiveHighlight;
{$ENDC }
			end;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then
				r.interactiveHighlight;
{$ENDC }
		end;
		l.free;
		simpleFirstIntersect := nil;
	end;


	type
		psfEventType = (startEvent, endEvent);

		psfEvent = object(obj)
				where: point2d;
				procedure processEvent (yTable: dictionary; var intersection: lineSegment);
				function getEventType: psfEventType;
			end;

		psfStartEvent = object(psfEvent)
				l: lineSegment;
				e: psfEndEvent;
				procedure processEvent (yTable: dictionary; var intersection: lineSegment);
				override;
				function getEventType: psfEventType;
				override;
			end;

		psfEndEvent = object(psfEvent)
				toKill: reference; (* into y-table *)
				procedure processEvent (yTable: dictionary; var intersection: lineSegment);
				override;
				function getEventType: psfEventType;
				override;
			end;

function psfEvent.getEventType: psfEventType;
	begin
		unimplementedMethod('psfEvent.getEventType');
		getEventType := startEvent;
	end;

function psfStartEvent.getEventType: psfEventType;
	begin
		getEventType := startEvent;
	end;

function psfEndEvent.getEventType: psfEventType;
	begin
		getEventType := endEvent;
	end;

function compareLineSegment (a, b: lineSegment): relationObj;
	begin
		case sign(b.whichSideX(a.p)) of
			-1: 
				compareLineSegment := greaterThanObj;
			0: 
				compareLineSegment := relationObj(1 + sign(b.whichSideX(a.q)));
			+1: 
				compareLineSegment := lessThanObj;
		end;
	end;

function lessThanPsfEvent (a, b: psfEvent): boolean;
	begin
		lessThanPsfEvent := pointLessThanXY(a.where, b.where) | a.where.equal(b.where) & (a.getEventType < b.getEventType);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure psfEvent.processEvent (yTable: dictionary; var intersection: lineSegment);
	begin
		unimplementedMethod('psfEvent.processEvent');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure testIntersection (d: dictionary; a, b: reference; var intersection: lineSegment);
	var
		s, r, x: lineSegment;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		l: list;
{$ENDC }
	begin
		if (a <> nil) & (b <> nil) then begin
			s := lineSegment(d.getObject(a));
			r := lineSegment(d.getObject(b));
			x := lineSegment(s.duplicate);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then begin
				l := newEmptyList;
				l.append(s);
				l.append(r);
				l.interactiveFlash;
				waitForClick(animationFlag[ALineSegmentIntersectionItem]);
				l.free;
			end;
{$ENDC }
			if s.intersectLineSegmentX(r, x) then begin
				intersection := x;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
				if animationFlag[ALineSegmentIntersectionItem] then begin
					intersection.interactiveHighlight;
					waitForClick(animationFlag[ALineSegmentIntersectionItem]);
				end;
{$ENDC }
			end
			else
				x.free;
		end;
	end;

procedure psfStartEvent.processEvent (yTable: dictionary; var intersection: lineSegment);
	var
		whereRef: reference;
		direction: directionType;
	begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then begin
			drawThickVertical(where.getX);
			l.interactiveFlash;
			l.interactiveHighlight;
			waitForClick(animationFlag[ALineSegmentIntersectionItem]);
			l.interactiveHighlight;
		end;
{$ENDC }
		if yTable.find(l, compareLineSegment, whereRef, direction) then begin
			intersection := lineSegment(l.duplicate);
			assert(l.intersectLineSegmentX(lineSegment(yTable.getObject(whereRef)), intersection), 'l intersect yTable.l');
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then begin
				intersection.interactiveHighlight;
				waitForClick(animationFlag[ALineSegmentIntersectionItem]);
			end;
{$ENDC }
		end
		else begin
			whereRef := yTable.insert(l, whereRef, direction);
			e.toKill := whereRef;
			testIntersection(yTable, whereRef, yTable.next(whereRef, left), intersection);
			if intersection = nil then
				testIntersection(yTable, whereRef, yTable.next(whereRef, right), intersection);
		end;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			drawThickVertical(where.getX);
{$ENDC }
	end;

procedure psfEndEvent.processEvent (yTable: dictionary; var intersection: lineSegment);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
	var
		l: lineSegment;
{$ENDC }
	begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then begin
			drawThickVertical(where.getX);
			l := lineSegment(yTable.getObject(toKill));
			l.interactiveFlash;
			l.interactiveHighlight;
			waitForClick(animationFlag[ALineSegmentIntersectionItem]);
			l.interactiveHighlight;
			l.interactiveOutput;
		end;
{$ENDC }
		testIntersection(yTable, yTable.next(toKill, left), yTable.next(toKill, right), intersection);
		yTable.delete(toKill);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			drawThickVertical(where.getX);
{$ENDC }
	end;

function lineSegmentVector.planeSweepFirstIntersect (ask: boolean): obj;
	var
		yTable: dictionary;
		dummy: priorityQueue;
		xQueue: vector;
		intersection: lineSegment;
		i: seqIndex1;

	function buildXQueue (lv: lineSegmentVector): vector;
		var
			xQueue: vector;
			i: seqIndex0;
			psfStartH: psfStartEvent;
			psfEndH: psfEndEvent;
			l: lineSegment;
		begin
			xQueue := newVector(2 * length);
			for i := 1 to lv.length do begin
				l := lineSegment(elements^^[i]);
				orderPointsGreaterThanXY(l.p, l.q);
				new(psfEndH);
				failNil(Handle(psfEndH));
				psfEndH.where := l.q;
				psfEndH.toKill := nil; (* will be set by processing the startEvent *)
				xQueue.append(psfEndH);
				new(psfStartH);
				failNil(Handle(psfStartH));
				psfStartH.where := l.p;
				psfStartH.l := l;
				psfStartH.e := psfEndH;
				xQueue.append(psfStartH);
			end;
			xQueue.sortAll(lessThanPsfEvent);
			buildXQueue := xQueue;
		end;

	begin
		if ask then begin
			if not getXY(0, false, dummy, length, yTable, false) then begin
				planeSweepFirstIntersect := nil;
				exit(planeSweepFirstIntersect);
			end;
		end
		else begin
			currentVectorLength := length;
			yTable := dictionary(newInitObj(avlTreeType));
		end;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			interactiveHighlight;
{$ENDC }
		xQueue := buildXQueue(self);
		intersection := nil;
		i := 1;
		while (i <= xQueue.length) & (intersection = nil) do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				xQueue.freeAll;
				yTable.free;
				planeSweepFirstIntersect := nil;
				exit(planeSweepFirstIntersect);
			end;
{$ENDC }
			psfEvent(xQueue.elements^^[i]).processEvent(yTable, intersection);
			i := i + 1;
		end;
		xQueue.freeAll;
		if yTable.infoString <> '' then
			displayInfoString(concat('Y', yTable.infoString));
		yTable.free;
		if (intersection = nil) | (not intersection.p.equalPoint(intersection.q)) then
			planeSweepFirstIntersect := intersection
		else begin
			planeSweepFirstIntersect := intersection.p.duplicate;
			intersection.free;
		end;
	end;


end.
