(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit floatPoint;

interface

uses
	basicGeometry;

implementation

function floatPoint.verticalDistance (o: obj): extended;
	begin
		verticalDistance := floatToReal(absF(subF(y, floatPoint(o).y)));
	end;

function floatPoint.distanceO (o: obj; var minP, minQ: obj): extended;
	begin
		floatPoint(minP).x := x;
		floatPoint(minP).y := y;
		floatPoint(minQ).x := floatPoint(o).x;
		floatPoint(minQ).y := floatPoint(o).y;
		distanceO := distanceX(point2d(o));
	end;

function floatPoint.getX: extended;
	begin
		getX := floatToReal(x);
	end;

procedure floatPoint.setX (newX: extended);
	begin
		realToFloat(newX, currentBase, currentPrecision, x);
	end;

function floatPoint.getY: extended;
	begin
		getY := floatToReal(y);
	end;

procedure floatPoint.setY (newY: extended);
	begin
		realToFloat(newY, currentBase, currentPrecision, y);
	end;

function floatPoint.xLessThan (p: point2d): boolean;
	begin
		xLessThan := compareFloat(x, floatPoint(p).x) = lessFloat;
	end;

function floatPoint.xEqual (p: point2d): boolean;
	begin
		xEqual := compareFloat(x, floatPoint(p).x) = equalFloat;
	end;

function floatPoint.yLessThan (p: point2d): boolean;
	begin
		yLessThan := compareFloat(y, floatPoint(p).y) = lessFloat;
	end;

function floatPoint.yEqual (p: point2d): boolean;
	begin
		yEqual := compareFloat(y, floatPoint(p).y) = equalFloat;
	end;

procedure floatPoint.textFileInput;
	begin
		readTextFloat(x);
		readTextFloat(y);
	end;

procedure floatPoint.textFileOutput;
	begin
		writeTextFloat(x);
		writeTextFloat(y);
	end;

procedure floatPoint.fileInput;
	begin
		readFloat(x);
		readFloat(y);
	end;

procedure floatPoint.fileOutput;
	begin
		writeFloat(x);
		writeFloat(y);
	end;

function floatPoint.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, floatPoint) & (compareFloat(floatPoint(o).x, x) = equalFloat) & (compareFloat(floatPoint(o).y, y) = equalFloat);
	end;

function floatPoint.equalPoint (p: point2d): boolean;
	begin
		equalPoint := (p <> nil) & member(p, floatPoint) & (compareFloat(floatPoint(p).x, x) = equalFloat) & (compareFloat(floatPoint(p).y, y) = equalFloat);
	end;

function floatPoint.memberType (o: obj): boolean;
	begin
		memberType := member(o, floatPoint) & (floatPoint(o).x.base = x.base) & (floatPoint(o).x.precision = x.precision);
	end;

function ws (p, q, r: floatPoint): float;
	begin
		ws := subF(mulF(subF(p.x, q.x), subF(q.y, r.y)), mulF(subF(p.y, q.y), subF(q.x, r.x)));
	end;

function whichSideFloat (p, q, r: floatPoint): float;
	var
		result: float;

	function whichSideFloatPQR (p, q, r: floatPoint): float;
		var
			dpqx, dpqy, dqrx, dqry, eps, errorFloat, result: float;
		begin
			subtractFloat(p.x, q.x, dpqx);
			subtractFloat(p.y, q.y, dpqy);
			subtractFloat(q.x, r.x, dqrx);
			subtractFloat(q.y, r.y, dqry);
			subtractFloat(mulF(dpqx, dqry), mulF(dpqy, dqrx), result);
			machineEpsilon(result.base, result.precision, eps);
			intToFloat(1, result.base, result.precision, errorFloat);
			if compareFloat(absF(result), mulF(mulF(eps, errorFloat), mulF(addF(absF(dpqx), absF(dpqy)), addF(absF(dqrx), absF(dqry))))) = lessFloat then
				intToFloat(0, result.base, result.precision, result);
			whichSideFloatPQR := result;
		end;

	begin
		result := whichSideFloatPQR(p, q, r);
		if result.sign = 0 then begin
			result := whichSideFloatPQR(q, p, r);
			if result.sign = 0 then begin
				result := whichSideFloatPQR(p, r, q);
			end;
			result.sign := -result.sign;
		end;
		whichSideFloat := result;
	end;

function floatPoint.whichSideX (p, q: point2d): extended;
	begin
		whichSideX := floatToReal(ws(floatPoint(p), self, floatPoint(q)));
	end;

function floatPoint.crossProductX (p, q, r: point2d): extended;
	var
		pr, qr, rr: floatPoint;
	begin
		pr := floatPoint(p);
		qr := floatPoint(q);
		rr := floatPoint(r);
		crossProductX := floatToReal(subF(mulF(subF(pr.x, x), subF(rr.y, qr.y)), mulF(subF(pr.y, y), subF(rr.x, qr.x))));
	end;

function floatPoint.scalarProductX (p: point2d): extended;  (* added by FB *)
	var
		pr: floatPoint;
	begin
		pr := floatPoint(p);
		scalarProductX := floatToReal(addF(mulF(pr.x, x), mulF(pr.y, y)));
	end;

function floatPoint.distanceX (p: point2d): extended;
	var
		pr: floatPoint;
	begin
		pr := floatPoint(p);
		distanceX := sqrt(floatToReal(addF(sqrF(subF(pr.x, x)), sqrF(subF(pr.y, y)))));
	end;

function floatPoint.squaredDxX (p: point2d): extended;
	begin
		squaredDxX := floatToReal(sqrF(subF(floatPoint(p).x, x)));
	end;

function floatPoint.squaredDyX (p: point2d): extended;
	begin
		squaredDyX := floatToReal(sqrF(subF(floatPoint(p).y, y)));
	end;

function floatPoint.squaredDistanceX (p: point2d): extended;
	var
		pr: floatPoint;
	begin
		pr := floatPoint(p);
		squaredDistanceX := floatToReal(addF(sqrF(subF(pr.x, x)), sqrF(subF(pr.y, y))));
	end;

function floatIntegerPower (x: float; p: integer): float;
	var
		t: float;
	begin
		case p of
			0:  begin
				intToFloat(1, x.base, x.precision, t);
				floatIntegerPower := t;
			end;
			1: 
				floatIntegerPower := x;
			otherwise begin
				t := sqrF(floatIntegerPower(x, p div 2));
				if p mod 2 = 0 then
					floatIntegerPower := t
				else
					floatIntegerPower := mulF(x, t);
			end;
		end;
	end;

function floatPthRoot (x: float; p: extended): float;
	begin
		if x.sign = +1 then
			floatPthRoot := realToFloatF(exp(ln(floatToReal(x)) / p), x.base, x.precision)
		else if x.sign = 0 then
			floatPthRoot := intToFloatF(0, x.base, x.precision)
		else begin
			assert(false, 'x>=0');
			floatPthRoot := intToFloatF(0, x.base, x.precision);
		end;
	end;

function floatPower (x: float; p: extended): float;
	begin
		if x.sign = +1 then
			floatPower := realToFloatF(exp(p * ln(floatToReal(x))), x.base, x.precision)
		else if x.sign = 0 then
			floatPower := intToFloatF(0, x.base, x.precision)
		else begin
			assert(false, 'x>=0');
			floatPower := intToFloatF(0, x.base, x.precision);
		end;
	end;

function floatPoint.distanceLpX (p: point2d): extended;
	begin
		if lpMetric = 2 then
			distanceLpX := distanceX(p)
		else if lpMetric = 1 then
			distanceLpX := floatToReal(addF(absF(subF(floatPoint(p).x, x)), absF(subF(floatPoint(p).y, y))))
		else if lpMetric = lpInfinity then
			distanceLpX := distancePLpX(p)
		else if trunc(lpMetric) = lpMetric then
{$IFC THINK_PASCAL }
			distanceLpX := floatToReal(floatPthRoot(addF(floatIntegerPower(absF(subF(x, floatPoint(p).x)), trunc(lpMetric)), floatIntegerPower(absF(subF(y, floatPoint(p).y)), trunc(lpMetric))), lpMetric))
{$ELSEC }
			distanceLpX := floatToReal(floatPthRoot(addF(floatIntegerPower(absF(subF(x, floatPoint(p).x)), truncLong(lpMetric)), floatIntegerPower(absF(subF(y, floatPoint(p).y)), truncLong(lpMetric))), lpMetric))
{$ENDC }
		else
			distanceLpX := floatToReal(floatPthRoot(addF(floatPower(absF(subF(x, floatPoint(p).x)), lpMetric), floatPower(absF(subF(y, floatPoint(p).y)), lpMetric)), lpMetric));
	end;

function floatPoint.distancePLpX (p: point2d): extended;
	begin
		if trunc(lpMetric) = lpMetric then
			if lpMetric = lpInfinity then
				distancePLpX := floatToReal(maxF(absF(subF(floatPoint(p).x, x)), absF(subF(floatPoint(p).y, y))))
			else
{$IFC THINK_PASCAL }
				distancePLpX := floatToReal(addF(floatIntegerPower(absF(subF(x, floatPoint(p).x)), trunc(lpMetric)), floatIntegerPower(absF(subF(y, floatPoint(p).y)), trunc(lpMetric))))
{$ELSEC }
				distancePLpX := floatToReal(addF(floatIntegerPower(absF(subF(x, floatPoint(p).x)), truncLong(lpMetric)), floatIntegerPower(absF(subF(y, floatPoint(p).y)), truncLong(lpMetric))))
{$ENDC }
		else
			distancePLpX := floatToReal(addF(floatPower(absF(subF(x, floatPoint(p).x)), lpMetric), floatPower(absF(subF(y, floatPoint(p).y)), lpMetric)));
	end;

function floatPoint.dxPLpX (p: point2d): extended;
	begin
		if trunc(lpMetric) = lpMetric then
			if lpMetric = lpInfinity then
				dxPLpX := floatToReal(absF(subF(floatPoint(p).x, x)))
			else
{$IFC THINK_PASCAL }
				dxPLpX := floatToReal(floatIntegerPower(absF(subF(x, floatPoint(p).x)), trunc(lpMetric)))
{$ELSEC }
				dxPLpX := floatToReal(floatIntegerPower(absF(subF(x, floatPoint(p).x)), truncLong(lpMetric)))
{$ENDC }
		else
			dxPLpX := floatToReal(floatPower(absF(subF(x, floatPoint(p).x)), lpMetric))
	end;

function floatPoint.dyPLpX (p: point2d): extended;
	begin
		if trunc(lpMetric) = lpMetric then
			if lpMetric = lpInfinity then
				dyPLpX := floatToReal(absF(subF(floatPoint(p).y, y)))
			else
{$IFC THINK_PASCAL }
				dyPLpX := floatToReal(floatIntegerPower(absF(subF(y, floatPoint(p).y)), trunc(lpMetric)))
{$ELSEC }
				dyPLpX := floatToReal(floatIntegerPower(absF(subF(y, floatPoint(p).y)), truncLong(lpMetric)))
{$ENDC }
		else
			dyPLpX := floatToReal(floatPower(absF(subF(y, floatPoint(p).y)), lpMetric))
	end;

procedure floatPoint.circleCenterX (p, q: point2d; var cx, cy: extended);
	var
		d, s1, s2, s3, const2: float;
		pr, qr: floatPoint;
	begin
		pr := floatPoint(p);
		qr := floatPoint(q);
		intToFloat(2, pr.x.base, pr.x.precision, const2);
		multiplyFloat(const2, subF(mulF(subF(pr.x, x), subF(pr.y, qr.y)), mulF(subF(pr.x, qr.x), subF(pr.y, y))), d);
		addFloat(sqrF(pr.x), sqrF(pr.y), s1);
		addFloat(sqrF(x), sqrF(y), s2);
		addFloat(sqrF(qr.x), sqrF(qr.y), s3);
		cx := floatToReal(divF(addF(mulF(s1, subF(y, qr.y)), addF(mulF(s2, subF(qr.y, pr.y)), mulF(s3, subF(pr.y, y)))), d));
		cy := floatToReal(divF(addF(mulF(s1, subF(qr.x, x)), addF(mulF(s2, subF(pr.x, qr.x)), mulF(s3, subF(x, pr.x)))), d));
	end;

function floatPoint.intersectLineSegmentX (q, r, s, left, right: point2d): boolean;

	function intersectLineSegment (r0, r1, s0, s1: floatPoint; left, right: floatPoint): boolean;
		var
			rs0, rs1, sr0, sr1, alpha: float;

		function pointOnSegment (p, r0, r1: floatPoint): boolean;
			function range (a, m, b: float): boolean;
				begin
					range := (compareFloat(a, m) <= equalFloat) & (compareFloat(m, b) <= equalFloat);
				end;

			begin
				pointOnSegment := range(r0.x, p.x, r1.x) & (range(r0.y, p.y, r1.y) | range(r1.y, p.y, r0.y));
			end;

		function oppositeSign (a, b: signType): boolean;
			begin
				oppositeSign := (a = +1) & (b = -1) | (a = -1) & (b = +1);
			end;

		begin
			intersectLineSegment := true;
			orderPointsGreaterThanXY(point2d(r0), point2d(r1));
			orderPointsGreaterThanXY(point2d(s0), point2d(s1));
			rs0 := whichSideFloat(r0, s0, r1);
			rs1 := whichSideFloat(r0, s1, r1);
			sr0 := whichSideFloat(s0, r0, s1);
			sr1 := whichSideFloat(s0, r1, s1);
			if oppositeSign(rs0.sign, rs1.sign) & oppositeSign(sr0.sign, sr1.sign) then begin
				divideFloat(sr0, subF(sr0, sr1), alpha);
				addFloat(r0.x, mulF(alpha, subF(r1.x, r0.x)), left.x);
				addFloat(r0.y, mulF(alpha, subF(r1.y, r0.y)), left.y);
				orderPointsGreaterThanXY(point2d(r0), point2d(s0));
				orderPointsGreaterThanXY(point2d(r1), point2d(s1));
				if pointLessThanXY(left, s0) then begin
					left.x := s0.x;
					left.y := s0.y;
				end
				else if pointGreaterThanXY(left, r1) then begin
					left.x := r1.x;
					left.y := r1.y;
				end;
				right.x := left.x;
				right.y := left.y;
			end
			else if (rs0.sign = 0) & pointOnSegment(s0, r0, r1) then begin
				left.x := s0.x;
				left.y := s0.y;
				if rs1.sign = 0 then begin
					orderPointsGreaterThanXY(point2d(s1), point2d(r1));
					right.x := s1.x;
					right.y := s1.y;
				end
				else begin
					right.x := s0.x;
					right.y := s0.y;
				end;
			end
			else if (rs1.sign = 0) & pointOnSegment(s1, r0, r1) then begin
				left.x := s1.x;
				left.y := s1.y;
				right.x := s1.x;
				right.y := s1.y;
			end
			else if (sr0.sign = 0) & pointOnSegment(r0, s0, s1) then begin
				left.x := r0.x;
				left.y := r0.y;
				if sr1.sign = 0 then begin
					orderPointsGreaterThanXY(point2d(s1), point2d(r1));
					right.x := s1.x;
					right.y := s1.y;
				end
				else begin
					right.x := r0.x;
					right.y := r0.y;
				end;
			end
			else if (sr1.sign = 0) & pointOnSegment(r1, s0, s1) then begin
				left.x := r1.x;
				left.y := r1.y;
				right.x := r1.x;
				right.y := r1.y;
			end
			else
				intersectLineSegment := false;
		end;

	begin
		intersectLineSegmentX := intersectLineSegment(self, floatPoint(q), floatPoint(r), floatPoint(s), floatPoint(left), floatPoint(right));
	end;

function intersectLine (p, q, r, s: floatPoint; left, right: floatPoint): boolean;
	var
		prq, psq, d: float;

	begin
		intersectLine := true;
		orderPointsGreaterThanXY(point2d(r), point2d(s));
		prq := ws(p, r, q);
		psq := ws(p, s, q);
		if compareFloat(prq, psq) = equalFloat then
			if psq.sign = 0 then begin
				left.x := r.x;
				left.y := r.y;
				right.x := q.x;
				right.y := q.y;
			end
			else
				intersectLine := false
		else if (prq.sign >= 0) & (psq.sign <= 0) | (prq.sign <= 0) & (psq.sign >= 0) then begin
			d := subF(prq, psq);
			left.x := divF(subF(mulF(prq, s.x), mulF(psq, r.x)), d);
			left.y := divF(subF(mulF(prq, s.y), mulF(psq, r.y)), d);
			right.x := left.x;
			right.y := left.y;
		end
		else
			intersectLine := false;
	end;

function floatPoint.intersectLineX (q, r, s, left, right: point2d): boolean;
	begin
		intersectLineX := intersectLine(self, floatPoint(q), floatPoint(r), floatPoint(s), floatPoint(left), floatPoint(right));
	end;

function intersectLineLine (p, q, r, s, intersection: floatPoint): boolean;
	var
		prq, psq, d: float;
	begin
		prq := ws(p, r, q);
		psq := ws(p, s, q);
		if compareFloat(prq, psq) = equalFloat then
			if psq.sign = 0 then begin
				intersection.x := p.x;
				intersection.y := p.y;
				intersectLineLine := true;
			end
			else
				intersectLineLine := false
		else begin
			d := subF(prq, psq);
			intersection.x := divF(subF(mulF(prq, s.x), mulF(psq, r.x)), d);
			intersection.y := divF(subF(mulF(prq, s.y), mulF(psq, r.y)), d);
			intersectLineLine := true;
		end;
	end;

function floatPoint.lineIntersectsLine (q, r, s, intersectionPoint: point2d): boolean;
	begin
		lineIntersectsLine := intersectLineLine(self, floatPoint(q), floatPoint(r), floatPoint(s), floatPoint(intersectionPoint));
	end;

procedure doProjectOnLine (p, q, r: floatPoint; var rProjectX, rProjectY: float; var distance, whichSide: extended; var lambdaSign: signType);
	var
		dxqp, dyqp, dxrp, dyrp, rv, ws, d: float;
	begin
		dxqp := subF(q.x, p.x);
		dyqp := subF(q.y, p.y);
		if (dxqp.sign = 0) & (dyqp.sign = 0) then begin
			rProjectX := p.x;
			rProjectY := p.y;
			distance := sqrt(floatToReal(addF(sqrF(subF(p.x, r.x)), sqrF(subF(p.y, r.y)))));
			whichSide := 0;
			lambdaSign := 0;
		end
		else begin
			dxrp := subF(r.x, p.x);
			dyrp := subF(r.y, p.y);
			rv := addF(mulF(dxqp, dxrp), mulF(dyqp, dyrp));
			ws := subF(mulF(dxrp, subF(q.y, r.y)), mulF(dyrp, subF(q.x, r.x)));
			d := addF(sqrF(dxqp), sqrF(dyqp));
			rProjectX := addF(p.x, divF(mulF(dxqp, rv), d));
			rProjectY := addF(p.y, divF(mulF(dyqp, rv), d));
			whichSide := floatToReal(ws);
			distance := abs(whichSide / sqrt(floatToReal(d)));
			if rv.sign < 0 then
				lambdaSign := -1
			else if compareFloat(rv, d) = greaterFloat then
				lambdaSign := +1
			else
				lambdaSign := 0;
		end;
	end;

procedure floatPoint.projectOnLine (endPoint, candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
	var
		px, py: float;
	begin
		if projectionPoint = nil then
			doProjectOnLine(self, floatPoint(endPoint), floatPoint(candidatePoint), px, py, distanceToLine, whichSide, lambdaSign)
		else
			doProjectOnLine(self, floatPoint(endPoint), floatPoint(candidatePoint), floatPoint(projectionPoint).x, floatPoint(projectionPoint).y, distanceToLine, whichSide, lambdaSign);
	end;

function floatPoint.distanceToLineSegmentX (p, q: point2d; var closestX, closestY: extended): extended;
	var
		rpx, rpy: float;
		d, w: extended;
		l: signType;
	begin
		doProjectOnLine(floatPoint(p), floatPoint(q), self, rpx, rpy, d, w, l);
		if l < 0 then begin
			closestX := p.getX;
			closestY := p.getY;
			distanceToLineSegmentX := distanceX(p);
		end
		else if l > 0 then begin
			closestX := q.getX;
			closestY := q.getY;
			distanceToLineSegmentX := distanceX(q);
		end
		else begin
			closestX := floatToReal(rpx);
			closestY := floatToReal(rpy);
			distanceToLineSegmentX := d;
		end;
	end;

function floatPoint.whichSideSign (p, q: point2d): signType;
	begin
		whichSideSign := whichSideFloat(floatPoint(p), self, floatPoint(q)).sign;
	end;

function floatPoint.xPlusYX: extended;
	begin
		xPlusYX := floatToReal(addF(x, y));
	end;

function floatPoint.xMinusYX: extended;
	begin
		xMinusYX := floatToReal(subF(x, y));
	end;

function floatPoint.getType: objectType;
	begin
		getType := floatPointType;
	end;

function floatPoint.getTextType: Str255;
	begin
		getTextType := 'Point(float)';
	end;

function floatPoint.middle (p: point2d): point2d;
	var
		t, pr: floatPoint;
		const2: float;
	begin
		pr := floatPoint(p);
		new(t);
		failNil(Handle(t));
		t.init;
		intToFloat(2, pr.x.base, pr.x.precision, const2);
		divideFloat(addF(x, pr.x), const2, t.x);
		divideFloat(addF(y, pr.y), const2, t.y);
		middle := t;
	end;

(* ------------------------- inserted by JM -------------------------------------------- *)

function floatPoint.DxX (p: point2d): extended;
	begin
		DxX := floatToReal(subF(floatPoint(p).x, x));
	end;

function floatPoint.DyX (p: point2d): extended;
	begin
		DyX := floatToReal(subF(floatPoint(p).y, y));
	end;

procedure floatPoint.setPoint (newp: point2d);
	begin
		x := floatPoint(newp).x;
		y := floatPoint(newp).y;
	end;

(* ------------------------------------------------------------------------------------------------ *)

end.
