(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit geoBenchEditMenu;

interface

uses
{$IFC ScriptEnabled }
	GeoScript, 
{$ENDC }
	geoBenchFileMenu, boundingBox, eyeInterface;

procedure DoEditMenu (menu, item: integer; when: longInt);

procedure DoTransMenu (menu, item: integer; when: longInt);

implementation

const
	draggedItem = 100;

procedure fixList (listToBeFixed: list; otherList: list; highlight: boolean; window: WindowPtr);
	begin
		if (listToBeFixed.length > 1) | (otherList.length > 0) then
			with undoPosition^^.value do begin
				interactiveOutput;
				if highlight then
					interactiveHighlight;
			end
		else
			clearWindow(window);
		undoPosition^^.value.freeAll;
		undoPosition^^.value := undoScrap.first^^.value;
		undoScrap.interactiveOutput;
		if highlight then
			undoScrap.interactiveHighlight;
		undoScrap.free;
	end;

function checkB (b: extended): boolean;
	begin
		checkB := (b <> 0) | (defaultA <> 0);
	end;

procedure moveToWindowPoint (s: sequence; p: Point);
	var
		cx, cy, px, py: extended;
	begin
		s.centerOfGravity(cx, cy);
		windToWorld(p.h, p.v, px, py);
		s.translate(px - cx, py - cy);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoEditMenu (menu, item: integer; when: longInt);
	var
		w: windowIndex0;
		fw: WindowPtr;
		consumedByDA: boolean;

	procedure zoomIn (w: windowIndex);
		var
			mouse: Point;
			r: rectangle;
			re: Rect;
			zoom: boolean;
		begin
			new(r);
			failNil(Handle(r));
			r.init;
			repeat
				writeCursorCoordinates(false);
			until Button;
			GetMouse(mouse);
			with wStatus^[w], frames do begin
				constrainedRectangle := true;
				with drawingRect(window) do
					widthByHeight := (right - left) / (bottom - top);
				r.windowInputMouse(mouse);
				constrainedRectangle := false;
				zoom := (r.left < r.right) & (r.bottom < r.top);
				if zoom then begin
					re := drawingRect(window);
					stackPointer := stackPointer + 1;
					createFrame(stack[stackPointer - 1], re.right - re.left, re.bottom - re.top, r.left, r.bottom, r.right, r.top, stack[stackPointer]);
					currentFrame := stack[stackPointer];
				end;
				zoomed := false;
			end;
			FlushEvents(mUpMask + mDownMask, 0);
			r.free;
			if zoom then
				redraw(w);
		end;

	procedure zoomOut (w: windowIndex);
		begin
			with wStatus^[w], frames do
				repeat
					deleteframe(stack[stackPointer]);
					stackPointer := stackPointer - 1;
					currentFrame := stack[stackPointer];
					with currentFrame.proD^ do begin
						eyeProHasChanged := true;
						lightHasChanged := true;
						visHasChanged := true;
						proHasChanged := true;
					end;
					zoomed := false;
				until not optionKeyPressed | (stackPointer = 1);
			redraw(w);
		end;

	procedure newScrap (window: WindowPtr);
		begin
			stopwatchFreeAll(obj(scrapBook));
			deleteframe(scrapFrame);
			copyFrame(currentFrame, scrapFrame);
			scrapRect := window^.portRect;
		end;

	procedure getValue (d: DialogPtr; item: integer; default: extended; var value: extended);
		var
			number: Str255;
			v: extended;
		begin
			GetDText(d, item, number);
			IOCheck(false);
			readString(number, v);
			if ioResult > 0 then begin
				SysBeep(1);
				value := default;
			end
			else
				value := v;
			IOCheck(true);
		end;

	procedure getProjSettings (var currentFrame: wFrame; win: windowIndex0);
		const
	(*object sketch; here a resource *)
	(*sidePictID = 400; *)
	(*topPictID = 401; *)
			DialogID = 9000;
		var
			oldEyePoint, oldProPoint: point3dT;
			timer: longInt;
			proDOld: projDescPtr;
			oldEye, theEye: EyeDesc;
			myCube: Cube;
			res: integer;
			Dlg: DialogPtr;
			savePort: GrafPtr;
			box: obj;
			change: boolean;
			l: list;
			mem1, mem2: longint;
		begin
			mem1 := FreeMem;
			l := newEmptyList;
			l.append(wStatus^[w].desktop);
			l.append(wStatus^[w].selection);
			box := l.getBoundingObject;
			l.free;
			box.freeAll;
			box := nil;
			mem2 := FreeMem;
			if mem1 > mem2 then
				SysBeep(2);
			if box <> nil then begin
	   (* check whether the projection settings can be displayed in the eyeinterface *)
				if box.getType = boundingBox2dType then
					with boundingBox2d(box) do begin
						if (abs(left) > 30000) | (abs(right) > 30000) | (abs(top) > 30000) | (abs(bottom) > 30000) then begin
							nonFatalError('Sorry, but your object is too big');
							box.freeAll;
							exit(getProjSettings);
						end;
					end
				else begin
					with boundingBox3d(box) do
						if (abs(xMin) > 30000) | (abs(xMax) > 30000) | (abs(yMin) > 30000) | (abs(yMax) > 30000) | (abs(zMin) > 30000) | (abs(zMax) > 30000) then begin
							nonFatalError('Sorry, but your object is too big');
							box.freeAll;
							exit(getProjSettings);
						end;
				end;
			end;
			new(proDOld);
			failNil(Handle(proDOld));
			proDOld^ := currentFrame.proD^;
			timer := TickCount;
			beginCursor(arrowCursor);
			Dlg := GetNewDialog(DialogID, nil, WindowPtr(-1));
			niceWindowPosition(WindowPtr(Dlg), false);
			if sPDialogPos.v > 0 then
				MoveWindow(Dlg, sPDialogPos.h, sPDialogPos.v, false);
			SelectWindow(Dlg);
			ShowWindow(Dlg);
			GetPort(savePort);
			SetPort(Dlg);
			with currentFrame.proD^ do begin
{$IFC THINK_PASCAL }
				theEye.x := round(eyePoint[1]);
				theEye.y := round(eyePoint[2]);
				theEye.z := round(eyePoint[3]);
				theEye.vx := round(proPoint[1]);
				theEye.vy := round(proPoint[2]);
				theEye.vz := round(proPoint[3]);
{$ELSEC }
				theEye.x := roundLong(eyePoint[1]);
				theEye.y := roundLong(eyePoint[2]);
				theEye.z := roundLong(eyePoint[3]);
				theEye.vx := roundLong(proPoint[1]);
				theEye.vy := roundLong(proPoint[2]);
				theEye.vz := roundLong(proPoint[3]);
{$ENDC }
				theEye.ProjectionType := 0;
				if projZ then
					theEye.ProjectionType := 1;
				theEye.hidden := visibility;
			end;
			if box = nil then begin
				myCube.xmin := 0;
				myCube.xmax := 100;
				myCube.ymin := 0;
				myCube.ymax := 100;
				myCube.zmin := 0;
				myCube.zmax := 100;
			end
			else begin
				if box.getType = boundingBox2dType then
					with boundingBox2d(box) do begin
{$IFC THINK_PASCAL }
						myCube.xmin := round(left);
						myCube.xmax := round(right);
						myCube.ymin := round(bottom);
						myCube.ymax := round(top);
{$ELSEC }
						myCube.xmin := roundLong(left);
						myCube.xmax := roundLong(right);
						myCube.ymin := roundLong(bottom);
						myCube.ymax := roundLong(top);
{$ENDC }
						myCube.zmin := 0;
						myCube.zmax := 100;
					end (* with *)
				else
					with boundingBox3d(box) do begin
{$IFC THINK_PASCAL }
						myCube.xmin := round(xMin);
						myCube.xmax := round(xMax);
						myCube.ymin := round(yMin);
						myCube.ymax := round(yMax);
						myCube.zmin := round(zMin);
						myCube.zmax := round(zMax);
{$ELSEC }
						myCube.xmin := roundLong(xMin);
						myCube.xmax := roundLong(xMax);
						myCube.ymin := roundLong(yMin);
						myCube.ymax := roundLong(yMax);
						myCube.zmin := roundLong(zMin);
						myCube.zmax := roundLong(zMax);
{$ENDC }
					end; (* with *)
				box.freeAll;
			end; (* else *)
			oldEye := theEye;
			res := SetEye(theEye, myCube, Dlg);
			while res = 1 do begin
				with currentFrame.proD^ do begin
					eyeProHasChanged := false;
					lightHasChanged := false;
					visHasChanged := false;
					proHasChanged := false;
					change := (theEye.x <> oldEye.x) | (theEye.y <> oldEye.y) | (theEye.z <> oldEye.z);
					change := change | (theEye.vx <> oldEye.vx) | (theEye.vy <> oldEye.vy) | (theEye.vz <> oldEye.vz);
					change := change | (oldEye.hidden <> theEye.hidden) | (oldEye.ProjectionType <> theEye.ProjectionType);
					if change then begin
						if theEye.ProjectionType <> oldEye.ProjectionType then begin
							projZ := not projZ;
							proHasChanged := true;
						end;
						if theEye.hidden <> oldEye.hidden then begin
							visibility := not visibility;
							visHasChanged := true;
						end;
						if (theEye.x <> oldEye.x) | (theEye.y <> oldEye.y) | (theEye.z <> oldEye.z) | (theEye.vx <> oldEye.vx) | (theEye.vy <> oldEye.vy) | (theEye.vz <> oldEye.vz) then begin
							oldProPoint := proPoint;
							proPoint[1] := theEye.vx;
							proPoint[2] := theEye.vy;
							proPoint[3] := theEye.vz;
							newProPoint(oldProPoint, proPoint, eyePoint, userTrafo, userTScal, projTrafo, viewTrafo);
							oldEyePoint := eyePoint;
							eyePoint[1] := theEye.x;
							eyePoint[2] := theEye.y;
							eyePoint[3] := theEye.z;
							newEyePoint(oldEyePoint, eyePoint, userTrafo, userTScal, projTrafo, viewTrafo);
							eyeProHasChanged := true;
						end;
					end;
					SetPort(savePort);
					redraw(win);
					SetPort(Dlg);
				end; (* with *)
				oldEye := theEye;
				res := SetEye(theEye, myCube, Dlg);
			end; (* while *)
			SetPort(savePort);
			sPDialogPos := WindowPeek(Dlg)^.contRgn^^.rgnBBox.topLeft;
{$IFC THINK_PASCAL }
			DisposDialog(Dlg);
{$ELSEC }
			DisposeDialog(Dlg);
{$ENDC }
			if (res = 0) then begin
				with currentFrame.proD^ do begin
					eyeProHasChanged := false;
					lightHasChanged := false;
					visHasChanged := false;
					proHasChanged := false;
					change := (theEye.x <> oldEye.x) | (theEye.y <> oldEye.y) | (theEye.z <> oldEye.z);
					change := change | (theEye.vx <> oldEye.vx) | (theEye.vy <> oldEye.vy) | (theEye.vz <> oldEye.vz);
					change := change | (oldEye.hidden <> theEye.hidden) | (oldEye.ProjectionType <> theEye.ProjectionType);
					if change then begin
						projZ := false;
						if theEye.ProjectionType <> oldEye.ProjectionType then begin
							projZ := not projZ;
							proHasChanged := true;
						end;
						if theEye.hidden <> oldEye.hidden then begin
							visibility := not visibility;
							visHasChanged := true;
						end;
						if (theEye.x <> oldEye.x) | (theEye.y <> oldEye.y) | (theEye.z <> oldEye.z) | (theEye.vx <> oldEye.vx) | (theEye.vy <> oldEye.vy) | (theEye.vz <> oldEye.vz) then begin
							oldProPoint := proPoint;
							proPoint[1] := theEye.vx;
							proPoint[2] := theEye.vy;
							proPoint[3] := theEye.vz;
							newProPoint(oldProPoint, proPoint, eyePoint, userTrafo, userTScal, projTrafo, viewTrafo);
							oldEyePoint := eyePoint;
							eyePoint[1] := theEye.x;
							eyePoint[2] := theEye.y;
							eyePoint[3] := theEye.z;
							newEyePoint(oldEyePoint, eyePoint, userTrafo, userTScal, projTrafo, viewTrafo);
							eyeProHasChanged := true;
						end;
					end;
					redraw(win);
				end; (* with *)
			end
			else begin
				with currentFrame.proD^ do begin
					proDOld^.lightHasChanged := false;
					proDOld^.eyeProHasChanged := (eyePoint[1] <> proDOld^.eyePoint[1]) | (eyePoint[2] <> proDOld^.eyePoint[2]) | (eyePoint[3] <> proDOld^.eyePoint[3]);
					proDOld^.eyeProHasChanged := proDOld^.eyeProHasChanged | (proPoint[1] <> proDOld^.proPoint[1]) | (proPoint[2] <> proDOld^.proPoint[2]) | (proPoint[3] <> proDOld^.proPoint[3]);
					proDOld^.visHasChanged := visibility <> proDOld^.visibility;
					proDOld^.proHasChanged := projZ <> proDOld^.projZ;
				end;
				currentFrame.proD^ := proDOld^;
				redraw(win);
			end;
			with currentFrame.proD^ do begin
				eyeProHasChanged := false;
				lightHasChanged := false;
				visHasChanged := false;
				proHasChanged := false;
			end;
			endCursor;
			startTime := startTime + TickCount - timer;
			dispose(proDOld);
		end; (* getEyePoint *)


	procedure setUserTransformation (var currentFrame: wFrame; win: windowIndex0);
		const
			uTResource = 36;
			uTOKItem = 1;
			uTCancelItem = 2;
			uTTItem = 3;
			uTSItem = 4;
			uTRItem = 5;
			uTTIncItem = 6;
			uTSIncItem = 7;
			uTRIncItem = 8;
			uTxUpItem = 9;
			uTxDoItem = 10;
			uTyUpItem = 11;
			uTyDoItem = 12;
			uTzUpitem = 13;
			uTzDoItem = 14;
			uTHiddenLineItem = 19;


			TDefault = 30;
			SDefault = 2;
			RDefault = 15;

		var
			uTDialog: DialogPtr;
			proDOld: projDescPtr;
			item, axis, dir: integer;
			timer: longInt;
			delta: point3dT;
			TInc, SInc, RInc: extended;
			op: (trans, scal, rot);

		begin
			with currentFrame do begin
				TInc := TDefault;
				SInc := SDefault;
				RInc := RDefault;
				op := trans;
				timer := TickCount;
				new(proDOld);
				failNil(Handle(proDOld));
				proDOld^ := proD^;
				beginCursor(arrowCursor);
				uTDialog := GetNewDialog(uTResource, nil, WindowPtr(-1));
				niceWindowPosition(WindowPtr(uTDialog), false);
				if uTDialogPos.v > 0 then
					MoveWindow(uTDialog, uTDialogPos.h, uTDialogPos.v, false);
				repeat
					with proD^ do begin
						eyeProHasChanged := false;
						lightHasChanged := false;
						visHasChanged := false;
						proHasChanged := false;
						SetDText(uTDialog, uTTIncItem, stringOf(TInc : 8 : 4));
						SetDText(uTDialog, uTSIncItem, stringOf(SInc : 8 : 4));
						SetDText(uTDialog, uTRIncItem, stringOf(RInc : 8 : 4));
						SetDCtl(uTDialog, uTTItem, op = trans);
						SetDCtl(uTDialog, uTSItem, op = scal);
						SetDCtl(uTDialog, uTRItem, op = rot);
						SetDCtl(uTDialog, uTHiddenLineItem, proD^.visibility);
						SelectWindow(uTDialog);
						ShowWindow(uTDialog);
						showDefaultButton(uTDialog, uTOKItem);
{$IFC THINK_PASCAL }
						ModalDialog(@dragDialogFilter, item);
{$ELSEC }
						ModalDialog(NewModalFilterProc(@dragDialogFilter), item);
{$ENDC }
						getValue(uTDialog, uTTIncItem, TDefault, TInc);
						getValue(uTDialog, uTSIncItem, SDefault, SInc);
						getValue(uTDialog, uTRIncItem, RDefault, RInc);
						case item of
							uTTItem: 
								op := trans;
							uTSItem: 
								op := scal;
							uTRItem: 
								op := rot;
							uTxUpItem..uTzDoItem:  begin
								axis := (item - uTxUpItem) div 2;
								dir := 1 - (item - uTxUpItem) mod 2 * 2;
								eyeProHasChanged := true;
								case op of
									trans:  begin
										delta[1] := 0;
										delta[2] := 0;
										delta[3] := 0;
										delta[4] := 1;
										delta[axis + 1] := dir * TInc;
										translation(delta, eyePoint, proPoint, lightPoint, userTrafo, userTScal, projTrafo, viewTrafo);
									end;
									rot: 
										rotation(axisT(axis), dir * RInc * pi / 180.0, eyePoint, proPoint, lightPoint, userTrafo, userTScal, projTrafo, viewTrafo);
									scal:  begin
										delta[1] := 1;
										delta[2] := 1;
										delta[3] := 1;
										delta[4] := 1;
										if dir < 0 then
											delta[axis + 1] := 1 / SInc
										else
											delta[axis + 1] := SInc;
										scaling(delta, eyePoint, proPoint, lightPoint, userTrafo, userTScal, projTrafo, viewTrafo);
									end;
								end;
							end;
							uTHiddenLineItem:  begin
								visHasChanged := true;
								visibility := not visibility;
								SetDCtl(uTDialog, uTHiddenLineItem, proD^.visibility);
							end;
							otherwise
								;
						end; (* case *)
						if not ((item = uTTItem) | (item = uTSItem) | (item = uTRItem) | (item = uTCancelItem) | (item = uTOKItem)) then begin
							redraw(win);
						end;
					end;
				until (item = uTCancelItem) | (item = uTOKItem);

				if item = uTCancelItem then begin
					proD^ := proDOld^;
					proD^.eyeProHasChanged := true;
					proD^.lightHasChanged := true;
					proD^.visHasChanged := true;
					proD^.proHasChanged := true;
					redraw(win);
				end
				else
					wStatus^[win].zoomed := false;
				uTDialogPos := WindowPeek(uTDialog)^.contRgn^^.rgnBBox.topLeft;
				HideWindow(uTDialog);
				endCursor;
{$IFC THINK_PASCAL }
				DisposDialog(uTDialog);
{$ELSEC }
				DisposeDialog(uTDialog);
{$ENDC }
				startTime := startTime + TickCount - timer;
				proD^.eyeProHasChanged := false;
				proD^.lightHasChanged := false;
				proD^.visHasChanged := false;
				proD^.proHasChanged := false;
			end; (* with *)
			dispose(proDOld);
		end; (* setUserTransformation *)

	procedure executeUndo;
(* assumes the current window's desktop and selection are exactly the same as immediately *)
(* after the undoable command was performed! So call forgetUndo often enough! *)
		begin
			with wStatus^[w] do begin
				case undoCmd of
					dragDesktopCmd: 
						fixList(desktop, selection, false, window);
					dragSelectionCmd: 
						fixList(selection, desktop, true, window);
					clearCmd:  begin
						selection.free;
						selection := undoScrap;
						selection.interactiveOutput;
						selection.interactiveHighlight;
					end;
					replaceSelectionCmd:  begin
						selection.interactiveHighlight;
						selection.interactiveOutput;
						stopwatchFreeAll(obj(selection));
						selection := undoScrap;
						selection.interactiveOutput;
						selection.interactiveHighlight;
					end;
					clearAllCmd:  begin
						selection.free;
						desktop.free;
						selection := list(undoScrap.first^^.value);
						undoScrap.pop;
						desktop := undoScrap;
						redraw(w);
					end;
					transformCmd, setInfoCmd, randomChangeCmd, ungroupCmd:  begin
						if desktop.length = 0 then
							clearWindow(window)
						else begin
							selection.interactiveOutput;
							selection.interactiveHighlight;
						end;
						stopwatchFreeAll(obj(selection));
						selection := undoScrap;
						selection.interactiveOutput;
						selection.interactiveHighlight;
					end;
				end;
				testObject(selection, names, types);
				buildOperationsMenu(names);
			end;
			undoScrap := nil;
		end;

	procedure duplicateCurrentSelection (w: windowIndex);
		var
			selectionCopy: list;
		begin
			forgetUndo;
			with wStatus^[w] do begin
				selection.interactiveHighlight;
				selectionCopy := list(stopwatchDuplicateAll(selection));
				selection.translate(defaultDx, defaultDy);
				selection.interactiveOutput;
				selection.interactiveHighlight;
				desktop.concatenateDispose(selectionCopy);
			end;
		end;

	begin
		fw := FrontWindow;
		case item of (* standard menus for desk accessories *)
			undoItem: 
				consumedByDA := SystemEdit(1);
			cutItem: 
				consumedByDA := SystemEdit(2);
			copyItem: 
				consumedByDA := SystemEdit(3);
			pasteItem: 
				consumedByDA := SystemEdit(4);
			clearItem: 
				consumedByDA := SystemEdit(5);
			otherwise
				consumedByDA := false;
		end;
		if not consumedByDA then begin
{$IFC ScriptEnabled }
			if not (currentScriptWindow < 0) then  (* delo new *)
				w := currentWindow;
			if currentScriptWindow < 0 then begin (* delo new *)
				EWindowEditOp(item);
			end
			else
{$ELSEC }
				w := currentWindow;
{$ENDC }
			with wStatus^[w] do begin
				modified := true;
				case item of
					undoItem: 
						executeUndo;
					cutItem:  begin
						forgetUndo;
						newScrap(window);
						scrapBook := selection;
						if desktop.length = 0 then
							clearWindow(window)
						else begin
							selection.interactiveHighlight;
							selection.interactiveOutput;
						end;
						selection := newEmptyList;
						testObject(selection, names, types);
						buildOperationsMenu(names);
					end;
					cutFirstItem:  begin
						forgetUndo;
						newScrap(window);
						scrapBook := newEmptyList;
						scrapBook.append(selection.first^^.value);
						selection.pop;
						scrapBook.interactiveHighlight;
						scrapBook.interactiveOutput;
						testObject(selection, names, types);
						buildOperationsMenu(names);
					end;
					copyFirstItem:  begin
						forgetUndo;
						newScrap(window);
						scrapBook := newEmptyList;
						scrapBook.append(stopwatchDuplicateAll(selection.first^^.value));
					end;
					selectNextItem:  begin
						forgetUndo;
						if shiftKeyPressed & not optionKeyPressed then
							with desktop.first^^ do begin
								value.interactiveHighlight;
								selection.append(value);
							end
						else if optionKeyPressed & not shiftKeyPressed then
							with selection.last^^ do begin
								value.interactiveHighlight;
								desktop.append(value);
								value := desktop.first^^.value;
								value.interactiveHighlight;
							end
						else begin
							selection.interactiveHighlight;
							desktop.concatenate(selection);
							selection.init;
							selection.append(desktop.first^^.value);
							selection.interactiveHighlight;
						end;
						desktop.pop;
						testObject(selection, names, types);
						buildOperationsMenu(names);
					end;
					duplicateItem: 
						duplicateCurrentSelection(w);
					copyItem:  begin
						forgetUndo;
						newScrap(window);
						scrapBook := list(stopwatchDuplicateAll(selection));
					end;
					pasteItem:  begin
						forgetUndo;
						selection.interactiveHighlight;
						desktop.concatenateDispose(selection);
						if scrapBook = nil then
							selection := loadScrapBook
(* not scrapBook := loadScrapBook, because the newest arithmetics and frame *)
(* settings should be used at each paste command when no XYGB entry is available *)
(* (see comment to the procedure loadScrapBook) *)
						else
							selection := list(stopwatchDuplicateAll(scrapBook));
						if optionKeyPressed then
							moveToWindowPoint(selection, previousClickPoint);
						selection.interactiveOutput;
						selection.interactiveHighlight;
						testObject(selection, names, types);
						buildOperationsMenu(names);
					end;
					clearItem:  begin
						forgetUndo;
						clearWindow(window);
						desktop.push(selection);
						undoCmd := clearAllCmd;
						undoScrap := desktop;
						if not undoEnabled then
							forgetUndo;
						desktop := newEmptyList;
						selection := newEmptyList;
						testObject(selection, names, types);
						buildOperationsMenu(names);
					end;
					selectAllItem:  begin
						forgetUndo;
						desktop.interactiveHighlight;
						selection.concatenate(desktop);
						desktop.init;
						testObject(selection, names, types);
						buildOperationsMenu(names);
						selectedType := lastObjectType;
						updatePalette(selectItem);
					end;
					redrawItem: 
						redraw(w);
					zoomInItem:  begin
						beginCursor(littleCrossCursor);
						zoomIn(w);
						endCursor;
					end;
					zoomOutItem: 
						zoomOut(w);
					eyePointItem: 
						getProjSettings(frames.stack[frames.stackPointer], currentWindow);
					setUserTItem: 
						setUserTransformation(frames.stack[frames.stackPointer], currentWindow);
					clearClipboardItem:  begin
						stopwatchFreeAll(obj(scrapBook));
						scrapBook := newEmptyList;
					end;
				end;
			end;
		end;
	end;

procedure DoTransMenu (menu, item: integer; when: longInt);
(* tranformations are undoable commands *)
	var
		reallyDoit: boolean;
	begin
		with wStatus^[currentWindow] do begin
			case item of
				translateItem: 
					reallyDoit := getUserParameter2('Delta(X)', 'Delta(Y)', defaultDx, defaultDy, defaultDx, defaultDy, nil, nil) & ((defaultDx <> 0) | (defaultDy <> 0));
				rotateItem: 
					reallyDoit := getUserParameter1('Angle (in degrees)', defaultAngle, defaultAngle, nil) & (defaultAngle <> 0);
				scaleItem: 
					reallyDoit := getUserParameter3('Lambda', defaultLambda, defaultLambda, nil, 'X-Origin', defaultOriginX, defaultOriginX, nil, 'Y-Origin', defaultOriginY, defaultOriginY, nil);
				reflectItem: 
					reallyDoit := getUserParameter3('A in AX+BY=C', defaultA, defaultA, nil, 'B in AX+BY=C', defaultB, defaultB, @checkB, 'C in AX+BY=C', defaultC, defaultC, nil);
			end;
			if reallyDoit then begin
				modified := true;
				forgetUndo;
				undoCmd := transformCmd;
				if undoEnabled then
					undoScrap := list(stopwatchDuplicateAll(selection));
				selection.interactiveHighlight;
				selection.interactiveOutput;
				case item of
					translateItem: 
						selection.translate(defaultDx, defaultDy);
					rotateItem: 
						if member(selection.ith(1), point2d) then
							selection.rotate(point2d(selection.ith(1)).getX, point2d(selection.ith(1)).getY, oneDegree * defaultAngle)
						else
							selection.ith(1).rotate(point2d(selection.ith(2)).getX, point2d(selection.ith(2)).getY, oneDegree * defaultAngle);
					scaleItem:  begin
						selection.translate(-defaultOriginX, -defaultOriginY);
						selection.scale(defaultLambda);
						selection.translate(defaultOriginX, defaultOriginY);
					end;
					reflectItem: 
						selection.reflect(defaultA, defaultB, defaultC);
				end;
				selection.interactiveOutput;
				selection.interactiveHighlight;
			end;
		end;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

end.
