(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit geoBenchFileMenu;

interface

uses
{$IFC ScriptEnabled }
	GeoScript, 
{$ENDC }
{$IFC Topography }
	swissTopo, 
{$ENDC }
{$IFC NOT THINK_PASCAL }
	SegLoad,
{$ENDC }
	browse, geoBenchPreferences, GFUtilities, Notification;

procedure DoFileMenu (menu, item: integer; when: longInt);

{$IFC Topography }
procedure DoTopoMenu (menu, item: integer; when: longInt);
{$ENDC }

{$IFC ScriptEnabled AND NetworkEnabled }
procedure DoCosimaMenu (menu, item: integer; when: longInt);					(* MS: 19.7.93 - Cosima Menu Handler *)
{$ENDC }

procedure DoWClobber;

procedure defaultModes (newTextSize: integer);

procedure buildOperationsMenu (names: list);

procedure activate (w: windowIndex; active: boolean);

procedure doObjectInfo (var selection: list);

procedure wakeupUser;

procedure removeNotification;

procedure inputFile (newName: Str255; volume: integer);

function makeWindow (r: Rect; oldName: Str255; sysName, modify: boolean; v: integer): windowIndex;

function niceRect (addToTheRight: boolean): Rect;

implementation

{$IFC NOT THINK_PASCAL }
uses StandardFile;
{$ENDC }

{$SETC TraceEvents = true }
{$SETC DebugWriteText = true }

{$IFC TraceEvents }
var
	lineCount: longInt;
{$ENDC }

const
	maxAlloc = 100000;

type
	OffScreenPort = object(obj) (* Offscreen drawing support, based on code by Cristoph Wirth. *)
			fTarget, fOSPort: GrafPtr;
			fArea: Rect;
			fTempRgn: RgnHandle;
			portIsOk: boolean;

			procedure free;
			override;

			procedure InitPort (target: GrafPtr; screenArea: Rect);
			procedure BeginOffScreen;
			procedure EndOffScreen;
		end;

procedure OffScreenPort.free;
	begin
		if portIsOk then begin
			DisposeRgn(fTempRgn);
{$IFC THINK_PASCAL }
			DisposPtr(fOSPort^.portBits.baseAddr);
{$ELSEC }
			DisposePtr(fOSPort^.portBits.baseAddr);
{$ENDC }
			fOSPort^.portBits.baseAddr := nil;
			ClosePort(fOSPort);
			Dispose(fOSPort);
		end;
		shallowFree;
	end;

procedure OffScreenPort.InitPort (target: GrafPtr; screenArea: Rect);
	var
		savePort: GrafPtr;
		bmap: BitMap;
		height, allocation: longInt;
	begin
		fTarget := target;
		fArea := screenArea;
		GetPort(savePort);
		NEW(fOSPort);
		failNil(Handle(fOSPort));
		OpenPort(fOSPort);
		defaultModes(target^.txSize);
		bmap.rowBytes := (screenArea.right - screenArea.left) div 8 + 1;
		if ODD(bmap.rowBytes) then
			bmap.rowBytes := bmap.rowBytes + 1;
		bmap.bounds := screenArea;
		height := screenArea.bottom - screenArea.top;
		allocation := bmap.rowBytes * height;
		if allocation > maxAlloc then
			portIsOk := false
		else begin
			bmap.baseAddr := NewPtr(bmap.rowBytes * height);
			portIsOk := bmap.baseAddr <> nil;
		end;
		if portIsOk then begin
			fOSPort^.portBits := bmap;
			fOSPort^.portRect := screenArea;
			RectRgn(fOSPort^.visRgn, screenArea);
			RectRgn(fOSPort^.clipRgn, screenArea);
			EraseRect(screenArea);
			fTempRgn := NewRgn;
			failNil(Handle(fTempRgn));
		end
		else begin
			ClosePort(fOSPort);
			Dispose(fOSPort);
		end;
		SetPort(savePort);
	end;

procedure OffScreenPort.BeginOffScreen;
	begin
		if portIsOk then begin
			offScreenId := actualWindow;
			SetPort(fOSPort);
			CopyBits(fTarget^.portBits, fOSPort^.portBits, fArea, fArea, srcCopy, nil);
		end;
	end;

procedure OffScreenPort.EndOffScreen;
	begin
		if portIsOk then begin
			SetPort(fTarget);
			CopyBits(fOSPort^.portBits, fTarget^.portBits, fArea, fArea, srcCopy, nil);
			offScreenId := 0;
		end;
	end;

procedure obj.dragObjectTransformTranslation (procedure whatToDoBefore (o: obj); procedure whatToDoAfter (o: obj); mouse: Point; procedure transformTranslation (var dx, dy: extended));
	var
		oldMouse: Point;
		dh, dv: integer;
		moveDx, moveDy: extended;
		offScreen: OffScreenPort;
	begin
		oldMouse := mouse;
		new(offScreen);
		failNil(Handle(offScreen));
		offScreen.InitPort(currentGeoBenchWindow, drawingRect(currentGeoBenchWindow));
		while Button do begin
			writeCursorCoordinates(false);
			GetMouse(mouse);
			dh := mouse.h - oldMouse.h;
			dv := mouse.v - oldMouse.v;
			if (dh <> 0) | (dv <> 0) then begin
				moveDx := dh / currentFrame.mx;
				moveDy := dv / currentFrame.my;
				transformTranslation(moveDx, moveDy);
				offScreen.BeginOffScreen;
				whatToDoBefore(self);
				interactiveOutput;
				translate(moveDx, moveDy);
				interactiveOutput;
				whatToDoAfter(self);
				offScreen.EndOffScreen;
				oldMouse := mouse;
			end;
		end;
		offScreen.free;
	end;

	var
		pageLength, scrollStep: integer;
		scrollFactor: real;	(*	= wStatus^[w].horFactor or verFactor													*)
		sheetPos: Point;		(*	position where the toolbox 'thinks' the sheet is, while running TrackControl,		*)
							(*	i.e. while pressing the mouse button														*)
		subMenus: array[firstSubMenuID..maxSubMenuID] of record
				entries: list;
				mainMenuEntry: integer;
			end;

	var
		myNotification: NMRec;
		wakeUpString: Str255;

procedure wakeupUser;
	begin
		if processInBackground & not notificationPending then begin
			with myNotification do begin
				qType := ORD(nmType);
				nmMark := 1;					(* add diamond-shaped mark in the Application menu *)
				nmIcon := GetResource('SICN', 1234);
				nmResp := nil;
				nmRefCon := 0;
				if waitLimit > 0 then
					nmSound := Handle(-1)		(* do beep *)
				else
					nmSound := nil;				(* be quiet *)
				if showWakeUpString then begin
					wakeUpString := concat('GeoBench operation ''', currentOperation, ''' has terminated.');
					nmStr := @wakeUpString;
				end
				else
					nmStr := nil;
			end;
			notificationPending := NMInstall(@myNotification) = noErr;
		end;
	end;

procedure removeNotification;
	var
		error: OSErr;
	begin
		if notificationPending then begin
			error := NMRemove(@myNotification);
			notificationPending := false;
		end;
	end;

procedure setNewName (newName: Str255; w: windowIndex);
	begin
		with wStatus^[w] do begin
			fileName := newName;
			name := GetFileName(newName);
			SetWTitle(window, name);
{$IFC THINK_PASCAL }
			SetItem(windowsMenu, w + lastWindowItem + 1, name);
{$ELSEC }
			SetMenuItemText(windowsMenu, w + lastWindowItem + 1, name);
{$ENDC }
			systemName := false;
		end;
	end;

procedure defaultModes (newTextSize: integer);
	begin
		PenMode(patXor);
{$IFC THINK_PASCAL }
		PenPat(black);
{$ELSEC }
		PenPat(qd.black);
{$ENDC }
		TextFont(geoBenchFont);
		TextFace([]);
		TextMode(srcXor);
		TextSize(newTextSize);
	end;

procedure DoWClobber;
	var
		thePort: GrafPtr;
	begin
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' Clobber');
{$ENDC }
		GetPort(thePort);	(* grafport of window to dispose of *)
		DisposeWindow(WindowPtr(thePort));
	end;

procedure doScroll (scroll: ControlHandle; value: integer);
	var
		vertical: boolean;
		w: windowIndex;
		r1, r2, cr: Rect;
	begin
		w := currentWindow;
		with wStatus^[w] do begin
			vertical := scroll = verScroll;
			if value < 0 then
				value := 0
			else if value > maxScrollValue then
				value := maxScrollValue;
			bigClip(window);
{$IFC THINK_PASCAL }
			SetCtlValue(scroll, value);
{$ELSEC }
			SetControlValue(scroll, value);
{$ENDC }
			r1 := drawingRect(window);
			r2 := r1;
			cr := r1;
			with frames, stack[stackPointer] do
				if vertical then begin
{$IFC THINK_PASCAL }
					value := sheetPos.v + round(value / verScrollFactor);
{$ELSEC }
					value := sheetPos.v + roundLong(value / verScrollFactor);
{$ENDC }
					dy := dy - value;
					sheetPos.v := sheetPos.v - value;
					if abs(value) < r1.bottom - r1.top then
						if value > 0 then begin
							r1.top := r1.top + value;
							r2.bottom := r2.bottom - value;
							cr.top := r2.bottom;
						end
						else begin
							r1.bottom := r1.bottom + value;
							r2.top := r2.top - value;
							cr.bottom := r2.top;
						end;
				end
				else begin
{$IFC THINK_PASCAL }
					value := sheetPos.h + round(value / horScrollFactor);
{$ELSEC }
					value := sheetPos.h + roundLong(value / horScrollFactor);
{$ENDC }
					dx := dx - value;
					sheetPos.h := sheetPos.h - value;
					if abs(value) < r1.right - r1.left then
						if value > 0 then begin
							r1.left := r1.left + value;
							r2.right := r2.right - value;
							cr.left := r2.right;
						end
						else begin
							r1.right := r1.right + value;
							r2.left := r2.left - value;
							cr.right := r2.left;
						end;
				end;
			if not EqualRect(r1, r2) then begin
				littleClip(window);
{$IFC THINK_PASCAL }
				CopyBits(thePort^.portBits, thePort^.portBits, r1, r2, srcCopy, nil);
{$ELSEC }
				CopyBits(qd.thePort^.portBits, qd.thePort^.portBits, r1, r2, srcCopy, nil);
{$ENDC }
			end;
			ClipRect(cr);

			sheet := drawingRect(window);
			sheetExtended := true;
			EraseRect(window^.portRect);
			with frames do begin
				currentFrame := stack[stackPointer];
				coordSysIsDrawn := false;
				desktop.interactiveOutput;
				selection.interactiveOutput;
				selection.interactiveHighlight;
			end;
			bigClip(window);
		end;
	end;

procedure trackScroll (scroll: ControlHandle; code: integer);
	var
		value: real;
	begin
{$IFC THINK_PASCAL }
		if (code = GetCRefCon(scroll)) then begin
{$ELSEC }
		if (code = GetControlReference(scroll)) then begin
{$ENDC }
			case code of
{$IFC THINK_PASCAL }
				inUpButton: 
					if GetCtlMin(scroll) <> GetCtlValue(scroll) then
{$ELSEC }
				kControlUpButtonPart:
					if GetControlMinimum(scroll) <> GetControlValue(scroll) then
{$ENDC }
						value := -scrollStep
					else
						value := 0;
{$IFC THINK_PASCAL }
				inDownButton: 
					if GetCtlMax(scroll) <> GetCtlValue(scroll) then
{$ELSEC }
				kControlDownButtonPart: 
					if GetControlMaximum(scroll) <> GetControlValue(scroll) then
{$ENDC }
						value := scrollStep
					else
						value := 0;
{$IFC THINK_PASCAL }
				inPageUp: 
{$ELSEC }
				kControlPageUpPart: 
{$ENDC }
					value := -pageLength;
{$IFC THINK_PASCAL }
				inPageDown: 
{$ELSEC }
				kControlPageDownPart: 
{$ENDC }
					value := pageLength;
				otherwise
					value := 0;
			end;
			if value <> 0 then begin
{$IFC THINK_PASCAL }
				value := value * scrollFactor + GetCtlValue(scroll);
{$ELSEC }
				value := value * scrollFactor + GetControlValue(scroll);
{$ENDC }
				if value > maxScrollValue then
					value := maxScrollValue
				else if value < 0 then
					value := 0;
{$IFC THINK_PASCAL }
				doScroll(scroll, round(value));
{$ELSEC }
				doScroll(scroll, roundLong(value));
{$ENDC }
			end;
		end;
	end;

procedure inScrollBar (wIndex: windowIndex; var scroll: ControlHandle; partCode: integer; pt: Point);
	var
		part: integer;
		pLength: integer;
		sStep: real;
		r: Rect;
	begin
		with wStatus^[wIndex] do begin
			sheetPos := sheet.topLeft;
			bigClip(window);
{$IFC THINK_PASCAL }
			if partCode = inThumb then begin
{$ELSEC }
			if partCode = kControlIndicatorPart then begin
{$ENDC }
				part := TrackControl(scroll, pt, nil);
				if part = partCode then
{$IFC THINK_PASCAL }
					doScroll(scroll, GetCtlValue(scroll));
{$ELSEC }
					doScroll(scroll, GetControlValue(scroll));
{$ENDC }
			end
			else if partCode <> 0 then begin
				r := drawingRect(window);
				if scroll = verScroll then begin
					pLength := r.bottom;
					scrollFactor := verScrollFactor;
				end
				else begin
					pLength := r.right;
					scrollFactor := horScrollFactor;
				end;
				sStep := pLength * scrollStepRatio;
{$IFC THINK_PASCAL }
				pageLength := round(pLength - sStep);
				scrollStep := round(sStep);
				SetCRefCon(scroll, longInt(partCode));
				part := TrackControl(scroll, pt, @trackScroll);
{$ELSEC }
				pageLength := roundLong(pLength - sStep);
				scrollStep := roundLong(sStep);
				SetControlReference(scroll, longInt(partCode));
				part := TrackControl(scroll, pt, NewControlActionProc(@trackScroll));
{$ENDC }
			end;
			littleClip(window);
		end;
	end;

procedure doObjectInfo (var selection: list);
	var
		status: browseStatus;
		oldPort: GrafPtr;
		newInfo: vector;
		changing: obj;
		newSel: list;
		d: integer;
	begin
		GetPort(oldPort);
		newSel := list(stopwatchDuplicateAll(selection));
		if selection.length = 1 then
			changing := newSel.first^^.value
		else
			changing := newSel;
		newInfo := vector(changing.getInfo);
		if newInfo = nil then begin
			d := CautionAlert(noInfoAlertResource, nil);
			stopwatchFreeAll(obj(newSel));
		end
		else begin
			enterBrowse;
			status := browse(newInfo, false, false, false);
			exitBrowse;
			changing.setInfo(newInfo);
			SetPort(oldPort);
			if (status = browseOk) & browseChange then begin
				selection.interactiveHighlight;
				selection.interactiveOutput;
				forgetUndo;
				undoCmd := setInfoCmd;
				undoScrap := selection;
				if not undoEnabled then
					forgetUndo;
				selection := newSel;
				selection.interactiveOutput;
				selection.interactiveHighlight;
			end
			else
				stopwatchFreeAll(obj(newSel));
			stopwatchFreeAll(obj(newInfo));
		end;
	end;

function determineUseOffScreen (o: obj; highlight: boolean): boolean;
	var
		complexity: longInt;
	begin
		if highlight then
			complexity := maxDrawingComplexity div 2
		else
			complexity := maxDrawingComplexity;
		determineUseOffScreen := o.drawingComplexity(complexity) >= complexity;
	end;

procedure DoWMouse (pt: Point; time: longInt; mods: integer);
	var
		w: windowIndex;
		r: rectangle;
		o: obj;
		newSelection, newDesktop: list;
		clickOnly: boolean;
		re: Rect;
		whichControl: ControlHandle;
		partCode: integer;
		dummy: elementH;

	procedure partition (source, yesPart, noPart: list; r: rectangle; highlight, selectOnlyOne: boolean; var where: elementH);
		var
			p: elementH;
			oneSelected: boolean;
		begin
			p := source.first;
			oneSelected := false;
			where := nil;
			while p <> nil do begin
				if not (oneSelected & selectOnlyOne) & p^^.value.isSelected(r) then begin
					oneSelected := true;
					where := p;
					yesPart.append(p^^.value);
					if highlight then
						p^^.value.interactiveHighlight;
				end
				else
					noPart.append(p^^.value);
				p := p^^.next;
			end;
		end;

	function getFirstSelected (source: list; r: rectangle): elementH;
		var
			p: elementH;
		begin
			p := source.first;
			while (p <> nil) & not p^^.value.isSelected(r) do
				p := p^^.next;
			getFirstSelected := p;
		end;

	function dragIntended (initialMouse: Point; w: windowIndex): boolean;
		var
			r: rectangle;
			newSelection, newDesktop: list;
			moved, useOffScreenDrawing, moveSelected: boolean;
			mouse, oldMouse: Point;
			offScreen: OffScreenPort;
			firstInSelection, firstInDesktop: elementH;
			objectToDrag: obj;
		begin
			new(r);
			failNil(Handle(r));
			with wStatus^[w] do begin
				getSelectionRectangle(initialMouse, r);
				firstInSelection := getFirstSelected(selection, r);
				firstInDesktop := getFirstSelected(desktop, r);
				if (firstInSelection = nil) & (firstInDesktop = nil) then begin
					dispose(r);
					dragIntended := false;
					exit(dragIntended);
				end;
				dragIntended := true;
				oldMouse := initialMouse;
				moved := false;
				currentGeoBenchWindow := window;
				while Button do begin
					writeCursorCoordinates(false);
					GetMouse(mouse);
					if not EqualPt(mouse, oldMouse) then begin
						if not moved then begin
							moveSelected := firstInSelection <> nil;
							if moveSelected then begin
								undoCmd := dragSelectionCmd;
								undoPosition := firstInSelection;
							end
							else begin
								undoCmd := dragDesktopCmd;
								undoPosition := firstInDesktop;
							end;
							objectToDrag := undoPosition^^.value;
							if undoEnabled then begin
								undoScrap := newEmptyList;
								undoScrap.append(objectToDrag.duplicateAll);
							end;
							offScreen := nil;
							useOffScreenDrawing := determineUseOffScreen(objectToDrag, moveSelected);
						end;
						if moved | not objectToDrag.dragVertex(r, initialMouse, moveSelected) then begin
							if useOffScreenDrawing & (offScreen = nil) then begin
								new(offScreen);
								failNil(Handle(offScreen));
								offScreen.InitPort(window, drawingRect(window));
							end;
							if useOffScreenDrawing then
								offScreen.BeginOffScreen;
							objectToDrag.interactiveOutput;
							if moveSelected then
								objectToDrag.interactiveHighlight;
							objectToDrag.translate((mouse.h - oldMouse.h) / currentFrame.mx, (mouse.v - oldMouse.v) / currentFrame.my);
							objectToDrag.interactiveOutput;
							if moveSelected then
								objectToDrag.interactiveHighlight;
							if useOffScreenDrawing then
								offScreen.EndOffScreen;
							oldMouse := mouse;
						end;
						moved := true;
						zoomed := false;
					end;
				end;
				if moved then begin
					if offScreen <> nil then
						offScreen.free;
				end
				else begin
					forgetUndo;
					newSelection := newEmptyList;
					newDesktop := newEmptyList;
					partition(desktop, newSelection, newDesktop, r, false, true, undoPosition);
					if newSelection.length = 0 then begin
						newDesktop.reset;
						partition(selection, newSelection, newDesktop, r, false, true, undoPosition);
						moveSelected := true;
					end
					else
						moveSelected := false;
					selection.interactiveHighlight;
					if moveSelected then begin
						newDesktop.concatenateDispose(desktop);
						selection.free;
					end
					else begin
						newDesktop.concatenateDispose(selection);
						desktop.free;
					end;
					selection := newSelection;
					desktop := newDesktop;
					selection.interactiveHighlight;
					testObject(selection, names, types);
					buildOperationsMenu(names);
				end;
			end;
			dispose(r);
		end;

	procedure inPalette (y: integer);
		var
			item, i: integer;
		begin
			item := 0;
			if y < paletteSelect.bottom then
				item := selectItem
			else begin
				i := firstObjectItem;
				repeat
					if y < paletteObjects[i].bottom then
						item := i;
					i := i + 1;
				until (i > lastObjectItem) | (item <> 0);
			end;
			case item of
				firstObjectItem..lastObjectItem:  begin
					selectedType := itemToType[item];
					updatePalette(item);
				end;
				selectItem:  begin
					selectedType := lastObjectType;
					updatePalette(selectItem);
				end;
				otherwise
					;
			end;
		end;

	begin (* DoWMouse *)
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' Mouse((', pt.h : 1, ',', pt.v : 1, '), ', time : 1, ',', mods : 1, ')');
{$ENDC }
		w := currentWindow;
		with wStatus^[w] do begin
			if pt.h < 0 then
				inPalette(pt.v)
			else begin
				re := drawingRect(window);
				if PtInRect(pt, re) then
					if (time - previousClickTime < GetDblTime) & EqualPt(previousClickPoint, pt) & (selection.length > 0) then
						doObjectInfo(selection)
					else begin
						previousClickPoint := pt;
						previousClickTime := time;
						modified := true;
						forgetUndo;
						if selectedType = lastObjectType then
							if (BAND(shiftKey, mods) = 0) & dragIntended(pt, w) then
							else begin
								new(r);
								failNil(Handle(r));
								r.windowInputMouse(pt);
								clickOnly := (r.left = r.right) & (r.top = r.bottom);
								if r.top - r.bottom < 2 * userTolerance then begin
									r.top := r.top + userTolerance;
									r.bottom := r.bottom - userTolerance;
								end;
								if r.right - r.left < 2 * userTolerance then begin
									r.right := r.right + userTolerance;
									r.left := r.left - userTolerance;
								end;
								r.windowToWorld;
								if BAND(shiftKey, mods) = 0 then begin
									selection.interactiveHighlight;
									desktop.concatenate(selection);
									selection.init;
								end;
								newSelection := newEmptyList;
								newDesktop := newEmptyList;
								if (BAND(shiftKey, mods) <> 0) & clickOnly then begin
									partition(desktop, newSelection, newDesktop, r, true, true, dummy);
									if newSelection.length > 0 then
										selection.concatenateDispose(newSelection)
									else begin
										partition(selection, newDesktop, newSelection, r, true, true, dummy);
										selection.free;
										selection := newSelection;
									end;
								end
								else begin
									partition(selection, newDesktop, newSelection, r, true, clickOnly, dummy);
									partition(desktop, newSelection, newDesktop, r, true, clickOnly, dummy);
									selection.free;
									selection := newSelection;
								end;
								desktop.free;
								desktop := newDesktop;
								r.free;
								testObject(selection, names, types);
								buildOperationsMenu(names);
							end
						else begin
							currentPointType := arithmeticsToPointTypeRange(currentArithmetic);
{$IFC THINK_PASCAL }
							currentDimension := trunc(defaultDimension);
							currentVertices := trunc(defaultVertices); (* for the proper initialization of 'o' *)
{$ELSEC }
							currentDimension := truncLong(defaultDimension);
							currentVertices := truncLong(defaultVertices); (* for the proper initialization of 'o' *)
{$ENDC }
							currentRegularNGonAngle := defaultRegularNGonAngle * oneDegree;
							currentAlignmentIsOn := defaultAlignmentIsOn;
							if selectedType = point2dType then
								o := newObj(currentPointType)
							else
								o := newObj(selectedType);
							o.interactiveInput(pt.h, pt.v);
							o.interactiveOutput;
							desktop.append(o);
							zoomed := false;
						end;
					end
				else begin
					partCode := FindControl(pt, window, whichControl);
					if whichControl <> nil then
						inScrollBar(w, whichControl, partCode, pt);
				end;
			end;
		end;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoWKey (c: char; mods: integer);
	begin
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' Key(''', c, ''', ', mods : 1, ')');
{$ENDC }
		with wStatus^[currentWindow] do
			if (c = DeleteKey) & (selection.length > 0) then begin
				if desktop.length = 0 then
					clearWindow(window)
				else begin
					selection.interactiveHighlight;
					selection.interactiveOutput;
				end;
				undoScrap := selection;
				undoCmd := clearCmd;
				if not undoEnabled then
					forgetUndo;
				selection := newEmptyList;
				testObject(selection, names, types);
				buildOperationsMenu(names);
			end
			else
				SysBeep(1);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure DoWGrow;
	var
		w: windowIndex;
		r, s, t: Rect;
	begin
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' Grow');
{$ENDC }
		w := actualWindow;
		with wStatus^[w] do begin
			t.left := verScroll^^.contrlRect.left;
			t.right := verScroll^^.contrlRect.right;
			t.top := horScroll^^.contrlRect.top;
			t.bottom := horScroll^^.contrlRect.bottom;
			HideControl(verScroll);
			HideControl(horScroll);
			r := window^.portRect;
			s.left := r.right - scrollBarWidth;
			s.bottom := r.bottom - scrollBarWidth + 1;
			s.top := r.top - 1;
			s.right := r.right + 1;
			SizeControl(verScroll, s.right - s.left, s.bottom - s.top);
			MoveControl(verScroll, s.left, s.top);
			s.left := r.left + paletteWidth - 1;
			s.top := r.bottom - scrollBarWidth;
			s.bottom := r.bottom + 1;
			s.right := r.right - scrollBarWidth + 1;
			SizeControl(horScroll, s.right - s.left, s.bottom - s.top);
			MoveControl(horScroll, s.left, s.top);
			ShowControl(verScroll);
			ShowControl(horScroll);
			InvalRect(t);
			drawGrow(window);
			zoomed := false;
		end;
		updateScrollBars(w);
	end;

procedure DoWZoom;
	var
		sheetSize, wPos: Point;
		r: Rect;
	begin
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' Zoom');
{$ENDC }
		with wStatus^[actualWindow] do begin
			r := drawingRect(window);
			if zoomed then begin
				SizeWindow(window, zoomRect.right - zoomRect.left, zoomRect.bottom - zoomRect.top, true);
				MoveWindow(window, zoomRect.left, zoomRect.top, false);
				with frames.stack[frames.stackPointer] do begin
					dx := dx + zoomOffset.h;
					dy := dy + zoomOffset.v;
				end;
				DoWGrow;
			end
			else begin
				sheetSize.h := sheet.right - sheet.left + scrollBarWidth + paletteWidth;
				sheetSize.v := sheet.bottom - sheet.top + scrollBarWidth;
				zoomRect := WindowPeek(window)^.contRgn^^.rgnBBox;
				wPos := zoomRect.topLeft;
{$IFC THINK_PASCAL }
				if (wPos.h + sheetSize.h > screenBits.bounds.right) | (wPos.v + sheetSize.v > screenBits.bounds.bottom) then
{$ELSEC }
				if (wPos.h + sheetSize.h > qd.screenBits.bounds.right) | (wPos.v + sheetSize.v > qd.screenBits.bounds.bottom) then
{$ENDC }
					MoveWindow(window, standardXOffset, standardYOffset, false);
				zoomOffset := sheet.topLeft;
{$IFC THINK_PASCAL }
				if sheetSize.h > screenBits.bounds.right - standardXOffset then begin
					sheetSize.h := screenBits.bounds.right - standardXOffset;
{$ELSEC }
				if sheetSize.h > qd.screenBits.bounds.right - standardXOffset then begin
					sheetSize.h := qd.screenBits.bounds.right - standardXOffset;
{$ENDC }
					zoomOffset.h := 0;
				end;
{$IFC THINK_PASCAL }
				if sheetSize.v > screenBits.bounds.bottom - standardYOffset then begin
					sheetSize.v := screenBits.bounds.bottom - standardYOffset;
{$ELSEC }
				if sheetSize.v > qd.screenBits.bounds.bottom - standardYOffset then begin
					sheetSize.v := qd.screenBits.bounds.bottom - standardYOffset;
{$ENDC }
					zoomOffset.v := 0;
				end;
				SizeWindow(window, sheetSize.h, sheetSize.v, true);
				with frames.stack[frames.stackPointer] do begin
					dx := dx - zoomOffset.h;
					dy := dy - zoomOffset.v;
				end;
				DoWGrow;
				zoomed := true;
			end;
			InvalRect(window^.portRect);
			if (zoomOffset.h = 0) & (zoomOffset.v = 0) then
				ValidRect(r);
			currentFrame := frames.stack[frames.stackPointer];
		end;
	end;

procedure DoWUpdate;
	begin
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' Update');
{$ENDC }
		redraw(actualWindow);
	end;

procedure activate (w: windowIndex; active: boolean);
	begin
		with wStatus^[w] do begin
			if active then begin
				if lastActiveWindow <> w then begin
					selectedType := lastObjectType;
					selectedItem := selectItem;
					lastActiveWindow := w;
				end;
				buildOperationsMenu(names);
				initPalette;
				bigClip(window);
				ShowControl(verScroll);
				ShowControl(horScroll);
				littleClip(window);
				currentFrame := frames.stack[frames.stackPointer];
				if w <> undoWindow then
					forgetUndo;
				undoWindow := w;
			end
			else begin
				buildOperationsMenu(nil);
				deactivatePalette;
				bigClip(window);
				HideControl(verScroll);
				HideControl(horScroll);
				littleClip(window);
			end;
			drawGrow(window);
		end;
	end;

procedure DoWActivate (active: boolean);
	begin
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' Activate(', active, ')');
{$ENDC }
		activate(actualWindow, active);
	end;

procedure longSave (w: windowIndex; fileName: Str255; volume: integer);
	var
		i: frameIndex;
	begin
		beginCursor(stopwatchCursor);
		openWriteStream(fileName, volume);
		with wStatus^[w] do begin
			writeLongInt(thisVersion);
			writeRect(getPortRect(window));
			with frames do begin
				writeInteger(stackPointer);
				for i := 1 to stackPointer do
					writeframe(stack[i]);
			end;
			desktop.fileOutput;
			selection.fileOutput;
			modified := false;
		end;
		closeWriteStream;
		endCursor;
	end;

procedure save (w: windowIndex);
	begin
		with wStatus^[w] do
			longSave(w, fileName, volume);
	end;

function saveAs (w: windowIndex): boolean;
	var
		newName: Str255;
	begin
		newName := NewFileName('Save As', wStatus^[w].name);
		if newName = '' then
			saveAs := false
		else begin
			saveAs := true;
			setNewName(newName, w);
			save(w);
		end;
	end;

procedure DoWSaveACopyAs (w: windowIndex);
	var
		newName: Str255;
	begin
		newName := NewFileName('Save a Copy As', wStatus^[w].name);
		if newName <> '' then
			longSave(w, newName, 0);
	end;

function closeGeoWindow (w: windowIndex; action: Str255): boolean;
	var
		done: boolean;
		i: integer;
	begin
		done := true;
		with wStatus^[w] do begin
			if modified & not shiftKeyPressed & ((desktop.length > 0) | (selection.length > 0)) then begin
				ParamText(name, action, '', '');
				case Alert(closeAlertResource, nil) of
					yesItem: 
						if systemName then
							done := saveAs(w)
						else
							save(w);
					noItem: 
						;
					cancelItem: 
						done := false;
				end;
			end;
			if done then begin
{$IFC ScriptEnabled }
				if script <> nil then begin
					script.removeScriptWin;
					script.free;
					script := nil;
				end;
{$ENDC }
				forgetUndo;
				SkelRmveWind(window);
				if operationsMenu <> nil then begin
					SkelRmveMenu(operationsMenu, true);
					operationsMenu := nil;
				end;
				beginCursor(stopwatchCursor);
				desktop.freeAll;
				selection.freeAll;
				types.freeAll;
				names.freeAll;
				window := nil;
				name := windowName(w);
{$IFC THINK_PASCAL }
				SetItem(windowsMenu, w + lastWindowItem + 1, name);
{$ELSEC }
				SetMenuItemText(windowsMenu, w + lastWindowItem + 1, name);
{$ENDC }
				windowCount := windowCount - 1;
				if windowCount = 0 then begin
					selectedType := lastObjectType;
					selectedItem := selectItem;
				end;
				endCursor;
				with frames do
					for i := 1 to stackPointer do
						deleteframe(stack[i]);
			end;
		end;
		closeGeoWindow := done;
		writeTiming;
	end;

function closeAllWindows (message: Str255): boolean;
	var
		continue, dummy: boolean;
		i: windowIndex0;
	begin
		i := currentWindow;
		continue := (i = 0) | closeGeoWindow(i, message);
		if continue then begin
			for i := 1 to maxWindow do
				with wStatus^[i] do
					if (window <> nil) & (not modified | (desktop.length = 0) & (selection.length = 0)) then
						dummy := closeGeoWindow(i, message);
			i := maxWindow;
			while continue & (i > 0) do begin
				continue := (wStatus^[i].window = nil) | closeGeoWindow(i, message);
				i := i - 1;
			end;
		end;
		closeAllWindows := continue;
	end;

procedure DoWClose;
	var
		dummy: boolean;
	begin
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' Close');
{$ENDC }
		if FrontWindow = infoWindow then begin
			HideWindow(infoWindow);
			if optionKeyPressed then
				dummy := closeAllWindows('closing');
		end
		else if optionKeyPressed then
			dummy := closeAllWindows('closing')
		else
			dummy := closeGeoWindow(currentWindow, 'closing');
	end;

function saveOne (w: windowIndex): boolean;
	begin
		if wStatus^[w].systemName then
			saveOne := saveAs(w)
		else begin
			save(w);
			saveOne := true;
		end;
	end;

procedure DoWSave;
	var
		w: windowIndex0;
		continue: boolean;
	begin
		if optionKeyPressed then begin
			for w := 1 to maxWindow do
				with wStatus^[w] do
					if (window <> nil) & not systemName then
						save(w);
			continue := true;
			w := maxWindow;
			while continue & (w > 0) do begin
				with wStatus^[w] do
					if (window <> nil) & systemName then begin
						SelectWindow(window);
						continue := saveAs(w);
					end;
				w := w - 1;
			end;
		end
		else begin
			w := currentWindow;
			if wStatus^[w].systemName then
				continue := saveAs(w)
			else
				save(w);
		end;
	end;

function niceRect (addToTheRight: boolean): Rect;
	var
		r, screen: Rect;
		w: WindowPtr;
		d: boolean;
		dx: integer;
	begin
		w := FrontWindow;
		if w = nil then
			SetRect(r, standardXOffset, standardYOffset, standardWidth + standardXOffset, standardHeight + standardYOffset)
		else begin
			r := WindowPeek(w)^.contRgn^^.rgnBBox;
			dx := r.right - r.left + 2;
{$IFC THINK_PASCAL }
			screen := screenBits.bounds;
{$ELSEC }
			screen := qd.screenBits.bounds;
{$ENDC }
			with screen do begin
				right := right - 1;
				bottom := bottom - 1;
			end;
			if addToTheRight & (r.left + dx < screen.right - minWindowWidth) then
				OffsetRect(r, dx, 0)
			else if (currentWindow = 0) | (WindowPeek(w)^.windowKind < 0) | (r.left > screen.right - minWindowWidth - standardDx) | (r.top > screen.bottom - minWindowHeight - standardDy) then
				SetRect(r, standardXOffset, standardYOffset, standardWidth + standardXOffset, standardHeight + standardYOffset)
			else
				OffsetRect(r, standardDx, standardDy);
			d := SectRect(r, screen, r);
		end;
		niceRect := r;
	end;

function makeWindow (r: Rect; oldName: Str255; sysName, modify: boolean; v: integer): windowIndex;
	var
		s: Rect;
		dummy: boolean;
		safe: GrafPtr;
		windowId: windowIndex;
		oldWindow: windowIndex0;
		i: frameIndex;
	begin
		sheetExtendProc := @ExtendSheet;
		oldWindow := currentWindow;
		windowCount := windowCount + 1;
		windowId := 1;
		while wStatus^[windowId].window <> nil do
			windowId := windowId + 1;
		with wStatus^[windowId] do begin
{$IFC ScriptEnabled }
			script := nil;
{$ENDC }
			GetPort(safe);
			if oldName <> '' then begin
				if modify then
					if sysName & (oldName[2] <> 'e') then
						oldName := concat('W', stringOf(windowId : 1), copy(oldName, pos(' ', oldName), length(oldName)))
					else
						oldName := concat('W', stringOf(windowId : 1), ' ', oldName);
				name := oldName;
				fileName := name;
			end;
			volume := v;
			if colorAvailable then
				window := NewCWindow(nil, r, name, true, zoomDocProc, WindowPtr(-1), true, windowId)
			else
				window := NewWindow(nil, r, name, true, zoomDocProc, WindowPtr(-1), true, windowId);
			failNil(Handle(window));
			if processInBackground then
				previousFrontWindow := window;
			dummy := SkelWindow(window, @DoWMouse, @DoWKey, @DoWUpdate, @DoWGrow, @DoWActivate, @DoWClose, @DoWClobber, @idleTask, false);
			SkelSetWindowZoom(window, @DoWZoom);
			zoomed := false;
{$IFC THINK_PASCAL }
			with screenBits.bounds do
{$ELSEC }
			with qd.screenBits.bounds do
{$ENDC }
				SkelGrowBounds(window, minWindowWidth, minWindowHeight, right - left, bottom - top - standardYOffset);
			SetOrigin(-paletteWidth, 0);
			defaultModes(geoTextSize);
			r := window^.portRect;
			s.left := r.right - scrollBarWidth;
			s.bottom := r.bottom - scrollBarWidth + 1;
			s.top := r.top - 1;
			s.right := r.right + 1;
			verScroll := NewControl(window, s, '', false, r.top, r.top, r.bottom - scrollBarWidth, scrollBarProc, 0);
{$IFC THINK_PASCAL }
			SetCtlMin(verScroll, 0);
			SetCtlMax(verScroll, maxScrollValue);
{$ELSEC }
			SetControlMinimum(verScroll, 0);
			SetControlMaximum(verScroll, maxScrollValue);
{$ENDC }
			s.left := r.left + paletteWidth - 1;
			s.top := r.bottom - scrollBarWidth;
			s.bottom := r.bottom + 1;
			s.right := r.right - scrollBarWidth + 1;
			horScroll := NewControl(window, s, '', false, r.left + paletteWidth, r.left + paletteWidth, r.right - scrollBarWidth, scrollBarProc, 0);
{$IFC THINK_PASCAL }
			SetCtlMin(horScroll, 0);
			SetCtlMax(horScroll, maxScrollValue);
{$ELSEC }
			SetControlMinimum(horScroll, 0);
			SetControlMaximum(horScroll, maxScrollValue);
{$ENDC }
			SetPort(safe);
			desktop := newEmptyList;
			selection := newEmptyList;
			names := newEmptyList;
			types := newEmptyList;
			modified := false;
			systemName := true;
			sheet := drawingRect(window);
			sheetExtended := true;
			testObject(selection, names, types);
			if oldWindow = 0 then
				with frames do begin
					stackPointer := 1;
					copyFrame(standardFrame, stack[stackPointer]);
				end
			else begin

				frames := wStatus^[oldWindow].frames;
				for i := 1 to wStatus^[oldWindow].frames.stackPointer do
					copyFrame(wStatus^[oldWindow].frames.stack[i], frames.stack[i]);

			end;
		end;
		makeWindow := windowId;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoOperationsMenu (menu, item: integer; when: longInt);
	var
		w, wResult: windowIndex;
		i: frameIndex;
		saveAnimationFlag: animationVector;
		r: rectangle;
		currentResultHandlingMode: resultHandlingModeType;
		resultRect: Rect;
		executeMenuEntryElement: menuEntryElement;
		executeObjectIndex: integer;
		savePort: GrafPtr;
		saveFrame: wFrame;
		realStartTime: longInt;

	procedure redrawIfNecessary;
		var
			savePort: GrafPtr;
		begin
			if animationIsOn | redrawScreen then begin
				GetPort(savePort);
				with wStatus^[w] do begin
					SetPort(window);
					redraw(w);
					ValidRect(window^.portRect);
				end;
				SetPort(savePort);
			end;
		end;

	begin
		beginCursor(stopwatchCursor);
		w := currentWindow;
		with wStatus^[w] do begin
			beginAnimation;
			displayInfoString('');
			currentArguments := selection;
			currentResult := newEmptyList;
			abortFlag := false;
			initMemoryMeter;
			saveAnimationFlag := animationFlag;
			animationIsOn := animationEnabled;
			currentPointType := arithmeticsToPointTypeRange(currentArithmetic);
			currentRegularNGonAngle := defaultRegularNGonAngle * oneDegree;
			currentAlignmentIsOn := defaultAlignmentIsOn;
{$IFC THINK_PASCAL }
			currentDimension := trunc(defaultDimension);
{$ELSEC }
			currentDimension := truncLong(defaultDimension);
{$ENDC }
			r := drawingRectangle(window);
			currentRandomConstraint := r;
			doBeep := true;
			rebuildOperationsMenu := false;
			redrawScreen := false;
			currentGeoBenchWindow := window;
			additionalDraw := nil;
			additionalHighlight := nil;
			forgetUndo;
			currentResultHandlingMode := resultHandlingMode;
			if resultHandlingModifiersAllowed then
				if optionKeyPressed & shiftKeyPressed then
					if (resultHandlingMode = createNewAdjacentWindowMode) & (windowCount < maxWindow) then
						currentResultHandlingMode := createNewWindowMode
					else
						currentResultHandlingMode := createNewAdjacentWindowMode
				else if optionKeyPressed then (* & not shiftKeyPressed *)
					if (resultHandlingMode = replaceSelectionMode) & (windowCount < maxWindow) then
						currentResultHandlingMode := createNewWindowMode
					else
						currentResultHandlingMode := replaceSelectionMode
				else if shiftKeyPressed then (* & not optionKeyPressed *)
					if (resultHandlingMode = addResultMode) & (windowCount < maxWindow) then
						currentResultHandlingMode := createNewWindowMode
					else
						currentResultHandlingMode := addResultMode;
			resultRect := niceRect(currentResultHandlingMode = createNewAdjacentWindowMode);
			if menu = operationsMenuID then begin
				executeMenuEntryElement := menuEntryElement(names.ith(item));
				executeObjectIndex := item;
				currentOperation := executeMenuEntryElement.entry
			end
			else
				with subMenus[menu] do begin
					executeMenuEntryElement := menuEntryElement(entries.ith(item));
					executeObjectIndex := mainMenuEntry;
					currentOperation := concat(subMenuElement(names.ith(mainMenuEntry)).title, ': ', executeMenuEntryElement.entry);
				end;
			geoBenchIsComputing := true;
			initTiming(realStartTime);
			prototypes[objectType(int(types.ith(executeObjectIndex)).value)].execute(executeMenuEntryElement.item);
			finishTiming(realStartTime);
			SetPort(FrontWindow);
			geoBenchIsComputing := false;
			additionalDraw := nil;
			additionalHighlight := nil;
			writeTiming;
			r.free;
			animationFlag := saveAnimationFlag;
			if abortFlag then
				if infoString = '' then
					displayInfoString('Operation aborted.')
				else
					displayInfoString(concat('Operation aborted. / ', infoString));
			modified := true;
			if (waitLimit > 0) & doBeep & not animationIsOn & not abortFlag & (cpuTimeUsedUpper > waitLimit) & not processInBackground then
				SysBeep(1);
			if animationIsOn then
				endAnimation;
			if rebuildOperationsMenu then begin
				testObject(selection, names, types);
				if window = FrontWindow then
					buildOperationsMenu(names);
			end;
		end;
		if currentResult <> nil then
			if (currentResult.length > 0) & not abortFlag then
				case currentResultHandlingMode of
					addResultMode:  begin
						redrawIfNecessary;
						with wStatus^[w] do begin
							GetPort(savePort);
							saveFrame := currentFrame;
							SetPort(window);
							currentFrame := frames.stack[frames.stackPointer];
							selection.interactiveHighlight;
							desktop.concatenateDispose(selection);
							selection := currentResult;
							selection.interactiveOutput;
							selection.interactiveHighlight;
							testObject(selection, names, types);
							SetPort(savePort);
							currentFrame := saveFrame;
							if window = FrontWindow then
								buildOperationsMenu(names);
							modified := true;
						end;
					end;
					replaceSelectionMode:  begin
						redrawIfNecessary;
						with wStatus^[w] do begin
							GetPort(savePort);
							saveFrame := currentFrame;
							SetPort(window);
							currentFrame := frames.stack[frames.stackPointer];
							selection.interactiveHighlight;
							selection.interactiveOutput;
							undoScrap := selection;
							undoCmd := replaceSelectionCmd;
							undoWindow := w;
							if not undoEnabled then
								forgetUndo;
							selection := currentResult;
							selection.interactiveOutput;
							selection.interactiveHighlight;
							testObject(selection, names, types);
							SetPort(savePort);
							currentFrame := saveFrame;
							if window = FrontWindow then
								buildOperationsMenu(names);
							modified := true;
						end;
					end;
					throwAwayResultMode:  begin
						currentResult.freeAll;
						redrawIfNecessary;
					end;
					createNewWindowMode, createNewAdjacentWindowMode:  begin
						if currentResultHandlingMode = createNewAdjacentWindowMode then
							redrawIfNecessary;
						with wStatus^[w] do
							wResult := makeWindow(resultRect, name, systemName, true, 0);
						with wStatus^[wResult] do begin
							for i := 1 to frames.stackPointer do
								deleteframe(frames.stack[i]);
							frames := wStatus^[w].frames;
							for i := 1 to wStatus^[w].frames.stackPointer do
								if wStatus^[w].frames.stack[i].proD <> nil then begin
									copyFrame(wStatus^[w].frames.stack[i], frames.stack[i]);
									if currentFrame.proD^.visibility then
										frames.stack[i].proD^.visibility := true;
								end;
							currentFrame := frames.stack[frames.stackPointer];
							selection.freeAll;
							selection := currentResult;
							testObject(selection, names, types);
							buildOperationsMenu(names);
							modified := true;
{$IFC THINK_PASCAL }
							SetItem(windowsMenu, wResult + lastWindowItem + 1, name);
{$ELSEC }
							SetMenuItemText(windowsMenu, wResult + lastWindowItem + 1, name);
{$ENDC }
						end; (* with wStatus^[wResult] do begin *)
					end;
				end (* case *)
			else begin
				currentResult.freeAll;
				redrawIfNecessary;
			end
		else
			redrawIfNecessary;
		endCursor;
		wakeupUser;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure buildOperationsMenu (names: list);
	var
		dummy: boolean;
		mainMenuIndex: integer;

	procedure buildAMenu (theMenu: MenuHandle; l: list);
		var
			e: elementH;
			subMenu: MenuHandle;
			el: subMenuElement;
		begin
			e := l.first;
			while e <> nil do begin
				if member(e^^.value, subMenuElement) then begin
					el := subMenuElement(e^^.value);
					lastSubMenuID := lastSubMenuID + 1;
					assert(lastSubMenuID <= maxSubMenuID, 'lastSubMenuID <= maxSubMenuID');
					AppendMenu(theMenu, stringOf(el.title, '/', escChar, '!', chr(lastSubMenuID)));
					if MemError <> noErr then
						failNil(nil);
					subMenu := NewMenu(lastSubMenuID, '');
					failNil(Handle(subMenu));
					with subMenus[lastSubMenuID] do begin
						entries := el.menu;
						mainMenuEntry := mainMenuIndex;
					end;
					buildAMenu(subMenu, el.menu);
					dummy := SkelMenu(subMenu, @DoOperationsMenu, @DoMClobber, false, true);
				end
				else begin
					AppendMenu(theMenu, menuEntryElement(e^^.value).entry);
					if MemError <> noErr then
						failNil(nil);
				end;
				if theMenu = operationsMenu then
					mainMenuIndex := mainMenuIndex + 1;
				e := e^^.next;
			end;
		end;

	begin  (* buildOperationsMenu *)
{$IFC TraceEvents AND NOT XYZApplication }
		lineCount := lineCount + 1;
		writeLn(lineCount : 1, ' BuildOperationsMenu(', (names <> nil) & (names.length > 0), ')');
{$ENDC }
		while lastSubMenuID >= firstSubMenuID do begin
{$IFC THINK_PASCAL }
			SkelRmveMenu(GetMHandle(lastSubMenuID), false);
{$ELSEC }
			SkelRmveMenu(GetMenuHandle(lastSubMenuID), false);
{$ENDC }
			lastSubMenuID := lastSubMenuID - 1;
		end;
		if operationsMenu <> nil then
			SkelRmveMenu(operationsMenu, true);
		if (names = nil) | (names.length = 0) then
			operationsMenu := nil
		else begin
			operationsMenu := NewMenu(operationsMenuID, 'Operations');
			failNil(Handle(operationsMenu));
			mainMenuIndex := 1;
			buildAMenu(operationsMenu, names);
			enableDisableOperationsMenu(false);
			dummy := SkelMenu(operationsMenu, @DoOperationsMenu, @DoMClobber, true, false);
		end;
		writeTiming;
	end;

procedure readFile (newName: Str255; volume: integer; w: windowIndex0);
	var
		r: Rect;
		t: integer;
		i: frameIndex;
		d, createWindow: boolean;
	begin
		readRect(r);
		createWindow := w = 0;
		if createWindow then begin
{$IFC THINK_PASCAL }
			if (r.left > screenBits.bounds.right - minWindowWidth) | (r.top > screenBits.bounds.bottom - minWindowHeight) then
				r := niceRect(false);
			d := SectRect(r, screenBits.bounds, r);
{$ELSEC }
			if (r.left > qd.screenBits.bounds.right - minWindowWidth) | (r.top > qd.screenBits.bounds.bottom - minWindowHeight) then
				r := niceRect(false);
			d := SectRect(r, qd.screenBits.bounds, r);
{$ENDC }
			w := makeWindow(r, GetFileName(newName), false, false, volume);
		end;
		with wStatus^[w] do begin
			with frames do begin
				for i := 1 to stackPointer do
					deleteframe(stack[i]);
				if currentVersion = version1 then begin
					stackPointer := 1;
					copyFrame(standardFrame, stack[1]);
				end
				else if currentVersion = version2 then begin
					readInteger(t);
					stackPointer := t;
					for i := 1 to stackPointer do
						oldReadFrame(stack[i]);
					stackPointer := 1;
					copyFrame(standardFrame, stack[1]);
				end
				else if (currentVersion = version3) | (currentVersion = version4) then begin
					readInteger(t);
					stackPointer := t;
					for i := 2 to stackPointer do
						oldReadFrame(stack[i]);
					copyFrame(standardFrame, stack[1]);
				end
				else if currentVersion = version5 then begin
					readInteger(t);
					stackPointer := t;
					for i := 1 to stackPointer do
						oldReadFrame(stack[i]);
				end
				else begin (*currentVersion = version6 | currentVersion = version7 | currentVersion = version8 | currentVersion = version9 | currentVersion = version10 | currentVersion = version10 *)
					readInteger(t);
					stackPointer := t;
					for i := 1 to stackPointer do
						readFrame(stack[i]);
				end;
			end;
			desktop.freeAll;
			selection.freeAll;
			new(desktop);
			failNil(Handle(desktop));
			desktop.fileInput;
			new(selection);
			failNil(Handle(selection));
			selection.fileInput;
			testObject(selection, names, types);
			modified := false;
			if createWindow then begin
				SetPort(window);
{$IFC THINK_PASCAL }
				SetItem(windowsMenu, w + lastWindowItem + 1, name);
{$ELSEC }
				SetMenuItemText(windowsMenu, w + lastWindowItem + 1, name);
{$ENDC }
				systemName := false;
				fileName := newName;
			end;
		end;
	end;

procedure inputFile (newName: Str255; volume: integer);
	begin
		beginCursor(stopwatchCursor);
		openReadStream(newName, volume);
		readLongInt(currentVersion);
		if goodVersion then
			readFile(newName, volume, 0)
		else
			wrongVersion;
		closeReadStream;
		endCursor;
	end;

function revert (w: windowIndex): boolean;
	const
		revertResource = 306;
	var
		safe: GrafPtr;
	begin
		GetPort(safe);
		with wStatus^[w] do begin
			ParamText(name, '', '', '');
			if shiftKeyPressed | (Alert(revertResource, nil) = okItem) then begin
				beginCursor(stopwatchCursor);
				openReadStream(fileName, volume);
				readLongInt(currentVersion);
				if goodVersion then
					readFile('', 0, w)
				else
					wrongVersion;
				closeReadStream;
				endCursor;
				SetPort(window);
				redraw(w);
				revert := true;
			end
			else
				revert := false;
		end;
		SetPort(safe);
	end;

procedure DoWRevert;
	var
		w: windowIndex;
		continue: boolean;
	begin
		if optionKeyPressed then begin
			for w := 1 to maxWindow do
				with wStatus^[w] do
					if (window <> nil) & modified & not revert(w) then
						leave;
		end
		else
			continue := revert(currentWindow);
	end;

function filter (pb: ParmBlkPtr): boolean;
	begin
		with pb^.ioFlFndrInfo do
			filter := not ((fdCreator = gbCreator) & (fdType = gbFileType) | (fdCreator = 'GBXY') & (fdType = 'DATA'));
	end;

function gbOldFileName (var shortName: Str255): Str255;
	var
		reply: SFReply;
		typeList: SFTypeList;
		p: Point;
	begin
		p.v := 40;
		p.h := 20;
		typeList[0] := 'DATA';
		typeList[1] := gbFileType;
{$IFC THINK_PASCAL }
		SFGetFile(p, '', @filter, 2, typeList, nil, reply);
{$ELSEC }
		SFGetFile(p, '', NewFileFilterProc(ProcPtr(@filter)), 2, @typeList, nil, reply);
{$ENDC }
		if reply.good then begin
			shortName := reply.fName;
			gbOldFileName := CONCAT(PathNameFromWD(reply.vRefNum), reply.fName);
		end
		else
			gbOldFileName := '';
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoFileMenu (menu, item: integer; when: longInt);
	var
		dummy: boolean;
		w: windowIndex;

	procedure restore;
		var
			newName, shortName: Str255;
		begin
			newName := gbOldFileName(shortName);
			if geoBenchIsComputing then
				nonFatalError(StringOf('Cannot open ''', shortName, ''' while computing.'))
			else if newName <> '' then
				inputFile(newName, 0);
		end;

{$IFC ScriptEnabled }
	procedure writeScript;
		var
			newName: Str255;
			vRefNum: integer;
			override: boolean;
			f: integer;
			count: longint;
			result, ignore: OSErr;
			hText: Handle;
			editTE: TEHandle;
		begin
			if SELECTnewFile('Write Text', '', newName, vRefNum, override) then begin
				result := FSOpen(newName, vRefNum, f);
				if result = fnfErr then begin
					result := Create(newName, vRefNum, 'MSWD', 'TEXT');
					if result = noErr then
						result := FSOpen(newName, vRefNum, f);
				end;
				if result <> noErr then
					nonFatalError('Can''t Create file')
				else begin
					beginCursor(stopwatchCursor);
					ignore := SetFPos(f, fsFromStart, longint(0));
					editTE := GetEWindowTE(wStatus^[-currentScriptWindow].script.stk.editWind);
					hText := editTE^^.hText;
					HLock(hText);
					count := editTE^^.selEnd - editTE^^.selStart;
					result := FSWrite(f, count, pointer(ord(hText^) + editTE^^.selStart));
					ignore := GetFPos(f, count);
					ignore := SetEOF(f, count);
					ignore := FSClose(f);
					setFileInfo(newName, vRefNum, 'TEXT', possibleCreatorTypes[textCreatorType]);
					ignore := FlushVol(nil, vRefNum);
					HUnlock(hText);
					endCursor;
					if result <> noErr then
						nonFatalError('Write error!');
				end;
			end;
		end; (* writeScript *)

	procedure readScript;
		var
			fileName: Str255;
			f, vRefNum, oldSelStart: integer;
			len: longint;
			h: Handle;
			editTE: TEHandle;
			ignore: OSErr;
		begin
			if SELECTfile(fileName, vRefNum, 'TEXT') then begin
				if FSOpen(fileName, vRefNum, f) <> noErr then
					nonFatalError('Couldn''t open file')
				else begin
					ignore := GetEOF(f, len);
					if len >= 32000 then
						nonFatalError('File is too big')
					else begin
						h := newHandleFailNil(len);
						HLock(h);
						ignore := FSRead(f, len, h^);
						HUnlock(h);
						editTE := GetEWindowTE(wStatus^[-currentScriptWindow].script.stk.editWind);
						oldSelStart := editTE^^.selStart;
						TEInsert(h^, len, editTE);
{$IFC THINK_PASCAL }
						DisposHandle(h);
{$ELSEC }
						DisposeHandle(h);
{$ENDC }
						TESetSelect(oldSelStart, oldSelStart + len, editTE);
					end;
					ignore := FSClose(f);
				end;
			end;
		end; (* readScript *)
{$ENDC }

	procedure writeSelectionAsText (l: list);
		var
			newName: Str255;
			vRefNum: integer;
			override: boolean;
		begin
			if SELECTnewFile('Write Text', '', newName, vRefNum, override) then begin
				openWriteStreamText(newName, vRefNum);
				beginCursor(stopwatchCursor);
{$IFC  DebugWriteText }
				currentTextFileOutputMethod := textFileOutputMethod;
{$ELSEC }
				currentTextFileOutputMethod := 0;
{$ENDC }
				writeTextObjNoCR(l);
				closeWriteStream;
				endCursor;
			end;
		end;

	procedure appendFromText (w: windowIndex);
		var
			o: obj;
			newName: Str255;
			vRefNum: integer;
		begin
			if SELECTfile(newName, vRefNum, 'TEXT') then begin
				openReadStream(newName, 0);
				beginCursor(stopwatchCursor);
				o := readTextObjSpecial;
				endCursor;
				closeReadStream;
				o.interactiveOutput;
				o.interactiveHighlight;
				with wStatus^[w] do begin
					if systemName & ((selection = nil) | (selection.length = 0)) then begin
						setNewName(newName, w);
						volume := vRefNum;
					end;
					if member(o, list) then
						if selection = nil then
							selection := list(o)
						else
							selection.concatenateDispose(list(o))
					else
						selection.append(o);
					testObject(selection, names, types);
				end;
			end;
		end;

	begin (* doFileMenu *)
		case item of
			newItem: 
				w := makeWindow(niceRect(false), '', false, false, 0);
			openItem: 
				restore;
			closeItem: 
				DoWClose;
			saveItem: 
				DoWSave;
			saveACopyAsItem: 
				DoWSaveACopyAs(currentWindow);
			revertItem: 
				DoWRevert;
			saveAsItem: 
				dummy := saveAs(currentWindow);
			writeTextItem: 
{$IFC ScriptEnabled }
				if (currentScriptWindow < 0) & (item = writeTextItem) then
					writeScript
				else
{$ENDC }
					writeSelectionAsText(wStatus^[currentWindow].selection);
			readTextItem: 
{$IFC ScriptEnabled }
				if currentScriptWindow < 0 then
					readScript
				else
{$ENDC }
					appendFromText(currentWindow);
			preferencesItem: 
				preferencesMenuItem;
			quitItem: 
				if closeAllWindows('quitting') then
					SkelWhoa;	(* Tell SkelMain to quit *)
		end;
	end;
{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }

{$IFC Topography }
procedure DoTopoMenu (menu, item: integer; when: longInt);

	procedure readPolygons;
		var
			w: windowIndex;
			l: list;
		begin
			l := newEmptyList;
			readSwissTopoPolygons(l);
			if l.length > 0 then begin
				w := makeWindow(niceRect(false), '', false, false, 0);
				with wStatus^[w] do begin
					desktop.freeAll;
					desktop := l;
					forgetUndo;
					modified := true;
				end;
			end
			else
				l.free;
		end;

	procedure readTriang;
		var
			w: windowIndex;
			l: list;
			t: triangulation3d;
		begin
			t := readSwissTriang;
			if t <> nil then begin
				w := makeWindow(niceRect(false), '', false, false, 0);
				l := newEmptyList;
				l.append(t);
				with wStatus^[w] do begin
					desktop.freeAll;
					desktop := l;
					forgetUndo;
					modified := true;
				end;
			end;
		end;

		var
			p: obj;

	begin
		if item = readPolygonsItem then
			readPolygons
		else if item = readTriangItem then
			readTriang
		else begin
			if item = bwMapItem then
				p := readBlackWhiteMap
			else
				p := readColorMap(item = grayScaleMapItem);
			if p <> nil then begin
				forgetUndo;
				wStatus^[currentWindow].selection.append(p);
				p.interactiveOutput;
				p.interactiveHighlight;
			end;
		end;
	end;
{$ENDC }

{$IFC ScriptEnabled AND NetworkEnabled }
procedure DoCosimaMenu (menu, item: integer; when: longInt);						(* MS: 19.7.93 - Cosima Menu Handler *)
	var
		aScript: geoScript;
		done: boolean;
	const
		openAsClient = 1;
		openAsServer = 2;
	begin
		case item of
			openAsServer:  begin
				if wStatus^[currentWindow].script = nil then begin
					new(aScript);
					failNil(Handle(aScript));
					aScript.init;
					wStatus^[currentWindow].script := aScript;
					aScript.createScriptWin(wStatus^[currentWindow].window, currentWindow, 'XYZ Geo Server Log-Window', Server, done);
					if not done then begin
						aScript.removeScriptWin;				(* MS: 27.7.93 - if createScriptWin not succesful close the window *)
						aScript.free;
						wStatus^[currentWindow].script := nil;
					end;
				end
			end;
			openAsClient:  begin
				new(aScript);
				failNil(Handle(aScript));
				aScript.init;
				wStatus^[currentWindow].script := aScript;
				aScript.createScriptWin(wStatus^[currentWindow].window, currentWindow, StringOf(wStatus^[currentWindow].name, '- XYZ Geo Client'), Client, done);
				if not done then begin
					aScript.removeScriptWin;				(* MS: 27.7.93 - if createScriptWin not succesful close the window *)
					aScript.free;
					wStatus^[currentWindow].script := nil;
				end;
			end;
		end;
	end;
{$ENDC }

end.
