(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

program geoBenchMain;

uses
	geoBenchFileMenu,
	geoBenchEditMenu,
	nearestNeighborInSector,
	dDimPointVector,
	upperMonotonePolygon,
	markedRealPoint,
	vector,
	infiniteUndirectedGraph,
	getTriangulation,
	generalExperiment,
	graph2,
	Fonts,
	GeoBenchAppleEvents,
{$IFC GFEnabled }
	pPoint, 
{$ENDC }
{$IFC Chemistry }
	moleculeDraw, moleculeIO, 
{$IFC ChemistryTriangulation }
	embed3dNew, 
{$ENDC }
{$ENDC }
{$IFC Topography }
	swissTopo, 
{$ENDC }
{$IFC RaumPlanung }
	raumPlanung, 
{$ENDC }
{$IFC ScriptEnabled }
	GeoScript, GeoScriptBrowser, 
{$ENDC }
	triangulation3d,
	polygonAreaExperiment,
	lineSegment;

const
	infoWindowTextSize = 12;

{$IFC ScriptEnabled }
procedure TextClose;
	var
		w: integer;
	begin
		w := currentScriptWindow;
		if w >= 0 then
			fatalError('Reference of window is > 0');
		wStatus^[-w].script.removeScriptWin;
		wStatus^[-w].script.free;
		wStatus^[-w].script := nil;
	end; (* TextClose *)

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure geoScriptKeyHandle (ch: char; mods: integer);   (* MS: 3.5.93 *)
	(* if the <enter> key is pressed act as the <opt>-<=> key [execute] is pressed *)
	const
		Enter = 3;   (* enter key *)
	var
		te: TEHandle;
		w: integer;
		i: integer;
	begin
		if ord(ch) = Enter then begin
			w := currentScriptWindow;
			te := GetEWindowTE(wStatus^[-w].script.stk.editWind);
			if te^^.selStart = te^^.selEnd then begin  (* no text is selected - current line is cmd line *)
				with te^^ do begin
					i := nLines - 1;
					while (i >= 0) and (lineStarts[i] > selStart) do
						i := i - 1;
					if i >= 0 then begin
						TESetSelect(lineStarts[i], selEnd, te);
					end;
				end;
			end;
			wStatus^[-w].script.executeSelection;
		end;
	end; (* geoScripKeyHandle  *)
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure geoScriptActivate (active: Boolean);   (* MS: 21.5.93 *)
	begin
		if active then begin
			wStatus^[-currentScriptWindow].script.stk.selectedList := wStatus^[-currentScriptWindow].selection;
		end;
	end; (* geoScriptActivate *)
{$ENDC }

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoWindowsMenu (menu, item: integer; when: longInt);
	var
		i: windowIndex;
		w, last: WindowPtr;
{$IFC ScriptEnabled }
		aScript: geoScript;
		aWin: WindowPtr;
		done: boolean;				(* MS: 27.7.93 *)
{$ENDC }
	begin
		case item of
			hideItem: 
				if optionKeyPressed then begin
					HideWindow(infoWindow);
					if continueButton <> nil then
						HideWindow(continueButton);
					for i := 1 to maxWindow do begin
						w := wStatus^[i].window;
						if w <> nil then
							HideWindow(w)
					end;
				end
				else
					HideWindow(FrontWindow);
			infoItem:  begin
				SelectWindow(infoWindow);
				ShowWindow(infoWindow);
			end;
			AnimationWindowItem:  begin
				SelectWindow(continueButton);
				ShowWindow(continueButton);
			end;
{$IFC ScriptEnabled }
			geoScriptItem:  begin
				if wStatus^[currentWindow].script = nil then begin
					new(aScript);
					failNil(Handle(aScript));
					aScript.init;
					wStatus^[currentWindow].script := aScript;
					aScript.createScriptWin(wStatus^[currentWindow].window, currentWindow, stringOf(wStatus^[currentWindow].name, ' - GeoScript'), Local, done);
				end
				else begin
					aWin := wStatus^[currentWindow].script.stk.editWind;
					SelectWindow(aWin);
					ShowWindow(aWin);
				end;
			end;
{$ENDC }
			ShowAllItem:  begin
				if not WindowPeek(infoWindow)^.visible then
					ShowWindow(infoWindow);
				last := infoWindow;
				for i := 1 to maxWindow do begin
					w := wStatus^[i].window;
					if (w <> nil) & not WindowPeek(w)^.visible then begin
						ShowWindow(w);
						last := w;
					end;
				end;
				SelectWindow(last);
				if (continueButton <> nil) & not WindowPeek(continueButton)^.visible then begin
					ShowWindow(continueButton);
					SelectWindow(continueButton);
				end;
			end;
			otherwise
				with wStatus^[item - lastWindowItem - 1] do begin
					SelectWindow(window);
					ShowWindow(window);
				end;
		end;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function defaultVerticesOk (x: extended): boolean;
	begin
		defaultVerticesOk := (3 <= x) & (x < maxN) & (trunc(x) = x);
	end;

function dimensionOk (x: extended): boolean;
	begin
		dimensionOk := (2 <= x) & (x <= maxDimension) & (trunc(x) = x);
	end;

function verticesOk (x: extended): boolean;
	begin
		verticesOk := (3 <= x) & (x <= maxVertices) & (trunc(x) = x);
	end;

function angleOk (x: extended): boolean;
	begin
		angleOk := (0 <= x) & (x <= 360)
	end;

function angleVisible: boolean;
	begin
		angleVisible := defaultAlignmentIsOn;
	end;

function hasSequence (x: obj): boolean;
	begin
		hasSequence := member(x, sequence);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoObjectsMenu (menu, item: integer; when: longInt);
	var
		s: vector;
		h: homogenVector;
{$IFC ScriptEnabled }
		gsbr_st: gsbr_status;
		newInfo: vector;
{$ENDC }

	procedure randomCreate (l: list; w: WindowPtr; t: objectType);
		var
			p: polyLine;
			x: obj;
			i, realStartTime: longInt;
			r: rectangle;
			oldFrame, thisFrame: wFrame;

		function getRegularNGonParameters2: boolean;
			var
				h: vector;
			begin
				h := newVector(5);
				h.append(getExtendedHandler(4, @defaultRegularNGonAngle, 0, @angleVisible, @angleOk));
				h.append(getBooleanHandler(7, @defaultAlignmentIsOn, false, nil));
				h.append(getStaticTextHandler(6, '', false, @angleVisible));
				h.append(getExtendedHandler(3, @defaultVertices, 0, nil, @verticesOk));
				h.append(getExtendedHandler(10, @defaultNumber, 0, nil, @isSeqIndex));
				getRegularNGonParameters2 := processDialogBox(132, 0, 8, 0, alwaysTrue, h);
			end;

		begin (* randomCreate *)
			oldFrame := currentFrame;
			beginCursor(stopwatchCursor);
			displayInfoString('');
			r := drawingRectangle(w);
			currentRandomConstraint := r;
			geoBenchIsComputing := true;
			if ((t = polygon2dType) | (t = convexPolygonType) | (t = polyLineType)) & getUserParameter2(concat('Number of ', typeToString(t), ' Vertices:'), 'Number of Objects:', defaultPolyVertices, defaultPolyObjects, defaultPolyVertices, defaultPolyObjects, @defaultVerticesOk, @isSeqIndex) then begin
				selectedType := lastObjectType;
				updatePalette(selectItem);
				abortFlag := false;
				initTiming(realStartTime);
{$IFC THINK_PASCAL }
				for i := 1 to trunc(defaultPolyObjects) do begin
{$ELSEC }
				for i := 1 to truncLong(defaultPolyObjects) do begin
{$ENDC }
					if userAbort then
						leave;
					p := polyLine(newObj(t));
					p.elements := nil;
{$IFC THINK_PASCAL }
					p.preset(trunc(defaultPolyVertices), arithmeticsToPointTypeRange(currentArithmetic));
{$ELSEC }
					p.preset(truncLong(defaultPolyVertices), arithmeticsToPointTypeRange(currentArithmetic));
{$ENDC }
					p.randomChange;
					thisFrame := currentFrame;
					currentFrame := oldFrame;
					p.interactiveOutput;
					p.interactiveHighlight;
					currentFrame := thisFrame;
					l.append(p);
				end;
				finishTiming(realStartTime);
				writeTiming;
				SetPort(FrontWindow);
			end
			else if (t = dDimPointType) & getUserParameter2('Number of D-Points', 'Number of Dimensions', defaultNumber, defaultDimension, defaultNumber, defaultDimension, @isSeqIndex, @dimensionOk) | (t = regularNGonType) & getRegularNGonParameters2 | ((t = point2dType) | (t = rectangleType) | (t = lineSegmentType) | (t = circleType)) & getUserParameter1(Concat(' Number of Random ', typeToString(t), 's'), defaultNumber, defaultNumber, @isSeqIndex) then begin
				selectedType := lastObjectType;
				updatePalette(selectItem);
				currentPointType := arithmeticsToPointTypeRange(currentArithmetic);
{$IFC THINK_PASCAL }
				currentDimension := trunc(defaultDimension);
				currentVertices := trunc(defaultVertices);
{$ELSEC }
				currentDimension := truncLong(defaultDimension);
				currentVertices := truncLong(defaultVertices);
{$ENDC }
				currentRegularNGonAngle := defaultRegularNGonAngle * oneDegree;
				currentAlignmentIsOn := defaultAlignmentIsOn;
				if t = point2dType then
					t := currentPointType;
				abortFlag := false;
				initTiming(realStartTime);
{$IFC THINK_PASCAL }
				for i := 1 to trunc(defaultNumber) do begin
{$ELSEC }
				for i := 1 to truncLong(defaultNumber) do begin
{$ENDC }
					if userAbort then
						leave;
					x := newInitObj(t);
					x.randomChange;
					thisFrame := currentFrame;
					currentFrame := oldFrame;
					x.interactiveOutput;
					x.interactiveHighlight;
					currentFrame := thisFrame;
					l.append(x);
				end;
				finishTiming(realStartTime);
				writeTiming;
				SetPort(FrontWindow);
			end;
			geoBenchIsComputing := false;
			r.free;
			endCursor;
		end;

	function getRegularNGonParameters: boolean;
		var
			h: vector;
		begin
			h := newVector(4);
			h.append(getExtendedHandler(4, @defaultRegularNGonAngle, 0, @angleVisible, @angleOk));
			h.append(getBooleanHandler(7, @defaultAlignmentIsOn, false, nil));
			h.append(getStaticTextHandler(6, '', false, @angleVisible));
			h.append(getExtendedHandler(3, @defaultVertices, 0, nil, @verticesOk));
			getRegularNGonParameters := processDialogBox(1010, 0, 8, 0, alwaysTrue, h);
		end;

	procedure randomChange (w: windowIndex);
		var
			r: rectangle;
		begin
			with wStatus^[w] do begin
				r := drawingRectangle(window);
				currentRandomConstraint := r;
				modified := true;
				selection.interactiveHighlight;
				selection.interactiveOutput;
				forgetUndo;
				undoCmd := randomChangeCmd;
				if undoEnabled then
					undoScrap := list(stopwatchDuplicateAll(selection));
				selection.randomChange;
				selection.interactiveOutput;
				selection.interactiveHighlight;
				testObject(selection, names, types);
				buildOperationsMenu(names);
			end;
			r.free;
		end;

	begin
{$IFC ScriptEnabled }
		if item = executeItem then  (* delo new *)
			wStatus^[-currentScriptWindow].script.executeSelection
		else if (item = objInfoItem) and (currentScriptWindow < 0) then begin (* MS: 11.5.93 - activate GeoScript Browser *)
			newInfo := vector(wStatus^[-currentScriptWindow].script.stk.getInfo);
			enterScriptBrowser;
			gsbr_st := scriptBrowser(newInfo, false);
			exitScriptBrowser;
		end
		else
{$ENDC }
			with wStatus^[currentWindow] do begin
				case item of
					firstObjectItem..lastSimpleObjectItem:  begin
						selectedType := itemToType[item];
						updatePalette(item);
					end;
					regularNGonItem: 
						if getRegularNGonParameters then begin
							selectedType := regularNGonType;
							updatePalette(item);
						end;
					dDimPointItem: 
						if getUserParameter1('Number of Dimensions', defaultDimension, defaultDimension, @dimensionOk) then begin
							selectedType := dDimPointType;
							updatePalette(item);
						end;
					selectItem:  begin
						selectedType := lastObjectType;
						updatePalette(selectItem);
					end;
					objInfoItem: 
						doObjectInfo(selection);
					groupItem:  begin
						beginCursor(stopwatchCursor);
						modified := true;
						forgetUndo;
						if selection.sameType then begin
							new(h);
							failNil(Handle(h));
							s := h;
						end
						else begin
							new(s);
							failNil(Handle(s));
						end;
						s.initFromList(selection);
						selection.free;
						selection := newEmptyList;
						selection.append(s);
						testObject(selection, names, types);
						buildOperationsMenu(names);
						endCursor;
					end;
					ungroupItem:  begin
						beginCursor(stopwatchCursor);
						modified := true;
						forgetUndo;
						undoCmd := ungroupCmd;
						if undoEnabled then
							undoScrap := list(selection.duplicateAll);
						selection.interactiveHighlight;
						selection.interactiveOutput;
						repeat
							selection.flatten;
						until not optionKeyPressed | (selection.exists(hasSequence) = 0);
						selection.interactiveOutput;
						selection.interactiveHighlight;
						testObject(selection, names, types);
						buildOperationsMenu(names);
						endCursor;
					end;
					randomChangeItem: 
						randomChange(currentWindow);
					randomItem:  begin
						modified := true;
						forgetUndo;
{$IFC THINK_PASCAL }
						GetItem(objectsMenu, randomItem, currentOperation);
{$ELSEC }
						GetMenuItemText(objectsMenu, randomItem, currentOperation);
{$ENDC }
						currentOperation := copy(currentOperation, 1, length(currentOperation) - 3); (* throw away '...' *)
						randomCreate(selection, window, selectedType);
						wakeupUser;
						testObject(selection, names, types);
						if window = FrontWindow then
							buildOperationsMenu(names);
						zoomed := false;
					end;
				end;
			end;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function isPositiveLongInt (x: extended): boolean;
	begin
		isPositiveLongInt := (trunc(x) = x) & (0 < x) & (x < 2147483647);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoRandomGeneratorMenu (menu, item: integer; when: longInt);
	var
		seed: extended;
	begin
		case item of
			randomizeSeedItem: 
				randomize;
			resetSeedItem: 
				initRandom;
			setSeedItem:  begin
				seed := getRandomSeed;
				if getUserParameter1('Seed (1  Seed  2147483646)', seed, seed, @isPositiveLongInt) then
					setRandomSeed(seed);
			end;
		end;
	end;

procedure DoAnimationMenu (menu, item: integer; when: longInt);
	begin
		if item <= allAnimationsOffItem then
			setAllAnimationFlag(item = allAnimationsOnItem)
		else
			animationFlag[item] := not animationFlag[item];
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function Strip (s: Str255; maxLength: integer): Str255;
	begin
		if length(s) <= maxLength then
			Strip := s
		else
			Strip := Concat(Copy(s, 1, maxLength - 1), '');
	end;

procedure updateCreationTime;
	var
		h: stringH;
	begin
		h := stringH(GetResource('STR ', 19861));
		if h^^ = '08/19/1961 at ??:??:??' then begin
			h^^ := getDateAndTimeString;
{$IFC XYZApplication }
			ChangedResource(Handle(h));
			WriteResource(Handle(h));
{$ENDC }
		end;
	end;

procedure setPriority;
	var
		currentAbortTime: extended;
	begin
		currentAbortTime := getAbortTime;
		if getUserParameter1('Max. Time (Ticks) Bef. Cont. Switch', currentAbortTime, currentAbortTime, @isPositiveOrZeroInteger) then
{$IFC THINK_PASCAL }
			setAbortTime(trunc(currentAbortTime));
{$ELSEC }
			setAbortTime(truncLong(currentAbortTime));
{$ENDC }
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoAbout (menu, item: integer; when: longInt);
	const
		moreId = 2;
		infoId = 6;
		cancelId = 4;
		backId = 4;
		fieldLength = 70;
		aboutAlertResource = 1001;
		moreAlertResource = 1007;
		InfoTimeAlertResource = 1021;
		setPriorityId = 5;
	var
		h: stringH;
		p: Str255;
		userButton: integer;

	begin
		beginCursor(arrowCursor);
		case item of
			aboutItem:  begin
				h := stringH(GetResource('STR ', 19861));
				if colorAvailable then
					p := ''
				else
					p := 'not ';
				repeat
					ParamText(h^^, p, '', '');
					userButton := Alert(aboutAlertResource, nil);
					if userButton = moreId then
						userButton := Alert(moreAlertResource, nil);
				until userButton <> backId;
			end;
			commandInfoItem:  begin
				ParamText(Strip(currentOperation, fieldLength), StringOf(max((when - startTime) / 60, 0) : 7 : 1, ' secs.'), '', '');
				userButton := Alert(InfoTimeAlertResource, nil);
				if userButton = cancelId then
					abortFlag := userAgreesToAbort
				else if userButton = setPriorityId then
					setPriority;
			end;
			otherwise
				assert(false, '1  item  2');
		end;
		endCursor;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure DoDeskAccOpen;
	begin
		saveScrapBook;
	end;

procedure DoResume (resume: boolean);
	var
		i: windowIndex0;
	begin
		processInBackground := not resume;
		if resume then begin
			removeNotification;
			for i := 1 to maxWindow do
				with wStatus^[i] do
					if (window = previousFrontWindow) & (window <> nil) then begin
						SetPort(window);
						activate(i, true);
						exit(DoResume);
					end;
		end
		else begin
			saveScrapBook;
			previousFrontWindow := FrontWindow;
			i := currentWindow;
			if i > 0 then
				with wStatus^[i] do begin
					SetPort(window);
					activate(i, false);
				end;
		end;
	end;

procedure installMenus;
	var
		dummy: boolean;
		i: integer;
		s: Str255;

	procedure AdjustString (var s: Str255; c1, c2, c3, c4: char);
		var
			i: integer;
		begin
			i := 1;
			while s[i] <> c1 do
				i := i + 1;
			s[i] := c2;
			while s[i] <> c3 do
				i := i + 1;
			s[i] := c4;
		end;

	begin
		fileMenu := NewMenu(fileMenuID, 'File');									(* Create Menu *)
		s := fileMenuString;
{$IFC NetworkEnabled }
		AdjustString(s, '', escChar, '', chr(cosimaMenuID));       (* MS: 19.7.93 - Menu for COSIMA server/client *)
{$ENDC }
{$IFC Topography }
		AdjustString(s, '', escChar, '', chr(topoMenuID));
{$ENDC }
		AppendMenu(fileMenu, s);
		dummy := SkelMenu(fileMenu, @DoFileMenu, @DoMClobber, false, false);		(* Tell Transkel to handle it *)

		editMenu := NewMenu(editMenuID, 'Edit');
		s := editMenuString;
		AdjustString(s, '', escChar, '', chr(transMenuID));
		AppendMenu(editMenu, s);
		dummy := SkelMenu(editMenu, @DoEditMenu, @DoMClobber, false, false);

		transMenu := NewMenu(transMenuID, '');
		AppendMenu(transMenu, transMenuString);
		dummy := SkelMenu(transMenu, @DoTransMenu, @DoMClobber, false, true);

		randomGeneratorMenu := NewMenu(randomGeneratorMenuID, '');
		AppendMenu(randomGeneratorMenu, randomGeneratorMenuString);
		dummy := SkelMenu(randomGeneratorMenu, @DoRandomGeneratorMenu, @DoMClobber, false, true);

		objectsMenu := NewMenu(objectsMenuID, 'Objects');
		s := objectsMenuString;
		AdjustString(s, '', escChar, '', chr(randomGeneratorMenuID));
		AppendMenu(objectsMenu, s);
		dummy := SkelMenu(objectsMenu, @DoObjectsMenu, @DoMClobber, false, false);

		arithmeticMenu := NewMenu(arithmeticMenuID, 'Arithmetric');
		AppendMenu(arithmeticMenu, arithmeticMenuString);
		dummy := SkelMenu(arithmeticMenu, @DoArithmeticMenu, @DoMClobber, false, false);

		animationMenu := NewMenu(animationMenuID, 'Animation');
		AppendMenu(animationMenu, animationMenuString);
		for i := minAnimation to maxAnimation do
			AppendMenu(animationMenu, 'none');
		for i := minAnimation to maxAnimation do
			DisableItem(animationMenu, i);
		dummy := SkelMenu(animationMenu, @DoAnimationMenu, @DoMClobber, false, false);
{$IFC NOT AnimationEnabled }
		DisableItem(animationMenu, 0);
{$ENDC }

{$IFC NetworkEnabled }
		cosimaMenu := NewMenu(cosimaMenuID, '');					(* MS: 19.7.93 - menu for COSIMA server/Client *)
		AppendMenu(cosimaMenu, cosimaMenuString);
		dummy := SkelMenu(cosimaMenu, @DoCosimaMenu, @DoMClobber, false, true);
{$ENDC }

{$IFC Topography }
		topoMenu := NewMenu(topoMenuID, '');
		AppendMenu(topoMenu, topoMenuString);
		dummy := SkelMenu(topoMenu, @DoTopoMenu, @DoMClobber, false, true);
{$ENDC }

		windowsMenu := NewMenu(windowsMenuID, 'Windows');
		AppendMenu(windowsMenu, windowsMenuString);
		for i := 1 to maxWindow do
			with wStatus^[i] do begin
				window := nil;
				name := windowName(i);
				fileName := name;
				if i <= 9 then
					AppendMenu(windowsMenu, concat(name, '/', chr(i + ord('0'))))
				else
					AppendMenu(windowsMenu, name);
			end;
		dummy := SkelMenu(windowsMenu, @DoWindowsMenu, @DoMClobber, true, false);

		operationsMenu := nil;
		lastSubMenuID := firstSubMenuID - 1;

		infoWindow := NewWindow(nil, infoWindowRect, 'Info', infoWindowVisible, documentProc, WindowPtr(-1), true, 0);
		dummy := SkelWindow(infoWindow, nil, nil, @DoCursorUpdate, nil, nil, nil, @DoWClobber, nil, false);

		defaultModes(infoWindowTextSize);
		PenMode(patOr);
		TextMode(srcCopy);

	end;

function eventHook (var theEvent: EventRecord): boolean;
	begin
		if (theEvent.what = mouseDown) | (theEvent.what = keyDown) then begin
			shiftKeyPressed := BitAnd(theEvent.modifiers, shiftKey) > 0;
			optionKeyPressed := BitAnd(theEvent.modifiers, optionKey) > 0;
		end;
		eventHook := false;
	end;

{$IFC NOT XYZApplication }
procedure TestResourcesReadable;
	begin
		if getResource('STR ', 19861) = nil then begin
			ShowText;
			writeLn('GeoBench.rsrc is not readable!');
			SysBeep(1);
			halt;
		end;
	end;
{$ENDC }

	const
		logoResource = 128;

	var
		logo: DialogPtr;

{$IFC OPTION(MC68881) or OPTION(MC68020) }
procedure TestMachine;
	const
		haltItem = 1;
		wrongMachineAlert = 1012;
	var
		error: OSErr;
		theWorld: SysEnvRec;
	begin
		error := SysEnvirons(1, theWorld);
		if (error <> noErr) | (theWorld.processor < env68020) | not theWorld.hasFPU then
			if false & (StopAlert(wrongMachineAlert, nil) = haltItem) then
				halt;
	end;

procedure InitFPState;
inline
	$42A7, $42A7, $F21F, $9800;

{$ENDC }

procedure initSystem;
	begin
{$IFC THINK_PASCAL }
		InitGraf(@thePort);
{$ELSEC }
		InitGraf(@qd.thePort);
{$ENDC }
		InitFonts;
		InitWindows;
		InitMenus;
		TEInit;
		InitDialogs(nil);
		InitCursor;
{$IFC XYZApplication }
		SetApplLimit(Ptr(LongInt(GetApplLimit) - 253952)); (* stack size is 8kB + 253952 Bytes = 256 kB *)
{$ELSEC }
		SetApplLimit(Ptr(LongInt(GetApplLimit) - 57344)); (* stack size is 8kB + 57344 Bytes = 64 kB *)
{$ENDC }
		MaxApplZone;
	end;

{$I- }

begin
{    InitProfile;}
{    %_PTrace := false;}
	initSystem;
{$IFC OPTION(MC68881) or OPTION(MC68020) }
	TestMachine;
	initFPState;
{$ENDC }
{$IFC NOT XYZApplication }
	TestResourcesReadable;
{$ENDC }
	initGeoBenchUtility;
	UTILITIESinit;
	DSInitGlobals;
	showCenteredDialog(logo, logoResource);
	initGeneralExperiment;
	new(wStatus);
	failNil(Handle(wStatus));
	initDialogEngine;
	initStreamIO;
	readPreferences;
	maxDrawingComplexity := GetConstResource(200);
	maxScrapBookComplexity := GetConstResource(201);
	maxStopWatchComplexity := GetConstResource(203);
	updateCreationTime;
	lastMemoryAlertPrinted := 0;
	memoryAlertEnabled := true;
	processInBackground := false;
	notificationPending := false;
	geoBenchIsComputing := false;
	previousFrontWindow := nil;
	beginCursor(stopwatchCursor);
{$IFC XYZApplication }
	SkelInit(GetConstResource(204), nil);	(* Initialize *)
{$ELSEC }
	SkelInit(128, nil);	(* Initialize *)
{$ENDC }
	SkelApple(appleMenuString, @DoAbout, @DoDeskAccOpen, @DoResume);
	SkelEventHook(@eventHook);
	windowCount := 0;
	garbageTimer := 0;
	scrapBook := nil;
	scrapFrame.proD := nil;
	undoScrap := nil;
	undoWindow := 0;
	offScreenId := 0;
	previousClickPoint.h := 0;
	previousClickPoint.v := 0;
	previousClickTime := 0;
	defaultNumber := initialDefaultNumber;
	defaultPolyVertices := initialDefaultNumber;
	defaultPolyObjects := 1;
	defaultAngle := 0;
	defaultDx := 10.0;
	defaultDy := 0;
	defaultLambda := 1.0;
	defaultOriginX := 0.0;
	defaultOriginY := 0.0;
	defaultA := 1.0;
	defaultB := -1.0;
	defaultC := 0.0;
	defaultDimension := 2;
	defaultVertices := 6;
	defaultAlignmentIsOn := false;
	defaultRegularNGonAngle := 0.0;
	cpuTimeUsedLower := 0.0;
	cpuTimeUsedUpper := 0.0;
	realTimeUsed := 0.0;
	sPDialogPos.v := 0;
	uTDialogPos.v := 0;
	selectedType := lastObjectType;
	selectedItem := selectItem;
	lastActiveWindow := 0;
	itemToType[pointItem] := point2dType;
	itemToType[rectangleItem] := rectangleType;
	itemToType[lineSegmentItem] := lineSegmentType;
	itemToType[polygonItem] := polygon2dType;
	itemToType[convexPolygonItem] := convexPolygonType;
	itemToType[circleItem] := circleType;
	itemToType[dDimPointItem] := dDimPointType;
	itemToType[polyLineItem] := polyLineType;
	itemToType[regularNGonItem] := regularNGonType;
	typeToItem[point2dType] := pointItem;
	typeToItem[rectangleType] := rectangleItem;
	typeToItem[lineSegmentType] := lineSegmentItem;
	typeToItem[polygon2dType] := polygonItem;
	typeToItem[convexPolygonType] := convexPolygonItem;
	typeToItem[circleType] := circleItem;
	typeToItem[dDimPointType] := dDimPointItem;
	typeToItem[polyLineType] := polyLineItem;
	typeToItem[regularNGonType] := regularNGonItem;
	installMenus;
	SkelSetMenuEnable(@enableItems);
	exclusionSet := [vectorType, searchTriangulation2dType, homogenVectorType, realPointType, floatPointType, longIntPointType, markedRealPointType, directedGraphType, QDPictureType, ColorQDPictureType, directedLineSegmentType, idRealPointType, sequenceType, listType, boundingBox2dType, boundingBox3dType, histogramType, gebaeudeLabelType];
	initAnimation;
	initGetUserParameter;
	initInfoWindow;
	initObj;
	initDictionary;
	initPriorityQueue;
	initGraphics;
	initBasicGeometry;
	initNearestNeighborInSector;
	initSequence;
	initVector;
	initRectangleVector;
	initLineSegmentVector;
	initLineSegment;
	initGraph;
	initGraph2;
	initTestGraph;
	initGetTriangulation;
	initPolygonAreaExperiment;
	initPointVector;
	initRegularNGon;
	initdDimPointVector;
	initPoint3d;
	initUpperMonotonePolygon;
	initTriangulation;
	initTriangulation3d;
	initSearchTriangulation2d;
	initSplinegon;
	initQDPicture;
	initMarkedRealPoint;
	initTextIO;
	initDoubleList;
	initTwoDTree;
	initQuadTree;
	initFreeListAVL;
	initLayer;
	initLayerVector;
	initBoundingBoxes;

{$IFC RaumPlanung }
	initRaumPlanung;
{$ENDC }
{$IFC Chemistry }
	initAtom;
	initMoleculeDraw;
	initMoleculeIO;
{$IFC ChemistryTriangulation }
	initEmbed3dNew;
{$ENDC }
{$ENDC }
{$IFC Topography }
	initSwissTopo;
{$ENDC }
{$IFC  GFEnabled }
	initpPoint;
	InitGridFile;
{$ENDC }
{$IFC ScriptEnabled }
	TransEditInit;
	SetEWindowProcs(nil, @geoScriptKeyHandle, @geoScriptActivate, @TextClose);  (* MS: 21.5.93 *)
	initGeoServer;
{$ENDC }
{$IFC NetworkEnabled }
	initCosimaGlobals;
{$ENDC }

{$IFC XYZApplication }
	initFastUndirectedGraph(GetConstResource(209), GetConstResource(210));
{$ELSEC }
	initFastUndirectedGraph(32, 256);
{$ENDC }
	startMemory := FreeMem + getScrapSize;
{$IFC XYZApplication }
	extendFreeList(GetConstResource(205));
	extendFreeListAVL(GetConstResource(206));
	extendFreeTreeList(GetConstResource(207));
	extendFreeDoubleList(GetConstResource(208));
{$ELSEC }
	extendFreeList(1024);
	extendFreeListAVL(1024);
	extendFreeTreeList(64);
	extendFreeDoubleList(64);
{$ENDC }
	endCursor;
	hideCenteredDialog(logo);
	SkelMain;												(* loop til quit selected *)
	beginCursor(stopwatchCursor);
	writePreferences;
	saveScrapBook;
	forgetUndo;
	SkelClobber;											(* clean up *)
	shrinkFreeList(freeListLength);
	shrinkFreeListAVL(freeListLengthAVL);
	shrinkFreeTreeList(freeTreeListLength);
	shrinkFreeDoubleList(freeDoubleListLength);
	exitTriangulation;
	exitNearestNeighborInSector;
	exitObj;
	exitGraph2;
	exitFastUndirectedGraph;
	exitVector;

{$IFC  GFEnabled }
	ExitGridFile;
{$ENDC }
{$IFC ScriptEnabled }
	exitGeoServer;
{$ENDC }
{$IFC NetworkEnabled }
	exitCosimaGlobals;
{$ENDC }
{$IFC Chemistry }
	exitMoleculeDraw;
	exitAtom;
	exitMoleculeIO;
{$ENDC }

	endCursor;
end.
