(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit geoBenchPreferences;

interface

uses
{$IFC ScriptEnabled }
	GeoScript, 
{$ENDC }
{$IFC Chemistry }
	atom, 
{$ENDC }
{$IFC THINK_PASCAL }
	SysEqu,
{$ENDC }
	geoBenchScrap;

var
	undoEnabled, showWakeUpString, infoWindowVisible: boolean;
	geoTextSize: integer;
	infoWindowRect: Rect;

procedure readPreferences;

procedure writePreferences;

procedure preferencesMenuItem;

implementation

(* code to be modified if preferences change begins here *)

const
	prefSignature = 1521986109;
	haltItem = 3;
	continueButDontSaveItem = 4;
	PrefFileType = 'Pref';
	PrefCreatorType = gbCreator;
	defaultUndoEnabled = true;
	defaultGeoTextSize = 24;
	defaultGlobalInitialGuessing = false;
	defaultPointRadius = 2;
	defaultPointRadiusHighlight = 4;
	defaultAcceptCircles = false;
	defaultResultHandlingMode = createNewWindowMode;
	defaultWaitLimit = 30;
	defaultGeoBenchFont = 3;
	defaultResultHandlingModifiersAllowed = false;
	defaultWriteTextToClipBoard = false;
	defaultTextFileOutputMethod = 0;
	defaultWriteGraphicsToClipBoard = true;
	defaultProcessPriorityUpdateEvents = true;
	defaultDrawLengthRatio = true;
	defaultTextCreatorType = 0;
	defaultShowWakeUpString = false;
	defaultInfoWindowVisible = true;
{$IFC Chemistry }
	defaultCurrentDrawEdgeMode = all;
{$ELSEC }
	defaultCurrentDrawEdgeMode = 0;
{$ENDC }


var
	currentSignature: longInt;
	saveAllowed: boolean;
	oldUndoEnabled: boolean;
	oldGeoTextSize: integer;
	oldGlobalInitialGuessing: boolean; (* 'globalInitialGuessing' is declared in 'proximity' *)
	oldPointRadius, oldPointRadiusHighlight: integer;
	oldAcceptCircles: boolean;
	oldResultHandlingMode: resultHandlingModeType;
	oldWaitLimit: integer;
	oldGeoBenchFont: integer;
	oldResultHandlingModifiersAllowed: boolean;
	oldWriteTextToClipBoard: boolean;
	oldTextFileOutputMethod: integer;
	oldWriteGraphicsToClipBoard: boolean;
	oldProcessPriorityUpdateEvents: boolean;
	oldDrawLengthRatio: boolean;
{$IFC Not Chemistry }
	drawLengthRatio: boolean;
	currentDrawEdgeMode: integer;
{$ENDC }
	oldTextCreatorType: integer;
	oldShowWakeUpString: boolean;
	oldInfoWindowVisible: boolean;
	oldInfoWindowRect: Rect;
	oldCurrentDrawEdgeMode: integer;

procedure oldIsCurrent;
	begin
		oldUndoEnabled := undoEnabled;
		oldGeoTextSize := geoTextSize;
		oldGlobalInitialGuessing := globalInitialGuessing;
		oldPointRadius := pointRadius;
		oldPointRadiusHighlight := pointRadiusHighlight;
		oldAcceptCircles := acceptCircles;
		oldResultHandlingMode := resultHandlingMode;
		oldWaitLimit := waitLimit;
		oldGeoBenchFont := geoBenchFont;
		oldResultHandlingModifiersAllowed := resultHandlingModifiersAllowed;
		oldWriteTextToClipBoard := writeTextToClipBoard;
		oldTextFileOutputMethod := textFileOutputMethod;
		oldWriteGraphicsToClipBoard := writeGraphicsToClipBoard;
		oldProcessPriorityUpdateEvents := processPriorityUpdateEvents;
		oldDrawLengthRatio := drawLengthRatio;
		oldTextCreatorType := textCreatorType;
		oldShowWakeUpString := showWakeUpString;
		oldInfoWindowVisible := infoWindowVisible;
		oldInfoWindowRect := infoWindowRect;
		oldCurrentDrawEdgeMode := currentDrawEdgeMode;
	end;

procedure currentIsOld;
	begin
		undoEnabled := oldUndoEnabled;
		geoTextSize := oldGeoTextSize;
		globalInitialGuessing := oldGlobalInitialGuessing;
		pointRadius := oldPointRadius;
		pointRadiusHighlight := oldPointRadiusHighlight;
		acceptCircles := oldAcceptCircles;
		resultHandlingMode := oldResultHandlingMode;
		waitLimit := oldWaitLimit;
		geoBenchFont := oldGeoBenchFont;
		resultHandlingModifiersAllowed := oldResultHandlingModifiersAllowed;
		writeTextToClipBoard := oldWriteTextToClipBoard;
		textFileOutputMethod := oldTextFileOutputMethod;
		writeGraphicsToClipBoard := oldWriteGraphicsToClipBoard;
		processPriorityUpdateEvents := oldProcessPriorityUpdateEvents;
		drawLengthRatio := oldDrawLengthRatio;
		textCreatorType := oldTextCreatorType;
		showWakeUpString := oldShowWakeUpString;
		infoWindowVisible := oldInfoWindowVisible;
		infoWindowRect := oldInfoWindowRect;
		currentDrawEdgeMode := oldCurrentDrawEdgeMode;
	end;

procedure writePrefData;
	begin
		writeRect(infoWindowRect);
		writeByte(ord(infoWindowVisible));
		writeByte(ord(undoEnabled));
		writeInteger(geoTextSize);
		writeByte(ord(globalInitialGuessing));
		writeInteger(pointRadius);
		writeInteger(pointRadiusHighlight);
		writeByte(ord(acceptCircles));
		writeByte(ord(resultHandlingMode));
		writeInteger(waitLimit);
		writeInteger(geoBenchFont);
		writeByte(ord(resultHandlingModifiersAllowed));
		writeByte(ord(writeTextToClipBoard));
		writeInteger(textFileOutputMethod);
		writeByte(ord(writeGraphicsToClipBoard));
		writeByte(ord(processPriorityUpdateEvents));
		writeByte(ord(drawLengthRatio));
		writeInteger(textCreatorType);
		writeByte(ord(showWakeUpString));
		writeInteger(currentDrawEdgeMode);
	end;

procedure defaultInfoWindowRect;
	begin
{$IFC THINK_PASCAL }
		SetRect(infoWindowRect, 1, screenBits.bounds.bottom - infoWindowHeight, infoWindowWidth, screenBits.bounds.bottom - 1);
{$ELSEC }
		SetRect(infoWindowRect, 1, qd.screenBits.bounds.bottom - infoWindowHeight, infoWindowWidth, qd.screenBits.bounds.bottom - 1);
{$ENDC }
	end;

procedure readPrefData;
	var
		b: Byte;
	begin
		readRect(infoWindowRect);
{$IFC THINK_PASCAL }
		if (infoWindowRect.right < screenBits.bounds.left) | (infoWindowRect.left > screenBits.bounds.right) | (infoWindowRect.top > screenBits.bounds.bottom) | (infoWindowRect.bottom < screenBits.bounds.top) then
{$ELSEC }
		if (infoWindowRect.right < qd.screenBits.bounds.left) | (infoWindowRect.left > qd.screenBits.bounds.right) | (infoWindowRect.top > qd.screenBits.bounds.bottom) | (infoWindowRect.bottom < qd.screenBits.bounds.top) then
{$ENDC }
			defaultInfoWindowRect;
		readByte(b);
		infoWindowVisible := b = 1;
		readByte(b);
		undoEnabled := b = 1;
		readInteger(geoTextSize);
		readByte(b);
		globalInitialGuessing := b = 1;
		readInteger(pointRadius);
		readInteger(pointRadiusHighlight);
		readByte(b);
		acceptCircles := b = 1;
		readByte(b);
		resultHandlingMode := resultHandlingModeType(b);
		readInteger(waitLimit);
		readInteger(geoBenchFont);
		readByte(b);
		resultHandlingModifiersAllowed := b = 1;
		readByte(b);
		writeTextToClipBoard := b = 1;
		readInteger(textFileOutputMethod);
{$IFC NOT Chemistry }
		if textFileOutputMethod > 2 then
			textFileOutputMethod := defaultTextFileOutputMethod;
{$ENDC }
		readByte(b);
		writeGraphicsToClipBoard := b = 1;
		readByte(b);
		processPriorityUpdateEvents := b = 1;
		readByte(b);
		drawLengthRatio := b = 1;
		readInteger(textCreatorType);
		readByte(b);
		showWakeUpString := b = 1;
		readInteger(currentDrawEdgeMode);
	end;

procedure defaultPrefData;
	begin
		defaultInfoWindowRect;
		infoWindowVisible := defaultInfoWindowVisible;
		undoEnabled := defaultUndoEnabled;
		geoTextSize := defaultGeoTextSize;
		globalInitialGuessing := defaultGlobalInitialGuessing;
		pointRadius := defaultPointRadius;
		pointRadiusHighlight := defaultPointRadiusHighlight;
		acceptCircles := defaultAcceptCircles;
		resultHandlingMode := defaultResultHandlingMode;
		waitLimit := defaultWaitLimit;
		geoBenchFont := defaultGeoBenchFont;
		resultHandlingModifiersAllowed := defaultResultHandlingModifiersAllowed;
		writeTextToClipBoard := defaultWriteTextToClipBoard;
		textFileOutputMethod := defaultTextFileOutputMethod;
		writeGraphicsToClipBoard := defaultWriteGraphicsToClipBoard;
		processPriorityUpdateEvents := defaultProcessPriorityUpdateEvents;
		drawLengthRatio := defaultDrawLengthRatio;
		textCreatorType := defaultTextCreatorType;
		showWakeUpString := defaultShowWakeUpString;
		currentDrawEdgeMode := defaultCurrentDrawEdgeMode;
	end;

function updateNecessary: boolean;
	var
		result: boolean;
	begin
		result := (undoEnabled <> oldUndoEnabled) | (geoTextSize <> oldGeoTextSize) | (globalInitialGuessing <> oldGlobalInitialGuessing) | (pointRadius <> oldPointRadius) | (pointRadiusHighlight <> oldPointRadiusHighlight) | (writeGraphicsToClipBoard <> oldWriteGraphicsToClipBoard) | (processPriorityUpdateEvents <> oldProcessPriorityUpdateEvents) | (drawLengthRatio <> oldDrawLengthRatio);
		result := result | (acceptCircles <> oldAcceptCircles) | (resultHandlingMode <> oldResultHandlingMode) | (waitLimit <> oldWaitLimit) | (geoBenchFont <> oldGeoBenchFont) | (resultHandlingModifiersAllowed <> oldResultHandlingModifiersAllowed) | (writeTextToClipBoard <> oldWriteTextToClipBoard) | (textFileOutputMethod <> oldTextFileOutputMethod) | (textCreatorType <> oldTextCreatorType) | (showWakeUpString <> oldShowWakeUpString);
		updateNecessary := result | (infoWindowVisible <> oldInfoWindowVisible) | (infoWindowRect.left <> oldInfoWindowRect.left) | (infoWindowRect.right <> oldInfoWindowRect.right) | (infoWindowRect.top <> oldInfoWindowRect.top) | (infoWindowRect.bottom <> oldInfoWindowRect.bottom) | (currentDrawEdgeMode <> oldCurrentDrawEdgeMode);
	end;

function checkFontSize (x: extended): boolean;
	begin
		checkFontSize := (5 <= x) & (x <= 72) & (trunc(x) = x);
	end;

function checkPointRadius (x: extended): boolean;
	begin
		checkPointRadius := (1 <= x) & (x <= 256) & (trunc(x) = x);
	end;

function checkWaitLimit (x: extended): boolean;
	begin
		checkWaitLimit := trunc(x) = x;
	end;

function checkFontNumber (x: extended): boolean;
	begin
		checkFontNumber := (trunc(x) = x) & (0 <= x) & (x <= 24);
	end;

function textFormatVisible: boolean;
	begin
		textFormatVisible := writeTextToClipBoard;
	end;

	var
		h: vector;
		geoTextSizeExtended, pointRadiusExtended, pointRadiusHighlightExtended, waitLimitExtended, geoBenchFontExtended: extended;
		resultHandlingModeInteger: integer;
		previousUndoEnabled: boolean;

procedure updateValues;
	begin
{$IFC THINK_PASCAL }
		geoTextSize := trunc(geoTextSizeExtended);
		pointRadius := trunc(pointRadiusExtended);
		pointRadiusHighlight := trunc(pointRadiusHighlightExtended);
		waitLimit := trunc(waitLimitExtended);
		geoBenchFont := trunc(geoBenchFontExtended);
{$ELSEC }
		geoTextSize := truncLong(geoTextSizeExtended);
		pointRadius := truncLong(pointRadiusExtended);
		pointRadiusHighlight := truncLong(pointRadiusHighlightExtended);
		waitLimit := truncLong(waitLimitExtended);
		geoBenchFont := truncLong(geoBenchFontExtended);
{$ENDC }
		resultHandlingMode := resultHandlingModeType(resultHandlingModeInteger);
		enableDisableOperationsMenu(true);
		if previousUndoEnabled & not undoEnabled then
			forgetUndo;
	end;

procedure defaultButton;
	var
		r: Rect;
	begin
		defaultInfoWindowRect;
		infoWindowVisible := defaultInfoWindowVisible;
		r := getPortRect(infoWindow);
		if (r.left <> infoWindowRect.left) | (r.top <> infoWindowRect.top) then
			MoveWindow(infoWindow, infoWindowRect.left, infoWindowRect.top, false);
		if infoWindowVisible then
			ShowWindow(infoWindow)
		else
			HideWindow(infoWindow);
	end;

procedure saveButton;
	var
		i: seqIndex1;
		canBeSaved: boolean;
	begin
		canBeSaved := true;
		for i := 1 to h.length do
			with interactiveHandler(h.elements^^[i]) do
				if not isAcceptable then begin
					restore;
					canBeSaved := false;
				end;
		if canBeSaved then begin
			updateValues;
			writePreferences
		end
		else
			for i := 1 to h.length do
				with interactiveHandler(h.elements^^[i]) do begin
					showValue;
					show;
				end;
	end;

function saveButtonVisible: boolean;
	begin
		saveButtonVisible := saveAllowed;
	end;

{$IFC Chemistry }
function objectExists (s1, s2: sequence; containing: objectTypeSet): boolean;

	function findObject (x: obj): boolean;
		begin
			findObject := (x.getType in containing) | member(x, sequence) & (sequence(x).exists(findObject) > 0);
		end;

	begin
		objectExists := (s1.exists(findObject) > 0) | (s2.exists(findObject) > 0);
	end;

procedure redrawAllWindows (containing: objectTypeSet);
	var
		i: windowIndex;
		safe: GrafPtr;
	begin
		GetPort(safe);
		for i := 1 to maxWindow do
			with wStatus^[i] do
				if (window <> nil) & objectExists(desktop, selection, containing) then begin
					SetPort(window);
					redraw(i);
				end;
		SetPort(safe);
	end;
{$ENDC }

procedure preferencesMenuItem;
{$IFC Chemistry }
	var
		previousCurrentDrawEdgeMode: integer;
		previousDrawLengthRatio: boolean;
{$ENDC }
	begin
{$IFC Chemistry }
		previousCurrentDrawEdgeMode := currentDrawEdgeMode;
		previousDrawLengthRatio := drawLengthRatio;
{$ENDC }
		geoTextSizeExtended := geoTextSize;
		pointRadiusExtended := pointRadius;
		pointRadiusHighlightExtended := pointRadiusHighlight;
		waitLimitExtended := waitLimit;
		geoBenchFontExtended := geoBenchFont;
		resultHandlingModeInteger := ord(resultHandlingMode);
		previousUndoEnabled := undoEnabled;
		h := newVector(32);
		h.append(getEnumerationMenuHandler(26, @textCreatorType, defaultTextCreatorType, nil, t('TeachText', t('BBEdit', t('MS Word', t('MacWrite II', nil))))));
		h.append(getExtendedHandler(7, @pointRadiusExtended, defaultPointRadius, nil, @checkPointRadius));
		h.append(getExtendedHandler(8, @pointRadiusHighlightExtended, defaultPointRadiusHighlight, nil, @checkPointRadius));
		h.append(getExtendedHandler(14, @waitLimitExtended, defaultWaitLimit, nil, @checkWaitLimit));
		h.append(getExtendedHandler(6, @geoBenchFontExtended, defaultGeoBenchFont, nil, @checkFontNumber));
		h.append(getExtendedHandler(4, @geoTextSizeExtended, defaultGeoTextSize, nil, @checkFontSize));
{$IFC Chemistry }
		h.append(getBooleanHandler(25, @drawLengthRatio, defaultDrawLengthRatio, nil));
		h.append(getEnumerationMenuHandler(32, @currentDrawEdgeMode, defaultCurrentDrawEdgeMode, nil, t('All', t('Bond', t('NMR', t('Calc', t('Angle', t('Other', t('Exact', nil)))))))));
{$ELSEC }
		h.append(getBooleanHandler(25, @drawLengthRatio, defaultDrawLengthRatio, @alwaysFalse));
		h.append(getEnumerationMenuHandler(32, @currentDrawEdgeMode, defaultCurrentDrawEdgeMode, @alwaysFalse, t('All', t('Bond', t('NMR', t('Calc', t('Angle', t('Other', t('Exact', nil)))))))));
		h.append(getStaticTextHandler(31, '', false, @alwaysFalse));
{$ENDC }
		h.append(getBooleanHandler(3, @undoEnabled, defaultUndoEnabled, nil));
		h.append(getBooleanHandler(5, @globalInitialGuessing, defaultGlobalInitialGuessing, nil));
		h.append(getBooleanHandler(12, @acceptCircles, defaultAcceptCircles, nil));
		h.append(getBooleanHandler(18, @resultHandlingModifiersAllowed, defaultResultHandlingModifiersAllowed, nil));
		h.append(getBooleanHandler(19, @writeTextToClipBoard, defaultWriteTextToClipBoard, nil));
		h.append(getBooleanHandler(30, @showWakeUpString, defaultShowWakeUpString, nil));
{$IFC Chemistry }
		h.append(getEnumerationMenuHandler(20, @textFileOutputMethod, defaultTextFileOutputMethod, @textFormatVisible, t('Standard', t('Point List', t('Length First', t('Metric Matrix', t('Exchange', t('Exchange All', t('Maple', t('Equations', nil))))))))));
{$ELSEC }
		h.append(getEnumerationMenuHandler(20, @textFileOutputMethod, defaultTextFileOutputMethod, @textFormatVisible, t('Standard', t('Point List', t('Length First', nil)))));
{$ENDC }
		h.append(getStaticTextHandler(28, '', false, @textFormatVisible));
		h.append(getBooleanHandler(23, @writeGraphicsToClipBoard, defaultWriteGraphicsToClipBoard, nil));
		h.append(getBooleanHandler(24, @processPriorityUpdateEvents, defaultProcessPriorityUpdateEvents, nil));
		h.append(getEnumerationMenuHandler(29, @resultHandlingModeInteger, ord(defaultResultHandlingMode), nil, t('New Window', t('New Adjacent Window', t('Add as Selection', t('Replace Selection', t('Discard Result', nil)))))));
		h.append(getButtonHandler(22, @saveButtonVisible, @saveButton));
		h.append(getButtonHandler(11, nil, @defaultButton));
		if processDialogBox(7200, 11, 21, 0, alwaysTrue, h) then begin
			updateValues;
{$IFC Chemistry }
			if (currentDrawEdgeMode <> previousCurrentDrawEdgeMode) | (drawLengthRatio <> previousDrawLengthRatio) then
				redrawAllWindows([moleculeType]);
{$ENDC }
		end;
	end;

(* code to be modified if preferences change ends here *)

{$IFC NOT THINK_PASCAL }
CONST
	BootDrive = $210;               {[GLOBAL VAR]  drive number of boot drive [word]}
{$ENDC }

procedure preferencesFileName (var name: Str255; var volume: integer);
	const
		errMsg = 'noErr in gbPref';
	type
		IntPtr = ^integer;
	var
		resultName, parentName: Str255;
		freeBytes: longInt;
		hpbRec: HParamBlockRec;
		cpbRec: CInfoPBRec;
	begin
		assert(GetVInfo(IntPtr(BootDrive)^, @resultName, volume, freeBytes) = noErr, errMsg);		(* get vRefNum *)
		with hpbRec do begin
			ioCompletion := nil;
			ioNamePtr := nil;
			ioVRefNum := volume;							(* get for default volume									*)
			ioVolIndex := 0;								(* were not making indexed calls						*)
{$IFC THINK_PASCAL }
			assert((PBHGetVInfo(@hpbRec, false) = noErr) & (ioVFndrInfo[1] <> 0) & (ioVFndrInfo[1] <> 2), errMsg);
{$ELSEC }
			assert((PBHGetVInfoSync(@hpbRec) = noErr) & (ioVFndrInfo[1] <> 0) & (ioVFndrInfo[1] <> 2), errMsg);
{$ENDC }
		end;
		with cpbRec do begin
			ioCompletion := nil;
			ioNamePtr := @resultName;
			ioVRefNum := volume;
			ioDrDirID := hpbRec.ioVFndrInfo[1]; 		(* the system folder id 									*)
			ioFDirIndex := -1;								(* get info about the directories 							*)
		end;
{$IFC THINK_PASCAL }
		assert(PBGetCatInfo(@cpbRec, false) = noErr, errMsg);
{$ELSEC }
		assert(PBGetCatInfoSync(@cpbRec) = noErr, errMsg);
{$ENDC }
		while cpbRec.ioDrParID > 1 do begin		(* in case the system folder is inside another folder	*)
			with cpbRec do begin
				ioCompletion := nil;
				ioNamePtr := @parentName;
				ioVRefNum := volume;
				ioDrDirID := ioDrParID; 						(* the parent folder id 										*)
				ioFDirIndex := -1;							(* get info about the directories 							*)
			end;
{$IFC THINK_PASCAL }
			assert(PBGetCatInfo(@cpbRec, false) = noErr, errMsg);
{$ELSEC }
			assert(PBGetCatInfoSync(@cpbRec) = noErr, errMsg);
{$ENDC }
			resultName := concat(parentName, ':', resultName);
		end;
		if isSystemSeven then
			resultName := concat(resultName, ':Preferences');
		name := concat(resultName, ':GeoBench Preferences');
	end;

procedure writePreferences;
	var
		volume: integer;
		name: Str255;
	begin
		infoWindowRect := getPortRect(infoWindow);
		infoWindowVisible := WindowPeek(infoWindow)^.visible;
		if saveAllowed & (updateNecessary | (currentSignature <> prefSignature)) then begin
			oldIsCurrent;
			defaultPrefData;
			preferencesFileName(name, volume);
			if updateNecessary then begin
				currentIsOld;
				openWriteStreamAppend(name, volume, PrefFileType, PrefCreatorType, false);
				writeLongInt(prefSignature);
				writePrefData;
				closeWriteStream;
			end
			else
				testOSError(FSDelete(name, volume), fnfErr);
		end;
	end;

procedure readPreferences;
	var
		volume, button: integer;
		name: Str255;
		preferenceFileExists: boolean;
	begin
		preferencesFileName(name, volume);
		openReadStreamTest(name, volume, saveAllowed, preferenceFileExists);
		if preferenceFileExists then begin
			readLongInt(currentSignature);
			if currentSignature = prefSignature then begin
				readPrefData;
				closeReadStream;
			end
			else begin
				closeReadStream;
				beginCursor(arrowCursor);
				button := CautionAlert(1020, nil);
				endCursor;
				if button = haltItem then
					halt;
				if button = continueButDontSaveItem then
					saveAllowed := false;
				defaultPrefData;
			end;
		end
		else
			defaultPrefData;
		oldIsCurrent;
	end;

end.
