(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1990 by Christoph Ammann  					*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.00: July 1991								*)

unit geoBenchScrap;

interface

uses
{$IFC ScriptEnabled }
	GeoScript, 
{$ENDC }
{$IFC Chemistry }
	MoleculeIO, 
{$ENDC }
	geoBenchGlobal, regularNGon, QDPicture;

function loadScrapBook: list;
(* Load external scrapbook and, if possible turn into a geoBench list. Used by the paste command.						*)
(* If a scrap entry of type XYGB is available, it is loaded using the 'fileInput' methods. If no XYGB,						*)
(* but a PICT (from a standard Macintosh drawing program like MacDraw) is available, it is converted in a			*)
(* reasonable way, e.g. line segment series are converted into line segments, polylines or polygons, whatever		*)
(* is appropriate. The conversion uses the current arithmetics setting of GeoBench to determine what					*)
(* arithmetic should be used to generate new points. The coordinates are converted using 'windowToWorld'			*)
(* i.e. using the current GeoBench frame settings.																				*)

procedure saveScrapBook;
(* Save both a XYGB and a PICT (QuickDraw) version of the variable 'scrapbook' to the external scrapbook,			*)
(* if the scrapbook is not already external (i.e. the variable 'scrapbook' is not set to 'nil').								*)
(* Standard Macintosh drawing programs can access the PICTs, and the GBXYs can later be reused by GeoBench.		*)
(* 'saveScrapBook' frees the variable 'scrapbook' and sets it to 'nil', which means the scrapbook is currently			*)
(* external and 'loadScrapBook' should be used to access it.																		*)
(* This procedure calls the 'fileOutput' methods of the current window's selected objects for the XYGB entry and		*)
(* the 'interactiveOutput' methods for the PICT entry.																			*)
(* This procedure should be called whenever one of the following occurs:														*)
(* -	a desk accessory is opened (especially under the old finder)																*)
(* -	a multifinder suspend event occurs (in System 6 (and older) simulated by a desk accessory						*)
(*  		open event; see notes about changes in TransSkel)																	*)
(* -	the GeoBench application is quit.																								*)

function arithmeticsToPointTypeRange (a: arithmetics): pointTypeRange;

function readTextObjSpecial: obj;

procedure showCenteredDialog (var myDialog: DialogPtr; myDialogResource: integer);

procedure hideCenteredDialog (myDialog: DialogPtr);

var
scrapFrame: wFrame; (* valid if scrapBook <> nil *)
scrapRect: Rect; (* valid if scrapBook <> nil *)
acceptCircles, writeTextToClipBoard, writeGraphicsToClipBoard: boolean;
textFileOutputMethod: integer;

implementation

{$IFC NOT THINK_PASCAL }
uses Scrap;
{$ENDC }

procedure doTextOutput (x: obj);
	begin
		if member(x, sequence) & not member(x, pointVector) then
			sequence(x).forAll(doTextOutput)
		else begin
			x.textFileOutput;
			writeSimpleCR;
		end;
	end;

function readTextObjSpecial: obj;

	function isStartOfNumber (ch: char): boolean;
		begin
			isStartOfNumber := ('0' <= ch) & (ch <= '9') | (ch = '.') | (ch = '+') | (ch = '-');
		end;

	function getCabriGraph: undirectedGraph;
		var
			s: Str255;
			result: undirectedGraph;
			p: realPoint;
			mrp: markedRealPoint;
			i, j, degree, startV, endV: longInt;
			x, y: real;
		begin
			repeat
				readRawString(s)
			until s[1] = '#';
			readTextLongInt(currentVertices);
			readRawString(s);
			assert(s[1] = '#', '# found');
			readTextLongInt(currentEdges);
			new(result);
			failNil(Handle(result));
			result.init;
			result.vertices.length := currentVertices;
			result.initAdjacency;
			readRawString(s);
			for i := 1 to currentVertices do begin
				IOCheck(false);
				readString(s, startV);
				assert(ioResult = 0, 'Number ok');
				IOCheck(true);
				startV := startV + 1;
				readRawString(s);
				assert(cap(s[1]) = 'D', 'D found');
				readTextLongInt(degree);
				for j := 1 to degree do begin
					readTextLongInt(endV);
					if startV <= endV then
						result.addNewEdge(startV, endV + 1);
				end;
				readRawString(s);
				assert(cap(s[1]) = 'C', 'C found');
				readTextReal(x);
				readTextReal(y);
				skipLine;
				readRawString(s);
				if cap(s[1]) = 'L' then begin
					new(mrp);
					failNil(Handle(mrp));
					mrp.init;
					readRawString(mrp.mark.value);
					skipLine;
					readRawString(s);
					p := mrp;
				end
				else begin
					new(p);
					failNil(Handle(p));
				end;
				p.setXY(x, y);
				result.vertices.elements^^[startV] := p;
			end;
			if s[1] = EOFChar then
				ch := EOFChar;
			getCabriGraph := result;
		end;

	function getPointList (resultType: objectType): pointVector;
		var
			pv: pointVector;
			l: list;
			p: realPoint;
		begin
			l := newEmptyList;
			skipWhiteSpace;
			while isStartOfNumber(ch) do begin
				new(p);
				failNil(Handle(p));
				readTextReal(p.x);
				readTextReal(p.y);
				l.append(p);
				skipWhiteSpace;
			end;
			pv := pointVector(newObj(resultType));
			pv.initFromList(l);
			l.free;
			getPointList := pv;
		end;

		var
			textTypeId: Str255;
			l: list;
	begin
		initLookAhead;
		skipWhiteSpace;
		if ch = '(' then
			readTextObjSpecial := readTextObj
		else begin
			l := newEmptyList;
			while ch <> EOFChar do begin
				if isStartOfNumber(ch) then
					l.append(getPointList(polygon2dType))
				else begin
					textTypeId := '';
					repeat
						textTypeId := concat(textTypeId, cap(ch));
						getCh;
					until isWhiteSpace(ch) | (ch = EOFChar);
					if (textTypeId = capString('PointList')) | (textTypeId = capString('PointVector')) then
						l.append(getPointList(pointVectorType))
					else if textTypeId = capString('Polygon') then
						l.append(getPointList(polygon2dType))
					else if textTypeId = capString('ConvexPolygon') then
						l.append(getPointList(convexPolygonType))
					else if textTypeId = capString('PolyLine') then
						l.append(getPointList(polyLineType))
					else if textTypeId = capString('CABRI') then
						l.append(getCabriGraph)
{$IFC Chemistry }
					else if textTypeId = capString('Molecule') then
						l.append(getMolecule)
{$ENDC }
					else
						fatalError('Illegal Text Type or ''('' expected');
					skipWhiteSpace;
				end;
			end;
			readTextObjSpecial := l;
		end;
	end;

	const
		stdPalette = 130;

	var
		clipboardHandle: Handle;
		clipboardIndex: longInt;

function arithmeticsToPointTypeRange (a: arithmetics): pointTypeRange;
	begin
		arithmeticsToPointTypeRange := pointTypeRange(ord(a) + ord(realPointType));
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure dummyExtend (x, y: integer);
	begin (* do nothing ! *)
	end;

function dummyFSOpen (fileName: Str255; vRefNum: integer; var refNum: integer): OSErr;
	begin
		dummyFSOpen := noErr;
	end;

function clipboardFSWrite (refNum: integer; var count: longInt; buffPtr: Ptr): OSErr;
	var
		size: longInt;
	begin
		if clipboardHandle = nil then begin
			size := startMemory div 20;
			if size < count then
				size := count;
			clipboardHandle := newHandleFailNil(size);
			clipboardIndex := 0;
		end
		else if clipboardIndex + count > GetHandleSize(clipboardHandle) then begin
			SetHandleSize(clipboardHandle, clipboardIndex + count);
			if MemError <> noErr then
				failNil(nil);
		end;
		BlockMove(buffPtr, Ptr(longInt(clipboardHandle^) + clipboardIndex), count);
		clipboardIndex := clipboardIndex + count;
		clipboardFSWrite := noErr;
	end;

function clipboardFSRead (refNum: integer; var count: longInt; buffPtr: Ptr): OSErr;
	begin
		if clipboardIndex + count > GetHandleSize(clipboardHandle) then
			count := GetHandleSize(clipboardHandle) - clipboardIndex;
		if count > 0 then
			BlockMove(Ptr(longInt(clipboardHandle^) + clipboardIndex), buffPtr, count);
		clipboardIndex := clipboardIndex + count;
		advanceStopWatchHandsNoMem;
		clipboardFSRead := noErr;
	end;

function dummyFSClose (refNum: integer): OSErr;
	begin
		dummyFSClose := noErr;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure showCenteredDialog (var myDialog: DialogPtr; myDialogResource: integer);
	var
		myDialogWidth, myDialogHeight, screenWidth, screenHeight: integer;
	begin
		myDialog := GetNewDialog(myDialogResource, nil, WindowPtr(-1));
		with GrafPtr(myDialog)^.portRect do begin
			myDialogWidth := right - left;
			myDialogHeight := bottom - top;
		end;
{$IFC THINK_PASCAL }
		with screenBits.bounds do begin
{$ELSEC }
		with qd.screenBits.bounds do begin
{$ENDC }
			screenWidth := right - left;
			screenHeight := bottom - top;
		end;
		MoveWindow(WindowPtr(myDialog), (screenWidth - myDialogWidth) div 2, (screenHeight - myDialogHeight) div 2, true);
		ShowWindow(myDialog);
		DrawDialog(myDialog);
	end;

procedure hideCenteredDialog (myDialog: DialogPtr);
	begin
		HideWindow(myDialog);
{$IFC THINK_PASCAL }
		DisposDialog(myDialog);
{$ELSEC }
		DisposeDialog(myDialog);
{$ENDC }
	end;

procedure saveScrapBook;
	const
		xSafe = 20;
		ySafe = 20;
		bigScrapBookDialogResource = 149;
	var
		picture: PicHandle;
		ignore, complexity: longInt;
		dummy: boolean;
		save: clipper;
		saveFrame: wFrame;
		gp: GrafPort;
		cgp: CGrafPort;
		savePort: GrafPtr;
		saveStatus: streamStatus;
		bigScrapBookDialog: DialogPtr;
	begin
		if scrapBook <> nil then begin
			ignore := ZeroScrap;
			if maxStopWatchComplexity > maxScrapBookComplexity then
				complexity := maxStopWatchComplexity
			else
				complexity := maxScrapBookComplexity;
			complexity := scrapBook.drawingComplexity(complexity);

			if complexity >= maxStopWatchComplexity then
				beginCursor(stopwatchCursor);
			if complexity >= maxScrapBookComplexity then
				showCenteredDialog(bigScrapBookDialog, bigScrapBookDialogResource);

			getStreamStatus(saveStatus);
			setStreamStatus(createStreamStatus(nil, @dummyFSOpen, nil, @clipboardFSWrite, @dummyFSClose, saveStatus.bufferSize));
			clipboardHandle := nil;
			openWriteStream('', 0);
			writeLongInt(thisVersion);
			scrapBook.fileOutput;
			closeWriteStream;
			ignore := PutScrap(clipboardIndex, gbFileType, clipboardHandle^);
{$IFC THINK_PASCAL }
			disposHandle(clipboardHandle);
{$ELSEC }
			DisposeHandle(clipboardHandle);
{$ENDC }
			if writeTextToClipBoard then begin
				clipboardHandle := nil;
				openWriteStream('', 0);
				currentTextFileOutputMethod := textFileOutputMethod;
				if textFileOutputMethod = 0 then
					writeTextObjNoCR(scrapBook)
				else
					doTextOutput(scrapBook);
				closeWriteStream;
				if clipboardHandle <> nil then begin
					ignore := PutScrap(clipboardIndex, 'TEXT', clipboardHandle^);
{$IFC THINK_PASCAL }
					disposHandle(clipboardHandle);
{$ELSEC }
					DisposeHandle(clipboardHandle);
{$ENDC }
				end;
			end;
			setStreamStatus(saveStatus);

			if writeGraphicsToClipBoard then begin
				save := clippingRectangle;
				saveFrame := currentFrame;
				currentFrame := scrapFrame;
				sheetExtendProc := @dummyExtend;
				GetPort(savePort);
				if colorAvailable then begin
					OpenCPort(@cgp);
					InitCPort(@cgp);
					SetPort(@cgp);
				end
				else begin
					OpenPort(@gp);
					InitPort(@gp);
					SetPort(@gp);
				end;
{$IFC THINK_PASCAL }
				thePort^.portRect := scrapRect;
{$ELSEC }
				qd.thePort^.portRect := scrapRect;
{$ENDC }
				picture := OpenPicture(scrapRect);
				scrapBook.interactiveOutput;
				ClosePicture;
				if colorAvailable then
					CloseCPort(@cgp)
				else
					ClosePort(@gp);
				SetPort(savePort);
				clippingRectangle := save;
				currentFrame := saveFrame;
				sheetExtendProc := @ExtendSheet;
				ignore := PutScrap(GetHandleSize(Handle(picture)), 'PICT', Ptr(picture^));
				KillPicture(picture);
			end;

			dummy := SystemEdit(3); (* avoid multifinder bug ! *)

			stopwatchFreeAll(obj(scrapBook));
			if complexity >= maxStopWatchComplexity then
				endCursor;
			if complexity >= maxScrapBookComplexity then
				hideCenteredDialog(bigScrapBookDialog);
		end;
	end;

	var
		currentPoly, currentScrapbook: list;

procedure closeCurrentPoly;
	var
		pl: polyLine;
		pg: polygon2d;
		ls: lineSegment;
		o: obj;
	begin
		if currentPoly.length > 0 then begin
			if currentPoly.length = 2 then begin
				new(ls);
				failNil(Handle(ls));
				ls.p := point2d(currentPoly.first^^.value);
				ls.q := point2d(currentPoly.last^^.value);
				o := ls;
			end
			else begin
				if point2d(currentPoly.last^^.value).equalPoint(point2d(currentPoly.first^^.value)) then begin
					new(pg);
					failNil(Handle(pg));
					o := pg;
					currentPoly.first^^.value.free;
					currentPoly.pop;
				end
				else begin
					new(pl);
					failNil(Handle(pl));
					o := pl;
				end;
				vector(o).initFromList(currentPoly);
			end;
			o.windowToWorld;
			if (currentScrapbook.length = 0) | not currentScrapbook.last^^.value.equal(o) then
				currentScrapbook.append(o) (* MacDraw II 1.1 sometimes produces subsequent identical polygons *)
			else
				o.freeAll;
			currentPoly.free;
			new(currentPoly);
			failNil(Handle(currentPoly));
			currentPoly.init;
		end;
	end;

	type
		charArray = packed array[0..255] of char;
		charPtr = ^charArray;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoScrapText (byteCount: integer; textBuf: charPtr; numer, denom: Point);
(* convert text lines into marked real points *)
	var
		mp: markedRealPoint;
		pos: Point;
		s: Str255;
		i, l: integer;
	begin
		GetPen(pos);
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.x := pos.h;
		mp.y := pos.v;
		s := '';
		if byteCount > 250 then
			l := 250
		else
			l := byteCount;
		for i := 0 to l - 1 do
			s := concat(s, textBuf^[i]);
		if byteCount > l then
			s[l] := '';
		mp.mark.value := s;
		currentScrapbook.append(mp);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure DoScrapLine (newPt: Point);
(* convert line segment series into line segments, polyLines or polygons *)
	var
		oldPt: Point;
		lastp: point2d;

	procedure appendPt (pt: Point);
		var
			p: point2d;
		begin
			p := point2d(newObj(currentPointType));
			p.setX(pt.h);
			p.setY(pt.v);
			currentPoly.append(p);
		end;

	begin
		GetPen(oldPt);
		if not EqualPt(oldPt, newPt) then begin
			if currentPoly.length = 0 then begin
				appendPt(oldPt);
				appendPt(newPt);
			end
			else begin
				lastp := point2d(currentPoly.last^^.value);
				if (lastp.getX <> oldPt.h) | (lastp.getY <> oldPt.v) then begin
					closeCurrentPoly;
					appendPt(oldPt);
				end;
				appendPt(newPt);
				if point2d(currentPoly.last^^.value).equalPoint(point2d(currentPoly.first^^.value)) then
					closeCurrentPoly;
			end;
		end;
		advanceStopWatchHandsNoMem;
	end;

procedure DoScrapOval (verb: GrafVerb; r: Rect);
(* convert framed ovals into geoBench circles and all other (especially filled) ovals into geoBench points ('point2d') *)
	var
		p: point2d;
		c: circle;
	begin
{$IFC THINK_PASCAL }
		if verb = frame then begin
{$ELSEC }
		if verb = kQDGrafVerbFrame then begin
{$ENDC }
			if acceptCircles then begin
				new(c);
				failNil(Handle(c));
				c.x := (r.left + r.right) / 2;
				c.y := (r.top + r.bottom) / 2;
				c.radius := (r.right - r.left + r.bottom - r.top) / 4;
				c.windowToWorld;
				currentScrapbook.append(c);
			end;
		end
		else begin
			p := point2d(newObj(currentPointType));
			p.setX((r.left + r.right) / 2);
			p.setY((r.top + r.bottom) / 2);
			p.windowToWorld;
			currentScrapbook.append(p);
		end;
		advanceStopWatchHandsNoMem;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoScrapArc (verb: GrafVerb; r: Rect; startAngle, endAngle: integer);
(* convert arcs as ovals *)
	begin
		DoScrapOval(verb, r);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure DoScrapRect (verb: GrafVerb; r: Rect);
(* convert rectangles into rectangles *)
	var
		r2d: rectangle;
	begin
{$IFC THINK_PASCAL }
		if verb = frame then begin
{$ELSEC }
		if verb = kQDGrafVerbFrame then begin
{$ENDC }
			new(r2d);
			failNil(Handle(r2d));
			r2d.top := r.bottom;
			r2d.bottom := r.top - 1;
			r2d.left := r.left;
			r2d.right := r.right - 1;
			r2d.windowToWorld;
			currentScrapbook.append(r2d);
		end;
		advanceStopWatchHandsNoMem;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure DoScrapRRect (verb: GrafVerb; r: Rect; ovalWidth, ovalHeight: integer);
(* convert rounded rectangles into rectangles *)
	begin
		DoScrapRect(verb, r);
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure DoScrapBits (var srcBits: BitMap; var srcRect, dstRect: Rect; mode: integer; maskRgn: RgnHandle);
(* convert bit images into GeoBench's bitmap QuickDraw pictures (type 'QDPicture') *)
	var
		grafProcs: QDProcsPtr;
	begin
{$IFC THINK_PASCAL }
		grafProcs := thePort^.grafProcs;
		thePort^.grafProcs := nil;
{$ELSEC }
		grafProcs := qd.thePort^.grafProcs;
		qd.thePort^.grafProcs := nil;
{$ENDC }
		ShowPen;
		if (srcBits.rowBytes < 0) & colorAvailable then begin
			if OpenColorBMQDPicture(dstRect.left, dstRect.top, dstRect.right, dstRect.bottom, stdPalette) then begin
{$IFC THINK_PASCAL }
				CopyBits(srcBits, thePort^.portBits, srcRect, dstRect, mode, maskRgn);
{$ELSEC }
				CopyBits(srcBits, qd.thePort^.portBits, srcRect, dstRect, mode, maskRgn);
{$ENDC }
				currentScrapbook.append(CloseColorBMQDPicture);
			end;
		end
		else if (srcBits.rowBytes > 0) & OpenBMQDPicture(dstRect.left, dstRect.top, dstRect.right, dstRect.bottom) then begin
{$IFC THINK_PASCAL }
			CopyBits(srcBits, thePort^.portBits, srcRect, dstRect, mode, maskRgn);
{$ELSEC }
			CopyBits(srcBits, qd.thePort^.portBits, srcRect, dstRect, mode, maskRgn);
{$ENDC }
			currentScrapbook.append(CloseBMQDPicture);
		end;
		HidePen;
{$IFC THINK_PASCAL }
		thePort^.grafProcs := grafProcs;
{$ELSEC }
		qd.thePort^.grafProcs := grafProcs;
{$ENDC }
	end;

function loadScrapBook: list;
	var
		length, offset: longInt;
		streamStatusSave: streamStatus;
		picture: PicHandle;
		grafProcs: QDProcs;
		cGrafProcs: CQDProcs;
		textObj: obj;

	procedure startRedirect (r: ResType);
		begin
			beginCursor(stopwatchCursor);
			clipboardHandle := newHandleFailNil(length);
			length := GetScrap(clipboardHandle, r, offset);
			clipboardIndex := 0;
			getStreamStatus(streamStatusSave);
			setStreamStatus(createStreamStatus(@dummyFSOpen, nil, @clipboardFSRead, nil, @dummyFSClose, streamStatusSave.bufferSize));
			openReadStream('', 0);
		end;

	procedure endRedirect;
		begin
			closeReadStream;
			setStreamStatus(streamStatusSave);
{$IFC THINK_PASCAL }
			disposHandle(clipboardHandle);
{$ELSEC }
			DisposeHandle(clipboardHandle);
{$ENDC }
			endCursor;
		end;

	begin
		currentScrapbook := nil;
		length := GetScrap(nil, gbFileType, offset);

		if length > 0 then begin
			startRedirect(gbFileType);
			readLongInt(currentVersion);
			if goodVersion then begin
				new(currentScrapbook);
				failNil(Handle(currentScrapbook));
				currentScrapbook.fileInput;
			end;
			endRedirect;
		end;

		if currentScrapbook = nil then begin
			length := GetScrap(nil, 'TEXT', offset);
			if length > 0 then begin
				startRedirect('TEXT');
				textObj := readTextObjSpecial;
				if member(textObj, list) then
					currentScrapbook := list(textObj)
				else begin
					new(currentScrapbook);
					failNil(Handle(currentScrapbook));
					currentScrapbook.append(textObj);
				end;
				endRedirect;
			end;
		end;

		if currentScrapbook = nil then begin
			length := GetScrap(nil, 'PICT', offset);
			currentScrapbook := newEmptyList;
			if length > 0 then begin
				beginCursor(stopwatchCursor);
				picture := PicHandle(newHandleFailNil(length));
				length := GetScrap(Handle(picture), 'PICT', offset);
				new(currentPoly);
				failNil(Handle(currentPoly));
				currentPoly.init;
				currentPointType := arithmeticsToPointTypeRange(currentArithmetic);
				if colorAvailable then begin
					SetStdCProcs(cGrafProcs);
{$IFC THINK_PASCAL }
					thePort^.grafProcs := @cGrafProcs;
					cGrafProcs.textProc := @DoScrapText;
					cGrafProcs.lineProc := @DoScrapLine;
					cGrafProcs.ovalProc := @DoScrapOval;
					cGrafProcs.rectProc := @DoScrapRect;
					cGrafProcs.rRectProc := @DoScrapRRect;
					cGrafProcs.arcProc := @DoScrapArc;
					cGrafProcs.bitsProc := @DoScrapBits;
{$ELSEC }
					qd.thePort^.grafProcs := @cGrafProcs;
					cGrafProcs.textProc := NewQDTextProc(ProcPtr(@DoScrapText));
					cGrafProcs.lineProc := NewQDLineProc(@DoScrapLine);
					cGrafProcs.ovalProc := NewQDOvalProc(ProcPtr(@DoScrapOval));
					cGrafProcs.rectProc := NewQDRectProc(ProcPtr(@DoScrapRect));
					cGrafProcs.rRectProc := NewQDRRectProc(ProcPtr(@DoScrapRRect));
					cGrafProcs.arcProc := NewQDArcProc(ProcPtr(@DoScrapArc));
					cGrafProcs.bitsProc := NewQDBitsProc(@DoScrapBits);
{$ENDC }
				end
				else begin
					SetStdProcs(grafProcs);
{$IFC THINK_PASCAL }
					thePort^.grafProcs := @grafProcs;
					grafProcs.textProc := @DoScrapText;
					grafProcs.lineProc := @DoScrapLine;
					grafProcs.ovalProc := @DoScrapOval;
					grafProcs.rectProc := @DoScrapRect;
					grafProcs.rRectProc := @DoScrapRRect;
					grafProcs.arcProc := @DoScrapArc;
					grafProcs.bitsProc := @DoScrapBits;
{$ELSEC }
					qd.thePort^.grafProcs := @grafProcs;
					grafProcs.textProc := NewQDTextProc(ProcPtr(@DoScrapText));
					grafProcs.lineProc := NewQDLineProc(@DoScrapLine);
					grafProcs.ovalProc := NewQDOvalProc(ProcPtr(@DoScrapOval));
					grafProcs.rectProc := NewQDRectProc(ProcPtr(@DoScrapRect));
					grafProcs.rRectProc := NewQDRRectProc(ProcPtr(@DoScrapRRect));
					grafProcs.arcProc := NewQDArcProc(ProcPtr(@DoScrapArc));
					grafProcs.bitsProc := NewQDBitsProc(@DoScrapBits);
{$ENDC }
				end;
				HidePen; (* don't really draw it *)
				DrawPicture(picture, picture^^.picFrame);
				ShowPen;
{$IFC THINK_PASCAL }
				thePort^.grafProcs := nil; (*standard output *)
{$ELSEC }
				qd.thePort^.grafProcs := nil; (*standard output *)
{$ENDC }
				closeCurrentPoly;
				currentPoly.free;
				KillPicture(picture);
				endCursor;
			end;
		end;

		loadScrapBook := currentScrapbook;
	end;

end.
