(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit getTriangulation;

interface

uses
	unionFind, testGraph, infiniteUndirectedGraph;

function getNumberList (s: Str255; var v: homogenVector; var k: longInt): boolean;

procedure initGetTriangulation;

implementation

{$SETC AgetTriangulation = TRUE }
{$SETC AMST = TRUE }
{$SETC ATravelingSalesman = TRUE }

function simpleUndirectedGraph.graphIsConnected: boolean;
	var
		ug: undirectedGraph;
		dg: directedGraph;
		startSet: list;
		additionOrder, reverseAdditionOrder: homogenVector;
	begin
		if vertices.length < 2 then
			graphIsConnected := true
		else begin
			startSet := s(1, nil);
			new(ug);
			failNil(Handle(ug));
			ug.initializeFromSimpleUndirectedGraph(self);
			dg := ug.getTriangulation(startSet, 1, chooseLast, additionOrder, reverseAdditionOrder);
			graphIsConnected := additionOrder.length = vertices.length;
			additionOrder.freeAll;
			reverseAdditionOrder.freeAll;
			dg.freeAll;
			ug.freeAll;
			startSet.freeAll;
		end;
	end;

function simpleUndirectedGraph.isConnected (s, e: seqIndex): boolean;

	function edgeExists (ge: graphEdge): boolean;
		begin
			edgeExists := (ge.startVertex = s) & (ge.endVertex = e) | (ge.startVertex = e) & (ge.endVertex = s);
		end;

	begin
		isConnected := edges.exists(edgeExists) > 0;
	end;

function directedGraph.isConnected (s, e: seqIndex): boolean;

	function edgeExists (ge: graphEdge): boolean;
		begin
			edgeExists := ge.endVertex = e;
		end;

	begin (* directedGraph.isConnected *)
		isConnected := list(adjacency.elements^^[s]).exists(edgeExists) > 0;
	end;

function simpleUndirectedGraph.minimumSpanningTree (k: seqIndex; var uf: unionFindSimple): spanningTree;
	label
		99;
	var
		mst: spanningTree;
		i, j: seqIndex0;
		e: graphEdge;
	begin
{$IFC AnimationEnabled AND AMST }
		if animationFlag[AMinimumSpanningTreeItem] then
			vertices.interactiveHighlight;
{$ENDC }
		currentVertices := vertices.length;
		currentEdges := vertices.length - 1;
		new(mst);
		failNil(Handle(mst));
		mst.init;
		for i := 1 to vertices.length do
			mst.vertices.elements^^[i] := vertices.elements^^[i].duplicateAll;
		mst.vertices.length := vertices.length;
		uf := createUnionFindSimple(vertices.length);
		j := 0;
		for i := 1 to vertices.length - k do begin
			repeat
{$IFC AnimationEnabled }
				if userAbort then begin
					uf.free;
					mst.freeAll;
					minimumSpanningTree := nil;
					exit(minimumSpanningTree);
				end;
{$ENDC }
				if j = edges.length then
					goto 99; (* no edges left means we are done *)
				j := j + 1;
				e := graphEdge(edges.elements^^[j]);
{$IFC AnimationEnabled AND AMST }
				if animationFlag[AMinimumSpanningTreeItem] then begin
					e.drawEdge(vertices, false);
					waitForClick(animationFlag[AMinimumSpanningTreeItem]);
					e.drawEdge(vertices, false);
				end;
{$ENDC }
			until uf.find(e.startVertex) <> uf.find(e.endVertex);
{$IFC AnimationEnabled AND AMST }
			if animationFlag[AMinimumSpanningTreeItem] then
				e.drawEdge(vertices, false);
{$ENDC }
			uf.union(e.startVertex, e.endVertex);
			mst.addEdge(graphEdge(e.duplicate));
		end;
99:
		if k = 1 then
			uf.free;
		minimumSpanningTree := mst;
	end;

function spanningTree.memberType (o: obj): boolean;
	begin
		memberType := member(o, spanningTree);
	end;

function spanningTree.TSPEMST: vector;
	var
		udg: undirectedGraph;
		result: vector;
		visitedH: booleanArrayH;
		i: seqIndex0;
{$IFC AnimationEnabled AND ATravelingSalesman }
		x1, y1, x2, y2: extended;
{$ENDC }

	procedure traverse (tree: undirectedGraph; father: seqIndex0; root: seqIndex; route: vector; var visited: booleanArray);
		var
			p: elementH;
			next: seqIndex;
{$IFC AnimationEnabled AND ATravelingSalesman }
			x1, y1, x2, y2: extended;
{$ENDC }

		begin
			p := list(tree.adjacency.elements^^[root]).first;
			while p <> nil do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					result.freeAll;
{$IFC THINK_PASCAL }
					disposHandle(Handle(visitedH));
{$ELSEC }
					DisposeHandle(Handle(visitedH));
{$ENDC }
					udg.freeAll;
					TSPEMST := nil;
					exit(TSPEMST);
				end;
{$ENDC }
				next := graphEdge(p^^.value).endVertex;
				if next <> father then begin
{$IFC AnimationEnabled AND ATravelingSalesman }
					if animationFlag[ATravelingSalesmanItem] then begin
						tree.vertices.elements^^[next].interactiveFlash;
						if visited[next] then
							waitForClick(animationFlag[ATravelingSalesmanItem]);
					end;
{$ENDC }
					if not visited[next] then begin
						route.append(tree.vertices.elements^^[next].duplicate);
						visited[next] := true;
{$IFC AnimationEnabled AND ATravelingSalesman }
						if animationFlag[ATravelingSalesmanItem] then begin
							route.elements^^[route.length].centerOfGravity(x1, y1);
							route.elements^^[route.length - 1].centerOfGravity(x2, y2);
							drawThickLineSegment(x1, y1, x2, y2);
							waitForClick(animationFlag[ATravelingSalesmanItem]);
						end;
{$ENDC }
					end;
					traverse(tree, root, next, route, visited);
{$IFC AnimationEnabled AND ATravelingSalesman }
					if animationFlag[ATravelingSalesmanItem] then begin
						tree.vertices.elements^^[root].interactiveFlash;
						if root > 1 then
							waitForClick(animationFlag[ATravelingSalesmanItem]);
					end;
{$ENDC }
				end;
				p := p^^.next;
			end;
		end;

	begin
{$IFC AnimationEnabled AND ATravelingSalesman }
		if animationFlag[ATravelingSalesmanItem] then
			interactiveHighlight;
{$ENDC }
		currentVertices := vertices.length;
		currentEdges := edges.length;
		new(udg);
		failNil(Handle(udg));
		udg.init;
		for i := 1 to vertices.length do
			udg.vertices.elements^^[i] := vertices.elements^^[i].duplicateAll;
		udg.vertices.length := vertices.length;
		for i := 1 to edges.length do
			udg.edges.elements^^[i] := edges.elements^^[i].duplicateAll;
		udg.edges.length := edges.length;
		udg.initAdjacency;
		result := newVector(vertices.length);
		visitedH := booleanArrayH(dynamicArray(vertices.length, sizeOf(boolean)));
		for i := 2 to vertices.length do
			visitedH^^[i] := false;
		result.append(udg.vertices.elements^^[1].duplicate);
		visitedH^^[1] := true;
{$IFC AnimationEnabled AND ATravelingSalesman }
		if animationFlag[ATravelingSalesmanItem] then begin
			vertices.elements^^[1].interactiveFlash;
			waitForClick(animationFlag[ATravelingSalesmanItem]);
		end;
{$ENDC }
		traverse(udg, 0, 1, result, visitedH^^);
{$IFC AnimationEnabled AND ATravelingSalesman }
		if animationFlag[ATravelingSalesmanItem] then begin
			result.elements^^[result.length].centerOfGravity(x1, y1);
			result.elements^^[1].centerOfGravity(x2, y2);
			drawThickLineSegment(x1, y1, x2, y2);
			waitForClick(animationFlag[ATravelingSalesmanItem]);
		end;
{$ENDC }
{$IFC THINK_PASCAL }
		disposHandle(Handle(visitedH));
{$ELSEC }
		DisposeHandle(Handle(visitedH));
{$ENDC }
		udg.freeAll;
		TSPEMST := result;
	end;

	const
		TravelingSalesmanC = 1;

procedure spanningTree.describe;
	begin
		with currentArguments do
			if (length = 1) & member(ith(1), spanningTree) then
				menuEntry('Traveling Salesman', TravelingSalesmanC);
	end;

procedure spanningTree.execute (item: integer);

	function callTravelingSalesman (s: spanningTree): vector;
		var
			v, poly: vector;
		begin
			v := s.TSPEMST;
			if (v <> nil) & (v.length > 0) & member(v.ith(1), point2d) then begin
				poly := vector(newObj(polygon2dType));
				poly.length := v.length;
				poly.elements := v.elements;
				dispose(v);
				callTravelingSalesman := poly;
			end
			else
				callTravelingSalesman := v;
		end;

	begin
		with currentArguments do
			case item of
				TravelingSalesmanC: 
					currentResult.appendNonNil(callTravelingSalesman(spanningTree(ith(1))));
			end;
	end;

procedure markedGraphEdge.init;
	var
		s: strings;
	begin
		new(s);
		failNil(Handle(s));
		s.init;
		mark := s;
	end;

procedure markedGraphEdge.free;
	begin
		mark.free;
		shallowFree;
	end;

function markedGraphEdge.duplicate: obj;
	var
		g: markedGraphEdge;
	begin
		g := markedGraphEdge(shallowDuplicate);
		g.mark := strings(mark.duplicate);
		duplicate := g;
	end;

procedure markedGraphEdge.textFileInput;
	var
		s: strings;
	begin
		inherited textFileInput;
		new(s);
		failNil(Handle(s));
		mark := s;
		mark.textFileInput;
	end;

procedure markedGraphEdge.textFileOutput;
	begin
		inherited textFileOutput;
		mark.textFileOutput;
	end;

procedure markedGraphEdge.fileInput;
	var
		s: strings;
	begin
		inherited fileInput;
		new(s);
		failNil(Handle(s));
		mark := s;
		mark.fileInput;
	end;

procedure markedGraphEdge.fileOutput;
	begin
		inherited fileOutput;
		mark.fileOutput;
	end;

function markedGraphEdge.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, markedGraphEdge) & inherited equal(o) & mark.equal(markedGraphEdge(o).mark);
	end;

function markedGraphEdge.memberType (o: obj): boolean;
	begin
		memberType := member(o, markedGraphEdge);
	end;

function markedGraphEdge.getType: objectType;
	begin
		getType := markedGraphEdgeType;
	end;

function markedGraphEdge.genericLessThan (o: obj): boolean;
	begin
		genericLessThan := inherited genericLessThan(o) | inherited equal(o) & (mark.value < markedGraphEdge(o).mark.value);
	end;

function markedGraphEdge.getTextType: Str255;
	begin
		getTextType := 'Marked Graph Edge';
	end;

function markedGraphEdge.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoObject(v, 'Mark', mark, false);
		getInfo := v;
	end;

procedure directedGraph.initAdjacency;
	var
		i: seqIndex0;
	begin
		adjacency.length := vertices.length;
		for i := 1 to vertices.length do
			adjacency.elements^^[i] := newEmptyList;
		if edges <> nil then
			for i := 1 to edges.length do
				addAdjacency(graphEdge(edges.elements^^[i]));
	end;

procedure directedGraph.exitAdjacency;
	begin
		adjacency.forAll(freeP);
	end;

procedure undirectedGraph.exitAdjacency;
	var
		tempEdges: vector;
	begin
		tempEdges := vector(edges.duplicateAll);
		adjacency.forAll(freeAllP);
		edges.free;
		edges := tempEdges;
	end;

procedure simpleUndirectedGraph.deleteVertex (v: seqIndex);
	var
		l: seqIndex0;
		i: seqIndex1;
		x: graphEdge;

	procedure fixEdge (x: graphEdge);
		begin
			if x.startVertex = l then
				x.startVertex := v;
			if x.endVertex = l then
				x.endVertex := v;
		end;

	begin
		i := 1;
		while i <= edges.length do begin
			x := graphEdge(edges.elements^^[i]);
			if (x.startVertex = v) | (x.endVertex = v) then
				deleteEdgeSimple(i)
			else
				i := i + 1;
		end;
		l := vertices.length;
		if v < l then begin
			vertices.swap(v, l);
			edges.forAll(fixEdge);
		end;
		vertices.elements^^[l].freeAll;
		vertices.length := l - 1;
	end;

procedure directedGraph.deleteVertex (v: seqIndex);
	begin
		exitAdjacency;
		inherited deleteVertex(v);
		initAdjacency;
	end;

procedure simpleUndirectedGraph.deleteEdgeSimple (e: seqIndex);
	begin
		edges.swap(e, edges.length);
		edges.elements^^[edges.length].freeAll;
		edges.length := edges.length - 1;
	end;

procedure simpleUndirectedGraph.deleteEdge (e: seqIndex);
	begin
		deleteEdgeSimple(e);
	end;

procedure directedGraph.deleteEdge (e: seqIndex);
	begin
		exitAdjacency;
		deleteEdgeSimple(e);
		initAdjacency;
	end;

function vector.interactiveSelect (unique, beep: boolean; function getIndex (x: seqIndex): seqIndex; low: seqIndex; high: seqIndex0): seqIndex0;
	const
		margin = 4;
	var
		p: realPoint;
		r: rectangle;
		count: seqIndex0;
		mouse: Point;
		i: seqIndex1;
		myPort, where: GrafPtr;
		userClicked: boolean;
	begin
		assert(high <= length, 'high  length');
		if low > high then begin
			interactiveSelect := 0;
			exit(interactiveSelect);
		end;
		new(p);
		failNil(Handle(p));
		new(r);
		failNil(Handle(r));
		beginCursor(littleCrossCursor);
		GetPort(myPort);
		repeat
			repeat
				if AbortInQueueUserClick(userClicked, where, mouse) | abortFlag then begin
					endCursor;
					r.free;
					p.free;
					interactiveSelect := 0;
					exit(interactiveSelect);
				end;
			until userClicked & (where = myPort);
			p.windowInputMouse(mouse);
			FlushEvents(mUpMask + mDownMask, 0);
			r.left := p.x - margin;
			r.right := p.x + margin;
			r.bottom := p.y - margin;
			r.top := p.y + margin;
			r.windowToWorld;
			count := 0;
			i := low;
			while (i <= high) & (unique | (count = 0)) do begin
				if elements^^[getIndex(i)].isSelected(r) then begin
					count := count + 1;
					interactiveSelect := i;
				end;
				i := i + 1;
			end;
			if beep & unique & (count > 1) then
				SysBeep(1);
		until unique & (count = 1) | not unique & (count > 0);
		endCursor;
		r.free;
		p.free;
	end;

function vector.interactiveSelectFromAll (unique, beep: boolean): seqIndex0;
	begin
		interactiveSelectFromAll := interactiveSelect(unique, beep, identityIndex, 1, length);
	end;

procedure revertEdge (x: graphEdge);
	var
		t: seqIndex;
	begin
		t := x.startVertex;
		x.startVertex := x.endVertex;
		x.endVertex := t;
	end;

procedure directedGraph.revertEdges;
	var
		tempEdges: vector;
	begin
		tempEdges := vector(edges.duplicateAll);
		adjacency.forAll(freeAllP);
		edges.free;
		edges := tempEdges;
		edges.forAll(revertEdge);
		initAdjacency;
	end;

function simpleUndirectedGraph.breadthFirstSearch (root: seqIndex): simpleUndirectedGraph;
	var
		queue: seqIndexArrayH;
		topQueue, bottomQueue: seqIndex1;
		distances: seqIndexArray0H;
		i, lengthQueue, gl, distance, diameter: seqIndex0;
		first: seqIndex;
		result: simpleUndirectedGraph;
		mp: markedRealPoint;
		p: point2d;
		pred, succ: homogenVector;

	procedure findMinDistance (x: graphEdge);
		var
			d: seqIndex0;
		begin
			d := distances^^[x.endVertex];
			if d < distance then
				distance := d;
		end;

	procedure findSucc (x: graphEdge);
		begin
			if distances^^[x.endVertex] = gl then begin
				topQueue := topQueue + 1;
				if topQueue > gl then
					topQueue := 1;
				queue^^[topQueue] := x.endVertex;
				lengthQueue := lengthQueue + 1;
			end;
		end;

	procedure setPredAndSucc (x: graphEdge);
		var
			rx: graphEdge;
		begin
			rx := graphEdge(x.duplicate);
			rx.startVertex := x.endVertex;
			rx.endVertex := x.startVertex;
			list(succ.elements^^[x.startVertex]).append(x.duplicate);
			list(pred.elements^^[rx.startVertex]).append(rx);
		end;

	begin
		gl := vertices.length;
		new(pred);
		failNil(Handle(pred));
		pred.elements := nil;
		pred.preset(gl, listType);
		succ := homogenVector(pred.duplicateAll);
		edges.forAll(setPredAndSucc);
		queue := seqIndexArrayH(dynamicArray(gl, sizeOf(seqIndex)));
		bottomQueue := 1;
		topQueue := 1;
		lengthQueue := 1;
		queue^^[1] := root;
		distances := seqIndexArray0H(dynamicArray(gl, sizeOf(longInt)));
		for i := 1 to gl do
			distances^^[i] := gl;

		while lengthQueue > 0 do begin
			first := queue^^[bottomQueue];
			bottomQueue := bottomQueue + 1;
			if bottomQueue > gl then
				bottomQueue := 1;
			lengthQueue := lengthQueue - 1;
			if distances^^[first] = gl then begin
				distance := gl;
				list(pred.elements^^[first]).forAll(findMinDistance);
				if first = root then
					distances^^[first] := 0
				else
					distances^^[first] := distance + 1;
				list(succ.elements^^[first]).forAll(findSucc);
			end;
		end;

		result := simpleUndirectedGraph(duplicateAll);
		diameter := 0;
		for i := 1 to gl do begin
			p := point2d(result.vertices.elements^^[i]);
			new(mp);
			failNil(Handle(mp));
			mp.init;
			distance := distances^^[i];
			if distance > diameter then
				diameter := distance;
			if distance = gl then
				mp.mark.value := ''
			else
				mp.mark.value := stringOf(distance : 1);
			mp.x := p.getX;
			mp.y := p.getY;
			p.free;
			result.vertices.elements^^[i] := mp;
		end;
		displayInfoString(StringOf('Diameter = ', diameter : 1));
		breadthFirstSearch := result;
		pred.freeAll;
		succ.freeAll;
{$IFC THINK_PASCAL }
		disposHandle(Handle(queue));
		disposHandle(Handle(distances));
{$ELSEC }
		DisposeHandle(Handle(queue));
		DisposeHandle(Handle(distances));
{$ENDC }
	end;

function listCompare (a, b: list): relationObj;
	begin
		if a.length < b.length then
			listCompare := greaterThanObj
		else if a.length > b.length then
			listCompare := lessThanObj
		else
			listCompare := equalObj;
	end;

function lexicographicGreater (a1, a2, b1, b2, c1, c2: longInt): boolean;
	begin
		lexicographicGreater := (a1 > a2) | (a1 = a2) & ((b1 > b2) | (b1 = b2) & (c1 > c2));
	end;

function directedGraph.getTriangulation (startSet: sequence; k: seqIndex; choiceMethod: integer; var additionOrder, reverseAdditionOrder: homogenVector): directedGraph;
	var
		result: directedGraph;
		visited, stack, levelSums: homogenVector;
		i, n, chosenIndex, thisLength, extremeLength, thisDegree, extremeDegree: seqIndex0;
		thisLevel, extremeLevel: longInt;

	function sumLevel (l: list): longInt;
		var
			result: longInt;

		procedure doSum (x: graphEdge);
			begin
				result := result + int(reverseAdditionOrder.elements^^[x.endVertex]).value;
			end;

		begin (* sumLevel *)
			result := 0;
			l.forAll(doSum);
			sumLevel := result;
		end;

	procedure processVertex (v: seqIndex);

		procedure doProcess (e: graphEdge);
			var
				l: list;
				g: graphEdge;
				w: seqIndex;
			begin
				if e.startVertex = v then
					w := e.endVertex
				else
					w := e.startVertex;
				if visited.elements^^[w] <> nil then begin
					l := list(visited.elements^^[w]);
					g := graphEdge(e.duplicate);
					g.startVertex := w;
					g.endVertex := v;
					l.append(g);
					if l.length = k then begin
						stack.append(l);
						levelSums.elements^^[w] := makeInt(sumLevel(l));
					end;
				end;
			end;

		procedure doAddEdge (e: graphEdge);
			begin
				result.addEdge(graphEdge(e.duplicate));
{$IFC AnimationEnabled AND AgetTriangulation }
				if animationFlag[ATriangulationItem] then begin
					PenSize(3, 3);
					e.drawEdge(vertices, false);
					PenSize(1, 1);
					e.drawEdge(vertices, false);
				end;
{$ENDC }
			end;

		begin (* processVertex *)
			additionOrder.appendInt(v);
			reverseAdditionOrder.elements^^[v] := makeInt(additionOrder.length);
			with list(visited.elements^^[v]) do begin
				forSome(doAddEdge, 1, k);
				freeAll;
			end;
			visited.elements^^[v] := nil;	(* i.e. was visited *)
			list(adjacency.elements^^[v]).forAll(doProcess);
		end;

	procedure processVertexObject (i: int);
		begin
			processVertex(i.value);
		end;

	function getLevelSum (i: seqIndex): longInt;
		begin
			getLevelSum := int(levelSums.elements^^[graphEdge(list(stack.elements^^[i]).ith(1)).startVertex]).value;
		end;

	function getDegree (i: seqIndex): seqIndex;
		begin
			getDegree := list(adjacency.elements^^[graphEdge(list(stack.elements^^[i]).ith(1)).startVertex]).length;
		end;

	function incidentVertices (l: list): Str255;
		var
			result: Str255;

		procedure appendToResult (e: graphEdge);
			begin
				if result = '' then
					result := StringOf(e.endVertex : 1)
				else
					result := concat(result, ',', StringOf(e.endVertex : 1));
			end;

		begin (* incidentVertices *)
			result := '';
			l.forAll(appendToResult);
			incidentVertices := result;
		end;

	begin (* directedGraph.getTriangulation *)
{$IFC AnimationEnabled AND AgetTriangulation }
		if animationFlag[ATriangulationItem] then
			interactiveHighlight;
{$ENDC }
		new(result);
		failNil(Handle(result));
		result.vertices := vector(vertices.duplicateAll);
		n := vertices.length;
		result.edges := newHomogenVector(k * n);
		result.adjacency := newHomogenVector(n);
		result.initAdjacency;
		additionOrder := newHomogenVector(n);
		reverseAdditionOrder := newHomogenVector(n);
		reverseAdditionOrder.length := n;
		levelSums := newHomogenVector(n);
		levelSums.length := n;
		for i := 1 to n do begin
			reverseAdditionOrder.elements^^[i] := nil;
			levelSums.elements^^[i] := nil;
		end;

		stack := newHomogenVector(n);
		visited := newHomogenVector(n);
		for i := 1 to vertices.length do
			visited.append(newEmptyList);
		startSet.forAll(processVertexObject);

		while stack.length > 0 do begin

			case choiceMethod of
				chooseMaxDegree:  begin
					extremeLength := list(stack.elements^^[1]).length;
					chosenIndex := 1;
					for i := 2 to stack.length do begin
						thisLength := list(stack.elements^^[i]).length;
						if thisLength > extremeLength then begin
							extremeLength := thisLength;
							chosenIndex := i;
						end;
					end;
				end;
				chooseBestLevel:  begin
					extremeLength := list(stack.elements^^[1]).length;
					extremeLevel := getLevelSum(1);
					extremeDegree := getDegree(1);
					chosenIndex := 1;
					for i := 2 to stack.length do begin
						thisLength := list(stack.elements^^[i]).length;
						thisLevel := getLevelSum(i);
						thisDegree := getDegree(i);
						if lexicographicGreater(thisLength, extremeLength, thisLevel, extremeLevel, extremeDegree, thisDegree) then begin
							extremeLength := thisLength;
							extremeLevel := thisLevel;
							extremeDegree := thisDegree;
							chosenIndex := i;
						end;
					end;
				end;
				chooseMinDegree:  begin
					extremeLength := list(stack.elements^^[1]).length;
					chosenIndex := 1;
					for i := 2 to stack.length do begin
						thisLength := list(stack.elements^^[i]).length;
						if thisLength < extremeLength then begin
							extremeLength := thisLength;
							chosenIndex := i;
						end;
					end;
				end;
				chooseLast: 
					chosenIndex := stack.length;
				chooseRandom: 
					chosenIndex := randomRange(1, stack.length);
			end;

			stack.swap(stack.length, chosenIndex);
			i := graphEdge(list(stack.elements^^[stack.length]).ith(1)).startVertex;
			stack.length := stack.length - 1;
{$IFC AnimationEnabled AND AgetTriangulation }
			if animationFlag[ATriangulationItem] then begin
				displayInfoString(StringOf('V=', i : 1, '. I=(', incidentVertices(list(visited.elements^^[i])), ').'));
			end;
{$ENDC }
			processVertex(i);
{$IFC AnimationEnabled AND AgetTriangulation }
			if animationFlag[ATriangulationItem] then
				waitForClick(animationFlag[ATriangulationItem]);
{$ENDC }
		end;


		stack.free;
		visited.freeAll;
		levelSums.freeAll;

		getTriangulation := result;
	end;

procedure pop (var s: Str255);
	var
		p: integer;
	begin
		p := pos(' ', s);
		if p = 0 then
			s := ''
		else
			delete(s, 1, pos(' ', s));
	end;

function getNumberList (s: Str255; var v: homogenVector; var k: longInt): boolean;
	var
		w: seqIndex;
		l: list;
	begin
		IOCheck(false);
		readString(s, k);
		if ioResult > 0 then
			getNumberList := false
		else begin
			l := newEmptyList;
			pop(s);
			while s <> '' do begin
				readString(s, w);
				pop(s);
				if ioResult > 0 then begin
					IOCheck(true);
					v.free;
					getNumberList := false;
					exit(getNumberList);
				end;
				l.appendInt(w);
			end;
			new(v);
			failNil(Handle(v));
			v.initFromList(l);
			l.free;
			getNumberList := true;
		end;
		IOCheck(true);
	end;

function doFindTriangulation (g: simpleUndirectedGraph; v: sequence; k: longInt): directedGraph;
	var
		order, reverseOrder: homogenVector;
		h: undirectedGraph;
		result: directedGraph;
	begin
		if member(g, directedGraph) then
			result := directedGraph(g).getTriangulation(v, k, chooseLast, order, reverseOrder)
		else begin
			new(h);
			failNil(Handle(h));
			h.initializeFromSimpleUndirectedGraph(g);
			result := h.getTriangulation(v, k, chooseLast, order, reverseOrder);
			h.freeAll;
		end;
		displayInfoString(StringOf('Visited ', order.length : 1, ' of ', result.vertices.length : 1, '.'));
		doFindTriangulation := result;
		v.freeAll;
		order.freeAll;
		reverseOrder.freeAll;
	end;

function callTestTriangulation (g: simpleUndirectedGraph; inString: Str255): directedGraph;
	var
		k: longInt;
		v: homogenVector;
	begin
		if getNumberList(inString, v, k) then
			if (v.length = 0) & (0 < k) & (k <= g.edges.length) then begin
				v.freeAll;
				with graphEdge(g.edges.ith(k)) do
					callTestTriangulation := doFindTriangulation(g, s(startVertex, s(endVertex, nil)), 2);
			end
			else
				callTestTriangulation := doFindTriangulation(g, v, k)
		else
			callTestTriangulation := nil;
	end;

function callTestTriangulationAsk (g: simpleUndirectedGraph): directedGraph;
	var
		sel: seqIndex0;
	begin
		setCurrentVerticesVector(g.vertices);
		displayInfoStringCommand('Click at an edge!');
		sel := g.edges.interactiveSelectFromAll(true, true);
		displayInfoString('');
		if sel = 0 then
			callTestTriangulationAsk := nil
		else
			with graphEdge(g.edges.ith(sel)) do
				callTestTriangulationAsk := doFindTriangulation(g, s(startVertex, s(endVertex, nil)), 2);
	end;

function callTestTriangulationEnumerate (g: simpleUndirectedGraph): vector;	(* in fact a pointVector *)
	var
		triangulationOrder, reverseTriangulationOrder, pv: homogenVector;
		h: undirectedGraph;
		dg, triangulationGraph: directedGraph;
		mrp: markedRealPoint;
		l, startSet: list;
		i: seqIndex0;
	begin
		if member(g, directedGraph) then
			dg := directedGraph(g)
		else begin
			new(h);
			failNil(Handle(h));
			h.initializeFromSimpleUndirectedGraph(g);
			dg := h;
		end;
		l := newEmptyList;
		for i := 1 to dg.edges.length do begin
			with graphEdge(dg.edges.ith(i)) do
				startSet := s(startVertex, s(endVertex, nil));
{$IFC AnimationEnabled AND AgetTriangulation }
			if animationFlag[ATriangulationItem] & (i > 1) then begin
				clearCurrentGeoBenchWindow;
				dg.interactiveOutput;
				dg.interactiveHighlight;
			end;
{$ENDC }
			triangulationGraph := dg.getTriangulation(startSet, 2, chooseLast, triangulationOrder, reverseTriangulationOrder);
			if triangulationOrder.length = dg.vertices.length then begin
				new(mrp);
				failNil(Handle(mrp));
				mrp.init;
				mrp.setXY(100, 100 + 30 * l.length);
				mrp.mark.value := StringOf(i : 1);
				l.append(mrp);
			end;
			startSet.freeAll;
			triangulationGraph.freeAll;
			triangulationOrder.freeAll;
			reverseTriangulationOrder.freeAll;
		end;
		if not member(g, directedGraph) then
			h.freeAll;
		pv := homogenVector(newObj(pointVectorType));
		pv.initFromList(l);
		l.free;
		callTestTriangulationEnumerate := pv;
	end;

function directedGraph.maxTriangulationOrder (var startEdge: seqIndex0): seqIndex0;
	var
		triangulationGraph: directedGraph;
		triangulationOrder, reverseTriangulationOrder: homogenVector;
		maxOrder: seqIndex0;
		startSet: list;
		i: seqIndex01;
	begin
		maxOrder := 0;
		i := 0;
		while (i < edges.length) & (maxOrder <> vertices.length) do begin
{$IFC AnimationEnabled }
			if userAbort then
				leave;
{$ENDC }
			i := i + 1;
			with graphEdge(edges.ith(i)) do
				startSet := s(startVertex, s(endVertex, nil));
			triangulationGraph := getTriangulation(startSet, 2, chooseLast, triangulationOrder, reverseTriangulationOrder);
			if triangulationOrder.length > maxOrder then begin
				maxOrder := triangulationOrder.length;
				startEdge := i;
			end;
			startSet.freeAll;
			triangulationGraph.freeAll;
			triangulationOrder.freeAll;
			reverseTriangulationOrder.freeAll;
		end;
		maxTriangulationOrder := maxOrder;
	end;

function sugOk (g: simpleUndirectedGraph): boolean;
	begin
		sugOk := (g.vertices.length > 0) & (g.edges.length = 2 * g.vertices.length - 3);
	end;

	var
		trialNumber, maxDegreeHennebergReduction: extended;


	const
		stdTrialNumber = 10;
		stdMaxDegree = 2;
		TestTriangulationC = 1;
		TestTriangulationAskC = 2;
		TestTriangulationEnumerateC = 3;
		IsGraphOkC = 0;
		HennebergSequenceAskC = 4;
		HennebergReductionAskC = 5;

procedure undirectedGraph.describe;
	begin
		with currentArguments do begin
			if (length = 2) & ((member(ith(1), directedGraph) | member(ith(1), simpleUndirectedGraph)) & member(ith(2), markedRealPoint) | (member(ith(2), directedGraph) | member(ith(2), simpleUndirectedGraph)) & member(ith(1), markedRealPoint)) then
				menuEntry('Test Triangulation', TestTriangulationC);
			if (length = 1) & (member(ith(1), directedGraph) | member(ith(1), simpleUndirectedGraph)) then begin
				menuEntry('Test Triangulation Ask...', TestTriangulationAskC);
				menuEntry('Test Triangulation Enumerate', TestTriangulationEnumerateC);
			end;
			if (length = 1) & member(ith(1), directedGraph) then
				menuEntry('Test Graph With Formula...', IsGraphOkC);
			if (length = 1) & member(ith(1), simpleUndirectedGraph) & sugOk(simpleUndirectedGraph(ith(1))) then
				menuEntry('Henneberg Sequence Ask...', HennebergSequenceAskC);
			if (length = 1) & member(ith(1), simpleUndirectedGraph) then
				menuEntry('Henneberg Reduction Ask...', HennebergReductionAskC);
		end;
	end;

procedure undirectedGraph.execute (item: integer);
	var
		errorPosition: integer;
		dummy: boolean;
	begin
		with currentArguments do
			case item of
				HennebergReductionAskC: 
					if getUserParameter1('Remove All Vertices With Degree ', maxDegreeHennebergReduction, maxDegreeHennebergReduction, @isPositiveOrZeroInteger) then
{$IFC THINK_PASCAL }
						currentResult.appendNonNil(computeHennebergKReduction(simpleUndirectedGraph(ith(1)), trunc(maxDegreeHennebergReduction)));
{$ELSEC }
						currentResult.appendNonNil(computeHennebergKReduction(simpleUndirectedGraph(ith(1)), truncLong(maxDegreeHennebergReduction)));
{$ENDC }
				HennebergSequenceAskC: 
					if getUserParameter1('Number of Trials', trialNumber, trialNumber, @isNatural) then
{$IFC THINK_PASCAL }
						currentResult.appendNonNil(findHennebergSequenceMaxTrials(simpleUndirectedGraph(ith(1)), trunc(trialNumber)));
{$ELSEC }
						currentResult.appendNonNil(findHennebergSequenceMaxTrials(simpleUndirectedGraph(ith(1)), truncLong(trialNumber)));
{$ENDC }
				IsGraphOkC: 
					if getUserString1('Formula', defaultFormula) then
						if isGraphOk(directedGraph(ith(1)), defaultFormula, true, errorPosition) then
							currentResult.append(ith(1).duplicateAll)
						else
							dummy := showErrorInDefaultFormula(errorPosition);
				TestTriangulationEnumerateC: 
					currentResult.appendNonNil(callTestTriangulationEnumerate(simpleUndirectedGraph(ith(1))));
				TestTriangulationAskC: 
					currentResult.appendNonNil(callTestTriangulationAsk(simpleUndirectedGraph(ith(1))));
				TestTriangulationC: 
					if member(ith(1), markedRealPoint) then
						currentResult.appendNonNil(callTestTriangulation(simpleUndirectedGraph(ith(2)), markedRealPoint(ith(1)).mark.value))
					else
						currentResult.appendNonNil(callTestTriangulation(simpleUndirectedGraph(ith(1)), markedRealPoint(ith(2)).mark.value));
			end;
	end;

procedure initGetTriangulation;
	begin
		trialNumber := stdTrialNumber;
		maxDegreeHennebergReduction := stdMaxDegree;
	end;

end.
