(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit getUserParameter;

interface

uses
	priorityQueue, dialogEngine, standardPredicates;

function getUserParameter1 (title: Str255; default: extended; var value: extended; check: ProcPtr): boolean;

function getUserParameter2 (title1, title2: Str255; default1, default2: extended; var value1, value2: extended; check1, check2: ProcPtr): boolean;

function getUserParameter3 (title1: Str255; default1: extended; var value1: extended; check1: ProcPtr; title2: Str255; default2: extended; var value2: extended; check2: ProcPtr; title3: Str255; default3: extended; var value3: extended; check3: ProcPtr): boolean;
(* 'title' specifies the appropriate title for the numerical value, 'default' gives the default value	*)
(* and function 'check' tests whether the user supplied value is acceptable. The boolean result	*)
(* TRUE indicates that the operation was performed successfully while FALSE indicates that the	*)
(* operation was canceled or some numerical value was not accepted by the 'check' function(s).	*)


function getUserString1 (title: Str255; var value: Str255): boolean;

function getUserString2 (title1, title2: Str255; var value1, value2: Str255): boolean;
(* 'title' specifies the appropriate title for the string value. 'value' contains at entry the	*)
(* default value and at exit the user supplied value.  The boolean result TRUE indicates		*)
(* that the operation was performed successfully while FALSE indicates that the			*)
(* operation was canceled. *)


function getXY (xLength: seqIndex0; mustBeDictPriorityQueue: boolean; var xQueue: priorityQueue; yLength: seqIndex0; var yTable: dictionary; initialGuessingValid: boolean): boolean;
(* If 'xLength' > 0 the user is asked to choose the data type of the priority queue			*)
(* 'xQueue'. The value of 'xLength' indicates the maximal number of entries in the			*)
(* priority queue. The value of 'mustBeDictPriorityQueue' specifies whether the priority	*)
(* queue must be realized as a dictionary ('mustBeDictPriorityQueue' = TRUE) or			*)
(* whether a heap implementation is admissible ('mustBeDictPriorityQueue' = FALSE).		*)
(* If 'yLength' > 0 the user is asked to choose the data type of the dictionary 'yTable'.		*)
(* The value of 'yLength' indicates the maximal number of entries in the dictionary.		*)
(* 'getXY' = false iff the operation was canceled by the user.									*)

function getXX (x1Length: seqIndex0; var x1Queue: priorityQueue; x2Length: seqIndex0; var x2Queue: priorityQueue; var statistics2: boolean; var xType2: priorityQueueTypeRange): boolean;
(* Lets the user choose the priority queues x1Queue and x2Queue								*)
(* 'getXX' = false iff the operation was canceled by the user									*)
(* statistics2 and x2Type are returned to make it possible to form the same structure of x2Queue out of this function. *)

procedure initGetUserParameter;

var
numberOfObjects, numberOfVertices, numberOfExperiments: extended;
globalInitialGuessing, stdInitialGuessing: boolean;

implementation

function getUserString1 (title: Str255; var value: Str255): boolean;
	var
		h: vector;
	begin
		h := newVector(2);
		h.append(getStringHandler(3, @value, '', nil));
		h.append(getStaticTextHandler(4, title, true, nil));
		getUserString1 := processDialogBox(1002, 0, 5, 0, alwaysTrue, h);
	end;

function getUserString2 (title1, title2: Str255; var value1, value2: Str255): boolean;
	var
		h: vector;
	begin
		h := newVector(4);
		h.append(getStringHandler(6, @value2, '', nil));
		h.append(getStaticTextHandler(5, title2, true, nil));
		h.append(getStringHandler(4, @value1, '', nil));
		h.append(getStaticTextHandler(3, title1, true, nil));
		getUserString2 := processDialogBox(1007, 0, 7, 0, alwaysTrue, h);
	end;

function getUserParameter1 (title: Str255; default: extended; var value: extended; check: ProcPtr): boolean;
	var
		h: vector;
	begin
		h := newVector(2);
		h.append(getExtendedHandler(3, @value, default, nil, check));
		h.append(getStaticTextHandler(4, title, true, nil));
		getUserParameter1 := processDialogBox(1002, 0, 5, 0, alwaysTrue, h);
	end;

function getUserParameter2 (title1, title2: Str255; default1, default2: extended; var value1, value2: extended; check1, check2: ProcPtr): boolean;
	var
		h: vector;
	begin
		h := newVector(4);
		h.append(getExtendedHandler(6, @value2, default2, nil, check2));
		h.append(getStaticTextHandler(5, title2, true, nil));
		h.append(getExtendedHandler(4, @value1, default1, nil, check1));
		h.append(getStaticTextHandler(3, title1, true, nil));
		getUserParameter2 := processDialogBox(1007, 0, 7, 0, alwaysTrue, h);
	end;

function getUserParameter3 (title1: Str255; default1: extended; var value1: extended; check1: ProcPtr; title2: Str255; default2: extended; var value2: extended; check2: ProcPtr; title3: Str255; default3: extended; var value3: extended; check3: ProcPtr): boolean;
	var
		h: vector;
	begin
		h := newVector(6);
		h.append(getExtendedHandler(6, @value2, default2, nil, check2));
		h.append(getStaticTextHandler(5, title2, true, nil));
		h.append(getExtendedHandler(8, @value3, default3, nil, check3));
		h.append(getStaticTextHandler(7, title3, true, nil));
		h.append(getExtendedHandler(4, @value1, default1, nil, check1));
		h.append(getStaticTextHandler(3, title1, true, nil));
		getUserParameter3 := processDialogBox(1009, 0, 9, 0, alwaysTrue, h);
	end;

	const
		avlTreeValue = 0;
		sortedListValue = 1;
		sortedArrayValue = 2;
		heapValue = 3;
	var
		xType, yType, x1Type, x2Type: integer;
		xStatistic, yStatistic, x1Statistic, x2Statistic, xActive, yActive: boolean;

function xVisible: boolean;
	begin
		xVisible := xActive;
	end;

function yVisible: boolean;
	begin
		yVisible := yActive;
	end;

function invisible: boolean;
	begin
		invisible := false;
	end;

function intToPrioType (x: integer): objectType;
	begin
		if x = heapValue then
			intToPrioType := heapPriorityQueueType
		else
			intToPrioType := dictPriorityQueueType
	end;

function intToDictType (x: integer): objectType;
	begin
		if x = avlTreeValue then
			intToDictType := avlTreeType
		else if x = sortedListValue then
			intToDictType := sortedListType
		else if x = sortedArrayValue then
			intToDictType := sortedVectorType
		else
			intToDictType := objType;
	end;

procedure getPriorityQueue (xLength: seqIndex0; xType: integer; xStatistic: boolean; var xQueue: priorityQueue);
	var
		xQS: priorityQueueStatistics;
	begin
		if xLength > 0 then begin
			currentVectorLength := xLength;
			currentDictionaryType := intToDictType(xType);
			if xStatistic then begin
				currentPriorityQueueType := intToPrioType(xType);
				new(xQS);
				failNil(Handle(xQS));
				xQS.init;
				xQueue := xQS;
			end
			else
				xQueue := priorityQueue(newInitObj(intToPrioType(xType)));
		end;
	end;

function getXX (x1Length: seqIndex0; var x1Queue: priorityQueue; x2Length: seqIndex0; var x2Queue: priorityQueue; var statistics2: boolean; var xType2: priorityQueueTypeRange): boolean;
	var
		h: vector;
	begin
		h := newVector(4);
		h.append(getEnumerationHandler(7, @x1Type, 0, nil, heapValue));
		h.append(getEnumerationHandler(11, @x2Type, 0, nil, heapValue));
		h.append(getBooleanHandler(15, @x1Statistic, false, nil));
		h.append(getBooleanHandler(16, @x2Statistic, false, nil));
		if processDialogBox(30565, 0, 17, 0, alwaysTrue, h) then begin
			getXX := true;
			getPriorityQueue(x1Length, x1Type, x1Statistic, x1Queue);
			getPriorityQueue(x2Length, x2Type, x2Statistic, x2Queue);
			statistics2 := x2Statistic;
			xType2 := intToPrioType(x2Type);
		end
		else
			getXX := false;
	end;

function getXY (xLength: seqIndex0; mustBeDictPriorityQueue: boolean; var xQueue: priorityQueue; yLength: seqIndex0; var yTable: dictionary; initialGuessingValid: boolean): boolean;
	var
		h: vector;
		yTS: dictionaryStatistics;
	begin
		xActive := xLength > 0;
		yActive := yLength > 0;
		h := newVector(6);
		h.append(getEnumerationHandler(10, @yType, 0, @yVisible, sortedArrayValue));
		h.append(getBooleanHandler(13, @yStatistic, false, @yVisible));
		if initialGuessingValid then
			h.append(getBooleanHandler(14, @stdInitialGuessing, false, nil))
		else
			h.append(getBooleanHandler(14, @stdInitialGuessing, false, @invisible));
		if mustBeDictPriorityQueue then begin
			h.append(getEnumerationHandler(5, @xType, 0, @xVisible, sortedArrayValue));
			h.append(getEnumerationHandler(8, @xType, 0, @invisible, 0));
			h.append(getBooleanHandler(9, @xStatistic, false, @invisible));
		end
		else begin
			h.append(getEnumerationHandler(5, @xType, 0, @xVisible, heapValue));
			h.append(getBooleanHandler(9, @xStatistic, false, @xVisible));
		end;
		if mustBeDictPriorityQueue & (xType = heapValue) then
			xType := avlTreeValue;
		if processDialogBox(1008, 0, 15, 0, alwaysTrue, h) then begin
			getXY := true;
			getPriorityQueue(xLength, xType, not mustBeDictPriorityQueue & xStatistic, xQueue);
			if yLength > 0 then begin
				currentVectorLength := yLength;
				currentDictionaryType := intToDictType(yType);
				if yStatistic then begin
					new(yTS);
					failNil(Handle(yTS));
					yTS.init;
					yTable := yTS;
				end
				else
					yTable := dictionary(newInitObj(intToDictType(yType)));
			end;
		end
		else
			getXY := false;
	end;

procedure initGetUserParameter;
	begin
		numberOfObjects := 10;
		numberOfVertices := 7;
		numberOfExperiments := 10;
		stdInitialGuessing := globalInitialGuessing;
		xType := heapValue;
		yType := avlTreeValue;
		xStatistic := false;
		yStatistic := false;
		x1Type := heapValue;
		x2Type := heapValue;
		x1Statistic := false;
		x2Statistic := false;
	end;

end.
