(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit graph;

interface

uses
{$IFC NOT THINK_PASCAL }
	GFUtilities,
{$ENDC }
	lineSegmentVector, point3d;

type
	unionFindSimple = object(obj)
			n: seqIndex;
			uf: seqIndexArrayH;
			procedure init;
			override;
(*	The global variable 'currentVectorLength' determines how many elements can be included.		*)
(*	The structure is initialized to contain the singleton sets [1], [2], ..., [n].							*)
			procedure free;
			override;
			procedure reset;
(*	The structure is initialized to contain the singleton sets [1], [2], ..., [n].							*)
			function find (startVertex: seqIndex): seqIndex;
(*	Return the index of the set containing 'startVertex'.													*)
			procedure union (startVertex, endVertex: seqIndex);
(*	Joins the sets containing 'startVertex' and 'endVertex'.												*)
			procedure compressAll;
(*	Compresses all paths.																						*)
			procedure addParts (s, e: seqIndex); (* private method *)
		end;

	unionFindCount = object(unionFindSimple)
			count: seqIndexArrayH;
			procedure init;
			override;
			procedure free;
			override;
			procedure reset;
			override;
			procedure addParts (s, e: seqIndex);
			override;
			function cardinality (s: seqIndex): seqIndex;
(*	Returns the cardinality of the component which contains 's'. *)
			function isOneComponent: boolean;
(*	Determines whether all elements are in one set. *)
		end;

type
	graphEdge = object(obj)
			startVertex, endVertex: seqIndex;	(* Pointers to the corresponding vector of vertices. *)
			primary: graphEdge;
(* Marco Ciot: primary is only used for undirectedEdges which consist of the *)
(* primary edge and the inverse edge that points to its primary! *)
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override; (* 'currentVerticesVector' must be set! *)
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			procedure drawEdge (v: vector; directed: boolean);
(* Displays the edge with vertices in 'v'. 'directed' specifies whether the edge	*)
(* should be drawn as a directed edge or not. *)

		end;

	markedGraphEdge = object(graphEdge)
			mark: strings;
			procedure init;
			override;
			function duplicate: obj;
			override;
			procedure free;
			override;
			procedure drawEdge (v: vector; directed: boolean);
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function getInfo: obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
		end;

	simpleUndirectedGraph = object(obj)
			vertices: vector;			(* Specifies the nodes. *)
			edges: vector;				(* A vector of 'graphEdge'. *)

			procedure init;
			override;
(* The global variables 'currentVertices' and 'currentEdges' determine how many vertices and edges are allocated. *)
			procedure free;
			override;
			procedure freeAll;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			function isSelected (where: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			function getInfo: obj;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			procedure randomChange;
			override;
			function getBoundingObject: obj;
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			function perimeter: extended;
			override;
			function convertToType (t: pointTypeRange): obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function verticesAndEdges: Str255;
			procedure deleteEdgeSimple (e: seqIndex);

			procedure drawEdges;
(* displays the edges of the graph. This method is used only in special cases. *)

			procedure addEdge (e: graphEdge);
(* Adds the edge 'e' to the graph. *)

			procedure addNewEdge (s, e: seqIndex);
(* Creates and adds the edge from 's' to 'e' to the graph. *)

			procedure deleteVertex (v: seqIndex);
(* Deletes vertex number 'v' and frees the vertex and all incident edges. *)

			procedure deleteEdge (e: seqIndex);
(* Deletes edge number 'e' and frees it. *)

			function minimumSpanningTree (k: seqIndex; var uf: unionFindSimple): spanningTree;
(* Computes a minimum spanning tree under the assumption that the edges are sorted by length into ascending order. *)

			function showNumbering: homogenVector;
			function getEdges: lineSegmentVector;
			function breadthFirstSearch (root: seqIndex): simpleUndirectedGraph;
	(* Computes the shortest distance for each node starting at the node 'root' in a breadth first manner. *)

			function hennebergSequence (var reduceDegreeThreeVertex: seqIndex0): homogenVector;
(*	A randomized algorithm that computes a Henneberg 2-sequence. 'reduceDegreeThreeVertex' is the number of			*)
(*	vertices of degree three that have been reduced. If this number is 0, the Henneberg sequence is 2-simple.				*)
(*	The algorithm might fail to find a Henneberg  2-sequence even if one exists and should be repeated in such a case.		*)

			function hennebergSequenceMaxTrials (maxTrials: longInt; var trials: longInt; var reduceDegreeThreeVertex: seqIndex0): homogenVector;
(*	Try 'hennebergSequence' 'maxTrials' time and report the longest sequence found.	*)

			function hennebergSequenceDeterministic (var reduceDegreeThreeVertex: seqIndex0): homogenVector;

			function isConnected (s, e: seqIndex): boolean;
(*	Determines whether there is an edge from 's' to 'e'. *)

			function graphIsConnected: boolean;
(*	Determines whether the graph is connected. *)

		end;

	spanningTree = object(simpleUndirectedGraph)
			function memberType (o: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function equal (o: obj): boolean;
			override;

			function TSPEMST: vector;
(* Traverses the spanning tree and produces in the Euclidean case as result a traveling salesman tour	*)
(* that is at most twice as long as the optimal tour. *)
		end;

const
	chooseLast = 0;
	chooseMaxDegree = 1;
	chooseMinDegree = 2;
	chooseRandom = 3;
	chooseBestLevel = 4;

type
	directedGraph = object(simpleUndirectedGraph)
			adjacency: homogenVector; (* Of list of graphEdge. *)
			procedure init;
			override;
(* The global variables 'currentVertices' and 'currentEdges' determine how many vertices and edges are allocated. *)
			procedure free;
			override;
			procedure freeAll;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			function equal (o: obj): boolean;
			override;
			procedure textFileInput;
			override;
			procedure fileInput;
			override;
			function memberType (o: obj): boolean;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure addEdge (e: graphEdge);
			override;
			procedure drawEdges;
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure deleteVertex (v: seqIndex);
			override;
			procedure deleteEdge (e: seqIndex);
			override;
			function isConnected (s, e: seqIndex): boolean;
			override;
			function convertToType (t: pointTypeRange): obj;
			override;

			procedure initAdjacency;
(* Constructs the adjacency lists for the already existing graph. This method can be used to	*)
(* transform a simple undirected graph into a directed graph. *)

			procedure exitAdjacency;
(* Removes the adjacency lists. *)

			procedure addAdjacency (e: graphEdge);
(* adds the adjacency information for 'e' to the graph. This method is used only for 'addEdge'. *)

			function getTriangulation (startSet: sequence; k: seqIndex; choiceMethod: integer; var additionOrder, reverseAdditionOrder: homogenVector): directedGraph;
(*	The 'startSet' is a homogen vector of 'int' describing a set of vertices. 'getTriangulation' determines the	*)
(*	maximal subgraph of 'self' which has the property that 'startSet' is included in it's vertices and each 		*)
(*	additional vertex can be reached by at least 'k' different edges. For k=1 this give the vertices reachable	*)
(*	from 'startSet', for k=2 this determines the set of vertices whose placement on the plane can be			*)
(*	determined, given the position of at least two vertices in 'startSet' is known. Analogous for k=3 in 3-d		*)
(*	space.  choiceMethod which must be between 0 and 4 determines which vertex is added next when			*)
(*	several possibilities exist. additionOrder contains as ints the order in which the vertices are added or	*)
(*	nil if it is not added. reverseAdditionOrder contains as ints for each vertex the position when it is	*)
(*	added or nil if it is not added.																							*)

			function maxTriangulationOrder (var startEdge: seqIndex0): seqIndex0;
(*	Determines the maximum number of vertices that can be reached with triangulation and gives an				*)
(*	appropriate edge in 'startEdge'. If the operation is aborted, the maximum number found so far is reported.	*)

			procedure revertEdges;
(*	Reverts all edges of the graph. *)

		end;

	undirectedGraph = object(directedGraph)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure drawEdges;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure addAdjacency (e: graphEdge);
			override;
			procedure exitAdjacency;
			override;
			function equal (o: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure initializeFromSimpleUndirectedGraph (sug: simpleUndirectedGraph);
	(*	Initializes the 'undirectedGraph' 'self' from the 'simpleUndirectedGraph' 'sug'. *)
		end;

(* begin Marco Ciot *)

	modifiableUndirectedGraph = object(undirectedGraph)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;

			procedure initializeFromSimpleUndirectedGraph (sug: simpleUndirectedGraph);
			override;

(* algorithms: *)
			procedure reduceToGGsimple;
			procedure reduceToRNG;
			procedure reduceToRNGfalse;

(* execute interfaces: *)
			procedure reduce2ggSimple;
			procedure reduce2rngSupowit;
			procedure reduce2rngFalse;
		end;

(* end Marco Ciot *)

function isGraph2d (sudg: simpleUndirectedGraph): boolean;

function isGraph3d (sudg: simpleUndirectedGraph): boolean;

function compareUndirectedGraphEdge (a, b: graphEdge): relationObj;

function compareDirectedGraphEdge (a, b: graphEdge): relationObj;

procedure initGraph;

procedure setCurrentVerticesVector (v: vector);

var
currentVerticesVector: vector;

implementation

{$SETC ReduceGraph = false }

var
	edgeNumber, vertexNumber, trialNumber: extended;

procedure setCurrentVerticesVector (v: vector);
	begin
		currentVerticesVector := v;
	end;

function compareDirectedGraphEdge (a, b: graphEdge): relationObj;
	begin
		if a.startVertex < b.startVertex then
			compareDirectedGraphEdge := lessThanObj
		else if a.startVertex > b.startVertex then
			compareDirectedGraphEdge := greaterThanObj
		else if a.endVertex < b.endVertex then
			compareDirectedGraphEdge := lessThanObj
		else if a.endVertex > b.endVertex then
			compareDirectedGraphEdge := greaterThanObj
		else
			compareDirectedGraphEdge := equalObj
	end;

function compareUndirectedGraphEdge (a, b: graphEdge): relationObj;
	var
		aStart, aEnd, bStart, bEnd: seqIndex;
	begin
		if a.startVertex < a.endVertex then begin
			aStart := a.startVertex;
			aEnd := a.endVertex;
		end
		else begin
			aStart := a.endVertex;
			aEnd := a.startVertex;
		end;
		if b.startVertex < b.endVertex then begin
			bStart := b.startVertex;
			bEnd := b.endVertex;
		end
		else begin
			bStart := b.endVertex;
			bEnd := b.startVertex;
		end;
		if aStart < bStart then
			compareUndirectedGraphEdge := lessThanObj
		else if aStart > bStart then
			compareUndirectedGraphEdge := greaterThanObj
		else if aEnd < bEnd then
			compareUndirectedGraphEdge := lessThanObj
		else if aEnd > bEnd then
			compareUndirectedGraphEdge := greaterThanObj
		else
			compareUndirectedGraphEdge := equalObj
	end;

function callFitToCircle (g: simpleUndirectedGraph; c: circle): simpleUndirectedGraph;
	var
		v: vector;
		result: simpleUndirectedGraph;
		p: point2d;
		i: seqIndex;
	begin
		if g.vertices.length > 1 then begin
			result := simpleUndirectedGraph(g.duplicateAll);
			v := vector(c.pointCover(g.vertices.length));
			for i := 1 to g.vertices.length do begin
				p := point2d(v.elements^^[i]);
				with point2d(result.vertices.elements^^[i]) do begin
					setX(p.getX);
					setY(p.getY);
				end;
			end;
			v.freeAll;
			callFitToCircle := result;
		end
		else
			callFitToCircle := nil;
	end;

procedure writeAsCabriDocument (g: simpleUndirectedGraph; offset: integer);
	var
		udg: undirectedGraph;
		name: Str255;
		t: text;
		i: seqIndex1;

	procedure writeAdjacency (x: graphEdge);
		begin
			with x do
				if startVertex = i then
					write(t, ' ', endVertex - offset : 1)
				else
					write(t, ' ', startVertex - offset : 1);
		end;

	procedure writeVertex (p: obj);
		var
			l: list;
			x, y: longInt;
		begin
			advanceStopWatchHands;
			l := list(udg.adjacency.elements^^[i]);
			write(t, i - offset : 1, ' degree: ', l.length : 2 - offset, ' ');
			l.forAll(writeAdjacency);
			writeLn(t);
			if member(p, point2d) then
				with point2d(p) do begin
{$IFC THINK_PASCAL }
					x := trunc(getX);
					y := trunc(getY);
{$ELSEC }
					x := truncLong(getX);
					y := truncLong(getY);
{$ENDC }
				end
			else if member(p, point3d) then
				with point3d(p) do begin
{$IFC THINK_PASCAL }
					x := trunc(getX);
					y := trunc(getY);
{$ELSEC }
					x := truncLong(getX);
					y := truncLong(getY);
{$ENDC }
				end
			else begin
				x := 0;
				y := 0;
			end;
			if offset = 1 then
				writeLn(t, '      coord.: ', x : 1, ' ', y : 1, ' val.: ', i - 1 : 1, ' shape:  round  size: 3 offsets: 7 4');
			i := i + 1;
		end;

	begin
		if offset = 1 then
			name := ' (CABRI)'
		else
			name := ' (AdjLst)';
		name := NewFileName('Save As', concat(WindowPeek(FrontWindow)^.titleHandle^^, name));
		if name <> '' then begin
			beginCursor(stopwatchCursor);
			if member(g, undirectedGraph) then
				udg := undirectedGraph(g)
			else begin
				new(udg);
				failNil(Handle(udg));
				udg.initializeFromSimpleUndirectedGraph(g);
			end;
			rewrite(t, name);
			if offset = 1 then begin
				writeLn(t, 'CABRI Version 3.0, june 1990');
				writeLn(t, 'grid : -20');
			end;
			writeLn(t, '#vertices: ', udg.vertices.length : 1, ' #edges: ', udg.edges.length : 1);
			writeLn(t);
			i := 1;
			udg.vertices.forAll(writeVertex);
			close(t);
			if not member(g, undirectedGraph) then
				udg.freeAll;
			endCursor;
		end;
	end;

function isGraph2d (sudg: simpleUndirectedGraph): boolean;
	begin
		with sudg do
			isGraph2d := (vertices.length > 0) & not member(vertices.elements^^[1], point3d);
	end;

function isGraph3d (sudg: simpleUndirectedGraph): boolean;
	begin
		with sudg do
			isGraph3d := (vertices.length > 0) & member(vertices.elements^^[1], point3d);
	end;

function isMarkedGraphEdge (x: obj): boolean;
	begin
		isMarkedGraphEdge := member(x, markedGraphEdge);
	end;

function callRemoveMarks (sudg: simpleUndirectedGraph): simpleUndirectedGraph;
	var
		result: simpleUndirectedGraph;
		i: seqIndex0;
	begin
		result := simpleUndirectedGraph(sudg.shallowDuplicate);
		result.vertices := vector(sudg.vertices.duplicateAll);
		result.edges := newHomogenVector(sudg.edges.length);
		with result.edges do begin
			length := sudg.edges.length;
			for i := 1 to length do
				elements^^[i] := sudg.edges.elements^^[i].duplicateTypeTransfer(graphEdgeType);
		end;
		if member(result, directedGraph) then
			with directedGraph(result) do begin
				adjacency := newHomogenVector(vertices.length);
				initAdjacency;
			end;
		callRemoveMarks := result;
	end;

function callAllDistances (g: simpleUndirectedGraph): simpleUndirectedGraph;
	var
		sel: seqIndex0;
		oldPort, thisPort: GrafPtr;
		oldFrame, thisFrame: wFrame;
	begin
		GetPort(oldPort);
		oldFrame := currentFrame;
		additionalHighlight := g;
		g.interactiveHighlight;
		displayInfoStringCommand('Select a vertex!');
		sel := g.vertices.interactiveSelectFromAll(true, true);
		displayInfoString('');
		additionalHighlight := nil;
		GetPort(thisPort);
		thisFrame := currentFrame;
		SetPort(oldPort);
		currentFrame := oldFrame;
		g.interactiveHighlight;
		SetPort(thisPort);
		currentFrame := thisFrame;
		if sel > 0 then
			callAllDistances := g.breadthFirstSearch(sel)
		else
			callAllDistances := nil;
	end;

function deleteVertexSelect (g: simpleUndirectedGraph): seqIndex0;
	begin
		deleteVertexSelect := g.vertices.interactiveSelectFromAll(true, true);
	end;

procedure deleteVertexChange (g: simpleUndirectedGraph; selected: seqIndex);
	begin
		g.deleteVertex(selected);
	end;

function deleteEdgeSelect (g: simpleUndirectedGraph): seqIndex0;
	begin
		setCurrentVerticesVector(g.vertices);
		deleteEdgeSelect := g.edges.interactiveSelectFromAll(true, true);
	end;

procedure deleteEdgeChange (g: simpleUndirectedGraph; selected: seqIndex);
	begin
		g.deleteEdge(selected);
	end;

procedure interactiveAddEdge (g: simpleUndirectedGraph);
	var
		s, e: seqIndex0;
		ed: graphEdge;
		oldPort, thisPort: GrafPtr;
		toHighlight: vector;
		oldFrame, thisFrame: wFrame;

	function edExists (ge: graphEdge): boolean;
		begin
			edExists := (ge.startVertex = ed.startVertex) & (ge.endVertex = ed.endVertex) | (ge.startVertex = ed.endVertex) & (ge.endVertex = ed.startVertex) & (not member(g, directedGraph) | member(g, undirectedGraph));
		end;

	begin (* interactiveAddEdge *)
		GetPort(oldPort);
		oldFrame := currentFrame;
		toHighlight := newVector(2);
		additionalHighlight := toHighlight;
		g.interactiveHighlight;
		toHighlight.append(g);
		setCurrentVerticesVector(g.vertices);
		repeat
			displayInfoStringCommand('Select START vertex!');
			s := g.vertices.interactiveSelectFromAll(true, true);
			if s = 0 then
				leave;
			g.vertices.elements^^[s].interactiveHighlight;
			toHighlight.append(g.vertices.elements^^[s]);
			displayInfoStringCommand('Select END vertex!');
			e := g.vertices.interactiveSelectFromAll(true, true);
			GetPort(thisPort);
			thisFrame := currentFrame;
			SetPort(oldPort);
			currentFrame := oldFrame;
			g.vertices.elements^^[s].interactiveHighlight;
			SetPort(thisPort);
			currentFrame := thisFrame;
			toHighlight.length := toHighlight.length - 1;
			if e = s then
				cycle;
			if e = 0 then
				leave;
			if g.edges.length = 0 then begin
				new(ed);
				failNil(Handle(ed));
			end
			else
				ed := graphEdge(g.edges.elements^^[1].duplicate);
			ed.startVertex := s;
			ed.endVertex := e;
			ed.init;
			ed.drawEdge(g.vertices, member(g, directedGraph) & not member(g, undirectedGraph));
			e := g.edges.exists(edExists);
			if e > 0 then begin
				ed.free;
				g.deleteEdge(e);
			end
			else
				g.addEdge(ed);
			rebuildOperationsMenu := true;
			doBeep := false;
		until false;
		toHighlight.free;
		additionalHighlight := nil;
		GetPort(thisPort);
		thisFrame := currentFrame;
		SetPort(oldPort);
		currentFrame := oldFrame;
		g.interactiveHighlight;
		SetPort(thisPort);
		currentFrame := thisFrame;
		displayInfoString('');
	end;

procedure interactiveAddVertex (g: simpleUndirectedGraph);
	var
		l: list;
		rp: realPoint;
		p: point2d;
		mouse: Point;
		oldPort, thisPort, where: GrafPtr;
		oldFrame, thisFrame: wFrame;
		userClicked: boolean;

	procedure doAppend (x: obj);
		begin
			g.vertices.append(x);
		end;

	begin (* interactiveAddVertex *)
		GetPort(oldPort);
		oldFrame := currentFrame;
		g.interactiveHighlight;
		l := newEmptyList;
		additionalHighlight := g;
		additionalDraw := l;
		beginCursor(littleCrossCursor);
		if g.vertices.length > 0 then
			p := point2d(g.vertices.ith(1).duplicate)
		else begin
			new(rp);
			failNil(Handle(rp));
			p := rp;
		end;
		displayInfoStringCommand('Click to create a new vertex!');
		repeat
			repeat
				if AbortInQueueUserClick(userClicked, where, mouse) | abortFlag then begin
					endCursor;
					if l.length > 0 then begin
						if member(g, directedGraph) then
							with directedGraph(g) do begin
								exitAdjacency;
								adjacency.allocateElements(vertices.length + l.length);
								adjacency.length := 0;
							end;
						g.vertices.allocateElements(g.vertices.length + l.length);
						l.forAll(doAppend);
						if member(g, directedGraph) then
							directedGraph(g).initAdjacency;
					end;
					additionalHighlight := nil;
					additionalDraw := nil;
					GetPort(thisPort);
					thisFrame := currentFrame;
					SetPort(oldPort);
					currentFrame := oldFrame;
					g.interactiveHighlight;
					SetPort(thisPort);
					currentFrame := thisFrame;
					l.free;
					p.free;
					FlushEvents(mUpMask + mDownMask, 0);
					displayInfoString('');
					exit(interactiveAddVertex);
				end;
			until userClicked & (where = oldPort);
			p.windowInputMouse(mouse);
			p.windowToWorld;
			p.interactiveOutput;
			l.append(p.duplicate);
		until false;
	end;

function addGraphs (l: list): simpleUndirectedGraph;
	var
		result: simpleUndirectedGraph;

	procedure addVerticesAndEdges (g: simpleUndirectedGraph);
		begin
			currentVertices := currentVertices + g.vertices.length;
			currentEdges := currentEdges + g.edges.length;
		end;

	procedure processGraph (g: simpleUndirectedGraph);
		procedure processVertex (o: obj);
			begin
				result.vertices.append(o.duplicateAll);
			end;

		procedure processEdge (e: graphEdge);
			begin
				with result.edges do begin
					append(e.duplicateAll);
					with graphEdge(elements^^[length]) do begin
						startVertex := startVertex + result.vertices.length;
						endVertex := endVertex + result.vertices.length;
					end;
				end;
			end;

		begin (* processGraph *)
			g.edges.forAll(processEdge);
			g.vertices.forAll(processVertex);
		end;

	begin (* addGraphs *)
		currentVertices := 0;
		currentEdges := 0;
		l.forAll(addVerticesAndEdges);
		result := simpleUndirectedGraph(newInitObj(l.ith(1).getType));
		l.forAll(processGraph);
		if member(result, directedGraph) then
			directedGraph(result).initAdjacency;
		addGraphs := result;
	end;

function vertexNumberIsOk (x: extended): boolean;
	begin
		with simpleUndirectedGraph(currentArguments.ith(1)) do
			vertexNumberIsOk := (x = trunc(x)) & (1 <= x) & (x <= vertices.length);
	end;

function edgeNumberIsOk (x: extended): boolean;
	begin
		with simpleUndirectedGraph(currentArguments.ith(1)) do
			edgeNumberIsOk := (x = trunc(x)) & (1 <= x) & (x <= edges.length);
	end;

	const
		AddVertexC = 0;
		HennebergSequenceC = 1;
		GetEdgesC = 2;
		ShowNumberingC = 3;
		GetVerticesC = 4;
		FitVerticesC = 5;
		WriteAsCabriDocumentC = 6;
		RemoveMarksC = 7;
		WriteAdjacencyListC = 8;
		AllDistancesFrom1C = 9;
		AllDistancesFromFromC = 10;
		RevertEdgesC = 11;
		DeleteVertexC = 12;
		DeleteEdgeC = 13;
		AddEdgeC = 14;
		AddGraphsC = 15;
		HennebergSequenceDeterministicC = 16;
		DeleteOneVertexC = 17;
		DeleteOneEdgeC = 18;
		HennebergSequenceAskC = 19;
(* begin Marco Ciot *)
{$IFC ReduceGraph }
		Reduce2rngSupowitC = HennebergSequenceAskC + 1;
		Reduce2ggSimpleC = Reduce2rngSupowitC + 1;
		GGsimpleC = Reduce2ggSimpleC + 1;
{$ELSEC }
		GGsimpleC = HennebergSequenceAskC + 1;
{$ENDC }
		RNGsupowitC = GGsimpleC + 1;
		RNGfalseC = RNGsupowitC + 1;
(* end Marco Ciot *)

procedure simpleUndirectedGraph.describe;
	begin
		with currentArguments do begin
			if (length >= 2) & memberTypeSet([simpleUndirectedGraphType, directedGraphType, undirectedGraphType, spanningTreeType]) then
				menuEntry('Add Graphs', AddGraphsC);
			if (length = 1) & member(ith(1), simpleUndirectedGraph) & (simpleUndirectedGraph(ith(1)).vertices.length > 0) then begin
				menuEntry('All Shortest Distances', AllDistancesFrom1C);
				menuEntry('All Shortest Distances From...', AllDistancesFromFromC);
				menuEntry('Find Henneberg 2-Sequence', HennebergSequenceC);
				menuEntry('Find Henneberg 2-Sequence Ask...', HennebergSequenceAskC);
				menuEntry('Find Henneberg 2-Sequence Deterministic', HennebergSequenceDeterministicC);
				if isGraph2d(simpleUndirectedGraph(ith(1))) then begin
					beginSubMenu('Interactive Operations');
					menuEntry('Add Vertices...', AddVertexC);
					menuEntry('Add Edges...', AddEdgeC);
					menuEntry('Delete Vertices...', DeleteVertexC);
					menuEntry('Delete Vertex...', DeleteOneVertexC);
					if simpleUndirectedGraph(ith(1)).edges.length > 0 then begin
						menuEntry('Delete Edges...', DeleteEdgeC);
						menuEntry('Delete Edge...', DeleteOneEdgeC);
					end;
					endSubMenu;
				end;
			end;
			if (length = 1) & member(ith(1), directedGraph) & not member(ith(1), undirectedGraph) & (directedGraph(ith(1)).edges.length > 0) then
				menuEntry('Revert Edges', RevertEdgesC);
			if (length = 1) & member(ith(1), simpleUndirectedGraph) & isGraph2d(simpleUndirectedGraph(ith(1))) then begin
				if simpleUndirectedGraph(ith(1)).edges.length > 0 then
					menuEntry('Get Edges', GetEdgesC);
				if simpleUndirectedGraph(ith(1)).vertices.length > 0 then begin
					menuEntry('Get Vertices', GetVerticesC);
					menuEntry('Show Numbering', ShowNumberingC);
				end;
			end;
			if (length = 1) & member(ith(1), simpleUndirectedGraph) & (simpleUndirectedGraph(ith(1)).vertices.length > 0) then begin
				if isGraph2d(simpleUndirectedGraph(ith(1))) | isGraph3d(simpleUndirectedGraph(ith(1))) then
					menuEntry('Write As CABRI Document...', WriteAsCabriDocumentC);
				menuEntry('Write Adjacency List...', WriteAdjacencyListC);
			end;
			if (length = 2) & (member(ith(1), simpleUndirectedGraph) & isGraph2d(simpleUndirectedGraph(ith(1))) & member(ith(2), circle) | member(ith(2), simpleUndirectedGraph) & isGraph2d(simpleUndirectedGraph(ith(2))) & member(ith(1), circle)) then
				menuEntry('Fit Vertices To Circle', FitVerticesC);
			if (length = 1) & member(ith(1), simpleUndirectedGraph) & (simpleUndirectedGraph(ith(1)).edges.exists(isMarkedGraphEdge) > 0) then
				menuEntry('Remove Edge Marks', RemoveMarksC);

(* begin Marco Ciot *)
{$IFC ReduceGraph }
			if (length = 1) & member(ith(1), modifiableUndirectedGraph) then begin
				beginSubMenu('Reduce Graph');
				menuEntry('To Gabriel Graph Simple', Reduce2ggSimpleC);
				menuEntry('To Relative Neighboorhood Graph Supowit', Reduce2rngSupowitC);
				endSubMenu;
			end;
{$ENDC }

			if (length = 1) & member(ith(1), simpleUndirectedGraph) then begin
				beginSubMenu('Subgraphs');
				menuEntry('Gabriel Graph Simple', GGsimpleC);
				menuEntry('Relative Neighborhood Graph Supowit', RNGsupowitC);
				menuEntry('Relative Neighborhood Graph False', RNGfalseC);
				endSubMenu;
			end;
(* end Marco Ciot *)

		end;
	end;


(* begin Marco Ciot *)

function computeSubgraph (g: simpleUndirectedGraph; item: integer): obj;
	var
		sg: modifiableUndirectedGraph;

	begin
		new(sg);
		failNil(Handle(sg));
		sg.initializeFromSimpleUndirectedGraph(g);
		case item of
			GGsimpleC: 
				sg.reduce2ggSimple;
			RNGsupowitC: 
				sg.reduce2rngSupowit;
			RNGfalseC: 
				sg.reduce2rngFalse;
		end;
		if userAbort then begin
			sg.freeAll;
			computeSubgraph := nil;
			exit(computeSubgraph);
		end;
		computeSubgraph := sg;
	end;
(* end Marco Ciot *)

procedure simpleUndirectedGraph.execute (item: integer);
	var
		dg: directedGraph;
		reduceDegreeThreeVertex: seqIndex0;
		sug: simpleUndirectedGraph;
		trialsUsed: longInt;
	begin
		with currentArguments do
			case item of
				DeleteOneEdgeC: 
					if getUserParameter1('Number of Edge', edgeNumber, edgeNumber, @edgeNumberIsOk) then begin
						sug := simpleUndirectedGraph(ith(1).duplicateAll);
{$IFC THINK_PASCAL }
						sug.deleteEdge(trunc(edgeNumber));
{$ELSEC }
						sug.deleteEdge(truncLong(edgeNumber));
{$ENDC }
						currentResult.append(sug);
					end;
				DeleteOneVertexC: 
					if getUserParameter1('Number of Vertex', vertexNumber, vertexNumber, @vertexNumberIsOk) then begin
						sug := simpleUndirectedGraph(ith(1).duplicateAll);
{$IFC THINK_PASCAL }
						sug.deleteVertex(trunc(vertexNumber));
{$ELSEC }
						sug.deleteVertex(truncLong(vertexNumber));
{$ENDC }
						currentResult.append(sug);
					end;
				RevertEdgesC:  begin
					dg := directedGraph(ith(1).duplicateAll);
					dg.revertEdges;
					currentResult.append(dg);
				end;
				HennebergSequenceDeterministicC: 
					currentResult.appendNonNil(simpleUndirectedGraph(ith(1)).hennebergSequenceDeterministic(reduceDegreeThreeVertex));
				HennebergSequenceAskC: 
					if getUserParameter1('Number of Trials', trialNumber, trialNumber, @isNatural) then begin
{$IFC THINK_PASCAL }
						currentResult.appendNonNil(simpleUndirectedGraph(ith(1)).hennebergSequenceMaxTrials(trunc(trialNumber), trialsUsed, reduceDegreeThreeVertex));
{$ELSEC }
						currentResult.appendNonNil(simpleUndirectedGraph(ith(1)).hennebergSequenceMaxTrials(truncLong(trialNumber), trialsUsed, reduceDegreeThreeVertex));
{$ENDC }
{$IFC AnimationEnabled }
						if userAbort then
							abortFlag := false (* preserve longest Henneberg 2-sequence found so far. *)
						else
							displayInfoString(StringOf('Trials = ', trialsUsed : 1));
{$ENDC }
					end;
				HennebergSequenceC:  begin
					currentResult.appendNonNil(simpleUndirectedGraph(ith(1)).hennebergSequenceMaxTrials(simpleUndirectedGraph(ith(1)).vertices.length + 3, trialsUsed, reduceDegreeThreeVertex));
{$IFC AnimationEnabled }
					displayInfoString(StringOf('Trials = ', trialsUsed : 1));
					if userAbort then
						abortFlag := false (* preserve longest Henneberg 2-sequence found so far. *)
					else
						displayInfoString(StringOf('Trials = ', trialsUsed : 1));
{$ENDC }
				end;
				AddGraphsC: 
					currentResult.append(addGraphs(currentArguments));
				AllDistancesFrom1C: 
					currentResult.appendNonNil(simpleUndirectedGraph(ith(1)).breadthFirstSearch(1));
				AllDistancesFromFromC: 
					currentResult.appendNonNil(callAllDistances(simpleUndirectedGraph(ith(1))));
				WriteAsCabriDocumentC: 
					writeAsCabriDocument(simpleUndirectedGraph(ith(1)), 1);
				WriteAdjacencyListC: 
					writeAsCabriDocument(simpleUndirectedGraph(ith(1)), 0);
				RemoveMarksC: 
					currentResult.append(callRemoveMarks(simpleUndirectedGraph(ith(1))));
				ShowNumberingC: 
					currentResult.append(simpleUndirectedGraph(ith(1)).showNumbering);
				GetVerticesC: 
					currentResult.append(simpleUndirectedGraph(ith(1)).vertices.duplicateAll);
				GetEdgesC: 
					currentResult.append(simpleUndirectedGraph(ith(1)).getEdges);
				AddVertexC: 
					interactiveAddVertex(simpleUndirectedGraph(ith(1)));
				DeleteEdgeC: 
					interactiveChange('Click at an edge to delete it!', simpleUndirectedGraph(ith(1)), deleteEdgeSelect, deleteEdgeChange);
				DeleteVertexC: 
					interactiveChange('Click at a vertex to delete it!', simpleUndirectedGraph(ith(1)), deleteVertexSelect, deleteVertexChange);
				AddEdgeC: 
					interactiveAddEdge(simpleUndirectedGraph(ith(1)));
				FitVerticesC: 
					if member(ith(1), circle) then
						currentResult.appendNonNil(callFitToCircle(simpleUndirectedGraph(ith(2)), circle(ith(1))))
					else
						currentResult.appendNonNil(callFitToCircle(simpleUndirectedGraph(ith(1)), circle(ith(2))));
(* begin Marco Ciot *)
{$IFC ReduceGraph }
(* Reduce2rngSupowitC and Reduce2ggSimpleC are not GeoBench Stile! They should be removed!!! *)
				Reduce2rngSupowitC: 
					modifiableUndirectedGraph(ith(1)).reduce2rngSupowit;
				Reduce2ggSimpleC: 
					modifiableUndirectedGraph(ith(1)).reduce2ggSimple;
{$ENDC }
				GGsimpleC, RNGsupowitC, RNGfalseC: 
					currentResult.appendNonNil(computeSubgraph(simpleUndirectedGraph(ith(1)), item));
(* end Marco Ciot *)
			end;
	end;

procedure initGraph;
	var
		s: simpleUndirectedGraph;
		g: directedGraph;
		h: undirectedGraph;
		e: graphEdge;
		t: spanningTree;
		m: markedGraphEdge;
	begin
		new(t);
		failNil(Handle(t));
		prototypes[spanningTreeType] := t;
		new(s);
		failNil(Handle(s));
		prototypes[simpleUndirectedGraphType] := s;
		new(g);
		failNil(Handle(g));
		prototypes[directedGraphType] := g;
		new(h);
		failNil(Handle(h));
		prototypes[undirectedGraphType] := h;
		new(e);
		failNil(Handle(e));
		prototypes[graphEdgeType] := e;
		new(m);
		failNil(Handle(m));
		prototypes[markedGraphEdgeType] := m;
		edgeNumber := 1;
		vertexNumber := 1;
		trialNumber := 1;
	end;

end.
