(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit graph2;

interface

uses
	graph;

procedure initGraph2;

procedure exitGraph2;

implementation

procedure graphEdge.centerOfGravity (var cx, cy: extended);
	var
		cxStart, cxEnd, cyStart, cyEnd: extended;
	begin
		currentVerticesVector.elements^^[startVertex].centerOfGravity(cxStart, cyStart);
		currentVerticesVector.elements^^[endVertex].centerOfGravity(cxEnd, cyEnd);
		cx := (cxStart + cxEnd) / 2.0;
		cy := (cyStart + cyEnd) / 2.0;
	end;

function graphEdge.getTextType: Str255;
	begin
		getTextType := 'Graph Edge';
	end;

function graphEdge.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(2);
		appendObjInfoExtended(v, 'Start Vertex', startVertex, false);
		appendObjInfoExtended(v, 'End   Vertex', endVertex, false);
		getInfo := v;
	end;

procedure graphEdge.setInfo (o: obj);
	var
		v: vector;
		se, ee: extended;
	begin
		v := vector(o);
		se := objInfoExtended(v.elements^^[1]).e;
		ee := objInfoExtended(v.elements^^[2]).e;
		if (se = trunc(se)) & (ee = trunc(ee)) & (se <> ee) & (1 <= se) & (se <= currentVerticesVector.length) & (1 <= ee) & (ee <= currentVerticesVector.length) then begin
{$IFC THINK_PASCAL }
			startVertex := trunc(se);
			endVertex := trunc(ee);
{$ELSEC }
			startVertex := truncLong(se);
			endVertex := truncLong(ee);
{$ENDC }
		end
		else
			SysBeep(1);
	end;

function simpleUndirectedGraph.showNumbering: homogenVector;
	var
		result: homogenVector;
		ch: char;

	procedure prePend (mp: markedRealPoint);
		begin
			mp.mark.value := concat(ch, mp.mark.value);
		end;

	begin
		result := newHomogenVector(2);
		setCurrentVerticesVector(vertices);
		result.append(vertices.showNumbering);
		result.append(edges.showNumbering);
		ch := 'V';
		sequence(result.elements^^[1]).forAll(prePend);
		ch := 'E';
		sequence(result.elements^^[2]).forAll(prePend);
		showNumbering := result;
	end;

function simpleUndirectedGraph.getInfo: obj;
	var
		v: vector;
	begin
		setCurrentVerticesVector(vertices);
		v := newVector(2);
		appendObjInfoObject(v, 'Vertices', vertices, false);
		appendObjInfoObject(v, 'Edges', edges, false);
		getInfo := v;
	end;

function directedGraph.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoObject(v, 'Adjacency', adjacency, false);
		getInfo := v;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure directedGraph.setInfo (o: obj);
	begin
		exitAdjacency;
		initAdjacency;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function graphEdge.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, graphEdge) & (startVertex = graphEdge(o).startVertex) & (endVertex = graphEdge(o).endVertex);
	end;

function simpleUndirectedGraph.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, simpleUndirectedGraph) & vertices.equal(simpleUndirectedGraph(o).vertices) & edges.equal(simpleUndirectedGraph(o).edges);
	end;

function spanningTree.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, spanningTree) & inherited equal(o);
	end;

function directedGraph.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, directedGraph) & inherited equal(o);
	end;

function undirectedGraph.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, undirectedGraph) & inherited equal(o);
	end;

function graphEdge.genericLessThan (o: obj): boolean;
	var
		e: graphEdge;
	begin
		e := graphEdge(o);
		genericLessThan := (startVertex < e.startVertex) | (startVertex = e.startVertex) & (endVertex < e.endVertex);
	end;

function simpleUndirectedGraph.genericLessThan (o: obj): boolean;
	var
		s: simpleUndirectedGraph;
	begin
		s := simpleUndirectedGraph(o);
		genericLessThan := vertices.genericLessThan(s.vertices) | vertices.equal(s.vertices) & edges.genericLessThan(s.edges);
	end;

procedure graphEdge.textFileInput;
	var
		x: longInt;
	begin
		readTextLongInt(x);
		startVertex := x;
		readTextLongInt(x);
		endVertex := x;
	end;

procedure graphEdge.textFileOutput;
	begin
		writeTextLongInt(startVertex);
		writeTextLongInt(endVertex);
	end;

procedure graphEdge.fileInput;
	var
		x: longInt;
	begin
		assert(currentVersion >= version7, 'Correct version for graphEdge');
		readLongInt(x);
		startVertex := x;
		readLongInt(x);
		endVertex := x;
	end;

procedure graphEdge.fileOutput;
	begin
		writeLongInt(startVertex);
		writeLongInt(endVertex);
	end;

function graphEdge.memberType (o: obj): boolean;
	begin
		memberType := member(o, graphEdge);
	end;

function graphEdge.getType: objectType;
	begin
		getType := graphEdgeType;
	end;

	var
		p: realPoint;

procedure markedGraphEdge.drawEdge (v: vector; directed: boolean);
	const
		k = 0.3;
	var
		startObj, endObj: obj;
		x1, x2, y1, y2: extended;
		ps: PenState;
	begin
		inherited drawEdge(v, directed);
		GetPenState(ps);
		if (ps.pnSize.h <> 3) | (ps.pnSize.v <> 3) then begin
			with v do
				if (startVertex < endVertex) | directed then begin
					startObj := elements^^[startVertex];
					endObj := elements^^[endVertex];
				end
				else begin
					startObj := elements^^[endVertex];
					endObj := elements^^[startVertex];
				end;
			startObj.centerOfGravity(x1, y1);
			endObj.centerOfGravity(x2, y2);
			with p do begin
				setX(x1 + k * (x2 - x1));
				setY(y1 + k * (y2 - y1));
				labelPoint(mark.value);
			end;
		end;
	end;

	var
		l: directedLineSegment;
		l2: directedLineSegment;
		p2, q2: realPoint;

procedure graphEdge.drawEdge (v: vector; directed: boolean);
	const
		offset = 3;
	var
		x1, y1, x2, y2: extended;
		startObj, endObj: obj;
	begin
		with v do begin
			startObj := elements^^[startVertex];
			endObj := elements^^[endVertex];
		end;
		if member(startObj, point2d) & member(endObj, point2d) then
			if directed then begin
				l.p := point2d(startObj);
				l.q := point2d(endObj);
				l.interactiveOutput;
			end
			else
				point2d(startObj).drawLine(point2d(endObj))
		else if member(startObj, point3d) & member(endObj, point3d) then begin
			assert(not directed, 'Undirected 3-d graphs only!');
			drawLineSegment3d(point3d(startObj).getX, point3d(startObj).getY, point3d(startObj).getZ, point3d(endObj).getX, point3d(endObj).getY, point3d(endObj).getZ);
		end
		else begin
			startObj.centerOfGravity(x1, y1);
			endObj.centerOfGravity(x2, y2);
			p2.setX(x1);
			p2.setY(y1);
			q2.setX(x2);
			q2.setY(y2);
			if directed then
				l2.interactiveOutput
			else
				p2.drawLine(q2);
		end;
	end;

procedure graphEdge.interactiveOutput;
	begin
		drawEdge(currentVerticesVector, false);
	end;

function graphEdge.isSelected (where: obj): boolean;
	var
		startObj, endObj: obj;
	begin
		with currentVerticesVector do begin
			startObj := elements^^[startVertex];
			endObj := elements^^[endVertex];
		end;
		if member(startObj, point2d) & member(endObj, point2d) then begin
			l.p := point2d(startObj);
			l.q := point2d(endObj);
			isSelected := l.isSelected(where);
		end
		else
			isSelected := false;
	end;

procedure simpleUndirectedGraph.init;
	begin
		vertices := newVector(currentVertices);
		edges := newVector(currentEdges);
	end;

procedure directedGraph.init;
	begin
		inherited init;
		adjacency := newHomogenVector(currentVertices);
	end;

procedure simpleUndirectedGraph.free;
	begin
		vertices.free;
		edges.freeAll;
		shallowFree;
	end;

procedure simpleUndirectedGraph.freeAll;
	begin
		vertices.freeAll;
		edges.freeAll;
		shallowFree;
	end;

procedure directedGraph.free;
	begin
		vertices.free;
		edges.free; (* the 'graphEdge's are destroyed with adjacency lists *)
		adjacency.freeAll;
		shallowFree;
	end;

procedure directedGraph.freeAll;
	begin
		vertices.freeAll;
		edges.free; (* the 'graphEdge's are destroyed with adjacency lists *)
		adjacency.freeAll;
		shallowFree;
	end;

function simpleUndirectedGraph.duplicate: obj;
	var
		g: simpleUndirectedGraph;
	begin
		g := simpleUndirectedGraph(shallowDuplicate);
		g.vertices := vector(vertices.duplicate);
		g.edges := vector(edges.duplicateAll);
		duplicate := g;
	end;

function simpleUndirectedGraph.convertToType (t: pointTypeRange): obj;
	var
		g: simpleUndirectedGraph;
		v: vector;
	begin
		v := vector(vertices.convertToType(t));
		if v.length = vertices.length then begin
			g := simpleUndirectedGraph(shallowDuplicate);
			g.vertices := v;
			g.edges := vector(edges.duplicateAll);
			convertToType := g;
		end
		else
			convertToType := nil;
	end;

function simpleUndirectedGraph.duplicateAll: obj;
	var
		g: simpleUndirectedGraph;
	begin
		g := simpleUndirectedGraph(shallowDuplicate);
		g.vertices := vector(vertices.duplicateAll);
		g.edges := vector(edges.duplicateAll);
		duplicateAll := g;
	end;

function directedGraph.duplicate: obj;
	var
		g: directedGraph;
	begin
		g := directedGraph(inherited duplicate);
		g.adjacency := newHomogenVector(vertices.length);
		g.initAdjacency;
		duplicate := g;
	end;

function directedGraph.convertToType (t: pointTypeRange): obj;
	var
		g: directedGraph;
	begin
		g := directedGraph(inherited convertToType(t));
		g.adjacency := newHomogenVector(vertices.length);
		g.initAdjacency;
		convertToType := g;
	end;

function directedGraph.duplicateAll: obj;
	var
		g: directedGraph;
	begin
		g := directedGraph(inherited duplicateAll);
		g.adjacency := newHomogenVector(vertices.length);
		g.initAdjacency;
		duplicateAll := g;
	end;

procedure simpleUndirectedGraph.drawEdges;
	var
		i: seqIndex0;
	begin
		for i := 1 to edges.length do
			graphEdge(edges.elements^^[i]).drawEdge(vertices, false);
	end;

procedure directedGraph.drawEdges;
	var
		i: seqIndex0;
	begin
		for i := 1 to edges.length do
			graphEdge(edges.elements^^[i]).drawEdge(vertices, true);
	end;

procedure undirectedGraph.drawEdges;
	var
		i: seqIndex0;
	begin
		for i := 1 to edges.length do
			graphEdge(edges.elements^^[i]).drawEdge(vertices, false);
	end;

procedure simpleUndirectedGraph.interactiveOutput;
	begin
		vertices.interactiveOutput;
		drawEdges;
	end;

procedure simpleUndirectedGraph.translate (dx, dy: extended);
	begin
		vertices.translate(dx, dy);
	end;

procedure simpleUndirectedGraph.rotate (xOrigin, yOrigin, alpha: extended);
	begin
		vertices.rotate(xOrigin, yOrigin, alpha);
	end;

procedure simpleUndirectedGraph.scale (lambda: extended);
	begin
		vertices.scale(lambda);
	end;

procedure simpleUndirectedGraph.reflect (a, b, c: extended);
	begin
		vertices.reflect(a, b, c);
	end;

procedure simpleUndirectedGraph.interactiveHighlight;
	begin
		vertices.interactiveHighlight;
		PenSize(3, 3);
		drawEdges;
		PenSize(1, 1);
	end;

procedure simpleUndirectedGraph.textFileInput;
	var
		oVertices, oedges: obj;
	begin
		oVertices := readTextObj;
		oedges := readTextObj;
		assert(member(oVertices, vector) & member(oedges, vector), 'Legal types in simpleUndirectedGraph');
		vertices := vector(oVertices);
		edges := vector(oedges);
	end;

procedure simpleUndirectedGraph.textFileOutput;
	begin
		writeTextObj(vertices);
		writeTextObj(edges);
	end;

procedure directedGraph.textFileInput;
	begin
		inherited textFileInput;
		adjacency := newHomogenVector(vertices.length);
		initAdjacency;
	end;

procedure simpleUndirectedGraph.fileInput;
	var
		h: homogenVector;
	begin
		vertices := vector(newObj(fileGetType));
		vertices.fileInput;
		if currentVersion >= version8 then
			edges := vector(newObj(fileGetType))
		else begin
			new(h);
			failNil(Handle(h));
			edges := h;
		end;
		edges.fileInput;
	end;

procedure simpleUndirectedGraph.fileOutput;
	begin
		vertices.fileOutputTag;
		edges.fileOutputTag;
	end;

procedure directedGraph.fileInput;
	begin
		inherited fileInput;
		adjacency := newHomogenVector(vertices.length);
		initAdjacency;
	end;

function simpleUndirectedGraph.isSelected (where: obj): boolean;
	begin
		isSelected := vertices.isSelected(where);
	end;

function simpleUndirectedGraph.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	var
		i: seqIndex0;
		l: list;

	function pointIsSelected (p: point2d): boolean;
		begin
			pointIsSelected := p.isSelected(where);
		end;

	procedure doDrawEdge (e: graphEdge);
		begin
			e.drawEdge(vertices, false);
			if highlight then begin
				PenSize(3, 3);
				e.drawEdge(vertices, false);
				PenSize(1, 1);
			end;
		end;

	procedure drawAdjacentEdges (o: obj);
		begin
			l.forAll(doDrawEdge);
			if highlight then
				o.interactiveHighlight;
		end;

	procedure buildAdj (e: graphEdge);
		begin
			if (e.startVertex = i) | (e.endVertex = i) then
				l.append(e);
		end;

	begin (* simpleUndirectedGraph.dragVertex *)
		GetKeys(globalKeyMap);
		dragVertex := false;
		if globalKeyMap[optionKeyValue] then begin
			i := vertices.exists(pointIsSelected);
			if i > 0 then begin
				dragVertex := true;
				l := newEmptyList;
				edges.forAll(buildAdj);
				vertices.elements^^[i].dragObject(drawAdjacentEdges, drawAdjacentEdges, mouse);
				l.free;
			end;
		end;
	end;

function directedGraph.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	var
		i: seqIndex0;
		l: list;

	function pointIsSelected (p: point2d): boolean;
		begin
			pointIsSelected := p.isSelected(where);
		end;

	procedure doDrawEdge (e: graphEdge);
		begin
			e.drawEdge(vertices, true);
			if highlight then begin
				PenSize(3, 3);
				e.drawEdge(vertices, true);
				PenSize(1, 1);
			end;
		end;

	procedure drawAdjacentEdges (o: obj);
		begin
			l.forAll(doDrawEdge);
			if highlight then
				o.interactiveHighlight;
		end;

	procedure buildAdj (e: graphEdge);
		begin
			if e.endVertex = i then
				l.append(e);
		end;

	begin
		GetKeys(globalKeyMap);
		dragVertex := false;
		if globalKeyMap[optionKeyValue] then begin
			i := vertices.exists(pointIsSelected);
			if i > 0 then begin
				dragVertex := true;
				l := list(adjacency.elements^^[i].duplicate);
				edges.forAll(buildAdj);
				vertices.elements^^[i].dragObject(drawAdjacentEdges, drawAdjacentEdges, mouse);
				l.free;
			end;
		end;
	end;

function undirectedGraph.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	var
		i: seqIndex0;

	function pointIsSelected (p: point2d): boolean;
		begin
			pointIsSelected := p.isSelected(where);
		end;

	procedure doDrawEdge (e: graphEdge);
		begin
			e.drawEdge(vertices, false);
			if highlight then begin
				PenSize(3, 3);
				e.drawEdge(vertices, false);
				PenSize(1, 1);
			end;
		end;

	procedure drawAdjacentEdges (o: obj);
		begin
			list(adjacency.elements^^[i]).forAll(doDrawEdge);
			if highlight then
				o.interactiveHighlight;
		end;

	begin
		GetKeys(globalKeyMap);
		dragVertex := false;
		if globalKeyMap[optionKeyValue] then begin
			i := vertices.exists(pointIsSelected);
			if i > 0 then begin
				dragVertex := true;
				vertices.elements^^[i].dragObject(drawAdjacentEdges, drawAdjacentEdges, mouse);
			end;
		end;
	end;

function simpleUndirectedGraph.memberType (o: obj): boolean;
	begin
		memberType := member(o, simpleUndirectedGraph);
	end;

function directedGraph.memberType (o: obj): boolean;
	begin
		memberType := member(o, directedGraph);
	end;

function undirectedGraph.memberType (o: obj): boolean;
	begin
		memberType := member(o, undirectedGraph);
	end;

function simpleUndirectedGraph.getType: objectType;
	begin
		getType := simpleUndirectedGraphType;
	end;

function simpleUndirectedGraph.verticesAndEdges: Str255;
	begin
		verticesAndEdges := stringOf('(V=', vertices.length : 1, ', E=', edges.length : 1, ')');
	end;

function simpleUndirectedGraph.getTextType: Str255;
	begin
		getTextType := concat('Simple Undirected Graph', verticesAndEdges);
	end;

function spanningTree.getType: objectType;
	begin
		getType := spanningTreeType;
	end;

function spanningTree.getTextType: Str255;
	begin
		getTextType := stringOf('Spanning Tree(V=', vertices.length : 1, ')');
	end;

function directedGraph.getType: objectType;
	begin
		getType := directedGraphType;
	end;

function directedGraph.getTextType: Str255;
	begin
		getTextType := concat('Directed Graph', verticesAndEdges);
	end;

function undirectedGraph.getType: objectType;
	begin
		getType := undirectedGraphType;
	end;

function undirectedGraph.getTextType: Str255;
	begin
		getTextType := concat('Undirected Graph', verticesAndEdges);
	end;

procedure simpleUndirectedGraph.addEdge (e: graphEdge);
	begin
		edges.appendSafe(e, 16);
	end;

procedure directedGraph.addEdge (e: graphEdge);
	begin
		edges.appendSafe(e, 16);
		addAdjacency(e);
	end;

procedure simpleUndirectedGraph.addNewEdge (s, e: seqIndex);
	var
		x: graphEdge;
	begin
		new(x);
		failNil(Handle(x));
		x.startVertex := s;
		x.endVertex := e;
		addEdge(x);
	end;

procedure directedGraph.addAdjacency (e: graphEdge);
	begin
		list(adjacency.elements^^[e.startVertex]).append(e);
	end;

procedure undirectedGraph.addAdjacency (e: graphEdge);
(* Modified by Marco Ciot, keeping all compatibility! *)
	var
		inverseEdge: graphEdge;
	begin
		e.primary := e;
		inherited addAdjacency(e);
		new(inverseEdge);
		inverseEdge.startVertex := e.endVertex;
		inverseEdge.endVertex := e.startVertex;
		inverseEdge.primary := e;
		inherited addAdjacency(inverseEdge);
	end;

function simpleUndirectedGraph.pointCover (n: longInt): obj;
	var
		lsv: lineSegmentVector;
		v: vector;
	begin
		lsv := getEdges;
		v := vector(lsv.pointCover(n));
		lsv.freeAll;
		if v <> nil then
			v.eliminateDuplicates;
		pointCover := v;
	end;

function simpleUndirectedGraph.getEdges: lineSegmentVector;
	var
		l: lineSegment;
		lsv: lineSegmentVector;
		i: seqIndex0;
	begin
		currentVectorLength := edges.length;
		new(lsv);
		failNil(Handle(lsv));
		lsv.init;
		for i := 1 to edges.length do begin
{$IFC AnimationEnabled }
			advanceStopWatchHands;
{$ENDC }
			new(l);
			failNil(Handle(l));
			with graphEdge(edges.elements^^[i]) do begin
				l.p := point2d(vertices.elements^^[startVertex].duplicate);
				l.q := point2d(vertices.elements^^[endVertex].duplicate);
			end;
			lsv.append(l);
		end;
		getEdges := lsv;
	end;

procedure simpleUndirectedGraph.randomChange;
	begin
		vertices.randomChange;
	end;

function simpleUndirectedGraph.perimeter: extended;
	var
		s: extended;
		i: seqIndex0;
	begin
		s := 0;
		for i := 1 to edges.length do begin
{$IFC AnimationEnabled }
			advanceStopWatchHands;
{$ENDC }
			with graphEdge(edges.elements^^[i]) do
				if member(vertices.elements^^[startVertex], point2d) & member(vertices.elements^^[endVertex], point2d) then
					s := s + point2d(vertices.elements^^[startVertex]).distanceLpX(point2d(vertices.elements^^[endVertex]));
		end;
		perimeter := s;
	end;

procedure undirectedGraph.initializeFromSimpleUndirectedGraph (sug: simpleUndirectedGraph);
	begin
		vertices := vector(sug.vertices.duplicateAll);
		edges := vector(sug.edges.duplicateAll);
		adjacency := newHomogenVector(sug.vertices.length);
		initAdjacency;
	end;

procedure initGraph2;
	begin
		new(l);
		failNil(Handle(l));
		new(l2);
		failNil(Handle(l2));
		new(p2);
		failNil(Handle(p2));
		new(q2);
		failNil(Handle(q2));
		l2.p := p2;
		l2.q := q2;
		new(p);
		failNil(Handle(p));
	end;

procedure exitGraph2;
	begin
		p.free;
		dispose(l);
	end;

end.
