(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit heapPriorityQueue;

interface

uses
	priorityQueue;

implementation

procedure heapPriorityQueue.reset;
	begin
		numberOfElements := 0;
		freeList := 1;
		masters^^[1].ref := 0;
	end;

procedure heapPriorityQueue.init;
	begin
		masters := refPairArrayH(dynamicArray(currentVectorLength + 1, sizeOf(refPair)));
		vector := seqIndex0ArrayH(dynamicArray(currentVectorLength + 1, sizeOf(seqIndex0)));
		reset;
	end;

procedure heapPriorityQueue.free;
	begin
{$IFC THINK_PASCAL }
		disposHandle(Handle(masters));
		disposHandle(Handle(vector));
{$ELSEC }
		DisposeHandle(Handle(masters));
		DisposeHandle(Handle(vector));
{$ENDC }
		shallowFree;
	end;

procedure heapPriorityQueue.freeAll;
	var
		i: seqIndex0;
	begin
		for i := 1 to numberOfElements do
			ithObj(i).freeAll;
		free;
	end;

procedure heapPriorityQueue.forAll (procedure whatToDo (x: obj));
	var
		i: seqIndex0;
	begin
		for i := 1 to numberOfElements do
			whatToDo(ithObj(i));
	end;

procedure heapPriorityQueue.interactiveOutput;
	var
		i: seqIndex0;
	begin
		for i := 1 to numberOfElements do
			ithObj(i).interactiveOutput;
	end;

procedure heapPriorityQueue.textOutput;
	var
		i: seqIndex0;
	begin
		for i := 1 to numberOfElements do begin
			ithObj(i).textOutput;
			writeLn(' [', vector^^[i] : 1, ']');
		end;
		writeLn;
	end;

function heapPriorityQueue.getObject (where: reference): obj;
	begin
		assert(where <> nil, 'where  nil in heapPriorityQueue.getObject');
		getObject := masters^^[seqIndex(where)].value;
	end;

{$SETC optimize=TRUE }

{$IFC optimize }
procedure heapPriorityQueue.pushUp (i: seqIndex; function compare (a, b: obj): relationObj);
	var
		x, t, idiv2: seqIndex0;
		xObj: obj;
	begin
		if i > 1 then begin
			x := vector^^[i];
			xObj := masters^^[x].value;
			idiv2 := i div 2;
			t := vector^^[idiv2];
			while compare(xObj, masters^^[t].value) = lessThanObj do begin
				vector^^[i] := t;
				masters^^[t].ref := i;
				i := idiv2;
				if idiv2 <= 1 then
					leave;
				idiv2 := i div 2;
				t := vector^^[idiv2];
			end;
			vector^^[i] := x;
			masters^^[x].ref := i;
		end;
	end;

procedure heapPriorityQueue.pushDown (i: seqIndex; function compare (a, b: obj): relationObj);
	var
		j, length2, x, t, tt: seqIndex0;
		xObj, tObj, ttObj: obj;
	begin
		length2 := numberOfElements div 2;
		if i <= length2 then begin
			x := vector^^[i];
			xObj := masters^^[x].value;
			repeat
				j := 2 * i;
				t := vector^^[j];
				tObj := masters^^[t].value;
				if j <> numberOfElements then begin
					tt := vector^^[j + 1];
					ttObj := masters^^[tt].value;
					if compare(tObj, ttObj) > lessThanObj then begin
						j := j + 1;
						t := tt;
						tObj := ttObj;
					end;
				end;
				if compare(xObj, tObj) <= equalObj then
					leave;
				vector^^[i] := t;
				masters^^[t].ref := i;
				i := j;
			until i > length2;
			vector^^[i] := x;
			masters^^[x].ref := i;
		end;
	end;

{$ELSEC }

procedure heapPriorityQueue.pushUp (i: seqIndex; function compare (a, b: obj): relationObj);
	begin
		while (i > 1) & (compare(ithObj(i), ithObj(i div 2)) = lessThanObj) do begin
			swap(i, i div 2);
			i := i div 2;
		end;
	end;

procedure heapPriorityQueue.pushDown (i: seqIndex; function compare (a, b: obj): relationObj);
	var
		j, length2: seqIndex0;
		continue: boolean;
	begin
		continue := true;
		length2 := numberOfElements div 2;
		while continue & (i <= length2) do begin
			j := 2 * i;
			if (j <> numberOfElements) & (compare(ithObj(j), ithObj(j + 1)) > lessThanObj) then
				j := j + 1;
			if compare(ithObj(i), ithObj(j)) = greaterThanObj then begin
				swap(i, j);
				i := j;
			end
			else
				continue := false;
		end;
	end;
{$ENDC }

function heapPriorityQueue.insert (x: obj; function compare (a, b: obj): relationObj): reference;
	begin
		insert := append(x);
		pushUp(numberOfElements, compare);
	end;

procedure heapPriorityQueue.delete (x: reference; function compare (a, b: obj): relationObj);
	var
		startPosition: seqIndex0;
	begin
		startPosition := masters^^[seqIndex(x)].ref;
		swap(startPosition, numberOfElements);
		masters^^[seqIndex(x)].ref := freeList;
		freeList := seqIndex(x);
		numberOfElements := numberOfElements - 1;
		if startPosition <= numberOfElements then begin
			pushDown(startPosition, compare);
			pushUp(startPosition, compare);
		end;
	end;

function heapPriorityQueue.isEmpty: boolean;
	begin
		isEmpty := numberOfElements = 0;
	end;

function heapPriorityQueue.minimum: reference;
	begin
		if numberOfElements = 0 then
			minimum := nil
		else
			minimum := reference(vector^^[1]);
	end;

function heapPriorityQueue.ith (i: longInt): reference;
	begin
		assert((1 <= i) & (i <= numberOfElements), 'Range ok');
		ith := reference(vector^^[i]);
	end;

function heapPriorityQueue.ithObj (i: seqIndex): obj;
	begin
		ithObj := masters^^[vector^^[i]].value;
	end;

function heapPriorityQueue.isHeap (root: seqIndex; function compare (a, b: obj): relationObj): boolean;
	var
		left, right: longInt;
	begin
		left := 2 * root;
		right := left + 1;
		isHeap := ((left > numberOfElements) | (compare(ithObj(left), ithObj(root)) > lessThanObj) & isHeap(left, compare)) & ((right > numberOfElements) | (compare(ithObj(right), ithObj(root)) > lessThanObj) & isHeap(right, compare));
	end;

procedure heapPriorityQueue.swap (p, q: seqIndex);
	var
		pm, qm: seqIndex;
	begin
		pm := vector^^[p];
		qm := vector^^[q];
		masters^^[pm].ref := q;
		masters^^[qm].ref := p;
		vector^^[p] := qm;
		vector^^[q] := pm;
	end;

procedure heapPriorityQueue.heapify (function compare (a, b: obj): relationObj);
	var
		l: seqIndex;
	begin
		freeList := numberOfElements + 1;
		masters^^[freeList].ref := 0;
		l := numberOfElements div 2 + 1;
		while l > 1 do begin
			l := l - 1;
			pushDown(l, compare);
		end;
	end;

function heapPriorityQueue.append (x: obj): reference;
	var
		i: seqIndex0;
	begin
		numberOfElements := numberOfElements + 1;
		i := freeList;
		if masters^^[freeList].ref = 0 then begin
			freeList := freeList + 1;
			masters^^[freeList].ref := 0;
		end
		else
			freeList := masters^^[freeList].ref;
		append := reference(i);
		with masters^^[i] do begin
			value := x;
			ref := numberOfElements;
		end;
		vector^^[numberOfElements] := i;
	end;

function heapPriorityQueue.simpleAppend (x: obj): reference;
	begin
		numberOfElements := numberOfElements + 1;
		simpleAppend := reference(numberOfElements);
		with masters^^[numberOfElements] do begin
			value := x;
			ref := numberOfElements;
		end;
		vector^^[numberOfElements] := numberOfElements;
	end;


end.
