(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit hennebergSequence;

interface

uses
	graph;

implementation

procedure check (e: homogenVector; g: undirectedGraph; a, b, c: seqIndex);
	begin
		if not g.isConnected(a, b) & (list(g.adjacency.elements^^[c]).length > 2) then begin
			e.length := e.length + 1;
			with graphEdge(e.elements^^[e.length]) do begin
				startVertex := a;
				endVertex := b;
			end;
		end;
	end;

procedure makeMRP (result: homogenVector; p: idRealPoint; v: vector; ge: graphEdge);
	var
		mrp: markedRealPoint;
	begin
		new(mrp);
		failNil(Handle(mrp));
		mrp.init;
		result.append(mrp);
		if ge = nil then
			mrp.initLocationMark(p, StringOf(result.length : 1, '=', p.id : 1), false)
		else
			mrp.initLocationMark(p, StringOf(result.length : 1, '=', p.id : 1, ' (', idRealPoint(v.elements^^[ge.startVertex]).id : 1, ',', idRealPoint(v.elements^^[ge.endVertex]).id : 1, ')'), false);
	end;

function degreeOneVertex (l: list): boolean;
	begin
		degreeOneVertex := l.length = 1;
	end;

function degreeTwoVertex (l: list): boolean;
	begin
		degreeTwoVertex := l.length = 2;
	end;

function degreeThreeVertex (l: list): boolean;
	begin
		degreeThreeVertex := l.length = 3;
	end;

function simpleUndirectedGraph.hennebergSequence (var reduceDegreeThreeVertex: seqIndex0): homogenVector;
	var
		g: undirectedGraph;
		result, e: homogenVector;
		i: seqIndex0;
		stop: boolean;
	begin
		result := newHomogenVector(vertices.length);
		hennebergSequence := result;
		if edges.length = 0 then
			exit(hennebergSequence);
		e := newHomogenVector(3);
		e.append(edges.elements^^[1].duplicate);
		e.append(edges.elements^^[1].duplicate);
		e.append(edges.elements^^[1].duplicate);
		new(g);
		failNil(Handle(g));
		g.initializeFromSimpleUndirectedGraph(self);
		g.vertices.freeAll;
		g.vertices := vertices.makeIdRealPointVector;
		reduceDegreeThreeVertex := 0;
		repeat
			stop := true;
			if g.adjacency.exists(degreeOneVertex) = 0 then begin
				i := g.adjacency.exists(degreeTwoVertex);
				if i > 0 then begin
					makeMRP(result, idRealPoint(g.vertices.elements^^[i]), nil, nil);
					g.deleteVertex(i);
					stop := false;
				end
				else begin
					i := g.adjacency.exists(degreeThreeVertex);
					if i > 0 then begin
						e.length := 0;
						with list(g.adjacency.elements^^[i]) do begin
							check(e, g, graphEdge(ith(1)).endVertex, graphEdge(ith(2)).endVertex, graphEdge(ith(3)).endVertex);
							check(e, g, graphEdge(ith(1)).endVertex, graphEdge(ith(3)).endVertex, graphEdge(ith(2)).endVertex);
							check(e, g, graphEdge(ith(2)).endVertex, graphEdge(ith(3)).endVertex, graphEdge(ith(1)).endVertex);
						end;
						if e.length > 0 then begin
							g.addEdge(graphEdge(e.randomChooseFromAll.duplicate));
							makeMRP(result, idRealPoint(g.vertices.elements^^[i]), g.vertices, graphEdge(g.edges.elements^^[g.edges.length]));
							reduceDegreeThreeVertex := reduceDegreeThreeVertex + 1;
							g.deleteVertex(i);
							stop := false;
						end;
					end;
				end;
			end;
		until stop;
		if g.edges.length = 1 then
			with graphEdge(g.edges.elements^^[1]) do begin
				makeMRP(result, idRealPoint(g.vertices.elements^^[startVertex]), nil, nil);
				makeMRP(result, idRealPoint(g.vertices.elements^^[endVertex]), nil, nil);
			end;
		g.freeAll;
		e.length := 3;
		e.freeAll;
	end;

function simpleUndirectedGraph.hennebergSequenceDeterministic (var reduceDegreeThreeVertex: seqIndex0): homogenVector;
	var
		g: undirectedGraph;
		result: homogenVector;
		i, bestDegreeThreeVertex, maxDegree, omitVertex: seqIndex0;
		stop: boolean;
		ge: graphEdge;

	procedure testVertex (a, b, c, d: seqIndex);
		begin
			if not g.isConnected(a, b) & (list(g.adjacency.elements^^[c]).length > maxDegree) then begin
				maxDegree := list(g.adjacency.elements^^[c]).length;
				bestDegreeThreeVertex := i;
				omitVertex := d;
			end;
		end;

	procedure findBestDegreeThreeVertex (l: list);
		begin
			i := i + 1;
			if l.length = 3 then begin
				testVertex(graphEdge(l.ith(1)).endVertex, graphEdge(l.ith(2)).endVertex, graphEdge(l.ith(3)).endVertex, 3);
				testVertex(graphEdge(l.ith(1)).endVertex, graphEdge(l.ith(3)).endVertex, graphEdge(l.ith(2)).endVertex, 2);
				testVertex(graphEdge(l.ith(2)).endVertex, graphEdge(l.ith(3)).endVertex, graphEdge(l.ith(1)).endVertex, 1);
			end;
		end;

	begin
		result := newHomogenVector(vertices.length);
		hennebergSequenceDeterministic := result;
		if edges.length = 0 then
			exit(hennebergSequenceDeterministic);
		new(g);
		failNil(Handle(g));
		g.initializeFromSimpleUndirectedGraph(self);
		g.vertices.freeAll;
		g.vertices := vertices.makeIdRealPointVector;
		reduceDegreeThreeVertex := 0;
		repeat
			stop := true;
			if g.adjacency.exists(degreeOneVertex) = 0 then begin
				i := g.adjacency.exists(degreeTwoVertex);
				if i > 0 then begin
					makeMRP(result, idRealPoint(g.vertices.elements^^[i]), nil, nil);
					g.deleteVertex(i);
					stop := false;
				end
				else begin
					bestDegreeThreeVertex := 0;
					maxDegree := 0;
					i := 0;
					g.adjacency.forAll(findBestDegreeThreeVertex);
					if bestDegreeThreeVertex > 0 then begin
						ge := graphEdge(edges.ith(1).duplicate);
						with list(g.adjacency.ith(bestDegreeThreeVertex)) do
							case omitVertex of
								1:  begin
									ge.startVertex := graphEdge(ith(2)).endVertex;
									ge.endVertex := graphEdge(ith(3)).endVertex;
								end;
								2:  begin
									ge.startVertex := graphEdge(ith(1)).endVertex;
									ge.endVertex := graphEdge(ith(3)).endVertex;
								end;
								3:  begin
									ge.startVertex := graphEdge(ith(1)).endVertex;
									ge.endVertex := graphEdge(ith(2)).endVertex;
								end;
							end;
						g.addEdge(ge);
						makeMRP(result, idRealPoint(g.vertices.elements^^[bestDegreeThreeVertex]), g.vertices, graphEdge(g.edges.elements^^[g.edges.length]));
						reduceDegreeThreeVertex := reduceDegreeThreeVertex + 1;
						g.deleteVertex(bestDegreeThreeVertex);
						stop := false;
					end;
				end;
			end;
		until stop;
		if g.edges.length = 1 then
			with graphEdge(g.edges.elements^^[1]) do begin
				makeMRP(result, idRealPoint(g.vertices.elements^^[startVertex]), nil, nil);
				makeMRP(result, idRealPoint(g.vertices.elements^^[endVertex]), nil, nil);
			end;
		g.freeAll;
	end;

function simpleUndirectedGraph.hennebergSequenceMaxTrials (maxTrials: longInt; var trials: longInt; var reduceDegreeThreeVertex: seqIndex0): homogenVector;
	var
		result, maxResult: homogenVector;
		resultReduceDegreeThreeVertex: seqIndex0;
	begin
		maxResult := hennebergSequence(reduceDegreeThreeVertex);
		trials := 1;
		while (trials <= maxTrials) & (maxResult.length < vertices.length) do begin
{$IFC AnimationEnabled }
			if userAbort then
				leave;
{$ENDC }
			result := hennebergSequence(resultReduceDegreeThreeVertex);
			if (result.length > maxResult.length) | (result.length = maxResult.length) & (resultReduceDegreeThreeVertex < reduceDegreeThreeVertex) then begin
				maxResult.freeAll;
				maxResult := result;
				reduceDegreeThreeVertex := resultReduceDegreeThreeVertex;
			end
			else
				result.freeAll;
			trials := trials + 1;
		end;
		hennebergSequenceMaxTrials := maxResult;
	end;

end.
