(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit hvLineSegment;

{$SETC ALineSegmentIntersection = TRUE }

interface

uses
	lineSegmentVector;

implementation

function lineSegmentIsHorizontalVertical (l: lineSegment): boolean;
	begin
		lineSegmentIsHorizontalVertical := (l.p.getX = l.q.getX) | (l.p.getY = l.q.getY);
	end;

function lineSegmentVector.isHorizontalVertical: boolean;
	begin
		isHorizontalVertical := sameProperty(lineSegmentIsHorizontalVertical);
	end;

function lineSegmentVector.projectOnX: lineSegmentVector;
	var
		r: lineSegmentVector;
		i: seqIndex0;
		l: lineSegment;
	begin
		r := lineSegmentVector(duplicateAll);
		for i := 1 to length do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				r.freeAll;
				projectOnX := nil;
				exit(projectOnX);
			end;
{$ENDC }
			l := lineSegment(r.elements^^[i]);
			l.q.setY(l.p.getY);
		end;
		projectOnX := r;
	end;

function lineSegmentVector.projectOnY: lineSegmentVector;
	var
		r: lineSegmentVector;
		i: seqIndex0;
		l: lineSegment;
	begin
		r := lineSegmentVector(duplicateAll);
		for i := 1 to length do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				r.freeAll;
				projectOnY := nil;
				exit(projectOnY);
			end;
{$ENDC }
			l := lineSegment(r.elements^^[i]);
			l.q.setX(l.p.getX);
		end;
		projectOnY := r;
	end;

function lineSegmentVector.projectOnXY: lineSegmentVector;
	var
		r: lineSegmentVector;
		i: seqIndex0;
		l: lineSegment;
	begin
		r := lineSegmentVector(duplicateAll);
		for i := 1 to length do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				r.freeAll;
				projectOnXY := nil;
				exit(projectOnXY);
			end;
{$ENDC }
			l := lineSegment(r.elements^^[i]);
			if abs(l.p.getX - l.q.getX) < abs(l.p.getY - l.q.getY) then
				l.q.setX(l.p.getX)
			else
				l.q.setY(l.p.getY);
		end;
		projectOnXY := r;
	end;

	type
		hvEventType = (startHorizontal, startVertical, endHorizontal);

		hvEvent = object(obj)
				where: point2d;
				procedure processEvent (yTable: dictionary; intersections: list; var vertical: lineSegment);
				function getEventType: hvEventType;
			end;

		hvEndHorizontal = object(hvEvent)
				toKill: reference; (* into y-table *)
				procedure processEvent (yTable: dictionary; intersections: list; var vertical: lineSegment);
				override;
				function getEventType: hvEventType;
				override;
			end;

		hvStartHorizontal = object(hvEvent)
				l: lineSegment;
				e: hvEndHorizontal;
				procedure processEvent (yTable: dictionary; intersections: list; var vertical: lineSegment);
				override;
				function getEventType: hvEventType;
				override;
			end;

		hvStartVertical = object(hvEvent)
				l: lineSegment;
				procedure processEvent (yTable: dictionary; intersections: list; var vertical: lineSegment);
				override;
				function getEventType: hvEventType;
				override;
			end;

		hvYElement = object(obj)
				l: lineSegment;
				e: hvEndHorizontal;
			end;

function hvEvent.getEventType: hvEventType;
	begin
		unimplementedMethod('hvEvent.getEventType');
		getEventType := startHorizontal;
	end;

function hvStartHorizontal.getEventType: hvEventType;
	begin
		getEventType := startHorizontal;
	end;

function hvStartVertical.getEventType: hvEventType;
	begin
		getEventType := startVertical;
	end;

function hvEndHorizontal.getEventType: hvEventType;
	begin
		getEventType := endHorizontal;
	end;

function compareHLineSegment (a, b: obj): relationObj;
	var
		al, bl: lineSegment;
	begin
		if member(a, lineSegment) then
			al := lineSegment(a)
		else
			al := hvYElement(a).l;
		if member(b, lineSegment) then
			bl := lineSegment(b)
		else
			bl := hvYElement(b).l;
		if al.p.getY < bl.p.getY then
			compareHLineSegment := lessThanObj
		else if al.p.getY > bl.p.getY then
			compareHLineSegment := greaterThanObj
		else
			compareHLineSegment := equalObj
	end;

function lessThanHVEvent (a, b: hvEvent): boolean;
	begin
		if a.where.getX < b.where.getX then
			lessThanHVEvent := true
		else if a.where.getX > b.where.getX then
			lessThanHVEvent := false
		else if a.getEventType < b.getEventType then
			lessThanHVEvent := true
		else if a.getEventType > b.getEventType then
			lessThanHVEvent := false
		else if a.where.getY < b.where.getY then
			lessThanHVEvent := true
		else
			lessThanHVEvent := false;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure hvEvent.processEvent (yTable: dictionary; intersections: list; var vertical: lineSegment);
	begin
		unimplementedMethod('hvEvent.processEvent');
	end;

procedure hvStartHorizontal.processEvent (yTable: dictionary; intersections: list; var vertical: lineSegment);
	var
		s: lineSegment;
		yEl, oldYEl: hvYElement;
		whereRef: reference;
		direction: directionType;
	begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then begin
			l.interactiveFlash;
			l.interactiveHighlight;
			waitForClick(animationFlag[ALineSegmentIntersectionItem]);
			l.interactiveHighlight;
		end;
{$ENDC }
		if yTable.find(l, compareHLineSegment, whereRef, direction) then begin
			oldYEl := hvYElement(yTable.getObject(whereRef));
			new(s);
			failNil(Handle(s));
			s.p := point2d(l.p.duplicate);
			if l.q.getX <= oldYEl.l.q.getX then begin
				s.q := point2d(l.q.duplicate);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
				if animationFlag[ALineSegmentIntersectionItem] then
					l.interactiveOutput;
{$ENDC }
			end
			else begin
				s.q := point2d(oldYEl.l.q.duplicate);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
				if animationFlag[ALineSegmentIntersectionItem] then
					oldYEl.l.interactiveOutput;
{$ENDC }
				e.toKill := whereRef;
				oldYEl.e.toKill := nil;
				oldYEl.l := l;
				oldYEl.e := e;
			end;
			intersections.append(s);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then
				s.interactiveHighlight;
{$ENDC }
		end
		else begin
			new(yEl);
			failNil(Handle(yEl));
			yEl.l := l;
			yEl.e := e;
			e.toKill := yTable.insert(yEl, whereRef, direction);
		end;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure hvStartVertical.processEvent (yTable: dictionary; intersections: list; var vertical: lineSegment);
	var
		k, ll: lineSegment;

	procedure report (k: hvYElement);
		var
			p: point2d;
		begin
			p := point2d(k.l.p.duplicate);
			p.setX(l.p.getX);
			intersections.append(p);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then begin
				k.l.interactiveFlash;
				k.l.interactiveHighlight;
				waitForClick(animationFlag[ALineSegmentIntersectionItem]);
				k.l.interactiveHighlight;
				p.interactiveHighlight;
			end;
{$ENDC }
		end;

	begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then begin
			l.interactiveFlash;
			l.interactiveHighlight;
			waitForClick(animationFlag[ALineSegmentIntersectionItem]);
			l.interactiveHighlight;
		end;
{$ENDC }
		new(ll);
		failNil(Handle(ll));
		ll.p := l.q;
		yTable.rangeForAll(l, ll, compareHLineSegment, report, left);
		dispose(ll);
		if (vertical <> nil) & (l.p.getX = vertical.p.getX) & (l.p.getY <= vertical.q.getY) then begin
			new(k);
			failNil(Handle(k));
			k.p := point2d(l.p.duplicate);
			if l.q.getY <= vertical.q.getY then
				k.q := point2d(l.q.duplicate)
			else begin
				k.q := point2d(vertical.q.duplicate);
				vertical := l;
			end;
			intersections.append(k);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then
				k.interactiveHighlight;
{$ENDC }
		end
		else
			vertical := l;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			l.interactiveOutput;
{$ENDC }
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure hvEndHorizontal.processEvent (yTable: dictionary; intersections: list; var vertical: lineSegment);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
	var
		l: lineSegment;
{$ENDC }
	begin
		if toKill <> nil then begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then begin
				l := hvYElement(yTable.getObject(toKill)).l;
				l.interactiveFlash;
				l.interactiveHighlight;
				waitForClick(animationFlag[ALineSegmentIntersectionItem]);
				l.interactiveHighlight;
				l.interactiveOutput;
			end;
{$ENDC }
			yTable.getObject(toKill).free;
			yTable.delete(toKill);
		end;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function lineSegmentVector.hvPlaneSweepAllIntersect (ask: boolean): vector;
	var
		yTable: dictionary;
		dummy: priorityQueue;
		xQueue: vector;
		intersections: list;
		i: seqIndex0;
		event: hvEvent;
		v: vector;
		currentVertical: lineSegment; (* the longest known vertical line segment at the current position of the sweep line *)

	function buildXQueue (lv: lineSegmentVector): vector;
		var
			xQueue: vector;
			i: seqIndex0;
			hvStartH: hvStartHorizontal;
			hvEndH: hvEndHorizontal;
			hvStartV: hvStartVertical;
			l: lineSegment;
		begin
			xQueue := newVector(2 * length);
			for i := 1 to lv.length do begin
				l := lineSegment(elements^^[i]);
				orderPointsGreaterThanXY(l.p, l.q);
				if l.p.getX = l.q.getX then begin (* vertical *)
					new(hvStartV);
					failNil(Handle(hvStartV));
					hvStartV.where := l.p;
					hvStartV.l := l;
					xQueue.append(hvStartV);
				end
				else (* horizontal *)
					begin
					new(hvEndH);
					failNil(Handle(hvEndH));
					hvEndH.where := l.q;
					hvEndH.toKill := nil; (* will be set by processing the startEvent *)
					xQueue.append(hvEndH);
					new(hvStartH);
					failNil(Handle(hvStartH));
					hvStartH.where := l.p;
					hvStartH.l := l;
					hvStartH.e := hvEndH;
					xQueue.append(hvStartH);
				end;
			end;
			xQueue.sortAll(lessThanHVEvent);
			buildXQueue := xQueue;
		end;

	begin
		if ask then begin
			if not getXY(0, false, dummy, length, yTable, false) then begin
				hvPlaneSweepAllIntersect := nil;
				exit(hvPlaneSweepAllIntersect);
			end;
		end
		else begin
			currentVectorLength := length;
			yTable := dictionary(newInitObj(avlTreeType));
		end;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			interactiveHighlight;
{$ENDC }
		xQueue := buildXQueue(self);
		intersections := newEmptyList;
		currentVertical := nil;
		for i := 1 to xQueue.length do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				xQueue.freeAll;
				yTable.freeAll;
				intersections.freeAll;
				hvPlaneSweepAllIntersect := nil;
				exit(hvPlaneSweepAllIntersect);
			end;
{$ENDC }
			event := hvEvent(xQueue.elements^^[i]);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then
				drawThickVertical(event.where.getX);
{$ENDC }
			event.processEvent(yTable, intersections, currentVertical);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then
				drawThickVertical(event.where.getX);
{$ENDC }
		end;
		xQueue.freeAll;
		if yTable.infoString <> '' then
			displayInfoString(concat('Y', yTable.infoString));
		yTable.free;
		if intersections.length > 0 then begin
			new(v);
			failNil(Handle(v));
			v.initFromList(intersections);
			hvPlaneSweepAllIntersect := v;
		end
		else
			hvPlaneSweepAllIntersect := nil;
		intersections.free;
	end;

end.
