(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit infoWindow;

interface

uses
	GeoBenchUtility, floatArithmetic, graphics;

type
	arithmetics = (realArithmetic, floatArithmetic, longIntArithmetic);

var
	infoString: Str255;
	currentArithmetic: arithmetics;
	currentBase, currentPrecision: integer;

procedure initInfoWindow;

procedure writeCursorCoordinates (always: boolean);
(* write the coordinates of the mouse to the 'infoWindow' *)

procedure displayInfoString (info: Str255);
(* Displays the string 'info' in the info window of the XYZ GeoBench. *)

procedure displayInfoStringCommand (info: Str255);

procedure displayInfoString2 (title1, info1, title2, info2: Str255);


implementation

var
	cursorX, cursorY: integer;

procedure initInfoWindow;
	begin
		cursorX := 0;
		cursorY := 0;
		infoString := '';
		currentBase := 10;
		currentPrecision := 2;
		currentArithmetic := realArithmetic;
	end;

procedure displayInfoString (info: Str255);
	var
		save: GrafPtr;
	begin
		GetPort(save);
		SetPort(infoWindow);
		MoveTo(15, 45);
{$IFC THINK_PASCAL }
		writeDraw(info, ' ' : 150);
{$ELSEC }
		DrawString(stringOf(info, ' ' : 150));
{$ENDC }
		infoString := info;
		SetPort(save);
	end;

procedure displayInfoStringCommand (info: Str255);
	begin
		displayInfoString(Concat(info, ' (Command ''.'' to exit)'));
	end;

procedure displayInfoString2 (title1, info1, title2, info2: Str255);
	begin
		if info1 = '' then begin
			if info2 <> '' then
				displayInfoString(concat(title2, info2));
		end
		else if info2 = '' then
			displayInfoString(concat(title1, info1))
		else
			displayInfoString(concat(title1, info1, ' ', title2, info2));
	end;

procedure writeCursorCoordinates (always: boolean);
	var
		mouse: Point;

	procedure writeCursor;
		var
			save: GrafPtr;

		function convert (x: extended): Str255;
			var
				i: longInt;
			begin
				case currentArithmetic of
					realArithmetic: 
						convert := stringOf(x : 10 : 6);
					floatArithmetic: 
						convert := floatToLine(realToFloatF(x, currentBase, currentPrecision));
					longIntArithmetic:  begin
{$IFC THINK_PASCAL }
						i := round(x);
{$ELSEC }
						i := roundLong(x);
{$ENDC }
						convert := stringOf(i : 1);
					end;
				end;
			end;

		begin (* writeCursor *)
			GetPort(save);
			SetPort(infoWindow);
			MoveTo(15, 15);
			with currentFrame do
{$IFC THINK_PASCAL }
				writeDraw('Cursor: (', convert((cursorX - dx) / mx), ',', convert((cursorY - dy) / my), ')');
			writeDraw(' ' : 30);
{$ELSEC }
				DrawString(stringOf('Cursor: (', convert((cursorX - dx) / mx), ',', convert((cursorY - dy) / my), ')'));
			DrawString(stringOf(' ' : 30));
{$ENDC }
			SetPort(save);
		end;

	begin(* writeCursorCoordinates *)
		if always then
			writeCursor
		else begin
			GetMouse(mouse);
			if (mouse.h <> cursorX) | (mouse.v <> cursorY) then begin
				cursorX := mouse.h;
				cursorY := mouse.v;
				writeCursor;
			end;
		end;
	end;

end.
