(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit layer;

interface

uses
	pointVector;

type
	layer = object(obj)
			z, thickness: extended;
			layElems: vector;
			procedure init;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			procedure free;
			override;
			procedure freeAll;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function getBoundingObject: obj;
			override;
			function equal (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
		end;

	polyLayer = object(layer)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
		end;

procedure initLayer;

implementation

const
	setLayersZC = 2;

function layer.equal (o: obj): boolean;
	var
		t: layer;
	begin
		if (o <> nil) & member(o, layer) then begin
			t := layer(o);
			equal := (z = t.z) & (thickness = t.thickness) & layElems.equal(t.layElems);
		end
		else
			equal := false;
	end;

function layer.genericLessThan (o: obj): boolean;
	var
		t: layer;
	begin
		if (o <> nil) & member(o, layer) then begin
			t := layer(o);
			genericLessThan := (z < t.z) | (z = t.z) & ((thickness < t.thickness) | (thickness = t.thickness) & layElems.genericLessThan(t.layElems));
		end
		else
			genericLessThan := false;
	end;

procedure layer.init;
	begin
		z := 0.0;
		thickness := 0.0;
		layElems := vector(newInitObj(vectorType));
	end;

function layer.duplicate: obj;
	var
		l: layer;
	begin
		l := layer(shallowDuplicate);
		l.layElems := vector(layElems.duplicate);
		duplicate := l;
	end;

function layer.duplicateAll: obj;
	var
		l: layer;
	begin
		l := layer(shallowDuplicate);
		l.layElems := vector(layElems.duplicateAll);
		duplicateAll := l;
	end;

procedure layer.free;
	begin
		if layElems <> nil then
			layElems.free;
		shallowFree;
	end;

procedure layer.freeAll;
	begin
		if layElems <> nil then
			layElems.freeAll;
		shallowFree;
	end;

function layer.getType;
	begin
		getType := layerType;
	end;

function layer.getTextType;
	begin
		getTextType := stringOf('Layer(', layElems.length : 1, ')');
	end;

procedure layer.fileInput;
	begin
		readExtended(z);
		readExtended(thickness);
		layElems := vector(newObj(fileGetType));
		layElems.fileInput;
	end;

procedure layer.fileOutput;
	begin
		writeExtended(z);
		writeExtended(thickness);
		layElems.fileOutputTag;
	end;

procedure layer.textFileInput;
	var
		oo: obj;
		r: real;
	begin
		readTextReal(r);
		z := r;
		readTextReal(r);
		thickness := r;
		oo := readTextObj;
		assert(member(oo, vector), 'Legal type in Layer');
		layElems := vector(oo);
	end;

procedure layer.textFileOutput;
	begin
		writeTextReal(z);
		writeTextReal(thickness);
		writeTextObj(layElems);
	end;

function layer.memberType (o: obj): boolean;
	begin
		memberType := member(o, layer);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure layer.translate (dx, dy: extended);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function layer.isSelected (where: obj): boolean;
	var
		l: lineSegment;
		ix1, ix2, iy1, iy2: integer;
		x1, x2, y1, y2: extended;
		dummy: boolean;
	begin
		l := lineSegment(newInitObj(lineSegmentType));
		dummy := worldToWindow3d(0, 0, z, ix1, iy1);
		dummy := worldToWindow3d(0, 0, z + thickness, ix2, iy2);
		windToWorld(ix1, iy1, x1, y1);
		windToWorld(ix2, iy2, x2, y2);
		l.p.setX(x1);
		l.p.setY(y1);
		l.q.setX(x2);
		l.q.setY(y2);
		isSelected := l.isSelected(where);
		l.freeAll;
	end;

function layer.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(3);
		appendObjInfoExtended(v, 'Z (Over XY-Plane)', z, false);
		appendObjInfoExtended(v, 'Thickness', thickness, false);
		appendObjInfoObject(v, 'Layer Elements', layElems, false);
		getInfo := v;
	end;

procedure layer.setInfo (o: obj);
	var
		v: vector;
	begin
		v := vector(o);
		z := objInfoExtended(v.elements^^[1]).e;
		thickness := objInfoExtended(v.elements^^[2]).e;
		layElems := vector(objInfoObj(v.elements^^[3]).o);
	end;

procedure layer.interactiveOutput;
	begin
		with currentFrame do begin
			currentFrame.z := self.z;
			currentFrame.proj3d := true;
			if (not coordSysIsDrawn) then begin
				coordSysIsDrawn := true;
				drawCoordSys;
			end;
			layElems.interactiveOutput;
			currentFrame.proj3d := false;
		end;
	end;

function layer.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := layElems.drawingComplexity(cutOff);
	end;

procedure layer.interactiveHighlight;
	begin
		with currentFrame do begin
			currentFrame.z := self.z;
			currentFrame.proj3d := true;
			if (not coordSysIsDrawn) then begin
				coordSysIsDrawn := true;
				drawCoordSys;
			end;
			layElems.interactiveHighlight;
			currentFrame.proj3d := false;
		end;
	end;

function polyLayer.getType: objectType;
	begin
		getType := polyLayerType;
	end;

function polyLayer.getTextType: Str255;
	begin
		getTextType := stringOf('PolyLayer(', layElems.length : 1, ')');
	end;

function polyLayer.memberType (o: obj): boolean;
	begin
		memberType := member(o, layer);
	end;


procedure polyLayer.interactiveOutput;

	const
		eps = 10E-8;


	type

		point3dExtT = record			(* for efficiency reasons *)
				p: point3dT;
				vis: array[1..4] of boolean;
			end;

	var
		oldZ: extended;
		po, fillPo: PolyHandle;
		neverMoved, fillIt, fillNeverMoved: boolean;
		p, pdot, clip: point3dT;
		pk: point3dExtT;
		pp, sp, fp: array[1..4] of point3dExtT;
		firstExist: array[1..4] of boolean;
		dist: extended;
		firstX, firstY, fillFirstX, fillFirstY: integer;
		oldPen, newPen: PenState;


	procedure filledPoly (poly: obj);

		var
			i: seqIndex0;
			k: extended;

		procedure intSec (var e: integer; var p, q, s: point3dExtT);
			var
				t: extended;
				ind1, ind2: integer;
			begin
				ind1 := (e mod 2);
				ind2 := 2 - ind1;
				ind1 := ind1 + 1;
				if currentFrame.proD^.projZ then begin
					t := (p.p[ind1] * dist - clip[e] * (dist + p.p[3])) / (dist * (p.p[ind1] - q.p[ind1]) + clip[e] * (q.p[3] - p.p[3]));
					if (t < 0.0) then begin
						SysBeep(1);
						t := 0.0;
					end
					else if t > 1.0 then begin
						SysBeep(1);
						t := 1.0;
					end;
					s.p[1] := p.p[1] + t * (q.p[1] - p.p[1]);
					s.p[2] := p.p[2] + t * (q.p[2] - p.p[2]);
					s.p[3] := p.p[3] + t * (q.p[3] - p.p[3]);
					s.p[4] := 1.0;
					t := 1.0 + s.p[3] / dist;
					s.vis[1] := s.p[2] >= t * clip[1];
					s.vis[2] := s.p[1] <= t * clip[2];
					s.vis[3] := s.p[2] <= t * clip[3];
					s.vis[4] := s.p[1] >= t * clip[4];
				end
				else begin
					t := (clip[e] - p.p[ind1]) / (q.p[ind1] - p.p[ind1]);
					if (t < 0.0) | (t > 1.0) then
						SysBeep(2);
					s.p[ind1] := clip[e];
					s.p[ind2] := p.p[ind2] + t * (q.p[ind2] - p.p[ind2]);
					s.p[3] := p.p[3] + t * (q.p[3] - p.p[3]);
					s.p[4] := 1.0;
					s.vis[1] := s.p[2] >= clip[1];
					s.vis[2] := s.p[1] <= clip[2];
					s.vis[3] := s.p[2] <= clip[3];
					s.vis[4] := s.p[1] >= clip[4];
				end;
			end;

		procedure prDrawLine (p: point3dExtT);
			var
				ix, iy: integer;
			begin
				if not currentFrame.proD^.projZ then begin
{$IFC THINK_PASCAL }
					ix := round(currentFrame.dx + (p.p[1] + transX) * currentFrame.mx);
					iy := round(currentFrame.dy + (p.p[2] + transY) * currentFrame.my);
{$ELSEC }
					ix := roundLong(currentFrame.dx + (p.p[1] + transX) * currentFrame.mx);
					iy := roundLong(currentFrame.dy + (p.p[2] + transY) * currentFrame.my);
{$ENDC }
				end
				else begin
					p.p[3] := p.p[3] + dist;
					if p.p[3] >= eps then begin
						p.p[3] := dist / p.p[3];
						p.p[1] := p.p[1] * p.p[3];
						p.p[2] := p.p[2] * p.p[3];
{$IFC THINK_PASCAL }
						ix := round(currentFrame.dx + (p.p[1] + transX) * currentFrame.mx);
						iy := round(currentFrame.dy + (p.p[2] + transY) * currentFrame.my);
{$ELSEC }
						ix := roundLong(currentFrame.dx + (p.p[1] + transX) * currentFrame.mx);
						iy := roundLong(currentFrame.dy + (p.p[2] + transY) * currentFrame.my);
{$ENDC }
					end
					else begin
						SysBeep(2);
						exit(prDrawLine);
					end;
				end;
				if neverMoved then begin
					firstX := ix;
					firstY := iy;
					if fillIt & fillNeverMoved then begin
						fillFirstX := ix;
						fillFirstY := iy;
						MoveTo(fillFirstX, fillFirstY);
						fillNeverMoved := false;
					end
					else if fillIt then
						LineTo(firstX, firstY)
					else
						MoveTo(firstX, firstY);
					neverMoved := false;
				end
				else
					LineTo(ix, iy);
				ExtendDrawingSheet(ix, iy);
			end;


		procedure prLineTo (e: integer; var p: point3dExtT);
			var
				iSp: point3dExtT;
			begin
				pp[e] := p;
				if firstExist[e] then begin
					if (sp[e].vis[e] & not pp[e].vis[e]) | (pp[e].vis[e] & not sp[e].vis[e]) then begin
						intSec(e, sp[e], pp[e], iSp);
						if e <> 4 then
							prLineTo(e + 1, iSp)
						else
							prDrawLine(iSp);
					end;
				end
				else begin
					fp[e] := p;
					firstExist[e] := true;
				end;
				sp[e] := pp[e];
				if sp[e].vis[e] then begin
					if e <> 4 then
						prLineTo(e + 1, sp[e])
					else
						prDrawLine(sp[e]);
				end;
			end;

		procedure last (e: integer);
			var
				iSp: point3dExtT;
			begin
				if firstExist[e] then begin
					if (sp[e].vis[e] & not fp[e].vis[e]) | (fp[e].vis[e] & not sp[e].vis[e]) then begin
						intSec(e, sp[e], fp[e], iSp);
						if e <> 4 then
							prLineTo(e + 1, iSp)
						else
							prDrawLine(iSp);
					end;
					if e <> 4 then
						last(e + 1);
				end;
			end;

		begin		(* filledPoly *)
			if not fillIt then
				po := OpenPoly;
			neverMoved := true;
			for i := 1 to 4 do
				firstExist[i] := false;
			for i := 1 to polygon2d(poly).length do begin
				with polygon2d(poly) do begin
					p[1] := point2d(elements^^[i]).getX;
					p[2] := point2d(elements^^[i]).getY;
					p[3] := currentFrame.z;
					p[4] := 1.0;
				end;
				mvm(currentFrame.proD^.viewTrafo, p, pdot);
				pk.p := pdot;
				if currentFrame.proD^.projZ then begin
					k := 1.0 + pk.p[3] / dist;
					pk.vis[1] := pk.p[2] >= k * clip[1];
					pk.vis[2] := pk.p[1] <= k * clip[2];
					pk.vis[3] := pk.p[2] <= k * clip[3];
					pk.vis[4] := pk.p[1] >= k * clip[4];
				end
				else begin
					pk.vis[1] := pk.p[2] >= clip[1];
					pk.vis[2] := pk.p[1] <= clip[2];
					pk.vis[3] := pk.p[2] <= clip[3];
					pk.vis[4] := pk.p[1] >= clip[4];
				end;
				prLineTo(1, pk);
			end;
			last(1);
			if not neverMoved then begin
				LineTo(firstX, firstY);
				if fillIt then
					if (firstX <> fillFirstX) | (firstY <> fillFirstY) then
						LineTo(fillFirstX, fillFirstY);
			end;
			if not fillIt then begin
				ClosePoly;
				FramePoly(po);
				KillPoly(po);
			end;
		end;

	begin	(*polyLayer.interactiveOutput *)
		with currentFrame do begin
			oldZ := currentFrame.z;
			currentFrame.z := self.z;
			if proD^.visibility then begin
				GetPenState(oldPen);										(* set paint mode 				*)
				newPen := oldPen;
				newPen.pnMode := patCopy;
				SetPenState(newPen);
				PenSize(1, 1);
			end;
			if not (coordSysIsDrawn | proD^.visibility) then begin
				coordSysIsDrawn := true;
				drawCoordSys;
			end;
			with clippingRectangle do begin
				windToWorld(left, bottom, clip[4], clip[1]);
				clip[4] := clip[4] - transX;
				clip[1] := clip[1] - transY;
				windToWorld(right, top, clip[2], clip[3]);
				clip[2] := clip[2] - transX;
				clip[3] := clip[3] - transY;
			end;
			mvm(proD^.viewTrafo, proD^.eyePoint, pdot);
			dist := -pdot[3];
			if proD^.visibility then begin
				fillIt := true;
				fillNeverMoved := true;
				fillPo := OpenPoly;
				layElems.forAll(filledPoly);
				ClosePoly;
{$IFC THINK_PASCAL }
				FillPoly(fillPo, white);
{$ELSEC }
				FillPoly(fillPo, qd.white);
{$ENDC }
				KillPoly(fillPo);
			end;
			fillIt := false;
			layElems.forAll(filledPoly);
			if proD^.visibility then
				SetPenState(oldPen);										(* reset paint mode 			*)
		end;
		currentFrame.z := oldZ;
	end;

procedure polyLayer.interactiveHighlight;
	begin
		PenSize(3, 3);
		interactiveOutput;
		PenSize(1, 1);
	end;

procedure initLayer;
	var
		l: layer;
		pl: polyLayer;
	begin
		new(l);
		failNil(Handle(l));
		prototypes[layerType] := l;
		new(pl);
		failNil(Handle(pl));
		prototypes[polyLayerType] := pl;
	end;

end.
