(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit layerVector;

interface

uses
	layer;

type
	layerVector = object(obj)
			dz: extended;
			layers: vector;
			procedure init;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			procedure free;
			override;
			procedure freeAll;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure setFirstZAndDz (firstZ, deltaZ: extended);
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function getBoundingObject: obj;
			override;
			function equal (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
		end;

	polyLayerVector = object(layerVector)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			procedure intersection (it: integer; arguments: list; var result: list);
		end;

procedure initLayerVector;

implementation

const
	setLayerVectDzC = 1;
	makePolyLayerVectorC = 2;
	makeRoadSegmentPlanesC = 3;
	polyLayerUnionC = 4;
	polyLayerIntersectionC = 5;
	polyLayer1diff2C = 6;
	polyLayer2diff1C = 7;
	groundPlanC = 8;
	makePlaneLayerVectorC = 9;


function layerVector.equal (o: obj): boolean;
	var
		t: layerVector;
	begin
		if (o <> nil) & member(o, layerVector) then begin
			t := layerVector(o);
			equal := (dz = t.dz) & layers.equal(t.layers);
		end
		else
			equal := false;
	end;

function layerVector.genericLessThan (o: obj): boolean;
	var
		t: layerVector;
	begin
		if (o <> nil) & member(o, layerVector) then begin
			t := layerVector(o);
			genericLessThan := (dz < t.dz) | (dz = t.dz) & layers.genericLessThan(t.layers);
		end
		else
			genericLessThan := false;
	end;

procedure layerVector.init;
	begin
		dz := 0.0;
		layers := vector(newInitObj(vectorType));
	end;

function layerVector.duplicate: obj;
	var
		l: layerVector;
	begin
		l := layerVector(shallowDuplicate);
		l.layers := vector(layers.duplicate);
		duplicate := l;
	end;

function layerVector.duplicateAll: obj;
	var
		l: layerVector;
	begin
		l := layerVector(shallowDuplicate);
		l.layers := vector(layers.duplicateAll);
		duplicateAll := l;
	end;

procedure layerVector.free;
	begin
		layers.free;
		shallowFree;
	end;

procedure layerVector.freeAll;
	begin
		layers.freeAll;
		shallowFree;
	end;

function layerVector.getType;
	begin
		getType := layerVectorType;
	end;

function layerVector.getTextType;
	begin
		getTextType := stringOf('Layer Vector(', layers.length : 1, ')');
	end;

procedure layerVector.fileInput;
	begin
		layers := vector(newObj(fileGetType));
		layers.fileInput;
		readExtended(dz);
	end;

procedure layerVector.fileOutput;
	begin
		layers.fileOutputTag;
		writeExtended(dz);
	end;

procedure layerVector.textFileInput;
	var
		oLayers: obj;
		r: real;
	begin
		readTextReal(r);
		dz := r;
		oLayers := readTextObj;
		assert(member(oLayers, vector), 'Legal type in layerVector');
		layers := vector(oLayers);
	end;

procedure layerVector.textFileOutput;
	begin
		writeTextReal(dz);
		writeTextObj(layers);
	end;

function layerVector.memberType (o: obj): boolean;
	begin
		memberType := member(o, layerVector);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure layerVector.translate (dx, dy: extended);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function layerVector.isSelected (where: obj): boolean;
	begin
		isSelected := layers.isSelected(where);
	end;

function layerVector.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(1);
		appendObjInfoObject(v, 'Layers', layers, false);
		getInfo := v;
	end;

procedure layerVector.setInfo (o: obj);
	var
		v: vector;
	begin
		v := vector(o);
		layers := vector(objInfoObj(v.elements^^[1]).o);
	end;

function is2d (o: obj): boolean;
	var
		still2d: boolean;
	procedure doCheck (o: obj);
		begin
			if not is2d(o) then
				still2d := false;
		end;
	begin
		if member(o, sequence) then begin
			still2d := true;
			sequence(o).forAll(doCheck);
			is2d := still2d;
		end
		else
			is2d := not member(o, layer);
	end;

procedure layerVector.describe;
	begin
		with currentArguments do
			if (length > 0) then
				if (length = 1) & member(ith(1), layerVector) then begin
					menuEntry('First Z And Distance Between Layers...', setLayerVectDzC);
					menuEntry('Make Ground Plan', groundPlanC);
				end;
	end;

procedure layerVector.execute (item: integer);
	var
		newDz, newZ: extended;
		dummy: boolean;
		v: vector;

	procedure appToV (o: obj);
		begin
			v.appendSafe(layer(o).layElems.duplicateAll, 1);
		end;

	begin
		with currentArguments do
			case item of
				setLayerVectDzC: 
					with layerVector(ith(1)) do begin
						dummy := getUserParameter2('z-Value of first layer', 'Distance between layers', 0, dz, newZ, newDz, nil, nil);
						layers.interactiveHighlight;
						layers.interactiveOutput;
						layerVector(ith(1)).setFirstZAndDz(newZ, newDz);
						layers.interactiveOutput;
						layers.interactiveHighlight;
					end;
				groundPlanC: 
					with layerVector(ith(1)) do begin
						v := vector(newInitObj(vectorType));
						layers.forAll(appToV);
						currentResult.append(v);
					end;
			end;
	end;

procedure layerVector.interactiveOutput;
	var
		i: seqIndex0;
		eyez: extended;
		notEnd: boolean;
	begin
		notEnd := true;
		if currentFrame.proD^.projZ then begin
			eyez := currentFrame.proD^.eyePoint[3];
			i := 1;
			while notEnd do begin
				notEnd := eyez > layer(layers.elements^^[i]).z;
				if notEnd then
					layer(layers.elements^^[i]).interactiveOutput;
				i := i + 1;
				notEnd := notEnd & (i <= layers.length);
			end;
			i := layers.length;
			notEnd := true;
			while notEnd do begin
				notEnd := eyez <= layer(layers.elements^^[i]).z;
				if notEnd then
					layer(layers.elements^^[i]).interactiveOutput;
				i := i - 1;
				notEnd := notEnd & (i >= 1);
			end;
		end
		else begin
			if currentFrame.proD^.eyePoint[3] > currentFrame.proD^.proPoint[3] then
				layers.interactiveOutput
			else begin
				i := layers.length;
				while i > 0 do begin
					layer(layers.elements^^[i]).interactiveOutput;
					i := i - 1;
				end;
			end;
		end;
	end;

function layerVector.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := layers.drawingComplexity(cutOff);
	end;

procedure layerVector.interactiveHighlight;
	begin
		if not currentFrame.proD^.visibility then begin
			layers.interactiveHighlight;
		end;
	end;

procedure layerVector.setFirstZAndDz (firstZ, deltaZ: extended);
	procedure setNewZ (o: obj);
		begin
			layer(o).z := firstZ;
			layer(o).thickness := deltaZ;
			firstZ := firstZ + deltaZ;
		end;
	begin
		dz := deltaZ;
		layers.forAll(setNewZ);
	end;

function polyLayerVector.getType: objectType;
	begin
		getType := polyLayerVectorType;
	end;

function polyLayerVector.getTextType: Str255;
	begin
		getTextType := stringOf('Polygon Layer Vector(', layers.length : 1, ')');
	end;

function polyLayerVector.memberType (o: obj): boolean;
	begin
		memberType := member(o, polyLayerVector);
	end;

function polyOk (x: obj): boolean;
	begin
		polyOk := member(x, polygon2d) & (polygon2d(x).length > 2);
	end;

procedure polyLayerVector.describe;
	var
		dummy: boolean;

	function isPolyVector (o: obj): boolean;
		begin
			isPolyVector := polyOk(o) | member(o, vector) & (vector(o).length > 0) & vector(o).sameProperty(polyOk);
		end;

	procedure testPV (o: obj);
		begin
			if not isPolyVector(o) then
				dummy := false;
		end;

	begin
		with currentArguments do begin
			if length = 2 then
				if member(ith(1), lineSegment) & member(ith(2), polygon2d) & (polygon2d(ith(2)).length > 2) | member(ith(2), lineSegment) & member(ith(1), polygon2d) & (polygon2d(ith(1)).length > 2) then begin
					menuEntry('Make Polygon Layer Plane...', makePlaneLayerVectorC);
					menuEntry('Make Road Segment Planes...', makeRoadSegmentPlanesC);
				end
				else if member(ith(1), polyLayerVector) & member(ith(2), polyLayerVector) then begin
					menuEntry('Union', polyLayerUnionC);
					menuEntry('Intersection', polyLayerIntersectionC);
					menuEntry('1 Difference 2', polyLayer1diff2C);
					menuEntry('2 Difference 1', polyLayer2diff1C);
				end;
			if (length > 0) then begin
				dummy := (length >= 2) & isPolyVector(ith(1)) & isPolyVector(ith(2));  (* for efficiency reasons we first test only two arguments *)
				if dummy & (length > 2) then
					currentArguments.forAll(testPV);  (* if the test for the first two elements successed we've got to test everything *)
				if dummy then
					menuEntry('Make Polygon Layer Vector', makePolyLayerVectorC);
			end;
		end;
	end;

function makePlaneLayerVector (poly: polygon2d; lineSeg: lineSegment; phi, deltaZ, maxZ: extended): polyLayerVector;
	const
		eps = 10E-8;
	var
		plV: polyLayerVector;
		dx, dy, deltaX, deltaY, fact, firstZ: extended;
		po: polygon2d;
		la: polyLayer;
		lS: lineSegment;
		dummy, clip: boolean;
	begin
		if abs(deltaZ) < eps then begin
			makePlaneLayerVector := nil;
			exit(makePlaneLayerVector);
		end;
		if abs(phi) < eps then
			clip := false
		else begin
			clip := true;
			if abs(phi - 90) < eps then begin
				dx := 0.0;
				dy := 0.0;
			end
			else begin
				deltaX := lineSeg.q.getX - lineSeg.p.getX;
				deltaY := lineSeg.q.getY - lineSeg.p.getY;
				fact := deltaZ / (lineSeg.length * tan(phi / 180.0 * pi));
				dx := -deltaY * fact;
				dy := deltaX * fact;
			end;
		end;
		firstZ := 0.0;
		lS := lineSegment(lineSeg.duplicateAll);
		dummy := true;
		while abs(firstZ) <= abs(maxZ) do begin
			if clip then
				po := poly.clip(lS)
			else
				po := polygon2d(poly.duplicateAll);
			if po.length > 2 then begin
				if dummy then begin
					new(plV);
					failNil(Handle(plV));
					plV.init;
					plV.dz := deltaZ;
					dummy := false;
				end;
				new(la);
				failNil(Handle(la));
				la.z := firstZ;
				la.thickness := deltaZ;
				la.layElems := vector(newInitObj(vectorType));
				la.layElems.appendSafe(po, 1);
				plV.layers.appendSafe(la, 1);
			end
			else
				po.freeAll;
			firstZ := firstZ + deltaZ;
			if clip then
				lS.translate(dx, dy);
		end;
		lS.freeAll;
		if dummy then
			plV := nil;
		makePlaneLayerVector := plV;
	end;

function readRoadSegParams (var theta, wideness, pZ, qZ, deltaZ, maxZ: extended): boolean;

	procedure getValue (d: DialogPtr; item: integer; default: extended; var value: extended);
		var
			number: Str255;
			v: extended;
		begin
			GetDText(d, item, number);
			IOCheck(false);
			readString(number, v);
			if ioResult > 0 then begin
				SysBeep(1);
				value := default;
			end
			else
				value := v;
			IOCheck(true);
		end;

		const
			rSResource = 37;
			rSOKItem = 1;
			rSCancelItem = 2;
			rSThetaItem = 3;
			rSWideItem = 4;
			rSpZItem = 5;
			rSqZItem = 6;
			rSdeltaZItem = 7;
			rSmaxZItem = 8;
			eps = 1.0E-08;

		var
			rSDialog: DialogPtr;
			item: integer;
			allowed: boolean;

	begin
		theta := 70.0;
		wideness := 50.0;
		pZ := 50.0;
		qZ := 70.0;
		deltaZ := 2.0;
		maxZ := 100.0;
		beginCursor(stopwatchCursor);
		beginCursor(arrowCursor);
		rSDialog := GetNewDialog(rSResource, nil, WindowPtr(-1));
		niceWindowPosition(WindowPtr(rSDialog), false);
		repeat
			SetDText(rSDialog, rSThetaItem, stringOf(theta : 8 : 4));
			SetDText(rSDialog, rSWideItem, stringOf(wideness : 8 : 4));
			SetDText(rSDialog, rSpZItem, stringOf(pZ : 8 : 4));
			SetDText(rSDialog, rSqZItem, stringOf(qZ : 8 : 4));
			SetDText(rSDialog, rSdeltaZItem, stringOf(deltaZ : 8 : 4));
			SetDText(rSDialog, rSmaxZItem, stringOf(maxZ : 8 : 4));
			SelectWindow(rSDialog);
			ShowWindow(rSDialog);
			showDefaultButton(rSDialog, rSOKItem);
			ModalDialog(nil, item);
			getValue(rSDialog, rSThetaItem, 70.0, theta);
			getValue(rSDialog, rSWideItem, 50.0, wideness);
			getValue(rSDialog, rSpZItem, 50.0, pZ);
			getValue(rSDialog, rSqZItem, 70.0, qZ);
			getValue(rSDialog, rSdeltaZItem, 2.0, deltaZ);
			getValue(rSDialog, rSmaxZItem, 100.0, maxZ);
			if (wideness < eps) | (abs(theta) < eps) | (pZ < 0) | (qZ < 0) | (abs(deltaZ) < eps) | (maxZ < 0.0) then begin
				SysBeep(2);
				allowed := false;
			end
			else
				allowed := true;
		until (item = rSCancelItem) | (item = rSOKItem) & allowed;
		readRoadSegParams := item = rSOKItem;
		HideWindow(rSDialog);
		endCursor;
{$IFC THINK_PASCAL }
		DisposDialog(rSDialog);
{$ELSEC }
		DisposeDialog(rSDialog);
{$ENDC }
		endCursor;
	end;


procedure makeRoadSegmentPlanes (poly: polygon2d; lineSeg: lineSegment; pZ, qZ, theta, wideness, dz, maxZ: extended; var res: list);
	const
		eps = 1.0E-08;
	var
		dist, deltaX, deltaY, deltaZ, cosXi, sinXi, tanTh, dx, dy, r, phi, rp, rq, tp, tq, ap, aq, xp, xq, yp, yq, rXp, rXq, rYp, rYq, dXp, dXq, dYp, dYq: extended;
		plV: polyLayerVector;
		lS: lineSegment;
	begin
		deltaZ := qZ - pZ;
		if (pZ > qZ) & not (abs(deltaZ) < eps) then begin
			new(lS);
			failNil(Handle(lS));
			lS.p := point2d(lineSeg.q.duplicateAll);
			lS.q := point2d(lineSeg.p.duplicateAll);
			makeRoadSegmentPlanes(poly, lS, qZ, pZ, theta, wideness, dz, maxZ, res);
			lS.freeAll;
		end
		else begin
			wideness := wideness / 2.0;
			dist := lineSeg.length;
			deltaX := lineSeg.q.getX - lineSeg.p.getX;
			deltaY := lineSeg.q.getY - lineSeg.p.getY;
			cosXi := deltaX / dist;
			sinXi := deltaY / dist;
			dx := wideness * sinXi;												(* construct road plane			*)
			dy := wideness * cosXi;
			lS := lineSegment(lineSeg.duplicateAll);
			if abs(deltaZ) > eps then begin
				r := pZ * dist / deltaZ;
				phi := arctan(deltaZ / dist) / pi * 180.0;
				lS.q.setX(lineSeg.p.getX - r * cosXi);
				lS.q.setY(lineSeg.p.getY - r * sinXi);
				lS.p.setX(lS.q.getX - dx);
				lS.p.setY(lS.q.getY + dy);
				lS.q.setX(lS.q.getX + dx);
				lS.q.setY(lS.q.getY - dy);
				plV := makePlaneLayerVector(poly, lS, phi, dz, maxZ);
			end
			else begin
				phi := 0.0;
				plV := makePlaneLayerVector(poly, lS, phi, dz, pZ);
			end;
			if plV <> nil then
				res.append(plV);
			tanTh := tan(theta / 180.0 * pi);									(* construct border planes	*)
			rp := pZ / tanTh;
			if abs(deltaZ) <= eps then begin
				rXq := 0.0;
				rYq := 0.0;
				rXp := 0.0;
				rYp := 0.0;
				dXp := (wideness - rp) * sinXi;
				dYp := (wideness - rp) * cosXi;
				dXq := dXp;
				dYq := dYp;
			end
			else begin
				rq := qZ / tanTh;
				ap := r;
				aq := ap + dist;
				tp := sqrt(sqr(ap) - sqr(rp));
				tq := sqrt(sqr(aq) - sqr(rq));
				yp := wideness - tp * rp / ap;
				yq := wideness - tq * rq / aq;
				xp := rp * sqrt(1.0 - sqr(tp) / sqr(ap));
				xq := rq * sqrt(1.0 - sqr(tq) / sqr(aq));
				dXp := yp * sinXi;
				dYp := yp * cosXi;
				dXq := yq * sinXi;
				dYq := yq * cosXi;
				rXp := xp * cosXi;
				rYp := xp * sinXi;
				rXq := xq * cosXi;
				rYq := xq * sinXi;
			end;
			lS.p.setX(lineSeg.p.getX - rXp - dXp);
			lS.p.setY(lineSeg.p.getY - rYp + dYp);
			lS.q.setX(lineSeg.q.getX - rXq - dXq);
			lS.q.setY(lineSeg.q.getY - rYq + dYq);
			plV := makePlaneLayerVector(poly, lS, theta, dz, maxZ);
			if plV <> nil then
				res.append(plV);
			lS.p.setX(lineSeg.q.getX - rXq + dXq);
			lS.p.setY(lineSeg.q.getY - rYq - dYq);
			lS.q.setX(lineSeg.p.getX - rXp + dXp);
			lS.q.setY(lineSeg.p.getY - rYp - dYp);
			plV := makePlaneLayerVector(poly, lS, theta, dz, maxZ);
			if plV <> nil then
				res.append(plV);
			lS.freeAll;
		end;
	end;

procedure polyLayerVector.execute (item: integer);

	var
		ang, deltaZ, maxZ, theta, wideness, pZ, qZ: extended;
		v: vector;
		pl: polyLayer;
		plV: polyLayerVector;
		i: integer;

	procedure appToPL (o: obj);
		begin
			new(pl);
			failNil(Handle(pl));
			if (o.getType = vectorType) | (o.getType = homogenVectorType) then begin
				pl.z := (i - 1) * plV.dz;
				pl.thickness := plV.dz;
				pl.layElems := vector(o.duplicateAll)
			end
			else begin
				pl.z := (i - 1) * plV.dz;
				pl.thickness := plV.dz;
				v := vector(newInitObj(vectorType));
				v.appendSafe(o.duplicateAll, 1);
				pl.layElems := v;
			end;
			vector(plV.layers).appendSafe(pl, 1);
			i := i + 1;
		end;

	begin
		with currentArguments do
			case item of
				makePolyLayerVectorC:  begin
					new(plV);
					failNil(Handle(plV));
					plV.init;
					plV.dz := 5;
					i := 1;
					currentArguments.forAll(appToPL);
					currentResult.append(plV);
				end;
				polyLayerUnionC, polyLayerIntersectionC, polyLayer1diff2C, polyLayer2diff1C: 
					intersection(item, currentArguments, currentResult);
				makePlaneLayerVectorC: 
					if getUserParameter3('Ascent angle', 90, ang, nil, 'Distance between layers', 5.0, deltaZ, nil, 'Maximum of z coordinate', 100.0, maxZ, nil) then begin
						if member(ith(1), polygon2d) then
							plV := makePlaneLayerVector(polygon2d(ith(1)), lineSegment(ith(2)), ang, deltaZ, maxZ)
						else
							plV := makePlaneLayerVector(polygon2d(ith(2)), lineSegment(ith(1)), ang, deltaZ, maxZ);
						if plV <> nil then
							currentResult.append(plV);
					end;
				makeRoadSegmentPlanesC: 
					if readRoadSegParams(theta, wideness, pZ, qZ, deltaZ, maxZ) then
						if member(ith(1), polygon2d) then
							makeRoadSegmentPlanes(polygon2d(ith(1)), lineSegment(ith(2)), pZ, qZ, theta, wideness, deltaZ, maxZ, currentResult)
						else
							makeRoadSegmentPlanes(polygon2d(ith(2)), lineSegment(ith(1)), pZ, qZ, theta, wideness, deltaZ, maxZ, currentResult);
			end;
	end;

procedure polyLayerVector.intersection (it: integer; arguments: list; var result: list);
	var
		lm1, lm2, resLayV: polyLayerVector;
		empty: polyLayer;
		i, j, whichZ: integer;
		endW: boolean;

	procedure doSetOp (pl1, pl2: polyLayer; var res: polyLayerVector);
		var
			pols: vector;
			resPolList: list;
			resPolVect: vector;
			resLay: polyLayer;
			i, j: integer;
			ref1, ref2: polSet;

		function nrOfEls (s: polSet): integer;
			var
				i, j: integer;
			begin
				j := 0;
				for i := 0 to pl1.layElems.length + pl2.layElems.length - 1 do
					if i in s then
						j := j + 1;
				nrOfEls := j;
			end;

		function intZoneQ (s: polSet): boolean;
			begin
				intZoneQ := odd(nrOfEls(s * ref1)) & odd(nrOfEls(s * ref2));
			end;

		function diff12ZoneQ (s: polSet): boolean;
			begin
				diff12ZoneQ := odd(nrOfEls(s * ref1)) & not odd(nrOfEls(s * ref2));
			end;

		function diff21ZoneQ (s: polSet): boolean;
			begin
				diff21ZoneQ := not odd(nrOfEls(s * ref1)) & odd(nrOfEls(s * ref2));
			end;

		function unionZoneQ (s: polSet): boolean;
			begin
				unionZoneQ := odd(nrOfEls(s * ref1)) | odd(nrOfEls(s * ref2));
			end;

		begin
			new(resLay);
			failNil(Handle(resLay));
			if pl1.layElems.length = 0 then
				case it of
					polyLayerUnionC:  begin
						resLay.z := pl2.z;
						resLay.layElems := vector(pl2.layElems.duplicateAll);
						res.layers.appendSafe(resLay, 10);
					end;
					polyLayerIntersectionC: 
						;
					polyLayer1diff2C: 
						;
					polyLayer2diff1C:  begin
						resLay.z := pl2.z;
						resLay.layElems := vector(pl2.layElems.duplicateAll);
						res.layers.appendSafe(resLay, 10);
					end;
				end
			else if pl2.layElems.length = 0 then
				case it of
					polyLayerUnionC:  begin
						resLay.z := pl1.z;
						resLay.layElems := vector(pl1.layElems.duplicateAll);
						res.layers.appendSafe(resLay, 10);
					end;
					polyLayerIntersectionC: 
						;
					polyLayer1diff2C:  begin
						resLay.z := pl1.z;
						resLay.layElems := vector(pl1.layElems.duplicateAll);
						res.layers.appendSafe(resLay, 10);
					end;
					polyLayer2diff1C: 
						;
				end
			else if member(pl1.layElems.elements^^[1], polygon2d) & member(pl2.layElems.elements^^[1], polygon2d) then begin
				i := pl1.layElems.length;
				ref1 := [0..(i - 1)];
				j := pl2.layElems.length - 1;
				ref2 := [i..i + j];
				pols := newVector(pl1.layElems.length);
				for i := 1 to pl1.layElems.length do
					pols.appendSafe(polygon2d(pl1.layElems.elements^^[i]).duplicateAll, 10);
				for i := 1 to pl2.layElems.length do
					pols.appendSafe(polygon2d(pl2.layElems.elements^^[i]).duplicateAll, 10);
				resPolList := newEmptyList;
{$IFC AnimationEnabled }
				abortEnable := false;
{$ENDC }
				case it of
					polyLayerUnionC: 
						polygon2d(pl1.layElems.elements^^[1]).intersection(pols, unionZoneQ, resPolList, false, nil);
					polyLayerIntersectionC: 
						polygon2d(pl1.layElems.elements^^[1]).intersection(pols, intZoneQ, resPolList, false, nil);
					polyLayer1diff2C: 
						polygon2d(pl1.layElems.elements^^[1]).intersection(pols, diff12ZoneQ, resPolList, false, nil);
					polyLayer2diff1C: 
						polygon2d(pl1.layElems.elements^^[1]).intersection(pols, diff21ZoneQ, resPolList, false, nil);
				end;
				if abortFlag then
					i := i + 1 - 1;
{$IFC AnimationEnabled }
				abortEnable := true;
{$ENDC }
				pols.freeAll;
				new(resPolVect);
				failNil(Handle(resPolVect));
				resPolVect.initFromList(resPolList);
				resPolList.free;
				if resPolVect.length = 0 then begin
					resPolVect.free;
					resLay.free;
				end
				else begin
					resLay.layElems := vector(resPolVect);
					if pl1.z > pl2.z then
						resLay.z := pl1.z
					else
						resLay.z := pl2.z;
					res.layers.appendSafe(resLay, 10);
				end;
			end;
		end;

	begin
		new(empty);
		failNil(Handle(empty));
		empty.init;
		lm1 := polyLayerVector(arguments.ith(1));
		lm2 := polyLayerVector(arguments.ith(2));
		currentVectorLength := lm1.layers.length;
		new(resLayV);
		failNil(Handle(resLayV));
		resLayV.init;
		j := 1;
		i := 1;
		endW := false;
		while not endW & not abortFlag do begin
			if (polyLayer(lm1.layers.elements^^[i]).z + polyLayer(lm1.layers.elements^^[i]).thickness) <= polyLayer(lm2.layers.elements^^[1]).z then begin
				doSetOp(polyLayer(lm1.layers.elements^^[i]), empty, resLayV);
				i := i + 1;
				if i > lm1.layers.length then
					endW := true;
			end
			else
				endW := true;
		end;
		endW := false;
		while not endW & not abortFlag do begin
			if (polyLayer(lm2.layers.elements^^[j]).z + polyLayer(lm2.layers.elements^^[j]).thickness) <= polyLayer(lm1.layers.elements^^[1]).z then begin
				doSetOp(empty, polyLayer(lm2.layers.elements^^[j]), resLayV);
				j := j + 1;
				if j > lm2.layers.length then
					endW := true;
			end
			else
				endW := true;
		end;
		if (i <= lm1.layers.length) & (j <= lm2.layers.length) & not abortFlag then
			if polyLayer(lm1.layers.elements^^[i]).z < polyLayer(lm2.layers.elements^^[j]).z then begin
				doSetOp(polyLayer(lm1.layers.elements^^[i]), empty, resLayV);
			end
			else if polyLayer(lm1.layers.elements^^[i]).z > polyLayer(lm2.layers.elements^^[j]).z then begin
				doSetOp(empty, polyLayer(lm2.layers.elements^^[j]), resLayV);
			end;
		if (i <= lm1.layers.length) & (j <= lm2.layers.length) & not abortFlag then
			if polyLayer(lm1.layers.elements^^[i]).z <= polyLayer(lm2.layers.elements^^[j]).z then
				whichZ := 2
			else
				whichZ := 1;
		while (i <= lm1.layers.length) & (j <= lm2.layers.length) & not abortFlag do begin
			doSetOp(polyLayer(lm1.layers.elements^^[i]), polyLayer(lm2.layers.elements^^[j]), resLayV);
			if polyLayer(lm1.layers.elements^^[i]).z = polyLayer(lm2.layers.elements^^[j]).z then begin
				if polyLayer(lm1.layers.elements^^[i]).thickness = polyLayer(lm2.layers.elements^^[j]).thickness then begin
					i := i + 1;
					j := j + 1;
				end
				else if polyLayer(lm1.layers.elements^^[i]).thickness > polyLayer(lm2.layers.elements^^[j]).thickness then
					j := j + 1
				else
					i := i + 1;
			end
			else begin
				if (polyLayer(lm1.layers.elements^^[i]).z + polyLayer(lm1.layers.elements^^[i]).thickness) = (polyLayer(lm2.layers.elements^^[j]).z + polyLayer(lm2.layers.elements^^[j]).thickness) then begin
					i := i + 1;
					j := j + 1;
				end
				else if (polyLayer(lm1.layers.elements^^[i]).z + polyLayer(lm1.layers.elements^^[i]).thickness) < (polyLayer(lm2.layers.elements^^[j]).z + polyLayer(lm2.layers.elements^^[j]).thickness) then begin
					i := i + 1;
					whichZ := 1;
				end
				else begin
					j := j + 1;
					whichZ := 2;
				end;
			end;
		end;
		whichZ := 1;
		while (i <= lm1.layers.length) & not abortFlag do begin
			if polyLayer(lm1.layers.elements^^[i]).z > (polyLayer(lm2.layers.elements^^[lm2.layers.length]).z + polyLayer(lm2.layers.elements^^[lm2.layers.length]).thickness) then
				doSetOp(polyLayer(lm1.layers.elements^^[i]), empty, resLayV);
			i := i + 1;
		end;
		whichZ := 2;
		while (j <= lm2.layers.length) & not abortFlag do begin
			if polyLayer(lm2.layers.elements^^[j]).z > (polyLayer(lm1.layers.elements^^[lm1.layers.length]).z + polyLayer(lm1.layers.elements^^[lm1.layers.length]).thickness) then
				doSetOp(empty, polyLayer(lm2.layers.elements^^[j]), resLayV);
			j := j + 1;
		end;
		if (resLayV.layers.length > 0) & not abortFlag then begin
			resLayV.dz := lm1.dz;
			for i := 1 to (resLayV.layers.length - 1) do
				polyLayer(resLayV.layers.elements^^[i]).thickness := polyLayer(resLayV.layers.elements^^[i + 1]).z - polyLayer(resLayV.layers.elements^^[i]).z;
			polyLayer(resLayV.layers.elements^^[resLayV.layers.length]).thickness := polyLayer(lm1.layers.elements^^[lm1.layers.length]).thickness;
			if polyLayer(lm1.layers.elements^^[lm1.layers.length]).thickness > polyLayer(lm2.layers.elements^^[lm2.layers.length]).thickness then
				polyLayer(resLayV.layers.elements^^[resLayV.layers.length]).thickness := polyLayer(lm2.layers.elements^^[lm1.layers.length]).thickness;
			result.append(resLayV);
		end
		else
			resLayV.freeAll;
		empty.freeAll;
	end;

procedure initLayerVector;
	var
		l: layerVector;
		pl: polyLayerVector;
	begin
		new(l);
		failNil(Handle(l));
		prototypes[layerVectorType] := l;
		new(pl);
		failNil(Handle(pl));
		prototypes[polyLayerVectorType] := pl;
	end;

end.
