(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit lineSegment;

interface

uses
{$IFC GFEnabled }
	GFValues, 
{$ENDC}
	regularNGon;

procedure initLineSegment;

implementation

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function lineSegment.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 2;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure lineSegment.init;
	begin
		p := point2d(newInitObj(currentPointType));
		q := point2d(newInitObj(currentPointType));
	end;

function lineSegment.duplicate: obj;
	var
		t: lineSegment;
	begin
		t := lineSegment(shallowDuplicate);
		t.p := point2d(p.duplicate);
		t.q := point2d(q.duplicate);
		duplicate := t;
	end;

procedure lineSegment.free;
	begin
		if p <> nil then
			p.free;
		if q <> nil then
			q.free;
		shallowFree;
	end;

procedure lineSegment.construct (px, py, newX, newY: integer);
	begin
		p.setX(px);
		p.setY(py);
		q.setX(newX);
		q.setY(newY);
	end;

procedure lineSegment.showObjectParameter;
	begin
		displayInfoString(StringOf('Length = ', p.distanceX(q) : 8 : 3, '. Lp Length = ', p.distanceLpX(q) : 8 : 3, '.'));
	end;

function lineSegment.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	var
		d, dxqp, dyqp: extended;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
	procedure doDraw (o: obj);
		begin
			if highlight then
				interactiveHighlight;
			interactiveOutput;
		end;

	procedure doDrawShowObjectParameter (o: obj);
		begin
			interactiveOutput;
			if highlight then
				interactiveHighlight;
			showObjectParameter;
		end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

	function processPoint (x: point2d): boolean;

		procedure transformTranslation (var dx, dy: extended);

			procedure newDxDyKeepLength (pp, qq: point2d);
				var
					drpx, drpy, lambda: extended;
				begin
					drpx := qq.getX + dx - pp.getX;
					drpy := qq.getY + dy - pp.getY;
					if (drpx = 0) & (drpy = 0) then begin
						dx := 0;
						dy := 0;
					end
					else begin
						lambda := d / sqrt(sqr(drpx) + sqr(drpy));
						dx := lambda * drpx + pp.getX - qq.getX;
						dy := lambda * drpy + pp.getY - qq.getY;
					end;
				end;

			procedure newDxDyKeepDirection (pp, qq: point2d; oldDxqp, oldDyqp: extended);
				var
					dxqp, dyqp, lambda: extended;
				begin
					if d = 0 then begin
						dx := 0;
						dy := 0;
					end
					else begin
						dxqp := qq.getX - pp.getX;
						dyqp := qq.getY - pp.getY;
						lambda := sqrt(sqr(dxqp + dx) + sqr(dyqp + dy)) / d;
						dx := lambda * oldDxqp - dxqp;
						dy := lambda * oldDyqp - dyqp;
					end;
				end;

			begin (* transformTranslation *)
				GetKeys(globalKeyMap);
				if globalKeyMap[optionKeyValue] then
					if x = p then
						newDxDyKeepLength(q, p)
					else
						newDxDyKeepLength(p, q)
				else if globalKeyMap[commandKeyValue] then
					if x = p then
						newDxDyKeepDirection(q, p, -dxqp, -dyqp)
					else
						newDxDyKeepDirection(p, q, dxqp, dyqp);
			end;

		begin (* processPoint *)
			if x.isSelected(where) then begin
				processPoint := true;
				x.interactiveOutput;
				x.dragObjectTransformTranslation(doDraw, doDrawShowObjectParameter, mouse, transformTranslation);
				x.interactiveOutput;
			end
			else
				processPoint := false;
		end;

	begin (* lineSegment.dragVertex *)
		d := p.distanceX(q);
		dxqp := q.getX - p.getX;
		dyqp := q.getY - p.getY;
		dragVertex := processPoint(p) | processPoint(q);
	end;

procedure lineSegment.windowToWorld;
	begin
		p.windowToWorld;
		q.windowToWorld;
	end;

function lineSegment.isSelected (where: obj): boolean;
	var
		result: boolean;
	function testEdges (px, py, qx, qy, left, right, bottom, top: extended): boolean;
		function eval (px, py, qx, qy, x: real): extended;
			begin
				eval := (x - px) * (qy - py) / (qx - px) + py;
			end;
		function inInterval (x, left, right: real): boolean;
			begin
				inInterval := (left <= x) & (x <= right) | (right <= x) & (x <= left);
			end;
		begin
			testEdges := (px <> qx) & (inInterval(left, px, qx) & inInterval(eval(px, py, qx, qy, left), bottom, top) | inInterval(right, px, qx) & inInterval(eval(px, py, qx, qy, right), bottom, top));
		end;
	begin
		if member(where, rectangle) then begin
			with rectangle(where) do
				result := testEdges(p.getX, p.getY, q.getX, q.getY, left, right, bottom, top) | testEdges(p.getY, p.getX, q.getY, q.getX, bottom, top, left, right);
			isSelected := result | p.isSelected(where);
		end
		else
			isSelected := false;
	end;

procedure lineSegment.windowOutput;
	begin
{$IFC THINK_PASCAL }
		MoveTo(round(p.getX), round(p.getY));
		LineTo(round(q.getX), round(q.getY));
{$ELSEC }
		MoveTo(roundLong(p.getX), roundLong(p.getY));
		LineTo(roundLong(q.getX), roundLong(q.getY));
{$ENDC }
	end;

procedure lineSegment.interactiveOutput;
	begin
		if p.equal(q) then
			p.interactiveOutput
		else
			p.drawLine(q);
	end;

procedure lineSegment.interactiveHighlight;
	begin
		if p.equalPoint(q) then
			p.interactiveHighlight
		else
			inherited interactiveHighlight;
	end;

procedure lineSegment.textFileInput;
	var
		op, oq: obj;
	begin
		op := readTextObj;
		oq := readTextObj;
		assert(member(op, point2d) & member(oq, point2d) & sameType2(op, oq), 'Legal type in line segment');
		p := point2d(op);
		q := point2d(oq);
	end;

procedure lineSegment.textFileOutput;
	begin
		writeTextObj(p);
		writeTextObj(q);
	end;

procedure lineSegment.fileInput;
	var
		t: objectType;
	begin
		t := fileGetType;
		p := point2d(newObj(t));
		q := point2d(newObj(t));
		p.fileInput;
		q.fileInput;
	end;

procedure lineSegment.fileOutput;
	begin
		p.fileOutputTag;
		q.fileOutput;
	end;

procedure lineSegment.textOutput;
	begin
		write('LS[');
		p.textOutput;
		write(';');
		q.textOutput;
		write('] ');
	end;

procedure lineSegment.randomChange;
	begin
		p.randomChange;
		q.randomChange;
	end;

function lineSegment.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, lineSegment) & p.equal(lineSegment(o).p) & q.equal(lineSegment(o).q);
	end;

function lineSegment.memberType (o: obj): boolean;
	begin
		memberType := member(o, lineSegment) & (lineSegment(o).p.memberType(p)) & p.memberType(lineSegment(o).p);
	end;

procedure lineSegment.translate (dx, dy: extended);
	begin
		p.translate(dx, dy);
		q.translate(dx, dy);
	end;

procedure lineSegment.rotate (xOrigin, yOrigin, alpha: extended);
	begin
		p.rotate(xOrigin, yOrigin, alpha);
		q.rotate(xOrigin, yOrigin, alpha);
	end;

procedure lineSegment.scale (lambda: extended);
	begin
		p.scale(lambda);
		q.scale(lambda);
	end;

procedure lineSegment.reflect (a, b, c: extended);
	begin
		p.reflect(a, b, c);
		q.reflect(a, b, c);
	end;

function lineSegment.pointCover (n: longInt): obj;
	var
		dx, dy, currentX, currentY: extended;
		r: point2d;
		v: vector;
		i: longInt;
	begin
		v := newPointVector(n);
		currentX := p.getX;
		currentY := p.getY;
		dx := (q.getX - currentX) / (n - 1);
		dy := (q.getY - currentY) / (n - 1);
		r := point2d(p.convertToType(currentPointType));
		v.append(r);
		for i := 1 to n - 2 do begin
			currentX := currentX + dx;
			currentY := currentY + dy;
			r := point2d(r.duplicate);
			r.setXY(currentX, currentY);
			v.append(r);
		end;
		v.append(q.convertToType(currentPointType));
		pointCover := v;
	end;

function lineSegment.randomPointCover (n: longInt): obj;
	var
		px, py, qx, qy, l: extended;
		r: point2d;
		v: vector;
		i: longInt;
	begin
		v := newPointVector(n);
		px := p.getX;
		py := p.getY;
		qx := q.getX;
		qy := q.getY;
		for i := 1 to n do begin
			r := point2d(newObj(currentPointType));
			l := randomReal;
			r.setXY(px + l * (qx - px), py + l * (qy - py));
			v.append(r);
		end;
		randomPointCover := v;
	end;

function lineSegment.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(2);
		appendObjInfoObject(v, 'P', p, false);
		appendObjInfoObject(v, 'Q', q, false);
		getInfo := v;
	end;

function lineSegment.getType: objectType;
	begin
		getType := lineSegmentType;
	end;

function lineSegment.getTextType: Str255;
	begin
		getTextType := concat('Line Segment(', p.getTextType, ')');
	end;

function lineSegment.genericLessThan (o: obj): boolean;
	begin
		genericLessThan := p.genericLessThan(lineSegment(o).p) | (p.equal(lineSegment(o).p) & q.genericLessThan(lineSegment(o).q));
	end;

function lineSegment.convertToType (t: pointTypeRange): obj;
	var
		result: lineSegment;
	begin
		new(result);
		failNil(Handle(result));
		result.p := point2d(p.convertToType(t));
		result.q := point2d(q.convertToType(t));
		convertToType := result;
	end;

procedure lineSegment.centerOfGravity (var cx, cy: extended);
	begin
		if p.getY > q.getY then begin
			cx := p.getX;
			cy := p.getY;
		end
		else begin
			cx := q.getX;
			cy := q.getY;
		end;
	end;

function lineSegment.whichSideX (r: point2d): extended;
	begin
		whichSideX := r.whichSideX(p, q);
	end;

function lineSegment.intersectLineSegmentX (r, l: lineSegment): boolean;
	begin
		if longInt(self) < longInt(r) then
			intersectLineSegmentX := p.intersectLineSegmentX(q, r.p, r.q, l.p, l.q)
		else
			intersectLineSegmentX := r.p.intersectLineSegmentX(r.q, p, q, l.p, l.q);
	end;

function lineSegment.intersectLineX (r, l: lineSegment): boolean;
	begin
		intersectLineX := r.p.intersectLineX(r.q, p, q, l.p, l.q);
	end;

function lineSegment.lineIntersectsLine (l: lineSegment; intersectionPoint: point2d): boolean;
	begin
		lineIntersectsLine := p.lineIntersectsLine(q, l.p, l.q, intersectionPoint);
	end;

function getLeft (ls: lineSegment): point2d;
	begin
		if ls.p.genericLessThan(ls.q) then
			getLeft := ls.p
		else
			getLeft := ls.q;
	end;

function getRight (ls: lineSegment): point2d;
	begin
		if ls.p.genericLessThan(ls.q) then
			getRight := ls.q
		else
			getRight := ls.p;
	end;

function lineSegment.HBGLessThan (o: obj): boolean; (* Eingefuegt von Michael Vollenbroeker *)
	var
		s: extended;
		ls: lineSegment;

	function Qquer (cp: lineSegment; s: extended): extended;  (* bestimmt die y-Koordinate des Schnittpunkts von Qquer(cp) mit der sweepline an der Stelle x=s *)

		begin
			if s >= point2d(getRight(cp)).getX then
				Qquer := point2d(getRight(cp)).getY
			else
				Qquer := point2d(getLeft(cp)).getY + (point2d(getRight(cp)).getY - point2d(getLeft(cp)).getY) * (s - point2d(getLeft(cp)).getX) / (point2d(getRight(cp)).getX - point2d(getLeft(cp)).getX);
		end;

	begin
		ls := lineSegment(o);
		if point2d(getLeft(self)).xLessThan(point2d(getLeft(ls))) then                       (* 'self.left' links von 'ls.left' *)
			s := point2d(getLeft(ls)).getX
		else
			s := point2d(getLeft(self)).getX;
		HBGLessThan := Qquer(self, s) < Qquer(ls, s);
	end;

function lineSegment.verticalDistance (o: obj): extended;  (* Eingefuegt von Michael Vollenbroeker 14.1.93 *)
	var
		hps, lps, hpo, lpo: point2d;
		ls: lineSegment;
	begin
		ls := lineSegment(o);
		if p.yLessThan(q) then begin
			hps := q;
			lps := p;
		end
		else begin
			hps := p;
			lps := q;
		end;
		if ls.p.yLessThan(ls.q) then begin
			hpo := ls.q;
			lpo := ls.p;
		end
		else begin
			hpo := ls.p;
			lpo := ls.q;
		end;
		if hpo.yLessThan(lps) then
			verticalDistance := lps.getY - hpo.getY
		else if hps.yLessThan(lpo) then
			verticalDistance := lpo.getY - hps.getY
		else
			verticalDistance := 0;
	end;

function lineSegment.getPointType: objectType;  (*  Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	begin
		getPointType := p.getType;
	end;

function lineSegment.distanceO (o: obj; var minP, minQ: obj): extended;
(* berechnet den Abstand von 'self' und 'o', der in p und q angenommen wird. *)
(* Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	var
		ls: lineSegment;
	begin
		new(ls);
		failNil(Handle(ls));
		ls.p := point2d(newObj(p.getType));
		ls.q := point2d(newObj(p.getType));
		if intersectLineSegmentX(lineSegment(o), ls) then begin
			point2d(minP).setPoint(ls.p);
			point2d(minQ).setPoint(point2d(minP));
			distanceO := 0;
		end
		else begin
			distanceO := distanceLineSegmentsX(lineSegment(o), ls);
			point2d(minP).setPoint(ls.p);
			point2d(minQ).setPoint(ls.q);
		end;
		ls.freeAll;
	end;

function lineSegment.getLeftPoint: obj;
(* liefert den am weitesten links liegenden Punkt von 'self'.  *)
(* Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	begin
		getLeftPoint := getLeft(self).duplicate;
	end;

function lineSegment.getRightPoint: obj;
(* liefert den am weitesten rechts liegenden Punkt von 'self'.  *)
(* Eingefuegt von Michael Vollenbroeker 12.2.93 *)
	begin
		getRightPoint := getRight(self).duplicate;
	end;

function lineSegment.length: extended;
	begin
		length := p.distanceX(q);
	end;

function lineSegment.perimeter: extended;
	begin
		perimeter := p.distanceLpX(q);
	end;

function lineSegment.lengthMP: markedRealPoint;
	var
		mp: markedRealPoint;
	begin
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.initLocationMark(self, stringOf(length : 8 : 5), true);
		lengthMP := mp;
	end;

procedure lineSegment.projectOnLine (candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
	begin
		p.projectOnLine(q, candidatePoint, projectionPoint, distanceToLine, whichSide, lambdaSign);
	end;

function lineSegment.distanceToPointX (candPoint: point2d; var closestX, closestY: extended): extended;    (* added by FB *)
	begin
		distanceToPointX := candPoint.distanceToLineSegmentX(p, q, closestX, closestY);
	end;

function lineSegment.distanceLineSegmentsX (l: lineSegment; var shortestConnection: lineSegment): extended; (* added by FB *)
	procedure update (distance: extended; clos: point2d; closestX, closestY: extended; var delta: extended);
		begin
			if distance < delta then begin
				delta := distance;
				shortestConnection.p.setX(closestX);
				shortestConnection.p.setY(closestY);
				shortestConnection.q.setX(clos.getX);
				shortestConnection.q.setY(clos.getY);
			end;
		end;

		var
			distance, delta, closestX, closestY: extended;
	begin
		delta := 1000000000;
		distance := self.distanceToPointX(point2d(l.p), closestX, closestY);
		update(distance, l.p, closestX, closestY, delta);
		distance := self.distanceToPointX(point2d(l.q), closestX, closestY);
		update(distance, l.q, closestX, closestY, delta);
		distance := l.distanceToPointX(self.p, closestX, closestY);
		update(distance, self.p, closestX, closestY, delta);
		distance := l.distanceToPointX(self.q, closestX, closestY);
		update(distance, self.q, closestX, closestY, delta);
		distanceLineSegmentsX := delta;
	end;

function lineSegment.collinear (l: lineSegment): boolean;
	begin
		collinear := (p.crossProductX(q, l.p, l.q) = 0) & (p.whichSideX(q, l.p) = 0);
	end;

{$IFC GFEnabled }
function lineSegment.convertToParameterPoint: obj;
	var
		pP: pPoint;
		r: Creal;
	begin
		NEW(pP);
		failNil(Handle(pP));
		pP.dimension := 4;
		NEW(r);
		failNil(Handle(r));
		r.value := p.getX;
		pP.fields[1] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := p.getY;
		pP.fields[2] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := q.getX;
		pP.fields[3] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := q.getY;
		pP.fields[4] := r;
		convertToParameterPoint := obj(pP);
	end; { convertToParameterPoint }

function lineSegment.convertFromParameterPoint (pP: obj): obj;
	var
		l: lineSegment;
	begin
		l := lineSegment(newInitObj(lineSegmentType));
		realPoint(l.p).x := Creal(pPoint(pP).fields[1]).value;
		realPoint(l.p).y := Creal(pPoint(pP).fields[2]).value;
		realPoint(l.q).x := Creal(pPoint(pP).fields[3]).value;
		realPoint(l.q).y := Creal(pPoint(pP).fields[4]).value;
		convertFromParameterPoint := obj(l);
	end; { convertFromParameterPoint }

procedure lineSegment.createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
	var
		l: lineSegment;
	begin
		whichAttributes[1] := true;
		whichAttributes[2] := true;
		whichAttributes[3] := false;
		whichAttributes[4] := false;
		l := lineSegment(newInitObj(lineSegmentType));
		realPoint(l.p).x := rectangle(rect).left;
		realPoint(l.p).y := rectangle(rect).bottom;
		lowpP := pPoint(l.convertToParameterPoint);
		realPoint(l.p).x := rectangle(rect).right;
		realPoint(l.p).y := rectangle(rect).top;
		highpP := pPoint(l.convertToParameterPoint);
		l.free;
	end; { createQuery }

function lineSegment.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := true;
	end; { canUseGridfile }
{$ENDC}

{***********************************************}

	const
		TwoPointsC = 0;
		ReflectC = 1;
		ToDirectedSegmentC = 2;
		RevertC = 3;
		ShowOrientationC = 4;
		EuclideanLengthC = 5;

procedure lineSegment.describe;
	begin
		with currentArguments do begin
			if (length = 1) & member(ith(1), lineSegment) then begin
				menuEntry('Euclidean Length', EuclideanLengthC);
				menuEntry('Two Points', TwoPointsC);
				menuEntry('Show Orientation', ShowOrientationC);
				menuEntry('Revert Orientation', RevertC);
				menuEntry('To Directed Segment', ToDirectedSegmentC);
			end;
			if (length = 2) & (member(ith(1), lineSegment) | member(ith(2), lineSegment)) then
				menuEntry('Reflect', ReflectC);
		end;
	end;

procedure revert (l: lineSegment);
	var
		t: point2d;
	begin
		with l do begin
			t := p;
			p := q;
			q := t;
		end;
	end;

function toDirectedSegment (l: lineSegment): directedLineSegment;
	var
		d: directedLineSegment;
	begin
		new(d);
		failNil(Handle(d));
		d.p := point2d(l.p.duplicate);
		d.q := point2d(l.q.duplicate);
		toDirectedSegment := d;
	end;

procedure lineSegment.execute (item: integer);
	var
		l1: lineSegment;

	function callReflect (l: lineSegment; o: obj): obj;
		var
			a, b: extended;
			toReflect: obj;
		begin
			toReflect := o.duplicateAll;
			a := l.q.getY - l.p.getY;
			b := l.p.getX - l.q.getX;
			toReflect.reflect(a, b, l.p.getY * b + l.p.getX * a);
			callReflect := toReflect;
		end;

	function showOrientation (l: lineSegment): vector;
		var
			poly: pointVector;
		begin
			poly := pointVector(l.pointCover(2));
			showOrientation := poly.showNumbering;
			poly.freeAll;
		end;

	begin
		with currentArguments do
			if member(ith(1), lineSegment) then
				l1 := lineSegment(ith(1))
			else
				l1 := nil;
		case item of
			RevertC: 
				revert(l1);
			ToDirectedSegmentC: 
				currentResult.append(toDirectedSegment(l1));
			EuclideanLengthC: 
				currentResult.append(l1.lengthMP);
			ShowOrientationC: 
				currentResult.append(showOrientation(l1));
			TwoPointsC: 
				with l1 do begin
					currentResult.append(p.duplicate);
					currentResult.append(q.duplicate);
				end;
			ReflectC: 
				if l1 = nil then
					currentResult.append(callReflect(lineSegment(currentArguments.ith(2)), currentArguments.ith(1)))
				else
					currentResult.append(callReflect(l1, currentArguments.ith(2)));
		end;
	end;

procedure directedLineSegment.windowOutput;
	var
		safe: wFrame;
	begin
		safe := currentFrame;
		currentFrame := standardFrame;
		interactiveOutput;
		currentFrame := safe;
	end;

	const
		k = 1.5;
	var
		cis : circle;
		
function directedLineSegment.isSelected (where: obj): boolean;
	var
		x1, y1, x2, y2: real;
	begin
		x1 := p.getX;
		y1 := p.getY;
		x2 := q.getX;
		y2 := q.getY;
		if (x1 = x2) & (y1 = y2) then
			isSelected := inherited isSelected(where)
		else begin
			cis.x := (x1 + x2) / 2 + k * (y1 - y2);
			cis.y := (y1 + y2) / 2 + k * (x2 - x1);
			cis.radius := sqrt(sqr(x1 - cis.x) + sqr(y1 - cis.y));
			isSelected := cis.isSelected(where);
		end;
	end;
	
procedure directedLineSegment.interactiveOutput; (* with code from PV *)
	const
		maxDistance = 5;
		conversion = 57.29577951; (* 180 / pi  *)
	var
		mx, my, x1, y1, x2, y2, radius, m, dx, dy, proportion, pqDistance, lambda, markDistanceFromQ, tangentX, tangentY, markX, markY, arrowVertex1X, arrowVertex1Y, arrowVertex2X, arrowVertex2Y: real;
		startAngle, endAngle, arc: integer;
	begin
		x1 := p.getX;
		y1 := p.getY;
		x2 := q.getX;
		y2 := q.getY;
		if (x1 = x2) & (y1 = y2) then
			p.interactiveOutput
		else begin
			mx := (x1 + x2) / 2 + k * (y1 - y2);
			my := (y1 + y2) / 2 + k * (x2 - x1);
			radius := sqrt(sqr(x1 - mx) + sqr(y1 - my));
{$IFC THINK_PASCAL }
			startAngle := round(argument(x1 - mx, y1 - my) * conversion);
			endAngle := round(argument(x2 - mx, y2 - my) * conversion);
{$ELSEC }
			startAngle := roundLong(argument(x1 - mx, y1 - my) * conversion);
			endAngle := roundLong(argument(x2 - mx, y2 - my) * conversion);
{$ENDC }
			arc := endAngle - startAngle;
			if arc > 180 then
				arc := arc - 360
			else if arc < -180 then
				arc := arc + 360;
			drawCircle(mx, my, radius, startAngle + 90, arc);
			proportion := 0.2;
	(* The arrow begins at (1-proportion) x pqDistance. *)
			pqDistance := p.distanceX(q);
			markDistanceFromQ := pqDistance * proportion;
			if markDistanceFromQ > maxDistance then
	(* Make new poportions. *)
				proportion := maxDistance / pqDistance;
			markX := proportion * x1 + (1 - proportion) * x2;
			markY := proportion * y1 + (1 - proportion) * y2;

			lambda := radius / pqDistance;
			tangentX := mx - lambda * (y1 - y2) - x2;
			tangentY := my - lambda * (x2 - x1) - y2;

	(* Draw arrow from mark until q. *)
			dx := x2 - markX;
			dy := markY - y2;
			if dx <> 0 then
				m := dy / dx
			else
				m := 0;  (* The gradient m is really infinite, but the arrow can be drawn like m was 0. *)
			dx := abs(dx);
			dy := abs(dy);
			if m <= 0 then begin
				arrowVertex1X := markX - dy;
				arrowVertex1Y := markY + dx;
				arrowVertex2X := markX + dy;
				arrowVertex2Y := markY - dx;
			end
			else begin
				arrowVertex1X := markX + dy;
				arrowVertex1Y := markY + dx;
				arrowVertex2X := markX - dy;
				arrowVertex2Y := markY - dx;
			end;
			drawLineSegment(x2 + tangentX, y2 + tangentY, arrowVertex1X + tangentX, arrowVertex1Y + tangentY);
			drawLineSegment(x2 + tangentX, y2 + tangentY, arrowVertex2X + tangentX, arrowVertex2Y + tangentY);
		end;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function directedLineSegment.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 4;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function directedLineSegment.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, directedLineSegment) & inherited equal(o);
	end;

function directedLineSegment.memberType (o: obj): boolean;
	begin
		memberType := member(o, directedLineSegment) & inherited memberType(o);
	end;

function directedLineSegment.getType: objectType;
	begin
		getType := directedLineSegmentType;
	end;

function directedLineSegment.getTextType: Str255;
	begin
		getTextType := concat('Directed Segment(', p.getTextType, ')');
	end;

procedure initLineSegment;
	begin
	new(cis);
	failNil(Handle(cis));
	end;

end.
