(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit longIntPoint;

interface

uses
	basicGeometry;

implementation

function ws (p, q, r: longIntPoint): longInt;
	begin
		ws := (p.x - q.x) * (q.y - r.y) - (p.y - q.y) * (q.x - r.x);
	end;

function longIntPoint.verticalDistance (o: obj): extended;
	begin
		verticalDistance := abs(y - longIntPoint(o).y);
	end;

function longIntPoint.distanceO (o: obj; var minP, minQ: obj): extended;
	begin
		longIntPoint(minP).x := x;
		longIntPoint(minP).y := y;
		longIntPoint(minQ).x := longIntPoint(o).x;
		longIntPoint(minQ).y := longIntPoint(o).y;
		distanceO := distanceX(point2d(o));
	end;

function longIntPoint.getX: extended;
	begin
		getX := x;
	end;

procedure longIntPoint.setX (newX: extended);
	begin
{$IFC THINK_PASCAL }
		x := round(newX);
{$ELSEC }
		x := roundLong(newX);
{$ENDC }
	end;

function longIntPoint.getY: extended;
	begin
		getY := y;
	end;

procedure longIntPoint.setY (newY: extended);
	begin
{$IFC THINK_PASCAL }
		y := round(newY);
{$ELSEC }
		y := roundLong(newY);
{$ENDC }
	end;

function longIntPoint.xLessThan (p: point2d): boolean;
	begin
		xLessThan := x < longIntPoint(p).x;
	end;

function longIntPoint.xEqual (p: point2d): boolean;
	begin
		xEqual := x = longIntPoint(p).x;
	end;

function longIntPoint.yLessThan (p: point2d): boolean;
	begin
		yLessThan := y < longIntPoint(p).y;
	end;

function longIntPoint.yEqual (p: point2d): boolean;
	begin
		yEqual := y = longIntPoint(p).y;
	end;

procedure longIntPoint.textFileInput;
	begin
		readTextLongInt(x);
		readTextLongInt(y);
	end;

procedure longIntPoint.textFileOutput;
	begin
		writeTextLongInt(x);
		writeTextLongInt(y);
	end;

procedure longIntPoint.fileInput;
	begin
		readLongInt(x);
		readLongInt(y);
	end;

procedure longIntPoint.fileOutput;
	begin
		writeLongInt(x);
		writeLongInt(y);
	end;

function longIntPoint.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, longIntPoint) & (longIntPoint(o).x = x) & (longIntPoint(o).y = y);
	end;

function longIntPoint.equalPoint (p: point2d): boolean;
	begin
		equalPoint := (p <> nil) & member(p, longIntPoint) & (longIntPoint(p).x = x) & (longIntPoint(p).y = y);
	end;

function longIntPoint.memberType (o: obj): boolean;
	begin
		memberType := member(o, longIntPoint);
	end;

function longIntPoint.middle (p: point2d): point2d;
	var
		t: longIntPoint;
	begin
		new(t);
		failNil(Handle(t));
		t.init;
		t.x := (x + longIntPoint(p).x) div 2;
		t.y := (y + longIntPoint(p).y) div 2;
		middle := t;
	end;

function longIntPoint.whichSideX (p, q: point2d): extended;
	var
		pr, qr: longIntPoint;
	begin
		pr := longIntPoint(p);
		qr := longIntPoint(q);
		whichSideX := (pr.x - x) * (pr.y - qr.y) - (pr.x - qr.x) * (pr.y - y);
	end;

function longIntPoint.crossProductX (p, q, r: point2d): extended;
	var
		pr, qr, rr: longIntPoint;
	begin
		pr := longIntPoint(p);
		qr := longIntPoint(q);
		rr := longIntPoint(r);
		crossProductX := (pr.x - x) * (rr.y - qr.y) - (pr.y - y) * (rr.x - qr.x);
	end;

function longIntPoint.scalarProductX (p: point2d): extended; (* added by FB *)
	var
		pr: longIntPoint;
	begin
		pr := longIntPoint(p);
		scalarProductX := (pr.x * x) + (pr.y * y);
	end;

function longIntPoint.distanceX (p: point2d): extended;
	var
		pr: longIntPoint;
	begin
		pr := longIntPoint(p);
		distanceX := round(sqrt(sqr(pr.x - x) + sqr(pr.y - y)));
	end;

function longIntPoint.squaredDxX (p: point2d): extended;
	begin
		squaredDxX := sqr(longIntPoint(p).x - x);
	end;

function longIntPoint.squaredDyX (p: point2d): extended;
	begin
		squaredDyX := sqr(longIntPoint(p).y - y);
	end;

function longIntPoint.squaredDistanceX (p: point2d): extended;
	var
		pr: longIntPoint;
	begin
		pr := longIntPoint(p);
		squaredDistanceX := sqr(pr.x - x) + sqr(pr.y - y);
	end;

function longIntIntegerPower (x: longInt; p: integer): longInt;
	var
		t: longInt;
	begin
		case p of
			0: 
				longIntIntegerPower := 1;
			1: 
				longIntIntegerPower := x;
			2: 
				longIntIntegerPower := sqr(x);
			3: 
				longIntIntegerPower := x * sqr(x);
			4: 
				longIntIntegerPower := sqr(sqr(x));
			otherwise begin
				t := sqr(longIntIntegerPower(x, p div 2));
				if p mod 2 = 0 then
					longIntIntegerPower := t
				else
					longIntIntegerPower := x * t;
			end;
		end;
	end;

function longIntPthRoot (x: longInt; p: extended): longInt;
	begin
		if x > 0 then
{$IFC THINK_PASCAL }
			longIntPthRoot := round(exp(ln(x) / p))
{$ELSEC }
			longIntPthRoot := roundLong(exp(ln(x) / p))
{$ENDC }
		else if x = 0 then
			longIntPthRoot := 0
		else begin
			assert(false, 'x>=0');
			longIntPthRoot := 0;
		end;
	end;

function longIntPower (x: longInt; p: extended): longInt;
	begin
		if x > 0 then
{$IFC THINK_PASCAL }
			longIntPower := round(exp(p * ln(x)))
{$ELSEC }
			longIntPower := roundLong(exp(p * ln(x)))
{$ENDC }
		else if x = 0 then
			longIntPower := 0
		else begin
			assert(false, 'x>=0');
			longIntPower := 0;
		end;
	end;

function longIntPoint.distanceLpX (p: point2d): extended;
	begin
		if lpMetric = 2 then
			distanceLpX := distanceX(p)
		else if lpMetric = 1 then
			distanceLpX := abs(longIntPoint(p).x - x) + abs(longIntPoint(p).y - y)
		else if lpMetric = lpInfinity then
			distanceLpX := distancePLpX(p)
		else
{$IFC THINK_PASCAL }
			distanceLpX := longIntPthRoot(trunc(distancePLpX(p)), lpMetric);
{$ELSEC }
			distanceLpX := longIntPthRoot(truncLong(distancePLpX(p)), lpMetric);
{$ENDC }
	end;

function longIntMax (a, b: longInt): longInt;
	begin
		if a > b then
			longIntMax := a
		else
			longIntMax := b;
	end;

function longIntPoint.distancePLpX (p: point2d): extended;
	var
		r, r1, r2: longInt;
	begin
{$IFC THINK_PASCAL }
		r1 := trunc(dxPLpX(p));
		r2 := trunc(dyPLpX(p));
{$ELSEC }
		r1 := truncLong(dxPLpX(p));
		r2 := truncLong(dyPLpX(p));
{$ENDC }
		if lpMetric = lpInfinity then
			r := longIntMax(r1, r2)
		else
			r := r1 + r2;
		distancePLpX := r;
	end;

function longIntPoint.dxPLpX (p: point2d): extended;
	begin
		if trunc(lpMetric) = lpMetric then
			if lpMetric = lpInfinity then
				dxPLpX := abs(x - longIntPoint(p).x)
			else
{$IFC THINK_PASCAL }
				dxPLpX := abs(longIntIntegerPower(x - longIntPoint(p).x, trunc(lpMetric)))
{$ELSEC }
				dxPLpX := abs(longIntIntegerPower(x - longIntPoint(p).x, truncLong(lpMetric)))
{$ENDC }
		else
			dxPLpX := longIntPower(abs(x - longIntPoint(p).x), lpMetric)
	end;

function longIntPoint.dyPLpX (p: point2d): extended;
	begin
		if trunc(lpMetric) = lpMetric then
			if lpMetric = lpInfinity then
				dyPLpX := abs(y - longIntPoint(p).y)
			else
{$IFC THINK_PASCAL }
				dyPLpX := abs(longIntIntegerPower(y - longIntPoint(p).y, trunc(lpMetric)))
{$ELSEC }
				dyPLpX := abs(longIntIntegerPower(y - longIntPoint(p).y, truncLong(lpMetric)))
{$ENDC }
		else
			dyPLpX := longIntPower(abs(y - longIntPoint(p).y), lpMetric)
	end;

procedure longIntPoint.circleCenterX (p, q: point2d; var cx, cy: extended);
	var
		d, s1, s2, s3: longInt;
		pr, qr: longIntPoint;
	begin
		pr := longIntPoint(p);
		qr := longIntPoint(q);
		d := 2 * ((pr.x - x) * (pr.y - qr.y) - (pr.x - qr.x) * (pr.y - y));
		s1 := sqr(pr.x) + sqr(pr.y);
		s2 := sqr(x) + sqr(y);
		s3 := sqr(qr.x) + sqr(qr.y);
		cx := (s1 * (y - qr.y) + s2 * (qr.y - pr.y) + s3 * (pr.y - y)) div d;
		cy := (s1 * (qr.x - x) + s2 * (pr.x - qr.x) + s3 * (x - pr.x)) div d;
	end;

function longIntPoint.intersectLineSegmentX (q, r, s, left, right: point2d): boolean;

	function intersectLineSegment (p, q, r, s: longIntPoint; left, right: longIntPoint): boolean;
		var
			prq, psq, rps, rqs, d: longInt;
		begin
			intersectLineSegment := true;
			orderPointsGreaterThanXY(point2d(p), point2d(q));
			orderPointsGreaterThanXY(point2d(r), point2d(s));
			prq := ws(p, r, q);
			psq := ws(p, s, q);
			rps := ws(r, p, s);
			rqs := ws(r, q, s);
			if prq = psq then
				if prq = 0 then begin
					orderPointsGreaterThanXY(point2d(p), point2d(r));
					orderPointsGreaterThanXY(point2d(q), point2d(s));
					if pointGreaterThanXY(point2d(r), point2d(q)) then
						intersectLineSegment := false
					else begin
						left.x := r.x;
						left.y := r.y;
						right.x := q.x;
						right.y := q.y;
					end;
				end
				else
					intersectLineSegment := false
			else if ((prq >= 0) & (psq <= 0) | (prq <= 0) & (psq >= 0)) & ((rps >= 0) & (rqs <= 0) | (rps <= 0) & (rqs >= 0)) then begin
				d := prq - psq;
				left.x := (prq * s.x - psq * r.x) div d;
				left.y := (prq * s.y - psq * r.y) div d;
				right.x := left.x;
				right.y := left.y;
			end
			else
				intersectLineSegment := false;
		end;

	begin
		intersectLineSegmentX := intersectLineSegment(self, longIntPoint(q), longIntPoint(r), longIntPoint(s), longIntPoint(left), longIntPoint(right));
	end;

function intersectLine (p, q, r, s: longIntPoint; left, right: longIntPoint): boolean;
	var
		prq, psq, d: longInt;
	begin
		intersectLine := true;
		orderPointsGreaterThanXY(point2d(r), point2d(s));
		prq := ws(p, r, q);
		psq := ws(p, s, q);
		if prq = psq then
			if psq = 0 then begin
				left.x := r.x;
				left.y := r.y;
				right.x := q.x;
				right.y := q.y;
			end
			else
				intersectLine := false
		else if (prq >= 0) & (psq <= 0) | (prq <= 0) & (psq >= 0) then begin
			d := prq - psq;
			left.x := (prq * s.x - psq * r.x) div d;
			left.y := (prq * s.y - psq * r.y) div d;
			right.x := left.x;
			right.y := left.y;
		end
		else
			intersectLine := false;
	end;

function longIntPoint.intersectLineX (q, r, s, left, right: point2d): boolean;
	begin
		intersectLineX := intersectLine(self, longIntPoint(q), longIntPoint(r), longIntPoint(s), longIntPoint(left), longIntPoint(right));
	end;

function intersectLineLine (p, q, r, s, intersection: longIntPoint): boolean;
	var
		prq, psq, d: longInt;
	begin
		prq := ws(p, r, q);
		psq := ws(p, s, q);
		if prq = psq then
			if psq = 0 then begin
				intersection.x := p.x;
				intersection.y := p.y;
				intersectLineLine := true;
			end
			else
				intersectLineLine := false
		else begin
			d := prq - psq;
			intersection.x := (prq * s.x - psq * r.x) div d;
			intersection.y := (prq * s.y - psq * r.y) div d;
			intersectLineLine := true;
		end;
	end;

function longIntPoint.lineIntersectsLine (q, r, s, intersectionPoint: point2d): boolean;
	begin
		lineIntersectsLine := intersectLineLine(self, longIntPoint(q), longIntPoint(r), longIntPoint(s), longIntPoint(intersectionPoint));
	end;

procedure doProjectOnLine (p, q, r: longIntPoint; var rProjectX, rProjectY: longInt; var distance, whichSide: extended; var lambdaSign: signType);
	var
		dxqp, dyqp, dxrp, dyrp, rv, ws, d: longInt;
	begin
		dxqp := q.x - p.x;
		dyqp := q.y - p.y;
		if (dxqp = 0) & (dyqp = 0) then begin
			rProjectX := p.x;
			rProjectY := p.y;
			distance := sqrt(sqr(p.x - r.x) + sqr(p.y - r.y));
			whichSide := 0;
			lambdaSign := 0;
		end
		else begin
			dxrp := r.x - p.x;
			dyrp := r.y - p.y;
			rv := dxqp * dxrp + dyqp * dyrp;
			ws := dxrp * (q.y - r.y) - dyrp * (q.x - r.x);
			d := sqr(dxqp) + sqr(dyqp);
			rProjectX := p.x + dxqp * rv div d;
			rProjectY := p.y + dyqp * rv div d;
			whichSide := ws;
			distance := abs(whichSide / sqrt(d));
			if rv < 0 then
				lambdaSign := -1
			else if rv > d then
				lambdaSign := +1
			else
				lambdaSign := 0;
		end;
	end;

procedure longIntPoint.projectOnLine (endPoint, candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
	var
		px, py: longInt;
	begin
		if projectionPoint = nil then
			doProjectOnLine(self, longIntPoint(endPoint), longIntPoint(candidatePoint), px, py, distanceToLine, whichSide, lambdaSign)
		else
			doProjectOnLine(self, longIntPoint(endPoint), longIntPoint(candidatePoint), longIntPoint(projectionPoint).x, longIntPoint(projectionPoint).y, distanceToLine, whichSide, lambdaSign);
	end;

function longIntPoint.distanceToLineSegmentX (p, q: point2d; var closestX, closestY: extended): extended; (* added by FB *)
	var
		rpx, rpy: longInt;
		d, w: extended;
		l: signType;
	begin
		doProjectOnLine(longIntPoint(p), longIntPoint(q), self, rpx, rpy, d, w, l);
		if l < 0 then begin
			closestX := longIntPoint(p).x;
			closestY := longIntPoint(p).y;
			distanceToLineSegmentX := distanceX(p);
		end
		else if l > 0 then begin
			closestX := longIntPoint(q).x;
			closestY := longIntPoint(q).y;
			distanceToLineSegmentX := distanceX(q);
		end
		else begin
			closestX := rpx;
			closestY := rpy;
			distanceToLineSegmentX := d;
		end;
	end;

function longIntPoint.xPlusYX: extended;
	begin
		xPlusYX := x + y;
	end;

function longIntPoint.xMinusYX: extended;
	begin
		xMinusYX := x - y;
	end;

function longIntPoint.getType: objectType;
	begin
		getType := longIntPointType;
	end;

function longIntPoint.getTextType: Str255;
	begin
		getTextType := 'Point(longInt)';
	end;

(* --------------------- inserted by JM ---------------------------------- *)

function longIntPoint.DxX (p: point2d): extended;
	begin
		DxX := x - longIntPoint(p).x;
	end;

function longIntPoint.DyX (p: point2d): extended;
	begin
		DyX := y - longIntPoint(p).y;
	end;

procedure longIntPoint.setPoint (newp: point2d);
	begin
		x := longIntPoint(newp).x;
		y := longIntPoint(newp).y;
	end;

(* --------------------------------------------------------------------------- *)

end.
