(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit lowerEnvelope;

interface

uses
	pointVector;

implementation

{$SETC ALowerEnvelope = TRUE }

function lineSegmentVector.lowerEnvelope: obj;
	var
		leftM, rightM, upperM, c1, c2: point2d;
		ls: lineSegment;
		i: seqIndex;

	procedure appendPoint (l: list; o: point2d; var beforeLast: point2d);
		var
			last: point2d;
		begin
			last := point2d(l.last^^.value);
			if not last.equalPoint(o) then begin
				if (beforeLast = nil) | (last.yLessThan(o)) & (last.yLessThan(beforeLast)) | (o.whichSideX(last, beforeLast) <> 0) then begin
{$IFC AnimationEnabled AND ALowerEnvelope }
					if animationFlag[ALowerEnvelopeItem] then
						o.interactiveHighlight;
{$ENDC }
					beforeLast := last;
					l.append(o.duplicate);
				end
				else begin
{$IFC AnimationEnabled AND ALowerEnvelope }
					if animationFlag[ALowerEnvelopeItem] then begin
						last.interactiveHighlight;
						o.interactiveHighlight;
					end;
{$ENDC }
					last.free;
					l.last^^.value := o.duplicate;
				end;
			end;
		end;

	function merge (chain1, chain2: polyLine; level: integer): polyLine;
		var
			resList: list;
			beforeLast: point2d;
			resPl: polyLine;
			i1, i2: seqIndex;
			p11, p12, p21, p22: point2d;
{$IFC AnimationEnabled AND ALowerEnvelope }
			ls1, ls2: lineSegment;
{$ENDC }

		begin
{$IFC AnimationEnabled }
			if abortFlag then begin
				if chain1 <> nil then
					chain1.freeAll;
				if chain2 <> nil then
					chain2.freeAll;
				merge := nil;
				exit(merge);
			end;
{$ENDC }
{$IFC AnimationEnabled AND ALowerEnvelope }
			if animationFlag[ALowerEnvelopeItem] then begin
				chain1.interactiveOutput;
				chain2.interactiveOutput;
				displayInfoString(stringOf('Merging: chain1.length = ', chain1.length : 1, ', chain2.length = ', chain2.length : 1, ', level = ', level : 1));
			end;
{$ENDC }
			resList := newEmptyList;
			i1 := 2;
			i2 := 2;
			beforeLast := nil;
			resList.append(leftM.duplicate);
{$IFC AnimationEnabled AND ALowerEnvelope }
			if animationFlag[ALowerEnvelopeItem] then
				leftM.interactiveHighlight;
{$ENDC }
			p11 := point2d(chain1.elements^^[1]);
			p12 := point2d(chain1.elements^^[2]);
			p21 := point2d(chain2.elements^^[1]);
			p22 := point2d(chain2.elements^^[2]);
			while true do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					resList.freeAll;
					chain1.freeAll;
					chain2.freeAll;
					merge := nil;
					exit(merge);
				end;
{$ENDC }
{$IFC AnimationEnabled AND ALowerEnvelope }
				if animationFlag[ALowerEnvelopeItem] then begin
					new(ls1);
					failNil(Handle(ls1));
					new(ls2);
					failNil(Handle(ls2));
					ls1.p := p11;
					ls1.q := p12;
					ls2.p := p21;
					ls2.q := p22;
					ls1.interactiveHighlight;
					ls2.interactiveHighlight;
					waitForClick(animationFlag[ALowerEnvelopeItem]);
					ls1.interactiveHighlight;
					ls2.interactiveHighlight;
					dispose(ls1);
					dispose(ls2);
				end;
{$ENDC }
				if p11.intersectLineSegmentX(p12, p21, p22, c1, c2) then
					appendPoint(resList, c2, beforeLast);
				if (i2 = chain2.length) | p12.xLessThan(p22) | p12.xEqual(p22) & (p12.whichSideX(p21, p22) > 0) then begin
					if p12.whichSideX(p21, p22) >= 0 then
						appendPoint(resList, p12, beforeLast);
					i1 := i1 + 1;
					if i1 > chain1.length then
						leave;
					p11 := p12;
					p12 := point2d(chain1.elements^^[i1]);
				end
				else begin
					if p22.whichSideX(p11, p12) >= 0 then
						appendPoint(resList, p22, beforeLast);
					i2 := i2 + 1;
					p21 := p22;
					p22 := point2d(chain2.elements^^[i2]);
				end;
			end;
			appendPoint(resList, rightM, beforeLast);
			new(resPl);
			failNil(Handle(resPl));
			resPl.initFromList(resList);
			merge := resPl;
{$IFC AnimationEnabled AND ALowerEnvelope }
			if animationFlag[ALowerEnvelopeItem] then begin
				resList.interactiveHighlight;
				resPl.interactiveHighlight;
				resPl.interactiveFlash;
				waitForClick(animationFlag[ALowerEnvelopeItem]);
				resPl.interactiveHighlight;
				chain1.interactiveOutput;
				chain2.interactiveOutput;
			end;
{$ENDC }
			resList.free;
			chain1.freeAll;
			chain2.freeAll;
		end;

	function recVis (f, t: seqIndex; level: integer): polyLine;
		var
			pl: polyLine;
			p1, p2, pp: point2d;
			m: seqIndex;
		begin
{$IFC AnimationEnabled AND ALowerEnvelope }
			if animationFlag[ALowerEnvelopeItem] then begin
			end;
{$ENDC }
			if f = t then begin
				with lineSegment(elements^^[f]) do
					if p.xLessThan(q) then begin
						p1 := p;
						p2 := q;
					end
					else begin
						p1 := q;
						p2 := p;
					end;
				currentVectorLength := 6;
				new(pl);
				failNil(Handle(pl));
				pl.init;
				pl.append(leftM.duplicate);
				if not p1.equalPoint(leftM) then begin
					if not p1.yEqual(leftM) & not p1.xEqual(leftM) then begin
						pp := point2d(p1.duplicate);
						pp.setY(leftM.getY);
						pl.append(pp);
					end;
					pl.append(p1.duplicate);
				end;
				if not p2.equalPoint(rightM) then begin
					pl.append(p2.duplicate);
					if not p2.yEqual(rightM) & not p2.xEqual(rightM) then begin
						pp := point2d(p2.duplicate);
						pp.setY(rightM.getY);
						pl.append(pp);
					end;
				end;
				pl.append(rightM.duplicate);
				recVis := pl;
			end
			else begin (* f < t *)
				m := (f + t) div 2;
				recVis := merge(recVis(f, m, level + 1), recVis(m + 1, t, level + 1), level);
			end;
		end;

	begin
		leftM := lineSegment(elements^^[1]).p;
		rightM := lineSegment(elements^^[1]).p;
		upperM := lineSegment(elements^^[1]).p;
		c1 := point2d(leftM.duplicate);
		c2 := point2d(leftM.duplicate);
		for i := 1 to length do begin
			ls := lineSegment(elements^^[i]);
			if ls.p.xLessThan(leftM) then
				leftM := ls.p;
			if ls.q.xLessThan(leftM) then
				leftM := ls.q;
			if rightM.xLessThan(ls.p) then
				rightM := ls.p;
			if rightM.xLessThan(ls.q) then
				rightM := ls.q;
			if upperM.yLessThan(ls.p) then
				upperM := ls.p;
			if upperM.yLessThan(ls.q) then
				upperM := ls.q;
		end;
		leftM := point2d(leftM.duplicate);
		rightM := point2d(rightM.duplicate);
		leftM.setY(upperM.getY);
		rightM.setY(upperM.getY);
{$IFC AnimationEnabled AND ALowerEnvelope }
		if animationFlag[ALowerEnvelopeItem] then begin
			interactiveOutput;
			interactiveHighlight;
		end;
{$ENDC }
		lowerEnvelope := recVis(1, length, 1);
		leftM.free;
		rightM.free;
		c1.free;
		c2.free;
	end;

end.
