(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit markedRealPoint;

interface

uses
	basicGeometry;

procedure initMarkedRealPoint;

implementation

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function markedRealPoint.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 2;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function obj.perimeterMP: obj;
	var
		mp: markedRealPoint;
	begin
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.initLocationMark(self, stringOf(perimeter : 8 : 5), true);
		perimeterMP := mp;
	end;

function obj.areaMP: obj;
	var
		mp: markedRealPoint;
	begin
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.initLocationMark(self, stringOf(area : 8 : 5), true);
		areaMP := mp;
	end;

procedure markedRealPoint.init;
	var
		s: strings;
	begin
		new(s);
		failNil(Handle(s));
		s.init;
		mark := s;
	end;

procedure markedRealPoint.free;
	begin
		mark.free;
		shallowFree;
	end;

function markedRealPoint.duplicate: obj;
	var
		p: markedRealPoint;
	begin
		p := markedRealPoint(shallowDuplicate);
		p.mark := strings(mark.duplicate);
		duplicate := p;
	end;

procedure markedRealPoint.interactiveOutput;
	var
		xi, yi: integer;
	begin
		if worldToWindow(x, y, xi, yi) then begin
			MoveTo(xi - 3, yi - 3);
			LineTo(xi + 3, yi + 3);
			MoveTo(xi - 3, yi + 3);
			LineTo(xi + 3, yi - 3);
			labelPoint(mark.value);
		end;
	end;

procedure markedRealPoint.textFileInput;
	var
		s: strings;
	begin
		inherited textFileInput;
		new(s);
		failNil(Handle(s));
		mark := s;
		mark.textFileInput;
	end;

procedure markedRealPoint.textFileOutput;
	begin
		case currentTextFileOutputMethod of
			0:  begin
				inherited textFileOutput;
				mark.textFileOutput;
			end;
			otherwise
				inherited textFileOutput;
		end;
	end;

procedure markedRealPoint.fileInput;
	var
		s: strings;
	begin
		inherited fileInput;
		new(s);
		failNil(Handle(s));
		mark := s;
		mark.fileInput;
	end;

procedure markedRealPoint.fileOutput;
	begin
		inherited fileOutput;
		mark.fileOutput;
	end;

function markedRealPoint.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, markedRealPoint) & inherited equal(o) & mark.equal(markedRealPoint(o).mark);
	end;

function markedRealPoint.equalPoint (p: point2d): boolean;
	begin
		equalPoint := (p <> nil) & member(p, markedRealPoint) & inherited equal(p);
	end;

function markedRealPoint.memberType (o: obj): boolean;
	begin
		memberType := member(o, markedRealPoint);
	end;

function markedRealPoint.getType: objectType;
	begin
		getType := markedRealPointType;
	end;

function markedRealPoint.getTextType: Str255;
	begin
		getTextType := 'Marked  Point(real)';
	end;

function markedRealPoint.genericLessThan (o: obj): boolean;
	var
		m: markedRealPoint;
	begin
		m := markedRealPoint(o);
		genericLessThan := (x < m.x) | (x = m.x) & ((y < m.y) | (y = m.y) & (mark.value < m.mark.value));
	end;

function markedRealPoint.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoObject(v, 'Mark', mark, false);
		getInfo := v;
	end;

procedure markedRealPoint.initLocationMark (o: obj; s: Str255; translation: boolean);
	var
		cx, cy: extended;
	begin
		o.centerOfGravity(cx, cy);
		if translation then
			setXY(cx + 25 / currentFrame.mx, cy + 25 / currentFrame.my)
		else
			setXY(cx, cy);
		mark.value := s
	end;

function markedRealPoint.toRealPoint: realPoint;
	var
		p: realPoint;
	begin
		new(p);
		failNil(Handle(p));
		p.x := x;
		p.y := y;
		toRealPoint := p;
	end;

function sequence.makeIdRealPointVector: homogenVector;
	var
		result: homogenVector;

	procedure transform (o: obj);
		var
			irp: idRealPoint;
			cx, cy: extended;
		begin
			new(irp);
			failNil(Handle(irp));
			o.centerOfGravity(cx, cy);
			irp.setXY(cx, cy);
			result.append(irp);
			irp.id := result.length;
		end;

	begin
		result := newHomogenVector(length);
		forAll(transform);
		makeIdRealPointVector := result;
	end;

function idRealPoint.getTextType: Str255;
	begin
		getTextType := 'Id Point(real)';
	end;

function idRealPoint.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoExtended(v, 'Index', id, false);
		getInfo := v;
	end;

procedure idRealPoint.setInfo (o: obj);
	var
		newId: extended;
	begin
		inherited setInfo(o);
		newId := objInfoExtended(vector(o).elements^^[3]).e;
		if (1 <= newId) & (newId < maxN) & (trunc(newId) = newId) then
{$IFC THINK_PASCAL }
			id := trunc(newId);
{$ELSEC }
			id := truncLong(newId);
{$ENDC }
	end;

procedure idRealPoint.textFileInput;
	var
		id4: longInt;
	begin
		inherited textFileInput;
		readTextLongInt(id4);
		id := id4;
	end;

procedure idRealPoint.textFileOutput;
	begin
		case currentTextFileOutputMethod of
			0:  begin
				inherited textFileOutput;
				writeTextLongInt(id);
			end;
			otherwise
				inherited textFileOutput;
		end;
	end;

procedure idRealPoint.fileInput;
	var
		id4: longInt;
	begin
		inherited fileInput;
		readLongInt(id4);
		id := id4;
	end;

procedure idRealPoint.fileOutput;
	begin
		inherited fileOutput;
		writeLongInt(id);
	end;

function idRealPoint.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, idRealPoint) & inherited equal(o) & (id = idRealPoint(o).id);
	end;

function idRealPoint.equalPoint (p: point2d): boolean;
	begin
		equalPoint := (p <> nil) & member(p, idRealPoint) & inherited equal(p);
	end;

function idRealPoint.memberType (o: obj): boolean;
	begin
		memberType := member(o, idRealPoint);
	end;

function idRealPoint.getType: objectType;
	begin
		getType := idRealPointType;
	end;

function idRealPoint.genericLessThan (o: obj): boolean;
	var
		i: idRealPoint;
	begin
		i := idRealPoint(o);
		genericLessThan := (x < i.x) | (x = i.x) & ((y < i.y) | (y = i.y) & (id < i.id));
	end;

{$IFC GFEnabled }
function idRealPoint.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := false;
	end;

{$ENDC}
	var
		path: Str255;

function sequence.showNumbering: vector;
	var
		result: vector;
		index: seqIndex0;
		oldPath: Str255;

	procedure createMarker (x: obj);
		var
			mp: markedRealPoint;
			cx, cy: extended;
		begin
			index := index + 1;
			if member(x, sequence) then begin
				oldPath := path;
				path := concat(path, StringOf(index : 1), '.');
				result.append(sequence(x).showNumbering);
				path := oldPath;
			end
			else begin
				x.centerOfGravity(cx, cy);
				new(mp);
				failNil(Handle(mp));
				mp.init;
				mp.x := cx;
				mp.y := cy;
				mp.mark.value := concat(path, StringOf(index : 1));
				result.append(mp);
			end;
		end;

	begin
		result := newVector(length);
		index := 0;
		forAll(createMarker);
		showNumbering := result;
	end;

procedure initMarkedRealPoint;
	begin
		path := '';
	end;


end.
