(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit miniDisk;

{$SETC AminimalDisk = TRUE }

interface

uses
	pointVector;

implementation

var
	miniDiskCounter2, miniDiskCounter3, inCircleCounter: longInt;

function md (l: doubleList; n: seqIndex0; r: list): circle;
(* function 'md'  produces the smallest circle which encloses the first 'n' *)
(* points in the double linked list 'l' where the points on 'r' lie on the boundary of the circle. *)
(* The list 'l' has links to its predecessor and successor in order to allow an *)
(* efficient 'moveToFront' operation *)
	var
		disk: circle; (* the result *)
		e: doubleElementH; (* an element of the double linked list *)
		i: seqIndex0;
	begin
		if r.length = 2 then
			miniDiskCounter2 := miniDiskCounter2 + 1
		else if r.length = 3 then
			miniDiskCounter3 := miniDiskCounter3 + 1;
		new(disk);
		failNil(Handle(disk));
		disk.makeDisk(r); (* creates the smallest disk containing the points from 'r' on the boundary *)
		if r.length <= 2 then begin
			e := l.first; (* set to first element in the list 'l' *)
			for i := 1 to n do (* process the first 'n' elements of 'l' *)
				begin
{$IFC AnimationEnabled }
				if userAbort then begin
					disk.free;
					md := nil;
					exit(md);
				end;
{$ENDC }
{$IFC AnimationEnabled AND AMinimalDisk }
				if animationFlag[AMinimalAreaDiskItem] & (r.length = 0) then
					with e^^.value do begin
						interactiveFlash;
						interactiveOutput;
						interactiveHighlight;
					end;
{$ENDC }
				inCircleCounter := inCircleCounter + 1;
				if not disk.inCircle(point2d(e^^.value)) then		(* the i-th point does not lie inside the current circle *)
					begin
{$IFC AnimationEnabled AND AMinimalDisk }
					if animationFlag[AMinimalAreaDiskItem] & (disk.radius > 0) & (r.length = 0) then
						disk.interactiveHighlight;
{$ENDC }
					r.push(e^^.value);										(* push the offending point on 'r' since it must lie on the boundary *)
					disk.free;													(* throw away the old disk *)
					disk := md(l, i - 1, r);									(* compute the new disk *)
{$IFC AnimationEnabled }
					if disk = nil then begin
						md := nil;
						exit(md);
					end;
{$ENDC }
{$IFC AnimationEnabled AND AMinimalDisk }
					if animationFlag[AMinimalAreaDiskItem] & (disk.radius > 0) then
						if r.length = 1 then begin
							disk.interactiveHighlight;
							waitForClick(animationFlag[AMinimalAreaDiskItem]);
						end
						else
							disk.interactiveOutput;
{$ENDC }
					r.pop;														(* restore the list 'r' *)
					l.moveToFront(e);										(* move the offending point to the front of the list 'l'. *)
				end;														(* This is a heuristic for greater speed and can be omitted *)
				e := e^^.succ;												(* process the next point *)
			end;
		end;
		md := disk; (* assign result *)
	end;

function pointVector.minimalDisk (verbose: boolean): circle;
	var
		e: list;
		l: doubleList;

	procedure moveToL (x: obj);
		begin
			l.append(x);
		end;

	begin
{$IFC AnimationEnabled AND AMinimalDisk }
		if animationFlag[AMinimalAreaDiskItem] then
			interactiveOutput;
{$ENDC }
		randomShuffleAll;
		e := newEmptyList;
		miniDiskCounter2 := 0;
		miniDiskCounter3 := 0;
		inCircleCounter := 0;
		new(l);
		failNil(Handle(l));
		l.init;
		forAll(moveToL);
		minimalDisk := md(l, length, e);
		l.free;
		e.free;
		if verbose then
			displayInfoString(stringOf('#miniDisk2=', miniDiskCounter2 : 1, ', #miniDisk3 = ', miniDiskCounter3 : 1, ', #inCircle = ', inCircleCounter : 1));
	end;

function pointVector.simpleMinimalDisk: circle;
	var
		c: circle;
		i: seqIndex0;
	begin
{$IFC AnimationEnabled AND AMinimalDisk }
		if animationFlag[AMinimalAreaDiskItem] then
			interactiveOutput;
{$ENDC }
		randomShuffleAll;
		new(c);
		failNil(Handle(c));
		c.init;
		c.makeFrom2Points(point2d(elements^^[1]), point2d(elements^^[2]));
{$IFC AnimationEnabled AND AMinimalDisk }
		if animationFlag[AMinimalAreaDiskItem] then begin
			c.interactiveHighlight;
			waitForClick(animationFlag[AMinimalAreaDiskItem]);
		end;
{$ENDC }
		for i := 3 to length do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				c.free;
				simpleMinimalDisk := nil;
				exit(simpleMinimalDisk);
			end;
{$ENDC }
{$IFC AnimationEnabled AND AMinimalDisk }
			if animationFlag[AMinimalAreaDiskItem] then
				elements^^[i].interactiveFlash;
{$ENDC }
			if not c.inCircle(point2d(elements^^[i])) then begin
{$IFC AnimationEnabled AND AMinimalDisk }
				if animationFlag[AMinimalAreaDiskItem] then
					c.interactiveHighlight;
{$ENDC }
				c.makeFromPointCircle(point2d(elements^^[i]), c);
{$IFC AnimationEnabled AND AMinimalDisk }
				if animationFlag[AMinimalAreaDiskItem] then begin
					c.interactiveHighlight;
					waitForClick(animationFlag[AMinimalAreaDiskItem]);
				end;
{$ENDC }
			end;
		end;
		simpleMinimalDisk := c;
	end;

function pointVector.containedInCircle (c: circle): boolean;
	var
		i: seqIndex1;
	begin
		i := 1;
		while (i <= length) & c.inCircle(point2d(elements^^[i])) do
			i := i + 1;
		containedInCircle := i > length;
	end;


end.
