(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit nearestNeighborSweep;

{$SETC AANN = TRUE }
interface

uses
	pointVector, proximity;

implementation

{$SETC IgnoreANNIntersectionEvent = FALSE }

function pointVector.nearestNeighborSweep (kind: neighborKind; ask: boolean): lineSegmentVector;
	var
		nearestNeighbors: lineSegmentVector;
		xQueue: priorityQueue;
		yTable: dictionary;
		currentIndex: seqIndex0;
		currentHandle: reference;
		r: refPoint;
		found: boolean;
		l: lineSegment;
		direction: directionType;
		sweepDirection: integer; (* actually, -1 or +1 *)
{$IFC AnimationEnabled AND AANN }
		i: seqIndex;
{$ENDC }

	procedure createIntersection (cx, cy: extended; toBeKilled: reference);
		var
			r: refPoint;
			where: point2d;
			toKill: refPoint;

		procedure makeR (pType: objectType);
			begin
				where := point2d(newInitObj(pType));
				where.setX(cx);
				where.setY(cy);
{$IFC AnimationEnabled AND AANN }
				if animationFlag[ANearestNeighborsItem] then
					where.interactiveOutput;
{$ENDC }
				new(r);
				failNil(Handle(r));
				r.p := where;
				r.ref := toBeKilled;
			end;

		begin
{$IFC NOT IgnoreANNIntersectionEvent }
			if (sweepDirection = 1) & (cx <= point2d(elements^^[length]).getX) | (sweepDirection = -1) & (cx >= point2d(elements^^[1]).getX) then begin
				toKill := refPoint(yTable.getObject(toBeKilled));
				if toKill.ref = nil then begin
					makeR(elements^^[length].getType);
					if sweepDirection = 1 then
						toKill.ref := xQueue.insert(r, refPointLessThanXY)
					else
						toKill.ref := xQueue.insert(r, refPointGreaterThanXY);
				end
				else if (sweepDirection = 1) & (cx < refPoint(xQueue.getObject(toKill.ref)).p.getX) | (sweepDirection = -1) & (cx > refPoint(xQueue.getObject(toKill.ref)).p.getX) then begin
					makeR(elements^^[length].getType);
{$IFC AnimationEnabled AND AANN }
					if animationFlag[ANearestNeighborsItem] then
						refPoint(xQueue.getObject(toKill.ref)).p.interactiveOutput;
{$ENDC }
					disposeRefPoint(refPoint(xQueue.getObject(toKill.ref)));
					if sweepDirection = 1 then begin
						xQueue.delete(toKill.ref, refPointLessThanXY);
						toKill.ref := xQueue.insert(r, refPointLessThanXY);
					end
					else begin
						xQueue.delete(toKill.ref, refPointGreaterThanXY);
						toKill.ref := xQueue.insert(r, refPointGreaterThanXY);
					end;
				end;
			end;
{$ENDC }
		end;

	procedure intersectionEvent (rp: refPoint);
		var
			cx, cy: extended;
			l, ll, r, rr: reference;
			lpt, llpt, rpt, rrpt: point2d;
{$IFC AnimationEnabled AND AANN }
			pTemp: point2d;
			radius: extended;
{$ENDC }
		begin
{$IFC AnimationEnabled AND AANN }
			if animationFlag[ANearestNeighborsItem] then begin
				drawVertical(rp.p.getX);
				pTemp := refPoint(yTable.getObject(rp.ref)).p;
				with pTemp do
					radius := sqrt(sqr(rp.p.getX - getX) + sqr(rp.p.getY - getY));
				drawCircle(rp.p.getX, rp.p.getY, radius, 0, 360);
				waitForClick(animationFlag[ANearestNeighborsItem]);
				pTemp.interactiveHighlight;
				drawCircle(rp.p.getX, rp.p.getY, radius, 0, 360);
				rp.p.interactiveOutput;
				drawVertical(rp.p.getX);
			end;
{$ENDC }
			l := yTable.next(rp.ref, left);
			lpt := refPoint(yTable.getObject(l)).p;
			ll := yTable.next(l, left);
			r := yTable.next(rp.ref, right);
			rpt := refPoint(yTable.getObject(r)).p;
			rr := yTable.next(r, right);
			if ll <> nil then begin
				llpt := refPoint(yTable.getObject(ll)).p;
				if rpt.whichSideX(lpt, llpt) <> 0 then begin
					rpt.circleCenterX(lpt, llpt, cx, cy);
					if (sweepDirection = 1) & (cx > rp.p.getX) | (sweepDirection = -1) & (cx < rp.p.getX) then
						createIntersection(cx, cy, l);
				end;
			end;
			if rr <> nil then begin
				rrpt := refPoint(yTable.getObject(rr)).p;
				if lpt.whichSideX(rpt, rrpt) <> 0 then begin
					lpt.circleCenterX(rpt, rrpt, cx, cy);
					if (sweepDirection = 1) & (cx > rp.p.getX) | (sweepDirection = -1) & (cx < rp.p.getX) then
						createIntersection(cx, cy, r);
				end;
			end;
			yTable.getObject(rp.ref).free;
			yTable.delete(rp.ref);
		end;

	function traverse (start: reference; direction: directionType; rightSign: integer; nearestNeighbor: point2d): point2d;
		var
			pointStart: point2d;
			next, nextNext, t: reference;
			cx, cy: extended;
{$IFC AnimationEnabled AND AANN }
			radius: extended;

		procedure drawSegment (p, q: point2d);
			var
				l: directedLineSegment;
			begin
				if kind = allNeighbors then begin
					new(l);
					failNil(Handle(l));
					l.p := p;
					l.q := q;
					l.interactiveOutput;
					l.shallowFree;
				end
				else
					p.drawLine(q);
			end;
{$ENDC }

		procedure min (newCandidate: reference; currentPoint: point2d; var nearestNeighbor: point2d);
			begin
				if newCandidate <> nil then
					if nearestNeighbor = nil then begin
						nearestNeighbor := refPoint(yTable.getObject(newCandidate)).p;
{$IFC AnimationEnabled AND AANN }
						if animationFlag[ANearestNeighborsItem] then begin
							drawSegment(pointStart, nearestNeighbor);
							waitForClick(animationFlag[ANearestNeighborsItem]);
						end;
{$ENDC }
					end
					else if currentPoint.squaredDistanceX(refPoint(yTable.getObject(newCandidate)).p) < currentPoint.squaredDistanceX(nearestNeighbor) then begin
{$IFC AnimationEnabled AND AANN }
						if animationFlag[ANearestNeighborsItem] then
							drawSegment(pointStart, nearestNeighbor);
{$ENDC }
						nearestNeighbor := refPoint(yTable.getObject(newCandidate)).p;
{$IFC AnimationEnabled AND AANN }
						if animationFlag[ANearestNeighborsItem] then begin
							drawSegment(pointStart, nearestNeighbor);
							waitForClick(animationFlag[ANearestNeighborsItem]);
						end;
{$ENDC }
					end;
			end;

		begin
			pointStart := refPoint(yTable.getObject(start)).p;
			nextNext := yTable.next(start, direction);
			repeat
				next := nextNext;
				min(next, pointStart, nearestNeighbor);
				nextNext := yTable.next(next, direction);
				if (nextNext = nil) | (sign(pointStart.whichSideX(refPoint(yTable.getObject(next)).p, refPoint(yTable.getObject(nextNext)).p)) <> rightSign) then
					leave
				else begin
					pointStart.circleCenterX(refPoint(yTable.getObject(next)).p, refPoint(yTable.getObject(nextNext)).p, cx, cy);
					if (sweepDirection = 1) & (cx > pointStart.getX) | (sweepDirection = -1) & (cx < pointStart.getX) then begin
						createIntersection(cx, cy, next);
						leave;
					end;
				end;
{$IFC AnimationEnabled AND AANN }
				if animationFlag[ANearestNeighborsItem] then begin
					radius := sqrt(sqr(cx - pointStart.getX) + sqr(cy - pointStart.getY));
					drawPoint(cx, cy, 3);
					drawCircle(cx, cy, radius, 0, 360);
					refPoint(yTable.getObject(next)).p.interactiveHighlight;
					waitForClick(animationFlag[ANearestNeighborsItem]);
					drawPoint(cx, cy, 3);
					drawCircle(cx, cy, radius, 0, 360);
				end;
{$ENDC }
				t := refPoint(yTable.getObject(next)).ref;
				if t <> nil then begin
{$IFC AnimationEnabled AND AANN }
					if animationFlag[ANearestNeighborsItem] then
						refPoint(xQueue.getObject(t)).p.interactiveOutput;
{$ENDC }
					disposeRefPoint(refPoint(xQueue.getObject(t)));
					if sweepDirection = 1 then
						xQueue.delete(t, refPointLessThanXY)
					else
						xQueue.delete(t, refPointGreaterThanXY)
				end;
				yTable.getObject(next).free;
				yTable.delete(next);
			until false;
			traverse := nearestNeighbor;
		end;

	begin
		if length > 1 then begin

			if ask then begin
				if not getXY(length, false, xQueue, length, yTable, false) then begin
					nearestNeighborSweep := nil;
					exit(nearestNeighborSweep);
				end;
			end
			else begin
				currentVectorLength := length;
				xQueue := priorityQueue(newInitObj(heapPriorityQueueType));
				yTable := dictionary(newInitObj(avlTreeType));
			end;

			sortAll(pointLessThanXY);
			nearestNeighbors := initNearestNeighbors(kind);
			if kind = rightNeighbors then
				sweepDirection := -1
			else
				sweepDirection := +1;

			repeat
				if sweepDirection = 1 then
					currentIndex := 1 { sweep to the right }
				else
					currentIndex := length; { sweep to the left beginning at the last element: elements^^[length]. }

				repeat
{$IFC AnimationEnabled }
					if userAbort then begin
						nearestNeighbors.nilQ;
						nearestNeighbors.freeAll;
						xQueue.forAll(disposeRefPoint);
						xQueue.free;
						yTable.forAll(disposeOnlyRefPoint, left);
						yTable.free;
						nearestNeighborSweep := nil;
						exit(nearestNeighborSweep);
					end;
{$ENDC }
					if (xQueue.minimum = nil) | (sweepDirection = 1) & (pointLessThanXY(point2d(elements^^[currentIndex]), refPoint(xQueue.getObject(xQueue.minimum)).p)) | (sweepDirection = -1) & (pointGreaterThanXY(point2d(elements^^[currentIndex]), refPoint(xQueue.getObject(xQueue.minimum)).p)) then begin
						new(r);
						failNil(Handle(r));
						r.p := point2d(elements^^[currentIndex]);
						r.ref := nil;
						found := yTable.find(r, refPointLessThanYX, currentHandle, direction);
						l := lineSegment(nearestNeighbors.elements^^[currentIndex]);
{$IFC AnimationEnabled AND AANN }
						if animationFlag[ANearestNeighborsItem] then begin
							drawThickVertical(r.p.getX);
							waitForClick(animationFlag[ANearestNeighborsItem]);
						end;
{$ENDC }
						if found then begin
							l.q := r.p
						end
						else begin
							currentHandle := yTable.insert(r, currentHandle, direction);
							l.q := traverse(currentHandle, left, -sweepDirection, traverse(currentHandle, right, +sweepDirection, l.q));
						end;
{$IFC AnimationEnabled AND AANN }
						if animationFlag[ANearestNeighborsItem] then
							drawThickVertical(r.p.getX);
{$ENDC }
						currentIndex := currentIndex + sweepDirection;
					end
					else begin
						r := refPoint(xQueue.getObject(xQueue.minimum));
						if sweepDirection = 1 then
							xQueue.delete(xQueue.minimum, refPointLessThanXY)
						else
							xQueue.delete(xQueue.minimum, refPointGreaterThanXY);
						intersectionEvent(r);
						disposeRefPoint(r);
					end;
				until (currentIndex > length) | (currentIndex < 1);

				if (sweepDirection = 1) & (kind = allNeighbors) then begin
{$IFC AnimationEnabled AND AANN }
					if animationFlag[ANearestNeighborsItem] then begin
			(* Redraw the points of self. *)
						for i := 1 to length do begin
							PenMode(patOr);
							point2d(elements^^[i]).interactiveHighlight;
							PenMode(patXor);
							point2d(elements^^[i]).interactiveOutput
						end
					end;
{$ENDC }
		(* The old structures can be kept and need only to be re-initialized. *)
					xQueue.forAll(disposeRefPoint);
					xQueue.reset; (* does not reset the counters in the statistics version *)
					yTable.forAll(disposeOnlyRefPoint, left);
					yTable.reset; (* does not reset the counters in the statistics version *)
				end;

				sweepDirection := -sweepDirection;
			until (kind = leftNeighbors) & (sweepDirection = -1) | (sweepDirection = +1);
			displayInfoString2('X', xQueue.infoString, 'Y', yTable.infoString);
			xQueue.forAll(disposeRefPoint);
			xQueue.free;
			yTable.forAll(disposeOnlyRefPoint, left);
			yTable.free;
			nearestNeighbors.finishNearestNeighbors(kind);
			nearestNeighborSweep := nearestNeighbors;
		end
		else
			nearestNeighborSweep := nil;
	end;

end.
