(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit obj;

interface

uses
	GeoBenchUtility, floatArithmetic, infoWindow;

const
	maxDimension = 21;

type
	dimensionRange = 1..maxDimension;
 {$IFC GFEnabled }
	tBooleanArray = packed array[dimensionRange] of BOOLEAN;
 {$ENDC}

	objectType = (intType, stringsType, point2dType, lineSegmentType, rectangleType, listType, vectorType,	(* *)
		homogenVectorType, pointVectorType, convexPolygonType, voronoiEdgeType, directedGraphType,			(* *)
		undirectedGraphType, graphEdgeType, voronoiDiagramType, polygon2dType, markedRealPointType,			(* *)
		simpleUndirectedGraphType, spanningTreeType, rectangleVectorType, dDimPointType, dDimCircleType,	(* *)
		dDimPointVectorType, realPointType, floatPointType, longIntPointType, sequenceType, avlTreeType,	(* *)
		sortedListType, sortedVectorType, lineSegmentVectorType, dictPriorityQueueType,						(* *)
		heapPriorityQueueType, circleType, splinegonType, spline2Type, straightEdgeType, polyLineType,		(* *)
		layerType, polyLayerType, layerVectorType, polyLayerVectorType, QDPictureType, ColorQDPictureType,	(* *)
		pPointType, twoDTreeType, twoDTreeNodeType, dasDBSType, massStorageType, dasGFType, dbsLayerType,	(* *)
		regularNGonType, doubleListType, attractorIdType, listMuchLinkedType, directedLineSegmentType,		(* *)
		idRealPointType, point3dType, realPoint3dType, triangle2dType, triangulation3dType,					(* *)
		triangulation2dType, graphEdgeTriangulationType, quadTreeType, quadTreeNodeType,					(* *)
		triangle3dType, virtualPoint3dType, virtualGraphEdgeTriangulationType, virtualTriangle3dType,		(* *)
		upperMonotonePolygonType, searchTriangulation2dType, boundingBox2dType, cube3dType,					(* *)
		boundingBox3dType, markedGraphEdgeType, atomType, moleculeType, moleculeEdgeType, triangleType,		(* *)
		raumPlanungType, infoValuesType, resInfoType, gebaeudeType, beduerfnisseType, resourcenType,		(* *)
		histogramType, gebaeudeLabelType, chiralityType, atomInformationType, incClosestPairType,			(* *)
		incObjectType, incAllIntersectType, incPolIntersectType, connectionDescType, objType);				(* *)
(* New types must be added just in front of objType in order to maintain upward compatibility. 		*)
(* This is of course only possible if the lower types don't change their implementation for			*)
(* 'fileInput', 'fileOutput', 'textFileInput' and 'textFileOutput'.												*)

	pointTypeRange = realPointType..longIntPointType;
	dictionaryTypeRange = avlTreeType..sortedVectorType;
	priorityQueueTypeRange = dictPriorityQueueType..heapPriorityQueueType;

	objectTypeSet = set of objectType;

	obj = object (* Root object. *)

(* Memory management *)
			procedure init;
(* Initializes the internal state of an object, i.e. all instance variables are set to a  defined value.	*)
(* This method should be called after creating an object with 'new'.  Instance variables being			*)
(* objects themselves are created and initialized since they belong to the internal state. 'init' may	*)
(* use global variables defined in 'obj' to determine types or other values. The appropriate			*)
(* variables are specified in the respective class interface. After 'init' has been called the object	*)
(* is ready to use. The default implementation is to do nothing.												*)

			procedure shallowFree;
(* Low level method to free an object, should not be overridden.											*)

			procedure free;
(* Reclaims the memory belonging to the object (the reverse operation to 'init'). Depending objects	*)
(* (objects not belonging to the state) are not removed. The object is removed; no dispose is			*)
(* necessary or even allowed. The default implementation calls 'shallowFree' which works			*)
(* correctly for all objects that do not have other objects belonging to their internal state.			*)

			procedure freeAll;
(* Like 'free', but all dependent objects are also reclaimed recursively. The default implementation	*)
(* calls 'free' which is correct for objects that do not have dependent objects.							*)

			function shallowDuplicate: obj;
(* Low level method to duplicate an object, should not be overriden.										*)

			procedure shallowInitializeFromObject (from: obj);
(*	Initilializes (copies) the instance variables of 'self' from the object 'from'. If 'self's class is a	*)
(*	subclass 'from's class then some instance variables are left uninitialized (type extension). If		*)
(*	'from's class is a subclass of 'self's class then only the common instance variables are copied	*)
(*	(type restriction). Should not be overridden.																*)

			function duplicate: obj;
(* Produces a copy of the object. Note that instance variables are also copied but objects not			*)
(* belonging to the state (independent objects) are not copied. The default implementation calls		*)
(* 'shallowDuplicate' which works correctly for all objects that do not have other objects			*)
(* belonging to their internal state.																				*)

			function duplicateAll: obj;
(* Like 'duplicate', but all dependent objects are copied recursively. The default implementation		*)
(* calls 'duplicate' which is correct for objects without dependent objects.								*)

			function duplicateTypeTransfer (t: objectType): obj;
(* Like 'duplicate', but the result is of type 't'.  The restrictions of 'shallowInitializeFromObject'	*)
(*	apply analogously.	Based on  'duplicate' and 'shallowInitializeFromObject', should not be			*)
(*	overridden.																										*)

			function duplicateAllTypeTransfer (t: objectType): obj;
(* Like 'duplicateAll', but the result is of type 't'. The restrictions of 'shallowInitializeFromObject'	*)
(*	apply analogously. Based on  'duplicateAll' and 'shallowInitializeFromObject', should not be		*)
(*	overridden.																										*)


(* Interactive I/O *)
			procedure interactiveOutput;
(* Draws the object on the current window. Because of the XOR mode an object can be removed		*)
(* from the window by a second call to 'interactiveOutput' which is useful in algorithm animation.	*)
(* The world coordinate system is used and the standard implementation raises an error condition.	*)
(*	*)
(*	Warning: interactiveOutput is also used to draw into the external scrapbook, e.g., if the user		*)
(*	selected Copy from the Edit menu and leaves the application. Do not assume that you are		*)
(*	necessarily drawing into a window when your interactiveOutput method is called. However,		*)
(*	you may use thePort^.portRect to determine the outline of thewindow you are drawing into. 	*)
(*	If you really need to know whether you are drawing or just writing to the scrapbook,				*)
(*	use thePort^.pnVis  (true means you are drawing in a window). See the implementation of			*)
(*	voronoiDiagram.interactiveOutput as an example.															*)

			procedure interactiveHighlight;
(* Draws the object on the current window in a highlighted fashion. The world coordinate system is	*)
(* used. The standard implementation enlarges the pen size and calls 'interactiveOutput'. 				*)

			procedure interactiveFlash;
(* Let the object flash in the current window. The object is not drawn. The world coordinate			*)
(* system is used and the default implementation is implemented completely in terms of				*)
(* 'interactiveHighlight'. 																							*)

			procedure interactiveInput (px, py: integer);
(* Like 'windowInput' but the world coordinate system is used. The default implementation is based	*)
(* completely on 'windowInput' and 'windowToWorld'. 														*)

			procedure windowOutput;
(* Draws  the internal state on the current window. The window coordinate system is used and the	*)
(* default implementation is based completely on 'interactiveOutput' although this is very				*)
(* inefficient in most cases. 																						*)

			procedure windowInput (px, py: integer);
(* Like 'init', but the state is read using the mouse. It gives visual feedback but leaves the drawing	*)
(* window unchanged after completion (XOR!). 'px' and 'py' are the coordinates of the first mouse	*)
(* click. The window coordinate system is used. The default implementation of 'windowInput' is		*)
(* based completely on 'construct' and 'windowOutput'.														*)

			procedure windowInputMouse (mouse: Point);
(*	Calls 'windowInput' with 'mouse.h' and 'mouse.v'. Should not be overwritten.						*)

			procedure construct (px, py, newX, newY: integer);
(* This method changes an existing object on which 'init' has been called according to the  four		*)
(* parameters formed by the first click location ('px', 'py') and the current click location				*)
(* ('newX', 'newY'). The window coordinate system is used. The default implementation raises an	*)
(* error condition.																									*)

			procedure constructWorld (mouse: Point);
(*	Similar to 'construct', but world coordinates are used and only the position of the current			*)
(*	click location ('mouse') is given.	Used by 'dragVertexWorld'. The default implementation raises	*)
(*	an error condition.																								*)

			procedure showObjectParameter;
(*	Displays parameters of the object in the info-window. This method is used when constructing		*)
(*	an object (via 'construct' and 'windowInput') or when 'dragVertex' is activated.	 The default		*)
(*	method does nothing.																							*)

			function drawingComplexity (cutOff: longInt): longInt;
(*	Determines the complexity of drawing the object on the screen. If the complexity is at least		*)
(*	'cutOff', the function returns a value which is at least 'cutOff' and at most the correct				*)
(*	complexity.	The default implementation returns 'cutOff'.												*)


(* Binary and text file I/O *)
			procedure fileOutput;
(* The whole object is written to 'currentStream' but no type identifier is written. The default		*)
(* implementation raises an error condition. 																	*)

			procedure fileOutputTag;
(* Like 'fileOutput', but  a type identifier is written first. The function 'fileGetType' in module 'obj'	*)
(* can be used to retrieve the type. The default implementation is based completely on  'getType'	*)
(* and 'fileOutput'. 																									*)

			procedure fileInput;
(* The complete object is read from 'currentStream', the first byte on f is the first byte of the		*)
(* internal state. The default implementation raises an error condition. 									*)

			procedure textFileOutput;
(* Writes the internal state in textual form to 'currentStream'. The first token written is always	*)
(* the type of the object. The grammar is given in the appendix. The global variable					*)
(*	'currentTextFileOutputMethod' determines which grammar is used for formatting the output.		*)
(*	The default implementation raises an error condition. 													*)

			procedure textFileInput;
(* Reads the internal state in textual form from 'currentStream'. The first token is always the		*)
(* type of the object. The grammar is given in the appendix. The default implementation raises an	*)
(* error condition. 																									*)

			function textFileOutputId: Str255;
(* Returns the type (keyword) of the object. *)


(* Description / execution. Whenever objects are selected by the user the interactive front end of	*)
(* the XYZ GeoBench calls for each class in the system the respective 'describe' method thereby	*)
(* asking each class whether an operation of this class can be executed on the selected objects.		*)
(* This is the way GeoBench determines which operations are appropriate for which objects.			*)
(* Whenever a new method should be made available to the end user the pair 'describe' / 'execute'	*)
(* must be changed to reflect the new algorithm.																*)

			procedure describe;
(* Computes the (possibly hierarchical) menu entries of operations applicable to the list of objects	*)
(* in the global variable 'currentArguments'. The procedures 'menuEntry' (plus 'beginSubMenu' and	*)
(* 'endSubMenu' for hierarchical menus) from the module 'sequence' are to be used. The default		*)
(* implementation does nothing.																					*)

			procedure execute (item: integer);
(* Executes the selected method 'item' with arguments in 'currentArguments' and appends the		*)
(* result to the global variable 'currentResult'. The default implementation raises an error			*)
(* condition. 																										*)


(* Geometric transformations *)
			procedure translate (dx, dy: extended);
(* Translates an object by (dx,dy). The default implementation raises an error	condition. 			*)

			procedure rotate (xOrigin, yOrigin, alpha: extended);
(* Rotate the object by 'alpha' (radians) around ('xOrigin', 'yOrigin'). The default implementation	*)
(* raises an error condition. 																						*)

			procedure scale (lambda: extended);
(* Scale the object by 'lambda'. The default implementation raises an error condition. 					*)

			procedure reflect (a, b, c: extended);
(* Reflect the object at the line  a*x+b*y = c, a*a+b*b  0. The default implementation raises an	*)
(* error condition. 																									*)

(*	Generic geoometric operations *)
			function area: extended;
(*	Computes the area of the geometric object, if applicable, 0 otherwise. The default					*)
(*	implementation returns 0. 																					*)

			function perimeter: extended;
(* Computes the perimeter of the geometric object, if applicable, 0 otherwise.	The Lp-metric		*)
(*	should be used.  The default implementation returns 0. 													*)

			function perimeterMP: obj; (* actually a markedRealPoint *)
(*	Creates a markedRealPoint which is marked with the 'perimeter' of the object.						*)

			function areaMP: obj; (* actually a markedRealPoint *)
(*	Creates a markedRealPoint which is marked with the 'area' of the object.								*)

			function convertToPolygon: obj;
(*	Transforms the object to a polygon (if possible). The default implementation calls					*)
(*	'convertToConvexPolygon'.																					*)

			function convertToPolygonPossible: boolean;
(*	Returns 'true' iff the conversion of the object to a polygon is possible. The default 					*)
(*	 implementation returns 'false'.																				*)

			function convertToConvexPolygon: obj;
(*	Transforms the object to a convex polygon (if possible). The default implementation returns		*)
(*	'nil'.																							*)

			function convertToConvexPolygonPossible: boolean;
(*	Returns 'true' iff the conversion of the object to a convex polygon is possible. The default			*)
(*	implementation calls 'convertToPolygonPossible'.															*)

			function getLeftPoint: obj;
(*	Computes the leftmost point of 'self'. Only defined for convex polygons, line segments, points	*)
(*	and circles. Introduced by Michael Vollenbroeker 12.2.93.												*)

			function getRightPoint: obj;
(*	Computes the rightmost point of 'self'. Only defined for convex polygons, line segments, points	*)
(*	and circles. Introduced by Michael Vollenbroeker 12.2.93.												*)

			function verticalDistance (o: obj): extended;
(* Computes the vertical distance between the objects 'self' and 'o'. Only defined for convex			*)
(*	polygons, line segments, circles and points. Introduced by Michael Vollenbroeker 14.1.93.			*)

			function distanceO (o: obj; var minP, minQ: obj): extended;
(*	Computes the distance between 'self' and 'o' which is achieved in 'minP' and 'minQ'. Only defined	*)
(*	for convex polygons, line segments, points and circles. Introduced by Michael Vollenbroeker		*)
(*	12.2.93.																											*)

			function closestPair: obj;
(*	Computes the closest pair of a configuration of convex objects using the plane sweep algorithm	*)
(*	by Hinrichs/Bartling/Graf. Works only with homogenous vectors of convex polygons, line			*)
(*	segments, circles and points.	ntroduced by Michael Vollenbroeker 3.12.92. 							*)


(* Other transformations *)
			procedure randomChange;
(* Changes the internal state randomly governed by 'currentRandomConstraint'. This global			*)
(* variable is usually interpreted as a rectangle, though it could be a more complex object.  The		*)
(* default implementation raises an error condition.															*)

			procedure windowToWorld;
(* Transforms the object to world coordinates. The default implementation raises an error			*)
(* condition. 																										*)


(* Type computation and comparison *)
			function getType: objectType;
(* Get 'self''s type.  The default implementation raises an error condition and returns 'objType'. 	*)

			function getTextType: Str255;
(* Computes a textual description of 'self''s type. The default implementation raises an error		*)
(* condition and returns the string 'Obj'. 																		*)

			function getPointType: objectType;
(*	Yields the point-type of 'self' if 'self' is a convex polygon, line segment, point or circle.			*)
(*	Introduced by Michael Vollenbroeker 12.2.93.																*)

			function getInfo: obj;
(* Displays in a Dialog-Box Informations about 'self'.  														*)

			procedure setInfo (o: obj);
(* Set information acording to the information in o.															*)

			function memberType (o: obj): boolean;
(* Tests whether 'o' is a descendant of 'self.'  The default implementation raises an error			*)
(* condition and returns 'TRUE'. 																					*)

			function convertToType (t: pointTypeRange): obj;
(* Produces a copy of 'self' based on type 't'. The default implementation returns nil.					*)

			function equal (o: obj): boolean;
(* Determines whether 'self' is identical to 'o'.  The default implementation raises an error			*)
(* condition and returns 'FALSE'. 																				*)

			function genericLessThan (o: obj): boolean;
(* Determines whether 'self' is less than 'o' where 'o.getType = self.getType'. This function		*)
(* imposes some canonical order on the objects of a type and is used to eliminate duplicates by		*)
(* sorting. The default implementation raises an error condition and returns TRUE iff 'o' and			*)
(* 'self' are different.																							*)

			function HBGLessThan (o: obj): boolean;
(*	Determines whether 'self' is less than 'o' in the order defined by Hinrichs/Bartling/Graf. The 	*)
(*	position of the sweepline is determined by the leftmost x-coordinate of the rightmost object.		*)
(*	Introduced by Michael Vollenbroeker 3.12.92.																*)

{$IFC GFEnabled }
			function convertToParameterPoint: obj;
(* Converts 'self' to the parameter point. The default implementation raises an error condition	*)
(* and returns the result of 'duplicateAll'. 																		*)

			function convertFromParameterPoint (pP: obj): obj;
(* Converts 'self' from the parameter point 'pP' . The default implementation raises an error condition	*)
(* and returns the result of 'duplicateAll'. 																		*)

			procedure createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
(* Returns in 'lowpP' and 'highpP' the rangequery. This must be parameter points. In '' stands, if a attribut *)
(* is defined. 	*)

			function canUseGridfile: BOOLEAN;
(* Returns true if the object can work with the gridfile. The default implementation returns 'FALSE'. 	*)
{$ENDC}

(* Selecting and dragging *)
			function isSelected (where: obj): boolean;
(* Determines whether the object is selected by the given rectangle 'where'. The default				*)
(* implementation raises an error condition and returns 'FALSE'. 											*)

			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
(* Tests whether the user attempted to drag a point (vertex) inside the rectangle 'where'. The		*)
(*	initial position of the mouse is 'mouse'.	If 'highlight' is true, the object is currently displayed in	*)
(*	highlighted mode. The default implementation returns 'FALSE'. 											*)

			procedure dragVertexWorld (mouse: Point; highlight: boolean);
(*	The user pressed the button at 'mouse' and drags a vertex of the object. Constructs a new		*)
(*	object in world coordinates based on the current mouse position. The implementation is based		*)
(*	on 'interactiveOutput', 'constructWorld' and 'showObjectParameter'.									*)

			procedure dragObjectTransformTranslation (procedure whatToDoBefore (o: obj); procedure whatToDoAfter (o: obj); mouse: Point; procedure transformTranslation (var dx, dy: extended));
(*	Drags the object on the screen while the button is pressed. Calls 'whatToDoBefore' before			*)
(*	moving the object and 'whatToDoAfter' after moving the object whenever the position of the		*)
(*	object changes.	The parameter 'o' contains 'self'. The initial position of the mouse is 'mouse'.	*)
(*	The procedure 'transformaTranslation' is called prior to the translation and can be used to		*)
(*	change the translation vector.																				*)

			procedure dragObject (procedure whatToDoBefore (o: obj); procedure whatToDoAfter (o: obj); mouse: Point);
(*	Calls 'dragObjectTransformTranslation' with a transformation which does nothing.					*)


(* Miscellaneous *)
			procedure textOutput;
(* Outputs the object in textual form in the THINK Pascal text window. Used for debugging. The		*)
(* default implementation raises an error condition. 															*)

			procedure centerOfGravity (var cx, cy: extended);
(* Returns a point ('cx', 'cy') close to the object. The center is used for drawing graphs. The			*)
(* default implementation raises an error condition and returns the point (0, 0). 						*)

			function pointCover (n: longInt): obj;
(* Covers the object with equally spaced points, n point for each atomic part. The result is a			*)
(*	'pointVector'. The type of	the points is determined by 'currentPointType'. The default				*)
(*	implementation returns 'nil'. 																					*)

			function randomPointCover (n: longInt): obj;
(* Covers the object with random points, n point for each atomic part. The result is a					*)
(*	'pointVector'. The type of	the points is determined by 'currentPointType'. The default				*)
(*	implementation returns 'nil'. 																					*)

			function getBoundingObject: obj;
(* Returns a 'simple' object that contains self, for example a rectangle for 2d objects or a cube	*)
(* for 3d objects. The default implementation raises an error condition and returns 'nil'.				*)

(* following function added by "Von0auf100", Thema QuadTrees *)
			function inclRectangle (bottomS, leftS, topS, rightS: extended): integer;
(*	Abstract method. Tests if 'self' lies in the rectangle given by bottomS, leftS, topS, rightS.		*)
(*	Returnvalue is						1	: if the rectangle is totally inside of the  'self'					*)
(*											0	: if the rectangle is partially inside of the  'self'				*)
(*											-1	: if the rectangle is totally outside of the  'self'				*)
(*	end of added function *)
		end;

	relationObj = (lessThanObj, equalObj, greaterThanObj, unEqualObj);

	objInfoElement = object(obj)
	(* public *)
			desc: string;
			protect: boolean;
	(* private *)
			cellInfo: string;
{$IFC ScriptEnabled }
			col2, col3: string; (* for compatibility reasons, ugly. *)
{$ENDC }
		end;

	objInfoExtended = object(objInfoElement)
			e: extended;
		end;

	objInfoString = object(objInfoElement)
			s: string;
		end;

	objInfoObj = object(objInfoElement)
			o: obj;
		end;

{$IFC NOT THINK_PASCAL }
	cleanUpT = procedure;
{$ENDC }

	objInfoTempObj = object(objInfoObj)
{$IFC THINK_PASCAL }
			cleanUp: ProcPtr;	(*	This procedure is called from 'free' after destroying 'o'. *)
{$ELSEC }
			cleanUp: cleanUpT;	(*	This procedure is called from 'free' after destroying 'o'. *)
{$ENDC }
			procedure free;
			override;
		end;

const
	firstObjectType = intType;
	lastObjectType = objType;

function newObj (t: objectType): obj;
(* like 'new()', create an object of type 't' *)

function newInitObj (t: objectType): obj;
(* like 'newObj', but object is already initialized *)

function fileGetType: objectType;
(* the next byte of the 'currentStream' is interpreted as an 'objectType' *)

function oppositeRelationObj (x: relationObj): relationObj;
(* 'lessThanObj' <==> 'greaterThanObj' *)

function sameType2 (x, y: obj): boolean;
(* tests whether 'x' and 'y' have identical types *)

function sameType3 (x, y, z: obj): boolean;
(* tests whether 'x', 'y' and 'z' have identical types *)

procedure doNothingObjP (o: obj);

procedure initObj;

procedure exitObj;

var
prototypes: array[objectType] of obj;					(* only new(), 'init' must be called later! *)
currentRandomConstraint: obj;								(* usually a rectangle *)
currentPointType: pointTypeRange;							(* which point to produce *)
currentDictionaryType: objectType;							(* which dictionary to produce *)
currentPriorityQueueType: priorityQueueTypeRange; 	(* which priority queue to choose *)
currentVectorLength: longInt;									(* how many elements to allocate initially for a 'vector' *)
currentEdges, currentVertices: longInt;						(* number of vertices and edges in graphs. 'currentVertices' is also used by 'regularNGon' for the number of vertices *)
currentDimension: dimensionRange;							(* dimension of 'dDimPoint' to be created *)
currentRegularNGonAngle: real;								(* angle of N-gon to be created *)
currentAlignmentIsOn: boolean;								(* determines whether alignment is turned on for regular n-gons *)
currentWidthPercentage: real;								(* for some 'currentRandomConstraint's *)
currentHeightPercentage: real; 								(* for some 'currentRandomConstraint's *)
currentTextFileOutputMethod: integer;						(*	0 = standard, 1= point lists, 2 = used by O'Rourke's convex intersecter *)
additionalDraw, additionalHighlight: obj;

implementation

const
	blinkNumber = 4;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure doNothingObjP (o: obj);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

{$IFC THINK_PASCAL }
procedure callP (myProc: ProcPtr);
inline
	$205f, $4e90;
{$ENDC }

procedure objInfoTempObj.free;
	begin
		o.freeAll;
{$IFC THINK_PASCAL }
		callP(cleanUp);
{$ELSEC }
		cleanUp;
{$ENDC }
		inherited free;
	end;

procedure obj.init; (* may be called *)
	begin
	end;

procedure obj.shallowFree;
	begin
{$IFC THINK_PASCAL }
		disposHandle(Handle(self));
{$ELSEC }
		DisposeHandle(Handle(self));
{$ENDC }
	end;

procedure obj.free; (* may be called *)
	begin
		shallowFree;
	end;

procedure obj.freeAll; (* may be called *)
	begin
		free;
	end;

function obj.shallowDuplicate: obj;
	var
		result: Handle;
	begin
		result := Handle(self);
		if HandToHand(result) <> noErr then
			failNil(nil); (* Out of memory! *)
		shallowDuplicate := obj(result);
	end;

procedure obj.shallowInitializeFromObject (from: obj);
	const
{$IFC THINK_PASCAL }
{$IFC FarCode }
		offset = 4;
{$ELSEC }
		offset = 2;
{$ENDC }
{$ELSEC }
		offset = 4;
{$ENDC }

	function pointerToInstanceVariables (x: obj): Ptr;
		begin
			pointerToInstanceVariables := Ptr(LongInt(Handle(x)^) + offset);
		end;

	begin
{$IFC THINK_PASCAL }
		BlockMove(pointerToInstanceVariables(from), pointerToInstanceVariables(self), trunc(min(GetHandleSize(Handle(from)), GetHandleSize(Handle(self)))) - offset);
{$ELSEC }
		BlockMove(pointerToInstanceVariables(from), pointerToInstanceVariables(self), truncLong(min(GetHandleSize(Handle(from)), GetHandleSize(Handle(self)))) - offset);
{$ENDC }
	end;

function obj.duplicateTypeTransfer (t: objectType): obj;
	var
		r, o: obj;
	begin
		r := newObj(t);
		o := duplicate; (* This is necessary for the case that 'self' contains dependent objects which belong to 'self's' state. *)
		r.shallowInitializeFromObject(o);
		o.shallowFree;
		duplicateTypeTransfer := r;
	end;

function obj.duplicateAllTypeTransfer (t: objectType): obj;
	var
		r, o: obj;
	begin
		r := newObj(t);
		o := duplicateAll;
		r.shallowInitializeFromObject(o);
		o.shallowFree;
		duplicateAllTypeTransfer := r;
	end;

function obj.duplicate: obj; (* default duplication *)
	begin
		duplicate := shallowDuplicate;
	end;

function obj.duplicateAll: obj; (* default duplication, assume no dependent objects *)
	begin
		duplicateAll := duplicate;
	end;

procedure obj.showObjectParameter;
	begin
	end;

function obj.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := cutOff;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure obj.construct (px, py, newX, newY: integer);
	begin
		unimplementedMethod('obj.construct');
	end;

procedure obj.constructWorld (mouse: Point);
	begin
		unimplementedMethod('obj.constructWorld');
	end;

procedure noTransformTranslation (var dx, dy: extended);
	begin
	end;

function obj.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	begin
		dragVertex := false;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure obj.dragObject (procedure whatToDoBefore (o: obj); procedure whatToDoAfter (o: obj); mouse: Point);
	begin
		dragObjectTransformTranslation(whatToDoBefore, whatToDoAfter, mouse, noTransformTranslation);
	end;

procedure obj.dragVertexWorld (mouse: Point; highlight: boolean);
	var
		currentMouse: Point;
	begin
		while Button do begin
			writeCursorCoordinates(false);
			GetMouse(currentMouse);
			if not EqualPt(mouse, currentMouse) then begin
				interactiveOutput;
				if highlight then
					interactiveHighlight;
				constructWorld(currentMouse);
				interactiveOutput;
				if highlight then
					interactiveHighlight;
				showObjectParameter;
				mouse := currentMouse;
			end;
		end;
	end;

procedure obj.textFileInput;
	begin
		unimplementedMethod('obj.textFileInput');
	end;

procedure obj.textFileOutput;
	begin
		unimplementedMethod('obj.textFileOutput');
	end;

function obj.textFileOutputId: Str255;
	begin
{$IFC THINK_PASCAL }
		textFileOutputId := StringOf(getType);
{$ELSEC }
		textFileOutputId := StringOf('Type',ord(getType));
{$ENDC }
	end;

procedure obj.fileInput;
	begin
		unimplementedMethod('obj.fileInput');
	end;

procedure obj.fileOutput;
	begin
		unimplementedMethod('obj.fileOutput');
	end;

procedure obj.fileOutputTag;
	begin
		writeByte(Byte(getType));
		fileOutput;
	end;

procedure obj.windowInput (px, py: integer);
	var
		mouse, oldMouse: Point;
	begin
		init;
		construct(px, py, px, py);
		windowOutput;
		oldMouse.h := px;
		oldMouse.v := py;
		while Button do begin
			writeCursorCoordinates(false);
			GetMouse(mouse);
			if not EqualPt(mouse, oldMouse) then begin
				windowOutput;
				construct(px, py, mouse.h, mouse.v);
				windowOutput;
				showObjectParameter;
				oldMouse := mouse;
			end;
		end;
		windowOutput;
	end;

procedure obj.windowInputMouse (mouse: Point);
	begin
		windowInput(mouse.h, mouse.v);
	end;

procedure obj.interactiveInput (px, py: integer);
	begin
		windowInput(px, py);
		windowToWorld;
	end;

procedure obj.windowOutput;
	var
		safe: wFrame;
	begin
		safe := currentFrame;
		currentFrame := standardFrame;
		interactiveOutput;
		currentFrame := safe;
	end;

procedure obj.interactiveOutput;
	begin
		unimplementedMethod('obj.interactiveOutput');
	end;

procedure obj.interactiveHighlight;
	begin
		PenSize(3, 3);
		interactiveOutput;
		PenSize(1, 1);
	end;

procedure obj.interactiveFlash;
	var
		i: 1..blinkNumber;
		t, blinkWait: longInt;
	begin
{$IFC THINK_PASCAL }
		blinkWait := trunc(movieWait / 6.0);
{$ELSEC }
		blinkWait := truncLong(movieWait / 6.0);
{$ENDC }
		if blinkWait < 3 then
			blinkWait := 3
		else if blinkWait > 10 then
			blinkWait := 10;
		for i := 1 to blinkNumber do begin
			interactiveHighlight;
			t := TickCount + blinkWait;
			repeat
			until TickCount >= t;
		end;
	end;

procedure obj.textOutput;
	begin
		unimplementedMethod('obj.textOutput');
	end;

procedure obj.windowToWorld;
	begin
		unimplementedMethod('obj.windowToWorld');
	end;

procedure obj.describe; (* default: nothing to describe *)
	begin
	end;

procedure obj.randomChange;
	begin
	end;

function obj.genericLessThan (o: obj): boolean;
	begin
		unimplementedMethod('obj.genericLessThan');
		genericLessThan := not equal(o);
	end;

function obj.memberType (o: obj): boolean;
	begin
		unimplementedMethod('obj.memberType');
		memberType := member(o, obj); (* true *)
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function obj.isSelected (where: obj): boolean;
	begin
		unimplementedMethod('obj.isSelected');
		isSelected := false;
	end;

procedure obj.execute (item: integer);
	begin
		unimplementedMethod('obj.execute');
	end;

function obj.equal (o: obj): boolean;
	begin
		unimplementedMethod('obj.equal');
		equal := false;
	end;

procedure obj.translate (dx, dy: extended);
	begin
		unimplementedMethod('obj.translate');
	end;

procedure obj.rotate (xOrigin, yOrigin, alpha: extended);
	begin
		unimplementedMethod('obj.rotate');
	end;

procedure obj.scale (lambda: extended);
	begin
		unimplementedMethod('obj.scale');
	end;

procedure obj.reflect (a, b, c: extended);
	begin
		unimplementedMethod('obj.reflect');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function obj.area: extended;
	begin
		area := 0.0;
	end;

function obj.perimeter: extended;
	begin
		perimeter := 0.0;
	end;

function obj.convertToPolygon: obj;
	begin
		convertToPolygon := convertToConvexPolygon;
	end;

function obj.convertToPolygonPossible: boolean;
	begin
		convertToPolygonPossible := false;
	end;

function obj.convertToConvexPolygon: obj;
	begin
		convertToConvexPolygon := nil;
	end;

function obj.convertToConvexPolygonPossible: boolean;
	begin
		convertToConvexPolygonPossible := convertToPolygonPossible;
	end;

function obj.getTextType: Str255;
	begin
		getTextType := 'Obj';
	end;

procedure obj.centerOfGravity (var cx, cy: extended);
	begin
		cx := 100;
		cy := 100;
	end;

function obj.randomPointCover (n: longInt): obj;
	var
		c: obj;
	begin
		if convertToConvexPolygonPossible then begin
			c := convertToConvexPolygon;
			randomPointCover := c.randomPointCover(n);
			c.freeAll;
		end
		else
			randomPointCover := nil;
	end;

function obj.getPointType: objectType;
	begin
		unimplementedMethod('obj.getPointType');
		getPointType := realPointType;
	end;

function obj.getLeftPoint: obj;
	begin
		unimplementedMethod('obj.getLeftPoint');
		getLeftPoint := nil;
	end;

function obj.getRightPoint: obj;
	begin
		unimplementedMethod('obj.getRightPoint');
		getRightPoint := nil;
	end;

function obj.getBoundingObject: obj;
	begin
		unimplementedMethod('obj.getBoundingObject');
		getBoundingObject := nil;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function obj.convertToType (t: pointTypeRange): obj;
	begin
		convertToType := nil;
	end;

function obj.pointCover (n: longInt): obj;
	begin
		pointCover := nil;
	end;

function obj.HBGLessThan (o: obj): boolean;
	begin
		unimplementedMethod('obj.HBGLessThan');
		HBGLessThan := false;
	end;

function obj.verticalDistance (o: obj): extended;   (* Eingefuegt von Michael Vollenbroeker 14.1.93 *)
	begin
		unimplementedMethod('obj.verticalDistance');
		verticalDistance := 0;
	end;

function obj.distanceO (o: obj; var minP, minQ: obj): extended;
	begin
		unimplementedMethod('obj.distanceO');
		minP := nil;
		minQ := nil;
		distanceO := 0;
	end;

function obj.inclRectangle (bottomS, leftS, topS, rightS: extended): integer;
	begin
		unimplementedMethod('obj.inclRectangle');
		inclRectangle := -1;
	end;

procedure obj.setInfo (o: obj);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function obj.getType: objectType;
	begin
		unimplementedMethod('obj.getType');
		getType := lastObjectType;
	end;

function obj.getInfo: obj;
	begin
		getInfo := nil;
	end;

{$IFC GFEnabled }
function obj.convertToParameterPoint: obj;
	begin
		unimplementedMethod('obj.convertToParameterPoint');
		convertToParameterPoint := nil;
	end; { convertToParameterPoint }

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function obj.convertFromParameterPoint (pP: obj): obj;
	begin
		unimplementedMethod('obj.convertFromParameterPoint');
		convertFromParameterPoint := nil;
	end; { convertFromParameterPoint }

procedure obj.createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
	begin
		unimplementedMethod('obj.createQuery');
	end; { createQuery }
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function obj.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := false;
	end;
{$ENDC}

function newObj (t: objectType): obj;
	begin
		if prototypes[t] = nil then begin
{$IFC THINK_PASCAL }
			unimplementedMethod(concat('newObj(', stringOf(t), ')'));
{$ELSEC }
			unimplementedMethod(concat('newObj(Type', StringOf(ord(t)), ')'));
{$ENDC }
			newObj := nil;
		end
		else
			newObj := prototypes[t].shallowDuplicate;
	end;

function newInitObj (t: objectType): obj;
	var
		o: obj;
	begin
		o := newObj(t);
		o.init;
		newInitObj := o;
	end;

{$SETC TransformHistoGramToGebaeudeLabel = false }

function fileGetType: objectType;
	const
		oldVersionAlert = 1016;
		haltItem = 1;
	var
		b: Byte;
	begin
		readByte(b);
{$IFC TransformHistoGramToGebaeudeLabel }
		if b = ord(histogramType) then
			b := ord(gebaeudeLabelType);
{$ENDC }
		if (b > ord(objType)) & (StopAlert(oldVersionAlert, nil) = haltItem) then begin
			closeReadStream;
			halt;
		end;
		fileGetType := objectType(b);
	end;

function oppositeRelationObj (x: relationObj): relationObj;
	begin
		oppositeRelationObj := relationObj(ord(greaterThanObj) - ord(x));
	end;

function sameType2 (x, y: obj): boolean;
	begin
		sameType2 := x.memberType(y) & y.memberType(x);
	end;

function sameType3 (x, y, z: obj): boolean;
	begin
		sameType3 := sameType2(x, y) & sameType2(y, z);
	end;

procedure initObj;
	var
		i: objectType;
		x: obj;
	begin
		for i := firstObjectType to lastObjectType do
			prototypes[i] := nil;
		new(x);
		failNil(Handle(x));
		prototypes[objType] := x;
		currentRandomConstraint := nil;
		currentPointType := realPointType;
		currentDictionaryType := avlTreeType;
		currentVectorLength := 0;
		currentPriorityQueueType := dictPriorityQueueType;
		currentEdges := 0;
		currentVertices := 0;
		currentDimension := 2;
		currentAlignmentIsOn := false;
		currentWidthPercentage := 0.0;
		currentHeightPercentage := 0.0;
		currentTextFileOutputMethod := 0;
		additionalDraw := nil;
		additionalHighlight := nil;
	end;

procedure exitObj;
	var
		i: objectType;
	begin
		for i := intType to objType do
			if prototypes[i] <> nil then
				dispose(prototypes[i]);
	end;

end.
