(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1991 by Philipp Meier						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.00:  27-Mai-91								*)

unit point3d;

interface

uses
	basicGeometry;

type
	point3d = object(obj)
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function getX: extended;					(* Get the value of the x-coordinate. *)
			procedure setX (newX: extended);			(* Set the value of the x-coordinate. *)
			function getY: extended;					(* Get the value of the y-coordinate. *)
			procedure setY (newY: extended);			(* Set the value of the y-coordinate. *)
			function getZ: extended;					(* Get the value of the z-coordinate. *)
			procedure setZ (newZ: extended);			(* Set the value of the z-coordinate. *)
			function xLessThan (p: point3d): boolean;	(* Is 'self.getX < p.getX' ? *)
			function xEqual (p: point3d): boolean;		(* Is 'self.getX = p.getX' ? *)
			function yLessThan (p: point3d): boolean;	(* Is 'self.getY < p.getY' ? *)
			function yEqual (p: point3d): boolean;		(* Is 'self.getY = p.getY' ? *)
			function zLessThan (p: point3d): boolean;	(* Is 'self.getZ < p.getZ' ? *)
			function zEqual (p: point3d): boolean;		(* Is 'self.getZ = p.getZ' ? *)
			procedure setPoint (p: point3d);

(* Drawing *)
			procedure drawLine (p: point3d);			(* Draws a line from 'self' to 'p'. *)
			procedure labelPoint (l: Str255; r: integer); (* Draws the string 'l' below 'self'. *)

			function distanceX (p: point3d): extended;
(* The Euclidean distance between 'self' and 'p'. *)

			function distanceLpX (p: point3d): extended;
(* The Lp-distance between 'self' and 'p'. *)

			function squaredDistanceX (p: point3d): extended;
(* The squared Euclidean distance between 'self' and 'p'. *)

			function squaredDxX (p: point3d): extended;
(* The squared difference in x-coordinates between 'self' and 'p'. *)

			function squaredDyX (p: point3d): extended;
(* The squared difference in y-coordinates between 'self' and 'p'. *)

			function squaredDzX (p: point3d): extended;
(* The squared difference in z-coordinates between 'self' and 'p'. *)

			function middle (p: point3d): point3d;
(* Computes the point ('self.getX + p.getX', 'self.getY + p.getY', 'SELV.getZ + p.getZ') / 2. *)

			function crossProduct (p, q, r: point3d): point3d;
(* The cross product XP  QR *)

			function innerProduct (p, q, r: point3d): extended;
(* The inner product of XP and QR *)

			function cosPhi (p: point3d): extended;
(* the two points self and p are looked at as vectors (the beginning is the origin) *)
(* this method calculates the cosinus of the angle between self and p *)

			procedure difference (p: point3d; var res: point3d);
(* res := p - self *)

		end; (* point3d *)


	realPoint3d = object(point3d)
			x, y, z: real;
			procedure drawLine (p: point3d); (* Draws a line from 'self' to 'p'. *)
			override;
			procedure labelPoint (l: Str255; r: integer); (* Draws the string 'l' below 'self'. *)
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function getBoundingObject: obj;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			procedure interactiveHighlight;
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure textOutput;
			override;
			procedure randomChange;
			override;
			function getType: objectType;
			override;
			procedure translate (dx, dy: extended);
			override;
			function getX: extended;
			override;
			procedure setX (newX: extended);
			override;
			function getY: extended;
			override;
			procedure setY (newY: extended);
			override;
			function getZ: extended;
			override;
			procedure setZ (newZ: extended);
			override;
			function xLessThan (p: point3d): boolean;
			override;
			function xEqual (p: point3d): boolean;
			override;
			function yLessThan (p: point3d): boolean;
			override;
			function yEqual (p: point3d): boolean;
			override;
			function zLessThan (p: point3d): boolean;
			override;
			function zEqual (p: point3d): boolean;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getTextType: Str255;
			override;
{$IFC GFEnabled }
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}
			function middle (p: point3d): point3d;
			override;
			function crossProduct (p, q, r: point3d): point3d;
			override;
			function innerProduct (p, q, r: point3d): extended;
			override;
			function cosPhi (p: point3d): extended;
			override;
			procedure difference (p: point3d; var res: point3d);
			override;
			function distanceX (p: point3d): extended;
			override;
			function distanceLpX (p: point3d): extended;
			override;
			function squaredDistanceX (p: point3d): extended;
			override;
			procedure setPoint (p: point3d);
			override;
		end; (* realPoint3d *)

procedure initPoint3d;

implementation

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function point3d.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 2;
	end;

procedure point3d.setPoint (p: point3d);
	begin
		unimplementedMethod('point3d.setPoint');
	end;

procedure point3d.setX (newX: extended);
	begin
		unimplementedMethod('point3d.setX');
	end;

procedure point3d.setY (newY: extended);
	begin
		unimplementedMethod('point3d.setY');
	end;

procedure point3d.setZ (newZ: extended);
	begin
		unimplementedMethod('point3d.setZ');
	end;

function point3d.xLessThan (p: point3d): boolean;
	begin
		unimplementedMethod('point3d.xLessThan');
		xLessThan := false;
	end;

function point3d.xEqual (p: point3d): boolean;			(* Is 'self.getX = p.getX' ? *)
	begin
		unimplementedMethod('point3d.xEqual');
		xEqual := false;
	end;

function point3d.yLessThan (p: point3d): boolean;	(* Is 'self.getY < p.getY' ? *)
	begin
		unimplementedMethod('point3d.yLessThan');
		yLessThan := false;
	end;

function point3d.yEqual (p: point3d): boolean;			(* Is 'self.getY = p.getY' ? *)
	begin
		unimplementedMethod('point3d.yEqual');
		yEqual := false;
	end;

function point3d.zLessThan (p: point3d): boolean;	(* Is 'self.getZ < p.getZ' ? *)
	begin
		unimplementedMethod('point3d.zLessThan');
		zLessThan := false;
	end;

function point3d.zEqual (p: point3d): boolean;			(* Is 'self.getZ = p.getZ' ? *)
	begin
		unimplementedMethod('point3d.zEqual');
		zEqual := false;
	end;

(* Drawing *)
procedure point3d.drawLine (p: point3d); (* Draws a line from 'self' to 'p'. *)
	begin
		unimplementedMethod('point3d.drawLine');
	end;

procedure point3d.labelPoint (l: Str255; r: integer); (* Draws the string 'l' below 'self'. *)
	begin
		unimplementedMethod('point3d.labelPoint');
	end;

function point3d.distanceX (p: point3d): extended;
(* The Euclidean distance between 'self' and 'p'. *)
	begin
		unimplementedMethod('point3d.distanceX');
		distanceX := 0;
	end;

function point3d.distanceLpX (p: point3d): extended;
(* The Lp-distance between 'self' and 'p'. *)
	begin
		unimplementedMethod('point3d.distanceLpX');
		distanceLpX := 0;
	end;

function point3d.squaredDistanceX (p: point3d): extended;
(* The squared Euclidean distance between 'self' and 'p'. *)
	begin
		unimplementedMethod('point3d.squaredDistanceX');
		squaredDistanceX := 0;
	end;

function point3d.squaredDxX (p: point3d): extended;
(* The squared difference in x-coordinates between 'self' and 'p'. *)
	begin
		unimplementedMethod('point3d.squaredDxX');
		squaredDxX := 0;
	end;

function point3d.squaredDyX (p: point3d): extended;
(* The squared difference in y-coordinates between 'self' and 'p'. *)
	begin
		unimplementedMethod('point3d.squaredDyX');
		squaredDyX := 0;
	end;

function point3d.squaredDzX (p: point3d): extended;
(* The squared difference in z-coordinates between 'self' and 'p'. *)
	begin
		unimplementedMethod('point3d.squaredDzX');
		squaredDzX := 0;
	end;

function point3d.middle (p: point3d): point3d;
(* Computes the point ('self.getX + p.getX', 'self.getY + p.getY', 'SELV.getZ + p.getZ') / 2. *)
	begin
		unimplementedMethod('point3d.middle');
		middle := nil;
	end;

function point3d.crossProduct (p, q, r: point3d): point3d;
	begin
		unimplementedMethod('point3d.crossProduct');
		crossProduct := nil;
	end;

function point3d.innerProduct (p, q, r: point3d): extended;
	begin
		unimplementedMethod('point3d.innerProduct');
		innerProduct := 0;
	end;


function point3d.cosPhi (p: point3d): extended;
	begin
		unimplementedMethod('point3d.cosPhi');
		cosPhi := 0;
	end;

procedure point3d.difference (p: point3d; var res: point3d);
	begin
		unimplementedMethod('point3d.difference');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function point3d.getX: extended;
	begin
		unimplementedMethod('point3d.getX');
		getX := 0;
	end;

function point3d.getY: extended;
	begin
		unimplementedMethod('point3d.getY');
		getY := 0;
	end;

function point3d.getZ: extended;
	begin
		unimplementedMethod('point3d.getZ');
		getZ := 0;
	end;

procedure initPoint3d;
	var
		rp: realPoint3d;
		p: point3d;
	begin
		new(rp);
		failNil(Handle(rp));
		prototypes[realPoint3dType] := rp;
		new(p);
		failNil(Handle(p));
		prototypes[point3dType] := p;
	end;


end.
