(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit polyLine;

{$SETC ApolyLine = TRUE }

interface

uses
	vector, pointVector;

implementation

{$SETC AConvexHull = TRUE }
function polyLine.convexHullLinear: convexPolygon;
	var
		deque: homogenVector;
		top, bottom, current: seqIndex0;
		result: convexPolygon;
		p: point2d;
		topWhichSideSign, bottomWhichSideSign: signType;
{$IFC AnimationEnabled AND AConvexHull }
		l: lineSegment;
{$ENDC }
	begin
{$IFC AnimationEnabled AND AConvexHull }
		new(l);
		failNil(Handle(l));
{$ENDC }
		if length > 2 then begin
			deque := newHomogenVector(2 * length + 3);
			deque.elements^^[length + 1] := elements^^[1];
			current := 2;
			repeat
				deque.elements^^[length + 2] := elements^^[current];
				deque.elements^^[length] := elements^^[current];
				current := current + 1;
			until (current > length) | (point2d(elements^^[current]).whichSideX(point2d(elements^^[1]), point2d(elements^^[current - 1])) <> 0);
			bottom := length;
			top := length + 2;
{$IFC AnimationEnabled AND AConvexHull }
			if animationFlag[AConvexHullItem] then begin
				interactiveHighlight;
				l.p := point2d(elements^^[1]);
				l.q := point2d(deque.elements^^[top]);
				l.interactiveHighlight;
			end;
{$ENDC }
			while current <= length do begin
				p := point2d(elements^^[current]);
{$IFC AnimationEnabled AND AConvexHull }
				if animationFlag[AConvexHullItem] then begin
					p.interactiveFlash;
					waitForClick(animationFlag[AConvexHullItem]);
				end;
{$ENDC }
				topWhichSideSign := p.whichSideSign(point2d(deque.elements^^[top - 1]), point2d(deque.elements^^[top]));
				bottomWhichSideSign := p.whichSideSign(point2d(deque.elements^^[bottom]), point2d(deque.elements^^[bottom + 1]));
				if (topWhichSideSign <= 0) | (bottomWhichSideSign <= 0) then begin
{$IFC AnimationEnabled AND AConvexHull }
					if animationFlag[AConvexHullItem] then
						p.interactiveHighlight;
{$ENDC }
					while (top > bottom + 1) & (topWhichSideSign <= 0) do begin
{$IFC AnimationEnabled AND AConvexHull }
						if animationFlag[AConvexHullItem] then begin
							point2d(deque.elements^^[top]).interactiveFlash;
							waitForClick(animationFlag[AConvexHullItem]);
						end;
{$ENDC }
						top := top - 1;
						if top > bottom then
							topWhichSideSign := p.whichSideSign(point2d(deque.elements^^[top - 1]), point2d(deque.elements^^[top]));
					end;
					while (top > bottom + 1) & (bottomWhichSideSign <= 0) do begin
{$IFC AnimationEnabled AND AConvexHull }
						if animationFlag[AConvexHullItem] then begin
							point2d(deque.elements^^[bottom]).interactiveFlash;
							waitForClick(animationFlag[AConvexHullItem]);
						end;
{$ENDC }
						bottom := bottom + 1;
						if top > bottom then
							bottomWhichSideSign := p.whichSideSign(point2d(deque.elements^^[bottom]), point2d(deque.elements^^[bottom + 1]));
					end;
{$IFC AnimationEnabled AND AConvexHull }
					if animationFlag[AConvexHullItem] then begin
						l.p := p;
						l.q := point2d(deque.elements^^[top]);
						l.interactiveHighlight;
						l.q := point2d(deque.elements^^[bottom]);
						l.interactiveHighlight;
						waitForClick(animationFlag[AConvexHullItem]);
					end;
{$ENDC }
					top := top + 1;
					bottom := bottom - 1;
					deque.elements^^[top] := p;
					deque.elements^^[bottom] := p;
				end
				else begin
{$IFC AnimationEnabled AND AConvexHull }
					if animationFlag[AConvexHullItem] then
						p.interactiveOutput;
{$ENDC }
				end;
				current := current + 1;
			end;
			if (top - bottom > 2) & (point2d(deque.elements^^[top]).whichSideSign(point2d(deque.elements^^[bottom + 1]), point2d(deque.elements^^[top - 1])) = 0) then
				top := top - 1;
			currentVectorLength := top - bottom;
			new(result);
			failNil(Handle(result));
			result.init;
			result.length := currentVectorLength;
			for current := 1 to currentVectorLength do
				result.elements^^[current] := deque.elements^^[bottom + current].duplicate;
			convexHullLinear := result;
			deque.free;
		end
		else
			convexHullLinear := convexPolygon(duplicateAllTypeTransfer(convexPolygonType));
{$IFC AnimationEnabled AND AConvexHull }
		dispose(l);
{$ENDC }
	end;

	const
		ConvexHullC = 0;
		StraightLineSegmentsC = 1;
		DirectedLineSegmentsC = 2;

procedure polyLine.describe;
	begin
		with currentArguments do
			if (length = 1) & member(ith(1), polyLine) & (polyLine(ith(1)).length >= 2) then begin
				beginSubMenu('To Line Segment Vector');
				menuEntry('Straight Line Segments', StraightLineSegmentsC);
				menuEntry('Directed Line Segments', DirectedLineSegmentsC);
				endSubMenu;
				menuEntry('Convex Hull', ConvexHullC);
			end;
	end;

procedure polyLine.execute (item: integer);
	begin
		with currentArguments do
			case item of
				StraightLineSegmentsC: 
					currentResult.appendNonNil(polyLine(ith(1)).toLineSegmentVector(lineSegmentType));
				DirectedLineSegmentsC: 
					currentResult.appendNonNil(polyLine(ith(1)).toLineSegmentVector(directedLineSegmentType));
				ConvexHullC: 
					currentResult.append(polyLine(ith(1)).convexHullLinear);
			end;
	end;

function polyLine.pointCover (n: longInt): obj;
	var
		lsv: lineSegmentVector;
		v: vector;
	begin
		if length > 1 then begin
			lsv := toLineSegmentVector(lineSegmentType);
			v := vector(lsv.pointCover(n));
			if v <> nil then
				v.eliminateDuplicates;
			pointCover := v;
			lsv.freeAll;
		end
		else if length = 1 then
			pointCover := elements^^[1].pointCover(n)
		else
			pointCover := nil;
	end;

function polyLine.randomPointCover (n: longInt): obj;
	var
		partialPerimeters: homogenVector; (* of realNumber *)
		totalPerimeter, position, d: extended;
		i: longInt;
		index: seqIndex;
		result: pointVector;
		p, q, r: point2d;

	procedure doDistance (i, j: seqIndex);
		var
			x: realNumber;
		begin
			new(x);
			failNil(Handle(x));
			totalPerimeter := totalPerimeter + point2d(elements^^[i]).distanceX(point2d(elements^^[j]));
			x.value := totalPerimeter;
			partialPerimeters.append(x);
		end;

	begin
		if length > 0 then begin
			partialPerimeters := newHomogenVector(length);
			totalPerimeter := 0;
			for i := 1 to length - 1 do
				doDistance(i, i + 1);
			if member(self, polygon2d) then begin
				doDistance(length, 1);
				n := length * n;
			end
			else
				n := (length - 1) * n;
			result := newPointVector(n);
			for i := 1 to n do begin
				position := randomReal * totalPerimeter;
				index := binaryFind(partialPerimeters, position);
				p := point2d(elements^^[index]);
				if index = length then
					q := point2d(elements^^[1])
				else
					q := point2d(elements^^[index + 1]);
				if index = 1 then
					d := position
				else
					d := position - realNumber(partialPerimeters.elements^^[index - 1]).value;
				d := d / p.distanceX(q);
				r := point2d(p.duplicate);
				r.setXY(p.getX + d * (q.getX - p.getX), p.getY + d * (q.getY - p.getY));
				result.append(r);
			end;
			partialPerimeters.freeAll;
		end
		else if length = 1 then begin
			result := newPointVector(n);
			for i := 1 to n do
				result.append(elements^^[1].duplicate);
		end
		else
			result := nil;
		randomPointCover := result;
	end;

function polyLine.getType: objectType;
	begin
		getType := polyLineType;
	end;

function polyLine.getTextType: Str255;
	begin
		getTextType := concat('Polyline', getTextBaseType);
	end;

function polyLine.memberType (o: obj): boolean;
	begin
		memberType := member(o, polyLine) & memberBaseType(o);
	end;

procedure polyLine.windowInput (px, py: integer);
	var
		ls: lineSegment;
		l: list;
		e: elementH;
	begin
		new(ls);
		failNil(Handle(ls));
		ls.init;
		ls.q.setX(px);
		ls.q.setY(py);
		l := newEmptyList;
		repeat
			while not Button do
				writeCursorCoordinates(false);
			l.append(ls.q.duplicate);
			ls.free;
			new(ls);
			failNil(Handle(ls));
			with point2d(l.last^^.value) do
{$IFC THINK_PASCAL }
				ls.windowInput(round(getX), round(getY));
{$ELSEC }
				ls.windowInput(roundLong(getX), roundLong(getY));
{$ENDC }
			ls.windowOutput;
		until ls.q.equalPoint(ls.p);
		while Button do
			writeCursorCoordinates(false);
		FlushEvents(mDownMask + mUpMask, 0);
		ls.windowOutput;
		e := l.first;
		while e^^.next <> nil do begin
			point2d(e^^.value).drawSimpleLine(point2d(e^^.next^^.value));
			e := e^^.next;
		end;
		initFromList(l);
		l.free;
		ls.free;
	end;

procedure polyLine.drawAdjacentEdges (i: seqIndex1; highlight: boolean);
	begin
		if i > 1 then
			point2d(elements^^[i - 1]).drawLineHighlight(point2d(elements^^[i]), highlight);
		if i < length then
			point2d(elements^^[i]).drawLineHighlight(point2d(elements^^[i + 1]), highlight);
	end;

function polyLine.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	var
		i: seqIndex0;
		p: point2d;

	function pointIsSelected (p: point2d): boolean;
		begin
			pointIsSelected := p.isSelected(where);
		end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
	procedure drawAdjacentEdgesP (o: obj);
		begin
			drawAdjacentEdges(i, highlight);
		end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

	begin (* polyLine.dragVertex *)
		dragVertex := false;
		if length > 1 then begin
			i := exists(pointIsSelected);
			if i > 0 then begin
				dragVertex := true;
				p := point2d(elements^^[i]);
				p.interactiveOutput;
				p.dragObject(drawAdjacentEdgesP, drawAdjacentEdgesP, mouse);
				p.interactiveOutput;
			end;
		end;
	end;

procedure polyLine.interactiveOutput;
	begin
		if length = 1 then
			inherited interactiveOutput
		else
			drawClippedPolyLine(self, member(self, polygon2d) & (length > 2))
	end;

procedure polyLine.interactiveHighlight;
	begin
		if length = 1 then
			inherited interactiveHighlight
		else begin
			PenSize(3, 3);
			interactiveOutput;
			PenSize(1, 1);
		end;
	end;

function polyLine.isSelected (where: obj): boolean;
	var
		l: lineSegment;
		found: boolean;
		i: seqIndex1;
	begin
		if length > 1 then begin
			new(l);
			failNil(Handle(l));
			l.q := point2d(elements^^[1]);
			found := false;
			i := 2;
			while not found & (i <= length) do begin
				l.p := l.q;
				l.q := point2d(elements^^[i]);
				i := i + 1;
				found := l.isSelected(where);
			end;
			isSelected := found;
			dispose(l);
		end
		else if length = 1 then
			isSelected := elements^^[1].isSelected(where)
		else
			isSelected := false;
	end;

function polyLine.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, polyLine) & inherited equal(o);
	end;

function polyLine.perimeter: extended;
	var
		i: seqIndex0;
		s: extended;
	begin
		s := 0;
		for i := 2 to length do
			s := s + point2d(elements^^[i]).distanceLpX(point2d(elements^^[i - 1]));
		perimeter := s;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function polyLine.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 2 * length;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function polyLine.toLineSegmentVector (segmentType: objectType): lineSegmentVector;
	var
		lv: lineSegmentVector;
		l: lineSegment;
		i: seqIndex;
	begin
		if length < 2 then
			toLineSegmentVector := nil
		else begin
			currentVectorLength := length;
			new(lv);
			failNil(Handle(lv));
			lv.init;
			lv.length := length - 1;
			for i := 1 to length - 1 do begin
				l := lineSegment(newObj(segmentType));
				l.p := point2d(elements^^[i].duplicate);
				l.q := point2d(elements^^[i + 1].duplicate);
				lv.elements^^[i] := l;
			end;
			toLineSegmentVector := lv;
		end;
	end;

function polyLine.closestPairPoints (ask: boolean): lineSegment;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		closestPairPoints := inherited closestPairPoints(ask);
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.closestPairModified (ask: boolean): lineSegment;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		closestPairModified := inherited closestPairModified(ask);
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.closestPairHeuristic: lineSegment;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		closestPairHeuristic := inherited closestPairHeuristic;
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.closestPairProbabilistic (ask: boolean): lineSegment;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		closestPairProbabilistic := inherited closestPairProbabilistic(ask);
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.nearestNeighborBoxes (ask: boolean): lineSegmentVector;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		nearestNeighborBoxes := inherited nearestNeighborBoxes(ask);
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.nearestNeighborSweep (kind: neighborKind; ask: boolean): lineSegmentVector;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		nearestNeighborSweep := inherited nearestNeighborSweep(kind, ask);
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.nearestNeighborProjection (kind: neighborKind; useY: boolean): lineSegmentVector;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		nearestNeighborProjection := inherited nearestNeighborProjection(kind, useY);
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.closestPairNewHeuristic: lineSegment;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		closestPairNewHeuristic := inherited closestPairNewHeuristic;
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.nearestNeighborInSector (ask: boolean; cosLower, sinLower, cosUpper, sinUpper: extended): lineSegmentVector;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		nearestNeighborInSector := inherited nearestNeighborInSector(ask, cosLower, sinLower, cosUpper, sinUpper);
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.minimalDisk (verbose: boolean): circle;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		minimalDisk := inherited minimalDisk(verbose);
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;

function polyLine.simpleMinimalDisk: circle;
	var
		t: Handle;
	begin
		t := Handle(elements);
		if HandToHand(t) <> noErr then
			failNil(nil);
		simpleMinimalDisk := inherited simpleMinimalDisk;
{$IFC THINK_PASCAL }
		disposHandle(Handle(elements));
{$ELSEC }
		DisposeHandle(Handle(elements));
{$ENDC }
		elements := objArrayH(t);
	end;


end.
