(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit polygon2d;

interface

uses
	pointVector;

implementation

{$SETC AWindingNumber = TRUE }
{$SETC ATravelingSalesman = TRUE }

const
{$IFC VorlesungGeoAlg }
	PointInPolygonNewC = 0;
{$ENDC }

	WindingNumberC = 1;
	TSPOptimize3C = 3;
	TSPOptimizeBigC = 4;
	TSPOptimizeAllC = 5;
	TSPOptimizeAskC = 6;

	InterUnionC = 7;
	IntersectionC = 8;
	Intersection1diff2C = 9;
	Intersection2diff1C = 10;
	InterUnionAskC = 11;
	IntersectionAskC = 12;
	Intersection1diff2AskC = 13;
	Intersection2diff1AskC = 14;
	SimplifyC = 15;
	SimplifyAskC = 16;

	SelfIntersectC = 17;
	SimpleSelfIntersectC = 18;

	ClipC = 19;
	PointsInPolygonC = 20;
	PointsInPolygonAskC = 21;
	DecomposeC = 22;
	RevertC = 23;
	WindingNumberDynamicC = 24;
	IsConvexC = 25;
	PointInPolygonC = 26;
	PointInPolygonDynamicC = 27;
	InterUnionDynamicC = 28;
	IntersectionDynamicC = 29;
	Intersection1diff2DynamicC = 30;
	Intersection2diff1DynamicC = 31;


function lengthOk (p: polygon2d): boolean;
	begin
		lengthOk := p.length > 2;
	end;

function pipOk (p: polygon2d; h: homogenVector): boolean;
	begin
		pipOk := lengthOk(p) & member(p.elements^^[1], realPoint) & member(h.elements^^[1], realPoint);
	end;

function clipOk (p: polygon2d; l: lineSegment): boolean;
	begin
		clipOk := (p.length > 2) & sameType2(p.elements^^[1], l.p);
	end;

function compatiblePoints (a, b: obj): boolean;
	begin
		compatiblePoints := sameType2(a, b) | member(a, realPoint) & member(b, realPoint);
	end;

procedure polygon2d.describe;
	var
		isOk: boolean;
	begin
		with currentArguments do begin
			if (length = 1) & member(ith(1), polygon2d) & (polygon2d(ith(1)).length > 0) then
				menuEntry('Is Convex', IsConvexC);
			if (length = 2) & (member(ith(1), polygon2d) & (polygon2d(ith(1)).length > 0) & compatiblePoints(polygon2d(ith(1)).elements^^[1], ith(2)) | member(ith(2), polygon2d) & (polygon2d(ith(2)).length > 0) & compatiblePoints(ith(1), polygon2d(ith(2)).elements^^[1])) then begin
				menuEntry('Winding Number', WindingNumberC);
				menuEntry('Winding Number Dynamic', WindingNumberDynamicC);
				menuEntry('Point In Polygon', PointInPolygonC);
				menuEntry('Point In Polygon Dynamic', PointInPolygonDynamicC);
{$IFC VorlesungGeoAlg }
				menuEntry('Point In Polygon New', PointInPolygonNewC);
{$ENDC }
			end
			else if (length = 2) & (member(ith(1), polygon2d) & member(ith(2), lineSegment) & clipOk(polygon2d(ith(1)), lineSegment(ith(2))) | member(ith(2), polygon2d) & member(ith(1), lineSegment) & clipOk(polygon2d(ith(2)), lineSegment(ith(1)))) then begin
				menuEntry('Clip', ClipC);
				menuEntry('Decompose', DecomposeC);
			end
			else if (length = 2) & (member(ith(1), polygon2d) & member(ith(2), homogenVector) & pipOk(polygon2d(ith(1)), homogenVector(ith(2))) | member(ith(2), polygon2d) & member(ith(1), homogenVector) & pipOk(polygon2d(ith(2)), homogenVector(ith(1)))) then begin
				menuEntry('Points In Polygon', PointsInPolygonC);
				menuEntry('Points In Polygon Ask...', PointsInPolygonAskC);
			end;
			if (length > 0) & member(ith(1), polygon2d) then begin
				isOk := currentArgumentsSameType & currentArguments.sameProperty(lengthOk);
				if (length = 2) & isOk then begin
					beginSubMenu('Set Operations');
					menuEntry('Intersection', IntersectionC);
					menuEntry('Union', InterUnionC);
					menuEntry('1 Difference 2', Intersection1diff2C);
					menuEntry('2 Difference 1', Intersection2diff1C);
					endSubMenu;
					beginSubMenu('Set Operations Ask');
					menuEntry('Intersection...', IntersectionAskC);
					menuEntry('Union...', InterUnionAskC);
					menuEntry('1 Difference 2...', Intersection1diff2AskC);
					menuEntry('2 Difference 1...', Intersection2diff1AskC);
					endSubMenu;
					beginSubMenu('Set Operations Dynamic');
					menuEntry('Intersection', IntersectionDynamicC);
					menuEntry('Union', InterUnionDynamicC);
					menuEntry('1 Difference 2', Intersection1diff2DynamicC);
					menuEntry('2 Difference 1', Intersection2diff1DynamicC);
					endSubMenu;
				end
				else if (length > 2) & (length <= maxPolys) & isOk then begin
					beginSubMenu('Set Operations');
					menuEntry('Intersection', IntersectionC);
					menuEntry('Union', InterUnionC);
					endSubMenu;
					beginSubMenu('Set Operations Ask');
					menuEntry('Intersection...', IntersectionAskC);
					menuEntry('Union...', InterUnionAskC);
					endSubMenu;
					beginSubMenu('Set Operations Dynamic');
					menuEntry('Intersection...', IntersectionDynamicC);
					menuEntry('Union...', InterUnionDynamicC);
					endSubMenu;
				end
				else if length = 1 then begin
					beginSubMenu('Traveling Salesman Optimize');
					menuEntry('3', TSPOptimize3C);
					menuEntry('Big', TSPOptimizeBigC);
					menuEntry('All', TSPOptimizeAllC);
					menuEntry('Ask...', TSPOptimizeAskC);
					endSubMenu;
					beginSubMenu('Self Intersect');
					menuEntry('Sophisticated', SelfIntersectC);
					menuEntry('Simple', SimpleSelfIntersectC);
					endSubMenu;
					if polygon2d(ith(1)).length > 2 then begin
						beginSubMenu('Set Operations');
						menuEntry('Decompose', InterUnionC);
						menuEntry('Simplify', SimplifyC);
						menuEntry('Decompose Ask...', InterUnionAskC);
						menuEntry('Simplify Ask...', SimplifyAskC);
						endSubMenu;
					end;
				end;
			end;
		end;
	end;

procedure computeWindingNumber (p: polygon2d; t: point2d; var mp: markedRealPoint);
	var
		onBoundary: boolean;
		w: longInt;
	begin
		w := p.windingNumber(t, onBoundary);
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.x := t.getX;
		mp.y := t.getY;
		mp.mark.value := stringOf(w : 1, ' ', onBoundary);
	end;

procedure callWindingNumberDynamic (p: polygon2d; t: point2d);
	begin
		dynamicExecution2(p, t, computeWindingNumber, false, true);
	end;

function translateToText (c: integer): Str255;
	begin
		case c of
			pointIsOutSide: 
				translateToText := 'Out';
			pointIsOnBoundary: 
				translateToText := 'On';
			pointIsInside: 
				translateToText := 'In';
		end;
	end;

procedure computePointInPolygon (p: polygon2d; t: point2d; var mp: markedRealPoint);
	begin
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.x := t.getX;
		mp.y := t.getY;
		mp.mark.value := translateToText(p.pointInPolygon(t));
	end;

procedure callPointInPolygonDynamic (p: polygon2d; t: point2d);
	begin
		dynamicExecution2(p, t, computePointInPolygon, false, false);
	end;

function computeResultInterUnion (l: list): obj;
	var
		result: list;

	function unionZoneQ (s: polSet): boolean;
		begin
			unionZoneQ := s <> [];
		end;

	begin (* computeResultInterUnion *)
		result := newEmptyList;
		polygon2d(l.ith(1)).intersection(l, unionZoneQ, result, false, nil);
		computeResultInterUnion := result;
	end;

function computeResultIntersection (l: list): obj;
	var
		result: list;
		refSet: polSet;

	function intZoneQ (s: polSet): boolean;
		begin
			intZoneQ := s = refSet;
		end;

	begin (* computeResultIntersection *)
		result := newEmptyList;
		refSet := [0..polRange(l.length - 1)];
		polygon2d(l.ith(1)).intersection(l, intZoneQ, result, false, nil);
		computeResultIntersection := result;
	end;

function computeResultIntersection1Diff2 (l: list): obj;
	var
		result: list;
		refSet: polSet;

	function intZoneQ (s: polSet): boolean;
		begin
			intZoneQ := s = refSet;
		end;

	begin (* computeResultIntersection1Diff2 *)
		result := newEmptyList;
		refSet := [0];
		polygon2d(l.ith(1)).intersection(l, intZoneQ, result, false, nil);
		computeResultIntersection1Diff2 := result;
	end;

function computeResultIntersection2Diff1 (l: list): obj;
	var
		result: list;
		refSet: polSet;

	function intZoneQ (s: polSet): boolean;
		begin
			intZoneQ := s = refSet;
		end;

	begin (* computeResultIntersection2Diff1 *)
		result := newEmptyList;
		refSet := [1];
		polygon2d(l.ith(1)).intersection(l, intZoneQ, result, false, nil);
		computeResultIntersection2Diff1 := result;
	end;

procedure polygon2d.execute (item: integer);

	function unionZoneQ (s: polSet): boolean;
		begin
			unionZoneQ := s <> [];
		end;

	procedure callWindingNumber (p: polygon2d; r: point2d);
		var
			q: markedRealPoint;
		begin
			computeWindingNumber(p, r, q);
			currentResult.append(q);
		end;

{$IFC VorlesungGeoAlg }
	procedure computePointInPolygonNew (p: polygon2d; t: point2d; var mp: markedRealPoint);
		begin
			new(mp);
			failNil(Handle(mp));
			mp.init;
			mp.x := t.getX;
			mp.y := t.getY;
			mp.mark.value := translateToText(p.pointInPolygonNew(t));
		end;

	procedure callPointInPolygonNew (p: polygon2d; r: point2d);
		var
			q: markedRealPoint;
		begin
			computePointInPolygonNew(p, r, q);
			currentResult.append(q);
		end;
{$ENDC }

	procedure callPointInPolygon (p: polygon2d; r: point2d);
		var
			q: markedRealPoint;
		begin
			computePointInPolygon(p, r, q);
			currentResult.append(q);
		end;

	function simpZoneQ (s: polSet): boolean;
		begin
			simpZoneQ := s = [];
		end;

	procedure callIntersect (i: integer);
		var
			refSet: polSet;

		function intZoneQ (s: polSet): boolean;
			begin
				intZoneQ := s = refSet;
			end;

		begin
			if (i = InterUnionC) | (i = InterUnionAskC) then
				intersection(currentArguments, unionZoneQ, currentResult, i = InterUnionAskC, nil)
			else begin
				case i of
					IntersectionC, IntersectionAskC: 
						refSet := [0..polRange(currentArguments.length - 1)];
					Intersection1diff2C, Intersection1diff2AskC: 
						refSet := [0];
					Intersection2diff1C, Intersection2diff1AskC: 
						refSet := [1];
				end;
				intersection(currentArguments, intZoneQ, currentResult, i >= IntersectionAskC, nil);
			end;
		end;

	function callIsConvex (poly: polygon2d): markedRealPoint;
		var
			result: markedRealPoint;
			resultString: Str255;
			markIndex: seqIndex1;
		begin
			if poly.length > 0 then begin
				new(result);
				failNil(Handle(result));
				result.init;
				case poly.isConvex(markIndex) of
					polygonIsConvexMinus: 
						resultString := 'Left Convex';
					polygonIsConvexDegenerate: 
						resultString := 'Convex Degenerate';
					polygonIsConvexPlus: 
						resultString := 'Right Convex';
					polygonIsNotConvex: 
						resultString := 'Not Convex';
					polygonIsNotConvexDegenerate: 
						resultString := 'Not Convex Degenerate';
				end;
				if markIndex > poly.length then
					markIndex := 1;
				result.initLocationMark(poly.elements^^[markIndex], resultString, true);
				callIsConvex := result;
			end
			else
				callIsConvex := nil;
		end;

	procedure callClip (poly: polygon2d; lS: lineSegment; result: list);
		var
			p: polygon2d;
		begin
			p := poly.clip(lS);
			if p.length > 2 then
				result.append(p)
			else
				p.freeAll;
		end;

	function callPointsInPolygon (poly: polygon2d; pts: homogenVector; ask: boolean): pointVector;
		var
			result: pointVector;
			l, res: list;

		procedure createVector (p: realPoint);
			var
				mp: markedRealPoint;
			begin
				new(mp);
				failNil(Handle(mp));
				mp.init;
				mp.mark.value := '';
				mp.x := p.x;
				mp.y := p.y;
				result.append(mp);
			end;

		begin (* callPointsInPolygon *)
			result := newPointVector(pts.length);
			pts.forAll(createVector);
			l := newEmptyList;
			l.append(poly);
			res := newEmptyList;
			intersection(l, unionZoneQ, res, ask, result);
			if (res = nil) | (res.length = 0) then begin
				result.freeAll;
				callPointsInPolygon := nil;
			end
			else begin
				callPointsInPolygon := result;
				res.freeAll;
			end;
			l.free;
		end;

	begin
		with currentArguments do
			case item of
				InterUnionDynamicC: 
					dynamicExecutionHighlightShowType(computeResultInterUnion);
				IntersectionDynamicC: 
					dynamicExecutionHighlightShowType(computeResultIntersection);
				Intersection1diff2DynamicC: 
					dynamicExecutionHighlightShowType(computeResultIntersection1Diff2);
				Intersection2diff1DynamicC: 
					dynamicExecutionHighlightShowType(computeResultIntersection2Diff1);
				IsConvexC: 
					currentResult.appendNonNil(callIsConvex(polygon2d(ith(1))));
				ClipC: 
					if member(ith(1), polygon2d) then
						callClip(polygon2d(ith(1)), lineSegment(ith(2)), currentResult)
					else
						callClip(polygon2d(ith(2)), lineSegment(ith(1)), currentResult);
				DecomposeC: 
					if member(ith(1), polygon2d) then
						currentResult.appendNonNil(polygon2d(ith(1)).decompose(lineSegment(ith(2))))
					else
						currentResult.appendNonNil(polygon2d(ith(2)).decompose(lineSegment(ith(1))));
				WindingNumberC: 
					if member(ith(1), polygon2d) then
						callWindingNumber(polygon2d(ith(1)), point2d(ith(2)))
					else
						callWindingNumber(polygon2d(ith(2)), point2d(ith(1)));
{$IFC VorlesungGeoAlg }
				PointInPolygonNewC: 
					if member(ith(1), polygon2d) then
						callPointInPolygonNew(polygon2d(ith(1)), point2d(ith(2)))
					else
						callPointInPolygonNew(polygon2d(ith(2)), point2d(ith(1)));
{$ENDC }
				PointInPolygonC: 
					if member(ith(1), polygon2d) then
						callPointInPolygon(polygon2d(ith(1)), point2d(ith(2)))
					else
						callPointInPolygon(polygon2d(ith(2)), point2d(ith(1)));
				WindingNumberDynamicC: 
					if member(ith(1), polygon2d) then
						callWindingNumberDynamic(polygon2d(ith(1)), point2d(ith(2)))
					else
						callWindingNumberDynamic(polygon2d(ith(2)), point2d(ith(1)));
				PointInPolygonDynamicC: 
					if member(ith(1), polygon2d) then
						callPointInPolygonDynamic(polygon2d(ith(1)), point2d(ith(2)))
					else
						callPointInPolygonDynamic(polygon2d(ith(2)), point2d(ith(1)));
				TSPOptimize3C, TSPOptimizeBigC, TSPOptimizeAllC: 
					currentResult.appendNonNil(polygon2d(ith(1)).TSPOptimize(item, false));
				TSPOptimizeAskC: 
					currentResult.appendNonNil(polygon2d(ith(1)).TSPOptimize(0, true));
				InterUnionC..Intersection2diff1AskC: 
					callIntersect(item);
				SelfIntersectC: 
					currentResult.appendNonNil(polygon2d(ith(1)).selfIntersect);
				SimpleSelfIntersectC: 
					currentResult.appendNonNil(polygon2d(ith(1)).simpleSelfIntersect);
				SimplifyAskC, SimplifyC: 
					intersection(currentArguments, simpZoneQ, currentResult, item = SimplifyAskC, nil);
				PointsInPolygonC, PointsInPolygonAskC: 
					if member(ith(1), polygon2d) & lengthOk(polygon2d(ith(1))) then
						currentResult.appendNonNil(callPointsInPolygon(polygon2d(ith(1)), homogenVector(ith(2)), item = PointsInPolygonAskC))
					else
						currentResult.appendNonNil(callPointsInPolygon(polygon2d(ith(2)), homogenVector(ith(1)), item = PointsInPolygonAskC));
			end;
	end;

function polygon2d.isSelected (where: obj): boolean;
	var
		l: lineSegment;
	begin
		if inherited isSelected(where) then
			isSelected := true
		else if length > 2 then begin
			new(l);
			failNil(Handle(l));
			l.p := point2d(elements^^[length]);
			l.q := point2d(elements^^[1]);
			isSelected := l.isSelected(where);
			dispose(l);
		end
		else
			isSelected := false;
	end;

procedure polygon2d.drawAdjacentEdges (i: seqIndex1; highlight: boolean);
	begin
		if length <= 2 then begin
			interactiveOutput;
			if highlight then
				interactiveHighlight;
		end
		else begin
			if i > 1 then
				point2d(elements^^[i - 1]).drawLineHighlight(point2d(elements^^[i]), highlight);
			if i < length then
				point2d(elements^^[i]).drawLineHighlight(point2d(elements^^[i + 1]), highlight);
			if (i = length) | (i = 1) then
				point2d(elements^^[length]).drawLineHighlight(point2d(elements^^[1]), highlight);
		end;
	end;

function polygon2d.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, polygon2d) & inherited equal(o);
	end;

function polygon2d.toLineSegmentVector (segmentType: objectType): lineSegmentVector;
	var
		lv: lineSegmentVector;
		l: lineSegment;
	begin
		lv := inherited toLineSegmentVector(segmentType);
		if length > 2 then begin
			l := lineSegment(newObj(segmentType));
			l.p := point2d(elements^^[length].duplicate);
			l.q := point2d(elements^^[1].duplicate);
			lv.append(l);
		end;
		toLineSegmentVector := lv;
	end;


function edgeLength (poly: polygon2d; index: seqIndex): real;
	begin
		edgeLength := point2d(poly.elements^^[(index - 1) mod poly.length + 1]).distanceX(point2d(poly.elements^^[index mod poly.length + 1]));
	end;

function pointDistance (poly: polygon2d; index1, index2: seqIndex): real;
	begin
		pointDistance := point2d(poly.elements^^[(index1 - 1) mod poly.length + 1]).distanceX(point2d(poly.elements^^[(index2 - 1) mod poly.length + 1]));
	end;

procedure copyNElements (poly: polygon2d; from: seqIndex; n: longInt; save: polygon2d);
	var
		i: longInt;
	begin
		for i := 1 to n do
			save.elements^^[i] := poly.elements^^[(from + i - 1) mod poly.length + 1];
	end;

procedure relocate (poly: polygon2d; whereFrom: seqIndex; n: longInt; whereTo: seqIndex);
(* whereFrom is last element not to be relocated , *)
(*whereTo is element after which sequence is to be inserted *)
	var
		save: polygon2d;
		i, j: seqIndex;
	begin
		save := newPolygon2d(n);
		copyNElements(poly, whereFrom, n, save);
		if (whereFrom < whereTo) then begin
			i := whereFrom + 1;
			for j := whereFrom + n + 1 to whereTo do begin
				poly.elements^^[i] := poly.elements^^[j];
				i := i + 1;
			end;
			whereTo := whereTo - n;
		end
		else (* whereFrom > whereTo *)
			begin
			if ((whereFrom + n) <= poly.length) then begin
				i := whereFrom;
				for j := whereFrom + n downto whereTo + n + 1 do begin
					poly.elements^^[j] := poly.elements^^[i];
					i := i - 1;
				end;
			end
			else begin
				i := 1;
				for j := (whereFrom + n) mod poly.length + 1 to whereTo do begin
					poly.elements^^[i] := poly.elements^^[j];
					i := i + 1;
				end;
				j := poly.length;
				for i := whereFrom downto whereTo + 1 do begin
					poly.elements^^[j] := poly.elements^^[i];
					j := j - 1;
				end;
				whereTo := whereTo - (whereFrom + n) mod poly.length;
			end;
		end;
		for i := 1 to n do
			poly.elements^^[(whereTo + i - 1) mod poly.length + 1] := save.elements^^[n + 1 - i];
		save.free;
	end;

	var
		globalLimit: longInt;

function testFirst (x: extended): boolean;
	begin
		testFirst := (1 <= x) & (x <= globalLimit) & (trunc(x) = x);
	end;

function testLast (x: extended): boolean;
	begin
		testLast := (1 <= x) & (x <= optimizeFirst) & (trunc(x) = x);
	end;

function polygon2d.TSPOptimize (whichN: longInt; ask: boolean): polygon2d; (* mostly by Martin Mueller *)
	var
		poly: polygon2d;
		i, j, ipn: seqIndex;
		n, first, last: longInt;
		iLen, ipnLen: real;
		unchanged: boolean;
		line1, line2, line3: lineSegment;
		seq: vector;
		pt: markedRealPoint;

	function validParameters (var first, last: longInt; limit: longInt): boolean;
		begin
			globalLimit := limit;
			if optimizeFirst > limit then
				optimizeFirst := limit;
			if getUserParameter2(concat('Try Longest Partial Tour (<= ', stringOf(limit : 1), ')'), 'Try Shortest Partial Tour', optimizeFirst, optimizeLast, optimizeFirst, optimizeLast, @testFirst, @testLast) then begin
				validParameters := true;
{$IFC THINK_PASCAL }
				first := trunc(optimizeFirst);
				last := trunc(optimizeLast);
{$ELSEC }
				first := truncLong(optimizeFirst);
				last := truncLong(optimizeLast);
{$ENDC }
			end
			else
				validParameters := false;
		end;

	function sum3 (a, b, c: real): real;
		begin
			if a < b then (* a < b *)
				if b < c then (* a < b, b < c *)
					sum3 := a + b + c
				else if a < c then (* a < b, c < b, a < c *)
					sum3 := a + c + b
				else (* a < b, c < b, c < a *)
					sum3 := c + a + b
			else if a < c then (* b < a, a < c *)
				sum3 := b + a + c
			else if b < c then (* b < a, c < a, b < c *)
				sum3 := b + c + a
			else (* b < a, c < a, c < b *)
				sum3 := c + b + a;
		end;

{$IFC AnimationEnabled }
		label
			9999;
{$ENDC }

	begin
		if ask & not validParameters(first, last, (length + 1) div 2) then begin
			TSPOptimize := nil;
			exit(TSPOptimize);
		end;
{$IFC AnimationEnabled AND ATravelingSalesman }
		if animationFlag[ATravelingSalesmanItem] then
			interactiveHighlight;
{$ENDC }
		poly := polygon2d(duplicateAll);
		new(line1);
		failNil(Handle(line1));
		new(line2);
		failNil(Handle(line2));
		new(line3);
		failNil(Handle(line3));
		seq := newVector(3);
		seq.append(line1);
		seq.append(line2);
		seq.append(line3);
		pt := nil;

		case whichN of
			TSPOptimize3C:  begin
				first := 3;
				last := 1;
			end;
			TSPOptimizeBigC:  begin
				first := (poly.length + 1) div 2;
				last := 4;
			end;
			TSPOptimizeAllC:  begin
				first := (poly.length + 1) div 2;
				last := 1;
			end;
			otherwise
		end;

		for n := first downto last do
			repeat
				unchanged := true;
				ipn := n + 1;
				i := 1;
				while unchanged & (i <= poly.length) do
(* consider string i+1,...,i+n: links are i--i+1 & i+n--i+n+1 *)
					begin
					iLen := edgeLength(poly, i);
					ipnLen := edgeLength(poly, ipn);
					for j := 1 to poly.length do begin
{$IFC AnimationEnabled }
						if userAbort then begin
							poly.freeAll;
							poly := nil;
							goto 9999;
						end;
{$ENDC }
						if unchanged & (((ipn > i) & ((j + 1 < i) | (j > ipn))) | ((ipn < i) & (j > ipn) & (j + 1 < i))) then begin
(* consider edge j--j+1 *)
(* now compute length of new edges *)
							if sum3(pointDistance(poly, j, ipn), pointDistance(poly, j + 1, i + 1), pointDistance(poly, i, ipn + 1)) < sum3(iLen, ipnLen, edgeLength(poly, j)) then begin
(* insert string between j, j+1 & link i with i+n+1 *)
{$IFC AnimationEnabled AND ATravelingSalesman }
								if animationFlag[ATravelingSalesmanItem] then begin
(* delete old lines *)
									line1.p := realPoint(poly.elements^^[j]);
									line1.q := realPoint(poly.elements^^[j mod poly.length + 1]);
									line2.p := realPoint(poly.elements^^[i]);
									line2.q := realPoint(poly.elements^^[i mod poly.length + 1]);
									line3.p := realPoint(poly.elements^^[ipn]);
									line3.q := realPoint(poly.elements^^[ipn mod poly.length + 1]);
									seq.interactiveFlash;
									seq.interactiveOutput;
									line1.p := realPoint(poly.elements^^[j]);
									line1.q := realPoint(poly.elements^^[ipn]);
									line2.p := realPoint(poly.elements^^[i mod poly.length + 1]);
									line2.q := realPoint(poly.elements^^[j mod poly.length + 1]);
									line3.p := realPoint(poly.elements^^[i]);
									line3.q := realPoint(poly.elements^^[ipn mod poly.length + 1]);
									seq.interactiveFlash;
									seq.interactiveOutput;
									if pt <> nil then begin
										pt.interactiveOutput;
										pt.free;
									end;
								end;
{$ENDC }

								relocate(poly, i, n, j);
								unchanged := false;
{$IFC AnimationEnabled AND ATravelingSalesman }
								if animationFlag[ATravelingSalesmanItem] then begin
									pt := markedRealPoint(poly.perimeterMP);
									pt.interactiveOutput;
									waitForClick(animationFlag[ATravelingSalesmanItem]);
								end;
{$ENDC }
							end;
						end;
					end; (* for j...  if ... *)
					i := i + 1;
					ipn := ipn mod poly.length + 1;
				end; (* while unchanged... *)
			until unchanged;

{$IFC AnimationEnabled }
9999:
{$ENDC }
		seq.free;
		dispose(line1);
		dispose(line2);
		dispose(line3);
		if pt <> nil then
			pt.free;
		TSPOptimize := poly;
	end;

function polygon2d.getType: objectType;
	begin
		getType := polygon2dType;
	end;

function polygon2d.getTextType: Str255;
	begin
		getTextType := concat('Polygon', getTextBaseType);
	end;

function polygon2d.memberType (o: obj): boolean;
	begin
		memberType := member(o, polygon2d) & memberBaseType(o);
	end;

function polygon2d.perimeter: extended;
	begin
		if length > 2 then
			perimeter := point2d(elements^^[1]).distanceLpX(point2d(elements^^[length])) + inherited perimeter
		else
			perimeter := inherited perimeter;
	end;

function isXUp (p: polygon2d): boolean;
	var
		i: seqIndex1;
	begin
		i := 2;
		with p do begin
			while (i <= length) & (point2d(elements^^[i - 1]).getX < point2d(elements^^[i]).getX) do
				i := i + 1;
			isXUp := i > length;
		end;
	end;

function isXDown (p: polygon2d): boolean;
	var
		i: seqIndex1;
	begin
		i := 2;
		with p do begin
			while (i <= length) & (point2d(elements^^[i - 1]).getX > point2d(elements^^[i]).getX) do
				i := i + 1;
			isXDown := i > length;
		end;
	end;

function isYUp (p: polygon2d): boolean;
	var
		i: seqIndex1;
	begin
		i := 2;
		with p do begin
			while (i <= length) & (point2d(elements^^[i - 1]).getY < point2d(elements^^[i]).getY) do
				i := i + 1;
			isYUp := i > length;
		end;
	end;

function isYDown (p: polygon2d): boolean;
	var
		i: seqIndex1;
	begin
		i := 2;
		with p do begin
			while (i <= length) & (point2d(elements^^[i - 1]).getY > point2d(elements^^[i]).getY) do
				i := i + 1;
			isYDown := i > length;
		end;
	end;

function polygon2d.isMonotone: boolean;
	var
		i: seqIndex1;
		s: signType;
		p, q: point2d;
	begin
		if length < 3 then
			isMonotone := false
		else if isXUp(self) | isXDown(self) | isYUp(self) | isYDown(self) then begin
			p := point2d(elements^^[1]);
			q := point2d(elements^^[length]);
			s := p.whichSideSign(point2d(elements^^[2]), q);
			i := 3;
			while (i < length) & (p.whichSideSign(point2d(elements^^[i]), q) = s) do
				i := i + 1;
			isMonotone := (i = length) & (s <> 0);
		end
		else
			isMonotone := false;
	end;

{$IFC AnimationEnabled AND AWindingNumber }
	var
		whichSideCount: longInt;
{$ENDC }

procedure updateX (left, right: extended; p, q, r: point2d; var c: integer);
	begin
		if q.getX < left then
			c := pointIsInside - c
		else if q.getX <= right then begin
{$IFC AnimationEnabled AND AWindingNumber }
			if animationFlag[AWindingNumberItem] then
				whichSideCount := whichSideCount + 1;
{$ENDC }
			case sign(q.whichSideX(p, r)) of
				-1: 
					c := pointIsInside - c;
				0: 
					c := pointIsOnBoundary;
				+1: 
			end;
		end;
	end;

procedure updateY (p, q, r: point2d; var c: integer);
	begin
		if p.getX < r.getX then
			updateX(p.getX, r.getX, p, q, r, c)
		else if r.getX < p.getX then
			updateX(r.getX, p.getX, p, q, r, c)
		else if q.getX < p.getX then
			c := pointIsInside - c
		else if q.getX = p.getX then
			c := pointIsOnBoundary;
	end;

function polygon2d.pointInPolygon (q: point2d): integer;
	var
		i: seqIndex1;
		p, r: point2d;
		c: integer;
	begin
{$IFC AnimationEnabled AND AWindingNumber }
		if animationFlag[AWindingNumberItem] then
			whichSideCount := 0;
{$ENDC }
		c := pointIsOutSide;
		if length > 2 then begin
			p := point2d(elements^^[1]);
			elements^^[length + 1] := p;
			for i := 2 to length + 1 do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					pointInPolygon := 0;
					exit(pointInPolygon);
				end;
{$ENDC }
				r := point2d(elements^^[i]);
				if (p.getY <= q.getY) & (q.getY < r.getY) then
					updateY(p, q, r, c)
				else if (r.getY <= q.getY) & (q.getY < p.getY) then
					updateY(r, q, p, c)
				else if (r.getY = q.getY) & (q.getY = p.getY) & ((p.getX <= q.getX) & (q.getX <= r.getX) | (r.getX <= q.getX) & (q.getX <= p.getX)) then
					c := pointIsOnBoundary;
				p := r;
			end;
		end;
		pointInPolygon := c;
{$IFC AnimationEnabled AND AWindingNumber }
		if animationFlag[AWindingNumberItem] then
			displayInfoString(StringOf(translateToText(c), ', Calls To WhichSide = ', whichSideCount : 1));
{$ENDC }
	end;

function polygon2d.windingNumber (r: point2d; var onBoundary: boolean): longInt;
	var
		wn: longInt;
		i: seqIndex1;
		p, q: point2d;
{$IFC AnimationEnabled AND AWindingNumber }
		l: lineSegment;
{$ENDC }
	begin
{$IFC AnimationEnabled AND AWindingNumber }
		new(l);
		failNil(Handle(l));
		if animationFlag[AWindingNumberItem] then
			interactiveHighlight;
{$ENDC }
		onBoundary := false;
		wn := 0;
		if length > 0 then begin
			p := point2d(elements^^[1]);
			elements^^[length + 1] := p;
			for i := 2 to length + 1 do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					windingNumber := 0;
					exit(windingNumber);
				end;
{$ENDC }
				q := point2d(elements^^[i]);
{$IFC AnimationEnabled AND AWindingNumber }
				if animationFlag[AWindingNumberItem] then begin
					p.interactiveHighlight;
					q.interactiveHighlight;
					waitForClick(animationFlag[AWindingNumberItem]);
					p.interactiveHighlight;
					q.interactiveHighlight;
				end;
{$ENDC }
				case sign(r.whichSideX(p, q)) of
					-1: 
						if (p.getX <= r.getX) & (r.getX < q.getX) then begin
							wn := wn - 1;
{$IFC AnimationEnabled AND AWindingNumber }
							if animationFlag[AWindingNumberItem] then begin
								l.p := p;
								l.q := q;
								l.interactiveFlash;
								waitForClick(animationFlag[AWindingNumberItem]);
							end;
{$ENDC }
						end;
					0: 
						onBoundary := onBoundary | inRange(p.getX, q.getX, r.getX) & inRange(p.getY, q.getY, r.getY);
					+1: 
						if (q.getX <= r.getX) & (r.getX < p.getX) then begin
							wn := wn + 1;
{$IFC AnimationEnabled AND AWindingNumber }
							if animationFlag[AWindingNumberItem] then begin
								l.p := p;
								l.q := q;
								l.interactiveFlash;
								waitForClick(animationFlag[AWindingNumberItem]);
							end;
{$ENDC }
						end;
				end;
				p := q;
			end;
		end;
		windingNumber := wn;
{$IFC AnimationEnabled AND AWindingNumber }
		Dispose(l);
{$ENDC }
	end; (* windingNumber *)


end.
