(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit polygonClip;

interface

uses
	pointVector;

implementation

{$SETC APolygonDecompose=TRUE }

function polygon2d.clip (lS: lineSegment): polygon2d; (* by Adrian Bruengger *)

	var
		poly: polygon2d;
		polyList: list;
		j: seqIndex;
		s, p, f, i: point2d;
		svis, pvis, fvis: boolean;

	procedure intSec (p, q, r, s: point2d; var i: point2d);
		var
			ldx, kdx, ldy, kdy, lkdx, lkdy, ldxkdy, ldykdx, det: extended;
		begin
			i := point2d(lS.p.duplicateAll);
			ldx := p.getX - q.getX;
			ldy := p.getY - q.getY;
			kdx := r.getX - s.getX;
			kdy := r.getY - s.getY;
			lkdx := p.getX - r.getX;
			lkdy := p.getY - r.getY;
			ldykdx := ldy * kdx;
			ldxkdy := ldx * kdy;
			det := ldxkdy - ldykdx;
			i.setX((kdx * ldx * lkdy + r.getX * ldxkdy - p.getX * ldykdx) / det);
			i.setY((kdy * ldy * lkdx + r.getY * ldykdx - p.getY * ldxkdy) / (-det));
		end;

	begin
		polyList := newEmptyList;
		for j := 1 to length do begin
			p := point2d(elements^^[j]);
			pvis := lS.whichSideX(p) < 0;
			if j = 1 then begin
				f := p;
				fvis := pvis;
			end
			else begin
				if svis <> pvis then begin
					intSec(lS.p, lS.q, s, p, i);
					polyList.append(i);
				end;
			end;
			s := p;
			svis := pvis;
			if svis then
				polyList.append(s.duplicateAll);
		end;
		if svis <> fvis then begin
			intSec(lS.p, lS.q, s, f, i);
			polyList.append(i);
		end;
		new(poly);
		failNil(Handle(poly));
		poly.initFromList(polyList);
		polyList.free;
		clip := poly;
	end;

	type
		decompIntersection = object(obj)
				where: point2d;
				successor: decompIntersection;
				pol, myIndex: seqIndex;
				visited: array[boolean] of boolean;
				procedure free;
				override;
{$IFC AnimationEnabled AND APolygonDecompose }
				procedure interactiveOutput;
				override;
				procedure interactiveHighlight;
				override;
{$ENDC }
			end;

{$IFC AnimationEnabled AND APolygonDecompose }
procedure decompIntersection.interactiveOutput;
	begin
		where.interactiveOutput;
	end;

procedure decompIntersection.interactiveHighlight;
	begin
		where.interactiveHighlight;
	end;
{$ENDC }

procedure decompIntersection.free;
	begin
		where.free;
		shallowFree;
	end;

function diLessThan (a, b: decompIntersection): boolean;
	begin
		diLessThan := pointLessThanXY(a.where, b.where);
	end;

procedure appendPoly (var poList: list; pvList: list);
	var
		po: polygon2d;
	begin
		if poList.length > 2 then begin
			new(po);
			failNil(Handle(po));
			po.initFromList(poList);
{$IFC AnimationEnabled AND APolygonDecompose }
			if animationFlag[APolygonIntersectionItem] then begin
				po.interactiveFlash;
				waitForClick(animationFlag[APolygonIntersectionItem]);
			end;
{$ENDC }
			poList.free;
			pvList.append(po);
		end
		else
			poList.freeAll;
	end;

procedure appendPoint (p: point2d; poList: list);
	begin
{$IFC AnimationEnabled AND APolygonDecompose }
		if animationFlag[APolygonIntersectionItem] then begin
			p.interactiveFlash;
			waitForClick(animationFlag[APolygonIntersectionItem]);
		end;
{$ENDC }
		if (poList.length = 0) | not p.equalPoint(point2d(poList.last^^.value)) then
			poList.append(p.duplicate);
	end;

procedure handleIntersection (i: seqIndex; direction: integer; isDown: boolean; intersections: vector; pvList: list; thisPolygon: polygon2d);
	var
		poList: list;
		j, k: seqIndex0;
		first, current: decompIntersection;
	begin
		current := decompIntersection(intersections.elements^^[i]);
		if not current.visited[isDown] then begin
			poList := newEmptyList;
			first := current;
			repeat
				appendPoint(current.where, poList);
				current.visited[isDown] := true;
				current.successor.visited[isDown] := true;
				j := current.pol;
				k := current.successor.pol;
				while j <> k do begin
					appendPoint(point2d(thisPolygon.elements^^[j]), poList);
					j := j + 1;
					if j > thisPolygon.length + 1 then
						j := j - thisPolygon.length;
				end;
				appendPoint(current.successor.where, poList);
				j := current.successor.myIndex + direction;
				assert((1 <= j) & (j <= intersections.length), 'range ok');
				current := decompIntersection(intersections.elements^^[j]);
			until current = first;
			appendPoly(poList, pvList);
		end;
	end;

function correctIntersection (l, clipper, result: lineSegment): boolean;
	begin
		if clipper.whichSideX(l.p) = 0 then
			if clipper.whichSideX(l.q) < 0 then begin
				correctIntersection := true;
				result.p.setX(l.p.getX);
				result.p.setY(l.p.getY);
			end
			else
				correctIntersection := false
		else if clipper.whichSideX(l.q) = 0 then
			if clipper.whichSideX(l.p) < 0 then begin
				correctIntersection := true;
				result.p.setX(l.q.getX);
				result.p.setY(l.q.getY);
			end
			else
				correctIntersection := false
		else
			correctIntersection := l.intersectLineX(clipper, result)
	end;

function polygon2d.decompose (clipper: lineSegment): vector;
	var
		l, result: lineSegment;
		intersections, pv, r: vector;
		i: seqIndex0;
		di: decompIntersection;
		pvList: list;
	begin
{$IFC AnimationEnabled AND APolygonDecompose }
		if animationFlag[APolygonIntersectionItem] then begin
			interactiveHighlight;
			clipper.interactiveHighlight;
		end;
{$ENDC }
		r := newVector(2);
		elements^^[length + 1] := elements^^[1];
		intersections := newVector(length);
		new(l);
		failNil(Handle(l));
		result := lineSegment(clipper.duplicate);
		for i := 1 to length do begin
			l.p := point2d(elements^^[i]);
			l.q := point2d(elements^^[i + 1]);
			if correctIntersection(l, clipper, result) then begin
				new(di);
				failNil(Handle(di));
				with di do begin
					where := point2d(result.p.duplicate);
					pol := i + 1;
					visited[true] := false;
					visited[false] := false;
				end;
				if intersections.length = 0 then
					intersections.append(di)
				else
					with decompIntersection(intersections.elements^^[intersections.length]) do
						if where.equalPoint(di.where) then begin
							free;
							di.free;
							intersections.length := intersections.length - 1;
						end
						else begin
							successor := di;
							intersections.append(di);
						end;
			end;
		end;
		result.freeAll;
		dispose(l);
		if intersections.length > 0 then
			di.successor := decompIntersection(intersections.elements^^[1]);
		intersections.sortAll(diLessThan);
{$IFC AnimationEnabled AND APolygonDecompose }
		if animationFlag[APolygonIntersectionItem] then begin
			intersections.interactiveFlash;
			intersections.interactiveHighlight;
			waitForClick(animationFlag[APolygonIntersectionItem]);
			intersections.interactiveHighlight;
		end;
{$ENDC }
		for i := 1 to intersections.length do
			with decompIntersection(intersections.elements^^[i]) do
				myIndex := i;

		pvList := newEmptyList;
		for i := 1 to intersections.length do
			handleIntersection(i, -1, true, intersections, pvList, self);
		new(pv);
		failNil(Handle(pv));
		pv.initFromList(pvList);
		pvList.free;
		r.append(pv);

		pvList := newEmptyList;
		for i := intersections.length downto 1 do
			handleIntersection(i, +1, false, intersections, pvList, self);
		new(pv);
		failNil(Handle(pv));
		pv.initFromList(pvList);
		pvList.free;
		r.append(pv);

		intersections.freeAll;
		decompose := r;
	end;

end.
