(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit probabilistic;

{$SETC AclosestPair = TRUE }

interface

uses
	pointVector;

implementation

const
	infinity = 10000;


function sizeOk (x: extended): boolean;
	begin
		sizeOk := (0 < x) & (x < 10000) & (trunc(x) = x);
	end;

function pointVector.closestPairProbabilistic (ask: boolean): lineSegment;
	type
		pair = record
				p, q: realPoint;
				delta: real;
			end;
	var
		l: lineSegment;
		delta2, logLength: real;
		limit23, limit49: seqIndex;
		p1, p: pair;
		hashtableSize: longInt;
		continue: boolean;
{$IFC AnimationEnabled AND AclosestPair }
		showMin: boolean;
{$ENDC }

{$IFC AnimationEnabled AND AclosestPair }
	procedure drawGrid (el: objArrayH; len: seqIndex; doubleWidth, dx, dy: real);
		var
			maxX, maxY, px, py: real;
			i: seqIndex;
			j, wx, wy: integer;
		begin
			maxX := 0;
			maxY := 0;
			for i := 1 to len do begin
				if realPoint(el^^[i]).x > maxX then
					maxX := realPoint(el^^[i]).x;
				if realPoint(el^^[i]).y > maxY then
					maxY := realPoint(el^^[i]).y;
			end;
{$IFC THINK_PASCAL }
			wx := trunc(maxX / doubleWidth) + 1;
			maxX := wx * doubleWidth + dx;
			wy := trunc(maxY / doubleWidth) + 1;
{$ELSEC }
			wx := truncLong(maxX / doubleWidth) + 1;
			maxX := wx * doubleWidth + dx;
			wy := truncLong(maxY / doubleWidth) + 1;
{$ENDC }
			maxY := wy * doubleWidth + dy;
			px := dx + 2 / currentFrame.mx;
			py := dy + 2 / currentFrame.mx;
			for j := 1 to wy do begin
				drawLineSegment(px, py, maxX, py);
				py := py + doubleWidth;
			end;
			drawLineSegment(px, py, maxX, py);
			px := dx + 2 / currentFrame.mx;
			py := dy + 2 / currentFrame.mx;
			for j := 1 to wx do begin
				drawLineSegment(px, py, px, maxY);
				px := px + doubleWidth;
			end;
			drawLineSegment(px, py, px, maxY);
		end;
{$ENDC }

	function fourGridHash (el: objArrayH; limit: seqIndex; doubleWidth: real): pair;
		var
			p: pair;

		function min (a, b: pair): pair;
			begin
				if (b.p = nil) | (a.p <> nil) & (a.delta < b.delta) then begin
					min := a;
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] & showMin & (b.p <> nil) & ((a.p <> b.p) | (a.q <> b.q)) then begin
						b.p.drawLine(b.q);
						waitForClick(animationFlag[AClosestPairItem]);
					end;
{$ENDC }
				end
				else begin
					min := b;
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] & showMin & (a.p <> nil) & ((a.p <> b.p) | (a.q <> b.q)) then begin
						a.p.drawLine(a.q);
						waitForClick(animationFlag[AClosestPairItem]);
					end;
{$ENDC }
				end;
			end;

		function oneGridHash (el: objArrayH; limit: seqIndex; doubleWidth, dx, dy: real): pair;
			var
				hashTable: homogenVector; (* of list of points *)
				i: seqIndex;
				p: pair;

			function hash (p: realPoint; doubleWidth, dx, dy: real): seqIndex;
				var
					x, y: longInt;
				begin
{$IFC THINK_PASCAL }
					x := trunc((p.x - dx) / doubleWidth);
					if x < 0 then
						x := 0;
					y := trunc((p.y - dy) / doubleWidth);
{$ELSEC }
					x := truncLong((p.x - dx) / doubleWidth);
					if x < 0 then
						x := 0;
					y := truncLong((p.y - dy) / doubleWidth);
{$ENDC }
					if y < 0 then
						y := 0;
					hash := ((x + y) * (x + y + 1) div 2 + y) mod hashtableSize + 1;
				end;

			function checkList (l: list): pair;
				var
					i, j: elementH;
					p: pair;
					q: realPoint;
{$IFC AnimationEnabled }
					k: longInt;
{$ENDC }
				begin
					if l.length < 2 then begin
						p.p := nil;
						p.delta := 0;
					end
					else begin
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] then
							l.interactiveHighlight;
{$ENDC }
						i := l.first;
						p.delta := infinity;
						while i <> nil do begin
							j := i^^.next;
							q := realPoint(i^^.value);
							while j <> nil do begin
{$IFC AnimationEnabled }
								if userAbort then begin
									for k := 1 to hashtableSize do
										hashTable.elements^^[k].free;
									hashTable.free;
									closestPairProbabilistic := nil;
									exit(closestPairProbabilistic);
								end;
{$ENDC }
								if q.distanceX(realPoint(j^^.value)) < p.delta then begin
{$IFC AnimationEnabled AND AclosestPair }
									if animationFlag[AClosestPairItem] & (p.delta <> infinity) then
										p.p.drawLine(p.q);
{$ENDC }
									p.delta := q.distanceX(realPoint(j^^.value));
									p.p := q;
									p.q := realPoint(j^^.value);
{$IFC AnimationEnabled AND AclosestPair }
									if animationFlag[AClosestPairItem] then begin
										p.p.drawLine(p.q);
										waitForClick(animationFlag[AClosestPairItem]);
									end;
{$ENDC }
								end;
								j := j^^.next;
							end;
							i := i^^.next;
						end;
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] then
							l.interactiveHighlight;
{$ENDC }
					end;
					checkList := p;
				end; (* of checkList *)

			begin (* of oneGridHash *)
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					showMin := true;
					drawGrid(el, length, doubleWidth, dx, dy);
				end;
{$ENDC }
				hashTable := newHomogenVector(hashtableSize);
				for i := 1 to hashtableSize do
					hashTable.elements^^[i] := newEmptyList;
				for i := 1 to limit do
					list(hashTable.elements^^[hash(realPoint(el^^[i]), doubleWidth, dx, dy)]).append(el^^[i]);
				p := checkList(list(hashTable.elements^^[1]));
				for i := 2 to hashtableSize do
					p := min(p, checkList(list(hashTable.elements^^[i])));
				for i := 1 to hashtableSize do
					hashTable.elements^^[i].free;
				hashTable.free;
				oneGridHash := p;
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					drawGrid(el, length, doubleWidth, dx, dy);
					if p.p <> nil then
						p.p.drawLine(p.q);
					showMin := false;
				end;
{$ENDC }
			end; (* of oneGridHash *)

		begin (* of fourGridHash *)
			p := oneGridHash(el, limit, doubleWidth, 0, 0);
			p := min(p, oneGridHash(el, limit, doubleWidth, doubleWidth / 2, 0));
			p := min(p, oneGridHash(el, limit, doubleWidth, 0, doubleWidth / 2));
			fourGridHash := min(p, oneGridHash(el, limit, doubleWidth, doubleWidth / 2, doubleWidth / 2));
		end; (* of fourGridHash *)

	function checkSequence (var el: objArray; limit: seqIndex): real;
		var
			i, j: seqIndex;
			delta: real;
		begin
			delta := infinity;
			for i := 1 to limit do
				for j := i + 1 to limit do
					if realPoint(el[i]).distanceX(realPoint(el[j])) < delta then
						delta := realPoint(el[i]).distanceX(realPoint(el[j]));
			checkSequence := delta;
		end;

	begin
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then
			interactiveHighlight;
{$ENDC }
		if ask then begin
			continue := getUserParameter1('Slots in the hashtable:', stdSize, stdSize, @sizeOk);
{$IFC THINK_PASCAL }
			hashtableSize := trunc(stdSize);
{$ELSEC }
			hashtableSize := truncLong(stdSize);
{$ENDC }
		end
		else begin
			hashtableSize := length div 4;
			if hashtableSize < 13 then
				hashtableSize := 13;
			continue := true;
		end;
		if continue then begin
			logLength := ln(length);
{$IFC THINK_PASCAL }
			limit23 := round(exp(2 * logLength / 3));
			limit49 := round(exp(4 * logLength / 9));
{$ELSEC }
			limit23 := roundLong(exp(2 * logLength / 3));
			limit49 := roundLong(exp(4 * logLength / 9));
{$ENDC }
			randomShuffle(limit23, length);
			randomShuffle(limit49, limit23);
			delta2 := checkSequence(elements^^, limit49);
			if delta2 = 0 then
				delta2 := 0.5;
			p1 := fourGridHash(elements, limit23, 2 * delta2);
			if p1.delta = 0 then begin
				new(l);
				failNil(Handle(l));
				l.p := realPoint(p1.p.duplicate);
				l.q := realPoint(p1.q.duplicate);
			end
			else begin
				p := fourGridHash(elements, length, 2 * p1.delta);
				new(l);
				failNil(Handle(l));
				l.p := realPoint(p.p.duplicate);
				l.q := realPoint(p.q.duplicate);
			end;
			closestPairProbabilistic := l;
		end
		else
			closestPairProbabilistic := nil;
	end;

end.
