(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit proximity;

{$SETC AANN = TRUE }
{$SETC AclosestPair = TRUE }

interface

uses
	pointVector;

type
	refPoint = object(obj)
			p: point2d;
			ref: reference; (* pointer into other table *)
		end;

function refPointLessThanXY (a, b: refPoint): relationObj;

function refPointLessThanYX (a, b: refPoint): relationObj;

procedure disposeRefPoint (r: refPoint);

procedure disposeOnlyRefPoint (r: refPoint);

function refPointGreaterThanXY (a, b: refPoint): relationObj;

implementation

function pointVector.initNearestNeighbors (kind: neighborKind): lineSegmentVector;
	var
		nearestNeighbors: lineSegmentVector;
		i: seqIndex;
		l: lineSegment;
		lDirected: directedLineSegment;
	begin
		currentVectorLength := length;
		new(nearestNeighbors);
		failNil(Handle(nearestNeighbors));
		nearestNeighbors.init;
		if kind = allNeighbors then begin
			new(lDirected);
			l := lDirected;
		end
		else
			new(l);
		failNil(Handle(l));
		for i := 1 to length do begin
			l.p := point2d(elements^^[i].duplicate);
			l.q := nil;
			nearestNeighbors.elements^^[i] := l.shallowDuplicate;
		end;
		l.shallowFree;
		nearestNeighbors.length := length;
		initNearestNeighbors := nearestNeighbors;
	end;

function pointVector.closestPairPoints (ask: boolean): lineSegment;
	type
		referenceArrayH = ^referenceArrayP;
		referenceArrayP = ^referenceArray;
		referenceArray = array[seqIndex] of reference;
	var
		dummy: priorityQueue;
		yTable: dictionary;
		squaredDelta: extended;
		currentPoint, closestLeft, closestRight: point2d;
		tail: seqIndex;
		current: seqIndex1;
		found, initialGuessing: boolean;
		references: referenceArrayH;
		l: lineSegment;
{$IFC AnimationEnabled AND AclosestPair }
		oldDelta: extended;

	procedure drawStatus (p: point2d; delta: extended);
		begin
			drawThickVertical(p.getX);
			drawVertical(p.getX - delta);
			drawCircle(p.getX, p.getY, delta - 1 / currentFrame.mx, 180, 180);
			drawLineSegment(p.getX, p.getY + delta, p.getX - delta, p.getY + delta);
			drawLineSegment(p.getX, p.getY - delta, p.getX - delta, p.getY - delta);
		end; (* drawStatus *)
{$ENDC }

	procedure traverse (check: reference; d: directionType);
		var
			newSquaredDelta: extended;
		begin
			repeat
				check := yTable.next(check, d);
				if check <> nil then begin
					newSquaredDelta := point2d(yTable.getObject(check)).distancePLpX(currentPoint);
					if newSquaredDelta < squaredDelta then begin
						squaredDelta := newSquaredDelta;
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] then begin
							waitForClick(animationFlag[AClosestPairItem]);
							closestRight.drawLine(closestLeft);
						end;
{$ENDC }
						closestLeft := point2d(yTable.getObject(check));
						closestRight := currentPoint;
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] then
							closestRight.drawLine(closestLeft);
{$ENDC }
					end;
				end;
			until (check = nil) | (currentPoint.dyPLpX(point2d(yTable.getObject(check))) > squaredDelta);
		end; (* traverse *)

       (* ------------------------------------------------------------------- *)


	procedure preprocessingStandard;
		begin
			closestLeft := point2d(elements^^[1]);
			closestRight := point2d(elements^^[2]);
			references^^[1] := yTable.insertObj(closestLeft, comparePointYX, found);
			references^^[2] := yTable.insertObj(closestRight, comparePointYX, found);
			if found then
				current := length + 1
			else begin
				squaredDelta := closestRight.distancePLpX(closestLeft);
				tail := 1;
				current := 3;
			end;

		end;

       (* ------------------------------------------------------------------- *)

	procedure preprocessingInitialGuessing;
		begin
			squaredDelta := 100000000;
			closestPairInitialGuessing(2, squaredDelta, closestLeft, closestRight);
			if squaredDelta = 0 then		(* 2 <=> 1-2,3-4,5-6,... *)
				current := length + 1			(* 1<=> 1-2,2-3,4-5,... *)
			else begin
				references^^[1] := yTable.insertObj(point2d(elements^^[1]), comparePointYX, found);
				tail := 1;						(* insert the first Point *)
				current := 2;
			end;
		end;

       (* ------------------------------------------------------------------- *)

	begin
		if length >= 2 then begin
			references := referenceArrayH(dynamicArray(length, sizeOf(reference)));	(* allocate memory for the refferenceArray *)

	(* initialize and select datastructure of the y-Table *)

			if ask then
				if getXY(0, false, dummy, length, yTable, true) then
					initialGuessing := stdInitialGuessing
				else begin
					closestPairPoints := nil;
					exit(closestPairPoints);
				end
			else begin
				initialGuessing := globalInitialGuessing;
				currentVectorLength := length;
				yTable := dictionary(newInitObj(avlTreeType));
			end;

	(*  preprocessing, like sorting and guessing of Delta *)
	(* --------------------------------------------- *)

			sortAll(pointLessThanXY); 		(* QuickSort the elements (increasing x) *)
               								(*    heapSort(pointGreaterThanXY);	--> use heapSort instead of QuickSort  *)

			if initialGuessing then
				preprocessingInitialGuessing	(* Procedurecall whithout used parameters ! sorry *)
			else
				preprocessingStandard;



{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then
				closestRight.drawLine(closestLeft);
{$ENDC }

			while current <= length do begin

{$IFC AnimationEnabled }
				if userAbort then begin
{$IFC THINK_PASCAL }
					disposHandle(Handle(references));
{$ELSEC }
					DisposeHandle(Handle(references));
{$ENDC }
					yTable.free;
					closestPairPoints := nil;
					exit(closestPairPoints);
				end;
{$ENDC }
				currentPoint := point2d(elements^^[current]);
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					oldDelta := sqrt(squaredDelta);
					drawStatus(currentPoint, oldDelta);
				end;
{$ENDC }
				while squaredDelta <= currentPoint.dxPLpX(point2d(elements^^[tail])) do begin
					yTable.delete(references^^[tail]);
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then begin
{$IFC THINK_PASCAL }
						PenPat(gray);
						elements^^[tail].interactiveOutput;
						PenPat(black);
{$ELSEC }
						PenPat(qd.gray);
						elements^^[tail].interactiveOutput;
						PenPat(qd.black);
{$ENDC }
					end;
{$ENDC }
					tail := tail + 1;
				end;
				references^^[current] := yTable.insertObj(currentPoint, comparePointYX, found);
				if found then begin
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then begin
						waitForClick(animationFlag[AClosestPairItem]);
						closestRight.drawLine(closestLeft);
					end;
{$ENDC }
					closestLeft := currentPoint;
					closestRight := point2d(yTable.getObject(references^^[current]));
					current := length;
				end
				else begin
					traverse(references^^[current], left);
					traverse(references^^[current], right);
				end;
				current := current + 1;
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					waitForClick(animationFlag[AClosestPairItem]);
					drawStatus(currentPoint, oldDelta);
				end;
{$ENDC }
			end;
{$IFC THINK_PASCAL }
			disposHandle(Handle(references));
{$ELSEC }
			DisposeHandle(Handle(references));
{$ENDC }
			if yTable.infoString <> '' then
				displayInfoString(concat('Y', yTable.infoString));
			yTable.free;
			new(l);
			failNil(Handle(l));
			l.p := point2d(closestLeft.duplicate);
			l.q := point2d(closestRight.duplicate);
			closestPairPoints := l;
		end;
	end;

function refPointLessThanXY (a, b: refPoint): relationObj;
	begin
		refPointLessThanXY := comparePointXY(a.p, b.p);
	end;

function refPointLessThanYX (a, b: refPoint): relationObj;
	begin
		refPointLessThanYX := comparePointYX(a.p, b.p);
	end;

procedure disposeRefPoint (r: refPoint);
	begin
		dispose(r.p);
		dispose(r);
	end;

procedure disposeOnlyRefPoint (r: refPoint);
	begin
		dispose(r);
	end;

function refPointGreaterThanXY (a, b: refPoint): relationObj;
	begin
		refPointGreaterThanXY := inverseComparePointXY(a.p, b.p);
	end;

function pointVector.closestPairHeuristic: lineSegment;
	var
		closestLeft, closestRight, currentPoint: point2d;
		delta: extended;
		i, j: seqIndex0;
		l: lineSegment;
	begin
		sortAll(pointLessThanXY);
		closestLeft := point2d(elements^^[1]);
		closestRight := point2d(elements^^[2]);
		delta := closestLeft.distancePLpX(closestRight);
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then
			closestLeft.drawLine(closestRight);
{$ENDC }
		for i := 3 to length do begin
			currentPoint := point2d(elements^^[i]);
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				drawThickVertical(currentPoint.getX);
				drawVertical(currentPoint.getX - sqrt(delta));
				waitForClick(animationFlag[AClosestPairItem]);
			end;
{$ENDC }
			j := i - 1;
			while (j > 0) & (currentPoint.dxPLpX(point2d(elements^^[j])) < delta) do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					closestPairHeuristic := nil;
					exit(closestPairHeuristic);
				end;
{$ENDC }
				if currentPoint.distancePLpX(point2d(elements^^[j])) < delta then begin
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then begin
						closestLeft.drawLine(closestRight);
						drawVertical(currentPoint.getX - sqrt(delta));
					end;
{$ENDC }
					closestLeft := point2d(elements^^[j]);
					delta := currentPoint.distancePLpX(closestLeft);
					closestRight := currentPoint;
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then begin
						closestLeft.drawLine(closestRight);
						drawVertical(currentPoint.getX - sqrt(delta));
						waitForClick(animationFlag[AClosestPairItem]);
					end;
{$ENDC }
				end;
				j := j - 1;
			end;
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				drawThickVertical(currentPoint.getX);
				drawVertical(currentPoint.getX - sqrt(delta));
			end;
{$ENDC }
		end;
		new(l);
		failNil(Handle(l));
		l.p := point2d(closestLeft.duplicate);
		l.q := point2d(closestRight.duplicate);
		closestPairHeuristic := l;
	end;

function pointVector.closestPairNewHeuristic: lineSegment;
	var
		sortedByY: pointVector;
		nearestLeft, nearestRight, currentPoint, t: point2d;
		dist: extended;
		i: seqIndex0;
		left, up, down: longInt;
		l: lineSegment;
		doneLeft, doneDown, doneUp: boolean;
	begin
		sortAll(pointLessThanXY);
		sortedByY := pointVector(duplicate);
		sortedByY.sortAll(pointLessThanYX);
		nearestLeft := point2d(elements^^[1]);
		nearestRight := point2d(elements^^[2]);
		dist := nearestLeft.distancePLpX(nearestRight);
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then begin
			nearestLeft.drawLine(nearestRight);
			waitForClick(animationFlag[AClosestPairItem]);
		end;
{$ENDC }
		for i := 3 to length do begin
			currentPoint := point2d(elements^^[i]);
			left := i;
			down := sortedByY.binarySearch(currentPoint, comparePointYX, 1, length);
			assert(down > 0, 'down > 0');
			up := down;
			doneLeft := false;
			doneDown := false;
			doneUp := false;
			repeat
{$IFC AnimationEnabled }
				if userAbort then begin
					sortedByY.free;
					closestPairNewHeuristic := nil;
					exit(closestPairNewHeuristic);
				end;
{$ENDC }
				left := left - 1;
				if left <= 0 then
					doneLeft := true
				else begin
					t := point2d(elements^^[left]);
					if (currentPoint.dxPLpX(t) >= dist) then
						doneLeft := true
					else if currentPoint.distancePLpX(t) < dist then begin
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] then begin
							nearestLeft.drawLine(nearestRight);
							currentPoint.drawLine(t);
							waitForClick(animationFlag[AClosestPairItem]);
						end;
{$ENDC }
						nearestLeft := t;
						nearestRight := currentPoint;
						dist := nearestLeft.distancePLpX(nearestRight);
					end;
				end;
				down := down - 1;
				if doneDown | (down <= 0) then
					doneDown := true
				else begin
					t := point2d(sortedByY.elements^^[down]);
					if currentPoint.dyPLpX(t) >= dist then
						doneDown := true
					else if currentPoint.distancePLpX(t) < dist then begin
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] then begin
							nearestLeft.drawLine(nearestRight);
							currentPoint.drawLine(t);
							waitForClick(animationFlag[AClosestPairItem]);
						end;
{$ENDC }
						nearestLeft := t;
						nearestRight := currentPoint;
						dist := nearestLeft.distancePLpX(nearestRight);
					end;
				end;
				up := up + 1;
				if doneUp | (up > length) then
					doneUp := true
				else begin
					t := point2d(sortedByY.elements^^[up]);
					if currentPoint.dyPLpX(t) >= dist then
						doneUp := true
					else if currentPoint.distancePLpX(t) < dist then begin
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] then begin
							nearestLeft.drawLine(nearestRight);
							currentPoint.drawLine(t);
							waitForClick(animationFlag[AClosestPairItem]);
						end;
{$ENDC }
						nearestLeft := t;
						nearestRight := currentPoint;
						dist := nearestLeft.distancePLpX(nearestRight);
					end;
				end;
			until doneLeft | (doneUp & doneDown);
		end;
		sortedByY.free;
		new(l);
		failNil(Handle(l));
		l.p := point2d(nearestLeft.duplicate);
		l.q := point2d(nearestRight.duplicate);
		closestPairNewHeuristic := l;
	end;

function pointVector.nearestNeighborProjection (kind: neighborKind; useY: boolean): lineSegmentVector;
	var
		sortedByY: pointVector;
		nearestNeighbors: lineSegmentVector;
		nearestNeighbor, currentPoint: point2d;
		dist: extended;
		i: seqIndex0;
		left, right, up, down: longInt;
		doneLeft, doneRight, doneDown, doneUp: boolean;

	procedure testPoint (t: point2d; isX: boolean; var done: boolean);

		function pointOk: boolean;
			var
				order: relationObj;
			begin
				order := comparePointXY(t, currentPoint);
				pointOk := (kind = leftNeighbors) & (order = lessThanObj) | (kind = rightNeighbors) & (order = greaterThanObj);
			end;

		begin
			if isX | (kind = allNeighbors) | pointOk then
				if nearestNeighbor = nil then begin
					nearestNeighbor := t;
					dist := currentPoint.distancePLpX(nearestNeighbor);
{$IFC AnimationEnabled AND AANN }
					if animationFlag[ANearestNeighborsItem] then begin
						currentPoint.drawLine(t);
						waitForClick(animationFlag[ANearestNeighborsItem]);
					end;
{$ENDC }
				end
				else begin
					if isX & (currentPoint.dxPLpX(t) >= dist) | not isX & (currentPoint.dyPLpX(t) >= dist) then
						done := true
					else if currentPoint.distancePLpX(t) < dist then begin
{$IFC AnimationEnabled AND AANN }
						if animationFlag[ANearestNeighborsItem] then begin
							currentPoint.drawLine(nearestNeighbor);
							currentPoint.drawLine(t);
							waitForClick(animationFlag[ANearestNeighborsItem]);
						end;
{$ENDC }
						nearestNeighbor := t;
						dist := currentPoint.distancePLpX(nearestNeighbor);
					end;
				end;
		end;

	begin
		if length > 1 then begin
			sortAll(pointLessThanXY);
			nearestNeighbors := initNearestNeighbors(kind);
			if useY then begin
				sortedByY := pointVector(duplicate);
				sortedByY.sortAll(pointLessThanYX);
			end;
			for i := 1 to length do begin
				currentPoint := point2d(elements^^[i]);
				left := i;
				right := i;
				if useY then begin
					down := sortedByY.binarySearch(currentPoint, comparePointYX, 1, length);
					assert(down > 0, 'down > 0');
					up := down;
					doneDown := false;
					doneUp := false;
				end
				else begin
					doneDown := true;
					doneUp := true;
				end;
				nearestNeighbor := nil;
				doneLeft := kind = rightNeighbors;
				doneRight := kind = leftNeighbors;
				repeat
{$IFC AnimationEnabled }
					if userAbort then begin
						if useY then
							sortedByY.free;
						nearestNeighbors.nilQ;
						nearestNeighbors.freeAll;
						nearestNeighborProjection := nil;
						exit(nearestNeighborProjection);
					end;
{$ENDC }
					left := left - 1;
					if doneLeft | (left <= 0) then
						doneLeft := true
					else
						testPoint(point2d(elements^^[left]), true, doneLeft);
					right := right + 1;
					if doneRight | (right > length) then
						doneRight := true
					else
						testPoint(point2d(elements^^[right]), true, doneRight);
					down := down - 1;
					if doneDown | (down <= 0) then
						doneDown := true
					else
						testPoint(point2d(sortedByY.elements^^[down]), false, doneDown);
					up := up + 1;
					if doneUp | (up > length) then
						doneUp := true
					else
						testPoint(point2d(sortedByY.elements^^[up]), false, doneUp);
				until (doneLeft & doneRight) | (useY & doneUp & doneDown);
				lineSegment(nearestNeighbors.elements^^[i]).q := nearestNeighbor;
			end;
			nearestNeighbors.finishNearestNeighbors(kind);
			if useY then
				sortedByY.free;
			nearestNeighborProjection := nearestNeighbors;
		end
		else
			nearestNeighborProjection := nil;
	end;

function pointVector.closestPairN2: lineSegment;
	var
		closestLeft, closestRight, current: point2d;
		l: lineSegment;
		i: seqIndex;
		j: seqIndex1;
		delta: extended;
	begin
		closestLeft := point2d(elements^^[1]);
		closestRight := point2d(elements^^[2]);
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then
			closestLeft.drawLine(closestRight);
{$ENDC }
		delta := closestLeft.distancePLpX(closestRight);
		for i := 1 to length do begin
			current := point2d(elements^^[i]);
			for j := i + 1 to length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					closestPairN2 := nil;
					exit(closestPairN2);
				end;
{$ENDC }
				if current.distancePLpX(point2d(elements^^[j])) < delta then begin
					delta := current.distancePLpX(point2d(elements^^[j]));
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then
						closestLeft.drawLine(closestRight);
{$ENDC }
					closestLeft := current;
					closestRight := point2d(elements^^[j]);
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then begin
						closestLeft.drawLine(closestRight);
						waitForClick(animationFlag[AClosestPairItem]);
					end;
{$ENDC }
				end;
			end;
		end;
		new(l);
		failNil(Handle(l));
		l.p := point2d(closestLeft.duplicate);
		l.q := point2d(closestRight.duplicate);
		closestPairN2 := l;
	end;

end.
