(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit realPoint;

interface

uses
{$IFC GFEnabled }
	GFValues, 
{$ENDC}
	basicGeometry;

function realIntegerPower (x: real; p: integer): real;

function realPower (x, p: real): real;

function realPthRoot (x, p: real): real;

implementation

function realMax (a, b: real): real;
	begin
		if a > b then
			realMax := a
		else
			realMax := b;
	end;

function realPoint.verticalDistance (o: obj): extended;
	begin
		verticalDistance := abs(y - realPoint(o).y);
	end;

function realPoint.distanceO (o: obj; var minP, minQ: obj): extended;
	begin
		realPoint(minP).x := x;
		realPoint(minP).y := y;
		realPoint(minQ).x := realPoint(o).x;
		realPoint(minQ).y := realPoint(o).y;
		distanceO := distanceX(point2d(o));
	end;

function realPoint.getX: extended;
	begin
		getX := x;
	end;

procedure realPoint.setX (newX: extended);
	begin
		x := newX;
	end;

function realPoint.getY: extended;
	begin
		getY := y;
	end;

procedure realPoint.setY (newY: extended);
	begin
		y := newY;
	end;

function realPoint.xLessThan (p: point2d): boolean;
	begin
		xLessThan := x < realPoint(p).x;
	end;

function realPoint.xEqual (p: point2d): boolean;
	begin
		xEqual := x = realPoint(p).x;
	end;

function realPoint.yLessThan (p: point2d): boolean;
	begin
		yLessThan := y < realPoint(p).y;
	end;

function realPoint.yEqual (p: point2d): boolean;
	begin
		yEqual := y = realPoint(p).y;
	end;

procedure realPoint.textFileInput;
	begin
		readTextReal(x);
		readTextReal(y);
	end;

procedure realPoint.textFileOutput;
	begin
		writeTextReal(x);
		writeTextReal(y);
	end;

procedure realPoint.fileInput;
	begin
		readReal(x);
		readReal(y);
	end;

procedure realPoint.fileOutput;
	begin
		writeReal(x);
		writeReal(y);
	end;

function realPoint.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, realPoint) & (realPoint(o).x = x) & (realPoint(o).y = y);
	end;

function realPoint.equalPoint (p: point2d): boolean;
	begin
		equalPoint := equal(p);
	end;

function realPoint.memberType (o: obj): boolean;
	begin
		memberType := member(o, realPoint);
	end;

function realPoint.middle (p: point2d): point2d;
	var
		t: realPoint;
	begin
		new(t);
		failNil(Handle(t));
		t.init;
		t.x := (x + realPoint(p).x) / 2;
		t.y := (y + realPoint(p).y) / 2;
		middle := t;
	end;

function realPoint.whichSideX (p, q: point2d): extended;
	var
		pr, qr: realPoint;
	begin
		pr := realPoint(p);
		qr := realPoint(q);
		whichSideX := (pr.x - x) * (pr.y - qr.y) - (pr.x - qr.x) * (pr.y - y);
	end;

function realPoint.crossProductX (p, q, r: point2d): extended;
	var
		pr, qr, rr: realPoint;
	begin
		pr := realPoint(p);
		qr := realPoint(q);
		rr := realPoint(r);
		crossProductX := (pr.x - x) * (rr.y - qr.y) - (pr.y - y) * (rr.x - qr.x);
	end;

function realPoint.scalarProductX (p: point2d): extended;  (* added by FB *)
	var
		pr: realPoint;
	begin
		pr := realPoint(p);
		scalarProductX := (pr.x * x) + (pr.y * y);
	end;

function realPoint.squaredDxX (p: point2d): extended;
	begin
		squaredDxX := sqr(realPoint(p).x - x);
	end;

function realPoint.squaredDyX (p: point2d): extended;
	begin
		squaredDyX := sqr(realPoint(p).y - y);
	end;

function realPoint.distanceX (p: point2d): extended;
	begin
		distanceX := sqrt(sqr(realPoint(p).x - x) + sqr(realPoint(p).y - y));
	end;

function realPoint.squaredDistanceX (p: point2d): extended;
	begin
		squaredDistanceX := sqr(realPoint(p).x - x) + sqr(realPoint(p).y - y);
	end;

function realIntegerPower (x: real; p: integer): real;
	var
		t: real;
	begin
		case p of
			0: 
				realIntegerPower := 1;
			1: 
				realIntegerPower := x;
			2: 
				realIntegerPower := sqr(x);
			3: 
				realIntegerPower := x * sqr(x);
			4: 
				realIntegerPower := sqr(sqr(x));
			otherwise begin
				t := sqr(realIntegerPower(x, p div 2));
				if p mod 2 = 0 then
					realIntegerPower := t
				else
					realIntegerPower := x * t;
			end;
		end;
	end;

function realPthRoot (x, p: real): real;
	begin
		if x > 0 then
			realPthRoot := exp(ln(x) / p)
		else if x = 0 then
			realPthRoot := 0
		else begin
			assert(false, 'x>=0');
			realPthRoot := 0;
		end;
	end;

function realPower (x, p: real): real;
	begin
		if x > 0 then
			realPower := exp(p * ln(x))
		else if x = 0 then
			realPower := 0
		else begin
			assert(false, 'x>=0');
			realPower := 0;
		end;
	end;

function realPoint.distanceLpX (p: point2d): extended;
	begin
		if lpMetric = 2 then
			distanceLpX := distanceX(p)
		else if lpMetric = 1 then
			distanceLpX := abs(realPoint(p).x - x) + abs(realPoint(p).y - y)
		else if lpMetric = lpInfinity then
			distanceLpX := distancePLpX(p)
		else
			distanceLpX := realPthRoot(distancePLpX(p), lpMetric);
	end;

function realPoint.distancePLpX (p: point2d): extended;
	var
		r, r1, r2: real;
	begin
		r1 := dxPLpX(p);
		r2 := dyPLpX(p);
		if lpMetric = lpInfinity then
			r := realMax(r1, r2)
		else
			r := r1 + r2;
		distancePLpX := r;
	end;

function realPoint.dxPLpX (p: point2d): extended;
	begin
		if trunc(lpMetric) = lpMetric then
			if lpMetric = lpInfinity then
				dxPLpX := abs(x - realPoint(p).x)
			else
{$IFC THINK_PASCAL }
				dxPLpX := realIntegerPower(abs(x - realPoint(p).x), trunc(lpMetric))
{$ELSEC }
				dxPLpX := realIntegerPower(abs(x - realPoint(p).x), truncLong(lpMetric))
{$ENDC }
		else
			dxPLpX := realPower(abs(x - realPoint(p).x), lpMetric)
	end;

function realPoint.dyPLpX (p: point2d): extended;
	begin
		if trunc(lpMetric) = lpMetric then
			if lpMetric = lpInfinity then
				dyPLpX := abs(y - realPoint(p).y)
			else
{$IFC THINK_PASCAL }
				dyPLpX := realIntegerPower(abs(y - realPoint(p).y), trunc(lpMetric))
{$ELSEC }
				dyPLpX := realIntegerPower(abs(y - realPoint(p).y), truncLong(lpMetric))
{$ENDC }
		else
			dyPLpX := realPower(abs(y - realPoint(p).y), lpMetric)
	end;

procedure realPoint.circleCenterX (p, q: point2d; var cx, cy: extended);
	const
		epsilon = 1E-6;
	var
		d, s1, s2, s3: extended;
		pr, qr: realPoint;
	begin
		pr := realPoint(p);
		qr := realPoint(q);
		d := 2 * whichSideX(p, q);
		if d = 0.0 then
			d := epsilon;
		s1 := sqr(pr.x) + sqr(pr.y);
		s2 := sqr(x) + sqr(y);
		s3 := sqr(qr.x) + sqr(qr.y);
		cx := (s1 * (y - qr.y) + s2 * (qr.y - pr.y) + s3 * (pr.y - y)) / d;
		cy := (s1 * (qr.x - x) + s2 * (pr.x - qr.x) + s3 * (x - pr.x)) / d;
	end;

function ws (p, q, r: realPoint): real;
	begin
		ws := (p.x - q.x) * (q.y - r.y) - (p.y - q.y) * (q.x - r.x);
	end;

function intersectLineLine (p, q, r, s, intersection: realPoint): boolean;
	var
		prq, psq, d: real;
	begin
		prq := ws(p, r, q);
		psq := ws(p, s, q);
		if prq = psq then
			if psq = 0 then begin
				intersection.x := p.x;
				intersection.y := p.y;
				intersectLineLine := true;
			end
			else
				intersectLineLine := false
		else begin
			d := prq - psq;
			intersection.x := (prq * s.x - psq * r.x) / d;
			intersection.y := (prq * s.y - psq * r.y) / d;
			intersectLineLine := true;
		end;
	end;

function realPoint.lineIntersectsLine (q, r, s, intersectionPoint: point2d): boolean;
	begin
		lineIntersectsLine := intersectLineLine(self, realPoint(q), realPoint(r), realPoint(s), realPoint(intersectionPoint));
	end;

function intersectLine (p, q, r, s: realPoint; left, right: realPoint): boolean;
	var
		prq, psq, d: real;
	begin
		intersectLine := true;
		orderPointsGreaterThanXY(point2d(r), point2d(s));
		prq := ws(p, r, q);
		psq := ws(p, s, q);
		if prq = psq then
			if psq = 0 then begin
				left.x := r.x;
				left.y := r.y;
				right.x := q.x;
				right.y := q.y;
			end
			else
				intersectLine := false
		else if (prq >= 0) & (psq <= 0) | (prq <= 0) & (psq >= 0) then begin
			d := prq - psq;
			left.x := (prq * s.x - psq * r.x) / d;
			left.y := (prq * s.y - psq * r.y) / d;
			right.x := left.x;
			right.y := left.y;
		end
		else
			intersectLine := false;
	end;

function realPoint.intersectLineX (q, r, s, left, right: point2d): boolean;
	begin
		intersectLineX := intersectLine(self, realPoint(q), realPoint(r), realPoint(s), realPoint(left), realPoint(right));
	end;

function realPoint.intersectLineSegmentX (q, r, s, left, right: point2d): boolean;
	function intersectLineSegment (p, q, r, s: realPoint; left, right: realPoint): boolean;
		var
			pqr, pqs, ldx, kdx, ldy, kdy, lkdx, lkdy, ldxkdy, ldykdx, det: real;

		begin
			intersectLineSegment := true;
			orderPointsGreaterThanXY(point2d(p), point2d(q));
			orderPointsGreaterThanXY(point2d(r), point2d(s));
			if p.equalPoint(r) then begin
				left.x := p.x;
				left.y := p.y;
				if q.equalPoint(s) | (q.whichSideX(r, s) = 0) then begin
					orderPointsGreaterThanXY(point2d(q), point2d(s));
					right.x := q.x;
					right.y := q.y;
				end
				else begin
					right.x := left.x;
					right.y := left.y;
				end;
			end
			else if p.equalPoint(s) then begin
				left.x := p.x;
				left.y := p.y;
				right.x := left.x;
				right.y := left.y;
			end
			else if q.equalPoint(r) then begin
				left.x := q.x;
				left.y := q.y;
				right.x := left.x;
				right.y := left.y;
			end
			else if q.equalPoint(s) then begin
				right.x := q.x;
				right.y := q.y;
				if p.whichSideX(r, s) = 0 then begin
					orderPointsGreaterThanXY(point2d(p), point2d(r));
					left.x := r.x;
					left.y := r.y;
				end
				else begin
					left.x := right.x;
					left.y := right.y;
				end;
			end
			else begin
				pqr := p.whichSideX(q, r);
				pqs := p.whichSideX(q, s);
				if (pqr = 0) & (pqs = 0) then begin
					orderPointsGreaterThanXY(point2d(p), point2d(r));
					orderPointsGreaterThanXY(point2d(q), point2d(s));
					if pointGreaterThanXY(point2d(r), point2d(q)) then
						intersectLineSegment := false
					else begin
						left.x := r.x;
						left.y := r.y;
						right.x := q.x;
						right.y := q.y;
					end;
				end
				else if (sign(pqr) <> sign(pqs)) & (sign(r.whichSideX(s, p)) <> sign(r.whichSideX(s, q))) then begin
					ldx := p.x - q.x;
					ldy := p.y - q.y;
					kdx := r.x - s.x;
					kdy := r.y - s.y;
					lkdx := p.x - r.x;
					lkdy := p.y - r.y;
					ldykdx := ldy * kdx;
					ldxkdy := ldx * kdy;
					det := ldxkdy - ldykdx;
					left.x := (kdx * ldx * lkdy + r.x * ldxkdy - p.x * ldykdx) / det;
					left.y := (kdy * ldy * lkdx + r.y * ldykdx - p.y * ldxkdy) / (-det);
					right.x := left.x;
					right.y := left.y;
				end
				else
					intersectLineSegment := false;
			end;
		end; (* intersectLineSegment *)

	begin
		intersectLineSegmentX := intersectLineSegment(self, realPoint(q), realPoint(r), realPoint(s), realPoint(left), realPoint(right));
	end;

procedure doProjectOnLine (p, q, r: realPoint; var rProjectX, rProjectY: real; var distance, whichSide: extended; var lambdaSign: signType);
	var
		dxqp, dyqp, dxrp, dyrp, rv, ws, d: real;
	begin
		dxqp := q.x - p.x;
		dyqp := q.y - p.y;
		if (dxqp = 0) & (dyqp = 0) then begin
			rProjectX := p.x;
			rProjectY := p.y;
			distance := sqrt(sqr(p.x - r.x) + sqr(p.y - r.y));
			whichSide := 0;
			lambdaSign := 0;
		end
		else begin
			dxrp := r.x - p.x;
			dyrp := r.y - p.y;
			rv := dxqp * dxrp + dyqp * dyrp;
			ws := dxrp * (q.y - r.y) - dyrp * (q.x - r.x);
			d := sqr(dxqp) + sqr(dyqp);
			rProjectX := p.x + dxqp * rv / d;
			rProjectY := p.y + dyqp * rv / d;
			whichSide := ws;
			distance := abs(whichSide / sqrt(d));
			if rv < 0 then
				lambdaSign := -1
			else if rv > d then
				lambdaSign := +1
			else
				lambdaSign := 0;
		end;
	end;

procedure realPoint.projectOnLine (endPoint, candidatePoint, projectionPoint: point2d; var distanceToLine, whichSide: extended; var lambdaSign: signType);
	var
		px, py: real;
	begin
		if projectionPoint = nil then
			doProjectOnLine(self, realPoint(endPoint), realPoint(candidatePoint), px, py, distanceToLine, whichSide, lambdaSign)
		else
			doProjectOnLine(self, realPoint(endPoint), realPoint(candidatePoint), realPoint(projectionPoint).x, realPoint(projectionPoint).y, distanceToLine, whichSide, lambdaSign);
	end;

function realPoint.distanceToLineSegmentX (p, q: point2d; var closestX, closestY: extended): extended;
	var
		rpx, rpy: real;
		d, w: extended;
		l: signType;
	begin
		doProjectOnLine(realPoint(p), realPoint(q), self, rpx, rpy, d, w, l);
		if l < 0 then begin
			closestX := realPoint(p).x;
			closestY := realPoint(p).y;
			distanceToLineSegmentX := distanceX(p);
		end
		else if l > 0 then begin
			closestX := realPoint(q).x;
			closestY := realPoint(q).y;
			distanceToLineSegmentX := distanceX(q);
		end
		else begin
			closestX := rpx;
			closestY := rpy;
			distanceToLineSegmentX := d;
		end;
	end;

function realPoint.xPlusYX: extended;
	begin
		xPlusYX := x + y;
	end;

function realPoint.xMinusYX: extended;
	begin
		xMinusYX := x - y;
	end;

function realPoint.rayEvalX (p: point2d; cosSlope, sinSlope: extended): extended;
	var
		pr: realPoint;
	begin
		pr := realPoint(p);
		rayEvalX := sinSlope * (x - pr.x) / cosSlope + pr.y;
	end;

function realPoint.bisectorEvalX (p, q: point2d): extended;
	var
		pr, qr: realPoint;
	begin
		pr := realPoint(p);
		qr := realPoint(q);
		bisectorEvalX := (sqr(pr.x) + sqr(pr.y) - sqr(qr.x) - sqr(qr.y) - 2 * x * (pr.x - qr.x)) / (2 * (pr.y - qr.y));
	end;

function realPoint.intersectRayBisectorX (cosSlope, sinSlope: extended; p, q, intersection: point2d): boolean;
	var
		pr, qr, intersectionr: realPoint;
	var
		a, b, c, d, t: real;
	begin
		if (p = nil) | (q = nil) then
			intersectRayBisectorX := false
		else begin
			pr := realPoint(p);
			qr := realPoint(q);
			a := 2 * (pr.x - qr.x);
			b := 2 * (pr.y - qr.y);
			c := sqr(pr.x) + sqr(pr.y) - sqr(qr.x) - sqr(qr.y);
			d := a * cosSlope + b * sinSlope;
			if d = 0 then
				intersectRayBisectorX := false
			else begin
				t := (c - a * x - b * y) / d;
				if t >= 0 then begin
					intersectRayBisectorX := true;
					intersectionr := realPoint(intersection);
					intersectionr.x := x + t * cosSlope;
					intersectionr.y := y + t * sinSlope;
				end
				else
					intersectRayBisectorX := false;
			end;
		end;
	end;

function realPoint.intersectBisectorBisectorX (p, q, r, intersection: point2d): boolean;
	var
		cx, cy: extended;
	begin
		if (p = nil) | (q = nil) | (r = nil) then
			intersectBisectorBisectorX := false
		else begin
			if (self = p) | (self = q) | (self = r) then
			else if (p = q) | (p = r) then
				p := self
			else if q = r then
				q := self
			else
				assert(false, '|{self, p, q, r}| = 3');
			if p.whichSideX(q, r) = 0 then
				intersectBisectorBisectorX := false
			else begin
				intersectBisectorBisectorX := true;
				p.circleCenterX(q, r, cx, cy);
				intersection.setX(cx);
				intersection.setY(cy);
			end;
		end;
	end;

function realPoint.intersectRayRayX (cosLower, sinLower: extended; p: point2d; cosUpper, sinUpper: extended; intersection: point2d): boolean;
	var
		s, t, dx, dy, d: real;
		pr, intersectionr: realPoint;
	begin
		pr := realPoint(p);
		dx := pr.x - x;
		dy := pr.y - y;
		d := cosUpper * sinLower - sinUpper * cosLower;
		if (d = 0) | (cosUpper = cosLower) & (sinUpper = sinLower) then
			intersectRayRayX := false
		else begin
			t := (cosLower * dy - sinLower * dx) / d;
			s := (cosUpper * dy - sinUpper * dx) / d;
			if (t >= 0) & (s >= 0) then begin
				intersectRayRayX := true;
				intersectionr := realPoint(intersection);
				intersectionr.x := x + s * cosLower;
				intersectionr.y := y + s * sinLower;
			end;
		end;
	end;

function realPoint.getType: objectType;
	begin
		getType := realPointType;
	end;

function realPoint.getTextType: Str255;
	begin
		getTextType := 'Point(real)';
	end;

{$IFC GFEnabled }
function realPoint.convertToParameterPoint: obj;
	var
		p: pPoint;
		r: Creal;
	begin
		NEW(p);
		failNil(Handle(p));
		p.dimension := 2;
		NEW(r);
		failNil(Handle(r));
		r.value := x;
		p.fields[1] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := y;
		p.fields[2] := r;
		convertToParameterPoint := obj(p);
	end; { convertToParameterPoint }

function realPoint.convertFromParameterPoint (pP: obj): obj;
	var
		r: realPoint;
	begin
		NEW(r);
		failNil(Handle(r));
		r.x := Creal(pPoint(pP).fields[1]).value;
		r.y := Creal(pPoint(pP).fields[2]).value;
		convertFromParameterPoint := obj(r);
	end; { convertFromParameterPoint }

procedure realPoint.createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
	var
		r: realPoint;
	begin
		whichAttributes[1] := true;
		whichAttributes[2] := true;
		NEW(r);
		failNil(Handle(r));
		r.x := rectangle(rect).left;
		r.y := rectangle(rect).bottom;
		lowpP := pPoint(r.convertToParameterPoint);
		r.x := rectangle(rect).right;
		r.y := rectangle(rect).top;
		highpP := pPoint(r.convertToParameterPoint);
		r.free;
	end; { createQuery }

function realPoint.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := true;
	end; { canUseGridfile }

{$ENDC}

 (* ----------------------- inserted by JM  ------------------------------------------------ *)

function realPoint.DxX (p: point2d): extended;
	begin
		DxX := x - realPoint(p).x;
	end;

function realPoint.DyX (p: point2d): extended;
	begin
		DyX := y - realPoint(p).y;
	end;

procedure realPoint.setPoint (newp: point2d);
	begin
		x := realPoint(newp).x;
		y := realPoint(newp).y;
	end;

(* ----------------------------------------------------------------------------------------------- *)
end.
