(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1991 by Philipp Meier						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.00:  27-Mai-91								*)

unit realPoint3d;

interface

uses
	point3d, lineSegmentVector, realPoint;


implementation

const
	pointRadius3d = 2; (* radius of 3-D-Points in the drawing window *)
	pointRadius3dHighlight = 4;

function realPoint3d.distanceX (p: point3d): extended;
	var
		rp: realPoint3d;
	begin
		rp := realPoint3d(p);
		distanceX := sqrt(sqr(x - rp.x) + sqr(y - rp.y) + sqr(z - rp.z));
	end;

function realPoint3d.distanceLpX (p: point3d): extended;
	var
		rp: realPoint3d;
		lpMetricInt: integer;
	begin
		if lpMetric = 2 then
			distanceLpX := distanceX(p)
		else begin
			rp := realPoint3d(p);
			if lpMetric = 1 then
				distanceLpX := abs(rp.x - x) + abs(rp.y - y) + abs(rp.z - z)
			else begin
{$IFC THINK_PASCAL }
				lpMetricInt := trunc(lpMetric);
{$ELSEC }
				lpMetricInt := truncLong(lpMetric);
{$ENDC }
				if lpMetricInt = lpMetric then
					distanceLpX := realPthRoot(realIntegerPower(abs(rp.x - x), lpMetricInt) + realIntegerPower(abs(rp.y - y), lpMetricInt) + realIntegerPower(abs(rp.z - z), lpMetricInt), lpMetric)
				else
					distanceLpX := realPthRoot(realPower(abs(rp.x - x), lpMetric) + realPower(abs(rp.y - y), lpMetric) + realPower(abs(rp.z - z), lpMetric), lpMetric)
			end;
		end;
	end;

function realPoint3d.getType: objectType;
	begin
		getType := realPoint3dType;
	end;

procedure realPoint3d.translate (dx, dy: extended);
	begin
		x := x + dx;
		y := y + dy;
	end;

function realPoint3d.getX: extended;
	begin
		getX := x;
	end;

procedure realPoint3d.setX (newX: extended);
	begin
		x := newX;
	end;

function realPoint3d.getY: extended;
	begin
		getY := y;
	end;

procedure realPoint3d.setY (newY: extended);
	begin
		y := newY;
	end;

function realPoint3d.getZ: extended;
	begin
		getZ := z;
	end;

procedure realPoint3d.setZ (newZ: extended);
	begin
		z := newZ;
	end;

procedure realPoint3d.setPoint (p: point3d);
	var
		rp: realPoint3d;
	begin
		rp := realPoint3d(p);
		x := rp.x;
		y := rp.y;
		z := rp.z;
	end;

function realPoint3d.xLessThan (p: point3d): boolean;
	begin
		xLessThan := x < realPoint3d(p).x;
	end;

function realPoint3d.xEqual (p: point3d): boolean;
	begin
		xEqual := x = realPoint3d(p).x;
	end;

function realPoint3d.yLessThan (p: point3d): boolean;
	begin
		yLessThan := y < realPoint3d(p).y;
	end;

function realPoint3d.yEqual (p: point3d): boolean;
	begin
		yEqual := y = realPoint3d(p).y;
	end;

function realPoint3d.zLessThan (p: point3d): boolean;
	begin
		zLessThan := z < realPoint3d(p).z;
	end;

function realPoint3d.zEqual (p: point3d): boolean;
	begin
		zEqual := z = realPoint3d(p).z;
	end;

procedure realPoint3d.textFileInput;
	begin
		readTextReal(x);
		readTextReal(y);
		readTextReal(z);
	end;

procedure realPoint3d.textFileOutput;
	begin
		writeTextReal(x);
		writeTextReal(y);
		writeTextReal(z);
	end;

procedure realPoint3d.fileInput;
	begin
		readReal(x);
		readReal(y);
		readReal(z);
	end;

procedure realPoint3d.fileOutput;
	begin
		writeReal(x);
		writeReal(y);
		writeReal(z);
	end;

function realPoint3d.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, realPoint3d) & (realPoint3d(o).x = x) & (realPoint3d(o).y = y) & (realPoint3d(o).z = z);
	end;

function realPoint3d.memberType (o: obj): boolean;
	begin
		memberType := member(o, realPoint3d);
	end;

function realPoint3d.getTextType: Str255;
	begin
		getTextType := 'Real Point 3d';
	end;

{$IFC GFEnabled }
function realPoint3d.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := false;
	end;
{$ENDC}

function realPoint3d.middle (p: point3d): point3d;
	var
		t: realPoint3d;
	begin
		t := realPoint3d(duplicate);
		t.x := (x + p.getX) / 2;
		t.y := (y + p.getY) / 2;
		t.z := (z + p.getZ) / 2;
		middle := point3d(t);
	end;

function realPoint3d.crossProduct (p, q, r: point3d): point3d;
	var
		v1, v2, v3, w1, w2, w3: extended;
		pp, pq, pr, res: realPoint3d;
	begin
		pp := realPoint3d(p);
		pq := realPoint3d(q);
		pr := realPoint3d(r);
		v1 := pp.x - x;
		v2 := pp.y - y;
		v3 := pp.z - z;
		w1 := pr.x - pq.x;
		w2 := pr.y - pq.y;
		w3 := pr.z - pq.z;
		res := realPoint3d(duplicate);
		res.x := v2 * w3 - w2 * v3;
		res.y := v3 * w1 - v1 * w3;
		res.z := v1 * w2 - v2 * w1;
		crossProduct := point3d(res);
	end;

function realPoint3d.innerProduct (p, q, r: point3d): extended;
	var
		pp, pq, pr: realPoint3d;
	begin
		pp := realPoint3d(p);
		pq := realPoint3d(q);
		pr := realPoint3d(r);
		innerProduct := (pp.x - x) * (pr.x - pq.x) + (pp.y - y) * (pr.y - pq.y) + (pp.z - z) * (pr.z - pq.z);
	end;

function realPoint3d.cosPhi (p: point3d): extended;
	var
		pp: realPoint3d;
	begin
		pp := realPoint3d(p);
		cosPhi := (x * pp.x + y * pp.y + z * pp.z) / sqrt(x * x + y * y + z * z) / sqrt(sqr(pp.x) + sqr(pp.y) + sqr(pp.z));
	end;

procedure realPoint3d.difference (p: point3d; var res: point3d);
	var
		pr, resr: realPoint3d;
	begin
		pr := realPoint3d(p);
		resr := realPoint3d(res);
		resr.x := pr.x - x;
		resr.y := pr.y - y;
		resr.z := pr.z - z;
	end;

procedure realPoint3d.interactiveOutput;
	begin
		drawPoint3d(x, y, z, pointRadius3d);
	end;

procedure realPoint3d.textOutput;
	begin
		write('(', x : 4 : 1, ',', y : 4 : 1, ',', z : 4 : 1, ')');
	end;

procedure realPoint3d.interactiveHighlight;
	begin
		drawPoint3d(x, y, z, pointRadius3dHighlight);
	end;

function realPoint3d.isSelected (where: obj): boolean;
	var
		p: realPoint;
		ix, iy: integer;
		rx, ry: extended;
	begin
		if worldToWindow3d(x, y, z, ix, iy) then begin
			windToWorld(ix, iy, rx, ry);
			new(p);
			failNil(Handle(p));
			p.x := rx;
			p.y := ry;
			isSelected := p.isSelected(where);
			p.free;
		end
		else
			isSelected := false;
	end;

function realPoint3d.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(3);
		appendObjInfoExtended(v, 'X-Coordinate', x, false);
		appendObjInfoExtended(v, 'Y-Coordinate', y, false);
		appendObjInfoExtended(v, 'Z-Coordinate', z, false);
		getInfo := v;
	end;

procedure realPoint3d.setInfo (o: obj);
	var
		v: vector;
	begin
		v := vector(o);
		x := objInfoExtended(v.elements^^[1]).e;
		y := objInfoExtended(v.elements^^[2]).e;
		z := objInfoExtended(v.elements^^[3]).e;
	end;

function realPoint3d.genericLessThan (o: obj): boolean;
	var
		p: realPoint3d;
	begin
		p := realPoint3d(o);
		genericLessThan := (x < p.x) | (x = p.x) & ((y < p.y) | (y = p.y) & (z < p.z));
	end;

function realPoint3d.pointCover (n: longInt): obj;
	var
		v: vector;
		i: seqIndex0;
		p: point2d;
	begin
		p := point2d(newObj(currentPointType));
		p.setXY(x, y);
		currentVectorLength := n;
		v := vector(newInitObj(pointVectorType));
		v.append(p);
		for i := 2 to n do
			v.append(p.duplicate);
		pointCover := v;
	end;

procedure realPoint3d.rotate (xOrigin, yOrigin, alpha: extended);
	var
		newX, newY: extended;
	begin
		newX := x;
		newY := y;
		rotatePoint(xOrigin, yOrigin, alpha, newX, newY);
		x := newX;
		y := newY;
	end;

procedure realPoint3d.scale (lambda: extended);
	begin
		x := lambda * x;
		y := lambda * y;
		z := lambda * z;
	end;

procedure realPoint3d.reflect (a, b, c: extended);
	var
		newX, newY: extended;
	begin
		newX := x;
		newY := y;
		reflectPoint(a, b, c, newX, newY);
		x := newX;
		y := newY;
	end;

procedure realPoint3d.centerOfGravity (var cx, cy: extended);
	begin
		cx := x;
		cy := y;
	end;

function realPoint3d.squaredDistanceX (p: point3d): extended;
	var
		rp: realPoint3d;
	begin
		rp := realPoint3d(p);
		squaredDistanceX := sqr(rp.x - x) + sqr(rp.y - y) + sqr(rp.z - z);
	end;

(* Drawing *)
procedure realPoint3d.drawLine (p: point3d); (* Draws a line from 'self' to 'p'. *)
	begin
		drawLineSegment3d(x, y, z, realPoint3d(p).x, realPoint3d(p).y, realPoint3d(p).z);
	end;

procedure realPoint3d.labelPoint (l: Str255; r: integer); (* Draws the string 'l' below 'self'. *)
	var
		ix, iy: integer;
	begin
		if worldToWindow3d(x, y, z, ix, iy) then begin
			MoveTo(ix, iy - r - 5);
{$IFC THINK_PASCAL }
			writeDraw(l);
			ExtendDrawingSheet(ix + StringWidth(l), iy - 5 - thePort^.txSize);
{$ELSEC }
			DrawString(l);
			ExtendDrawingSheet(ix + StringWidth(l), iy - 5 - qd.thePort^.txSize);
{$ENDC }
		end;
	end;



end.
