(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit rectangle;

interface

uses
{$IFC GFEnabled }
	GFValues, 
{$ENDC}
	basicGeometry;

implementation

type
	cornerType = (leftBottom, leftTop, rightBottom, rightTop);

var
	corner: cornerType;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function rectangle.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 4;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure order (var x, y: real);
	var
		t: real;
	begin
		if x > y then begin
			t := x;
			x := y;
			y := t;
		end;
	end;

procedure rectangle.centerOfGravity (var cx, cy: extended);
	begin
		cx := (left + right) / 2.0;
		cy := (bottom + top) / 2.0;
	end;

function rectangle.area: extended;
	begin
		area := (right - left) * (top - bottom);
	end;

function rectangle.perimeter: extended;
	begin
		perimeter := 2 * (right - left + top - bottom);
	end;

procedure rectangle.orderVertices;
	begin
		order(left, right);
		order(bottom, top);
	end;

function rectangle.pointCover (n: longInt): obj;
	var
		v: vector;
		p: realPoint;
	begin
		currentVectorLength := 4;
		v := vector(newInitObj(polygon2dType));
		new(p);
		failNil(Handle(p));
		p.x := left;
		p.y := bottom;
		v.append(p.duplicate);
		p.x := right;
		p.y := bottom;
		v.append(p.duplicate);
		p.x := right;
		p.y := top;
		v.append(p.duplicate);
		p.x := left;
		p.y := top;
		v.append(p);
		pointCover := v.pointCover(n);
		v.freeAll;
	end;

function rectangle.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(4);
		appendObjInfoExtended(v, 'Left', left, false);
		appendObjInfoExtended(v, 'Bottom', bottom, false);
		appendObjInfoExtended(v, 'Right', right, false);
		appendObjInfoExtended(v, 'Top', top, false);
		getInfo := v;
	end;

procedure rectangle.setInfo (o: obj);
	var
		v: vector;
	begin
		v := vector(o);
		left := objInfoExtended(v.elements^^[1]).e;
		bottom := objInfoExtended(v.elements^^[2]).e;
		right := objInfoExtended(v.elements^^[3]).e;
		top := objInfoExtended(v.elements^^[4]).e;
		orderVertices;
	end;

procedure rectangle.construct (px, py, newX, newY: integer);
	var
		dx, dy, t: integer;
	begin
		GetKeys(globalKeyMap);
		if constrainedRectangle & not globalKeyMap[optionKeyValue] then begin
			dx := abs(newX - px);
			dy := abs(newY - py);
{$IFC THINK_PASCAL }
			t := trunc(widthByHeight * dy);
{$ELSEC }
			t := truncLong(widthByHeight * dy);
{$ENDC }
			if dx < t then
				if newX > px then
					newX := t + px
				else
					newX := px - t
			else begin
{$IFC THINK_PASCAL }
				t := trunc(dx / widthByHeight);
{$ELSEC }
				t := truncLong(dx / widthByHeight);
{$ENDC }
				if newY > py then
					newY := t + py
				else
					newY := py - t;
			end;
		end
		else if globalKeyMap[optionKeyValue] & not constrainedRectangle then begin
			dx := abs(newX - px);
			dy := abs(newY - py);
			if dx < dy then
				if newX > px then
					newX := dy + px
				else
					newX := px - dy
			else if newY > py then
				newY := dx + py
			else
				newY := py - dx;
		end;
		left := px;
		bottom := py;
		right := newX;
		top := newY;
		orderVertices;
	end;

procedure rectangle.constructWorld (mouse: Point);
	var
		newX, newY, dx, dy, px, py: extended;
	begin
		windToWorld(mouse.h, mouse.v, newX, newY);
		GetKeys(globalKeyMap);
		if globalKeyMap[optionKeyValue] then begin
			case corner of
				leftBottom:  begin
					px := right;
					py := top;
				end;
				leftTop:  begin
					px := right;
					py := bottom;
				end;
				rightBottom:  begin
					px := left;
					py := top;
				end;
				rightTop:  begin
					px := left;
					py := bottom;
				end;
			end;
			dx := abs(newX - px);
			dy := abs(newY - py);
			if dx < dy then
				if newX > px then
					newX := dy + px
				else
					newX := px - dy
			else if newY > py then
				newY := dx + py
			else
				newY := py - dx;
		end;
		case corner of
			leftBottom:  begin
				left := newX;
				bottom := newY;
			end;
			leftTop:  begin
				left := newX;
				top := newY;
			end;
			rightBottom:  begin
				right := newX;
				bottom := newY;
			end;
			rightTop:  begin
				right := newX;
				top := newY;
			end;
		end;
	end;

procedure rectangle.showObjectParameter;
	begin
		displayInfoString(StringOf('Width = ', abs(right - left) : 8 : 3, '. Height = ', abs(top - bottom) : 8 : 3, '.'));
	end;

procedure rectangle.windowToWorld;
	begin
		with currentFrame do begin
			left := (left - dx) / mx;
			right := (right - dx) / mx;
			top := (top - dy) / my;
			bottom := (bottom - dy) / my;
		end;
	end;

function rectangle.isSelected (where: obj): boolean;
	var
		vertical, horizontal, bottomInRange, leftInRange, result: boolean;
	begin
		if member(where, rectangle) then
			with rectangle(where) do begin
				bottomInRange := (bottom <= self.bottom) & (self.bottom <= top);
				leftInRange := (left <= self.left) & (self.left <= right);
				horizontal := (bottomInRange | (bottom <= self.top) & (self.top <= top)) & ((self.left <= left) & (left <= self.right) | (self.left <= right) & (right <= self.right));
				vertical := (leftInRange | (left <= self.right) & (self.right <= right)) & ((self.bottom <= bottom) & (bottom <= self.top) | (self.bottom <= top) & (top <= self.top));
				result := horizontal | vertical | bottomInRange & leftInRange;
			end
		else
			result := false;
		isSelected := result;
	end;

function rectangle.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;

	function drag (var x, y: real; ct: cornerType): boolean;
		begin
			if (rectangle(where).left <= x) & (x <= rectangle(where).right) & (rectangle(where).bottom <= y) & (y <= rectangle(where).top) then begin
				drag := true;
				corner := ct;
				dragVertexWorld(mouse, highlight);
				orderVertices;
			end
			else
				drag := false;
		end;

	begin (* rectangle.dragVertex *)
		dragVertex := drag(left, top, leftTop) | drag(left, bottom, leftBottom) | drag(right, top, rightTop) | drag(right, bottom, rightBottom);
	end;

procedure rectangle.windowOutput;
	begin
{$IFC THINK_PASCAL }
		FrameRect(trunc(bottom), trunc(left), trunc(top) + 1, trunc(right) + 1);
{$ELSEC }
		FrameRect2(truncLong(bottom), truncLong(left), truncLong(top) + 1, truncLong(right) + 1);
{$ENDC }
	end;

procedure rectangle.interactiveOutput;
	begin
		if (left = right) & (bottom = top) then
			drawPoint(left, bottom, pointRadius)
		else
			drawRectangle2d(left, bottom, right, top);
	end;

procedure rectangle.interactiveHighlight;
	begin
		if (left = right) & (bottom = top) then
			drawPoint(left, bottom, pointRadiusHighlight)
		else
			inherited interactiveHighlight;
	end;

procedure rectangle.textFileInput;
	begin
		readTextReal(left);
		readTextReal(bottom);
		readTextReal(right);
		readTextReal(top);
	end;

procedure rectangle.textFileOutput;
	begin
		writeTextReal(left);
		writeTextReal(bottom);
		writeTextReal(right);
		writeTextReal(top);
	end;

procedure rectangle.fileInput;
	begin
		readReal(left);
		readReal(bottom);
		readReal(right);
		readReal(top);
	end;

procedure rectangle.fileOutput;
	begin
		writeReal(left);
		writeReal(bottom);
		writeReal(right);
		writeReal(top);
	end;

procedure rectangle.initFromOppositePoints (p, q: point2d);
	begin
		left := p.getX;
		bottom := p.getY;
		right := q.getX;
		top := q.getY;
		orderVertices;
	end;

procedure rectangle.randomChange;
	var
		p, q, r, s: realPoint;
		c: circle;
		widthWeight, heightWeight, width, height: real;
	begin
		r := nil;
		if currentRandomConstraint <> nil then begin
			p := realPoint(newObj(realPointType));
			q := realPoint(newObj(realPointType));
			if member(currentRandomConstraint, circle) then begin
				r := realPoint(newObj(realPointType));
				s := realPoint(newObj(realPointType));
				c := circle(currentRandomConstraint);
				repeat
					p.randomChange;
					q.randomChange;
					initFromOppositePoints(p, q);
					p.x := left;
					p.y := bottom;
					q.x := left;
					q.y := top;
					r.x := right;
					r.y := bottom;
					s.x := right;
					s.y := top;
				until c.inside(p) & c.inside(q) & c.inside(r) & c.inside(s);
				s.free;
				r.free;
			end
			else if member(currentRandomConstraint, rectangle) & (currentWidthPercentage > 0) & (currentHeightPercentage > 0) then begin
				with rectangle(currentRandomConstraint) do begin
					widthWeight := currentWidthPercentage * (left - right) / 2;
					heightWeight := currentHeightPercentage * (top - bottom) / 2;
				end;
				repeat
					with r do begin
						width := randomReal * widthWeight;
						height := randomReal * heightWeight;
					end;
					p.randomChange;
					p.x := p.x - width;
					p.y := p.y - height;
					q.x := p.x + 2 * width;
					q.y := p.y + 2 * height;
				until p.isSelected(currentRandomConstraint) & q.isSelected(currentRandomConstraint);
				initFromOppositePoints(p, q);
			end
			else begin
				p.randomChange;
				q.randomChange;
				initFromOppositePoints(p, q);
			end;
			p.free;
			q.free;
		end;
	end;

function rectangle.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, rectangle) & (rectangle(o).left = left) & (rectangle(o).bottom = bottom) & (rectangle(o).right = right) & (rectangle(o).top = top);
	end;

function rectangle.memberType (o: obj): boolean;
	begin
		memberType := member(o, rectangle);
	end;

procedure rectangle.translate (dx, dy: extended);
	begin
		left := left + dx;
		right := right + dx;
		bottom := bottom + dy;
		top := top + dy;
	end;

procedure rectangle.rotate (xOrigin, yOrigin, alpha: extended);
	var
		x, y: extended;
	begin
		x := left;
		y := bottom;
		rotatePoint(xOrigin, yOrigin, alpha, x, y);
		left := x;
		bottom := y;
		x := right;
		y := top;
		rotatePoint(xOrigin, yOrigin, alpha, x, y);
		right := x;
		top := y;
		orderVertices;
	end;

procedure rectangle.scale (lambda: extended);
	begin
		left := lambda * left;
		right := lambda * right;
		bottom := lambda * bottom;
		top := lambda * top;
		orderVertices;
	end;

procedure rectangle.reflect (a, b, c: extended);
	var
		x, y: extended;
	begin
		x := left;
		y := bottom;
		reflectPoint(a, b, c, x, y);
		left := x;
		bottom := y;
		x := right;
		y := top;
		reflectPoint(a, b, c, x, y);
		right := x;
		top := y;
		orderVertices;
	end;

(* following function added by "Von0auf100", Thema QuadTrees *)
function rectangle.inclRectangle (bottomS, leftS, topS, rightS: extended): integer;
	const
		all = 1;
		part = 0;
		no = -1;
	begin (* rectangle.inclRectangle *)
		if (left <= leftS) & (bottom <= bottomS) & (right >= rightS) & (top >= topS) then
			inclRectangle := all
		else if (top <= bottomS) | (bottom >= topS) | (right <= leftS) | (left >= rightS) then
			inclRectangle := no
		else
			inclRectangle := part
	end; (* rectangle.inclRectangle *)
(* end of added function *)

function rectangle.getType: objectType;
	begin
		getType := rectangleType;
	end;

function rectangle.getTextType: Str255;
	begin
		getTextType := 'Rectangle';
	end;

function rectangle.genericLessThan (o: obj): boolean;
	var
		r: rectangle;
	begin
		r := rectangle(o);
		genericLessThan := (left < r.left) | (left = r.left) & ((right < r.right) | (right = r.right) & ((top < r.top) | (top = r.top) & (bottom < r.bottom)));
	end;


{$IFC GFEnabled }
function rectangle.convertToParameterPoint: obj;
	var
		p: pPoint;
		r: Creal;
	begin
		NEW(p);
		failNil(Handle(p));
		p.dimension := 4;
		NEW(r);
		failNil(Handle(r));
		r.value := left;
		p.fields[1] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := top;
		p.fields[2] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := right;
		p.fields[3] := r;
		NEW(r);
		failNil(Handle(r));
		r.value := bottom;
		p.fields[4] := r;
		convertToParameterPoint := obj(p);
	end; { convertToParameterPoint }

function rectangle.convertFromParameterPoint (pP: obj): obj;
	var
		r: rectangle;
	begin
		NEW(r);
		failNil(Handle(r));
		r.left := Creal(pPoint(pP).fields[1]).value;
		r.top := Creal(pPoint(pP).fields[2]).value;
		r.right := Creal(pPoint(pP).fields[3]).value;
		r.bottom := Creal(pPoint(pP).fields[4]).value;
		convertFromParameterPoint := obj(r);
	end; { convertFromParameterPoint }

procedure rectangle.createQuery (rect: obj; var lowpP, highpP: obj; var whichAttributes: tBooleanArray);
	var
		r: rectangle;
	begin
		whichAttributes[1] := true;
		whichAttributes[2] := true;
		whichAttributes[3] := true;
		whichAttributes[4] := true;
		NEW(r);
		failNil(Handle(r));
		r.left := -3.0E38;
		r.top := rectangle(rect).bottom;
		r.right := rectangle(rect).left;
		r.bottom := -3.0E38;
		lowpP := pPoint(r.convertToParameterPoint);

		r.left := rectangle(rect).right;
		r.top := 3.0E38;
		r.right := 3.0E38;
		r.bottom := rectangle(rect).top;
		highpP := pPoint(r.convertToParameterPoint);

		r.free;
	end; { createQuery }

function rectangle.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := true;
	end; { canUseGridfile }
{$ENDC}

{**********************************************}

procedure callAlign (l: list; isTranslate: boolean);
	var
		r: rectangle;
		p: point2d;

	procedure doAlign (o: obj);
		var
			modified: boolean;
			oCopy: sequence;

		procedure doAlignSequence (x: obj);
			var
				xp: point2d;
			begin
				if (not modified | not isTranslate) & member(x, point2d) & x.isSelected(r) then begin
					xp := point2d(x);
					modified := true;
					if p = nil then
						p := xp
					else begin
						if isTranslate then
							oCopy.translate(p.getX - xp.getX, p.getY - xp.getY);
						xp.setPoint(p);
					end;
				end;
			end;

		begin (* doAlign *)
			if member(o, sequence) then begin
				modified := false;
				oCopy := sequence(o.duplicateAll);
				oCopy.forAll(doAlignSequence);
				if modified then
					currentResult.append(oCopy)
				else
					oCopy.freeAll;
			end
			else if member(o, point2d) & (p = nil) then
				p := point2d(o);
		end; (* doAlign *)

	begin (* callAlign *)
		r := rectangle(l.ith(1));
		p := nil;
		l.forSome(doAlign, 2, l.length);
	end; (* callAlign *)

	const
		AlignC = 0;
		AlignMoveC = 1;

procedure rectangle.describe;
	begin
		with currentArguments do
			if (length > 1) & member(ith(1), rectangle) then begin
				beginSubMenu('Align To First Point');
				menuEntry('Translate Containing Object', AlignC);
				menuEntry('Move Only Point', AlignMoveC);
				endSubMenu;
			end;
	end;

procedure rectangle.execute (item: integer);
	begin
		case item of
			AlignC, AlignMoveC: 
				callAlign(currentArguments, item = AlignC);
		end;
	end;

end.
