(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(*	This module implemented by Marco Ciot				*)

(* ----------------------------------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------------------------------- *)

unit relativeNeighborhoodGraph;

interface

uses
	pointVector, graph;

type
	seqIndexObj = object(obj)
			index, sortIndex: seqIndex;
			value: obj;			(* this way the type of point is still variable! *)
			procedure free;		(* value.free; *)
			override;
		end;

	deletableGraphEdge = object(graphEdge)
			deleted: boolean;
		end;

(* ----------------------------------------------------------------------------------------------- *)

implementation

type
	edgeAngleSO = object(obj)
			key: extended;
			edge: graphEdge;
		end;


procedure seqIndexObj.free;
	begin
		value.free;
		shallowFree;
	end;


function compareRotatedXY (a, b: seqIndexObj): relationObj;	(* NOT USED ! *)
	var
		p, q: point2d;
	begin
		p := point2d(a.value);
		q := point2d(b.value);
		compareRotatedXY := comparePointXY(p, q);
	end;

function compareRotatedYX (a, b: seqIndexObj): relationObj;
	var
		p, q: point2d;
	begin
		p := point2d(a.value);
		q := point2d(b.value);
(* comparePointYX *)
		if p.yLessThan(q) then
			compareRotatedYX := lessThanObj
		else if q.yLessThan(p) then
			compareRotatedYX := greaterThanObj
		else if p.xLessThan(q) then
			compareRotatedYX := lessThanObj
		else if q.xLessThan(p) then
			compareRotatedYX := greaterThanObj
		else
			compareRotatedYX := equalObj;
	end;

function rotatedPointGreaterThanXY (a, b: seqIndexObj): boolean;
	var
		p, q: point2d;
	begin
		p := point2d(a.value);
		q := point2d(b.value);
		rotatedPointGreaterThanXY := q.xLessThan(p) | q.xEqual(p) & q.yLessThan(p);
	end;

function edgeAngleLessThan (a, b: edgeAngleSO): boolean;
	begin
		edgeAngleLessThan := (a.key < b.key);
	end;
(* ----------------------------------------------------------------------------------------------- *)

procedure modifiableUndirectedGraph.initializeFromSimpleUndirectedGraph (sug: simpleUndirectedGraph);
	var
		i: seqIndex0;
		edge: deletableGraphEdge;
	begin
		vertices := vector(sug.vertices.duplicateAll);
		edges := vector(sug.edges.duplicate);
		i := 0;
		while i < edges.length do begin
			i := i + 1;
			new(edge);
			failNil(Handle(edge));
			with graphEdge(sug.edges.elements^^[i]) do begin
				edge.startVertex := startVertex;
				edge.endVertex := endVertex;
			end;
			edge.deleted := false;
			edges.elements^^[i] := edge;
		end;
		adjacency := newHomogenVector(sug.vertices.length);
		initAdjacency;
	end;

(* ----------------------------------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------------------------------- *)

procedure drawLuneOfPoints (p, q: point2d);
	var
		a, b: point2d;
		alpha: integer;
		dx, tanalpha: extended;
		curDiameter: extended;
	begin
		if p.xLessThan(q) then begin
			a := p;
			b := q;
		end
		else begin
			b := p;
			a := q;
		end;
		curDiameter := a.distanceX(b);
		dx := b.DxX(a);
(* There should be a test on division by zero here!						*)
(* But it works without, as the result of a division by zero is INF		*)
(* and the following arctan(INF) gives the correct result: PI/2 !!!		*)
		tanalpha := a.DyX(b) / dx;
{$IFC THINK_PASCAL }
		alpha := round(arctan(tanalpha) / oneDegree);
{$ELSEC }
		alpha := roundLong(arctan(tanalpha) / oneDegree);
{$ENDC }
		drawCircle(a.getX, a.getY, curDiameter, 20 - alpha, 140);
		drawCircle(b.getX, b.getY, curDiameter, 200 - alpha, 140);
	end;
(* ----------------------------------------------------------------------------------------------- *)

procedure modifiableUndirectedGraph.reduceToGGsimple;

(* animation precondition: graph is displayed but not highlighted! *)
	var
		curEdge: deletableGraphEdge;			(* Current Edge that is checked to be removed *)
		startVertex, endVertex, tv: seqIndex;	(* Start and End of curEdge *)
		startVertexObj, endVertexObj: point2d;
		curSquaredDiameter: extended;			(* Length of curEdge *)
		i: seqIndex0;
		adjacentEdges, tl: list;
		c: circle;

	function checkCircle (edge: graphEdge): boolean;
(* Checks whether edge connects to a vertex that lies inside the circle of curEdge. *)
		var
			d1, d2: extended;
		begin
			if graphEdge(edge).endVertex <> endVertex then begin
				with point2d(vertices.elements^^[graphEdge(edge).endVertex]) do begin
					d1 := squaredDistanceX(startVertexObj);
					d2 := squaredDistanceX(endVertexObj);
					checkCircle := (d1 + d2 > curSquaredDiameter);		(* not delete *)
				end;
			end
			else begin
				checkCircle := true;
			end;
		end;

{$IFC AnimationEnabled }
	function animCheckCircle (edge: graphEdge): boolean;
(* Animated Version *)
		begin
			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				if graphEdge(edge).endVertex <> endVertex then begin
					edge.interactiveFlash;
					waitForClick(animationFlag[ARelativeNeighborhoodItem]);
				end;
			end;
			animCheckCircle := checkCircle(edge);
		end;
{$ENDC }

	begin (* reduceToGGsimple *)
		setCurrentVerticesVector(vertices);
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			new(c);
			failNil(Handle(c));
		end
		else begin
			c := nil;
		end;
{$ENDC }
		i := 0;
		while i < edges.length do begin
			i := i + 1;
			curEdge := deletableGraphEdge(edges.elements^^[i]);
			startVertex := curEdge.startVertex;
			endVertex := curEdge.endVertex;
			with adjacency do begin
				adjacentEdges := list(elements^^[startVertex]);
				tl := list(elements^^[endVertex]);
				if tl.length < adjacentEdges.length then begin
					adjacentEdges := tl;
					tv := startVertex;
					startVertex := endVertex;
					endVertex := tv;
				end;
			end;
			startVertexObj := point2d(vertices.elements^^[startVertex]);
			endVertexObj := point2d(vertices.elements^^[endVertex]);
			curSquaredDiameter := startVertexObj.squaredDistanceX(endVertexObj);
{$IFC AnimationEnabled }
			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				curEdge.interactiveHighlight;
				c.makeFrom2Points(startVertexObj, endVertexObj);
				c.interactiveOutput;
				curEdge.deleted := not adjacentEdges.sameProperty(animCheckCircle);
				curEdge.interactiveHighlight;
				if curEdge.deleted then begin
					curEdge.interactiveOutput;
					waitForClick(animationFlag[ARelativeNeighborhoodItem]);
				end;
				c.interactiveOutput;
			end
			else
{$ENDC }
				curEdge.deleted := not adjacentEdges.sameProperty(checkCircle);
		end;
		exitAdjacency;
		i := edges.length;
		while i > 0 do begin
			if deletableGraphEdge(edges.elements^^[i]).deleted then begin
				deleteEdgeSimple(i);
			end;
			i := i - 1;
		end;
		initAdjacency;
{$IFC AnimationEnabled }
		if c <> nil then begin
			c.freeAll;
		end;
{$ENDC }
	end;

(* ----------------------------------------------------------------------------------------------- *)

procedure modifiableUndirectedGraph.reduceToRNGfalse;
	var
		curEdge: deletableGraphEdge;			(* Current Edge that is checked to be removed *)
		startVertex, endVertex, tv: seqIndex;	(* Start and End of curEdge *)
		startVertexObj, endVertexObj: point2d;
		curSquaredDiameter: extended;			(* Length of curEdge *)
		i: seqIndex0;
		adjacentEdges, tl: list;

	procedure drawLune;
		begin
			drawLuneOfPoints(startVertexObj, endVertexObj);
		end;

	function checkLune (edge: graphEdge): boolean;
(* Checks whether edge connects to a vertex that lies inside the lune of curEdge. *)
		var
			d1, d2: extended;
		begin
			if edge.endVertex <> endVertex then begin
				with point2d(vertices.elements^^[edge.endVertex]) do begin
					d1 := squaredDistanceX(startVertexObj);
					d2 := squaredDistanceX(endVertexObj);
					checkLune := (d1 >= curSquaredDiameter) or (d2 >= curSquaredDiameter);		(* not delete *)
				end;
			end
			else begin
				checkLune := true;
			end;
		end;

{$IFC AnimationEnabled }
	function animCheckLune (edge: graphEdge): boolean;
(* Animated Version *)
		begin
			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				if edge.endVertex <> endVertex then begin
					edge.interactiveFlash;
					waitForClick(animationFlag[ARelativeNeighborhoodItem]);
				end;
			end;
			animCheckLune := checkLune(edge);
		end;
{$ENDC }

	begin (* reduceToRNGfalse *)
		setCurrentVerticesVector(vertices);
		i := 0;
		while i < edges.length do begin
			i := i + 1;
			curEdge := deletableGraphEdge(edges.elements^^[i]);
			startVertex := curEdge.startVertex;
			endVertex := curEdge.endVertex;
			with adjacency do begin
				adjacentEdges := list(elements^^[startVertex]);
				tl := list(elements^^[endVertex]);
				if tl.length < adjacentEdges.length then begin
					adjacentEdges := tl;
					tv := startVertex;
					startVertex := endVertex;
					endVertex := tv;
				end;
			end;
			startVertexObj := point2d(vertices.elements^^[startVertex]);
			endVertexObj := point2d(vertices.elements^^[endVertex]);
			curSquaredDiameter := startVertexObj.squaredDistanceX(endVertexObj);
{$IFC AnimationEnabled }
			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				curEdge.interactiveHighlight;
				drawLune;
				curEdge.deleted := not adjacentEdges.sameProperty(animCheckLune);
				curEdge.interactiveHighlight;
				if curEdge.deleted then begin
					curEdge.interactiveOutput;
					waitForClick(animationFlag[ARelativeNeighborhoodItem]);
				end;
				drawLune;
			end
			else
{$ENDC }
				curEdge.deleted := not adjacentEdges.sameProperty(checkLune);
		end;
		exitAdjacency;
		i := edges.length;
		while i > 0 do begin
			if deletableGraphEdge(edges.elements^^[i]).deleted then begin
				deleteEdgeSimple(i);
			end;
			i := i - 1;
		end;
		initAdjacency;
	end;
(* ----------------------------------------------------------------------------------------------- *)

procedure modifiableUndirectedGraph.reduceToRNG;	(* Supowit *)

(* animation precondition: graph is displayed and not highlighted *)
	const
		numDirections = 6;

	var
		edgeAngleAllocated: seqIndex0;
		i: seqIndex0;
		V: vector;			(* vector of seqIndexObj: sorted by value.x *)
		T: dictionary;		(* balanced tree of seqIndexObj: sorted by value.y *)
		Tdeleted: list;
		dir: 0..numDirections;
		el: elementH;
		curEdge: graphEdge;	(* Current Edge that is checked to be removed *)
		curVo, endVo: point2d;
		sortCurV, sortEndV: seqIndex0;
		sortCurVio, sortEndVio: seqIndexObj;
		sortCurVo, sortEndVo: point2d;
		curSquaredDiameter: extended;	(* Length of curEdge *)

		edgeAngles: vector;
		e: graphEdge;
		eso: edgeAngleSO;
		anExtended: extended;

		anObj: obj;
		aReference: reference;
		aDirection: directionType;
		sweepLine, startLine, endLine: lineSegment;
		animationAllocated: boolean;

		direction: array[1..numDirections] of extended;

	procedure initDirectionArray;
		var
			i: 0..numDirections;
		begin
			i := 0;
			while i < numDirections do begin
				i := i + 1;
				direction[i] := (i - 1) * 2 * pi / numDirections;
			end;
		end;

	procedure initRotatedVector (V: vector; k: integer);
		var
			i, Vlength: seqIndex0;
			entry: seqIndexObj;
			curGraphVertex: point2d;
			Velements, curGraphVerticesElements: objArrayH;
		begin
			Vlength := V.length;
			Velements := V.elements;
			curGraphVerticesElements := vertices.elements;
			i := 0;
			while i < Vlength do begin
				i := i + 1;
				entry := seqIndexObj(Velements^^[i]);
				entry.index := i;
				with point2d(entry.value) do begin
					curGraphVertex := point2d(curGraphVerticesElements^^[i]);
					setX(curGraphVertex.getX);
					setY(curGraphVertex.getY);
					rotate(0, 0, direction[k]);
				end;
			end;
			V.sortAll(rotatedPointGreaterThanXY);
			i := 0;
			while i < Vlength do begin
				i := i + 1;
				entry := seqIndexObj(Velements^^[i]);
				seqIndexObj(Velements^^[entry.index]).sortIndex := i;
			end;
		end;

	procedure myUnhighlight (x: seqIndexObj);
		begin
			vertices.elements^^[x.index].interactiveHighlight;
		end;

	procedure drawSweepLine (r: point2d);
		begin
			with sweepLine do begin
				p.setX(r.getX);
				q.setX(r.getX);
				p.setY(r.getY + 1000);
				q.setY(r.getY - 1000);
				rotate(r.getX, r.getY, -direction[dir]);
				interactiveOutput;
				waitForClick(animationFlag[ARelativeNeighborhoodItem]);
			end;
		end;

	procedure drawScanArea;
		begin
			with startLine do begin
				p.setX(curVo.getX);
				p.setY(curVo.getY);
				q.setX(curVo.getX + 1000);
				q.setY(curVo.getY);
				rotate(p.getX, p.getY, -direction[dir]);
				interactiveOutput;
			end;
			with endLine do begin
				p.setX(endVo.getX);
				p.setY(endVo.getY);
				q.setX(endVo.getX + 1000);
				q.setY(endVo.getY);
				rotate(p.getX, p.getY, -direction[dir]);
				interactiveOutput;
			end;
		end;

	function checkItOut (v: seqIndexObj): boolean;
		var
			d1, d2: extended;
			deleteCriterium: boolean;
		begin
			if (v <> sortCurVio) and (v.index <> curEdge.endVertex) then begin
				with point2d(vertices.elements^^[v.index]) do begin
					d1 := squaredDistanceX(curVo);
					d2 := squaredDistanceX(endVo);
					deleteCriterium := (d1 < curSquaredDiameter) and (d2 < curSquaredDiameter);
					if deleteCriterium then begin
						deletableGraphEdge(curEdge.primary).deleted := true;
						checkItOut := false;
					end
					else begin
						Tdeleted.push(v);
						checkItOut := true;
					end;
				end;
			end
			else begin
				checkItOut := true;
			end;
		end;

{$IFC AnimationEnabled }
	function animCheckItOut (v: seqIndexObj): boolean;
		var
			d1, d2: extended;
			deleteCriterium: boolean;
		begin
			if (v <> sortCurVio) and (v.index <> curEdge.endVertex) then begin
				with point2d(vertices.elements^^[v.index]) do begin
					if (animationFlag[ARelativeNeighborhoodItem]) then begin
						interactiveFlash;
					end;
					d1 := squaredDistanceX(curVo);
					d2 := squaredDistanceX(endVo);
					deleteCriterium := (d1 < curSquaredDiameter) and (d2 < curSquaredDiameter);
					if deleteCriterium then begin
						deletableGraphEdge(curEdge.primary).deleted := true;
						if (animationFlag[ARelativeNeighborhoodItem]) then begin
							curEdge.interactiveFlash;
							curEdge.interactiveOutput;
						end;
						animCheckItOut := false;
					end
					else begin
						Tdeleted.push(v);
						if (animationFlag[ARelativeNeighborhoodItem]) then begin
							interactiveHighlight;
						end;
						animCheckItOut := true;
					end;
					if (animationFlag[ARelativeNeighborhoodItem]) then begin
						waitForClick(animationFlag[ARelativeNeighborhoodItem]);
					end;
				end;
			end
			else begin
				animCheckItOut := true;
			end;
		end;
{$ENDC }

	procedure freeAllHere;
		begin
			T.free;
			Tdeleted.free;
			V.freeAll;
			edgeAngles.length := edgeAngleAllocated;
			edgeAngles.freeAll;
{$IFC AnimationEnabled }
			if animationAllocated then begin
				sweepLine.freeAll;
				startLine.freeAll;
				endLine.freeAll;
			end;
{$ENDC }
		end;

	begin (* reduceToRNG *)
		setCurrentVerticesVector(vertices);
		initDirectionArray;

		V := newHomogenVector(vertices.length);
		V.length := vertices.length;
		anObj := vertices.elements^^[1];
		with V do begin
			i := 0;
			while i < length do begin
				i := i + 1;
				new(seqIndexObj(elements^^[i]));
				failNil(Handle(elements^^[i]));
				seqIndexObj(elements^^[i]).value := anObj.duplicate;	(* to have the same object type (Problem: duplicateTypeTransfer? *)
			end;
		end;

		new(avlTree(T));
		failNil(Handle(T));
		T.init;
		new(Tdeleted);
		failNil(Handle(Tdeleted));
		Tdeleted.init;

		edgeAngleAllocated := edges.length;
		edgeAngles := newHomogenVector(edgeAngleAllocated);
		i := 0;
		while i < edgeAngleAllocated do begin
			i := i + 1;
			new(eso);
			failNil(Handle(eso));
			edgeAngles.elements^^[i] := eso;
		end;

{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			new(sweepLine);
			failNil(Handle(sweepLine));
			sweepLine.init;
			new(startLine);
			failNil(Handle(startLine));
			startLine.init;
			new(endLine);
			failNil(Handle(endLine));
			endLine.init;
			animationAllocated := true;
		end
		else begin
			animationAllocated := false;
		end;
{$ENDC }

		dir := 0;
		while dir < numDirections do begin
			dir := dir + 1;
			initRotatedVector(V, dir);
			sortCurV := 0;
			while sortCurV < V.length do begin
				if userAbort then begin
					freeAllHere;
					exit(reduceToRNG);
				end;
				sortCurV := sortCurV + 1;
				sortCurVio := seqIndexObj(V.elements^^[sortCurV]);
				sortCurVo := point2d(sortCurVio.value);
				curVo := point2d(vertices.elements^^[sortCurVio.index]);
{$IFC AnimationEnabled }
				if (animationFlag[ARelativeNeighborhoodItem]) then begin
					curVo.interactiveHighlight;	(* vertices.elements^^[sortCurVio.index].interactiveHighlight; *)
					drawSweepLine(curVo);
				end;
{$ENDC }
				T.insertNewObj(sortCurVio, compareRotatedYX);

				i := 1;
				el := list(adjacency.elements^^[sortCurVio.index]).first;
				while el <> nil do begin
					e := graphEdge(el^^.value);
					if not deletableGraphEdge(e.primary).deleted then begin
						sortEndVo := point2d(seqIndexObj(V.elements^^[seqIndexObj(V.elements^^[e.endVertex]).sortIndex]).value);
						anExtended := sortCurVo.DxX(sortEndVo);
						if anExtended <= 0 then begin		(* edge must be tested*)
							edgeAngleSO(edgeAngles.elements^^[i]).key := abs(sortCurVo.DyX(sortEndVo) / anExtended);
							edgeAngleSO(edgeAngles.elements^^[i]).edge := e;
							i := i + 1;
						end;
					end;
					el := el^^.next;
				end;
				edgeAngles.length := i - 1;
				edgeAngles.sortAll(edgeAngleLessThan);

				for i := 1 to edgeAngles.length do begin
					curEdge := edgeAngleSO(edgeAngles.elements^^[i]).edge;
					with curEdge do begin
						endVo := point2d(vertices.elements^^[endVertex]);
						curSquaredDiameter := curVo.squaredDistanceX(endVo);
						sortEndV := seqIndexObj(V.elements^^[endVertex]).sortIndex;
						sortEndVio := seqIndexObj(V.elements^^[sortEndV]);
{$IFC AnimationEnabled }
						if (animationFlag[ARelativeNeighborhoodItem]) then begin
							curEdge.interactiveHighlight;
							drawScanArea;
							drawLuneOfPoints(curVo, endVo);
							waitForClick(animationFlag[ARelativeNeighborhoodItem]);

							if point2d(sortCurVio.value).yLessThan(point2d(sortEndVio.value)) then begin
								T.rangeForAllWhile(sortCurVio, sortEndVio, compareRotatedYX, animCheckItOut, right);
							end
							else begin
								T.rangeForAllWhile(sortEndVio, sortCurVio, compareRotatedYX, animCheckItOut, right);
							end;
						end
						else begin
{$ENDC }
							if point2d(sortCurVio.value).yLessThan(point2d(sortEndVio.value)) then begin
								T.rangeForAllWhile(sortCurVio, sortEndVio, compareRotatedYX, checkItOut, right);
							end
							else begin
								T.rangeForAllWhile(sortEndVio, sortCurVio, compareRotatedYX, checkItOut, right);
							end;
{$IFC AnimationEnabled }
						end;
{$ENDC }
						with Tdeleted do begin
							while length > 0 do begin
								if T.find(first^^.value, compareRotatedYX, aReference, aDirection) then begin	(* it simpy must be found!!! *)
									T.delete(aReference);
								end;
								pop;
							end;
						end;
{$IFC AnimationEnabled }
						if (animationFlag[ARelativeNeighborhoodItem]) then begin
							curEdge.interactiveHighlight;
							drawLuneOfPoints(curVo, endVo);
							startLine.interactiveOutput;
							endLine.interactiveOutput;
						end;
{$ENDC }
					end;
				end;
{$IFC AnimationEnabled }
				if (animationFlag[ARelativeNeighborhoodItem]) then begin
					curVo.interactiveOutput;
					sweepLine.interactiveOutput;
				end;
{$ENDC }
			end;
{$IFC AnimationEnabled }
			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				T.forAll(myUnhighlight, right);
				vertices.interactiveOutput;
			end;
{$ENDC }
			T.reset;
		end;

		exitAdjacency;
		i := edges.length;
		while i > 0 do begin
			if deletableGraphEdge(edges.elements^^[i]).deleted then begin
				deleteEdgeSimple(i);
(* The Updating of the Adjacencies is done very inefficently!!! --> IMPROVE IT!!! *)
			end;
			i := i - 1;
		end;
		initAdjacency;

		freeAllHere;
	end;

(* ----------------------------------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------------------------------- *)

function voronoiDiagram.modifiableDT: modifiableUndirectedGraph;

(* animation postcondition: delaunay triangulation is displayed *)

	var
		dt: modifiableUndirectedGraph;

	procedure createGraphEdge (x: voronoiEdge);
		var
			e: deletableGraphEdge;
		begin
			new(e);
			failNil(Handle(e));
			e.startVertex := idRealPoint(x.p1).id;
			e.endVertex := idRealPoint(x.p2).id;
			dt.edges.append(e);
		end;

	begin
		new(dt);
		failNil(Handle(dt));
		dt.vertices := vector(points.duplicateAll);
		dt.edges := newHomogenVector(edges.length);
		edges.forAll(createGraphEdge);
		dt.adjacency := newHomogenVector(dt.vertices.length);
		dt.initAdjacency;
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			dt.interactiveOutput;
			waitForClick(animationFlag[ARelativeNeighborhoodItem]);
		end;
{$ENDC }
		if userAbort then begin
			dt.freeAll;
			modifiableDT := nil;
			exit(modifiableDT);
		end;
		modifiableDT := dt;
	end;


function voronoiDiagram.gabrielGraph: modifiableUndirectedGraph;

(* animation precondition: voronoiDiagram is displayed but not highlighted *)
	var
		gg: modifiableUndirectedGraph;
		c: circle;

	procedure checkItOut (x: voronoiEdge);
		var
			e: deletableGraphEdge;	(* the potential delaunay edge *)
			d: extended;
			ok: boolean;
		begin
			new(e);
			failNil(Handle(e));
			e.startVertex := idRealPoint(x.p1).id;
			e.endVertex := idRealPoint(x.p2).id;

			d := x.p1.squaredDistanceX(x.p2);
			ok := true;
			if not (x.eType = bisector) then begin
				ok := (x.v1.squaredDistanceX(x.p1) + x.v1.squaredDistanceX(x.p2) > d);
				if ok and (x.eType = segment) then begin
					ok := (x.v2.squaredDistanceX(x.p1) + x.v2.squaredDistanceX(x.p2) > d);
				end;
			end;

			if ok then begin
				gg.edges.append(e);
			end
			else begin
				e.free;
			end;
		end;

{$IFC AnimationEnabled }
	procedure animCheckItOut (x: voronoiEdge);
		var
			e: deletableGraphEdge;	(* the potential delaunay edge *)
			d: extended;
			ok: boolean;
		begin
			new(e);
			failNil(Handle(e));
			e.startVertex := idRealPoint(x.p1).id;
			e.endVertex := idRealPoint(x.p2).id;

			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				x.interactiveHighlight;
				e.interactiveHighlight;
				c.makeFrom2Points(x.p1, x.p2);
				c.interactiveOutput;
				waitForClick(animationFlag[ARelativeNeighborhoodItem]);
			end;
			d := x.p1.squaredDistanceX(x.p2);
			ok := true;
			if not (x.eType = bisector) then begin
				ok := (x.v1.squaredDistanceX(x.p1) + x.v1.squaredDistanceX(x.p2) > d);
				if (animationFlag[ARelativeNeighborhoodItem]) then begin
					x.v1.interactiveFlash;
				end;
				if ok and (x.eType = segment) then begin
					ok := (x.v2.squaredDistanceX(x.p1) + x.v2.squaredDistanceX(x.p2) > d);
					if (animationFlag[ARelativeNeighborhoodItem]) then begin
						x.v2.interactiveFlash;
					end;
				end;
			end;

			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				x.interactiveHighlight;
				x.interactiveOutput;
				c.interactiveOutput;
			end;
			if ok then begin
				gg.edges.append(e);
			end
			else begin
				if (animationFlag[ARelativeNeighborhoodItem]) then begin
					e.interactiveHighlight;
				end;
				e.free;
			end;
		end;
{$ENDC }

	begin (* gabrielGraph *)
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			new(c);
			failNil(Handle(c));
		end
		else begin
			c := nil;
		end;
{$ENDC }
		new(gg);
		failNil(Handle(gg));
		gg.vertices := vector(points.duplicateAll);
		setCurrentVerticesVector(gg.vertices);
		gg.edges := newHomogenVector(edges.length);
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			edges.forAll(animCheckItOut);
		end
		else
{$ENDC }
			edges.forAll(checkItOut);

{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			gg.edges.interactiveHighlight;
			gg.edges.interactiveOutput;
			waitForClick(animationFlag[ARelativeNeighborhoodItem]);
		end;
		if c <> nil then begin
			c.freeAll;
		end;
{$ENDC }
		gg.adjacency := newHomogenVector(gg.vertices.length);
		gg.initAdjacency;
		if userAbort then begin
			gg.freeAll;
			gabrielGraph := nil;
			exit(gabrielGraph);
		end;
		gabrielGraph := gg;
	end;


function voronoiDiagram.RNGfalse: modifiableUndirectedGraph;

(* animation precondition: voronoiDiagram is displayed but not highlighted *)
	var
		gg: modifiableUndirectedGraph;

	procedure checkItOut (x: voronoiEdge);
		var
			e: deletableGraphEdge;	(* the potential delaunay edge *)
			d: extended;
			ok: boolean;
		begin
			new(e);
			failNil(Handle(e));
			e.startVertex := idRealPoint(x.p1).id;
			e.endVertex := idRealPoint(x.p2).id;

			d := x.p1.squaredDistanceX(x.p2);
			ok := true;
			if not (x.eType = bisector) then begin
				ok := (x.v1.squaredDistanceX(x.p1) >= d) or (x.v1.squaredDistanceX(x.p2) >= d);
				if ok and (x.eType = segment) then begin
					ok := (x.v2.squaredDistanceX(x.p1) >= d) or (x.v2.squaredDistanceX(x.p2) >= d);
				end;
			end;

			if ok then begin
				gg.edges.append(e);
			end
			else begin
				e.free;
			end;
		end;

{$IFC AnimationEnabled }
	procedure animCheckItOut (x: voronoiEdge);
		var
			e: deletableGraphEdge;	(* the potential delaunay edge *)
			d: extended;
			ok: boolean;
		begin
			new(e);
			failNil(Handle(e));
			e.startVertex := idRealPoint(x.p1).id;
			e.endVertex := idRealPoint(x.p2).id;

			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				x.interactiveHighlight;
				e.interactiveHighlight;
				drawLuneOfPoints(x.p1, x.p2);
				waitForClick(animationFlag[ARelativeNeighborhoodItem]);
			end;
			d := x.p1.squaredDistanceX(x.p2);
			ok := true;
			if not (x.eType = bisector) then begin
				ok := (x.v1.squaredDistanceX(x.p1) >= d) or (x.v1.squaredDistanceX(x.p2) >= d);
				if (animationFlag[ARelativeNeighborhoodItem]) then begin
					x.v1.interactiveFlash;
				end;
				if ok and (x.eType = segment) then begin
					ok := (x.v2.squaredDistanceX(x.p1) >= d) or (x.v2.squaredDistanceX(x.p2) >= d);
{$IFC AnimationEnabled }
					if (animationFlag[ARelativeNeighborhoodItem]) then begin
						x.v2.interactiveFlash;
					end;
				end;
			end;

			if (animationFlag[ARelativeNeighborhoodItem]) then begin
				x.interactiveHighlight;
				x.interactiveOutput;
				drawLuneOfPoints(x.p1, x.p2);
			end;
			if ok then begin
				gg.edges.append(e);
			end
			else begin
				if (animationFlag[ARelativeNeighborhoodItem]) then begin
					e.interactiveHighlight;
				end;
				e.free;
			end;
		end;
{$ENDC }

	begin (* RNGfalse *)
		new(gg);
		failNil(Handle(gg));
		gg.vertices := vector(points.duplicateAll);
		setCurrentVerticesVector(gg.vertices);
		gg.edges := newHomogenVector(edges.length);
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			edges.forAll(animCheckItOut);
		end
		else
{$ENDC }
			edges.forAll(checkItOut);

{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			gg.edges.interactiveHighlight;
			gg.edges.interactiveOutput;
			waitForClick(animationFlag[ARelativeNeighborhoodItem]);
		end;
{$ENDC }
		gg.adjacency := newHomogenVector(gg.vertices.length);
		gg.initAdjacency;
		if userAbort then begin
			gg.freeAll;
			RNGfalse := nil;
			exit(RNGfalse);
		end;
		RNGfalse := gg;
	end;
(* ----------------------------------------------------------------------------------------------- *)

procedure modifiableUndirectedGraph.reduce2ggSimple;

(* animation precondition: graph is displayed and highlighted *)
	begin
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			interactiveHighlight;	(* unhighlight graph *)
		end;
{$ENDC }

		reduceToGGsimple;
	end;


procedure modifiableUndirectedGraph.reduce2rngSupowit;

(* precondition: graph must be subgraph of gg *)
(* yet2do: an new method (reduce2rng2) that works, even if it isn't any subgraph of gg *)
(*				very easy: calculate GGsimple before!!! *)

(* animation precondition: graph is displayed and highlighted *)
	begin
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			interactiveHighlight;	(* unhighlight graph *)
		end;
{$ENDC }

		reduceToRNG;
	end;


procedure modifiableUndirectedGraph.reduce2rngFalse;

(* animation precondition: graph is displayed and highlighted *)
	begin
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			interactiveHighlight;	(* unhighlight graph *)
		end;
{$ENDC }

		reduceToRNGfalse;
	end;
(* ----------------------------------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------------------------------- *)

function pointVector.ggSimple: modifiableUndirectedGraph;

(* animation precondition: points are displayed and highlighted *)
	var
		vd: voronoiDiagram;
		gg: modifiableUndirectedGraph;

	begin
		vd := voronoiDiagramSweepLineSeidel(false);
		if userAbort then begin
			ggSimple := nil;
			exit(ggSimple);
		end;
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin	(* undisplay points *)
			interactiveOutput;
			interactiveHighlight;
		end;
{$ENDC }
		gg := modifiableUndirectedGraph(vd.modifiableDT);
		vd.freeAll;
		if userAbort then begin
			ggSimple := nil;
			exit(ggSimple);
		end;

(* gg = Delaunay Triangulation *)

		gg.reduceToGGsimple;
		if userAbort then begin
			gg.freeAll;
			ggSimple := nil;
			exit(ggSimple);
		end;
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			waitForClick(animationFlag[ARelativeNeighborhoodItem]);
		end;
{$ENDC }
		ggSimple := gg;
	end;

function pointVector.rngFalse: modifiableUndirectedGraph;

(* animation precondition: points are displayed and highlighted *)
	var
		vd: voronoiDiagram;
		gg: modifiableUndirectedGraph;

	begin
		vd := voronoiDiagramSweepLineSeidel(false);
		if userAbort then begin
			rngFalse := nil;
			exit(rngFalse);
		end;
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin	(* undisplay points *)
			interactiveOutput;
			interactiveHighlight;
		end;
{$ENDC }
		gg := modifiableUndirectedGraph(vd.modifiableDT);
		vd.freeAll;
		if userAbort then begin
			rngFalse := nil;
			exit(rngFalse);
		end;

(* gg = Delaunay Triangulation *)

		gg.reduceToRNGfalse;
		if userAbort then begin
			gg.freeAll;
			rngFalse := nil;
			exit(rngFalse);
		end;
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			waitForClick(animationFlag[ARelativeNeighborhoodItem]);
		end;
{$ENDC }
		rngFalse := gg;
	end;


function pointVector.ggMatulaSokal: modifiableUndirectedGraph;

(* animation precondition: points are displayed and highlighted *)
	var
		vd: voronoiDiagram;

	begin
		vd := voronoiDiagramSweepLineSeidel(false);
		if userAbort then begin
			ggMatulaSokal := nil;
			exit(ggMatulaSokal);
		end;
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			interactiveOutput;		(* unhighlight points *)
			interactiveHighlight;
			vd.interactiveOutput;	(* display voronoi diagram *)
		end;
{$ENDC }
		ggMatulaSokal := vd.gabrielGraph;
		vd.freeAll;
	end;


function pointVector.rngSupowit: modifiableUndirectedGraph;

(* animation precondition: points are displayed and highlighted *)
	var
		rng: modifiableUndirectedGraph;
		vd: voronoiDiagram;

	begin
		vd := voronoiDiagramSweepLineSeidel(false);
		if userAbort then begin
			rngSupowit := nil;
			exit(rngSupowit);
		end;
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			interactiveOutput;		(* unhighlight points *)
			interactiveHighlight;
			vd.interactiveOutput;	(* display voronoi diagram *)
		end;
{$ENDC }
		rng := vd.RNGfalse;
		vd.freeAll;
		if userAbort then begin
			rngSupowit := nil;
			exit(rngSupowit);
		end;

		rng.reduceToRNG;
		if userAbort then begin
			rng.freeAll;
			rngSupowit := nil;
			exit(rngSupowit);
		end;
		rngSupowit := rng;
	end;

(* ----------------------------------------------------------------------------------------------- *)

function modifiableUndirectedGraph.memberType (o: obj): boolean;
	begin
		memberType := member(o, modifiableUndirectedGraph);
	end;

function modifiableUndirectedGraph.getType: objectType;
	begin
		getType := undirectedGraphType;
	end;

function modifiableUndirectedGraph.getTextType: Str255;
	begin
		getTextType := concat('Modifiable Undirected Graph', verticesAndEdges);
	end;

(* ----------------------------------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------------------------------- *)
end.