(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit saveDrawing;

interface

uses
	Types;

const
	paletteWidth = 18;		(* for THINK Pascal defined in GeoBenchUtility	*)
	scrollBarWidth = 15;	(* for THINK Pascal defined in GeoBenchUtility	*)
	gbCreator = 'GBXY';		(* for THINK Pascal defined in streamIO			*)
	gbFileType = 'DATA';	(* for for THINK Pascal defined in streamIO		*)

procedure SaveDrawing (title: Str255);

implementation

uses
	Files, QuickDraw, Standardfile, Script;

function SavePicture (picH: PicHandle; title: Str255): OSErr;
	label
		8, 9;
	var
		myReply: StandardFileReply;
		err, ignore: OSErr;
		pictFileRefNum: Integer;
		dataLength, zeroData, count: LongInt;
	begin							{display the default Save dialog box}
		StandardPutFile('Save picture as:', title, myReply);
		err := noErr;				{return noErr if the user cancels}
		if myReply.sfGood then begin
			if not myReply.sfReplacing then	{create the file if it doesn't exist}
				err := FSpCreate(myReply.sfFile, gbCreator, 'PICT', smSystemScript);
			if err <> noErr then
				goto 9;
			err := FSpOpenDF(myReply.sfFile, fsRdWrPerm, pictFileRefNum);																				{open file}
			if err <> noErr then
				goto 8;
			zeroData := 0;
			dataLength := 4;
			for count := 1 to 512 div dataLength do	begin				{write the PICT file header}
				err := FSWrite(pictFileRefNum, dataLength, @zeroData);	{for this app, put 0's in header}
				if err <> noErr then
					goto 8;
			end;
			dataLength := GetHandleSize(Handle(picH));
			HLock(Handle(picH));										{lock picture handle before writing data}
			err := FSWrite(pictFileRefNum, dataLength, Ptr(picH^));		{data to file}
			HUnlock(Handle(picH));										{unlock picture handle after writing data}
		end;
8:		ignore := FSClose(pictFileRefNum);								{close the file}
9:		SavePicture := err;
	end;

function createPicture (sourcePort: GrafPtr): PicHandle;
	var
		myOpenCPicParams: OpenCPicParams;
		myPic: PicHandle;
		r: Rect;
	begin
		r := sourcePort^.portRect;
		r.bottom := r.bottom - scrollBarWidth;
		r.left := r.left + paletteWidth;
		r.right := r.right - scrollBarWidth;
		with myOpenCPicParams do begin
			srcRect := r;							{best rectangle for displaying this picture}
			hRes := $00480000;						{horizontal resolution}
			vRes := $00480000;						{vertical resolution}
			version := -2;							{always set this field to -2}
			reserved1 := 0;							{this field is unused}
			reserved2 := 0;							{this field is unused}
		end;
		myPic := OpenCPicture(myOpenCPicParams);	{start creating the picture}
		ClipRect(r);								{always set a valid clip region}
		CopyBits(sourcePort^.portBits, qd.thePort^.portBits, r, r, srcCopy, nil);
		ClosePicture;								{finish the picture}
		createPicture := myPic;
	end;

procedure SaveDrawing (title: Str255);
	var
		err: OSErr;
		picture: PicHandle;
	begin
		picture := createPicture(qd.thePort);
		err := SavePicture(picture, title);
		KillPicture(picture);
	end;

end.
