(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit selfIntersect;

(* Copyright 1990 by Christoph Ammann  ammann@inf.ethz.ch	*)
(* Informatik, ETH											*)
(* CH-8092 Zrich											*)
(* Switzerland												*)
(* Version 0.10: 18-Sep-90									*)

interface

{$SETC AselfIntersect = TRUE }

uses
	pointVector;

implementation

var
	resP, resQ: point2d;

function polygon2d.selfIntersect;

	type
		windIndex = 1..maxInt;
		windIndex0 = 0..maxInt;
		monoElArr = array[1..maxInt] of point2d;
		seqRec = record
				par, antiPar: seqIndex;
				target: seqIndex0; (* was seqIndex, P. Schorn, 9-May-1998 *)
			end;
		seqArr = array[windIndex] of seqRec;

	var
		seq: ^seqArr;
		(* invariant: for all i: par[i] <= target[i] < antiPar[i] *)
		w: windIndex0;
		resList: list;
		resVec: vector;
		eye: seqIndex;
		monoLen: seqIndex0;
		monoEl: ^monoElArr;
{$IFC AnimationEnabled AND AselfIntersect }
		animCut: list;
{$ENDC }

	function plusOne (i: seqIndex): seqIndex;
		begin
			if i = monoLen then
				plusOne := 1
			else
				plusOne := i + 1;
		end;{plusOne}

{$IFC AnimationEnabled AND AselfIntersect }
	procedure animate;
		procedure drawIt;
			var
				i: windIndex0;
			begin
				monoEl^[eye].drawLine(monoEl^[plusOne(eye)]);
				for i := 1 to w do
					with seq^[i] do begin
						monoEl^[par].interactiveHighlight;
{$IFC THINK_PASCAL }
						PenPat(gray);
{$ELSEC }
						PenPat(qd.gray);
{$ENDC }
						monoEl^[antiPar].interactiveHighlight;
						if target <> 0 then
							point2d(monoEl^[target]).drawLine(monoEl^[plusOne(target)]);
{$IFC THINK_PASCAL }
						PenPat(black);
{$ELSEC }
						PenPat(qd.black);
{$ENDC }
					end;
			end;
		begin
			PenSize(3, 3);
			drawIt;
			if animCut.length > 0 then begin
				animCut.interactiveFlash;
				animCut.reset;
			end;
			waitForClick(animationFlag[APolygonIntersectionItem]);
			drawIt;
			PenSize(1, 1);
		end; {animate}
{$ENDC }

	function monotonize: boolean;
		var
			ip1: seqIndex;
			pi, pim1, pip1: point2d;
			whs: extended;
		procedure putPoint (p: point2d);
			begin
				monoLen := monoLen + 1;
				monoEl^[monoLen] := p;
			end;
		begin
			pim1 := point2d(elements^^[length]);
			pi := point2d(elements^^[1]);
			ip1 := 2;
			monoLen := 0;
			repeat
				pip1 := point2d(elements^^[ip1]);
				whs := pip1.whichSideX(pim1, pi);
				if whs <> 0 then begin
					putPoint(pi);
					if whs < 0 then
						putPoint(pi);
					pim1 := pi;
				end;
				pi := pip1;
				if ip1 < length then
					ip1 := ip1 + 1
				else
					ip1 := 1;
			until ip1 = 2;
			monotonize := true;
		end; {monotonize}

	function whichside (i, j: seqIndex): signType;
		var
			p, q, r: point2d;
		begin
			p := monoEl^[i];
			q := monoEl^[plusOne(i)];
			r := monoEl^[j];
			if (q = r) | (p = q) then
				whichside := 1
			else
				whichside := sign(r.whichSideX(p, q));
		end;{whichside}

	(*	s.whichSideX(p,q) - r.whichSideX(p,q) = s.crossProductX(r,q,p) *)
	(*	s.whichsideX(p,q) = s.crossProductX(p,q,p) *)

	function crossProd (i, j: seqIndex): signType;
		(* = whichside (i,j+1) - whichside (i,j) *)
		var
			p, q, r, s: point2d;
		begin
			p := monoEl^[i];
			q := monoEl^[plusOne(i)];
			r := monoEl^[j];
			s := monoEl^[plusOne(j)];
			if p = q then begin
				p := monoEl^[plusOne(i + 1)];
				if i = 1 then
					q := monoEl^[monoLen]
				else
					q := monoEl^[i - 1];
			end;
			if r = s then begin
				r := monoEl^[plusOne(j + 1)];
				if j = 1 then
					s := monoEl^[monoLen]
				else
					s := monoEl^[j - 1];
			end;
			crossProd := sign(s.crossProductX(r, q, p));
		end;

	function intersect (i, j: seqIndex; var result: obj): boolean;
		var
			p1, q1, p2, q2: point2d;
			ls: lineSegment;
		begin
			p1 := monoEl^[i];
			q1 := monoEl^[plusOne(i)];
			p2 := monoEl^[j];
			q2 := monoEl^[plusOne(j)];
			if (p1 <> q2) & p1.intersectLineSegmentX(q1, p2, q2, resP, resQ) then begin
				if resP.equalPoint(resQ) then
					result := resP.duplicate
				else begin
					new(ls);
					failNil(Handle(ls));
					ls.p := point2d(resP.duplicate);
					ls.q := point2d(resQ.duplicate);
					result := ls;
				end;
				intersect := true;
			end
			else
				intersect := false;
		end;{intersect}

	procedure adjustTargets;
		var
			am1: seqIndex;
			i: windIndex0;
			result: obj;
			wep, wea: signType;
		begin
			for i := 1 to w do
				with seq^[i] do begin
					wep := whichside(eye, par);
					wea := whichside(eye, antiPar);

					if (wep = 0) & (eye < par) & (whichside(eye, plusOne(par)) = 0) & intersect(eye, par, result) then begin
{$IFC AnimationEnabled AND AselfIntersect }
						if animationFlag[APolygonIntersectionItem] then
							animCut.append(result);
{$ENDC }
						resList.append(result);
					end;

					if (wea = 0) & (eye < antiPar) then begin
						if antiPar > 1 then
							am1 := antiPar - 1
						else
							am1 := monoLen;
						if (whichside(eye, am1) = 0) & intersect(eye, am1, result) then begin
{$IFC AnimationEnabled AND AselfIntersect }
							if animationFlag[APolygonIntersectionItem] then
								animCut.append(result);
{$ENDC }
							resList.append(result);
						end;
					end;

					if (wep < 0) & (wea >= 0) then begin
						while whichside(eye, plusOne(target)) < 0 do
							target := plusOne(target);
						if whichside(target, eye) <= 0 then
							par := target
						else if intersect(eye, target, result) then begin
{$IFC AnimationEnabled AND AselfIntersect }
							if animationFlag[APolygonIntersectionItem] then
								animCut.append(result);
{$ENDC }
							resList.append(result);
						end;
					end;
				end;
		end; {adjustTargets}

	function initArrays: boolean; (* O(n) *)
		var
			i: seqIndex;
		begin
			i := 3;
			while crossProd(1, i) > 0 do
				i := plusOne(i);
			w := 0;
			while true do begin
				while (i < monoLen) & (crossProd(1, i) < 0) do
					i := plusOne(i);
				if i = monoLen then
					leave;
				w := w + 1;
				with seq^[w] do begin
					par := i;
					target := i;
					while crossProd(1, i) >= 0 do
						i := plusOne(i);
					antiPar := i;
				end;
			end;
			initArrays := true;
		end; {initArrays}

	procedure step;
		var
			i: windIndex0;
			pp1: seqIndex;
		begin
			for i := 1 to w do
				with seq^[i] do begin
					while crossProd(eye, par) < 0 do begin
						pp1 := plusOne(par);
						if target = par then
							target := pp1;
						par := pp1;
					end;
					while crossProd(eye, antiPar) >= 0 do
						antiPar := plusOne(antiPar);
				end;
		end; {step}

	begin
		selfIntersect := nil;
		resList := newEmptyList;
		monoEl := pointer(NewPtr(2 * length * sizeOf(point2d)));
		failNil(Handle(monoEl));
		seq := pointer(NewPtr(2 * length * sizeOf(seqRec)));
		failNil(Handle(seq));
		if (length >= 3) & monotonize & initArrays then begin
			displayInfoString(stringOf('w =', w : 2));
			resP := point2d(elements^^[1].duplicate);
			resQ := point2d(resP.duplicate);
{$IFC AnimationEnabled AND AselfIntersect }
			if animationFlag[APolygonIntersectionItem] then
				interactiveHighlight;
			animCut := newEmptyList;
{$ENDC }
			eye := 1;
			repeat
				adjustTargets;
{$IFC AnimationEnabled AND AselfIntersect }
				if animationFlag[APolygonIntersectionItem] then
					animate;
{$ENDC }
				eye := plusOne(eye);
				step;
			until eye = 1;
{$IFC AnimationEnabled AND AselfIntersect }
			animCut.free;
{$ENDC }
			resP.free;
			resQ.free;
		end;
		if resList.length > 0 then begin
			new(resVec);
			failNil(Handle(resVec));
			resVec.initFromList(resList);
			selfIntersect := resVec;
		end;
{$IFC THINK_PASCAL }
		disposPtr(pointer(monoEl));
		disposPtr(pointer(seq));
{$ELSEC }
		DisposePtr(pointer(monoEl));
		DisposePtr(pointer(seq));
{$ENDC }
		resList.free;
	end; {polygon2d.monoIntersect}

(* ------------------------------------------- *)

function plusOne (length: integer; i: seqIndex): seqIndex;
	begin
		if i < length then
			plusOne := i + 1
		else
			plusOne := 1;
	end;

function minusOne (length: integer; i: seqIndex): seqIndex;
	begin
		if i > 1 then
			minusOne := i - 1
		else
			minusOne := length;
	end;

function polygon2d.lastVertexNumber (vertexNumber: seqIndex): seqIndex;
(* liefert den Index der Vorgaengerecke von Ecke mit vertexnumber *)
(* Eingefuegt von Michael Vollenbroeker *)
	begin
		lastVertexNumber := minusOne(length, vertexNumber);
	end;

function polygon2d.nextVertexNumber (vertexNumber: seqIndex): seqIndex;
(* liefert den Index der Nachfolgerecke von Ecke mit vertexnumber *)
(* Eingefuegt von Michael Vollenbroeker *)
	begin
		nextVertexNumber := plusOne(length, vertexNumber);
	end;

function polygon2d.middleVertex (vertIndex1, vertIndex2: seqIndex): seqIndex;
(* liefert den Index der mittleren Ecke zwischen den Ecken mit den Nummern vertindex1 und vertindex2 *)
(* Eingefuegt von Michael Vollenbroeker *)
	begin
		if vertIndex1 <= vertIndex2 then
			middleVertex := (vertIndex1 + vertIndex2 + 1) div 2
		else
			middleVertex := (((vertIndex1 + vertIndex2 + length + 1) div 2 - 1) mod length) + 1
	end;

function polygon2d.numberOfVertices (m, gm: seqIndex): seqIndex0;
(* liefert die Anzahl der Ecken, die im Polygon zwischen den Ecken mit den Nummern m und gm liegen *)
(* Eingefuegt von Michael Vollenbroeker *)
	begin
		if m < gm then
			numberOfVertices := gm - m - 1
		else if gm < m then
			numberOfVertices := gm + length - m - 1
		else (* m=gm *)
			numberOfVertices := 0; (* -1 ist nicht sinnvoll *)
	end;


function intersect (poly: polygon2d; i, j: seqIndex; var result: obj): boolean;

	var
		p1, q1, p2, q2: point2d;
		ls: lineSegment;

	begin
		with poly do begin
			p1 := point2d(elements^^[i]);
			q1 := point2d(elements^^[plusOne(length, i)]);
			p2 := point2d(elements^^[j]);
			q2 := point2d(elements^^[plusOne(length, j)]);
			if p1.intersectLineSegmentX(q1, p2, q2, resP, resQ) then begin
				if resP.equalPoint(resQ) then
					result := resP.duplicate
				else begin
					new(ls);
					failNil(Handle(ls));
					ls.p := point2d(resP.duplicate);
					ls.q := point2d(resQ.duplicate);
					result := ls;
				end;
				intersect := true;
			end
			else
				intersect := false;
		end;
	end;

function polygon2d.simpleSelfIntersect: vector;
	var
		resList: list;
		resVec: vector;
		result: obj;
		i, j: integer;
	begin
		resList := newEmptyList;
		if length >= 3 then begin
			resP := point2d(elements^^[1].duplicate);
			resQ := point2d(resP.duplicate);
			for i := 3 to length - 1 do
				for j := 1 to i - 2 do
					if intersect(self, i, j, result) then
						resList.append(result);
			for j := 2 to length - 2 do
				if intersect(self, length, j, result) then
					resList.append(result);
			resP.free;
			resQ.free;
		end;
		if resList.length = 0 then
			simpleSelfIntersect := nil
		else begin
			new(resVec);
			failNil(Handle(resVec));
			resVec.initFromList(resList);
			simpleSelfIntersect := resVec;
		end;
		resList.free;
	end;

end.
