(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit sequence;

interface

uses
	obj, textIO;

const
	maxN = 100000;
	maxN1 = maxN + 1;
(* inserted by PV *)
	powerOfTwo = 4;
(* inserted by JM *)
	maxStep = 10;
	maxN2 = maxN + maxStep;
(* --- *)

type
	seqIndex = 1..maxN;
	seqIndex0 = 0..maxN;
	seqIndex1 = 1..maxN1;
	seqIndex01 = 0..maxN1;
(* inserted by PV *)
	markIndex = 1..powerOfTwo;
	markIndex0 = 0..powerOfTwo;
(* *)
(* inserted by JM *)
	stepIndex = 1..maxStep;
	seqIndex2 = 1..maxN2;
(* --- *)

	sequence = object(obj)
			length: seqIndex0; (* Number of elements in the sequence. *)
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure textFileOutput;
			override;
			procedure fileOutput;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			procedure randomChange; (* global variable 'currentRandomConstraint' is handed down *)
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure windowToWorld;
			override;
			procedure freeAll;
			override;
			function getInfo: obj;
			override;
			function getBoundingObject: obj;
			override;
			function pointCover (n: longInt): obj;
			override;
			function randomPointCover (n: longInt): obj;
			override;
			function isSelected (where: obj): boolean;
			override;
			function area: extended;
			override;
			function perimeter: extended;
			override;
			function convertToType (t: pointTypeRange): obj;
			override;
			function convertToConvexPolygon: obj;
			override;
			function convertToPolygonPossible: boolean;
			override;
			function convertToPolygon: obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			procedure forAll (procedure whatToDo (x: obj));
(* Perform the procedure 'whatToDo' on all objects. *)

			procedure forAllDown (procedure whatToDo (x: obj));
(* Perform the procedure 'whatToDo' on all objects, beginning with the last. *)

			procedure forSome (procedure whatToDo (x: obj); l: seqIndex; r: seqIndex0);
(*	Perform the procedure 'whatToDo' on the objects with indices 'l'..'r'.		*)

			procedure forRandomSubset (cardinality: seqIndex0; procedure whatToDo (x: obj));
(*	Perform 'whatToDo' on a random subset of the elements of the sequenece with given 'cardinality' . *)

			function sameProperty (function property (x: obj): boolean): boolean;
(* Tests whether all elements of the sequence have 'property'. *)

			function exists (function property (x: obj): boolean): seqIndex0;
(*	Determines whether some element of the sequence has 'property' and returns its index, 0 otherwise. *)

			function existsObject (function property (x: obj): boolean): obj;
(*	Determines whether some element of the sequence has 'property' and returns the object, nil if none exists. *)

			procedure append (x: obj);
(* Appends 'x' at the end of the sequence. *)

			procedure appendInt (x: longInt);
(* Appends an 'int' with value 'x' at the end of the sequence. *)

			procedure appendSafe (x: obj; increment: seqIndex0);
(* Like append, but the sequence is extended (if necessary) by 'increment' if there is not sufficient			*)
(*	space. The default implementation calls 'append'. *)

			procedure appendNonNil (x: obj);
(* Appends 'x' if 'x  nil'. *)

			procedure revert;
(* Reverts the order of the elements in the sequence. *)

			procedure removeLast;
(* Removes the last element from the sequence without destroying it. *)

			procedure randomShuffle (n, l: seqIndex0);
(* Chooses randomly 'n' elements from the elements in 1..'l' and places them into 1..'n'. *)
(*	The default implementation returns an error. The implementation for lists maybe not efficient. *)

			procedure randomShuffleAll;
(* Shuffles the elements in the sequence randomly. Equivalent to a call 'randomShuffle(length, length)'.	*)

			function removeRepeatedElements: sequence;
(* Removes from the sequence all adjacent repeated elements. Postcondition for the result:				*)
(*	not elements^^[i].equal(elements^^[i+1] for 1  i < length													*)

			function sameType: boolean;
(* Tests whether all elements of the sequence are of the same type. *)

			function createRandomObjects (t: objectType; n: longInt): homogenVector;
(*	Creates a vector with 'n*length' elements of type 't' such that elements (i-1)*n+1 .. i*n, 0 < i  n, are	*)
(*	randomly chosen by interpreting the i'th element of 'self' as a randomconstraint *)

			function memberTypeSet (types: objectTypeSet): boolean;
(*	Tests whether all elements of the sequence have types in 'types' *)

			function ith (i: seqIndex): obj;
(* Produces the ith element of the sequence. Although this method is also available for vectors we			*)
(* recommend for efficiency reasons the use of elements^^[i] instead of the more expensive method call. 	*)

			function cyclicIth (i: longInt): obj;
(*	Produces the ith element of the sequence by taking it modulo the length.										*)

			function showNumbering: vector;
(*	Creates a vector of 'markedRealPoint's at the object's center of gravity marked with its index.			*)

			procedure cyclicShift (newFirstElement: seqIndex);
(* Shifts the elements cyclically such that 'newFirstElement' becomes the new first element in the sequence. *)

			function createGraph (resultType: objectType): obj;
(* Creates a graph where the vertices are the elements and the edges go from 'i' to 'i+1'. *)

			function arrange (objectsPerLine: longInt): vector;
(*	Arranges the objects without overlap on a rectangular grid with 'objectsPerLine' per row. The object *)
(*	'self' is destroyed. *)

			function randomChooseFromRange (l: seqIndex; r: seqIndex0): obj;
(*	Chooses uniformly randomly an object from the range 'l' to 'r'.			*)

			function randomChooseFromAll: obj;
(*	Chooses uniformly randomly an object from the range 1 to 'length'.	*)

			function makeIdRealPointVector: homogenVector;
(*	Creates a vector of 'idRealPoint' where the 'id' is the position in the vector.	*)

			procedure dynamicExecution (function computeResult (objects: sequence): obj; highlightResult, showType: boolean);

			procedure dynamicExecutionHighlightShowType (function computeResult (objects: sequence): obj);

			procedure dynamicExecutionNoHighlightShowType (function computeResult (objects: sequence): obj);

(* inserted by PV *)
			function getLength: seqIndex0;

		end;

	objArrayH = ^objArrayP;
(* Handles (i.e. pointers to pointers)  are more efficient than mere pointers for the Macintosh memory		*)
(* manager since they can be moved after they have been allocated while pointers stay fixed. 				*)
	objArrayP = ^objArray;
	objArray = array[seqIndex] of obj;

	booleanArrayH = ^booleanArrayP;
	booleanArrayP = ^booleanArray;
	booleanArray = array[seqIndex] of boolean;

	integerArrayH = ^integerArrayP;
	integerArrayP = ^integerArray;
	integerArray = array[seqIndex] of integer;

	realArrayH = ^realArrayP;
	realArrayP = ^realArray;
	realArray = array[seqIndex] of real;

	seqIndexArrayH = ^seqIndexArrayP;
	seqIndexArrayP = ^seqIndexArray;
	seqIndexArray = array[seqIndex] of seqIndex;

	seqIndexArray0H = ^seqIndexArray0P;
	seqIndexArray0P = ^seqIndexArray0;
	seqIndexArray0 = array[seqIndex] of seqIndex0;

	vector = object(sequence)
			elements: objArrayH;		(* Handle to an array containing the elements. *)
			procedure free;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			procedure textFileInput;
			override;
			procedure fileInput;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			procedure forSome (procedure whatToDo (x: obj); l: seqIndex; r: seqIndex0);
			override;
			procedure forAllDown (procedure whatToDo (x: obj));
			override;
			function sameProperty (function property (x: obj): boolean): boolean;
			override;
			function exists (function property (x: obj): boolean): seqIndex0;
			override;
			procedure append (x: obj);
			override;
			procedure appendSafe (x: obj; increment: seqIndex0);
			override;
			function sameType: boolean;
			override;
			function ith (i: seqIndex): obj;
			override;
			procedure revert;
			override;
			procedure cyclicShift (newFirstElement: seqIndex);
			override;
			function existsObject (function property (x: obj): boolean): obj;
			override;
			procedure randomShuffle (n, l: seqIndex0);
			override;

(* Memory management *)
			procedure init;
			override;
(* The global variable 'currentVectorLength' determines how many elements are allocated and must		*)
(* be set before a call to init may occur. 																			*)

			procedure initFromList (l: list);
(* Creates a vector from the elements in the list 'l'. No call to 'init' is necessary. *)

			procedure initFromListPartial (l: list; len: seqIndex0);
(*	Creates a vector from the first 'len' elements in the list if they exist. No call to 'init' is necessary. *)

			procedure allocateElements (number: seqIndex0);
(* Reserves space for at most 'number' elements. This method can be used to dynamically extend or	*)
(* shrink a vector although it should be used cautiously because it involves copying of the used part of	*)
(* the 'elements' array. *)

			function allocatedElements: seqIndex0;
(* Returns the number of currently allocated elements. *)

(* Rearranging the elements *)

			procedure swap (i, j: seqIndex);
(* exchanges the 'i'th and the 'j'th element. *)

			procedure compact (mayRearrange: boolean);
(*	Remove all objects from the vector which are 'nil'. If 'mayRearrange' = false then the relative order	*)
(*	of the other elements does not change otherwise it may change.												*)

			procedure deleteElement (i: seqIndex);
(*	Deletes the i'th element from the vector. The order of the elements changes and the deleted object is not	*)
(*	destroyed.																													*)

			procedure deleteElementFree (i: seqIndex);
(*	Deletes the i'th element from the vector. The order of the elements changes and the deleted object is		*)
(*	freed with a call to 'free'.																									*)

			procedure deleteElementFreeAll (i: seqIndex);
(*	Deletes the i'th element from the vector. The order of the elements changes and the deleted object is		*)
(*	freed with a call to 'freeAll'.																								*)


(* Sorting and searching *)
			procedure sort (function lessThan (p, q: obj): boolean; l: seqIndex; r: seqIndex0);
(* Sorts 'elements' between 'l' and 'r' using the comparison function 'lessThan'. *)

			procedure sortAll (function lessThan (p, q: obj): boolean);
(* Sorts 'elements' between 1 and 'length' using the comparison function 'lessThan'. *)

			function binarySearch (x: obj; function compare (a, b: obj): relationObj; l: seqIndex; r: seqIndex0): seqIndex0;
(* Finds 'x' between 'l' and 'r' in the vector which is sorted in ascending order. 'binarySearch = 0'		*)
(* means that 'x' was not found, otherwise 'binarySearch' gives the location of  'x' in the vector. 		*)

			function findExtreme (function lessThan (p, q: obj): boolean): seqIndex0;
(* Finds the smallest object according to the comparison function 'lessThan'. *)

(* Heap operations *)
			procedure heapify (function lessThan (p, q: obj): boolean);
(* Produces a heap using 'lessThan'. The first element is the minimum. *)

			procedure sift (function lessThan (p, q: obj): boolean; l, r: seqIndex);
(* Sifts element 'l' into the heap which ranges from 'l'+1 to 'r'. *)

			procedure nextMin (function lessThan (p, q: obj): boolean);
(* Creates a new heap by removing the first element. *)

			procedure insert (x: obj; function lessThan (p, q: obj): boolean);
(* Inserts an element into the heap. *)

			procedure heapSort (function greaterThan (p, q: obj): boolean);
(* Sorts the whole vector using heapsort into ascending order. Note that the usual lessThan function	*)
(* will sort the wrong way! *)

			function isHeap (root: seqIndex; function lessThan (p, q: obj): boolean): boolean;
(* Tests whether the vector with the given 'root' (usually 1) fulfills the heap property. *)

(* Miscellaneous *)
			procedure preset (number: seqIndex0; t: objectType);
(* Creates a vector of 'number' objects of type 't'. No call to init is necessary. *)

			procedure eliminateDuplicates;
(* Removes from 'self' all duplicated elements and destroys them with 'freeAll'. *)

			procedure forAllPermutations (n: seqIndex0; procedure whatToDo (v: vector));
(* Executes the procedure 'whatToDo' for all permutations of the first 'n' elements. *)

			procedure readLength;
(* Reads the length of the vector from 'currentStream.' *)

			procedure partialInteractiveFlash (l: seqIndex; r: seqIndex0);
(*	like 'interactiveFlash', but only elements from 'l' to 'r' participate. 	*)

			function interactiveSelect (unique, beep: boolean; function getIndex (x: seqIndex): seqIndex; low: seqIndex; high: seqIndex0): seqIndex0;
(* Lets the user interactively select an object from the vector. If 'unique=true' only one object may	*)
(* be selected. The result is 0 if the operation was cancelled or there is nothing to select, otherwise	*)
(*	the result is an i such that getIndex(i) is the index of the	*)
(* selected object. The function 'getIndex' is used to generate the indices of the objects to be tested.	*)
(* The range of the argument is 'low..high'. The identity index function 'identityIndex' is provided.	If	*)
(*	beep=true and unique=true and more than one object was selected then an audible alarm rings.		*)

			function interactiveSelectFromAll (unique, beep: boolean): seqIndex0;
(*	Same as 'interactiveSelect' but 'getIndex = identityIndex', 'low = 1', and 'high = length'. *)

		end;

	homogenVector = object(vector) (* all objects are of the same type *)
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberTypeSet (types: objectTypeSet): boolean;
			override;
			function sameType: boolean;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function getTextBaseType: Str255;
			function memberBaseType (o: obj): boolean;
		end;

	elementH = ^elementP; (* handle *)
	elementP = ^element;
	element = record
			value: obj; (* the object *)
			next: elementH; (* the successor *)
		end;

	list = object(sequence) (* linked list *)
			first, last: elementH; (* The first and the last element in the linked list. *)
			procedure init;
			override;
			procedure free;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			procedure textFileInput;
			override;
			procedure fileInput;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function sameType: boolean;
			override;
			procedure append (x: obj);
			override;
			procedure removeLast;
			override;
			function ith (i: seqIndex): obj;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			procedure forAll (procedure whatToDo (x: obj));
			override;
			procedure forSome (procedure whatToDo (x: obj); l: seqIndex; r: seqIndex0);
			override;
			function sameProperty (function property (x: obj): boolean): boolean;
			override;
			function exists (function property (x: obj): boolean): seqIndex0;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure revert;
			override;
			procedure cyclicShift (newFirstElement: seqIndex);
			override;
			function existsObject (function property (x: obj): boolean): obj;
			override;
			procedure randomShuffle (n, l: seqIndex0);
			override;

			procedure reset;
(*	Resets the list to the state after 'init' was called. The objects in the list are not free'd.						*)

			procedure concatenate (x: list);
(* Appends the list 'x' at the end of the list. Note: no duplication is taking place and the list 'x' is left intact. 	*)

			procedure concatenateDispose (x: list);
(* Appends the list 'x' at the end of the list. Note: no duplication is taking place but the 'x' is destroyed via	*)
(*	a call to 'dispose'.																										 	*)

			procedure flatten;
(* Appends to the list recursively all elements which are member of a sequence which is a member of 'self'.	*)
(* The member sequences of 'self' are destroyed. *)

			procedure push (x: obj);
(* Appends the element 'x' at the front of the list. *)

			procedure pop;
(* Removes the first element from the list. *)

			procedure delete (e: elementH);
(* Delete the element in the list which follows 'e' and must exist. *)

		end;

	strings = object(obj)
			value: Str255;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
		end;

	int = object(obj)
			value: longInt;
			procedure interactiveInput (px, py: integer);
			override;
			procedure interactiveOutput;
			override;
			procedure textOutput;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure randomChange; (* global variable 'currentRandomConstraint' is interpreted as an 'int' *)
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
		end;

	realNumber = object(obj)
			value: real;
			function getTextType: Str255;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
		end;

	sequenceIndex = object(obj)   (* Eingefuegt von Michael Vollenbroeker 30.11.92 *)
			value: seqIndex;
			procedure setvalue (n: seqIndex);
			function getvalue: seqIndex;
		end;

function intCompare (x, y: int): relationObj;

function intLessThan (x, y: int): boolean;

function intGreaterThan (x, y: int): boolean;

function makeInt (x: longInt): int;

function lessThanRealNumber (a, b: realNumber): boolean;

(* inserted by PV *)
type
treeElementH = ^treeElementP; (* handle *)
treeElementP = ^treeElement;
treeElement = record
		el: obj; (* the object *)
		mark: markIndex0;
		left, ancestor, second, right: treeElementH;  (* the predecessor, ancestor, second coordinate and successor *)
	end;

listMuchLinked = object(sequence)
		first, last: treeElementH;
		procedure init;
		override;
		function duplicate: obj;
		override;
		procedure delete (x: treeElementH);
		procedure forAll (procedure whatToDo (x: obj));
		override;
		procedure append (x: obj);
		override;
		function appendWithPointerBack (x: obj): treeElementH;
		procedure appendAndLinkWithOtherList (x: obj; p: treeElementH);
		function buildBinaryTree: treeElementH;
		function embeddInBinaryTree: binTreeWithList;
	end;

succArray = array[markIndex] of listMuchLinked;

binTreeWithList = object(obj)		(* binary tree with embedded list *)
		root: treeElementH;
		list: listMuchLinked;			(* list much linked, def. in sequence *)
		procedure traverse (var newLists: succArray);
		procedure freeAll;
		override;
		procedure specialPrepareForFreeing;
	end;

handleAsObject = object(obj)
		tr: treeElementH;
	end;

markHandleAsObject = object(handleAsObject)
		k: markIndex0;
	end;

(* *)

doubleElementH = ^doubleElementP;
doubleElementP = ^doubleElement;
doubleElement = record
		value: obj;
		pred, succ: doubleElementH;
	end;

doubleList = object(obj)
		first, last: doubleElementH;
		procedure init;
		override;
		procedure free;
		override;
		procedure append (x: obj);
		procedure moveToFront (var d: doubleElementH);
(* inserted by PV *)
		function appendWithPointer (x: obj): doubleElementH;
		procedure delete (d: doubleElementH);
		procedure setLast (x: obj);
		procedure forAll (procedure whatToDo (x: obj));
		procedure freeAll;
		override;
(* *)
	end;

menuEntryElement = object(obj)
		entry: Str255;
		item: integer;
	end;

subMenuElement = object(obj)
		title: Str255;
		menu: list;
		procedure free;
		override;
		procedure freeAll;
		override;
	end;

{$IFC ScriptEnabled }
var
applicableFcts: list;
calledFromGeoScript: boolean;

type
executableFct = object(obj)		(* M_SCM, 25.11.93 *)
		menu: Str255;
		item: integer;
		title: Str255;
	end;

procedure initApplicableFcts;		(* M_SCM, 25.11.93 *)
{$ENDC }

procedure initCurrentMenu;

procedure menuEntry (entry: Str255; item: integer);

procedure beginSubMenu (title: Str255);

procedure endSubMenu;

procedure freeAllP (x: obj);

procedure freeP (x: obj);

procedure interactiveHighlightP (x: obj);

function dynamicArray (number, size: longInt): Handle;

function drawingComplexity2 (a, b: obj; cutOff: longInt): longInt;

function dynamicArrayNoCheck (number, size: longInt): Handle;

function newVector (length: longInt): vector;

function newHomogenVector (length: longInt): homogenVector;

procedure appendObjInfoObject (s: sequence; name: string; value: obj; protection: boolean);
{$IFC THINK_PASCAL }
procedure appendObjInfoTempObject (s: sequence; name: string; value: obj; protection: boolean; doCleanUp: ProcPtr);
{$ELSEC }
procedure appendObjInfoTempObject (s: sequence; name: string; value: obj; protection: boolean; doCleanUp: cleanUpT);
{$ENDC }
procedure appendObjInfoExtended (s: sequence; name: string; value: extended; protection: boolean);
procedure appendObjInfoString (s: sequence; name: string; value: string; protection: boolean);

function identityIndex (x: seqIndex): seqIndex;

procedure initSequence;


function newEmptyList: list;

procedure extendFreeList (number: longInt);

procedure shrinkFreeList (number: longInt);

function freeListLength: longInt;

function freeListSize: longInt; (* number of bytes in the 'freeList' *)

function newElement: elementH;

procedure disposeElement (e: elementH);

(* inserted by PV *)

procedure extendFreeTreeList (number: longInt);

procedure shrinkFreeTreeList (number: longInt);

function freeTreeListLength: longInt;

function freeTreeListSize: longInt; (* number of bytes in the 'freeTreeList' *)

function newTreeElement: treeElementH;

procedure disposeTreeElement (e: treeElementH);


procedure initDoubleList;

procedure extendFreeDoubleList (number: longInt);

procedure shrinkFreeDoubleList (number: longInt);

function freeDoubleListLength: longInt;

function freeDoubleListSize: longInt; (* number of bytes in the 'freeDoubleList' *)

function newDoubleElement: doubleElementH;

procedure disposeDoubleElement (e: doubleElementH);


(* *)

var
currentArguments, currentMenu, currentResult: list;
currentArgumentsSameType, currentArguments1IsHomogenVector, currentArguments2IsHomogenVector: boolean;

implementation

function newEmptyList: list;
	var
		result: list;
	begin
		new(result);
		failNil(Handle(result));
		result.init;
		newEmptyList := result;
	end;

function drawingComplexity2 (a, b: obj; cutOff: longInt): longInt;
	var
		c: longInt;
	begin
		c := a.drawingComplexity(cutOff);
		if cutOff > c then
			drawingComplexity2 := c + b.drawingComplexity(cutOff - c)
		else
			drawingComplexity2 := c;
	end;

function intCompare (x, y: int): relationObj;
	begin
		if x.value < y.value then
			intCompare := lessThanObj
		else if x.value > y.value then
			intCompare := greaterThanObj
		else
			intCompare := equalObj;
	end;

function intLessThan (x, y: int): boolean;
	begin
		intLessThan := x.value < y.value;
	end;

function intGreaterThan (x, y: int): boolean;
	begin
		intGreaterThan := x.value > y.value;
	end;

function lessThanRealNumber (a, b: realNumber): boolean;
	begin
		lessThanRealNumber := a.value < b.value;
	end;

function identityIndex (x: seqIndex): seqIndex;
	begin
		identityIndex := x;
	end;

procedure interactiveOutputP (x: obj);
	begin
		x.interactiveOutput;
	end;

procedure interactiveHighlightP (x: obj);
	begin
		x.interactiveHighlight;
	end;

procedure fileOutputP (x: obj);
	begin
		x.fileOutputTag;
	end;

procedure textFileOutputP (x: obj);
	begin
		writeTextObj(x);
	end;

procedure windowToWorldP (x: obj);
	begin
		x.windowToWorld;
	end;

procedure randomChangeP (x: obj);
	begin
		x.randomChange;
	end;

procedure freeAllP (x: obj);
	begin
		if x <> nil then
			x.freeAll;
	end;

procedure freeP (x: obj);
	begin
		if x <> nil then
			x.free;
	end;

function sequence.area: extended;
	var
		s: extended;

	procedure sumArea (x: obj);
		begin
			s := s + x.area;
		end;

	begin (* sequence.area *)
		s := 0;
		forAll(sumArea);
		area := s;
	end;

function sequence.perimeter: extended;
	var
		s: extended;

	procedure sumPerimeter (x: obj);
		begin
			s := s + x.perimeter;
		end;

	begin (* sequence.perimeter *)
		s := 0;
		forAll(sumPerimeter);
		perimeter := s;
	end;

procedure sequence.interactiveOutput;
	begin
		forAll(interactiveOutputP);
	end;

procedure sequence.interactiveHighlight;
	begin
		forAll(interactiveHighlightP);
	end;

function sequence.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	var
		objectToDrag: obj;

	function dragVertexP (o: obj): boolean;
		begin
			dragVertexP := not o.dragVertex(where, mouse, highlight);
		end;

	function selectedObj (o: obj): boolean;
		begin
			selectedObj := o.isSelected(where);
		end;

	begin (* sequence.dragVertex *)
		GetKeys(globalKeyMap);
		if globalKeyMap[optionKeyValue] & isSelected(where) then begin
			objectToDrag := existsObject(selectedObj);
			if objectToDrag = nil then
				dragVertex := false
			else begin
				dragVertex := true;
				if highlight then
					objectToDrag.dragObject(interactiveHighlightP, interactiveHighlightP, mouse)
				else
					objectToDrag.dragObject(doNothingObjP, doNothingObjP, mouse);
			end;
		end
		else
			dragVertex := not sameProperty(dragVertexP);
	end;

procedure sequence.textFileOutput;
	begin
		writeTextLongInt(length);
		forAll(textFileOutputP);
	end;

procedure sequence.fileOutput;
	begin
		writeLongInt(length);
		forAll(fileOutputP);
	end;

procedure sequence.randomChange;
	begin
		forAll(randomChangeP);
	end;

procedure sequence.freeAll;
	begin
		forAll(freeAllP);
		free;
	end;

procedure sequence.translate (dx, dy: extended);
	procedure translateP (x: obj);
		begin
			x.translate(dx, dy);
		end;
	begin
		forAll(translateP);
	end;

procedure sequence.rotate (xOrigin, yOrigin, alpha: extended);
	procedure rotateP (x: obj);
		begin
			x.rotate(xOrigin, yOrigin, alpha);
		end;
	begin
		forAll(rotateP);
	end;

procedure sequence.reflect (a, b, c: extended);
	procedure reflectP (x: obj);
		begin
			x.reflect(a, b, c);
		end;
	begin
		forAll(reflectP);
	end;

procedure sequence.scale (lambda: extended);
	procedure scaleP (x: obj);
		begin
			x.scale(lambda);
		end;
	begin
		forAll(scaleP);
	end;

procedure sequence.windowToWorld;
	begin
		forAll(windowToWorldP);
	end;

procedure sequence.forAll (procedure whatToDo (x: obj));
	begin
		forSome(whatToDo, 1, length);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure sequence.forSome (procedure whatToDo (x: obj); l: seqIndex; r: seqIndex0);
	begin
		unimplementedMethod('sequence.forSome');
	end;

procedure sequence.forAllDown (procedure whatToDo (x: obj));
	begin
		unimplementedMethod('sequence.forAllDown');
	end;

function sequence.sameProperty (function property (x: obj): boolean): boolean;
	begin
		unimplementedMethod('sequence.sameProperty');
		sameProperty := false;
	end;

function sequence.exists (function property (x: obj): boolean): seqIndex0;
	begin
		unimplementedMethod('sequence.exists');
		exists := 0;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function sequence.existsObject (function property (x: obj): boolean): obj;
	var
		i: seqIndex0;
	begin
		i := exists(property);
		if i = 0 then
			existsObject := nil
		else
			existsObject := ith(i);
	end;

function sequence.memberTypeSet (types: objectTypeSet): boolean;

	function typeOk (o: obj): boolean;
		begin
			typeOk := o.getType in types;
		end;

	begin
		memberTypeSet := sameProperty(typeOk);
	end;

function sequence.isSelected (where: obj): boolean;

	function notSelected (o: obj): boolean;
		begin
			notSelected := not o.isSelected(where);
		end;

	begin
		isSelected := not sameProperty(notSelected);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure sequence.append (x: obj);
	begin
		unimplementedMethod('sequence.append');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function makeInt (x: longInt): int;
	var
		result: int;
	begin
		new(result);
		failNil(Handle(result));
		result.value := x;
		makeInt := result;
	end;

procedure sequence.appendInt (x: longInt);
	begin
		append(makeInt(x));
	end;

procedure sequence.revert;
	begin
		unimplementedMethod('sequence.revert');
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure sequence.cyclicShift (newFirstElement: seqIndex);
	begin
		unimplementedMethod('sequence.cyclicShift');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure sequence.centerOfGravity (var cx, cy: extended);

	procedure findMaxCy (x: obj);
		var
			xcx, xcy: extended;
		begin
			x.centerOfGravity(xcx, xcy);
			if xcy > cy then begin
				cx := xcx;
				cy := xcy;
			end;
		end;

	begin (* sequence.centerOfGravity *)
		if length > 0 then begin
			ith(1).centerOfGravity(cx, cy);
			forAll(findMaxCy);
		end
		else begin
			cx := 0;
			cy := 0;
		end;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure sequence.appendSafe (x: obj; increment: seqIndex0);
	begin
		append(x);
	end;

procedure sequence.randomShuffle (n, l: seqIndex0);
	begin
		unimplementedMethod('sequence.randomShuffle');
	end;

function sequence.ith (i: seqIndex): obj;
	begin
		unimplementedMethod('sequence.ith');
		ith := nil;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure sequence.removeLast;
	begin
		if length > 0 then
			length := length - 1
		else
			fatalError('Nothing to remove in sequence.removeLast');
	end;

procedure sequence.randomShuffleAll;
	begin
		randomShuffle(length, length);
	end;

function sequence.removeRepeatedElements: sequence;
	var
		result: sequence;
		last: obj;

	procedure processObj (x: obj);
		begin
			if (last = nil) | not x.equal(last) then begin
				last := x;
				result.append(x.duplicateAll);
			end;
		end;

	begin
		result := sequence(newObj(getType));
		currentVectorLength := length;
		result.init;
		last := nil;
		forAll(processObj);
		removeRepeatedElements := result;
	end;

procedure sequence.appendNonNil (x: obj);
	begin
		if x <> nil then
			append(x);
	end;

function sequence.sameType: boolean;
	begin
		unimplementedMethod('sequence.sameType');
		sameType := false;
	end;

function sequence.cyclicIth (i: longInt): obj;
	begin
		assert(length > 0, 'Length > 0');
		if i > length then
			repeat
				i := i - length;
			until i <= length
		else if i < 1 then
			repeat
				i := i + length;
			until i > 0;
		cyclicIth := ith(i);
	end;

function sequence.drawingComplexity (cutOff: longInt): longInt;
	var
		currentComplexity: longInt;
		dummy: boolean;

	function testComplexity (x: obj): boolean;
		begin
			if cutOff > currentComplexity then begin
				currentComplexity := currentComplexity + x.drawingComplexity(cutOff - currentComplexity);
				testComplexity := true;
			end
			else
				testComplexity := false;
		end;

	begin
		currentComplexity := 0;
		dummy := sameProperty(testComplexity);
		drawingComplexity := currentComplexity;
	end;

function sequence.createRandomObjects (t: objectType; n: longInt): homogenVector;
	var
		r: homogenVector;
		k: longInt;

	procedure create (x: obj);
		var
			j: longInt;
		begin
			currentRandomConstraint := x;
			for j := 1 to n do begin
				r.elements^^[k].randomChange;
				k := k + 1;
			end;
		end;

	begin
		new(r);
		failNil(Handle(r));
		r.elements := nil;
		r.preset(n * length, t);
		k := 1;
		forAll(create);
		createRandomObjects := r;
	end;

function sequence.getInfo: obj;
	var
		v: vector;
		i: seqIndex0;
	procedure getInfoP (o: obj);
		begin
			appendObjInfoObject(v, stringOf(i : 2, '-Element'), o, false);
			i := i + 1;
		end;
	begin
		v := newVector(length);
		i := 1;
		forAll(getInfoP);
		getInfo := v;
	end;

function doPointCover (s: sequence; n: longInt; isRandom: boolean): obj;
	var
		v: vector;
		l: list;

	procedure cover (x: obj);
		var
			r: vector;

		procedure appendL (x: obj);
			begin
				l.appendNonNil(x);
			end;

		begin
			if isRandom then
				r := vector(x.randomPointCover(n))
			else
				r := vector(x.pointCover(n));
			if r <> nil then begin
				r.forAll(appendL);
				r.free;
			end;
		end;

	begin (* doPointCover *)
		l := newEmptyList;
		s.forAll(cover);
		if l.length = 0 then
			doPointCover := nil
		else begin
			v := vector(newObj(pointVectorType));
			v.initFromList(l);
			doPointCover := v;
		end;
		l.free;
	end;

function sequence.pointCover (n: longInt): obj;
	begin
		pointCover := doPointCover(self, n, false);
	end;

function sequence.randomPointCover (n: longInt): obj;
	begin
		randomPointCover := doPointCover(self, n, true);
	end;

function sequence.randomChooseFromRange (l: seqIndex; r: seqIndex0): obj;
	begin
		if l < r then
			randomChooseFromRange := ith(randomRange(l, r + 1))
		else if l > r then
			randomChooseFromRange := nil
		else
			randomChooseFromRange := ith(l);
	end;

function sequence.randomChooseFromAll: obj;
	begin
		randomChooseFromAll := randomChooseFromRange(1, length);
	end;

(* inserted by PV *)

function sequence.getLength: seqIndex0;
	begin
		getLength := length
	end;   { getLength }

(* *)

(* inserted by PV *)

procedure listMuchLinked.init;
	begin
		first := nil;
		last := nil;
		length := 0;
	end;

function listMuchLinked.duplicate: obj;
	var
		l: listMuchLinked;
		e: treeElementH;
	begin
		new(l);
		failNil(Handle(l));
		l.init;
		e := first;
		while e <> nil do begin
			l.append(e^^.el);
			e := e^^.right;
		end;
		duplicate := l;
	end;   { duplicate }

procedure listMuchLinked.delete (x: treeElementH);
	begin
		if x = first then
			first := x^^.right;
		if x = last then
			last := x^^.left;
		if treeElementH(x^^.ancestor) <> nil then
			if treeElementH(x^^.ancestor)^^.left = x then
				treeElementH(x^^.ancestor)^^.left := nil
			else
				treeElementH(x^^.ancestor)^^.right := nil;
		treeElementH(x^^.left)^^.right := x^^.right;
		treeElementH(x^^.right)^^.left := x^^.left;
		length := length - 1;
		disposeTreeElement(x)
	end;   { delete }

procedure listMuchLinked.forAll (procedure whatToDo (x: obj));
	var
		el: treeElementH;
		i: seqIndex0;
	begin
		el := first;
		for i := 1 to length do begin
			whatToDo(el^^.el);
			el := el^^.right;
		end;
	end;

procedure listMuchLinked.append (x: obj);
	var
		e: treeElementH;
	begin
		e := newTreeElement;
		e^^.ancestor := nil;
		e^^.second := nil;
		e^^.mark := 0;
		e^^.el := x;
		length := length + 1;
		if first = nil then begin
			e^^.left := e;
			e^^.right := e;
			first := e
		end
		else begin
			e^^.left := last;
			e^^.right := first;
			first^^.left := e;
			last^^.right := e
		end;
		last := e;
	end;   { append }


function listMuchLinked.appendWithPointerBack (x: obj): treeElementH;
	begin
(* Returns a pointer to the last added element. *)
		self.append(x);
		appendWithPointerBack := last
	end;   { appendWithPointerBack }

procedure listMuchLinked.appendAndLinkWithOtherList (x: obj; p: treeElementH);
	begin
		append(x);
		last^^.second := p;
		p^^.second := last;
		last^^.mark := 0
	end;   { appendAndLinkWithOtherList }

function listMuchLinked.buildBinaryTree: treeElementH;
	var
		q, listPointer: treeElementH;

	function bin (l, r: seqIndex): treeElementH;
		var
			hs: seqIndex0;
			newL, newR: seqIndex;
			p, pBin: treeElementH;
		begin
			if l = r then begin
				p := listPointer;
				listPointer := listPointer^^.right
			end
			else begin  (* l<>r *)
				p := newTreeElement;
				hs := (r - l) div 2;
				newR := l + hs;
				newL := newR + 1;
				pBin := bin(l, newR);
				pBin^^.ancestor := p;
				p^^.left := pBin;
				p^^.second := nil;
				p^^.mark := 0;
				pBin := bin(newL, r);
				pBin^^.ancestor := p;
				p^^.right := pBin
			end;
			p^^.ancestor := nil;
			bin := p
		end;   { bin }

	begin
(* The tree is built from bottom to top, started out from the list-elements. *)
		listPointer := first;
		q := bin(1, length);
 (* q points now to the root of the binary tree *)
		buildBinaryTree := q
	end;   { buildBinaryTree }

function listMuchLinked.embeddInBinaryTree: binTreeWithList;
	var
		tree: binTreeWithList;
	begin
(* The list self is embedded in the leaves of a complete binary-search-tree. *)
(* If list.length=1 then tree.root=list.first. *)
		new(tree);
		failNil(Handle(tree));
		tree.list := self;
		if length = 1 then
			tree.root := first
		else
			tree.root := self.buildBinaryTree;
		embeddInBinaryTree := tree
	end;   { embeddInBinaryTree }

procedure binTreeWithList.traverse (var newLists: succArray);
	var
		stack: list;
		stackIn, stackOut: markHandleAsObject;
		node: treeElementH;

	procedure getNewLists (x: treeElementH);
		var
			succNumber: markIndex;
		begin
(* Delete x from the old list and put it in one of the successors' list. *)
			succNumber := x^^.mark;
			x^^.mark := 0;
			newLists[succNumber].appendAndLinkWithOtherList(x^^.el, x^^.second);
			list.delete(x);
		end;   { getNewLists }

	begin (* binTreeWithList.traverse *)
(* The binary tree self is traversed in bottom-up order *)
(* 1. left subtree, 2. right subtree, 3. visit root. *)
(* Maximal length of the stack = 2 x list.length *)
		stack := newEmptyList;
		new(stackIn);
		failNil(Handle(stackIn));
		stackIn.tr := root;
		stackIn.k := 1;
		stack.push(stackIn);
		while stack.getLength <> 0 do begin
			stackOut := markHandleAsObject(stack.ith(1));
			stack.pop;   (* only removes but not returns first element *)
			node := stackOut.tr;
			case stackOut.k of
				1:   (* left subtree *)
					begin
					new(stackIn);
					failNil(Handle(stackIn));
					stackIn.tr := node;
					stackIn.k := 2;
					stack.push(stackIn);
					if ((node^^.second = nil) & (node^^.left <> nil)) & (treeElementH(node^^.left)^^.mark <> 0) then (* traverse only marked (=5) points *)
						begin
						new(stackIn);
						failNil(Handle(stackIn));
						stackIn.tr := node^^.left;
						stackIn.k := 1;
						stack.push(stackIn)
					end
				end;
				2:   (* right subtree *)
					begin
					new(stackIn);
					failNil(Handle(stackIn));
					stackIn.tr := node;
					stackIn.k := 3;
					stack.push(stackIn);
					if ((node^^.second = nil) & (node^^.right <> nil)) & (treeElementH(node^^.right)^^.mark <> 0) then  (* traverse only marked (=5) points *)
						begin
						new(stackIn);
						failNil(Handle(stackIn));
						stackIn.tr := node^^.right;
						stackIn.k := 1;
						stack.push(stackIn)
					end
				end;
				3:    (* visit node *)
					if node^^.second <> nil then
			(* Only such nodes are considered, which belong to the list embedded in the tree. *)
						getNewLists(node)
					else begin
						if (node^^.right = nil) & (node^^.left = nil) then begin
				(* If a node has no successors it is deleted from the tree. *)
							if treeElementH(node^^.ancestor)^^.left = node then
								treeElementH(node^^.ancestor)^^.left := nil
							else
								treeElementH(node^^.ancestor)^^.right := nil;
							disposeTreeElement(node)
						end
						else
							node^^.mark := 0;  (* unlabeling *)
					end
			end;
			stackOut.free
		end;  (* while stack.length <> 0 *)
(* If only one element x remains, the tree is a list with first element = root and last element = x. *)
		stack.free
	end;   { traverse }

procedure binTreeWithList.freeAll;
	var
		node, ancestor: treeElementH;
	begin
(* It is supposed that only one element is inserted in the tree's list. *)
(* The path to it and the list is freed. *)
		node := treeElementH(list.first);
		repeat
			ancestor := treeElementH(node^^.ancestor);
			disposeTreeElement(node);
			node := ancestor
		until node = nil;
		list.free;
		shallowFree;
	end;   { freeAll }

procedure binTreeWithList.specialPrepareForFreeing;
	var
		is: seqIndex;
		t, anc, ancanc: treeElementH;
	begin
(* It is supposed that all elements inserted in the trees list are the same onees; *)
(* the elements except the first one are removed and with tem the paths from the root to them. *)
		for is := 2 to listMuchLinked(self.list).length do begin
			t := treeElementH(listMuchLinked(self.list).last);
			anc := treeElementH(t^^.ancestor);
			listMuchLinked(self.list).delete(t);
			while (anc^^.right = nil) & (anc^^.left = nil) do begin
	(* If a node has no successors it is deleted from the tree. *)
				ancanc := treeElementH(anc^^.ancestor);
				if ancanc^^.left = anc then
					treeElementH(ancanc)^^.left := nil
				else
					treeElementH(ancanc)^^.right := nil;
				disposeTreeElement(anc);
				anc := ancanc
			end
		end
	end;

(* *)

procedure list.init;
	begin
		first := nil;
		last := nil;
		length := 0;
	end;

procedure list.reset;
	var
		t, tt: elementH;
	begin
		t := first;
		while t <> nil do begin
			tt := t;
			t := t^^.next;
			disposeElement(tt);
		end;
		init;
	end;

procedure list.free;
	begin
		reset;
		shallowFree;
	end;

function list.duplicate: obj;
	var
		l: list;
		e: elementH;
	begin
		l := newEmptyList;
		e := first;
		while e <> nil do begin
			l.append(e^^.value);
			e := e^^.next;
		end;
		duplicate := l;
	end;

function list.duplicateAll: obj;
	var
		l: list;
		e: elementH;
	begin
		l := newEmptyList;
		e := first;
		while e <> nil do begin
			l.append(e^^.value.duplicateAll);
			e := e^^.next;
		end;
		duplicateAll := l;
	end;

procedure list.fileInput;
	var
		i, l: seqIndex0;
		x: obj;
		r: longint;
	begin
		init;
		readLongInt(r);
		l := r;
		for i := 1 to l do begin
{$IFC AnimationEnabled }
			advanceStopWatchHandsNoMem;
{$ENDC }
			x := newObj(fileGetType);
			x.fileInput;
			append(x);
		end;
	end;

procedure list.textFileInput;
	var
		i, l: seqIndex0;
		r: longint;
	begin
		init;
		readTextLongInt(r);
		l := r;
		for i := 1 to l do begin
{$IFC AnimationEnabled }
			advanceStopWatchHandsNoMem;
{$ENDC }
			append(readTextObj);
		end;
	end;

function list.memberType (o: obj): boolean;
	begin
		memberType := member(o, list);
	end;

function list.sameType: boolean;
	var
		e: elementH;
		f: obj;
	begin
		if length <= 1 then
			sameType := true
		else begin
			f := first^^.value;
			e := first^^.next;
			while (e <> nil) & sameType2(f, e^^.value) do
				e := e^^.next;
			sameType := e = nil;
		end;
	end;

procedure list.append (x: obj);
	var
		e: elementH;
	begin
		e := newElement;
		e^^.next := nil;
		e^^.value := x;
		length := length + 1;
		if first = nil then
			first := e
		else
			last^^.next := e;
		last := e;
	end;

procedure list.revert;
	var
		result, current, temp: elementH;
	begin
		result := nil;
		current := first;
		while current <> nil do begin
			temp := current;
			current := current^^.next;
			temp^^.next := result;
			result := temp;
		end;
		last := first;
		first := result;
	end;

procedure list.removeLast;
	var
		e: elementH;
	begin
		inherited removeLast;
		if length = 0 then begin
			disposeElement(first);
			first := nil;
			last := nil;
		end
		else begin
			e := first;
			while e^^.next <> last do
				e := e^^.next;
			last := e;
			disposeElement(e^^.next);
			e^^.next := nil;
		end;
	end;

procedure list.push (x: obj);
	var
		e: elementH;
	begin
		e := newElement;
		e^^.next := first;
		e^^.value := x;
		length := length + 1;
		if last = nil then
			last := e;
		first := e;
	end;

procedure list.pop;
	var
		e: elementH;
	begin
		e := first;
		first := e^^.next;
		if last = e then
			last := nil;
		length := length - 1;
		disposeElement(e);
	end;

procedure list.delete (e: elementH);
	var
		toDelete: elementH;
(* Delete then element in the list which follows 'e' and must exist. *)
	begin
		toDelete := e^^.next;
		assert(toDelete <> nil, 'toDelete <> nil');
		e^^.next := e^^.next^^.next;
		if e^^.next = nil then
			last := e;
		length := length - 1;
		disposeElement(toDelete);
	end;

procedure list.concatenate (x: list);
	begin
		if x.length > 0 then begin
			length := length + x.length;
			if first = nil then
				first := x.first
			else
				last^^.next := x.first;
			last := x.last;
		end;
	end;

procedure list.concatenateDispose (x: list);
	begin
		concatenate(x);
		dispose(x);
	end;

function list.ith (i: seqIndex): obj;
	var
		j: longInt;
		e: elementH;
	begin
		if i >= 50 then
			fatalError(StringOf('list.ith  50. Got ', i : 1, '.'));
		e := first;
		for j := 1 to i - 1 do
			e := e^^.next;
		ith := e^^.value;
	end;


procedure list.forSome (procedure whatToDo (x: obj); l: seqIndex; r: seqIndex0);
	var
		el: elementH;
		i: seqIndex1;
	begin
		if l <= r then begin
			el := first;
			i := 1;
			while (i < l) & (el <> nil) do begin
				el := el^^.next;
				i := i + 1;
			end;
			while (i <= r) & (el <> nil) do begin
				whatToDo(el^^.value);
				el := el^^.next;
				i := i + 1;
			end;
		end;
	end;

procedure list.forAll (procedure whatToDo (x: obj));
	var
		el: elementH;
	begin
		el := first;
		while el <> nil do begin
			whatToDo(el^^.value);
			el := el^^.next;
		end;
	end;

function list.sameProperty (function property (x: obj): boolean): boolean;
	var
		el: elementH;
	begin
		el := first;
		while (el <> nil) & property(el^^.value) do
			el := el^^.next;
		sameProperty := el = nil;
	end;

function list.exists (function property (x: obj): boolean): seqIndex0;
	var
		el: elementH;
		result: seqIndex1;
	begin
		el := first;
		result := 1;
		while (el <> nil) & not property(el^^.value) do begin
			el := el^^.next;
			result := result + 1;
		end;
		if el = nil then
			exists := 0
		else
			exists := result;
	end;

function list.existsObject (function property (x: obj): boolean): obj;
	var
		el: elementH;
	begin
		el := first;
		while (el <> nil) & not property(el^^.value) do
			el := el^^.next;
		if el = nil then
			existsObject := nil
		else
			existsObject := el^^.value;
	end;

function list.getType: objectType;
	begin
		getType := listType;
	end;

function list.getTextType: Str255;
	begin
		getTextType := stringOf('List(', length : 1, ')');
	end;

procedure list.cyclicShift (newFirstElement: seqIndex);
	var
		i: seqIndex0;
	begin
		assert((1 <= newFirstElement) & (newFirstElement <= length), 'newFirstElement in range');
		for i := 1 to newFirstElement - 1 do begin
			last^^.next := first;
			first := first^^.next;
			last := last^^.next;
		end;
		last^^.next := nil;
	end;

procedure doubleList.init;
	begin
		first := nil;
		last := nil;
	end;

procedure doubleList.free;
	var
		t, tt: doubleElementH;
	begin
		t := first;
		while t <> nil do begin
			tt := t;
			t := t^^.succ;
			disposeDoubleElement(tt);
		end;
		shallowFree;
	end;

procedure doubleList.append (x: obj);
	var
		newD: doubleElementH;
	begin
(* changed by PV *)
		newD := newDoubleElement;
		with newD^^ do begin
			value := x;
			pred := last;
			succ := nil;
		end;
		if first = nil then
			first := newD
		else
			last^^.succ := newD;
		last := newD;
	end;

procedure doubleList.moveToFront (var d: doubleElementH);
	var
		succD, predD: doubleElementH;
	begin
		if d^^.pred <> nil then begin
			predD := d^^.pred;
			succD := d^^.succ;
			predD^^.succ := succD;
			if succD = nil then
				last := predD
			else
				succD^^.pred := predD;
			d^^.succ := first;
			d^^.pred := nil;
			first := d;
			d := predD;
		end;
	end;

(* Inserted by PV *)

procedure doubleList.freeAll;
	begin
		forAll(freeAllP);
		free;
	end;   { freeAll }

function doubleList.appendWithPointer (x: obj): doubleElementH;
	begin
(* Returns apointer to the last added element. *)
		self.append(x);
		appendWithPointer := last;
	end;   { appendWithPointer }

procedure doubleList.delete (d: doubleElementH);
	var
		succD, predD: doubleElementH;
	begin
		if d^^.pred <> nil then begin
			predD := d^^.pred;
			succD := d^^.succ;
			predD^^.succ := succD;
			if succD = nil then
				last := predD
			else
				succD^^.pred := predD;
			disposeDoubleElement(d);
		end
		else begin
			first := d^^.succ;
			if first <> nil then
				first^^.pred := nil;
(* **************** *)
			if last = d then    (* 20/SEP/93 RICDIC: d is first and last element *)
				last := nil;         (* without this : last will point to somewhere, the next append (see above) pred := last *)
(* ********************* *)
			disposeDoubleElement(d);
		end;
	end;   { delete }

procedure doubleList.setLast (x: obj);
	begin
		last^^.value := x;
	end;   { setLast }

procedure doubleList.forAll (procedure whatToDo (x: obj));
	var
		el: doubleElementH;
	begin
		el := first;
		while el <> nil do begin
			whatToDo(el^^.value);
			el := el^^.succ;
		end;
	end;

(* *)

procedure subMenuElement.free;
	begin
		menu.free;
		shallowFree;
	end;

procedure subMenuElement.freeAll;
	begin
		menu.freeAll;
		shallowFree;
	end;

	var
		freeList: elementH;
		freeListN: longInt;
(* inserted by PV *)
		freeDoubleList: doubleElementH;
		freeDoubleListN: longInt;
		freeTreeListN: longInt;
		freeTreeList: treeElementH;
(* *)

function newElement: elementH;
	var
		t: elementH;
	begin
		if freeList = nil then begin
			t := elementH(newHandleFailNil(sizeOf(element)));
			newElement := t;
		end
		else begin
			newElement := freeList;
			freeList := freeList^^.next;
			freeListN := freeListN - 1;
		end;
	end;

procedure disposeElement (e: elementH);
	begin
		e^^.next := freeList;
		freeList := e;
		freeListN := freeListN + 1;
	end;

function freeListSize: longInt;
	begin
		freeListSize := freeListN * (sizeOf(element) + 8);
	(* '+ 8' is necessary to reflect the correct amount of memory used. '+0' causes "loss" of memory. *)
	end;

function freeListLength: longInt;
	begin
		freeListLength := freeListN;
	end;

procedure extendFreeList (number: longInt);
	var
		i: longInt;
		e: elementH;
	begin
		for i := 1 to number do begin
{$IFC AnimationEnabled }
			advanceStopWatchHandsNoMem;
{$ENDC }
			e := elementH(newHandleFailNil(sizeOf(element)));
			e^^.next := freeList;
			freeList := e;
		end;
		freeListN := freeListN + number;
	end;

procedure shrinkFreeList (number: longInt);
	var
		i: longInt;
		e: elementH;
	begin
		if number > freeListN then
			number := freeListN;
		for i := 1 to number do begin
{$IFC AnimationEnabled }
			advanceStopWatchHandsNoMem;
{$ENDC }
			e := freeList;
			freeList := freeList^^.next;
{$IFC THINK_PASCAL }
			disposHandle(Handle(e));
{$ELSEC }
			DisposeHandle(Handle(e));
{$ENDC }
		end;
		freeListN := freeListN - number;
	end;

(* Inserted by PV *)

function newTreeElement: treeElementH;
	begin
		if freeTreeList = nil then
			newTreeElement := treeElementH(newHandleFailNil(sizeof(treeElement)))
		else begin
			newTreeElement := freeTreeList;
			freeTreeList := freeTreeList^^.right;
			freeTreeListN := freeTreeListN - 1;
		end;
	end;

procedure disposeTreeElement (e: treeElementH);
	begin
		e^^.right := freeTreeList;
		freeTreeList := e;
		freeTreeListN := freeTreeListN + 1;
	end;

function freeTreeListSize: longInt;
	begin
		freeTreeListSize := freeTreeListN * (sizeOf(treeElement) + 8);
	end;

function freeTreeListLength: longInt;
	begin
		freeTreeListLength := freeTreeListN;
	end;

procedure extendFreeTreeList (number: longInt);
	var
		i: longInt;
		e: treeElementH;
	begin
		for i := 1 to number do begin
			e := treeElementH(newHandleFailNil(sizeof(treeElement)));
			e^^.right := freeTreeList;
			freeTreeList := e;
		end;
		freeTreeListN := freeTreeListN + number;
	end;

procedure shrinkFreeTreeList (number: longInt);
	var
		i: longInt;
		e: treeElementH;
	begin
		if number > freeTreeListN then
			number := freeTreeListN;
		for i := 1 to number do begin
			e := freeTreeList;
			freeTreeList := freeTreeList^^.right;
{$IFC THINK_PASCAL }
			disposHandle(Handle(e));
{$ELSEC }
			DisposeHandle(Handle(e));
{$ENDC }
		end;
		freeTreeListN := freeTreeListN - number;
	end;



function newDoubleElement: doubleElementH;
	begin
		if freeDoubleList = nil then
			newDoubleElement := doubleElementH(newHandleFailNil(sizeof(doubleElement)))
		else begin
			newDoubleElement := freeDoubleList;
			freeDoubleList := freeDoubleList^^.succ;
			freeDoubleListN := freeDoubleListN - 1;
		end;
	end;

procedure disposeDoubleElement (e: doubleElementH);
	begin
		e^^.succ := freeDoubleList;
		freeDoubleList := e;
		freeDoubleListN := freeDoubleListN + 1;
	end;

function freeDoubleListSize: longInt;
	begin
		freeDoubleListSize := freeDoubleListN * (sizeOf(doubleElement) + 8);
	end;

function freeDoubleListLength: longInt;
	begin
		freeDoubleListLength := freeDoubleListN;
	end;

procedure extendFreeDoubleList (number: longInt);
	var
		i: longInt;
		e: doubleElementH;
	begin
		for i := 1 to number do begin
			e := doubleElementH(newHandleFailNil(sizeof(doubleElement)));
			e^^.succ := freeDoubleList;
			freeDoubleList := e;
		end;
		freeDoubleListN := freeDoubleListN + number;
	end;

procedure shrinkFreeDoubleList (number: longInt);
	var
		i: longInt;
		e: doubleElementH;
	begin
		if number > freeDoubleListN then
			number := freeDoubleListN;
		for i := 1 to number do begin
			e := freeDoubleList;
			freeDoubleList := freeDoubleList^^.succ;
{$IFC THINK_PASCAL }
			disposHandle(Handle(e));
{$ELSEC }
			DisposeHandle(Handle(e));
{$ENDC }
		end;
		freeDoubleListN := freeDoubleListN - number;
	end;

(* *)

function newVector (length: longInt): vector;
	var
		v: vector;
	begin
		currentVectorLength := length;
		new(v);
		failNil(Handle(v));
		v.init;
		newVector := v;
	end;

function newHomogenVector (length: longInt): homogenVector;
	var
		v: homogenVector;
	begin
		currentVectorLength := length;
		new(v);
		failNil(Handle(v));
		v.init;
		newHomogenVector := v;
	end;

procedure appendObjInfoObject (s: sequence; name: string; value: obj; protection: boolean);
	var
		descO: objInfoObj;
	begin
		new(descO);
		failNil(Handle(descO));
		with descO do begin
			desc := name;
			o := value;
			protect := protection;
		end;
		s.appendSafe(descO, 1);
	end;

{$IFC THINK_PASCAL }
procedure appendObjInfoTempObject (s: sequence; name: string; value: obj; protection: boolean; doCleanUp: ProcPtr);
{$ELSEC }
procedure appendObjInfoTempObject (s: sequence; name: string; value: obj; protection: boolean; doCleanUp: cleanUpT);
{$ENDC }
	var
		descO: objInfoTempObj;
	begin
		new(descO);
		failNil(Handle(descO));
		with descO do begin
			desc := name;
			o := value;
			protect := protection;
		end;
		descO.cleanUp := doCleanUp;
		s.appendSafe(descO, 1);
	end;

procedure appendObjInfoExtended (s: sequence; name: string; value: extended; protection: boolean);
	var
		descE: objInfoExtended;
	begin
		new(descE);
		failNil(Handle(descE));
		with descE do begin
			desc := name;
			e := value;
			protect := protection;
		end;
		s.appendSafe(descE, 1);
	end;


procedure appendObjInfoString (s: sequence; name: string; value: string; protection: boolean);
	var
		descS: objInfoString;
	begin
		new(descS);
		failNil(Handle(descS));
		with descS do begin
			desc := name;
			s := value;
			protect := protection;
		end;
		s.appendSafe(descS, 1);
	end;

function dynamicArray (number, size: longInt): Handle;
	begin
		dynamicArray := failNilF(dynamicArrayNoCheck(number, size));
	end;

function dynamicArrayNoCheck (number, size: longInt): Handle;
	begin
		dynamicArrayNoCheck := NewHandle(number * size);
	end;

	const
		maxMenuStack = 10;

	var
		menuStack: array[1..maxMenuStack] of list;
		menuStackPointer: 0..maxMenuStack;
{$IFC ScriptEnabled }
		menuName: Str255;
{$ENDC }

procedure initCurrentMenu;
	begin
		new(currentMenu);
		failNil(Handle(currentMenu));
		currentMenu.init;
		menuStackPointer := 1;
		menuStack[1] := currentMenu;
	end;

{$IFC ScriptEnabled }
procedure initApplicableFcts; (* M_SCM *)
	begin
		menuName := 'menu';
		new(applicableFcts);
		failNil(Handle(applicableFcts));
		applicableFcts.init;
	end;
{$ENDC }

procedure menuEntry (entry: Str255; item: integer); (* M_SCM *)
	var
		m: menuEntryElement;
{$IFC ScriptEnabled }
		eF: executableFct;
{$ENDC }
	begin
{$IFC ScriptEnabled }
		if calledFromGeoScript then begin
			new(eF);
			failNil(Handle(eF));
			eF.init;
			eF.menu := entry;
			eF.item := item;
			eF.title := menuName;
			list(applicableFcts).append(eF);
		end
		else begin
{$ENDC }
			new(m);
			failNil(Handle(m));
			m.entry := entry;
			m.item := item;
			menuStack[menuStackPointer].append(m);
{$IFC ScriptEnabled }
		end;
{$ENDC }
	end;

procedure beginSubMenu (title: Str255);
	var
		s: subMenuElement;
		l: list;
	begin
{$IFC ScriptEnabled }
		if calledFromGeoScript then
			menuName := title
		else begin
{$ENDC }
			l := newEmptyList;
			new(s);
			failNil(Handle(s));
			s.title := title;
			s.menu := l;
			menuStack[menuStackPointer].append(s);
			assert(menuStackPointer <= maxMenuStack, 'menuStackPointer <= maxMenuStack');
			menuStackPointer := menuStackPointer + 1;
			menuStack[menuStackPointer] := l
{$IFC ScriptEnabled }
		end;
{$ENDC }
	end;

procedure endSubMenu; (* M_SCM *)
	begin
	{$IFC ScriptEnabled }
		if calledFromGeoScript then
			menuName := 'menu'
		else begin
		{$ENDC }
			assert(menuStackPointer > 0, 'menuStackPointer > 0');
			menuStackPointer := menuStackPointer - 1;
	{$IFC ScriptEnabled }
		end;
		{$ENDC }
	end;

procedure initSequence;
	var
		s: sequence;
		x: vector;
		y: homogenVector;
		z: list;
		a: listMuchLinked;
	begin
		freeList := nil;
		freeListN := 0;
(* inserted by PV *)
		freeTreeList := nil;
		freeTreeListN := 0;
(* *)
		currentMenu := nil;
		new(s);
		failNil(Handle(s));
		prototypes[sequenceType] := s;
		new(x);
		failNil(Handle(x));
		prototypes[vectorType] := x;
		new(y);
		failNil(Handle(y));
		prototypes[homogenVectorType] := y;
		new(z);
		failNil(Handle(z));
		prototypes[listType] := z;
(* inserted by PV *)
		new(a);
		failNil(Handle(a));
		prototypes[listMuchLinkedType] := a;
	end;


(* inserted by PV *)

procedure initDoubleList;
	var
		d: doubleList;
	begin
		freeDoubleList := nil;
		freeDoubleListN := 0;
		new(d);
		failNil(Handle(d));
		prototypes[doubleListType] := d
	end;   { initDoubleList }

(* *)


end.
