(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit sortedList;

interface

uses
	dictionary;

implementation

procedure traverseFree (p: sortedListNodeH);
	var
		pp: sortedListNodeH;
	begin
		while p <> nil do begin
			pp := p^^.sons[right];
{$IFC THINK_PASCAL }
			disposHandle(Handle(p));
{$ELSEC }
			DisposeHandle(Handle(p));
{$ENDC }
			p := pp;
		end;
	end;

procedure sortedList.reset;
	begin
		traverseFree(minimum);
		init;
	end;

procedure sortedList.init;
	begin
		minimum := nil;
		maximum := nil;
		inherited init;
	end;

procedure sortedList.free;
	begin
		traverseFree(minimum);
		shallowFree;
	end;

procedure sortedList.interactiveOutput;
	var
		p: sortedListNodeH;
	begin
		p := minimum;
		if p = nil then
			writeLn('minimum = NIL')
		else
			while p <> nil do begin
				if p = minimum then
					write('MIN ')
				else if p = maximum then
					write('MAX ')
				else
					write('    ');
				p^^.value.interactiveOutput;
				writeLn;
				p := p^^.sons[right];
			end;
		writeLn;
	end;

procedure sortedList.textOutput;
	var
		p: sortedListNodeH;
	begin
		p := minimum;
		while p <> nil do begin
			p^^.value.textOutput;
			writeLn;
			p := p^^.sons[right];
		end;
	end;

function sortedList.getObject (where: reference): obj;
	begin
		assert(where <> nil, 'where  nil in sortedList.getObject');
		getObject := sortedListNodeH(where)^^.value;
	end;

procedure sortedList.setObject (where: reference; newValue: obj);
	begin
		sortedListNodeH(where)^^.value := newValue;
	end;

function sortedList.findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
	var
		p: sortedListNodeH;
	begin
		p := minimum;
		position := 0;
		while (p <> nil) & (compare(x, p^^.value) = greaterThanObj) do begin
			p := p^^.sons[right];
			position := position + 1;
		end;
		if p = nil then begin
			where := reference(maximum);
			direction := right;
			findPosition := false;
		end
		else begin
			where := reference(p);
			direction := left;
			if compare(x, p^^.value) = equalObj then begin
				findPosition := true;
				position := position + 1;
			end
			else
				findPosition := false;
		end;
	end;

function sortedList.getPosition (x: reference): longInt;
	var
		p: sortedListNodeH;
		pos: longInt;
	begin
		p := minimum;
		pos := 1;
		while p <> sortedListNodeH(x) do begin
			p := p^^.sons[right];
			pos := pos + 1;
		end;
		getPosition := pos;
	end;

function sortedList.ith (i: longInt): reference;
	var
		p: sortedListNodeH;
	begin
		assert((1 <= i) & (i <= numberOfElements), 'Range ok');
		p := minimum;
		while i > 1 do begin
			p := p^^.sons[right];
			i := i - 1;
		end;
		ith := reference(p);
	end;

function sortedList.insert (x: obj; where: reference; direction: directionType): reference;
	var
		p, t: sortedListNodeH;
	begin
		t := sortedListNodeH(newHandleFailNil(sizeOf(sortedListNode)));
		insert := reference(t);
		t^^.value := x;
		t^^.sons[switchDirection(direction)] := sortedListNodeH(where);
		if where = nil then begin
			t^^.sons[direction] := nil;
			minimum := t;
			maximum := t;
		end
		else begin
			p := sortedListNodeH(where)^^.sons[direction];
			t^^.sons[direction] := p;
			sortedListNodeH(where)^^.sons[direction] := t;
			if p <> nil then
				p^^.sons[switchDirection(direction)] := t
			else if direction = left then
				minimum := t
			else
				maximum := t;
		end;
		numberOfElements := numberOfElements + 1;
	end;

procedure sortedList.delete (x: reference);
	begin
		if (sortedListNodeH(x) = minimum) & (sortedListNodeH(x) = maximum) then begin
			minimum := nil;
			maximum := nil;
		end
		else if sortedListNodeH(x) = minimum then begin
			minimum := sortedListNodeH(x)^^.sons[right];
			minimum^^.sons[left] := nil;
		end
		else if sortedListNodeH(x) = maximum then begin
			maximum := sortedListNodeH(x)^^.sons[left];
			maximum^^.sons[right] := nil;
		end
		else begin
			sortedListNodeH(x)^^.sons[left]^^.sons[right] := sortedListNodeH(x)^^.sons[right];
			sortedListNodeH(x)^^.sons[right]^^.sons[left] := sortedListNodeH(x)^^.sons[left];
		end;
{$IFC THINK_PASCAL }
		disposHandle(Handle(x));
{$ELSEC }
		DisposeHandle(Handle(x));
{$ENDC }
		numberOfElements := numberOfElements - 1;
	end;

function sortedList.next (x: reference; direction: directionType): reference;
	begin
		if x = nil then
			next := nil
		else
			next := reference(sortedListNodeH(x)^^.sons[direction])
	end;

function sortedList.extreme (direction: directionType): reference;
	begin
		if direction = left then
			extreme := reference(minimum)
		else
			extreme := reference(maximum)
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure sortedList.sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
	begin
		unimplementedMethod('sortedList.sequenceToDictionary');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure sortedList.swap (p, q: reference);

	procedure exchangeDisjoint (p, q: sortedListNodeH);
		var
			pLeft, pRight, qLeft, qRight: sortedListNodeH;

		procedure halfExchange (x, y, xLeft, xRight, yLeft, yRight: sortedListNodeH);
			begin
				if xLeft = nil then
					minimum := y
				else
					xLeft^^.sons[right] := y;
				x^^.sons[left] := yLeft;
				x^^.sons[right] := yRight;
				if xRight = nil then
					maximum := y
				else
					xRight^^.sons[left] := y;
			end;

		begin
			pLeft := p^^.sons[left];
			pRight := p^^.sons[right];
			qLeft := q^^.sons[left];
			qRight := q^^.sons[right];
			halfExchange(p, q, pLeft, pRight, qLeft, qRight);
			halfExchange(q, p, qLeft, qRight, pLeft, pRight);
		end;

	procedure exchangeNonDisjoint (p, q: sortedListNodeH);
		var
			pLeft, qRight: sortedListNodeH;
		begin
			pLeft := p^^.sons[left];
			qRight := q^^.sons[right];
			if pLeft = nil then
				minimum := q
			else
				pLeft^^.sons[right] := q;
			p^^.sons[left] := q;
			p^^.sons[right] := qRight;
			q^^.sons[left] := pLeft;
			q^^.sons[right] := p;
			if qRight = nil then
				maximum := p
			else
				qRight^^.sons[left] := p;
		end;

	begin
		if sortedListNodeH(p)^^.sons[left] = sortedListNodeH(q) then
			exchangeNonDisjoint(sortedListNodeH(q), sortedListNodeH(p))
		else if sortedListNodeH(p)^^.sons[right] = sortedListNodeH(q) then
			exchangeNonDisjoint(sortedListNodeH(p), sortedListNodeH(q))
		else
			exchangeDisjoint(sortedListNodeH(p), sortedListNodeH(q));
	end;

procedure sortedList.rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
	var
		p: sortedListNodeH;
	begin
		if direction = left then begin
			p := minimum;
			while (p <> nil) & (leftBound <> nil) & (compare(p^^.value, leftBound) = lessThanObj) do
				p := p^^.sons[right];
			while (p <> nil) & ((rightBound = nil) | (compare(p^^.value, rightBound) < greaterThanObj)) do begin
				whatToDo(p^^.value);
				p := p^^.sons[right];
			end;
		end
		else begin
			p := maximum;
			while (p <> nil) & (rightBound <> nil) & (compare(p^^.value, rightBound) = greaterThanObj) do
				p := p^^.sons[left];
			while (p <> nil) & ((leftBound = nil) | (compare(p^^.value, leftBound) > lessThanObj)) do begin
				whatToDo(p^^.value);
				p := p^^.sons[left];
			end;
		end;
	end;

procedure sortedList.forAll (procedure whatToDo (x: obj); direction: directionType);
	var
		p: sortedListNodeH;
	begin
		if direction = left then
			p := minimum
		else
			p := maximum;
		direction := switchDirection(direction);
		while p <> nil do begin
			whatToDo(p^^.value);
			p := p^^.sons[direction];
		end;
	end;

function sortedList.exists (function property (x: obj): boolean): reference;
	var
		p: sortedListNodeH;
	begin
		p := minimum;
		while (p <> nil) & not property(p^^.value) do
			p := p^^.sons[right];
		exists := reference(p);
	end;

function sortedList.compareReference (p, q: reference): relationObj;
	var
		t: sortedListNodeH;
	begin
		if p = q then
			compareReference := equalObj
		else begin
			t := minimum;
			while (t <> sortedListNodeH(p)) & (t <> sortedListNodeH(q)) do
				t := t^^.sons[right];
			if t = sortedListNodeH(p) then
				compareReference := lessThanObj
			else
				compareReference := greaterThanObj;
		end;
	end;


end.
