(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit sortedVector;

interface

uses
	dictionary;

implementation

procedure sortedVector.allocateElements (number: seqIndex0);
	begin
		masters := refPairArrayH(dynamicArray(number + 1, sizeOf(refPair)));
		vector := seqIndex0ArrayH(dynamicArray(number + 1, sizeOf(seqIndex0)));
		masters^^[1].ref := 0;
	end;

procedure sortedVector.reset;
	begin
		freeList := 1;
		masters^^[1].ref := 0;
		inherited init;
	end;

procedure sortedVector.init;
	begin
		capacity := currentVectorLength;
		freeList := 1;
		allocateElements(currentVectorLength);
		inherited init;
	end;

procedure sortedVector.free;
	begin
		if masters <> nil then
{$IFC THINK_PASCAL }
			disposHandle(Handle(masters));
		if vector <> nil then
			disposHandle(Handle(vector));
{$ELSEC }
			DisposeHandle(Handle(masters));
		if vector <> nil then
			DisposeHandle(Handle(vector));
{$ENDC }
		shallowFree;
	end;

procedure sortedVector.interactiveOutput;
	var
		i: seqIndex0;
	begin
		for i := 1 to numberOfElements do
			ithObject(i).interactiveOutput;
	end;

procedure sortedVector.textOutput;
	var
		i: seqIndex0;
	begin
		for i := 1 to numberOfElements do begin
			ithObject(i).textOutput;
			writeLn;
		end;
		writeLn;
	end;

function sortedVector.getObject (where: reference): obj;
	begin
		assert(where <> nil, 'where  nil in sortedVector.getObject');
		getObject := masters^^[seqIndex(where)].value;
	end;

procedure sortedVector.setObject (where: reference; newValue: obj);
	begin
		masters^^[seqIndex(where)].value := newValue;
	end;

function sortedVector.findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
	var
		i: seqIndex1;
	begin
		if numberOfElements = 0 then begin
			where := nil;
			findPosition := false
		end
		else begin
			i := 1;
			while (i <= numberOfElements) & (compare(x, ithObject(i)) = greaterThanObj) do
				i := i + 1;
			position := i - 1;
			if i > numberOfElements then begin
				where := reference(vector^^[numberOfElements]);
				direction := right;
				findPosition := false;
			end
			else begin
				where := reference(vector^^[i]);
				direction := left;
				if compare(x, ithObject(i)) = equalObj then begin
					position := position + 1;
					findPosition := true;
				end
				else
					findPosition := false;
			end;
		end;
	end;

function sortedVector.getPosition (x: reference): longInt;
	begin
		getPosition := masters^^[seqIndex(x)].ref;
	end;

function sortedVector.insert (x: obj; where: reference; direction: directionType): reference;
	var
		i, startPosition: seqIndex0;
	procedure increment (var x: seqIndex0);
		begin
			x := x + 1;
		end;
	begin
		assert(numberOfElements < capacity, 'Dictionary has space');
		if where = nil then
			startPosition := 1
		else
			startPosition := seqIndex(masters^^[seqIndex(where)].ref) + ord(direction);
		for i := numberOfElements downto startPosition do begin
			increment(masters^^[vector^^[i]].ref);
			vector^^[i + 1] := vector^^[i];
		end;
		i := freeList;
		if masters^^[freeList].ref = 0 then begin
			freeList := freeList + 1;
			masters^^[freeList].ref := 0;
		end
		else
			freeList := masters^^[freeList].ref;
		insert := reference(i);
		with masters^^[i] do begin
			value := x;
			ref := startPosition;
		end;
		vector^^[startPosition] := i;
		numberOfElements := numberOfElements + 1;
	end;

procedure sortedVector.delete (x: reference);
	var
		i, startPosition: seqIndex0;
	procedure decrement (var x: seqIndex0);
		begin
			x := x - 1;
		end;
	begin
		startPosition := masters^^[seqIndex(x)].ref;
		for i := startPosition + 1 to numberOfElements do begin
			decrement(masters^^[vector^^[i]].ref);
			vector^^[i - 1] := vector^^[i];
		end;
		masters^^[seqIndex(x)].ref := freeList;
		freeList := seqIndex(x);
		numberOfElements := numberOfElements - 1;
	end;

function sortedVector.next (x: reference; direction: directionType): reference;
	var
		t: seqIndex;
	begin
		if x = nil then
			next := nil
		else begin
			t := masters^^[seqIndex(x)].ref;
			if direction = left then
				if t = 1 then
					next := nil
				else
					next := reference(vector^^[t - 1])
			else if t = numberOfElements then
				next := nil
			else
				next := reference(vector^^[t + 1]);
		end;
	end;

function sortedVector.extreme (direction: directionType): reference;
	begin
		if numberOfElements = 0 then
			extreme := nil
		else if direction = left then
			extreme := reference(vector^^[1])
		else
			extreme := reference(vector^^[numberOfElements]);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure sortedVector.sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
	begin
		unimplementedMethod('sortedVector.sequenceToDictionary');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure sortedVector.swap (p, q: reference);
	var
		pRef, qRef: seqIndex;
	begin
		pRef := masters^^[seqIndex(p)].ref;
		qRef := masters^^[seqIndex(q)].ref;
		vector^^[pRef] := seqIndex(q);
		vector^^[qRef] := seqIndex(p);
		masters^^[seqIndex(p)].ref := qRef;
		masters^^[seqIndex(q)].ref := pRef;
	end;

procedure sortedVector.rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
	var
		i: seqIndex01;
	begin
		if direction = left then begin
			i := 1;
			while (i <= numberOfElements) & (leftBound <> nil) & (compare(ithObject(i), leftBound) = lessThanObj) do
				i := i + 1;
			while (i <= numberOfElements) & ((rightBound = nil) | (compare(ithObject(i), rightBound) < greaterThanObj)) do begin
				whatToDo(ithObject(i));
				i := i + 1;
			end;
		end
		else begin
			i := numberOfElements;
			while (i > 0) & (rightBound <> nil) & (compare(ithObject(i), rightBound) = greaterThanObj) do
				i := i - 1;
			while (i > 0) & ((leftBound = nil) | (compare(ithObject(i), leftBound) > lessThanObj)) do begin
				whatToDo(ithObject(i));
				i := i - 1;
			end;
		end;
	end;

procedure sortedVector.forAll (procedure whatToDo (x: obj); direction: directionType);
	var
		i: seqIndex0;
	begin
		if direction = left then
			for i := numberOfElements downto 1 do
				whatToDo(ithObject(i))
		else
			for i := 1 to numberOfElements do
				whatToDo(ithObject(i));
	end;

function sortedVector.exists (function property (x: obj): boolean): reference;
	var
		i: seqIndex1;
	begin
		i := 1;
		while (i <= numberOfElements) & not property(ithObject(i)) do
			i := i + 1;
		if i > numberOfElements then
			exists := nil
		else
			exists := ith(i);
	end;

function sortedVector.compareReference (p, q: reference): relationObj;
	begin
		if masters^^[seqIndex(p)].ref < masters^^[seqIndex(q)].ref then
			compareReference := lessThanObj
		else if masters^^[seqIndex(p)].ref > masters^^[seqIndex(q)].ref then
			compareReference := greaterThanObj
		else
			compareReference := equalObj;
	end;

function sortedVector.ithObject (i: seqIndex): obj;
	begin
		ithObject := masters^^[vector^^[i]].value;
	end;

function sortedVector.ith (i: longInt): reference;
	begin
		assert((1 <= i) & (i <= numberOfElements), 'Range ok');
		ith := reference(vector^^[i]);
	end;

end.
