(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit splinegon;

{$SETC ASplinegon = TRUE }

interface

uses
	pointVector;

type
	splinegon = object(vector)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function duplicateAll: obj;
			override;
			procedure reflect (a, b, c: extended);{specially for simple splinegons}
			override;
			function equal (o: obj): boolean;
			override;
			procedure textFileInput;
			override;
			procedure fileInput;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function sameType: boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			function memberType (o: obj): boolean;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
			function carrierPolygon: polygon2d;
			function boundingPolygon: polygon2d;
			function convexHull: splinegon;
		end;

	curvedEdge = object(obj)
			v1, v2, pv: realPoint;
			procedure free;
			override;
			function duplicate: obj;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function getTextType: Str255;
			override;
			function getInfo: obj;
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure tangent (p: realPoint; b: boolean; tv, tpv1, tpv2: realPoint);
{b determines whether the tangent point closer to v2 is taken}
			procedure commonTangent (c: curvedEdge; tv1, tv2, tpv11, tpv12, tpv21, tpv22: realPoint);
			function xExtremeX: extended;
		end;

	spline2 = object(curvedEdge)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure interactiveOutput;
			override;
			procedure tangent (p: realPoint; b: boolean; tv, tpv1, tpv2: realPoint);
			override;
			function xExtremeX: extended;
			override;
			function memberType (o: obj): boolean;
			override;
		end;

	straightEdge = object(curvedEdge)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure interactiveOutput;
			override;
			function isSelected (where: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
		end;

procedure initSplinegon;

implementation

const
	smoothC = 0;
	carrierPolygonC = 1;
	boundingPolygonC = 2;
	spline2gonC = 3;
	ConvexHullC = 4;

	renumberMinusC = 5;
	renumberPlusC = 6;
	revertC = 7;

function splinegon.memberType (o: obj): boolean;
	begin
		memberType := member(o, splinegon);
	end;

function curvedEdge.memberType (o: obj): boolean;
	begin
		memberType := member(o, curvedEdge);
	end;

function spline2.memberType (o: obj): boolean;
	begin
		memberType := member(o, spline2);
	end;

function straightEdge.memberType (o: obj): boolean;
	begin
		memberType := member(o, straightEdge);
	end;

function curvedEdge.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(3);
		appendObjInfoObject(v, 'V1', v1, true);
		appendObjInfoObject(v, 'V2', v2, true);
		appendObjInfoObject(v, 'PV', pv, true);
		getInfo := v;
	end;

function curvedEdge.getTextType: Str255;
	begin
		getTextType := 'Curved Edge';
	end;

function spline2.getTextType: Str255;
	begin
		getTextType := 'Spline2';
	end;

function straightEdge.getTextType: Str255;
	begin
		getTextType := 'Straight Edge';
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function splinegon.pointCover (n: longInt): obj;
	begin
		pointCover := nil;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function splinegon.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, splinegon) & inherited equal(o);
	end;

function curvedEdge.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, curvedEdge) & v1.equal(curvedEdge(o).v1) & v2.equal(curvedEdge(o).v2) & pv.equal(curvedEdge(o).pv);
	end;

function curvedEdge.genericLessThan (o: obj): boolean;
	var
		c: curvedEdge;
	begin
		c := curvedEdge(o);
		genericLessThan := v1.genericLessThan(c.v1) | v1.equal(c.v1) & (v2.genericLessThan(c.v2) | v2.equal(c.v2) & pv.genericLessThan(c.pv));
	end;

procedure connectEdges (s: splinegon);
	var
		i: seqIndex0;
	begin
		for i := 2 to s.length do
			curvedEdge(s.elements^^[i]).v1 := curvedEdge(s.elements^^[i - 1]).v2;
		curvedEdge(s.elements^^[1]).v1 := curvedEdge(s.elements^^[s.length]).v2;
	end;

procedure connectEdgesBack (s: splinegon);
	var
		i: seqIndex0;
	begin
		for i := 2 to s.length do
			curvedEdge(s.elements^^[i - 1]).v2 := curvedEdge(s.elements^^[i]).v1;
		curvedEdge(s.elements^^[s.length]).v2 := curvedEdge(s.elements^^[1]).v1;
	end;

procedure renumberMinus (s: splinegon);
	var
		ce: obj;
		i: integer;
	begin
		ce := s.elements^^[1];
		for i := 1 to s.length - 1 do
			s.elements^^[i] := s.elements^^[i + 1];
		s.elements^^[s.length] := ce;
	end;

procedure renumberPlus (s: splinegon);
	var
		ce: obj;
		i: integer;
	begin
		ce := s.elements^^[s.length];
		for i := s.length downto 2 do
			s.elements^^[i] := s.elements^^[i - 1];
		s.elements^^[1] := ce;
	end;

procedure revertSG (s: splinegon);
	var
		tmp: obj;
		i: integer;
	begin
		for i := 1 to s.length do
			with curvedEdge(s.elements^^[i]) do begin
				tmp := v1;
				v1 := v2;
				v2 := realPoint(tmp);
			end;
		for i := 1 to s.length div 2 do begin
			tmp := s.elements^^[i];
			s.elements^^[i] := s.elements^^[s.length + 1 - i];
			s.elements^^[s.length + 1 - i] := tmp;
		end;
	end;

(*  *)

function splinegon.getType: objectType;
	begin
		getType := splinegonType;
	end;

function splinegon.getTextType: Str255;
	begin
		getTextType := stringOf('Splinegon(', length : 1, ')');
	end;

function splinegon.duplicateAll: obj;
	var
		s: splinegon;
	begin
		s := splinegon(inherited duplicateAll);
		connectEdges(s);
		duplicateAll := s;
	end;{splinegon.duplicateAll}

procedure splinegon.textFileInput;
	begin
		inherited textFileInput;
		connectEdges(self);
	end;

procedure splinegon.fileInput;
	begin
		inherited fileInput;
		connectEdges(self);
	end;

procedure splinegon.reflect (a, b, c: extended);{specially for simple splinegons}
	begin
		inherited reflect(a, b, c);
		revertSG(self);
	end;

procedure splinegon.describe;
	begin
		with currentArguments do
			if length = 1 then
				if member(ith(1), polygon2d) & (polygon2d(ith(1)).length > 2) & member(polygon2d(ith(1)).elements^^[1], realPoint) then begin
					beginSubMenu('Create Splinegon');
					menuEntry('Contiguous', smoothC);
					menuEntry('Piecewise', spline2gonC);
					endSubMenu;
				end
				else if member(ith(1), splinegon) then begin
					beginSubMenu('Polygons');
					menuEntry('Carrier Polygon', carrierPolygonC);
					menuEntry('Bounding Polygon', boundingPolygonC);
					endSubMenu;
					beginSubMenu('Renumberings');
					menuEntry('Renumber -', renumberMinusC);
					menuEntry('Renumber +', renumberPlusC);
					menuEntry('Revert', revertC);
					endSubMenu;
					menuEntry('Convex Hull', ConvexHullC);
				end;
	end; {splinegon.describe}

procedure splinegon.execute (item: integer);

	function smooth (p: polygon2d): splinegon;
		var
			s: splinegon;
			s2: spline2;
			i, ip1: integer;
		begin
			new(s);
			failNil(Handle(s));
			currentVectorLength := p.length;
			s.init;
			s.length := currentVectorLength;
			for i := 1 to s.length do begin
				new(s2);
				failNil(Handle(s2));
				s.elements^^[i] := s2;
				s2.pv := realPoint(point2d(p.elements^^[i]).duplicate);
				if i = s.length then
					ip1 := 1
				else
					ip1 := i + 1;
				s2.v2 := realPoint(point2d(p.elements^^[i]).middle(point2d(p.elements^^[ip1])));
			end;
			connectEdges(s);
			smooth := s;
		end;{smooth}

	function spline2gon (p: polygon2d): splinegon;
		var
			s: splinegon;
			s2: spline2;
			i: integer;
		begin
			new(s);
			failNil(Handle(s));
			currentVectorLength := p.length div 2;
			s.init;
			s.length := currentVectorLength;
			for i := 1 to s.length do begin
				new(s2);
				failNil(Handle(s2));
				s.elements^^[i] := s2;
				s2.v1 := realPoint(point2d(p.elements^^[2 * i - 1]).duplicate);
				s2.pv := realPoint(point2d(p.elements^^[2 * i]).duplicate);
			end;
			connectEdgesBack(s);
			spline2gon := s;
		end;{spline2gon}

	begin
		with currentArguments do
			case item of
				smoothC: 
					currentResult.appendNonNil(smooth(polygon2d(ith(1))));
				carrierPolygonC: 
					currentResult.append(splinegon(ith(1)).carrierPolygon);
				boundingPolygonC: 
					currentResult.append(splinegon(ith(1)).boundingPolygon);
				spline2gonC: 
					currentResult.append(spline2gon(polygon2d(ith(1))));
				ConvexHullC: 
					currentResult.append(splinegon(ith(1)).convexHull);
				renumberMinusC: 
					renumberMinus(splinegon(ith(1)));
				renumberPlusC: 
					renumberPlus(splinegon(ith(1)));
				revertC: 
					revertSG(splinegon(ith(1)));
			end;
	end;{splinegon.execute}

function splinegon.sameType: boolean;
	begin
		sameType := false;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function splinegon.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 5 * length;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function splinegon.carrierPolygon: polygon2d;
	var
		p: polygon2d;
		i: integer;
	begin
		new(p);
		failNil(Handle(p));
		p.length := length;
		p.elements := nil;
		p.allocateElements(length);
		for i := 1 to length do
			p.elements^^[i] := curvedEdge(elements^^[i]).v1.duplicate;
		carrierPolygon := p;
	end;{splinegon.carrierPolygon}

function splinegon.boundingPolygon: polygon2d;
	var
		p: polygon2d;
		i: integer;
	begin
		new(p);
		failNil(Handle(p));
		p.length := 2 * length;
		p.elements := nil;
		p.allocateElements(2 * length);
		for i := 1 to length do begin
			p.elements^^[2 * i - 1] := curvedEdge(elements^^[i]).v1.duplicate;
			p.elements^^[2 * i] := curvedEdge(elements^^[i]).pv.duplicate;
		end;
		boundingPolygon := p;
	end;

(*  *)

procedure curvedEdge.free;
	begin
		v2.free;
		pv.free;
		shallowFree;
	end;

function curvedEdge.duplicate;
	var
		c: curvedEdge;
	begin
		c := curvedEdge(shallowDuplicate);
		c.v2 := realPoint(v2.duplicate);
		c.pv := realPoint(pv.duplicate);
		duplicate := c;
	end;

procedure curvedEdge.textFileInput;
	var
		ov2, opv: obj;
	begin
		ov2 := readTextObj;
		opv := readTextObj;
		assert(member(ov2, realPoint) & member(opv, realPoint), 'Legal types in curvedEdge');
		v2 := realPoint(ov2);
		pv := realPoint(opv);
	end;

procedure curvedEdge.textFileOutput;
	begin
		writeTextObj(v2);
		writeTextObj(pv);
	end;

procedure curvedEdge.fileInput;
	begin
		new(v2);
		failNil(Handle(v2));
		v2.fileInput;
		new(pv);
		failNil(Handle(pv));
		pv.fileInput;
	end;

procedure curvedEdge.fileOutput;
	begin
		v2.fileOutput;
		pv.fileOutput;
	end;

function curvedEdge.isSelected (where: obj): boolean;
	begin
		isSelected := v2.isSelected(where);
	end;

procedure curvedEdge.translate (dx, dy: extended);
	begin
		v2.translate(dx, dy);
		pv.translate(dx, dy);
	end;

procedure curvedEdge.reflect (a, b, c: extended);
	begin
		v2.reflect(a, b, c);
		pv.reflect(a, b, c);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure curvedEdge.tangent (p: realPoint; b: boolean; tv, tpv1, tpv2: realPoint);
	begin
		unimplementedMethod('curvedEdge.tangent');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function curvedEdge.xExtremeX: extended;
	begin
		unimplementedMethod('curvedEdge.xExtremeX');
		xExtremeX := 0;
	end;

procedure curvedEdge.commonTangent (c: curvedEdge; tv1, tv2, tpv11, tpv12, tpv21, tpv22: realPoint);
	const
		eps = 1e-5;
	var
		oldTv2: realPoint;
		count: integer;
	begin
		new(oldTv2);
		failNil(Handle(oldTv2));
		tv2.x := c.v1.x + 0.1 * (c.pv.x - c.v1.x);
		tv2.y := c.v1.y + 0.1 * (c.pv.y - c.v1.y);
		count := 0;
		repeat
			oldTv2.x := tv2.x;
			oldTv2.y := tv2.y;
			tangent(tv2, false, tv1, tpv11, tpv12);
			c.tangent(tv1, true, tv2, tpv21, tpv22);
			count := count + 1;
		until (tv2.squaredDistanceX(oldTv2) < eps) | (count > 20);
		if count > 20 then begin
			interactiveFlash;
			c.interactiveFlash;
		end;
		dispose(oldTv2);
	end;

(*  *)

function spline2.getType: objectType;
	begin
		getType := spline2Type;
	end;

procedure spline2.tangent (p: realPoint; b: boolean; tv, tpv1, tpv2: realPoint);
	var
		t, t1, wpv1, tpv1x, tpv1y, tpv2x, tpv2y: extended;
	function sqEq (a, b, c: extended; var x1, x2: extended): boolean;
		var
			d: extended;
		begin
			d := b * b - 4 * a * c;
			sqEq := d >= 0;
			if d < 0 then
				d := 0
			else
				d := sqrt(d);
			x1 := (-b - d) / (2 * a);
			x2 := (-b + d) / (2 * a);
		end;
	begin
		wpv1 := p.whichSideX(pv, v1);
		assert(sqEq(v2.whichSideX(v1, pv), p.crossProductX(v1, v2, pv) + wpv1, -wpv1, t, t1), 'square equation');
		if b & (t < t1) | not b & (t > t1) then
			t := t1;
		if t < 0 then
			t := 0
		else if t > 1 then
			t := 1;
		t1 := 1 - t;
		tpv1x := v1.x * t1 + pv.x * t;
		tpv1y := v1.y * t1 + pv.y * t;
		tpv2x := pv.x * t1 + v2.x * t;
		tpv2y := pv.y * t1 + v2.y * t;
		tv.x := tpv1x * t1 + tpv2x * t;
		tv.y := tpv1y * t1 + tpv2y * t;
		tpv1.x := tpv1x;
		tpv1.y := tpv1y;
		tpv2.x := tpv2x;
		tpv2.y := tpv2y;
	end;

function spline2.xExtremeX: extended;
	begin
		xExtremeX := (v1.x * v2.x - sqr(pv.x)) / (v2.x - 2 * pv.x + v1.x);
	end;

procedure spline2.interactiveOutput;
	const
		d = 8;
		minN = 5;
	var
		sx, sy, v1x, v1y, pvx, pvy, v2x, v2y, i, n, n2: integer;
		t, tt, t1t, t1t1: real;
		v1b, v2b, pvb, moved: boolean;
	begin
		v1b := worldToWindow(v1.x, v1.y, v1x, v1y);
		pvb := worldToWindow(pv.x, pv.y, pvx, pvy);
		v2b := worldToWindow(v2.x, v2.y, v2x, v2y);
		if v1b & pvb then
			n := (abs(v1x - pvx) + abs(v1y - pvy)) div d;
		if v2b & pvb then begin
			n2 := (abs(v2x - pvx) + abs(v2y - pvy)) div d;
			if n2 > n then
				n := n2;
		end;
		if not (v1b & pvb & v2b) then
			n := minN;
		if n < 1 then
			n := 1;

		moved := false;
		for i := 0 to n do begin
			t := i / n;
			tt := t * t;
			t1t := 2 * t * (1 - t);
			t1t1 := (1 - t) * (1 - t);
			if worldToWindow(v1.x * t1t1 + pv.x * t1t + v2.x * tt, v1.y * t1t1 + pv.y * t1t + v2.y * tt, sx, sy) then
				if moved then begin
					LineTo(sx, sy);
					LineTo(sx, sy);(*twice ! *)
				end
				else begin
					MoveTo(sx, sy);
					moved := true;
				end
			else
				moved := false;
		end;
	end;

(*  *)

function straightEdge.getType: objectType;
	begin
		getType := straightEdgeType;
	end;

procedure straightEdge.interactiveOutput;
	begin
		v1.drawLine(v2);
	end;

function straightEdge.isSelected (where: obj): boolean;
	var
		ls: lineSegment;
	begin
		new(ls);
		failNil(Handle(ls));
		ls.p := v1;
		ls.q := v2;
		isSelected := ls.isSelected(where);
		dispose(ls);
	end;

(*  *)

procedure copyRealPoint (s, t: realPoint);
	begin
		t.x := s.x;
		t.y := s.y;
	end;

function approxWhichSideSign (s, p, q: realPoint): signType;
	const
		wsEps = 0.1;
	var
		ws: extended;
	begin
		ws := s.whichSideX(p, q) / p.distanceX(q);
		if abs(ws) <= wsEps then
			approxWhichSideSign := 0
		else
			approxWhichSideSign := sign(ws);
	end;

function splinegon.convexHull: splinegon;
	type
		vertex = integer; (* 1..2*length *)
		stackArr = array[1..maxInt] of vertex;
	var
		minXv, maxXv: vertex;
		tv1, tv2, tpv11, tpv12, tpv21, tpv22: realPoint;
		stack: ^stackArr;
		stackPtr: integer;

	procedure push (vx: vertex);
		begin
			stackPtr := stackPtr + 1;
			stack^[stackPtr] := vx;
		end;

	procedure findExtremes;
		var
			i: integer;
			minX, maxX, thisX: extended;
		begin
			minX := curvedEdge(elements^^[1]).v1.x;
			maxX := minX;
			minXv := 1;
			maxXv := 1;
			for i := 1 to length do
				with curvedEdge(elements^^[i]) do begin
					if v2.x >= maxX then begin
						maxXv := 2 * i + 1;
						maxX := v2.x;
					end;
					if pv.x > maxX then begin
						thisX := xExtremeX;
						if thisX > maxX then begin
							maxXv := 2 * i;
							maxX := thisX;
						end;
					end;

					if v2.x <= minX then begin
						minXv := 2 * i + 1;
						minX := v2.x;
					end;
					if pv.x < minX then begin
						thisX := xExtremeX;
						if thisX < minX then begin
							minXv := 2 * i;
							minX := thisX;
						end;
					end;
				end;
		end;{findExtremes}

	function getIndexedCE (vx: vertex): obj;
		begin
			getIndexedCE := curvedEdge(elements^^[(vx div 2 - 1) mod length + 1]);
		end;

	function getIndexedPoint (vx: vertex): realPoint;
		begin
			if odd(vx) then
				getIndexedPoint := curvedEdge(elements^^[vx div 2 mod length + 1]).v1
			else
				getIndexedPoint := curvedEdge(elements^^[(vx div 2 - 1) mod length + 1]).pv;
		end;

	procedure lineSegDir (vx1, vx2: vertex; var p, q: realPoint);
		var
			ce1, ce2: curvedEdge;

		begin{lineSegDir}
			if vx2 - vx1 = 1 then begin
				p := getIndexedPoint(vx1);
				q := getIndexedPoint(vx2);
			end
			else if odd(vx1) then begin
				p := getIndexedPoint(vx1);
				if odd(vx2) then
					q := getIndexedPoint(vx2)
				else begin {1}
					ce2 := curvedEdge(getIndexedCE(vx2));
					q := ce2.v2;
					if approxWhichSideSign(ce2.v1, p, q) < 0 then
						q := ce2.v1;
					if approxWhichSideSign(ce2.pv, p, q) < 0 then begin
						ce2.tangent(p, true, tv2, tpv21, tpv22);
						q := tv2;
					end;
				end;
			end
			else begin
				ce1 := curvedEdge(getIndexedCE(vx1));
				if odd(vx2) then begin {2}
					q := getIndexedPoint(vx2);
					p := ce1.v1;
					if approxWhichSideSign(ce1.v2, p, q) < 0 then
						p := ce1.v2;
					if approxWhichSideSign(ce1.pv, p, q) < 0 then begin
						ce1.tangent(q, false, tv1, tpv11, tpv12);
						p := tv1;
					end;
				end
				else begin {3}
					ce2 := curvedEdge(getIndexedCE(vx2));
					lineSegDir(vx1 - 1, vx2, p, q);
					if (approxWhichSideSign(ce1.pv, p, q) < 0) | (approxWhichSideSign(ce1.v2, p, q) < 0) then begin
						lineSegDir(vx1, vx2 + 1, p, q);
						if (approxWhichSideSign(ce2.pv, p, q) < 0) | (approxWhichSideSign(ce2.v1, p, q) < 0) then begin
							lineSegDir(vx1 + 1, vx2, p, q);
							if (approxWhichSideSign(ce1.pv, p, q) < 0) | (approxWhichSideSign(ce1.v1, p, q) < 0) then begin
								lineSegDir(vx1, vx2 - 1, p, q);
								if (approxWhichSideSign(ce2.pv, p, q) < 0) | (approxWhichSideSign(ce2.v2, p, q) < 0) then begin
									ce1.commonTangent(ce2, tv1, tv2, tpv11, tpv12, tpv21, tpv22);
									p := tv1;
									q := tv2;
								end;
							end;
						end;
					end;
				end;
			end;
		end;{lineSegDir}

	procedure lineSeg (vx1, vx2: vertex; var p, q: realPoint);
		begin
			if vx2 - vx1 = 1 then begin
				if odd(vx1) then
					p := getIndexedPoint(vx1)
				else
					p := getIndexedPoint(vx2);
				q := p;
			end
			else
				lineSegDir(vx1, vx2, p, q);
		end;

	function whichside (vx1, vx2, vx3: vertex): extended;
		var
			p, q: realPoint;
			ce: curvedEdge;

		begin{whichside}
			lineSegDir(vx2, vx3, p, q);
			if odd(vx1) then
				whichside := getIndexedPoint(vx1).whichSideX(p, q)
			else begin
				ce := curvedEdge(getIndexedCE(vx1));
				if ce.v1.whichSideX(p, q) < 0 then
					whichside := -1
				else if ce.v2.whichSideX(p, q) < 0 then
					whichside := -1
				else if approxWhichSideSign(ce.pv, p, q) < 0 then begin
					p := realPoint(p.duplicate);
					ce.tangent(p, true, tv1, tpv11, tpv12);
					whichside := tv1.whichSideX(p, q);
					p.free;
				end
				else
					whichside := 1;
			end;
		end;{whichside}

	procedure hull (from, upto: vertex; dir: signType);
		var
			i, j, start: integer;
			p, q: realPoint;
		begin{hull}

			if stackPtr = 0 then
				push(from);
			start := stackPtr;
			push(from + 1);

			for i := from + 2 to upto do begin

				while (stackPtr > start) & (whichside(i, stack^[stackPtr - 1], stack^[stackPtr]) <= 0) do
					stackPtr := stackPtr - 1;

				lineSegDir(stack^[stackPtr], i, p, q);
{$IFC true }
				if (sign(q.x - p.x) <> -dir) & (not odd(stack^[stackPtr]) | (approxWhichSideSign(q, getIndexedPoint(stack^[stackPtr] - 1), getIndexedPoint(stack^[stackPtr])) >= 0)) then
					push(i);
{$ELSEC }
				if sign(q.x - p.x) <> -dir then
					push(i);
{$ENDC }

{$IFC AnimationEnabled AND ASplinegon }
				if animationFlag[AConvexHullItem] then begin
{$IFC THINK_PASCAL }
					showText;
{$ENDC }
					write((stack^[start] - 1) mod (2 * length) + 1 : 3);
					for j := start + 1 to stackPtr do begin
						lineSeg(stack^[j - 1], stack^[j], p, q);
						p.drawLine(q);
						write((stack^[j] - 1) mod (2 * length) + 1 : 3);
					end;
					writeLn;
					waitForClick(animationFlag[AConvexHullItem]);
					for j := start + 1 to stackPtr do begin
						lineSeg(stack^[j - 1], stack^[j], p, q);
						p.drawLine(q);
					end;
				end;
{$ENDC }

			end;{for}
		end;{hull}

	function assemble: splinegon;
		var
			result: splinegon;
			ce, oldce: curvedEdge;
			se: straightEdge;
			i: integer;
			vx, lastvx: vertex;
			oldPt, pt: realPoint;

		procedure initStraightEdge (se: straightEdge);
			begin
				se.pv := realPoint(se.v1.duplicate);
(*se.pv := realPoint(se.v1.middle(se.v2)); *)
			end;

		begin{assemble}
			new(result);
			failNil(Handle(result));
			currentVectorLength := 2 * length;
			result.init;

			if odd(minXv) then
				oldPt := realPoint(getIndexedPoint(minXv).duplicate)
			else begin
				ce := curvedEdge(getIndexedCE(minXv).duplicate);
				ce.v1 := realPoint(ce.v1.duplicate);
				result.append(ce);
			end;

			vx := minXv;
			for i := 2 to stackPtr do begin
				lastvx := vx;
				vx := stack^[i];
				if result.length > 0 then begin
					oldce := curvedEdge(result.elements^^[result.length]);
					oldPt := oldce.v2;
				end;
				if odd(vx) then begin
					pt := getIndexedPoint(vx);
					if vx - lastvx <> 1 then begin
						if not odd(lastvx) then begin
							oldce.tangent(pt, false, tv1, tpv11, tpv12);
							copyRealPoint(tv1, oldce.v2);
							copyRealPoint(tpv11, oldce.pv);
							oldPt := oldce.v2;
						end;
						new(se);
						failNil(Handle(se));
						se.v1 := oldPt;
						se.v2 := realPoint(pt.duplicate);
						initStraightEdge(se);
						result.append(se);
					end;
				end
				else begin
					ce := curvedEdge(getIndexedCE(vx));
					if vx - lastvx <> 1 then begin
						if odd(lastvx) then
							ce.tangent(getIndexedPoint(lastvx), true, tv2, tpv21, tpv22)
						else begin
							oldce.commonTangent(ce, tv1, tv2, tpv11, tpv12, tpv21, tpv22);
							copyRealPoint(tv1, oldce.v2);
							copyRealPoint(tpv11, oldce.pv);
							oldPt := oldce.v2;
						end;
						new(se);
						failNil(Handle(se));
						se.v1 := oldPt;
						se.v2 := realPoint(tv2.duplicate);
						initStraightEdge(se);
						result.append(se);
						ce := curvedEdge(ce.duplicate);
						ce.v1 := se.v2;
						copyRealPoint(tpv22, ce.pv);
					end
					else begin
						ce := curvedEdge(ce.duplicate);
						ce.v1 := oldPt;
					end;
					result.append(ce);
				end;{if}
			end;{for}

			oldce := curvedEdge(result.elements^^[result.length]);
			if odd(minXv) then begin
				oldce.v2.free;
				oldce.v2 := curvedEdge(result.elements^^[1]).v1;
			end
			else begin
				ce := curvedEdge(result.elements^^[1]);
				if vx - lastvx <> 1 then begin
					se := straightEdge(result.elements^^[result.length - 1]);
					ce.tangent(se.v1, true, tv1, tpv11, tpv12);
					copyRealPoint(tpv12, ce.pv);
				end;
				ce.v1.free;
				ce.v1 := oldce.v1;
				oldce.free;
				result.removeLast;
			end;

			result.allocateElements(result.length);
			assemble := result;
		end;{assemble}

	begin{splinegon.convexHull}
		findExtremes;

		new(tv1);
		failNil(Handle(tv1));
		new(tv2);
		failNil(Handle(tv2));
		new(tpv11);
		failNil(Handle(tpv11));
		new(tpv12);
		failNil(Handle(tpv12));
		new(tpv21);
		failNil(Handle(tpv21));
		new(tpv22);
		failNil(Handle(tpv22));

		stack := pointer(NewPtr(4 * length * sizeOf(vertex)));
		failNil(Handle(stack));
		stackPtr := 0;

		if maxXv < minXv then
			maxXv := maxXv + 2 * length;
		hull(minXv, maxXv, 1);
		hull(maxXv, minXv + 2 * length, -1);
		convexHull := assemble;

{$IFC THINK_PASCAL }
		disposPtr(pointer(stack));
{$ELSEC }
		DisposePtr(pointer(stack));
{$ENDC }

		dispose(tv1);
		dispose(tv2);
		dispose(tpv11);
		dispose(tpv12);
		dispose(tpv21);
		dispose(tpv22);

	end;{splinegon.convexHull}

(*  *)

procedure initSplinegon;
	var
		s: splinegon;
		s2: spline2;
		se: straightEdge;
	begin
		new(s);
		failNil(Handle(s));
		prototypes[splinegonType] := s;
		new(s2);
		failNil(Handle(s2));
		prototypes[spline2Type] := s2;
		new(se);
		failNil(Handle(se));
		prototypes[straightEdgeType] := se;
	end; {initSplinegon}

end.
