(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit standardPredicates;

interface

uses
	sequence;

function noConstraint (x: extended): boolean;
(* This function returns always true and can be used if a numerical value is unconstrained. *)

function nonTrivialSeqIndex (x: extended): boolean;

function isPositive (x: extended): boolean;

function isPositiveOrZero (x: extended): boolean;

function isPositiveOrZeroInteger (x: extended): boolean;

function isNonZeroInteger (x: extended): boolean;

function isNatural (x: extended): boolean;

function isSeqIndex (x: extended): boolean;

function isSeqIndex0 (x: extended): boolean;

function upperBoundedNatural (x: extended): boolean;

function isInteger (x: extended): boolean;

function percentageOk (x: extended): boolean;

function alwaysTrue: boolean;

function alwaysFalse: boolean;

var
upperBound: extended;

implementation

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function noConstraint (x: extended): boolean;
	begin
		noConstraint := true;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function isNonZeroInteger (x: extended): boolean;
	begin
		isNonZeroInteger := (trunc(x) = x) & (x <> 0);
	end;

function nonTrivialSeqIndex (x: extended): boolean;
	begin
		nonTrivialSeqIndex := (trunc(x) = x) & (1 < x) & (x < maxN);
	end;

function isPositive (x: extended): boolean;
	begin
		isPositive := x > 0;
	end;

function isPositiveOrZero (x: extended): boolean;
	begin
		isPositiveOrZero := x >= 0;
	end;

function isPositiveOrZeroInteger (x: extended): boolean;
	begin
		isPositiveOrZeroInteger := (x >= 0) & (trunc(x) = x);
	end;

function isNatural (x: extended): boolean;
	begin
		isNatural := (x > 0) & (trunc(x) = x);
	end;

function isSeqIndex (x: extended): boolean;
	begin
		isSeqIndex := (x = trunc(x)) & (0 < x) & (x < maxN);
	end;

function isSeqIndex0 (x: extended): boolean;
	begin
		isSeqIndex0 := (x = trunc(x)) & (0 <= x) & (x < maxN);
	end;

function upperBoundedNatural (x: extended): boolean;
	begin
		upperBoundedNatural := (1 <= x) & (x <= upperBound) & (x = trunc(x));
	end;

function isInteger (x: extended): boolean;
	begin
		isInteger := trunc(x) = x;
	end;

function percentageOk (x: extended): boolean;
	begin
		percentageOk := (0.0 <= x) & (x <= 100.0);
	end;

function alwaysTrue: boolean;
	begin
		alwaysTrue := true;
	end;

function alwaysFalse: boolean;
	begin
		alwaysFalse := false;
	end;


end.
