(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit streamIO;

interface

uses
	floatArithmetic;

{$IFC THINK_PASCAL }
type
	streamStatus = record
			FSOpenRProc, FSOpenWProc, FSReadProc, FSWriteProc, FSCloseProc: ProcPtr;
			bufferSize: longInt;
		end;

function createStreamStatus (FSOpenRProc, FSOpenWProc, FSReadProc, FSWriteProc, FSCloseProc: ProcPtr; bufferSize: longInt): streamStatus;
{$ELSEC }
type
	FSOpenProcT = function (fileName: Str255; vRefNum: integer; var refNum: integer): OSErr;
	FSReadProcT = function (refNum: integer; var count: longInt; buffPtr: Ptr): OSErr;
 	FSWriteProcT = function (refNum: integer; var count: longInt; buffPtr: Ptr): OSErr;
 	FSCloseProcT = function (refNum: integer): OSErr;
	
	streamStatus = record
			FSOpenRProc: FSOpenProcT;
			FSOpenWProc: FSOpenProcT;
			FSReadProc:  FSReadProcT;
			FSWriteProc: FSWriteProcT;
			FSCloseProc: FSCloseProcT;
			bufferSize: longInt;
		end;

function createStreamStatus (FSOpenRProc: FSOpenProcT; FSOpenWProc: FSOpenProcT; FSReadProc:  FSReadProcT; FSWriteProc: FSWriteProcT; FSCloseProc: FSCloseProcT; bufferSize: longInt): streamStatus;
{$ENDC }

procedure getStreamStatus (var s: streamStatus);

procedure setStreamStatus (s: streamStatus);

procedure openReadStreamTest (name: Str255; volume: integer; var directoryExists, fileExists: boolean);

procedure openReadStream (name: Str255; volume: integer);

procedure closeReadStream;

procedure openWriteStream (name: Str255; volume: integer);

procedure openWriteStreamAppend (name: Str255; volume: integer; fileType, creator: OSType; doAppend: boolean);

procedure flushStream;

procedure closeWriteStream;

procedure testPathName (pathName: Str255; volume: integer; var directoryExists, fileExists: boolean);

procedure readByte (var b: Byte);

procedure writeByte (b: Byte);

procedure readInteger (var x: integer);

procedure writeInteger (x: integer);

procedure readReal (var x: real);

procedure writeReal (x: real);

procedure readLongInt (var x: longInt);

procedure writeLongInt (x: longInt);

procedure readExtended (var e: extended);

procedure writeExtended (e: extended);

procedure readRect (var r: Rect);

procedure writeRect (r: Rect);

procedure readFloat (var x: float);

procedure writeFloat (x: float);

procedure readHandle (var h: Handle);

procedure writeHandle (h: Handle);

procedure readStr255 (var x: Str255);

procedure writeStr255 (x: Str255);

procedure setFileInfo (name: Str255; volume: integer; newFileType, newCreatorType: OSType);

function streamFSOpen (fileName: Str255; vRefNum: integer; var refNum: integer): OSErr;

function streamFSRead (refNum: integer; var count: longInt; buffPtr: Ptr): OSErr;

function streamFSWrite (refNum: integer; var count: longInt; buffPtr: Ptr): OSErr;

function streamFSClose (refNum: integer): OSErr;

procedure initStreamIO;

const
maxBufferSize = 2048;

{$IFC THINK_PASCAL }
gbCreator = 'GBXY';		(*	for Metrowerks defined in saveWindow	*)
gbFileType = 'DATA';	(*	for Metrowerks defined in saveWindow	*)
{$ENDC }


implementation

const
	fileProblemAlert = 1002;
	haltItem = 3;
	maxBufferSize1 = maxBufferSize - 1;
	sizeOfInteger = sizeof(integer);
	sizeOfLongInt = sizeof(longInt);
	sizeOfReal = sizeof(real);
{$IFC OPTION(MC68881) }
	sizeOfExtended = sizeof(extended80);
{$ELSEC }
	sizeOfExtended = sizeof(extended);
{$ENDC }

type
	bufferIndex = 0..maxBufferSize1;
	bufferIndex1 = 0..maxBufferSize;
	bufferType = packed array[bufferIndex] of Byte;
	byteStream = record
			fileName: Str255;
			buffer: bufferType;
			f: integer; (* refNum *)
			filePosition: bufferIndex1;
			s: streamStatus;
		end;
	integerPtr = ^integer;

	indexInteger = 1..sizeOfInteger;
	numberInteger = (integerT, byteIntegerT);
	numberIntegerEqv = record
			case numberInteger of
				integerT: (
						x: integer;
				);
				byteIntegerT: (
						y: packed array[indexInteger] of Byte;
				);
		end;

	indexLongint = 1..sizeOfLongInt;
	numberLongint = (longIntT, byteLongintT);
	numberLongIntEqv = record
			case numberLongint of
				longIntT: (
						l: longInt;
				);
				byteLongintT: (
						y: packed array[indexLongint] of Byte;
				);
		end;
	
	indexReal = 1..sizeOfReal;
	numberReal = (realT, byteRealT);
	numberRealEqv = record
			case numberReal of
				realT: (
						x: real;
				);
				byteRealT: (
						y: packed array[indexReal] of Byte;
				);
		end;

	indexExtended = 1..sizeOfExtended;
	numberExtended = (extendedT, byteExtendedT);
	numberExtendedEqv = record
			case numberExtended of
				extendedT: (
{$IFC OPTION(MC68881) }
						e: extended80;
{$ELSEC }
						e: extended;
{$ENDC }
				);
				byteExtendedT: (
						y: packed array[indexExtended] of Byte;
				);
		end;

var
	currentStream: byteStream;

{$IFC THINK_PASCAL }
function callFSOpenProc (fileName: Str255; vRefNum: integer; var refNum: integer; p: ProcPtr): OSErr;
inline
	$205f, $4e90;

function callFSReadProc (refNum: integer; var count: longInt; buffPtr: Ptr; p: ProcPtr): OSErr;
inline
	$205f, $4e90;

function callFSWriteProc (refNum: integer; var count: longInt; buffPtr: Ptr; p: ProcPtr): OSErr;
inline
	$205f, $4e90;

function callFSCloseProc (refNum: integer; p: ProcPtr): OSErr;
inline
	$205f, $4e90;
{$ENDC }

function streamFSOpen (fileName: Str255; vRefNum: integer; var refNum: integer): OSErr;
	begin
		streamFSOpen := FSOpen(fileName, vRefNum, refNum)
	end;

function streamFSRead (refNum: integer; var count: longInt; buffPtr: Ptr): OSErr;
	begin
		streamFSRead := FSRead(refNum, count, buffPtr);
	end;

function streamFSWrite (refNum: integer; var count: longInt; buffPtr: Ptr): OSErr;
	begin
		streamFSWrite := FSWrite(refNum, count, buffPtr);
	end;

function streamFSClose (refNum: integer): OSErr;
	begin
		streamFSClose := FSClose(refNum);
	end;

procedure openReadStream (name: Str255; volume: integer);
	var
		directoryExists, fileExists: boolean;
	begin
		openReadStreamTest(name, volume, directoryExists, fileExists);
	end;

procedure openReadStreamTest (name: Str255; volume: integer; var directoryExists, fileExists: boolean);
	var
		err: OSErr;
	begin
		with currentStream, s do begin
			fileName := name;
{$IFC THINK_PASCAL }
			err := callFSOpenProc(fileName, volume, f, FSOpenRProc);
{$ELSEC }
			err := FSOpenRProc(fileName, volume, f);
{$ENDC }
			if err = fnfErr then begin
				directoryExists := true;
				fileExists := false;
			end
			else if (err = dirNFErr) | (err = nsvErr) then begin
				directoryExists := false;
				fileExists := false;
			end
			else begin
				directoryExists := true;
				fileExists := true;
				testOSError(err, noErr);
			end;
			filePosition := bufferSize;
		end;
	end;

procedure testPathName (pathName: Str255; volume: integer; var directoryExists, fileExists: boolean);
	begin
		openReadStreamTest(pathName, volume, directoryExists, fileExists);
		if directoryExists & fileExists then
			closeReadStream;
	end;

function getParentID (name: Str255; volume: integer): longInt;
	var
		cpbRec: CInfoPBRec;
	begin
		with cpbRec do begin
			ioCompletion := nil;
			ioNamePtr := @name;
			ioVRefNum := volume;
			ioFDirIndex := 0;
			ioDirID := 0;
		end;
{$IFC THINK_PASCAL }
		assert(PBGetCatInfo(@cpbRec, false) = noErr, 'getParentId ok');
{$ELSEC }
		assert(PBGetCatInfoSync(@cpbRec) = noErr, 'getParentId ok');
{$ENDC }
		getParentID := cpbRec.ioFlParID;
	end;

procedure touchFolder (volume: integer; directoryId: longInt);
	var
		cpbRec: CInfoPBRec;
	begin
		with cpbRec do begin
			ioCompletion := nil;
			ioNamePtr := nil;
			ioVRefNum := volume;
			ioDirID := directoryId;
			ioFDirIndex := -1;
			ioFVersNum := 0;
		end;
{$IFC THINK_PASCAL }
		assert(PBGetCatInfo(@cpbRec, false) = noErr, 'touchFolder getInfo ok');
{$ELSEC }
		assert(PBGetCatInfoSync(@cpbRec) = noErr, 'touchFolder getInfo ok');
{$ENDC }
		with cpbRec do begin
			ioCompletion := nil;
			ioNamePtr := nil;
			ioVRefNum := volume;
			ioDirID := directoryId;
			ioFDirIndex := -1;
			ioFVersNum := 0;
			GetDateTime(ioDrMdDat);
		end;
{$IFC THINK_PASCAL }
		assert(PBSetCatInfo(@cpbRec, false) = noErr, 'touchFolder setInfo ok');
{$ELSEC }
		assert(PBSetCatInfoSync(@cpbRec) = noErr, 'touchFolder setInfo ok');
{$ENDC }
	end;

procedure setFileInfo (name: Str255; volume: integer; newFileType, newCreatorType: OSType);
	var
		info: FInfo;
	begin
		testOSError(GetFInfo(name, volume, info), noErr);
		info.fdType := newFileType;
		info.fdCreator := newCreatorType;
		testOSError(SetFInfo(name, volume, info), noErr);
		touchFolder(volume, getParentID(name, volume));
	end;

procedure openWriteStreamAppend (name: Str255; volume: integer; fileType, creator: OSType; doAppend: boolean);
	var
		err: OSErr;
	begin
		with currentStream, s do begin
			fileName := name;
{$IFC THINK_PASCAL }
			err := callFSOpenProc(fileName, volume, f, FSOpenWProc);
			if FSOpenWProc = @streamFSOpen then begin
{$ELSEC }
			err := FSOpenWProc(fileName, volume, f);
			if FSOpenWProc = streamFSOpen then begin
{$ENDC }
				if err = fnfErr then begin
					testOSError(Create(fileName, volume, creator, fileType), noErr);
					testOSError(FSOpen(fileName, volume, f), noErr);
				end
				else begin
					testOSError(err, noErr);
					if doAppend then
						testOSError(SetFPos(f, fsFromLEOF, 0), noErr)
					else
						testOSError(SetEOF(f, 0), noErr);
					setFileInfo(fileName, volume, fileType, creator);
				end;
			end;
			filePosition := 0;
		end;
	end;

procedure openWriteStream (name: Str255; volume: integer);
	begin
		openWriteStreamAppend(name, volume, gbFileType, gbCreator, false);
	end;

procedure closeReadStream;
	begin
{$IFC THINK_PASCAL }
		testOSError(callFSCloseProc(currentStream.f, currentStream.s.FSCloseProc), noErr);
{$ELSEC }
		testOSError( currentStream.s.FSCloseProc(currentStream.f), noErr);
{$ENDC }
	end;

procedure flushStream;
	var
		count: longInt;
	begin
		with currentStream, s do
			if filePosition > 0 then begin
				count := filePosition;
				filePosition := 0;
{$IFC THINK_PASCAL }
				testOSError(callFSWriteProc(f, count, @buffer, FSWriteProc), noErr);
{$ELSEC }
				testOSError(FSWriteProc(f, count, @buffer), noErr);
{$ENDC }
			end;
	end;

procedure closeWriteStream;
	begin
		with currentStream, s do begin
			flushStream;
{$IFC THINK_PASCAL }
			testOSError(callFSCloseProc(f, FSCloseProc), noErr);
{$ELSEC }
			testOSError(FSCloseProc(f), noErr);
{$ENDC }
		end;
	end;

procedure readByte (var b: Byte);
	var
		count, i: longInt;
	begin
		with currentStream, s do begin
			if filePosition = bufferSize then begin
				filePosition := 0;
				count := bufferSize;
{$IFC THINK_PASCAL }
				testOSError(callFSReadProc(f, count, @buffer, FSReadProc), eofErr);
{$ELSEC }
				testOSError(FSReadProc(f, count, @buffer), eofErr);
{$ENDC }
				for i := count to bufferSize - 1 do
					buffer[count] := 0;
			end;
			b := buffer[filePosition];
			filePosition := filePosition + 1;
		end;
	end;

{$SETC DebugWriteByte = false }

procedure writeByte (b: Byte);
	begin
{$IFC DebugWriteByte }
		write(chr(b));
{$ENDC }
		with currentStream, s do begin
			if filePosition = bufferSize then
				flushStream;
			buffer[filePosition] := b;
			filePosition := filePosition + 1;
		end;
	end;

procedure readInteger (var x: integer);
	var
		eqv: numberIntegerEqv;
		i: indexInteger;
		b: Byte;
	begin
		for i := 1 to sizeOfInteger do begin
			readByte(b);
			eqv.y[i] := b;
		end;
		x := eqv.x;
	end;

procedure writeInteger (x: integer);
	var
		eqv: numberIntegerEqv;
		i: indexInteger;
	begin
		eqv.x := x;
		for i := 1 to sizeOfInteger do
			writeByte(eqv.y[i]);
	end;

procedure readLongInt (var x: longInt);
	var
		eqv: numberLongIntEqv;
		i: indexLongint;
		b: Byte;
	begin
		for i := 1 to sizeOfLongInt do begin
			readByte(b);
			eqv.y[i] := b;
		end;
		x := eqv.l;
	end;

procedure writeLongInt (x: longInt);
	var
		eqv: numberLongIntEqv;
		i: indexLongint;
	begin
		eqv.l := x;
		for i := 1 to sizeOfLongInt do
			writeByte(eqv.y[i]);
	end;

procedure readReal (var x: real);
	var
		eqv: numberRealEqv;
		i: indexReal;
		b: Byte;
	begin
		for i := 1 to sizeOfReal do begin
			readByte(b);
			eqv.y[i] := b;
		end;
		x := eqv.x;
	end;

procedure writeReal (x: real);
	var
		eqv: numberRealEqv;
		i: indexReal;
	begin
		eqv.x := x;
		for i := 1 to sizeOfReal do
			writeByte(eqv.y[i]);
	end;

{$IFC OPTION(MC68881) }
procedure readExtended (var e: extended);
	var
		eqv: numberExtendedEqv;
		i: indexExtended;
		b: Byte;
	begin
		for i := 1 to sizeOfExtended do begin
			readByte(b);
			eqv.y[i] := b;
		end;
		e := X80toX96(eqv.e);
	end;

procedure writeExtended (e: extended);
	var
		eqv: numberExtendedEqv;
		i: indexExtended;
	begin
		eqv.e := X96toX80(e);
		for i := 1 to sizeOfExtended do
			writeByte(eqv.y[i]);
	end;

{$ELSEC }
procedure readExtended (var e: extended);
	var
		eqv: numberExtendedEqv;
		i: indexExtended;
		b: Byte;
	begin
		for i := 1 to sizeOfExtended do begin
			readByte(b);
			eqv.y[i] := b;
		end;
		e := eqv.e;
	end;

procedure writeExtended (e: extended);
	var
		eqv: numberExtendedEqv;
		i: indexExtended;
	begin
		eqv.e := e;
		for i := 1 to sizeOfExtended do
			writeByte(eqv.y[i]);
	end;

{$ENDC }

procedure readStr255 (var x: Str255);
	var
		c: char;
		i: integer;
		l: Byte;
	begin
		readByte(l);
		x := '';
		for i := 1 to l do begin
			readByte(Byte(c));
			x := concat(x, c);
		end;
	end;

procedure writeStr255 (x: Str255);
	var
		i: integer;
	begin
		writeByte(Byte(length(x)));
		for i := 1 to length(x) do
			writeByte(Byte(x[i]));
	end;

procedure readRect (var r: Rect);
	begin
		with r do begin
			readInteger(top);
			readInteger(left);
			readInteger(bottom);
			readInteger(right);
		end;
	end;

procedure writeRect (r: Rect);
	begin
		with r do begin
			writeInteger(top);
			writeInteger(left);
			writeInteger(bottom);
			writeInteger(right);
		end;
	end;

procedure readFloat (var x: float);
	var
		i: integer;
		b: Byte;
	begin
		readInteger(i);
		x.sign := i;
		readByte(b);
		x.base := b;
		readInteger(x.precision);
		readInteger(x.exponent);
		for i := 1 to x.precision do begin
			readByte(b);
			x.mantissa[i] := b;
		end;
	end;

procedure writeFloat (x: float);
	var
		i: integer;
	begin
		writeInteger(x.sign);
		writeByte(x.base);
		writeInteger(x.precision);
		writeInteger(x.exponent);
		for i := 1 to x.precision do
			writeByte(x.mantissa[i]);
	end;

procedure readHandle (var h: Handle);
	var
		length, i, upto: longInt;
	begin
		readLongInt(length);
		if odd(length) then
			length := length + 1;
		h := newHandleFailNil(length);
		i := longInt(h^);
		upto := i + length;
		while i < upto do begin
			readInteger(integerPtr(i)^);
			i := i + 2;
		end;
	end;

procedure writeHandle (h: Handle);
	var
		length, i, upto: longInt;
	begin
		length := GetHandleSize(h);
		writeLongInt(length);
		if odd(length) then
			length := length + 1;
		i := longInt(h^);
		upto := i + length;
		while i < upto do begin
			writeInteger(integerPtr(i)^);
			i := i + 2;
		end;
	end;

{$IFC THINK_PASCAL }
function createStreamStatus (FSOpenRProc, FSOpenWProc, FSReadProc, FSWriteProc, FSCloseProc: ProcPtr; bufferSize: longInt): streamStatus;
{$ELSEC }
function createStreamStatus (FSOpenRProc: FSOpenProcT; FSOpenWProc: FSOpenProcT; FSReadProc:  FSReadProcT; FSWriteProc: FSWriteProcT; FSCloseProc: FSCloseProcT; bufferSize: longInt): streamStatus;
{$ENDC }
	var
		result: streamStatus;
	begin
		assert((0 < bufferSize) & (bufferSize <= maxBufferSize), StringOf('(0 < ', bufferSize : 1, ') & (', bufferSize : 1, '  ', maxBufferSize : 1, ')'));
		result.FSOpenRProc := FSOpenRProc;
		result.FSOpenWProc := FSOpenWProc;
		result.FSReadProc := FSReadProc;
		result.FSWriteProc := FSWriteProc;
		result.FSCloseProc := FSCloseProc;
		result.bufferSize := bufferSize;
		createStreamStatus := result;
	end;

procedure setStreamStatus (s: streamStatus);
	begin
		currentStream.s := s;
	end;

procedure getStreamStatus (var s: streamStatus);
	begin
		s := currentStream.s;
	end;

procedure initStreamIO;
	begin
		setStreamStatus(createStreamStatus(@streamFSOpen, @streamFSOpen, @streamFSRead, @streamFSWrite, @streamFSClose, maxBufferSize));
	end;

end.
