(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit textIO;

interface

uses
	floatArithmetic, obj;

procedure initLookAhead;

(* Low level routines *)

const
EOFChar = chr(0);
CR = chr(13);
TAB = chr(9);
maxCreators = 4;
maxCreators1 = maxCreators - 1;

var
ch: char;
textCreatorType: integer;
possibleCreatorTypes: array[0..maxCreators1] of OSType;

procedure openWriteStreamTextAppend (name: Str255; volume: integer; doAppend: boolean);

procedure openWriteStreamText (name: Str255; volume: integer);

function isWhiteSpace (x: char): boolean;

procedure skipWhiteSpace;

procedure skipLine;

procedure getCh;

procedure readRawString (var x: Str255);

procedure writeRawString (x: Str255);

procedure writeRawStringCR (x: Str255); (* with indented CR *)

procedure writeSimpleCR;

procedure writeIndentedCR;

procedure testString (var x: Str255);

(* High level routines *)
procedure writeTextString (x: Str255);
procedure readTextString (var x: Str255);

procedure writeTextLongInt (x: longInt);
procedure readTextLongInt (var x: longInt);

procedure writeTextInteger (x: integer);
procedure readTextInteger (var x: integer);

procedure writeTextReal (x: real);
procedure readTextReal (var x: real);

procedure writeTextExtended (x: extended);
procedure readTextExtended (var x: extended);

procedure writeTextFloat (x: float);
procedure readTextFloat (var x: float);

procedure writeTextHandle (x: Handle);
procedure readTextHandle (var x: Handle);

procedure writeTextObj (x: obj);
procedure writeTextObjNoCR (x: obj);
function readTextObj: obj;

procedure dump (fileName: Str255; o: obj);

function truncate (s: Str255; l: integer): Str255;

procedure initTextIO;

implementation

{$SETC ShowReadInput = False }

var
	indent: integer;

procedure openWriteStreamTextAppend (name: Str255; volume: integer; doAppend: boolean);
	begin
		openWriteStreamAppend(name, volume, 'TEXT', possibleCreatorTypes[textCreatorType], doAppend);
	end;

procedure openWriteStreamText (name: Str255; volume: integer);
	begin
		openWriteStreamTextAppend(name, volume, false);
	end;

procedure writeSimpleCR;
	begin
		writeByte(Byte(CR));
	end;

procedure writeIndentedCR;
	var
		i: integer;
	begin
		writeSimpleCR;
		for i := 1 to indent do
			writeByte(Byte(TAB));
	end;

procedure testString (var x: Str255);
	begin
		if length(x) = 250 then
			fatalError(concat('String ''', copy(x, 1, 9), ''' too long.'));
	end;

procedure getCh;
	begin
		readByte(Byte(ch));
{$IFC ShowReadInput }
		write(ch);
{$ENDC }
	end;

procedure writeRawString (x: Str255);
	var
		i: integer;
	begin
		for i := 1 to length(x) do
			writeByte(Byte(x[i]));
	end;

procedure writeRawStringCR (x: Str255);
	begin
		writeRawString(x);
		writeIndentedCR;
	end;

function isWhiteSpace (x: char): boolean;
	begin
		isWhiteSpace := (x = ' ') | (x = TAB) | (x = CR);
	end;

procedure skipWhiteSpace;
	var
		bracketCount: longInt;
	begin
		repeat
			while (ch = ' ') | (ch = TAB) | (ch = CR) do
				getCh;
			if (ch = '%') | (ch = '#') then begin
				repeat
					getCh;
				until ch = CR;
				getCh;
			end
			else if ch = '{' then begin
				bracketCount := 1;
				repeat
					getCh;
					case ch of
						EOFChar: 
							exit(skipWhiteSpace);
						'}':  begin
							bracketCount := bracketCount - 1;
							if bracketCount = 0 then
								leave;
						end;
						'{': 
							bracketCount := bracketCount + 1;
						otherwise
					end;
				until false;
				getCh;
			end
			else
				exit(skipWhiteSpace);
		until false;
	end;

	var
		stopCharacters: set of char;

procedure readRawString (var x: Str255);
	begin
		skipWhiteSpace;
		x := ch;
		getCh;
		while not (ch in stopCharacters) & (length(x) < 250) do begin
			x := concat(x, ch);
			getCh;
		end;
		testString(x);
	end;

procedure initLookAhead;
	begin
		ch := ' ';
	end;

	type
		string3 = string[3];

function objectTypeToString (t: objectType): string3;
	begin
		case t of
			intType: 
				objectTypeToString := 'INT';
			stringsType: 
				objectTypeToString := 'STR';
			point2dType: 
				objectTypeToString := '---';
			lineSegmentType: 
				objectTypeToString := 'LSG';
			rectangleType: 
				objectTypeToString := 'REC';
			listType: 
				objectTypeToString := 'LST';
			vectorType: 
				objectTypeToString := 'VEC';
			homogenVectorType: 
				objectTypeToString := 'HVC';
			pointVectorType: 
				objectTypeToString := 'PVC';
			convexPolygonType: 
				objectTypeToString := 'CPL';
			voronoiEdgeType: 
				objectTypeToString := 'VED';
			directedGraphType: 
				objectTypeToString := 'DGR';
			undirectedGraphType: 
				objectTypeToString := 'UDG';
			graphEdgeType: 
				objectTypeToString := 'GED';
			voronoiDiagramType: 
				objectTypeToString := 'VDG';
			polygon2dType: 
				objectTypeToString := 'POL';
			markedRealPointType: 
				objectTypeToString := 'MRP';
			simpleUndirectedGraphType: 
				objectTypeToString := 'SUG';
			spanningTreeType: 
				objectTypeToString := 'SPT';
			rectangleVectorType: 
				objectTypeToString := 'RVC';
			dDimPointType: 
				objectTypeToString := 'DPT';
			dDimCircleType: 
				objectTypeToString := 'DCI';
			dDimPointVectorType: 
				objectTypeToString := 'DPV';
			realPointType: 
				objectTypeToString := 'RPT';
			floatPointType: 
				objectTypeToString := 'FPT';
			longIntPointType: 
				objectTypeToString := 'IPT';
			sequenceType: 
				objectTypeToString := '---';
			avlTreeType: 
				objectTypeToString := '---';
			sortedListType: 
				objectTypeToString := '---';
			sortedVectorType: 
				objectTypeToString := '---';
			lineSegmentVectorType: 
				objectTypeToString := 'LVC';
			dictPriorityQueueType: 
				objectTypeToString := '---';
			heapPriorityQueueType: 
				objectTypeToString := '---';
			circleType: 
				objectTypeToString := 'CIR';
			splinegonType: 
				objectTypeToString := 'SPL';
			spline2Type: 
				objectTypeToString := 'SP2';
			straightEdgeType: 
				objectTypeToString := 'SED';
			polyLineType: 
				objectTypeToString := 'PLI';
			layerType: 
				objectTypeToString := 'LAY';
			polyLayerType: 
				objectTypeToString := 'PLA';
			layerVectorType: 
				objectTypeToString := 'LAV';
			polyLayerVectorType: 
				objectTypeToString := 'PLV';
			QDPictureType: 
				objectTypeToString := 'QDP';
			ColorQDPictureType: 
				objectTypeToString := 'CQP';
			pPointType: 
				objectTypeToString := '---';
			twoDTreeType: 
				objectTypeToString := 'T2T';
			twoDTreeNodeType: 
				objectTypeToString := 'T2N';
			dasDBSType: 
				objectTypeToString := '---';
			massStorageType: 
				objectTypeToString := '---';
			dasGFType: 
				objectTypeToString := '---';
			dbsLayerType: 
				objectTypeToString := '---';
			regularNGonType: 
				objectTypeToString := 'RNG';
			doubleListType: 
				objectTypeToString := '---';
			attractorIdType: 
				objectTypeToString := '---';
			listMuchLinkedType: 
				objectTypeToString := '---';
			directedLineSegmentType: 
				objectTypeToString := 'DST';
			idRealPointType: 
				objectTypeToString := 'IDP';
			point3dType: 
				objectTypeToString := '---';
			realPoint3dType: 
				objectTypeToString := 'RP3';
			triangle2dType: 
				objectTypeToString := 'TG2';
			triangulation3dType: 
				objectTypeToString := 'TR3';
			triangulation2dType: 
				objectTypeToString := 'TR2';
			graphEdgeTriangulationType: 
				objectTypeToString := 'GET';
			quadTreeType: 
				objectTypeToString := 'QTT';
			quadTreeNodeType: 
				objectTypeToString := 'QTN';
			triangle3dType: 
				objectTypeToString := 'TG3';
			virtualPoint3dType: 
				objectTypeToString := 'VP3';
			virtualGraphEdgeTriangulationType: 
				objectTypeToString := 'VE3';
			virtualTriangle3dType: 
				objectTypeToString := 'VT3';
			upperMonotonePolygonType: 
				objectTypeToString := '---';
			searchTriangulation2dType: 
				objectTypeToString := '---';
			boundingBox2dType: 
				objectTypeToString := 'BB2';
			cube3dType: 
				objectTypeToString := 'CUB';
			boundingBox3dType: 
				objectTypeToString := 'BB3';
			markedGraphEdgeType: 
				objectTypeToString := 'MGE';
			atomType: 
				objectTypeToString := 'ATO';
			moleculeType: 
				objectTypeToString := 'MOL';
			moleculeEdgeType: 
				objectTypeToString := 'MOE';
			triangleType: 
				objectTypeToString := '---';
			raumPlanungType: 
				objectTypeToString := 'RAU';
			infoValuesType: 
				objectTypeToString := 'IVL';
			resInfoType: 
				objectTypeToString := 'REI';
			gebaeudeType: 
				objectTypeToString := 'GEB';
			beduerfnisseType: 
				objectTypeToString := 'BED';
			resourcenType: 
				objectTypeToString := 'RES';
			histogramType: 
				objectTypeToString := 'HST';
			gebaeudeLabelType: 
				objectTypeToString := '---';
			chiralityType: 
				objectTypeToString := 'CHI';
			atomInformationType: 
				objectTypeToString := 'ATI';
			objType: 
				objectTypeToString := '---';
			otherwise begin
{$IFC THINK_PASCAL }
				objectTypeToString := copy(StringOf(t, '---'), 1, 3);
				fatalError(StringOf('Cannot convert ''', t, ''' to type identifier.'));
{$ELSEC }
				objectTypeToString := copy(StringOf(ord(t), '---'), 1, 3);
				fatalError(StringOf('Cannot convert ''', ord(t), ''' to type identifier.'));
{$ENDC }
			end;
		end;
	end;

procedure writeTextType (t: objectType);
	var
		id: string3;
	begin
		id := objectTypeToString(t);
		writeByte(Byte(id[1]));
		writeByte(Byte(id[2]));
		writeByte(Byte(id[3]));
		if id[1] = '-' then
{$IFC THINK_PASCAL }
			fatalError(StringOf('Try to write an object of type ''', t, ''' as text object.'));
{$ELSEC }
			fatalError(StringOf('Try to write an object of type ''', ord(t), ''' as text object.'));
{$ENDC }
	end;

procedure skipLine;
	begin
		while ch <> CR do
			getCh;
	end;

function readTextType: objectType;
	var
		ch1, ch2, ch3: char;

	procedure illegalTypeError;
		begin
			fatalError(concat('Type identifier expected, got: ''', ch1, ch2, ch3, '''.'));
		end;

	begin
		readTextType := objType;
		skipWhiteSpace;
		ch1 := cap(ch);
		getCh;
		ch2 := cap(ch);
		getCh;
		ch3 := cap(ch);
		getCh;
		case ch1 of
			'A': 
				if (ch2 = 'T') & (ch3 = 'O') then
					readTextType := atomType
				else if (ch2 = 'T') & (ch3 = 'I') then
					readTextType := atomInformationType
				else
					illegalTypeError;
			'B': 
				if (ch2 = 'B') & (ch3 = '2') then
					readTextType := boundingBox2dType
				else if (ch2 = 'B') & (ch3 = '3') then
					readTextType := boundingBox3dType
				else if (ch2 = 'E') & (ch3 = 'D') then
					readTextType := beduerfnisseType
				else
					illegalTypeError;
			'C': 
				if (ch2 = 'I') & (ch3 = 'R') then
					readTextType := circleType
				else if (ch2 = 'P') & (ch3 = 'L') then
					readTextType := convexPolygonType
				else if (ch2 = 'Q') & (ch3 = 'P') then
					readTextType := ColorQDPictureType
				else if (ch2 = 'U') & (ch3 = 'B') then
					readTextType := cube3dType
				else if (ch2 = 'H') & (ch3 = 'I') then
					readTextType := chiralityType
				else
					illegalTypeError;
			'D': 
				if (ch2 = 'C') & (ch3 = 'I') then
					readTextType := dDimCircleType
				else if (ch2 = 'G') & (ch3 = 'R') then
					readTextType := directedGraphType
				else if (ch2 = 'P') & (ch3 = 'T') then
					readTextType := dDimPointType
				else if (ch2 = 'P') & (ch3 = 'V') then
					readTextType := dDimPointVectorType
				else if (ch2 = 'S') & (ch3 = 'T') then
					readTextType := directedLineSegmentType
				else
					illegalTypeError;
			'F': 
				if (ch2 = 'P') & (ch3 = 'T') then
					readTextType := floatPointType
				else
					illegalTypeError;
			'G': 
				if (ch2 = 'E') & (ch3 = 'D') then
					readTextType := graphEdgeType
				else if (ch2 = 'E') & (ch3 = 'T') then
					readTextType := graphEdgeTriangulationType
				else if (ch2 = 'E') & (ch3 = 'B') then
					readTextType := gebaeudeType
				else
					illegalTypeError;
			'H': 
				if (ch2 = 'V') & (ch3 = 'C') then
					readTextType := homogenVectorType
				else if (ch2 = 'S') & (ch3 = 'T') then
					readTextType := histogramType
				else
					illegalTypeError;
			'I': 
				if (ch2 = 'N') & (ch3 = 'T') then
					readTextType := intType
				else if (ch2 = 'P') & (ch3 = 'T') then
					readTextType := longIntPointType
				else if (ch2 = 'D') & (ch3 = 'P') then
					readTextType := idRealPointType
				else if (ch2 = 'V') & (ch3 = 'L') then
					readTextType := infoValuesType
				else
					illegalTypeError;
			'L': 
				if (ch2 = 'S') & (ch3 = 'G') then
					readTextType := lineSegmentType
				else if (ch2 = 'S') & (ch3 = 'T') then
					readTextType := listType
				else if (ch2 = 'V') & (ch3 = 'C') then
					readTextType := lineSegmentVectorType
				else if (ch2 = 'A') & (ch3 = 'V') then
					readTextType := layerVectorType
				else if (ch2 = 'A') & (ch3 = 'Y') then
					readTextType := layerType
				else
					illegalTypeError;
			'M': 
				if (ch2 = 'R') & (ch3 = 'P') then
					readTextType := markedRealPointType
				else if (ch2 = 'O') & (ch3 = 'L') then
					readTextType := moleculeType
				else if (ch2 = 'O') & (ch3 = 'E') then
					readTextType := moleculeEdgeType
				else if (ch2 = 'G') & (ch3 = 'E') then
					readTextType := markedGraphEdgeType
				else
					illegalTypeError;
			'P': 
				if (ch2 = 'O') & (ch3 = 'L') then
					readTextType := polygon2dType
				else if (ch2 = 'L') & (ch3 = 'I') then
					readTextType := polyLineType
				else if (ch2 = 'L') & (ch3 = 'A') then
					readTextType := polyLayerType
				else if (ch2 = 'L') & (ch3 = 'V') then
					readTextType := polyLayerVectorType
				else if (ch2 = 'V') & (ch3 = 'C') then
					readTextType := pointVectorType
				else
					illegalTypeError;
			'Q': 
				if (ch2 = 'D') & (ch3 = 'P') then
					readTextType := QDPictureType
				else if (ch2 = 'T') & (ch3 = 'T') then
					readTextType := quadTreeType
				else if (ch2 = 'T') & (ch3 = 'N') then
					readTextType := quadTreeNodeType;
			'R': 
				if (ch2 = 'P') & (ch3 = 'T') then
					readTextType := realPointType
				else if (ch2 = 'E') & (ch3 = 'C') then
					readTextType := rectangleType
				else if (ch2 = 'V') & (ch3 = 'C') then
					readTextType := rectangleVectorType
				else if (ch2 = 'N') & (ch3 = 'G') then
					readTextType := regularNGonType
				else if (ch2 = 'P') & (ch3 = '3') then
					readTextType := realPoint3dType
				else if (ch2 = 'A') & (ch3 = 'U') then
					readTextType := raumPlanungType
				else if (ch2 = 'E') & (ch3 = 'I') then
					readTextType := resInfoType
				else if (ch2 = 'E') & (ch3 = 'S') then
					readTextType := resourcenType
				else
					illegalTypeError;
			'S': 
				if (ch2 = 'E') & (ch3 = 'D') then
					readTextType := straightEdgeType
				else if (ch2 = 'T') & (ch3 = 'R') then
					readTextType := stringsType
				else if (ch2 = 'P') & (ch3 = '2') then
					readTextType := spline2Type
				else if (ch2 = 'P') & (ch3 = 'L') then
					readTextType := splinegonType
				else if (ch2 = 'P') & (ch3 = 'T') then
					readTextType := spanningTreeType
				else if (ch2 = 'U') & (ch3 = 'G') then
					readTextType := simpleUndirectedGraphType
				else
					illegalTypeError;
			'T': 
				if (ch2 = '2') & (ch3 = 'T') then
					readTextType := twoDTreeType
				else if (ch2 = '2') & (ch3 = 'N') then
					readTextType := twoDTreeNodeType
				else if (ch2 = 'G') & (ch3 = '2') then
					readTextType := triangle2dType
				else if (ch2 = 'G') & (ch3 = '3') then
					readTextType := triangle3dType
				else if (ch2 = 'R') & (ch3 = '3') then
					readTextType := triangulation3dType
				else if (ch2 = 'R') & (ch3 = '2') then
					readTextType := triangulation2dType
				else
					illegalTypeError;
			'U': 
				if (ch2 = 'D') & (ch3 = 'G') then
					readTextType := undirectedGraphType
				else
					illegalTypeError;
			'V': 
				if (ch2 = 'E') & (ch3 = 'C') then
					readTextType := vectorType
				else if (ch2 = 'E') & (ch3 = 'D') then
					readTextType := voronoiEdgeType
				else if (ch2 = 'D') & (ch3 = 'G') then
					readTextType := voronoiDiagramType
				else if (ch2 = 'P') & (ch3 = '3') then
					readTextType := virtualPoint3dType
				else if (ch2 = 'E') & (ch3 = '3') then
					readTextType := virtualGraphEdgeTriangulationType
				else if (ch2 = 'T') & (ch3 = '3') then
					readTextType := virtualTriangle3dType
				else
					illegalTypeError;
			otherwise
				illegalTypeError;
		end;
	end;

procedure writeTextString (x: Str255);
	var
		i: integer;
	begin
		writeByte(Byte(' '));
		writeByte(Byte(''''));
		for i := 1 to length(x) do
			if x[i] = '''' then begin
				writeByte(Byte(''''));
				writeByte(Byte(''''));
			end
			else
				writeByte(Byte(x[i]));
		writeByte(Byte(''''));
	end;

procedure readTextString (var x: Str255);
	var
		done: boolean;
	begin
		skipWhiteSpace;
		if ch <> '''' then
			fatalError(concat(''' expected, got: ''', ch, '''.'));
		done := false;
		x := '';
		repeat
			getCh;
			if ch = '''' then begin
				getCh;
				if ch = '''' then
					x := concat(x, ch)
				else
					done := true
			end
			else
				x := concat(x, ch)
		until done | (length(x) = 250);
		testString(x);
	end;

procedure writeTextReal (x: real);
	begin
		if trunc(x) = x then
{$PUSH }
{$R- }
			writeRawString(stringOf(' ', x : 1 : 0))
{$POP }
		else
			writeRawString(stringOf(' ', x : 1 : 5));
	end;

procedure writeTextExtended (x: extended);
	begin
		if trunc(x) = x then
{$PUSH }
{$R- }
			writeRawString(stringOf(' ', x : 1 : 0))
{$POP }
		else
			writeRawString(stringOf(' ', x : 1 : 8));
	end;

function truncate (s: Str255; l: integer): Str255;
	begin
		if length(s) <= l then
			truncate := s
		else
			truncate := concat(copy(s, 1, l - 1), '');
	end;

procedure readTextReal (var x: real);
	var
		s: Str255;
	begin
		readRawString(s);
		IOCheck(false);
		readString(s, x);
		if ioResult > 0 then begin
			fatalError(concat('Illegal real number, got ''', truncate(s, 10), '''.'));
			x := 0;
		end;
		IOCheck(true);
	end;

procedure readTextExtended (var x: extended);
	var
		s: Str255;
	begin
		readRawString(s);
		IOCheck(false);
		readString(s, x);
		if ioResult > 0 then begin
			fatalError(concat('Illegal extended number, got ''', truncate(s, 10), '''.'));
			x := 0;
		end;
		IOCheck(true);
	end;

procedure writeTextLongInt (x: longInt);
	begin
		writeRawString(stringOf(' ', x : 1));
	end;

procedure readTextLongInt (var x: longInt);
	var
		s: Str255;
	begin
		readRawString(s);
		IOCheck(false);
		readString(s, x);
		if ioResult > 0 then begin
			fatalError(concat('Illegal longInt number, got ''', truncate(s, 10), '''.'));
			x := 0;
		end;
		IOCheck(true);
	end;

procedure writeTextInteger (x: integer);
	begin
		writeRawString(stringOf(' ', x : 1));
	end;

procedure readTextInteger (var x: integer);
	var
		s: Str255;
	begin
		readRawString(s);
		IOCheck(false);
		readString(s, x);
		if ioResult > 0 then begin
			fatalError(concat('Illegal integer number, got ''', truncate(s, 10), '''.'));
			x := 0;
		end;
		IOCheck(true);
	end;

procedure writeTextFloat (x: float);
	begin
		writeRawString(stringOf(' ', x.base : 1, ' ', x.precision : 1, ' ', floatToLine(x)));
	end;

procedure readTextFloat (var x: float);
	var
		base, precision: longInt;
		s: Str255;
	begin
		readTextLongInt(base);
		readTextLongInt(precision);
		readRawString(s);
		lineToFloat(s, base, precision, x);
	end;

procedure readTextHandle (var x: Handle);
	var
		length, i, upto: longInt;
	begin
		readTextLongInt(length);
		if odd(length) then
			length := length + 1;
		x := newHandleFailNil(length);
		i := longInt(x^);
		upto := i + length;
		while i < upto do begin
			readTextInteger(IntegerPtr(i)^);
			i := i + 2;
		end;
	end;

procedure writeTextHandle (x: Handle);
	const
		integersPerLine = 10;
	var
		length, i, upto: longInt;
		k: 0..integersPerLine;
	begin
		length := GetHandleSize(x);
		writeTextLongInt(length);
		if odd(length) then
			length := length + 1;
		i := longInt(x^);
		upto := i + length;
		k := 0;
		while i < upto do begin
			writeTextInteger(IntegerPtr(i)^);
			i := i + 2;
			k := k + 1;
			if (k = integersPerLine) & (i < upto) then begin
				k := 0;
				writeIndentedCR;
			end;
		end;
	end;

function readTextObj: obj;
	var
		o: obj;
	begin
		skipWhiteSpace;
		if ch <> '(' then
			fatalError(concat('( expected, got: ''', ch, '''.'));
		getCh;
		o := newObj(readTextType);
		o.textFileInput;
		readTextObj := o;
		skipWhiteSpace;
		if ch <> ')' then
			fatalError(concat(') expected, got: ''', ch, '''.'));
		getCh;
	end;

procedure writeTextObjGeneral (x: obj; withCR: boolean);
	begin
		if withCR then
			writeIndentedCR;
		writeByte(Byte('('));
		indent := indent + 1;
		writeTextType(x.getType);
		x.textFileOutput;
		indent := indent - 1;
		if prototypes[listType].memberType(x) | prototypes[vectorType].memberType(x) then
			writeIndentedCR;
		writeByte(Byte(')'));
	end;

procedure writeTextObj (x: obj);
	begin
		writeTextObjGeneral(x, true);
	end;

procedure writeTextObjNoCR (x: obj);
	begin
		writeTextObjGeneral(x, false);
	end;

procedure dump (fileName: Str255; o: obj);
	begin
		openWriteStreamTextAppend(fileName, 0, true);
		writeTextObjNoCR(o);
		closeWriteStream;
	end;

procedure initTextIO;
	begin
		initLookAhead;
		indent := 0;
		stopCharacters := [' ', TAB, CR, ')', '{', EOFChar, '%', '#'];
		possibleCreatorTypes[0] := 'ttxt';
		possibleCreatorTypes[1] := 'R*ch';
		possibleCreatorTypes[2] := 'MSWD';
		possibleCreatorTypes[3] := 'MWII';
	end;

end.
