(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit travelingOptimalSimple;

interface

uses
	pointVector;

implementation

{$SETC ATravelingSalesman = TRUE }

const
	maxPts = 66;

type

	tourT = record
			nrOfPoints, curSon: integer;
			points: array[1..maxPts] of integer;
			twoFlag, threeFlag: boolean;
			upperSortBound: integer;
			dist, bestRest: real;
		end;

	stackT = record
			ptr: integer;
			entries: array[1..maxPts] of tourT;
		end;
	stackP = ^stackT;
	stackH = ^stackP;

	distMatrT = array[1..maxPts, 1..maxPts] of real;
	distMatrP = ^distMatrT;
	distMatrH = ^distMatrP;

	twoNNT = array[1..maxPts] of real;
	twoNNP = ^twoNNT;
	twoNNH = ^twoNNP;

procedure InitDistanceMatrix (pv: pointVector; var dM: distMatrT);
	var
		i, j: integer;
		x1, y1, x2, y2: real;
		sqX, sqY: array[1..maxPts] of real;
	begin
		for i := 1 to pv.length do begin
			x1 := point2d(pv.elements^^[i]).getX;
			y1 := point2d(pv.elements^^[i]).getY;
			sqX[i] := x1 * x1;
			sqY[i] := y1 * y1;
		end;
		for i := 1 to pv.length do begin
			dM[i, i] := 0.0;
			for j := i + 1 to pv.length do begin
				x1 := point2d(pv.elements^^[i]).getX;
				y1 := point2d(pv.elements^^[i]).getY;
				x2 := point2d(pv.elements^^[j]).getX;
				y2 := point2d(pv.elements^^[j]).getY;
				dM[i, j] := sqrt(sqX[i] + sqX[j] + sqY[i] + sqY[j] - 2.0 * (x1 * x2 + y1 * y2));
				dM[j, i] := dM[i, j];
			end;
		end;
	end;

function pointVector.TSPSimple: polygon2d;
	var
		stack: stackH;
		i: integer;
		distMatr: distMatrH;
		shTour, curTour: tourT;
		shPoly: polygon2d;
		pt: pointVector;
		cancel: boolean;

	procedure findMinTour (var stack: stackT; var shTour: tourT; totalPts: integer; var distMatr: distMatrT);
		var
			tour: tourT;
			i, j, k, twoCutOff: integer;
			odd: boolean;
			nrOfTours: longint;
{$IFC AnimationEnabled AND ATravelingSalesman }
			grTour, grShTour: polyLine;
{$ENDC }

		begin
{$IFC AnimationEnabled AND ATravelingSalesman }
			currentVectorLength := length;
			new(grTour);
			failNil(Handle(grTour));
			grTour.init;
			grTour.appendSafe(elements^^[1], 1);
			new(grShTour);
			failNil(Handle(grShTour));
			grShTour.init;
			for i := 1 to totalPts do
				grShTour.appendSafe(elements^^[shTour.points[i]], 1);
			grShTour.appendSafe(elements^^[1], 1);
			if animationFlag[ATravelingSalesmanItem] then
				grShTour.interactiveOutput;
{$ENDC }
			twoCutOff := totalPts div 2;
			odd := (totalPts mod 2) = 1;
			nrOfTours := 0;
			while (stack.ptr <> 0) do begin
		(* pop tour *)
				tour := stack.entries[stack.ptr];
				stack.ptr := stack.ptr - 1;
{$IFC AnimationEnabled AND ATravelingSalesman }
				if animationFlag[ATravelingSalesmanItem] then begin
					j := grTour.length;
					for i := tour.nrOfPoints to (j - 1) do
						grTour.removeLast;
				end;
{$ENDC }
				if tour.nrOfPoints + tour.curSon < totalPts then begin
		(* push tour *)
					stack.ptr := stack.ptr + 1;
					stack.entries[stack.ptr] := tour;
					stack.entries[stack.ptr].curSon := tour.curSon + 1;
				end;
		(* append son *)
				k := tour.points[tour.nrOfPoints + tour.curSon];
				tour.twoFlag := tour.twoFlag or (k = 2);
				tour.threeFlag := tour.threeFlag or (k = 3);
				tour.nrOfPoints := tour.nrOfPoints + 1;
				tour.points[tour.nrOfPoints - 1 + tour.curSon] := tour.points[tour.nrOfPoints];
				tour.points[tour.nrOfPoints] := k;
				tour.dist := tour.dist + distMatr[k, tour.points[tour.nrOfPoints - 1]];
{$IFC AnimationEnabled AND ATravelingSalesman }
				if animationFlag[ATravelingSalesmanItem] then begin
					grTour.appendSafe(elements^^[k], 1);
					grTour.interactiveHighlight;
					displayInfoString(stringOf('best: ', shTour.dist : 5 : 2, '     current: ', tour.dist : 5 : 2));
					waitForClick(animationFlag[ATravelingSalesmanItem]);
					grTour.interactiveHighlight;
				end;
{$ENDC }
				if tour.nrOfPoints < (totalPts - 1) then begin
					tour.curSon := 1;
					if (tour.nrOfPoints = twoCutOff) and not tour.twoFlag then begin
		(* next Point has to be 2, for an even nr of Pts 3 has to be visited already *)
						if odd or tour.threeFlag then begin
							tour.dist := tour.dist + distMatr[2, tour.points[tour.nrOfPoints]];
{$IFC AnimationEnabled AND ATravelingSalesman }
							if animationFlag[ATravelingSalesmanItem] then begin
								grTour.appendSafe(elements^^[2], 1);
								grTour.interactiveHighlight;
								displayInfoString(stringOf('best: ', shTour.dist : 5 : 2, '     current: ', tour.dist : 5 : 2));
								waitForClick(animationFlag[ATravelingSalesmanItem]);
								grTour.interactiveHighlight;
							end;
{$ENDC }
							i := tour.nrOfPoints + 1;
							tour.nrOfPoints := i;
							while tour.points[i] <> 2 do
								i := i + 1;
							j := tour.points[tour.nrOfPoints];
							tour.points[tour.nrOfPoints] := 2;
							tour.points[i] := j;
							tour.twoFlag := true;
			(* push tour  *)
							stack.ptr := stack.ptr + 1;
							stack.entries[stack.ptr] := tour;
						end;
					end
					else begin
		(* push tour *)
						stack.ptr := stack.ptr + 1;
						stack.entries[stack.ptr] := tour;
					end;
				end
				else begin
		(* complete tour *)
					nrOfTours := nrOfTours + 1;
					k := tour.points[totalPts];
					tour.dist := tour.dist + distMatr[tour.points[tour.nrOfPoints], k] + distMatr[1, k];
{$IFC AnimationEnabled AND ATravelingSalesman }
					if animationFlag[ATravelingSalesmanItem] then begin
						grTour.appendSafe(elements^^[k], 1);
						grTour.appendSafe(elements^^[1], 1);
						grTour.interactiveHighlight;
						displayInfoString(stringOf('best: ', shTour.dist : 5 : 2, '     current: ', tour.dist : 5 : 2));
						waitForClick(animationFlag[ATravelingSalesmanItem]);
						grTour.interactiveHighlight;
					end;
{$ENDC }
					cancel := userAbort;
{$IFC AnimationEnabled AND ATravelingSalesman }
					if cancel then begin
						grTour.free;
						grShTour.free;
					end;
{$ENDC }
					if cancel then
						exit(findMinTour);
{$IFC AnimationEnabled AND ATravelingSalesman }
					if animationFlag[ATravelingSalesmanItem] then begin
						if tour.dist < shTour.dist then begin
							grShTour.interactiveOutput;
							grShTour.length := 0;
							for i := 1 to totalPts do
								grShTour.appendSafe(elements^^[tour.points[i]], 1);
							grShTour.appendSafe(elements^^[1], 1);
							grShTour.interactiveOutput;
						end;
					end;
{$ENDC }
					if tour.dist < shTour.dist then
						shTour := tour;
				end;
			end; (*while *)
			displayInfoString(stringOf('Number of calculated full tours: ', nrOfTours : 1));
{$IFC AnimationEnabled AND ATravelingSalesman }
			grTour.free;
			grShTour.free;
{$ENDC }
		end;

	begin
		if (length > maxPts) | (length <= 2) then begin
			nonFatalError('Too few or too many points...sorry');
			TSPSimple := nil;
		end
		else begin
			cancel := false;
			new(stack);
			failNil(Handle(stack));
			new(stack^);
			failNil(Handle(stack^));
			new(distMatr);
			failNil(Handle(distMatr));
			new(distMatr^);
			failNil(Handle(distMatr^));
(* determine an initial shortest tour *)
			currentVectorLength := length;
			new(shPoly);
			failNil(Handle(shPoly));
			shPoly.init;
			for i := 1 to length do begin
				shTour.points[i] := i;
				shPoly.append(elements^^[i].duplicateAll);
			end;
			shPoly.length := length;
			shTour.dist := shPoly.perimeter;
			stack^^.ptr := 1;
			curTour.nrOfPoints := 1;
			curTour.curSon := 1;
			curTour.twoFlag := false;
			curTour.threeFlag := false;
			curTour.dist := 0.0;
			for i := 1 to length do
				curTour.points[i] := i;
(* initializing distance matrix *)
			InitDistanceMatrix(pointVector(shPoly), distMatr^^);
			stack^^.entries[1] := curTour;
(* search the tree! *)
			findMinTour(stack^^, shTour, length, distMatr^^);
			if cancel then
				abortFlag := false;
(* construct the polygon *)
			currentVectorLength := length;
			new(pt);
			failNil(Handle(pt));
			pt.init;
			for i := 2 to length do
				pt.elements^^[i] := shPoly.elements^^[shTour.points[i]];
			for i := 2 to length do
				shPoly.elements^^[i] := pt.elements^^[i];
			TSPSimple := shPoly;
			pt.free;
			dispose(stack^);
			dispose(stack);
			dispose(distMatr^);
			dispose(distMatr);
		end;
	end;

{$IFC false }
function pointVector.TSPSimpleBranch (initTour: boolean): polygon2d; (* by Adrian Brngger *)
	var
		stack: stackH;
		i: integer;
		distMatr: distMatrH;
		shTour, curTour: tourT;
		shPoly: polygon2d;
		pt: pointVector;
		cancel, saveAbortEnable: boolean;

	procedure findMinTour (var stack: stackT; var shTour: tourT; totalPts: integer; var distMatr: distMatrT);
		var
			tour: tourT;
			i, j, k, l, twoCutOff: integer;
			odd: boolean;
			nrOfTours: longint;

		begin
			twoCutOff := totalPts div 2;
			odd := (totalPts mod 2) = 1;
			nrOfTours := 0;
			while (stack.ptr <> 0) do begin
		(* pop tour *)
				tour := stack.entries[stack.ptr];
				stack.ptr := stack.ptr - 1;
				if tour.nrOfPoints + tour.curSon < totalPts then begin
		(* push tour *)
					stack.ptr := stack.ptr + 1;
					stack.entries[stack.ptr] := tour;
					stack.entries[stack.ptr].curSon := tour.curSon + 1;
				end;
		(* append son *)
				k := tour.points[tour.nrOfPoints + tour.curSon];
				tour.twoFlag := tour.twoFlag or (k = 2);
				tour.threeFlag := tour.threeFlag or (k = 3);
				tour.nrOfPoints := tour.nrOfPoints + 1;
				tour.points[tour.nrOfPoints - 1 + tour.curSon] := tour.points[tour.nrOfPoints];
				tour.points[tour.nrOfPoints] := k;
				tour.dist := tour.dist + distMatr[k, tour.points[tour.nrOfPoints - 1]];
				if tour.nrOfPoints < (totalPts - 1) then begin
					tour.curSon := 1;
					if (tour.nrOfPoints = twoCutOff) and not tour.twoFlag then begin
		(* next Point has to be 2, for an even nr of Pts 3 has to be visited already *)
						if odd or tour.threeFlag then begin
							tour.dist := tour.dist + distMatr[2, tour.points[tour.nrOfPoints]];
							if tour.dist < shTour.dist then begin
								i := tour.nrOfPoints + 1;
								tour.nrOfPoints := i;
								while tour.points[i] <> 2 do
									i := i + 1;
								j := tour.points[tour.nrOfPoints];
								tour.points[tour.nrOfPoints] := 2;
								tour.points[i] := j;
								tour.twoFlag := true;
			(* push tour  *)
								stack.ptr := stack.ptr + 1;
								stack.entries[stack.ptr] := tour;
							end;
						end;
					end
					else if tour.dist < shTour.dist then begin
		(* push tour *)
						stack.ptr := stack.ptr + 1;
						stack.entries[stack.ptr] := tour;
					end;
				end
				else begin
		(* complete tour *)
					nrOfTours := nrOfTours + 1;
					k := tour.points[totalPts];
					tour.dist := tour.dist + distMatr[tour.points[tour.nrOfPoints], k] + distMatr[1, k];
					cancel := userAbort;
					if cancel then
						exit(findMinTour);
					if tour.dist < shTour.dist then begin
						shTour := tour;
(* construct the polygon *)
						shPoly.interactiveOutput;
						for l := 2 to length do
							pt.elements^^[l] := shPoly.elements^^[shTour.points[l]];
						for l := 2 to length do
							shPoly.elements^^[l] := pt.elements^^[l];
						shPoly.interactiveOutput;
					end;
				end;
			end; (*while *)
			displayInfoString(stringOf('Number of calculated full tours: ', nrOfTours : 1));
		end;

	begin
		if (length > MaxPts) | (length <= 2) then begin
			nonFatalError('Too few or too many points...sorry');
			TSPSimpleBranch := nil;
		end
		else begin
			cancel := false;
			new(stack);
			failNil(Handle(stack));
			new(stack^);
			failNil(Handle(stack^));
			new(distMatr);
			failNil(Handle(distMatr));
			new(distMatr^);
			failNil(Handle(distMatr^));
(* determine an initial shortest tour *)
			if initTour then begin
				saveAbortEnable := abortEnable;
				abortEnable := false;
				shPoly := TSPConvexHull;
				abortEnable := saveAbortEnable;
				shTour.dist := shPoly.perimeter;
				shPoly.interactiveOutput;
				for i := 1 to length do
					shTour.points[i] := i;
			end
			else begin
				currentVectorLength := length;
				new(shPoly);
				failNil(Handle(shPoly));
				shPoly.init;
				for i := 1 to length do begin
					shTour.points[i] := i;
					shPoly.append(elements^^[i].duplicateAll);
				end;
				shPoly.length := length;
				shTour.dist := shPoly.perimeter;
			end;
			stack^^.ptr := 1;
			curTour.nrOfPoints := 1;
			curTour.curSon := 1;
			curTour.twoFlag := false;
			curTour.threeFlag := false;
			curTour.dist := 0.0;
			for i := 1 to length do
				curTour.points[i] := i;
(* initializing distance matrix *)
			InitDistanceMatrix(pointVector(shPoly), distMatr^^);
			stack^^.entries[1] := curTour;
			currentVectorLength := length;
			new(pt);
			failNil(Handle(pt));
			pt.init;
    (* search the tree! *)
			findMinTour(stack^^, shTour, length, distMatr^^);
			shPoly.interactiveOutput;
			pt.free;
			if not cancel then begin
(* construct the polygon *)
				TSPSimpleBranch := shPoly;
			end
			else begin
				TSPSimpleBranch := nil;
				shPoly.freeAll;
			end;
			dispose(stack^);
			dispose(stack);
			dispose(distMatr^);
			dispose(distMatr);
		end;
	end;


function pointVector.TSPConsConvHullOrder: polygon2d; (* by Adrian Brngger *)
	var
		stack: stackH;
		i, j, l: integer;
		distMatr: distMatrH;
		shTour, curTour: tourT;
		shPoly, pt: polygon2d;
		cH, cHH: pointVector;
		hullLength, eliminated: seqIndex0;
		cancel, saveAbortEnable: boolean;

	procedure SortSons (var tour: tourT; var distMatr: distMatrT; totalPtsToSort: integer);
		var
			i, j, k, l: integer;
			d: real;
		begin
(* sort tour.points[tour.nrOfPoints+1...maxPts] according to shortest Distance to tour.points[tour.nrOfPoints] *)
			with tour do begin
				l := points[nrOfPoints];
				for i := nrOfPoints + 2 to totalPtsToSort do begin
					k := points[i];
					j := i - 1;
					d := distMatr[k, l];
					while (distMatr[points[j], l] > d) and (j > nrOfPoints) do begin
						points[j + 1] := points[j];
						j := j - 1;
					end;
					points[j + 1] := k;
				end;
			end;
		end;

	procedure findMinTour (var stack: stackT; var shTour: tourT; totalPts: integer; var distMatr: distMatrT);
		var
			tour: tourT;
			k, j, l: integer;
			nrOfTours: longint;
		begin
			nrOfTours := 0;
			while stack.ptr <> 0 do begin
		(* pop tour *)
				tour := stack.entries[stack.ptr];
				stack.ptr := stack.ptr - 1;
				if tour.nrOfPoints = (tour.upperSortBound - 1) then begin
		(* complete tour *)
					nrOfTours := nrOfTours + 1;
					for j := tour.nrOfPoints to totalPts - 1 do begin
						k := tour.points[j];
						l := tour.points[j + 1];
						tour.dist := tour.dist + distMatr[l, k];
					end;
					tour.nrOfPoints := totalPts;
					tour.dist := tour.dist + distMatr[1, l];
					cancel := userAbort;
					if cancel then
						exit(findMinTour);
		(* is it the shortest? *)
					if tour.dist < shTour.dist then begin
						shTour := tour;
						pt.interactiveOutput;
						for l := 1 to length do
							pt.elements^^[l] := cH.elements^^[shTour.points[l]];
						pt.length := length;
						pt.interactiveOutput;
					end;
				end
				else begin
					tour.curSon := tour.curSon + 1;
					if (tour.curSon + tour.nrOfPoints) < tour.upperSortBound then begin
		(* push tour *)
						stack.ptr := stack.ptr + 1;
						stack.entries[stack.ptr] := tour;
					end;
		(* append son to tour *)
					k := tour.points[tour.nrOfPoints + tour.curSon];
					tour.nrOfPoints := tour.nrOfPoints + 1;
					j := tour.nrOfPoints;
					tour.points[j - 1 + tour.curSon] := tour.points[j];
					tour.points[j] := k;
					tour.dist := tour.dist + distMatr[k, tour.points[j - 1]];
		(* should we go further with this tour? *)
					if tour.dist > shTour.dist then begin
		(* pop tour because sons are sorted! *)
						if stack.ptr > 0 then begin
							tour := stack.entries[stack.ptr];
							stack.ptr := stack.ptr - 1;
						end;
					end
					else begin
		(* if the new son is on the convex Hull adjust sortBound *)
						if k < hullLength then
							tour.upperSortBound := tour.upperSortBound + 1;
		(* sort sons if there are at least 2 sons *)
						if tour.nrOfPoints < (tour.upperSortBound - 1) then
							SortSons(tour, distMatr, tour.upperSortBound);
						tour.curSon := 0;
		(* push tour *)
						stack.ptr := stack.ptr + 1;
						stack.entries[stack.ptr] := tour;
					end;
				end;
			end; (*while *)
			displayInfoString(stringOf('Number of calculated full tours: ', nrOfTours : 1));
		end;

	begin
		if (length > 13) | (length <= 2) then begin
			nonFatalError('Too few or too many points...sorry');
			TSPConsConvHullOrder := nil;
		end
		else begin
(* determine the convex hull *)
			saveAbortEnable := abortEnable;
			abortEnable := false;
			cH := pointVector(duplicateAll);
			cH.convexHull(true, false, hullLength, eliminated);
			abortEnable := saveAbortEnable;
			if hullLength < length then begin
				new(stack);
				failNil(Handle(stack));
				new(stack^);
				failNil(Handle(stack^));
				new(distMatr);
				failNil(Handle(distMatr));
				new(distMatr^);
				failNil(Handle(distMatr^));
(* determine start tour *)
				curTour.nrOfPoints := 1;
				curTour.curSon := 0;
				curTour.twoFlag := false;
				curTour.dist := 0.0;
				curTour.points[1] := 1;
				j := length - hullLength;
				curTour.upperSortBound := j + 2;
				for i := 1 to j do
					curTour.points[i + 1] := i + hullLength;
				for i := j + 2 to length do
					curTour.points[i] := i - j;

(* determine an initial shortest tour *)
				currentVectorLength := length;
				saveAbortEnable := abortEnable;
				abortEnable := false;
				cHH := pointVector(cH.duplicateAll);    	(* cHH is necessary because TSPConvexHull changes the	*)
				shPoly := cHH.TSPConvexHull;				(* order of the points!												*)
				shPoly.interactiveOutput;
				cHH.freeAll;
				abortEnable := saveAbortEnable;
				shTour.dist := shPoly.perimeter;
				for i := 1 to length do begin
					j := 1;
					while not realPoint(shPoly.elements^^[i]).equal(cH.elements^^[j]) do
						j := j + 1;
					shTour.points[i] := j;
				end;
				shPoly.freeAll;
(* initializing distance matrix *)
				InitDistanceMatrix(pointVector(cH), distMatr^^);
(* sorting sons of start tour and pushing it *)
				SortSons(curTour, distMatr^^, curTour.upperSortBound);
				stack^^.ptr := 1;
				stack^^.entries[1] := curTour;
(* search the tree! *)
				currentVectorLength := length;
				new(pt);
				failNil(Handle(pt));
				pt.init;
				for l := 1 to length do
					pt.elements^^[l] := cH.elements^^[shTour.points[l]];
				pt.length := length;
				findMinTour(stack^^, shTour, length, distMatr^^);
				pt.interactiveOutput;
				dispose(stack^);
				dispose(stack);
				dispose(distMatr^);
				dispose(distMatr);
				if not cancel then begin
(* construct the polygon *)
					TSPConsConvHullOrder := pt;
					cH.free;
				end
				else begin
					TSPConsConvHullOrder := nil;
					cH.freeAll;
					pt.free;
				end;
			end
			else begin
(* cH is the optimal tour *)
				currentVectorLength := length;
				new(pt);
				failNil(Handle(pt));
				pt.init;
				for i := 1 to length do
					pt.elements^^[i] := cH.elements^^[i];
				pt.length := length;
				TSPConsConvHullOrder := pt;
				cH.free;
			end;
		end;
	end;

function pointVector.TSPNoInters: polygon2d;
	var
		stack: stackH;
		i: integer;
		distMatr: distMatrH;
		shTour, curTour: tourT;
		shPoly: polygon2d;
		pt: pointVector;
		cancel, saveAbortEnable, noInters: boolean;

	procedure findMinTour (var stack: stackT; var shTour: tourT; totalPts: integer; var distMatr: distMatrT);
		const
			IntersFromBeginning = 6;
			IntersFromEnd = 50;

		var
			tour: tourT;
			i, j, k, twoCutOff: integer;
			odd: boolean;
			nrOfTours: longint;
			left, right: realPoint;
{$IFC AnimationEnabled AND ATravelingSalesman }
			grTour, grShTour: polyLine;
			shPerim, curPerim: markedRealPoint;
			locShPerim, locCurPerim: realPoint;
{$ENDC }

		function TestIntersection (var tour: tourT): boolean;
(* returns true iff there are no intersections between the last line segment in the tour and the beginning of the tour *)
			var
				p, q, r, s: realPoint;
				i: integer;
				nextP, res: boolean;
			begin
				if tour.nrOfPoints < 4 then
					TestIntersection := true
				else begin
					p := realPoint(elements^^[tour.points[1]]);
					q := realPoint(elements^^[tour.points[2]]);
					r := realPoint(elements^^[tour.points[tour.nrOfPoints]]);
					s := realPoint(elements^^[tour.points[tour.nrOfPoints - 1]]);
					i := 3;
					nextP := true;
					res := p.intersectLineSegmentX(q, r, s, left, right);
					while (i < (tour.nrOfPoints - 1)) and (i < (3 + IntersFromBeginning)) and not res do begin
						if nextP then
							p := realPoint(elements^^[tour.points[i]])
						else
							q := realPoint(elements^^[tour.points[i]]);
						i := i + 1;
						nextP := not (nextP);
						res := p.intersectLineSegmentX(q, r, s, left, right);
					end;
					TestIntersection := not res;
				end;
			end;

		begin
{$IFC AnimationEnabled AND ATravelingSalesman }
			currentVectorLength := length;
			new(grTour);
			failNil(Handle(grTour));
			grTour.init;
			grTour.appendSafe(elements^^[1], 1);
			new(shPerim);
			failNil(Handle(shPerim));
			new(curPerim);
			failNil(Handle(curPerim));
			curPerim.init;
			new(locShPerim);
			failNil(Handle(locShPerim));
			locShPerim.init;
			locShPerim.setX(20.0);
			locShPerim.setY(20.0);
			new(locCurPerim);
			failNil(Handle(locCurPerim));
			locCurPerim.init;
			locCurPerim.setX(20.0);
			locCurPerim.setY(50.0);
			shPerim.init;
			shPerim.initLocationMark(locShPerim, stringOf(shTour.dist : 5 : 2), false);
			new(grShTour);
			failNil(Handle(grShTour));
			grShTour.init;
			for i := 1 to totalPts do
				grShTour.appendSafe(elements^^[shTour.points[i]], 1);
			grShTour.appendSafe(elements^^[1], 1);
			if animationFlag[ATravelingSalesmanItem] then
				grShTour.interactiveOutput;
{$ENDC }
			twoCutOff := totalPts div 2;
			odd := (totalPts mod 2) = 1;
			nrOfTours := 0;
			new(left);
			failNil(Handle(left));
			new(right);
			failNil(Handle(right));
			left.init;
			right.init;
			while (stack.ptr <> 0) do begin
		(* pop tour *)
				tour := stack.entries[stack.ptr];
				stack.ptr := stack.ptr - 1;
{$IFC AnimationEnabled AND ATravelingSalesman }
				if animationFlag[ATravelingSalesmanItem] then begin
					j := grTour.length;
					for i := tour.nrOfPoints to (j - 1) do
						grTour.removeLast;
				end;
{$ENDC }
				if tour.nrOfPoints + tour.curSon < totalPts then begin
		(* push tour *)
					stack.ptr := stack.ptr + 1;
					stack.entries[stack.ptr] := tour;
					stack.entries[stack.ptr].curSon := tour.curSon + 1;
				end;
		(* append son *)
				k := tour.points[tour.nrOfPoints + tour.curSon];
				tour.twoFlag := tour.twoFlag or (k = 2);
				tour.threeFlag := tour.threeFlag or (k = 3);
				tour.nrOfPoints := tour.nrOfPoints + 1;
				tour.points[tour.nrOfPoints - 1 + tour.curSon] := tour.points[tour.nrOfPoints];
				tour.points[tour.nrOfPoints] := k;
				tour.dist := tour.dist + distMatr[k, tour.points[tour.nrOfPoints - 1]];
				noInters := true;
				noInters := TestIntersection(tour);
{$IFC AnimationEnabled AND ATravelingSalesman }
				if animationFlag[ATravelingSalesmanItem] then begin
					grTour.appendSafe(elements^^[k], 1);
					grTour.interactiveHighlight;
					curPerim.initLocationMark(locCurPerim, stringOf(tour.dist : 5 : 2), false);
					curPerim.interactiveOutput;
					shPerim.interactiveOutput;
					waitForClick(animationFlag[ATravelingSalesmanItem]);
					curPerim.interactiveOutput;
					shPerim.interactiveOutput;
					grTour.interactiveHighlight;
				end;
{$ENDC }
				if tour.nrOfPoints < (totalPts - 1) then begin
					tour.curSon := 1;
					if (tour.nrOfPoints = twoCutOff) and not tour.twoFlag then begin
		(* next Point has to be 2, for an even nr of Pts 3 has to be visited already *)
						if odd or tour.threeFlag then begin
							tour.dist := tour.dist + distMatr[2, tour.points[tour.nrOfPoints]];
{$IFC AnimationEnabled AND ATravelingSalesman }
							if animationFlag[ATravelingSalesmanItem] then begin
								grTour.appendSafe(elements^^[2], 1);
								grTour.interactiveHighlight;
								curPerim.initLocationMark(locCurPerim, stringOf(tour.dist : 5 : 2), false);
								curPerim.interactiveOutput;
								shPerim.interactiveOutput;
								waitForClick(animationFlag[ATravelingSalesmanItem]);
								curPerim.interactiveOutput;
								shPerim.interactiveOutput;
								grTour.interactiveHighlight;
							end;
{$ENDC }
							i := tour.nrOfPoints + 1;
							tour.nrOfPoints := i;
							while tour.points[i] <> 2 do
								i := i + 1;
							j := tour.points[tour.nrOfPoints];
							tour.points[tour.nrOfPoints] := 2;
							tour.points[i] := j;
							tour.twoFlag := true;
			(* push tour  *)
							if TestIntersection(tour) and noInters then begin
								stack.ptr := stack.ptr + 1;
								stack.entries[stack.ptr] := tour;
							end;
						end;
					end
					else if noInters then begin
		(* push tour *)
						stack.ptr := stack.ptr + 1;
						stack.entries[stack.ptr] := tour;
					end;
				end
				else begin
		(* complete tour *)
					nrOfTours := nrOfTours + 1;
					k := tour.points[totalPts];
					tour.dist := tour.dist + distMatr[tour.points[tour.nrOfPoints], k] + distMatr[1, k];
{$IFC AnimationEnabled AND ATravelingSalesman }
					if animationFlag[ATravelingSalesmanItem] then begin
						grTour.appendSafe(elements^^[k], 1);
						grTour.appendSafe(elements^^[1], 1);
						grTour.interactiveHighlight;
						curPerim.initLocationMark(locCurPerim, stringOf(tour.dist : 5 : 2), false);
						curPerim.interactiveOutput;
						shPerim.interactiveOutput;
						waitForClick(animationFlag[ATravelingSalesmanItem]);
						curPerim.interactiveOutput;
						shPerim.interactiveOutput;
						grTour.interactiveHighlight;
					end;
{$ENDC }
					cancel := userAbort;
{$IFC AnimationEnabled AND ATravelingSalesman }
					if cancel then begin
						grTour.free;
						grShTour.free;
						shPerim.freeAll;
						curPerim.freeAll;
						locShPerim.freeAll;
						locCurPerim.freeAll;
					end;
{$ENDC }
					if cancel then begin
						left.freeAll;
						right.freeAll;
						exit(findMinTour);
					end;
{$IFC AnimationEnabled AND ATravelingSalesman }
					if animationFlag[ATravelingSalesmanItem] then begin
						if tour.dist < shTour.dist then begin
							SysBeep(2);
							shPerim.initLocationMark(locShPerim, stringOf(tour.dist : 5 : 2), false);
							grShTour.interactiveOutput;
							grShTour.length := 0;
							for i := 1 to totalPts do
								grShTour.appendSafe(elements^^[tour.points[i]], 1);
							grShTour.appendSafe(elements^^[1], 1);
							grShTour.interactiveOutput;
						end;
					end;
{$ENDC }
					if tour.dist < shTour.dist then
						shTour := tour;
				end;
			end; (*while *)
			left.freeAll;
			right.freeAll;
			displayInfoString(stringOf('Number of calculated full tours: ', nrOfTours : 1));
{$IFC AnimationEnabled AND ATravelingSalesman }
			grTour.free;
			grShTour.free;
			shPerim.freeAll;
			curPerim.freeAll;
			locShPerim.freeAll;
			locCurPerim.freeAll;
{$ENDC }
		end;

	begin
		if (length <= maxPts) and (length > 2) then begin
			cancel := false;
			new(stack);
			failNil(Handle(stack));
			new(stack^);
			failNil(Handle(stack^));
			new(distMatr);
			failNil(Handle(distMatr));
			new(distMatr^);
			failNil(Handle(distMatr^));
(* determine an initial shortest tour *)
			currentVectorLength := length;
			new(shPoly);
			failNil(Handle(shPoly));
			shPoly.init;
			for i := 1 to length do begin
				shTour.points[i] := i;
				shPoly.append(elements^^[i].duplicateAll);
			end;
			shPoly.length := length;
			shTour.dist := shPoly.perimeter;
			stack^^.ptr := 1;
			curTour.nrOfPoints := 1;
			curTour.curSon := 1;
			curTour.twoFlag := false;
			curTour.threeFlag := false;
			curTour.dist := 0.0;
			for i := 1 to length do
				curTour.points[i] := i;
(* initializing distance matrix *)
			InitDistanceMatrix(pointVector(shPoly), distMatr^^);
			stack^^.entries[1] := curTour;
(* search the tree! *)
			findMinTour(stack^^, shTour, length, distMatr^^);
			if not cancel then begin
(* construct the polygon *)
				currentVectorLength := length;
				new(pt);
				failNil(Handle(pt));
				pt.init;
				for i := 2 to length do
					pt.elements^^[i] := shPoly.elements^^[shTour.points[i]];
				for i := 2 to length do
					shPoly.elements^^[i] := pt.elements^^[i];
				TSPNoInters := shPoly;
				pt.free;
			end
			else begin
				TSPNoInters := nil;
				shPoly.freeAll;
			end;
			dispose(stack^);
			dispose(stack);
			dispose(distMatr^);
			dispose(distMatr);
		end;
	end;


function pointVector.TSP2NNBranchAndBound: polygon2d;
	var
		stack: stackH;
		i, j: integer;
		distMatr: distMatrH;
		twoNN: twoNNH;
		shTour, curTour: tourT;
		shPoly, sPoly: polygon2d;
		pt: pointVector;
		cancel, saveAbortEnable: boolean;
		min1, min2: real;

	procedure SortSons (var tour: tourT; var distMatr: distMatrT; totalPts: integer);
		var
			i, j, k, l: integer;
			d: real;
		begin
(* sort tour.points[tour.nrOfPoints+1...maxPts] according to shortest Distance to tour.points[tour.nrOfPoints] *)
			with tour do begin
				l := points[nrOfPoints];
				for i := nrOfPoints + 2 to totalPts do begin
					k := points[i];
					j := i - 1;
					d := distMatr[k, l];
					while (distMatr[points[j], l] > d) and (j > nrOfPoints) do begin
						points[j + 1] := points[j];
						j := j - 1;
					end;
					points[j + 1] := k;
				end;
			end;
		end;



	procedure findMinTour (var stack: stackT; var shTour: tourT; totalPts: integer; var distMatr: distMatrT; var twoNN: twoNNT);
		var
			tour: tourT;
			i, j, k, twoCutOff: integer;
			odd: boolean;
			nrOfTours: longint;
			pot: pointVector;

		procedure ShowShTour;
			var
				i: integer;
			begin
				sPoly.interactiveOutput;
				for i := 2 to length do
					pot.elements^^[i] := shPoly.elements^^[shTour.points[i]];
				for i := 2 to length do
					sPoly.elements^^[i] := pot.elements^^[i];
				sPoly.interactiveOutput;
			end;

		begin
			nrOfTours := 0;
			twoCutOff := totalPts div 2;
			odd := (totalPts mod 2) = 1;
			currentVectorLength := length;
			new(pot);
			failNil(Handle(pot));
			pot.init;
			while (stack.ptr <> 0) do begin
		(* pop tour *)
				tour := stack.entries[stack.ptr];
				stack.ptr := stack.ptr - 1;
				if tour.nrOfPoints = (totalPts - 1) then begin
		(* complete tour *)
					nrOfTours := nrOfTours + 1;
					tour.nrOfPoints := totalPts;
					k := tour.points[totalPts];
					tour.dist := tour.dist + distMatr[tour.points[totalPts - 1], k] + distMatr[1, k];
					tour.bestRest := tour.bestRest - twoNN[k];
					cancel := userAbort;
					if cancel then begin
						exit(findMinTour);
						ShowShTour;
					end;
		(* is it the shortest? *)
					if tour.dist < shTour.dist then begin
						shTour := tour;
						ShowShTour;
					end;
				end
				else begin
					tour.curSon := tour.curSon + 1;
					if (tour.curSon + tour.nrOfPoints) < totalPts then begin
		(* push tour *)
						stack.ptr := stack.ptr + 1;
						stack.entries[stack.ptr] := tour;
					end;
		(* append son to tour *)
					k := tour.points[tour.nrOfPoints + tour.curSon];
					tour.twoFlag := tour.twoFlag or (k = 2);
					tour.threeFlag := tour.threeFlag or (k = 3);
					tour.nrOfPoints := tour.nrOfPoints + 1;
					j := tour.nrOfPoints;
					tour.points[j - 1 + tour.curSon] := tour.points[j];
					tour.points[j] := k;
					tour.dist := tour.dist + distMatr[k, tour.points[j - 1]];
					tour.bestRest := tour.bestRest - twoNN[k];
		(* should we go further with this tour? *)
					if (tour.dist + tour.bestRest) > shTour.dist then begin
		(* pop tour because sons are sorted! *)
						if (tour.nrOfPoints + tour.curSon) <> (totalPts + 1) then
							stack.ptr := stack.ptr - 1;
					end
					else begin
						tour.curSon := 0;
						if (tour.nrOfPoints = twoCutOff) and not tour.twoFlag then begin
		(* next Point has to be 2, for an even nr of Pts 3 has to be visited already *)
							if odd or tour.threeFlag then begin
								tour.dist := tour.dist + distMatr[2, tour.points[tour.nrOfPoints]];
								tour.bestRest := tour.bestRest - twoNN[2];
								i := tour.nrOfPoints + 1;
								tour.nrOfPoints := i;
								while tour.points[i] <> 2 do
									i := i + 1;
								j := tour.points[tour.nrOfPoints];
								tour.points[tour.nrOfPoints] := 2;
								tour.points[i] := j;
								tour.twoFlag := true;
			(* sort sons if there are at least 2 sons *)
								if tour.nrOfPoints < (totalPts - 1) then
									SortSons(tour, distMatr, totalPts);
       		(* push tour  *)
								stack.ptr := stack.ptr + 1;
								stack.entries[stack.ptr] := tour;
							end;
						end
						else begin
		(* sort sons if there are at least 2 sons *)
							if tour.nrOfPoints < (totalPts - 1) then
								SortSons(tour, distMatr, totalPts);
       	(* push tour *)
							stack.ptr := stack.ptr + 1;
							stack.entries[stack.ptr] := tour;
						end;
					end;
				end;
			end; (*while *)
			displayInfoString(stringOf('Number of calculated full tours: ', nrOfTours : 1));
			pot.free;
		end;


	begin
		if (length > 14) | (length <= 2) then begin
			nonFatalError('Too few or too many points...sorry');
			TSP2NNBranchAndBound := nil;
		end
		else begin
			cancel := false;
			new(stack);
			failNil(Handle(stack));
			new(stack^);
			failNil(Handle(stack^));
			new(distMatr);
			failNil(Handle(distMatr));
			new(distMatr^);
			failNil(Handle(distMatr^));
			new(twoNN);
			failNil(Handle(twoNN));
			new(twoNN^);
			failNil(Handle(twoNN^));
(* determine an initial shortest tour *)
			currentVectorLength := length;
			new(shPoly);
			failNil(Handle(shPoly));
			shPoly.init;
			for i := 1 to length do begin
				shTour.points[i] := i;
				shPoly.append(elements^^[i].duplicateAll);
			end;
			shPoly.interactiveOutput;
			sPoly := polygon2d(shPoly.duplicateAll);
			shTour.dist := shPoly.perimeter;
			curTour.nrOfPoints := 1;
			curTour.curSon := 0;
			curTour.twoFlag := false;
			curTour.threeFlag := false;
			curTour.dist := 0.0;
			for i := 1 to length do
				curTour.points[i] := i;
(* initializing distance matrix *)
			InitDistanceMatrix(pointVector(shPoly), distMatr^^);
			SortSons(curTour, distMatr^^, length);
(* initializing two nearast neighbors *)
			i := 1;
			while i <= length do begin
				min1 := distMatr^^[i, i mod length + 1];
				min2 := distMatr^^[i, (i mod length + 1) mod length + 1];
				j := ((i mod length + 1) mod length + 1) mod length + 1;
				while j <> i do begin
					if distMatr^^[i, j] < min1 then begin
						min2 := min1;
						min1 := distMatr^^[i, j];
					end
					else if distMatr^^[i, j] < min2 then
						min2 := distMatr^^[i, j];
					j := j mod length + 1;
				end;
				twoNN^^[i] := (min1 + min2) / 2.0;
				i := i + 1;
			end;
			curTour.bestRest := 0.0;
			for i := 2 to length do
				curTour.bestRest := curTour.bestRest + twoNN^^[i];
			stack^^.ptr := 1;
			stack^^.entries[1] := curTour;
(* search the tree! *)
			findMinTour(stack^^, shTour, length, distMatr^^, twoNN^^);
			if not cancel then begin
(* construct the polygon *)
				currentVectorLength := length;
				new(pt);
				failNil(Handle(pt));
				pt.init;
				for i := 2 to length do
					pt.elements^^[i] := shPoly.elements^^[shTour.points[i]];
				for i := 2 to length do
					shPoly.elements^^[i] := pt.elements^^[i];
				TSP2NNBranchAndBound := shPoly;
				pt.free;
			end
			else begin
				TSP2NNBranchAndBound := nil;
				shPoly.freeAll;
			end;
			dispose(stack^);
			dispose(stack);
			dispose(distMatr^);
			dispose(distMatr);
			dispose(twoNN^);
			dispose(twoNN);
		end;
	end;
{$ENDC }
end.
