(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1992 by Bert Leukens						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.0:  Aug-02-92								*)

unit triangulatePlaneSweep3d;

interface

uses
{    Profiler, }
	triangulatePlaneSweep, triangulation3d;

implementation

type

	virtualTriangulation3d = object(triangulation)
			origP, origE, origT: vector;
			procedure init;
			override;
			procedure free;
			override;
			function addNewEdge (startVertexNew, endVertexNew: seqIndex): graphEdgeTriangulation;
			override;
			procedure addNewTriangle (edgeANew, edgeBNew, edgeCNew: graphEdgeTriangulation);
			override;
		end;

	triangulationPolygonInfoDirectory3d = object(triangulationPolygonInfoDirectory)
			firstInLayer: seqIndexArrayH;	(* dynamic array of indices to first point in every layer *)
			curLayer: seqIndex0;				(* index to current layer *)
			maxLayers: seqIndex0;				(* max. number of layers *)
			currentZ: real;
			minY, maxY: real;

			procedure init;
			override;
			procedure free;
			override;
			procedure initialize (tri: triangulation; vertices, outerPolygons, innerPolygons: seqIndex0);
			override;
			procedure addNewPoint (p: point2d);
			override;
			procedure appendPolygons (v: vector);
			override;
			procedure initLayers (layers: seqIndex);
	(* defines how many layers are going to be triangulated *)
		end;

	triangulationSweepLine3d = object(triangulationSweepLine)
			pipe1, pipe2: seqIndex0;				(* index to first events in event pipe one and two *)
			pipe1End, pipe2End: seqIndex0;		(* last valid index in event pipe to be processed *)

			procedure nextEvent;
			override;
			function colinearCheck (p, q: seqIndex): boolean;
			override;
{$IFC AnimationEnabled AND ATriangulation }
			procedure animateNextEvent;
			override;
			procedure animateNewEdge;
			override;
			procedure animateVirtualRectangle;
{$ENDC }
			procedure initPipes (firstLayer, secondLayer: seqIndex0);
		(* defines pipes for two layers *)
			procedure startVirtualSweep (vt: virtualTriangulation3d);
		(* same sweep algorithm, but virtual edges and triangles will be appended *)
		end;


procedure virtualTriangulation3d.init;
	begin
		inherited init;
		origP := points;
		origE := edges;
		origT := triangles;
		points := currentTriangulation.points;
		edges := currentTriangulation.edges;
		triangles := currentTriangulation.triangles;
	end;

procedure virtualTriangulation3d.free;
	begin
		points := origP;
		edges := origE;
		triangles := origT;
		inherited free
	end;

function virtualTriangulation3d.addNewEdge (startVertexNew, endVertexNew: seqIndex): graphEdgeTriangulation;
	var
		e: virtualGraphEdgeTriangulation;
	begin
		new(e);
		failNil(Handle(e));
		e.init;
		edges.append(e);
		addNewEdge := inherited addNewEdge(startVertexNew, endVertexNew);
	end;

procedure virtualTriangulation3d.addNewTriangle (edgeANew, edgeBNew, edgeCNew: graphEdgeTriangulation);
	var
		t: virtualTriangle3d;
	begin
		new(t);
		failNil(Handle(t));
		t.init;
		triangles.append(t);
		inherited addNewTriangle(edgeANew, edgeBNew, edgeCNew)
	end;

function virtualPoint3d.getType: objectType;
	begin
		getType := virtualPoint3dType;
	end;

function virtualPoint3d.getTextType: Str255;
	begin
		getTextType := stringOf('Virtual Point 3d');
	end;

function virtualGraphEdgeTriangulation.getType: objectType;
	begin
		getType := virtualGraphEdgeTriangulationType;
	end;

function virtualGraphEdgeTriangulation.getTextType: Str255;
	begin
		getTextType := stringOf('Virtual GraphEdgeTriangulation');
	end;

function virtualTriangle3d.getType: objectType;
	begin
		getType := virtualTriangle3dType;
	end;

function virtualTriangle3d.getTextType: Str255;
	begin
		getTextType := stringOf('Virtual Triangle 3d');
	end;

procedure virtualTriangle3d.drawTriangle;
	begin
{    inherited drawTriangle;}
	(* don't draw anything *)
	end;

procedure virtualTriangle3d.calcColor;
	begin
{    inherited calcColor;}
	(* don't do anything *)
	end;


procedure triangulationPolygonInfoDirectory3d.init;
	begin
		firstInLayer := nil;
		curLayer := 1;
		maxLayers := 0;
		minY := 1e38;
		maxY := 1e-44;
		inherited init
	end;

procedure triangulationPolygonInfoDirectory3d.free;
	var
		i: seqIndex;
	begin
		DisposeHandle(Handle(firstInLayer));
		for i := 1 to 4 do
			pInfo^^[i].p.free;
		inherited free;
	end;

procedure triangulationPolygonInfoDirectory3d.initLayers (layers: seqIndex);
	begin
		maxLayers := layers;
		firstInLayer := seqIndexArrayH(dynamicArray(maxLayers + 1, sizeOf(seqIndex)));
		failNil(Handle(firstInLayer));
		firstInLayer^^[maxLayers + 1] := firstInPoly^^[maxPolygons + 1];		(* define first layer sentinel *)
	end;

procedure triangulationPolygonInfoDirectory3d.initialize (tri: triangulation; (* *)
								vertices, outerPolygons, innerPolygons: seqIndex0);
	begin
		maxVertices := vertices;
		maxPolygons := innerPolygons + outerPolygons;		(* total amount of polygons *)
		maxEdges := 3 * maxVertices - 7;
		maxTriangles := 2 * maxVertices - 6;

		t := tri;
		init;
	end;

procedure triangulationPolygonInfoDirectory3d.addNewPoint (p: point2d);
	var
		p3d: realPoint3d;
	begin
		new(p3d);
		failNil(Handle(p3d));
		with p3d do begin
			x := p.getX;
			y := p.getY;
			z := currentZ
		end;
		t.points.append(p3d);								(* generate new point of triangulation *)
	end;

procedure triangulationPolygonInfoDirectory3d.appendPolygons (v: vector);
	begin
		if curLayer > maxLayers then
			fatalError('Layers overflow');
		firstInLayer^^[curLayer] := t.points.length + 1;
		inherited appendPolygons(v);
		sort(firstInLayer^^[curLayer], t.points.length);		(* sort all points in layer *)
		curLayer := curLayer + 1;
	end;


procedure triangulationSweepLine3d.nextEvent;
	var
		p, q: point2d;
		e1, e2: seqIndex;
		whichPipe: integer;

	procedure handleSpecialCase;
		var
			sector: triangulationSector;
			where, nextLeft, nextRight: reference;
			direction, otherdir: directionType;
		begin
			with triangulationPolygonInfoDirectory3d(polygonInfo), pInfo^^[e1], sectors do begin
				if typ < pInfo^^[e2].typ then
					whichPipe := 1
				else if pInfo^^[e2].typ < typ then
					whichPipe := 2
				else begin
					currentEvent := pipe1;
					currentPoint := e1;
					case typ of
						closeSectorC:  begin
							if findSector(true, where, direction) then
								sector := triangulationSector(sectors.getObject(where));
							nextLeft := next(where, left);
							nextRight := next(where, right);
							if (sector.nextVertex[left] = currentPoint) & (nextRight <> nil) & (triangulationSector(getObject(nextRight)).nextVertex[right] = currentPoint) then
								whichPipe := 1
							else if (sector.nextVertex[right] = currentPoint) & (nextLeft <> nil) & (triangulationSector(getObject(nextLeft)).nextVertex[left] = currentPoint) then
								whichPipe := 1
							else
								whichPipe := 2;
						end;
						continueSectorC: 
							whichPipe := 1;
						openSectorC: 
							with t.points do
								if sectors.numberOfElements = 0 then
									if point3d(elements^^[e1]).getZ < point3d(elements^^[e2]).getZ then
										whichPipe := 1
									else
										whichPipe := 2
								else if findSector(false, where, direction) then begin
									sector := triangulationSector(sectors.getObject(where));
									if point3d(elements^^[e1]).getZ = point3d(elements^^[sector.s]).getZ then
										whichPipe := 1
									else
										whichPipe := 2
								end
								else begin
									if direction = right then
										otherdir := left
									else
										otherdir := right;
									sector := triangulationSector(sectors.getObject(where));
									if point3d(elements^^[e1]).getZ = point3d(elements^^[sector.nextVertex[otherdir]]).getZ then
										whichPipe := 1
									else
										whichPipe := 2
								end
					end;
				end;
			end;
		end;

	begin
		with triangulationPolygonInfoDirectory3d(polygonInfo) do begin
			if pipe1 > 0 then begin
				if pipe2 > 0 then begin
					e1 := events^^[pipe1];
					e2 := events^^[pipe2];
					p := pInfo^^[e1].p;
					q := pInfo^^[e2].p;
					case comparePointXY(p, q) of
						lessThanObj: 
							whichPipe := 1;
						greaterThanObj: 
							whichPipe := 2;
						equalObj: 
							handleSpecialCase
					end
				end
				else
					whichPipe := 1;
			end
			else if pipe2 > 0 then
				whichPipe := 2
			else
				whichPipe := 0;
			case whichPipe of
				0: 
					currentEvent := 0;
				1:  begin
					currentEvent := pipe1;
					if pipe1 = pipe1End then
						pipe1 := 0
					else
						pipe1 := pipe1 + 1;
				end;
				2:  begin
					currentEvent := pipe2;
					if pipe2 = pipe2End then
						pipe2 := 0
					else
						pipe2 := pipe2 + 1;
				end;
			end;
			if currentEvent > 0 then begin
				currentPoint := events^^[currentEvent];
				with point3d(t.points.elements^^[currentPoint]) do
					if getY > maxY then
						maxY := getY
					else if getY < minY then
						minY := getY;
{$IFC AnimationEnabled AND ATriangulation }
				animateNextEvent;
				if userAbort then
					currentEvent := 0;
{$ENDC }
			end;
		end;
	end;

function triangulationSweepLine3d.colinearCheck (p, q: seqIndex): boolean;
	begin
		colinearCheck := point3d(t.points.elements^^[p]).getZ <> point3d(t.points.elements^^[q]).getZ
	end;


{$IFC AnimationEnabled AND ATriangulation }
procedure triangulationSweepLine3d.animateNextEvent;
	begin
		if animationFlag[ATriangulationItem] then begin
			t.points.elements^^[currentPoint].interactiveFlash;
			waitForClick(animationFlag[ATriangulationItem]);
		end;
	end;

procedure triangulationSweepLine3d.animateNewEdge;
	var
		x1, y1, z1, x2, y2, z2: extended;
		e: graphEdgeTriangulation;
		v: homogenVector;
	begin
		if animationFlag[ATriangulationItem] then
			with t.points do begin
				v := newHomogenVector(2);
				e := graphEdgeTriangulation(t.edges.elements^^[t.edges.length]);
				v.append(elements^^[e.startVertex]);
				v.append(elements^^[e.endVertex]);
				v.interactiveFlash;
				v.free;
				with point3d(elements^^[e.startVertex]) do begin
					x1 := getX;
					y1 := getY;
					z1 := getZ;
				end;
				with point3d(elements^^[e.endVertex]) do begin
					x2 := getX;
					y2 := getY;
					z2 := getZ;
				end;
				drawLineSegment3d(x1, y1, z1, x2, y2, z2);
				waitForClick(animationFlag[ATriangulationItem]);
			end;
	end;

procedure triangulationSweepLine3d.animateVirtualRectangle;
	var
		oldLength: seqIndex0;
		i: integer;
	begin
		with t.edges do begin
			oldLength := length;
			for i := 1 to 4 do begin
				length := i;
				animateNewEdge
			end;
			length := oldLength
		end
	end;
{$ENDC }

procedure triangulationSweepLine3d.initPipes (firstLayer, secondLayer: seqIndex0);
	begin
		pipe1 := triangulationPolygonInfoDirectory3d(polygonInfo).firstInLayer^^[firstLayer];
		pipe1End := triangulationPolygonInfoDirectory3d(polygonInfo).firstInLayer^^[firstLayer + 1] - 1;
		if secondLayer > 0 then begin
			pipe2 := triangulationPolygonInfoDirectory3d(polygonInfo).firstInLayer^^[secondLayer];
			pipe2End := triangulationPolygonInfoDirectory3d(polygonInfo).firstInLayer^^[secondLayer + 1] - 1;
		end
		else begin
			pipe2 := 0;
			pipe2End := 0;
		end;
	end;

procedure triangulationSweepLine3d.startVirtualSweep (vt: virtualTriangulation3d);
	var
		oldT: triangulation;
	begin
		oldT := t;
		t := vt;
{$IFC AnimationEnabled AND ATriangulation }
		animateVirtualRectangle;
{$ENDC }
		startSweep;
		t := oldT
	end;



procedure triangulation3d.initializeFromLayerVectorPlaneSweep (v: polyLayerVector);
	var
		i, j: seqIndex;
		vt: virtualTriangulation3d;		(* virtual triangulation with same data as instance variables of  self *)
		polyInfo: triangulationPolygonInfoDirectory3d;
		sweepLine: triangulationSweepLine3d;
		sumOfLayers, sumOfPolygons, sumOfVertices: seqIndex;

	procedure appendVirtualPolygon;
		var
			e: graphEdgeTriangulation;
		procedure newPoint (pT: pointType);
			var
				vp: virtualPoint3d;
				p2d: realPoint;
			begin
				new(vp);
				failNil(Handle(vp));
				vp.init;
				new(p2d);
				failNil(Handle(p2d));
				p2d.init;
				with polyInfo do begin
					t.points.append(vp);
					with pInfo^^[t.points.length] do begin
						p := p2d;
						typ := pT;
						poly := 1;
					end;
				end
			end;

		begin
			with polyInfo do begin
				firstInLayer^^[1] := 1;
				firstInPoly^^[1] := 1;
				curPolygon := 2;
				curLayer := 2;
			end;
			newPoint(openSector);
			newPoint(continueSector);
			newPoint(continueSector);
			newPoint(closeSector);
			with vt do begin
				e := addNewEdge(1, 2);
				e := addNewEdge(2, 3);
				e := addNewEdge(3, 4);
				e := addNewEdge(4, 1);
			end;
		end;

	procedure setVirtualPoints;
		const
			diff = 10;
		var
			minX, maxX, newZ: real;
		procedure setPoint (id: seqIndex; newX, newY: real);
			begin
				with polyInfo.pInfo^^[id].p do begin
					setX(newX);
					setY(newY);
				end;
				with point3d(points.elements^^[id]) do begin
					setX(newX);
					setY(newY);
					setZ(newZ);
				end;
			end;

		begin
			with points, polyInfo do begin
				newZ := point3d(elements^^[5]).getZ - 25;
				minX := point3d(elements^^[events^^[firstInLayer^^[2]]]).getX;
				maxX := point3d(elements^^[events^^[firstInLayer^^[3] - 1]]).getX;
				setPoint(1, minX - diff, minY - diff);
				setPoint(2, minX - diff, maxY + diff);
				setPoint(3, maxX + diff, maxY + diff);
				setPoint(4, maxX + diff, minY - diff);
			end
		end;

	begin

{    %_PTrace := true;}
{    ResetProfile;}

		currentTriangulation := self;
		new(polyInfo);
		failNil(Handle(polyInfo));
		new(sweepLine);
		failNil(Handle(sweepLine));
		sweepLine.init;

		sumOfLayers := 1;				(* virtual *)
		sumOfPolygons := 1;			(* virtual *)
		sumOfVertices := 4;				(* virtual *)

		for i := 1 to v.layers.length do
			with layer(v.layers.elements^^[i]).layElems do begin
				sumOfLayers := sumOfLayers + 1;
				for j := 1 to length do begin
					sumOfPolygons := sumOfPolygons + 1;
					sumOfVertices := sumOfVertices + polygon2d(elements^^[j]).length;
				end;
			end;

		polyInfo.initialize(self, sumOfVertices, 1, sumOfPolygons - 1);
		polyInfo.initLayers(sumOfLayers);
		sweepLine.initialize(polyInfo, self);

		new(vt);
		failNil(Handle(vt));
		vt.init;

		appendVirtualPolygon;

		for i := 1 to v.layers.length do
			with layer(v.layers.elements^^[i]) do begin
				polyInfo.currentZ := z;
				polyInfo.appendPolygons(layElems);
			end;


		with sweepLine do begin
			for i := 2 to sumOfLayers - 1 do begin
				initPipes(i, i + 1);
				startSweep
			end;
			initPipes(sumOfLayers, 0);
			startSweep;

			setVirtualPoints;
			initPipes(1, 2);
			startVirtualSweep(vt);

			free;					(* release data structures *)
		end;
		polyInfo.free;
		vt.free;

{    DumpProfileToFile('Profile3D');}
{    %_PTrace := false;}

	end;

end.
