(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1992 by 	Michael Greminger				*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.0:  31-Jan-92								*)
(*														*)
(* Copyright 1992 by Bert Leukens						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.1:  May-13-92								*)
(* Version 1.2:  July-29-92								*)

unit triangulation;

interface

uses
	pointVector;

type

	graphEdgeTriangulation = object(graphEdge)
			triangleA, triangleB: seqIndex0;
		  (* indices of triangles adjacent to edge, *)
		  (* 0 if triangle does not exist. *)
			id: seqIndex0;
		  (* position of this edge in 'triangulation.edges' *)
	(* Memory management *)
			procedure init;
			override;
	(* Binary and text file I/O *)
			procedure fileOutput;
			override;
			procedure fileInput;
			override;
			procedure textFileOutput;
			override;
			procedure textFileInput;
			override;
	(* Type computation and comparison *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function getInfo: obj;
			override;
			function memberType (o: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			procedure interactiveOutput;
			override; (* 'currentTriangulation' must be set! *)
			function isSelected (where: obj): boolean;
			override; (* 'currentTriangulation' must be set! *)
	(* Miscellaneous *)
			procedure appendTriangle (triangleID: seqIndex);
	      (* appends a triangle to the edge *)
			procedure changeTriangle (oldTriangle, newTriangle: seqIndex);
	(*	Changes either 'triangleA' or 'triangleB' from 'oldTriangle' to 'newTriangle'. 'oldTriangle'			*)
	(*	must exist.																										*)

			procedure changeVertex (oldVertex, newVertex: seqIndex);
	(*	Changes either 'startVertex' or 'endVertex' from 'oldVertex' to 'newVertex'. 'oldVertex' need	*)
	(*	not exist.																											*)

			procedure getOtherPointsId (var a, b: seqIndex0);
	(*	Computes the other points of triangle A and triangle B, if they exist.									*)

		end;

	triangle = object(obj)
			edgeA, edgeB, edgeC: seqIndex;
		  (* indices of edges adjacent to triangle. *)
	(* Binary and text file I/O *)
			procedure fileOutput;
			override;
			procedure fileInput;
			override;
			procedure textFileOutput;
			override;
			procedure textFileInput;
			override;
	(* Type computation and comparison *)
			function getInfo: obj;
			override;
			function memberType (o: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;

	(* Miscellaneous *)
			procedure getEdges (var refEdgeA, refEdgeB, refEdgeC: graphEdgeTriangulation);
	      (* returns the references to the three edges A, B, C of the triangle *)
		  (* PRECONDITION: the global variable currentTriangulation must be set. *)

			procedure getPointIDs (var pointA, pointB, pointC: seqIndex);
	      (* returns the three points A, B, C of the triangle *)
		  (* PRECONDITION: the global variable currentTriangulation must be set. *)
	      (* POSTCONDITION: point A is the point opposite of edgeA *)
	      (*                           point B is the point opposite of edgeB *)
	      (*                           point C is the point opposite of edgeC *)

			procedure getThirdPoint (refEdge: graphEdgeTriangulation; var refPoint: obj);
	      (* returns the reference of the point opposite of the referenced edge *)
		  (* PRECONDITION: the global variable currentTriangulation must be set. refEdge is edge of triangle *)
	      (* POSTCONDITION: point opposite of edge *)

			function getThirdPointId (refEdge: graphEdgeTriangulation): seqIndex;
	      (* returns the index of the point opposite of the referenced edge *)
		  (* PRECONDITION: the global variable currentTriangulation must be set. refEdge is edge of triangle *)
	      (* POSTCONDITION: index of point opposite of edge *)

			procedure getPoints (var refPointA, refPointB, refPointC: obj);
	      (* returns the references to the three points A, B, C of the triangle *)
		  (* PRECONDITION: the global variable currentTriangulation must be set. *)
	      (* POSTCONDITION: point A is the point opposite of edgeA *)
	      (*                           point B is the point opposite of edgeB *)
	      (*                           point C is the point opposite of edgeC *)

			procedure changeEdge (oldEdge, newEdge: seqIndex);
	(*	Changes the edge-index 'oldEdge' to 'newEdge'. 'oldEdge' must exist among 'edgeA', 'edgeB' and 'edgeC'. *)


		end;

	triangle2d = object(triangle)
	(* Type computation and comparison *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
	(* Miscellaneous *)
			procedure centerOfGravity (var cx, cy: extended);
 	     (* PRECONDITION: the global variable currentTriangulation must be set. *)
			override;
			function isSelected (where: obj): boolean;
			override;
		end;


	triangulation = object(obj)
			points, edges, triangles: vector;
	(* Memory management *)
			procedure init;
			override;
			procedure free;
			override;
			procedure freeAll;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
	(* Binary and text file I/O *)
			procedure fileOutput;
			override;
			procedure fileInput;
			override;
			procedure textFileOutput;
			override;
			procedure textFileInput;
			override;
	(* Geometric transformations *)
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
	(* Type computation and comparison *)
			function getInfo: obj;
			override;
			function equal (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;
	(* Miscellaneous *)
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			procedure initialize (nPoints, nEdges, nTriangles: longInt);
		  (* Initialize new vectors *)

			procedure initializeEuler (vertices, holes: longInt);
		  (* Initialize new vectors *)
		  (* PRECONDITION: #bounding edges (incl. those of holes) = #vertices *)

			procedure initializeFromGraph (udg: simpleUndirectedGraph; function pointTransformation (p: point2d): obj);
		  (* Initialize triangulation from a simple unidirected graph. The function pointTransformation should *)
		  (* generate and initialize a point2d or point3d from the given point2d. *)

			function addNewEdge (startVertexNew, endVertexNew: seqIndex): graphEdgeTriangulation;
		  (* Generates and initializes a new edge, appends it to the triangulaton and returns its reference. *)
		  (* must be overriden. *)

			procedure addNewTriangle (edgeANew, edgeBNew, edgeCNew: graphEdgeTriangulation);
		  (* Generates and initializes a new triangle, appends it to the triangulation and updates the *)
		  (* adjacent edges with its index.  Must be overriden. *)

			procedure flip (x: graphEdgeTriangulation);
	(*	Flips the edge 'x' of the graph. 'x' must belong to two triangles.												*)

			procedure randomFlip (n: longInt);
	(*	Calls 'n' times flip on a random edge.																				*)

			function optimize (function flipOk (x: graphEdgeTriangulation): boolean; container: edgeContainer): longInt;
	(*	Performs the flip operations as long as there are edges for which 'flipOk' yields true.						*)

			function optimize2 (function flipOk (x: graphEdgeTriangulation): boolean; container: edgeContainerX): longInt;

			procedure updateEdge (x: graphEdgeTriangulation; triangleId: seqIndex);
	(*	Removes from 'x' the triangle 'triangleId' (which must exist) and removes 'x' if 'triangleId' was the	*)
	(*	last triangle.																												*)

			procedure deleteEdge (x: graphEdgeTriangulation);
	(*	Removes 'x' from the triangulation.	No references to 'x' are allowed to exist.								*)

			procedure deleteTriangle (triangleId: seqIndex);
	(*	Deletes the triangle 'triangleId' and all associated edges from the triangulation. 								*)

			procedure deleteVertex (vertexId: seqIndex);
	(*	Deletes the point 'vertexId' and all associated triangles and edges from the triangulation.					*)

			function toUndirectedGraph: undirectedGraph;
	(*	Transforms the triangulation to an undirected graph.																*)

		end;

	triangulation2d = object(triangulation)
	(* Interactive I/O *)
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
	(* Description / execution *)
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
	(* Type computation and comparison *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
	(* Selecting and dragging *)
			function isSelected (where: obj): boolean;
			override;
	(* Miscellaneous *)
			function pointCover (n: longInt): obj;
			override;
			function addNewEdge (startVertexNew, endVertexNew: seqIndex): graphEdgeTriangulation;
			override;
			procedure addNewTriangle (edgeANew, edgeBNew, edgeCNew: graphEdgeTriangulation);
			override;
			function getBoundingObject: obj;
			override;
			procedure randomFlip (n: longInt);
			override;
			function area: extended;
			override;

			function flipOk (x: graphEdgeTriangulation): boolean;
	(*	Determines if a flip is legal. *)

			function showNumbering: homogenVector;

			procedure initializeFromPolygon2d (polygon: polygon2d);
	  (* Initialize triangulation from a 'polygon2d'. *)

			function generateUniformRandomPoints (n: seqIndex0): homogenVector;
	(*	Generates 'n' random points, uniformly distributed inside the triangles. *)

			function createLineSegments: lineSegmentVector;
	(*	Creates a vector of line segments corresponding to the edges of the graph.	*)

			function findMinimalAngleTriangle: pointVector;
	(* Finds the triangle with smallest angle. The result is a vector of three points forming the angle. *)

		end;

	edgeContainer = object(obj)
			stackH: seqIndexArrayH;
			onStackH: booleanArrayH;
			function applicableFlips (function flipOk (x: graphEdgeTriangulation): boolean): seqIndex0;
			procedure initContainer (length: seqIndex);
			procedure addEdge (e: seqIndex);
			procedure addTriangle (triangleIndex, exceptionEdge: seqIndex);
			function getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
			function isNotEmpty: boolean;
			procedure exitContainer;
		end;

	stackC = object(edgeContainer)
			stackPointer: seqIndex0;
			function applicableFlips (function flipOk (x: graphEdgeTriangulation): boolean): seqIndex0;
			override;
			procedure initContainer (length: seqIndex);
			override;
			procedure addEdge (e: seqIndex);
			override;
			function getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
			override;
			function isNotEmpty: boolean;
			override;
		end;

	randomC = object(stackC)
			function getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
			override;
		end;

	degreeC = object(stackC)
			function getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
			override;
		end;

	minAngleC = object(stackC)
			function getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
			override;
			function getDelta (a, b, c, d: point2d): extended;
	(*	Measures the quality of a flip. Before flip: triangles abc and abd. After flip: triangles acd and bcd. *)
		end;

	minAngle2C = object(minAngleC)
			function getDelta (a, b, c, d: point2d): extended;
	(*	Measures the quality of a flip. Before flip: triangles abc and abd. After flip: triangles acd and bcd. *)
			override;
		end;

	queueC = object(edgeContainer)
			top, bottom, l: seqIndex;
			count: seqIndex0;
			function applicableFlips (function flipOk (x: graphEdgeTriangulation): boolean): seqIndex0;
			override;
			procedure initContainer (length: seqIndex);
			override;
			procedure addEdge (e: seqIndex);
			override;
			function getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
			override;
			function isNotEmpty: boolean;
			override;
		end;

const
	stackSel = 0;
	queueSel = 1;
	deltaSel = 2;
	angleSel = 3;
	degreeSel = 4;

function getMethodSelection: boolean;

type
edgeContainerX = object(obj)
		length: seqIndex0;
		function applicableFlips: seqIndex0;
		procedure initContainer (len: seqIndex);
		procedure updateContainerEdge (e: seqIndex; function flipOk (x: graphEdgeTriangulation): boolean);
		procedure update (x: graphEdgeTriangulation; function flipOk (x: graphEdgeTriangulation): boolean);
		procedure updateTriangle (triangleIndex, exceptionEdge: seqIndex; function flipOk (x: graphEdgeTriangulation): boolean);
		function getEdge: graphEdgeTriangulation;
		function isNotEmpty: boolean;
		procedure exitContainer;
	end;

stackCX = object(edgeContainerX)
		stackH: seqIndexArrayH;
		stackLocationH: seqIndexArray0H;
		procedure initContainer (len: seqIndex);
		override;
		procedure updateContainerEdge (e: seqIndex; function flipOk (x: graphEdgeTriangulation): boolean);
		override;
		function getEdge: graphEdgeTriangulation;
		override;
		procedure exitContainer;
		override;
		procedure exchangeWithTop (newPos: seqIndex);
		procedure updateEdgeGraphics (x: graphEdgeTriangulation);
		procedure updateGraphicsStatus (e: seqIndex; x: graphEdgeTriangulation);
		procedure deselectEdge (x: graphEdgeTriangulation);
	end;

randomCX = object(stackCX)
		function getEdge: graphEdgeTriangulation;
		override;
	end;

interactiveCX = object(stackCX)
		animationStatus: boolean;
		procedure initContainer (len: seqIndex);
		override;
		function getEdge: graphEdgeTriangulation;
		override;
		procedure update (x: graphEdgeTriangulation; function flipOk (x: graphEdgeTriangulation): boolean);
		override;
		procedure updateEdgeGraphics (x: graphEdgeTriangulation);
		override;
		procedure exitContainer;
		override;
	end;

interactiveFlipCX = object(interactiveCX)
		delaunayEdgeH: booleanArrayH;
		procedure initContainer (len: seqIndex);
		override;
		procedure exitContainer;
		override;
		procedure updateTriangle (triangleIndex, exceptionEdge: seqIndex; function flipOk (x: graphEdgeTriangulation): boolean);
		override;
		procedure updateGraphicsStatus (e: seqIndex; x: graphEdgeTriangulation);
		override;
		procedure deselectEdge (x: graphEdgeTriangulation);
		override;
	end;

procedure initTriangulation;
		  (* Generates global prototypes of graphEdgeTriangulation, triangle2d and triangulation2d. *)

procedure exitTriangulation;

var
deleteWhichVertex, deleteWhichTriangle, minimalDifference: extended;
sel1, sel2: integer;
currentTriangulation: triangulation;
		  (* Global variable to give access to the vectors of points, edges and triangles *)


implementation

procedure graphEdgeTriangulation.changeTriangle (oldTriangle, newTriangle: seqIndex);
	begin
		assert((triangleA = oldTriangle) | (triangleB = oldTriangle), 'oldTriangle exists');
		if triangleA = oldTriangle then
			triangleA := newTriangle
		else
			triangleB := newTriangle;
	end;

procedure graphEdgeTriangulation.changeVertex (oldVertex, newVertex: seqIndex);
	begin
		if startVertex = oldVertex then
			startVertex := newVertex
		else if endVertex = oldVertex then
			endVertex := newVertex;
	end;

procedure graphEdgeTriangulation.getOtherPointsId (var a, b: seqIndex0);
	begin
		with currentTriangulation.triangles do begin
			a := triangle(elements^^[triangleA]).getThirdPointId(self);
			if triangleB > 0 then
				b := triangle(elements^^[triangleB]).getThirdPointId(self)
			else
				b := 0;
		end;
	end;

procedure triangle.changeEdge (oldEdge, newEdge: seqIndex);
	begin
		assert((edgeA = oldEdge) | (edgeB = oldEdge) | (edgeC = oldEdge), 'oldEdge exists');
		if edgeA = oldEdge then
			edgeA := newEdge
		else if edgeB = oldEdge then
			edgeB := newEdge
		else
			edgeC := newEdge
	end;

procedure triangulation.updateEdge (x: graphEdgeTriangulation; triangleId: seqIndex);
	begin
		assert((x.triangleA = triangleId) | (x.triangleB = triangleId), 'triangleId exists');
		with x do
			if triangleB > 0 then begin
				if triangleA = triangleId then
					triangleA := triangleB;
				triangleB := 0;
			end
			else
				deleteEdge(x);
	end;

procedure triangulation.deleteEdge (x: graphEdgeTriangulation);
	var
		lastEdge: graphEdgeTriangulation;
	begin
		if x.id < edges.length then begin
			lastEdge := graphEdgeTriangulation(edges.elements^^[edges.length]);
			with lastEdge do begin
				triangle(triangles.elements^^[triangleA]).changeEdge(id, x.id);
				if triangleB > 0 then
					triangle(triangles.elements^^[triangleB]).changeEdge(id, x.id);
			end;
			edges.elements^^[x.id] := lastEdge;
			lastEdge.id := x.id;
		end;
		edges.length := edges.length - 1;
		x.free;
	end;

procedure triangulation.deleteTriangle (triangleId: seqIndex);
	var
		thisTriangle, lastTriangle: triangle;
	begin
		thisTriangle := triangle(triangles.elements^^[triangleId]);
		with thisTriangle, edges do begin
			updateEdge(graphEdgeTriangulation(elements^^[edgeA]), triangleId);
			updateEdge(graphEdgeTriangulation(elements^^[edgeB]), triangleId);
			updateEdge(graphEdgeTriangulation(elements^^[edgeC]), triangleId);
		end;
		if triangleId < triangles.length then begin
			lastTriangle := triangle(triangles.elements^^[triangles.length]);
			with lastTriangle, edges do begin
				graphEdgeTriangulation(elements^^[edgeA]).changeTriangle(triangles.length, triangleId);
				graphEdgeTriangulation(elements^^[edgeB]).changeTriangle(triangles.length, triangleId);
				graphEdgeTriangulation(elements^^[edgeC]).changeTriangle(triangles.length, triangleId);
			end;
			triangles.elements^^[triangleId] := lastTriangle;
		end;
		triangles.length := triangles.length - 1;
		thisTriangle.free;
	end;

procedure triangulation.deleteVertex (vertexId: seqIndex);
	var
		i: seqIndex01;
		pointA, pointB, pointC: seqIndex;
		thisPoint: obj;
	begin
		currentTriangulation := self;
		i := 1;
		while i <= triangles.length do begin
			triangle(triangles.elements^^[i]).getPointIDs(pointA, pointB, pointC);
			if (pointA = vertexId) | (pointB = vertexId) | (pointC = vertexId) then
				deleteTriangle(i)
			else
				i := i + 1;
		end;
		thisPoint := points.elements^^[vertexId];
		if vertexId < points.length then begin
			for i := 1 to edges.length do
				graphEdgeTriangulation(edges.elements^^[i]).changeVertex(points.length, vertexId);
			points.elements^^[vertexId] := points.elements^^[points.length];
		end;
		points.length := points.length - 1;
		thisPoint.free;
	end;

procedure graphEdgeTriangulation.init;
	begin
		triangleA := 0;
		triangleB := 0;
		id := 0;
	end;

procedure graphEdgeTriangulation.fileOutput;
	begin
		inherited fileOutput;
		writeLongInt(triangleA);
		writeLongInt(triangleB);
		writeLongInt(id);
	end;

procedure graphEdgeTriangulation.fileInput;
	var
		l: longint;
	begin
		inherited fileInput;
		readLongInt(l);
		triangleA := l;
		readLongInt(l);
		triangleB := l;
		readLongInt(l);
		id := l;
	end;

procedure graphEdgeTriangulation.textFileOutput;
	begin
		inherited textFileOutput;
		writeTextLongInt(triangleA);
		writeTextLongInt(triangleB);
		writeTextLongInt(id);
	end;

procedure graphEdgeTriangulation.textFileInput;
	var
		l: longint;
	begin
		inherited textFileInput;
		readTextLongInt(l);
		triangleA := l;
		readTextLongInt(l);
		triangleB := l;
		readTextLongInt(l);
		id := l;
	end;

function graphEdgeTriangulation.getType: objectType;
	begin
		getType := graphEdgeTriangulationType;
	end;

function graphEdgeTriangulation.getTextType: Str255;
	begin
		getTextType := 'Graph Edge Triangulation';
	end;

function graphEdgeTriangulation.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoExtended(v, 'Triangle A', triangleA, true);
		appendObjInfoExtended(v, 'Triangle B', triangleB, true);
		appendObjInfoExtended(v, 'Index', id, true);
		getInfo := v;
	end;

function graphEdgeTriangulation.memberType (o: obj): boolean;
	begin
		memberType := member(o, graphEdgeTriangulation);
	end;

function graphEdgeTriangulation.equal (o: obj): boolean;
	var
		e: graphEdgeTriangulation;
	begin
		if (o <> nil) & member(o, graphEdgeTriangulation) then begin
			e := graphEdgeTriangulation(o);
			equal := inherited equal(o) & (triangleA = e.triangleA) & (triangleB = e.triangleB) & (id = e.id)
		end
		else
			equal := false;
	end;

function graphEdgeTriangulation.genericLessThan (o: obj): boolean;
	var
		e: graphEdgeTriangulation;
	begin
		e := graphEdgeTriangulation(o);
		genericLessThan := inherited genericLessThan(o) | inherited equal(o) & ((triangleA < e.triangleA) | (triangleA = e.triangleA) & ((triangleB < e.triangleB) | (triangleB = e.triangleB) & (id < e.id)));
	end;

procedure graphEdgeTriangulation.appendTriangle (triangleID: seqIndex);
	begin
		if triangleA = 0 then
			triangleA := triangleID
		else begin
			assert(triangleB = 0, 'Two triangles per edge');
			triangleB := triangleID;
		end;
	end;


procedure triangle.fileOutput;
	begin
		writeLongInt(edgeA);
		writeLongInt(edgeB);
		writeLongInt(edgeC);
	end;

procedure triangle.fileInput;
	var
		l: longInt;
	begin
		readLongInt(l);
		edgeA := l;
		readLongInt(l);
		edgeB := l;
		readLongInt(l);
		edgeC := l;
	end;

procedure triangle.textFileOutput;
	begin
		writeTextLongInt(edgeA);
		writeTextLongInt(edgeB);
		writeTextLongInt(edgeC);
	end;

procedure triangle.textFileInput;
	var
		l: longInt;
	begin
		readTextLongInt(l);
		edgeA := l;
		readTextLongInt(l);
		edgeB := l;
		readTextLongInt(l);
		edgeC := l;
	end;

function triangle.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(3);
		appendObjInfoExtended(v, 'Edge A', edgeA, true);
		appendObjInfoExtended(v, 'Edge B', edgeB, true);
		appendObjInfoExtended(v, 'Edge C', edgeC, true);
		getInfo := v;
	end;

function triangle.memberType (o: obj): boolean;
	begin
		memberType := member(o, triangle);
	end;

function triangle.equal (o: obj): boolean;
	var
		t: triangle;
	begin
		if (o <> nil) & member(o, triangle) then begin
			t := triangle(o);
			equal := (edgeA = t.edgeA) & (edgeB = t.edgeB) & (edgeC = t.edgeC);
		end
		else
			equal := false;
	end;

function triangle.genericLessThan (o: obj): boolean;
	var
		t: triangle;
	begin
		if (o <> nil) & member(o, triangle) then begin
			t := triangle(o);
			genericLessThan := (edgeA < t.edgeA) | (edgeA = t.edgeA) & ((edgeB < t.edgeB) | (edgeB = t.edgeB) & ((edgeC < t.edgeC)));
		end
		else
			genericLessThan := false;
	end;

procedure triangle.getEdges (var refEdgeA, refEdgeB, refEdgeC: graphEdgeTriangulation);
	var
		elements: objArrayH;
	begin
		elements := currentTriangulation.edges.elements;
		refEdgeA := graphEdgeTriangulation(elements^^[edgeA]);
		refEdgeB := graphEdgeTriangulation(elements^^[edgeB]);
		refEdgeC := graphEdgeTriangulation(elements^^[edgeC]);
	end;

procedure triangle.getPointIDs (var pointA, pointB, pointC: seqIndex);
	var
		refEdgeA, refEdgeB, refEdgeC: graphEdgeTriangulation;
		As, Ae, Bs, Be, Cs, Ce: seqIndex;
	begin
		getEdges(refEdgeA, refEdgeB, refEdgeC);
		As := refEdgeA.startVertex;
		Ae := refEdgeA.endVertex;
		Bs := refEdgeB.startVertex;
		Be := refEdgeB.endVertex;
		Cs := refEdgeC.startVertex;
		Ce := refEdgeC.endVertex;
		if (As = Cs) | (As = Ce) then begin
			pointB := As;
			pointC := Ae;
		end
		else begin
			pointB := Ae;
			pointC := As;
		end;
		if (Bs = Cs) | (Bs = Ce) then
			pointA := Bs
		else
			pointA := Be;
	end;

procedure triangle.getThirdPoint (refEdge: graphEdgeTriangulation; var refPoint: obj);
	begin
		refPoint := currentTriangulation.points.elements^^[getThirdPointId(refEdge)];
	end;

function triangle.getThirdPointId (refEdge: graphEdgeTriangulation): seqIndex;
	var
		refEdge2: graphEdgeTriangulation;
	begin
		if edgeA = refEdge.id then
			refEdge2 := graphEdgeTriangulation(currentTriangulation.edges.elements^^[edgeB])
		else
			refEdge2 := graphEdgeTriangulation(currentTriangulation.edges.elements^^[edgeA]);
		if (refEdge2.startVertex = refEdge.startVertex) | (refEdge2.startVertex = refEdge.endVertex) then
			getThirdPointId := refEdge2.endVertex
		else
			getThirdPointId := refEdge2.startVertex
	end;

procedure triangle.getPoints (var refPointA, refPointB, refPointC: obj);
	var
		elements: objArrayH;
		pointA, pointB, pointC: seqIndex;
	begin
		getPointIDs(pointA, pointB, pointC);
		elements := currentTriangulation.points.elements;
		refPointA := elements^^[pointA];
		refPointB := elements^^[pointB];
		refPointC := elements^^[pointC];
	end;

function triangulation.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	var
		i: seqIndex0;
		l: list;

	function pointIsSelected (p: point2d): boolean;
		begin
			pointIsSelected := p.isSelected(where);
		end;

	procedure drawAdjacentEdges (o: obj);
		begin
			l.interactiveOutput;
			if highlight then begin
				l.interactiveHighlight;
				o.interactiveHighlight;
			end;
		end;

	procedure buildAdj (e: graphEdgeTriangulation);
		begin
			if (e.startVertex = i) | (e.endVertex = i) then
				l.append(e);
		end;

	begin (* triangulation.dragVertex *)
		GetKeys(globalKeyMap);
		dragVertex := false;
		if globalKeyMap[optionKeyValue] then begin
			i := points.exists(pointIsSelected);
			if i > 0 then begin
				currentTriangulation := self;
				dragVertex := true;
				l := newEmptyList;
				edges.forAll(buildAdj);
				points.elements^^[i].dragObject(drawAdjacentEdges, drawAdjacentEdges, mouse);
				l.free;
			end;
		end;
	end;


procedure triangulation.init;
	begin
		initializeEuler(3, 0);
	end;

procedure triangulation.free;
	begin
		points.free;
		edges.free;
		triangles.free;
		shallowFree;
	end;

procedure triangulation.freeAll;
	begin
		points.freeAll;
		edges.freeAll;
		triangles.freeAll;
		shallowFree;
	end;

function triangulation.duplicate: obj;
	var
		t: triangulation;
	begin
		t := triangulation(shallowDuplicate);
		t.points := vector(points.duplicate);
		t.edges := vector(edges.duplicateAll);
		t.triangles := vector(triangles.duplicateAll);
		duplicate := t;
	end;

function triangulation.duplicateAll: obj;
	var
		t: triangulation;
		i: seqIndex0;
	begin
		t := triangulation(duplicate);
		for i := 1 to points.length do
			points.elements^^[i] := points.elements^^[i].duplicate;
		duplicateAll := t;
	end;

procedure triangulation.fileOutput;
	begin
		points.fileOutputTag;
		edges.fileOutputTag;
		triangles.fileOutputTag;
	end;

procedure triangulation.fileInput;
	begin
		points := vector(newObj(fileGetType));
		points.fileInput;
		edges := vector(newObj(fileGetType));
		edges.fileInput;
		triangles := vector(newObj(fileGetType));
		triangles.fileInput;
	end;

procedure triangulation.textFileOutput;
	begin
		writeTextObj(points);
		writeTextObj(edges);
		writeTextObj(triangles);
	end;

procedure triangulation.textFileInput;
	begin
		points := vector(readTextObj);
		edges := vector(readTextObj);
		triangles := vector(readTextObj);
	end;

procedure triangulation.translate (dx, dy: extended);
	begin
		points.translate(dx, dy);
	end;

procedure triangulation.rotate (xOrigin, yOrigin, alpha: extended);
	begin
		points.rotate(xOrigin, yOrigin, alpha);
	end;

procedure triangulation.scale (lambda: extended);
	begin
		points.scale(lambda);
	end;

procedure triangulation.reflect (a, b, c: extended);
	begin
		points.reflect(a, b, c);
	end;

function triangulation.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(3);
		appendObjInfoObject(v, 'Points', points, false);
		appendObjInfoObject(v, 'Edges', edges, true);
		appendObjInfoObject(v, 'Triangles', triangles, false);
		getInfo := v;
	end;

function triangulation.equal (o: obj): boolean;
	var
		t: triangulation;
	begin
		if (o <> nil) & member(o, triangulation) then begin
			t := triangulation(o);
			equal := points.equal(t.points) & edges.equal(t.edges) & triangles.equal(t.triangles);
		end
		else
			equal := false;
	end;

function triangulation.genericLessThan (o: obj): boolean;
	var
		t: triangulation;
	begin
		if (o <> nil) & member(o, triangulation) then begin
			t := triangulation(o);
			genericLessThan := points.genericLessThan(t.points) | points.equal(t.points) & (edges.genericLessThan(t.edges) | edges.equal(t.edges) & triangles.genericLessThan(t.triangles));
		end
		else
			genericLessThan := false;
	end;

procedure triangulation.centerOfGravity (var cx, cy: extended);
	begin
		points.centerOfGravity(cx, cy);
	end;

procedure triangulation.initialize (nPoints, nEdges, nTriangles: longInt);
	begin
		points := newVector(nPoints);
		edges := newVector(nEdges);
		triangles := newVector(nTriangles);
	end;

procedure triangulation.initializeEuler (vertices, holes: longInt);
	begin
		holes := holes - 1;
		initialize(vertices, 2 * vertices + 3 * holes, vertices + 2 * holes);
	end;

procedure triangulation.initializeFromGraph (udg: simpleUndirectedGraph; function pointTransformation (p: point2d): obj);
	var
		p, q, s, center: point2d;
		i, j, k, centerIndex: seqIndex;
		adjV, edgV: vector;
		edgL: list;
		g, curEdge1, curEdge2, curEdge3: graphEdgeTriangulation;
		e: graphEdge;
		curind: seqIndex;	(* counts the edges, which are inserted *)
		found: boolean;
		cornerB, cornerC, lengthOfVector: seqIndex; 		(* stores the corners from the current triangle *)

	function angleLessThan (m, n: graphEdge): boolean;
		var
			w: extended;
			p, q: point2d;
		begin
			if m.startVertex = centerIndex then
				p := point2d(udg.vertices.elements^^[m.endVertex])
			else
				p := point2d(udg.vertices.elements^^[m.startVertex]);
			if n.startVertex = centerIndex then
				q := point2d(udg.vertices.elements^^[n.endVertex])
			else
				q := point2d(udg.vertices.elements^^[n.startVertex]);
			w := center.whichSideX(p, q);
			if w = 0 then
				angleLessThan := center.distanceX(p) < center.distanceX(q)
			else
				angleLessThan := w < 0;
		end;

	function checkIndex (curind: seqIndex; d: longint; low, high: seqIndex): seqIndex;
		(* avoids under and overflow in indexes, d is the increment *)
		begin
			if d = 1 then
				if curind < high then
					checkIndex := curind + 1
				else
					checkIndex := low
			else if curind > low then
				checkIndex := curind - 1
			else
				checkIndex := high;
		end;

	function samePoint (edge1, edge2: graphEdge): boolean;
		(* checks whether edge1 and edge2 have common points *)
		begin
			samePoint := (edge1.startVertex = edge2.startVertex) | (edge1.startVertex = edge2.endVertex) | (edge1.endVertex = edge2.startVertex) | (edge1.endVertex = edge2.endVertex);
		end;

	begin		(* initializeFromGraph *)
		initialize(udg.vertices.length, udg.edges.length, 1 + udg.edges.length - udg.vertices.length);
		for i := 1 to udg.edges.length do begin
			new(g);
			failNil(Handle(g));
			g.init;
			with graphEdge(udg.edges.elements^^[i]) do
				if startVertex < endVertex then begin
					g.startVertex := startVertex;
					g.endVertex := endVertex;
				end
				else begin
					g.startVertex := endVertex;
					g.endVertex := startVertex;
				end;
			edges.elements^^[i] := g;
			g.id := i;
		end;
		edges.length := udg.edges.length;

		for i := 1 to udg.vertices.length do
			points.append(pointTransformation(point2d(udg.vertices.elements^^[i])));

	(* build adj lists *)
		adjV := newVector(udg.vertices.length);
		for i := 1 to udg.vertices.length do (* builds to each pointnr a vector of edges *)
			adjV.append(newEmptyList);
		for i := 1 to udg.edges.length do begin	(* inserts the edges for each pointnr *)
			e := graphEdge(edges.elements^^[i]);
			sequence(adjV.elements^^[e.startVertex]).append(e);
			sequence(adjV.elements^^[e.endVertex]).append(e);
		end;
		for i := 1 to points.length do begin
			new(edgV);
			failNil(Handle(edgV));
			edgL := list(adjV.elements^^[i]);
			edgV.initFromList(edgL);
			edgL.free;
			adjV.elements^^[i] := edgV;
		end;

	(* sorting the edges for each point in clockwise order *)
		for k := 1 to adjV.length do begin
{$IFC AnimationEnabled }
			advanceStopWatchHands;
{$ENDC }
			center := point2d(udg.vertices.elements^^[k]);
			centerIndex := k;
			i := 1;
			j := vector(adjV.elements^^[k]).length + 1;
			while i < j do begin
				with graphEdge(vector(adjV.elements^^[k]).elements^^[i]) do
					if centerIndex = startVertex then
						p := point2d(udg.vertices.elements^^[endVertex])
					else
						p := point2d(udg.vertices.elements^^[startVertex]);
				if pointLessThanXY(p, center) then
					i := i + 1
				else begin
					j := j - 1;
					vector(adjV.elements^^[k]).swap(i, j);
				end;
			end;
			vector(adjV.elements^^[k]).sort(angleLessThan, 1, i - 1);
			vector(adjV.elements^^[k]).sort(angleLessThan, i, vector(adjV.elements^^[k]).length);
		end;

	(* construction of triangles *)
		curind := 1;
		for i := 1 to adjV.length do begin
{$IFC AnimationEnabled }
			advanceStopWatchHands;
{$ENDC }
			p := point2d(udg.vertices.elements^^[i]);
			for j := 1 to vector(adjV.elements^^[i]).length do begin
				curEdge1 := graphEdgeTriangulation(vector(adjV.elements^^[i]).elements^^[j]);
				if j = vector(adjV.elements^^[i]).length then
					curEdge2 := graphEdgeTriangulation(vector(adjV.elements^^[i]).elements^^[1])
				else
					curEdge2 := graphEdgeTriangulation(vector(adjV.elements^^[i]).elements^^[j + 1]);
				found := (curEdge1.startVertex < i) | (curEdge2.startVertex < i);
				if not found then begin
					if curEdge1.startVertex = i then
						cornerB := curEdge1.endVertex
					else
						cornerB := curEdge1.startVertex;
					if curEdge2.startVertex = i then
						cornerC := curEdge2.endVertex
					else
						cornerC := curEdge2.startVertex;
					q := point2d(udg.vertices.elements^^[cornerB]);
					s := point2d(udg.vertices.elements^^[cornerC]);
					if s.whichSideX(p, q) <= 0 then begin	(* tm. ok else border Triangle, will be later or was already visited *)
						lengthOfVector := vector(adjV.elements^^[cornerB]).length;
						k := 1;
						while (k <= lengthOfVector) & (graphEdge(vector(adjV.elements^^[cornerB]).elements^^[k]) <> curEdge1) do
							k := k + 1;
						if k > lengthOfVector then begin
							nonFatalError('Error in Triangulation');
							exit(initializeFromGraph);
						end
						else begin
							curEdge3 := graphEdgeTriangulation(vector(adjV.elements^^[cornerB]).elements^^[checkIndex(k, -1, 1, lengthOfVector)]);
							if not samePoint(curEdge3, curEdge2) then begin
								curEdge3 := graphEdgeTriangulation(vector(adjV.elements^^[cornerB]).elements^^[checkIndex(k, 1, 1, lengthOfVector)]);
								if not samePoint(curEdge3, curEdge2) then begin
									nonFatalError('Error in the graph. Cannot triangulate!');
									exit(initializeFromGraph);
								end;
							end;
							addNewTriangle(curEdge1, curEdge2, curEdge3);
						end;
					end;
				end;
			end; (* for j *)
		end; (* for i *)
		for i := 1 to adjV.length do begin
			vector(adjV.elements^^[i]).free;
		end;
		vector(adjV).free;
	end; (* initializeFromGraph *)

function triangulation.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := drawingComplexity2(points, edges, cutOff);
	end;

function triangulation.addNewEdge (startVertexNew, endVertexNew: seqIndex): graphEdgeTriangulation;
	var
		e: graphEdgeTriangulation;
	begin
		e := graphEdgeTriangulation(edges.elements^^[edges.length]);
		with e do begin
			startVertex := startVertexNew;
			endVertex := endVertexNew;
			triangleA := 0;
			triangleB := 0;
			id := edges.length;
		end;
		addNewEdge := e;
	end;

procedure triangulation.addNewTriangle (edgeANew, edgeBNew, edgeCNew: graphEdgeTriangulation);
	var
		t: triangle;
		id: seqIndex0;

	begin
		id := triangles.length;
		t := triangle(triangles.elements^^[id]);
		with t do begin
			edgeA := edgeANew.id;
			edgeB := edgeBNew.id;
			edgeC := edgeCNew.id;
		end;
		edgeANew.appendTriangle(id);
		edgeBNew.appendTriangle(id);
		edgeCNew.appendTriangle(id);
	end;

function activateOkButton: boolean;
	begin
		activateOkButton := sel1 <> sel2;
	end;

function getMethodSelection: boolean;
	const
		highestMethod = 4;
	var
		h: vector;
	begin
		h := newVector(4);
		h.append(getEnumerationHandler(5, @sel1, 0, nil, highestMethod));
		h.append(getEnumerationHandler(10, @sel2, 0, nil, highestMethod));
		h.append(getExtendedHandler(15, @minimalDifference, minimalDifference, nil, @isInteger));
		h.append(getExtendedHandler(16, @numberOfObjects, numberOfObjects, nil, @isSeqIndex));
		getMethodSelection := processDialogBox(129, 0, 19, 0, activateOkButton, h);
	end;

	var
		pv: polygon2d;
		rp: realPoint;

function triangulation2d.flipOk (x: graphEdgeTriangulation): boolean;
	var
		hullLength, eliminated: seqIndex0;
{$IFC AnimationEnabled }
		abortEnableSave, animationEnableSave: boolean;
{$ENDC }
	begin
		if x.triangleB > 0 then begin
			triangle(triangles.elements^^[x.triangleA]).getPoints(pv.elements^^[1], pv.elements^^[2], pv.elements^^[3]);
			triangle(triangles.elements^^[x.triangleB]).getThirdPoint(x, pv.elements^^[4]);
			pv.length := 4;
{$IFC AnimationEnabled }
			abortEnableSave := abortEnable;
			abortEnable := false;
			animationEnableSave := animationFlag[AConvexHullItem];
			animationFlag[AConvexHullItem] := false;
{$ENDC }
			pv.convexHull(false, false, hullLength, eliminated);
{$IFC AnimationEnabled }
			abortEnable := abortEnableSave;
			animationFlag[AConvexHullItem] := animationEnableSave;
{$ENDC }
			flipOk := hullLength = 4;
		end
		else
			flipOk := false;
	end;

function triangle2d.isSelected (where: obj): boolean;
	var
		cx, cy: extended;
{$IFC AnimationEnabled }
		abortEnableSave, animationEnableSave: boolean;
{$ENDC }
		onBoundary: boolean;
	begin
		getPoints(pv.elements^^[1], pv.elements^^[2], pv.elements^^[3]);
		pv.length := 3;
		where.centerOfGravity(cx, cy);
		rp.x := cx;
		rp.y := cy;
{$IFC AnimationEnabled }
		abortEnableSave := abortEnable;
		abortEnable := false;
		animationEnableSave := animationFlag[AWindingNumberItem];
		animationFlag[AWindingNumberItem] := false;
{$ENDC }
		isSelected := odd(pv.windingNumber(rp, onBoundary));
{$IFC AnimationEnabled }
		abortEnable := abortEnableSave;
		animationFlag[AWindingNumberItem] := animationEnableSave;
{$ENDC }
	end;


procedure initTriangulation;
	var
		g: graphEdgeTriangulation;
		t: triangle2d;
		t2d: triangulation2d;
	begin
		new(g);
		failNil(Handle(g));
		prototypes[graphEdgeTriangulationType] := g;
		new(t);
		failNil(Handle(t));
		prototypes[triangle2dType] := t;
		new(t2d);
		failNil(Handle(t2d));
		prototypes[triangulation2dType] := t2d;
		deleteWhichVertex := 1;
		deleteWhichTriangle := 1;
		minimalDifference := 1;
		sel1 := stackSel;
		sel2 := queueSel;
		pv := newPolygon2d(4);
		new(rp);
		failNil(Handle(rp));
	end;

procedure exitTriangulation;
	begin
		pv.free;
		rp.free;
	end;



end.
