(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1992 by 	Michael Greminger				*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.0:  31-Jan-92								*)
(*														*)
(* Copyright 1992 by Bert Leukens						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.1:  May-15-92								*)

unit triangulation2d;

interface

uses
{    Profiler, }
	vector, randomizedTriangulation, triangulation2d2, triangulatePlaneSweep, exploreGraph;

implementation

function findEdge (startVertex, endVertex: seqIndex; t: triangle): graphEdgeTriangulation;
	var
		e: graphEdgeTriangulation;
	begin
		with currentTriangulation.edges do begin
			e := graphEdgeTriangulation(elements^^[t.edgeA]);
			if (startVertex = e.startVertex) & (endVertex = e.endVertex) | (startVertex = e.endVertex) & (endVertex = e.startVertex) then
				findEdge := e
			else begin
				e := graphEdgeTriangulation(elements^^[t.edgeB]);
				if (startVertex = e.startVertex) & (endVertex = e.endVertex) | (startVertex = e.endVertex) & (endVertex = e.startVertex) then
					findEdge := e
				else
					findEdge := graphEdgeTriangulation(elements^^[t.edgeC]);
			end;
		end;
	end;

procedure setEdge (t: triangle; oldValue, newValue: seqIndex);
	begin
		with t do
			if edgeA = oldValue then
				edgeA := newValue
			else if edgeB = oldValue then
				edgeB := newValue
			else
				edgeC := newValue;
	end;

procedure triangulation.flip (x: graphEdgeTriangulation);
	var
		pLeft, pRight: seqIndex;
		triangA, triangB: triangle;
		leftEdge, rightEdge: graphEdgeTriangulation;
	begin
		assert(x.triangleB > 0, 'Triangle B exists');
		triangA := triangle(triangles.elements^^[x.triangleA]);
		triangB := triangle(triangles.elements^^[x.triangleB]);
		pLeft := triangA.getThirdPointId(x);
		pRight := triangB.getThirdPointId(x);
		leftEdge := findEdge(x.endVertex, pLeft, triangA);
		rightEdge := findEdge(x.startVertex, pRight, triangB);
		with leftEdge do
			if triangleA = x.triangleA then
				triangleA := x.triangleB
			else
				triangleB := x.triangleB;
		with rightEdge do
			if triangleA = x.triangleB then
				triangleA := x.triangleA
			else
				triangleB := x.triangleA;
		setEdge(triangA, leftEdge.id, rightEdge.id);
		setEdge(triangB, rightEdge.id, leftEdge.id);
		x.startVertex := pLeft;
		x.endVertex := pRight;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure triangulation.randomFlip (n: longInt);
	begin
		unimplementedMethod('triangulation.randomFlip');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure triangulation2d.randomFlip (n: longInt);
	var
		i: longInt;
		x: graphEdgeTriangulation;
{$IFC AnimationEnabled }
		convexHullAnimation: boolean;
{$ENDC }
	begin
		if triangles.length > 1 then begin
			currentTriangulation := self;
{$IFC AnimationEnabled }
			convexHullAnimation := animationFlag[AConvexHullItem];
			animationFlag[AConvexHullItem] := false;
{$ENDC }
{$IFC AnimationEnabled AND ATriangulation }
			if animationFlag[ATriangulationItem] then
				interactiveHighlight;
{$ENDC }
			for i := 1 to n do begin
				repeat
{$IFC AnimationEnabled }
					if userAbort then begin
						animationFlag[AConvexHullItem] := convexHullAnimation;
						exit(randomFlip);
					end;
{$ENDC }
{$IFC THINK_PASCAL }
					x := graphEdgeTriangulation(edges.elements^^[trunc(randomReal * edges.length) + 1]);
{$ELSEC }
					x := graphEdgeTriangulation(edges.elements^^[truncLong(randomReal * edges.length) + 1]);
{$ENDC }
				until flipOk(x);
{$IFC AnimationEnabled AND ATriangulation }
				if animationFlag[ATriangulationItem] then begin
					x.interactiveFlash;
					x.interactiveOutput;
				end;
{$ENDC }
				flip(x);
{$IFC AnimationEnabled AND ATriangulation }
				if animationFlag[ATriangulationItem] then begin
					x.interactiveOutput;
					waitForClick(animationFlag[ATriangulationItem]);
				end;
{$ENDC }
			end;
{$IFC AnimationEnabled }
			animationFlag[AConvexHullItem] := convexHullAnimation;
{$ENDC }
		end;
	end;

function triangulation2d.findMinimalAngleTriangle: pointVector;
	var
		result: pointVector;
		minSin: extended;
		mp: markedRealPoint;

	procedure findMin (tr: triangle);
		var
			p, q, r: point2d;
			a, b, c, area2, t: extended;
		begin
			tr.getPoints(obj(p), obj(q), obj(r));
			a := q.distanceX(r);
			b := r.distanceX(p);
			c := p.distanceX(q);
			area2 := abs(p.whichSideX(q, r));
			if (area2 > 0) & (a > 0) & (b > 0) & (c > 0) then begin
				t := area2 / (b * c);
				if t < minSin then begin
					result.elements^^[1] := r;
					result.elements^^[2] := p;
					result.elements^^[3] := q;
					minSin := t;
				end;
				t := area2 / (a * c);
				if t < minSin then begin
					result.elements^^[1] := p;
					result.elements^^[2] := q;
					result.elements^^[3] := r;
					minSin := t;
				end;
				t := area2 / (a * b);
				if t < minSin then begin
					result.elements^^[1] := q;
					result.elements^^[2] := r;
					result.elements^^[3] := p;
					minSin := t;
				end;
			end;
		end;

	begin
		result := newPointVector(3);
		minSin := 4.0;
		currentTriangulation := self;
		triangles.forAll(findMin);
		if minSin < 4.0 then begin
			new(mp);
			failNil(Handle(mp));
			mp.init;
			with point2d(result.elements^^[2]) do begin
				mp.x := getX;
				mp.y := getY;
			end;
			mp.mark.value := StringOf('Angle = ', arctan(minSin / sqrt(1 - sqr(minSin))) / oneDegree : 13 : 8, '');
			result.elements^^[1] := result.elements^^[1].duplicate;
			result.elements^^[2] := mp;
			result.elements^^[3] := result.elements^^[3].duplicate;
			result.length := 3;
			findMinimalAngleTriangle := result;
		end
		else
			findMinimalAngleTriangle := nil;
	end;

function triangle2d.getType: objectType;
	begin
		getType := triangle2dType;
	end;

function triangle2d.getTextType: Str255;
	begin
		getTextType := stringOf('Triangle 2d');
	end;

function triangle2d.memberType (o: obj): boolean;
	begin
		memberType := member(o, triangle2d);
	end;

procedure triangle2d.centerOfGravity (var cx, cy: extended);
	var
		cax, cbx, ccx, cay, cby, ccy: extended;
		refEdgeA, refEdgeB, refEdgeC: graphEdgeTriangulation;
	begin
		getEdges(refEdgeA, refEdgeB, refEdgeC);
		refEdgeA.centerOfGravity(cax, cay);
		refEdgeB.centerOfGravity(cbx, cby);
		refEdgeC.centerOfGravity(ccx, ccy);
		cx := (cax + cbx + ccx) / 3.0;
		cy := (cay + cby + ccy) / 3.0;
	end;

procedure triangulation2d.interactiveOutput;
	begin
		currentTriangulation := self;
		edges.interactiveOutput;
		points.interactiveOutput;
	end;

procedure triangulation2d.interactiveHighlight;
	begin
		currentTriangulation := self;
		edges.interactiveHighlight;
		points.interactiveHighlight;
	end;

function pointInTriangle (t: triangle): point2d;
	var
		result, v1, v2, v3: point2d;
		u, v, temp: extended;
	begin
		result := point2d(newObj(currentPointType));
		t.getPoints(obj(v1), obj(v2), obj(v3));
		u := randomReal;
		v := randomReal;
		if u > v then begin
			temp := u;
			u := v;
			v := temp;
		end;
		with result do begin
			setX(u * v1.getX + (v - u) * v2.getX + (1 - v) * v3.getX);
			setY(u * v1.getY + (v - u) * v2.getY + (1 - v) * v3.getY);
		end;
		pointInTriangle := result;
	end;

function triangulation2d.area: extended;
	var
		totalArea: extended;

	procedure computeArea (t: triangle);
		var
			v1, v2, v3: point2d;
		begin
			t.getPoints(obj(v1), obj(v2), obj(v3));
			totalArea := totalArea + abs(v1.whichSideX(v2, v3) / 2.0);
		end;

	begin
		currentTriangulation := self;
		totalArea := 0;
		triangles.forAll(computeArea);
		area := totalArea;
	end;

function triangulation2d.generateUniformRandomPoints (n: seqIndex0): homogenVector;
	var
		result, areas: homogenVector; (* of realNumber *)
		totalArea: real;
		i: seqIndex0;

	procedure computeArea (t: triangle);
		var
			x: realNumber;
			v1, v2, v3: point2d;
		begin
			new(x);
			failNil(Handle(x));
			t.getPoints(obj(v1), obj(v2), obj(v3));
			totalArea := totalArea + abs(v1.whichSideX(v2, v3));
			x.value := totalArea;
			areas.append(x);
		end;

	begin
		currentTriangulation := self;
		areas := newHomogenVector(triangles.length);
		totalArea := 0;
		triangles.forAll(computeArea);
		result := newHomogenVector(n);
		for i := 1 to n do
			result.append(pointInTriangle(triangle(triangles.elements^^[binaryFind(areas, randomReal * totalArea)])));
		areas.freeAll;
		generateUniformRandomPoints := result;
	end;

	const
		graphToTriangulation2dC = 1;
		trapezC = 2;
		TriangulateMonotoneC = 3;
		FlipEdgesC = 4;
		ShowNumberingC = 5;
		planeSweepC = 6;
		RandomPointsC = 7;
		TriangulateC = 8;
		ToLineSegmentsC = 9;
		FindTriangulationC = 10;
		ExtractPointsC = 11;
		FindSmallestAngleC = 12;
		TriangulateNoShuffleC = 13;
		RandomFlipEdgesC = 14;
		DeleteTriangleC = 15;
		DeleteVertexC = 16;
		DeleteBoundingTriangleC = 17;
		ToDelaunayTriangulationRandomizedC = 18;
		ToDelaunayTriangulationQueueC = 19;
		ToDelaunayTriangulationDegreeC = 20;
		ToDelaunayTriangulationStackXC = 21;
		ToDelaunayTriangulationRandomizedXC = 22;
		ToDelaunayTriangulationAngleC = 23;
		ToDelaunayTriangulationInteractiveXC = 24;
		ToDelaunayTriangulationAngle2C = 25;
		ToDelaunayTriangulationStackC = 26;
		ToDelaunayReverseStackC = 27;
		ToDelaunayReverseRandomizedC = 28;
		ToDelaunayReverseQueueC = 29;
		ToDelaunayReverseDegreeC = 30;
		ToDelaunayReverseStackXC = 31;
		ToDelaunayReverseRandomizedXC = 32;
		ToDelaunayReverseAngleC = 33;
		ToDelaunayReverseInteractiveXC = 34;
		ToDelaunayReverseAngle2C = 35;
		ComputeAreaC = 36;
		ExploreGraphDelaunayC = 37;
		ExploreGraphAllC = 38;
		ExploreGraphDelaunayReverseC = 39;
		ExploreGraphDelaunayDirectedC = 40;
		ExploreGraphAllDirectedC = 41;
		ExploreGraphDelaunayReverseDirectedC = 42;
		ExploreGraphDelaunayNoLabelC = 43;
		ExploreGraphAllNoLabelC = 44;
		ExploreGraphDelaunayReverseNoLabelC = 45;
		ExploreGraphDelaunayDirectedNoLabelC = 46;
		ExploreGraphAllDirectedNoLabelC = 47;
		ExploreGraphDelaunayReverseDirectedNoLabelC = 48;
		DeleteTriangleInteractiveC = 49;
		DeleteVertexInteractiveC = 50;
		ToUndirectedGraphC = 51;

function callRandomPoints (t: triangulation2d): homogenVector;
	begin
		if getUserParameter1('Number of points', numberOfObjects, numberOfObjects, @isSeqIndex) then
{$IFC THINK_PASCAL }
			callRandomPoints := t.generateUniformRandomPoints(trunc(numberOfObjects))
{$ELSEC }
			callRandomPoints := t.generateUniformRandomPoints(truncLong(numberOfObjects))
{$ENDC }
		else
			callRandomPoints := nil;
	end;

function callRandomFlipEdges (t: triangulation): triangulation;
	var
		result: triangulation;
	begin
		if getUserParameter1('Number of flips', numberOfExperiments, numberOfExperiments, @isNatural) then begin
			result := triangulation(t.duplicateAll);
{$IFC THINK_PASCAL }
			result.randomFlip(trunc(numberOfExperiments));
{$ELSEC }
			result.randomFlip(truncLong(numberOfExperiments));
{$ENDC }
			callRandomFlipEdges := result;
		end
		else
			callRandomFlipEdges := nil;
	end;


function deleteVertexSelect (t: triangulation): seqIndex0;
	begin
		deleteVertexSelect := t.points.interactiveSelectFromAll(true, true);
	end;

procedure deleteVertexChange (t: triangulation; selected: seqIndex);
	begin
		t.deleteVertex(selected);
	end;

function deleteTriangleSelect (t: triangulation): seqIndex0;
	begin
		currentTriangulation := t;
		deleteTriangleSelect := t.triangles.interactiveSelectFromAll(true, true);
	end;

procedure deleteTriangleChange (t: triangulation; selected: seqIndex);
	begin
		t.deleteTriangle(selected);
	end;

function callDeleteVertex (t: triangulation): triangulation;
	var
		result: triangulation;
	begin
		upperBound := t.points.length;
		if deleteWhichVertex > t.points.length then
			deleteWhichVertex := upperBound;
		if getUserParameter1('Number of vertex to delete', deleteWhichVertex, deleteWhichVertex, @upperBoundedNatural) then begin
			result := triangulation(t.duplicateAll);
{$IFC THINK_PASCAL }
			result.deleteVertex(trunc(deleteWhichVertex));
{$ELSEC }
			result.deleteVertex(truncLong(deleteWhichVertex));
{$ENDC }
			callDeleteVertex := result;
		end
		else
			callDeleteVertex := nil;
	end;

function callDeleteTriangle (t: triangulation): triangulation;
	var
		result: triangulation;
	begin
		upperBound := t.triangles.length;
		if numberOfVertices > t.triangles.length then
			numberOfVertices := upperBound;
		if getUserParameter1('Number of triangle to delete', deleteWhichTriangle, deleteWhichTriangle, @upperBoundedNatural) then begin
			result := triangulation(t.duplicateAll);
{$IFC THINK_PASCAL }
			result.deleteTriangle(trunc(deleteWhichTriangle));
{$ELSEC }
			result.deleteTriangle(truncLong(deleteWhichTriangle));
{$ENDC }
			callDeleteTriangle := result;
		end
		else
			callDeleteTriangle := nil;
	end;

procedure callFlipEdges (t: triangulation2d);
	var
		int: interactiveFlipCX;
		flips: longInt;
	begin
		currentTriangulation := t;
		new(int);
		failNil(Handle(int));
		flips := t.optimize2(justFlipOk, int);
		displayInfoString(StringOf('Interactively Flipped Edges = ', flips : 1));
	end;

procedure triangulation2d.describe;
	var
		isOpen: boolean;

	function possibleUdgTriangulation (udg: simpleUndirectedGraph): boolean;
		begin
			with udg do
				possibleUdgTriangulation := (edges.length >= vertices.length) & (vertices.length > 0) & member(vertices.elements^^[1], point2d);
		end;

	procedure addMenuEntry (entry: Str255; item: integer);
		begin
			if not isOpen then begin
				beginSubMenu('Triangulation 2d');
				isOpen := true
			end;
			menuEntry(entry, item);
		end;

	begin {triangulation2d.describe}
		isOpen := false;
		with currentArguments do begin
			if analyseInput <> 0 then begin
				addMenuEntry('Plane Sweep', planeSweepC);
				addMenuEntry('Trapez', trapezC);
			end;
			if (length = 1) & member(ith(1), simpleUndirectedGraph) & possibleUdgTriangulation(simpleUndirectedGraph(ith(1))) then
				addMenuEntry('Convert Graph', graphToTriangulation2dC);
			if (length = 1) & member(ith(1), polygon2d) & (polygon2d(ith(1)).length > 2) & member(polygon2d(ith(1)).elements^^[1], realPoint) then
				addMenuEntry('Triangulate Monotone', TriangulateMonotoneC);
			if (length = 1) & currentArguments1IsHomogenVector & (vector(ith(1)).length > 2) & member(vector(ith(1)).ith(1), point2d) then begin
				addMenuEntry('Randomized', TriangulateC);
{$IFC OptionalCommand }
				menuEntry('Randomized No Shuffle', TriangulateNoShuffleC);
{$ENDC }
			end;
			if isOpen then
				endSubMenu;
			if (length = 1) & member(ith(1), triangulation) then
				menuEntry('To Undirected Graph', ToUndirectedGraphC);
			if (length = 1) & member(ith(1), triangulation2d) & (triangulation2d(ith(1)).triangles.length > 0) then begin
				menuEntry('Show Numbering', ShowNumberingC);
				menuEntry('Random Points...', RandomPointsC);
				menuEntry('To Line Segments', ToLineSegmentsC);
{$IFC OptionalCommand }
				beginSubMenu('Explore Graph');
				menuEntry('Delaunay Flips', ExploreGraphDelaunayC);
				menuEntry('Delaunay Flips No Label', ExploreGraphDelaunayNoLabelC);
				menuEntry('Delaunay Flips Directed', ExploreGraphDelaunayDirectedC);
				menuEntry('Delaunay Flips Directed No Label', ExploreGraphDelaunayDirectedNoLabelC);
				menuEntry('Delaunay Reverse Flips', ExploreGraphDelaunayReverseC);
				menuEntry('Delaunay Reverse Flips No Label', ExploreGraphDelaunayReverseNoLabelC);
				menuEntry('Delaunay Reverse Flips Directed', ExploreGraphDelaunayReverseDirectedC);
				menuEntry('Delaunay Reverse Flips Directed No Label', ExploreGraphDelaunayReverseDirectedNoLabelC);
				menuEntry('All Flips', ExploreGraphAllC);
				menuEntry('All Flips No Label', ExploreGraphAllNoLabelC);
				menuEntry('All Flips Directed', ExploreGraphAllDirectedC);
				menuEntry('All Flips Directed No Label', ExploreGraphAllDirectedNoLabelC);
				endSubMenu;
				beginSubMenu('To Delaunay Triangulation');
				menuEntry('Stack', ToDelaunayTriangulationStackC);
				menuEntry('Queue', ToDelaunayTriangulationQueueC);
				menuEntry('Max Delta', ToDelaunayTriangulationAngleC);
				menuEntry('Max Angle', ToDelaunayTriangulationAngle2C);
				menuEntry('Dec Degree', ToDelaunayTriangulationDegreeC);
				menuEntry('Randomized', ToDelaunayTriangulationRandomizedC);
				menuEntry('Stack X', ToDelaunayTriangulationStackXC);
				menuEntry('Randomized X', ToDelaunayTriangulationRandomizedXC);
				menuEntry('Interactive', ToDelaunayTriangulationInteractiveXC);
				endSubMenu;
				beginSubMenu('To Delaunay Triangulation Reverse');
				menuEntry('Stack', ToDelaunayReverseStackC);
				menuEntry('Queue', ToDelaunayReverseQueueC);
				menuEntry('Max Delta', ToDelaunayReverseAngleC);
				menuEntry('Max Angle', ToDelaunayReverseAngle2C);
				menuEntry('Dec Degree', ToDelaunayReverseDegreeC);
				menuEntry('Randomized', ToDelaunayReverseRandomizedC);
				menuEntry('Stack X', ToDelaunayReverseStackXC);
				menuEntry('Randomized X', ToDelaunayReverseRandomizedXC);
				menuEntry('Interactive', ToDelaunayReverseInteractiveXC);
				endSubMenu;
{$ELSEC }
				menuEntry('To Delaunay Triangulation', ToDelaunayTriangulationStackC);
{$ENDC }
				beginSubMenu('Interactive Operations');
				menuEntry('Flip Edges...', FlipEdgesC);
				menuEntry('Delete Triangles...', DeleteTriangleInteractiveC);
				menuEntry('Delete Vertices...', DeleteVertexInteractiveC);
				endSubMenu;
				menuEntry('Find Most Acute Triangle', FindSmallestAngleC);
				menuEntry('Extract Points', ExtractPointsC);
			end;
			if (length = 1) & member(ith(1), triangulation) & (triangulation(ith(1)).triangles.length > 0) then begin
				menuEntry('Random Flip Edges...', RandomFlipEdgesC);
				menuEntry('Delete Triangle...', DeleteTriangleC);
				menuEntry('Delete Vertex...', DeleteVertexC);
			end;
{$IFC OptionalCommand }
			if (length = 1) & member(ith(1), rectangle) then
				menuEntry('Find Triangulation...', FindTriangulationC);
{$ENDC }
			if (length = 1) & member(ith(1), triangulation2d) & (triangulation(ith(1)).points.length >= 3) then
				menuEntry('Delete Bounding Triangle', DeleteBoundingTriangleC);
		end;
	end; {triangulation2d.describe}

function graphToTriangulation2d (udg: simpleUndirectedGraph): triangulation2d;
	var
		t: triangulation2d;

	function copyTransformation (p: point2d): obj;
		begin
			copyTransformation := p.duplicate;
		end;

	begin
		new(t);
		failNil(Handle(t));
		t.initializeFromGraph(udg, copyTransformation);
		graphToTriangulation2d := t;
	end; {graphToTriangulation2d}

function callInitializeFromPolygon2d (p: polygon2d): triangulation2d;
	var
		timer: longInt;
		t: triangulation2d;
	begin
		new(t);
		failNil(Handle(t));
		timer := TickCount;
		if not p.isMonotone then
			nonFatalError('Polygon is not monotone. Triangulation maybe incorrect.');
		t.initializeFromPolygon2d(p);
		startTime := startTime + TickCount - timer;
		callInitializeFromPolygon2d := t;
	end;

function triangulate2dByTrapez: triangulation2d;
	var
		t: triangulation2d;
		dummyTree, dummyTree2: avlTree;
		outerPolygons, innerPolygons: vector;
		dummyVector: homogenVector;
		globalShift: integer;

	function newTriangle: triangle;
		var
			t: triangle2d;
		begin
			new(t);
			newTriangle := t
		end;

	begin
		globalShift := 0;
		dummyTree := nil;
		dummyTree2 := nil;
		doWholeTriangulation(t, false, globalShift, dummyVector, dummyTree, dummyTree2, outerPolygons, innerPolygons, newTriangle);
		triangulate2dByTrapez := t;
	end; {triangulate2dByTrapez}


function triangulate2dByPlaneSweep: triangulation2d;
	var
		t2d: triangulation2d;
		polyInfo: triangulationPolygonInfoDirectory;
		sweepLine: triangulationSweepLine;
		inputType: integer;

		iPolys, oPolys: vector;
		iPoly, oPoly: polygon2d;

	function sumVertices (v: vector): seqIndex0;
		var
			l: seqIndex;
			sum: seqIndex0;
		begin
			sum := 0;
			with v do
				for l := 1 to length do
					sum := sum + polygon2d(elements^^[l]).length;
			sumVertices := sum
		end;

	begin {triangulate2dByPlaneSweep}
		new(t2d);									(* initialize data structures *)
		failNil(Handle(t2d));

{    %_PTrace := true;}
{    ResetProfile;}

		new(polyInfo);
		failNil(Handle(polyInfo));
		new(sweepLine);
		failNil(Handle(sweepLine));
		sweepLine.init;

		inputType := analyseInput;
		case inputType of
			1, 3, 4: 
				oPoly := polygon2d(list(currentArguments).ith(1));
			2, 5, 6: 
				oPolys := vector(list(currentArguments).ith(1));
		end;
		case inputType of
			1, 2: 
				;
			3, 5: 
				iPoly := polygon2d(list(currentArguments).ith(2));
			4, 6: 
				iPolys := vector(list(currentArguments).ith(2));
		end;
		case inputType of
			1:  begin
				polyInfo.initialize(t2d, oPoly.length, 1, 0);
				polyInfo.appendPolygon(oPoly);
			end;
			2:  begin
				polyInfo.initialize(t2d, sumVertices(oPolys), oPolys.length, 0);
				polyInfo.appendPolygons(oPolys);
			end;
			3:  begin
				polyInfo.initialize(t2d, oPoly.length + iPoly.length, 1, 1);
				polyInfo.appendPolygon(oPoly);
				polyInfo.appendPolygon(iPoly);
			end;
			4:  begin
				polyInfo.initialize(t2d, oPoly.length + sumVertices(iPolys), 1, iPolys.length);
				polyInfo.appendPolygon(oPoly);
				polyInfo.appendPolygons(iPolys);
			end;
			5:  begin
				polyInfo.initialize(t2d, sumVertices(oPolys) + iPoly.length, oPolys.length, 1);
				polyInfo.appendPolygons(oPolys);
				polyInfo.appendPolygon(iPoly);
			end;
			6:  begin
				polyInfo.initialize(t2d, sumVertices(oPolys) + sumVertices(iPolys), oPolys.length, iPolys.length);
				polyInfo.appendPolygons(oPolys);
				polyInfo.appendPolygons(iPolys);
			end;
		end;

		polyInfo.sort(1, t2d.points.length);
		with sweepLine do begin
			initialize(polyInfo, t2d);
			startSweep;
			free;					(* release data structures *)
		end;
		polyInfo.free;

{    DumpProfileToFile('Profile2D');}
{    TerminateProfile;}

		triangulate2dByPlaneSweep := t2d
	end; {triangulate2dByPlaneSweep}

function callTriangulation (v: vector; shuffle: boolean): searchTriangulation2d;
	var
		result: searchTriangulation2d;
	begin
		new(result);
		failNil(Handle(result));
		result.initFromPointVector(v, shuffle);
		callTriangulation := result;
	end;

function callDeleteBoundingTriangle (t: triangulation): triangulation;
	var
		result: triangulation;
	begin
		result := triangulation(t.duplicateAll);
		result.deleteVertex(3);
		result.deleteVertex(2);
		result.deleteVertex(1);
		callDeleteBoundingTriangle := result;
	end;

function getContainer (item: integer): edgeContainer;
	var
		s: stackC;
		r: randomC;
		q: queueC;
		d: degreeC;
		a: minAngleC;
		a2: minAngle2C;
		result: edgeContainer;
	begin
		result := nil;
		case item of
			ToDelaunayTriangulationStackC, ToDelaunayReverseStackC:  begin
				new(s);
				result := s;
			end;
			ToDelaunayTriangulationQueueC, ToDelaunayReverseQueueC:  begin
				new(q);
				result := q;
			end;
			ToDelaunayTriangulationRandomizedC, ToDelaunayReverseRandomizedC:  begin
				new(r);
				result := r;
			end;
			ToDelaunayTriangulationDegreeC, ToDelaunayReverseDegreeC:  begin
				new(d);
				result := d;
			end;
			ToDelaunayTriangulationAngle2C, ToDelaunayReverseAngle2C:  begin
				new(a2);
				result := a2;
			end;
			ToDelaunayTriangulationAngleC, ToDelaunayReverseAngleC:  begin
				new(a);
				result := a;
			end;
		end;
		failNil(Handle(result));
		getContainer := result;
	end;

function getContainerX (item: integer): edgeContainerX;
	var
		s: stackCX;
		r: randomCX;
		i: interactiveCX;
		result: edgeContainerX;
	begin
		result := nil;
		case item of
			ToDelaunayTriangulationStackXC, ToDelaunayReverseStackXC:  begin
				new(s);
				result := s;
			end;
			ToDelaunayTriangulationRandomizedXC, ToDelaunayReverseRandomizedXC:  begin
				new(r);
				result := r;
			end;
			ToDelaunayTriangulationInteractiveXC, ToDelaunayReverseInteractiveXC:  begin
				new(i);
				result := i;
			end;
		end;
		failNil(Handle(result));
		getContainerX := result;
	end;

function findTriangulation (n, minDifference: longInt; r: rectangle; betterContainer, goodContainer: edgeContainer; var tries: longInt): triangulation2d;
	var
		pv: pointVector;
		tGood, tBetter, result: searchTriangulation2d;
		cGood, cBetter: edgeContainer;
		flipsGood, flipsBetter: longInt;
		found: boolean;
	begin
		pv := createRealPointVector(n);
		currentRandomConstraint := r;
		tries := 0;
		repeat
			tries := tries + 1;
			pv.randomChange;
			new(result);
			failNil(Handle(result));
			with result do begin
				initFromPointVector(pv, false);
				if userAbort then begin
					result.freeAll;
					pv.freeAll;
					dispose(goodContainer);
					dispose(betterContainer);
					findTriangulation := nil;
					exit(findTriangulation);
				end;
				deleteVertex(1);
				deleteVertex(2);
				deleteVertex(3);
			end;
			tGood := searchTriangulation2d(result.duplicateAll);
			tBetter := searchTriangulation2d(result.duplicateAll);
			cGood := edgeContainer(goodContainer.shallowDuplicate);
			cBetter := edgeContainer(betterContainer.shallowDuplicate);
			flipsGood := tGood.optimize(delaunayFlipOk, cGood);
			flipsBetter := tBetter.optimize(delaunayFlipOk, cBetter);
			tGood.freeAll;
			tBetter.freeAll;
			found := minDifference <= flipsGood - flipsBetter;
			if not found then
				result.freeAll;
		until found | userAbort;
		if userAbort then begin
			if found then
				result.freeAll;
			findTriangulation := nil;
		end
		else
			findTriangulation := result;
		pv.freeAll;
		dispose(goodContainer);
		dispose(betterContainer);
	end;

function callFindTriangulation (r: rectangle): triangulation2d;
	var
		tries: longInt;

	function getContainer (item: integer): edgeContainer;
		var
			s: stackC;
			q: queueC;
			d: degreeC;
			a: minAngleC;
			a2: minAngle2C;
			result: edgeContainer;
		begin
			result := nil;
			case item of
				stackSel:  begin
					new(s);
					result := s;
				end;
				queueSel:  begin
					new(q);
					result := q;
				end;
				degreeSel:  begin
					new(d);
					result := d;
				end;
				angleSel:  begin
					new(a2);
					result := a2;
				end;
				deltaSel:  begin
					new(a);
					result := a;
				end;
			end;
			failNil(Handle(result));
			getContainer := result;
		end;

	begin
		if getMethodSelection then begin
{$IFC THINK_PASCAL }
			callFindTriangulation := findTriangulation(trunc(numberOfObjects), trunc(minimalDifference), r, getContainer(sel1), getContainer(sel2), tries);
{$ELSEC }
			callFindTriangulation := findTriangulation(truncLong(numberOfObjects), truncLong(minimalDifference), r, getContainer(sel1), getContainer(sel2), tries);
{$ENDC }
			displayInfoString(StringOf('Number Of Tries = ', tries : 1));
		end
		else
			callFindTriangulation := nil;
	end;

procedure triangulation2d.execute (item: integer);
	var
		t: triangulation2d;
		flips: longInt;

	procedure getT;
		begin
			t := triangulation2d(currentArguments.ith(1).duplicateAll);
		end;

	procedure processT;
		begin
			if abortFlag then
				t.freeAll
			else begin
				displayInfoString(StringOf('Number Of Flips = ', flips : 1));
				if flips = 0 then
					t.freeAll
				else
					currentResult.append(t);
			end;
		end;

	begin
		with currentArguments do begin
			case item of
				TriangulateC, TriangulateNoShuffleC: 
					currentResult.appendNonNil(callTriangulation(vector(ith(1)), item = TriangulateC));
				ToLineSegmentsC: 
					currentResult.appendNonNil(triangulation2d(ith(1)).createLineSegments);
				ToUndirectedGraphC: 
					currentResult.appendNonNil(triangulation2d(ith(1)).toUndirectedGraph);
				ExtractPointsC: 
					currentResult.append(triangulation2d(ith(1)).points.duplicateAll);
				ExploreGraphDelaunayC, ExploreGraphDelaunayDirectedC, ExploreGraphDelaunayNoLabelC, ExploreGraphDelaunayDirectedNoLabelC: 
					currentResult.appendNonNil(exploreGraph(triangulation2d(ith(1)), delaunayFlipOk, (item = ExploreGraphDelaunayDirectedC) | (item = ExploreGraphDelaunayDirectedNoLabelC), (item = ExploreGraphDelaunayC) | (item = ExploreGraphDelaunayDirectedC)));
				ExploreGraphAllC, ExploreGraphAllDirectedC, ExploreGraphAllNoLabelC, ExploreGraphAllDirectedNoLabelC: 
					currentResult.appendNonNil(exploreGraph(triangulation2d(ith(1)), justFlipOk, (item = ExploreGraphAllDirectedC) | (item = ExploreGraphAllDirectedNoLabelC), (item = ExploreGraphAllC) | (item = ExploreGraphAllDirectedC)));
				ExploreGraphDelaunayReverseC, ExploreGraphDelaunayReverseDirectedC, ExploreGraphDelaunayReverseNoLabelC, ExploreGraphDelaunayReverseDirectedNoLabelC: 
					currentResult.appendNonNil(exploreGraph(triangulation2d(ith(1)), delaunayFlipOkReverse, (item = ExploreGraphDelaunayReverseDirectedC) | (item = ExploreGraphDelaunayReverseDirectedNoLabelC), (item = ExploreGraphDelaunayReverseC) | (item = ExploreGraphDelaunayReverseDirectedC)));
				FlipEdgesC: 
					callFlipEdges(triangulation2d(ith(1)));
				ToDelaunayTriangulationStackC, ToDelaunayTriangulationQueueC, ToDelaunayTriangulationRandomizedC, ToDelaunayTriangulationAngleC, ToDelaunayTriangulationAngle2C, ToDelaunayTriangulationDegreeC:  begin
					getT;
					flips := t.optimize(delaunayFlipOk, getContainer(item));
					processT;
				end;
				ToDelaunayTriangulationStackXC, ToDelaunayTriangulationRandomizedXC, ToDelaunayTriangulationInteractiveXC:  begin
					getT;
					flips := t.optimize2(delaunayFlipOk, getContainerX(item));
					processT;
				end;
				ToDelaunayReverseStackC, ToDelaunayReverseQueueC, ToDelaunayReverseRandomizedC, ToDelaunayReverseAngleC, ToDelaunayReverseAngle2C, ToDelaunayReverseDegreeC:  begin
					getT;
					flips := t.optimize(delaunayFlipOkReverse, getContainer(item));
					processT;
				end;
				ToDelaunayReverseStackXC, ToDelaunayReverseRandomizedXC, ToDelaunayReverseInteractiveXC:  begin
					getT;
					flips := t.optimize2(delaunayFlipOkReverse, getContainerX(item));
					processT;
				end;
				DeleteBoundingTriangleC: 
					currentResult.append(callDeleteBoundingTriangle(triangulation2d(ith(1))));
				FindTriangulationC: 
					currentResult.appendNonNil(callFindTriangulation(rectangle(ith(1))));
				trapezC: 
					currentResult.append(triangulate2dByTrapez);
				planeSweepC: 
					currentResult.append(triangulate2dByPlaneSweep);
				graphToTriangulation2dC: 
					currentResult.append(graphToTriangulation2d(simpleUndirectedGraph(ith(1))));
				TriangulateMonotoneC: 
					currentResult.appendNonNil(callInitializeFromPolygon2d(polygon2d(ith(1))));
				ShowNumberingC: 
					currentResult.append(triangulation2d(ith(1)).showNumbering);
				FindSmallestAngleC: 
					currentResult.appendNonNil(triangulation2d(ith(1)).findMinimalAngleTriangle);
				RandomPointsC: 
					currentResult.appendNonNil(callRandomPoints(triangulation2d(ith(1))));
				RandomFlipEdgesC: 
					currentResult.appendNonNil(callRandomFlipEdges(triangulation(ith(1))));
				DeleteTriangleC: 
					currentResult.appendNonNil(callDeleteTriangle(triangulation(ith(1))));
				DeleteVertexC: 
					currentResult.appendNonNil(callDeleteVertex(triangulation(ith(1))));
				DeleteTriangleInteractiveC: 
					interactiveChange('Click at a triangle to delete it!', triangulation2d(ith(1)), deleteTriangleSelect, deleteTriangleChange);
				DeleteVertexInteractiveC: 
					interactiveChange('Click at a vertex to delete it!', triangulation(ith(1)), deleteVertexSelect, deleteVertexChange);
			end;
		end;
	end; {triangulation2d.execute}

function triangulation2d.getType: objectType;
	begin
		getType := triangulation2dType;
	end;

function triangulation2d.getTextType: Str255;
	begin
		getTextType := stringOf('Triangulation 2d (V=', points.length : 1, ',E=', edges.length : 1, ',T=', triangles.length : 1, ')');
	end;

function triangulation2d.memberType (o: obj): boolean;
	begin
		memberType := member(o, triangulation2d);
	end;

function triangulation2d.isSelected (where: obj): boolean;
	begin
		isSelected := points.isSelected(where);
	end;

function triangulation2d.pointCover (n: longInt): obj;
	var
		udg: simpleUndirectedGraph;
		homogenEdges: homogenVector;
	begin
		new(homogenEdges);
		failNil(Handle(homogenEdges));
		homogenEdges.shallowInitializeFromObject(edges);
		new(udg);
		failNil(Handle(udg));
		udg.vertices := points;
		udg.edges := homogenEdges;
		pointCover := udg.pointCover(n);
		dispose(udg);
		dispose(homogenEdges);
	end;

function triangulation2d.addNewEdge (startVertexNew, endVertexNew: seqIndex): graphEdgeTriangulation;
	var
		e: graphEdgeTriangulation;
	begin
		new(e);
		failNil(Handle(e));
		e.init;
		edges.append(e);
		addNewEdge := inherited addNewEdge(startVertexNew, endVertexNew);
	end;

procedure triangulation2d.addNewTriangle (edgeANew, edgeBNew, edgeCNew: graphEdgeTriangulation);
	var
		t: triangle2d;
	begin
		new(t);
		failNil(Handle(t));
		t.init;
		triangles.append(t);
		inherited addNewTriangle(edgeANew, edgeBNew, edgeCNew)
	end;

function triangulation2d.showNumbering: homogenVector;
	var
		result: homogenVector;
		ch: char;

	procedure prePend (mp: markedRealPoint);
		begin
			mp.mark.value := concat(ch, mp.mark.value);
		end;

	begin
		currentTriangulation := self;
		setCurrentVerticesVector(points);
		result := newHomogenVector(3);
		result.append(points.showNumbering);
		result.append(edges.showNumbering);
		result.append(triangles.showNumbering);
		ch := 'V';
		sequence(result.elements^^[1]).forAll(prePend);
		ch := 'E';
		sequence(result.elements^^[2]).forAll(prePend);
		ch := 'T';
		sequence(result.elements^^[3]).forAll(prePend);
		showNumbering := result;
	end;

procedure triangulation2d.initializeFromPolygon2d (polygon: polygon2d);
	var
		concaveChain: homogenVector; (* of graphEdgeTriangulation *)
		polyIndex: seqIndex1;
		pq, qr, rp: graphEdgeTriangulation;
		idPoints: pointVector;
		p, q, r: idRealPoint;
		concaveSign: signType;
{$IFC AnimationEnabled AND ATriangulation }
		flash: homogenVector;
{$ENDC }

	procedure advanceQR;
		begin
			q := r;
			polyIndex := polyIndex + 1;
			if polyIndex <= polygon.length then begin
				r := idRealPoint(idPoints.elements^^[polyIndex]);
				concaveChain.append(addNewEdge(q.id, r.id));
			end;
		end;

	begin
		idPoints := polygon.makeIdPoints;
		initializeEuler(polygon.length, 0);
		with idPoints do begin
			p := idRealPoint(elements^^[1]);
			q := idRealPoint(elements^^[2]);
			r := idRealPoint(elements^^[3]);
		end;
		concaveChain := newHomogenVector(polygon.length);
		concaveChain.append(addNewEdge(1, 2));
		concaveChain.append(addNewEdge(2, 3));
		for polyIndex := 1 to polygon.length do
			points.elements^^[polyIndex] := polygon.elements^^[polyIndex].duplicateAll;
		points.length := polygon.length;
		polyIndex := 3;
		concaveSign := p.whichSideSign(point2d(idPoints.elements^^[polygon.length]), q);
		while polyIndex <= polygon.length do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				freeAll;
				idPoints.freeAll;
				concaveChain.free;
				exit(initializeFromPolygon2d);
			end;
{$ENDC }
{$IFC AnimationEnabled AND ATriangulation }
			if animationFlag[ATriangulationItem] then begin
				flash := newHomogenVector(3);
				flash.append(p);
				flash.append(q);
				flash.append(r);
				flash.interactiveFlash;
				waitForClick(animationFlag[ATriangulationItem]);
				flash.free;
			end;
{$ENDC }
			if r.whichSideSign(p, q) = concaveSign then begin (* add to concave chain *)
				p := q;
				advanceQR;
			end
			else begin (* produce triangle *)
				rp := addNewEdge(p.id, r.id);
{$IFC AnimationEnabled AND ATriangulation }
				if animationFlag[ATriangulationItem] then begin
					r.drawLine(p);
					waitForClick(animationFlag[ATriangulationItem]);
				end;
{$ENDC }
				with concaveChain do begin
					pq := graphEdgeTriangulation(elements^^[length - 1]);
					qr := graphEdgeTriangulation(elements^^[length]);
					length := length - 1;
					elements^^[length] := rp;
				end;
				addNewTriangle(pq, qr, rp);
				if concaveChain.length > 1 then begin
					q := p;
					with concaveChain do
						p := idRealPoint(idPoints.elements^^[graphEdgeTriangulation(elements^^[length - 1]).startVertex]);
				end
				else
					advanceQR;
			end;
		end;
		idPoints.freeAll;
		concaveChain.free;
	end;


end.
