(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(*														*)
(* Copyright 1992 by Michael Greminger					*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.0:  31-Jan-92								*)
(*														*)
(* Copyright 1992 by Bert Leukens						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.1:  May-13-92								*)

unit triangulation2d2;

interface

uses
	triangulation;

function analyseInput: integer;

procedure doWholeTriangulation (var t: triangulation2d; fromLayer: boolean; var globalShift: integer;									(* the global shift of the already allocated vertices *)
							var outerPolGlobalIndexes: homogenVector; fromLastLayerEdgeTree: avlTree; var newLayerEdgeTree: avlTree;					(* stores the edges for the next layer *)
							var outerPolygons, innerPolygons: vector; function newTriangle: triangle);


implementation

function analyseInput: integer;	(* changed by B. Leukens *)

	function polygonIsOK (o: obj): boolean;
		begin
			polygonIsOK := member(o, polygon2d) & (polygon2d(o).length > 2) & member(polygon2d(o).elements^^[1], realPoint)
		end;

	function vectorIsOK (o: obj): boolean;
		var
			ok: boolean;
			l: seqIndex0;
		begin
			ok := member(o, vector) & (vector(o).length > 0);
			if ok then begin
				if member(o, homogenVector) then
					ok := polygonIsOK(homogenVector(o).elements^^[1])
				else begin
					l := vector(o).length;
					while ok & (l > 0) do begin
						ok := polygonIsOK(vector(o).elements^^[l]);
						l := l - 1
					end;
				end;
			end;
			vectorIsOK := ok;
		end;

	begin
		analyseInput := 0;
		with currentArguments do
			if length = 1 then begin
				if polygonIsOK(ith(1)) then
					analyseInput := 1
				else if vectorIsOK(ith(1)) then
					analyseInput := 2;
			end
			else if length = 2 then begin
				if polygonIsOK(ith(1)) & polygonIsOK(ith(2)) then
					analyseInput := 3
				else if polygonIsOK(ith(1)) & vectorIsOK(ith(2)) then
					analyseInput := 4
				else if vectorIsOK(ith(1)) & polygonIsOK(ith(2)) then
					analyseInput := 5
				else if vectorIsOK(ith(1)) & vectorIsOK(ith(2)) then
					analyseInput := 6;
			end
	end;


	type
		polyPoint = object(obj)
				vertex: point2d;
				duplInd: seqIndex0;
		  (* if duplInd<>0 then it means that this vertex is only a temporary copy of duplInd *)
				prev, next: seqIndex0;
				trapez1Point2: seqIndex0;
				edge1L, edge1R: graphEdge;
				trapez2Point2: seqIndex0;
				edge2L, edge2R: graphEdge;
			end;

		trapezoid = object(obj)
				vertex1, vertex2: seqIndex0;
				edgeL, edgeR: graphEdge;
			end;

		triangPoly = object(obj)
				startInd, endInd: seqIndex0;
				done: boolean;
			end;

function compareIndexes (a: int; b: triangPoly): relationObj;
	begin
		if a.value < b.startInd then
			compareIndexes := lessThanObj
		else if a.value > b.endInd then
			compareIndexes := greaterThanObj
		else
			compareIndexes := equalObj;
	end;

function compareGlobalIndexes (a: int; b: graphEdge): relationObj;
	begin
		if a.value < b.startVertex then
			compareGlobalIndexes := lessThanObj
		else if a.value > b.endVertex then
			compareGlobalIndexes := greaterThanObj
		else
			compareGlobalIndexes := equalObj;
	end;

procedure triangulateFirst (first, last: seqIndex; shift, globalShift: seqIndex0; outerPolGlobalIndexes, sortedInnerPolGlobalIndexes: homogenVector; fromLayer: boolean; currentOuterPolygonNr: seqIndex; fromLastLayerEdgeTree: avlTree; globalStartInPolInd, localStartInPolInd: integer; var newLayerEdgeTree: avlTree; var pointVect: homogenVector; var t: triangulation2d; function newTriangle: triangle);
	var
		edgesTree: avlTree;
		toSearch: int;
		bufGraphEdge: graphEdgeTriangulation;	(* Buffer avoids splitting of memory *)

	procedure triangulate (first: seqIndex);
		var
			curVertex, botVertex: polyPoint;
			botVertexInd, curVertexInd: seqIndex0;
			saveNext, savePrev: seqIndex;
			isUnimonotone, firstTime, stop: boolean;

		function diagonal (pInd: seqIndex): seqIndex0;
			var
				p: polyPoint;

			begin	{diagonal}
				p := polyPoint(pointVect.elements^^[pInd]);
				diagonal := 0;
				if p.trapez1Point2 <> 0 then begin
					if p.trapez2Point2 = 0 then begin	(* only one trapezoid in this vertex *)
						if (not (((p.edge1L.startVertex = pInd) & (p.edge1L.endVertex = p.trapez1Point2)) | ((p.edge1R.startVertex = pInd) & (p.edge1R.endVertex = p.trapez1Point2)))) then begin
							diagonal := p.trapez1Point2;
							p.trapez1Point2 := 0;
						end
					end
					else begin
						if (not (((p.edge1L.startVertex = pInd) & (p.edge1L.endVertex = p.trapez1Point2)) | ((p.edge1R.startVertex = pInd) & (p.edge1R.endVertex = p.trapez1Point2)))) then begin
							diagonal := p.trapez1Point2;
							p.edge1L := p.edge2L;
							p.edge1R := p.edge2R;
							p.trapez1Point2 := p.trapez2Point2;
							p.trapez2Point2 := 0;
						end
						else begin
							if (not (((p.edge2L.startVertex = pInd) & (p.edge2L.endVertex = p.trapez2Point2)) | ((p.edge2R.startVertex = pInd) & (p.edge2R.endVertex = p.trapez2Point2)))) then begin
								diagonal := p.trapez2Point2;
								p.trapez2Point2 := 0;
							end
						end;
					end;
				end;
			end;

		procedure triangulateMonoton (first: seqIndex);
			var
				numberVertices: integer;
				startVertexInd, curVertexInd, nextInd, nextNextInd, prevInd: seqIndex;
				curVertex: polyPoint;

			procedure determineStartVertex (first: seqIndex; var startVertexInd: seqIndex; var numberVertices: integer);
				var
					curInd, lowestYInd, highestYInd, startHigh, endHigh, startLow, endLow: seqIndex;
					lowestY, highestY, y: extended;

				procedure horizontalChain (ind: seqIndex; var startInd, endInd: seqIndex);
					var
						p, temp: polyPoint;
						y, lastY: extended;
						tempInd: seqIndex;
					begin
						p := polyPoint(pointVect.elements^^[ind]);
						lastY := p.vertex.getY;
						tempInd := p.next;
						temp := polyPoint(pointVect.elements^^[p.next]);
						while tempInd <> ind do begin
							temp := polyPoint(pointVect.elements^^[tempInd]);
							y := temp.vertex.getY;
							if y <> lastY then begin
								lastY := y;
								startInd := tempInd
							end;
							tempInd := temp.next;
						end;
						if y <> p.vertex.getY then
							startInd := ind;
						y := p.vertex.getY;
						tempInd := ind;
						temp := p;
						repeat
							endInd := tempInd;
							tempInd := temp.next;
							temp := polyPoint(pointVect.elements^^[tempInd]);
						until temp.vertex.getY <> y;
					end;

				begin {determineStartVertex}
					lowestY := polyPoint(pointVect.elements^^[first]).vertex.getY;
					highestY := lowestY;
					lowestYInd := first;
					highestYInd := first;
					curInd := polyPoint(pointVect.elements^^[first]).next;
					numberVertices := 1;
					while curInd <> first do begin
						y := polyPoint(pointVect.elements^^[curInd]).vertex.getY;
						if y > highestY then begin
							highestY := y;
							highestYInd := curInd;
						end
						else if y < lowestY then begin
							lowestY := y;
							lowestYInd := curInd;
						end;
						numberVertices := numberVertices + 1;
						curInd := polyPoint(pointVect.elements^^[curInd]).next;
					end;
					if polyPoint(pointVect.elements^^[lowestYInd]).next = highestYInd then
						startVertexInd := highestYInd		(* monotone chain is on the right *)
					else if polyPoint(pointVect.elements^^[highestYInd]).next = lowestYInd then
						startVertexInd := lowestYInd		(* monotone chain is on the left *)
					else begin		(* there is a horizontal edge *)
						horizontalChain(highestYInd, startHigh, endHigh);
						horizontalChain(lowestYInd, startLow, endLow);
						if polyPoint(pointVect.elements^^[endLow]).next = startHigh then
							startVertexInd := startHigh		(* monotone chain is on the right *)
						else if polyPoint(pointVect.elements^^[endHigh]).next = startLow then
							startVertexInd := startLow		(* monotone chain is on the left *)
						else
							fatalError('Startindex Error');
					end;
				end;	(* determineStartVertex *)

			function isConvex (prevP, curP, nextP: point2d): boolean;
				begin
					isConvex := nextP.whichSideSign(prevP, curP) = 1;
				end;

			function compareEdges (a, b: graphEdge): relationObj;
				begin
					if a.startVertex < b.startVertex then
						compareEdges := lessThanObj
					else if a.startVertex > b.startVertex then
						compareEdges := greaterThanObj
					else if a.endVertex < b.endVertex then
						compareEdges := lessThanObj
					else if a.endVertex > b.endVertex then
						compareEdges := greaterThanObj
					else
						compareEdges := equalObj;
				end;

			procedure triAppend (p1, p2, p3: seqIndex);
				var
					e1, e2, e3: graphEdgeTriangulation;
					b1, b2, b3: boolean;
					triang: triangle;

				procedure sortPoints (var p1, p2: seqIndex; var b1, b2: boolean);
					var
						temp: seqIndex;
						tempB: boolean;
					begin
						if p1 > p2 then begin
							temp := p1;
							tempB := b1;
							p1 := p2;
							b1 := b2;
							p2 := temp;
							b2 := tempB;
						end;
					end;

				procedure edgAppend (p1, p2: seqIndex; b1, b2: boolean; var e: graphEdgeTriangulation; triangleId: seqIndex);
					var
						dummy: directionType;
						where: reference;

					function areNeighbourVertices1 (p1, p2: seqIndex): boolean;
						var
							where: integer;
						begin
							if sortedInnerPolGlobalIndexes.length = 1 then
								areNeighbourVertices1 := (p1 + 1 = p2) | ((p1 = graphEdge(sortedInnerPolGlobalIndexes.elements^^[1]).startVertex) & (p2 = graphEdge(sortedInnerPolGlobalIndexes.elements^^[1]).endVertex))
							else begin
								toSearch.value := p1;
								where := sortedInnerPolGlobalIndexes.binarySearch(toSearch, compareGlobalIndexes, 1, sortedInnerPolGlobalIndexes.length);
								areNeighbourVertices1 := (p2 <= graphEdge(sortedInnerPolGlobalIndexes.elements^^[where]).endVertex) & ((p1 + 1 = p2) | ((p1 = graphEdge(sortedInnerPolGlobalIndexes.elements^^[where]).startVertex) & (p2 = graphEdge(sortedInnerPolGlobalIndexes.elements^^[where]).endVertex)));
							end;
						end;

					procedure testEdgesTree;
						var
							dummy: directionType;
							where: reference;
						begin
							if edgesTree.find(e, compareEdges, where, dummy) then begin
								bufGraphEdge := e;
								e := graphEdgeTriangulation(edgesTree.getObject(where));
								e.triangleB := triangleId;
								exit(edgAppend);
							end;
						end;

					begin
						if bufGraphEdge = nil then begin
							new(e);
							failNil(Handle(e));
							e.init;
						end
						else begin
							e := bufGraphEdge;
							bufGraphEdge := nil;
						end;
						e.startVertex := p1;
						e.endVertex := p2;
						if fromLayer then
							if (b1 & b2 & (globalShift <> 0) & ((p1 + 1 = p2) | ((p1 = graphEdge(outerPolGlobalIndexes.elements^^[currentOuterPolygonNr]).startVertex) & (p2 = graphEdge(outerPolGlobalIndexes.elements^^[currentOuterPolygonNr]).endVertex)))) then begin
								if not fromLastLayerEdgeTree.find(e, compareEdges, where, dummy) then
									fatalError('Error in edgeAppend');
								bufGraphEdge := e;
								e := graphEdgeTriangulation(fromLastLayerEdgeTree.getObject(where));
								e.triangleB := triangleId;
								exit(edgAppend);
							end
							else if (not (b1) & not (b2) & areNeighbourVertices1(p1, p2)) then
								newLayerEdgeTree.insertNewObj(e, compareEdges)
							else
								testEdgesTree
						else
							testEdgesTree;
						t.edges.appendSafe(e, 10);
						edgesTree.insertNewObj(e, compareEdges);
						e.id := t.edges.length;
						e.triangleA := triangleId;
						e.triangleB := 0;
					end;

				procedure eliminateDupl (var pInd: seqIndex);
					begin
						if polyPoint(pointVect.elements^^[pInd]).duplInd <> 0 then
							pInd := polyPoint(pointVect.elements^^[pInd]).duplInd;
					end;

				procedure adjustIndex (var pInd: seqIndex; var adjudsted: boolean);
					begin
						adjudsted := false;
						if (pInd < localStartInPolInd) then begin
							if globalShift <> 0 then
								pInd := graphEdge(outerPolGlobalIndexes.elements^^[currentOuterPolygonNr]).endVertex - pInd + 1
							else
								pInd := pInd + shift;
							adjudsted := true;
						end
						else if globalShift <> 0 then
							pInd := pInd - localStartInPolInd + globalStartInPolInd
						else
							pInd := pInd + shift;
					end;

				begin
{$IFC AnimationEnabled }
					advanceStopWatchHands;
{$ENDC }
					eliminateDupl(p1);
					eliminateDupl(p2);
					eliminateDupl(p3);
					if fromLayer then begin
						adjustIndex(p1, b1);
						adjustIndex(p2, b2);
						adjustIndex(p3, b3);
					end
					else begin
						p1 := p1 + shift;
						p2 := p2 + shift;
						p3 := p3 + shift;
						b1 := false;
						b2 := false;
						b3 := false;
					end;
					sortPoints(p1, p2, b1, b2);
					sortPoints(p2, p3, b2, b3);
					sortPoints(p1, p2, b1, b2);
					triang := newTriangle;
					failNil(Handle(triang));
					triang.init;
					t.triangles.appendSafe(triang, 5);
					edgAppend(p1, p2, b1, b2, e1, t.triangles.length);
					edgAppend(p1, p3, b1, b3, e2, t.triangles.length);
					edgAppend(p2, p3, b2, b3, e3, t.triangles.length);
					triang.edgeA := e1.id;
					triang.edgeB := e2.id;
					triang.edgeC := e3.id;
				end;

			begin		(* triangulateMonoton *)
				determineStartVertex(first, startVertexInd, numberVertices);
				curVertexInd := startVertexInd;
				while numberVertices >= 3 do begin
					curVertex := polyPoint(pointVect.elements^^[curVertexInd]);
					nextInd := curVertex.next;
					nextNextInd := polyPoint(pointVect.elements^^[curVertex.next]).next;
					if isConvex(curVertex.vertex, polyPoint(pointVect.elements^^[nextInd]).vertex, polyPoint(pointVect.elements^^[nextNextInd]).vertex) then begin
						nextInd := curVertex.next;
						prevInd := curVertex.prev;
						triAppend(curVertexInd, nextInd, nextNextInd);
						polyPoint(pointVect.elements^^[curVertexInd]).next := nextNextInd;
						polyPoint(pointVect.elements^^[nextNextInd]).prev := curVertexInd;
						curVertexInd := startVertexInd;			(* avoids recursive procedure *)
						numberVertices := numberVertices - 1;
					end
					else begin
						curVertexInd := polyPoint(pointVect.elements^^[curVertexInd]).next;
					end;
				end;
			end;

		begin		(* triangulate *)
			curVertexInd := first;
			curVertex := polyPoint(pointVect.elements^^[curVertexInd]);
			isUnimonotone := true;
			firstTime := true;
			stop := false;
			while ((curVertexInd <> first) | firstTime) & not (stop) do begin
				firstTime := false;
				botVertexInd := diagonal(curVertexInd);
				if botVertexInd <> 0 then begin
					isUnimonotone := false;
					botVertex := polyPoint(pointVect.elements^^[botVertexInd]);
{    botVertex.vertex.drawLine(curVertex.vertex);}
					saveNext := curVertex.next;
					savePrev := botVertex.prev;
					curVertex.next := botVertexInd;
					botVertex.prev := curVertexInd;
					triangulate(botVertexInd);
					curVertex.next := saveNext;
					botVertex.prev := savePrev;
					botVertex.next := curVertexInd;
					curVertex.prev := botVertexInd;
					triangulate(curVertexInd);
					stop := true;
				end
				else begin
					curVertexInd := curVertex.next;
					curVertex := polyPoint(pointVect.elements^^[curVertexInd]);
				end;
			end;
			if isUnimonotone then
				triangulateMonoton(first);
		end;

	begin
		if not (fromLayer) then
			localStartInPolInd := last + 1;
		bufGraphEdge := nil;
		new(toSearch);			(* global search structure *)
		failNil(Handle(toSearch));
		toSearch.init;
		new(edgesTree);		(* stores the already allocated edges, if a new triangle is appended *)
		failNil(Handle(edgesTree));
		edgesTree.init;
		triangulate(first);
		edgesTree.free;
		toSearch.free;
		if bufGraphEdge <> nil then
			bufGraphEdge.free;
	end;

function determineType (pointVect: homogenVector; ver: polyPoint; var horizontal: boolean): integer;
	var
		pr, ne: polyPoint;
		y: extended;
		diffNext, diffPrev: signType;
	begin
		horizontal := false;
		pr := polyPoint(pointVect.elements^^[ver.prev]);
		ne := polyPoint(pointVect.elements^^[ver.next]);
		y := ver.vertex.getY;
		diffNext := sign(y - ne.vertex.getY);
		diffPrev := sign(y - pr.vertex.getY);
		if diffNext * diffPrev > 0 then
			if diffNext > 0 then
				determineType := 2
			else
				determineType := 3
		else begin
			if diffNext = 0 then begin
				horizontal := true;
				if diffPrev = sign(y - polyPoint(pointVect.elements^^[ne.next]).vertex.getY) then
					if diffPrev = 1 then
						determineType := 2
					else
						determineType := 3
				else
					determineType := 1;
			end
			else begin
				if diffPrev = 0 then begin
					horizontal := true;
					if diffNext = sign(y - polyPoint(pointVect.elements^^[pr.prev]).vertex.getY) then
						if diffNext = 1 then
							determineType := 2
						else
							determineType := 3
					else
						determineType := 1;
				end
				else
					determineType := 1;
			end;
		end;
	end;

procedure insertDividingEdges (requiredEdges, numberOfAllVertices: integer; innerPolInd: homogenVector; var pointVect: homogenVector; var numberOfVertices: integer);
	var
		curVertex: polyPoint;
		newEdges, outerPolVertices, polNr: integer;
		horizontal: boolean;
		toSearch: int;
		where, i: seqIndex0;
		inserted: boolean;

	procedure secondTrapezPoints (toSearch: seqIndex; var oneVer, twoVer: seqIndex0);
		var
			i: seqIndex;
		begin
			oneVer := 0;
			twoVer := 0;
			for i := 1 to numberOfVertices do begin
				if polyPoint(pointVect.elements^^[i]).trapez1Point2 = toSearch then
					if oneVer <> 0 then begin
						twoVer := i;
						exit(secondTrapezPoints);
					end
					else
						oneVer := i
				else if polyPoint(pointVect.elements^^[i]).trapez2Point2 = toSearch then begin
					if oneVer <> 0 then begin
						twoVer := i;
						exit(secondTrapezPoints);
					end
					else
						oneVer := i;
				end
			end;
		end;

	procedure replaceBotTrapPoint (toReplace, toReplaceWith: seqIndex);
		var
			i: seqIndex;
		begin
			for i := 1 to numberOfVertices do begin
				if polyPoint(pointVect.elements^^[i]).trapez1Point2 = toReplace then begin
					polyPoint(pointVect.elements^^[i]).trapez1Point2 := toReplaceWith;
					exit(replaceBotTrapPoint);
				end;
				if polyPoint(pointVect.elements^^[i]).trapez2Point2 = toReplace then begin
					polyPoint(pointVect.elements^^[i]).trapez2Point2 := toReplaceWith;
					exit(replaceBotTrapPoint);
				end
			end;
		end;

	procedure replaceBotTrapPoint2 (toReplace, toReplaceWith, firstEdgeP, secondEdgeP: seqIndex; isLeftEdge: boolean);
		var
			i: seqIndex;
		begin
			for i := 1 to numberOfVertices do begin
				with polyPoint(pointVect.elements^^[i]) do begin
					if polyPoint(pointVect.elements^^[i]).trapez1Point2 = toReplace then begin
						if (trapez1Point2 <> 0) & ((isLeftEdge & (edge1L.startVertex = firstEdgeP) & (edge1L.endVertex = secondEdgeP)) | (not (isLeftEdge) & (edge1R.startVertex = firstEdgeP) & (edge1R.endVertex = secondEdgeP))) then begin
							polyPoint(pointVect.elements^^[i]).trapez1Point2 := toReplaceWith;
							exit(replaceBotTrapPoint2);
						end;
						if (trapez2Point2 <> 0) & ((isLeftEdge & (edge2L.startVertex = firstEdgeP) & (edge2L.endVertex = secondEdgeP)) | (not (isLeftEdge) & (edge2R.startVertex = firstEdgeP) & (edge2R.endVertex = secondEdgeP))) then begin
							polyPoint(pointVect.elements^^[i]).trapez2Point2 := toReplaceWith;
							exit(replaceBotTrapPoint2);
						end;
					end;	(* endif *)
				end;	(* with *)
			end;	(* for *)
		end;

	procedure searchAndReplaceEdge (firstInd, secondInd, newfirstInd: seqIndex);
		var
			i: seqIndex;
		begin
			for i := 1 to pointVect.length do begin
				with polyPoint(pointVect.elements^^[i]) do begin
					if trapez1Point2 <> 0 then begin
						if ((firstInd = edge1L.startVertex) & (secondInd = edge1L.endVertex)) then
							edge1L.startVertex := newfirstInd
						else if ((firstInd = edge1R.startVertex) & (secondInd = edge1R.endVertex)) then
							edge1R.startVertex := newfirstInd
						else if ((secondInd = edge1R.startVertex) & (firstInd = edge1R.endVertex)) then
							edge1R.endVertex := newfirstInd
						else if ((secondInd = edge1L.startVertex) & (firstInd = edge1L.endVertex)) then
							edge1L.endVertex := newfirstInd;
					end;
					if trapez2Point2 <> 0 then begin
						if ((firstInd = edge2L.startVertex) & (secondInd = edge2L.endVertex)) then
							edge2L.startVertex := newfirstInd
						else if ((firstInd = edge2R.startVertex) & (secondInd = edge2R.endVertex)) then
							edge2R.startVertex := newfirstInd
						else if ((secondInd = edge2R.startVertex) & (firstInd = edge2R.endVertex)) then
							edge2R.endVertex := newfirstInd
						else if ((secondInd = edge2L.startVertex) & (firstInd = edge2L.endVertex)) then
							edge2L.endVertex := newfirstInd;
					end;
				end;
			end;
		end;

	procedure insertNewEdge (curVertexInd: seqIndex; oneOrTwo: seqIndex; var numberOfVertices: integer);
		var
			curVertex, twoVertex, newP1, newP2: polyPoint;
			newP1Ind, newP2Ind, twoVertexInd, endOfEdgeR: seqIndex;
			twoType, curType: integer;
		begin
			curVertex := polyPoint(pointVect.elements^^[curVertexInd]);
			if oneOrTwo = 1 then begin
				twoVertexInd := curVertex.trapez1Point2;
				curVertex.trapez1Point2 := 0;
				endOfEdgeR := curVertex.edge1R.endVertex;
			end
			else begin
				twoVertexInd := curVertex.trapez2Point2;
				curVertex.trapez2Point2 := 0;
				endOfEdgeR := curVertex.edge2R.endVertex;
			end;
			twoVertex := polyPoint(pointVect.elements^^[twoVertexInd]);	(* second trapezoid point *)
{    penSize(3, 3);}
{    twoVertex.vertex.drawLine(curVertex.vertex);}
{    penSize(1, 1);}
			new(newP1);
			failNil(Handle(newP1));
			new(newP2);
			failNil(Handle(newP2));
			newP1.init;
			newP2.init;
			numberOfVertices := numberOfVertices + 2;
			pointVect.appendSafe(newP1, 2);
			newP1Ind := numberOfVertices - 1;
			pointVect.appendSafe(newP2, 1);
			newP2Ind := numberOfVertices;
			newP1.trapez1Point2 := 0;
			newP1.trapez2Point2 := 0;
			curType := determineType(pointVect, curVertex, horizontal);
			twoType := determineType(pointVect, twoVertex, horizontal);
			if (((endOfEdgeR = twoVertexInd) & (twoType <> 2)) | (twoType = 2) & (curVertex.vertex.getX < twoVertex.vertex.getX)) then begin
(* if curVertex.vertex.getX < twoVertex.vertex.getx then begin *)
		(* if hole on the right *)
				searchAndReplaceEdge(twoVertexInd, twoVertex.next, newP2Ind);
				searchAndReplaceEdge(curVertexInd, curVertex.prev, newP1Ind);
				if curType = 3 then
					replaceBotTrapPoint2(curVertexInd, newP1Ind, curVertex.prev, newP1Ind, false);
				with newP1 do begin
					prev := curVertex.prev;
					next := newP2Ind;
					vertex := curVertex.vertex;
					duplInd := curVertexInd;
				end;
				polyPoint(pointVect.elements^^[curVertex.prev]).next := newP1Ind;
				with newP2 do begin
					prev := newP1Ind;
					next := twoVertex.next;
					trapez1Point2 := twoVertex.trapez1Point2;
					edge1L := twoVertex.edge1L;
					edge1R := twoVertex.edge1R;
					if twoType = 2 then begin
						twoVertex.trapez1Point2 := twoVertex.trapez2Point2;
						twoVertex.edge1L := twoVertex.edge2L;
						twoVertex.edge1R := twoVertex.edge2R;
					end
					else
						twoVertex.trapez1Point2 := 0;
					vertex := twoVertex.vertex;
					duplInd := twoVertexInd;
				end;
				polyPoint(pointVect.elements^^[twoVertex.next]).prev := newP2Ind;
				twoVertex.next := curVertexInd;
				curVertex.prev := twoVertexInd;
			end
			else begin															(* if hole on the left *)
				searchAndReplaceEdge(twoVertexInd, twoVertex.prev, newP2Ind);
				searchAndReplaceEdge(curVertexInd, curVertex.next, newP1Ind);
				if curType = 3 then
					replaceBotTrapPoint2(curVertexInd, newP1Ind, curVertex.next, newP1Ind, true);
				with newP2 do begin
					next := newP1Ind;
					prev := twoVertex.prev;
					if twoType = 2 then begin
						trapez1Point2 := twoVertex.trapez2Point2;
						edge1L := twoVertex.edge2L;
						edge1R := twoVertex.edge2R;
					end
					else begin
						trapez1Point2 := twoVertex.trapez1Point2;
						edge1L := twoVertex.edge1L;
						edge1R := twoVertex.edge1R;
						twoVertex.trapez1Point2 := 0;
					end;
					vertex := twoVertex.vertex;
					duplInd := twoVertexInd;
				end;
				polyPoint(pointVect.elements^^[twoVertex.prev]).next := newP2Ind;
				with newP1 do begin
					next := curVertex.next;
					prev := newP2Ind;
					vertex := curVertex.vertex;
					duplInd := curVertexInd;
				end;
				polyPoint(pointVect.elements^^[curVertex.next]).prev := newP1Ind;
				curVertex.next := twoVertexInd;
				twoVertex.prev := curVertexInd;
			end;
			if (oneOrTwo = 1) & (curVertex.trapez2Point2 <> 0) then
				with curVertex do begin
					trapez1Point2 := trapez2Point2;
					edge1L := edge2L;
					edge1R := edge2R;
				end;
			curVertex.trapez2Point2 := 0;
			twoVertex.trapez2Point2 := 0;
			replaceBotTrapPoint(twoVertexInd, newP2Ind);
		end;

	function checkVertex (alreadyConnected: boolean): boolean;
		begin
			checkVertex := false;
			if curVertex.trapez1Point2 <> 0 then begin
				toSearch.value := curVertex.trapez1Point2 + numberOfAllVertices;
				where := innerPolInd.binarySearch(toSearch, compareIndexes, 1, innerPolInd.length);
				if (where <> 0) & ((not (triangPoly(innerPolInd.elements^^[where]).done) & alreadyConnected) | (triangPoly(innerPolInd.elements^^[where]).done & not (alreadyConnected))) then begin
					insertNewEdge(i, 1, numberOfVertices);
					newEdges := newEdges + 1;
					triangPoly(innerPolInd.elements^^[where]).done := true;
					checkVertex := true;
					exit(checkVertex);
				end;
			end;
			if curVertex.trapez2Point2 <> 0 then begin
				toSearch.value := curVertex.trapez2Point2 + numberOfAllVertices;
				where := innerPolInd.binarySearch(toSearch, compareIndexes, 1, innerPolInd.length);
				if (where <> 0) & ((not (triangPoly(innerPolInd.elements^^[where]).done) & alreadyConnected) | (triangPoly(innerPolInd.elements^^[where]).done & not (alreadyConnected))) then begin
					insertNewEdge(i, 2, numberOfVertices);
					newEdges := newEdges + 1;
					triangPoly(innerPolInd.elements^^[where]).done := true;
					checkVertex := true;
				end;
			end;
		end;

	begin		(* insertDividingEdges *)
		outerPolVertices := triangPoly(innerPolInd.elements^^[1]).startInd - 1;
		newEdges := 0;
		i := 1;
		new(toSearch);
		failNil(Handle(toSearch));
		toSearch.init;
		while (newEdges < requiredEdges) & (i <= outerPolVertices) do begin
			curVertex := polyPoint(pointVect.elements^^[i]);
			inserted := checkVertex(true);
			i := i + 1;
		end;

		polNr := 1;
		while newEdges < requiredEdges do begin	(* need to connect hole polygons *)
			i := triangPoly(innerPolInd.elements^^[polNr]).startInd - numberOfAllVertices;
			inserted := false;
			while (newEdges < requiredEdges) & (not (inserted)) & ((i + numberOfAllVertices) <= triangPoly(innerPolInd.elements^^[polNr]).endInd) do begin
				curVertex := polyPoint(pointVect.elements^^[i]);
				inserted := checkVertex(triangPoly(innerPolInd.elements^^[polNr]).done);
				if inserted then begin
					triangPoly(innerPolInd.elements^^[polNr]).done := true;
					polNr := 0;
				end;
				i := i + 1;
			end;
			polNr := polNr + 1;
		end;
		toSearch.free;
	end;		(* insertDividingEdges *)

procedure prepareInput (var outerPolygons, innerPolygons: vector);
	var
		i: seqIndex;
	begin
		case analyseInput of
			1:  begin
				outerPolygons.appendSafe(polygon2d(list(currentArguments).ith(1)).duplicateAll, 1);
			end;
			2:  begin
				for i := 1 to homogenVector(list(currentArguments).ith(1)).length do
					outerPolygons.appendSafe(vector(list(currentArguments).ith(1)).elements^^[i].duplicateAll, 1);
			end;
			3:  begin
				outerPolygons.appendSafe(polygon2d(list(currentArguments).ith(1)).duplicateAll, 1);
				innerPolygons.appendSafe(polygon2d(list(currentArguments).ith(2)).duplicateAll, 1)
			end;
			4:  begin
				outerPolygons.appendSafe(polygon2d(list(currentArguments).ith(1)).duplicateAll, 1);
				for i := 1 to homogenVector(list(currentArguments).ith(2)).length do
					innerPolygons.appendSafe(vector(list(currentArguments).ith(2)).elements^^[i].duplicateAll, 1);
			end;
			5:  begin
				for i := 1 to homogenVector(list(currentArguments).ith(1)).length do
					outerPolygons.appendSafe(vector(list(currentArguments).ith(1)).elements^^[i].duplicateAll, 1);
				innerPolygons.appendSafe(polygon2d(list(currentArguments).ith(2)).duplicateAll, 1);
			end;
			6:  begin
				for i := 1 to homogenVector(list(currentArguments).ith(1)).length do
					outerPolygons.appendSafe(vector(list(currentArguments).ith(1)).elements^^[i].duplicateAll, 1);
				for i := 1 to homogenVector(list(currentArguments).ith(2)).length do
					innerPolygons.appendSafe(vector(list(currentArguments).ith(2)).elements^^[i].duplicateAll, 1);
			end;
		end;
	end;



function getTotalVertices (outerPolygons, innerPolygons: vector): longInt;
	var
		i: seqIndex0;
		s: longInt;
	begin
		s := 0;
		with outerPolygons do
			for i := 1 to length do
				s := s + polygon2d(elements^^[i]).length;
		with innerPolygons do
			for i := 1 to length do
				s := s + polygon2d(elements^^[i]).length;
		getTotalVertices := s;
	end;

procedure doWholeTriangulation (var t: triangulation2d; fromLayer: boolean; var globalShift: integer;								(* the global shift of the already allocated vertices *)
								var outerPolGlobalIndexes: homogenVector; fromLastLayerEdgeTree: avlTree; var newLayerEdgeTree: avlTree;					(* stores the edges for the next layer *)
								var outerPolygons, innerPolygons: vector; function newTriangle: triangle);
	var
		highYInd: seqIndex;
		i, j: seqIndex0;
		globalStartInPolInd: integer;		(* stores the global index of the first inner vertex *)
		localStartInPolInd: integer;		(* stores the local index of the first inner vertex *)
		shiftOfInnerVertices: integer;	(* counts the vertices of the outerPolygons which were already allocated in the triangulation3d structure *)
		shiftOfOuterVertices: integer;
		numberOfPolVertices: integer;	(* counts the vertices of one polygon with the corresponding holes, already allocated vertices are also counted *)
		numberOfAllVertices: integer;	(* is the sum of all numberOfPolVertices, tm. the sum of all vertices of outerPolygons and innerPolygons *)
													(* is updated after the triangulation of one polygon *)
		innerPolInd: homogenVector;		(* corresponding to one outer polygon, the beginning indexes of the inner polygons are stored *)
		innerPolGlobalIndexes: homogenVector;	(* stores the global indexes of the inner polygons of one layer, the length of this vector *)
																(* is always the same as the number of innerPolygons in the layer *)
		sortedInnerPolGlobalIndexes: homogenVector;
		currentOuterPolygonNr: seqIndex;		(* stores the index of the outerPolygon in the outerPol vector *)
		tPoly: triangPoly;
		tEdge: graphEdge;
		dummy: boolean;
		pointVect, sortedVect: homogenVector;
		requiredNewEdges: integer;
		totalNumberOfVertices: longInt;

	function isLeft (mid, left, right: point2d): boolean;
		var
			x, y, factor, divisor, dividend: extended;
		begin
			x := mid.getX;
			y := mid.getY;
			dividend := left.getY - y;
			divisor := right.getY - y;
			if (divisor <> 0) then begin
				factor := dividend / divisor;
				if (left.getX - x) < (factor * (right.getX - x)) then
					isLeft := true
				else
					isLeft := false;
			end
			else begin
				if divisor = 0 then
					if right.getX < x then
						isLeft := false
					else
						isLeft := true
				else if left.getX < x then
					isLeft := true
				else
					isLeft := false
			end;
		end;

	procedure doOneTriangulation (shift: seqIndex0);
		var
			polyVertex: polyPoint;
			activeLTrapez, activeRTrapez: avlTree;
			curTrapez, curTrapez2, searchTrapez, searchTrap2: trapezoid;	(* searchTrap2 is used in delete *)
			curEdge, curEdge2: graphEdge;
			where, where2: reference;
			direction, dir: directionType;
			upperInd, lowerInd, leftInd, rightInd, realInd, i: seqIndex0;
			found, found1L, horizontal: boolean;
			numberOfVertices: integer;
			trapezEdges: list;

		function comparePolyPoints (a, b: int): boolean;
			begin
				comparePolyPoints := false;
				if polyPoint(pointVect.elements^^[a.value]).vertex.getY < polyPoint(pointVect.elements^^[b.value]).vertex.getY then
					comparePolyPoints := true
				else if (polyPoint(pointVect.elements^^[a.value]).vertex.getY = polyPoint(pointVect.elements^^[b.value]).vertex.getY) & (polyPoint(pointVect.elements^^[a.value]).vertex.getX > polyPoint(pointVect.elements^^[b.value]).vertex.getX) then
					comparePolyPoints := true
			end;

		function compTra (a1, a2, b1, b2: point2d): relationObj;
			var
				side1, side2: signType;
			begin
				side1 := a1.whichSideSign(b1, b2);
				side2 := a2.whichSideSign(b1, b2);
				if (side1 <> 1) & (side2 <> 1) then
					compTra := lessThanObj
				else begin
					if (side1 <> -1) & (side2 <> -1) then
						compTra := greaterThanObj
					else begin
						side1 := b1.whichSideSign(a1, a2);
						side2 := b2.whichSideSign(a1, a2);
						if (side1 <> 1) & (side2 <> 1) then
							compTra := greaterThanObj
						else begin
							if (side1 <> -1) & (side2 <> -1) then
								compTra := lessThanObj
							else begin
								fatalError('Compare Error');
								compTra := lessThanObj;
							end;
						end;
					end;
				end;
			end;

		function compareLTrapez (a, b: trapezoid): relationObj;
			var
				a1, a2, b1, b2: point2d;
			begin
				if (a.edgeL.startVertex = b.edgeL.startVertex) & (a.edgeL.endVertex = b.edgeL.endVertex) then
					compareLTrapez := equalObj
				else begin
					a1 := polyPoint(pointVect.elements^^[a.edgeL.startVertex]).vertex;
					a2 := polyPoint(pointVect.elements^^[a.edgeL.endVertex]).vertex;
					b1 := polyPoint(pointVect.elements^^[b.edgeL.startVertex]).vertex;
					b2 := polyPoint(pointVect.elements^^[b.edgeL.endVertex]).vertex;
					compareLTrapez := compTra(a1, a2, b1, b2);
				end;
			end;

		function compareRTrapez (a, b: trapezoid): relationObj;
			var
				a1, a2, b1, b2: point2d;
			begin
				if (a.edgeR.startVertex = b.edgeR.startVertex) & (a.edgeR.endVertex = b.edgeR.endVertex) then
					compareRTrapez := equalObj
				else begin
					a1 := polyPoint(pointVect.elements^^[a.edgeR.startVertex]).vertex;
					a2 := polyPoint(pointVect.elements^^[a.edgeR.endVertex]).vertex;
					b1 := polyPoint(pointVect.elements^^[b.edgeR.startVertex]).vertex;
					b2 := polyPoint(pointVect.elements^^[b.edgeR.endVertex]).vertex;
					compareRTrapez := compTra(a1, a2, b1, b2);
				end;
			end;

		procedure upperLowerPoint (ver: polyPoint; var upperInd, lowerInd: seqIndex0);
			var
				p: polyPoint;
				y: extended;
			begin
				y := ver.vertex.getY;
				p := polyPoint(pointVect.elements^^[ver.prev]);
				if p.vertex.getY > y then begin
					upperInd := ver.prev;
					lowerInd := ver.next;
				end
				else begin
					if p.vertex.getY = y then begin
						if polyPoint(pointVect.elements^^[ver.next]).vertex.getY > y then begin
							upperInd := ver.next;
							lowerInd := ver.prev;
						end
						else begin
							upperInd := ver.prev;
							lowerInd := ver.next;
						end;
					end
					else begin
						upperInd := ver.next;
						lowerInd := ver.prev;
					end;
				end;
			end;

		procedure leftRightPoint (ver: polyPoint; point1, point2: seqIndex0; var leftInd, rightInd: seqIndex0);
			begin
				if isLeft(ver.vertex, polyPoint(pointVect.elements^^[point1]).vertex, polyPoint(pointVect.elements^^[point2]).vertex) then begin
					leftInd := point1;
					rightInd := point2;
				end
				else begin
					leftInd := point2;
					rightInd := point1;
				end;
			end;

		procedure leftRightVertex (ver: polyPoint; var leftInd, rightInd: seqIndex0);
			begin
				leftRightPoint(ver, ver.prev, ver.next, leftInd, rightInd);
			end;

		function withinTrapez (i: seqIndex0; trap: trapezoid): boolean;
			var
				p: point2d;
			begin
				p := polyPoint(pointVect.elements^^[i]).vertex;
				withinTrapez := (p.whichSideSign(polyPoint(pointVect.elements^^[trap.edgeL.startVertex]).vertex, polyPoint(pointVect.elements^^[trap.edgeL.endVertex]).vertex) <> p.whichSideSign(polyPoint(pointVect.elements^^[trap.edgeR.startVertex]).vertex, polyPoint(pointVect.elements^^[trap.edgeR.endVertex]).vertex));
			end;

		procedure searchEdge (upperInd, lowerInd: seqIndex0; var where: reference; var direc: directionType);
			var
				found: boolean;
				dummy: directionType;
			begin
				searchTrapez.edgeL.startVertex := upperInd;
				searchTrapez.edgeL.endVertex := lowerInd;
				found := activeLTrapez.find(searchTrapez, compareLTrapez, where, dummy);
				if not (found) then begin
					searchTrapez.edgeR.startVertex := upperInd;
					searchTrapez.edgeR.endVertex := lowerInd;
					found := activeRTrapez.find(searchTrapez, compareRTrapez, where, dummy);
					if not (found) then
						where := nil
					else
						direc := right;
				end
				else
					direc := left;
			end;

		procedure completeTrapez (p1, p2: seqIndex0; eL, eR: graphEdge);
			var
				point: polyPoint;
				leftInd, rightInd: seqIndex0;
			begin
				point := polyPoint(pointVect.elements^^[p1]);
				if point.trapez1Point2 <> 0 then begin
					leftRightPoint(point, point.trapez1Point2, p2, leftInd, rightInd);
					if leftInd = p2 then begin	(* the leftmost trapezoid must be the first *)
						with point do begin
							trapez2Point2 := trapez1Point2;
							edge2L := edge1L;
							edge2R := edge1R;
							trapez1Point2 := p2;
							edge1L := eL;
							edge1R := eR;
						end;
					end
					else begin
						with point do begin
							trapez2Point2 := p2;
							edge2L := eL;
							edge2R := eR;
						end;
					end;
				end
				else begin
					with point do begin
						trapez1Point2 := p2;
						edge1L := eL;
						edge1R := eR;
					end;
				end;
			end;

		procedure insertTrapez (p1: seqIndex; leftEdge, rightEdge: graphEdge);
		(* inserts a trapez into the avlTree of active trapezs *)
			var
				newTrapez: trapezoid;
			begin
				new(newTrapez);
				failNil(Handle(newTrapez));
				newTrapez.init;
				with newTrapez do begin
					vertex1 := p1;
					vertex2 := 0;
					edgeL := leftEdge;
					edgeR := rightEdge;
				end;
				activeLTrapez.insertNewObj(newTrapez, compareLTrapez);
				activeRTrapez.insertNewObj(newTrapez, compareRTrapez);
			end;

		procedure deleteTrapez (where: reference; dir: directionType; otherEdge: graphEdge);
			var
				where2: reference;
				found: boolean;
				dummy: directionType;
			begin
				if dir = left then begin
					searchTrap2.edgeR := otherEdge;
					found := activeRTrapez.find(searchTrap2, compareRTrapez, where2, dummy);
					activeLTrapez.delete(where);
					activeRTrapez.delete(where2);
				end
				else begin
					searchTrap2.edgeL := otherEdge;
					found := activeLTrapez.find(searchTrap2, compareLTrapez, where2, dummy);
					activeLTrapez.delete(where2);
					activeRTrapez.delete(where);
				end;
			end;

		begin	(* doOneTriangulation *)
(* now the effective algorithm starts *)
			new(trapezEdges);
			failNil(Handle(trapezEdges));
			trapezEdges.init;
			new(activeLTrapez);
			failNil(Handle(activeLTrapez));
			activeLTrapez.init;
			new(activeRTrapez);
			failNil(Handle(activeRTrapez));
			activeRTrapez.init;
			new(searchTrapez);	(* only for searching, may not be deallocated *)
			failNil(Handle(searchTrapez));
			searchTrapez.init;
			new(searchTrapez.edgeL);
			failNil(Handle(searchTrapez.edgeL));
			new(searchTrapez.edgeR);
			failNil(Handle(searchTrapez.edgeR));
			searchTrapez.edgeL.init;
			searchTrapez.edgeR.init;
			new(searchTrap2);		(* is used in deleteTrapez, may not be deallocated *)
			failNil(Handle(searchTrap2));
			searchTrap2.init;
			sortedVect.sortAll(comparePolyPoints);
			numberOfVertices := pointVect.length;
			leftRightVertex(polyPoint(pointVect.elements^^[int(sortedVect.elements^^[numberOfVertices]).value]), leftInd, rightInd);
			if rightInd = polyPoint(pointVect.elements^^[int(sortedVect.elements^^[numberOfVertices]).value]).prev then
				fatalError('Order Error');		(* the polygon has to be in clockwise order, in the window however it must be counterclock *)
			horizontal := false;
			for i := numberOfVertices downto 1 do begin
				if not (horizontal) then begin
					realInd := int(sortedVect.elements^^[i]).value;
					polyVertex := polyPoint(pointVect.elements^^[realInd]);
{$IFC AnimationEnabled }
					advanceStopWatchHands;
{$ENDC }
					case determineType(pointVect, polyVertex, horizontal) of
						1:  begin
							upperLowerPoint(polyVertex, upperInd, lowerInd);
							searchEdge(upperInd, realInd, where, dir);
							if where = nil then begin
								if polyVertex.next = upperInd then begin
									searchEdge(polyPoint(pointVect.elements^^[upperInd]).next, upperInd, where, dir);
									curTrapez := trapezoid(activeLTrapez.getObject(where));
									completeTrapez(curTrapez.vertex1, upperInd, curTrapez.edgeL, curTrapez.edgeR);
								end
								else begin
									searchEdge(polyPoint(pointVect.elements^^[upperInd]).prev, upperInd, where, dir);
									curTrapez := trapezoid(activeLTrapez.getObject(where));
									completeTrapez(curTrapez.vertex1, realInd, curTrapez.edgeL, curTrapez.edgeR);
								end;
							end
							else begin
								curTrapez := trapezoid(activeLTrapez.getObject(where));
								completeTrapez(curTrapez.vertex1, realInd, curTrapez.edgeL, curTrapez.edgeR);
							end;

							if dir = left then
								deleteTrapez(where, left, curTrapez.edgeR)
							else
								deleteTrapez(where, right, curTrapez.edgeL);

							new(curEdge);
							failNil(Handle(curEdge));
							curEdge.init;
							trapezEdges.append(curEdge);
							if not (horizontal) then begin
								curEdge.startVertex := realInd;
								curEdge.endVertex := lowerInd;
								if dir = left then
									insertTrapez(realInd, curEdge, curTrapez.edgeR)
								else
									insertTrapez(realInd, curTrapez.edgeL, curEdge);
							end
							else begin
								leftRightVertex(polyVertex, leftInd, rightInd);
								if ((dir = right) & (lowerInd = rightInd)) then begin
									curEdge.startVertex := lowerInd;
									curEdge.endVertex := polyPoint(pointVect.elements^^[lowerInd]).next;
									insertTrapez(realInd, curTrapez.edgeL, curEdge);
								end
								else if dir = right then begin
									curEdge.startVertex := realInd;
									curEdge.endVertex := leftInd;
									insertTrapez(realInd, curTrapez.edgeL, curEdge);
								end
								else if ((dir = left) & (lowerInd = rightInd)) then begin
									curEdge.startVertex := lowerInd;
									curEdge.endVertex := polyPoint(pointVect.elements^^[lowerInd]).prev;
									insertTrapez(rightInd, curEdge, curTrapez.edgeR);
								end
								else if dir = left then begin
									curEdge.startVertex := realInd;
									curEdge.endVertex := polyVertex.prev;
									insertTrapez(rightInd, curEdge, curTrapez.edgeR);
								end;
							end;
							curTrapez.free;
						end;
						2:  begin
							searchTrapez.edgeL.startVertex := realInd;
							searchTrapez.edgeL.endVertex := realInd;
							found := activeLTrapez.find(searchTrapez, compareLTrapez, where, direction);
							if where <> nil then begin
								if direction = right then begin
									where2 := activeLTrapez.next(where, direction);
									where := where2;
								end;
								if where <> nil then begin
									curTrapez := trapezoid(activeLTrapez.getObject(where));
									if withinTrapez(realInd, curTrapez) then begin

										completeTrapez(curTrapez.vertex1, realInd, curTrapez.edgeL, curTrapez.edgeR);

										deleteTrapez(where, left, curTrapez.edgeR);
										new(curEdge);
										failNil(Handle(curEdge));
										curEdge.init;
										trapezEdges.append(curEdge);
										curEdge.startVertex := realInd;
										if not (horizontal) then begin
											leftRightVertex(polyVertex, leftInd, rightInd);
											curEdge.endVertex := leftInd;
											insertTrapez(realInd, curTrapez.edgeL, curEdge);

											new(curEdge);
											failNil(Handle(curEdge));
											curEdge.init;
											curEdge.startVertex := realInd;
											curEdge.endVertex := rightInd;
											insertTrapez(realInd, curEdge, curTrapez.edgeR);
										end
										else begin	(* horizontal edge *)
											curEdge.endVertex := polyVertex.next;
											insertTrapez(realInd, curTrapez.edgeL, curEdge);
											new(curEdge);
											failNil(Handle(curEdge));
											curEdge.init;
											curEdge.startVertex := polyVertex.prev;
											curEdge.endVertex := polyPoint(pointVect.elements^^[polyVertex.prev]).prev;
											insertTrapez(polyVertex.prev, curEdge, curTrapez.edgeR);
										end;
										trapezEdges.append(curEdge);
										curTrapez.free;
									end
									else
										where := nil;
								end;
							end;
							if where = nil then begin
								new(curEdge);
								failNil(Handle(curEdge));
								curEdge.init;
								trapezEdges.append(curEdge);
								new(curEdge2);
								failNil(Handle(curEdge2));
								curEdge2.init;
								trapezEdges.append(curEdge2);
								curEdge.startVertex := realInd;
								if not (horizontal) then begin
									leftRightVertex(polyVertex, leftInd, rightInd);
									curEdge.endVertex := leftInd;
									curEdge2.startVertex := realInd;
									curEdge2.endVertex := rightInd;
									insertTrapez(realInd, curEdge, curEdge2);
								end
								else begin
									curEdge.endVertex := polyVertex.prev;
									curEdge2.startVertex := polyVertex.next;
									curEdge2.endVertex := polyPoint(pointVect.elements^^[polyVertex.next]).next;
									insertTrapez(realInd, curEdge, curEdge2);
								end;
							end;
						end;
						3:  begin
							leftRightVertex(polyVertex, leftInd, rightInd);
							searchTrapez.edgeL.startVertex := leftInd;
							searchTrapez.edgeL.endVertex := realInd;
							found1L := activeLTrapez.find(searchTrapez, compareLTrapez, where, direction);
							if found1L then
								curTrapez := trapezoid(activeLTrapez.getObject(where))
							else begin
								searchTrapez.edgeR.startVertex := leftInd;
								searchTrapez.edgeR.endVertex := realInd;
								found := activeRTrapez.find(searchTrapez, compareRTrapez, where, direction);
								if not (found1L) & not (found) then
									nonFatalError('Trapez Error')
								else begin
									curTrapez := trapezoid(activeRTrapez.getObject(where));
								end;
							end;

							if found1L then begin		(* t.m. the side which was found is on the left side of the trapezoid *)
								if not (horizontal) then begin
									searchTrapez.edgeR.startVertex := rightInd;
									searchTrapez.edgeR.endVertex := realInd;
								end
								else begin
									searchTrapez.edgeR.startVertex := polyPoint(pointVect.elements^^[rightInd]).prev;
									searchTrapez.edgeR.endVertex := rightInd;
								end;
								found := activeRTrapez.find(searchTrapez, compareRTrapez, where2, direction);
								curTrapez2 := trapezoid(activeRTrapez.getObject(where2));
							end
							else begin
								if not (horizontal) then begin
									searchTrapez.edgeL.startVertex := rightInd;
									searchTrapez.edgeL.endVertex := realInd;
								end
								else begin
									searchTrapez.edgeL.startVertex := polyPoint(pointVect.elements^^[rightInd]).next;
									searchTrapez.edgeL.endVertex := rightInd;
								end;
								found := activeLTrapez.find(searchTrapez, compareLTrapez, where2, direction);
								curTrapez2 := trapezoid(activeLTrapez.getObject(where2));
							end;

							if curTrapez = curTrapez2 then begin		(* edges belong to same trapezoid *)
								completeTrapez(curTrapez.vertex1, realInd, curTrapez.edgeL, curTrapez2.edgeR);
								if found1L then
									deleteTrapez(where, left, curTrapez.edgeR)
								else
									deleteTrapez(where, right, curTrapez.edgeL);
							end
							else begin
								if found1L then begin
									deleteTrapez(where, left, curTrapez.edgeR);
									deleteTrapez(where2, right, curTrapez2.edgeL);
								end
								else begin
									deleteTrapez(where, right, curTrapez.edgeL);
									deleteTrapez(where2, left, curTrapez2.edgeR);
								end;
								completeTrapez(curTrapez.vertex1, realInd, curTrapez.edgeL, curTrapez.edgeR);
								if not (horizontal) then
									completeTrapez(curTrapez2.vertex1, realInd, curTrapez2.edgeL, curTrapez2.edgeR)
								else
									completeTrapez(curTrapez2.vertex1, rightInd, curTrapez2.edgeL, curTrapez2.edgeR);
								if found1L then begin	(* t.m. the side of the first trapezoid which was found is the left side of the trapezoid *)
									insertTrapez(realInd, curTrapez.edgeR, curTrapez2.edgeL);
									nonFatalError('Trapez Error')
								end
								else if not (horizontal) then
									insertTrapez(realInd, curTrapez.edgeL, curTrapez2.edgeR)
								else
									insertTrapez(rightInd, curTrapez.edgeL, curTrapez2.edgeR);
								curTrapez2.free;
							end;
							curTrapez.free;
						end;
					end;
				end	(* end not horizontal *)
				else
					horizontal := false;
			end; (* endfor *)
			searchTrapez.edgeL.free;
			searchTrapez.edgeR.free;
			searchTrapez.free;
			searchTrap2.free;
			activeLTrapez.free;
			activeRTrapez.free;
			sortedVect.freeAll;

			if requiredNewEdges > 0 then
				insertDividingEdges(requiredNewEdges, numberOfAllVertices, innerPolInd, pointVect, numberOfVertices);

			triangulateFirst(1, numberOfVertices, shift, globalShift, outerPolGlobalIndexes, sortedInnerPolGlobalIndexes, fromLayer, currentOuterPolygonNr, fromLastLayerEdgeTree, globalStartInPolInd, localStartInPolInd, newLayerEdgeTree, pointVect, t, newTriangle);
			trapezEdges.freeAll;
			for i := 1 to pointVect.length do
				polyPoint(pointVect.elements^^[i]).free;
			pointVect.free;
			numberOfVertices := numberOfVertices - 2 * requiredNewEdges;
		end;	(* doOneTriangulation *)

	function searchHighestY (var pol: polygon2d): seqIndex;
		var
			highY: extended;
			i: seqIndex;
		begin
			highY := -10000000.0;
			for i := 1 to pol.length do begin
				if point2d(pol.elements^^[i]).getY > highY then begin
					highY := point2d(pol.elements^^[i]).getY;
					searchHighestY := i;
				end;
			end;
		end;

	procedure orderPolygon (po: polygon2d; si: signType);
		(* si = +1 clockwise otherwise counterclock *)
		var
			prev, next: seqIndex0;
		begin
			highYInd := searchHighestY(po);
			if highYInd <> 1 then
				prev := highYInd - 1
			else
				prev := po.length;
			if highYInd <> po.length then
				next := highYInd + 1
			else
				next := 1;
			if isLeft(point2d(po.elements^^[highYInd]), point2d(po.elements^^[prev]), point2d(po.elements^^[next])) then begin
				if si <> 1 then
					po.revert;
			end
			else begin
				if si = 1 then
					po.revert;
			end;
		end;

	procedure buildStructure (pol: polygon2d; shift: seqIndex0; insertInTriangStru: boolean);
		var
			i, last: seqIndex;
			index: int;
			polyVertex: polyPoint;

		procedure revertVect (ve: vector; from, ende: seqIndex0);
			begin
				while from < ende do begin
					ve.swap(from, ende);
					from := from + 1;
					ende := ende - 1;
				end;
			end;

		begin
			last := pol.length;
			new(index);
			failNil(Handle(index));
			index.value := 1 + shift;
			sortedVect.appendSafe(index, last);		(* for efficiency reason *)
			new(polyVertex);
			failNil(Handle(polyVertex));
			with polyVertex do begin
				init;
				prev := last + shift;
				next := shift + 2;
				vertex := point2d(pol.elements^^[1]);
				trapez1Point2 := 0;
				edge1L := nil;
				edge1R := nil;
				trapez2Point2 := 0;
				edge2L := nil;
				edge2R := nil;
				duplInd := 0;
			end;
			pointVect.appendSafe(polyVertex, last);	(* for efficiency reason *)
			if insertInTriangStru then
				t.points.appendSafe(polyVertex.vertex, last);	(* for efficiency reason *)

			for i := 2 to last do begin
				new(index);
				failNil(Handle(index));
				index.value := i + shift;
				sortedVect.appendSafe(index, last);
				new(polyVertex);
				failNil(Handle(polyVertex));
				with polyVertex do begin
					init;
					prev := i - 1 + shift;
					if i <> last then
						next := i + 1 + shift
					else
						next := 1 + shift;
					vertex := point2d(pol.elements^^[i]);
					trapez1Point2 := 0;
					edge1L := nil;
					edge1R := nil;
					trapez2Point2 := 0;
					edge2L := nil;
					edge2R := nil;
					duplInd := 0;
				end;
				pointVect.appendSafe(polyVertex, last);
				if insertInTriangStru then
					t.points.appendSafe(polyVertex.vertex, last);
			end;
		end;


	begin		(* doWholeTriangulation *)
		currentVectorLength := 1;
		if fromLayer then
			totalNumberOfVertices := 1
		else begin
			new(t);
			failNil(Handle(t));
			new(outerPolygons);
			failNil(Handle(outerPolygons));
			new(innerPolygons);
			failNil(Handle(innerPolygons));
			outerPolygons.init;
			innerPolygons.init;
			prepareInput(outerPolygons, innerPolygons);
			totalNumberOfVertices := getTotalVertices(outerPolygons, innerPolygons);
			t.initializeEuler(totalNumberOfVertices, innerPolygons.length);
			totalNumberOfVertices := totalNumberOfVertices + 2 * innerPolygons.length;
		end;
		if innerPolygons.length <> 0 then begin
			innerPolInd := newHomogenVector(innerPolygons.length);
			if fromLayer then begin
				new(innerPolGlobalIndexes);
				failNil(Handle(innerPolGlobalIndexes));
				innerPolGlobalIndexes.init;
				tEdge := nil;
				for i := 1 to innerPolygons.length do
					innerPolGlobalIndexes.append(tEdge);
			end;
		end;
		for i := 1 to outerPolygons.length do
			orderPolygon(polygon2d(outerPolygons.elements^^[i]), +1);
		for i := 1 to innerPolygons.length do
			orderPolygon(polygon2d(innerPolygons.elements^^[i]), -1);

		numberOfAllVertices := 0;
		shiftOfInnerVertices := 0;
		shiftOfOuterVertices := 0;
		for i := 1 to outerPolygons.length do begin
			pointVect := newHomogenVector(totalNumberOfVertices);
			sortedVect := newHomogenVector(totalNumberOfVertices);
			requiredNewEdges := 0;
			currentOuterPolygonNr := i;

			buildStructure(polygon2d(outerPolygons.elements^^[i]), 0, (globalShift = 0));

			numberOfPolVertices := polygon2d(outerPolygons.elements^^[i]).length;
			localStartInPolInd := numberOfPolVertices + 1;
			if globalShift <> 0 then
				shiftOfInnerVertices := shiftOfInnerVertices - numberOfPolVertices;

			globalStartInPolInd := 0;
			if fromLayer then
				sortedInnerPolGlobalIndexes := newHomogenVector(1);
			for j := 1 to innerPolygons.length do begin
				if polygon2d(outerPolygons.elements^^[i]).windingNumber(point2d(polygon2d(innerPolygons.elements^^[j]).elements^^[1]), dummy) <> 0 then begin
					new(tPoly);
					failNil(Handle(tPoly));
					tPoly.startInd := numberOfAllVertices + numberOfPolVertices + 1;
					buildStructure(polygon2d(innerPolygons.elements^^[j]), numberOfPolVertices, true);
					numberOfPolVertices := numberOfPolVertices + polygon2d(innerPolygons.elements^^[j]).length;
					tPoly.endInd := numberOfAllVertices + numberOfPolVertices;
					tPoly.done := false;
					innerPolInd.appendSafe(tPoly, 2);
					if fromLayer then begin
						new(tEdge);
						failNil(Handle(tEdge));
						tEdge.init;
						tEdge.startVertex := tPoly.startInd + globalShift + shiftOfInnerVertices;
						tEdge.endVertex := tPoly.endInd + globalShift + shiftOfInnerVertices;
						innerPolGlobalIndexes.elements^^[j] := tEdge;
						sortedInnerPolGlobalIndexes.appendSafe(tEdge, 2);
						if globalStartInPolInd = 0 then
							globalStartInPolInd := tEdge.startVertex;
					end;
					requiredNewEdges := requiredNewEdges + 1;
				end;
			end;
			doOneTriangulation(numberOfAllVertices);

			if fromLayer then
				sortedInnerPolGlobalIndexes.free;
			numberOfAllVertices := numberOfAllVertices + numberOfPolVertices;
		end;

		if innerPolygons.length <> 0 then
			innerPolInd.freeAll;
		if not (fromLayer) then begin
			for i := 1 to outerPolygons.length do
				polygon2d(outerPolygons.elements^^[i]).free;
			for i := 1 to innerPolygons.length do
				polygon2d(innerPolygons.elements^^[i]).free;
			outerPolygons.free;
			innerPolygons.free;
		end
		else begin
			if outerPolGlobalIndexes <> nil then
				outerPolGlobalIndexes.freeAll;
			outerPolGlobalIndexes := innerPolGlobalIndexes;
		end;
		globalShift := globalShift + numberOfAllVertices + shiftOfInnerVertices + shiftOfOuterVertices;
	end;	(* doWholeTriangulation *)


end.
