(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1991 by Philipp Meier						*)
(* Copyright 1992 by Michael Greminger					*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.0:  31-Jan-92								*)
(*														*)
(* Copyright 1992 by Bert Leukens						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.4:  July-29-92								*)

unit triangulation3d;

interface

uses
	QDPicture, point3d, layerVector, randomizedTriangulation, triangulation2d2;

type

	clipVertexKind = (noVertex, innerVertex, Xtop, tlVertex, Xleft, blVertex, Xbottom, brVertex, Xright, trVertex);
		  (* clipping a polygon on a rectangle results in a new polygon. Its vertices are of the kind defined above. *)

	clippingPolygon = record
			vertexKinds: longint;
		  (* binary coded  array [1..7] of clipVertexKind; Bits [......,  8..11,4..7, 0..3] *)
			ixy: array[1..6] of integer;
		  (* screen coordinates for QuickDraw procedures *)
		end;

	triangle3d = object(triangle)
			color: real;
		  (* used for shading: index to array of pattern in unit graphics *)
			polygon: clippingPolygon;
		  (*  used for quick interactive I/O *)

	(* Type computation and comparison *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function getInfo: obj;
			override;
			function memberType (o: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;

	(* Miscellaneous *)
			procedure drawTriangle;
	      (* PRECONDITION: the global variable currentTriangulation must be set. *)
			procedure getNormalVector (var normalVector: point3d);
		  (* calculate normal vector *)
	      (* PRECONDITION: the global variable currentTriangulation must be set. *)
			procedure calcColor;
		  (* calculate shading color *)
	      (* PRECONDITION: the global variable currentTriangulation must be set. *)
		end;


	triangulation3d = object(triangulation)
	(* Interactive I/O *)
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
	(* Description / execution *)
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
	(* Type computation and comparison *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
	(* Selecting and dragging *)
			function isSelected (where: obj): boolean;
			override;
	(* Miscellaneous *)
			function getBoundingObject: obj;
			override;
			function addNewEdge (startVertexNew, endVertexNew: seqIndex): graphEdgeTriangulation;
			override;
			procedure addNewTriangle (edgeANew, edgeBNew, edgeCNew: graphEdgeTriangulation);
			override;
			procedure randomFlip (n: longInt);
			override;

			procedure calcColor;
		  (* calculate shading color of triangles *)
			procedure initializeFromLayerVectorTrapez (v: layerVector);
		  (* Generate triangulation 3d from a vector of layers by trapez method *)
			procedure initializeFromLayerVectorPlaneSweep (v: polyLayerVector);
		  (* Generate triangulation 3d from a vector of layers by plane sweep method *)
		end;

	virtualPoint3d = object(realPoint3d)
	(* Type computation and comparison *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
		end;

	virtualGraphEdgeTriangulation = object(graphEdgeTriangulation)
	(* Type computation and comparison *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
		end;

	virtualTriangle3d = object(triangle3d)
	(* Type computation and comparison *)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
	(* Miscellaneous *)
			procedure drawTriangle;
			override;
			procedure calcColor;
			override;
		end;

procedure initTriangulation3d;
		  (* Generate global prototypes of triangle3d and triangulation3d. *)

var
light, eye, norm: realPoint3d;			(* used for temporary calculations; only allocated once *)

implementation

{$SETC ATriangulation = TRUE }

procedure triangulation3d.randomFlip (n: longInt);
	var
		i: longInt;
		x: graphEdgeTriangulation;
		pv: pointVector;
		hullLength, eliminated: seqIndex0;
{$IFC AnimationEnabled }
		abortEnableSave, convexHullAnimation: boolean;
{$ENDC }
		flipOk: boolean;
		a, b, c, d: realPoint;
		a3d, b3d, c3d, d3d: point3d;
	begin
		if triangles.length > 1 then begin
			currentTriangulation := self;
			new(a);
			failNil(Handle(a));
			new(b);
			failNil(Handle(b));
			new(c);
			failNil(Handle(c));
			new(d);
			failNil(Handle(d));
{$IFC AnimationEnabled }
			convexHullAnimation := animationFlag[AConvexHullItem];
			animationFlag[AConvexHullItem] := false;
{$ENDC }
{$IFC AnimationEnabled AND ATriangulation }
			if animationFlag[ATriangulationItem] then
				interactiveHighlight;
{$ENDC }
			pv := newPointVector(4);
			pv.append(a);
			pv.append(b);
			pv.append(c);
			pv.append(d);
			for i := 1 to n do begin
				flipOk := false;
				repeat
{$IFC AnimationEnabled }
					if userAbort then begin
						pv.freeAll;
						animationFlag[AConvexHullItem] := convexHullAnimation;
						exit(randomFlip);
					end;
{$ENDC }
{$IFC THINK_PASCAL }
					x := graphEdgeTriangulation(edges.elements^^[trunc(randomReal * edges.length) + 1]);
{$ELSEC }
					x := graphEdgeTriangulation(edges.elements^^[truncLong(randomReal * edges.length) + 1]);
{$ENDC }
					if x.triangleB > 0 then begin
						triangle(triangles.elements^^[x.triangleA]).getPoints(obj(a3d), obj(b3d), obj(c3d));
						triangle(triangles.elements^^[x.triangleB]).getThirdPoint(x, obj(d3d));
{$IFC AnimationEnabled }
						abortEnableSave := abortEnable;
						abortEnable := false;
{$ENDC }
						a.x := a3d.getX;
						a.y := a3d.getY;
						b.x := b3d.getX;
						b.y := b3d.getY;
						c.x := c3d.getX;
						c.y := c3d.getY;
						d.x := d3d.getX;
						d.y := d3d.getY;
						pv.convexHull(false, false, hullLength, eliminated);
{$IFC AnimationEnabled }
						abortEnable := abortEnableSave;
{$ENDC }
						flipOk := hullLength = 4;
					end;
				until flipOk;
{$IFC AnimationEnabled AND ATriangulation }
				if animationFlag[ATriangulationItem] then begin
					x.interactiveFlash;
					x.interactiveOutput;
				end;
{$ENDC }
				flip(x);
{$IFC AnimationEnabled AND ATriangulation }
				if animationFlag[ATriangulationItem] then begin
					x.interactiveOutput;
					waitForClick(animationFlag[ATriangulationItem]);
				end;
{$ENDC }
			end;
{$IFC AnimationEnabled }
			animationFlag[AConvexHullItem] := convexHullAnimation;
{$ENDC }
			pv.freeAll;
		end;
	end;

	var
		pixPat: PixPatHandle; {1}

function triangle3d.getType: objectType;
	begin
		getType := triangle3dType;
	end;

function triangle3d.getTextType: Str255;
	begin
		getTextType := stringOf('Triangle 3d');
	end;

function triangle3d.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoExtended(v, 'Color', color, true);
		getInfo := v;
	end;

function triangle3d.memberType (o: obj): boolean;
	begin
		memberType := member(o, triangle3d);
	end;

function triangle3d.equal (o: obj): boolean;
	var
		t: triangle3d;
	begin
		if (o <> nil) & member(o, triangle3d) then begin
			t := triangle3d(o);
			equal := inherited equal(t) & (color = t.color);
		end
		else
			equal := false;
	end;

function triangle3d.genericLessThan (o: obj): boolean;
	var
		t: triangle3d;
	begin
		if (o <> nil) & member(o, triangle3d) then begin
			t := triangle3d(o);
			genericLessThan := inherited genericLessThan(t) | (inherited equal(t) & (color < t.color));
		end
		else
			genericLessThan := false;
	end;

procedure triangle3d.drawTriangle;
	var
		ix, iy, ixfirst, iyfirst, u: integer;
		vk: clipVertexKind;
		vks: longint;
		polyH: PolyHandle;
		rgb: RGBColor; {1}

	function nextXY: boolean;
		begin
			vk := clipVertexKind(BAND(vks, 15)); {mask lower four bits}
			nextXY := vk <> noVertex;
			if vk <> noVertex then
				with clippingRectangle, polygon do begin
					vks := BSR(vks, 4);
					case vk of
						innerVertex:  begin
							ix := ixy[u];
							iy := ixy[u + 1];
							u := u + 2;
						end;
						Xtop:  begin
							ix := ixy[u];
							iy := top;
							u := u + 1;
						end;
						trVertex:  begin
							ix := right;
							iy := top;
						end;
						Xright:  begin
							ix := right;
							iy := ixy[u];
							u := u + 1;
						end;
						brVertex:  begin
							ix := right;
							iy := bottom;
						end;
						Xbottom:  begin
							ix := ixy[u];
							iy := bottom;
							u := u + 1;
						end;
						blVertex:  begin
							ix := left;
							iy := bottom;
						end;
						Xleft:  begin
							ix := left;
							iy := ixy[u];
							u := u + 1;
						end;
						tlVertex:  begin
							ix := left;
							iy := top;
						end
					end
				end
		end;

	begin	 {triangle3d.drawTriangle}
		vks := polygon.vertexKinds;
		u := 1;
		if nextXY then begin
			ixfirst := ix;
			iyfirst := iy;
			polyH := OpenPoly;
			MoveTo(ixfirst, iyfirst);
			while nextXY do
				LineTo(ix, iy);
			LineTo(ixfirst, iyfirst);
			ClosePoly;
			if currentFrame.proD^.visibility then begin
				HSBColor(0.3346303502, color, 1.0 - color, rgb);{1}
				MakeRGBPat(pixPat, rgb);{1}
				FillCPoly(polyH, pixPat);
			end
			else
				FramePoly(polyH);
			KillPoly(polyH);
		end;
	end;

procedure triangle3d.getNormalVector (var normalVector: point3d);
	var
		p1, p2, p3: point3d;
	begin
		getPoints(obj(p1), obj(p2), obj(p3));
		normalVector.setX((p1.getY - p2.getY) * (p3.getZ - p2.getZ) - (p1.getZ - p2.getZ) * (p3.getY - p2.getY));
		normalVector.setY((p1.getZ - p2.getZ) * (p3.getX - p2.getX) - (p1.getX - p2.getX) * (p3.getZ - p2.getZ));
		normalVector.setZ((p1.getX - p2.getX) * (p3.getY - p2.getY) - (p1.getY - p2.getY) * (p3.getX - p2.getX));
	end;

procedure triangle3d.calcColor;
	const
		cblack = 1.0;
		clight = 0.1;
		cdark = 0.9;
	var
		v: extended;
	begin
		getNormalVector(point3d(norm));
		with norm do begin
			if z < 0.0 then begin
				x := -x;
				y := -y;
				z := -z;
			end;
			if (x = 0.0) & (y = 0.0) & (z = 0.0) then
				color := 1.0
			else begin
				v := norm.cosPhi(light);
				if v < 0.0 then
					v := -v;
				color := cdark - (cdark - clight) * v;
			end;
		end;
	end;



	const
		graphToTriangulation3dC = 1;
		initFromLayerVectorTrapezC = 2;
		initFromLayerVectorPlaneSweepC = 3;
		ToDelaunayTriangulationC = 4;
		ToDelaunayTriangulationKeepEdgesC = 5;

procedure triangulation3d.describe;
	var
		isOpen: boolean;

	function possibleUdgTriangulation (udg: simpleUndirectedGraph): boolean;
		begin
			possibleUdgTriangulation := (udg.edges.length >= udg.vertices.length) & (udg.edges.length > 2);
		end;

	procedure addMenuEntry (entry: Str255; item: integer);
		begin
			if not isOpen then begin
				beginSubMenu('Triangulation 3d');
				isOpen := true
			end;
			menuEntry(entry, item);
		end;

	begin {describe}
		isOpen := false;
		with currentArguments do
			if length = 1 then begin
				if member(ith(1), triangulation3d) & (triangulation3d(ith(1)).triangles.length > 1) then begin
					addMenuEntry('To Delaunay Triangulation', ToDelaunayTriangulationC);
					addMenuEntry('To Del. Triang., Keep Org. Edges', ToDelaunayTriangulationKeepEdgesC);
				end;
				if member(ith(1), simpleUndirectedGraph) & possibleUdgTriangulation(simpleUndirectedGraph(ith(1))) then
					addMenuEntry('Convert graph', graphToTriangulation3dC)
				else if member(ith(1), polyLayerVector) then begin
					addMenuEntry('Plane Sweep', initFromLayerVectorPlaneSweepC);
					addMenuEntry('Trapez', initFromLayerVectorTrapezC);
				end;
			end;
		if isOpen then
			endSubMenu;
	end;

function graphToTriangulation3d (udg: simpleUndirectedGraph; function f (x, y: extended): extended): triangulation3d;
	var
		t: triangulation3d;

	function extendTransformation (p: point2d): obj;
		var
			r: realPoint3d;
		begin
			new(r);
			failNil(Handle(r));
			r.init;
			with p do begin
				r.setX(getX);
				r.setY(getY);
				r.setZ(f(getX, getY));
			end;
			extendTransformation := r;
		end;

	begin
		new(t);
		failNil(Handle(t));
		t.initializeFromGraph(udg, extendTransformation);
		t.calcColor;
		graphToTriangulation3d := t;
	end;

function layerToTriangulation3dTrapez (v: polyLayerVector): triangulation3d;
	var
		t: triangulation3d;
	begin
		new(t);
		failNil(Handle(t));
		t.init;
		t.initializeFromLayerVectorTrapez(v);
		layerToTriangulation3dTrapez := t;
	end;

function layerToTriangulation3dPlaneSweep (v: polyLayerVector): triangulation3d;
	var
		t: triangulation3d;
	begin
		new(t);
		failNil(Handle(t));
		t.initializeFromLayerVectorPlaneSweep(v);
		layerToTriangulation3dPlaneSweep := t;
	end;

function inCircle3d (pTest, p1, p2, p3: point3d): boolean;
	begin
		inCircle3d := inCircle(p1.getX, p1.getY, p2.getX, p2.getY, p3.getX, p3.getY, pTest.getX, pTest.getY);
	end;

function delaunayFlipOk3d (x: graphEdgeTriangulation): boolean;
	var
		pStart, pEnd, pLeft, pRight: point3d;
	begin
		with currentTriangulation.triangles, x do
			if (triangleB > 0) & not member(x, virtualGraphEdgeTriangulation) & not member(elements^^[triangleA], virtualTriangle3d) & not member(elements^^[triangleB], virtualTriangle3d) then begin
				with currentTriangulation.points do begin
					pStart := point3d(elements^^[startVertex]);
					pEnd := point3d(elements^^[endVertex]);
				end;
				triangle(elements^^[triangleA]).getThirdPoint(x, obj(pLeft));
				triangle(elements^^[triangleB]).getThirdPoint(x, obj(pRight));
				delaunayFlipOk3d := inCircle3d(pLeft, pStart, pEnd, pRight) | inCircle3d(pRight, pStart, pEnd, pLeft);
			end
			else
				delaunayFlipOk3d := false;
	end;

function delaunayFlipOkKeepEdges3d (x: graphEdgeTriangulation): boolean;
	var
		pStart, pEnd, pLeft, pRight: point3d;
	begin
		with currentTriangulation.triangles, x do
			if (triangleB > 0) & not member(x, virtualGraphEdgeTriangulation) & not member(elements^^[triangleA], virtualTriangle3d) & not member(elements^^[triangleB], virtualTriangle3d) then begin
				with currentTriangulation.points do begin
					pStart := point3d(elements^^[startVertex]);
					pEnd := point3d(elements^^[endVertex]);
				end;
				if pStart.getZ = pEnd.getZ then begin
					triangle(elements^^[triangleA]).getThirdPoint(x, obj(pLeft));
					triangle(elements^^[triangleB]).getThirdPoint(x, obj(pRight));
					delaunayFlipOkKeepEdges3d := inCircle3d(pLeft, pStart, pEnd, pRight) | inCircle3d(pRight, pStart, pEnd, pLeft);
				end
				else
					delaunayFlipOkKeepEdges3d := false;
			end
			else
				delaunayFlipOkKeepEdges3d := false;
	end;


procedure triangulation3d.execute (item: integer);
	var
		t: triangulation3d;
		s: stackC;
		flips: longInt;

	function f (x, y: extended): extended;
		var
			xd, yd: extended;
		begin
			xd := (x - 150) * pi / 62.5;
			yd := (y - 150) * pi / 62.5;
			f := 30 * cos(sqrt(xd * xd + yd * yd));
		end;

	begin (* triangulation3d.execute *)
		with currentArguments do
			case item of
				graphToTriangulation3dC: 
					currentResult.append(graphToTriangulation3d(simpleUndirectedGraph(ith(1)), f));
				initFromLayerVectorTrapezC: 
					currentResult.append(layerToTriangulation3dTrapez(polyLayerVector(ith(1))));
				initFromLayerVectorPlaneSweepC: 
					currentResult.append(layerToTriangulation3dPlaneSweep(polyLayerVector(ith(1))));
				ToDelaunayTriangulationC, ToDelaunayTriangulationKeepEdgesC:  begin
					t := triangulation3d(ith(1).duplicateAll);
					new(s);
					failNil(Handle(s));
					if item = ToDelaunayTriangulationC then
						flips := t.optimize(delaunayFlipOk3d, s)
					else
						flips := t.optimize(delaunayFlipOkKeepEdges3d, s);
					if abortFlag then
						t.freeAll
					else
						currentResult.append(t);
				end;
			end;
	end;

function triangulation3d.getType: objectType;
	begin
		getType := triangulation3dType;
	end;

function triangulation3d.getTextType: Str255;
	begin
		getTextType := stringOf('Triangulation 3d (V=', points.length : 1, ',E=', edges.length : 1, ',T=', triangles.length : 1, ')');
	end;

function triangulation3d.memberType (o: obj): boolean;
	begin
		memberType := member(o, triangulation3d);
	end;

function triangulation3d.isSelected (where: obj): boolean;
	var
		p: point2d;
		ix1, iy1: integer;
		x1, y1: extended;
		dummy: boolean;
	begin
		p := realPoint(newInitObj(realPointType));
		dummy := worldToWindow3d(0, 0, 0, ix1, iy1);
		windToWorld(ix1, iy1, x1, y1);
		p.setX(x1);
		p.setY(y1);
		isSelected := p.isSelected(where);
		p.freeAll;
	end;

function triangulation3d.addNewEdge (startVertexNew, endVertexNew: seqIndex): graphEdgeTriangulation;
	var
		e: graphEdgeTriangulation;
	begin
		new(e);
		failNil(Handle(e));
		e.init;
		edges.append(e);
		addNewEdge := inherited addNewEdge(startVertexNew, endVertexNew);
	end;

procedure triangulation3d.addNewTriangle (edgeANew, edgeBNew, edgeCNew: graphEdgeTriangulation);
	var
		t: triangle3d;
	begin
		new(t);
		failNil(Handle(t));
		t.init;
		triangles.append(t);
		inherited addNewTriangle(edgeANew, edgeBNew, edgeCNew);
	end;

procedure triangulation3d.calcColor;

	procedure calcTriangleColor (t: triangle3d);
		begin
			t.calcColor;
		end;

	begin
		currentTriangulation := self;
		with currentFrame.proD^ do begin
			light.setX(lightPoint[1] - proPoint[1]);
			light.setY(lightPoint[2] - proPoint[2]);
			light.setZ(lightPoint[3] - proPoint[3]);
			eye.setX(eyePoint[1]);
			eye.setY(eyePoint[2]);
			eye.setZ(eyePoint[3]);
			triangles.forAll(calcTriangleColor);
		end;
	end;

procedure triangulation3d.initializeFromLayerVectorTrapez (v: layerVector);
	var
		i, j: seqIndex;
		t: triangulation2d;
		p3: realPoint3d;
		globalShift, sumOfLayerVertices, layerVertices: integer;
		oldEdgeTree, newEdgeTree: avlTree;
		oldLayer, newLayer: vector;
		innerPolGlobalIndexes: homogenVector;
		currentZ: extended;
		layers: vector;

	function newTriangle: triangle;
		var
			t: triangle3d;
		begin
			new(t);
			failNil(Handle(t));
			newTriangle := t;
		end;


	begin
		currentTriangulation := self;
		new(t);
		failNil(Handle(t));
		t.init;
		layers := v.layers;
		globalShift := 0;
		innerPolGlobalIndexes := nil;
		oldLayer := vector(vector(layer(layers.elements^^[1]).layElems).duplicateAll);
		for i := 1 to layers.length - 1 do begin
			newLayer := vector(vector(layer(layers.elements^^[i + 1]).layElems).duplicateAll);
			new(newEdgeTree);
			failNil(Handle(newEdgeTree));
			newEdgeTree.init;
			doWholeTriangulation(t, true, globalShift, innerPolGlobalIndexes, oldEdgeTree, newEdgeTree, oldLayer, newLayer, newTriangle);
			if i <> 1 then
				oldEdgeTree.free;
			oldEdgeTree := newEdgeTree;
			oldLayer.free;
			oldLayer := newLayer;
		end;
		new(newLayer);
		failNil(Handle(newLayer));
		newLayer.init;
		doWholeTriangulation(t, true, globalShift, innerPolGlobalIndexes, oldEdgeTree, newEdgeTree, oldLayer, newLayer, newTriangle);
		oldEdgeTree.free;
		newLayer.free;
		oldLayer.free;
		sumOfLayerVertices := 0;
		for i := 1 to layers.length do begin
			layerVertices := 0;
			newLayer := vector(layer(layers.elements^^[i]).layElems);
			currentZ := layer(layers.elements^^[i]).z;
			for j := 1 to newLayer.length do
				layerVertices := layerVertices + polygon2d(newLayer.elements^^[j]).length;
			for j := (sumOfLayerVertices + 1) to (sumOfLayerVertices + layerVertices) do begin
				new(p3);
				failNil(Handle(p3));
				p3.init;
				p3.setX(point2d(t.points.elements^^[j]).getX);
				p3.setY(point2d(t.points.elements^^[j]).getY);
				point2d(t.points.elements^^[j]).free;
				p3.setZ(currentZ);
				points.appendSafe(p3, 10);
			end;
			sumOfLayerVertices := sumOfLayerVertices + layerVertices;
		end;
		edges := t.edges;
		triangles := t.triangles;
		calcColor;
		t.points.free;
		t.shallowFree;
	end;


procedure initTriangulation3d;
	var
		t: triangle3d;
		t3d: triangulation3d;
		vp3d: virtualPoint3d;
		ve: virtualGraphEdgeTriangulation;
		vt: virtualTriangle3d;
	begin
		new(t);
		failNil(Handle(t));
		t.init;
		prototypes[triangle3dType] := t;
		new(t3d);
		failNil(Handle(t3d));
		t3d.init;
		prototypes[triangulation3dType] := t3d;
		new(vp3d);
		failNil(Handle(vp3d));
		vp3d.init;
		prototypes[virtualPoint3dType] := vp3d;
		new(ve);
		failNil(Handle(ve));
		ve.init;
		prototypes[virtualGraphEdgeTriangulationType] := ve;
		new(vt);
		failNil(Handle(vt));
		vt.init;
		prototypes[virtualTriangle3dType] := vt;
		if colorAvailable then
			pixPat := NewPixPat;
    (* allocate memory for the eye point, light point and the normal vector  *)
		new(eye);
		failNil(Handle(eye));
		eye.init;
		new(light);
		failNil(Handle(light));
		light.init;
		new(norm);
		failNil(Handle(norm));
		norm.init;
	end;
end.
