(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1991 by 	Philipp Meier					*)
(* Copyright 1992 by 	Michael Greminger				*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.0:  31-Jan-92								*)
(*														*)
(* Copyright 1992 by Bert Leukens						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.4:  Aug-02-92								*)

unit triangulation3dVisibility;

interface

uses
{    Profiler, }
	triangulation3d;

implementation

procedure triangulation3d.interactiveOutput;

	const
		pointCacheSize = 51;
		edgeCacheSize = 71;
		pointCacheSize_1 = pointCacheSize - 1;
		edgeCacheSize_1 = edgeCacheSize - 1;

	type
		positionSet = set of clipVertexKind;

		pointCE = record (* vertex cache entry *)
				id: seqIndex0;
				viewPoint: point3dT;
				ps: positionSet;
				ix, iy: integer;
			end;

		edgeKind = (notvisible, clipped, visible);
		edgePoint = record
				id: seqIndex;
				vk: clipVertexKind;
				ix, iy: integer;
			end;
		edgeCE = record (* edge cache entry *)
				id: seqIndex0;
				ek: edgeKind;
				startVertex: 0..1;
				vertex: array[0..1] of edgePoint;
			end;

	var
		oldPen, newPen: PenState;
		clippingPyramide: record
				top, left, bottom, right: extended;
				eye: point3dT;
				d: extended; (* d = -eye[3] *)
			end;
		M: trafoMatrT; (* copy of currentFrame.proD^.viewTrafo *)
		l: seqIndex0;
		pointCache: array[0..pointCacheSize_1] of pointCE;
		edgeCache: array[0..edgeCacheSize_1] of edgeCE;
		pointHit, edgeHit: seqIndex0;
		changed: boolean;

	procedure drawTriangle (t: triangle3d);
		begin
			t.drawTriangle;
		end;

	procedure newColor (t: triangle3d);
		begin
			t.calcColor;
			t.drawTriangle;
		end;

	procedure clipTriangle (t: triangle3d);
		const
			eps = 1e-8;
		var
			viewPoints: array[0..2] of pointCE;
			viewEdges: array[0..2] of edgeCE;
			i, u, swap: integer;
			l: longint;
			firstVertex, prevVertex: clipVertexKind;
			p, q: point3dT;

		procedure positionCode (var p: point3dT; var ps: positionSet);
			var
				k: extended;
			begin
				ps := [];
				with clippingPyramide do
					if currentFrame.proD^.projZ then begin {central proj}
						k := (d + p[3]) / d;
						if p[1] < k * left then
							ps := ps + [Xleft];
						if p[1] > k * right then
							ps := ps + [Xright];
						if p[2] < k * bottom then
							ps := ps + [Xbottom];
						if p[2] > k * top then
							ps := ps + [Xtop];
					end
					else begin {parallel proj}
						if p[1] < left then
							ps := ps + [Xleft];
						if p[1] > right then
							ps := ps + [Xright];
						if p[2] < bottom then
							ps := ps + [Xbottom];
						if p[2] > top then
							ps := ps + [Xtop];
					end;
			end;

		procedure viewToWindow (var p: point3dT; var ix, iy: integer);
			var
				k: extended;
			begin
				with clippingPyramide, currentFrame do
					if proD^.projZ then begin {central proj}
						k := d / (d + p[3]);
{$IFC THINK_PASCAL }
						ix := round(dx + (p[1] * k + transX) * mx);
						iy := round(dy + (p[2] * k + transY) * my);
{$ELSEC }
						ix := roundLong(dx + (p[1] * k + transX) * mx);
						iy := roundLong(dy + (p[2] * k + transY) * my);
{$ENDC }
					end
					else begin {parallel proj}
{$IFC THINK_PASCAL }
						ix := round(dx + (p[1] + transX) * mx);
						iy := round(dy + (p[2] + transY) * my);
{$ELSEC }
						ix := roundLong(dx + (p[1] + transX) * mx);
						iy := roundLong(dy + (p[2] + transY) * my);
{$ENDC }
					end;
			end;

		function clipEdge (where: clipVertexKind; var v: edgePoint): boolean;
			const
				xcoordinate = 1;
				ycoordinate = 2;
			var
				clip, k, t: extended;
				ind: integer;
				r: point3dT;
				code: positionSet;
			begin
				with clippingPyramide do begin
					case where of
						Xtop: 
							clip := top;
						Xbottom: 
							clip := bottom;
						Xleft: 
							clip := left;
						Xright: 
							clip := right;
					end;
					case where of
						Xtop, Xbottom: 
							ind := ycoordinate;
						Xleft, Xright: 
							ind := xcoordinate;
					end;
					if currentFrame.proD^.projZ then {central proj}
						begin
						k := clip * (q[3] - p[3]) - d * (q[ind] - p[ind]);
						if k = 0.0 then
							k := eps;
						t := (p[ind] * d - clip * (p[3] + d)) / k;
					end
					else  {parallel proj}
						begin
						k := q[ind] - p[ind];
						if k = 0.0 then
							k := eps;
						t := (clip - p[ind]) / k;
					end;
					r[1] := p[1] + t * (q[1] - p[1]);
					r[2] := p[2] + t * (q[2] - p[2]);
					r[3] := p[3] + t * (q[3] - p[3]);
					positionCode(r, code);
					code := code - [where];
					if code = [] then
						with v do begin
							vk := where;
							viewToWindow(r, ix, iy);
							clipEdge := true;
						end
					else
						clipEdge := false;
				end;
			end;

		procedure getPoint (var p: pointCE);
			var
				hashIndex: seqIndex0;
				worldPoint: point3d;
				x, y, z: extended;
			begin
				with p do begin
					hashIndex := id mod pointCacheSize;
					if pointCache[hashIndex].id = id then begin
						pointHit := pointHit + 1;
						p := pointCache[hashIndex];
					end
					else begin
						worldPoint := point3d(points.elements^^[id]);
						x := worldPoint.getX;
						y := worldPoint.getY;
						z := worldPoint.getZ;
						viewPoint[1] := M[1, 1] * x + M[1, 2] * y + M[1, 3] * z + M[1, 4]; (* M = currentFrame.proD^.viewTrafo *)
						viewPoint[2] := M[2, 1] * x + M[2, 2] * y + M[2, 3] * z + M[2, 4];
						viewPoint[3] := M[3, 1] * x + M[3, 2] * y + M[3, 3] * z + M[3, 4];
						positionCode(viewPoint, ps);
						if ps = [] then
							viewToWindow(viewPoint, ix, iy);
						pointCache[hashIndex] := p;
					end; {else}
				end; {with}
			end;

		procedure getEdge (c: integer; edgeId: seqIndex);
			var
				hashIndex: seqIndex0;
				i, h: integer;
				ps: positionSet;
			begin
				hashIndex := edgeId mod edgeCacheSize;
				if edgeCache[hashIndex].id = edgeId then begin
					edgeHit := edgeHit + 1;
					viewEdges[c] := edgeCache[hashIndex];
				end
				else begin
					viewEdges[c].id := edgeId;
					with viewEdges[c] do begin
						startVertex := 0;
						ek := notvisible;
						for i := 0 to 1 do begin
							with vertex[i] do begin
								h := (c + 1 + i) mod 3;
								id := viewPoints[h].id;
								if viewPoints[h].ps = [] then begin
									vk := innerVertex;
									ix := viewPoints[h].ix;
									iy := viewPoints[h].iy;
									ek := succ(ek);
								end;
							end;
						end;
						if (viewPoints[(c + 1) mod 3].ps * viewPoints[(c + 2) mod 3].ps = []) & (ek <> visible) then begin {clipEdge}
							for i := 0 to 1 do begin
								h := (c + 1 + i) mod 3;
								ps := viewPoints[h].ps;
								if ps <> [] then begin
									p := viewPoints[h].viewPoint;
									q := viewPoints[(c + 2 - i) mod 3].viewPoint;
									if ps = [Xtop, Xleft, Xbottom, Xright] then
										with clippingPyramide do begin
											if p[1] * (eye[3] - q[3]) - q[1] * (eye[3] - p[3]) > 0.0 then
												ps := [Xleft]
											else
												ps := [Xright];
											if p[2] * (eye[3] - q[3]) - q[2] * (eye[3] - p[3]) > 0.0 then
												ps := ps + [Xbottom]
											else
												ps := ps + [Xtop];
										end;
									vertex[i].vk := noVertex;
									if (Xtop in ps) & clipEdge(Xtop, vertex[i]) then
										ps := [];
									if (Xbottom in ps) & clipEdge(Xbottom, vertex[i]) then
										ps := [];
									if (Xleft in ps) & clipEdge(Xleft, vertex[i]) then
										ps := [];
									if (Xright in ps) & clipEdge(Xright, vertex[i]) then
										ps := [];
									if ps = [] then
										ek := clipped;
								end;
							end;
							if (ek = clipped) & ((vertex[0].vk = noVertex) | (vertex[1].vk = noVertex)) then begin
								if vertex[0].vk = noVertex then
									i := 0
								else
									i := 1;
								if vertex[1 - i].vk = innerVertex then begin {this edge must be clipped}
									p := viewPoints[(c + 1 + i) mod 3].viewPoint;
									q := viewPoints[(c + 2 - i) mod 3].viewPoint;
									with clippingRectangle, vertex[i] do begin
										if p[1] < 0 then
											vk := Xleft
										else
											vk := Xright;
										if p[2] > 0 then
											iy := top
										else
											iy := bottom;
									end; {with}
								end
								else {do not accept clipping result of this edge}
									ek := notvisible;
							end;
						end;
					end;
					edgeCache[hashIndex] := viewEdges[c];
				end;
			end;

		procedure appendVertex (where: clipVertexKind; ix, iy: integer);
			begin
				if prevVertex = noVertex then
					firstVertex := where;
				with t.polygon do begin
					case where of
						innerVertex:  begin
							ixy[u] := ix;
							ixy[u + 1] := iy;
							u := u + 1;
						end;
						Xtop, Xbottom: 
							ixy[u] := ix;
						Xleft, Xright: 
							ixy[u] := iy;
					end;
					vertexKinds := BOR(BSL(vertexKinds, 4), ord(where));
					u := u + 1;
					prevVertex := where;
				end;
			end;

		procedure closePolygon (where: clipVertexKind);
			begin
				with t.polygon do
					if not (prevVertex in [noVertex, innerVertex]) & not (where in [noVertex, innerVertex]) then
						while prevVertex <> where do begin
							vertexKinds := BOR(BSL(vertexKinds, 4), (ord(prevVertex) - ord(Xtop) + 1) mod 8 + ord(Xtop));
							prevVertex := clipVertexKind((ord(prevVertex) - ord(Xtop) + 2) mod 8 + ord(Xtop));
						end;
			end;

		function surfaceIsVisible (var a, b, c: point3dT): boolean;
			begin
				with norm do begin
					x := (c[2] - a[2]) * (b[3] - a[3]) - (b[2] - a[2]) * (c[3] - a[3]);
					y := (c[3] - a[3]) * (b[1] - a[1]) - (b[3] - a[3]) * (c[1] - a[1]);
					z := (c[1] - a[1]) * (b[2] - a[2]) - (b[1] - a[1]) * (c[2] - a[2]);
					if z > 0.0 then begin
						x := -x;
						y := -y;
						z := -z;
					end;
					surfaceIsVisible := x * a[1] + y * a[2] + z * (a[3] - clippingPyramide.eye[3]) < 0.0
				end;
			end;

		function counterclockwise (var a, b, c: point3dT): boolean;
			var
				result: boolean;
			begin
				result := ((b[1] - a[1]) * (b[2] - c[2]) - (b[1] - c[1]) * (b[2] - a[2])) > 0.0;
				if not surfaceIsVisible(a, b, c) then
					result := not result;
				counterclockwise := result;
			end;

		begin {clipTriangle}
			t.getPointIDs(seqIndex(viewPoints[0].id), seqIndex(viewPoints[1].id), seqIndex(viewPoints[2].id));
			for i := 0 to 2 do
				getPoint(viewPoints[i]);

			getEdge(0, t.edgeA);
			getEdge(1, t.edgeB);
			getEdge(2, t.edgeC);

	  (* set clockwise orientation *)
			if counterclockwise(viewPoints[0].viewPoint, viewPoints[1].viewPoint, viewPoints[2].viewPoint) then
				swap := 2  (* do swap edges B and C *)
			else
				swap := 1;  (* don't swap edges B and C *)
		(* all startVertex values are set to 0 by default *)
			with viewEdges[0] do (* check edge A *)
				if vertex[0].id <> viewPoints[swap].id then
					startVertex := 1;
			with viewEdges[swap] do (* check edge B *)
				if vertex[1].id <> viewPoints[0].id then
					startVertex := 1;
			with viewEdges[3 - swap] do (* check edge C *)
				if vertex[0].id <> viewPoints[0].id then
					startVertex := 1;

	  (* connect points to polygon *)
			firstVertex := noVertex;
			prevVertex := noVertex;
			t.polygon.vertexKinds := 0;
			u := 1; {counter of polygon coordinates}
			for i := 0 to 2 do
				with viewEdges[(i * swap) mod 3] do begin
					case ek of
						visible: 
							with vertex[startVertex] do
								appendVertex(innerVertex, ix, iy);
						clipped:  begin
							with vertex[startVertex] do
								if vk <> noVertex then begin
									closePolygon(vk);
									appendVertex(vk, ix, iy);
								end;
							with vertex[1 - startVertex] do
								if (vk <> innerVertex) & (vk <> noVertex) then
									appendVertex(vk, ix, iy);
						end;
						notvisible: 
					end;
				end;
			closePolygon(firstVertex);

			if u = 1 then
		(* none of the edges is visible, but triangle could cut viewing pyramide *)
				if viewPoints[0].ps <> [Xtop, Xright, Xbottom, Xleft] then
					if viewPoints[1].ps <> [Xtop, Xright, Xbottom, Xleft] then
						if viewPoints[2].ps <> [Xtop, Xright, Xbottom, Xleft] then
							if viewPoints[0].ps + viewPoints[1].ps + viewPoints[2].ps = [Xtop, Xright, Xbottom, Xleft] then
								if surfaceIsVisible(viewPoints[0].viewPoint, viewPoints[swap].viewPoint, viewPoints[3 - swap].viewPoint) then
									with t.polygon do begin
										vertexKinds := ord(trVertex);
										vertexKinds := BOR(BSL(vertexKinds, 4), ord(brVertex));
										vertexKinds := BOR(BSL(vertexKinds, 4), ord(blVertex));
										vertexKinds := BOR(BSL(vertexKinds, 4), ord(tlVertex));
									end;
			with t.polygon do begin {bring the 'clipVertexKind' bit vector into reverse order}
				l := 0;
				while vertexKinds <> 0 do begin
					l := BOR(BSL(l, 4), BAND(vertexKinds, 15));
					vertexKinds := BSR(vertexKinds, 4)
				end;
				vertexKinds := l;
			end;
			if currentFrame.proD^.lightHasChanged then
				t.calcColor;
			t.drawTriangle;
		end;

	procedure sortTriangles;
		type
			integerP = reference;
			infoRec = record
					refCount: integer; 		(* reference counter *)
					verticalRef: integerP;	(* one reference counter vertical planes *)
					done: boolean; 				(* true, if all three edges are done *)
					newID: seqIndex; 			(* triangle's reference after sort *)
				end;
			sortInfoH = ^sortInfoP;
			sortInfoP = ^sortInfo;
			sortInfo = array[seqIndex] of infoRec;
		var
			si: sortInfoH;
			sortedTriangles: vector;
			e: array[0..2] of graphEdgeTriangulation;
			t: triangle3d;
			pEye: point3d;
			curID, sortedID: seqIndex0;
			curVerticalPlane: integerP;
			tdx, tdy, tdz: real;
			i: integer;

		procedure calcVisibility (var t: triangle3d);
			var
				p: array[0..2] of point3d;
				a, b, c: integer;
				otherID: seqIndex0;
				s: point3d;

			function whichSide (var p, q, r: point3d): extended;
				begin
					whichSide := (q.getY - p.getY) * (r.getX - p.getX) - (r.getY - p.getY) * (q.getX - p.getX);
				end;

			procedure setArrow (var e: graphEdgeTriangulation; var curID, otherID: seqIndex0; var p, q, r: point3d);
				var
					k1, k2: extended;
					AtoB: boolean;
				begin
					if not currentFrame.proD^.projZ then
						with pEye do begin
							setX(p.getX + tdx);
							setY(p.getY + tdy);
							setZ(p.getZ + tdz)
						end;
					with e do begin
						k1 := whichSide(p, q, r);
						k2 := whichSide(p, q, pEye);
						if k2 = 0.0 then
							AtoB := k1 < 0.0
						else
							AtoB := (k1 < 0.0) = (k2 < 0.0);
						if AtoB then begin
							triangleA := otherID;
							triangleB := curID
						end
						else begin
							triangleB := otherID;
							triangleA := curID
						end;
						with si^^[triangleB] do begin
							refCount := refCount + 1;
							if verticalRef <> nil then
								verticalRef^ := verticalRef^ + 1
						end;
					end;
				end;

			procedure verticalVisibility (curID: seqIndex0);
				var
					c: integer;
					e: array[0..2] of graphEdgeTriangulation;
					p: array[0..2] of point3d;
				begin
					with triangle3d(triangles.elements^^[curID]) do begin
						getPoints(obj(p[0]), obj(p[1]), obj(p[2]));
						getEdges(e[0], e[1], e[2]);
					end;
					with si^^[curID] do begin
						done := true;
						verticalRef := curVerticalPlane;
					end;
					curVerticalPlane^ := curVerticalPlane^ + si^^[curID].refCount;
					for c := 0 to 2 do
						with e[c] do begin
							if triangleA = curID then
								otherID := triangleB
							else
								otherID := triangleA;
							if (otherID > 0) & not si^^[otherID].done then begin
								triangle3d(triangles.elements^^[otherID]).getThirdPoint(e[c], obj(s));
								a := (c + 1) mod 3;
								b := (a + 1) mod 3;
								if (p[a].getX - p[b].getX = 0.0) & (p[a].getY - p[b].getY = 0.0) then  (* vertical edge *)
									if whichSide(p[a], p[c], s) = 0.0 then
										verticalVisibility(otherID) (* triangles are colinear *)
									else
										setArrow(e[c], curID, otherID, p[a], s, p[c])
								else if whichSide(p[a], p[b], s) <> 0.0 then  (* other triangle is normal triangle *)
									setArrow(e[c], otherID, curID, p[a], p[b], s)
								else  (* don't reference triangles of same plane *)
									verticalVisibility(otherID);
							end
						end;
				end;

			begin {calcVisibility}
				t.getPoints(obj(p[0]), obj(p[1]), obj(p[2]));
				t.getEdges(e[0], e[1], e[2]);
				if whichSide(p[0], p[1], p[2]) = 0.0 then begin (* special case: vertical triangle *)
					new(curVerticalPlane);
					failNil(Handle(curVerticalPlane));
					curVerticalPlane^ := 0;
					verticalVisibility(curID);
				end
				else begin (* normal case *)
					si^^[curID].done := true;
					for c := 0 to 2 do
						with e[c] do begin
							if triangleA = curID then
								otherID := triangleB
							else
								otherID := triangleA;
							if (otherID > 0) & not si^^[otherID].done then begin
								a := (c + 1) mod 3;
								b := (a + 1) mod 3;
								setArrow(e[c], curID, otherID, p[a], p[b], p[c]);
							end;
						end;
				end;
			end; {calcVisibility}

		procedure appendToSortedList (curID: seqIndex);
			begin
				with si^^[curID] do
					if refCount = 0 then begin
						newID := sortedID;
						done := false;
						verticalRef := nil;
						sortedTriangles.elements^^[sortedID] := triangles.elements^^[curID];
						sortedID := sortedID + 1;
					end;
			end;

		procedure appendPlane (curID: seqIndex);
			var
				otherID: seqIndex0;
				e: array[0..2] of graphEdgeTriangulation;
				i: integer;
			begin
				appendToSortedList(curID);
				triangle3d(triangles.elements^^[curID]).getEdges(e[0], e[1], e[2]);
				for i := 0 to 2 do
					with e[i] do begin
						if triangleA = curID then
							otherID := triangleB
						else
							otherID := triangleA;
						if (otherID > 0) & si^^[otherID].done & (si^^[otherID].verticalRef = curVerticalPlane) then
							appendPlane(otherID);
					end;
			end; {appendPlane}

		procedure dereferenceTriangle (curID: seqIndex);
			begin
				with si^^[curID] do
					if done then begin
						refCount := refCount - 1;
						if verticalRef = nil then
							appendToSortedList(curID)
						else begin
							verticalRef^ := verticalRef^ - 1;
							if verticalRef^ = 0 then begin
								curVerticalPlane := verticalRef;
								appendPlane(curID);
								dispose(curVerticalPlane);
							end
						end
					end;
			end;

		begin {sortTriangles}
			with currentFrame.proD^ do
				if projZ then
					pEye := eye
				else begin
					pEye := norm;
					tdx := eyePoint[1] - proPoint[1];
					tdy := eyePoint[2] - proPoint[2];
					tdz := eyePoint[3] - proPoint[3]
				end;
			si := sortInfoH(dynamicArray(triangles.length, sizeof(infoRec)));
			sortedTriangles := newVector(triangles.length);
			sortedTriangles.length := triangles.length;

			with triangles do begin
				for curID := 1 to length do
					with si^^[curID] do begin
						refCount := 0;
						done := false;
						verticalRef := nil;
					end;

				for curID := length downto 1 do
					if not si^^[curID].done then
						calcVisibility(triangle3d(elements^^[curID]));

		(* NOTE: done-flags are all true now. All following done-checks will have an inverted sense! *)
				sortedID := 1;
				for curID := length downto 1 do {find unreferenced triangles}
					with si^^[curID] do
						if done then
							if verticalRef <> nil then begin
								if verticalRef^ = 0 then begin
									curVerticalPlane := verticalRef;
									appendPlane(curID);
									dispose(curVerticalPlane)
								end
							end
							else
								appendToSortedList(curID);
			end;

			curID := 1;
			with sortedTriangles do {dereference all dependend triangles}
				while curID < sortedID do begin
					t := triangle3d(elements^^[curID]);
					if changed then
						clipTriangle(t)
					else
						t.drawTriangle;
					t.getEdges(e[0], e[1], e[2]);
					for i := 0 to 2 do
						with e[i] do
							if triangleB > 0 then
								dereferenceTriangle(triangleB);
					curID := curID + 1
				end;

			if curID > sortedTriangles.length then begin
				with edges do {update all edges}
					for curID := 1 to length do
						with graphEdgeTriangulation(elements^^[curID]) do begin
							triangleA := si^^[triangleA].newID;
							if triangleB <> 0 then
								triangleB := si^^[triangleB].newID;
						end;
				triangles.free;
				triangles := sortedTriangles
			end
			else begin
				nonFatalError('Vertical plane blocks graph of dependencies; calculation of visibility is not possible');
				currentFrame.proD^.visibility := false;
				sortedTriangles.free;
				triangles.forAll(clipTriangle);
			end;

			DisposeHandle(Handle(si));
		end; {sortTriangles}

	begin (* triangulation3d.interactiveOutput *)

{    %_PTrace := true;}
{    ResetProfile;}

		currentTriangulation := self;
		GetPenState(oldPen);
		newPen := oldPen;
		newPen.pnMode := patCopy;(* set paint mode *)
		SetPenState(newPen);

		with currentFrame.proD^ do begin
			changed := eyeProHasChanged | proHasChanged;
			if not visibility then begin
				drawCoordSys;
				coordSysIsDrawn := true
			end;
			with light do begin
				x := lightPoint[1] - proPoint[1];
				y := lightPoint[2] - proPoint[2];
				z := lightPoint[3] - proPoint[3];
			end;
			with eye do begin
				x := eyePoint[1];
				y := eyePoint[2];
				z := eyePoint[3];
			end;
			if changed then begin
				M := viewTrafo;
				with clippingPyramide do begin
					windToWorld(clippingRectangle.left, clippingRectangle.top, left, top);
					windToWorld(clippingRectangle.right, clippingRectangle.bottom, right, bottom);
					left := left - transX;
					right := right - transX;
					top := top - transY;
					bottom := bottom - transY;
					mvm(viewTrafo, eyePoint, eye);
					d := -eye[3]; (* Distance between eye and projection plane *)
				end;

		  (* reset caches *)
				for l := 0 to pointCacheSize_1 do
					pointCache[l].id := 0;
				for l := 0 to edgeCacheSize_1 do
					edgeCache[l].id := 0;
				pointHit := 0;
				edgeHit := 0;

				if visibility then
					sortTriangles
				else
					triangles.forAll(clipTriangle);

{$IFC XYZApplication }
{$ELSEC }
				displayInfoString2('Caches used:  edges=', concat(stringOf(100.0 * edgeHit / (3.0 * triangles.length - edges.length) : 2 : 1), '%'), '    points=', concat(stringOf(100.0 * pointHit / (3.0 * triangles.length - points.length) : 2 : 1), '%'));
{$ENDC XYZApplication }
			end
			else if visHasChanged & visibility then
				sortTriangles
			else if lightHasChanged then
				triangles.forAll(newColor)
			else
				triangles.forAll(drawTriangle);
		end;
		SetPenState(oldPen);

{    DumpProfileToFile('ProfileVisibility3D');}
{    TerminateProfile;}

	end;

procedure triangulation3d.interactiveHighlight;
	begin
	end;


end.
