(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Version history *)

(*	20-May-94, PS	*)
(*	-	bug fix: if a dynamic execution window was hidden and the dynamic execution aborted, an error resulted.	*)
(*	-	bug fix: refresh of dynamic execution was incorrect.																	*)

(*	10-May-94, PS	*)
(*	-	added the method 'sequence.forRandomSubset'.																		*)
(*	-	changed the menu command 'Select First' to 'Select Next'.															*)
(*	-	added the menu commands 'Add Next' and 'Replace Next'.															*)

(*	5-May-94, PS	*)
(*	-	the bounding box of a markedRealPoint now considers the marking string.											*)
(*	-	added the method 'list.reset'.																								*)
(*	-	changed the definition of 'obj.dragVertex', 'obj.dragObjectTransformTranslation', 'obj.dragVertexWorld'	*)
(*		and 'obj.dragObject'.																										*)
(*	-	added the method 'point2d.drawLineHighlight'.																			*)
(*	-	one can now drag selected objects.																						*)
(*	-	dragging an object does not change its position in the selected or unselected list.								*)

(*	4-May-94, PS	*)
(*	-	fixed a bug in the method 'pointVector.getBoundingObject' which caused a crash in the case the method		*)
(*		was called with a pointVector of length zero.																			*)
(*	-	made sure that member(currentRandomConstraint, ...) is only called if the random constraint is not nil.		*)
(*	-	fixed a bug in 'sequence.centerOfGravity' which caused the center coordinates to be undefined if the			*)
(*		length of the sequence was zero.																							*)
(*	-	fixed a bug which prevented the textual output of a histogram.														*)

(*	25-Apr-94, PS	*)
(*	-	changed the result type of 'randomReal' from 'real' to 'extended' in order to gurantee that the result		*)
(*		is less than one.																												*)

(*	19-Apr-94, PS	*)
(*	-	added the method 'vector.initFromListPartial'.																			*)
(*	-	moved the methods 'vector.randomShuffleAll' and 'vector.randomShuffle' to the class 'sequence'.			*)

(*	14-Apr-94, PS	*)
(*	-	added the method 'triangulation.toUndirectedGraph.																	*)
(*	-	added the command 'To Undirected Graph' for triangulations.														*)

(*	11-Apr-94, PS	*)
(*	-	fixed a bug where it was possible do open a GeoBench document with a double click while the GeoBench		*)
(*		was computing in the background.																						*)
(*	-	fixed a bug in the Raumplanungs-Software (NS).																		*)
(*	-	fixed a bug in the exact computation for the Traveling Salesman Problem (the answer did sometimes		*)
(*		depend on the order of the points, AB).																					*)
(*	-	fixed a bug where the user preferences could not be found if the system folder was not on the top level.	*)
(*	-	added a button in the operation info menu entry (in the apple menu) to inspect and set the background		*)
(*		priority (i.e. the minimal number of ticks the program will computer un-interrupted in the background.		*)

(*	4-Feb-94, PS	*)
(*	-	added the methods 'vector.deleteElement', 'vector.deleteElementFree' and 'vector.deleteElementFreeAll'.	*)

(*	3-Feb-94, PS	*)
(*	-	added the commands 'Make: Full Graph' and 'Make: Full Directed Graph'.											*)
(*	-	added the commands 'Random: Perturb in Circle...' and 'Random: Perturb in Rectangle...'.						*)
(*	-	fixed a bug which prevented the output method 'Length First' to work properly.									*)
(*	-	added the method 'dictionary.exists'.																					*)
(*	-	changed the definition of 'infiniteUndirectedGraph.fromSimpleUndirectedGraph'.									*)
(*	-	changed the definition of 'infiniteUndirectedGraph.hennebergKReduction'.											*)
(*	-	changed the comment for 'infiniteVertex'.																				*)
(*	-	added the method 'infiniteUndirectedGraph.fromSimpleUndirectedGraphSpecial'.									*)
(*	-	added the method 'infiniteUndirectedGraph.countEdgesSpecial'.														*)

(*	19-Jan-94, PS	*)
(*	-	added the menu commands 'Edit: Paste Special' and 'Objects: Ungroup All'.										*)
(*	-	changed the implementation of 'list.flatten' to flatten included lists as well.										*)
(*	-	added the commands 'Euclidean Distance' and 'Lp Distance' for two points.										*)

(*	11-Jan-94, PS	*)
(*	-	added the command 'Which Side' and 'Which Side Sign' for three points (previously only for a point and		*)
(*		a line segment.																												*)
(*	-	added the method 'vector.compact'.																						*)
(*	-	changed the definition of 'infiniteUndirectedGraph.fromSimpleUndirectedGraph'.									*)

(*	3-Jan-94, PS	*)
(*	-	added the method 'infiniteUndirectedGraph.hennebergKReduction'.													*)
(*	-	added the command 'Henneberg Reduction Ask...'.																		*)

(*	22-Dec-93, PS	*)
(*	-	added the field 'cleanUp' to the object 'objInfoTempObj'.																*)
(*	-	fixed a bug in 'polyLayerVector.describe'.																				*)

(*	12-Nov-93, PS	*)
(*	-	added the method 'obj.showObjectParameter', 'obj.dragVertexWorld' and 'obj.constructWorld'.				*)
(*	-	manual is uptodate.																											*)

(*	4-Nov-93, PS	*)
(*	-	added the file menu commands 'Save a Copy As...', 'Revert', 'Revert All', 'Revert All Without Asking'.		*)
(*	-	changed the text file format for regular n-gons.																		*)
(*	-	changed the definition of 'pointVector.cluster'.																			*)
(*	-	changed the command 'Clustering: Components...' to 'Clustering: Circle Components...'.							*)
(*	-	added the command 'Clustering: Convex Polygon Components...'.													*)
(*	-	changed the definition of 'convexPolygon.intersect'.																	*)

(*	29-Oct-93, PS	*)
(*	-	changed the definition of 'pointVector.minimalDisk'.																	*)
(*	-	changed the definition of 'dDimPointVector.minimalDisk'.																*)
(*	-	added the command 'Cluster: Smallest Area'.																			*)
(*	-	added the command 'Cluster: Smallest Area Intersect'.																*)
(*	-	added the command 'Cluster: Smallest Area...'.																		*)
(*	-	added the command 'Cluster: Smallest Area Intersect...'.															*)
(*	-	added the command 'Cluster: Smallest Size'.																			*)
(*	-	added the command 'Cluster: Components...'.																			*)
(*	-	changed the definition of 'pointVector.computeSmallestAreaCluster'.												*)

(*	22-Oct-93, PS	*)
(*	-	fixed a bug in 'pointVector.closestPairNewHeuristic'.																	*)
(*	-	fixed a bug in 'pointVector.shortestPath'.																				*)
(*	-	changed the definition of 'pointVector.convexHull'.																	*)
(*	-	fixed a bug in 'convexPolygon.randomChange'.																			*)
(*	-	added the method 'sequence.cyclicIth'.																					*)
(*	-	fixed a bug in 'unionFind.find'.																								*)

(*	21-Oct-93, PS	*)
(*	-	added a range to the number of vertices in the random creation of polylines, polygons and convex			*)
(*		polygons.																														*)
(*	-	changed the definition of 'pointVector.EMST', 'voronoiDiagram.EMST' and											*)
(*		'simpleUndirectedGraph.minimumSpanningTree'.																		*)
(*	-	added the method 'unionFindSimple.compressAll'.																		*)
(*	-	added the method 'pointVector.cluster'.																					*)
(*	-	added the method 'pointVector.computeSmallestAreaCluster'.														*)
(*	-	added the method 'pointVector.computeSmallestSizeCluster'.														*)

(*	30-Sep-93, PS	*)
(*	-	added keyboard equivalents for the buttons in the animation dialog box.											*)
(*	-	the animation dialog box can be moved to front with the 'Animation' command in the windows menu.			*)

(*	22-Sep-93, PS	*)
(*	-	added the commads 'Voronoi Diagram: Delaunay Triangulation Grid' and											*)
(*		'Voronoi Diagram: Delaunay Triangulation Grid Dynamic'.															*)
(*	-	added the method 'pointVector.gridTriangulation'.																		*)

(*	14-Sep-93, PS	*)
(*	-	the manual is uptodate.																										*)
(*	-	fixed a bug where menus could be activated while computing.														*)
(*	-	'Operation Info...' gives also information about the currently executing operation.								*)
(*	-	random create operations are timed.																						*)
(*	-	improved 'Randomize Seed...'.																								*)
(*	-	added the command 'Minimal Area Disk: Approximate Dynamic'.													*)
(*	-	added the command 'Minimal Area Disk: Dynamic' (computes exact and approximate solution).				*)
(*	-	changed the command 'Minimal Area Disk: Dynamic' to 'Minimal Area Disk: Randomized Dynamic'.			*)
(*	-	improved timing.																											*)
(*	-	Version 4.4.4 released.																									*)

(*	27-Aug-93, PS	*)
(*	-	GeoScript code updated.																									*)
(*	-	Version 4.4.4 Beta released.																								*)

(*	19-Aug-93, PS	*)
(*	-	added the method 'list.delete'.																								*)
(*	-	added the method 'obj.drawingComplexity'.																				*)

(*	12-Aug-93, PS	*)
(*	-	added the commands 'Voronoi Diagram: Delaunay Triangulation' and												*)
(*		'Voronoi Diagram: Delaunay Triangulation Dynamic' for point sets.													*)

(*	6-Aug-93, PS	*)
(*	-	changed the method 'sequence.dynamicExecutionNoAnimation' to													*)
(*		'sequence.dynamicExecutionHighlightShowType' and  'dynamicExecutionNoHighlightShowType'.				*)
(*	-	changed the command 'Align To First Point' to 'Align To First Point: Translate Containing Object'.			*)
(*	-	added the command 'Align To First Point: Move Only Point'.															*)

(*	4-Aug-93, PS	*)
(*	-	changed the definition of 'point2d.distancePLpX', 'point2d.dxPLpX' and 'point2d.dyPLpX'.						*)
(*	-	the manual is uptodate.																										*)
(*	-	version 4.4.3 released.																									*)

(*	2-Aug-93, PS	*)
(*	-	added the commands 'Nearest Neighbors: Left Plane Sweep Dynamic'												*)
(*		'Nearest Neighbors: Right Plane Sweep Dynamic', 'Nearest Neighbors: All Plane Sweep Dynamic'				*)
(*		'Nearest Neighbors: All Boxes Dynamic', 'Nearest Neighbors: Left In Sector Dynamic'							*)
(*		'Convex Hull: Dynamic'. 'Minimal Area Disk: Dynamic', 'Contour Of Rectangles Dynamic',						*)
(*		'All Intersect: Plane Sweep Dynamic', 'All Intersect: Horizontal Vertical Dynamic'								*)
(*		'First Intersect: Plane Sweep Dynamic', 'Lower Envelope Dynamic'												*)
(*	-	added the method 'sequence.dynamicExecutionNoAnimation'.															*)

(*	30-Jul-93, PS	*)
(*	-	changed the command 'Euclidean Minimum Spanning Tree' to 														*)
(*		'Euclidean Minimum Spanning Tree: Nearest Neighbor In Octant'.													*)
(*	-	added the commands 'Euclidean Minimum Spanning Tree: Nearest Neighbor In Octant Dynamic',				*)
(*		'Euclidean Minimum Spanning Tree: Voronoi Diagram',																*)
(*		'Euclidean Minimum Spanning Tree: Voronoi Diagram Dynamic', and												*)
(*		'Traveling Salesman Optimal: 1 Tree Dynamic'.																		*)

(*	29-Jul-93, PS	*)
(*	-	added the method 'pointVector.TSPSimple' by Adrian Brngger.														*)
(*	-	added the method 'pointVector.TSP1TreeBranchAndBound' by Adrian Brngger.									*)
(*	-	added the command 'Traveling Salesman Optimal: Simple'.															*)
(*	-	added the command 'Traveling Salesman Optimal: 1 Tree Branch and Bound'.										*)
(*	-	the manual is uptodate.																										*)

(*	28-Jul-93, PS	*)
(*	-	changed the definition of 'obj.dragObject' and 'obj.dragVertex'.														*)
(*	-	added the method 'obj.windowInputMouse'.																				*)
(*	-	added the method 'obj.dragObjectTransformTranslation'.																*)
(*	-	changed the comment for 'obj.dragObject'.																				*)
(*	-	drag vertex for line segments now listens to option key and the command key.									*)

(*	22-Jul-93, PS	*)
(*	-	added the method 'pointVector.orderTypeOne'.																			*)
(*	-	added the command 'Order Type'.																							*)
(*	-	added the command 'Create Star Bushel'.																				*)

(*	8-Jul-93, PS	*)
(*	-	removed 'point2d.signDx' and 'point2d.signDy'.																		*)
(*	-	added 'point2d.DxX' and 'point2d.DyX'.																					*)
(*	-	added the command 'Random Point Cover...'.																			*)
(*	-	added the method 'obj.randomPointCover...'.																			*)
(*	-	removed the commands 'Intersection Detection', 'Find Y-Extreme Vertices' and									*)
(*		'Find X-Extreme Vertices', 'Minimal Distance: Binary 2' and 'Minimal Distance: Binary 2 Dynamic'.			*)
(*	-	added the method 'circle.getCenter'.																						*)

(*	20-May-93, PS	*)
(*	-	enlarged the result type of 'polygon2d.isConvex'.																		*)
(*	-	added the method 'obj.convertToConvexPolygonPossible'.															*)
(*	-	removed the method 'polygon2d.makePolygon'.																			*)
(*	-	added the methods 'obj.convertToPolygon' and 'obj.convertToPolygonPossible'.									*)
(*	-	added the command 'Conversion: To Polygon'.																			*)

(*	14-May-93, PS	*)
(*	-	added the command 'Make: Simple Polyline' for a set of line segments.											*)
(*	-	the menu command 'Ungroup' ungroups line segments as well.														*)
(*	-	the user can now specify the output method for text output in the preferences.									*)

(*	10-May-93, PS	*)
(*	-	added the command 'Convex Intersection All Shifts'.																	*)
(*	-	changed the definition of 'obj.textFileOutput'.																			*)
(*	-	removed the method 'obj.textFileOutputSpecial'.																		*)

(*	6-May-93, PS	*)
(*	-	added the method 'pointVector.convexHullGrahamScan'.																*)
(*	-	changed the command 'To Point Vector' to 'Conversion: To Point Vector'.											*)
(*	-	changed the command 'Show Numbering' to 'Numbering: Show'.														*)
(*	-	changed the command 'Revert Numbering' to 'Numbering: Revert'.													*)
(*	-	changed the command 'Cyclic Shift...' to 'Numbering: Cyclic Shift...'.												*)
(*	-	changed the command 'Arrange Objects...' to 'Arrange Objects: On Box Grid...'.									*)
(*	-	changed the command 'Eliminate Repeated Elements' to 'Eliminate: Repeated Elements'.							*)
(*	-	changed the command 'Eliminate Duplicates' to 'Eliminate: Duplicates'.												*)
(*	-	changed the command 'Random Shuffle' to 'Random: Shuffle'.														*)
(*	-	changed the command 'Random Change' to 'Random: Change'.															*)
(*	-	changed the command 'Perimeter' to 'Universal Geometry: Perimeter'.											*)
(*	-	changed the command 'Area' to 'Universal Geometry: Area'.														*)
(*	-	added the command  'Arrange Objects: With Same Center Of Gravity'.												*)
(*	-	added the command 'Closest Pair: Plane Sweep Dynamic' (for line segments).									*)

(*	3-May-93, PS	*)
(*	-	added the methods 'obj.HBGLessThan', 'obj.verticalDistance', 'obj.closestPair','obj.getPointType',			*)
(*		'obj.distanceO', 'obj.getLeftPoint' and 'obj.getRightPoint'.															*)
(*	-	added the commands 'Plane Sweep Closest Pair', 'Intersection Detection', 'Find Y-Extreme Vertices' and	*)
(*		'Find X-Extreme Vertices', 'Minimal Distance: Binary 2' and 'Minimal Distance: Binary 2 Dynamic'.			*)
(*	-	changed 'pointVector.closestPair' to 'pointVector.closestPairPoints'.												*)
(*	-	added the command 'Intersection Detection Dynamic' and 'Plane Sweep Closest Pair Dynamic'.				*)
(*	-	added the method 'obj.convertToConvexPolygon'																		*)
(*	-	changed the definition of 'convexPolygon.tangent' to accomodate the case where no tangents exist.			*)
(*	-	changed the command 'Center Of Gravity' to 'Conversion: To Center Of Gravity'.								*)
(*	-	changed the command 'To Real' to 'Conversion: To Real'.																*)
(*	-	changed the command 'To LongInt' to 'Conversion: To LongInt'.														*)
(*	-	changed the command 'To Float' to 'Conversion: To Float'.															*)
(*	-	added the methods 'point2d.leftOf', 'point2d.rightOf', 'point2d.onLine', 'point2d.orthogonalProjection',		*)
(*		'point2d.lineIntersectsLine', 'point2d.projectOnLine'.																	*)
(*	-	removed the method 'point2d.perpendicularDistanceX' and 'lineSegment.perpendicularDistanceX'.				*)
(*	-	added the methods 'lineSegment.lineIntersectsLine' and 'lineSegment.projectOnLine'.							*)

(*	27-Apr-93, PS	*)
(*	-	added the command 'Align To First Point'.																				*)

(*	22-Apr-93, PS	*)
(*	-	added the commands 'Point In Polygon' and 'Point In Polygon Dynamic'.											*)
(*	-	changed the definition of 'vector.dynamicExecution' and moved it to 'sequence.dynamicExecution'.			*)
(*	-	changed the definition of 'dynamicExecution2'.																			*)
(*	-	added the method 'polygon2d.pointInPolygon'.																			*)
(*	-	added the command 'Center Of Gravity'.																					*)
(*	-	added the method 'sequence.existsObject'.																				*)
(*	-	added the commands 'Set Operations Dynamic: Intersection', 'Set Operations Dynamic: Union'					*)
(*		'Set Operations Dynamic: 1 Difference 2' and 'Set Operations Dynamic: 2 Difference 1'.						*)
(*	-	added the commands 'Tangent Dynamic' and 'Convex Intersection Dynamic' for convex polygons.				*)

(*	7-Apr-93, PS	*)
(*	-	added the commands 'Interactive Operations: Delete Vertex' and 'Interactive Operations: Delete Edge'.		*)

(*	1-Apr-93, PS	*)
(*	-	added the method 'vector.dynamicExecution'.																			*)
(*	-	added the commands 'Voronoi Diagram: Dynamic Fortune', 'Voronoi Diagram: Dynamic Seidel' and			*)
(*		'Closest Pair: Dynamic'.																									*)

(*	31-Mar-93, PS	*)
(*	-	added the methods 'point2d.distanceLpX', 'point2d.distancePLpX', 'point2d.dxPLpX', 'point2d.dyPLpX'.		*)
(*	-	added the method 'pointVector.voronoiDiagramSweepLineSeidel'.													*)
(*	-	added a menu entry 'Lp metric' in the 'Arithmetric'-menu.															*)
(*	-	added the method 'obj.perimeter' which computes the perimeter in the Lp-metric.								*)
(*	-	added the method 'obj.perimeterMP'.																					*)
(*	-	added the method 'obj.areaMP'.																							*)
(*	-	the commands 'Area' and 'Perimeter' are now universal.																*)

(*	25-Mar-93, PS	*)
(*	-	added a function which tests whether a graph satisfies a fairly complex boolean formula.						*)
(*	-	added a command 'Test Graph With Formula...'.																			*)
(*	-	added the command 'Find: Graph Satisfying Formula...'.																*)

(*	18-Mar-93, PS	*)
(*	-	added the method 'simpleUndirectedGraph.hennebergSequenceMaxTrials'.											*)
(*	-	changed the definition of 'simpleUndirectedGraph.hennebergSequence'.												*)
(*	-	changed the definition of 'directedGraph.maxTriangulationOrder'.													*)
(*	-	added the command 'Find: Random Triangulable Graph'.																*)
(*	-	added the command 'Find: Random Rigid Graph'.																		*)
(*	-	added the command 'Find: Random Rigid Not Triangulable Graph'.													*)

(*	15-Mar-93, PS	*)
(*	-	added the method 'directedGraph.maxTriangulationOrder'.															*)
(*	-	added the method 'simpleUndirectedGraph.graphIsConnected'.														*)

(*	11-Mar-93, PS	*)
(*	-	added the methods 'sequence.randomChooseFromRange', 'sequence.randomChooseFromAll'						*)
(*		'simpleUndirectedGraph.isConnected' and 'simpleUndirectedGraph.hennebergSequence'.							*)
(*	-	added the commands 'Make: Empty Graph' and 'Make: Empty Directed Graph'.									*)
(*	-	added the command 'Find Henneberg 2-Sequence'.																		*)
(*	-	changed the definition of 'markedRealPoint.initialize' and renamed it to											*)
(*		'markedRealPoint.initLocationMark'.																						*)
(*	-	changed 'simpleUndirectedGraph.deleteVertex' and 'simpleUndirectedGraph.deleteEdge' to free the objects	*)
(*		which cannot longer be accessed.																							*)
(*	-	added the method 'sequence.makeIdRealPointVector'.																	*)
(*	-	the manual is up-to-date.																									*)

(*	10-Mar-93, PS	*)
(*	-	added 'RaumPlanung' code by Roland Siposs (internal version only).												*)
(*	-	manual is up-to-date.																										*)

(*	5-Mar-93, PS	*)
(*	-	fixed a bug in 'interactiveAddVertex'.																					*)
(*	-	removed an extraneous 'Make: Star-Shaped Polygon' command.													*)
(*	-	added balloon help for the icon.																							*)

(*	25-Feb-93, PS	*)
(*	-	added the command 'Arrange Objects...'.																				*)
(*	-	added the method 'sequence.arrange'.																					*)

(*	18-Feb-93, PS	*)
(*	-	extended the preferences to include the type of the text files to be created.										*)
(*	-	changed the definition of 'directedGraph.getTriangulation'.															*)
(*	-	the command 'Interactive Operations: Add Edges...' can also delete an edge by selecting an existing edge.	*)

(*	9-Feb-93, PS	*)
(*	-	added the method 'obj.area'.																								*)
(*	-	added the method 'polygon2d.triangulated2d'.																			*)
(*	-	added the command 'To Convex Polygon' for a set of rectangles.													*)
(*	-	the command 'To Convex Polygon' works now also for a set of regular n-gons.									*)

(*	28-Jan-93, PS	*)
(*	-	added the method 'dictionary.insertIfNew'.																				*)

(*	26-Jan-93, PS	*)
(*	-	removed the file menu entry 'Write Text Simple...'.																	*)

(*	22-Jan-93, PS	*)
(*	-	added the data type 'union find'.																							*)
(*	-	changed the definition of 'pointVector.createRandomGraph' to include connectedness.							*)

(*	20-Jan-93, PS	*)
(*	-	added the method 'sequence.forSome'.																					*)
(*	-	added the method 'sequence.appendInt'.																					*)
(*	-	changed the definition of 'directedGraph.getTriangulation'.															*)
(*	-	added the command 'Interactive Operations: Add Vertices...'.														*)

(*	12-Jan-93, PS	*)
(*	-	added the method 'pointVector.createRandomGraph'.																	*)

(*	6-Jan-93, PS	*)
(*	-	added the method 'point2d.setXY'.																						*)
(*	-	added the method 'point2d.getXY'.																						*)
(*	-	the result of 'getTriangulation' is always a directed graph.															*)
(*	-	the command 'Test Triangulation Ask..' was added.																	*)
(*	-	added the ability to read CABRI graph documents.																		*)
(*	-	version 4.4.2 released.																									*)

(*	18-Dec-92, PS	*)
(*	-	added the command 'Add Graphs' for a set of graphs.																	*)
(*	-	the menu entries 'Close' and 'Save' specify the name of the window to be closed resp. saved.					*)
(*	-	a preferences entry for chemistry was added.																			*)
(*	-	the dialog engine was improved to handle popup-menus (using code by Markus Julen).							*)
(*	-	a simplified method is provided for specifying non contiguous radio-buttons for the implementation of		*)
(*		enumeration types.																											*)
(*	-	a tutorial method was added (c.f. compiler variable Tutorial).														*)

(*	20-Nov-92, PS	*)
(*	-	added the command 'Make: Random Graph...' for a set of points.														*)
(*	-	added the command 'Interactive Operations: Add Edges...'.															*)

(*	12-Nov-92, PS	*)
(*	-	Version 4.4.1 released.																									*)
(*	-	changed icon of preferences file.																							*)
(*	-	no attempts to interpret preferences file as geometry data.															*)
(*	-	changed 'about' dialog box and initial dialog box to show 'Freeware'.												*)
(*	-	default preference for initialization heuristic (closest pair) is now 'off'.											*)
(*	-	the command 'Eliminate Repeated Elements' does not return a list any more.										*)
(*	-	the keyboard short cut for GeoScript windows is now 'B' instead of 'M' which was already in use.			*)

(*	3-Nov-92, PS	*)
(*	-	fixed a bug in 'polyLine.linearConvexHull' regarding degenerate configurations.									*)
(*	-	addded the commands 'Which Side Sign Dynamic' and 'Which Side Dynamic'.										*)
(*	-	changed the definition of 'polygonKind'.																					*)
(*	-	'convexPolygon.intersect' makes sure that the polygons are left turning.											*)
(*	-	fixed a orientation dependent bug in 'convexPolygon.diameter'.														*)

(*	27-Oct-92, PS	*)
(*	-	result handling modifiers reverse mode if the mode is already present.											*)
(*	-	added 'obj.textFileOutputSpecial' and 'obj.textFileOutputSpecialId'.													*)
(*	-	added a 'Write Text Simple...' in the file menu.																			*)
(*	-	added a boolean in the preferences that specifies whether a textual version of the selection should be put	*)
(*		into the clipboard.																											*)
(*	-	added a boolean in the preferences that specifies the format of the textual version in the clipboard.			*)

(*	23-Oct-92, PS *)
(*	-	changed the parameters of 'polygon2d.isConvex'.																		*)
(*	-	added 'sequence.removeRepeatedElements'.																				*)
(*	-	added a command 'Eliminate Repeated Elements'.																		*)
(*	-	added 'simpleUndirectedGraph.addNewEdge'																				*)
(*	-	added 'sequence.createGraph'.																							*)
(*	-	added the command 'Make: Directed Graph' for point vectors.														*)

(*	21-Oct-92, PS	*)
(*	-	added the method 'polygon2d.isConvex'.																					*)
(*	-	added new key-words for text file input: pointlist and polygon for reading a list of polygons and point		*)
(*		sets.																															*)
(*	-	added some range checks for lengths on strings created in 'textIO'.												*)
(*	-	filled the whole buffer with 0's after an incomplete read operation.												*)

(*	14-Oct-92, PS *)
(*	-	added 'vector.interactiveSelect'.																							*)
(*	-	added 'directedGraph.reverseEdges'.																						*)
(*	-	added a command to revert the edges of a directed graph.															*)
(*	-	added 'simpleUndirectedGraph.breadthFirstSearch' for finding the distance of all nodes from a given node	*)
(*		using breadth first search.																								*)
(*	-	added a command to find the distances from node 1.																	*)
(*	-	added a command to find the distances from an arbitrary user selectable node.									*)
(*	-	use 'vector.interactiveSelect' in 'interactiveCX.getEdge'.															*)
(*	-	added 'vector.interactiveSelectFromAll'.																				*)
(*	-	made 'Delete Vertex...' and 'Delete Triangle...' interactive.															*)
(*	-	implemented 'triangle2d.isSelected'.																						*)
(*	-	implemented 'graphEdge.isSelected'.																						*)
(*	-	changed the definition of 'graphEdge.drawEdge'.																		*)

(*	13-Oct-92, PS	*)
(*	-	new commands 'Align Points X' and 'Align Points Y' for sets of points.												*)
(*	-	interactively flipping shows whether an edge is Delaunay or not.													*)
(*	-	commands for exploring the graph induced by flipping.																*)
(*	-	introduced a new type 'markedGraphEdge'.																				*)
(*	-	added a command to write graphs in adjacency list form to file and to write them in CABRI graph format.	*)
(*	-	changed 'getInfo' and 'setInfo' for graphs such that changes are now allowed.										*)
(*	-	introduced a new GeoBench file version id: 'version8' ist now current.											*)
(*	-	changed the type declaration for 'simpleUndirectedGraph': the edges are now of type 'vector'.				*)
(*	-	changed the file format for graphs (compatible with old version).													*)
(*	-	the font number is now in the preferences.																				*)
(*	-	added the boolean 'Accept Result Handling Modifiers' in the preferences dialog box. If this box is checked	*)
(*		the shift-key and the option-key are interpreted as modifiers in the operations menu. Their meaning is:		*)
(*		If the shift-key is pressed and not the option-key then the result is added as a selection to the current		*)
(*		window. If the option-key is pressed and not the shift-key then the result replaces the selection as a new	*)
(*		selection.																														*)

(*	6-Oct-92, PS	*)
(*	-	wait limit for the beep time after long operations is now in the preferences.										*)
(*	-	new result handling method: delete result.																				*)
(*	-	improved drawing of directed line segments.																			*)
(*	-	new menu commands: 'Show All', 'Save All' and 'Hide All' if the option-key is pressed.							*)

(*	24-Sep-92, PS	*)
(*	-	fixed a bug which caused incorrect windownames for 'W##' windows.												*)
(*	-	fixed the function which determined whether system 7 is around.													*)
(*	-	expanded the submenu 'To Delaunay Triangulation'.																	*)
(*	-	added the command 'Flip Edges Interactively' to flip any kind of edge interactively.								*)
(*	-	added the methods 'triangulation.flipOk'.																				*)
(*	-	added the command 'Find Triangulation...' to find triangulation on which the flip-strategies work with the	*)
(*		prescribed difference in number of flips.																				*)
(*	-	changed the preferences to allow three differenct modes for handling the result of a computation:			*)
(*		1) put the result into a new window (as before), 2) add the result as a new selection to the current			*)
(*		window and 3) the result replaces the selection (undoable).															*)
(*	-	changed 'triangulation2d.interactiveOutput' and 'triangulation.interactiveHighlight' to also display the		*)
(*		vertices.																														*)

(*	20-Aug-92, PS	*)
(*	-	'Duplicate' command in edit menu to duplicate the selection.															*)
(*	-	added 'list.concatenateDispose' to concatenate lists and to dispose the wrapping object of the concatenated	*)
(*		list.																															*)
(*	-	added 'graphEdgeTriangulation.changeVertex' and 'graphEdgeTriangulation.changeTriangle'.					*)
(*	-	added 'triangle.changeEdge'.																								*)
(*	-	added 'triangulation.updateEdge', 'triangulation.deleteEdge', 'triangulation.deleteTriangle' and					*)
(*		'triangulation.deleteVertex'.																								*)
(*	-	added the command 'Delete Triangle...' to delete a specified triangle from a triangulation.						*)
(*	-	added the command 'Delete Vertex...' to delete a specified vertex from a triangulation.							*)
(*	-	added the command 'Delete Bounding Triangle' to delete the bounding triangle from a							*)
(*		'searchTriangulation2d'.																									*)
(*	-	changed defintion of 'triangulation.optimize'.																			*)
(*	-	added command 'To Delaunay Triangulation Randomized'.																*)

(*	18-Aug-92, PS	*)
(*	-	changed the definition of 'sequence.showNumbering'. Elements which are sequences themselves are			*)
(*		recursively expanded.																										*)
(*	-	added a method 'triangulation.randomFlip' for randomly flipping edges in a triangulation keeping the			*)
(*		triangulation planar.																										*)
(*	-	added a command 'Randomized No Shuffle' for testing the randomized triangulation algorithm without		*)
(*		shuffling taking place.																										*)
(*	-	added a command 'Random Flip Edges...' for randomly flipping edges in a triangulation.							*)

(*	14-Aug-92, PS	*)
(*	-	added the method 'triangulation2d.findMinimalAngleTriangle' for finding the most acute triangle.				*)
(*		added the command 'Find Most Acute Triangle' for finding the most acute triangle in a 2-d triangulation.	*)

(*	11-Aug-92, PS	*)
(*	-	changed the name of the method 'convexPolygon.closestPair' to														*)
(*		'convexPolygon.closestPairConvexPolygons' to enable the computation of the closest pair of the vertices	*)
(*		of a convex polygon.																										*)
(*	-	implemented 'polyLine.closestPairModified' in order to prevent a change of the original input data.			*)
(*	-	changed the implementation of 'pointVector.makeIdPoints' such that the result is always of type				*)
(*		'pointVector' and not of a derived class.																					*)
(*	-	changed the execution of the 'Cyclic Shift' command to work properly for poly lines.							*)

(*	5-Aug-92, PS	*)
(*	-	added the boolean 'acceptCircles' to the preferences. If set to 'true', frames are interpreted as circles,	*)
(*		otherwise they are discarded.																							*)

(*	3-Aug-92, PS	*)
(*	-	integrated new software from Bert Leukens (triangulation code).													*)
(*	-	fixed a bug in 'triangulation2d.optimize'.																				*)

(*	30-Jul-92, PS	*)
(*	-	added 'simpleUndirectedGraph.getEdges' to compute the edges of a graph as a line segment vector.			*)
(*	-	added commands 'Get Edges' and 'Get Vertices' for graphs.															*)
(*	-	added the option 'Zoom Out All' when the shift-key is pressed.														*)
(*	-	added 'voronoiDiagram.getFiniteEdges' and 'voronoiDiagram.getVertices'.											*)
(*	-	added commands 'Point Cover...' for Voronoi diagrams and 'Get Finite Voronoi Edges' and						*)
(*		'Get Voronoi Vertices'.																										*)

(*	28-Jul-92, PS	*)
(*	-	changed the datastructure for 'triangle', added methods 'triangle.getEdges', 'triangle.getPointIDs',			*)
(*		'triangle.getThirdPoint', 'triangle.getThirdPointId',' triangle.getPoints' (Code by Bert Leukens).				*)
(*	-	added 'triangulation.addNewEdge' and 'triangulation.addNewTriangle' (code by Bert Leukens).					*)
(*	-	added code for bounding boxes by Adrian Bruengger.																	*)
(*	-	added an improved 3d user interface (code by Christian Franz).														*)
(*	-	added shortest path code by Peter Bodenmann (probably too buggy to survive).									*)
(*	-	added randomized triangulation and some operations on triangulations: 'triangulation2d.flip', 					*)
(*		'triangulation2d.initializeFromPolygon2d',  'triangulation2d.generateUniformRandomPoints',					*)
(*		'triangulation2d.createLineSegments',  'triangulation2d.optimize'. The initialization procedure used to be	*)
(*		the method in 'polygon2d' to triangulate simple polygons.															*)
(*		New commands are on planar triangulations are 'To Line Segments', 'To Delaunay Triangulation',	 and		*)
(*		'Extract Points'.																												*)

(*	8-Jul-92, PS	*)
(*	-	added 'pointRadius' and 'pointRadiusHighlight' to the preferences (useful for preparing slides).				*)
(*	-	added the compile option 'OptionalCommand' to remove optional operations from the operations menu.		*)

(*	1-Jul-92, PS	*)
(*	-	made 'obj.shallowInitializeFromObject' more flexible.																*)
(*	-	added an object-type of x-monotone chains and some operations on them.  The operations and					*)
(*		accompanying commands are optional commands and can be omitted when they are not needed.				*)
(*	-	added 'triangulation2d.createUniformRandomPoints' for the creation of uniformly distributed random 		*)
(*		points inside the triangles of a triangulation.																			*)
(*	-	added the command 'Random Points...'.																					*)

(*	24-Jun-92, PS	*)
(*	-	added the boolean 'rebuildOperationsMenu' for commands that modify 'currentArguments'.					*)
(*	-	improved 'graphEdge.drawEdge' such that it can handle graphs in three dimensions.								*)

(*	22-Jun-92, PS	*)
(*	-	added 'pointVector.shortestPath' to find the shortest path between two points among vertical					*)
(*		linesegments as obstacles.																									*)
(*	-	added corresponding command 'Shortest Path'.																			*)

(*	19-Jun-92, PS	*)
(*	-	added undirectedGraph.initializeFromSimpleUndirectedGraph.														*)
(*	-	added directedGraph.getTriangulation.																					*)
(*	-	added command 'Test Triangulation...'.																					*)

(*	4-Jun-92, PS	*)
(*	-	added 'sequence.exists' to test whether an element satisfying a predicate exists in a sequence.				*)
(*	-	the vertices of a graph can be dragged by pressing the option key.													*)
(*	-	changed the parameters of 'obj.dragObject' to allow more flexibility while dragging.							*)
(*	-	'randomChange' now works for graphs.																					*)

(*	2-Jun-92, PS	*)
(*	-	added 'polyLine.convexHullLinear' for computing the convex hull of a simple poly line.							*)
(*	-	added the command 'Convex Hull' for poly lines.																		*)

(*	26-May-92, PS	*)
(*	-	added 'lineSegmentVector.clip' by Rolf Heckendorn and Niklaus Ruess and a corresponding command.			*)
(*	-	removed 'convexPolygon.minimumDistance' and the corresponding command.										*)
(*	-	extended the animation menu to accomodate 'Clip With Rectangle'.													*)
(*	-	removed a bug in 'quadTree.describe' which caused conversion to quadtree to be made available for			*)
(*		line segments.																												*)
(*	-	removed a bug in 'analyzeInput' which caused problems when a vector of length 0 was passed to the			*)
(*		routine.																														*)

(*	8-May-92, PS	*)
(*	-	added two routines for performing a simple closest pair among convex polygons.								*)
(*	-	added a routine to compute the simple minimal distance between two convex polygons (by JM).				*)
(*	-	added hierarchical commands to use the distance functions and the closest pair functions for convex			*)
(*		polygons.																														*)
(*	-	added animation to 'Winding Number'	.																					*)
(*	-	added the command 'Winding Number Dynamic'	.																		*)
(*	-	'convexPolygon.diameter' now also computes the 'height' of a convex polygon.									*)

(*	4-May-92, PS	*)
(*	-	added a command to create random (convex) polygons inside rectangles.											*)
(*	-	added a method for dragging an object ('obj.dragObject').															*)
(*	-	added 'point2d.signDx', 'point2d.signDy' and 'point2d.setPoint'.														*)
(*	-	added distance function for convex polygons and a closest pair function for convex polygons					*)
(*		(implementations by Joerg Mrkwitschka (JM)).																		*)

(*	22-Apr-92, PS	*)
(*	-	added the global variable 'isSystemSeven' to indicate which version of the MacOS is running.					*)
(*	-	added improved checking for 'command' 'period' with code adapted from the THINK Class Library.			*)

(*	3-Apr-92, PS	*)
(*	-	timing is suspended while the is-no-monotone-polygon alert is shown.												*)
(*	-	added sort by XY and sort YX in the sort submenu.																	*)

(*	2-Apr-92, PS	*)
(*	-	added a call to 'ExtendDrawingSheet' in the drawing procedures for quad-trees.									*)
(*	-	the operations 'Rotate...' and 'Scale...' are now available if the second argument is a point.					*)
(*	-	the method 'regularNGon.construct' now computes the radius such that the cursor is on the perimeter		*)
(*		if 'currentAlignmentIsOn' = true.																							*)

(*	31-Mar-92, PS	*)
(*	-	added the quadtree package by Rolf Heckendorn and Niklaus Ruess													*)
(*	-	new operations: Rectangle To QuadTree:Maximal Depth, Rectangle To QuadTree: Chosen Depth				*)
(*		Circle To QuadTree:Maximal Depth, Circle To QuadTree: Chosen Depth											*)
(*		Grouped Objects To QuadTree:Maximal Depth, Grouped Objects To QuadTree To QuadTree: Chosen Depth	*)
(*		Intersection:Tree1 * Tree2, 	Intersection:NOT Tree1 * Tree2, Intersection:Tree1 * NOT Tree2				*)
(*		Intersection:NOT Tree1 * NOT Tree2																					*)
(*		Union:Tree1 * Tree2, 	Union:NOT Tree1 * Tree2, Union:Tree1 * NOT Tree2										*)
(*		Union:NOT Tree1 * NOT Tree2, Difference, Negation																	*)
(*	-	added 'polygon2d.triangulateMonotone' for the triangulation of monotone polygons.								*)
(*	-	new operations:	Triangulate Monotone																					*)
(*	-	added 'sequence.cyclicShift' to shift a sequence cyclically.															*)
(*	-	new operation: Cyclic Shift...																								*)
(*	-	added 'vector.partialFlash' to flash only part of a vector.															*)
(*	-	higlight the default buttons in dialog boxes.																				*)
(*	-	the 'zoom in' rectangle is now constrained to the original drawing rectangle by default.						*)
(*	-	the command 'Random Change' is now undoable.																		*)
(*	-	new operation 'Make:Circles' to create a set of circles from a set of points.										*)

(*	23-Mar-92, PS	*)
(*	-	moved 'vector.revert' to 'sequence.revert' and implemented 'list.revert'.										*)
(*	-	the call  to 'NewPixPat' in 'triangle3d' is now made only on color machines.										*)
(*	-	a new window is automatically opened at startup, if no files have been opened.									*)
(*	-	'Show Numbering' and 'Revert Numbering' are now applicable to all sequences.									*)

(*	2-Mar-92, PS	*)
(*	-	added 'obj.shallowInitializeFromObject', 'obj.duplicateTypeTransfer' and 'obj.duplicateAllTypeTransfer'	*)
(*	- 	added the operation 'Random Change' to 'vector'.																		*)
(*	-	renamed 'Randomize' and 'Reset Random'																				*)
(*	-	added 'point2d.equalPoint', 'realPoint.equalPoint', l'ongIntPoint.equalPoint', 'floatPoint.equalPoint'.			*)
(*	-	added a menu command 'Random Change' in the objects menu.														*)

(*	13-Feb-92, PS	*)
(*	-	added 'triangulation.initializeFromGraph'.																				*)
(*	-	added 'sequence.showNumbering'.																						*)

(*	12-Feb-92, PS	*)
(*	-	a rectangle is now constrained to a square by pressing the 'option' key instead of the 'shift' key which		*)
(*		caused a conflict during selection.																						*)
(*	-	added the classes 'triangulation', 'triangulation2d', 'triangulation3d', 'graphEdgeTriangulation' and			*)
(*		fixed several problems related to the inefficient use of 'appendSafe'.												*)
(*	-	added a operation to create a 'rectangleVector' from two 'pointVector's using the newly created				*)
(*		'pointVector.createRectangleVector'.																					*)
(*	-	added animation to the divide and conquer version of the closest pair algorithm.									*)
(*	-	changed the file and textual representation of Voronoi-diagrams for higher efficiency.							*)

(*	17-Jan-92, PS	*)
(*	-	fixed a bug in 'markedRealPoint.genericLessThan' (uninitialized variable 'm').									*)
(*	-	the type of the points created by 'obj.pointCover' is determined by 'currentPointType'.						*)
(*	-	the class 'idRealPoint' has been promoted to a I/O-able class.														*)
(*	-	the divide and conquer algorithm for computing the Voronoi-diagram has been removed.						*)
(*	-	added a method 'voronoiDiagram.polygonArea' which computes a vector of 'realNumber' representing		*)
(*		the size of the Voronoi-region associated with each site.																*)
(*	-	simplified the Voronoi-diagram datastructure by removing the redundant 'neighbors'-list.						*)
(*	-	the instance variables 'p1', 'p2', 'v1', 'v2' in 'voronoiEdge' are supposed to have type 'idRealPoint'.			*)

(*	10-Jan-92, PS	*)
(*	-	rewrote 'lineSegmentVector.isHorizontalVertical' to use 'sequence.sameProperty'								*)
(*	-	removed some calls to 'advanceStopwatchHandleNoMem' from the modules 'sequence' and 'vector'			*)

(*	19-Dec-91, PS	*)
(*	-	removed 'initInt' which was only declared, but never implemented or used.										*)

(*	18-Dec-91, PS	*)
(*	-	improved 'rectangleVector.describe' and 'rectangleVector.execute' to give a meaningful description in		*)
(*		the dialog boxes for random creation of objects.																		*)
(*	-	changed 'rectangle.construct' to allow the construction of squares if the shift key is pressed.					*)
(*	-	moved declaration of 'dDimCircle' to 'basicGeometry'.																*)
(*	-	added 'dDimPoint.randomInCircle' for uniform random generation inside a circle.								*)
(*	-	changed 'dDimPoint.randomChange' to accomodate (d-dimensional) circles.										*)

(*	13-Dec-91, PS	*)
(*	-	result type of 'lineSegment.length' is 'extended'.																		*)
(*	-	added a command to revert the numbering of a 'pointVector'.														*)
(*	-	added a command to display the numbering of a 'pointVector'.														*)
(*	-	added 'directedLineSegmentType' in 'obj' and added 'equal', 'memberType', 'getType'							*)
(*		and 'getTextType' as methods to the class 'directedLineSegment'.													*)
(*	-	added 'DST' as text identifier for 'directedLineSegment'.																*)
(*	-	added 'directedLineSegmentType' to the exception set of types which are not examined for 'describe'.		*)
(*	-	added a parameter 'segmentType' to the method 'polyLine.toLineSegmentVector'.								*)
(*	-	added a method 'pointVector.showNumbering' which maps each point to a 'markedRealPoint' marked			*)
(*		with its respective index in the 'pointVector'.																			*)
(*	-	added a command to display the orientation of a line segment.														*)
(*	-	added a command to revert the orientation of a line segment.														*)
(*	-	changed the rotate transformation such that it accepts either a point and a list of objects or an object		*)
(*		and a point.																													*)
(*	-	removed the overridden implementations of 'polyLine.makePolyLine' and 'polygon2d.makePolygon'.			*)
(*	-	changed the result type of 'polyLine.perimenter' from 'real' to 'extended'.										*)
(*	-	added 'circle.inside' and 'circle.distanceToPerimeterX' 																*)
(*	-	changed 'point2d.randomChange', 'rectange.randomChange' and 'circle.randomChange' to accomodate		*)
(*		regular n-gons as random constraints.																					*)

(*	11-Dec-91, PS	*)
(*	-	fixed a bug in 'lineSegment': 'moveTop' was set to false by dragging a line segment which subsequently		*)
(*		caused 'polyLine.windowInput' to hang in an infinite loop.																*)
(*	-	fixed 'polygon2d.clip': 'currentVectorLength' was not set properly causing unnecessary memory				*)
(*		manager operations.																										*)
(*	-	added 'point2d.intersectLineX', 'realPoint.intersectLineX', 'floatPoint.intersectLineX' and						*)
(*		'lineSegment.intersectLineX' for intersecting a line segment and a (infinite) line.								*)
(*	-	added 'polygon2d.decompose' in order to cut a simple polygon along a line.										*)
(*	-	made the operation 'reflect' at a line segment symmetric, i.e. the line segment can be the first or the		*)
(*		second argument.																											*)

(*	6-Dec-91, PS	*)
(*	- spanningTree.execute: the resulting traveling salesman tour is converted to a 'polygon2d', if possible.		*)

{$IFC AnimationEnabled }
{$ENDC }
{$IFC ChecksEnabled }
{$ENDC }
{$IFC Chemistry }
{$ENDC }
{$IFC ChemistryTriangulation }
{$ENDC }
{$IFC THINK_PASCAL }
{$IFC Elems881 }
{$ENDC }
{$IFC FarCode }
{$ENDC }
{$ENDC }
{$IFC GFEnabled }
{$ENDC }
{$IFC NetworkEnabled }
{$ENDC }
{$IFC OptionalCommand }
{$ENDC }
{$IFC RaumPlanung }
{$ENDC }
{$IFC ScriptEnabled }
{$ENDC }
{$IFC THINK_PASCAL }
{$ENDC }
{$IFC Topography }
{$ENDC }
{$IFC Tutorial }
{$ENDC }
{$IFC VorlesungGeoAlg }
{$ENDC }
{$IFC XYZApplication }
{$ENDC }


{$IFC false }
(* Defaults *)
AnimationEnabled = TRUE;
ChecksEnabled = FALSE;
Chemistry = FALSE;
ChemistryTriangulation = FALSE;
Elems881 = TRUE;
FarCode = TRUE;
GFEnabled = FALSE;
NetworkEnabled = FALSE;
OptionalCommand = FALSE;
Raumplanung = FALSE;
ScriptEnabled = FALSE;
THINK_PASCAL = TRUE;
Topography = FALSE;
Tutorial = FALSE;
VorlesungGeoAlg = FALSE;
XYZApplication = FALSE;
{$ENDC }


unit versionHistory;
interface
implementation
end.
