(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit voronoiDiagramSL;

(* A sweepline algorithm for computing the Voronoi Diagram mostly by Beat Fawer *)

(* 18.3.93  *)
(* pSiteLast ist nicht mehr vom Typ Obj sondern vom Typ realPoint, so entfaellt duplicate *)

{$SETC AvoronoiDiagramSL = TRUE }

(************************************************************************		*)
(****** Silke Schweifer, Universitt Mnster, 27.3.93 ***********************************		*)
(************************************************************************		*)
(*																																			*)
(* Folgende Fehler wurden bei dem Algorithmus von Peter Schorn entdeckt:                           							*)
(*																																			*)
(* 1.) Bei einer Verteilung von n>3 Datenpunkten auf einem Kreis,wobei einer dieser Datenpunkte auf  					*)
(*dem "obersten"  Punkt des Kreises liegt, kann unter Umstnden (je nachdem in welcher Reihenfolge die 				*)
(* Knotenduplikate bzw. der Datenpunkt aus der queue Q entnommen wird) ein falsches Voronoi Diagramm erstellt	*)
(* werden. Zwei korrekt erstellte Grenzen werden wieder gelscht (d.h. haben eine Lnge gleich Null) und				*)
(* eine falsche Voronoi Grenze wird erzeugt.																						*)
(*																																			*)
(* Der Fehler kommt folgendermaen zustande:    																				*)
(*																																			*)
(* Sei q der Datenpunkt, der auf dem "obersten" Punkt des Kreises liegt, p1...pm die brigen cocircularen				*)
(* Datenpunkte. Unmittelbar unter q hat man, bei geeigneter Nummerierung, folgende Belegung:                      			*)
(* L=(...R1, C(2,1), R2, ... , Rm-1, C(m-1,m), Rm... ) 													     					*)
(* Q= (q, S(1,3), S(2,4), ..., S(m-2,m))  , wobei S(i-2,i) der Schnittpunkt der Grenzen C(i-2,i-1) und					*)
(* C(i-1,i) ist.																															*)
(*																																			*)
(* Die Reihenfolge der Bearbeitung der Duplikate in Q ist unerheblich (nach Skript Fortune). Bearbeite 					*)
(* den Datenpunkt q als nchstes Element. q liege in der (transformierten) Region Ri. Dann gilt:           					*)
(*																																			*)
(* L= (R1, C(2,1), ..., C(i-1,i), Ri, C(i,q), Rq, C(q,i), Ri, C(i,i+1), ..., Rm)													*)
(*							     |----?----| |-----?---|																					*)
(* lsche aus Q:  S(i-1,i+1)																											*)
(* fge in Q ein:  S(i-1,q)  und  S(q,i+1)																							*)
(* Dann ist:  Q= (S(1,3), S(2,4), ..., S(i-2,i), S(i-1,q), S(q,i+1), S(i,i+2), ..., S(m-2,m)									*)
(*																																			*)
(* Bearbeite nun S(i-1,q). Dann gilt:																									*)
(* L= (R1, C(2,1), ... C(i-2,i-1), Ri-1, C(i-1,q), Rq, C(i,q), Ri, C(i,i+1), ... Rm)                          						*)
(*								 |----?------| |----?------|																			*)
(* lsche aus Q:  S(i-2,i)				  																								*)
(* fge in Q ein:  S(i-2,q)  																											*)
(*																																			*)
(* Frage: mu der Schnittpunkt S(i-1,i) in Q eingefgt werden ?																*)
(*																																			*)
(* Betrachte dazu folgende Prozedur:																								*)
(*																																			*)
(************************************************************************		*)
(*    function sameThreePoints (p1, p2, p3: realPoint; v: queueEl): boolean;													*)
(*																																			*)
(*    function isIn (p: realPoint; v: queueEl): boolean;																				*)
(*    begin																																*)
(*    isIn := (p = v.q) | (p = v.r) | (p = v.s);																							*)
(*    end; 																												    		 		*)
(*																																			*)
(*    begin 																																*)
(*    sameThreePoints := isIn(p1, v) & isIn(p2, v) & isIn(p3, v);																*)
(*    end; 																																*)
(************************************************************************		*)
(*																																			*)
(*	Die Variablen in der Funktion 'sameThreePoints' haben folgende Belegung:													*)
(*	p1= i-1          v.q= i-1  																							     			*)
(*	p2= q             v.r= i    																					     		     			*)
(*	p3= i              v.s= q    													     									     			*)
(* d.h. sameThreePoints = true, und der Knoten S(i-1,i) wird nicht in Q eingefgt. (Vergleiche 	      	     			*)
(* Prozedur 'insertIntoQueueIfIntersectBoundary')														 	   	     			*)
(*	Dann ist: Q= (S(1,3), ..., S(i-3,i-1), S(i-2,q), S(q,i+1), S(i,i+2),... S(m-2,m)) 										*)
(*																																			*)
(* Bearbeite nun S(q,i+1). Dann gilt:																								*)
(* L= (R1, C(2,1), ... C(i-2,i-1), Ri-1, C(i-1,q), Rq, C(q,i+1), Ri+1, C(i+1,i+2), ... Rm)         		       				*)
(*												    |----?-----|  |-----?-------|														*)
(* lsche aus Q:  S(i,i+2)				  																								*)
(* fge in Q ein:  S(q,i+2)  																											*)
(*																																			*)
(* Frage A: mu der Schnittpunkt S(i-1,i+1) in Q eingefgt werden ?															*)
(*																																			*)
(*	Die Variablen in der Funktion 'sameThreePoints' haben folgende Belegung:													*)
(*	p1= i-1          v.q= q  																								   				*)
(*	p2= q             v.r= i    																								   				*)
(*	p3= i+1          v.s= i+1    													     													*)
(*	Dann ist  isIn(p1,S(q,i+1)) = false, d.h. sameThreePoints = false, und der Knoten S(i-1,i) wird  in	     			*)
(* Q eingefgt.																															*)
(*																																			*)
(* Hier liegt der Fehler!!!!!!!!!	, denn Q hat nun folgende Belegung:																*)
(* Q= (S(1,3), ..., S(i-3,i-1), S(i-2,q), S(i-1,i+1), S(q,i+2), S(i+1,i+3), ... S(m-2,m)) 									*)
(*																																			*)
(* Bearbeite nun S(i-1,i+1). Dann gilt:																								*)
(* L= (R1, C(2,1), ... C(i-2,i-1), Ri-1, C(i-1,i+1), Ri+1, C(i+1,i+2), ... Rm)        				     		   				*)
(*								 |------?-------|  |------?------|																	*)
(* lsche aus Q:  S(i-2,q) und S(q,i+2)				  																				*)
(* fge in Q ein:  S(i-2,i+1) und S(i-1,i+2)  																						*)
(*	Dann ist: Q= (S(1,3), ..., S(i-3,i-1), S(i-2,i+1), S(i-1,i+2),... S(m-2,m)) 					     		 				*)
(*																																			*)
(* ... fahre so fort bis Q= ( ):																											*)
(* L reduziert sich zu (...R1, C(1,m), Rm...) . Korrekt wre aber folgende Belegung fr L:									*)
(* (... R1, C(1,q), Rq, C(q,m), Rm...) 																				  				*)
(*																																			*)
(* Diesen Fehler kann man folgendermaen beheben:																				*)
(* Der jeweils letztbearbeitete Datenpunkt wird in der global  definierten Variablen 'pSiteLast' gespeichert 			*)
(* ( siehe Prozedur 'voronoiDiagramSl').																							*)
(* Markiert p in einem Schleifendurchgang einen Datenpunkt, wird der Algorithmus wie bisher durchlaufen. 			*)
(* (pSiteLast wird entsprechend aktualisiert). 																					*)
(* Markiert p aber ein Knoten, so wird in der Prozedur 'insertIntoQueueIfIntersectBoundary' berprft,				*)
(* ob der gemeinsam definierende Site (gespeichert in der lokalen Variablen 'r') der zwei, sich "eventuell"				*)
(* schneidenden Grenzen dem 'pSiteLast' entspricht. Der Schnittpunkt wird nur dann in Q eingetragen,					*)
(* wenn dies nicht der Fall ist.																										*)
(* 																																			*)
(*************************************************************************	*)
(* procedure insertIntoQueueIfIntersectBoundary (b1,b2:reference; v: queueEl);											*)
(* begin 																																	*)
(*    if (b1 <> nil) & (b2 <> nil) then																									*)
(* 	   begin																																*)
(*       activeEl(active.getObject(b1)).e.commonPoints(activeEl(active.getObject(b2)).e, q, r, s, ok);					*)
(*       if (q.whichSideX(r, s) <> 0) then																								*)
(*       begin																																*)
(*          new(c);																														*)
(*          failNil(Handle(c));																											*)
(*          q.circleCenterX(r, s, cx, cy);																								*)
(*          c.x := cx;																														*)
(*          c.y := cy;																														*)
(*          if v.qType = vertex then																									*)
(*          begin																															*)
(********************* 13.11.92 **********************												*)
(*             if (r.x <> realPoint(pSiteLast).x) | (r.y <> realPoint(pSiteLast).y) then											*)
(*             begin																															*)
(********************* 13.11.92 **********************												*)
(*                if not sameThreePoints(q, r, s, v) & pointOfBoundary(c, activeEl(active.getObject(b1)), false) & 		*)
(*                                    pointOfBoundary(c, activeEl(active.getObject(b2)), false)  then 								*)
(*                  insertVertexIntoQueue(c, q, r, s, b1, b2)																			*)
(*                else																														*)
(*                  dispose(c);																												*)
(*             end;																															*)
(*          end																																*)
(*          else if pointOfBoundary(c, activeEl(active.getObject(b1)), true) & 													*)
(*                                                                 pointOfBoundary(c, activeEl(active.getObject(b2)), true) then			*)
(*                    insertVertexIntoQueue(c, q, r, s, b1, b2)																		*)
(*                else																														*)
(*                   dispose(c);																												*)
(*       end;																																*)
(*    end;																																	*)
(* end; 																																	*)
(*************************************************************************	*)
(* 																																			*)
(* Fr das obige Beispiel bedeutet das, eingesprungen bei der Frage A: mu der Schnittpunkt S(i-1,i+1) in 				*)
(* Q eingefgt werden ?																												*)
(*																																			*)
(*	Die Variablen in der Funktion 'insertIntoQueueIfIntersectBoundary' haben folgende Belegung:							*)
(* r= q           pSiteLast= q  (q ist der zuletzt bearbeitete Datenpunkt gewesen, siehe oben)								*)
(* d.h. die Prozedur 'sameThreePoints' wird nicht angesprungen, und der Knoten S(i-1,i) wird nicht in 					*)
(* Q eingetragen.																														*)
(*																																			*)
(* Fhrt man so fort, hat man nach Bearbeitung aller Knoten, die auf q fallen, somit folgende korrekte 					*)
(* Belegungen:  		L= (...R1, C(1,q), Rq, C(q,m), Rm ...)  																		*)
(* 																																			*)
(*																																			*)
(* 2.) Bei einer collinearen Verteilung von n Datenpunkten werden einige Voronoi Kanten korrekt gezeichnet,			*)
(* andere ntige Kanten fehlen. Nheres dazu in 'voronoiDiagram' bei der entsprechenden "fehlerhaften"				*)
(* Prozedur 'voronoiEdge.interactivOutput" bzw. 'voronoiEdge.toPoints'.														*)
(*																																			*)
(*************************************************************************	*)
(*************************************************************************	*)



interface

uses
	pointVector;

implementation

const
	absTol = 0.01;

type
	edgeSL = object(voronoiEdge)
			d2: double;
			procedure textOutput;
			override;
		end;

	activeType = (boundary, newPoint);
	activeEl = object(obj)
			aType: activeType;
			e: edgeSL;					(* aType = boundary only *)
			side: directionType;		(* aType = boundary only *)
			p: realPoint;				(* aType = newPoint only *)
			procedure textOutput;
			override;
		end;

	queueType = (site, vertex);
	queueEl = object(obj)
			qType: queueType;
			q, r, s: realPoint;		(* r is the common point, if qType=site then q is this site *)
			v: realPoint;				(* qType = vertex only *)
			yTransf: real;				(* qType = vertex only *)
			bqr, brs: reference;		(* qType = vertex only *)
			procedure free;
			override;
			procedure freeAll;
			override;
			procedure textOutput;	(* implementation correct if queue is avlTree or sortedList *)
			override;
			procedure interactiveOutput;
			override;
		end;

procedure edgeSL.textOutput;
	begin
		inherited textOutput;
		write(d2 : 5 : 1, ' ');
	end;

procedure activeEl.textOutput;
	begin
{$IFC THINK_PASCAL }
		write(aType, ' ');
{$ELSEC }
		write(ord(aType), ' ');
{$ENDC }
		if aType = boundary then begin
			e.textOutput;
{$IFC THINK_PASCAL }
			write(side, ' ');
{$ELSEC }
			write(ord(side), ' ');
{$ENDC }
		end
		else
			p.textOutput;
	end;

procedure queueEl.textOutput;
	begin
{$IFC THINK_PASCAL }
		write(qType, ' ');
{$ELSEC }
		write(ord(qType), ' ');
{$ENDC }
		q.textOutput;
		if qType = vertex then begin
			r.textOutput;
			s.textOutput;
			v.textOutput;
			write(yTransf : 4 : 1, ' ');
			if bqr <> nil then
				avlNodeH(bqr)^^.value.textOutput; (* may not work for other references! *)
			if brs <> nil then
				avlNodeH(brs)^^.value.textOutput; (* may not work for other references! *)
		end;
	end;

procedure queueEl.free;
	begin
		shallowFree;
	end;

procedure queueEl.freeAll;
	begin
		if qType = vertex then
			dispose(v);
		shallowFree;
	end;

procedure queueEl.interactiveOutput;
	begin
		q.interactiveOutput;
		if qType = vertex then begin
			r.interactiveOutput;
			s.interactiveOutput;
			v.interactiveOutput;
		end;
	end;

procedure updateIds (p: pointVector);
	var
		i: seqIndex0;
	begin
		for i := 1 to p.length do
			with idRealPoint(p.elements^^[i]) do
				id := i;
	end;

function compareQueueInsert (a, b: queueEl): relationObj;
(* compare-procedure for inserting an element in the queue-list *)
	begin
		if a.yTransf < b.yTransf then
			compareQueueInsert := lessThanObj
		else if a.yTransf > b.yTransf then
			compareQueueInsert := greaterThanObj
		else if a.r.y < b.r.y then (* special hack to make VD 9 work, algorithm wrong? *)
			compareQueueInsert := lessThanObj
		else if a.r.y > b.r.y then
			compareQueueInsert := greaterThanObj
		else
			compareQueueInsert := equalObj;
	end; (* compareQueueInsert *)

function compareQueueFind (a, b: queueEl): relationObj;
(* compare-procedure for finding an element in the queue-list.													*)
(* there can be several elements with the same key (almost the same y-coordinate) in the queue-list.	*)
	begin
		if a.yTransf < b.yTransf - absTol then
			compareQueueFind := lessThanObj
		else if a.yTransf > b.yTransf + absTol then
			compareQueueFind := greaterThanObj
		else
			compareQueueFind := equalObj;
	end; (* compareQueueFind *)

function pointVector.voronoiDiagramSweepLine (ask: boolean): voronoiDiagram;
	var
		active: dictionary;
		queue: dictPriorityQueue;
		yActive: real;
		firstReg: realPoint;
		siteIndex: seqIndex;
		vor: voronoiDiagram;
(************************************************************************	*)
(****** Silke Schweifer, Universitt Mnster, 27.3.93 ***********************************	*)
(************************************************************************	*)
		pSiteLast: realPoint;
(********************* 13.11.92 ********************** *)

	procedure xOfEdge (e: edgeSL; var xLeft, xRight: double);
(* calculates the x-positions where the horizontal line with y-coordinate yActive (global variable)	*)
(* intersects the transformed bisector of the edge e.															*)
		var
			a2, b, c, d, r: double;
		begin
			a2 := 2 * (e.p1.y - e.p2.y);
			if a2 = 0 then begin
				xLeft := (e.p1.x + e.p2.x) / 2;
				xRight := xLeft;
			end
			else if e.p2.y = yActive then begin
				xLeft := e.p2.x;
				xRight := xLeft;
			end
			else begin
				b := 2 * ((e.p1.x - e.p2.x) * (e.p2.y - yActive) - e.p2.x * (e.p1.y - e.p2.y));
				c := (e.p1.y - e.p2.y) * ((sqr(e.p2.y) - sqr(yActive)) + sqr(e.p2.x)) - e.d2 * (e.p2.y - yActive);
				d := sqr(b) - 2 * a2 * c;
				if d < 0 then  (* numerical error *)
					begin
					xLeft := e.p2.x;
					xRight := xLeft;
				end
				else begin
					r := sqrt(d);
					xLeft := (-b + r) / a2;
					xRight := (-b - r) / a2;
				end;
			end;
		end; (* xOfEdge *)

	function compareActive (a, b: activeEl): relationObj;
(* compare-procedure for inserting an element in the active-list *)
		var
			xa, xb: double;

		procedure xOfActive (bound: activeEl; var x: double);
(* calculates the x-positions where the horizontal line with y-coordinate yActive (global variable)	*)
(* intersects the boundary.																						*)
			var
				xL, xR: double;
			begin
				if bound.aType = boundary then begin
					xOfEdge(bound.e, xL, xR);
					if bound.side = left then
						x := xL
					else
						x := xR
				end
				else (* if aType = newPoint then *)
					x := bound.p.x;
			end; (* xOfActive *)

		begin
			xOfActive(a, xa);
			xOfActive(b, xb);
			if xa < xb then
				compareActive := lessThanObj
			else if xa > xb then
				compareActive := greaterThanObj
			else if b.aType = boundary then
				compareActive := lessThanObj
			else
				compareActive := greaterThanObj;
		end; (* compareActive *)

	function extractMinQueue: queueEl;
(* extracts the minimum element of the queue-list and the original sites.									*)
(* the minimum applies to the y-coordinate of a site or the transformed y-coordinate of an inter-	*)
(* section between two boundaries.																				*)
		var
			vertexMin: reference;
			qEl: queueEl;

		function almostEqualPoints (p, q: realPoint; tol: real): boolean;
			begin
				almostEqualPoints := (abs(p.x - q.x) <= tol) & (abs(p.y - q.y) <= tol)
			end; (* almostEqualPoints *)

		begin
			vertexMin := queue.minimum;
			if (vertexMin = nil) | (siteIndex <= vor.points.length) & (realPoint(vor.points.elements^^[siteIndex]).y <= queueEl(queue.getObject(vertexMin)).yTransf) then begin
				new(qEl);
				failNil(Handle(qEl));
				qEl.qType := site;
				qEl.q := realPoint(vor.points.elements^^[siteIndex]);
				siteIndex := siteIndex + 1;
				while (siteIndex <= vor.points.length) & almostEqualPoints(realPoint(vor.points.elements^^[siteIndex - 1]), realPoint(vor.points.elements^^[siteIndex]), 0.001) do
					siteIndex := siteIndex + 1;
				extractMinQueue := qEl;
			end
			else begin
				extractMinQueue := queueEl(queue.getObject(vertexMin));
				queue.delete(vertexMin, compareQueueFind);
			end;
		end; (* extractMinQueue *)

	function getRegion (p: realPoint; var region: reference): realPoint;
(* finds the region (labeled by a site) of point p. *)
		var
			regHandle: reference;
			np: activeEl;
			found: boolean;

		function createNewPoint (p: realPoint): activeEl;
			var
				np: activeEl;
			begin
				new(np);
				failNil(Handle(np));
				np.aType := newPoint;
				np.p := p;
				createNewPoint := np;
			end; (* createNewPoint *)

		function regOf (rightBound: activeEl): realPoint;
(* determines the region whose right boundary is b *)
			begin
				if (rightBound.side = left) | (edgeSL(rightBound.e).p1.y = edgeSL(rightBound.e).p2.y) then
					regOf := edgeSL(rightBound.e).p1
				else
					regOf := edgeSL(rightBound.e).p2;
			end; (*  regOf *)

		begin (* getRegion *)
			np := createNewPoint(p);
			regHandle := active.insertObj(np, compareActive, found);
			region := active.next(regHandle, right);
			if region = nil then begin
				region := active.next(regHandle, left);
				if region = nil then
					getRegion := firstReg
				else begin
					getRegion := activeEl(active.getObject(region)).e.otherPointOfEdge(regOf(activeEl(active.getObject(region))));
					region := nil;
				end;
			end
			else
				getRegion := regOf(activeEl(active.getObject(region)));
			region := regHandle;
		end; (* getRegion *)

	function createBisector (p, q: realPoint): activeEl;
		var
			e: edgeSL;
			b: activeEl;
		begin
			new(e);
			failNil(Handle(e));
			e.init;
			new(b);
			failNil(Handle(b));
			e.eType := bisector;
			if pointLessThanYX(p, q) then begin
				e.p1 := p;
				e.p2 := q;
			end
			else begin
				e.p1 := q;
				e.p2 := p;
			end;
			e.d2 := (e.p1.x + e.p2.x) * (e.p1.x - e.p2.x) + (e.p1.y + e.p2.y) * (e.p1.y - e.p2.y);
			vor.edges.append(e);
			b.aType := boundary;
			b.e := e;
			b.side := right;
			createBisector := b
		end; (* createBisector *)

	procedure createBisector1 (p, q: realPoint; var bLeft, bRight: activeEl);
(* creates the left and right boundary for the bisector of points p and q *)
		begin
			bRight := createBisector(p, q);
			new(bLeft);
			failNil(Handle(bLeft));
			bLeft.aType := bRight.aType;
			bLeft.e := bRight.e;
			bLeft.side := left;
		end; (* createBisector1 *)

	procedure createBisector2 (v: queueEl; var b: activeEl);
(* creates the boundary for the new bisector at the position of vertex v *)
		var
			pLow: realPoint;
		begin
			b := createBisector(v.q, v.s);
			if edgeSL(b.e).p1.y = edgeSL(b.e).p2.y then
				b.side := right
			else if v.v.x > edgeSL(b.e).p2.x then
				b.side := right
			else if v.v.x < edgeSL(b.e).p2.x then
				b.side := left
			else if edgeSL(b.e).p2.x = v.v.x then begin
				if v.q.y > v.s.y then
					pLow := v.s
				else
					pLow := v.q;
				if pointLessThanYX(pLow, v.r) then
					if v.r.x < v.v.x then
						b.side := right
					else
						b.side := left
				else if pLow.x < v.v.x then
					b.side := left
				else
					b.side := right
			end;
		end; (* createBisector2 *)

	procedure updateL1 (region: reference; bRecLeft, bRecRight: activeEl; var bLeft, bRight: reference);
(* updates the active-list with the boundaries of a new region *)
		begin
			active.getObject(region).free;
			bRight := region;
			active.setObject(bRight, bRecRight);
			if bRecLeft.e.p1.y = bRecLeft.e.p2.y then begin
				dispose(bRecLeft);
				bLeft := nil;
			end
			else
				bLeft := active.insert(bRecLeft, bRight, left);
		end; (* updateL1 *)

	procedure updateL2 (bqr, brs: reference; bqsRec: activeEl; var bqs: reference);
(* updates the active-list with the new boundary of a vertex-event *)
		begin
			active.getObject(bqr).free;
			active.getObject(brs).free;
			active.setObject(bqr, bqsRec);
			active.delete(brs);
			bqs := bqr;
		end; (* updateL2 *)

	procedure deleteFromQueue (b1, b2: reference);
(* deletes  from the queue-listthe intersection of the boundaries of b1 and b2, if any *)
		var
			q: reference;
			v, x: queueEl;
			deleted, ok: boolean;
			p1, p2, p3, c: realPoint;
			dir: directionType;
			cx, cy: extended;
		begin
			if (b1 <> nil) & (b2 <> nil) & (activeEl(active.getObject(b1)).e <> activeEl(active.getObject(b2)).e) then begin
				activeEl(active.getObject(b1)).e.commonPoints(activeEl(active.getObject(b2)).e, p1, p2, p3, ok);
				if p1.whichSideX(p2, p3) <> 0 then begin
					new(c);
					failNil(Handle(c));
					p1.circleCenterX(p2, p3, cx, cy);
					c.x := cx;
					c.y := cy;
					deleted := false;
					new(x);
					failNil(Handle(x));
					x.qType := vertex;
					x.yTransf := c.y + sqrt(c.squaredDistanceX(p1));
					if queue.find(x, compareQueueFind, q, dir) then begin
(* check if q is left of all duplicates *)
						while (queue.next(q, left) <> nil) & (compareQueueFind(queueEl(queue.getObject(queue.next(q, left))), queueEl(queue.getObject(q))) > lessThanObj) do
							q := queue.next(q, left);
						while not deleted & (q <> nil) do begin
							v := queueEl(queue.getObject(q));
							if (v.qType = vertex) & ((b1 = v.bqr) & (b2 = v.brs) | (b2 = v.bqr) & (b1 = v.brs)) then begin (* delete q from queue *)
								queue.getObject(q).freeAll;
								queue.delete(q, compareQueueFind);
								deleted := true;
							end
							else begin
								q := queue.next(q, right);
								if (q <> nil) & (compareQueueFind(x, queueEl(queue.getObject(q))) = lessThanObj) then
									q := nil;
							end;
						end;
					end;
					dispose(c);
					dispose(x);
				end;
			end;
		end; (* deleteFromQueue *)

	procedure insertIntoQueueIfIntersectBoundary (b1, b2: reference; v: queueEl);
(* inserts into the queue-list the intersection of the boundaries of b1 and b2, if any *)
		var
			q, r, s, c: realPoint;
			ok: boolean;
			cx, cy: extended;

		function pointOfBoundary (p: realPoint; bound: activeEl; site: boolean): boolean;
(* p has to be a point of the bisector bound^^.e. This procedure checks only if the x-coordinate is valid. *)
			var
				pob: boolean;
				intX, intY: extended;
			begin
				if bound.e.p1.y = bound.e.p2.y then
					pob := site | (bound.side = right)
				else begin
					pob := (bound.side = left) & (p.x <= bound.e.p2.x) | (bound.side = right) & (p.x >= bound.e.p2.x) | (abs(p.x - bound.e.p2.x) < absTol);
					if pob & (bound.e.eType = ray) & (bound.e.v1.whichSideX(bound.e.p1, bound.e.p2) <> 0) then begin
						bound.e.v1.circleCenterX(bound.e.p1, bound.e.p2, intX, intY);
						pob := (bound.side = left) & (p.x - intX <= absTol) | (bound.side = right) & (intX - p.x <= absTol);
					end
				end;
				pointOfBoundary := pob;
			end; (* pointOfBoundary *)

		procedure insertVertexIntoQueue (c, q, r, s: realPoint; bqr, brs: reference);
			var
				qEl: queueEl;
				newQueueEl: reference;
			begin
				new(qEl);
				failNil(Handle(qEl));
				qEl.qType := vertex;
				qEl.v := c;
				qEl.yTransf := c.y + sqrt(q.squaredDistanceX(c));
				qEl.r := r;
				qEl.q := q;
				qEl.s := s;
				qEl.bqr := bqr;
				qEl.brs := brs;
				newQueueEl := queue.insert(qEl, compareQueueInsert);
			end; (* insertVertexIntoQueue *)

		begin (* insertIntoQueueIfIntersectBoundary *)
			if (b1 <> nil) & (b2 <> nil) then begin
				activeEl(active.getObject(b1)).e.commonPoints(activeEl(active.getObject(b2)).e, q, r, s, ok);
				if q.whichSideX(r, s) <> 0 then begin
					new(c);
					failNil(Handle(c));
					q.circleCenterX(r, s, cx, cy);
					c.x := cx;
					c.y := cy;
					if v.qType = vertex then begin
(************************************************************************	*)
(****** Silke Schweifer, Universitt Mnster, 27.3.93 ***********************************	*)
(************************************************************************	*)
						if not (r = pSiteLast) & pointOfBoundary(c, activeEl(active.getObject(b1)), false) & pointOfBoundary(c, activeEl(active.getObject(b2)), false) then
(********************* 13.11.92 ********************** *)
							insertVertexIntoQueue(c, q, r, s, b1, b2)
						else
							dispose(c);
					end
					else if pointOfBoundary(c, activeEl(active.getObject(b1)), true) & pointOfBoundary(c, activeEl(active.getObject(b2)), true) then begin
						insertVertexIntoQueue(c, q, r, s, b1, b2);
					end
					else
						dispose(c);
				end;
			end;
		end; (* insertIntoQueueIfIntersectBoundary *)

	procedure markEndPoint (p: realPoint; e: edgeSL);
(* marks point p as an endpoint of edge e *)
		begin
			if e.eType = bisector then begin
				e.eType := ray;
				e.v1 := p;
			end
			else if e.eType = ray then begin
				e.eType := segment;
				e.v2 := p;
			end;
		end; (* markEndPoint *)

{$IFC AnimationEnabled AND AvoronoiDiagramSL }
	procedure voronoiDiagramAnimation (eList: list);
		const
			minX = 0;
			maxX = 500;
			maxY = 1000;
			incX = 5;
			maxYOut = 5000;

		procedure output (x: obj);
			var
				e: edgeSL;
				p, q, t: realPoint;
				xEnd, xL, xR: double;
			begin
				e := edgeSL(x);
				e.toPoints(p, q, 1000);
				xOfEdge(e, xL, xR);
				if pointLessThanXY(q, p) then begin
					t := q;
					q := p;
					p := t;
				end;
				if p.x < minX then begin
					p.x := minX;
					p.y := e.bisecY(p.x);
				end;
				if q.x > maxX then begin
					q.x := maxX;
					q.y := e.bisecY(q.x);
				end;
				if p.x = q.x then begin
					p.y := p.y + p.distanceX(e.p2);
					q.y := q.y + q.distanceX(e.p2);
					if p.y > maxYOut then
						p.y := maxYOut;
					if q.y > maxYOut then
						q.y := maxYOut;
					p.drawLine(q);
				end
				else begin
					xEnd := q.x;
					p.y := p.y + p.distanceX(e.p2);
					while p.y > maxY do begin
						p.x := p.x + incX;
						p.y := e.bisecY(p.x);
						p.y := p.y + p.distanceX(e.p2);
					end;
					while (p.x < xEnd) & (p.y < maxY) do begin
						q.x := p.x + incX;
						if q.x > xEnd then
							q.x := xEnd;
						q.y := e.bisecY(q.x);
						q.y := q.y + q.distanceX(e.p2);
						if (p.y <= maxY) & (q.y < maxYOut) then
							p.drawLine(q);
						p.x := q.x;
						p.y := q.y;
					end;
				end;
				dispose(p);
				dispose(q);
			end; (* output *)

		begin
			if yActive <= maxY then
				eList.forAll(output);
		end; (* voronoiDiagramAnimation *)
{$ENDC }

		var
			p: queueEl;
			region, bLeft, bRight, bqs: reference;
			bRecLeft, bRecRight, bqsRec: activeEl;
			reg: realPoint;
			r: pointVector;

{$IFC AnimationEnabled AND AvoronoiDiagramSL }
			sweepLineP, sweepLineQ: realPoint;
{$ENDC }

	begin (* voronoiDiagramSL *)
		if self.length > 0 then begin

			if ask then begin
				if not getXY(length, true, priorityQueue(queue), 2 * length, active, false) then begin
					voronoiDiagramSweepLine := nil;
					exit(voronoiDiagramSweepLine);
				end;
			end
			else begin
				currentDictionaryType := avlTreeType;
				queue := dictPriorityQueue(newInitObj(dictPriorityQueueType));
				active := dictionary(newInitObj(avlTreeType));
			end;

{$IFC AnimationEnabled AND AvoronoiDiagramSL }
			new(sweepLineP);
			failNil(Handle(sweepLineP));
			new(sweepLineQ);
			failNil(Handle(sweepLineQ));
			sweepLineP.x := -1000;
			sweepLineQ.x := +1000;
{$ENDC }

			new(vor);
			failNil(Handle(vor));
			vor.init;
			r := makeIdPoints;
			r.sortAll(pointLessThanYX);
			updateIds(r);
			vor.points := r;
			new(vor.edges);
			failNil(Handle(vor.edges));
			vor.edges.init;
			firstReg := realPoint(vor.points.elements^^[1]);
			siteIndex := 2;

			while not queue.isEmpty | (siteIndex <= vor.points.length) do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					queue.freeAll;
					active.freeAll;
					vor.freeAll;
					voronoiDiagramSweepLine := nil;
{$IFC AvoronoiDiagramSL }
					dispose(sweepLineP);
					dispose(sweepLineQ);
{$ENDC }
					exit(voronoiDiagramSweepLine);
				end;
{$ENDC }
				p := extractMinQueue;
				case p.qType of
					site:  begin
						yActive := p.q.y;
						reg := getRegion(p.q, region);
						createBisector1(p.q, reg, bRecLeft, bRecRight);
						deleteFromQueue(active.next(region, left), active.next(region, right));
						updateL1(region, bRecLeft, bRecRight, bLeft, bRight);
						insertIntoQueueIfIntersectBoundary(active.next(bLeft, left), bLeft, p);
						insertIntoQueueIfIntersectBoundary(bRight, active.next(bRight, right), p);
(************************************************************************	*)
(****** Silke Schweifer, Universitt Mnster, 27.3.93 ***********************************	*)
(************************************************************************	*)
						pSiteLast := realPoint(p.q);
(************** 13.11.92 ********************************* *)
					end; (* site *)
					vertex:  begin

						if (active.next(p.bqr, right) <> nil) & (active.next(p.brs, left) <> nil) then (*true | (active.next(p.bqr, right) <> nil) & (active.next(p.brs, left) <> nil) *)
							begin
							yActive := p.yTransf;
							createBisector2(p, bqsRec);
							deleteFromQueue(active.next(p.bqr, left), p.bqr);
							deleteFromQueue(p.brs, active.next(p.brs, right));
							markEndPoint(p.s, activeEl(active.getObject(p.bqr)).e);
							markEndPoint(p.q, activeEl(active.getObject(p.brs)).e);
							markEndPoint(p.r, bqsRec.e);
							updateL2(p.bqr, p.brs, bqsRec, bqs);
							insertIntoQueueIfIntersectBoundary(active.next(bqs, left), bqs, p);
							insertIntoQueueIfIntersectBoundary(bqs, active.next(bqs, right), p);

						end; (* vertex *)
					end;
				end; (* case *)

{$IFC AnimationEnabled AND AvoronoiDiagramSL }
				if animationFlag[AVoronoiDiagramItem] & (yActive <= 2000) then begin
					sweepLineP.y := yActive;
					sweepLineQ.y := yActive;
					sweepLineP.drawLine(sweepLineQ);
					p.interactiveOutput;
					voronoiDiagramAnimation(vor.edges);
					waitForClick(animationFlag[AVoronoiDiagramItem]);
					sweepLineP.drawLine(sweepLineQ);
					p.interactiveOutput;
					voronoiDiagramAnimation(vor.edges);
				end;
{$ENDC }

				p.freeAll;
			end; (* while queueRoot <> nil *)
			if active.infoString <> '' then
				displayInfoString(concat('Active ', active.infoString));
			queue.free;
			active.freeAll;
			voronoiDiagramSweepLine := vor;

{$IFC AnimationEnabled AND AvoronoiDiagramSL }
			dispose(sweepLineP);
			dispose(sweepLineQ);
{$ENDC }

		end (* if vor.points.n > 0 *)
		else
			voronoiDiagramSweepLine := nil;
	end; (* pointVector.voronoiDiagramSweepLine *)

end. (* voronoiDiagramSL *)
